<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>MemoteReportApp</title>
  <base href="/">

  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="icon" type="image/x-icon" href="data:image/vnd.microsoft.icon;base64, ">
</head>
<body style="background-color: rgba(177, 212, 237, 0.1)">
  <app-root></app-root>
  <script>window.data = $results;
          window.reportType = $report_type;</script>
<script type="text/javascript">
!function(e){var n=window.webpackJsonp;window.webpackJsonp=function(r,c,i){for(var u,a,f,l=0,s=[];l<r.length;l++)a=r[l],t[a]&&s.push(t[a][0]),t[a]=0;for(u in c)Object.prototype.hasOwnProperty.call(c,u)&&(e[u]=c[u]);for(n&&n(r,c,i);s.length;)s.shift()();if(i)for(l=0;l<i.length;l++)f=o(o.s=i[l]);return f};var r={},t={inline:0};function o(n){if(r[n])return r[n].exports;var t=r[n]={i:n,l:!1,exports:{}};return e[n].call(t.exports,t,t.exports,o),t.l=!0,t.exports}o.e=function(e){var n=t[e];if(0===n)return new Promise(function(e){e()});if(n)return n[2];var r=new Promise(function(r,o){n=t[e]=[r,o]});n[2]=r;var c=document.getElementsByTagName("head")[0],i=document.createElement("script");i.type="text/javascript",i.charset="utf-8",i.async=!0,i.timeout=12e4,o.nc&&i.setAttribute("nonce",o.nc),i.src=o.p+""+e+".chunk.js";var u=setTimeout(a,12e4);function a(){i.onerror=i.onload=null,clearTimeout(u);var n=t[e];0!==n&&(n&&n[1](new Error("Loading chunk "+e+" failed.")),t[e]=void 0)}return i.onerror=i.onload=a,c.appendChild(i),r},o.m=e,o.c=r,o.d=function(e,n,r){o.o(e,n)||Object.defineProperty(e,n,{configurable:!1,enumerable:!0,get:r})},o.n=function(e){var n=e&&e.__esModule?function(){return e.default}:function(){return e};return o.d(n,"a",n),n},o.o=function(e,n){return Object.prototype.hasOwnProperty.call(e,n)},o.p="",o.oe=function(e){throw console.error(e),e}}([]);
</script><script type="text/javascript">
webpackJsonp(["polyfills"],{"../../../../../src/polyfills.ts":function(e,t,r){"use strict";Object.defineProperty(t,"__esModule",{value:!0});var o=r("../../../../core-js/es7/reflect.js"),n=(r.n(o),r("../../../../zone.js/dist/zone.js"));r.n(n)},"../../../../core-js/es7/reflect.js":function(e,t,r){r("../../../../core-js/modules/es7.reflect.define-metadata.js"),r("../../../../core-js/modules/es7.reflect.delete-metadata.js"),r("../../../../core-js/modules/es7.reflect.get-metadata.js"),r("../../../../core-js/modules/es7.reflect.get-metadata-keys.js"),r("../../../../core-js/modules/es7.reflect.get-own-metadata.js"),r("../../../../core-js/modules/es7.reflect.get-own-metadata-keys.js"),r("../../../../core-js/modules/es7.reflect.has-metadata.js"),r("../../../../core-js/modules/es7.reflect.has-own-metadata.js"),r("../../../../core-js/modules/es7.reflect.metadata.js"),e.exports=r("../../../../core-js/modules/_core.js").Reflect},"../../../../core-js/modules/_a-function.js":function(e,t){e.exports=function(e){if("function"!=typeof e)throw TypeError(e+" is not a function!");return e}},"../../../../core-js/modules/_an-instance.js":function(e,t){e.exports=function(e,t,r,o){if(!(e instanceof t)||void 0!==o&&o in e)throw TypeError(r+": incorrect invocation!");return e}},"../../../../core-js/modules/_an-object.js":function(e,t,r){var o=r("../../../../core-js/modules/_is-object.js");e.exports=function(e){if(!o(e))throw TypeError(e+" is not an object!");return e}},"../../../../core-js/modules/_array-from-iterable.js":function(e,t,r){var o=r("../../../../core-js/modules/_for-of.js");e.exports=function(e,t){var r=[];return o(e,!1,r.push,r,t),r}},"../../../../core-js/modules/_array-includes.js":function(e,t,r){var o=r("../../../../core-js/modules/_to-iobject.js"),n=r("../../../../core-js/modules/_to-length.js"),s=r("../../../../core-js/modules/_to-absolute-index.js");e.exports=function(e){return function(t,r,i){var a,c=o(t),u=n(c.length),l=s(i,u);if(e&&r!=r){for(;u>l;)if((a=c[l++])!=a)return!0}else for(;u>l;l++)if((e||l in c)&&c[l]===r)return e||l||0;return!e&&-1}}},"../../../../core-js/modules/_array-methods.js":function(e,t,r){var o=r("../../../../core-js/modules/_ctx.js"),n=r("../../../../core-js/modules/_iobject.js"),s=r("../../../../core-js/modules/_to-object.js"),i=r("../../../../core-js/modules/_to-length.js"),a=r("../../../../core-js/modules/_array-species-create.js");e.exports=function(e,t){var r=1==e,c=2==e,u=3==e,l=4==e,f=6==e,d=5==e||f,p=t||a;return function(t,a,h){for(var j,m,v=s(t),_=n(v),y=o(a,h,3),g=i(_.length),b=0,k=r?p(t,g):c?p(t,0):void 0;g>b;b++)if((d||b in _)&&(m=y(j=_[b],b,v),e))if(r)k[b]=m;else if(m)switch(e){case 3:return!0;case 5:return j;case 6:return b;case 2:k.push(j)}else if(l)return!1;return f?-1:u||l?l:k}}},"../../../../core-js/modules/_array-species-constructor.js":function(e,t,r){var o=r("../../../../core-js/modules/_is-object.js"),n=r("../../../../core-js/modules/_is-array.js"),s=r("../../../../core-js/modules/_wks.js")("species");e.exports=function(e){var t;return n(e)&&("function"!=typeof(t=e.constructor)||t!==Array&&!n(t.prototype)||(t=void 0),o(t)&&null===(t=t[s])&&(t=void 0)),void 0===t?Array:t}},"../../../../core-js/modules/_array-species-create.js":function(e,t,r){var o=r("../../../../core-js/modules/_array-species-constructor.js");e.exports=function(e,t){return new(o(e))(t)}},"../../../../core-js/modules/_classof.js":function(e,t,r){var o=r("../../../../core-js/modules/_cof.js"),n=r("../../../../core-js/modules/_wks.js")("toStringTag"),s="Arguments"==o(function(){return arguments}());e.exports=function(e){var t,r,i;return void 0===e?"Undefined":null===e?"Null":"string"==typeof(r=function(e,t){try{return e[t]}catch(e){}}(t=Object(e),n))?r:s?o(t):"Object"==(i=o(t))&&"function"==typeof t.callee?"Arguments":i}},"../../../../core-js/modules/_cof.js":function(e,t){var r={}.toString;e.exports=function(e){return r.call(e).slice(8,-1)}},"../../../../core-js/modules/_collection-strong.js":function(e,t,r){"use strict";var o=r("../../../../core-js/modules/_object-dp.js").f,n=r("../../../../core-js/modules/_object-create.js"),s=r("../../../../core-js/modules/_redefine-all.js"),i=r("../../../../core-js/modules/_ctx.js"),a=r("../../../../core-js/modules/_an-instance.js"),c=r("../../../../core-js/modules/_for-of.js"),u=r("../../../../core-js/modules/_iter-define.js"),l=r("../../../../core-js/modules/_iter-step.js"),f=r("../../../../core-js/modules/_set-species.js"),d=r("../../../../core-js/modules/_descriptors.js"),p=r("../../../../core-js/modules/_meta.js").fastKey,h=r("../../../../core-js/modules/_validate-collection.js"),j=d?"_s":"size",m=function(e,t){var r,o=p(t);if("F"!==o)return e._i[o];for(r=e._f;r;r=r.n)if(r.k==t)return r};e.exports={getConstructor:function(e,t,r,u){var l=e(function(e,o){a(e,l,t,"_i"),e._t=t,e._i=n(null),e._f=void 0,e._l=void 0,e[j]=0,void 0!=o&&c(o,r,e[u],e)});return s(l.prototype,{clear:function(){for(var e=h(this,t),r=e._i,o=e._f;o;o=o.n)o.r=!0,o.p&&(o.p=o.p.n=void 0),delete r[o.i];e._f=e._l=void 0,e[j]=0},delete:function(e){var r=h(this,t),o=m(r,e);if(o){var n=o.n,s=o.p;delete r._i[o.i],o.r=!0,s&&(s.n=n),n&&(n.p=s),r._f==o&&(r._f=n),r._l==o&&(r._l=s),r[j]--}return!!o},forEach:function(e){h(this,t);for(var r,o=i(e,arguments.length>1?arguments[1]:void 0,3);r=r?r.n:this._f;)for(o(r.v,r.k,this);r&&r.r;)r=r.p},has:function(e){return!!m(h(this,t),e)}}),d&&o(l.prototype,"size",{get:function(){return h(this,t)[j]}}),l},def:function(e,t,r){var o,n,s=m(e,t);return s?s.v=r:(e._l=s={i:n=p(t,!0),k:t,v:r,p:o=e._l,n:void 0,r:!1},e._f||(e._f=s),o&&(o.n=s),e[j]++,"F"!==n&&(e._i[n]=s)),e},getEntry:m,setStrong:function(e,t,r){u(e,t,function(e,r){this._t=h(e,t),this._k=r,this._l=void 0},function(){for(var e=this._k,t=this._l;t&&t.r;)t=t.p;return this._t&&(this._l=t=t?t.n:this._t._f)?l(0,"keys"==e?t.k:"values"==e?t.v:[t.k,t.v]):(this._t=void 0,l(1))},r?"entries":"values",!r,!0),f(t)}}},"../../../../core-js/modules/_collection-weak.js":function(e,t,r){"use strict";var o=r("../../../../core-js/modules/_redefine-all.js"),n=r("../../../../core-js/modules/_meta.js").getWeak,s=r("../../../../core-js/modules/_an-object.js"),i=r("../../../../core-js/modules/_is-object.js"),a=r("../../../../core-js/modules/_an-instance.js"),c=r("../../../../core-js/modules/_for-of.js"),u=r("../../../../core-js/modules/_array-methods.js"),l=r("../../../../core-js/modules/_has.js"),f=r("../../../../core-js/modules/_validate-collection.js"),d=u(5),p=u(6),h=0,j=function(e){return e._l||(e._l=new m)},m=function(){this.a=[]},v=function(e,t){return d(e.a,function(e){return e[0]===t})};m.prototype={get:function(e){var t=v(this,e);if(t)return t[1]},has:function(e){return!!v(this,e)},set:function(e,t){var r=v(this,e);r?r[1]=t:this.a.push([e,t])},delete:function(e){var t=p(this.a,function(t){return t[0]===e});return~t&&this.a.splice(t,1),!!~t}},e.exports={getConstructor:function(e,t,r,s){var u=e(function(e,o){a(e,u,t,"_i"),e._t=t,e._i=h++,e._l=void 0,void 0!=o&&c(o,r,e[s],e)});return o(u.prototype,{delete:function(e){if(!i(e))return!1;var r=n(e);return!0===r?j(f(this,t)).delete(e):r&&l(r,this._i)&&delete r[this._i]},has:function(e){if(!i(e))return!1;var r=n(e);return!0===r?j(f(this,t)).has(e):r&&l(r,this._i)}}),u},def:function(e,t,r){var o=n(s(t),!0);return!0===o?j(e).set(t,r):o[e._i]=r,e},ufstore:j}},"../../../../core-js/modules/_collection.js":function(e,t,r){"use strict";var o=r("../../../../core-js/modules/_global.js"),n=r("../../../../core-js/modules/_export.js"),s=r("../../../../core-js/modules/_redefine.js"),i=r("../../../../core-js/modules/_redefine-all.js"),a=r("../../../../core-js/modules/_meta.js"),c=r("../../../../core-js/modules/_for-of.js"),u=r("../../../../core-js/modules/_an-instance.js"),l=r("../../../../core-js/modules/_is-object.js"),f=r("../../../../core-js/modules/_fails.js"),d=r("../../../../core-js/modules/_iter-detect.js"),p=r("../../../../core-js/modules/_set-to-string-tag.js"),h=r("../../../../core-js/modules/_inherit-if-required.js");e.exports=function(e,t,r,j,m,v){var _=o[e],y=_,g=m?"set":"add",b=y&&y.prototype,k={},w=function(e){var t=b[e];s(b,e,"delete"==e?function(e){return!(v&&!l(e))&&t.call(this,0===e?0:e)}:"has"==e?function(e){return!(v&&!l(e))&&t.call(this,0===e?0:e)}:"get"==e?function(e){return v&&!l(e)?void 0:t.call(this,0===e?0:e)}:"add"==e?function(e){return t.call(this,0===e?0:e),this}:function(e,r){return t.call(this,0===e?0:e,r),this})};if("function"==typeof y&&(v||b.forEach&&!f(function(){(new y).entries().next()}))){var T=new y,E=T[g](v?{}:-0,1)!=T,O=f(function(){T.has(1)}),x=d(function(e){new y(e)}),S=!v&&f(function(){for(var e=new y,t=5;t--;)e[g](t,t);return!e.has(-0)});x||((y=t(function(t,r){u(t,y,e);var o=h(new _,t,y);return void 0!=r&&c(r,m,o[g],o),o})).prototype=b,b.constructor=y),(O||S)&&(w("delete"),w("has"),m&&w("get")),(S||E)&&w(g),v&&b.clear&&delete b.clear}else y=j.getConstructor(t,e,m,g),i(y.prototype,r),a.NEED=!0;return p(y,e),k[e]=y,n(n.G+n.W+n.F*(y!=_),k),v||j.setStrong(y,e,m),y}},"../../../../core-js/modules/_core.js":function(e,t){var r=e.exports={version:"2.5.1"};"number"==typeof __e&&(__e=r)},"../../../../core-js/modules/_ctx.js":function(e,t,r){var o=r("../../../../core-js/modules/_a-function.js");e.exports=function(e,t,r){if(o(e),void 0===t)return e;switch(r){case 1:return function(r){return e.call(t,r)};case 2:return function(r,o){return e.call(t,r,o)};case 3:return function(r,o,n){return e.call(t,r,o,n)}}return function(){return e.apply(t,arguments)}}},"../../../../core-js/modules/_defined.js":function(e,t){e.exports=function(e){if(void 0==e)throw TypeError("Can't call method on  "+e);return e}},"../../../../core-js/modules/_descriptors.js":function(e,t,r){e.exports=!r("../../../../core-js/modules/_fails.js")(function(){return 7!=Object.defineProperty({},"a",{get:function(){return 7}}).a})},"../../../../core-js/modules/_dom-create.js":function(e,t,r){var o=r("../../../../core-js/modules/_is-object.js"),n=r("../../../../core-js/modules/_global.js").document,s=o(n)&&o(n.createElement);e.exports=function(e){return s?n.createElement(e):{}}},"../../../../core-js/modules/_enum-bug-keys.js":function(e,t){e.exports="constructor,hasOwnProperty,isPrototypeOf,propertyIsEnumerable,toLocaleString,toString,valueOf".split(",")},"../../../../core-js/modules/_export.js":function(e,t,r){var o=r("../../../../core-js/modules/_global.js"),n=r("../../../../core-js/modules/_core.js"),s=r("../../../../core-js/modules/_hide.js"),i=r("../../../../core-js/modules/_redefine.js"),a=r("../../../../core-js/modules/_ctx.js"),c="prototype",u=function(e,t,r){var l,f,d,p,h=e&u.F,j=e&u.G,m=e&u.S,v=e&u.P,_=e&u.B,y=j?o:m?o[t]||(o[t]={}):(o[t]||{})[c],g=j?n:n[t]||(n[t]={}),b=g[c]||(g[c]={});j&&(r=t);for(l in r)d=((f=!h&&y&&void 0!==y[l])?y:r)[l],p=_&&f?a(d,o):v&&"function"==typeof d?a(Function.call,d):d,y&&i(y,l,d,e&u.U),g[l]!=d&&s(g,l,p),v&&b[l]!=d&&(b[l]=d)};o.core=n,u.F=1,u.G=2,u.S=4,u.P=8,u.B=16,u.W=32,u.U=64,u.R=128,e.exports=u},"../../../../core-js/modules/_fails.js":function(e,t){e.exports=function(e){try{return!!e()}catch(e){return!0}}},"../../../../core-js/modules/_for-of.js":function(e,t,r){var o=r("../../../../core-js/modules/_ctx.js"),n=r("../../../../core-js/modules/_iter-call.js"),s=r("../../../../core-js/modules/_is-array-iter.js"),i=r("../../../../core-js/modules/_an-object.js"),a=r("../../../../core-js/modules/_to-length.js"),c=r("../../../../core-js/modules/core.get-iterator-method.js"),u={},l={};(t=e.exports=function(e,t,r,f,d){var p,h,j,m,v=d?function(){return e}:c(e),_=o(r,f,t?2:1),y=0;if("function"!=typeof v)throw TypeError(e+" is not iterable!");if(s(v)){for(p=a(e.length);p>y;y++)if((m=t?_(i(h=e[y])[0],h[1]):_(e[y]))===u||m===l)return m}else for(j=v.call(e);!(h=j.next()).done;)if((m=n(j,_,h.value,t))===u||m===l)return m}).BREAK=u,t.RETURN=l},"../../../../core-js/modules/_global.js":function(e,t){var r=e.exports="undefined"!=typeof window&&window.Math==Math?window:"undefined"!=typeof self&&self.Math==Math?self:Function("return this")();"number"==typeof __g&&(__g=r)},"../../../../core-js/modules/_has.js":function(e,t){var r={}.hasOwnProperty;e.exports=function(e,t){return r.call(e,t)}},"../../../../core-js/modules/_hide.js":function(e,t,r){var o=r("../../../../core-js/modules/_object-dp.js"),n=r("../../../../core-js/modules/_property-desc.js");e.exports=r("../../../../core-js/modules/_descriptors.js")?function(e,t,r){return o.f(e,t,n(1,r))}:function(e,t,r){return e[t]=r,e}},"../../../../core-js/modules/_html.js":function(e,t,r){var o=r("../../../../core-js/modules/_global.js").document;e.exports=o&&o.documentElement},"../../../../core-js/modules/_ie8-dom-define.js":function(e,t,r){e.exports=!r("../../../../core-js/modules/_descriptors.js")&&!r("../../../../core-js/modules/_fails.js")(function(){return 7!=Object.defineProperty(r("../../../../core-js/modules/_dom-create.js")("div"),"a",{get:function(){return 7}}).a})},"../../../../core-js/modules/_inherit-if-required.js":function(e,t,r){var o=r("../../../../core-js/modules/_is-object.js"),n=r("../../../../core-js/modules/_set-proto.js").set;e.exports=function(e,t,r){var s,i=t.constructor;return i!==r&&"function"==typeof i&&(s=i.prototype)!==r.prototype&&o(s)&&n&&n(e,s),e}},"../../../../core-js/modules/_iobject.js":function(e,t,r){var o=r("../../../../core-js/modules/_cof.js");e.exports=Object("z").propertyIsEnumerable(0)?Object:function(e){return"String"==o(e)?e.split(""):Object(e)}},"../../../../core-js/modules/_is-array-iter.js":function(e,t,r){var o=r("../../../../core-js/modules/_iterators.js"),n=r("../../../../core-js/modules/_wks.js")("iterator"),s=Array.prototype;e.exports=function(e){return void 0!==e&&(o.Array===e||s[n]===e)}},"../../../../core-js/modules/_is-array.js":function(e,t,r){var o=r("../../../../core-js/modules/_cof.js");e.exports=Array.isArray||function(e){return"Array"==o(e)}},"../../../../core-js/modules/_is-object.js":function(e,t){e.exports=function(e){return"object"==typeof e?null!==e:"function"==typeof e}},"../../../../core-js/modules/_iter-call.js":function(e,t,r){var o=r("../../../../core-js/modules/_an-object.js");e.exports=function(e,t,r,n){try{return n?t(o(r)[0],r[1]):t(r)}catch(t){var s=e.return;throw void 0!==s&&o(s.call(e)),t}}},"../../../../core-js/modules/_iter-create.js":function(e,t,r){"use strict";var o=r("../../../../core-js/modules/_object-create.js"),n=r("../../../../core-js/modules/_property-desc.js"),s=r("../../../../core-js/modules/_set-to-string-tag.js"),i={};r("../../../../core-js/modules/_hide.js")(i,r("../../../../core-js/modules/_wks.js")("iterator"),function(){return this}),e.exports=function(e,t,r){e.prototype=o(i,{next:n(1,r)}),s(e,t+" Iterator")}},"../../../../core-js/modules/_iter-define.js":function(e,t,r){"use strict";var o=r("../../../../core-js/modules/_library.js"),n=r("../../../../core-js/modules/_export.js"),s=r("../../../../core-js/modules/_redefine.js"),i=r("../../../../core-js/modules/_hide.js"),a=r("../../../../core-js/modules/_has.js"),c=r("../../../../core-js/modules/_iterators.js"),u=r("../../../../core-js/modules/_iter-create.js"),l=r("../../../../core-js/modules/_set-to-string-tag.js"),f=r("../../../../core-js/modules/_object-gpo.js"),d=r("../../../../core-js/modules/_wks.js")("iterator"),p=!([].keys&&"next"in[].keys()),h="values",j=function(){return this};e.exports=function(e,t,r,m,v,_,y){u(r,t,m);var g,b,k,w=function(e){if(!p&&e in x)return x[e];switch(e){case"keys":case h:return function(){return new r(this,e)}}return function(){return new r(this,e)}},T=t+" Iterator",E=v==h,O=!1,x=e.prototype,S=x[d]||x["@@iterator"]||v&&x[v],P=S||w(v),D=v?E?w("entries"):P:void 0,Z="Array"==t?x.entries||S:S;if(Z&&(k=f(Z.call(new e)))!==Object.prototype&&k.next&&(l(k,T,!0),o||a(k,d)||i(k,d,j)),E&&S&&S.name!==h&&(O=!0,P=function(){return S.call(this)}),o&&!y||!p&&!O&&x[d]||i(x,d,P),c[t]=P,c[T]=j,v)if(g={values:E?P:w(h),keys:_?P:w("keys"),entries:D},y)for(b in g)b in x||s(x,b,g[b]);else n(n.P+n.F*(p||O),t,g);return g}},"../../../../core-js/modules/_iter-detect.js":function(e,t,r){var o=r("../../../../core-js/modules/_wks.js")("iterator"),n=!1;try{var s=[7][o]();s.return=function(){n=!0},Array.from(s,function(){throw 2})}catch(e){}e.exports=function(e,t){if(!t&&!n)return!1;var r=!1;try{var s=[7],i=s[o]();i.next=function(){return{done:r=!0}},s[o]=function(){return i},e(s)}catch(e){}return r}},"../../../../core-js/modules/_iter-step.js":function(e,t){e.exports=function(e,t){return{value:t,done:!!e}}},"../../../../core-js/modules/_iterators.js":function(e,t){e.exports={}},"../../../../core-js/modules/_library.js":function(e,t){e.exports=!1},"../../../../core-js/modules/_meta.js":function(e,t,r){var o=r("../../../../core-js/modules/_uid.js")("meta"),n=r("../../../../core-js/modules/_is-object.js"),s=r("../../../../core-js/modules/_has.js"),i=r("../../../../core-js/modules/_object-dp.js").f,a=0,c=Object.isExtensible||function(){return!0},u=!r("../../../../core-js/modules/_fails.js")(function(){return c(Object.preventExtensions({}))}),l=function(e){i(e,o,{value:{i:"O"+ ++a,w:{}}})},f=e.exports={KEY:o,NEED:!1,fastKey:function(e,t){if(!n(e))return"symbol"==typeof e?e:("string"==typeof e?"S":"P")+e;if(!s(e,o)){if(!c(e))return"F";if(!t)return"E";l(e)}return e[o].i},getWeak:function(e,t){if(!s(e,o)){if(!c(e))return!0;if(!t)return!1;l(e)}return e[o].w},onFreeze:function(e){return u&&f.NEED&&c(e)&&!s(e,o)&&l(e),e}}},"../../../../core-js/modules/_metadata.js":function(e,t,r){var o=r("../../../../core-js/modules/es6.map.js"),n=r("../../../../core-js/modules/_export.js"),s=r("../../../../core-js/modules/_shared.js")("metadata"),i=s.store||(s.store=new(r("../../../../core-js/modules/es6.weak-map.js"))),a=function(e,t,r){var n=i.get(e);if(!n){if(!r)return;i.set(e,n=new o)}var s=n.get(t);if(!s){if(!r)return;n.set(t,s=new o)}return s};e.exports={store:i,map:a,has:function(e,t,r){var o=a(t,r,!1);return void 0!==o&&o.has(e)},get:function(e,t,r){var o=a(t,r,!1);return void 0===o?void 0:o.get(e)},set:function(e,t,r,o){a(r,o,!0).set(e,t)},keys:function(e,t){var r=a(e,t,!1),o=[];return r&&r.forEach(function(e,t){o.push(t)}),o},key:function(e){return void 0===e||"symbol"==typeof e?e:String(e)},exp:function(e){n(n.S,"Reflect",e)}}},"../../../../core-js/modules/_object-assign.js":function(e,t,r){"use strict";var o=r("../../../../core-js/modules/_object-keys.js"),n=r("../../../../core-js/modules/_object-gops.js"),s=r("../../../../core-js/modules/_object-pie.js"),i=r("../../../../core-js/modules/_to-object.js"),a=r("../../../../core-js/modules/_iobject.js"),c=Object.assign;e.exports=!c||r("../../../../core-js/modules/_fails.js")(function(){var e={},t={},r=Symbol(),o="abcdefghijklmnopqrst";return e[r]=7,o.split("").forEach(function(e){t[e]=e}),7!=c({},e)[r]||Object.keys(c({},t)).join("")!=o})?function(e,t){for(var r=i(e),c=arguments.length,u=1,l=n.f,f=s.f;c>u;)for(var d,p=a(arguments[u++]),h=l?o(p).concat(l(p)):o(p),j=h.length,m=0;j>m;)f.call(p,d=h[m++])&&(r[d]=p[d]);return r}:c},"../../../../core-js/modules/_object-create.js":function(e,t,r){var o=r("../../../../core-js/modules/_an-object.js"),n=r("../../../../core-js/modules/_object-dps.js"),s=r("../../../../core-js/modules/_enum-bug-keys.js"),i=r("../../../../core-js/modules/_shared-key.js")("IE_PROTO"),a=function(){},c="prototype",u=function(){var e,t=r("../../../../core-js/modules/_dom-create.js")("iframe"),o=s.length;for(t.style.display="none",r("../../../../core-js/modules/_html.js").appendChild(t),t.src="javascript:",(e=t.contentWindow.document).open(),e.write("<script>document.F=Object<\/script>"),e.close(),u=e.F;o--;)delete u[c][s[o]];return u()};e.exports=Object.create||function(e,t){var r;return null!==e?(a[c]=o(e),r=new a,a[c]=null,r[i]=e):r=u(),void 0===t?r:n(r,t)}},"../../../../core-js/modules/_object-dp.js":function(e,t,r){var o=r("../../../../core-js/modules/_an-object.js"),n=r("../../../../core-js/modules/_ie8-dom-define.js"),s=r("../../../../core-js/modules/_to-primitive.js"),i=Object.defineProperty;t.f=r("../../../../core-js/modules/_descriptors.js")?Object.defineProperty:function(e,t,r){if(o(e),t=s(t,!0),o(r),n)try{return i(e,t,r)}catch(e){}if("get"in r||"set"in r)throw TypeError("Accessors not supported!");return"value"in r&&(e[t]=r.value),e}},"../../../../core-js/modules/_object-dps.js":function(e,t,r){var o=r("../../../../core-js/modules/_object-dp.js"),n=r("../../../../core-js/modules/_an-object.js"),s=r("../../../../core-js/modules/_object-keys.js");e.exports=r("../../../../core-js/modules/_descriptors.js")?Object.defineProperties:function(e,t){n(e);for(var r,i=s(t),a=i.length,c=0;a>c;)o.f(e,r=i[c++],t[r]);return e}},"../../../../core-js/modules/_object-gopd.js":function(e,t,r){var o=r("../../../../core-js/modules/_object-pie.js"),n=r("../../../../core-js/modules/_property-desc.js"),s=r("../../../../core-js/modules/_to-iobject.js"),i=r("../../../../core-js/modules/_to-primitive.js"),a=r("../../../../core-js/modules/_has.js"),c=r("../../../../core-js/modules/_ie8-dom-define.js"),u=Object.getOwnPropertyDescriptor;t.f=r("../../../../core-js/modules/_descriptors.js")?u:function(e,t){if(e=s(e),t=i(t,!0),c)try{return u(e,t)}catch(e){}if(a(e,t))return n(!o.f.call(e,t),e[t])}},"../../../../core-js/modules/_object-gops.js":function(e,t){t.f=Object.getOwnPropertySymbols},"../../../../core-js/modules/_object-gpo.js":function(e,t,r){var o=r("../../../../core-js/modules/_has.js"),n=r("../../../../core-js/modules/_to-object.js"),s=r("../../../../core-js/modules/_shared-key.js")("IE_PROTO"),i=Object.prototype;e.exports=Object.getPrototypeOf||function(e){return e=n(e),o(e,s)?e[s]:"function"==typeof e.constructor&&e instanceof e.constructor?e.constructor.prototype:e instanceof Object?i:null}},"../../../../core-js/modules/_object-keys-internal.js":function(e,t,r){var o=r("../../../../core-js/modules/_has.js"),n=r("../../../../core-js/modules/_to-iobject.js"),s=r("../../../../core-js/modules/_array-includes.js")(!1),i=r("../../../../core-js/modules/_shared-key.js")("IE_PROTO");e.exports=function(e,t){var r,a=n(e),c=0,u=[];for(r in a)r!=i&&o(a,r)&&u.push(r);for(;t.length>c;)o(a,r=t[c++])&&(~s(u,r)||u.push(r));return u}},"../../../../core-js/modules/_object-keys.js":function(e,t,r){var o=r("../../../../core-js/modules/_object-keys-internal.js"),n=r("../../../../core-js/modules/_enum-bug-keys.js");e.exports=Object.keys||function(e){return o(e,n)}},"../../../../core-js/modules/_object-pie.js":function(e,t){t.f={}.propertyIsEnumerable},"../../../../core-js/modules/_property-desc.js":function(e,t){e.exports=function(e,t){return{enumerable:!(1&e),configurable:!(2&e),writable:!(4&e),value:t}}},"../../../../core-js/modules/_redefine-all.js":function(e,t,r){var o=r("../../../../core-js/modules/_redefine.js");e.exports=function(e,t,r){for(var n in t)o(e,n,t[n],r);return e}},"../../../../core-js/modules/_redefine.js":function(e,t,r){var o=r("../../../../core-js/modules/_global.js"),n=r("../../../../core-js/modules/_hide.js"),s=r("../../../../core-js/modules/_has.js"),i=r("../../../../core-js/modules/_uid.js")("src"),a=Function.toString,c=(""+a).split("toString");r("../../../../core-js/modules/_core.js").inspectSource=function(e){return a.call(e)},(e.exports=function(e,t,r,a){var u="function"==typeof r;u&&(s(r,"name")||n(r,"name",t)),e[t]!==r&&(u&&(s(r,i)||n(r,i,e[t]?""+e[t]:c.join(String(t)))),e===o?e[t]=r:a?e[t]?e[t]=r:n(e,t,r):(delete e[t],n(e,t,r)))})(Function.prototype,"toString",function(){return"function"==typeof this&&this[i]||a.call(this)})},"../../../../core-js/modules/_set-proto.js":function(e,t,r){var o=r("../../../../core-js/modules/_is-object.js"),n=r("../../../../core-js/modules/_an-object.js"),s=function(e,t){if(n(e),!o(t)&&null!==t)throw TypeError(t+": can't set as prototype!")};e.exports={set:Object.setPrototypeOf||("__proto__"in{}?function(e,t,o){try{(o=r("../../../../core-js/modules/_ctx.js")(Function.call,r("../../../../core-js/modules/_object-gopd.js").f(Object.prototype,"__proto__").set,2))(e,[]),t=!(e instanceof Array)}catch(e){t=!0}return function(e,r){return s(e,r),t?e.__proto__=r:o(e,r),e}}({},!1):void 0),check:s}},"../../../../core-js/modules/_set-species.js":function(e,t,r){"use strict";var o=r("../../../../core-js/modules/_global.js"),n=r("../../../../core-js/modules/_object-dp.js"),s=r("../../../../core-js/modules/_descriptors.js"),i=r("../../../../core-js/modules/_wks.js")("species");e.exports=function(e){var t=o[e];s&&t&&!t[i]&&n.f(t,i,{configurable:!0,get:function(){return this}})}},"../../../../core-js/modules/_set-to-string-tag.js":function(e,t,r){var o=r("../../../../core-js/modules/_object-dp.js").f,n=r("../../../../core-js/modules/_has.js"),s=r("../../../../core-js/modules/_wks.js")("toStringTag");e.exports=function(e,t,r){e&&!n(e=r?e:e.prototype,s)&&o(e,s,{configurable:!0,value:t})}},"../../../../core-js/modules/_shared-key.js":function(e,t,r){var o=r("../../../../core-js/modules/_shared.js")("keys"),n=r("../../../../core-js/modules/_uid.js");e.exports=function(e){return o[e]||(o[e]=n(e))}},"../../../../core-js/modules/_shared.js":function(e,t,r){var o=r("../../../../core-js/modules/_global.js"),n="__core-js_shared__",s=o[n]||(o[n]={});e.exports=function(e){return s[e]||(s[e]={})}},"../../../../core-js/modules/_to-absolute-index.js":function(e,t,r){var o=r("../../../../core-js/modules/_to-integer.js"),n=Math.max,s=Math.min;e.exports=function(e,t){return(e=o(e))<0?n(e+t,0):s(e,t)}},"../../../../core-js/modules/_to-integer.js":function(e,t){var r=Math.ceil,o=Math.floor;e.exports=function(e){return isNaN(e=+e)?0:(e>0?o:r)(e)}},"../../../../core-js/modules/_to-iobject.js":function(e,t,r){var o=r("../../../../core-js/modules/_iobject.js"),n=r("../../../../core-js/modules/_defined.js");e.exports=function(e){return o(n(e))}},"../../../../core-js/modules/_to-length.js":function(e,t,r){var o=r("../../../../core-js/modules/_to-integer.js"),n=Math.min;e.exports=function(e){return e>0?n(o(e),9007199254740991):0}},"../../../../core-js/modules/_to-object.js":function(e,t,r){var o=r("../../../../core-js/modules/_defined.js");e.exports=function(e){return Object(o(e))}},"../../../../core-js/modules/_to-primitive.js":function(e,t,r){var o=r("../../../../core-js/modules/_is-object.js");e.exports=function(e,t){if(!o(e))return e;var r,n;if(t&&"function"==typeof(r=e.toString)&&!o(n=r.call(e)))return n;if("function"==typeof(r=e.valueOf)&&!o(n=r.call(e)))return n;if(!t&&"function"==typeof(r=e.toString)&&!o(n=r.call(e)))return n;throw TypeError("Can't convert object to primitive value")}},"../../../../core-js/modules/_uid.js":function(e,t){var r=0,o=Math.random();e.exports=function(e){return"Symbol(".concat(void 0===e?"":e,")_",(++r+o).toString(36))}},"../../../../core-js/modules/_validate-collection.js":function(e,t,r){var o=r("../../../../core-js/modules/_is-object.js");e.exports=function(e,t){if(!o(e)||e._t!==t)throw TypeError("Incompatible receiver, "+t+" required!");return e}},"../../../../core-js/modules/_wks.js":function(e,t,r){var o=r("../../../../core-js/modules/_shared.js")("wks"),n=r("../../../../core-js/modules/_uid.js"),s=r("../../../../core-js/modules/_global.js").Symbol,i="function"==typeof s;(e.exports=function(e){return o[e]||(o[e]=i&&s[e]||(i?s:n)("Symbol."+e))}).store=o},"../../../../core-js/modules/core.get-iterator-method.js":function(e,t,r){var o=r("../../../../core-js/modules/_classof.js"),n=r("../../../../core-js/modules/_wks.js")("iterator"),s=r("../../../../core-js/modules/_iterators.js");e.exports=r("../../../../core-js/modules/_core.js").getIteratorMethod=function(e){if(void 0!=e)return e[n]||e["@@iterator"]||s[o(e)]}},"../../../../core-js/modules/es6.map.js":function(e,t,r){"use strict";var o=r("../../../../core-js/modules/_collection-strong.js"),n=r("../../../../core-js/modules/_validate-collection.js");e.exports=r("../../../../core-js/modules/_collection.js")("Map",function(e){return function(){return e(this,arguments.length>0?arguments[0]:void 0)}},{get:function(e){var t=o.getEntry(n(this,"Map"),e);return t&&t.v},set:function(e,t){return o.def(n(this,"Map"),0===e?0:e,t)}},o,!0)},"../../../../core-js/modules/es6.set.js":function(e,t,r){"use strict";var o=r("../../../../core-js/modules/_collection-strong.js"),n=r("../../../../core-js/modules/_validate-collection.js");e.exports=r("../../../../core-js/modules/_collection.js")("Set",function(e){return function(){return e(this,arguments.length>0?arguments[0]:void 0)}},{add:function(e){return o.def(n(this,"Set"),e=0===e?0:e,e)}},o)},"../../../../core-js/modules/es6.weak-map.js":function(e,t,r){"use strict";var o,n=r("../../../../core-js/modules/_array-methods.js")(0),s=r("../../../../core-js/modules/_redefine.js"),i=r("../../../../core-js/modules/_meta.js"),a=r("../../../../core-js/modules/_object-assign.js"),c=r("../../../../core-js/modules/_collection-weak.js"),u=r("../../../../core-js/modules/_is-object.js"),l=r("../../../../core-js/modules/_fails.js"),f=r("../../../../core-js/modules/_validate-collection.js"),d=i.getWeak,p=Object.isExtensible,h=c.ufstore,j={},m=function(e){return function(){return e(this,arguments.length>0?arguments[0]:void 0)}},v={get:function(e){if(u(e)){var t=d(e);return!0===t?h(f(this,"WeakMap")).get(e):t?t[this._i]:void 0}},set:function(e,t){return c.def(f(this,"WeakMap"),e,t)}},_=e.exports=r("../../../../core-js/modules/_collection.js")("WeakMap",m,v,c,!0,!0);l(function(){return 7!=(new _).set((Object.freeze||Object)(j),7).get(j)})&&(a((o=c.getConstructor(m,"WeakMap")).prototype,v),i.NEED=!0,n(["delete","has","get","set"],function(e){var t=_.prototype,r=t[e];s(t,e,function(t,n){if(u(t)&&!p(t)){this._f||(this._f=new o);var s=this._f[e](t,n);return"set"==e?this:s}return r.call(this,t,n)})}))},"../../../../core-js/modules/es7.reflect.define-metadata.js":function(e,t,r){var o=r("../../../../core-js/modules/_metadata.js"),n=r("../../../../core-js/modules/_an-object.js"),s=o.key,i=o.set;o.exp({defineMetadata:function(e,t,r,o){i(e,t,n(r),s(o))}})},"../../../../core-js/modules/es7.reflect.delete-metadata.js":function(e,t,r){var o=r("../../../../core-js/modules/_metadata.js"),n=r("../../../../core-js/modules/_an-object.js"),s=o.key,i=o.map,a=o.store;o.exp({deleteMetadata:function(e,t){var r=arguments.length<3?void 0:s(arguments[2]),o=i(n(t),r,!1);if(void 0===o||!o.delete(e))return!1;if(o.size)return!0;var c=a.get(t);return c.delete(r),!!c.size||a.delete(t)}})},"../../../../core-js/modules/es7.reflect.get-metadata-keys.js":function(e,t,r){var o=r("../../../../core-js/modules/es6.set.js"),n=r("../../../../core-js/modules/_array-from-iterable.js"),s=r("../../../../core-js/modules/_metadata.js"),i=r("../../../../core-js/modules/_an-object.js"),a=r("../../../../core-js/modules/_object-gpo.js"),c=s.keys,u=s.key,l=function(e,t){var r=c(e,t),s=a(e);if(null===s)return r;var i=l(s,t);return i.length?r.length?n(new o(r.concat(i))):i:r};s.exp({getMetadataKeys:function(e){return l(i(e),arguments.length<2?void 0:u(arguments[1]))}})},"../../../../core-js/modules/es7.reflect.get-metadata.js":function(e,t,r){var o=r("../../../../core-js/modules/_metadata.js"),n=r("../../../../core-js/modules/_an-object.js"),s=r("../../../../core-js/modules/_object-gpo.js"),i=o.has,a=o.get,c=o.key,u=function(e,t,r){if(i(e,t,r))return a(e,t,r);var o=s(t);return null!==o?u(e,o,r):void 0};o.exp({getMetadata:function(e,t){return u(e,n(t),arguments.length<3?void 0:c(arguments[2]))}})},"../../../../core-js/modules/es7.reflect.get-own-metadata-keys.js":function(e,t,r){var o=r("../../../../core-js/modules/_metadata.js"),n=r("../../../../core-js/modules/_an-object.js"),s=o.keys,i=o.key;o.exp({getOwnMetadataKeys:function(e){return s(n(e),arguments.length<2?void 0:i(arguments[1]))}})},"../../../../core-js/modules/es7.reflect.get-own-metadata.js":function(e,t,r){var o=r("../../../../core-js/modules/_metadata.js"),n=r("../../../../core-js/modules/_an-object.js"),s=o.get,i=o.key;o.exp({getOwnMetadata:function(e,t){return s(e,n(t),arguments.length<3?void 0:i(arguments[2]))}})},"../../../../core-js/modules/es7.reflect.has-metadata.js":function(e,t,r){var o=r("../../../../core-js/modules/_metadata.js"),n=r("../../../../core-js/modules/_an-object.js"),s=r("../../../../core-js/modules/_object-gpo.js"),i=o.has,a=o.key,c=function(e,t,r){if(i(e,t,r))return!0;var o=s(t);return null!==o&&c(e,o,r)};o.exp({hasMetadata:function(e,t){return c(e,n(t),arguments.length<3?void 0:a(arguments[2]))}})},"../../../../core-js/modules/es7.reflect.has-own-metadata.js":function(e,t,r){var o=r("../../../../core-js/modules/_metadata.js"),n=r("../../../../core-js/modules/_an-object.js"),s=o.has,i=o.key;o.exp({hasOwnMetadata:function(e,t){return s(e,n(t),arguments.length<3?void 0:i(arguments[2]))}})},"../../../../core-js/modules/es7.reflect.metadata.js":function(e,t,r){var o=r("../../../../core-js/modules/_metadata.js"),n=r("../../../../core-js/modules/_an-object.js"),s=r("../../../../core-js/modules/_a-function.js"),i=o.key,a=o.set;o.exp({metadata:function(e,t){return function(r,o){a(e,t,(void 0!==o?n:s)(r),i(o))}}})},"../../../../webpack/buildin/global.js":function(e,t){var r;r=function(){return this}();try{r=r||Function("return this")()||(0,eval)("this")}catch(e){"object"==typeof window&&(r=window)}e.exports=r},"../../../../zone.js/dist/zone.js":function(e,t,r){(function(e){(function(){"use strict";!function(e){var t=e.performance;function r(e){t&&t.mark&&t.mark(e)}function o(e,r){t&&t.measure&&t.measure(e,r)}if(r("Zone"),e.Zone)throw new Error("Zone already loaded.");var n=function(){function t(e,t){this._properties=null,this._parent=e,this._name=t?t.name||"unnamed":"<root>",this._properties=t&&t.properties||{},this._zoneDelegate=new a(this,this._parent&&this._parent._zoneDelegate,t)}return t.assertZonePatched=function(){if(e.Promise!==O.ZoneAwarePromise)throw new Error("Zone.js has detected that ZoneAwarePromise `(window|global).Promise` has been overwritten.\nMost likely cause is that a Promise polyfill has been loaded after Zone.js (Polyfilling Promise api is not necessary when zone.js is loaded. If you must load one, do so before loading zone.js.)")},Object.defineProperty(t,"root",{get:function(){for(var e=t.current;e.parent;)e=e.parent;return e},enumerable:!0,configurable:!0}),Object.defineProperty(t,"current",{get:function(){return S.zone},enumerable:!0,configurable:!0}),Object.defineProperty(t,"currentTask",{get:function(){return P},enumerable:!0,configurable:!0}),t.__load_patch=function(n,s){if(O.hasOwnProperty(n))throw Error("Already loaded patch: "+n);if(!e["__Zone_disable_"+n]){var i="Zone:"+n;r(i),O[n]=s(e,t,x),o(i,i)}},Object.defineProperty(t.prototype,"parent",{get:function(){return this._parent},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"name",{get:function(){return this._name},enumerable:!0,configurable:!0}),t.prototype.get=function(e){var t=this.getZoneWith(e);if(t)return t._properties[e]},t.prototype.getZoneWith=function(e){for(var t=this;t;){if(t._properties.hasOwnProperty(e))return t;t=t._parent}return null},t.prototype.fork=function(e){if(!e)throw new Error("ZoneSpec required!");return this._zoneDelegate.fork(this,e)},t.prototype.wrap=function(e,t){if("function"!=typeof e)throw new Error("Expecting function got: "+e);var r=this._zoneDelegate.intercept(this,e,t),o=this;return function(){return o.runGuarded(r,this,arguments,t)}},t.prototype.run=function(e,t,r,o){void 0===t&&(t=void 0),void 0===r&&(r=null),void 0===o&&(o=null),S={parent:S,zone:this};try{return this._zoneDelegate.invoke(this,e,t,r,o)}finally{S=S.parent}},t.prototype.runGuarded=function(e,t,r,o){void 0===t&&(t=null),void 0===r&&(r=null),void 0===o&&(o=null),S={parent:S,zone:this};try{try{return this._zoneDelegate.invoke(this,e,t,r,o)}catch(e){if(this._zoneDelegate.handleError(this,e))throw e}}finally{S=S.parent}},t.prototype.runTask=function(e,t,r){if(e.zone!=this)throw new Error("A task can only be run in the zone of creation! (Creation: "+(e.zone||m).name+"; Execution: "+this.name+")");if(!(e.state===v)||e.type!==E){var o=e.state!=g;o&&e._transitionTo(g,y),e.runCount++;var n=P;P=e,S={parent:S,zone:this};try{e.type==T&&e.data&&!e.data.isPeriodic&&(e.cancelFn=null);try{return this._zoneDelegate.invokeTask(this,e,t,r)}catch(e){if(this._zoneDelegate.handleError(this,e))throw e}}finally{e.state!==v&&e.state!==k&&(e.type==E||e.data&&e.data.isPeriodic?o&&e._transitionTo(y,g):(e.runCount=0,this._updateTaskCount(e,-1),o&&e._transitionTo(v,g,v))),S=S.parent,P=n}}},t.prototype.scheduleTask=function(e){if(e.zone&&e.zone!==this)for(var t=this;t;){if(t===e.zone)throw Error("can not reschedule task to "+this.name+" which is descendants of the original zone "+e.zone.name);t=t.parent}e._transitionTo(_,v);var r=[];e._zoneDelegates=r,e._zone=this;try{e=this._zoneDelegate.scheduleTask(this,e)}catch(t){throw e._transitionTo(k,_,v),this._zoneDelegate.handleError(this,t),t}return e._zoneDelegates===r&&this._updateTaskCount(e,1),e.state==_&&e._transitionTo(y,_),e},t.prototype.scheduleMicroTask=function(e,t,r,o){return this.scheduleTask(new c(w,e,t,r,o,null))},t.prototype.scheduleMacroTask=function(e,t,r,o,n){return this.scheduleTask(new c(T,e,t,r,o,n))},t.prototype.scheduleEventTask=function(e,t,r,o,n){return this.scheduleTask(new c(E,e,t,r,o,n))},t.prototype.cancelTask=function(e){if(e.zone!=this)throw new Error("A task can only be cancelled in the zone of creation! (Creation: "+(e.zone||m).name+"; Execution: "+this.name+")");e._transitionTo(b,y,g);try{this._zoneDelegate.cancelTask(this,e)}catch(t){throw e._transitionTo(k,b),this._zoneDelegate.handleError(this,t),t}return this._updateTaskCount(e,-1),e._transitionTo(v,b),e.runCount=0,e},t.prototype._updateTaskCount=function(e,t){var r=e._zoneDelegates;-1==t&&(e._zoneDelegates=null);for(var o=0;o<r.length;o++)r[o]._updateTaskCount(e.type,t)},t}();n.__symbol__=z;var s,i={name:"",onHasTask:function(e,t,r,o){return e.hasTask(r,o)},onScheduleTask:function(e,t,r,o){return e.scheduleTask(r,o)},onInvokeTask:function(e,t,r,o,n,s){return e.invokeTask(r,o,n,s)},onCancelTask:function(e,t,r,o){return e.cancelTask(r,o)}},a=function(){function e(e,t,r){this._taskCounts={microTask:0,macroTask:0,eventTask:0},this.zone=e,this._parentDelegate=t,this._forkZS=r&&(r&&r.onFork?r:t._forkZS),this._forkDlgt=r&&(r.onFork?t:t._forkDlgt),this._forkCurrZone=r&&(r.onFork?this.zone:t.zone),this._interceptZS=r&&(r.onIntercept?r:t._interceptZS),this._interceptDlgt=r&&(r.onIntercept?t:t._interceptDlgt),this._interceptCurrZone=r&&(r.onIntercept?this.zone:t.zone),this._invokeZS=r&&(r.onInvoke?r:t._invokeZS),this._invokeDlgt=r&&(r.onInvoke?t:t._invokeDlgt),this._invokeCurrZone=r&&(r.onInvoke?this.zone:t.zone),this._handleErrorZS=r&&(r.onHandleError?r:t._handleErrorZS),this._handleErrorDlgt=r&&(r.onHandleError?t:t._handleErrorDlgt),this._handleErrorCurrZone=r&&(r.onHandleError?this.zone:t.zone),this._scheduleTaskZS=r&&(r.onScheduleTask?r:t._scheduleTaskZS),this._scheduleTaskDlgt=r&&(r.onScheduleTask?t:t._scheduleTaskDlgt),this._scheduleTaskCurrZone=r&&(r.onScheduleTask?this.zone:t.zone),this._invokeTaskZS=r&&(r.onInvokeTask?r:t._invokeTaskZS),this._invokeTaskDlgt=r&&(r.onInvokeTask?t:t._invokeTaskDlgt),this._invokeTaskCurrZone=r&&(r.onInvokeTask?this.zone:t.zone),this._cancelTaskZS=r&&(r.onCancelTask?r:t._cancelTaskZS),this._cancelTaskDlgt=r&&(r.onCancelTask?t:t._cancelTaskDlgt),this._cancelTaskCurrZone=r&&(r.onCancelTask?this.zone:t.zone),this._hasTaskZS=null,this._hasTaskDlgt=null,this._hasTaskDlgtOwner=null,this._hasTaskCurrZone=null;var o=r&&r.onHasTask,n=t&&t._hasTaskZS;(o||n)&&(this._hasTaskZS=o?r:i,this._hasTaskDlgt=t,this._hasTaskDlgtOwner=this,this._hasTaskCurrZone=e,r.onScheduleTask||(this._scheduleTaskZS=i,this._scheduleTaskDlgt=t,this._scheduleTaskCurrZone=this.zone),r.onInvokeTask||(this._invokeTaskZS=i,this._invokeTaskDlgt=t,this._invokeTaskCurrZone=this.zone),r.onCancelTask||(this._cancelTaskZS=i,this._cancelTaskDlgt=t,this._cancelTaskCurrZone=this.zone))}return e.prototype.fork=function(e,t){return this._forkZS?this._forkZS.onFork(this._forkDlgt,this.zone,e,t):new n(e,t)},e.prototype.intercept=function(e,t,r){return this._interceptZS?this._interceptZS.onIntercept(this._interceptDlgt,this._interceptCurrZone,e,t,r):t},e.prototype.invoke=function(e,t,r,o,n){return this._invokeZS?this._invokeZS.onInvoke(this._invokeDlgt,this._invokeCurrZone,e,t,r,o,n):t.apply(r,o)},e.prototype.handleError=function(e,t){return!this._handleErrorZS||this._handleErrorZS.onHandleError(this._handleErrorDlgt,this._handleErrorCurrZone,e,t)},e.prototype.scheduleTask=function(e,t){var r=t;if(this._scheduleTaskZS)this._hasTaskZS&&r._zoneDelegates.push(this._hasTaskDlgtOwner),(r=this._scheduleTaskZS.onScheduleTask(this._scheduleTaskDlgt,this._scheduleTaskCurrZone,e,t))||(r=t);else if(t.scheduleFn)t.scheduleFn(t);else{if(t.type!=w)throw new Error("Task is missing scheduleFn.");h(t)}return r},e.prototype.invokeTask=function(e,t,r,o){return this._invokeTaskZS?this._invokeTaskZS.onInvokeTask(this._invokeTaskDlgt,this._invokeTaskCurrZone,e,t,r,o):t.callback.apply(r,o)},e.prototype.cancelTask=function(e,t){var r;if(this._cancelTaskZS)r=this._cancelTaskZS.onCancelTask(this._cancelTaskDlgt,this._cancelTaskCurrZone,e,t);else{if(!t.cancelFn)throw Error("Task is not cancelable");r=t.cancelFn(t)}return r},e.prototype.hasTask=function(e,t){try{return this._hasTaskZS&&this._hasTaskZS.onHasTask(this._hasTaskDlgt,this._hasTaskCurrZone,e,t)}catch(t){this.handleError(e,t)}},e.prototype._updateTaskCount=function(e,t){var r=this._taskCounts,o=r[e],n=r[e]=o+t;if(n<0)throw new Error("More tasks executed then were scheduled.");if(0==o||0==n){var s={microTask:r.microTask>0,macroTask:r.macroTask>0,eventTask:r.eventTask>0,change:e};this.hasTask(this.zone,s)}},e}(),c=function(){function t(r,o,n,s,i,a){this._zone=null,this.runCount=0,this._zoneDelegates=null,this._state="notScheduled",this.type=r,this.source=o,this.data=s,this.scheduleFn=i,this.cancelFn=a,this.callback=n;var c=this;r===E&&s&&s.isUsingGlobalCallback?this.invoke=t.invokeTask:this.invoke=function(){return t.invokeTask.apply(e,[c,this,arguments])}}return t.invokeTask=function(e,t,r){e||(e=this),D++;try{return e.runCount++,e.zone.runTask(e,t,r)}finally{1==D&&j(),D--}},Object.defineProperty(t.prototype,"zone",{get:function(){return this._zone},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"state",{get:function(){return this._state},enumerable:!0,configurable:!0}),t.prototype.cancelScheduleRequest=function(){this._transitionTo(v,_)},t.prototype._transitionTo=function(e,t,r){if(this._state!==t&&this._state!==r)throw new Error(this.type+" '"+this.source+"': can not transition to '"+e+"', expecting state '"+t+"'"+(r?" or '"+r+"'":"")+", was '"+this._state+"'.");this._state=e,e==v&&(this._zoneDelegates=null)},t.prototype.toString=function(){return this.data&&void 0!==this.data.handleId?this.data.handleId:Object.prototype.toString.call(this)},t.prototype.toJSON=function(){return{type:this.type,state:this.state,source:this.source,zone:this.zone.name,invoke:this.invoke,scheduleFn:this.scheduleFn,cancelFn:this.cancelFn,runCount:this.runCount,callback:this.callback}},t}(),u=z("setTimeout"),l=z("Promise"),f=z("then"),d=[],p=!1;function h(t){0===D&&0===d.length&&(s||e[l]&&(s=e[l].resolve(0)),s?s[f](j):e[u](j,0)),t&&d.push(t)}function j(){if(!p){for(p=!0;d.length;){var e=d;d=[];for(var t=0;t<e.length;t++){var r=e[t];try{r.zone.runTask(r,null,null)}catch(e){x.onUnhandledError(e)}}}n[z("ignoreConsoleErrorUncaughtError")];x.microtaskDrainDone(),p=!1}}var m={name:"NO ZONE"},v="notScheduled",_="scheduling",y="scheduled",g="running",b="canceling",k="unknown",w="microTask",T="macroTask",E="eventTask",O={},x={symbol:z,currentZoneFrame:function(){return S},onUnhandledError:Z,microtaskDrainDone:Z,scheduleMicroTask:h,showUncaughtError:function(){return!n[z("ignoreConsoleErrorUncaughtError")]},patchEventTarget:function(){return[]},patchOnProperties:Z,patchMethod:function(){return Z},setNativePromise:function(e){s=e.resolve(0)}},S={parent:null,zone:new n(null,null)},P=null,D=0;function Z(){}function z(e){return"__zone_symbol__"+e}o("Zone","Zone"),e.Zone=n}("undefined"!=typeof window&&window||"undefined"!=typeof self&&self||e);var t=function(e,t){var r="function"==typeof Symbol&&e[Symbol.iterator];if(!r)return e;var o,n,s=r.call(e),i=[];try{for(;(void 0===t||t-- >0)&&!(o=s.next()).done;)i.push(o.value)}catch(e){n={error:e}}finally{try{o&&!o.done&&(r=s.return)&&r.call(s)}finally{if(n)throw n.error}}return i},r=function(e){var t="function"==typeof Symbol&&e[Symbol.iterator],r=0;return t?t.call(e):{next:function(){return e&&r>=e.length&&(e=void 0),{value:e&&e[r++],done:!e}}}};Zone.__load_patch("ZoneAwarePromise",function(e,o,n){var s=n.symbol,i=[],a=s("Promise"),c=s("then");n.onUnhandledError=function(e){if(n.showUncaughtError()){var t=e&&e.rejection;t?console.error("Unhandled Promise rejection:",t instanceof Error?t.message:t,"; Zone:",e.zone.name,"; Task:",e.task&&e.task.source,"; Value:",t,t instanceof Error?t.stack:void 0):console.error(e)}},n.microtaskDrainDone=function(){for(;i.length;)for(var e=function(){var e=i.shift();try{e.zone.runGuarded(function(){throw e})}catch(e){l(e)}};i.length;)e()};var u=s("unhandledPromiseRejectionHandler");function l(e){n.onUnhandledError(e);try{var t=o[u];t&&"function"==typeof t&&t.apply(this,[e])}catch(e){}}function f(e){return e&&e.then}function d(e){return e}function p(e){return Z.reject(e)}var h=s("state"),j=s("value"),m="Promise.then",v=null,_=!0,y=!1,g=0;function b(e,t){return function(r){try{x(e,t,r)}catch(t){x(e,!1,t)}}}var k=function(){var e=!1;return function(t){return function(){e||(e=!0,t.apply(null,arguments))}}},w="Promise resolved with itself",T="object",E="function",O=s("currentTask");function x(e,t,r){var s=k();if(e===r)throw new TypeError(w);if(e[h]===v){var a=null;try{typeof r!==T&&typeof r!==E||(a=r&&r.then)}catch(t){return s(function(){x(e,!1,t)})(),e}if(t!==y&&r instanceof Z&&r.hasOwnProperty(h)&&r.hasOwnProperty(j)&&r[h]!==v)P(r),x(e,r[h],r[j]);else if(t!==y&&typeof a===E)try{a.apply(r,[s(b(e,t)),s(b(e,!1))])}catch(t){s(function(){x(e,!1,t)})()}else{e[h]=t;var c=e[j];e[j]=r,t===y&&r instanceof Error&&(r[O]=o.currentTask);for(var u=0;u<c.length;)D(e,c[u++],c[u++],c[u++],c[u++]);if(0==c.length&&t==y){e[h]=g;try{throw new Error("Uncaught (in promise): "+r+(r&&r.stack?"\n"+r.stack:""))}catch(t){var l=t;l.rejection=r,l.promise=e,l.zone=o.current,l.task=o.currentTask,i.push(l),n.scheduleMicroTask()}}}}return e}var S=s("rejectionHandledHandler");function P(e){if(e[h]===g){try{var t=o[S];t&&typeof t===E&&t.apply(this,[{rejection:e[j],promise:e}])}catch(e){}e[h]=y;for(var r=0;r<i.length;r++)e===i[r].promise&&i.splice(r,1)}}function D(e,t,r,o,n){P(e);var s=e[h]?typeof o===E?o:d:typeof n===E?n:p;t.scheduleMicroTask(m,function(){try{x(r,!0,t.run(s,void 0,[e[j]]))}catch(e){x(r,!1,e)}})}var Z=function(){function e(t){if(!(this instanceof e))throw new Error("Must be an instanceof Promise.");this[h]=v,this[j]=[];try{t&&t(b(this,_),b(this,y))}catch(e){x(this,!1,e)}}return e.toString=function(){return"function ZoneAwarePromise() { [native code] }"},e.resolve=function(e){return x(new this(null),_,e)},e.reject=function(e){return x(new this(null),y,e)},e.race=function(e){var o,n,s,i,a=new this(function(e,r){var s;s=t([e,r],2),o=s[0],n=s[1]});function c(e){a&&(a=o(e))}function u(e){a&&(a=n(e))}try{for(var l=r(e),d=l.next();!d.done;d=l.next()){var p=d.value;f(p)||(p=this.resolve(p)),p.then(c,u)}}catch(e){s={error:e}}finally{try{d&&!d.done&&(i=l.return)&&i.call(l)}finally{if(s)throw s.error}}return a},e.all=function(e){var t,o,n,s,i=new this(function(e,r){t=e,o=r}),a=0,c=[];try{for(var u=r(e),l=u.next();!l.done;l=u.next()){var d=l.value;f(d)||(d=this.resolve(d)),d.then(function(e){return function(r){c[e]=r,--a||t(c)}}(a),o),a++}}catch(e){n={error:e}}finally{try{l&&!l.done&&(s=u.return)&&s.call(u)}finally{if(n)throw n.error}}return a||t(c),i},e.prototype.then=function(e,t){var r=new this.constructor(null),n=o.current;return this[h]==v?this[j].push(n,r,e,t):D(this,n,r,e,t),r},e.prototype.catch=function(e){return this.then(null,e)},e}();Z.resolve=Z.resolve,Z.reject=Z.reject,Z.race=Z.race,Z.all=Z.all;var z=e[a]=e.Promise,M=o.__symbol__("ZoneAwarePromise"),C=Object.getOwnPropertyDescriptor(e,"Promise");C&&!C.configurable||(C&&delete C.writable,C&&delete C.value,C||(C={configurable:!0,enumerable:!0}),C.get=function(){return e[M]?e[M]:e[a]},C.set=function(t){t===Z?e[M]=t:(e[a]=t,t.prototype[c]||R(t),n.setNativePromise(t))},Object.defineProperty(e,"Promise",C)),e.Promise=Z;var L,I=s("thenPatched");function R(e){var t=e.prototype,r=t.then;t[c]=r;var o=Object.getOwnPropertyDescriptor(e.prototype,"then");o&&!1===o.writable&&o.configurable&&Object.defineProperty(e.prototype,"then",{writable:!0}),e.prototype.then=function(e,t){var o=this;return new Z(function(e,t){r.call(o,e,t)}).then(e,t)},e[I]=!0}if(z){R(z);var F=e.fetch;typeof F==E&&(e.fetch=(L=F,function(){var e=L.apply(this,arguments);if(e instanceof Z)return e;var t=e.constructor;return t[I]||R(t),e}))}return Promise[o.__symbol__("uncaughtPromiseErrors")]=i,Z});var o=Zone.__symbol__,n="object"==typeof window&&window||"object"==typeof self&&self||e,s="function",i="undefined",a="removeAttribute";function c(e,t){for(var r=e.length-1;r>=0;r--)typeof e[r]===s&&(e[r]=Zone.current.wrap(e[r],t+"_"+r));return e}function u(e){return!e||!1!==e.writable&&(typeof e.get!==s||typeof e.set!==i)}var l="undefined"!=typeof WorkerGlobalScope&&self instanceof WorkerGlobalScope,f=!("nw"in n)&&void 0!==n.process&&"[object process]"==={}.toString.call(n.process),d=!f&&!l&&!("undefined"==typeof window||!window.HTMLElement),p=void 0!==n.process&&"[object process]"==={}.toString.call(n.process)&&!l&&!("undefined"==typeof window||!window.HTMLElement),h={},j=function(e){if(e=e||n.event){var t=h[e.type];t||(t=h[e.type]=o("ON_PROPERTY"+e.type));var r=(this||e.target||n)[t],s=r&&r.apply(this,arguments);return void 0==s||s||e.preventDefault(),s}};function m(e,t,r){var i=Object.getOwnPropertyDescriptor(e,t);!i&&r&&(Object.getOwnPropertyDescriptor(r,t)&&(i={enumerable:!0,configurable:!0}));if(i&&i.configurable){delete i.writable,delete i.value;var c=i.get,u=t.substr(2),l=h[u];l||(l=h[u]=o("ON_PROPERTY"+u)),i.set=function(t){var r=this;(r||e!==n||(r=n),r)&&(r[l]&&r.removeEventListener(u,j),"function"==typeof t?(r[l]=t,r.addEventListener(u,j,!1)):r[l]=null)},i.get=function(){var r=this;if(r||e!==n||(r=n),!r)return null;var o=r[l];if(o)return o;if(c){var u=c&&c.apply(this);if(u)return i.set.apply(this,[u]),typeof r[a]===s&&r.removeAttribute(t),u}return null},Object.defineProperty(e,t,i)}}function v(e,t,r){if(t)for(var o=0;o<t.length;o++)m(e,"on"+t[o],r);else{var n=[];for(var s in e)"on"==s.substr(0,2)&&n.push(s);for(var i=0;i<n.length;i++)m(e,n[i],r)}}var _=o("originalInstance");function y(e){var t=n[e];if(t){n[o(e)]=t,n[e]=function(){var r=c(arguments,e);switch(r.length){case 0:this[_]=new t;break;case 1:this[_]=new t(r[0]);break;case 2:this[_]=new t(r[0],r[1]);break;case 3:this[_]=new t(r[0],r[1],r[2]);break;case 4:this[_]=new t(r[0],r[1],r[2],r[3]);break;default:throw new Error("Arg list too long.")}},b(n[e],t);var r,s=new t(function(){});for(r in s)"XMLHttpRequest"===e&&"responseBlob"===r||function(t){"function"==typeof s[t]?n[e].prototype[t]=function(){return this[_][t].apply(this[_],arguments)}:Object.defineProperty(n[e].prototype,t,{set:function(r){"function"==typeof r?(this[_][t]=Zone.current.wrap(r,e+"."+t),b(this[_][t],r)):this[_][t]=r},get:function(){return this[_][t]}})}(r);for(r in t)"prototype"!==r&&t.hasOwnProperty(r)&&(n[e][r]=t[r])}}function g(e,t,r){for(var n=e;n&&!n.hasOwnProperty(t);)n=Object.getPrototypeOf(n);!n&&e[t]&&(n=e);var s,i=o(t);if(n&&!(s=n[i])&&(s=n[i]=n[t],u(n&&Object.getOwnPropertyDescriptor(n,t)))){var a=r(s,i,t);n[t]=function(){return a(this,arguments)},b(n[t],s)}return s}function b(e,t){e[o("OriginalDelegate")]=t}var k=!1,w=!1;Zone.__load_patch("toString",function(e,t,r){var n=t.__zone_symbol__originalToString=Function.prototype.toString,s=o("OriginalDelegate"),i=o("Promise"),a=o("Error");Function.prototype.toString=function(){if("function"==typeof this){var t=this[s];if(t)return"function"==typeof t?n.apply(this[s],arguments):Object.prototype.toString.call(t);if(this===Promise){var r=e[i];if(r)return n.apply(r,arguments)}if(this===Error){var o=e[a];if(o)return n.apply(o,arguments)}}return n.apply(this,arguments)};var c=Object.prototype.toString;Object.prototype.toString=function(){return this instanceof Promise?"[object Promise]":c.apply(this,arguments)}});var T=function(e,t){var r="function"==typeof Symbol&&e[Symbol.iterator];if(!r)return e;var o,n,s=r.call(e),i=[];try{for(;(void 0===t||t-- >0)&&!(o=s.next()).done;)i.push(o.value)}catch(e){n={error:e}}finally{try{o&&!o.done&&(r=s.return)&&r.call(s)}finally{if(n)throw n.error}}return i},E=function(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(T(arguments[t]));return e},O="true",x="false",S={isUsingGlobalCallback:!0},P={},D={},Z="name",z="function",M="object",C="__zone_symbol__",L=/^__zone_symbol__(\w+)(true|false)$/,I="__zone_symbol__propagationStopped";function R(e,t,r){var n=r&&r.addEventListenerFnName||"addEventListener",s=r&&r.removeEventListenerFnName||"removeEventListener",i=r&&r.listenersFnName||"eventListeners",a=r&&r.removeAllFnName||"removeAllListeners",c=o(n),u="."+n+":",l="prependListener",f="."+l+":",d=function(e,t,r){if(!e.isRemoved){var o=e.callback;typeof o===M&&o.handleEvent&&(e.callback=function(e){return o.handleEvent(e)},e.originalDelegate=o),e.invoke(e,t,[r]);var n=e.options;if(n&&"object"==typeof n&&n.once){var i=e.originalDelegate?e.originalDelegate:e.callback;t[s].apply(t,[r.type,i,n])}}},p=function(t){if(t=t||e.event){var r=this||t.target||e,o=r[P[t.type][x]];if(o)if(1===o.length)d(o[0],r,t);else for(var n=o.slice(),s=0;s<n.length&&(!t||!0!==t[I]);s++)d(n[s],r,t)}},h=function(t){if(t=t||e.event){var r=this||t.target||e,o=r[P[t.type][O]];if(o)if(1===o.length)d(o[0],r,t);else for(var n=o.slice(),s=0;s<n.length&&(!t||!0!==t[I]);s++)d(n[s],r,t)}};function j(t,r){if(!t)return!1;var d=!0;r&&void 0!==r.useGlobalCallback&&(d=r.useGlobalCallback);var j=r&&r.validateHandler,m=!0;r&&void 0!==r.checkDuplicate&&(m=r.checkDuplicate);var v=!1;r&&void 0!==r.returnTarget&&(v=r.returnTarget);for(var _=t;_&&!_.hasOwnProperty(n);)_=Object.getPrototypeOf(_);if(!_&&t[n]&&(_=t),!_)return!1;if(_[c])return!1;var y,g={},k=_[c]=_[n],w=_[o(s)]=_[s],T=_[o(i)]=_[i],I=_[o(a)]=_[a];r&&r.prependEventListenerFnName&&(y=_[o(r.prependEventListenerFnName)]=_[r.prependEventListenerFnName]);var R=d?function(e){if(!g.isExisting)return k.apply(g.target,[g.eventName,g.capture?h:p,g.options])}:function(e){return k.apply(g.target,[g.eventName,e.invoke,g.options])},H=d?function(e){if(!e.isRemoved){var t=P[e.eventName],r=void 0;t&&(r=t[e.capture?O:x]);var o=r&&e.target[r];if(o)for(var n=0;n<o.length;n++)if(o[n]===e){o.splice(n,1),e.isRemoved=!0,0===o.length&&(e.allRemoved=!0,e.target[r]=null);break}}if(e.allRemoved)return w.apply(e.target,[e.eventName,e.capture?h:p,e.options])}:function(e){return w.apply(e.target,[e.eventName,e.invoke,e.options])},q=r&&r.compareTaskCallbackVsDelegate?r.compareTaskCallbackVsDelegate:function(e,t){var r=typeof t;return r===z&&e.callback===t||r===M&&e.originalDelegate===t},A=function(t,r,o,n,s,i){return void 0===s&&(s=!1),void 0===i&&(i=!1),function(){var a=this||e,c=(Zone.current,arguments[1]);if(!c)return t.apply(this,arguments);var u=!1;if(typeof c!==z){if(!c.handleEvent)return t.apply(this,arguments);u=!0}if(!j||j(t,c,a,arguments)){var l,f=arguments[0],p=arguments[2],h=!1;void 0===p?l=!1:!0===p?l=!0:!1===p?l=!1:(l=!!p&&!!p.capture,h=!!p&&!!p.once);var v,_=Zone.current,y=P[f];if(y)v=y[l?O:x];else{var b=C+(f+x),k=C+(f+O);P[f]={},P[f][x]=b,P[f][O]=k,v=l?k:b}var w,T=a[v],E=!1;if(T){if(E=!0,m)for(var M=0;M<T.length;M++)if(q(T[M],c))return}else T=a[v]=[];var L=a.constructor[Z],I=D[L];I&&(w=I[f]),w||(w=L+r+f),g.options=p,h&&(g.options.once=!1),g.target=a,g.capture=l,g.eventName=f,g.isExisting=E;var R=d?S:null,F=_.scheduleEventTask(w,c,R,o,n);return h&&(p.once=!0),F.options=p,F.target=a,F.capture=l,F.eventName=f,u&&(F.originalDelegate=c),i?T.unshift(F):T.push(F),s?a:void 0}}};return _[n]=A(k,u,R,H,v),y&&(_[l]=A(y,f,function(e){return y.apply(g.target,[g.eventName,e.invoke,g.options])},H,v,!0)),_[s]=function(){var t,r=this||e,o=arguments[0],n=arguments[2];t=void 0!==n&&(!0===n||!1!==n&&(!!n&&!!n.capture));var s=arguments[1];if(!s)return w.apply(this,arguments);if(!j||j(w,s,r,arguments)){var i,a=P[o];a&&(i=a[t?O:x]);var c=i&&r[i];if(c)for(var u=0;u<c.length;u++){var l=c[u];if(q(l,s))return c.splice(u,1),l.isRemoved=!0,0===c.length&&(l.allRemoved=!0,r[i]=null),void l.zone.cancelTask(l)}}},_[i]=function(){for(var t=[],r=F(this||e,arguments[0]),o=0;o<r.length;o++){var n=r[o],s=n.originalDelegate?n.originalDelegate:n.callback;t.push(s)}return t},_[a]=function(){var t=this||e,r=arguments[0];if(r){var o=P[r];if(o){var n=o[x],i=o[O],c=t[n],u=t[i];if(c){var l=E(c);for(h=0;h<l.length;h++){var f=(d=l[h]).originalDelegate?d.originalDelegate:d.callback;this[s].apply(this,[r,f,d.options])}}if(u)for(l=E(u),h=0;h<l.length;h++){var d;f=(d=l[h]).originalDelegate?d.originalDelegate:d.callback;this[s].apply(this,[r,f,d.options])}}}else{for(var p=Object.keys(t),h=0;h<p.length;h++){var j=p[h],m=L.exec(j),v=m&&m[1];v&&"removeListener"!==v&&this[a].apply(this,[v])}this[a].apply(this,["removeListener"])}},b(_[n],k),b(_[s],w),I&&b(_[a],I),T&&b(_[i],T),!0}for(var m=[],v=0;v<t.length;v++)m[v]=j(t[v],r);return m}function F(e,t){var r=[];for(var o in e){var n=L.exec(o),s=n&&n[1];if(s&&(!t||s===t)){var i=e[o];if(i)for(var a=0;a<i.length;a++)r.push(i[a])}}return r}var H=o("zoneTask");function q(e,t,r,o){var n=null,s=null;r+=o;var i={},a="number";function c(t){var r=t.data;return r.args[0]=function(){try{t.invoke.apply(this,arguments)}finally{typeof r.handleId===a?delete i[r.handleId]:r.handleId&&(r.handleId[H]=null)}},r.handleId=n.apply(e,r.args),t}function u(e){return s(e.data.handleId)}n=g(e,t+=o,function(r){return function(n,s){if("function"==typeof s[0]){var l=Zone.current,f={handleId:null,isPeriodic:"Interval"===o,delay:"Timeout"===o||"Interval"===o?s[1]||0:null,args:s},d=l.scheduleMacroTask(t,s[0],f,c,u);if(!d)return d;var p=d.data.handleId;return typeof p===a?i[p]=d:p&&(p[H]=d),p&&p.ref&&p.unref&&"function"==typeof p.ref&&"function"==typeof p.unref&&(d.ref=p.ref.bind(p),d.unref=p.unref.bind(p)),typeof p===a||p?p:d}return r.apply(e,s)}}),s=g(e,r,function(t){return function(r,o){var n,s=o[0];typeof s===a?n=i[s]:(n=s&&s[H])||(n=s),n&&"string"==typeof n.type?"notScheduled"!==n.state&&(n.cancelFn&&n.data.isPeriodic||0===n.runCount)&&(typeof s===a?delete i[s]:s&&(s[H]=null),n.zone.cancelTask(n)):t.apply(e,o)}})}var A=Object[o("defineProperty")]=Object.defineProperty,N=Object[o("getOwnPropertyDescriptor")]=Object.getOwnPropertyDescriptor,B=Object.create,W=o("unconfigurables"),X="prototype",U="object",G="undefined";function K(e,t){return e&&e[W]&&e[W][t]}function V(e,t,r){return r.configurable=!0,r.configurable||(e[W]||A(e,W,{writable:!0,value:{}}),e[W][t]=!0),r}function Y(e,t,r,o){try{return A(e,t,r)}catch(s){if(!r.configurable)throw s;typeof o==G?delete r.configurable:r.configurable=o;try{return A(e,t,r)}catch(o){var n=null;try{n=JSON.stringify(r)}catch(e){n=n.toString()}console.log("Attempting to configure '"+t+"' with descriptor '"+n+"' on object '"+e+"' and got error, giving up: "+o)}}}var J=["absolutedeviceorientation","afterinput","afterprint","appinstalled","beforeinstallprompt","beforeprint","beforeunload","devicelight","devicemotion","deviceorientation","deviceorientationabsolute","deviceproximity","hashchange","languagechange","message","mozbeforepaint","offline","online","paint","pageshow","pagehide","popstate","rejectionhandled","storage","unhandledrejection","unload","userproximity","vrdisplyconnected","vrdisplaydisconnected","vrdisplaypresentchange"],Q=["encrypted","waitingforkey","msneedkey","mozinterruptbegin","mozinterruptend"],$=["load"],ee=["blur","error","focus","load","resize","scroll","messageerror"],te=["bounce","finish","start"],re=["loadstart","progress","abort","error","load","progress","timeout","loadend","readystatechange"],oe=["upgradeneeded","complete","abort","success","error","blocked","versionchange","close"],ne=["close","error","open","message"],se=["error","message"],ie=["abort","animationcancel","animationend","animationiteration","auxclick","beforeinput","blur","cancel","canplay","canplaythrough","change","compositionstart","compositionupdate","compositionend","cuechange","click","close","contextmenu","curechange","dblclick","drag","dragend","dragenter","dragexit","dragleave","dragover","drop","durationchange","emptied","ended","error","focus","focusin","focusout","gotpointercapture","input","invalid","keydown","keypress","keyup","load","loadstart","loadeddata","loadedmetadata","lostpointercapture","mousedown","mouseenter","mouseleave","mousemove","mouseout","mouseover","mouseup","mousewheel","orientationchange","pause","play","playing","pointercancel","pointerdown","pointerenter","pointerleave","pointerlockchange","mozpointerlockchange","webkitpointerlockerchange","pointerlockerror","mozpointerlockerror","webkitpointerlockerror","pointermove","pointout","pointerover","pointerup","progress","ratechange","reset","resize","scroll","seeked","seeking","select","selectionchange","selectstart","show","sort","stalled","submit","suspend","timeupdate","volumechange","touchcancel","touchmove","touchstart","touchend","transitioncancel","transitionend","waiting","wheel"].concat(["webglcontextrestored","webglcontextlost","webglcontextcreationerror"],["autocomplete","autocompleteerror"],["toggle"],["afterscriptexecute","beforescriptexecute","DOMContentLoaded","fullscreenchange","mozfullscreenchange","webkitfullscreenchange","msfullscreenchange","fullscreenerror","mozfullscreenerror","webkitfullscreenerror","msfullscreenerror","readystatechange","visibilitychange"],J,["beforecopy","beforecut","beforepaste","copy","cut","paste","dragstart","loadend","animationstart","search","transitionrun","transitionstart","webkitanimationend","webkitanimationiteration","webkitanimationstart","webkittransitionend"],["activate","afterupdate","ariarequest","beforeactivate","beforedeactivate","beforeeditfocus","beforeupdate","cellchange","controlselect","dataavailable","datasetchanged","datasetcomplete","errorupdate","filterchange","layoutcomplete","losecapture","move","moveend","movestart","propertychange","resizeend","resizestart","rowenter","rowexit","rowsdelete","rowsinserted","command","compassneedscalibration","deactivate","help","mscontentzoom","msmanipulationstatechanged","msgesturechange","msgesturedoubletap","msgestureend","msgesturehold","msgesturestart","msgesturetap","msgotpointercapture","msinertiastart","mslostpointercapture","mspointercancel","mspointerdown","mspointerenter","mspointerhover","mspointerleave","mspointermove","mspointerout","mspointerover","mspointerup","pointerout","mssitemodejumplistitemremoved","msthumbnailclick","stop","storagecommit"]);function ae(e,t,r,o){v(e,function(e,t,r){if(!r)return t;var o=r.filter(function(t){return t.target===e});if(!o||0===o.length)return t;var n=o[0].ignoreProperties;return t.filter(function(e){return-1===n.indexOf(e)})}(e,t,r),o)}function ce(e,t){if(!f||p){var r="undefined"!=typeof WebSocket;if(function(){if((d||p)&&!Object.getOwnPropertyDescriptor(HTMLElement.prototype,"onclick")&&"undefined"!=typeof Element){var e=Object.getOwnPropertyDescriptor(Element.prototype,"onclick");if(e&&!e.configurable)return!1}var t=Object.getOwnPropertyDescriptor(XMLHttpRequest.prototype,"onreadystatechange");if(t){Object.defineProperty(XMLHttpRequest.prototype,"onreadystatechange",{enumerable:!0,configurable:!0,get:function(){return!0}});var r=new XMLHttpRequest,n=!!r.onreadystatechange;return Object.defineProperty(XMLHttpRequest.prototype,"onreadystatechange",t||{}),n}var s=o("fakeonreadystatechange");Object.defineProperty(XMLHttpRequest.prototype,"onreadystatechange",{enumerable:!0,configurable:!0,get:function(){return this[s]},set:function(e){this[s]=e}});var r=new XMLHttpRequest,i=function(){};r.onreadystatechange=i;var n=r[s]===i;return r.onreadystatechange=null,n}()){var n=t.__Zone_ignore_on_properties;if(d){ae(window,ie.concat(["messageerror"]),n,Object.getPrototypeOf(window)),ae(Document.prototype,ie,n),void 0!==window.SVGElement&&ae(window.SVGElement.prototype,ie,n),ae(Element.prototype,ie,n),ae(HTMLElement.prototype,ie,n),ae(HTMLMediaElement.prototype,Q,n),ae(HTMLFrameSetElement.prototype,J.concat(ee),n),ae(HTMLBodyElement.prototype,J.concat(ee),n),ae(HTMLFrameElement.prototype,$,n),ae(HTMLIFrameElement.prototype,$,n);var s=window.HTMLMarqueeElement;s&&ae(s.prototype,te,n);var i=window.Worker;i&&ae(i.prototype,se,n)}ae(XMLHttpRequest.prototype,re,n);var a=t.XMLHttpRequestEventTarget;a&&ae(a&&a.prototype,re,n),"undefined"!=typeof IDBIndex&&(ae(IDBIndex.prototype,oe,n),ae(IDBRequest.prototype,oe,n),ae(IDBOpenDBRequest.prototype,oe,n),ae(IDBDatabase.prototype,oe,n),ae(IDBTransaction.prototype,oe,n),ae(IDBCursor.prototype,oe,n)),r&&ae(WebSocket.prototype,ne,n)}else!function(){for(var e=function(e){var t=ie[e],r="on"+t;self.addEventListener(t,function(e){var t,o,n=e.target;for(o=n?n.constructor.name+"."+r:"unknown."+r;n;)n[r]&&!n[r][ue]&&((t=Zone.current.wrap(n[r],o))[ue]=n[r],n[r]=t),n=n.parentElement},!0)},t=0;t<ie.length;t++)e(t)}(),y("XMLHttpRequest"),r&&function(e,t){var r=t.WebSocket;t.EventTarget||R(t,[r.prototype]),t.WebSocket=function(e,t){var o,n,s=arguments.length>1?new r(e,t):new r(e),i=Object.getOwnPropertyDescriptor(s,"onmessage");return i&&!1===i.configurable?(o=Object.create(s),n=s,["addEventListener","removeEventListener","send","close"].forEach(function(e){o[e]=function(){var t=Array.prototype.slice.call(arguments);if("addEventListener"===e||"removeEventListener"===e){var r=t.length>0?t[0]:void 0;if(r){var n=Zone.__symbol__("ON_PROPERTY"+r);s[n]=o[n]}}return s[e].apply(s,t)}})):o=s,v(o,["close","error","message","open"],n),o};var o=t.WebSocket;for(var n in r)o[n]=r[n]}(0,t)}}var ue=o("unbound");function le(e,t){var r="Anchor,Area,Audio,BR,Base,BaseFont,Body,Button,Canvas,Content,DList,Directory,Div,Embed,FieldSet,Font,Form,Frame,FrameSet,HR,Head,Heading,Html,IFrame,Image,Input,Keygen,LI,Label,Legend,Link,Map,Marquee,Media,Menu,Meta,Meter,Mod,OList,Object,OptGroup,Option,Output,Paragraph,Pre,Progress,Quote,Script,Select,Source,Span,Style,TableCaption,TableCell,TableCol,Table,TableRow,TableSection,TextArea,Title,Track,UList,Unknown,Video",o="ApplicationCache,EventSource,FileReader,InputMethodContext,MediaController,MessagePort,Node,Performance,SVGElementInstance,SharedWorker,TextTrack,TextTrackCue,TextTrackList,WebKitNamedFlow,Window,Worker,WorkerGlobalScope,XMLHttpRequest,XMLHttpRequestEventTarget,XMLHttpRequestUpload,IDBRequest,IDBOpenDBRequest,IDBDatabase,IDBTransaction,IDBCursor,DBIndex,WebSocket".split(","),n=[],s=e.wtf,i=r.split(",");s?n=i.map(function(e){return"HTML"+e+"Element"}).concat(o):e.EventTarget?n.push("EventTarget"):n=o;for(var a=e.__Zone_disable_IE_check||!1,c=e.__Zone_enable_cross_context_check||!1,u=function(){if(k)return w;k=!0;try{var e=window.navigator.userAgent;return e.indexOf("MSIE "),-1===e.indexOf("MSIE ")&&-1===e.indexOf("Trident/")&&-1===e.indexOf("Edge/")||(w=!0),w}catch(e){}}(),l="[object FunctionWrapper]",f="function __BROWSERTOOLS_CONSOLE_SAFEFUNC() { [native code] }",d=0;d<ie.length;d++){var p=ie[d],h=C+(p+x),j=C+(p+O);P[p]={},P[p][x]=h,P[p][O]=j}for(d=0;d<r.length;d++)for(var m=i[d],v=D[m]={},_=0;_<ie.length;_++){v[p=ie[_]]=m+".addEventListener:"+p}var y=[];for(d=0;d<n.length;d++){var g=e[n[d]];y.push(g&&g.prototype)}return R(e,y,{validateHandler:function(e,t,r,o){if(!a&&u){if(c)try{var n;if((n=t.toString())===l||n==f)return e.apply(r,o),!1}catch(t){return e.apply(r,o),!1}else if((n=t.toString())===l||n==f)return e.apply(r,o),!1}else if(c)try{t.toString()}catch(t){return e.apply(r,o),!1}return!0}}),t.patchEventTarget=R,!0}function fe(e,t){var r,o;r=t,(o=e.Event)&&o.prototype&&r.patchMethod(o.prototype,"stopImmediatePropagation",function(e){return function(e,t){e[I]=!0}})}function de(e){if((d||p)&&"registerElement"in e.document){var t=document.registerElement,r=["createdCallback","attachedCallback","detachedCallback","attributeChangedCallback"];document.registerElement=function(e,o){return o&&o.prototype&&r.forEach(function(e){var t,r,n,s,i="Document.registerElement::"+e;if(o.prototype.hasOwnProperty(e)){var a=Object.getOwnPropertyDescriptor(o.prototype,e);a&&a.value?(a.value=Zone.current.wrap(a.value,i),t=o.prototype,r=e,s=(n=a).configurable,Y(t,r,n=V(t,r,n),s)):o.prototype[e]=Zone.current.wrap(o.prototype[e],i)}else o.prototype[e]&&(o.prototype[e]=Zone.current.wrap(o.prototype[e],i))}),t.apply(document,[e,o])},b(document.registerElement,t)}}Zone.__load_patch("util",function(e,t,r){r.patchOnProperties=v,r.patchMethod=g}),Zone.__load_patch("timers",function(e,t,r){q(e,"set","clear","Timeout"),q(e,"set","clear","Interval"),q(e,"set","clear","Immediate")}),Zone.__load_patch("requestAnimationFrame",function(e,t,r){q(e,"request","cancel","AnimationFrame"),q(e,"mozRequest","mozCancel","AnimationFrame"),q(e,"webkitRequest","webkitCancel","AnimationFrame")}),Zone.__load_patch("blocking",function(e,t,r){for(var o=["alert","prompt","confirm"],n=0;n<o.length;n++){g(e,o[n],function(r,o,n){return function(o,s){return t.current.run(r,e,s,n)}})}}),Zone.__load_patch("EventTarget",function(e,t,r){fe(e,r),le(e,r);var o=e.XMLHttpRequestEventTarget;o&&o.prototype&&r.patchEventTarget(e,[o.prototype]),y("MutationObserver"),y("WebKitMutationObserver"),y("IntersectionObserver"),y("FileReader")}),Zone.__load_patch("on_property",function(e,t,r){ce(0,e),Object.defineProperty=function(e,t,r){if(K(e,t))throw new TypeError("Cannot assign to read only property '"+t+"' of "+e);var o=r.configurable;return t!==X&&(r=V(e,t,r)),Y(e,t,r,o)},Object.defineProperties=function(e,t){return Object.keys(t).forEach(function(r){Object.defineProperty(e,r,t[r])}),e},Object.create=function(e,t){return typeof t!==U||Object.isFrozen(t)||Object.keys(t).forEach(function(r){t[r]=V(e,r,t[r])}),B(e,t)},Object.getOwnPropertyDescriptor=function(e,t){var r=N(e,t);return K(e,t)&&(r.configurable=!1),r},de(e)}),Zone.__load_patch("canvas",function(e,t,r){var o=e.HTMLCanvasElement;void 0!==o&&o.prototype&&o.prototype.toBlob&&function(e,t,r){var o=null;function n(e){var t=e.data;return t.args[t.callbackIndex]=function(){e.invoke.apply(this,arguments)},o.apply(t.target,t.args),e}o=g(e,t,function(e){return function(t,o){var s=r(t,o);return s.callbackIndex>=0&&"function"==typeof o[s.callbackIndex]?Zone.current.scheduleMacroTask(s.name,o[s.callbackIndex],s,n,null):e.apply(t,o)}})}(o.prototype,"toBlob",function(e,t){return{name:"HTMLCanvasElement.toBlob",target:e,callbackIndex:0,args:t}})}),Zone.__load_patch("XHR",function(e,t,r){!function(e){var r=o("addEventListener"),u=o("removeEventListener"),l=XMLHttpRequest.prototype[r],f=XMLHttpRequest.prototype[u];if(!l){var d=e.XMLHttpRequestEventTarget;d&&(l=d.prototype[r],f=d.prototype[u])}var p="readystatechange",h="scheduled";function j(e){XMLHttpRequest[a]=!1;var t=e.data,o=t.target,s=o[i];l||(l=o[r],f=o[u]),s&&f.apply(o,[p,s]);var c=o[i]=function(){o.readyState===o.DONE&&!t.aborted&&XMLHttpRequest[a]&&e.state===h&&e.invoke()};l.apply(o,[p,c]);var d=o[n];return d||(o[n]=e),y.apply(o,t.args),XMLHttpRequest[a]=!0,e}function m(){}function v(e){var t=e.data;return t.aborted=!0,b.apply(t.target,t.args)}var _=g(e.XMLHttpRequest.prototype,"open",function(){return function(e,t){return e[s]=0==t[2],e[c]=t[1],_.apply(e,t)}}),y=g(e.XMLHttpRequest.prototype,"send",function(){return function(e,r){var o=t.current;if(e[s])return y.apply(e,r);var n={target:e,url:e[c],isPeriodic:!1,delay:null,args:r,aborted:!1};return o.scheduleMacroTask("XMLHttpRequest.send",m,n,j,v)}}),b=g(e.XMLHttpRequest.prototype,"abort",function(e){return function(e,t){var r=e[n];if(r&&"string"==typeof r.type){if(null==r.cancelFn||r.data&&r.data.aborted)return;r.zone.cancelTask(r)}}})}(e);var n=o("xhrTask"),s=o("xhrSync"),i=o("xhrListener"),a=o("xhrScheduled"),c=o("xhrURL")}),Zone.__load_patch("geolocation",function(e,t,r){e.navigator&&e.navigator.geolocation&&function(e,t){for(var r=e.constructor.name,o=function(o){var n,s,i=t[o],a=e[i];if(a){if(!u(Object.getOwnPropertyDescriptor(e,i)))return"continue";e[i]=(b(s=function(){return n.apply(this,c(arguments,r+"."+i))},n=a),s)}},n=0;n<t.length;n++)o(n)}(e.navigator.geolocation,["getCurrentPosition","watchPosition"])}),Zone.__load_patch("PromiseRejectionEvent",function(e,t,r){function n(t){return function(r){F(e,t).forEach(function(o){var n=e.PromiseRejectionEvent;if(n){var s=new n(t,{promise:r.promise,reason:r.rejection});o.invoke(s)}})}}e.PromiseRejectionEvent&&(t[o("unhandledPromiseRejectionHandler")]=n("unhandledrejection"),t[o("rejectionHandledHandler")]=n("rejectionhandled"))})})()}).call(t,r("../../../../webpack/buildin/global.js"))},4:function(e,t,r){e.exports=r("../../../../../src/polyfills.ts")}},[4]);
</script><script type="text/javascript">
webpackJsonp(["styles"],{"../../../../../src/styles.scss":function(a,t,n){var e=n('../../../../css-loader/index.js?{"sourceMap":false,"importLoaders":1}!../../../../postcss-loader/index.js?{"ident":"postcss"}!../../../../sass-loader/lib/loader.js?{"sourceMap":false,"precision":8,"includePaths":[]}!../../../../../src/styles.scss');"string"==typeof e&&(e=[[a.i,e,""]]);n("../../../../style-loader/addStyles.js")(e,{});e.locals&&(a.exports=e.locals)},'../../../../css-loader/index.js?{"sourceMap":false,"importLoaders":1}!../../../../postcss-loader/index.js?{"ident":"postcss"}!../../../../sass-loader/lib/loader.js?{"sourceMap":false,"precision":8,"includePaths":[]}!../../../../../src/styles.scss':function(a,t,n){(a.exports=n("../../../../css-loader/lib/css-base.js")(!1)).push([a.i,'/**\n * Applies styles for users in high contrast mode. Note that this only applies\n * to Microsoft browsers. Chrome can be included by checking for the `html[hc]`\n * attribute, however Chrome handles high contrast differently.\n */\n/* Theme for the ripple elements.*/\n/** The mixins below are shared between mat-menu and mat-select */\n/**\n * This mixin adds the correct panel transform styles based\n * on the direction that the menu panel opens.\n */\n/* stylelint-disable material/no-prefixes */\n/* stylelint-enable */\n/**\n * This mixin contains shared option styles between the select and\n * autocomplete components.\n */\n.mat-elevation-z0 {\n  box-shadow: 0px 0px 0px 0px rgba(0, 0, 0, 0.2), 0px 0px 0px 0px rgba(0, 0, 0, 0.14), 0px 0px 0px 0px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z1 {\n  box-shadow: 0px 2px 1px -1px rgba(0, 0, 0, 0.2), 0px 1px 1px 0px rgba(0, 0, 0, 0.14), 0px 1px 3px 0px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z2 {\n  box-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z3 {\n  box-shadow: 0px 3px 3px -2px rgba(0, 0, 0, 0.2), 0px 3px 4px 0px rgba(0, 0, 0, 0.14), 0px 1px 8px 0px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z4 {\n  box-shadow: 0px 2px 4px -1px rgba(0, 0, 0, 0.2), 0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z5 {\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 5px 8px 0px rgba(0, 0, 0, 0.14), 0px 1px 14px 0px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z6 {\n  box-shadow: 0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 6px 10px 0px rgba(0, 0, 0, 0.14), 0px 1px 18px 0px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z7 {\n  box-shadow: 0px 4px 5px -2px rgba(0, 0, 0, 0.2), 0px 7px 10px 1px rgba(0, 0, 0, 0.14), 0px 2px 16px 1px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z8 {\n  box-shadow: 0px 5px 5px -3px rgba(0, 0, 0, 0.2), 0px 8px 10px 1px rgba(0, 0, 0, 0.14), 0px 3px 14px 2px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z9 {\n  box-shadow: 0px 5px 6px -3px rgba(0, 0, 0, 0.2), 0px 9px 12px 1px rgba(0, 0, 0, 0.14), 0px 3px 16px 2px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z10 {\n  box-shadow: 0px 6px 6px -3px rgba(0, 0, 0, 0.2), 0px 10px 14px 1px rgba(0, 0, 0, 0.14), 0px 4px 18px 3px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z11 {\n  box-shadow: 0px 6px 7px -4px rgba(0, 0, 0, 0.2), 0px 11px 15px 1px rgba(0, 0, 0, 0.14), 0px 4px 20px 3px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z12 {\n  box-shadow: 0px 7px 8px -4px rgba(0, 0, 0, 0.2), 0px 12px 17px 2px rgba(0, 0, 0, 0.14), 0px 5px 22px 4px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z13 {\n  box-shadow: 0px 7px 8px -4px rgba(0, 0, 0, 0.2), 0px 13px 19px 2px rgba(0, 0, 0, 0.14), 0px 5px 24px 4px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z14 {\n  box-shadow: 0px 7px 9px -4px rgba(0, 0, 0, 0.2), 0px 14px 21px 2px rgba(0, 0, 0, 0.14), 0px 5px 26px 4px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z15 {\n  box-shadow: 0px 8px 9px -5px rgba(0, 0, 0, 0.2), 0px 15px 22px 2px rgba(0, 0, 0, 0.14), 0px 6px 28px 5px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z16 {\n  box-shadow: 0px 8px 10px -5px rgba(0, 0, 0, 0.2), 0px 16px 24px 2px rgba(0, 0, 0, 0.14), 0px 6px 30px 5px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z17 {\n  box-shadow: 0px 8px 11px -5px rgba(0, 0, 0, 0.2), 0px 17px 26px 2px rgba(0, 0, 0, 0.14), 0px 6px 32px 5px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z18 {\n  box-shadow: 0px 9px 11px -5px rgba(0, 0, 0, 0.2), 0px 18px 28px 2px rgba(0, 0, 0, 0.14), 0px 7px 34px 6px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z19 {\n  box-shadow: 0px 9px 12px -6px rgba(0, 0, 0, 0.2), 0px 19px 29px 2px rgba(0, 0, 0, 0.14), 0px 7px 36px 6px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z20 {\n  box-shadow: 0px 10px 13px -6px rgba(0, 0, 0, 0.2), 0px 20px 31px 3px rgba(0, 0, 0, 0.14), 0px 8px 38px 7px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z21 {\n  box-shadow: 0px 10px 13px -6px rgba(0, 0, 0, 0.2), 0px 21px 33px 3px rgba(0, 0, 0, 0.14), 0px 8px 40px 7px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z22 {\n  box-shadow: 0px 10px 14px -6px rgba(0, 0, 0, 0.2), 0px 22px 35px 3px rgba(0, 0, 0, 0.14), 0px 8px 42px 7px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z23 {\n  box-shadow: 0px 11px 14px -7px rgba(0, 0, 0, 0.2), 0px 23px 36px 3px rgba(0, 0, 0, 0.14), 0px 9px 44px 8px rgba(0, 0, 0, 0.12); }\n\n.mat-elevation-z24 {\n  box-shadow: 0px 11px 15px -7px rgba(0, 0, 0, 0.2), 0px 24px 38px 3px rgba(0, 0, 0, 0.14), 0px 9px 46px 8px rgba(0, 0, 0, 0.12); }\n\n.mat-h1, .mat-headline, .mat-typography h1 {\n  font: 400 24px/32px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n\n.mat-h2, .mat-title, .mat-typography h2 {\n  font: 500 20px/32px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n\n.mat-h3, .mat-subheading-2, .mat-typography h3 {\n  font: 400 16px/28px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n\n.mat-h4, .mat-subheading-1, .mat-typography h4 {\n  font: 400 15px/24px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 16px; }\n\n.mat-h5, .mat-typography h5 {\n  font-size: 11.62px;\n  font-weight: 400;\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  line-height: 20px;\n  margin: 0 0 12px; }\n\n.mat-h6, .mat-typography h6 {\n  font-size: 9.38px;\n  font-weight: 400;\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  line-height: 20px;\n  margin: 0 0 12px; }\n\n.mat-body-strong, .mat-body-2 {\n  font: 500 14px/24px Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-body, .mat-body-1, .mat-typography {\n  font: 400 14px/20px Roboto, "Helvetica Neue", sans-serif; }\n  .mat-body p, .mat-body-1 p, .mat-typography p {\n    margin: 0 0 12px; }\n\n.mat-small, .mat-caption {\n  font: 400 12px/20px Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-display-4, .mat-typography .mat-display-4 {\n  font: 300 112px/112px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 56px;\n  letter-spacing: -0.05em; }\n\n.mat-display-3, .mat-typography .mat-display-3 {\n  font: 400 56px/56px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 64px;\n  letter-spacing: -0.02em; }\n\n.mat-display-2, .mat-typography .mat-display-2 {\n  font: 400 45px/48px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 64px;\n  letter-spacing: -0.005em; }\n\n.mat-display-1, .mat-typography .mat-display-1 {\n  font: 400 34px/40px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0 0 64px; }\n\n.mat-button, .mat-raised-button, .mat-icon-button, .mat-fab, .mat-mini-fab {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 500; }\n\n.mat-button-toggle {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-card {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-card-title {\n  font-size: 24px;\n  font-weight: 400; }\n\n.mat-card-subtitle,\n.mat-card-content,\n.mat-card-header .mat-card-title {\n  font-size: 14px; }\n\n.mat-checkbox {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-checkbox-layout .mat-checkbox-label {\n  line-height: 24px; }\n\n.mat-chip {\n  font-size: 13px;\n  line-height: 18px; }\n  .mat-chip .mat-chip-remove.mat-icon {\n    font-size: 18px; }\n\n.mat-table {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-header-cell {\n  font-size: 12px;\n  font-weight: 500; }\n\n.mat-cell {\n  font-size: 14px; }\n\n.mat-calendar {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-calendar-body {\n  font-size: 13px; }\n\n.mat-calendar-body-label,\n.mat-calendar-period-button {\n  font-size: 14px;\n  font-weight: 500; }\n\n.mat-calendar-table-header th {\n  font-size: 11px;\n  font-weight: 400; }\n\n.mat-dialog-title {\n  font: 500 20px/32px Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-expansion-panel-header {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 15px;\n  font-weight: 400; }\n\n.mat-expansion-panel-content {\n  font: 400 14px/20px Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-form-field {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: inherit;\n  font-weight: 400;\n  line-height: 1.125; }\n\n.mat-form-field-wrapper {\n  padding-bottom: 1.25em; }\n\n.mat-form-field-prefix .mat-icon,\n.mat-form-field-suffix .mat-icon {\n  font-size: 150%;\n  line-height: 1.125; }\n\n.mat-form-field-prefix .mat-icon-button,\n.mat-form-field-suffix .mat-icon-button {\n  height: 1.5em;\n  width: 1.5em; }\n  .mat-form-field-prefix .mat-icon-button .mat-icon,\n  .mat-form-field-suffix .mat-icon-button .mat-icon {\n    height: 1.125em;\n    line-height: 1.125; }\n\n.mat-form-field-infix {\n  padding: 0.4375em 0;\n  border-top: 0.84375em solid transparent; }\n\n.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,\n.mat-form-field-can-float .mat-input-server:focus + .mat-form-field-label-wrapper .mat-form-field-label {\n  -webkit-transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.001px);\n          transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.001px);\n  -ms-transform: translateY(-1.28125em) scale(0.75);\n  width: 133.33333333%; }\n\n.mat-form-field-can-float .mat-form-field-autofill-control:-webkit-autofill + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00101px);\n          transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00101px);\n  -ms-transform: translateY(-1.28124em) scale(0.75);\n  width: 133.33334333%; }\n\n.mat-form-field-can-float .mat-input-server[label]:not(:label-shown) + .mat-form-field-label-wrapper\n.mat-form-field-label {\n  -webkit-transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00102px);\n          transform: translateY(-1.28125em) scale(0.75) perspective(100px) translateZ(0.00102px);\n  -ms-transform: translateY(-1.28123em) scale(0.75);\n  width: 133.33335333%; }\n\n.mat-form-field-label-wrapper {\n  top: -0.84375em;\n  padding-top: 0.84375em; }\n\n.mat-form-field-label {\n  top: 1.28125em; }\n\n.mat-form-field-underline {\n  bottom: 1.25em; }\n\n.mat-form-field-subscript-wrapper {\n  font-size: 75%;\n  margin-top: 0.54166667em;\n  top: calc(100% - 1.66666667em); }\n\n.mat-grid-tile-header,\n.mat-grid-tile-footer {\n  font-size: 14px; }\n  .mat-grid-tile-header .mat-line,\n  .mat-grid-tile-footer .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n    .mat-grid-tile-header .mat-line:nth-child(n+2),\n    .mat-grid-tile-footer .mat-line:nth-child(n+2) {\n      font-size: 12px; }\n\ninput.mat-input-element {\n  margin-top: -0.0625em; }\n\n.mat-menu-item {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 16px;\n  font-weight: 400; }\n\n.mat-paginator,\n.mat-paginator-page-size .mat-select-trigger {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 12px; }\n\n.mat-radio-button {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-select {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-select-trigger {\n  height: 1.125em; }\n\n.mat-slide-toggle-content {\n  font: 400 14px/20px Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-slider-thumb-label-text {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 12px;\n  font-weight: 500; }\n\n.mat-stepper-vertical, .mat-stepper-horizontal {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-step-label {\n  font-size: 14px;\n  font-weight: 400; }\n\n.mat-step-label-selected {\n  font-size: 14px;\n  font-weight: 500; }\n\n.mat-tab-group {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-tab-label, .mat-tab-link {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 500; }\n\n.mat-toolbar,\n.mat-toolbar h1,\n.mat-toolbar h2,\n.mat-toolbar h3,\n.mat-toolbar h4,\n.mat-toolbar h5,\n.mat-toolbar h6 {\n  font: 500 20px/32px Roboto, "Helvetica Neue", sans-serif;\n  margin: 0; }\n\n.mat-tooltip {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 10px;\n  padding-top: 6px;\n  padding-bottom: 6px; }\n\n.mat-list-item {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-list-option {\n  font-family: Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-list .mat-list-item, .mat-nav-list .mat-list-item, .mat-selection-list .mat-list-item {\n  font-size: 16px; }\n  .mat-list .mat-list-item .mat-line, .mat-nav-list .mat-list-item .mat-line, .mat-selection-list .mat-list-item .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n    .mat-list .mat-list-item .mat-line:nth-child(n+2), .mat-nav-list .mat-list-item .mat-line:nth-child(n+2), .mat-selection-list .mat-list-item .mat-line:nth-child(n+2) {\n      font-size: 14px; }\n\n.mat-list .mat-list-option, .mat-nav-list .mat-list-option, .mat-selection-list .mat-list-option {\n  font-size: 16px; }\n  .mat-list .mat-list-option .mat-line, .mat-nav-list .mat-list-option .mat-line, .mat-selection-list .mat-list-option .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n    .mat-list .mat-list-option .mat-line:nth-child(n+2), .mat-nav-list .mat-list-option .mat-line:nth-child(n+2), .mat-selection-list .mat-list-option .mat-line:nth-child(n+2) {\n      font-size: 14px; }\n\n.mat-list .mat-subheader, .mat-nav-list .mat-subheader, .mat-selection-list .mat-subheader {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px;\n  font-weight: 500; }\n\n.mat-list[dense] .mat-list-item, .mat-nav-list[dense] .mat-list-item, .mat-selection-list[dense] .mat-list-item {\n  font-size: 12px; }\n  .mat-list[dense] .mat-list-item .mat-line, .mat-nav-list[dense] .mat-list-item .mat-line, .mat-selection-list[dense] .mat-list-item .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n    .mat-list[dense] .mat-list-item .mat-line:nth-child(n+2), .mat-nav-list[dense] .mat-list-item .mat-line:nth-child(n+2), .mat-selection-list[dense] .mat-list-item .mat-line:nth-child(n+2) {\n      font-size: 12px; }\n\n.mat-list[dense] .mat-list-option, .mat-nav-list[dense] .mat-list-option, .mat-selection-list[dense] .mat-list-option {\n  font-size: 12px; }\n  .mat-list[dense] .mat-list-option .mat-line, .mat-nav-list[dense] .mat-list-option .mat-line, .mat-selection-list[dense] .mat-list-option .mat-line {\n    white-space: nowrap;\n    overflow: hidden;\n    text-overflow: ellipsis;\n    display: block;\n    box-sizing: border-box; }\n    .mat-list[dense] .mat-list-option .mat-line:nth-child(n+2), .mat-nav-list[dense] .mat-list-option .mat-line:nth-child(n+2), .mat-selection-list[dense] .mat-list-option .mat-line:nth-child(n+2) {\n      font-size: 12px; }\n\n.mat-list[dense] .mat-subheader, .mat-nav-list[dense] .mat-subheader, .mat-selection-list[dense] .mat-subheader {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 12px;\n  font-weight: 500; }\n\n.mat-option {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 16px; }\n\n.mat-optgroup-label {\n  font: 500 14px/24px Roboto, "Helvetica Neue", sans-serif; }\n\n.mat-simple-snackbar {\n  font-family: Roboto, "Helvetica Neue", sans-serif;\n  font-size: 14px; }\n\n.mat-simple-snackbar-action {\n  line-height: 1;\n  font-family: inherit;\n  font-size: inherit;\n  font-weight: 500; }\n\n.mat-ripple {\n  overflow: hidden; }\n  @media screen and (-ms-high-contrast: active) {\n    .mat-ripple {\n      display: none; } }\n\n.mat-ripple.mat-ripple-unbounded {\n  overflow: visible; }\n\n.mat-ripple-element {\n  position: absolute;\n  border-radius: 50%;\n  pointer-events: none;\n  transition: opacity, -webkit-transform 0ms cubic-bezier(0, 0, 0.2, 1);\n  transition: opacity, transform 0ms cubic-bezier(0, 0, 0.2, 1);\n  transition: opacity, transform 0ms cubic-bezier(0, 0, 0.2, 1), -webkit-transform 0ms cubic-bezier(0, 0, 0.2, 1);\n  -webkit-transform: scale(0);\n          transform: scale(0); }\n\n.mat-option {\n  white-space: nowrap;\n  overflow: hidden;\n  text-overflow: ellipsis;\n  display: block;\n  line-height: 48px;\n  height: 48px;\n  padding: 0 16px;\n  text-align: left;\n  text-decoration: none;\n  position: relative;\n  cursor: pointer;\n  outline: none;\n  display: -webkit-box;\n  display: -ms-flexbox;\n  display: flex;\n  -webkit-box-orient: horizontal;\n  -webkit-box-direction: normal;\n      -ms-flex-direction: row;\n          flex-direction: row;\n  max-width: 100%;\n  box-sizing: border-box;\n  -webkit-box-align: center;\n      -ms-flex-align: center;\n          align-items: center; }\n  .mat-option[disabled] {\n    cursor: default; }\n  [dir=\'rtl\'] .mat-option {\n    text-align: right; }\n  .mat-option .mat-icon {\n    margin-right: 16px; }\n    [dir=\'rtl\'] .mat-option .mat-icon {\n      margin-left: 16px;\n      margin-right: 0; }\n  .mat-option[aria-disabled=\'true\'] {\n    -webkit-user-select: none;\n    -moz-user-select: none;\n    -ms-user-select: none;\n    user-select: none;\n    cursor: default; }\n  .mat-optgroup .mat-option:not(.mat-option-multiple) {\n    padding-left: 32px; }\n    [dir=\'rtl\'] .mat-optgroup .mat-option:not(.mat-option-multiple) {\n      padding-left: 16px;\n      padding-right: 32px; }\n\n.mat-option-text {\n  display: inline-block;\n  -webkit-box-flex: 1;\n      -ms-flex-positive: 1;\n          flex-grow: 1;\n  overflow: hidden;\n  text-overflow: ellipsis; }\n\n.mat-option-ripple {\n  top: 0;\n  left: 0;\n  right: 0;\n  bottom: 0;\n  position: absolute;\n  pointer-events: none; }\n  @media screen and (-ms-high-contrast: active) {\n    .mat-option-ripple {\n      opacity: 0.5; } }\n\n.mat-option-pseudo-checkbox {\n  margin-right: 8px; }\n  [dir=\'rtl\'] .mat-option-pseudo-checkbox {\n    margin-left: 8px;\n    margin-right: 0; }\n\n.mat-optgroup-label {\n  white-space: nowrap;\n  overflow: hidden;\n  text-overflow: ellipsis;\n  display: block;\n  line-height: 48px;\n  height: 48px;\n  padding: 0 16px;\n  text-align: left;\n  text-decoration: none;\n  -webkit-user-select: none;\n  -moz-user-select: none;\n  -ms-user-select: none;\n  user-select: none;\n  cursor: default; }\n  .mat-optgroup-label[disabled] {\n    cursor: default; }\n  [dir=\'rtl\'] .mat-optgroup-label {\n    text-align: right; }\n  .mat-optgroup-label .mat-icon {\n    margin-right: 16px; }\n    [dir=\'rtl\'] .mat-optgroup-label .mat-icon {\n      margin-left: 16px;\n      margin-right: 0; }\n\n.cdk-visually-hidden {\n  border: 0;\n  clip: rect(0 0 0 0);\n  height: 1px;\n  margin: -1px;\n  overflow: hidden;\n  padding: 0;\n  position: absolute;\n  width: 1px; }\n\n.cdk-overlay-container, .cdk-global-overlay-wrapper {\n  pointer-events: none;\n  top: 0;\n  left: 0;\n  height: 100%;\n  width: 100%; }\n\n.cdk-overlay-container {\n  position: fixed;\n  z-index: 1000; }\n\n.cdk-global-overlay-wrapper {\n  display: -webkit-box;\n  display: -ms-flexbox;\n  display: flex;\n  position: absolute;\n  z-index: 1000; }\n\n.cdk-overlay-pane {\n  position: absolute;\n  pointer-events: auto;\n  box-sizing: border-box;\n  z-index: 1000; }\n\n.cdk-overlay-backdrop {\n  position: absolute;\n  top: 0;\n  bottom: 0;\n  left: 0;\n  right: 0;\n  z-index: 1000;\n  pointer-events: auto;\n  -webkit-tap-highlight-color: transparent;\n  transition: opacity 400ms cubic-bezier(0.25, 0.8, 0.25, 1);\n  opacity: 0; }\n  .cdk-overlay-backdrop.cdk-overlay-backdrop-showing {\n    opacity: 0.48; }\n\n.cdk-overlay-dark-backdrop {\n  background: rgba(0, 0, 0, 0.6); }\n\n.cdk-overlay-transparent-backdrop {\n  background: none; }\n\n.cdk-global-scrollblock {\n  position: fixed;\n  width: 100%;\n  overflow-y: scroll; }\n\n.mat-ripple-element {\n  background-color: rgba(0, 0, 0, 0.1); }\n\n.mat-option {\n  color: rgba(0, 0, 0, 0.87); }\n  .mat-option:hover:not(.mat-option-disabled), .mat-option:focus:not(.mat-option-disabled) {\n    background: rgba(0, 0, 0, 0.04); }\n  .mat-primary .mat-option.mat-selected:not(.mat-option-disabled) {\n    color: #2a7bb8; }\n  .mat-accent .mat-option.mat-selected:not(.mat-option-disabled) {\n    color: #81c6fa; }\n  .mat-warn .mat-option.mat-selected:not(.mat-option-disabled) {\n    color: #f44336; }\n  .mat-option.mat-selected:not(.mat-option-multiple):not(.mat-option-disabled) {\n    background: rgba(0, 0, 0, 0.04); }\n  .mat-option.mat-active {\n    background: rgba(0, 0, 0, 0.04);\n    color: rgba(0, 0, 0, 0.87); }\n  .mat-option.mat-option-disabled {\n    color: rgba(0, 0, 0, 0.38); }\n\n.mat-optgroup-label {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-optgroup-disabled .mat-optgroup-label {\n  color: rgba(0, 0, 0, 0.38); }\n\n.mat-pseudo-checkbox {\n  color: rgba(0, 0, 0, 0.54); }\n  .mat-pseudo-checkbox::after {\n    color: #fafafa; }\n\n.mat-pseudo-checkbox-checked,\n.mat-pseudo-checkbox-indeterminate,\n.mat-accent .mat-pseudo-checkbox-checked,\n.mat-accent .mat-pseudo-checkbox-indeterminate {\n  background: #81c6fa; }\n\n.mat-primary .mat-pseudo-checkbox-checked,\n.mat-primary .mat-pseudo-checkbox-indeterminate {\n  background: #2a7bb8; }\n\n.mat-warn .mat-pseudo-checkbox-checked,\n.mat-warn .mat-pseudo-checkbox-indeterminate {\n  background: #f44336; }\n\n.mat-pseudo-checkbox-checked.mat-pseudo-checkbox-disabled,\n.mat-pseudo-checkbox-indeterminate.mat-pseudo-checkbox-disabled {\n  background: #b0b0b0; }\n\n.mat-app-background {\n  background-color: #fafafa; }\n\n.mat-theme-loaded-marker {\n  display: none; }\n\n.mat-autocomplete-panel {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n  .mat-autocomplete-panel .mat-option.mat-selected:not(.mat-active):not(:hover) {\n    background: white; }\n    .mat-autocomplete-panel .mat-option.mat-selected:not(.mat-active):not(:hover):not(.mat-option-disabled) {\n      color: rgba(0, 0, 0, 0.87); }\n\n.mat-button, .mat-icon-button {\n  background: transparent; }\n  .mat-button.mat-primary .mat-button-focus-overlay, .mat-icon-button.mat-primary .mat-button-focus-overlay {\n    background-color: rgba(42, 123, 184, 0.12); }\n  .mat-button.mat-accent .mat-button-focus-overlay, .mat-icon-button.mat-accent .mat-button-focus-overlay {\n    background-color: rgba(129, 198, 250, 0.12); }\n  .mat-button.mat-warn .mat-button-focus-overlay, .mat-icon-button.mat-warn .mat-button-focus-overlay {\n    background-color: rgba(244, 67, 54, 0.12); }\n  .mat-button[disabled] .mat-button-focus-overlay, .mat-icon-button[disabled] .mat-button-focus-overlay {\n    background-color: transparent; }\n  .mat-button.mat-primary, .mat-icon-button.mat-primary {\n    color: #2a7bb8; }\n  .mat-button.mat-accent, .mat-icon-button.mat-accent {\n    color: #81c6fa; }\n  .mat-button.mat-warn, .mat-icon-button.mat-warn {\n    color: #f44336; }\n  .mat-button.mat-primary[disabled], .mat-button.mat-accent[disabled], .mat-button.mat-warn[disabled], .mat-button[disabled][disabled], .mat-icon-button.mat-primary[disabled], .mat-icon-button.mat-accent[disabled], .mat-icon-button.mat-warn[disabled], .mat-icon-button[disabled][disabled] {\n    color: rgba(0, 0, 0, 0.38); }\n\n.mat-raised-button, .mat-fab, .mat-mini-fab {\n  color: rgba(0, 0, 0, 0.87);\n  background-color: white; }\n  .mat-raised-button.mat-primary, .mat-fab.mat-primary, .mat-mini-fab.mat-primary {\n    color: white; }\n  .mat-raised-button.mat-accent, .mat-fab.mat-accent, .mat-mini-fab.mat-accent {\n    color: black; }\n  .mat-raised-button.mat-warn, .mat-fab.mat-warn, .mat-mini-fab.mat-warn {\n    color: white; }\n  .mat-raised-button.mat-primary[disabled], .mat-raised-button.mat-accent[disabled], .mat-raised-button.mat-warn[disabled], .mat-raised-button[disabled][disabled], .mat-fab.mat-primary[disabled], .mat-fab.mat-accent[disabled], .mat-fab.mat-warn[disabled], .mat-fab[disabled][disabled], .mat-mini-fab.mat-primary[disabled], .mat-mini-fab.mat-accent[disabled], .mat-mini-fab.mat-warn[disabled], .mat-mini-fab[disabled][disabled] {\n    color: rgba(0, 0, 0, 0.38); }\n  .mat-raised-button.mat-primary, .mat-fab.mat-primary, .mat-mini-fab.mat-primary {\n    background-color: #2a7bb8; }\n  .mat-raised-button.mat-accent, .mat-fab.mat-accent, .mat-mini-fab.mat-accent {\n    background-color: #81c6fa; }\n  .mat-raised-button.mat-warn, .mat-fab.mat-warn, .mat-mini-fab.mat-warn {\n    background-color: #f44336; }\n  .mat-raised-button.mat-primary[disabled], .mat-raised-button.mat-accent[disabled], .mat-raised-button.mat-warn[disabled], .mat-raised-button[disabled][disabled], .mat-fab.mat-primary[disabled], .mat-fab.mat-accent[disabled], .mat-fab.mat-warn[disabled], .mat-fab[disabled][disabled], .mat-mini-fab.mat-primary[disabled], .mat-mini-fab.mat-accent[disabled], .mat-mini-fab.mat-warn[disabled], .mat-mini-fab[disabled][disabled] {\n    background-color: rgba(0, 0, 0, 0.12); }\n  .mat-raised-button.mat-primary .mat-ripple-element, .mat-fab.mat-primary .mat-ripple-element, .mat-mini-fab.mat-primary .mat-ripple-element {\n    background-color: rgba(255, 255, 255, 0.2); }\n  .mat-raised-button.mat-accent .mat-ripple-element, .mat-fab.mat-accent .mat-ripple-element, .mat-mini-fab.mat-accent .mat-ripple-element {\n    background-color: rgba(0, 0, 0, 0.2); }\n  .mat-raised-button.mat-warn .mat-ripple-element, .mat-fab.mat-warn .mat-ripple-element, .mat-mini-fab.mat-warn .mat-ripple-element {\n    background-color: rgba(255, 255, 255, 0.2); }\n\n.mat-button.mat-primary .mat-ripple-element {\n  background-color: rgba(42, 123, 184, 0.1); }\n\n.mat-button.mat-accent .mat-ripple-element {\n  background-color: rgba(129, 198, 250, 0.1); }\n\n.mat-button.mat-warn .mat-ripple-element {\n  background-color: rgba(244, 67, 54, 0.1); }\n\n.mat-icon-button.mat-primary .mat-ripple-element {\n  background-color: rgba(42, 123, 184, 0.2); }\n\n.mat-icon-button.mat-accent .mat-ripple-element {\n  background-color: rgba(129, 198, 250, 0.2); }\n\n.mat-icon-button.mat-warn .mat-ripple-element {\n  background-color: rgba(244, 67, 54, 0.2); }\n\n.mat-button-toggle {\n  color: rgba(0, 0, 0, 0.38); }\n  .mat-button-toggle.cdk-focused .mat-button-toggle-focus-overlay {\n    background-color: rgba(0, 0, 0, 0.06); }\n\n.mat-button-toggle-checked {\n  background-color: #e0e0e0;\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-button-toggle-disabled {\n  background-color: #eeeeee;\n  color: rgba(0, 0, 0, 0.38); }\n  .mat-button-toggle-disabled.mat-button-toggle-checked {\n    background-color: #bdbdbd; }\n\n.mat-card {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-card-subtitle {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-checkbox-frame {\n  border-color: rgba(0, 0, 0, 0.54); }\n\n.mat-checkbox-checkmark {\n  fill: #fafafa; }\n\n.mat-checkbox-checkmark-path {\n  stroke: #fafafa !important; }\n\n.mat-checkbox-mixedmark {\n  background-color: #fafafa; }\n\n.mat-checkbox-indeterminate.mat-primary .mat-checkbox-background, .mat-checkbox-checked.mat-primary .mat-checkbox-background {\n  background-color: #2a7bb8; }\n\n.mat-checkbox-indeterminate.mat-accent .mat-checkbox-background, .mat-checkbox-checked.mat-accent .mat-checkbox-background {\n  background-color: #81c6fa; }\n\n.mat-checkbox-indeterminate.mat-warn .mat-checkbox-background, .mat-checkbox-checked.mat-warn .mat-checkbox-background {\n  background-color: #f44336; }\n\n.mat-checkbox-disabled.mat-checkbox-checked .mat-checkbox-background, .mat-checkbox-disabled.mat-checkbox-indeterminate .mat-checkbox-background {\n  background-color: #b0b0b0; }\n\n.mat-checkbox-disabled:not(.mat-checkbox-checked) .mat-checkbox-frame {\n  border-color: #b0b0b0; }\n\n.mat-checkbox-disabled .mat-checkbox-label {\n  color: #b0b0b0; }\n\n.mat-checkbox:not(.mat-checkbox-disabled).mat-primary .mat-checkbox-ripple .mat-ripple-element {\n  background-color: rgba(42, 123, 184, 0.26); }\n\n.mat-checkbox:not(.mat-checkbox-disabled).mat-accent .mat-checkbox-ripple .mat-ripple-element {\n  background-color: rgba(129, 198, 250, 0.26); }\n\n.mat-checkbox:not(.mat-checkbox-disabled).mat-warn .mat-checkbox-ripple .mat-ripple-element {\n  background-color: rgba(244, 67, 54, 0.26); }\n\n.mat-chip:not(.mat-basic-chip) {\n  background-color: #e0e0e0;\n  color: rgba(0, 0, 0, 0.87); }\n  .mat-chip:not(.mat-basic-chip) .mat-chip-remove {\n    color: rgba(0, 0, 0, 0.87);\n    opacity: 0.4; }\n  .mat-chip:not(.mat-basic-chip) .mat-chip-remove:hover {\n    opacity: 0.54; }\n\n.mat-chip.mat-chip-selected.mat-primary {\n  background-color: #2a7bb8;\n  color: white; }\n  .mat-chip.mat-chip-selected.mat-primary .mat-chip-remove {\n    color: white;\n    opacity: 0.4; }\n  .mat-chip.mat-chip-selected.mat-primary .mat-chip-remove:hover {\n    opacity: 0.54; }\n\n.mat-chip.mat-chip-selected.mat-warn {\n  background-color: #f44336;\n  color: white; }\n  .mat-chip.mat-chip-selected.mat-warn .mat-chip-remove {\n    color: white;\n    opacity: 0.4; }\n  .mat-chip.mat-chip-selected.mat-warn .mat-chip-remove:hover {\n    opacity: 0.54; }\n\n.mat-chip.mat-chip-selected.mat-accent {\n  background-color: #81c6fa;\n  color: black; }\n  .mat-chip.mat-chip-selected.mat-accent .mat-chip-remove {\n    color: black;\n    opacity: 0.4; }\n  .mat-chip.mat-chip-selected.mat-accent .mat-chip-remove:hover {\n    opacity: 0.54; }\n\n.mat-table {\n  background: white; }\n\n.mat-row, .mat-header-row {\n  border-bottom-color: rgba(0, 0, 0, 0.12); }\n\n.mat-header-cell {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-cell {\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-datepicker-content {\n  background-color: white;\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-calendar-arrow {\n  border-top-color: rgba(0, 0, 0, 0.54); }\n\n.mat-calendar-next-button,\n.mat-calendar-previous-button {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-calendar-table-header {\n  color: rgba(0, 0, 0, 0.38); }\n\n.mat-calendar-table-header-divider::after {\n  background: rgba(0, 0, 0, 0.12); }\n\n.mat-calendar-body-label {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-calendar-body-cell-content {\n  color: rgba(0, 0, 0, 0.87);\n  border-color: transparent; }\n  .mat-calendar-body-disabled > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected) {\n    color: rgba(0, 0, 0, 0.38); }\n\n:not(.mat-calendar-body-disabled):hover > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected),\n.cdk-keyboard-focused .mat-calendar-body-active > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected),\n.cdk-program-focused .mat-calendar-body-active > .mat-calendar-body-cell-content:not(.mat-calendar-body-selected) {\n  background-color: rgba(0, 0, 0, 0.04); }\n\n.mat-calendar-body-selected {\n  background-color: #2a7bb8;\n  color: white; }\n\n.mat-calendar-body-disabled > .mat-calendar-body-selected {\n  background-color: rgba(42, 123, 184, 0.4); }\n\n.mat-calendar-body-today:not(.mat-calendar-body-selected) {\n  border-color: rgba(0, 0, 0, 0.38); }\n\n.mat-calendar-body-today.mat-calendar-body-selected {\n  box-shadow: inset 0 0 0 1px white; }\n\n.mat-calendar-body-disabled > .mat-calendar-body-today:not(.mat-calendar-body-selected) {\n  border-color: rgba(0, 0, 0, 0.18); }\n\n.mat-dialog-container {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-expansion-panel {\n  background: white;\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-action-row {\n  border-top-color: rgba(0, 0, 0, 0.12); }\n\n.mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=\'true\']).cdk-keyboard-focused, .mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=\'true\']).cdk-program-focused, .mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=\'true\']):hover {\n  background: rgba(0, 0, 0, 0.04); }\n\n.mat-expansion-panel-header-title {\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-expansion-panel-header-description,\n.mat-expansion-indicator::after {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-expansion-panel-header[aria-disabled=\'true\'] {\n  color: rgba(0, 0, 0, 0.38); }\n  .mat-expansion-panel-header[aria-disabled=\'true\'] .mat-expansion-panel-header-title,\n  .mat-expansion-panel-header[aria-disabled=\'true\'] .mat-expansion-panel-header-description {\n    color: inherit; }\n\n.mat-form-field-label {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-hint {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-focused .mat-form-field-label {\n  color: #2a7bb8; }\n  .mat-focused .mat-form-field-label.mat-accent {\n    color: #81c6fa; }\n  .mat-focused .mat-form-field-label.mat-warn {\n    color: #f44336; }\n\n.mat-focused .mat-form-field-required-marker {\n  color: #81c6fa; }\n\n.mat-form-field-underline {\n  background-color: rgba(0, 0, 0, 0.42); }\n\n.mat-form-field-disabled .mat-form-field-underline {\n  background-image: linear-gradient(to right, rgba(0, 0, 0, 0.42) 0%, rgba(0, 0, 0, 0.42) 33%, transparent 0%);\n  background-size: 4px 1px;\n  background-repeat: repeat-x; }\n\n.mat-form-field-ripple {\n  background-color: #2a7bb8; }\n  .mat-form-field-ripple.mat-accent {\n    background-color: #81c6fa; }\n  .mat-form-field-ripple.mat-warn {\n    background-color: #f44336; }\n\n.mat-form-field-invalid .mat-form-field-label {\n  color: #f44336; }\n  .mat-form-field-invalid .mat-form-field-label.mat-accent,\n  .mat-form-field-invalid .mat-form-field-label .mat-form-field-required-marker {\n    color: #f44336; }\n\n.mat-form-field-invalid .mat-form-field-ripple {\n  background-color: #f44336; }\n\n.mat-error {\n  color: #f44336; }\n\n.mat-icon.mat-primary {\n  color: #2a7bb8; }\n\n.mat-icon.mat-accent {\n  color: #81c6fa; }\n\n.mat-icon.mat-warn {\n  color: #f44336; }\n\n.mat-input-element:disabled {\n  color: rgba(0, 0, 0, 0.42); }\n\n.mat-input-element::-webkit-input-placeholder {\n  color: rgba(0, 0, 0, 0.42); }\n\n.mat-input-element:-ms-input-placeholder {\n  color: rgba(0, 0, 0, 0.42); }\n\n.mat-input-element::placeholder {\n  color: rgba(0, 0, 0, 0.42); }\n\n.mat-input-element::-moz-placeholder {\n  color: rgba(0, 0, 0, 0.42); }\n\n.mat-input-element::-webkit-input-placeholder {\n  color: rgba(0, 0, 0, 0.42); }\n\n.mat-input-element:-ms-input-placeholder {\n  color: rgba(0, 0, 0, 0.42); }\n\n.mat-list .mat-list-item, .mat-nav-list .mat-list-item, .mat-selection-list .mat-list-item {\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-list .mat-list-option, .mat-nav-list .mat-list-option, .mat-selection-list .mat-list-option {\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-list .mat-subheader, .mat-nav-list .mat-subheader, .mat-selection-list .mat-subheader {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-list-item-disabled {\n  background-color: #eeeeee; }\n\n.mat-divider {\n  border-top-color: rgba(0, 0, 0, 0.12); }\n\n.mat-nav-list .mat-list-item {\n  outline: none; }\n  .mat-nav-list .mat-list-item:hover, .mat-nav-list .mat-list-item.mat-list-item-focus {\n    background: rgba(0, 0, 0, 0.04); }\n\n.mat-list-option {\n  outline: none; }\n  .mat-list-option:hover, .mat-list-option.mat-list-item-focus {\n    background: rgba(0, 0, 0, 0.04); }\n\n.mat-menu-panel {\n  background: white; }\n\n.mat-menu-item {\n  background: transparent;\n  color: rgba(0, 0, 0, 0.87); }\n  .mat-menu-item[disabled] {\n    color: rgba(0, 0, 0, 0.38); }\n\n.mat-menu-item .mat-icon:not([color]),\n.mat-menu-item-submenu-trigger::after {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-menu-item:hover:not([disabled]),\n.mat-menu-item:focus:not([disabled]),\n.mat-menu-item-highlighted:not([disabled]) {\n  background: rgba(0, 0, 0, 0.04); }\n\n.mat-paginator {\n  background: white; }\n\n.mat-paginator,\n.mat-paginator-page-size .mat-select-trigger {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-paginator-increment,\n.mat-paginator-decrement {\n  border-top: 2px solid rgba(0, 0, 0, 0.54);\n  border-right: 2px solid rgba(0, 0, 0, 0.54); }\n\n.mat-icon-button[disabled] .mat-paginator-increment,\n.mat-icon-button[disabled] .mat-paginator-decrement {\n  border-color: rgba(0, 0, 0, 0.38); }\n\n.mat-progress-bar-background {\n  background-image: url("data:image/svg+xml;charset=UTF-8,%3Csvg%20version%3D%271.1%27%20xmlns%3D%27http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%27%20xmlns%3Axlink%3D%27http%3A%2F%2Fwww.w3.org%2F1999%2Fxlink%27%20x%3D%270px%27%20y%3D%270px%27%20enable-background%3D%27new%200%200%205%202%27%20xml%3Aspace%3D%27preserve%27%20viewBox%3D%270%200%205%202%27%20preserveAspectRatio%3D%27none%20slice%27%3E%3Ccircle%20cx%3D%271%27%20cy%3D%271%27%20r%3D%271%27%20fill%3D%27%23b1d4ed%27%2F%3E%3C%2Fsvg%3E"); }\n\n.mat-progress-bar-buffer {\n  background-color: #b1d4ed; }\n\n.mat-progress-bar-fill::after {\n  background-color: #2a7bb8; }\n\n.mat-progress-bar.mat-accent .mat-progress-bar-background {\n  background-image: url("data:image/svg+xml;charset=UTF-8,%3Csvg%20version%3D%271.1%27%20xmlns%3D%27http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%27%20xmlns%3Axlink%3D%27http%3A%2F%2Fwww.w3.org%2F1999%2Fxlink%27%20x%3D%270px%27%20y%3D%270px%27%20enable-background%3D%27new%200%200%205%202%27%20xml%3Aspace%3D%27preserve%27%20viewBox%3D%270%200%205%202%27%20preserveAspectRatio%3D%27none%20slice%27%3E%3Ccircle%20cx%3D%271%27%20cy%3D%271%27%20r%3D%271%27%20fill%3D%27%23b1d4ed%27%2F%3E%3C%2Fsvg%3E"); }\n\n.mat-progress-bar.mat-accent .mat-progress-bar-buffer {\n  background-color: #b1d4ed; }\n\n.mat-progress-bar.mat-accent .mat-progress-bar-fill::after {\n  background-color: #81c6fa; }\n\n.mat-progress-bar.mat-warn .mat-progress-bar-background {\n  background-image: url("data:image/svg+xml;charset=UTF-8,%3Csvg%20version%3D%271.1%27%20xmlns%3D%27http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%27%20xmlns%3Axlink%3D%27http%3A%2F%2Fwww.w3.org%2F1999%2Fxlink%27%20x%3D%270px%27%20y%3D%270px%27%20enable-background%3D%27new%200%200%205%202%27%20xml%3Aspace%3D%27preserve%27%20viewBox%3D%270%200%205%202%27%20preserveAspectRatio%3D%27none%20slice%27%3E%3Ccircle%20cx%3D%271%27%20cy%3D%271%27%20r%3D%271%27%20fill%3D%27%23ffcdd2%27%2F%3E%3C%2Fsvg%3E"); }\n\n.mat-progress-bar.mat-warn .mat-progress-bar-buffer {\n  background-color: #ffcdd2; }\n\n.mat-progress-bar.mat-warn .mat-progress-bar-fill::after {\n  background-color: #f44336; }\n\n.mat-progress-spinner circle, .mat-spinner circle {\n  stroke: #2a7bb8; }\n\n.mat-progress-spinner.mat-accent circle, .mat-spinner.mat-accent circle {\n  stroke: #81c6fa; }\n\n.mat-progress-spinner.mat-warn circle, .mat-spinner.mat-warn circle {\n  stroke: #f44336; }\n\n.mat-radio-outer-circle {\n  border-color: rgba(0, 0, 0, 0.54); }\n\n.mat-radio-disabled .mat-radio-outer-circle {\n  border-color: rgba(0, 0, 0, 0.38); }\n\n.mat-radio-disabled .mat-radio-ripple .mat-ripple-element, .mat-radio-disabled .mat-radio-inner-circle {\n  background-color: rgba(0, 0, 0, 0.38); }\n\n.mat-radio-disabled .mat-radio-label-content {\n  color: rgba(0, 0, 0, 0.38); }\n\n.mat-radio-button.mat-primary.mat-radio-checked .mat-radio-outer-circle {\n  border-color: #2a7bb8; }\n\n.mat-radio-button.mat-primary .mat-radio-inner-circle {\n  background-color: #2a7bb8; }\n\n.mat-radio-button.mat-primary .mat-radio-ripple .mat-ripple-element {\n  background-color: rgba(42, 123, 184, 0.26); }\n\n.mat-radio-button.mat-accent.mat-radio-checked .mat-radio-outer-circle {\n  border-color: #81c6fa; }\n\n.mat-radio-button.mat-accent .mat-radio-inner-circle {\n  background-color: #81c6fa; }\n\n.mat-radio-button.mat-accent .mat-radio-ripple .mat-ripple-element {\n  background-color: rgba(129, 198, 250, 0.26); }\n\n.mat-radio-button.mat-warn.mat-radio-checked .mat-radio-outer-circle {\n  border-color: #f44336; }\n\n.mat-radio-button.mat-warn .mat-radio-inner-circle {\n  background-color: #f44336; }\n\n.mat-radio-button.mat-warn .mat-radio-ripple .mat-ripple-element {\n  background-color: rgba(244, 67, 54, 0.26); }\n\n.mat-select-disabled .mat-select-value,\n.mat-select-arrow {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-select-content, .mat-select-panel-done-animating {\n  background: white; }\n\n.mat-select-value {\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-select-panel .mat-option.mat-selected:not(.mat-option-multiple) {\n  background: rgba(0, 0, 0, 0.12); }\n\n.mat-form-field.mat-focused.mat-primary .mat-select-arrow {\n  color: #2a7bb8; }\n\n.mat-form-field.mat-focused.mat-accent .mat-select-arrow {\n  color: #81c6fa; }\n\n.mat-form-field.mat-focused.mat-warn .mat-select-arrow {\n  color: #f44336; }\n\n.mat-form-field .mat-select.mat-select-invalid .mat-select-arrow {\n  color: #f44336; }\n\n.mat-form-field .mat-select.mat-select-disabled .mat-select-arrow {\n  color: rgba(0, 0, 0, 0.54); }\n\n.mat-select.mat-select-disabled .mat-select-arrow {\n  color: #f44336; }\n\n.mat-select-placeholder {\n  color: rgba(0, 0, 0, 0.42); }\n\n.mat-drawer-container {\n  background-color: #fafafa;\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-drawer {\n  background-color: white;\n  color: rgba(0, 0, 0, 0.87); }\n  .mat-drawer.mat-drawer-push {\n    background-color: white; }\n\n.mat-drawer-backdrop.mat-drawer-shown {\n  background-color: rgba(0, 0, 0, 0.6); }\n\n.mat-slide-toggle.mat-checked:not(.mat-disabled) .mat-slide-toggle-thumb {\n  background-color: #2a7bb8; }\n\n.mat-slide-toggle.mat-checked:not(.mat-disabled) .mat-slide-toggle-bar {\n  background-color: rgba(42, 123, 184, 0.5); }\n\n.mat-slide-toggle:not(.mat-checked) .mat-ripple-element {\n  background-color: rgba(0, 0, 0, 0.06); }\n\n.mat-slide-toggle .mat-ripple-element {\n  background-color: rgba(42, 123, 184, 0.12); }\n\n.mat-slide-toggle.mat-primary.mat-checked:not(.mat-disabled) .mat-slide-toggle-thumb {\n  background-color: #2a7bb8; }\n\n.mat-slide-toggle.mat-primary.mat-checked:not(.mat-disabled) .mat-slide-toggle-bar {\n  background-color: rgba(42, 123, 184, 0.5); }\n\n.mat-slide-toggle.mat-primary:not(.mat-checked) .mat-ripple-element {\n  background-color: rgba(0, 0, 0, 0.06); }\n\n.mat-slide-toggle.mat-primary .mat-ripple-element {\n  background-color: rgba(42, 123, 184, 0.12); }\n\n.mat-slide-toggle.mat-warn.mat-checked:not(.mat-disabled) .mat-slide-toggle-thumb {\n  background-color: #f44336; }\n\n.mat-slide-toggle.mat-warn.mat-checked:not(.mat-disabled) .mat-slide-toggle-bar {\n  background-color: rgba(244, 67, 54, 0.5); }\n\n.mat-slide-toggle.mat-warn:not(.mat-checked) .mat-ripple-element {\n  background-color: rgba(0, 0, 0, 0.06); }\n\n.mat-slide-toggle.mat-warn .mat-ripple-element {\n  background-color: rgba(244, 67, 54, 0.12); }\n\n.mat-disabled .mat-slide-toggle-thumb {\n  background-color: #bdbdbd; }\n\n.mat-disabled .mat-slide-toggle-bar {\n  background-color: rgba(0, 0, 0, 0.1); }\n\n.mat-slide-toggle-thumb {\n  background-color: #fafafa; }\n\n.mat-slide-toggle-bar {\n  background-color: rgba(0, 0, 0, 0.38); }\n\n.mat-slider-track-background {\n  background-color: rgba(0, 0, 0, 0.26); }\n\n.mat-primary .mat-slider-track-fill,\n.mat-primary .mat-slider-thumb,\n.mat-primary .mat-slider-thumb-label {\n  background-color: #2a7bb8; }\n\n.mat-primary .mat-slider-thumb-label-text {\n  color: white; }\n\n.mat-accent .mat-slider-track-fill,\n.mat-accent .mat-slider-thumb,\n.mat-accent .mat-slider-thumb-label {\n  background-color: #81c6fa; }\n\n.mat-accent .mat-slider-thumb-label-text {\n  color: black; }\n\n.mat-warn .mat-slider-track-fill,\n.mat-warn .mat-slider-thumb,\n.mat-warn .mat-slider-thumb-label {\n  background-color: #f44336; }\n\n.mat-warn .mat-slider-thumb-label-text {\n  color: white; }\n\n.mat-slider-focus-ring {\n  background-color: rgba(129, 198, 250, 0.2); }\n\n.mat-slider:hover .mat-slider-track-background,\n.cdk-focused .mat-slider-track-background {\n  background-color: rgba(0, 0, 0, 0.38); }\n\n.mat-slider-disabled .mat-slider-track-background,\n.mat-slider-disabled .mat-slider-track-fill,\n.mat-slider-disabled .mat-slider-thumb {\n  background-color: rgba(0, 0, 0, 0.26); }\n\n.mat-slider-disabled:hover .mat-slider-track-background {\n  background-color: rgba(0, 0, 0, 0.26); }\n\n.mat-slider-min-value .mat-slider-focus-ring {\n  background-color: rgba(0, 0, 0, 0.12); }\n\n.mat-slider-min-value.mat-slider-thumb-label-showing .mat-slider-thumb,\n.mat-slider-min-value.mat-slider-thumb-label-showing .mat-slider-thumb-label {\n  background-color: rgba(0, 0, 0, 0.87); }\n\n.mat-slider-min-value.mat-slider-thumb-label-showing.cdk-focused .mat-slider-thumb,\n.mat-slider-min-value.mat-slider-thumb-label-showing.cdk-focused .mat-slider-thumb-label {\n  background-color: rgba(0, 0, 0, 0.26); }\n\n.mat-slider-min-value:not(.mat-slider-thumb-label-showing) .mat-slider-thumb {\n  border-color: rgba(0, 0, 0, 0.26);\n  background-color: transparent; }\n\n.mat-slider-min-value:not(.mat-slider-thumb-label-showing):hover .mat-slider-thumb, .mat-slider-min-value:not(.mat-slider-thumb-label-showing).cdk-focused .mat-slider-thumb {\n  border-color: rgba(0, 0, 0, 0.38); }\n\n.mat-slider-min-value:not(.mat-slider-thumb-label-showing):hover.mat-slider-disabled .mat-slider-thumb, .mat-slider-min-value:not(.mat-slider-thumb-label-showing).cdk-focused.mat-slider-disabled .mat-slider-thumb {\n  border-color: rgba(0, 0, 0, 0.26); }\n\n.mat-slider-has-ticks .mat-slider-wrapper::after {\n  border-color: rgba(0, 0, 0, 0.7); }\n\n.mat-slider-horizontal .mat-slider-ticks {\n  background-image: repeating-linear-gradient(to right, rgba(0, 0, 0, 0.7), rgba(0, 0, 0, 0.7) 2px, transparent 0, transparent);\n  background-image: -moz-repeating-linear-gradient(0.0001deg, rgba(0, 0, 0, 0.7), rgba(0, 0, 0, 0.7) 2px, transparent 0, transparent); }\n\n.mat-slider-vertical .mat-slider-ticks {\n  background-image: repeating-linear-gradient(to bottom, rgba(0, 0, 0, 0.7), rgba(0, 0, 0, 0.7) 2px, transparent 0, transparent); }\n\n.mat-step-header.cdk-keyboard-focused, .mat-step-header.cdk-program-focused, .mat-step-header:hover {\n  background-color: rgba(0, 0, 0, 0.04); }\n\n.mat-step-header .mat-step-label,\n.mat-step-header .mat-step-optional {\n  color: rgba(0, 0, 0, 0.38); }\n\n.mat-step-header .mat-step-icon {\n  background-color: #2a7bb8;\n  color: white; }\n\n.mat-step-header .mat-step-icon-not-touched {\n  background-color: rgba(0, 0, 0, 0.38);\n  color: white; }\n\n.mat-step-header .mat-step-label.mat-step-label-active {\n  color: rgba(0, 0, 0, 0.87); }\n\n.mat-stepper-horizontal, .mat-stepper-vertical {\n  background-color: white; }\n\n.mat-stepper-vertical-line::before {\n  border-left-color: rgba(0, 0, 0, 0.12); }\n\n.mat-stepper-horizontal-line {\n  border-top-color: rgba(0, 0, 0, 0.12); }\n\n.mat-tab-nav-bar,\n.mat-tab-header {\n  border-bottom: 1px solid rgba(0, 0, 0, 0.12); }\n\n.mat-tab-group-inverted-header .mat-tab-nav-bar,\n.mat-tab-group-inverted-header .mat-tab-header {\n  border-top: 1px solid rgba(0, 0, 0, 0.12);\n  border-bottom: none; }\n\n.mat-tab-label, .mat-tab-link {\n  color: rgba(0, 0, 0, 0.87); }\n  .mat-tab-label.mat-tab-disabled, .mat-tab-link.mat-tab-disabled {\n    color: rgba(0, 0, 0, 0.38); }\n\n.mat-tab-header-pagination-chevron {\n  border-color: rgba(0, 0, 0, 0.87); }\n\n.mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(0, 0, 0, 0.38); }\n\n.mat-tab-group[class*=\'mat-background-\'] .mat-tab-header,\n.mat-tab-nav-bar[class*=\'mat-background-\'] {\n  border-bottom: none;\n  border-top: none; }\n\n.mat-tab-group.mat-primary .mat-tab-label:focus, .mat-tab-group.mat-primary .mat-tab-link:focus, .mat-tab-nav-bar.mat-primary .mat-tab-label:focus, .mat-tab-nav-bar.mat-primary .mat-tab-link:focus {\n  background-color: rgba(177, 212, 237, 0.3); }\n\n.mat-tab-group.mat-primary .mat-ink-bar, .mat-tab-nav-bar.mat-primary .mat-ink-bar {\n  background-color: #2a7bb8; }\n\n.mat-tab-group.mat-primary.mat-background-primary .mat-ink-bar, .mat-tab-nav-bar.mat-primary.mat-background-primary .mat-ink-bar {\n  background-color: white; }\n\n.mat-tab-group.mat-accent .mat-tab-label:focus, .mat-tab-group.mat-accent .mat-tab-link:focus, .mat-tab-nav-bar.mat-accent .mat-tab-label:focus, .mat-tab-nav-bar.mat-accent .mat-tab-link:focus {\n  background-color: rgba(177, 212, 237, 0.3); }\n\n.mat-tab-group.mat-accent .mat-ink-bar, .mat-tab-nav-bar.mat-accent .mat-ink-bar {\n  background-color: #81c6fa; }\n\n.mat-tab-group.mat-accent.mat-background-accent .mat-ink-bar, .mat-tab-nav-bar.mat-accent.mat-background-accent .mat-ink-bar {\n  background-color: black; }\n\n.mat-tab-group.mat-warn .mat-tab-label:focus, .mat-tab-group.mat-warn .mat-tab-link:focus, .mat-tab-nav-bar.mat-warn .mat-tab-label:focus, .mat-tab-nav-bar.mat-warn .mat-tab-link:focus {\n  background-color: rgba(255, 205, 210, 0.3); }\n\n.mat-tab-group.mat-warn .mat-ink-bar, .mat-tab-nav-bar.mat-warn .mat-ink-bar {\n  background-color: #f44336; }\n\n.mat-tab-group.mat-warn.mat-background-warn .mat-ink-bar, .mat-tab-nav-bar.mat-warn.mat-background-warn .mat-ink-bar {\n  background-color: white; }\n\n.mat-tab-group.mat-background-primary .mat-tab-label:focus, .mat-tab-group.mat-background-primary .mat-tab-link:focus, .mat-tab-nav-bar.mat-background-primary .mat-tab-label:focus, .mat-tab-nav-bar.mat-background-primary .mat-tab-link:focus {\n  background-color: rgba(177, 212, 237, 0.3); }\n\n.mat-tab-group.mat-background-primary .mat-tab-header, .mat-tab-group.mat-background-primary .mat-tab-links, .mat-tab-nav-bar.mat-background-primary .mat-tab-header, .mat-tab-nav-bar.mat-background-primary .mat-tab-links {\n  background-color: #2a7bb8; }\n\n.mat-tab-group.mat-background-primary .mat-tab-label, .mat-tab-group.mat-background-primary .mat-tab-link, .mat-tab-nav-bar.mat-background-primary .mat-tab-label, .mat-tab-nav-bar.mat-background-primary .mat-tab-link {\n  color: white; }\n  .mat-tab-group.mat-background-primary .mat-tab-label.mat-tab-disabled, .mat-tab-group.mat-background-primary .mat-tab-link.mat-tab-disabled, .mat-tab-nav-bar.mat-background-primary .mat-tab-label.mat-tab-disabled, .mat-tab-nav-bar.mat-background-primary .mat-tab-link.mat-tab-disabled {\n    color: rgba(255, 255, 255, 0.4); }\n\n.mat-tab-group.mat-background-primary .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-primary .mat-tab-header-pagination-chevron {\n  border-color: white; }\n\n.mat-tab-group.mat-background-primary .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-primary .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(255, 255, 255, 0.4); }\n\n.mat-tab-group.mat-background-primary .mat-ripple-element, .mat-tab-nav-bar.mat-background-primary .mat-ripple-element {\n  background-color: rgba(255, 255, 255, 0.12); }\n\n.mat-tab-group.mat-background-accent .mat-tab-label:focus, .mat-tab-group.mat-background-accent .mat-tab-link:focus, .mat-tab-nav-bar.mat-background-accent .mat-tab-label:focus, .mat-tab-nav-bar.mat-background-accent .mat-tab-link:focus {\n  background-color: rgba(177, 212, 237, 0.3); }\n\n.mat-tab-group.mat-background-accent .mat-tab-header, .mat-tab-group.mat-background-accent .mat-tab-links, .mat-tab-nav-bar.mat-background-accent .mat-tab-header, .mat-tab-nav-bar.mat-background-accent .mat-tab-links {\n  background-color: #81c6fa; }\n\n.mat-tab-group.mat-background-accent .mat-tab-label, .mat-tab-group.mat-background-accent .mat-tab-link, .mat-tab-nav-bar.mat-background-accent .mat-tab-label, .mat-tab-nav-bar.mat-background-accent .mat-tab-link {\n  color: black; }\n  .mat-tab-group.mat-background-accent .mat-tab-label.mat-tab-disabled, .mat-tab-group.mat-background-accent .mat-tab-link.mat-tab-disabled, .mat-tab-nav-bar.mat-background-accent .mat-tab-label.mat-tab-disabled, .mat-tab-nav-bar.mat-background-accent .mat-tab-link.mat-tab-disabled {\n    color: rgba(0, 0, 0, 0.4); }\n\n.mat-tab-group.mat-background-accent .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-accent .mat-tab-header-pagination-chevron {\n  border-color: black; }\n\n.mat-tab-group.mat-background-accent .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-accent .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(0, 0, 0, 0.4); }\n\n.mat-tab-group.mat-background-accent .mat-ripple-element, .mat-tab-nav-bar.mat-background-accent .mat-ripple-element {\n  background-color: rgba(0, 0, 0, 0.12); }\n\n.mat-tab-group.mat-background-warn .mat-tab-label:focus, .mat-tab-group.mat-background-warn .mat-tab-link:focus, .mat-tab-nav-bar.mat-background-warn .mat-tab-label:focus, .mat-tab-nav-bar.mat-background-warn .mat-tab-link:focus {\n  background-color: rgba(255, 205, 210, 0.3); }\n\n.mat-tab-group.mat-background-warn .mat-tab-header, .mat-tab-group.mat-background-warn .mat-tab-links, .mat-tab-nav-bar.mat-background-warn .mat-tab-header, .mat-tab-nav-bar.mat-background-warn .mat-tab-links {\n  background-color: #f44336; }\n\n.mat-tab-group.mat-background-warn .mat-tab-label, .mat-tab-group.mat-background-warn .mat-tab-link, .mat-tab-nav-bar.mat-background-warn .mat-tab-label, .mat-tab-nav-bar.mat-background-warn .mat-tab-link {\n  color: white; }\n  .mat-tab-group.mat-background-warn .mat-tab-label.mat-tab-disabled, .mat-tab-group.mat-background-warn .mat-tab-link.mat-tab-disabled, .mat-tab-nav-bar.mat-background-warn .mat-tab-label.mat-tab-disabled, .mat-tab-nav-bar.mat-background-warn .mat-tab-link.mat-tab-disabled {\n    color: rgba(255, 255, 255, 0.4); }\n\n.mat-tab-group.mat-background-warn .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-warn .mat-tab-header-pagination-chevron {\n  border-color: white; }\n\n.mat-tab-group.mat-background-warn .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron, .mat-tab-nav-bar.mat-background-warn .mat-tab-header-pagination-disabled .mat-tab-header-pagination-chevron {\n  border-color: rgba(255, 255, 255, 0.4); }\n\n.mat-tab-group.mat-background-warn .mat-ripple-element, .mat-tab-nav-bar.mat-background-warn .mat-ripple-element {\n  background-color: rgba(255, 255, 255, 0.12); }\n\n.mat-toolbar {\n  background: whitesmoke;\n  color: rgba(0, 0, 0, 0.87); }\n  .mat-toolbar.mat-primary {\n    background: #2a7bb8;\n    color: white; }\n  .mat-toolbar.mat-accent {\n    background: #81c6fa;\n    color: black; }\n  .mat-toolbar.mat-warn {\n    background: #f44336;\n    color: white; }\n\n.mat-tooltip {\n  background: rgba(97, 97, 97, 0.9); }\n\n.mat-snack-bar-container {\n  background: #323232;\n  color: white; }\n\n.mat-simple-snackbar-action {\n  color: #81c6fa; }\n\n.fixed-header {\n  position: fixed;\n  top: 0;\n  left: 0;\n  z-index: 2;\n  width: 100%  !important; }\n\nbody {\n  padding-top: 3%; }\n\ntextarea {\n  color: Gray;\n  border: none;\n  background-color: #f0f6fc;\n  overflow-wrap: break-word;\n  width: 100%; }\n\n.test-title {\n  color: black;\n  font-weight: 500; }\n',""])},"../../../../css-loader/lib/css-base.js":function(a,t){a.exports=function(a){var t=[];return t.toString=function(){return this.map(function(t){var n=function(a,t){var n=a[1]||"",e=a[3];if(!e)return n;if(t&&"function"==typeof btoa){var o=(i=e,"/*# sourceMappingURL=data:application/json;charset=utf-8;base64,"+btoa(unescape(encodeURIComponent(JSON.stringify(i))))+" */"),r=e.sources.map(function(a){return"/*# sourceURL="+e.sourceRoot+a+" */"});return[n].concat(r).concat([o]).join("\n")}var i;return[n].join("\n")}(t,a);return t[2]?"@media "+t[2]+"{"+n+"}":n}).join("")},t.i=function(a,n){"string"==typeof a&&(a=[[null,a,""]]);for(var e={},o=0;o<this.length;o++){var r=this[o][0];"number"==typeof r&&(e[r]=!0)}for(o=0;o<a.length;o++){var i=a[o];"number"==typeof i[0]&&e[i[0]]||(n&&!i[2]?i[2]=n:n&&(i[2]="("+i[2]+") and ("+n+")"),t.push(i))}},t}},"../../../../style-loader/addStyles.js":function(a,t){var n={},e=function(a){var t;return function(){return void 0===t&&(t=a.apply(this,arguments)),t}},o=e(function(){return/msie [6-9]\b/.test(self.navigator.userAgent.toLowerCase())}),r=e(function(){return document.head||document.getElementsByTagName("head")[0]}),i=null,m=0,l=[];function b(a,t){for(var e=0;e<a.length;e++){var o=a[e],r=n[o.id];if(r){r.refs++;for(var i=0;i<r.parts.length;i++)r.parts[i](o.parts[i]);for(;i<o.parts.length;i++)r.parts.push(g(o.parts[i],t))}else{var m=[];for(i=0;i<o.parts.length;i++)m.push(g(o.parts[i],t));n[o.id]={id:o.id,refs:1,parts:m}}}}function c(a){for(var t=[],n={},e=0;e<a.length;e++){var o=a[e],r=o[0],i={css:o[1],media:o[2],sourceMap:o[3]};n[r]?n[r].parts.push(i):t.push(n[r]={id:r,parts:[i]})}return t}function d(a,t){var n=r(),e=l[l.length-1];if("top"===a.insertAt)e?e.nextSibling?n.insertBefore(t,e.nextSibling):n.appendChild(t):n.insertBefore(t,n.firstChild),l.push(t);else{if("bottom"!==a.insertAt)throw new Error("Invalid value for parameter 'insertAt'. Must be 'top' or 'bottom'.");n.appendChild(t)}}function s(a){a.parentNode.removeChild(a);var t=l.indexOf(a);t>=0&&l.splice(t,1)}function p(a){var t=document.createElement("style");return t.type="text/css",d(a,t),t}function g(a,t){var n,e,o,r,l;if(t.singleton){var b=m++;n=i||(i=p(t)),e=x.bind(null,n,b,!1),o=x.bind(null,n,b,!0)}else a.sourceMap&&"function"==typeof URL&&"function"==typeof URL.createObjectURL&&"function"==typeof URL.revokeObjectURL&&"function"==typeof Blob&&"function"==typeof btoa?(r=t,(l=document.createElement("link")).rel="stylesheet",d(r,l),e=function(a,t){var n=t.css,e=t.sourceMap;e&&(n+="\n/*# sourceMappingURL=data:application/json;base64,"+btoa(unescape(encodeURIComponent(JSON.stringify(e))))+" */");var o=new Blob([n],{type:"text/css"}),r=a.href;a.href=URL.createObjectURL(o),r&&URL.revokeObjectURL(r)}.bind(null,n=l),o=function(){s(n),n.href&&URL.revokeObjectURL(n.href)}):(n=p(t),e=function(a,t){var n=t.css,e=t.media;e&&a.setAttribute("media",e);if(a.styleSheet)a.styleSheet.cssText=n;else{for(;a.firstChild;)a.removeChild(a.firstChild);a.appendChild(document.createTextNode(n))}}.bind(null,n),o=function(){s(n)});return e(a),function(t){if(t){if(t.css===a.css&&t.media===a.media&&t.sourceMap===a.sourceMap)return;e(a=t)}else o()}}a.exports=function(a,t){if("undefined"!=typeof DEBUG&&DEBUG&&"object"!=typeof document)throw new Error("The style-loader cannot be used in a non-browser environment");void 0===(t=t||{}).singleton&&(t.singleton=o()),void 0===t.insertAt&&(t.insertAt="bottom");var e=c(a);return b(e,t),function(a){for(var o=[],r=0;r<e.length;r++){var i=e[r];(m=n[i.id]).refs--,o.push(m)}a&&b(c(a),t);for(r=0;r<o.length;r++){var m;if(0===(m=o[r]).refs){for(var l=0;l<m.parts.length;l++)m.parts[l]();delete n[m.id]}}}};var u,f=(u=[],function(a,t){return u[a]=t,u.filter(Boolean).join("\n")});function x(a,t,n,e){var o=n?"":e.css;if(a.styleSheet)a.styleSheet.cssText=f(t,o);else{var r=document.createTextNode(o),i=a.childNodes;i[t]&&a.removeChild(i[t]),i.length?a.insertBefore(r,i[t]):a.appendChild(r)}}},5:function(a,t,n){a.exports=n("../../../../../src/styles.scss")}},[5]);
</script><script type="text/javascript">
webpackJsonp(["vendor"],{"../../../../base64-js/index.js":function(t,e,n){"use strict";e.byteLength=function(t){return 3*t.length/4-u(t)},e.toByteArray=function(t){var e,n,r,s,a,c=t.length;s=u(t),a=new o(3*c/4-s),n=s>0?c-4:c;var l=0;for(e=0;e<n;e+=4)r=i[t.charCodeAt(e)]<<18|i[t.charCodeAt(e+1)]<<12|i[t.charCodeAt(e+2)]<<6|i[t.charCodeAt(e+3)],a[l++]=r>>16&255,a[l++]=r>>8&255,a[l++]=255&r;2===s?(r=i[t.charCodeAt(e)]<<2|i[t.charCodeAt(e+1)]>>4,a[l++]=255&r):1===s&&(r=i[t.charCodeAt(e)]<<10|i[t.charCodeAt(e+1)]<<4|i[t.charCodeAt(e+2)]>>2,a[l++]=r>>8&255,a[l++]=255&r);return a},e.fromByteArray=function(t){for(var e,n=t.length,i=n%3,o="",s=[],a=0,c=n-i;a<c;a+=16383)s.push(l(t,a,a+16383>c?c:a+16383));1===i?(e=t[n-1],o+=r[e>>2],o+=r[e<<4&63],o+="=="):2===i&&(e=(t[n-2]<<8)+t[n-1],o+=r[e>>10],o+=r[e>>4&63],o+=r[e<<2&63],o+="=");return s.push(o),s.join("")};for(var r=[],i=[],o="undefined"!=typeof Uint8Array?Uint8Array:Array,s="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/",a=0,c=s.length;a<c;++a)r[a]=s[a],i[s.charCodeAt(a)]=a;function u(t){var e=t.length;if(e%4>0)throw new Error("Invalid string. Length must be a multiple of 4");return"="===t[e-2]?2:"="===t[e-1]?1:0}function l(t,e,n){for(var i,o,s=[],a=e;a<n;a+=3)i=(t[a]<<16)+(t[a+1]<<8)+t[a+2],s.push(r[(o=i)>>18&63]+r[o>>12&63]+r[o>>6&63]+r[63&o]);return s.join("")}i["-".charCodeAt(0)]=62,i["_".charCodeAt(0)]=63},"../../../../buffer/index.js":function(t,e,n){"use strict";(function(t){var r=n("../../../../base64-js/index.js"),i=n("../../../../ieee754/index.js"),o=n("../../../../isarray/index.js");function s(){return c.TYPED_ARRAY_SUPPORT?2147483647:1073741823}function a(t,e){if(s()<e)throw new RangeError("Invalid typed array length");return c.TYPED_ARRAY_SUPPORT?(t=new Uint8Array(e)).__proto__=c.prototype:(null===t&&(t=new c(e)),t.length=e),t}function c(t,e,n){if(!(c.TYPED_ARRAY_SUPPORT||this instanceof c))return new c(t,e,n);if("number"==typeof t){if("string"==typeof e)throw new Error("If encoding is specified then the first argument must be a string");return p(this,t)}return u(this,t,e,n)}function u(t,e,n,r){if("number"==typeof e)throw new TypeError('"value" argument must not be a number');return"undefined"!=typeof ArrayBuffer&&e instanceof ArrayBuffer?function(t,e,n,r){if(e.byteLength,n<0||e.byteLength<n)throw new RangeError("'offset' is out of bounds");if(e.byteLength<n+(r||0))throw new RangeError("'length' is out of bounds");e=void 0===n&&void 0===r?new Uint8Array(e):void 0===r?new Uint8Array(e,n):new Uint8Array(e,n,r);c.TYPED_ARRAY_SUPPORT?(t=e).__proto__=c.prototype:t=d(t,e);return t}(t,e,n,r):"string"==typeof e?function(t,e,n){"string"==typeof n&&""!==n||(n="utf8");if(!c.isEncoding(n))throw new TypeError('"encoding" must be a valid string encoding');var r=0|f(e,n),i=(t=a(t,r)).write(e,n);i!==r&&(t=t.slice(0,i));return t}(t,e,n):function(t,e){if(c.isBuffer(e)){var n=0|h(e.length);return 0===(t=a(t,n)).length?t:(e.copy(t,0,0,n),t)}if(e){if("undefined"!=typeof ArrayBuffer&&e.buffer instanceof ArrayBuffer||"length"in e)return"number"!=typeof e.length||(r=e.length)!=r?a(t,0):d(t,e);if("Buffer"===e.type&&o(e.data))return d(t,e.data)}var r;throw new TypeError("First argument must be a string, Buffer, ArrayBuffer, Array, or array-like object.")}(t,e)}function l(t){if("number"!=typeof t)throw new TypeError('"size" argument must be a number');if(t<0)throw new RangeError('"size" argument must not be negative')}function p(t,e){if(l(e),t=a(t,e<0?0:0|h(e)),!c.TYPED_ARRAY_SUPPORT)for(var n=0;n<e;++n)t[n]=0;return t}function d(t,e){var n=e.length<0?0:0|h(e.length);t=a(t,n);for(var r=0;r<n;r+=1)t[r]=255&e[r];return t}function h(t){if(t>=s())throw new RangeError("Attempt to allocate Buffer larger than maximum size: 0x"+s().toString(16)+" bytes");return 0|t}function f(t,e){if(c.isBuffer(t))return t.length;if("undefined"!=typeof ArrayBuffer&&"function"==typeof ArrayBuffer.isView&&(ArrayBuffer.isView(t)||t instanceof ArrayBuffer))return t.byteLength;"string"!=typeof t&&(t=""+t);var n=t.length;if(0===n)return 0;for(var r=!1;;)switch(e){case"ascii":case"latin1":case"binary":return n;case"utf8":case"utf-8":case void 0:return N(t).length;case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return 2*n;case"hex":return n>>>1;case"base64":return L(t).length;default:if(r)return N(t).length;e=(""+e).toLowerCase(),r=!0}}function m(t,e,n){var r=t[e];t[e]=t[n],t[n]=r}function g(t,e,n,r,i){if(0===t.length)return-1;if("string"==typeof n?(r=n,n=0):n>2147483647?n=2147483647:n<-2147483648&&(n=-2147483648),n=+n,isNaN(n)&&(n=i?0:t.length-1),n<0&&(n=t.length+n),n>=t.length){if(i)return-1;n=t.length-1}else if(n<0){if(!i)return-1;n=0}if("string"==typeof e&&(e=c.from(e,r)),c.isBuffer(e))return 0===e.length?-1:v(t,e,n,r,i);if("number"==typeof e)return e&=255,c.TYPED_ARRAY_SUPPORT&&"function"==typeof Uint8Array.prototype.indexOf?i?Uint8Array.prototype.indexOf.call(t,e,n):Uint8Array.prototype.lastIndexOf.call(t,e,n):v(t,[e],n,r,i);throw new TypeError("val must be string, number or Buffer")}function v(t,e,n,r,i){var o,s=1,a=t.length,c=e.length;if(void 0!==r&&("ucs2"===(r=String(r).toLowerCase())||"ucs-2"===r||"utf16le"===r||"utf-16le"===r)){if(t.length<2||e.length<2)return-1;s=2,a/=2,c/=2,n/=2}function u(t,e){return 1===s?t[e]:t.readUInt16BE(e*s)}if(i){var l=-1;for(o=n;o<a;o++)if(u(t,o)===u(e,-1===l?0:o-l)){if(-1===l&&(l=o),o-l+1===c)return l*s}else-1!==l&&(o-=o-l),l=-1}else for(n+c>a&&(n=a-c),o=n;o>=0;o--){for(var p=!0,d=0;d<c;d++)if(u(t,o+d)!==u(e,d)){p=!1;break}if(p)return o}return-1}function y(t,e,n,r){n=Number(n)||0;var i=t.length-n;r?(r=Number(r))>i&&(r=i):r=i;var o=e.length;if(o%2!=0)throw new TypeError("Invalid hex string");r>o/2&&(r=o/2);for(var s=0;s<r;++s){var a=parseInt(e.substr(2*s,2),16);if(isNaN(a))return s;t[n+s]=a}return s}function b(t,e,n,r){return F(function(t){for(var e=[],n=0;n<t.length;++n)e.push(255&t.charCodeAt(n));return e}(e),t,n,r)}function _(t,e,n){return 0===e&&n===t.length?r.fromByteArray(t):r.fromByteArray(t.slice(e,n))}function j(t,e,n){n=Math.min(t.length,n);for(var r=[],i=e;i<n;){var o,s,a,c,u=t[i],l=null,p=u>239?4:u>223?3:u>191?2:1;if(i+p<=n)switch(p){case 1:u<128&&(l=u);break;case 2:128==(192&(o=t[i+1]))&&(c=(31&u)<<6|63&o)>127&&(l=c);break;case 3:o=t[i+1],s=t[i+2],128==(192&o)&&128==(192&s)&&(c=(15&u)<<12|(63&o)<<6|63&s)>2047&&(c<55296||c>57343)&&(l=c);break;case 4:o=t[i+1],s=t[i+2],a=t[i+3],128==(192&o)&&128==(192&s)&&128==(192&a)&&(c=(15&u)<<18|(63&o)<<12|(63&s)<<6|63&a)>65535&&c<1114112&&(l=c)}null===l?(l=65533,p=1):l>65535&&(l-=65536,r.push(l>>>10&1023|55296),l=56320|1023&l),r.push(l),i+=p}return function(t){var e=t.length;if(e<=x)return String.fromCharCode.apply(String,t);var n="",r=0;for(;r<e;)n+=String.fromCharCode.apply(String,t.slice(r,r+=x));return n}(r)}e.Buffer=c,e.SlowBuffer=function(t){+t!=t&&(t=0);return c.alloc(+t)},e.INSPECT_MAX_BYTES=50,c.TYPED_ARRAY_SUPPORT=void 0!==t.TYPED_ARRAY_SUPPORT?t.TYPED_ARRAY_SUPPORT:function(){try{var t=new Uint8Array(1);return t.__proto__={__proto__:Uint8Array.prototype,foo:function(){return 42}},42===t.foo()&&"function"==typeof t.subarray&&0===t.subarray(1,1).byteLength}catch(t){return!1}}(),e.kMaxLength=s(),c.poolSize=8192,c._augment=function(t){return t.__proto__=c.prototype,t},c.from=function(t,e,n){return u(null,t,e,n)},c.TYPED_ARRAY_SUPPORT&&(c.prototype.__proto__=Uint8Array.prototype,c.__proto__=Uint8Array,"undefined"!=typeof Symbol&&Symbol.species&&c[Symbol.species]===c&&Object.defineProperty(c,Symbol.species,{value:null,configurable:!0})),c.alloc=function(t,e,n){return r=null,o=e,s=n,l(i=t),i<=0?a(r,i):void 0!==o?"string"==typeof s?a(r,i).fill(o,s):a(r,i).fill(o):a(r,i);var r,i,o,s},c.allocUnsafe=function(t){return p(null,t)},c.allocUnsafeSlow=function(t){return p(null,t)},c.isBuffer=function(t){return!(null==t||!t._isBuffer)},c.compare=function(t,e){if(!c.isBuffer(t)||!c.isBuffer(e))throw new TypeError("Arguments must be Buffers");if(t===e)return 0;for(var n=t.length,r=e.length,i=0,o=Math.min(n,r);i<o;++i)if(t[i]!==e[i]){n=t[i],r=e[i];break}return n<r?-1:r<n?1:0},c.isEncoding=function(t){switch(String(t).toLowerCase()){case"hex":case"utf8":case"utf-8":case"ascii":case"latin1":case"binary":case"base64":case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return!0;default:return!1}},c.concat=function(t,e){if(!o(t))throw new TypeError('"list" argument must be an Array of Buffers');if(0===t.length)return c.alloc(0);var n;if(void 0===e)for(e=0,n=0;n<t.length;++n)e+=t[n].length;var r=c.allocUnsafe(e),i=0;for(n=0;n<t.length;++n){var s=t[n];if(!c.isBuffer(s))throw new TypeError('"list" argument must be an Array of Buffers');s.copy(r,i),i+=s.length}return r},c.byteLength=f,c.prototype._isBuffer=!0,c.prototype.swap16=function(){var t=this.length;if(t%2!=0)throw new RangeError("Buffer size must be a multiple of 16-bits");for(var e=0;e<t;e+=2)m(this,e,e+1);return this},c.prototype.swap32=function(){var t=this.length;if(t%4!=0)throw new RangeError("Buffer size must be a multiple of 32-bits");for(var e=0;e<t;e+=4)m(this,e,e+3),m(this,e+1,e+2);return this},c.prototype.swap64=function(){var t=this.length;if(t%8!=0)throw new RangeError("Buffer size must be a multiple of 64-bits");for(var e=0;e<t;e+=8)m(this,e,e+7),m(this,e+1,e+6),m(this,e+2,e+5),m(this,e+3,e+4);return this},c.prototype.toString=function(){var t=0|this.length;return 0===t?"":0===arguments.length?j(this,0,t):function(t,e,n){var r=!1;if((void 0===e||e<0)&&(e=0),e>this.length)return"";if((void 0===n||n>this.length)&&(n=this.length),n<=0)return"";if((n>>>=0)<=(e>>>=0))return"";for(t||(t="utf8");;)switch(t){case"hex":return C(this,e,n);case"utf8":case"utf-8":return j(this,e,n);case"ascii":return w(this,e,n);case"latin1":case"binary":return O(this,e,n);case"base64":return _(this,e,n);case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return k(this,e,n);default:if(r)throw new TypeError("Unknown encoding: "+t);t=(t+"").toLowerCase(),r=!0}}.apply(this,arguments)},c.prototype.equals=function(t){if(!c.isBuffer(t))throw new TypeError("Argument must be a Buffer");return this===t||0===c.compare(this,t)},c.prototype.inspect=function(){var t="",n=e.INSPECT_MAX_BYTES;return this.length>0&&(t=this.toString("hex",0,n).match(/.{2}/g).join(" "),this.length>n&&(t+=" ... ")),"<Buffer "+t+">"},c.prototype.compare=function(t,e,n,r,i){if(!c.isBuffer(t))throw new TypeError("Argument must be a Buffer");if(void 0===e&&(e=0),void 0===n&&(n=t?t.length:0),void 0===r&&(r=0),void 0===i&&(i=this.length),e<0||n>t.length||r<0||i>this.length)throw new RangeError("out of range index");if(r>=i&&e>=n)return 0;if(r>=i)return-1;if(e>=n)return 1;if(e>>>=0,n>>>=0,r>>>=0,i>>>=0,this===t)return 0;for(var o=i-r,s=n-e,a=Math.min(o,s),u=this.slice(r,i),l=t.slice(e,n),p=0;p<a;++p)if(u[p]!==l[p]){o=u[p],s=l[p];break}return o<s?-1:s<o?1:0},c.prototype.includes=function(t,e,n){return-1!==this.indexOf(t,e,n)},c.prototype.indexOf=function(t,e,n){return g(this,t,e,n,!0)},c.prototype.lastIndexOf=function(t,e,n){return g(this,t,e,n,!1)},c.prototype.write=function(t,e,n,r){if(void 0===e)r="utf8",n=this.length,e=0;else if(void 0===n&&"string"==typeof e)r=e,n=this.length,e=0;else{if(!isFinite(e))throw new Error("Buffer.write(string, encoding, offset[, length]) is no longer supported");e|=0,isFinite(n)?(n|=0,void 0===r&&(r="utf8")):(r=n,n=void 0)}var i=this.length-e;if((void 0===n||n>i)&&(n=i),t.length>0&&(n<0||e<0)||e>this.length)throw new RangeError("Attempt to write outside buffer bounds");r||(r="utf8");for(var o,s,a,c,u,l,p,d,h,f=!1;;)switch(r){case"hex":return y(this,t,e,n);case"utf8":case"utf-8":return d=e,h=n,F(N(t,(p=this).length-d),p,d,h);case"ascii":return b(this,t,e,n);case"latin1":case"binary":return b(this,t,e,n);case"base64":return c=this,u=e,l=n,F(L(t),c,u,l);case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return s=e,a=n,F(function(t,e){for(var n,r,i,o=[],s=0;s<t.length&&!((e-=2)<0);++s)n=t.charCodeAt(s),r=n>>8,i=n%256,o.push(i),o.push(r);return o}(t,(o=this).length-s),o,s,a);default:if(f)throw new TypeError("Unknown encoding: "+r);r=(""+r).toLowerCase(),f=!0}},c.prototype.toJSON=function(){return{type:"Buffer",data:Array.prototype.slice.call(this._arr||this,0)}};var x=4096;function w(t,e,n){var r="";n=Math.min(t.length,n);for(var i=e;i<n;++i)r+=String.fromCharCode(127&t[i]);return r}function O(t,e,n){var r="";n=Math.min(t.length,n);for(var i=e;i<n;++i)r+=String.fromCharCode(t[i]);return r}function C(t,e,n){var r=t.length;(!e||e<0)&&(e=0),(!n||n<0||n>r)&&(n=r);for(var i="",o=e;o<n;++o)i+=R(t[o]);return i}function k(t,e,n){for(var r=t.slice(e,n),i="",o=0;o<r.length;o+=2)i+=String.fromCharCode(r[o]+256*r[o+1]);return i}function E(t,e,n){if(t%1!=0||t<0)throw new RangeError("offset is not uint");if(t+e>n)throw new RangeError("Trying to access beyond buffer length")}function S(t,e,n,r,i,o){if(!c.isBuffer(t))throw new TypeError('"buffer" argument must be a Buffer instance');if(e>i||e<o)throw new RangeError('"value" argument is out of bounds');if(n+r>t.length)throw new RangeError("Index out of range")}function P(t,e,n,r){e<0&&(e=65535+e+1);for(var i=0,o=Math.min(t.length-n,2);i<o;++i)t[n+i]=(e&255<<8*(r?i:1-i))>>>8*(r?i:1-i)}function M(t,e,n,r){e<0&&(e=4294967295+e+1);for(var i=0,o=Math.min(t.length-n,4);i<o;++i)t[n+i]=e>>>8*(r?i:3-i)&255}function A(t,e,n,r,i,o){if(n+r>t.length)throw new RangeError("Index out of range");if(n<0)throw new RangeError("Index out of range")}function T(t,e,n,r,o){return o||A(t,0,n,4),i.write(t,e,n,r,23,4),n+4}function D(t,e,n,r,o){return o||A(t,0,n,8),i.write(t,e,n,r,52,8),n+8}c.prototype.slice=function(t,e){var n,r=this.length;if(t=~~t,e=void 0===e?r:~~e,t<0?(t+=r)<0&&(t=0):t>r&&(t=r),e<0?(e+=r)<0&&(e=0):e>r&&(e=r),e<t&&(e=t),c.TYPED_ARRAY_SUPPORT)(n=this.subarray(t,e)).__proto__=c.prototype;else{var i=e-t;n=new c(i,void 0);for(var o=0;o<i;++o)n[o]=this[o+t]}return n},c.prototype.readUIntLE=function(t,e,n){t|=0,e|=0,n||E(t,e,this.length);for(var r=this[t],i=1,o=0;++o<e&&(i*=256);)r+=this[t+o]*i;return r},c.prototype.readUIntBE=function(t,e,n){t|=0,e|=0,n||E(t,e,this.length);for(var r=this[t+--e],i=1;e>0&&(i*=256);)r+=this[t+--e]*i;return r},c.prototype.readUInt8=function(t,e){return e||E(t,1,this.length),this[t]},c.prototype.readUInt16LE=function(t,e){return e||E(t,2,this.length),this[t]|this[t+1]<<8},c.prototype.readUInt16BE=function(t,e){return e||E(t,2,this.length),this[t]<<8|this[t+1]},c.prototype.readUInt32LE=function(t,e){return e||E(t,4,this.length),(this[t]|this[t+1]<<8|this[t+2]<<16)+16777216*this[t+3]},c.prototype.readUInt32BE=function(t,e){return e||E(t,4,this.length),16777216*this[t]+(this[t+1]<<16|this[t+2]<<8|this[t+3])},c.prototype.readIntLE=function(t,e,n){t|=0,e|=0,n||E(t,e,this.length);for(var r=this[t],i=1,o=0;++o<e&&(i*=256);)r+=this[t+o]*i;return r>=(i*=128)&&(r-=Math.pow(2,8*e)),r},c.prototype.readIntBE=function(t,e,n){t|=0,e|=0,n||E(t,e,this.length);for(var r=e,i=1,o=this[t+--r];r>0&&(i*=256);)o+=this[t+--r]*i;return o>=(i*=128)&&(o-=Math.pow(2,8*e)),o},c.prototype.readInt8=function(t,e){return e||E(t,1,this.length),128&this[t]?-1*(255-this[t]+1):this[t]},c.prototype.readInt16LE=function(t,e){e||E(t,2,this.length);var n=this[t]|this[t+1]<<8;return 32768&n?4294901760|n:n},c.prototype.readInt16BE=function(t,e){e||E(t,2,this.length);var n=this[t+1]|this[t]<<8;return 32768&n?4294901760|n:n},c.prototype.readInt32LE=function(t,e){return e||E(t,4,this.length),this[t]|this[t+1]<<8|this[t+2]<<16|this[t+3]<<24},c.prototype.readInt32BE=function(t,e){return e||E(t,4,this.length),this[t]<<24|this[t+1]<<16|this[t+2]<<8|this[t+3]},c.prototype.readFloatLE=function(t,e){return e||E(t,4,this.length),i.read(this,t,!0,23,4)},c.prototype.readFloatBE=function(t,e){return e||E(t,4,this.length),i.read(this,t,!1,23,4)},c.prototype.readDoubleLE=function(t,e){return e||E(t,8,this.length),i.read(this,t,!0,52,8)},c.prototype.readDoubleBE=function(t,e){return e||E(t,8,this.length),i.read(this,t,!1,52,8)},c.prototype.writeUIntLE=function(t,e,n,r){(t=+t,e|=0,n|=0,r)||S(this,t,e,n,Math.pow(2,8*n)-1,0);var i=1,o=0;for(this[e]=255&t;++o<n&&(i*=256);)this[e+o]=t/i&255;return e+n},c.prototype.writeUIntBE=function(t,e,n,r){(t=+t,e|=0,n|=0,r)||S(this,t,e,n,Math.pow(2,8*n)-1,0);var i=n-1,o=1;for(this[e+i]=255&t;--i>=0&&(o*=256);)this[e+i]=t/o&255;return e+n},c.prototype.writeUInt8=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,1,255,0),c.TYPED_ARRAY_SUPPORT||(t=Math.floor(t)),this[e]=255&t,e+1},c.prototype.writeUInt16LE=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,2,65535,0),c.TYPED_ARRAY_SUPPORT?(this[e]=255&t,this[e+1]=t>>>8):P(this,t,e,!0),e+2},c.prototype.writeUInt16BE=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,2,65535,0),c.TYPED_ARRAY_SUPPORT?(this[e]=t>>>8,this[e+1]=255&t):P(this,t,e,!1),e+2},c.prototype.writeUInt32LE=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,4,4294967295,0),c.TYPED_ARRAY_SUPPORT?(this[e+3]=t>>>24,this[e+2]=t>>>16,this[e+1]=t>>>8,this[e]=255&t):M(this,t,e,!0),e+4},c.prototype.writeUInt32BE=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,4,4294967295,0),c.TYPED_ARRAY_SUPPORT?(this[e]=t>>>24,this[e+1]=t>>>16,this[e+2]=t>>>8,this[e+3]=255&t):M(this,t,e,!1),e+4},c.prototype.writeIntLE=function(t,e,n,r){if(t=+t,e|=0,!r){var i=Math.pow(2,8*n-1);S(this,t,e,n,i-1,-i)}var o=0,s=1,a=0;for(this[e]=255&t;++o<n&&(s*=256);)t<0&&0===a&&0!==this[e+o-1]&&(a=1),this[e+o]=(t/s>>0)-a&255;return e+n},c.prototype.writeIntBE=function(t,e,n,r){if(t=+t,e|=0,!r){var i=Math.pow(2,8*n-1);S(this,t,e,n,i-1,-i)}var o=n-1,s=1,a=0;for(this[e+o]=255&t;--o>=0&&(s*=256);)t<0&&0===a&&0!==this[e+o+1]&&(a=1),this[e+o]=(t/s>>0)-a&255;return e+n},c.prototype.writeInt8=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,1,127,-128),c.TYPED_ARRAY_SUPPORT||(t=Math.floor(t)),t<0&&(t=255+t+1),this[e]=255&t,e+1},c.prototype.writeInt16LE=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,2,32767,-32768),c.TYPED_ARRAY_SUPPORT?(this[e]=255&t,this[e+1]=t>>>8):P(this,t,e,!0),e+2},c.prototype.writeInt16BE=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,2,32767,-32768),c.TYPED_ARRAY_SUPPORT?(this[e]=t>>>8,this[e+1]=255&t):P(this,t,e,!1),e+2},c.prototype.writeInt32LE=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,4,2147483647,-2147483648),c.TYPED_ARRAY_SUPPORT?(this[e]=255&t,this[e+1]=t>>>8,this[e+2]=t>>>16,this[e+3]=t>>>24):M(this,t,e,!0),e+4},c.prototype.writeInt32BE=function(t,e,n){return t=+t,e|=0,n||S(this,t,e,4,2147483647,-2147483648),t<0&&(t=4294967295+t+1),c.TYPED_ARRAY_SUPPORT?(this[e]=t>>>24,this[e+1]=t>>>16,this[e+2]=t>>>8,this[e+3]=255&t):M(this,t,e,!1),e+4},c.prototype.writeFloatLE=function(t,e,n){return T(this,t,e,!0,n)},c.prototype.writeFloatBE=function(t,e,n){return T(this,t,e,!1,n)},c.prototype.writeDoubleLE=function(t,e,n){return D(this,t,e,!0,n)},c.prototype.writeDoubleBE=function(t,e,n){return D(this,t,e,!1,n)},c.prototype.copy=function(t,e,n,r){if(n||(n=0),r||0===r||(r=this.length),e>=t.length&&(e=t.length),e||(e=0),r>0&&r<n&&(r=n),r===n)return 0;if(0===t.length||0===this.length)return 0;if(e<0)throw new RangeError("targetStart out of bounds");if(n<0||n>=this.length)throw new RangeError("sourceStart out of bounds");if(r<0)throw new RangeError("sourceEnd out of bounds");r>this.length&&(r=this.length),t.length-e<r-n&&(r=t.length-e+n);var i,o=r-n;if(this===t&&n<e&&e<r)for(i=o-1;i>=0;--i)t[i+e]=this[i+n];else if(o<1e3||!c.TYPED_ARRAY_SUPPORT)for(i=0;i<o;++i)t[i+e]=this[i+n];else Uint8Array.prototype.set.call(t,this.subarray(n,n+o),e);return o},c.prototype.fill=function(t,e,n,r){if("string"==typeof t){if("string"==typeof e?(r=e,e=0,n=this.length):"string"==typeof n&&(r=n,n=this.length),1===t.length){var i=t.charCodeAt(0);i<256&&(t=i)}if(void 0!==r&&"string"!=typeof r)throw new TypeError("encoding must be a string");if("string"==typeof r&&!c.isEncoding(r))throw new TypeError("Unknown encoding: "+r)}else"number"==typeof t&&(t&=255);if(e<0||this.length<e||this.length<n)throw new RangeError("Out of range index");if(n<=e)return this;var o;if(e>>>=0,n=void 0===n?this.length:n>>>0,t||(t=0),"number"==typeof t)for(o=e;o<n;++o)this[o]=t;else{var s=c.isBuffer(t)?t:N(new c(t,r).toString()),a=s.length;for(o=0;o<n-e;++o)this[o+e]=s[o%a]}return this};var I=/[^+\/0-9A-Za-z-_]/g;function R(t){return t<16?"0"+t.toString(16):t.toString(16)}function N(t,e){var n;e=e||1/0;for(var r=t.length,i=null,o=[],s=0;s<r;++s){if((n=t.charCodeAt(s))>55295&&n<57344){if(!i){if(n>56319){(e-=3)>-1&&o.push(239,191,189);continue}if(s+1===r){(e-=3)>-1&&o.push(239,191,189);continue}i=n;continue}if(n<56320){(e-=3)>-1&&o.push(239,191,189),i=n;continue}n=65536+(i-55296<<10|n-56320)}else i&&(e-=3)>-1&&o.push(239,191,189);if(i=null,n<128){if((e-=1)<0)break;o.push(n)}else if(n<2048){if((e-=2)<0)break;o.push(n>>6|192,63&n|128)}else if(n<65536){if((e-=3)<0)break;o.push(n>>12|224,n>>6&63|128,63&n|128)}else{if(!(n<1114112))throw new Error("Invalid code point");if((e-=4)<0)break;o.push(n>>18|240,n>>12&63|128,n>>6&63|128,63&n|128)}}return o}function L(t){return r.toByteArray(function(t){var e;if((t=(e=t,e.trim?e.trim():e.replace(/^\s+|\s+$/g,"")).replace(I,"")).length<2)return"";for(;t.length%4!=0;)t+="=";return t}(t))}function F(t,e,n,r){for(var i=0;i<r&&!(i+n>=e.length||i>=t.length);++i)e[i+n]=t[i];return i}}).call(e,n("../../../../webpack/buildin/global.js"))},"../../../../css-loader/lib/css-base.js":function(t,e){t.exports=function(t){var e=[];return e.toString=function(){return this.map(function(e){var n=function(t,e){var n=t[1]||"",r=t[3];if(!r)return n;if(e&&"function"==typeof btoa){var i=(s=r,"/*# sourceMappingURL=data:application/json;charset=utf-8;base64,"+btoa(unescape(encodeURIComponent(JSON.stringify(s))))+" */"),o=r.sources.map(function(t){return"/*# sourceURL="+r.sourceRoot+t+" */"});return[n].concat(o).concat([i]).join("\n")}var s;return[n].join("\n")}(e,t);return e[2]?"@media "+e[2]+"{"+n+"}":n}).join("")},e.i=function(t,n){"string"==typeof t&&(t=[[null,t,""]]);for(var r={},i=0;i<this.length;i++){var o=this[i][0];"number"==typeof o&&(r[o]=!0)}for(i=0;i<t.length;i++){var s=t[i];"number"==typeof s[0]&&r[s[0]]||(n&&!s[2]?s[2]=n:n&&(s[2]="("+s[2]+") and ("+n+")"),e.push(s))}},e}},"../../../../d3-array/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/src/bisect.js");n.d(e,"b",function(){return r.c}),n.d(e,"d",function(){return r.b}),n.d(e,"c",function(){return r.a});var i=n("../../../../d3-array/src/ascending.js");n.d(e,"a",function(){return i.a});var o=n("../../../../d3-array/src/bisector.js");n.d(e,"e",function(){return o.a});n("../../../../d3-array/src/cross.js"),n("../../../../d3-array/src/descending.js"),n("../../../../d3-array/src/deviation.js");var s=n("../../../../d3-array/src/extent.js");n.d(e,"f",function(){return s.a});n("../../../../d3-array/src/histogram.js"),n("../../../../d3-array/src/threshold/freedmanDiaconis.js"),n("../../../../d3-array/src/threshold/scott.js");var a=n("../../../../d3-array/src/threshold/sturges.js");n.d(e,"p",function(){return a.a});var c=n("../../../../d3-array/src/max.js");n.d(e,"g",function(){return c.a});var u=n("../../../../d3-array/src/mean.js");n.d(e,"h",function(){return u.a});var l=n("../../../../d3-array/src/median.js");n.d(e,"i",function(){return l.a});var p=n("../../../../d3-array/src/merge.js");n.d(e,"j",function(){return p.a});var d=n("../../../../d3-array/src/min.js");n.d(e,"k",function(){return d.a});n("../../../../d3-array/src/pairs.js");var h=n("../../../../d3-array/src/permute.js");n.d(e,"l",function(){return h.a});var f=n("../../../../d3-array/src/quantile.js");n.d(e,"m",function(){return f.a});var m=n("../../../../d3-array/src/range.js");n.d(e,"n",function(){return m.a});n("../../../../d3-array/src/scan.js"),n("../../../../d3-array/src/shuffle.js");var g=n("../../../../d3-array/src/sum.js");n.d(e,"o",function(){return g.a});var v=n("../../../../d3-array/src/ticks.js");n.d(e,"s",function(){return v.a}),n.d(e,"q",function(){return v.b}),n.d(e,"r",function(){return v.c});n("../../../../d3-array/src/transpose.js");var y=n("../../../../d3-array/src/variance.js");n.d(e,"t",function(){return y.a});n("../../../../d3-array/src/zip.js")},"../../../../d3-array/src/array.js":function(t,e,n){"use strict";n.d(e,"b",function(){return i}),n.d(e,"a",function(){return o});var r=Array.prototype,i=r.slice,o=r.map},"../../../../d3-array/src/ascending.js":function(t,e,n){"use strict";e.a=function(t,e){return t<e?-1:t>e?1:t>=e?0:NaN}},"../../../../d3-array/src/bisect.js":function(t,e,n){"use strict";n.d(e,"b",function(){return s}),n.d(e,"a",function(){return a});var r=n("../../../../d3-array/src/ascending.js"),i=n("../../../../d3-array/src/bisector.js"),o=Object(i.a)(r.a),s=o.right,a=o.left;e.c=s},"../../../../d3-array/src/bisector.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/src/ascending.js");e.a=function(t){var e;return 1===t.length&&(e=t,t=function(t,n){return Object(r.a)(e(t),n)}),{left:function(e,n,r,i){for(null==r&&(r=0),null==i&&(i=e.length);r<i;){var o=r+i>>>1;t(e[o],n)<0?r=o+1:i=o}return r},right:function(e,n,r,i){for(null==r&&(r=0),null==i&&(i=e.length);r<i;){var o=r+i>>>1;t(e[o],n)>0?i=o:r=o+1}return r}}}},"../../../../d3-array/src/constant.js":function(t,e,n){"use strict";e.a=function(t){return function(){return t}}},"../../../../d3-array/src/cross.js":function(t,e,n){"use strict";n("../../../../d3-array/src/pairs.js")},"../../../../d3-array/src/descending.js":function(t,e,n){"use strict"},"../../../../d3-array/src/deviation.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/src/variance.js");e.a=function(t,e){var n=Object(r.a)(t,e);return n?Math.sqrt(n):n}},"../../../../d3-array/src/extent.js":function(t,e,n){"use strict";e.a=function(t,e){var n,r,i,o=t.length,s=-1;if(null==e){for(;++s<o;)if(null!=(n=t[s])&&n>=n)for(r=i=n;++s<o;)null!=(n=t[s])&&(r>n&&(r=n),i<n&&(i=n))}else for(;++s<o;)if(null!=(n=e(t[s],s,t))&&n>=n)for(r=i=n;++s<o;)null!=(n=e(t[s],s,t))&&(r>n&&(r=n),i<n&&(i=n));return[r,i]}},"../../../../d3-array/src/histogram.js":function(t,e,n){"use strict";n("../../../../d3-array/src/array.js"),n("../../../../d3-array/src/bisect.js"),n("../../../../d3-array/src/constant.js"),n("../../../../d3-array/src/extent.js"),n("../../../../d3-array/src/identity.js"),n("../../../../d3-array/src/range.js"),n("../../../../d3-array/src/ticks.js"),n("../../../../d3-array/src/threshold/sturges.js")},"../../../../d3-array/src/identity.js":function(t,e,n){"use strict";e.a=function(t){return t}},"../../../../d3-array/src/max.js":function(t,e,n){"use strict";e.a=function(t,e){var n,r,i=t.length,o=-1;if(null==e){for(;++o<i;)if(null!=(n=t[o])&&n>=n)for(r=n;++o<i;)null!=(n=t[o])&&n>r&&(r=n)}else for(;++o<i;)if(null!=(n=e(t[o],o,t))&&n>=n)for(r=n;++o<i;)null!=(n=e(t[o],o,t))&&n>r&&(r=n);return r}},"../../../../d3-array/src/mean.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/src/number.js");e.a=function(t,e){var n,i=t.length,o=i,s=-1,a=0;if(null==e)for(;++s<i;)isNaN(n=Object(r.a)(t[s]))?--o:a+=n;else for(;++s<i;)isNaN(n=Object(r.a)(e(t[s],s,t)))?--o:a+=n;if(o)return a/o}},"../../../../d3-array/src/median.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/src/ascending.js"),i=n("../../../../d3-array/src/number.js"),o=n("../../../../d3-array/src/quantile.js");e.a=function(t,e){var n,s=t.length,a=-1,c=[];if(null==e)for(;++a<s;)isNaN(n=Object(i.a)(t[a]))||c.push(n);else for(;++a<s;)isNaN(n=Object(i.a)(e(t[a],a,t)))||c.push(n);return Object(o.a)(c.sort(r.a),.5)}},"../../../../d3-array/src/merge.js":function(t,e,n){"use strict";e.a=function(t){for(var e,n,r,i=t.length,o=-1,s=0;++o<i;)s+=t[o].length;for(n=new Array(s);--i>=0;)for(e=(r=t[i]).length;--e>=0;)n[--s]=r[e];return n}},"../../../../d3-array/src/min.js":function(t,e,n){"use strict";e.a=function(t,e){var n,r,i=t.length,o=-1;if(null==e){for(;++o<i;)if(null!=(n=t[o])&&n>=n)for(r=n;++o<i;)null!=(n=t[o])&&r>n&&(r=n)}else for(;++o<i;)if(null!=(n=e(t[o],o,t))&&n>=n)for(r=n;++o<i;)null!=(n=e(t[o],o,t))&&r>n&&(r=n);return r}},"../../../../d3-array/src/number.js":function(t,e,n){"use strict";e.a=function(t){return null===t?NaN:+t}},"../../../../d3-array/src/pairs.js":function(t,e,n){"use strict";e.a=r;function r(t,e){return[t,e]}},"../../../../d3-array/src/permute.js":function(t,e,n){"use strict";e.a=function(t,e){for(var n=e.length,r=new Array(n);n--;)r[n]=t[e[n]];return r}},"../../../../d3-array/src/quantile.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/src/number.js");e.a=function(t,e,n){if(null==n&&(n=r.a),i=t.length){if((e=+e)<=0||i<2)return+n(t[0],0,t);if(e>=1)return+n(t[i-1],i-1,t);var i,o=(i-1)*e,s=Math.floor(o),a=+n(t[s],s,t);return a+(+n(t[s+1],s+1,t)-a)*(o-s)}}},"../../../../d3-array/src/range.js":function(t,e,n){"use strict";e.a=function(t,e,n){t=+t,e=+e,n=(i=arguments.length)<2?(e=t,t=0,1):i<3?1:+n;for(var r=-1,i=0|Math.max(0,Math.ceil((e-t)/n)),o=new Array(i);++r<i;)o[r]=t+r*n;return o}},"../../../../d3-array/src/scan.js":function(t,e,n){"use strict";n("../../../../d3-array/src/ascending.js")},"../../../../d3-array/src/shuffle.js":function(t,e,n){"use strict"},"../../../../d3-array/src/sum.js":function(t,e,n){"use strict";e.a=function(t,e){var n,r=t.length,i=-1,o=0;if(null==e)for(;++i<r;)(n=+t[i])&&(o+=n);else for(;++i<r;)(n=+e(t[i],i,t))&&(o+=n);return o}},"../../../../d3-array/src/threshold/freedmanDiaconis.js":function(t,e,n){"use strict";n("../../../../d3-array/src/array.js"),n("../../../../d3-array/src/ascending.js"),n("../../../../d3-array/src/number.js"),n("../../../../d3-array/src/quantile.js")},"../../../../d3-array/src/threshold/scott.js":function(t,e,n){"use strict";n("../../../../d3-array/src/deviation.js")},"../../../../d3-array/src/threshold/sturges.js":function(t,e,n){"use strict";e.a=function(t){return Math.ceil(Math.log(t.length)/Math.LN2)+1}},"../../../../d3-array/src/ticks.js":function(t,e,n){"use strict";e.b=s,e.c=function(t,e,n){var s=Math.abs(e-t)/Math.max(0,n),a=Math.pow(10,Math.floor(Math.log(s)/Math.LN10)),c=s/a;c>=r?a*=10:c>=i?a*=5:c>=o&&(a*=2);return e<t?-a:a};var r=Math.sqrt(50),i=Math.sqrt(10),o=Math.sqrt(2);function s(t,e,n){var s=(e-t)/Math.max(0,n),a=Math.floor(Math.log(s)/Math.LN10),c=s/Math.pow(10,a);return a>=0?(c>=r?10:c>=i?5:c>=o?2:1)*Math.pow(10,a):-Math.pow(10,-a)/(c>=r?10:c>=i?5:c>=o?2:1)}e.a=function(t,e,n){var r,i,o,a,c=-1;if(n=+n,(t=+t)===(e=+e)&&n>0)return[t];if((r=e<t)&&(i=t,t=e,e=i),0===(a=s(t,e,n))||!isFinite(a))return[];if(a>0)for(t=Math.ceil(t/a),e=Math.floor(e/a),o=new Array(i=Math.ceil(e-t+1));++c<i;)o[c]=(t+c)*a;else for(t=Math.floor(t*a),e=Math.ceil(e*a),o=new Array(i=Math.ceil(t-e+1));++c<i;)o[c]=(t-c)/a;return r&&o.reverse(),o}},"../../../../d3-array/src/transpose.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/src/min.js");function i(t){return t.length}e.a=function(t){if(!(s=t.length))return[];for(var e=-1,n=Object(r.a)(t,i),o=new Array(n);++e<n;)for(var s,a=-1,c=o[e]=new Array(s);++a<s;)c[a]=t[a][e];return o}},"../../../../d3-array/src/variance.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/src/number.js");e.a=function(t,e){var n,i,o=t.length,s=0,a=-1,c=0,u=0;if(null==e)for(;++a<o;)isNaN(n=Object(r.a)(t[a]))||(u+=(i=n-c)*(n-(c+=i/++s)));else for(;++a<o;)isNaN(n=Object(r.a)(e(t[a],a,t)))||(u+=(i=n-c)*(n-(c+=i/++s)));if(s>1)return u/(s-1)}},"../../../../d3-array/src/zip.js":function(t,e,n){"use strict";n("../../../../d3-array/src/transpose.js")},"../../../../d3-collection/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-collection/src/nest.js");n.d(e,"b",function(){return r.a});n("../../../../d3-collection/src/set.js");var i=n("../../../../d3-collection/src/map.js");n.d(e,"a",function(){return i.a});n("../../../../d3-collection/src/keys.js"),n("../../../../d3-collection/src/values.js"),n("../../../../d3-collection/src/entries.js")},"../../../../d3-collection/src/entries.js":function(t,e,n){"use strict"},"../../../../d3-collection/src/keys.js":function(t,e,n){"use strict"},"../../../../d3-collection/src/map.js":function(t,e,n){"use strict";n.d(e,"b",function(){return r});var r="$";function i(){}function o(t,e){var n=new i;if(t instanceof i)t.each(function(t,e){n.set(e,t)});else if(Array.isArray(t)){var r,o=-1,s=t.length;if(null==e)for(;++o<s;)n.set(o,t[o]);else for(;++o<s;)n.set(e(r=t[o],o,t),r)}else if(t)for(var a in t)n.set(a,t[a]);return n}i.prototype=o.prototype={constructor:i,has:function(t){return r+t in this},get:function(t){return this[r+t]},set:function(t,e){return this[r+t]=e,this},remove:function(t){var e=r+t;return e in this&&delete this[e]},clear:function(){for(var t in this)t[0]===r&&delete this[t]},keys:function(){var t=[];for(var e in this)e[0]===r&&t.push(e.slice(1));return t},values:function(){var t=[];for(var e in this)e[0]===r&&t.push(this[e]);return t},entries:function(){var t=[];for(var e in this)e[0]===r&&t.push({key:e.slice(1),value:this[e]});return t},size:function(){var t=0;for(var e in this)e[0]===r&&++t;return t},empty:function(){for(var t in this)if(t[0]===r)return!1;return!0},each:function(t){for(var e in this)e[0]===r&&t(this[e],e.slice(1),this)}},e.a=o},"../../../../d3-collection/src/nest.js":function(t,e,n){"use strict";var r=n("../../../../d3-collection/src/map.js");function i(){return{}}function o(t,e,n){t[e]=n}function s(){return Object(r.a)()}function a(t,e,n){t.set(e,n)}e.a=function(){var t,e,n,c=[],u=[];function l(n,i,o,s){if(i>=c.length)return null!=t&&n.sort(t),null!=e?e(n):n;for(var a,u,p,d=-1,h=n.length,f=c[i++],m=Object(r.a)(),g=o();++d<h;)(p=m.get(a=f(u=n[d])+""))?p.push(u):m.set(a,[u]);return m.each(function(t,e){s(g,e,l(t,i,o,s))}),g}return n={object:function(t){return l(t,0,i,o)},map:function(t){return l(t,0,s,a)},entries:function(t){return function t(n,r){if(++r>c.length)return n;var i,o=u[r-1];return null!=e&&r>=c.length?i=n.entries():(i=[],n.each(function(e,n){i.push({key:n,values:t(e,r)})})),null!=o?i.sort(function(t,e){return o(t.key,e.key)}):i}(l(t,0,s,a),0)},key:function(t){return c.push(t),n},sortKeys:function(t){return u[c.length-1]=t,n},sortValues:function(e){return t=e,n},rollup:function(t){return e=t,n}}}},"../../../../d3-collection/src/set.js":function(t,e,n){"use strict";var r=n("../../../../d3-collection/src/map.js");function i(){}var o=r.a.prototype;function s(t,e){var n=new i;if(t instanceof i)t.each(function(t){n.add(t)});else if(t){var r=-1,o=t.length;if(null==e)for(;++r<o;)n.add(t[r]);else for(;++r<o;)n.add(e(t[r],r,t))}return n}i.prototype=s.prototype={constructor:i,has:o.has,add:function(t){return t+="",this[r.b+t]=t,this},remove:o.remove,clear:o.clear,values:o.keys,size:o.size,empty:o.empty,each:o.each}},"../../../../d3-collection/src/values.js":function(t,e,n){"use strict"},"../../../../d3-color/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-color/src/color.js");n.d(e,"a",function(){return r.e}),n.d(e,"f",function(){return r.g}),n.d(e,"d",function(){return r.f});var i=n("../../../../d3-color/src/lab.js");n.d(e,"e",function(){return i.a}),n.d(e,"c",function(){return i.b});var o=n("../../../../d3-color/src/cubehelix.js");n.d(e,"b",function(){return o.a})},"../../../../d3-color/src/color.js":function(t,e,n){"use strict";e.a=i,n.d(e,"d",function(){return o}),n.d(e,"c",function(){return s}),e.e=b,e.h=x,e.g=w,e.b=O,e.f=k;var r=n("../../../../d3-color/src/define.js");function i(){}var o=.7,s=1/o,a="\\s*([+-]?\\d+)\\s*",c="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)\\s*",u="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)%\\s*",l=/^#([0-9a-f]{3})$/,p=/^#([0-9a-f]{6})$/,d=new RegExp("^rgb\\("+[a,a,a]+"\\)$"),h=new RegExp("^rgb\\("+[u,u,u]+"\\)$"),f=new RegExp("^rgba\\("+[a,a,a,c]+"\\)$"),m=new RegExp("^rgba\\("+[u,u,u,c]+"\\)$"),g=new RegExp("^hsl\\("+[c,u,u]+"\\)$"),v=new RegExp("^hsla\\("+[c,u,u,c]+"\\)$"),y={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};function b(t){var e;return t=(t+"").trim().toLowerCase(),(e=l.exec(t))?new O((e=parseInt(e[1],16))>>8&15|e>>4&240,e>>4&15|240&e,(15&e)<<4|15&e,1):(e=p.exec(t))?_(parseInt(e[1],16)):(e=d.exec(t))?new O(e[1],e[2],e[3],1):(e=h.exec(t))?new O(255*e[1]/100,255*e[2]/100,255*e[3]/100,1):(e=f.exec(t))?j(e[1],e[2],e[3],e[4]):(e=m.exec(t))?j(255*e[1]/100,255*e[2]/100,255*e[3]/100,e[4]):(e=g.exec(t))?C(e[1],e[2]/100,e[3]/100,1):(e=v.exec(t))?C(e[1],e[2]/100,e[3]/100,e[4]):y.hasOwnProperty(t)?_(y[t]):"transparent"===t?new O(NaN,NaN,NaN,0):null}function _(t){return new O(t>>16&255,t>>8&255,255&t,1)}function j(t,e,n,r){return r<=0&&(t=e=n=NaN),new O(t,e,n,r)}function x(t){return t instanceof i||(t=b(t)),t?new O((t=t.rgb()).r,t.g,t.b,t.opacity):new O}function w(t,e,n,r){return 1===arguments.length?x(t):new O(t,e,n,null==r?1:r)}function O(t,e,n,r){this.r=+t,this.g=+e,this.b=+n,this.opacity=+r}function C(t,e,n,r){return r<=0?t=e=n=NaN:n<=0||n>=1?t=e=NaN:e<=0&&(t=NaN),new E(t,e,n,r)}function k(t,e,n,r){return 1===arguments.length?function(t){if(t instanceof E)return new E(t.h,t.s,t.l,t.opacity);if(t instanceof i||(t=b(t)),!t)return new E;if(t instanceof E)return t;var e=(t=t.rgb()).r/255,n=t.g/255,r=t.b/255,o=Math.min(e,n,r),s=Math.max(e,n,r),a=NaN,c=s-o,u=(s+o)/2;return c?(a=e===s?(n-r)/c+6*(n<r):n===s?(r-e)/c+2:(e-n)/c+4,c/=u<.5?s+o:2-s-o,a*=60):c=u>0&&u<1?0:a,new E(a,c,u,t.opacity)}(t):new E(t,e,n,null==r?1:r)}function E(t,e,n,r){this.h=+t,this.s=+e,this.l=+n,this.opacity=+r}function S(t,e,n){return 255*(t<60?e+(n-e)*t/60:t<180?n:t<240?e+(n-e)*(240-t)/60:e)}Object(r.a)(i,b,{displayable:function(){return this.rgb().displayable()},toString:function(){return this.rgb()+""}}),Object(r.a)(O,w,Object(r.b)(i,{brighter:function(t){return t=null==t?s:Math.pow(s,t),new O(this.r*t,this.g*t,this.b*t,this.opacity)},darker:function(t){return t=null==t?o:Math.pow(o,t),new O(this.r*t,this.g*t,this.b*t,this.opacity)},rgb:function(){return this},displayable:function(){return 0<=this.r&&this.r<=255&&0<=this.g&&this.g<=255&&0<=this.b&&this.b<=255&&0<=this.opacity&&this.opacity<=1},toString:function(){var t=this.opacity;return(1===(t=isNaN(t)?1:Math.max(0,Math.min(1,t)))?"rgb(":"rgba(")+Math.max(0,Math.min(255,Math.round(this.r)||0))+", "+Math.max(0,Math.min(255,Math.round(this.g)||0))+", "+Math.max(0,Math.min(255,Math.round(this.b)||0))+(1===t?")":", "+t+")")}})),Object(r.a)(E,k,Object(r.b)(i,{brighter:function(t){return t=null==t?s:Math.pow(s,t),new E(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?o:Math.pow(o,t),new E(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=this.h%360+360*(this.h<0),e=isNaN(t)||isNaN(this.s)?0:this.s,n=this.l,r=n+(n<.5?n:1-n)*e,i=2*n-r;return new O(S(t>=240?t-240:t+120,i,r),S(t,i,r),S(t<120?t+240:t-120,i,r),this.opacity)},displayable:function(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1}}))},"../../../../d3-color/src/cubehelix.js":function(t,e,n){"use strict";e.a=d;var r=n("../../../../d3-color/src/define.js"),i=n("../../../../d3-color/src/color.js"),o=n("../../../../d3-color/src/math.js"),s=-.29227,a=-.90649,c=1.97294,u=c*a,l=1.78277*c,p=1.78277*s- -.14861*a;function d(t,e,n,r){return 1===arguments.length?function(t){if(t instanceof h)return new h(t.h,t.s,t.l,t.opacity);t instanceof i.b||(t=Object(i.h)(t));var e=t.r/255,n=t.g/255,r=t.b/255,d=(p*r+u*e-l*n)/(p+u-l),f=r-d,m=(c*(n-d)-s*f)/a,g=Math.sqrt(m*m+f*f)/(c*d*(1-d)),v=g?Math.atan2(m,f)*o.b-120:NaN;return new h(v<0?v+360:v,g,d,t.opacity)}(t):new h(t,e,n,null==r?1:r)}function h(t,e,n,r){this.h=+t,this.s=+e,this.l=+n,this.opacity=+r}Object(r.a)(h,d,Object(r.b)(i.a,{brighter:function(t){return t=null==t?i.c:Math.pow(i.c,t),new h(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?i.d:Math.pow(i.d,t),new h(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=isNaN(this.h)?0:(this.h+120)*o.a,e=+this.l,n=isNaN(this.s)?0:this.s*e*(1-e),r=Math.cos(t),u=Math.sin(t);return new i.b(255*(e+n*(-.14861*r+1.78277*u)),255*(e+n*(s*r+a*u)),255*(e+n*(c*r)),this.opacity)}}))},"../../../../d3-color/src/define.js":function(t,e,n){"use strict";e.b=function(t,e){var n=Object.create(t.prototype);for(var r in e)n[r]=e[r];return n},e.a=function(t,e,n){t.prototype=e.prototype=n,n.constructor=t}},"../../../../d3-color/src/lab.js":function(t,e,n){"use strict";e.a=f,e.b=_;var r=n("../../../../d3-color/src/define.js"),i=n("../../../../d3-color/src/color.js"),o=n("../../../../d3-color/src/math.js"),s=.95047,a=1,c=1.08883,u=4/29,l=6/29,p=3*l*l,d=l*l*l;function h(t){if(t instanceof m)return new m(t.l,t.a,t.b,t.opacity);if(t instanceof j){var e=t.h*o.a;return new m(t.l,Math.cos(e)*t.c,Math.sin(e)*t.c,t.opacity)}t instanceof i.b||(t=Object(i.h)(t));var n=b(t.r),r=b(t.g),u=b(t.b),l=g((.4124564*n+.3575761*r+.1804375*u)/s),p=g((.2126729*n+.7151522*r+.072175*u)/a);return new m(116*p-16,500*(l-p),200*(p-g((.0193339*n+.119192*r+.9503041*u)/c)),t.opacity)}function f(t,e,n,r){return 1===arguments.length?h(t):new m(t,e,n,null==r?1:r)}function m(t,e,n,r){this.l=+t,this.a=+e,this.b=+n,this.opacity=+r}function g(t){return t>d?Math.pow(t,1/3):t/p+u}function v(t){return t>l?t*t*t:p*(t-u)}function y(t){return 255*(t<=.0031308?12.92*t:1.055*Math.pow(t,1/2.4)-.055)}function b(t){return(t/=255)<=.04045?t/12.92:Math.pow((t+.055)/1.055,2.4)}function _(t,e,n,r){return 1===arguments.length?function(t){if(t instanceof j)return new j(t.h,t.c,t.l,t.opacity);t instanceof m||(t=h(t));var e=Math.atan2(t.b,t.a)*o.b;return new j(e<0?e+360:e,Math.sqrt(t.a*t.a+t.b*t.b),t.l,t.opacity)}(t):new j(t,e,n,null==r?1:r)}function j(t,e,n,r){this.h=+t,this.c=+e,this.l=+n,this.opacity=+r}Object(r.a)(m,f,Object(r.b)(i.a,{brighter:function(t){return new m(this.l+18*(null==t?1:t),this.a,this.b,this.opacity)},darker:function(t){return new m(this.l-18*(null==t?1:t),this.a,this.b,this.opacity)},rgb:function(){var t=(this.l+16)/116,e=isNaN(this.a)?t:t+this.a/500,n=isNaN(this.b)?t:t-this.b/200;return t=a*v(t),e=s*v(e),n=c*v(n),new i.b(y(3.2404542*e-1.5371385*t-.4985314*n),y(-.969266*e+1.8760108*t+.041556*n),y(.0556434*e-.2040259*t+1.0572252*n),this.opacity)}})),Object(r.a)(j,_,Object(r.b)(i.a,{brighter:function(t){return new j(this.h,this.c,this.l+18*(null==t?1:t),this.opacity)},darker:function(t){return new j(this.h,this.c,this.l-18*(null==t?1:t),this.opacity)},rgb:function(){return h(this).rgb()}}))},"../../../../d3-color/src/math.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r}),n.d(e,"b",function(){return i});var r=Math.PI/180,i=180/Math.PI},"../../../../d3-contour/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-contour/src/contours.js");n.d(e,"b",function(){return r.a});var i=n("../../../../d3-contour/src/density.js");n.d(e,"a",function(){return i.a})},"../../../../d3-contour/src/area.js":function(t,e,n){"use strict";e.a=function(t){for(var e=0,n=t.length,r=t[n-1][1]*t[0][0]-t[n-1][0]*t[0][1];++e<n;)r+=t[e-1][1]*t[e][0]-t[e-1][0]*t[e][1];return r}},"../../../../d3-contour/src/array.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=Array.prototype.slice},"../../../../d3-contour/src/ascending.js":function(t,e,n){"use strict";e.a=function(t,e){return t-e}},"../../../../d3-contour/src/blur.js":function(t,e,n){"use strict";e.a=function(t,e,n){for(var r=t.width,i=t.height,o=1+(n<<1),s=0;s<i;++s)for(var a=0,c=0;a<r+n;++a)a<r&&(c+=t.data[a+s*r]),a>=n&&(a>=o&&(c-=t.data[a-o+s*r]),e.data[a-n+s*r]=c/Math.min(a+1,r-1+o-a,o))},e.b=function(t,e,n){for(var r=t.width,i=t.height,o=1+(n<<1),s=0;s<r;++s)for(var a=0,c=0;a<i+n;++a)a<i&&(c+=t.data[s+a*r]),a>=n&&(a>=o&&(c-=t.data[s+(a-o)*r]),e.data[s+(a-n)*r]=c/Math.min(a+1,i-1+o-a,o))}},"../../../../d3-contour/src/constant.js":function(t,e,n){"use strict";e.a=function(t){return function(){return t}}},"../../../../d3-contour/src/contains.js":function(t,e,n){"use strict";function r(t,e){for(var n,r,i,o,s,a,c,u,l,p,d=e[0],h=e[1],f=-1,m=0,g=t.length,v=g-1;m<g;v=m++){var y=t[m],b=y[0],_=y[1],j=t[v],x=j[0],w=j[1];if(p=void 0,o=c=y,a=l=e,((s=u=j)[0]-o[0])*(a[1]-o[1])==(a[0]-o[0])*(s[1]-o[1])&&(n=c[p=+(c[0]===u[0])],r=l[p],i=u[p],n<=r&&r<=i||i<=r&&r<=n))return 0;_>h!=w>h&&d<(x-b)*(h-_)/(w-_)+b&&(f=-f)}return f}e.a=function(t,e){for(var n,i=-1,o=e.length;++i<o;)if(n=r(t,e[i]))return n;return 0}},"../../../../d3-contour/src/contours.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-contour/src/array.js"),o=n("../../../../d3-contour/src/ascending.js"),s=n("../../../../d3-contour/src/area.js"),a=n("../../../../d3-contour/src/constant.js"),c=n("../../../../d3-contour/src/contains.js"),u=n("../../../../d3-contour/src/noop.js"),l=[[],[[[1,1.5],[.5,1]]],[[[1.5,1],[1,1.5]]],[[[1.5,1],[.5,1]]],[[[1,.5],[1.5,1]]],[[[1,1.5],[.5,1]],[[1,.5],[1.5,1]]],[[[1,.5],[1,1.5]]],[[[1,.5],[.5,1]]],[[[.5,1],[1,.5]]],[[[1,1.5],[1,.5]]],[[[.5,1],[1,.5]],[[1.5,1],[1,1.5]]],[[[1.5,1],[1,.5]]],[[[.5,1],[1.5,1]]],[[[1,1.5],[1.5,1]]],[[[.5,1],[1,1.5]]],[]];e.a=function(){var t=1,e=1,n=r.p,p=f;function d(i){var a=n(i);if(Array.isArray(a))a=a.slice().sort(o.a);else{var u=Object(r.f)(i),d=u[0],f=u[1];a=Object(r.r)(d,f,a),a=Object(r.n)(Math.floor(d/a)*a,Math.floor(f/a)*a,a)}return a.map(function(n){var r=[],o=[];return function(n,r,i){var o,s,a,c,u,p,d=new Array,f=new Array;o=s=-1,c=n[0]>=r,l[c<<1].forEach(m);for(;++o<t-1;)a=c,c=n[o+1]>=r,l[a|c<<1].forEach(m);l[c<<0].forEach(m);for(;++s<e-1;){for(o=-1,c=n[s*t+t]>=r,u=n[s*t]>=r,l[c<<1|u<<2].forEach(m);++o<t-1;)a=c,c=n[s*t+t+o+1]>=r,p=u,u=n[s*t+o+1]>=r,l[a|c<<1|u<<2|p<<3].forEach(m);l[c|u<<3].forEach(m)}o=-1,u=n[s*t]>=r,l[u<<2].forEach(m);for(;++o<t-1;)p=u,u=n[s*t+o+1]>=r,l[u<<2|p<<3].forEach(m);function m(t){var e,n,r=[t[0][0]+o,t[0][1]+s],a=[t[1][0]+o,t[1][1]+s],c=h(r),u=h(a);(e=f[c])?(n=d[u])?(delete f[e.end],delete d[n.start],e===n?(e.ring.push(a),i(e.ring)):d[e.start]=f[n.end]={start:e.start,end:n.end,ring:e.ring.concat(n.ring)}):(delete f[e.end],e.ring.push(a),f[e.end=u]=e):(e=d[u])?(n=f[c])?(delete d[e.start],delete f[n.end],e===n?(e.ring.push(a),i(e.ring)):d[n.start]=f[e.end]={start:n.start,end:e.end,ring:n.ring.concat(e.ring)}):(delete d[e.start],e.ring.unshift(r),d[e.start=c]=e):d[c]=f[u]={start:c,end:u,ring:[r,a]}}l[u<<3].forEach(m)}(i,n,function(t){p(t,i,n),Object(s.a)(t)>0?r.push([t]):o.push(t)}),o.forEach(function(t){for(var e,n=0,i=r.length;n<i;++n)if(-1!==Object(c.a)((e=r[n])[0],t))return void e.push(t)}),r}).map(function(t,e){return{type:"MultiPolygon",value:a[e],coordinates:t}})}function h(e){return 2*e[0]+e[1]*(t+1)*4}function f(n,r,i){n.forEach(function(n){var o,s=n[0],a=n[1],c=0|s,u=0|a,l=r[u*t+c];s>0&&s<t&&c===s&&(o=r[u*t+c-1],n[0]=s+(i-o)/(l-o)-.5),a>0&&a<e&&u===a&&(o=r[(u-1)*t+c],n[1]=a+(i-o)/(l-o)-.5)})}return d.size=function(n){if(!arguments.length)return[t,e];var r=Math.ceil(n[0]),i=Math.ceil(n[1]);if(!(r>0&&i>0))throw new Error("invalid size");return t=r,e=i,d},d.thresholds=function(t){return arguments.length?(n="function"==typeof t?t:Array.isArray(t)?Object(a.a)(i.a.call(t)):Object(a.a)(t),d):n},d.smooth=function(t){return arguments.length?(p=t?f:u.a,d):p===f},d}},"../../../../d3-contour/src/density.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-contour/src/array.js"),o=n("../../../../d3-contour/src/blur.js"),s=n("../../../../d3-contour/src/constant.js"),a=n("../../../../d3-contour/src/contours.js");function c(t){return t[0]}function u(t){return t[1]}e.a=function(){var t=c,e=u,n=960,l=500,p=20,d=2,h=3*p,f=n+2*h>>d,m=l+2*h>>d,g=Object(s.a)(20);function v(n){var i=new Float32Array(f*m),s=new Float32Array(f*m);n.forEach(function(n,r,o){var s=t(n,r,o)+h>>d,a=e(n,r,o)+h>>d;s>=0&&s<f&&a>=0&&a<m&&++i[s+a*f]}),Object(o.a)({width:f,height:m,data:i},{width:f,height:m,data:s},p>>d),Object(o.b)({width:f,height:m,data:s},{width:f,height:m,data:i},p>>d),Object(o.a)({width:f,height:m,data:i},{width:f,height:m,data:s},p>>d),Object(o.b)({width:f,height:m,data:s},{width:f,height:m,data:i},p>>d),Object(o.a)({width:f,height:m,data:i},{width:f,height:m,data:s},p>>d),Object(o.b)({width:f,height:m,data:s},{width:f,height:m,data:i},p>>d);var c=g(i);if(!Array.isArray(c)){var u=Object(r.g)(i);c=Object(r.r)(0,u,c),(c=Object(r.n)(0,Math.floor(u/c)*c,c)).shift()}return Object(a.a)().thresholds(c).size([f,m])(i).map(y)}function y(t){return t.value*=Math.pow(2,-2*d),t.coordinates.forEach(b),t}function b(t){t.forEach(_)}function _(t){t.forEach(j)}function j(t){t[0]=t[0]*Math.pow(2,d)-h,t[1]=t[1]*Math.pow(2,d)-h}function x(){return f=n+2*(h=3*p)>>d,m=l+2*h>>d,v}return v.x=function(e){return arguments.length?(t="function"==typeof e?e:Object(s.a)(+e),v):t},v.y=function(t){return arguments.length?(e="function"==typeof t?t:Object(s.a)(+t),v):e},v.size=function(t){if(!arguments.length)return[n,l];var e=Math.ceil(t[0]),r=Math.ceil(t[1]);if(!(e>=0||e>=0))throw new Error("invalid size");return n=e,l=r,x()},v.cellSize=function(t){if(!arguments.length)return 1<<d;if(!((t=+t)>=1))throw new Error("invalid cell size");return d=Math.floor(Math.log(t)/Math.LN2),x()},v.thresholds=function(t){return arguments.length?(g="function"==typeof t?t:Array.isArray(t)?Object(s.a)(i.a.call(t)):Object(s.a)(t),v):g},v.bandwidth=function(t){if(!arguments.length)return Math.sqrt(p*(p+1));if(!((t=+t)>=0))throw new Error("invalid bandwidth");return p=Math.round((Math.sqrt(4*t*t+1)-1)/2),x()},v}},"../../../../d3-contour/src/noop.js":function(t,e,n){"use strict";e.a=function(){}},"../../../../d3-dispatch/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-dispatch/src/dispatch.js");n.d(e,"a",function(){return r.a})},"../../../../d3-dispatch/src/dispatch.js":function(t,e,n){"use strict";var r={value:function(){}};function i(){for(var t,e=0,n=arguments.length,r={};e<n;++e){if(!(t=arguments[e]+"")||t in r)throw new Error("illegal type: "+t);r[t]=[]}return new o(r)}function o(t){this._=t}function s(t,e){for(var n,r=0,i=t.length;r<i;++r)if((n=t[r]).name===e)return n.value}function a(t,e,n){for(var i=0,o=t.length;i<o;++i)if(t[i].name===e){t[i]=r,t=t.slice(0,i).concat(t.slice(i+1));break}return null!=n&&t.push({name:e,value:n}),t}o.prototype=i.prototype={constructor:o,on:function(t,e){var n,r,i=this._,o=(r=i,(t+"").trim().split(/^|\s+/).map(function(t){var e="",n=t.indexOf(".");if(n>=0&&(e=t.slice(n+1),t=t.slice(0,n)),t&&!r.hasOwnProperty(t))throw new Error("unknown type: "+t);return{type:t,name:e}})),c=-1,u=o.length;if(!(arguments.length<2)){if(null!=e&&"function"!=typeof e)throw new Error("invalid callback: "+e);for(;++c<u;)if(n=(t=o[c]).type)i[n]=a(i[n],t.name,e);else if(null==e)for(n in i)i[n]=a(i[n],t.name,null);return this}for(;++c<u;)if((n=(t=o[c]).type)&&(n=s(i[n],t.name)))return n},copy:function(){var t={},e=this._;for(var n in e)t[n]=e[n].slice();return new o(t)},call:function(t,e){if((n=arguments.length-2)>0)for(var n,r,i=new Array(n),o=0;o<n;++o)i[o]=arguments[o+2];if(!this._.hasOwnProperty(t))throw new Error("unknown type: "+t);for(o=0,n=(r=this._[t]).length;o<n;++o)r[o].value.apply(e,i)},apply:function(t,e,n){if(!this._.hasOwnProperty(t))throw new Error("unknown type: "+t);for(var r=this._[t],i=0,o=r.length;i<o;++i)r[i].value.apply(e,n)}},e.a=i},"../../../../d3-dsv/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-dsv/src/dsv.js");n.d(e,"b",function(){return r.a});var i=n("../../../../d3-dsv/src/csv.js");n.d(e,"a",function(){return i.a});var o=n("../../../../d3-dsv/src/tsv.js");n.d(e,"c",function(){return o.a})},"../../../../d3-dsv/src/csv.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../d3-dsv/src/dsv.js"),i=Object(r.a)(","),o=i.parse;i.parseRows,i.format,i.formatRows},"../../../../d3-dsv/src/dsv.js":function(t,e,n){"use strict";var r={},i={},o=34,s=10,a=13;function c(t){return new Function("d","return {"+t.map(function(t,e){return JSON.stringify(t)+": d["+e+"]"}).join(",")+"}")}e.a=function(t){var e=new RegExp('["'+t+"\n\r]"),n=t.charCodeAt(0);function u(t,e){var c,u=[],l=t.length,p=0,d=0,h=l<=0,f=!1;function m(){if(h)return i;if(f)return f=!1,r;var e,c,u=p;if(t.charCodeAt(u)===o){for(;p++<l&&t.charCodeAt(p)!==o||t.charCodeAt(++p)===o;);return(e=p)>=l?h=!0:(c=t.charCodeAt(p++))===s?f=!0:c===a&&(f=!0,t.charCodeAt(p)===s&&++p),t.slice(u+1,e-1).replace(/""/g,'"')}for(;p<l;){if((c=t.charCodeAt(e=p++))===s)f=!0;else if(c===a)f=!0,t.charCodeAt(p)===s&&++p;else if(c!==n)continue;return t.slice(u,e)}return h=!0,t.slice(u,l)}for(t.charCodeAt(l-1)===s&&--l,t.charCodeAt(l-1)===a&&--l;(c=m())!==i;){for(var g=[];c!==r&&c!==i;)g.push(c),c=m();e&&null==(g=e(g,d++))||u.push(g)}return u}function l(e){return e.map(p).join(t)}function p(t){return null==t?"":e.test(t+="")?'"'+t.replace(/"/g,'""')+'"':t}return{parse:function(t,e){var n,r,i=u(t,function(t,i){if(n)return n(t,i-1);var o,s,a;r=t,n=e?(s=e,a=c(o=t),function(t,e){return s(a(t),e,o)}):c(t)});return i.columns=r||[],i},parseRows:u,format:function(e,n){var r,i,o;return null==n&&(r=e,i=Object.create(null),o=[],r.forEach(function(t){for(var e in t)e in i||o.push(i[e]=e)}),n=o),[n.map(p).join(t)].concat(e.map(function(e){return n.map(function(t){return p(e[t])}).join(t)})).join("\n")},formatRows:function(t){return t.map(l).join("\n")}}}},"../../../../d3-dsv/src/tsv.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../d3-dsv/src/dsv.js"),i=Object(r.a)("\t"),o=i.parse;i.parseRows,i.format,i.formatRows},"../../../../d3-force/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-force/src/center.js");n.d(e,"a",function(){return r.a});var i=n("../../../../d3-force/src/collide.js");n.d(e,"b",function(){return i.a});var o=n("../../../../d3-force/src/link.js");n.d(e,"c",function(){return o.a});var s=n("../../../../d3-force/src/manyBody.js");n.d(e,"d",function(){return s.a});n("../../../../d3-force/src/radial.js");var a=n("../../../../d3-force/src/simulation.js");n.d(e,"e",function(){return a.a});var c=n("../../../../d3-force/src/x.js");n.d(e,"f",function(){return c.a});var u=n("../../../../d3-force/src/y.js");n.d(e,"g",function(){return u.a})},"../../../../d3-force/src/center.js":function(t,e,n){"use strict";e.a=function(t,e){var n;function r(){var r,i,o=n.length,s=0,a=0;for(r=0;r<o;++r)s+=(i=n[r]).x,a+=i.y;for(s=s/o-t,a=a/o-e,r=0;r<o;++r)(i=n[r]).x-=s,i.y-=a}return null==t&&(t=0),null==e&&(e=0),r.initialize=function(t){n=t},r.x=function(e){return arguments.length?(t=+e,r):t},r.y=function(t){return arguments.length?(e=+t,r):e},r}},"../../../../d3-force/src/collide.js":function(t,e,n){"use strict";var r=n("../../../../d3-force/src/constant.js"),i=n("../../../../d3-force/src/jiggle.js"),o=n("../../../../d3-quadtree/index.js");function s(t){return t.x+t.vx}function a(t){return t.y+t.vy}e.a=function(t){var e,n,c=1,u=1;function l(){for(var t,r,l,d,h,f,m,g=e.length,v=0;v<u;++v)for(r=Object(o.a)(e,s,a).visitAfter(p),t=0;t<g;++t)l=e[t],f=n[l.index],m=f*f,d=l.x+l.vx,h=l.y+l.vy,r.visit(y);function y(t,e,n,r,o){var s=t.data,a=t.r,u=f+a;if(!s)return e>d+u||r<d-u||n>h+u||o<h-u;if(s.index>l.index){var p=d-s.x-s.vx,g=h-s.y-s.vy,v=p*p+g*g;v<u*u&&(0===p&&(v+=(p=Object(i.a)())*p),0===g&&(v+=(g=Object(i.a)())*g),v=(u-(v=Math.sqrt(v)))/v*c,l.vx+=(p*=v)*(u=(a*=a)/(m+a)),l.vy+=(g*=v)*u,s.vx-=p*(u=1-u),s.vy-=g*u)}}}function p(t){if(t.data)return t.r=n[t.data.index];for(var e=t.r=0;e<4;++e)t[e]&&t[e].r>t.r&&(t.r=t[e].r)}function d(){if(e){var r,i,o=e.length;for(n=new Array(o),r=0;r<o;++r)i=e[r],n[i.index]=+t(i,r,e)}}return"function"!=typeof t&&(t=Object(r.a)(null==t?1:+t)),l.initialize=function(t){e=t,d()},l.iterations=function(t){return arguments.length?(u=+t,l):u},l.strength=function(t){return arguments.length?(c=+t,l):c},l.radius=function(e){return arguments.length?(t="function"==typeof e?e:Object(r.a)(+e),d(),l):t},l}},"../../../../d3-force/src/constant.js":function(t,e,n){"use strict";e.a=function(t){return function(){return t}}},"../../../../d3-force/src/jiggle.js":function(t,e,n){"use strict";e.a=function(){return 1e-6*(Math.random()-.5)}},"../../../../d3-force/src/link.js":function(t,e,n){"use strict";var r=n("../../../../d3-force/src/constant.js"),i=n("../../../../d3-force/src/jiggle.js"),o=n("../../../../d3-collection/index.js");function s(t){return t.index}function a(t,e){var n=t.get(e);if(!n)throw new Error("missing: "+e);return n}e.a=function(t){var e,n,c,u,l,p=s,d=function(t){return 1/Math.min(u[t.source.index],u[t.target.index])},h=Object(r.a)(30),f=1;function m(r){for(var o=0,s=t.length;o<f;++o)for(var a,c,u,p,d,h,m,g=0;g<s;++g)c=(a=t[g]).source,p=(u=a.target).x+u.vx-c.x-c.vx||Object(i.a)(),d=u.y+u.vy-c.y-c.vy||Object(i.a)(),p*=h=((h=Math.sqrt(p*p+d*d))-n[g])/h*r*e[g],d*=h,u.vx-=p*(m=l[g]),u.vy-=d*m,c.vx+=p*(m=1-m),c.vy+=d*m}function g(){if(c){var r,i,s=c.length,d=t.length,h=Object(o.a)(c,p);for(r=0,u=new Array(s);r<d;++r)(i=t[r]).index=r,"object"!=typeof i.source&&(i.source=a(h,i.source)),"object"!=typeof i.target&&(i.target=a(h,i.target)),u[i.source.index]=(u[i.source.index]||0)+1,u[i.target.index]=(u[i.target.index]||0)+1;for(r=0,l=new Array(d);r<d;++r)i=t[r],l[r]=u[i.source.index]/(u[i.source.index]+u[i.target.index]);e=new Array(d),v(),n=new Array(d),y()}}function v(){if(c)for(var n=0,r=t.length;n<r;++n)e[n]=+d(t[n],n,t)}function y(){if(c)for(var e=0,r=t.length;e<r;++e)n[e]=+h(t[e],e,t)}return null==t&&(t=[]),m.initialize=function(t){c=t,g()},m.links=function(e){return arguments.length?(t=e,g(),m):t},m.id=function(t){return arguments.length?(p=t,m):p},m.iterations=function(t){return arguments.length?(f=+t,m):f},m.strength=function(t){return arguments.length?(d="function"==typeof t?t:Object(r.a)(+t),v(),m):d},m.distance=function(t){return arguments.length?(h="function"==typeof t?t:Object(r.a)(+t),y(),m):h},m}},"../../../../d3-force/src/manyBody.js":function(t,e,n){"use strict";var r=n("../../../../d3-force/src/constant.js"),i=n("../../../../d3-force/src/jiggle.js"),o=n("../../../../d3-quadtree/index.js"),s=n("../../../../d3-force/src/simulation.js");e.a=function(){var t,e,n,a,c=Object(r.a)(-30),u=1,l=1/0,p=.81;function d(r){var i,a=t.length,c=Object(o.a)(t,s.b,s.c).visitAfter(f);for(n=r,i=0;i<a;++i)e=t[i],c.visit(m)}function h(){if(t){var e,n,r=t.length;for(a=new Array(r),e=0;e<r;++e)n=t[e],a[n.index]=+c(n,e,t)}}function f(t){var e,n,r,i,o,s=0,c=0;if(t.length){for(r=i=o=0;o<4;++o)(e=t[o])&&(n=Math.abs(e.value))&&(s+=e.value,c+=n,r+=n*e.x,i+=n*e.y);t.x=r/c,t.y=i/c}else{(e=t).x=e.data.x,e.y=e.data.y;do{s+=a[e.data.index]}while(e=e.next)}t.value=s}function m(t,r,o,s){if(!t.value)return!0;var c=t.x-e.x,d=t.y-e.y,h=s-r,f=c*c+d*d;if(h*h/p<f)return f<l&&(0===c&&(f+=(c=Object(i.a)())*c),0===d&&(f+=(d=Object(i.a)())*d),f<u&&(f=Math.sqrt(u*f)),e.vx+=c*t.value*n/f,e.vy+=d*t.value*n/f),!0;if(!(t.length||f>=l)){(t.data!==e||t.next)&&(0===c&&(f+=(c=Object(i.a)())*c),0===d&&(f+=(d=Object(i.a)())*d),f<u&&(f=Math.sqrt(u*f)));do{t.data!==e&&(h=a[t.data.index]*n/f,e.vx+=c*h,e.vy+=d*h)}while(t=t.next)}}return d.initialize=function(e){t=e,h()},d.strength=function(t){return arguments.length?(c="function"==typeof t?t:Object(r.a)(+t),h(),d):c},d.distanceMin=function(t){return arguments.length?(u=t*t,d):Math.sqrt(u)},d.distanceMax=function(t){return arguments.length?(l=t*t,d):Math.sqrt(l)},d.theta=function(t){return arguments.length?(p=t*t,d):Math.sqrt(p)},d}},"../../../../d3-force/src/radial.js":function(t,e,n){"use strict";n("../../../../d3-force/src/constant.js")},"../../../../d3-force/src/simulation.js":function(t,e,n){"use strict";e.b=function(t){return t.x},e.c=function(t){return t.y};var r=n("../../../../d3-dispatch/index.js"),i=n("../../../../d3-collection/index.js"),o=n("../../../../d3-timer/index.js");var s=10,a=Math.PI*(3-Math.sqrt(5));e.a=function(t){var e,n=1,c=.001,u=1-Math.pow(c,1/300),l=0,p=.6,d=Object(i.a)(),h=Object(o.a)(m),f=Object(r.a)("tick","end");function m(){g(),f.call("tick",e),n<c&&(h.stop(),f.call("end",e))}function g(){var e,r,i=t.length;for(n+=(l-n)*u,d.each(function(t){t(n)}),e=0;e<i;++e)null==(r=t[e]).fx?r.x+=r.vx*=p:(r.x=r.fx,r.vx=0),null==r.fy?r.y+=r.vy*=p:(r.y=r.fy,r.vy=0)}function v(){for(var e,n=0,r=t.length;n<r;++n){if((e=t[n]).index=n,isNaN(e.x)||isNaN(e.y)){var i=s*Math.sqrt(n),o=n*a;e.x=i*Math.cos(o),e.y=i*Math.sin(o)}(isNaN(e.vx)||isNaN(e.vy))&&(e.vx=e.vy=0)}}function y(e){return e.initialize&&e.initialize(t),e}return null==t&&(t=[]),v(),e={tick:g,restart:function(){return h.restart(m),e},stop:function(){return h.stop(),e},nodes:function(n){return arguments.length?(t=n,v(),d.each(y),e):t},alpha:function(t){return arguments.length?(n=+t,e):n},alphaMin:function(t){return arguments.length?(c=+t,e):c},alphaDecay:function(t){return arguments.length?(u=+t,e):+u},alphaTarget:function(t){return arguments.length?(l=+t,e):l},velocityDecay:function(t){return arguments.length?(p=1-t,e):1-p},force:function(t,n){return arguments.length>1?(null==n?d.remove(t):d.set(t,y(n)),e):d.get(t)},find:function(e,n,r){var i,o,s,a,c,u=0,l=t.length;for(null==r?r=1/0:r*=r,u=0;u<l;++u)(s=(i=e-(a=t[u]).x)*i+(o=n-a.y)*o)<r&&(c=a,r=s);return c},on:function(t,n){return arguments.length>1?(f.on(t,n),e):f.on(t)}}}},"../../../../d3-force/src/x.js":function(t,e,n){"use strict";var r=n("../../../../d3-force/src/constant.js");e.a=function(t){var e,n,i,o=Object(r.a)(.1);function s(t){for(var r,o=0,s=e.length;o<s;++o)(r=e[o]).vx+=(i[o]-r.x)*n[o]*t}function a(){if(e){var r,s=e.length;for(n=new Array(s),i=new Array(s),r=0;r<s;++r)n[r]=isNaN(i[r]=+t(e[r],r,e))?0:+o(e[r],r,e)}}return"function"!=typeof t&&(t=Object(r.a)(null==t?0:+t)),s.initialize=function(t){e=t,a()},s.strength=function(t){return arguments.length?(o="function"==typeof t?t:Object(r.a)(+t),a(),s):o},s.x=function(e){return arguments.length?(t="function"==typeof e?e:Object(r.a)(+e),a(),s):t},s}},"../../../../d3-force/src/y.js":function(t,e,n){"use strict";var r=n("../../../../d3-force/src/constant.js");e.a=function(t){var e,n,i,o=Object(r.a)(.1);function s(t){for(var r,o=0,s=e.length;o<s;++o)(r=e[o]).vy+=(i[o]-r.y)*n[o]*t}function a(){if(e){var r,s=e.length;for(n=new Array(s),i=new Array(s),r=0;r<s;++r)n[r]=isNaN(i[r]=+t(e[r],r,e))?0:+o(e[r],r,e)}}return"function"!=typeof t&&(t=Object(r.a)(null==t?0:+t)),s.initialize=function(t){e=t,a()},s.strength=function(t){return arguments.length?(o="function"==typeof t?t:Object(r.a)(+t),a(),s):o},s.y=function(e){return arguments.length?(t="function"==typeof e?e:Object(r.a)(+e),a(),s):t},s}},"../../../../d3-format/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-format/src/defaultLocale.js");n.d(e,"a",function(){return r.a}),n.d(e,"b",function(){return r.b});n("../../../../d3-format/src/locale.js");var i=n("../../../../d3-format/src/formatSpecifier.js");n.d(e,"c",function(){return i.a});var o=n("../../../../d3-format/src/precisionFixed.js");n.d(e,"d",function(){return o.a});var s=n("../../../../d3-format/src/precisionPrefix.js");n.d(e,"e",function(){return s.a});var a=n("../../../../d3-format/src/precisionRound.js");n.d(e,"f",function(){return a.a})},"../../../../d3-format/src/defaultLocale.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i}),n.d(e,"b",function(){return o});var r,i,o,s,a=n("../../../../d3-format/src/locale.js");s={decimal:".",thousands:",",grouping:[3],currency:["$",""]},r=Object(a.a)(s),i=r.format,o=r.formatPrefix},"../../../../d3-format/src/exponent.js":function(t,e,n){"use strict";var r=n("../../../../d3-format/src/formatDecimal.js");e.a=function(t){return(t=Object(r.a)(Math.abs(t)))?t[1]:NaN}},"../../../../d3-format/src/formatDecimal.js":function(t,e,n){"use strict";e.a=function(t,e){if((n=(t=e?t.toExponential(e-1):t.toExponential()).indexOf("e"))<0)return null;var n,r=t.slice(0,n);return[r.length>1?r[0]+r.slice(2):r,+t.slice(n+1)]}},"../../../../d3-format/src/formatDefault.js":function(t,e,n){"use strict";e.a=function(t,e){t=t.toPrecision(e);t:for(var n,r=t.length,i=1,o=-1;i<r;++i)switch(t[i]){case".":o=n=i;break;case"0":0===o&&(o=i),n=i;break;case"e":break t;default:o>0&&(o=0)}return o>0?t.slice(0,o)+t.slice(n+1):t}},"../../../../d3-format/src/formatGroup.js":function(t,e,n){"use strict";e.a=function(t,e){return function(n,r){for(var i=n.length,o=[],s=0,a=t[0],c=0;i>0&&a>0&&(c+a+1>r&&(a=Math.max(1,r-c)),o.push(n.substring(i-=a,i+a)),!((c+=a+1)>r));)a=t[s=(s+1)%t.length];return o.reverse().join(e)}}},"../../../../d3-format/src/formatNumerals.js":function(t,e,n){"use strict";e.a=function(t){return function(e){return e.replace(/[0-9]/g,function(e){return t[+e]})}}},"../../../../d3-format/src/formatPrefixAuto.js":function(t,e,n){"use strict";n.d(e,"b",function(){return r});var r,i=n("../../../../d3-format/src/formatDecimal.js");e.a=function(t,e){var n=Object(i.a)(t,e);if(!n)return t+"";var o=n[0],s=n[1],a=s-(r=3*Math.max(-8,Math.min(8,Math.floor(s/3))))+1,c=o.length;return a===c?o:a>c?o+new Array(a-c+1).join("0"):a>0?o.slice(0,a)+"."+o.slice(a):"0."+new Array(1-a).join("0")+Object(i.a)(t,Math.max(0,e+a-1))[0]}},"../../../../d3-format/src/formatRounded.js":function(t,e,n){"use strict";var r=n("../../../../d3-format/src/formatDecimal.js");e.a=function(t,e){var n=Object(r.a)(t,e);if(!n)return t+"";var i=n[0],o=n[1];return o<0?"0."+new Array(-o).join("0")+i:i.length>o+1?i.slice(0,o+1)+"."+i.slice(o+1):i+new Array(o-i.length+2).join("0")}},"../../../../d3-format/src/formatSpecifier.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../d3-format/src/formatTypes.js"),i=/^(?:(.)?([<>=^]))?([+\-\( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?([a-z%])?$/i;function o(t){return new s(t)}function s(t){if(!(e=i.exec(t)))throw new Error("invalid format: "+t);var e,n=e[1]||" ",o=e[2]||">",s=e[3]||"-",a=e[4]||"",c=!!e[5],u=e[6]&&+e[6],l=!!e[7],p=e[8]&&+e[8].slice(1),d=e[9]||"";"n"===d?(l=!0,d="g"):r.a[d]||(d=""),(c||"0"===n&&"="===o)&&(c=!0,n="0",o="="),this.fill=n,this.align=o,this.sign=s,this.symbol=a,this.zero=c,this.width=u,this.comma=l,this.precision=p,this.type=d}o.prototype=s.prototype,s.prototype.toString=function(){return this.fill+this.align+this.sign+this.symbol+(this.zero?"0":"")+(null==this.width?"":Math.max(1,0|this.width))+(this.comma?",":"")+(null==this.precision?"":"."+Math.max(0,0|this.precision))+this.type}},"../../../../d3-format/src/formatTypes.js":function(t,e,n){"use strict";var r=n("../../../../d3-format/src/formatDefault.js"),i=n("../../../../d3-format/src/formatPrefixAuto.js"),o=n("../../../../d3-format/src/formatRounded.js");e.a={"":r.a,"%":function(t,e){return(100*t).toFixed(e)},b:function(t){return Math.round(t).toString(2)},c:function(t){return t+""},d:function(t){return Math.round(t).toString(10)},e:function(t,e){return t.toExponential(e)},f:function(t,e){return t.toFixed(e)},g:function(t,e){return t.toPrecision(e)},o:function(t){return Math.round(t).toString(8)},p:function(t,e){return Object(o.a)(100*t,e)},r:o.a,s:i.a,X:function(t){return Math.round(t).toString(16).toUpperCase()},x:function(t){return Math.round(t).toString(16)}}},"../../../../d3-format/src/identity.js":function(t,e,n){"use strict";e.a=function(t){return t}},"../../../../d3-format/src/locale.js":function(t,e,n){"use strict";var r=n("../../../../d3-format/src/exponent.js"),i=n("../../../../d3-format/src/formatGroup.js"),o=n("../../../../d3-format/src/formatNumerals.js"),s=n("../../../../d3-format/src/formatSpecifier.js"),a=n("../../../../d3-format/src/formatTypes.js"),c=n("../../../../d3-format/src/formatPrefixAuto.js"),u=n("../../../../d3-format/src/identity.js"),l=["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"];e.a=function(t){var e=t.grouping&&t.thousands?Object(i.a)(t.grouping,t.thousands):u.a,n=t.currency,p=t.decimal,d=t.numerals?Object(o.a)(t.numerals):u.a,h=t.percent||"%";function f(t){var r=(t=Object(s.a)(t)).fill,i=t.align,o=t.sign,u=t.symbol,f=t.zero,m=t.width,g=t.comma,v=t.precision,y=t.type,b="$"===u?n[0]:"#"===u&&/[boxX]/.test(y)?"0"+y.toLowerCase():"",_="$"===u?n[1]:/[%p]/.test(y)?h:"",j=a.a[y],x=!y||/[defgprs%]/.test(y);function w(t){var n,s,a,u=b,h=_;if("c"===y)h=j(t)+h,t="";else{var w=(t=+t)<0;if(t=j(Math.abs(t),v),w&&0==+t&&(w=!1),u=(w?"("===o?o:"-":"-"===o||"("===o?"":o)+u,h=("s"===y?l[8+c.b/3]:"")+h+(w&&"("===o?")":""),x)for(n=-1,s=t.length;++n<s;)if(48>(a=t.charCodeAt(n))||a>57){h=(46===a?p+t.slice(n+1):t.slice(n))+h,t=t.slice(0,n);break}}g&&!f&&(t=e(t,1/0));var O=u.length+t.length+h.length,C=O<m?new Array(m-O+1).join(r):"";switch(g&&f&&(t=e(C+t,C.length?m-h.length:1/0),C=""),i){case"<":t=u+t+h+C;break;case"=":t=u+C+t+h;break;case"^":t=C.slice(0,O=C.length>>1)+u+t+h+C.slice(O);break;default:t=C+u+t+h}return d(t)}return v=null==v?y?6:12:/[gprs]/.test(y)?Math.max(1,Math.min(21,v)):Math.max(0,Math.min(20,v)),w.toString=function(){return t+""},w}return{format:f,formatPrefix:function(t,e){var n=f(((t=Object(s.a)(t)).type="f",t)),i=3*Math.max(-8,Math.min(8,Math.floor(Object(r.a)(e)/3))),o=Math.pow(10,-i),a=l[8+i/3];return function(t){return n(o*t)+a}}}}},"../../../../d3-format/src/precisionFixed.js":function(t,e,n){"use strict";var r=n("../../../../d3-format/src/exponent.js");e.a=function(t){return Math.max(0,-Object(r.a)(Math.abs(t)))}},"../../../../d3-format/src/precisionPrefix.js":function(t,e,n){"use strict";var r=n("../../../../d3-format/src/exponent.js");e.a=function(t,e){return Math.max(0,3*Math.max(-8,Math.min(8,Math.floor(Object(r.a)(e)/3)))-Object(r.a)(Math.abs(t)))}},"../../../../d3-format/src/precisionRound.js":function(t,e,n){"use strict";var r=n("../../../../d3-format/src/exponent.js");e.a=function(t,e){return t=Math.abs(t),e=Math.abs(e)-t,Math.max(0,Object(r.a)(e)-Object(r.a)(t))+1}},"../../../../d3-geo/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/area.js");n.d(e,"c",function(){return r.c});var i=n("../../../../d3-geo/src/bounds.js");n.d(e,"f",function(){return i.a});var o=n("../../../../d3-geo/src/centroid.js");n.d(e,"g",function(){return o.a});n("../../../../d3-geo/src/circle.js"),n("../../../../d3-geo/src/clip/antimeridian.js"),n("../../../../d3-geo/src/clip/circle.js"),n("../../../../d3-geo/src/clip/extent.js"),n("../../../../d3-geo/src/clip/rectangle.js"),n("../../../../d3-geo/src/contains.js"),n("../../../../d3-geo/src/distance.js");var s=n("../../../../d3-geo/src/graticule.js");n.d(e,"m",function(){return s.a});n("../../../../d3-geo/src/interpolate.js"),n("../../../../d3-geo/src/length.js");var a=n("../../../../d3-geo/src/path/index.js");n.d(e,"p",function(){return a.a});var c=n("../../../../d3-geo/src/projection/albers.js");n.d(e,"a",function(){return c.a});var u=n("../../../../d3-geo/src/projection/albersUsa.js");n.d(e,"b",function(){return u.a});var l=n("../../../../d3-geo/src/projection/azimuthalEqualArea.js");n.d(e,"d",function(){return l.a});var p=n("../../../../d3-geo/src/projection/azimuthalEquidistant.js");n.d(e,"e",function(){return p.a});var d=n("../../../../d3-geo/src/projection/conicConformal.js");n.d(e,"h",function(){return d.a});var h=n("../../../../d3-geo/src/projection/conicEqualArea.js");n.d(e,"i",function(){return h.a});var f=n("../../../../d3-geo/src/projection/conicEquidistant.js");n.d(e,"j",function(){return f.a});var m=n("../../../../d3-geo/src/projection/equirectangular.js");n.d(e,"k",function(){return m.a});var g=n("../../../../d3-geo/src/projection/gnomonic.js");n.d(e,"l",function(){return g.a});n("../../../../d3-geo/src/projection/identity.js"),n("../../../../d3-geo/src/projection/index.js");var v=n("../../../../d3-geo/src/projection/mercator.js");n.d(e,"n",function(){return v.a});n("../../../../d3-geo/src/projection/naturalEarth1.js");var y=n("../../../../d3-geo/src/projection/orthographic.js");n.d(e,"o",function(){return y.a});var b=n("../../../../d3-geo/src/projection/stereographic.js");n.d(e,"q",function(){return b.a});var _=n("../../../../d3-geo/src/projection/transverseMercator.js");n.d(e,"r",function(){return _.a});n("../../../../d3-geo/src/rotation.js"),n("../../../../d3-geo/src/stream.js"),n("../../../../d3-geo/src/transform.js")},"../../../../d3-geo/src/adder.js":function(t,e,n){"use strict";function r(){this.reset()}e.a=function(){return new r},r.prototype={constructor:r,reset:function(){this.s=this.t=0},add:function(t){o(i,t,this.t),o(this,i.s,this.s),this.s?this.t+=i.t:this.s=i.t},valueOf:function(){return this.s}};var i=new r;function o(t,e,n){var r=t.s=e+n,i=r-e,o=r-i;t.t=e-o+(n-i)}},"../../../../d3-geo/src/area.js":function(t,e,n){"use strict";n.d(e,"a",function(){return d}),n.d(e,"b",function(){return f});var r,i,o,s,a,c=n("../../../../d3-geo/src/adder.js"),u=n("../../../../d3-geo/src/math.js"),l=n("../../../../d3-geo/src/noop.js"),p=n("../../../../d3-geo/src/stream.js"),d=Object(c.a)(),h=Object(c.a)(),f={point:l.a,lineStart:l.a,lineEnd:l.a,polygonStart:function(){d.reset(),f.lineStart=m,f.lineEnd=g},polygonEnd:function(){var t=+d;h.add(t<0?u.w+t:t),this.lineStart=this.lineEnd=this.point=l.a},sphere:function(){h.add(u.w)}};function m(){f.point=v}function g(){y(r,i)}function v(t,e){f.point=y,r=t,i=e,t*=u.r,e*=u.r,o=t,s=Object(u.g)(e=e/2+u.q),a=Object(u.t)(e)}function y(t,e){t*=u.r,e=(e*=u.r)/2+u.q;var n=t-o,r=n>=0?1:-1,i=r*n,c=Object(u.g)(e),l=Object(u.t)(e),p=a*l,h=s*c+p*Object(u.g)(i),f=p*r*Object(u.t)(i);d.add(Object(u.e)(f,h)),o=t,s=c,a=l}e.c=function(t){return h.reset(),Object(p.a)(t,f),2*h}},"../../../../d3-geo/src/bounds.js":function(t,e,n){"use strict";var r,i,o,s,a,c,u,l,p,d,h=n("../../../../d3-geo/src/adder.js"),f=n("../../../../d3-geo/src/area.js"),m=n("../../../../d3-geo/src/cartesian.js"),g=n("../../../../d3-geo/src/math.js"),v=n("../../../../d3-geo/src/stream.js"),y=Object(h.a)(),b={point:_,lineStart:x,lineEnd:w,polygonStart:function(){b.point=O,b.lineStart=C,b.lineEnd=k,y.reset(),f.b.polygonStart()},polygonEnd:function(){f.b.polygonEnd(),b.point=_,b.lineStart=x,b.lineEnd=w,f.a<0?(r=-(o=180),i=-(s=90)):y>g.i?s=90:y<-g.i&&(i=-90),d[0]=r,d[1]=o}};function _(t,e){p.push(d=[r=t,o=t]),e<i&&(i=e),e>s&&(s=e)}function j(t,e){var n=Object(m.a)([t*g.r,e*g.r]);if(l){var c=Object(m.c)(l,n),u=[c[1],-c[0],0],h=Object(m.c)(u,c);Object(m.e)(h),h=Object(m.g)(h);var f,v=t-a,y=v>0?1:-1,b=h[0]*g.h*y,_=Object(g.a)(v)>180;_^(y*a<b&&b<y*t)?(f=h[1]*g.h)>s&&(s=f):_^(y*a<(b=(b+360)%360-180)&&b<y*t)?(f=-h[1]*g.h)<i&&(i=f):(e<i&&(i=e),e>s&&(s=e)),_?t<a?E(r,t)>E(r,o)&&(o=t):E(t,o)>E(r,o)&&(r=t):o>=r?(t<r&&(r=t),t>o&&(o=t)):t>a?E(r,t)>E(r,o)&&(o=t):E(t,o)>E(r,o)&&(r=t)}else p.push(d=[r=t,o=t]);e<i&&(i=e),e>s&&(s=e),l=n,a=t}function x(){b.point=j}function w(){d[0]=r,d[1]=o,b.point=_,l=null}function O(t,e){if(l){var n=t-a;y.add(Object(g.a)(n)>180?n+(n>0?360:-360):n)}else c=t,u=e;f.b.point(t,e),j(t,e)}function C(){f.b.lineStart()}function k(){O(c,u),f.b.lineEnd(),Object(g.a)(y)>g.i&&(r=-(o=180)),d[0]=r,d[1]=o,l=null}function E(t,e){return(e-=t)<0?e+360:e}function S(t,e){return t[0]-e[0]}function P(t,e){return t[0]<=t[1]?t[0]<=e&&e<=t[1]:e<t[0]||t[1]<e}e.a=function(t){var e,n,a,c,u,l,h;if(s=o=-(r=i=1/0),p=[],Object(v.a)(t,b),n=p.length){for(p.sort(S),e=1,u=[a=p[0]];e<n;++e)P(a,(c=p[e])[0])||P(a,c[1])?(E(a[0],c[1])>E(a[0],a[1])&&(a[1]=c[1]),E(c[0],a[1])>E(a[0],a[1])&&(a[0]=c[0])):u.push(a=c);for(l=-1/0,e=0,a=u[n=u.length-1];e<=n;a=c,++e)c=u[e],(h=E(a[1],c[0]))>l&&(l=h,r=c[0],o=a[1])}return p=d=null,r===1/0||i===1/0?[[NaN,NaN],[NaN,NaN]]:[[r,i],[o,s]]}},"../../../../d3-geo/src/cartesian.js":function(t,e,n){"use strict";e.g=function(t){return[Object(r.e)(t[1],t[0]),Object(r.c)(t[2])]},e.a=function(t){var e=t[0],n=t[1],i=Object(r.g)(n);return[i*Object(r.g)(e),i*Object(r.t)(e),Object(r.t)(n)]},e.d=function(t,e){return t[0]*e[0]+t[1]*e[1]+t[2]*e[2]},e.c=function(t,e){return[t[1]*e[2]-t[2]*e[1],t[2]*e[0]-t[0]*e[2],t[0]*e[1]-t[1]*e[0]]},e.b=function(t,e){t[0]+=e[0],t[1]+=e[1],t[2]+=e[2]},e.f=function(t,e){return[t[0]*e,t[1]*e,t[2]*e]},e.e=function(t){var e=Object(r.u)(t[0]*t[0]+t[1]*t[1]+t[2]*t[2]);t[0]/=e,t[1]/=e,t[2]/=e};var r=n("../../../../d3-geo/src/math.js")},"../../../../d3-geo/src/centroid.js":function(t,e,n){"use strict";var r,i,o,s,a,c,u,l,p,d,h,f,m,g,v,y,b=n("../../../../d3-geo/src/math.js"),_=n("../../../../d3-geo/src/noop.js"),j=n("../../../../d3-geo/src/stream.js"),x={sphere:_.a,point:w,lineStart:C,lineEnd:S,polygonStart:function(){x.lineStart=P,x.lineEnd=M},polygonEnd:function(){x.lineStart=C,x.lineEnd=S}};function w(t,e){t*=b.r,e*=b.r;var n=Object(b.g)(e);O(n*Object(b.g)(t),n*Object(b.t)(t),Object(b.t)(e))}function O(t,e,n){o+=(t-o)/++r,s+=(e-s)/r,a+=(n-a)/r}function C(){x.point=k}function k(t,e){t*=b.r,e*=b.r;var n=Object(b.g)(e);g=n*Object(b.g)(t),v=n*Object(b.t)(t),y=Object(b.t)(e),x.point=E,O(g,v,y)}function E(t,e){t*=b.r,e*=b.r;var n=Object(b.g)(e),r=n*Object(b.g)(t),o=n*Object(b.t)(t),s=Object(b.t)(e),a=Object(b.e)(Object(b.u)((a=v*s-y*o)*a+(a=y*r-g*s)*a+(a=g*o-v*r)*a),g*r+v*o+y*s);i+=a,c+=a*(g+(g=r)),u+=a*(v+(v=o)),l+=a*(y+(y=s)),O(g,v,y)}function S(){x.point=w}function P(){x.point=A}function M(){T(f,m),x.point=w}function A(t,e){f=t,m=e,t*=b.r,e*=b.r,x.point=T;var n=Object(b.g)(e);g=n*Object(b.g)(t),v=n*Object(b.t)(t),y=Object(b.t)(e),O(g,v,y)}function T(t,e){t*=b.r,e*=b.r;var n=Object(b.g)(e),r=n*Object(b.g)(t),o=n*Object(b.t)(t),s=Object(b.t)(e),a=v*s-y*o,f=y*r-g*s,m=g*o-v*r,_=Object(b.u)(a*a+f*f+m*m),j=Object(b.c)(_),x=_&&-j/_;p+=x*a,d+=x*f,h+=x*m,i+=j,c+=j*(g+(g=r)),u+=j*(v+(v=o)),l+=j*(y+(y=s)),O(g,v,y)}e.a=function(t){r=i=o=s=a=c=u=l=p=d=h=0,Object(j.a)(t,x);var e=p,n=d,f=h,m=e*e+n*n+f*f;return m<b.j&&(e=c,n=u,f=l,i<b.i&&(e=o,n=s,f=a),(m=e*e+n*n+f*f)<b.j)?[NaN,NaN]:[Object(b.e)(n,e)*b.h,Object(b.c)(f/Object(b.u)(m))*b.h]}},"../../../../d3-geo/src/circle.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../d3-geo/src/cartesian.js"),i=(n("../../../../d3-geo/src/constant.js"),n("../../../../d3-geo/src/math.js"));n("../../../../d3-geo/src/rotation.js");function o(t,e,n,o,a,c){if(n){var u=Object(i.g)(e),l=Object(i.t)(e),p=o*n;null==a?(a=e+o*i.w,c=e-p/2):(a=s(u,a),c=s(u,c),(o>0?a<c:a>c)&&(a+=o*i.w));for(var d,h=a;o>0?h>c:h<c;h-=p)d=Object(r.g)([u,-l*Object(i.g)(h),-l*Object(i.t)(h)]),t.point(d[0],d[1])}}function s(t,e){(e=Object(r.a)(e))[0]-=t,Object(r.e)(e);var n=Object(i.b)(-e[1]);return((-e[2]<0?-n:n)+i.w-i.i)%i.w}},"../../../../d3-geo/src/clip/antimeridian.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/clip/index.js"),i=n("../../../../d3-geo/src/math.js");e.a=Object(r.a)(function(){return!0},function(t){var e,n=NaN,r=NaN,o=NaN;return{lineStart:function(){t.lineStart(),e=1},point:function(s,a){var c,u,l,p,d,h,f,m=s>0?i.o:-i.o,g=Object(i.a)(s-n);Object(i.a)(g-i.o)<i.i?(t.point(n,r=(r+a)/2>0?i.l:-i.l),t.point(o,r),t.lineEnd(),t.lineStart(),t.point(m,r),t.point(s,r),e=0):o!==m&&g>=i.o&&(Object(i.a)(n-o)<i.i&&(n-=o*i.i),Object(i.a)(s-m)<i.i&&(s-=m*i.i),c=n,u=r,l=s,p=a,f=Object(i.t)(c-l),r=Object(i.a)(f)>i.i?Object(i.d)((Object(i.t)(u)*(h=Object(i.g)(p))*Object(i.t)(l)-Object(i.t)(p)*(d=Object(i.g)(u))*Object(i.t)(c))/(d*h*f)):(u+p)/2,t.point(o,r),t.lineEnd(),t.lineStart(),t.point(m,r),e=0),t.point(n=s,r=a),o=m},lineEnd:function(){t.lineEnd(),n=r=NaN},clean:function(){return 2-e}}},function(t,e,n,r){var o;if(null==t)o=n*i.l,r.point(-i.o,o),r.point(0,o),r.point(i.o,o),r.point(i.o,0),r.point(i.o,-o),r.point(0,-o),r.point(-i.o,-o),r.point(-i.o,0),r.point(-i.o,o);else if(Object(i.a)(t[0]-e[0])>i.i){var s=t[0]<e[0]?i.o:-i.o;o=n*s/2,r.point(-s,o),r.point(0,o),r.point(s,o)}else r.point(e[0],e[1])},[-i.o,-i.l])},"../../../../d3-geo/src/clip/buffer.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/noop.js");e.a=function(){var t,e=[];return{point:function(e,n){t.push([e,n])},lineStart:function(){e.push(t=[])},lineEnd:r.a,rejoin:function(){e.length>1&&e.push(e.pop().concat(e.shift()))},result:function(){var n=e;return e=[],t=null,n}}}},"../../../../d3-geo/src/clip/circle.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/cartesian.js"),i=n("../../../../d3-geo/src/circle.js"),o=n("../../../../d3-geo/src/math.js"),s=n("../../../../d3-geo/src/pointEqual.js"),a=n("../../../../d3-geo/src/clip/index.js");e.a=function(t){var e=Object(o.g)(t),n=6*o.r,c=e>0,u=Object(o.a)(e)>o.i;function l(t,n){return Object(o.g)(t)*Object(o.g)(n)>e}function p(t,n,i){var s=Object(r.a)(t),a=Object(r.a)(n),c=[1,0,0],u=Object(r.c)(s,a),l=Object(r.d)(u,u),p=u[0],d=l-p*p;if(!d)return!i&&t;var h=e*l/d,f=-e*p/d,m=Object(r.c)(c,u),g=Object(r.f)(c,h),v=Object(r.f)(u,f);Object(r.b)(g,v);var y=m,b=Object(r.d)(g,y),_=Object(r.d)(y,y),j=b*b-_*(Object(r.d)(g,g)-1);if(!(j<0)){var x=Object(o.u)(j),w=Object(r.f)(y,(-b-x)/_);if(Object(r.b)(w,g),w=Object(r.g)(w),!i)return w;var O,C=t[0],k=n[0],E=t[1],S=n[1];k<C&&(O=C,C=k,k=O);var P=k-C,M=Object(o.a)(P-o.o)<o.i,A=M||P<o.i;if(!M&&S<E&&(O=E,E=S,S=O),A?M?E+S>0^w[1]<(Object(o.a)(w[0]-C)<o.i?E:S):E<=w[1]&&w[1]<=S:P>o.o^(C<=w[0]&&w[0]<=k)){var T=Object(r.f)(y,(-b+x)/_);return Object(r.b)(T,g),[w,Object(r.g)(T)]}}}function d(e,n){var r=c?t:o.o-t,i=0;return e<-r?i|=1:e>r&&(i|=2),n<-r?i|=4:n>r&&(i|=8),i}return Object(a.a)(l,function(t){var e,n,r,i,a;return{lineStart:function(){i=r=!1,a=1},point:function(h,f){var m,g=[h,f],v=l(h,f),y=c?v?0:d(h,f):v?d(h+(h<0?o.o:-o.o),f):0;if(!e&&(i=r=v)&&t.lineStart(),v!==r&&(!(m=p(e,g))||Object(s.a)(e,m)||Object(s.a)(g,m))&&(g[0]+=o.i,g[1]+=o.i,v=l(g[0],g[1])),v!==r)a=0,v?(t.lineStart(),m=p(g,e),t.point(m[0],m[1])):(m=p(e,g),t.point(m[0],m[1]),t.lineEnd()),e=m;else if(u&&e&&c^v){var b;y&n||!(b=p(g,e,!0))||(a=0,c?(t.lineStart(),t.point(b[0][0],b[0][1]),t.point(b[1][0],b[1][1]),t.lineEnd()):(t.point(b[1][0],b[1][1]),t.lineEnd(),t.lineStart(),t.point(b[0][0],b[0][1])))}!v||e&&Object(s.a)(e,g)||t.point(g[0],g[1]),e=g,r=v,n=y},lineEnd:function(){r&&t.lineEnd(),e=null},clean:function(){return a|(i&&r)<<1}}},function(e,r,o,s){Object(i.a)(s,t,n,o,e,r)},c?[0,-t]:[-o.o,t-o.o])}},"../../../../d3-geo/src/clip/extent.js":function(t,e,n){"use strict";n("../../../../d3-geo/src/clip/rectangle.js")},"../../../../d3-geo/src/clip/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/clip/buffer.js"),i=n("../../../../d3-geo/src/clip/rejoin.js"),o=n("../../../../d3-geo/src/math.js"),s=n("../../../../d3-geo/src/polygonContains.js"),a=n("../../../../d3-array/index.js");function c(t){return t.length>1}function u(t,e){return((t=t.x)[0]<0?t[1]-o.l-o.i:o.l-t[1])-((e=e.x)[0]<0?e[1]-o.l-o.i:o.l-e[1])}e.a=function(t,e,n,o){return function(l){var p,d,h,f=e(l),m=Object(r.a)(),g=e(m),v=!1,y={point:b,lineStart:j,lineEnd:x,polygonStart:function(){y.point=w,y.lineStart=O,y.lineEnd=C,d=[],p=[]},polygonEnd:function(){y.point=b,y.lineStart=j,y.lineEnd=x,d=Object(a.j)(d);var t=Object(s.a)(p,o);d.length?(v||(l.polygonStart(),v=!0),Object(i.a)(d,u,t,n,l)):t&&(v||(l.polygonStart(),v=!0),l.lineStart(),n(null,null,1,l),l.lineEnd()),v&&(l.polygonEnd(),v=!1),d=p=null},sphere:function(){l.polygonStart(),l.lineStart(),n(null,null,1,l),l.lineEnd(),l.polygonEnd()}};function b(e,n){t(e,n)&&l.point(e,n)}function _(t,e){f.point(t,e)}function j(){y.point=_,f.lineStart()}function x(){y.point=b,f.lineEnd()}function w(t,e){h.push([t,e]),g.point(t,e)}function O(){g.lineStart(),h=[]}function C(){w(h[0][0],h[0][1]),g.lineEnd();var t,e,n,r,i=g.clean(),o=m.result(),s=o.length;if(h.pop(),p.push(h),h=null,s)if(1&i){if((e=(n=o[0]).length-1)>0){for(v||(l.polygonStart(),v=!0),l.lineStart(),t=0;t<e;++t)l.point((r=n[t])[0],r[1]);l.lineEnd()}}else s>1&&2&i&&o.push(o.pop().concat(o.shift())),d.push(o.filter(c))}return y}}},"../../../../d3-geo/src/clip/line.js":function(t,e,n){"use strict";e.a=function(t,e,n,r,i,o){var s,a=t[0],c=t[1],u=0,l=1,p=e[0]-a,d=e[1]-c;if(s=n-a,p||!(s>0)){if(s/=p,p<0){if(s<u)return;s<l&&(l=s)}else if(p>0){if(s>l)return;s>u&&(u=s)}if(s=i-a,p||!(s<0)){if(s/=p,p<0){if(s>l)return;s>u&&(u=s)}else if(p>0){if(s<u)return;s<l&&(l=s)}if(s=r-c,d||!(s>0)){if(s/=d,d<0){if(s<u)return;s<l&&(l=s)}else if(d>0){if(s>l)return;s>u&&(u=s)}if(s=o-c,d||!(s<0)){if(s/=d,d<0){if(s>l)return;s>u&&(u=s)}else if(d>0){if(s<u)return;s<l&&(l=s)}return u>0&&(t[0]=a+u*p,t[1]=c+u*d),l<1&&(e[0]=a+l*p,e[1]=c+l*d),!0}}}}}},"../../../../d3-geo/src/clip/rectangle.js":function(t,e,n){"use strict";e.a=function(t,e,n,l){function p(r,i){return t<=r&&r<=n&&e<=i&&i<=l}function d(r,i,o,s){var a=0,c=0;if(null==r||(a=h(r,o))!==(c=h(i,o))||m(r,i)<0^o>0)do{s.point(0===a||3===a?t:n,a>1?l:e)}while((a=(a+o+4)%4)!==c);else s.point(i[0],i[1])}function h(i,o){return Object(r.a)(i[0]-t)<r.i?o>0?0:3:Object(r.a)(i[0]-n)<r.i?o>0?2:1:Object(r.a)(i[1]-e)<r.i?o>0?1:0:o>0?3:2}function f(t,e){return m(t.x,e.x)}function m(t,e){var n=h(t,1),r=h(e,1);return n!==r?n-r:0===n?e[1]-t[1]:1===n?t[0]-e[0]:2===n?t[1]-e[1]:e[0]-t[0]}return function(r){var h,m,g,v,y,b,_,j,x,w,O,C=r,k=Object(i.a)(),E={point:S,lineStart:function(){E.point=P,m&&m.push(g=[]);w=!0,x=!1,_=j=NaN},lineEnd:function(){h&&(P(v,y),b&&x&&k.rejoin(),h.push(k.result()));E.point=S,x&&C.lineEnd()},polygonStart:function(){C=k,h=[],m=[],O=!0},polygonEnd:function(){var e=function(){for(var e=0,n=0,r=m.length;n<r;++n)for(var i,o,s=m[n],a=1,c=s.length,u=s[0],p=u[0],d=u[1];a<c;++a)i=p,o=d,u=s[a],p=u[0],d=u[1],o<=l?d>l&&(p-i)*(l-o)>(d-o)*(t-i)&&++e:d<=l&&(p-i)*(l-o)<(d-o)*(t-i)&&--e;return e}(),n=O&&e,i=(h=Object(a.j)(h)).length;(n||i)&&(r.polygonStart(),n&&(r.lineStart(),d(null,null,1,r),r.lineEnd()),i&&Object(s.a)(h,f,e,d,r),r.polygonEnd());C=r,h=m=g=null}};function S(t,e){p(t,e)&&C.point(t,e)}function P(r,i){var s=p(r,i);if(m&&g.push([r,i]),w)v=r,y=i,b=s,w=!1,s&&(C.lineStart(),C.point(r,i));else if(s&&x)C.point(r,i);else{var a=[_=Math.max(u,Math.min(c,_)),j=Math.max(u,Math.min(c,j))],d=[r=Math.max(u,Math.min(c,r)),i=Math.max(u,Math.min(c,i))];Object(o.a)(a,d,t,e,n,l)?(x||(C.lineStart(),C.point(a[0],a[1])),C.point(d[0],d[1]),s||C.lineEnd(),O=!1):s&&(C.lineStart(),C.point(r,i),O=!1)}_=r,j=i,x=s}return E}};var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/clip/buffer.js"),o=n("../../../../d3-geo/src/clip/line.js"),s=n("../../../../d3-geo/src/clip/rejoin.js"),a=n("../../../../d3-array/index.js"),c=1e9,u=-c},"../../../../d3-geo/src/clip/rejoin.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/pointEqual.js");function i(t,e,n,r){this.x=t,this.z=e,this.o=n,this.e=r,this.v=!1,this.n=this.p=null}function o(t){if(e=t.length){for(var e,n,r=0,i=t[0];++r<e;)i.n=n=t[r],n.p=i,i=n;i.n=n=t[0],n.p=i}}e.a=function(t,e,n,s,a){var c,u,l=[],p=[];if(t.forEach(function(t){if(!((e=t.length-1)<=0)){var e,n,o=t[0],s=t[e];if(Object(r.a)(o,s)){for(a.lineStart(),c=0;c<e;++c)a.point((o=t[c])[0],o[1]);a.lineEnd()}else l.push(n=new i(o,t,null,!0)),p.push(n.o=new i(o,null,n,!1)),l.push(n=new i(s,t,null,!1)),p.push(n.o=new i(s,null,n,!0))}}),l.length){for(p.sort(e),o(l),o(p),c=0,u=p.length;c<u;++c)p[c].e=n=!n;for(var d,h,f=l[0];;){for(var m=f,g=!0;m.v;)if((m=m.n)===f)return;d=m.z,a.lineStart();do{if(m.v=m.o.v=!0,m.e){if(g)for(c=0,u=d.length;c<u;++c)a.point((h=d[c])[0],h[1]);else s(m.x,m.n.x,1,a);m=m.n}else{if(g)for(d=m.p.z,c=d.length-1;c>=0;--c)a.point((h=d[c])[0],h[1]);else s(m.x,m.p.x,-1,a);m=m.p}d=(m=m.o).z,g=!g}while(!m.v);a.lineEnd()}}}},"../../../../d3-geo/src/compose.js":function(t,e,n){"use strict";e.a=function(t,e){function n(n,r){return n=t(n,r),e(n[0],n[1])}return t.invert&&e.invert&&(n.invert=function(n,r){return(n=e.invert(n,r))&&t.invert(n[0],n[1])}),n}},"../../../../d3-geo/src/constant.js":function(t,e,n){"use strict";e.a=function(t){return function(){return t}}},"../../../../d3-geo/src/contains.js":function(t,e,n){"use strict";n("../../../../d3-geo/src/polygonContains.js"),n("../../../../d3-geo/src/distance.js"),n("../../../../d3-geo/src/math.js")},"../../../../d3-geo/src/distance.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/length.js"),i=[null,null],o={type:"LineString",coordinates:i};e.a=function(t,e){return i[0]=t,i[1]=e,Object(r.a)(o)}},"../../../../d3-geo/src/graticule.js":function(t,e,n){"use strict";e.a=a;var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-geo/src/math.js");function o(t,e,n){var o=Object(r.n)(t,e-i.i,n).concat(e);return function(t){return o.map(function(e){return[t,e]})}}function s(t,e,n){var o=Object(r.n)(t,e-i.i,n).concat(e);return function(t){return o.map(function(e){return[e,t]})}}function a(){var t,e,n,a,c,u,l,p,d,h,f,m,g=10,v=g,y=90,b=360,_=2.5;function j(){return{type:"MultiLineString",coordinates:x()}}function x(){return Object(r.n)(Object(i.f)(a/y)*y,n,y).map(f).concat(Object(r.n)(Object(i.f)(p/b)*b,l,b).map(m)).concat(Object(r.n)(Object(i.f)(e/g)*g,t,g).filter(function(t){return Object(i.a)(t%y)>i.i}).map(d)).concat(Object(r.n)(Object(i.f)(u/v)*v,c,v).filter(function(t){return Object(i.a)(t%b)>i.i}).map(h))}return j.lines=function(){return x().map(function(t){return{type:"LineString",coordinates:t}})},j.outline=function(){return{type:"Polygon",coordinates:[f(a).concat(m(l).slice(1),f(n).reverse().slice(1),m(p).reverse().slice(1))]}},j.extent=function(t){return arguments.length?j.extentMajor(t).extentMinor(t):j.extentMinor()},j.extentMajor=function(t){return arguments.length?(a=+t[0][0],n=+t[1][0],p=+t[0][1],l=+t[1][1],a>n&&(t=a,a=n,n=t),p>l&&(t=p,p=l,l=t),j.precision(_)):[[a,p],[n,l]]},j.extentMinor=function(n){return arguments.length?(e=+n[0][0],t=+n[1][0],u=+n[0][1],c=+n[1][1],e>t&&(n=e,e=t,t=n),u>c&&(n=u,u=c,c=n),j.precision(_)):[[e,u],[t,c]]},j.step=function(t){return arguments.length?j.stepMajor(t).stepMinor(t):j.stepMinor()},j.stepMajor=function(t){return arguments.length?(y=+t[0],b=+t[1],j):[y,b]},j.stepMinor=function(t){return arguments.length?(g=+t[0],v=+t[1],j):[g,v]},j.precision=function(r){return arguments.length?(_=+r,d=o(u,c,90),h=s(e,t,_),f=o(p,l,90),m=s(a,n,_),j):_},j.extentMajor([[-180,-90+i.i],[180,90-i.i]]).extentMinor([[-180,-80-i.i],[180,80+i.i]])}},"../../../../d3-geo/src/identity.js":function(t,e,n){"use strict";e.a=function(t){return t}},"../../../../d3-geo/src/interpolate.js":function(t,e,n){"use strict";n("../../../../d3-geo/src/math.js")},"../../../../d3-geo/src/length.js":function(t,e,n){"use strict";var r,i,o,s=n("../../../../d3-geo/src/adder.js"),a=n("../../../../d3-geo/src/math.js"),c=n("../../../../d3-geo/src/noop.js"),u=n("../../../../d3-geo/src/stream.js"),l=Object(s.a)(),p={sphere:c.a,point:c.a,lineStart:function(){p.point=h,p.lineEnd=d},lineEnd:c.a,polygonStart:c.a,polygonEnd:c.a};function d(){p.point=p.lineEnd=c.a}function h(t,e){t*=a.r,e*=a.r,r=t,i=Object(a.t)(e),o=Object(a.g)(e),p.point=f}function f(t,e){t*=a.r,e*=a.r;var n=Object(a.t)(e),s=Object(a.g)(e),c=Object(a.a)(t-r),u=Object(a.g)(c),p=s*Object(a.t)(c),d=o*n-i*s*u,h=i*n+o*s*u;l.add(Object(a.e)(Object(a.u)(p*p+d*d),h)),r=t,i=n,o=s}e.a=function(t){return l.reset(),Object(u.a)(t,p),+l}},"../../../../d3-geo/src/math.js":function(t,e,n){"use strict";n.d(e,"i",function(){return r}),n.d(e,"j",function(){return i}),n.d(e,"o",function(){return o}),n.d(e,"l",function(){return s}),n.d(e,"q",function(){return a}),n.d(e,"w",function(){return c}),n.d(e,"h",function(){return u}),n.d(e,"r",function(){return l}),n.d(e,"a",function(){return p}),n.d(e,"d",function(){return d}),n.d(e,"e",function(){return h}),n.d(e,"g",function(){return f}),n.d(e,"f",function(){return m}),n.d(e,"k",function(){return g}),n.d(e,"n",function(){return v}),n.d(e,"p",function(){return y}),n.d(e,"t",function(){return b}),n.d(e,"s",function(){return _}),n.d(e,"u",function(){return j}),n.d(e,"v",function(){return x}),e.b=function(t){return t>1?0:t<-1?o:Math.acos(t)},e.c=function(t){return t>1?s:t<-1?-s:Math.asin(t)},e.m=function(t){return(t=b(t/2))*t};var r=1e-6,i=1e-12,o=Math.PI,s=o/2,a=o/4,c=2*o,u=180/o,l=o/180,p=Math.abs,d=Math.atan,h=Math.atan2,f=Math.cos,m=Math.ceil,g=Math.exp,v=(Math.floor,Math.log),y=Math.pow,b=Math.sin,_=Math.sign||function(t){return t>0?1:t<0?-1:0},j=Math.sqrt,x=Math.tan},"../../../../d3-geo/src/noop.js":function(t,e,n){"use strict";e.a=function(){}},"../../../../d3-geo/src/path/area.js":function(t,e,n){"use strict";var r,i,o,s,a=n("../../../../d3-geo/src/adder.js"),c=n("../../../../d3-geo/src/math.js"),u=n("../../../../d3-geo/src/noop.js"),l=Object(a.a)(),p=Object(a.a)(),d={point:u.a,lineStart:u.a,lineEnd:u.a,polygonStart:function(){d.lineStart=h,d.lineEnd=g},polygonEnd:function(){d.lineStart=d.lineEnd=d.point=u.a,l.add(Object(c.a)(p)),p.reset()},result:function(){var t=l/2;return l.reset(),t}};function h(){d.point=f}function f(t,e){d.point=m,r=o=t,i=s=e}function m(t,e){p.add(s*t-o*e),o=t,s=e}function g(){m(r,i)}e.a=d},"../../../../d3-geo/src/path/bounds.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/noop.js"),i=1/0,o=i,s=-i,a=s,c={point:function(t,e){t<i&&(i=t);t>s&&(s=t);e<o&&(o=e);e>a&&(a=e)},lineStart:r.a,lineEnd:r.a,polygonStart:r.a,polygonEnd:r.a,result:function(){var t=[[i,o],[s,a]];return s=a=-(o=i=1/0),t}};e.a=c},"../../../../d3-geo/src/path/centroid.js":function(t,e,n){"use strict";var r,i,o,s,a=n("../../../../d3-geo/src/math.js"),c=0,u=0,l=0,p=0,d=0,h=0,f=0,m=0,g=0,v={point:y,lineStart:b,lineEnd:x,polygonStart:function(){v.lineStart=w,v.lineEnd=O},polygonEnd:function(){v.point=y,v.lineStart=b,v.lineEnd=x},result:function(){var t=g?[f/g,m/g]:h?[p/h,d/h]:l?[c/l,u/l]:[NaN,NaN];return c=u=l=p=d=h=f=m=g=0,t}};function y(t,e){c+=t,u+=e,++l}function b(){v.point=_}function _(t,e){v.point=j,y(o=t,s=e)}function j(t,e){var n=t-o,r=e-s,i=Object(a.u)(n*n+r*r);p+=i*(o+t)/2,d+=i*(s+e)/2,h+=i,y(o=t,s=e)}function x(){v.point=y}function w(){v.point=C}function O(){k(r,i)}function C(t,e){v.point=k,y(r=o=t,i=s=e)}function k(t,e){var n=t-o,r=e-s,i=Object(a.u)(n*n+r*r);p+=i*(o+t)/2,d+=i*(s+e)/2,h+=i,f+=(i=s*t-o*e)*(o+t),m+=i*(s+e),g+=3*i,y(o=t,s=e)}e.a=v},"../../../../d3-geo/src/path/context.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/noop.js");function o(t){this._context=t}o.prototype={_radius:4.5,pointRadius:function(t){return this._radius=t,this},polygonStart:function(){this._line=0},polygonEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){0===this._line&&this._context.closePath(),this._point=NaN},point:function(t,e){switch(this._point){case 0:this._context.moveTo(t,e),this._point=1;break;case 1:this._context.lineTo(t,e);break;default:this._context.moveTo(t+this._radius,e),this._context.arc(t,e,this._radius,0,r.w)}},result:i.a}},"../../../../d3-geo/src/path/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/identity.js"),i=n("../../../../d3-geo/src/stream.js"),o=n("../../../../d3-geo/src/path/area.js"),s=n("../../../../d3-geo/src/path/bounds.js"),a=n("../../../../d3-geo/src/path/centroid.js"),c=n("../../../../d3-geo/src/path/context.js"),u=n("../../../../d3-geo/src/path/measure.js"),l=n("../../../../d3-geo/src/path/string.js");e.a=function(t,e){var n,p,d=4.5;function h(t){return t&&("function"==typeof d&&p.pointRadius(+d.apply(this,arguments)),Object(i.a)(t,n(p))),p.result()}return h.area=function(t){return Object(i.a)(t,n(o.a)),o.a.result()},h.measure=function(t){return Object(i.a)(t,n(u.a)),u.a.result()},h.bounds=function(t){return Object(i.a)(t,n(s.a)),s.a.result()},h.centroid=function(t){return Object(i.a)(t,n(a.a)),a.a.result()},h.projection=function(e){return arguments.length?(n=null==e?(t=null,r.a):(t=e).stream,h):t},h.context=function(t){return arguments.length?(p=null==t?(e=null,new l.a):new c.a(e=t),"function"!=typeof d&&p.pointRadius(d),h):e},h.pointRadius=function(t){return arguments.length?(d="function"==typeof t?t:(p.pointRadius(+t),+t),h):d},h.projection(t).context(e)}},"../../../../d3-geo/src/path/measure.js":function(t,e,n){"use strict";var r,i,o,s,a,c=n("../../../../d3-geo/src/adder.js"),u=n("../../../../d3-geo/src/math.js"),l=n("../../../../d3-geo/src/noop.js"),p=Object(c.a)(),d={point:l.a,lineStart:function(){d.point=h},lineEnd:function(){r&&f(i,o),d.point=l.a},polygonStart:function(){r=!0},polygonEnd:function(){r=null},result:function(){var t=+p;return p.reset(),t}};function h(t,e){d.point=f,i=s=t,o=a=e}function f(t,e){s-=t,a-=e,p.add(Object(u.u)(s*s+a*a)),s=t,a=e}e.a=d},"../../../../d3-geo/src/path/string.js":function(t,e,n){"use strict";function r(){this._string=[]}function i(t){return"m0,"+t+"a"+t+","+t+" 0 1,1 0,"+-2*t+"a"+t+","+t+" 0 1,1 0,"+2*t+"z"}e.a=r,r.prototype={_radius:4.5,_circle:i(4.5),pointRadius:function(t){return(t=+t)!==this._radius&&(this._radius=t,this._circle=null),this},polygonStart:function(){this._line=0},polygonEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){0===this._line&&this._string.push("Z"),this._point=NaN},point:function(t,e){switch(this._point){case 0:this._string.push("M",t,",",e),this._point=1;break;case 1:this._string.push("L",t,",",e);break;default:null==this._circle&&(this._circle=i(this._radius)),this._string.push("M",t,",",e,this._circle)}},result:function(){if(this._string.length){var t=this._string.join("");return this._string=[],t}return null}}},"../../../../d3-geo/src/pointEqual.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js");e.a=function(t,e){return Object(r.a)(t[0]-e[0])<r.i&&Object(r.a)(t[1]-e[1])<r.i}},"../../../../d3-geo/src/polygonContains.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/adder.js"),i=n("../../../../d3-geo/src/cartesian.js"),o=n("../../../../d3-geo/src/math.js"),s=Object(r.a)();e.a=function(t,e){var n=e[0],r=e[1],a=[Object(o.t)(n),-Object(o.g)(n),0],c=0,u=0;s.reset();for(var l=0,p=t.length;l<p;++l)if(h=(d=t[l]).length)for(var d,h,f=d[h-1],m=f[0],g=f[1]/2+o.q,v=Object(o.t)(g),y=Object(o.g)(g),b=0;b<h;++b,m=j,v=w,y=O,f=_){var _=d[b],j=_[0],x=_[1]/2+o.q,w=Object(o.t)(x),O=Object(o.g)(x),C=j-m,k=C>=0?1:-1,E=k*C,S=E>o.o,P=v*w;if(s.add(Object(o.e)(P*k*Object(o.t)(E),y*O+P*Object(o.g)(E))),c+=S?C+k*o.w:C,S^m>=n^j>=n){var M=Object(i.c)(Object(i.a)(f),Object(i.a)(_));Object(i.e)(M);var A=Object(i.c)(a,M);Object(i.e)(A);var T=(S^C>=0?-1:1)*Object(o.c)(A[2]);(r>T||r===T&&(M[0]||M[1]))&&(u+=S^C>=0?1:-1)}}return(c<-o.i||c<o.i&&s<-o.i)^1&u}},"../../../../d3-geo/src/projection/albers.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/projection/conicEqualArea.js");e.a=function(){return Object(r.a)().parallels([29.5,45.5]).scale(1070).translate([480,250]).rotate([96,0]).center([-.6,38.7])}},"../../../../d3-geo/src/projection/albersUsa.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/albers.js"),o=n("../../../../d3-geo/src/projection/conicEqualArea.js"),s=n("../../../../d3-geo/src/projection/fit.js");e.a=function(){var t,e,n,a,c,u,l=Object(i.a)(),p=Object(o.a)().rotate([154,0]).center([-2,58.5]).parallels([55,65]),d=Object(o.a)().rotate([157,0]).center([-3,19.9]).parallels([8,18]),h={point:function(t,e){u=[t,e]}};function f(t){var e=t[0],r=t[1];return u=null,n.point(e,r),u||(a.point(e,r),u)||(c.point(e,r),u)}function m(){return t=e=null,f}return f.invert=function(t){var e=l.scale(),n=l.translate(),r=(t[0]-n[0])/e,i=(t[1]-n[1])/e;return(i>=.12&&i<.234&&r>=-.425&&r<-.214?p:i>=.166&&i<.234&&r>=-.214&&r<-.115?d:l).invert(t)},f.stream=function(n){return t&&e===n?t:(r=[l.stream(e=n),p.stream(n),d.stream(n)],i=r.length,t={point:function(t,e){for(var n=-1;++n<i;)r[n].point(t,e)},sphere:function(){for(var t=-1;++t<i;)r[t].sphere()},lineStart:function(){for(var t=-1;++t<i;)r[t].lineStart()},lineEnd:function(){for(var t=-1;++t<i;)r[t].lineEnd()},polygonStart:function(){for(var t=-1;++t<i;)r[t].polygonStart()},polygonEnd:function(){for(var t=-1;++t<i;)r[t].polygonEnd()}});var r,i},f.precision=function(t){return arguments.length?(l.precision(t),p.precision(t),d.precision(t),m()):l.precision()},f.scale=function(t){return arguments.length?(l.scale(t),p.scale(.35*t),d.scale(t),f.translate(l.translate())):l.scale()},f.translate=function(t){if(!arguments.length)return l.translate();var e=l.scale(),i=+t[0],o=+t[1];return n=l.translate(t).clipExtent([[i-.455*e,o-.238*e],[i+.455*e,o+.238*e]]).stream(h),a=p.translate([i-.307*e,o+.201*e]).clipExtent([[i-.425*e+r.i,o+.12*e+r.i],[i-.214*e-r.i,o+.234*e-r.i]]).stream(h),c=d.translate([i-.205*e,o+.212*e]).clipExtent([[i-.214*e+r.i,o+.166*e+r.i],[i-.115*e-r.i,o+.234*e-r.i]]).stream(h),m()},f.fitExtent=function(t,e){return Object(s.a)(f,t,e)},f.fitSize=function(t,e){return Object(s.c)(f,t,e)},f.fitWidth=function(t,e){return Object(s.d)(f,t,e)},f.fitHeight=function(t,e){return Object(s.b)(f,t,e)},f.scale(1070)}},"../../../../d3-geo/src/projection/azimuthal.js":function(t,e,n){"use strict";e.b=function(t){return function(e,n){var i=Object(r.g)(e),o=Object(r.g)(n),s=t(i*o);return[s*o*Object(r.t)(e),s*Object(r.t)(n)]}},e.a=function(t){return function(e,n){var i=Object(r.u)(e*e+n*n),o=t(i),s=Object(r.t)(o),a=Object(r.g)(o);return[Object(r.e)(e*s,i*a),Object(r.c)(i&&n*s/i)]}};var r=n("../../../../d3-geo/src/math.js")},"../../../../d3-geo/src/projection/azimuthalEqualArea.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/azimuthal.js"),o=n("../../../../d3-geo/src/projection/index.js"),s=Object(i.b)(function(t){return Object(r.u)(2/(1+t))});s.invert=Object(i.a)(function(t){return 2*Object(r.c)(t/2)}),e.a=function(){return Object(o.a)(s).scale(124.75).clipAngle(179.999)}},"../../../../d3-geo/src/projection/azimuthalEquidistant.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/azimuthal.js"),o=n("../../../../d3-geo/src/projection/index.js"),s=Object(i.b)(function(t){return(t=Object(r.b)(t))&&t/Object(r.t)(t)});s.invert=Object(i.a)(function(t){return t}),e.a=function(){return Object(o.a)(s).scale(79.4188).clipAngle(179.999)}},"../../../../d3-geo/src/projection/conic.js":function(t,e,n){"use strict";e.a=function(t){var e=0,n=r.o/3,o=Object(i.b)(t),s=o(e,n);return s.parallels=function(t){return arguments.length?o(e=t[0]*r.r,n=t[1]*r.r):[e*r.h,n*r.h]},s};var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/index.js")},"../../../../d3-geo/src/projection/conicConformal.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/conic.js"),o=n("../../../../d3-geo/src/projection/mercator.js");function s(t){return Object(r.v)((r.l+t)/2)}function a(t,e){var n=Object(r.g)(t),i=t===e?Object(r.t)(t):Object(r.n)(n/Object(r.g)(e))/Object(r.n)(s(e)/s(t)),a=n*Object(r.p)(s(t),i)/i;if(!i)return o.c;function c(t,e){a>0?e<-r.l+r.i&&(e=-r.l+r.i):e>r.l-r.i&&(e=r.l-r.i);var n=a/Object(r.p)(s(e),i);return[n*Object(r.t)(i*t),a-n*Object(r.g)(i*t)]}return c.invert=function(t,e){var n=a-e,o=Object(r.s)(i)*Object(r.u)(t*t+n*n);return[Object(r.e)(t,Object(r.a)(n))/i*Object(r.s)(n),2*Object(r.d)(Object(r.p)(a/o,1/i))-r.l]},c}e.a=function(){return Object(i.a)(a).scale(109.5).parallels([30,30])}},"../../../../d3-geo/src/projection/conicEqualArea.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/conic.js"),o=n("../../../../d3-geo/src/projection/cylindricalEqualArea.js");function s(t,e){var n=Object(r.t)(t),i=(n+Object(r.t)(e))/2;if(Object(r.a)(i)<r.i)return Object(o.a)(t);var s=1+n*(2*i-n),a=Object(r.u)(s)/i;function c(t,e){var n=Object(r.u)(s-2*i*Object(r.t)(e))/i;return[n*Object(r.t)(t*=i),a-n*Object(r.g)(t)]}return c.invert=function(t,e){var n=a-e;return[Object(r.e)(t,Object(r.a)(n))/i*Object(r.s)(n),Object(r.c)((s-(t*t+n*n)*i*i)/(2*i))]},c}e.a=function(){return Object(i.a)(s).scale(155.424).center([0,33.6442])}},"../../../../d3-geo/src/projection/conicEquidistant.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/conic.js"),o=n("../../../../d3-geo/src/projection/equirectangular.js");function s(t,e){var n=Object(r.g)(t),i=t===e?Object(r.t)(t):(n-Object(r.g)(e))/(e-t),s=n/i+t;if(Object(r.a)(i)<r.i)return o.b;function a(t,e){var n=s-e,o=i*t;return[n*Object(r.t)(o),s-n*Object(r.g)(o)]}return a.invert=function(t,e){var n=s-e;return[Object(r.e)(t,Object(r.a)(n))/i*Object(r.s)(n),s-Object(r.s)(i)*Object(r.u)(t*t+n*n)]},a}e.a=function(){return Object(i.a)(s).scale(131.154).center([0,13.9389])}},"../../../../d3-geo/src/projection/cylindricalEqualArea.js":function(t,e,n){"use strict";e.a=function(t){var e=Object(r.g)(t);function n(t,n){return[t*e,Object(r.t)(n)/e]}return n.invert=function(t,n){return[t/e,Object(r.c)(n*e)]},n};var r=n("../../../../d3-geo/src/math.js")},"../../../../d3-geo/src/projection/equirectangular.js":function(t,e,n){"use strict";e.b=i;var r=n("../../../../d3-geo/src/projection/index.js");function i(t,e){return[t,e]}i.invert=i,e.a=function(){return Object(r.a)(i).scale(152.63)}},"../../../../d3-geo/src/projection/fit.js":function(t,e,n){"use strict";e.a=s,e.c=function(t,e,n){return s(t,[[0,0],e],n)},e.d=function(t,e,n){return o(t,function(n){var r=+e,i=r/(n[1][0]-n[0][0]),o=(r-i*(n[1][0]+n[0][0]))/2,s=-i*n[0][1];t.scale(150*i).translate([o,s])},n)},e.b=function(t,e,n){return o(t,function(n){var r=+e,i=r/(n[1][1]-n[0][1]),o=-i*n[0][0],s=(r-i*(n[1][1]+n[0][1]))/2;t.scale(150*i).translate([o,s])},n)};var r=n("../../../../d3-geo/src/stream.js"),i=n("../../../../d3-geo/src/path/bounds.js");function o(t,e,n){var o=t.clipExtent&&t.clipExtent();return t.scale(150).translate([0,0]),null!=o&&t.clipExtent(null),Object(r.a)(n,t.stream(i.a)),e(i.a.result()),null!=o&&t.clipExtent(o),t}function s(t,e,n){return o(t,function(n){var r=e[1][0]-e[0][0],i=e[1][1]-e[0][1],o=Math.min(r/(n[1][0]-n[0][0]),i/(n[1][1]-n[0][1])),s=+e[0][0]+(r-o*(n[1][0]+n[0][0]))/2,a=+e[0][1]+(i-o*(n[1][1]+n[0][1]))/2;t.scale(150*o).translate([s,a])},n)}},"../../../../d3-geo/src/projection/gnomonic.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/azimuthal.js"),o=n("../../../../d3-geo/src/projection/index.js");function s(t,e){var n=Object(r.g)(e),i=Object(r.g)(t)*n;return[n*Object(r.t)(t)/i,Object(r.t)(e)/i]}s.invert=Object(i.a)(r.d),e.a=function(){return Object(o.a)(s).scale(144.049).clipAngle(60)}},"../../../../d3-geo/src/projection/identity.js":function(t,e,n){"use strict";n("../../../../d3-geo/src/clip/rectangle.js"),n("../../../../d3-geo/src/identity.js"),n("../../../../d3-geo/src/transform.js"),n("../../../../d3-geo/src/projection/fit.js")},"../../../../d3-geo/src/projection/index.js":function(t,e,n){"use strict";e.a=function(t){return f(function(){return t})()},e.b=f;var r=n("../../../../d3-geo/src/clip/antimeridian.js"),i=n("../../../../d3-geo/src/clip/circle.js"),o=n("../../../../d3-geo/src/clip/rectangle.js"),s=n("../../../../d3-geo/src/compose.js"),a=n("../../../../d3-geo/src/identity.js"),c=n("../../../../d3-geo/src/math.js"),u=n("../../../../d3-geo/src/rotation.js"),l=n("../../../../d3-geo/src/transform.js"),p=n("../../../../d3-geo/src/projection/fit.js"),d=n("../../../../d3-geo/src/projection/resample.js"),h=Object(l.a)({point:function(t,e){this.stream.point(t*c.r,e*c.r)}});function f(t){var e,n,f,m,g,v,y,b,_,j,x=150,w=480,O=250,C=0,k=0,E=0,S=0,P=0,M=null,A=r.a,T=null,D=a.a,I=.5,R=Object(d.a)(F,I);function N(t){return[(t=g(t[0]*c.r,t[1]*c.r))[0]*x+n,f-t[1]*x]}function L(t){return(t=g.invert((t[0]-n)/x,(f-t[1])/x))&&[t[0]*c.h,t[1]*c.h]}function F(t,r){return[(t=e(t,r))[0]*x+n,f-t[1]*x]}function V(){g=Object(s.a)(m=Object(u.b)(E,S,P),e);var t=e(C,k);return n=w-t[0]*x,f=O+t[1]*x,z()}function z(){return _=j=null,N}return N.stream=function(t){return _&&j===t?_:_=h((e=m,Object(l.a)({point:function(t,n){var r=e(t,n);return this.stream.point(r[0],r[1])}}))(A(R(D(j=t)))));var e},N.preclip=function(t){return arguments.length?(A=t,M=void 0,z()):A},N.postclip=function(t){return arguments.length?(D=t,T=v=y=b=null,z()):D},N.clipAngle=function(t){return arguments.length?(A=+t?Object(i.a)(M=t*c.r):(M=null,r.a),z()):M*c.h},N.clipExtent=function(t){return arguments.length?(D=null==t?(T=v=y=b=null,a.a):Object(o.a)(T=+t[0][0],v=+t[0][1],y=+t[1][0],b=+t[1][1]),z()):null==T?null:[[T,v],[y,b]]},N.scale=function(t){return arguments.length?(x=+t,V()):x},N.translate=function(t){return arguments.length?(w=+t[0],O=+t[1],V()):[w,O]},N.center=function(t){return arguments.length?(C=t[0]%360*c.r,k=t[1]%360*c.r,V()):[C*c.h,k*c.h]},N.rotate=function(t){return arguments.length?(E=t[0]%360*c.r,S=t[1]%360*c.r,P=t.length>2?t[2]%360*c.r:0,V()):[E*c.h,S*c.h,P*c.h]},N.precision=function(t){return arguments.length?(R=Object(d.a)(F,I=t*t),z()):Object(c.u)(I)},N.fitExtent=function(t,e){return Object(p.a)(N,t,e)},N.fitSize=function(t,e){return Object(p.c)(N,t,e)},N.fitWidth=function(t,e){return Object(p.d)(N,t,e)},N.fitHeight=function(t,e){return Object(p.b)(N,t,e)},function(){return e=t.apply(this,arguments),N.invert=e.invert&&L,V()}}},"../../../../d3-geo/src/projection/mercator.js":function(t,e,n){"use strict";e.c=s,e.b=a;var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/rotation.js"),o=n("../../../../d3-geo/src/projection/index.js");function s(t,e){return[t,Object(r.n)(Object(r.v)((r.l+e)/2))]}function a(t){var e,n,a,c=Object(o.a)(t),u=c.center,l=c.scale,p=c.translate,d=c.clipExtent,h=null;function f(){var o=r.o*l(),u=c(Object(i.a)(c.rotate()).invert([0,0]));return d(null==h?[[u[0]-o,u[1]-o],[u[0]+o,u[1]+o]]:t===s?[[Math.max(u[0]-o,h),e],[Math.min(u[0]+o,n),a]]:[[h,Math.max(u[1]-o,e)],[n,Math.min(u[1]+o,a)]])}return c.scale=function(t){return arguments.length?(l(t),f()):l()},c.translate=function(t){return arguments.length?(p(t),f()):p()},c.center=function(t){return arguments.length?(u(t),f()):u()},c.clipExtent=function(t){return arguments.length?(null==t?h=e=n=a=null:(h=+t[0][0],e=+t[0][1],n=+t[1][0],a=+t[1][1]),f()):null==h?null:[[h,e],[n,a]]},f()}s.invert=function(t,e){return[t,2*Object(r.d)(Object(r.k)(e))-r.l]},e.a=function(){return a(s).scale(961/r.w)}},"../../../../d3-geo/src/projection/naturalEarth1.js":function(t,e,n){"use strict";n("../../../../d3-geo/src/projection/index.js");var r=n("../../../../d3-geo/src/math.js");function i(t,e){var n=e*e,r=n*n;return[t*(.8707-.131979*n+r*(r*(.003971*n-.001529*r)-.013791)),e*(1.007226+n*(.015085+r*(.028874*n-.044475-.005916*r)))]}i.invert=function(t,e){var n,i=e,o=25;do{var s=i*i,a=s*s;i-=n=(i*(1.007226+s*(.015085+a*(.028874*s-.044475-.005916*a)))-e)/(1.007226+s*(.045255+a*(.259866*s-.311325-.005916*11*a)))}while(Object(r.a)(n)>r.i&&--o>0);return[t/(.8707+(s=i*i)*(s*(s*s*s*(.003971-.001529*s)-.013791)-.131979)),i]}},"../../../../d3-geo/src/projection/orthographic.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/azimuthal.js"),o=n("../../../../d3-geo/src/projection/index.js");function s(t,e){return[Object(r.g)(e)*Object(r.t)(t),Object(r.t)(e)]}s.invert=Object(i.a)(r.c),e.a=function(){return Object(o.a)(s).scale(249.5).clipAngle(90+r.i)}},"../../../../d3-geo/src/projection/resample.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/cartesian.js"),i=n("../../../../d3-geo/src/math.js"),o=n("../../../../d3-geo/src/transform.js"),s=16,a=Object(i.g)(30*i.r);e.a=function(t,e){return+e?function(t,e){function n(r,o,s,c,u,l,p,d,h,f,m,g,v,y){var b=p-r,_=d-o,j=b*b+_*_;if(j>4*e&&v--){var x=c+f,w=u+m,O=l+g,C=Object(i.u)(x*x+w*w+O*O),k=Object(i.c)(O/=C),E=Object(i.a)(Object(i.a)(O)-1)<i.i||Object(i.a)(s-h)<i.i?(s+h)/2:Object(i.e)(w,x),S=t(E,k),P=S[0],M=S[1],A=P-r,T=M-o,D=_*A-b*T;(D*D/j>e||Object(i.a)((b*A+_*T)/j-.5)>.3||c*f+u*m+l*g<a)&&(n(r,o,s,c,u,l,P,M,E,x/=C,w/=C,O,v,y),y.point(P,M),n(P,M,E,x,w,O,p,d,h,f,m,g,v,y))}}return function(e){var i,o,a,c,u,l,p,d,h,f,m,g,v={point:y,lineStart:b,lineEnd:j,polygonStart:function(){e.polygonStart(),v.lineStart=x},polygonEnd:function(){e.polygonEnd(),v.lineStart=b}};function y(n,r){n=t(n,r),e.point(n[0],n[1])}function b(){d=NaN,v.point=_,e.lineStart()}function _(i,o){var a=Object(r.a)([i,o]),c=t(i,o);n(d,h,p,f,m,g,d=c[0],h=c[1],p=i,f=a[0],m=a[1],g=a[2],s,e),e.point(d,h)}function j(){v.point=y,e.lineEnd()}function x(){b(),v.point=w,v.lineEnd=O}function w(t,e){_(i=t,e),o=d,a=h,c=f,u=m,l=g,v.point=_}function O(){n(d,h,p,f,m,g,o,a,i,c,u,l,s,e),v.lineEnd=j,j()}return v}}(t,e):(n=t,Object(o.a)({point:function(t,e){t=n(t,e),this.stream.point(t[0],t[1])}}));var n}},"../../../../d3-geo/src/projection/stereographic.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/azimuthal.js"),o=n("../../../../d3-geo/src/projection/index.js");function s(t,e){var n=Object(r.g)(e),i=1+Object(r.g)(t)*n;return[n*Object(r.t)(t)/i,Object(r.t)(e)/i]}s.invert=Object(i.a)(function(t){return 2*Object(r.d)(t)}),e.a=function(){return Object(o.a)(s).scale(250).clipAngle(142)}},"../../../../d3-geo/src/projection/transverseMercator.js":function(t,e,n){"use strict";var r=n("../../../../d3-geo/src/math.js"),i=n("../../../../d3-geo/src/projection/mercator.js");function o(t,e){return[Object(r.n)(Object(r.v)((r.l+e)/2)),-t]}o.invert=function(t,e){return[-e,2*Object(r.d)(Object(r.k)(t))-r.l]},e.a=function(){var t=Object(i.b)(o),e=t.center,n=t.rotate;return t.center=function(t){return arguments.length?e([-t[1],t[0]]):[(t=e())[1],-t[0]]},t.rotate=function(t){return arguments.length?n([t[0],t[1],t.length>2?t[2]+90:90]):[(t=n())[0],t[1],t[2]-90]},n([0,0,90]).scale(159.155)}},"../../../../d3-geo/src/rotation.js":function(t,e,n){"use strict";e.b=s;var r=n("../../../../d3-geo/src/compose.js"),i=n("../../../../d3-geo/src/math.js");function o(t,e){return[t>i.o?t-i.w:t<-i.o?t+i.w:t,e]}function s(t,e,n){return(t%=i.w)?e||n?Object(r.a)(c(t),u(e,n)):c(t):e||n?u(e,n):o}function a(t){return function(e,n){return[(e+=t)>i.o?e-i.w:e<-i.o?e+i.w:e,n]}}function c(t){var e=a(t);return e.invert=a(-t),e}function u(t,e){var n=Object(i.g)(t),r=Object(i.t)(t),o=Object(i.g)(e),s=Object(i.t)(e);function a(t,e){var a=Object(i.g)(e),c=Object(i.g)(t)*a,u=Object(i.t)(t)*a,l=Object(i.t)(e),p=l*n+c*r;return[Object(i.e)(u*o-p*s,c*n-l*r),Object(i.c)(p*o+u*s)]}return a.invert=function(t,e){var a=Object(i.g)(e),c=Object(i.g)(t)*a,u=Object(i.t)(t)*a,l=Object(i.t)(e),p=l*o-u*s;return[Object(i.e)(u*o+l*s,c*n+p*r),Object(i.c)(p*n-c*r)]},a}o.invert=o,e.a=function(t){function e(e){return(e=t(e[0]*i.r,e[1]*i.r))[0]*=i.h,e[1]*=i.h,e}return t=s(t[0]*i.r,t[1]*i.r,t.length>2?t[2]*i.r:0),e.invert=function(e){return(e=t.invert(e[0]*i.r,e[1]*i.r))[0]*=i.h,e[1]*=i.h,e},e}},"../../../../d3-geo/src/stream.js":function(t,e,n){"use strict";function r(t,e){t&&o.hasOwnProperty(t.type)&&o[t.type](t,e)}var i={Feature:function(t,e){r(t.geometry,e)},FeatureCollection:function(t,e){for(var n=t.features,i=-1,o=n.length;++i<o;)r(n[i].geometry,e)}},o={Sphere:function(t,e){e.sphere()},Point:function(t,e){t=t.coordinates,e.point(t[0],t[1],t[2])},MultiPoint:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)t=n[r],e.point(t[0],t[1],t[2])},LineString:function(t,e){s(t.coordinates,e,0)},MultiLineString:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)s(n[r],e,0)},Polygon:function(t,e){a(t.coordinates,e)},MultiPolygon:function(t,e){for(var n=t.coordinates,r=-1,i=n.length;++r<i;)a(n[r],e)},GeometryCollection:function(t,e){for(var n=t.geometries,i=-1,o=n.length;++i<o;)r(n[i],e)}};function s(t,e,n){var r,i=-1,o=t.length-n;for(e.lineStart();++i<o;)r=t[i],e.point(r[0],r[1],r[2]);e.lineEnd()}function a(t,e){var n=-1,r=t.length;for(e.polygonStart();++n<r;)s(t[n],e,1);e.polygonEnd()}e.a=function(t,e){t&&i.hasOwnProperty(t.type)?i[t.type](t,e):r(t,e)}},"../../../../d3-geo/src/transform.js":function(t,e,n){"use strict";e.a=r;function r(t){return function(e){var n=new i;for(var r in t)n[r]=t[r];return n.stream=e,n}}function i(){}i.prototype={constructor:i,point:function(t,e){this.stream.point(t,e)},sphere:function(){this.stream.sphere()},lineStart:function(){this.stream.lineStart()},lineEnd:function(){this.stream.lineEnd()},polygonStart:function(){this.stream.polygonStart()},polygonEnd:function(){this.stream.polygonEnd()}}},"../../../../d3-hierarchy/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-hierarchy/src/cluster.js");n.d(e,"a",function(){return r.a});var i=n("../../../../d3-hierarchy/src/hierarchy/index.js");n.d(e,"b",function(){return i.c});var o=n("../../../../d3-hierarchy/src/pack/index.js");n.d(e,"c",function(){return o.a});n("../../../../d3-hierarchy/src/pack/siblings.js"),n("../../../../d3-hierarchy/src/pack/enclose.js");var s=n("../../../../d3-hierarchy/src/partition.js");n.d(e,"d",function(){return s.a});var a=n("../../../../d3-hierarchy/src/stratify.js");n.d(e,"e",function(){return a.a});var c=n("../../../../d3-hierarchy/src/tree.js");n.d(e,"f",function(){return c.a});var u=n("../../../../d3-hierarchy/src/treemap/index.js");n.d(e,"g",function(){return u.a});var l=n("../../../../d3-hierarchy/src/treemap/binary.js");n.d(e,"h",function(){return l.a});var p=n("../../../../d3-hierarchy/src/treemap/dice.js");n.d(e,"i",function(){return p.a});var d=n("../../../../d3-hierarchy/src/treemap/slice.js");n.d(e,"k",function(){return d.a});var h=n("../../../../d3-hierarchy/src/treemap/sliceDice.js");n.d(e,"l",function(){return h.a});var f=n("../../../../d3-hierarchy/src/treemap/squarify.js");n.d(e,"m",function(){return f.a});var m=n("../../../../d3-hierarchy/src/treemap/resquarify.js");n.d(e,"j",function(){return m.a})},"../../../../d3-hierarchy/src/accessors.js":function(t,e,n){"use strict";function r(t){if("function"!=typeof t)throw new Error;return t}e.a=function(t){return null==t?null:r(t)},e.b=r},"../../../../d3-hierarchy/src/array.js":function(t,e,n){"use strict";n.d(e,"b",function(){return r}),e.a=function(t){var e,n,r=t.length;for(;r;)n=Math.random()*r--|0,e=t[r],t[r]=t[n],t[n]=e;return t};var r=Array.prototype.slice},"../../../../d3-hierarchy/src/cluster.js":function(t,e,n){"use strict";function r(t,e){return t.parent===e.parent?1:2}function i(t,e){return t+e.x}function o(t,e){return Math.max(t,e.y)}e.a=function(){var t=r,e=1,n=1,s=!1;function a(r){var a,c=0;r.eachAfter(function(e){var n,r=e.children;r?(e.x=(n=r).reduce(i,0)/n.length,e.y=1+r.reduce(o,0)):(e.x=a?c+=t(e,a):0,e.y=0,a=e)});var u=function(t){for(var e;e=t.children;)t=e[0];return t}(r),l=function(t){for(var e;e=t.children;)t=e[e.length-1];return t}(r),p=u.x-t(u,l)/2,d=l.x+t(l,u)/2;return r.eachAfter(s?function(t){t.x=(t.x-r.x)*e,t.y=(r.y-t.y)*n}:function(t){t.x=(t.x-p)/(d-p)*e,t.y=(1-(r.y?t.y/r.y:1))*n})}return a.separation=function(e){return arguments.length?(t=e,a):t},a.size=function(t){return arguments.length?(s=!1,e=+t[0],n=+t[1],a):s?null:[e,n]},a.nodeSize=function(t){return arguments.length?(s=!0,e=+t[0],n=+t[1],a):s?[e,n]:null},a}},"../../../../d3-hierarchy/src/constant.js":function(t,e,n){"use strict";e.a=function(){return 0},e.b=function(t){return function(){return t}}},"../../../../d3-hierarchy/src/hierarchy/ancestors.js":function(t,e,n){"use strict";e.a=function(){for(var t=this,e=[t];t=t.parent;)e.push(t);return e}},"../../../../d3-hierarchy/src/hierarchy/count.js":function(t,e,n){"use strict";function r(t){var e=0,n=t.children,r=n&&n.length;if(r)for(;--r>=0;)e+=n[r].value;else e=1;t.value=e}e.a=function(){return this.eachAfter(r)}},"../../../../d3-hierarchy/src/hierarchy/descendants.js":function(t,e,n){"use strict";e.a=function(){var t=[];return this.each(function(e){t.push(e)}),t}},"../../../../d3-hierarchy/src/hierarchy/each.js":function(t,e,n){"use strict";e.a=function(t){var e,n,r,i,o=this,s=[o];do{for(e=s.reverse(),s=[];o=e.pop();)if(t(o),n=o.children)for(r=0,i=n.length;r<i;++r)s.push(n[r])}while(s.length);return this}},"../../../../d3-hierarchy/src/hierarchy/eachAfter.js":function(t,e,n){"use strict";e.a=function(t){for(var e,n,r,i=this,o=[i],s=[];i=o.pop();)if(s.push(i),e=i.children)for(n=0,r=e.length;n<r;++n)o.push(e[n]);for(;i=s.pop();)t(i);return this}},"../../../../d3-hierarchy/src/hierarchy/eachBefore.js":function(t,e,n){"use strict";e.a=function(t){for(var e,n,r=this,i=[r];r=i.pop();)if(t(r),e=r.children)for(n=e.length-1;n>=0;--n)i.push(e[n]);return this}},"../../../../d3-hierarchy/src/hierarchy/index.js":function(t,e,n){"use strict";e.c=f,e.b=v,e.a=y;var r=n("../../../../d3-hierarchy/src/hierarchy/count.js"),i=n("../../../../d3-hierarchy/src/hierarchy/each.js"),o=n("../../../../d3-hierarchy/src/hierarchy/eachBefore.js"),s=n("../../../../d3-hierarchy/src/hierarchy/eachAfter.js"),a=n("../../../../d3-hierarchy/src/hierarchy/sum.js"),c=n("../../../../d3-hierarchy/src/hierarchy/sort.js"),u=n("../../../../d3-hierarchy/src/hierarchy/path.js"),l=n("../../../../d3-hierarchy/src/hierarchy/ancestors.js"),p=n("../../../../d3-hierarchy/src/hierarchy/descendants.js"),d=n("../../../../d3-hierarchy/src/hierarchy/leaves.js"),h=n("../../../../d3-hierarchy/src/hierarchy/links.js");function f(t,e){var n,r,i,o,s,a=new y(t),c=+t.value&&(a.value=t.value),u=[a];for(null==e&&(e=m);n=u.pop();)if(c&&(n.value=+n.data.value),(i=e(n.data))&&(s=i.length))for(n.children=new Array(s),o=s-1;o>=0;--o)u.push(r=n.children[o]=new y(i[o])),r.parent=n,r.depth=n.depth+1;return a.eachBefore(v)}function m(t){return t.children}function g(t){t.data=t.data.data}function v(t){var e=0;do{t.height=e}while((t=t.parent)&&t.height<++e)}function y(t){this.data=t,this.depth=this.height=0,this.parent=null}y.prototype=f.prototype={constructor:y,count:r.a,each:i.a,eachAfter:s.a,eachBefore:o.a,sum:a.a,sort:c.a,path:u.a,ancestors:l.a,descendants:p.a,leaves:d.a,links:h.a,copy:function(){return f(this).eachBefore(g)}}},"../../../../d3-hierarchy/src/hierarchy/leaves.js":function(t,e,n){"use strict";e.a=function(){var t=[];return this.eachBefore(function(e){e.children||t.push(e)}),t}},"../../../../d3-hierarchy/src/hierarchy/links.js":function(t,e,n){"use strict";e.a=function(){var t=this,e=[];return t.each(function(n){n!==t&&e.push({source:n.parent,target:n})}),e}},"../../../../d3-hierarchy/src/hierarchy/path.js":function(t,e,n){"use strict";e.a=function(t){for(var e=this,n=function(t,e){if(t===e)return t;var n=t.ancestors(),r=e.ancestors(),i=null;t=n.pop(),e=r.pop();for(;t===e;)i=t,t=n.pop(),e=r.pop();return i}(e,t),r=[e];e!==n;)e=e.parent,r.push(e);for(var i=r.length;t!==n;)r.splice(i,0,t),t=t.parent;return r}},"../../../../d3-hierarchy/src/hierarchy/sort.js":function(t,e,n){"use strict";e.a=function(t){return this.eachBefore(function(e){e.children&&e.children.sort(t)})}},"../../../../d3-hierarchy/src/hierarchy/sum.js":function(t,e,n){"use strict";e.a=function(t){return this.eachAfter(function(e){for(var n=+t(e.data)||0,r=e.children,i=r&&r.length;--i>=0;)n+=r[i].value;e.value=n})}},"../../../../d3-hierarchy/src/pack/enclose.js":function(t,e,n){"use strict";var r=n("../../../../d3-hierarchy/src/array.js");function i(t,e){var n,r;if(a(e,t))return[e];for(n=0;n<t.length;++n)if(o(e,t[n])&&a(u(t[n],e),t))return[t[n],e];for(n=0;n<t.length-1;++n)for(r=n+1;r<t.length;++r)if(o(u(t[n],t[r]),e)&&o(u(t[n],e),t[r])&&o(u(t[r],e),t[n])&&a(l(t[n],t[r],e),t))return[t[n],t[r],e];throw new Error}function o(t,e){var n=t.r-e.r,r=e.x-t.x,i=e.y-t.y;return n<0||n*n<r*r+i*i}function s(t,e){var n=t.r-e.r+1e-6,r=e.x-t.x,i=e.y-t.y;return n>0&&n*n>r*r+i*i}function a(t,e){for(var n=0;n<e.length;++n)if(!s(t,e[n]))return!1;return!0}function c(t){switch(t.length){case 1:return{x:(e=t[0]).x,y:e.y,r:e.r};case 2:return u(t[0],t[1]);case 3:return l(t[0],t[1],t[2])}var e}function u(t,e){var n=t.x,r=t.y,i=t.r,o=e.x,s=e.y,a=e.r,c=o-n,u=s-r,l=a-i,p=Math.sqrt(c*c+u*u);return{x:(n+o+c/p*l)/2,y:(r+s+u/p*l)/2,r:(p+i+a)/2}}function l(t,e,n){var r=t.x,i=t.y,o=t.r,s=e.x,a=e.y,c=e.r,u=n.x,l=n.y,p=n.r,d=r-s,h=r-u,f=i-a,m=i-l,g=c-o,v=p-o,y=r*r+i*i-o*o,b=y-s*s-a*a+c*c,_=y-u*u-l*l+p*p,j=h*f-d*m,x=(f*_-m*b)/(2*j)-r,w=(m*g-f*v)/j,O=(h*b-d*_)/(2*j)-i,C=(d*v-h*g)/j,k=w*w+C*C-1,E=2*(o+x*w+O*C),S=x*x+O*O-o*o,P=-(k?(E+Math.sqrt(E*E-4*k*S))/(2*k):S/E);return{x:r+x+w*P,y:i+O+C*P,r:P}}e.a=function(t){for(var e,n,o=0,a=(t=Object(r.a)(r.b.call(t))).length,u=[];o<a;)e=t[o],n&&s(n,e)?++o:(n=c(u=i(u,e)),o=0);return n}},"../../../../d3-hierarchy/src/pack/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-hierarchy/src/pack/siblings.js"),i=n("../../../../d3-hierarchy/src/accessors.js"),o=n("../../../../d3-hierarchy/src/constant.js");function s(t){return Math.sqrt(t.value)}function a(t){return function(e){e.children||(e.r=Math.max(0,+t(e)||0))}}function c(t,e){return function(n){if(i=n.children){var i,o,s,a=i.length,c=t(n)*e||0;if(c)for(o=0;o<a;++o)i[o].r+=c;if(s=Object(r.a)(i),c)for(o=0;o<a;++o)i[o].r-=c;n.r=s+c}}}function u(t){return function(e){var n=e.parent;e.r*=t,n&&(e.x=n.x+t*e.x,e.y=n.y+t*e.y)}}e.a=function(){var t=null,e=1,n=1,r=o.a;function l(i){return i.x=e/2,i.y=n/2,t?i.eachBefore(a(t)).eachAfter(c(r,.5)).eachBefore(u(1)):i.eachBefore(a(s)).eachAfter(c(o.a,1)).eachAfter(c(r,i.r/Math.min(e,n))).eachBefore(u(Math.min(e,n)/(2*i.r))),i}return l.radius=function(e){return arguments.length?(t=Object(i.a)(e),l):t},l.size=function(t){return arguments.length?(e=+t[0],n=+t[1],l):[e,n]},l.padding=function(t){return arguments.length?(r="function"==typeof t?t:Object(o.b)(+t),l):r},l}},"../../../../d3-hierarchy/src/pack/siblings.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../d3-hierarchy/src/pack/enclose.js");function i(t,e,n){var r=t.x,i=t.y,o=e.r+n.r,s=t.r+n.r,a=e.x-r,c=e.y-i,u=a*a+c*c;if(u){var l=.5+((s*=s)-(o*=o))/(2*u),p=Math.sqrt(Math.max(0,2*o*(s+u)-(s-=u)*s-o*o))/(2*u);n.x=r+l*a+p*c,n.y=i+l*c-p*a}else n.x=r+s,n.y=i}function o(t,e){var n=e.x-t.x,r=e.y-t.y,i=t.r+e.r;return i*i-1e-6>n*n+r*r}function s(t){var e=t._,n=t.next._,r=e.r+n.r,i=(e.x*n.r+n.x*e.r)/r,o=(e.y*n.r+n.y*e.r)/r;return i*i+o*o}function a(t){this._=t,this.next=null,this.previous=null}function c(t){if(!(u=t.length))return 0;var e,n,c,u,l,p,d,h,f,m,g;if((e=t[0]).x=0,e.y=0,!(u>1))return e.r;if(n=t[1],e.x=-n.r,n.x=e.r,n.y=0,!(u>2))return e.r+n.r;i(n,e,c=t[2]),e=new a(e),n=new a(n),c=new a(c),e.next=c.previous=n,n.next=e.previous=c,c.next=n.previous=e;t:for(d=3;d<u;++d){i(e._,n._,c=t[d]),c=new a(c),h=n.next,f=e.previous,m=n._.r,g=e._.r;do{if(m<=g){if(o(h._,c._)){n=h,e.next=n,n.previous=e,--d;continue t}m+=h._.r,h=h.next}else{if(o(f._,c._)){(e=f).next=n,n.previous=e,--d;continue t}g+=f._.r,f=f.previous}}while(h!==f.next);for(c.previous=e,c.next=n,e.next=n.previous=n=c,l=s(e);(c=c.next)!==n;)(p=s(c))<l&&(e=c,l=p);n=e.next}for(e=[n._],c=n;(c=c.next)!==n;)e.push(c._);for(c=Object(r.a)(e),d=0;d<u;++d)e=t[d],e.x-=c.x,e.y-=c.y;return c.r}},"../../../../d3-hierarchy/src/partition.js":function(t,e,n){"use strict";var r=n("../../../../d3-hierarchy/src/treemap/round.js"),i=n("../../../../d3-hierarchy/src/treemap/dice.js");e.a=function(){var t=1,e=1,n=0,o=!1;function s(s){var a,c,u=s.height+1;return s.x0=s.y0=n,s.x1=t,s.y1=e/u,s.eachBefore((a=e,c=u,function(t){t.children&&Object(i.a)(t,t.x0,a*(t.depth+1)/c,t.x1,a*(t.depth+2)/c);var e=t.x0,r=t.y0,o=t.x1-n,s=t.y1-n;o<e&&(e=o=(e+o)/2),s<r&&(r=s=(r+s)/2),t.x0=e,t.y0=r,t.x1=o,t.y1=s})),o&&s.eachBefore(r.a),s}return s.round=function(t){return arguments.length?(o=!!t,s):o},s.size=function(n){return arguments.length?(t=+n[0],e=+n[1],s):[t,e]},s.padding=function(t){return arguments.length?(n=+t,s):n},s}},"../../../../d3-hierarchy/src/stratify.js":function(t,e,n){"use strict";var r=n("../../../../d3-hierarchy/src/accessors.js"),i=n("../../../../d3-hierarchy/src/hierarchy/index.js"),o="$",s={depth:-1},a={};function c(t){return t.id}function u(t){return t.parentId}e.a=function(){var t=c,e=u;function n(n){var r,c,u,l,p,d,h,f=n.length,m=new Array(f),g={};for(c=0;c<f;++c)r=n[c],p=m[c]=new i.a(r),null!=(d=t(r,c,n))&&(d+="")&&(g[h=o+(p.id=d)]=h in g?a:p);for(c=0;c<f;++c)if(p=m[c],null!=(d=e(n[c],c,n))&&(d+="")){if(!(l=g[o+d]))throw new Error("missing: "+d);if(l===a)throw new Error("ambiguous: "+d);l.children?l.children.push(p):l.children=[p],p.parent=l}else{if(u)throw new Error("multiple roots");u=p}if(!u)throw new Error("no root");if(u.parent=s,u.eachBefore(function(t){t.depth=t.parent.depth+1,--f}).eachBefore(i.b),u.parent=null,f>0)throw new Error("cycle");return u}return n.id=function(e){return arguments.length?(t=Object(r.b)(e),n):t},n.parentId=function(t){return arguments.length?(e=Object(r.b)(t),n):e},n}},"../../../../d3-hierarchy/src/tree.js":function(t,e,n){"use strict";var r=n("../../../../d3-hierarchy/src/hierarchy/index.js");function i(t,e){return t.parent===e.parent?1:2}function o(t){var e=t.children;return e?e[0]:t.t}function s(t){var e=t.children;return e?e[e.length-1]:t.t}function a(t,e){this._=t,this.parent=null,this.children=null,this.A=null,this.a=this,this.z=0,this.m=0,this.c=0,this.s=0,this.t=null,this.i=e}a.prototype=Object.create(r.a.prototype),e.a=function(){var t=i,e=1,n=1,r=null;function c(i){var o=function(t){for(var e,n,r,i,o,s=new a(t,0),c=[s];e=c.pop();)if(r=e._.children)for(e.children=new Array(o=r.length),i=o-1;i>=0;--i)c.push(n=e.children[i]=new a(r[i],i)),n.parent=e;return(s.parent=new a(null,0)).children=[s],s}(i);if(o.eachAfter(u),o.parent.m=-o.z,o.eachBefore(l),r)i.eachBefore(p);else{var s=i,c=i,d=i;i.eachBefore(function(t){t.x<s.x&&(s=t),t.x>c.x&&(c=t),t.depth>d.depth&&(d=t)});var h=s===c?1:t(s,c)/2,f=h-s.x,m=e/(c.x+h+f),g=n/(d.depth||1);i.eachBefore(function(t){t.x=(t.x+f)*m,t.y=t.depth*g})}return i}function u(e){var n=e.children,r=e.parent.children,i=e.i?r[e.i-1]:null;if(n){!function(t){for(var e,n=0,r=0,i=t.children,o=i.length;--o>=0;)(e=i[o]).z+=n,e.m+=n,n+=e.s+(r+=e.c)}(e);var a=(n[0].z+n[n.length-1].z)/2;i?(e.z=i.z+t(e._,i._),e.m=e.z-a):e.z=a}else i&&(e.z=i.z+t(e._,i._));e.parent.A=function(e,n,r){if(n){for(var i,a=e,c=e,u=n,l=a.parent.children[0],p=a.m,d=c.m,h=u.m,f=l.m;u=s(u),a=o(a),u&&a;)l=o(l),(c=s(c)).a=e,(i=u.z+h-a.z-p+t(u._,a._))>0&&(_=e,j=r,m=(b=u).a.parent===_.parent?b.a:j,void 0,y=(v=i)/((g=e).i-m.i),g.c-=y,g.s+=v,m.c+=y,g.z+=v,g.m+=v,p+=i,d+=i),h+=u.m,p+=a.m,f+=l.m,d+=c.m;u&&!s(c)&&(c.t=u,c.m+=h-d),a&&!o(l)&&(l.t=a,l.m+=p-f,r=e)}var m,g,v,y;var b,_,j;return r}(e,i,e.parent.A||r[0])}function l(t){t._.x=t.z+t.parent.m,t.m+=t.parent.m}function p(t){t.x*=e,t.y=t.depth*n}return c.separation=function(e){return arguments.length?(t=e,c):t},c.size=function(t){return arguments.length?(r=!1,e=+t[0],n=+t[1],c):r?null:[e,n]},c.nodeSize=function(t){return arguments.length?(r=!0,e=+t[0],n=+t[1],c):r?[e,n]:null},c}},"../../../../d3-hierarchy/src/treemap/binary.js":function(t,e,n){"use strict";e.a=function(t,e,n,r,i){var o,s,a=t.children,c=a.length,u=new Array(c+1);for(u[0]=s=o=0;o<c;++o)u[o+1]=s+=a[o].value;!function t(e,n,r,i,o,s,c){if(e>=n-1){var l=a[e];return l.x0=i,l.y0=o,l.x1=s,void(l.y1=c)}var p=u[e],d=r/2+p,h=e+1,f=n-1;for(;h<f;){var m=h+f>>>1;u[m]<d?h=m+1:f=m}d-u[h-1]<u[h]-d&&e+1<h&&--h;var g=u[h]-p,v=r-g;if(s-i>c-o){var y=(i*v+s*g)/r;t(e,h,g,i,o,y,c),t(h,n,v,y,o,s,c)}else{var b=(o*v+c*g)/r;t(e,h,g,i,o,s,b),t(h,n,v,i,b,s,c)}}(0,c,t.value,e,n,r,i)}},"../../../../d3-hierarchy/src/treemap/dice.js":function(t,e,n){"use strict";e.a=function(t,e,n,r,i){for(var o,s=t.children,a=-1,c=s.length,u=t.value&&(r-e)/t.value;++a<c;)(o=s[a]).y0=n,o.y1=i,o.x0=e,o.x1=e+=o.value*u}},"../../../../d3-hierarchy/src/treemap/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-hierarchy/src/treemap/round.js"),i=n("../../../../d3-hierarchy/src/treemap/squarify.js"),o=n("../../../../d3-hierarchy/src/accessors.js"),s=n("../../../../d3-hierarchy/src/constant.js");e.a=function(){var t=i.a,e=!1,n=1,a=1,c=[0],u=s.a,l=s.a,p=s.a,d=s.a,h=s.a;function f(t){return t.x0=t.y0=0,t.x1=n,t.y1=a,t.eachBefore(m),c=[0],e&&t.eachBefore(r.a),t}function m(e){var n=c[e.depth],r=e.x0+n,i=e.y0+n,o=e.x1-n,s=e.y1-n;o<r&&(r=o=(r+o)/2),s<i&&(i=s=(i+s)/2),e.x0=r,e.y0=i,e.x1=o,e.y1=s,e.children&&(n=c[e.depth+1]=u(e)/2,r+=h(e)-n,i+=l(e)-n,o-=p(e)-n,s-=d(e)-n,o<r&&(r=o=(r+o)/2),s<i&&(i=s=(i+s)/2),t(e,r,i,o,s))}return f.round=function(t){return arguments.length?(e=!!t,f):e},f.size=function(t){return arguments.length?(n=+t[0],a=+t[1],f):[n,a]},f.tile=function(e){return arguments.length?(t=Object(o.b)(e),f):t},f.padding=function(t){return arguments.length?f.paddingInner(t).paddingOuter(t):f.paddingInner()},f.paddingInner=function(t){return arguments.length?(u="function"==typeof t?t:Object(s.b)(+t),f):u},f.paddingOuter=function(t){return arguments.length?f.paddingTop(t).paddingRight(t).paddingBottom(t).paddingLeft(t):f.paddingTop()},f.paddingTop=function(t){return arguments.length?(l="function"==typeof t?t:Object(s.b)(+t),f):l},f.paddingRight=function(t){return arguments.length?(p="function"==typeof t?t:Object(s.b)(+t),f):p},f.paddingBottom=function(t){return arguments.length?(d="function"==typeof t?t:Object(s.b)(+t),f):d},f.paddingLeft=function(t){return arguments.length?(h="function"==typeof t?t:Object(s.b)(+t),f):h},f}},"../../../../d3-hierarchy/src/treemap/resquarify.js":function(t,e,n){"use strict";var r=n("../../../../d3-hierarchy/src/treemap/dice.js"),i=n("../../../../d3-hierarchy/src/treemap/slice.js"),o=n("../../../../d3-hierarchy/src/treemap/squarify.js");e.a=function t(e){function n(t,n,s,a,c){if((u=t._squarify)&&u.ratio===e)for(var u,l,p,d,h,f=-1,m=u.length,g=t.value;++f<m;){for(p=(l=u[f]).children,d=l.value=0,h=p.length;d<h;++d)l.value+=p[d].value;l.dice?Object(r.a)(l,n,s,a,s+=(c-s)*l.value/g):Object(i.a)(l,n,s,n+=(a-n)*l.value/g,c),g-=l.value}else t._squarify=u=Object(o.c)(e,t,n,s,a,c),u.ratio=e}return n.ratio=function(e){return t((e=+e)>1?e:1)},n}(o.b)},"../../../../d3-hierarchy/src/treemap/round.js":function(t,e,n){"use strict";e.a=function(t){t.x0=Math.round(t.x0),t.y0=Math.round(t.y0),t.x1=Math.round(t.x1),t.y1=Math.round(t.y1)}},"../../../../d3-hierarchy/src/treemap/slice.js":function(t,e,n){"use strict";e.a=function(t,e,n,r,i){for(var o,s=t.children,a=-1,c=s.length,u=t.value&&(i-n)/t.value;++a<c;)(o=s[a]).x0=e,o.x1=r,o.y0=n,o.y1=n+=o.value*u}},"../../../../d3-hierarchy/src/treemap/sliceDice.js":function(t,e,n){"use strict";var r=n("../../../../d3-hierarchy/src/treemap/dice.js"),i=n("../../../../d3-hierarchy/src/treemap/slice.js");e.a=function(t,e,n,o,s){(1&t.depth?i.a:r.a)(t,e,n,o,s)}},"../../../../d3-hierarchy/src/treemap/squarify.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o}),e.c=s;var r=n("../../../../d3-hierarchy/src/treemap/dice.js"),i=n("../../../../d3-hierarchy/src/treemap/slice.js"),o=(1+Math.sqrt(5))/2;function s(t,e,n,o,s,a){for(var c,u,l,p,d,h,f,m,g,v,y,b=[],_=e.children,j=0,x=0,w=_.length,O=e.value;j<w;){l=s-n,p=a-o;do{d=_[x++].value}while(!d&&x<w);for(h=f=d,y=d*d*(v=Math.max(p/l,l/p)/(O*t)),g=Math.max(f/y,y/h);x<w;++x){if(d+=u=_[x].value,u<h&&(h=u),u>f&&(f=u),y=d*d*v,(m=Math.max(f/y,y/h))>g){d-=u;break}g=m}b.push(c={value:d,dice:l<p,children:_.slice(j,x)}),c.dice?Object(r.a)(c,n,o,s,O?o+=p*d/O:a):Object(i.a)(c,n,o,O?n+=l*d/O:s,a),O-=d,j=x}return b}e.a=function t(e){function n(t,n,r,i,o){s(e,t,n,r,i,o)}return n.ratio=function(e){return t((e=+e)>1?e:1)},n}(o)},"../../../../d3-interpolate/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../d3-interpolate/src/value.js");n.d(e,"interpolate",function(){return r.a});var i=n("../../../../d3-interpolate/src/array.js");n.d(e,"interpolateArray",function(){return i.a});var o=n("../../../../d3-interpolate/src/basis.js");n.d(e,"interpolateBasis",function(){return o.b});var s=n("../../../../d3-interpolate/src/basisClosed.js");n.d(e,"interpolateBasisClosed",function(){return s.a});var a=n("../../../../d3-interpolate/src/date.js");n.d(e,"interpolateDate",function(){return a.a});var c=n("../../../../d3-interpolate/src/number.js");n.d(e,"interpolateNumber",function(){return c.a});var u=n("../../../../d3-interpolate/src/object.js");n.d(e,"interpolateObject",function(){return u.a});var l=n("../../../../d3-interpolate/src/round.js");n.d(e,"interpolateRound",function(){return l.a});var p=n("../../../../d3-interpolate/src/string.js");n.d(e,"interpolateString",function(){return p.a});var d=n("../../../../d3-interpolate/src/transform/index.js");n.d(e,"interpolateTransformCss",function(){return d.a}),n.d(e,"interpolateTransformSvg",function(){return d.b});var h=n("../../../../d3-interpolate/src/zoom.js");n.d(e,"interpolateZoom",function(){return h.a});var f=n("../../../../d3-interpolate/src/rgb.js");n.d(e,"interpolateRgb",function(){return f.a}),n.d(e,"interpolateRgbBasis",function(){return f.b}),n.d(e,"interpolateRgbBasisClosed",function(){return f.c});var m=n("../../../../d3-interpolate/src/hsl.js");n.d(e,"interpolateHsl",function(){return m.a}),n.d(e,"interpolateHslLong",function(){return m.b});var g=n("../../../../d3-interpolate/src/lab.js");n.d(e,"interpolateLab",function(){return g.a});var v=n("../../../../d3-interpolate/src/hcl.js");n.d(e,"interpolateHcl",function(){return v.a}),n.d(e,"interpolateHclLong",function(){return v.b});var y=n("../../../../d3-interpolate/src/cubehelix.js");n.d(e,"interpolateCubehelix",function(){return y.b}),n.d(e,"interpolateCubehelixLong",function(){return y.a});var b=n("../../../../d3-interpolate/src/quantize.js");n.d(e,"quantize",function(){return b.a})},"../../../../d3-interpolate/src/array.js":function(t,e,n){"use strict";var r=n("../../../../d3-interpolate/src/value.js");e.a=function(t,e){var n,i=e?e.length:0,o=t?Math.min(i,t.length):0,s=new Array(o),a=new Array(i);for(n=0;n<o;++n)s[n]=Object(r.a)(t[n],e[n]);for(;n<i;++n)a[n]=e[n];return function(t){for(n=0;n<o;++n)a[n]=s[n](t);return a}}},"../../../../d3-interpolate/src/basis.js":function(t,e,n){"use strict";function r(t,e,n,r,i){var o=t*t,s=o*t;return((1-3*t+3*o-s)*e+(4-6*o+3*s)*n+(1+3*t+3*o-3*s)*r+s*i)/6}e.a=r,e.b=function(t){var e=t.length-1;return function(n){var i=n<=0?n=0:n>=1?(n=1,e-1):Math.floor(n*e),o=t[i],s=t[i+1],a=i>0?t[i-1]:2*o-s,c=i<e-1?t[i+2]:2*s-o;return r((n-i/e)*e,a,o,s,c)}}},"../../../../d3-interpolate/src/basisClosed.js":function(t,e,n){"use strict";var r=n("../../../../d3-interpolate/src/basis.js");e.a=function(t){var e=t.length;return function(n){var i=Math.floor(((n%=1)<0?++n:n)*e),o=t[(i+e-1)%e],s=t[i%e],a=t[(i+1)%e],c=t[(i+2)%e];return Object(r.a)((n-i/e)*e,o,s,a,c)}}},"../../../../d3-interpolate/src/color.js":function(t,e,n){"use strict";e.c=function(t,e){var n=e-t;return n?i(t,n>180||n<-180?n-360*Math.round(n/360):n):Object(r.a)(isNaN(t)?e:t)},e.b=function(t){return 1==(t=+t)?o:function(e,n){return n-e?(i=e,o=n,s=t,i=Math.pow(i,s),o=Math.pow(o,s)-i,s=1/s,function(t){return Math.pow(i+t*o,s)}):Object(r.a)(isNaN(e)?n:e);var i,o,s}},e.a=o;var r=n("../../../../d3-interpolate/src/constant.js");function i(t,e){return function(n){return t+n*e}}function o(t,e){var n=e-t;return n?i(t,n):Object(r.a)(isNaN(t)?e:t)}},"../../../../d3-interpolate/src/constant.js":function(t,e,n){"use strict";e.a=function(t){return function(){return t}}},"../../../../d3-interpolate/src/cubehelix.js":function(t,e,n){"use strict";n.d(e,"a",function(){return s});var r=n("../../../../d3-color/index.js"),i=n("../../../../d3-interpolate/src/color.js");function o(t){return function e(n){function o(e,o){var s=t((e=Object(r.b)(e)).h,(o=Object(r.b)(o)).h),a=Object(i.a)(e.s,o.s),c=Object(i.a)(e.l,o.l),u=Object(i.a)(e.opacity,o.opacity);return function(t){return e.h=s(t),e.s=a(t),e.l=c(Math.pow(t,n)),e.opacity=u(t),e+""}}return n=+n,o.gamma=e,o}(1)}e.b=o(i.c);var s=o(i.a)},"../../../../d3-interpolate/src/date.js":function(t,e,n){"use strict";e.a=function(t,e){var n=new Date;return e-=t=+t,function(r){return n.setTime(t+e*r),n}}},"../../../../d3-interpolate/src/hcl.js":function(t,e,n){"use strict";n.d(e,"b",function(){return s});var r=n("../../../../d3-color/index.js"),i=n("../../../../d3-interpolate/src/color.js");function o(t){return function(e,n){var o=t((e=Object(r.c)(e)).h,(n=Object(r.c)(n)).h),s=Object(i.a)(e.c,n.c),a=Object(i.a)(e.l,n.l),c=Object(i.a)(e.opacity,n.opacity);return function(t){return e.h=o(t),e.c=s(t),e.l=a(t),e.opacity=c(t),e+""}}}e.a=o(i.c);var s=o(i.a)},"../../../../d3-interpolate/src/hsl.js":function(t,e,n){"use strict";n.d(e,"b",function(){return s});var r=n("../../../../d3-color/index.js"),i=n("../../../../d3-interpolate/src/color.js");function o(t){return function(e,n){var o=t((e=Object(r.d)(e)).h,(n=Object(r.d)(n)).h),s=Object(i.a)(e.s,n.s),a=Object(i.a)(e.l,n.l),c=Object(i.a)(e.opacity,n.opacity);return function(t){return e.h=o(t),e.s=s(t),e.l=a(t),e.opacity=c(t),e+""}}}e.a=o(i.c);var s=o(i.a)},"../../../../d3-interpolate/src/lab.js":function(t,e,n){"use strict";e.a=function(t,e){var n=Object(i.a)((t=Object(r.e)(t)).l,(e=Object(r.e)(e)).l),o=Object(i.a)(t.a,e.a),s=Object(i.a)(t.b,e.b),a=Object(i.a)(t.opacity,e.opacity);return function(e){return t.l=n(e),t.a=o(e),t.b=s(e),t.opacity=a(e),t+""}};var r=n("../../../../d3-color/index.js"),i=n("../../../../d3-interpolate/src/color.js")},"../../../../d3-interpolate/src/number.js":function(t,e,n){"use strict";e.a=function(t,e){return e-=t=+t,function(n){return t+e*n}}},"../../../../d3-interpolate/src/object.js":function(t,e,n){"use strict";var r=n("../../../../d3-interpolate/src/value.js");e.a=function(t,e){var n,i={},o={};null!==t&&"object"==typeof t||(t={}),null!==e&&"object"==typeof e||(e={});for(n in e)n in t?i[n]=Object(r.a)(t[n],e[n]):o[n]=e[n];return function(t){for(n in i)o[n]=i[n](t);return o}}},"../../../../d3-interpolate/src/quantize.js":function(t,e,n){"use strict";e.a=function(t,e){for(var n=new Array(e),r=0;r<e;++r)n[r]=t(r/(e-1));return n}},"../../../../d3-interpolate/src/rgb.js":function(t,e,n){"use strict";n.d(e,"b",function(){return c}),n.d(e,"c",function(){return u});var r=n("../../../../d3-color/index.js"),i=n("../../../../d3-interpolate/src/basis.js"),o=n("../../../../d3-interpolate/src/basisClosed.js"),s=n("../../../../d3-interpolate/src/color.js");function a(t){return function(e){var n,i,o=e.length,s=new Array(o),a=new Array(o),c=new Array(o);for(n=0;n<o;++n)i=Object(r.f)(e[n]),s[n]=i.r||0,a[n]=i.g||0,c[n]=i.b||0;return s=t(s),a=t(a),c=t(c),i.opacity=1,function(t){return i.r=s(t),i.g=a(t),i.b=c(t),i+""}}}e.a=function t(e){var n=Object(s.b)(e);function i(t,e){var i=n((t=Object(r.f)(t)).r,(e=Object(r.f)(e)).r),o=n(t.g,e.g),a=n(t.b,e.b),c=Object(s.a)(t.opacity,e.opacity);return function(e){return t.r=i(e),t.g=o(e),t.b=a(e),t.opacity=c(e),t+""}}return i.gamma=t,i}(1);var c=a(i.b),u=a(o.a)},"../../../../d3-interpolate/src/round.js":function(t,e,n){"use strict";e.a=function(t,e){return e-=t=+t,function(n){return Math.round(t+e*n)}}},"../../../../d3-interpolate/src/string.js":function(t,e,n){"use strict";var r=n("../../../../d3-interpolate/src/number.js"),i=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,o=new RegExp(i.source,"g");e.a=function(t,e){var n,s,a,c,u,l=i.lastIndex=o.lastIndex=0,p=-1,d=[],h=[];for(t+="",e+="";(n=i.exec(t))&&(s=o.exec(e));)(a=s.index)>l&&(a=e.slice(l,a),d[p]?d[p]+=a:d[++p]=a),(n=n[0])===(s=s[0])?d[p]?d[p]+=s:d[++p]=s:(d[++p]=null,h.push({i:p,x:Object(r.a)(n,s)})),l=o.lastIndex;return l<e.length&&(a=e.slice(l),d[p]?d[p]+=a:d[++p]=a),d.length<2?h[0]?(u=h[0].x,function(t){return u(t)+""}):(c=e,function(){return c}):(e=h.length,function(t){for(var n,r=0;r<e;++r)d[(n=h[r]).i]=n.x(t);return d.join("")})}},"../../../../d3-interpolate/src/transform/decompose.js":function(t,e,n){"use strict";n.d(e,"b",function(){return i});var r=180/Math.PI,i={translateX:0,translateY:0,rotate:0,skewX:0,scaleX:1,scaleY:1};e.a=function(t,e,n,i,o,s){var a,c,u;return(a=Math.sqrt(t*t+e*e))&&(t/=a,e/=a),(u=t*n+e*i)&&(n-=t*u,i-=e*u),(c=Math.sqrt(n*n+i*i))&&(n/=c,i/=c,u/=c),t*i<e*n&&(t=-t,e=-e,u=-u,a=-a),{translateX:o,translateY:s,rotate:Math.atan2(e,t)*r,skewX:Math.atan(u)*r,scaleX:a,scaleY:c}}},"../../../../d3-interpolate/src/transform/index.js":function(t,e,n){"use strict";n.d(e,"a",function(){return s}),n.d(e,"b",function(){return a});var r=n("../../../../d3-interpolate/src/number.js"),i=n("../../../../d3-interpolate/src/transform/parse.js");function o(t,e,n,i){function o(t){return t.length?t.pop()+" ":""}return function(s,a){var c,u,l,p,d,h,f=[],m=[];return s=t(s),a=t(a),function(t,i,o,s,a,c){if(t!==o||i!==s){var u=a.push("translate(",null,e,null,n);c.push({i:u-4,x:Object(r.a)(t,o)},{i:u-2,x:Object(r.a)(i,s)})}else(o||s)&&a.push("translate("+o+e+s+n)}(s.translateX,s.translateY,a.translateX,a.translateY,f,m),c=s.rotate,u=a.rotate,l=f,c!==u?(c-u>180?u+=360:u-c>180&&(c+=360),m.push({i:l.push(o(l)+"rotate(",null,i)-2,x:Object(r.a)(c,u)})):u&&l.push(o(l)+"rotate("+u+i),p=s.skewX,d=a.skewX,h=f,p!==d?m.push({i:h.push(o(h)+"skewX(",null,i)-2,x:Object(r.a)(p,d)}):d&&h.push(o(h)+"skewX("+d+i),function(t,e,n,i,s,a){if(t!==n||e!==i){var c=s.push(o(s)+"scale(",null,",",null,")");a.push({i:c-4,x:Object(r.a)(t,n)},{i:c-2,x:Object(r.a)(e,i)})}else 1===n&&1===i||s.push(o(s)+"scale("+n+","+i+")")}(s.scaleX,s.scaleY,a.scaleX,a.scaleY,f,m),s=a=null,function(t){for(var e,n=-1,r=m.length;++n<r;)f[(e=m[n]).i]=e.x(t);return f.join("")}}}var s=o(i.a,"px, ","px)","deg)"),a=o(i.b,", ",")",")")},"../../../../d3-interpolate/src/transform/parse.js":function(t,e,n){"use strict";e.a=function(t){if("none"===t)return a.b;r||(r=document.createElement("DIV"),i=document.documentElement,o=document.defaultView);return r.style.transform=t,t=o.getComputedStyle(i.appendChild(r),null).getPropertyValue("transform"),i.removeChild(r),t=t.slice(7,-1).split(","),Object(a.a)(+t[0],+t[1],+t[2],+t[3],+t[4],+t[5])},e.b=function(t){if(null==t)return a.b;s||(s=document.createElementNS("http://www.w3.org/2000/svg","g"));return s.setAttribute("transform",t),(t=s.transform.baseVal.consolidate())?(t=t.matrix,Object(a.a)(t.a,t.b,t.c,t.d,t.e,t.f)):a.b};var r,i,o,s,a=n("../../../../d3-interpolate/src/transform/decompose.js")},"../../../../d3-interpolate/src/value.js":function(t,e,n){"use strict";var r=n("../../../../d3-color/index.js"),i=n("../../../../d3-interpolate/src/rgb.js"),o=n("../../../../d3-interpolate/src/array.js"),s=n("../../../../d3-interpolate/src/date.js"),a=n("../../../../d3-interpolate/src/number.js"),c=n("../../../../d3-interpolate/src/object.js"),u=n("../../../../d3-interpolate/src/string.js"),l=n("../../../../d3-interpolate/src/constant.js");e.a=function(t,e){var n,p=typeof e;return null==e||"boolean"===p?Object(l.a)(e):("number"===p?a.a:"string"===p?(n=Object(r.a)(e))?(e=n,i.a):u.a:e instanceof r.a?i.a:e instanceof Date?s.a:Array.isArray(e)?o.a:"function"!=typeof e.valueOf&&"function"!=typeof e.toString||isNaN(e)?c.a:a.a)(t,e)}},"../../../../d3-interpolate/src/zoom.js":function(t,e,n){"use strict";var r=Math.SQRT2;function i(t){return((t=Math.exp(t))+1/t)/2}e.a=function(t,e){var n,o,s=t[0],a=t[1],c=t[2],u=e[0],l=e[1],p=e[2],d=u-s,h=l-a,f=d*d+h*h;if(f<1e-12)o=Math.log(p/c)/r,n=function(t){return[s+t*d,a+t*h,c*Math.exp(r*t*o)]};else{var m=Math.sqrt(f),g=(p*p-c*c+4*f)/(2*c*2*m),v=(p*p-c*c-4*f)/(2*p*2*m),y=Math.log(Math.sqrt(g*g+1)-g),b=Math.log(Math.sqrt(v*v+1)-v);o=(b-y)/r,n=function(t){var e,n,u=t*o,l=i(y),p=c/(2*m)*(l*(n=r*u+y,((n=Math.exp(2*n))-1)/(n+1))-(e=y,((e=Math.exp(e))-1/e)/2));return[s+p*d,a+p*h,c*l/i(r*u+y)]}}return n.duration=1e3*o,n}},"../../../../d3-path/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-path/src/path.js");n.d(e,"a",function(){return r.a})},"../../../../d3-path/src/path.js":function(t,e,n){"use strict";var r=Math.PI,i=2*r,o=i-1e-6;function s(){this._x0=this._y0=this._x1=this._y1=null,this._=""}function a(){return new s}s.prototype=a.prototype={constructor:s,moveTo:function(t,e){this._+="M"+(this._x0=this._x1=+t)+","+(this._y0=this._y1=+e)},closePath:function(){null!==this._x1&&(this._x1=this._x0,this._y1=this._y0,this._+="Z")},lineTo:function(t,e){this._+="L"+(this._x1=+t)+","+(this._y1=+e)},quadraticCurveTo:function(t,e,n,r){this._+="Q"+ +t+","+ +e+","+(this._x1=+n)+","+(this._y1=+r)},bezierCurveTo:function(t,e,n,r,i,o){this._+="C"+ +t+","+ +e+","+ +n+","+ +r+","+(this._x1=+i)+","+(this._y1=+o)},arcTo:function(t,e,n,i,o){t=+t,e=+e,n=+n,i=+i,o=+o;var s=this._x1,a=this._y1,c=n-t,u=i-e,l=s-t,p=a-e,d=l*l+p*p;if(o<0)throw new Error("negative radius: "+o);if(null===this._x1)this._+="M"+(this._x1=t)+","+(this._y1=e);else if(d>1e-6)if(Math.abs(p*c-u*l)>1e-6&&o){var h=n-s,f=i-a,m=c*c+u*u,g=h*h+f*f,v=Math.sqrt(m),y=Math.sqrt(d),b=o*Math.tan((r-Math.acos((m+d-g)/(2*v*y)))/2),_=b/y,j=b/v;Math.abs(_-1)>1e-6&&(this._+="L"+(t+_*l)+","+(e+_*p)),this._+="A"+o+","+o+",0,0,"+ +(p*h>l*f)+","+(this._x1=t+j*c)+","+(this._y1=e+j*u)}else this._+="L"+(this._x1=t)+","+(this._y1=e);else;},arc:function(t,e,n,s,a,c){t=+t,e=+e;var u=(n=+n)*Math.cos(s),l=n*Math.sin(s),p=t+u,d=e+l,h=1^c,f=c?s-a:a-s;if(n<0)throw new Error("negative radius: "+n);null===this._x1?this._+="M"+p+","+d:(Math.abs(this._x1-p)>1e-6||Math.abs(this._y1-d)>1e-6)&&(this._+="L"+p+","+d),n&&(f<0&&(f=f%i+i),f>o?this._+="A"+n+","+n+",0,1,"+h+","+(t-u)+","+(e-l)+"A"+n+","+n+",0,1,"+h+","+(this._x1=p)+","+(this._y1=d):f>1e-6&&(this._+="A"+n+","+n+",0,"+ +(f>=r)+","+h+","+(this._x1=t+n*Math.cos(a))+","+(this._y1=e+n*Math.sin(a))))},rect:function(t,e,n,r){this._+="M"+(this._x0=this._x1=+t)+","+(this._y0=this._y1=+e)+"h"+ +n+"v"+ +r+"h"+-n+"Z"},toString:function(){return this._}},e.a=a},"../../../../d3-quadtree/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-quadtree/src/quadtree.js");n.d(e,"a",function(){return r.a})},"../../../../d3-quadtree/src/add.js":function(t,e,n){"use strict";function r(t,e,n,r){if(isNaN(e)||isNaN(n))return t;var i,o,s,a,c,u,l,p,d,h=t._root,f={data:r},m=t._x0,g=t._y0,v=t._x1,y=t._y1;if(!h)return t._root=f,t;for(;h.length;)if((u=e>=(o=(m+v)/2))?m=o:v=o,(l=n>=(s=(g+y)/2))?g=s:y=s,i=h,!(h=h[p=l<<1|u]))return i[p]=f,t;if(a=+t._x.call(null,h.data),c=+t._y.call(null,h.data),e===a&&n===c)return f.next=h,i?i[p]=f:t._root=f,t;do{i=i?i[p]=new Array(4):t._root=new Array(4),(u=e>=(o=(m+v)/2))?m=o:v=o,(l=n>=(s=(g+y)/2))?g=s:y=s}while((p=l<<1|u)==(d=(c>=s)<<1|a>=o));return i[d]=h,i[p]=f,t}e.a=function(t){var e,n,i,o,s=t.length,a=new Array(s),c=new Array(s),u=1/0,l=1/0,p=-1/0,d=-1/0;for(n=0;n<s;++n)isNaN(i=+this._x.call(null,e=t[n]))||isNaN(o=+this._y.call(null,e))||(a[n]=i,c[n]=o,i<u&&(u=i),i>p&&(p=i),o<l&&(l=o),o>d&&(d=o));p<u&&(u=this._x0,p=this._x1);d<l&&(l=this._y0,d=this._y1);for(this.cover(u,l).cover(p,d),n=0;n<s;++n)r(this,a[n],c[n],t[n]);return this},e.b=function(t){var e=+this._x.call(null,t),n=+this._y.call(null,t);return r(this.cover(e,n),e,n,t)}},"../../../../d3-quadtree/src/cover.js":function(t,e,n){"use strict";e.a=function(t,e){if(isNaN(t=+t)||isNaN(e=+e))return this;var n=this._x0,r=this._y0,i=this._x1,o=this._y1;if(isNaN(n))i=(n=Math.floor(t))+1,o=(r=Math.floor(e))+1;else{if(!(n>t||t>i||r>e||e>o))return this;var s,a,c=i-n,u=this._root;switch(a=(e<(r+o)/2)<<1|t<(n+i)/2){case 0:do{s=new Array(4),s[a]=u,u=s}while(o=r+(c*=2),t>(i=n+c)||e>o);break;case 1:do{s=new Array(4),s[a]=u,u=s}while(o=r+(c*=2),(n=i-c)>t||e>o);break;case 2:do{s=new Array(4),s[a]=u,u=s}while(r=o-(c*=2),t>(i=n+c)||r>e);break;case 3:do{s=new Array(4),s[a]=u,u=s}while(r=o-(c*=2),(n=i-c)>t||r>e)}this._root&&this._root.length&&(this._root=u)}return this._x0=n,this._y0=r,this._x1=i,this._y1=o,this}},"../../../../d3-quadtree/src/data.js":function(t,e,n){"use strict";e.a=function(){var t=[];return this.visit(function(e){if(!e.length)do{t.push(e.data)}while(e=e.next)}),t}},"../../../../d3-quadtree/src/extent.js":function(t,e,n){"use strict";e.a=function(t){return arguments.length?this.cover(+t[0][0],+t[0][1]).cover(+t[1][0],+t[1][1]):isNaN(this._x0)?void 0:[[this._x0,this._y0],[this._x1,this._y1]]}},"../../../../d3-quadtree/src/find.js":function(t,e,n){"use strict";var r=n("../../../../d3-quadtree/src/quad.js");e.a=function(t,e,n){var i,o,s,a,c,u,l,p=this._x0,d=this._y0,h=this._x1,f=this._y1,m=[],g=this._root;for(g&&m.push(new r.a(g,p,d,h,f)),null==n?n=1/0:(p=t-n,d=e-n,h=t+n,f=e+n,n*=n);u=m.pop();)if(!(!(g=u.node)||(o=u.x0)>h||(s=u.y0)>f||(a=u.x1)<p||(c=u.y1)<d))if(g.length){var v=(o+a)/2,y=(s+c)/2;m.push(new r.a(g[3],v,y,a,c),new r.a(g[2],o,y,v,c),new r.a(g[1],v,s,a,y),new r.a(g[0],o,s,v,y)),(l=(e>=y)<<1|t>=v)&&(u=m[m.length-1],m[m.length-1]=m[m.length-1-l],m[m.length-1-l]=u)}else{var b=t-+this._x.call(null,g.data),_=e-+this._y.call(null,g.data),j=b*b+_*_;if(j<n){var x=Math.sqrt(n=j);p=t-x,d=e-x,h=t+x,f=e+x,i=g.data}}return i}},"../../../../d3-quadtree/src/quad.js":function(t,e,n){"use strict";e.a=function(t,e,n,r,i){this.node=t,this.x0=e,this.y0=n,this.x1=r,this.y1=i}},"../../../../d3-quadtree/src/quadtree.js":function(t,e,n){"use strict";e.a=m;var r=n("../../../../d3-quadtree/src/add.js"),i=n("../../../../d3-quadtree/src/cover.js"),o=n("../../../../d3-quadtree/src/data.js"),s=n("../../../../d3-quadtree/src/extent.js"),a=n("../../../../d3-quadtree/src/find.js"),c=n("../../../../d3-quadtree/src/remove.js"),u=n("../../../../d3-quadtree/src/root.js"),l=n("../../../../d3-quadtree/src/size.js"),p=n("../../../../d3-quadtree/src/visit.js"),d=n("../../../../d3-quadtree/src/visitAfter.js"),h=n("../../../../d3-quadtree/src/x.js"),f=n("../../../../d3-quadtree/src/y.js");function m(t,e,n){var r=new g(null==e?h.b:e,null==n?f.b:n,NaN,NaN,NaN,NaN);return null==t?r:r.addAll(t)}function g(t,e,n,r,i,o){this._x=t,this._y=e,this._x0=n,this._y0=r,this._x1=i,this._y1=o,this._root=void 0}function v(t){for(var e={data:t.data},n=e;t=t.next;)n=n.next={data:t.data};return e}var y=m.prototype=g.prototype;y.copy=function(){var t,e,n=new g(this._x,this._y,this._x0,this._y0,this._x1,this._y1),r=this._root;if(!r)return n;if(!r.length)return n._root=v(r),n;for(t=[{source:r,target:n._root=new Array(4)}];r=t.pop();)for(var i=0;i<4;++i)(e=r.source[i])&&(e.length?t.push({source:e,target:r.target[i]=new Array(4)}):r.target[i]=v(e));return n},y.add=r.b,y.addAll=r.a,y.cover=i.a,y.data=o.a,y.extent=s.a,y.find=a.a,y.remove=c.a,y.removeAll=c.b,y.root=u.a,y.size=l.a,y.visit=p.a,y.visitAfter=d.a,y.x=h.a,y.y=f.a},"../../../../d3-quadtree/src/remove.js":function(t,e,n){"use strict";e.b=function(t){for(var e=0,n=t.length;e<n;++e)this.remove(t[e]);return this},e.a=function(t){if(isNaN(o=+this._x.call(null,t))||isNaN(s=+this._y.call(null,t)))return this;var e,n,r,i,o,s,a,c,u,l,p,d,h=this._root,f=this._x0,m=this._y0,g=this._x1,v=this._y1;if(!h)return this;if(h.length)for(;;){if((u=o>=(a=(f+g)/2))?f=a:g=a,(l=s>=(c=(m+v)/2))?m=c:v=c,e=h,!(h=h[p=l<<1|u]))return this;if(!h.length)break;(e[p+1&3]||e[p+2&3]||e[p+3&3])&&(n=e,d=p)}for(;h.data!==t;)if(r=h,!(h=h.next))return this;return(i=h.next)&&delete h.next,r?(i?r.next=i:delete r.next,this):e?(i?e[p]=i:delete e[p],(h=e[0]||e[1]||e[2]||e[3])&&h===(e[3]||e[2]||e[1]||e[0])&&!h.length&&(n?n[d]=h:this._root=h),this):(this._root=i,this)}},"../../../../d3-quadtree/src/root.js":function(t,e,n){"use strict";e.a=function(){return this._root}},"../../../../d3-quadtree/src/size.js":function(t,e,n){"use strict";e.a=function(){var t=0;return this.visit(function(e){if(!e.length)do{++t}while(e=e.next)}),t}},"../../../../d3-quadtree/src/visit.js":function(t,e,n){"use strict";var r=n("../../../../d3-quadtree/src/quad.js");e.a=function(t){var e,n,i,o,s,a,c=[],u=this._root;for(u&&c.push(new r.a(u,this._x0,this._y0,this._x1,this._y1));e=c.pop();)if(!t(u=e.node,i=e.x0,o=e.y0,s=e.x1,a=e.y1)&&u.length){var l=(i+s)/2,p=(o+a)/2;(n=u[3])&&c.push(new r.a(n,l,p,s,a)),(n=u[2])&&c.push(new r.a(n,i,p,l,a)),(n=u[1])&&c.push(new r.a(n,l,o,s,p)),(n=u[0])&&c.push(new r.a(n,i,o,l,p))}return this}},"../../../../d3-quadtree/src/visitAfter.js":function(t,e,n){"use strict";var r=n("../../../../d3-quadtree/src/quad.js");e.a=function(t){var e,n=[],i=[];for(this._root&&n.push(new r.a(this._root,this._x0,this._y0,this._x1,this._y1));e=n.pop();){var o=e.node;if(o.length){var s,a=e.x0,c=e.y0,u=e.x1,l=e.y1,p=(a+u)/2,d=(c+l)/2;(s=o[0])&&n.push(new r.a(s,a,c,p,d)),(s=o[1])&&n.push(new r.a(s,p,c,u,d)),(s=o[2])&&n.push(new r.a(s,a,d,p,l)),(s=o[3])&&n.push(new r.a(s,p,d,u,l))}i.push(e)}for(;e=i.pop();)t(e.node,e.x0,e.y0,e.x1,e.y1);return this}},"../../../../d3-quadtree/src/x.js":function(t,e,n){"use strict";e.b=function(t){return t[0]},e.a=function(t){return arguments.length?(this._x=t,this):this._x}},"../../../../d3-quadtree/src/y.js":function(t,e,n){"use strict";e.b=function(t){return t[1]},e.a=function(t){return arguments.length?(this._y=t,this):this._y}},"../../../../d3-request/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-request/src/request.js");n.d(e,"a",function(){return r.a});n("../../../../d3-request/src/html.js"),n("../../../../d3-request/src/json.js"),n("../../../../d3-request/src/text.js"),n("../../../../d3-request/src/xml.js"),n("../../../../d3-request/src/csv.js"),n("../../../../d3-request/src/tsv.js")},"../../../../d3-request/src/csv.js":function(t,e,n){"use strict";var r=n("../../../../d3-dsv/index.js"),i=n("../../../../d3-request/src/dsv.js");Object(i.a)("text/csv",r.a)},"../../../../d3-request/src/dsv.js":function(t,e,n){"use strict";var r=n("../../../../d3-request/src/request.js");e.a=function(t,e){return function(n,i,o){arguments.length<3&&(o=i,i=null);var s=Object(r.a)(n).mimeType(t);return s.row=function(t){return arguments.length?s.response((n=e,r=i=t,function(t){return n(t.responseText,r)})):i;var n,r},s.row(i),o?s.get(o):s}}},"../../../../d3-request/src/html.js":function(t,e,n){"use strict";var r=n("../../../../d3-request/src/type.js");Object(r.a)("text/html",function(t){return document.createRange().createContextualFragment(t.responseText)})},"../../../../d3-request/src/json.js":function(t,e,n){"use strict";var r=n("../../../../d3-request/src/type.js");Object(r.a)("application/json",function(t){return JSON.parse(t.responseText)})},"../../../../d3-request/src/request.js":function(t,e,n){"use strict";var r=n("../../../../d3-collection/index.js"),i=n("../../../../d3-dispatch/index.js");e.a=function(t,e){var n,o,s,a,c=Object(i.a)("beforesend","progress","load","error"),u=Object(r.a)(),l=new XMLHttpRequest,p=null,d=null,h=0;function f(t){var e,r,i,o=l.status;if(!o&&((i=(r=l).responseType)&&"text"!==i?r.response:r.responseText)||o>=200&&o<300||304===o){if(s)try{e=s.call(n,l)}catch(t){return void c.call("error",n,t)}else e=l;c.call("load",n,e)}else c.call("error",n,t)}if("undefined"==typeof XDomainRequest||"withCredentials"in l||!/^(http(s)?:)?\/\//.test(t)||(l=new XDomainRequest),"onload"in l?l.onload=l.onerror=l.ontimeout=f:l.onreadystatechange=function(t){l.readyState>3&&f(t)},l.onprogress=function(t){c.call("progress",n,t)},n={header:function(t,e){return t=(t+"").toLowerCase(),arguments.length<2?u.get(t):(null==e?u.remove(t):u.set(t,e+""),n)},mimeType:function(t){return arguments.length?(o=null==t?null:t+"",n):o},responseType:function(t){return arguments.length?(a=t,n):a},timeout:function(t){return arguments.length?(h=+t,n):h},user:function(t){return arguments.length<1?p:(p=null==t?null:t+"",n)},password:function(t){return arguments.length<1?d:(d=null==t?null:t+"",n)},response:function(t){return s=t,n},get:function(t,e){return n.send("GET",t,e)},post:function(t,e){return n.send("POST",t,e)},send:function(e,r,i){var s;return l.open(e,t,!0,p,d),null==o||u.has("accept")||u.set("accept",o+",*/*"),l.setRequestHeader&&u.each(function(t,e){l.setRequestHeader(e,t)}),null!=o&&l.overrideMimeType&&l.overrideMimeType(o),null!=a&&(l.responseType=a),h>0&&(l.timeout=h),null==i&&"function"==typeof r&&(i=r,r=null),null!=i&&1===i.length&&(s=i,i=function(t,e){s(null==t?e:null)}),null!=i&&n.on("error",i).on("load",function(t){i(null,t)}),c.call("beforesend",n,l),l.send(null==r?null:r),n},abort:function(){return l.abort(),n},on:function(){var t=c.on.apply(c,arguments);return t===c?n:t}},null!=e){if("function"!=typeof e)throw new Error("invalid callback: "+e);return n.get(e)}return n}},"../../../../d3-request/src/text.js":function(t,e,n){"use strict";var r=n("../../../../d3-request/src/type.js");Object(r.a)("text/plain",function(t){return t.responseText})},"../../../../d3-request/src/tsv.js":function(t,e,n){"use strict";var r=n("../../../../d3-dsv/index.js"),i=n("../../../../d3-request/src/dsv.js");Object(i.a)("text/tab-separated-values",r.c)},"../../../../d3-request/src/type.js":function(t,e,n){"use strict";var r=n("../../../../d3-request/src/request.js");e.a=function(t,e){return function(n,i){var o=Object(r.a)(n).mimeType(t).response(e);if(null!=i){if("function"!=typeof i)throw new Error("invalid callback: "+i);return o.get(i)}return o}}},"../../../../d3-request/src/xml.js":function(t,e,n){"use strict";var r=n("../../../../d3-request/src/type.js");Object(r.a)("application/xml",function(t){var e=t.responseXML;if(!e)throw new Error("parse error");return e})},"../../../../d3-scale-chromatic/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../d3-scale-chromatic/src/categorical/category10.js");n.d(e,"schemeCategory10",function(){return r.a});var i=n("../../../../d3-scale-chromatic/src/categorical/Accent.js");n.d(e,"schemeAccent",function(){return i.a});var o=n("../../../../d3-scale-chromatic/src/categorical/Dark2.js");n.d(e,"schemeDark2",function(){return o.a});var s=n("../../../../d3-scale-chromatic/src/categorical/Paired.js");n.d(e,"schemePaired",function(){return s.a});var a=n("../../../../d3-scale-chromatic/src/categorical/Pastel1.js");n.d(e,"schemePastel1",function(){return a.a});var c=n("../../../../d3-scale-chromatic/src/categorical/Pastel2.js");n.d(e,"schemePastel2",function(){return c.a});var u=n("../../../../d3-scale-chromatic/src/categorical/Set1.js");n.d(e,"schemeSet1",function(){return u.a});var l=n("../../../../d3-scale-chromatic/src/categorical/Set2.js");n.d(e,"schemeSet2",function(){return l.a});var p=n("../../../../d3-scale-chromatic/src/categorical/Set3.js");n.d(e,"schemeSet3",function(){return p.a});var d=n("../../../../d3-scale-chromatic/src/diverging/BrBG.js");n.d(e,"interpolateBrBG",function(){return d.a}),n.d(e,"schemeBrBG",function(){return d.b});var h=n("../../../../d3-scale-chromatic/src/diverging/PRGn.js");n.d(e,"interpolatePRGn",function(){return h.a}),n.d(e,"schemePRGn",function(){return h.b});var f=n("../../../../d3-scale-chromatic/src/diverging/PiYG.js");n.d(e,"interpolatePiYG",function(){return f.a}),n.d(e,"schemePiYG",function(){return f.b});var m=n("../../../../d3-scale-chromatic/src/diverging/PuOr.js");n.d(e,"interpolatePuOr",function(){return m.a}),n.d(e,"schemePuOr",function(){return m.b});var g=n("../../../../d3-scale-chromatic/src/diverging/RdBu.js");n.d(e,"interpolateRdBu",function(){return g.a}),n.d(e,"schemeRdBu",function(){return g.b});var v=n("../../../../d3-scale-chromatic/src/diverging/RdGy.js");n.d(e,"interpolateRdGy",function(){return v.a}),n.d(e,"schemeRdGy",function(){return v.b});var y=n("../../../../d3-scale-chromatic/src/diverging/RdYlBu.js");n.d(e,"interpolateRdYlBu",function(){return y.a}),n.d(e,"schemeRdYlBu",function(){return y.b});var b=n("../../../../d3-scale-chromatic/src/diverging/RdYlGn.js");n.d(e,"interpolateRdYlGn",function(){return b.a}),n.d(e,"schemeRdYlGn",function(){return b.b});var _=n("../../../../d3-scale-chromatic/src/diverging/Spectral.js");n.d(e,"interpolateSpectral",function(){return _.a}),n.d(e,"schemeSpectral",function(){return _.b});var j=n("../../../../d3-scale-chromatic/src/sequential-multi/BuGn.js");n.d(e,"interpolateBuGn",function(){return j.a}),n.d(e,"schemeBuGn",function(){return j.b});var x=n("../../../../d3-scale-chromatic/src/sequential-multi/BuPu.js");n.d(e,"interpolateBuPu",function(){return x.a}),n.d(e,"schemeBuPu",function(){return x.b});var w=n("../../../../d3-scale-chromatic/src/sequential-multi/GnBu.js");n.d(e,"interpolateGnBu",function(){return w.a}),n.d(e,"schemeGnBu",function(){return w.b});var O=n("../../../../d3-scale-chromatic/src/sequential-multi/OrRd.js");n.d(e,"interpolateOrRd",function(){return O.a}),n.d(e,"schemeOrRd",function(){return O.b});var C=n("../../../../d3-scale-chromatic/src/sequential-multi/PuBuGn.js");n.d(e,"interpolatePuBuGn",function(){return C.a}),n.d(e,"schemePuBuGn",function(){return C.b});var k=n("../../../../d3-scale-chromatic/src/sequential-multi/PuBu.js");n.d(e,"interpolatePuBu",function(){return k.a}),n.d(e,"schemePuBu",function(){return k.b});var E=n("../../../../d3-scale-chromatic/src/sequential-multi/PuRd.js");n.d(e,"interpolatePuRd",function(){return E.a}),n.d(e,"schemePuRd",function(){return E.b});var S=n("../../../../d3-scale-chromatic/src/sequential-multi/RdPu.js");n.d(e,"interpolateRdPu",function(){return S.a}),n.d(e,"schemeRdPu",function(){return S.b});var P=n("../../../../d3-scale-chromatic/src/sequential-multi/YlGnBu.js");n.d(e,"interpolateYlGnBu",function(){return P.a}),n.d(e,"schemeYlGnBu",function(){return P.b});var M=n("../../../../d3-scale-chromatic/src/sequential-multi/YlGn.js");n.d(e,"interpolateYlGn",function(){return M.a}),n.d(e,"schemeYlGn",function(){return M.b});var A=n("../../../../d3-scale-chromatic/src/sequential-multi/YlOrBr.js");n.d(e,"interpolateYlOrBr",function(){return A.a}),n.d(e,"schemeYlOrBr",function(){return A.b});var T=n("../../../../d3-scale-chromatic/src/sequential-multi/YlOrRd.js");n.d(e,"interpolateYlOrRd",function(){return T.a}),n.d(e,"schemeYlOrRd",function(){return T.b});var D=n("../../../../d3-scale-chromatic/src/sequential-single/Blues.js");n.d(e,"interpolateBlues",function(){return D.a}),n.d(e,"schemeBlues",function(){return D.b});var I=n("../../../../d3-scale-chromatic/src/sequential-single/Greens.js");n.d(e,"interpolateGreens",function(){return I.a}),n.d(e,"schemeGreens",function(){return I.b});var R=n("../../../../d3-scale-chromatic/src/sequential-single/Greys.js");n.d(e,"interpolateGreys",function(){return R.a}),n.d(e,"schemeGreys",function(){return R.b});var N=n("../../../../d3-scale-chromatic/src/sequential-single/Purples.js");n.d(e,"interpolatePurples",function(){return N.a}),n.d(e,"schemePurples",function(){return N.b});var L=n("../../../../d3-scale-chromatic/src/sequential-single/Reds.js");n.d(e,"interpolateReds",function(){return L.a}),n.d(e,"schemeReds",function(){return L.b});var F=n("../../../../d3-scale-chromatic/src/sequential-single/Oranges.js");n.d(e,"interpolateOranges",function(){return F.a}),n.d(e,"schemeOranges",function(){return F.b});var V=n("../../../../d3-scale-chromatic/src/sequential-multi/cubehelix.js");n.d(e,"interpolateCubehelixDefault",function(){return V.a});var z=n("../../../../d3-scale-chromatic/src/sequential-multi/rainbow.js");n.d(e,"interpolateRainbow",function(){return z.b}),n.d(e,"interpolateWarm",function(){return z.c}),n.d(e,"interpolateCool",function(){return z.a});var B=n("../../../../d3-scale-chromatic/src/sequential-multi/viridis.js");n.d(e,"interpolateViridis",function(){return B.a}),n.d(e,"interpolateMagma",function(){return B.c}),n.d(e,"interpolateInferno",function(){return B.b}),n.d(e,"interpolatePlasma",function(){return B.d})},"../../../../d3-scale-chromatic/src/categorical/Accent.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale-chromatic/src/colors.js");e.a=Object(r.a)("7fc97fbeaed4fdc086ffff99386cb0f0027fbf5b17666666")},"../../../../d3-scale-chromatic/src/categorical/Dark2.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale-chromatic/src/colors.js");e.a=Object(r.a)("1b9e77d95f027570b3e7298a66a61ee6ab02a6761d666666")},"../../../../d3-scale-chromatic/src/categorical/Paired.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale-chromatic/src/colors.js");e.a=Object(r.a)("a6cee31f78b4b2df8a33a02cfb9a99e31a1cfdbf6fff7f00cab2d66a3d9affff99b15928")},"../../../../d3-scale-chromatic/src/categorical/Pastel1.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale-chromatic/src/colors.js");e.a=Object(r.a)("fbb4aeb3cde3ccebc5decbe4fed9a6ffffcce5d8bdfddaecf2f2f2")},"../../../../d3-scale-chromatic/src/categorical/Pastel2.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale-chromatic/src/colors.js");e.a=Object(r.a)("b3e2cdfdcdaccbd5e8f4cae4e6f5c9fff2aef1e2cccccccc")},"../../../../d3-scale-chromatic/src/categorical/Set1.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale-chromatic/src/colors.js");e.a=Object(r.a)("e41a1c377eb84daf4a984ea3ff7f00ffff33a65628f781bf999999")},"../../../../d3-scale-chromatic/src/categorical/Set2.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale-chromatic/src/colors.js");e.a=Object(r.a)("66c2a5fc8d628da0cbe78ac3a6d854ffd92fe5c494b3b3b3")},"../../../../d3-scale-chromatic/src/categorical/Set3.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale-chromatic/src/colors.js");e.a=Object(r.a)("8dd3c7ffffb3bebadafb807280b1d3fdb462b3de69fccde5d9d9d9bc80bdccebc5ffed6f")},"../../../../d3-scale-chromatic/src/categorical/category10.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale-chromatic/src/colors.js");e.a=Object(r.a)("1f77b4ff7f0e2ca02cd627289467bd8c564be377c27f7f7fbcbd2217becf")},"../../../../d3-scale-chromatic/src/colors.js":function(t,e,n){"use strict";e.a=function(t){for(var e=t.length/6|0,n=new Array(e),r=0;r<e;)n[r]="#"+t.slice(6*r,6*++r);return n}},"../../../../d3-scale-chromatic/src/diverging/BrBG.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("d8b365f5f5f55ab4ac","a6611adfc27d80cdc1018571","a6611adfc27df5f5f580cdc1018571","8c510ad8b365f6e8c3c7eae55ab4ac01665e","8c510ad8b365f6e8c3f5f5f5c7eae55ab4ac01665e","8c510abf812ddfc27df6e8c3c7eae580cdc135978f01665e","8c510abf812ddfc27df6e8c3f5f5f5c7eae580cdc135978f01665e","5430058c510abf812ddfc27df6e8c3c7eae580cdc135978f01665e003c30","5430058c510abf812ddfc27df6e8c3f5f5f5c7eae580cdc135978f01665e003c30").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/diverging/PRGn.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("af8dc3f7f7f77fbf7b","7b3294c2a5cfa6dba0008837","7b3294c2a5cff7f7f7a6dba0008837","762a83af8dc3e7d4e8d9f0d37fbf7b1b7837","762a83af8dc3e7d4e8f7f7f7d9f0d37fbf7b1b7837","762a839970abc2a5cfe7d4e8d9f0d3a6dba05aae611b7837","762a839970abc2a5cfe7d4e8f7f7f7d9f0d3a6dba05aae611b7837","40004b762a839970abc2a5cfe7d4e8d9f0d3a6dba05aae611b783700441b","40004b762a839970abc2a5cfe7d4e8f7f7f7d9f0d3a6dba05aae611b783700441b").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/diverging/PiYG.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e9a3c9f7f7f7a1d76a","d01c8bf1b6dab8e1864dac26","d01c8bf1b6daf7f7f7b8e1864dac26","c51b7de9a3c9fde0efe6f5d0a1d76a4d9221","c51b7de9a3c9fde0eff7f7f7e6f5d0a1d76a4d9221","c51b7dde77aef1b6dafde0efe6f5d0b8e1867fbc414d9221","c51b7dde77aef1b6dafde0eff7f7f7e6f5d0b8e1867fbc414d9221","8e0152c51b7dde77aef1b6dafde0efe6f5d0b8e1867fbc414d9221276419","8e0152c51b7dde77aef1b6dafde0eff7f7f7e6f5d0b8e1867fbc414d9221276419").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/diverging/PuOr.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("998ec3f7f7f7f1a340","5e3c99b2abd2fdb863e66101","5e3c99b2abd2f7f7f7fdb863e66101","542788998ec3d8daebfee0b6f1a340b35806","542788998ec3d8daebf7f7f7fee0b6f1a340b35806","5427888073acb2abd2d8daebfee0b6fdb863e08214b35806","5427888073acb2abd2d8daebf7f7f7fee0b6fdb863e08214b35806","2d004b5427888073acb2abd2d8daebfee0b6fdb863e08214b358067f3b08","2d004b5427888073acb2abd2d8daebf7f7f7fee0b6fdb863e08214b358067f3b08").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/diverging/RdBu.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ef8a62f7f7f767a9cf","ca0020f4a58292c5de0571b0","ca0020f4a582f7f7f792c5de0571b0","b2182bef8a62fddbc7d1e5f067a9cf2166ac","b2182bef8a62fddbc7f7f7f7d1e5f067a9cf2166ac","b2182bd6604df4a582fddbc7d1e5f092c5de4393c32166ac","b2182bd6604df4a582fddbc7f7f7f7d1e5f092c5de4393c32166ac","67001fb2182bd6604df4a582fddbc7d1e5f092c5de4393c32166ac053061","67001fb2182bd6604df4a582fddbc7f7f7f7d1e5f092c5de4393c32166ac053061").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/diverging/RdGy.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ef8a62ffffff999999","ca0020f4a582bababa404040","ca0020f4a582ffffffbababa404040","b2182bef8a62fddbc7e0e0e09999994d4d4d","b2182bef8a62fddbc7ffffffe0e0e09999994d4d4d","b2182bd6604df4a582fddbc7e0e0e0bababa8787874d4d4d","b2182bd6604df4a582fddbc7ffffffe0e0e0bababa8787874d4d4d","67001fb2182bd6604df4a582fddbc7e0e0e0bababa8787874d4d4d1a1a1a","67001fb2182bd6604df4a582fddbc7ffffffe0e0e0bababa8787874d4d4d1a1a1a").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/diverging/RdYlBu.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fc8d59ffffbf91bfdb","d7191cfdae61abd9e92c7bb6","d7191cfdae61ffffbfabd9e92c7bb6","d73027fc8d59fee090e0f3f891bfdb4575b4","d73027fc8d59fee090ffffbfe0f3f891bfdb4575b4","d73027f46d43fdae61fee090e0f3f8abd9e974add14575b4","d73027f46d43fdae61fee090ffffbfe0f3f8abd9e974add14575b4","a50026d73027f46d43fdae61fee090e0f3f8abd9e974add14575b4313695","a50026d73027f46d43fdae61fee090ffffbfe0f3f8abd9e974add14575b4313695").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/diverging/RdYlGn.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fc8d59ffffbf91cf60","d7191cfdae61a6d96a1a9641","d7191cfdae61ffffbfa6d96a1a9641","d73027fc8d59fee08bd9ef8b91cf601a9850","d73027fc8d59fee08bffffbfd9ef8b91cf601a9850","d73027f46d43fdae61fee08bd9ef8ba6d96a66bd631a9850","d73027f46d43fdae61fee08bffffbfd9ef8ba6d96a66bd631a9850","a50026d73027f46d43fdae61fee08bd9ef8ba6d96a66bd631a9850006837","a50026d73027f46d43fdae61fee08bffffbfd9ef8ba6d96a66bd631a9850006837").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/diverging/Spectral.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fc8d59ffffbf99d594","d7191cfdae61abdda42b83ba","d7191cfdae61ffffbfabdda42b83ba","d53e4ffc8d59fee08be6f59899d5943288bd","d53e4ffc8d59fee08bffffbfe6f59899d5943288bd","d53e4ff46d43fdae61fee08be6f598abdda466c2a53288bd","d53e4ff46d43fdae61fee08bffffbfe6f598abdda466c2a53288bd","9e0142d53e4ff46d43fdae61fee08be6f598abdda466c2a53288bd5e4fa2","9e0142d53e4ff46d43fdae61fee08bffffbfe6f598abdda466c2a53288bd5e4fa2").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/ramp.js":function(t,e,n){"use strict";var r=n("../../../../d3-interpolate/index.js");e.a=function(t){return Object(r.interpolateRgbBasis)(t[t.length-1])}},"../../../../d3-scale-chromatic/src/sequential-multi/BuGn.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e5f5f999d8c92ca25f","edf8fbb2e2e266c2a4238b45","edf8fbb2e2e266c2a42ca25f006d2c","edf8fbccece699d8c966c2a42ca25f006d2c","edf8fbccece699d8c966c2a441ae76238b45005824","f7fcfde5f5f9ccece699d8c966c2a441ae76238b45005824","f7fcfde5f5f9ccece699d8c966c2a441ae76238b45006d2c00441b").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/BuPu.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e0ecf49ebcda8856a7","edf8fbb3cde38c96c688419d","edf8fbb3cde38c96c68856a7810f7c","edf8fbbfd3e69ebcda8c96c68856a7810f7c","edf8fbbfd3e69ebcda8c96c68c6bb188419d6e016b","f7fcfde0ecf4bfd3e69ebcda8c96c68c6bb188419d6e016b","f7fcfde0ecf4bfd3e69ebcda8c96c68c6bb188419d810f7c4d004b").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/GnBu.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e0f3dba8ddb543a2ca","f0f9e8bae4bc7bccc42b8cbe","f0f9e8bae4bc7bccc443a2ca0868ac","f0f9e8ccebc5a8ddb57bccc443a2ca0868ac","f0f9e8ccebc5a8ddb57bccc44eb3d32b8cbe08589e","f7fcf0e0f3dbccebc5a8ddb57bccc44eb3d32b8cbe08589e","f7fcf0e0f3dbccebc5a8ddb57bccc44eb3d32b8cbe0868ac084081").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/OrRd.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fee8c8fdbb84e34a33","fef0d9fdcc8afc8d59d7301f","fef0d9fdcc8afc8d59e34a33b30000","fef0d9fdd49efdbb84fc8d59e34a33b30000","fef0d9fdd49efdbb84fc8d59ef6548d7301f990000","fff7ecfee8c8fdd49efdbb84fc8d59ef6548d7301f990000","fff7ecfee8c8fdd49efdbb84fc8d59ef6548d7301fb300007f0000").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/PuBu.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ece7f2a6bddb2b8cbe","f1eef6bdc9e174a9cf0570b0","f1eef6bdc9e174a9cf2b8cbe045a8d","f1eef6d0d1e6a6bddb74a9cf2b8cbe045a8d","f1eef6d0d1e6a6bddb74a9cf3690c00570b0034e7b","fff7fbece7f2d0d1e6a6bddb74a9cf3690c00570b0034e7b","fff7fbece7f2d0d1e6a6bddb74a9cf3690c00570b0045a8d023858").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/PuBuGn.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ece2f0a6bddb1c9099","f6eff7bdc9e167a9cf02818a","f6eff7bdc9e167a9cf1c9099016c59","f6eff7d0d1e6a6bddb67a9cf1c9099016c59","f6eff7d0d1e6a6bddb67a9cf3690c002818a016450","fff7fbece2f0d0d1e6a6bddb67a9cf3690c002818a016450","fff7fbece2f0d0d1e6a6bddb67a9cf3690c002818a016c59014636").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/PuRd.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e7e1efc994c7dd1c77","f1eef6d7b5d8df65b0ce1256","f1eef6d7b5d8df65b0dd1c77980043","f1eef6d4b9dac994c7df65b0dd1c77980043","f1eef6d4b9dac994c7df65b0e7298ace125691003f","f7f4f9e7e1efd4b9dac994c7df65b0e7298ace125691003f","f7f4f9e7e1efd4b9dac994c7df65b0e7298ace125698004367001f").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/RdPu.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fde0ddfa9fb5c51b8a","feebe2fbb4b9f768a1ae017e","feebe2fbb4b9f768a1c51b8a7a0177","feebe2fcc5c0fa9fb5f768a1c51b8a7a0177","feebe2fcc5c0fa9fb5f768a1dd3497ae017e7a0177","fff7f3fde0ddfcc5c0fa9fb5f768a1dd3497ae017e7a0177","fff7f3fde0ddfcc5c0fa9fb5f768a1dd3497ae017e7a017749006a").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/YlGn.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("f7fcb9addd8e31a354","ffffccc2e69978c679238443","ffffccc2e69978c67931a354006837","ffffccd9f0a3addd8e78c67931a354006837","ffffccd9f0a3addd8e78c67941ab5d238443005a32","ffffe5f7fcb9d9f0a3addd8e78c67941ab5d238443005a32","ffffe5f7fcb9d9f0a3addd8e78c67941ab5d238443006837004529").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/YlGnBu.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("edf8b17fcdbb2c7fb8","ffffcca1dab441b6c4225ea8","ffffcca1dab441b6c42c7fb8253494","ffffccc7e9b47fcdbb41b6c42c7fb8253494","ffffccc7e9b47fcdbb41b6c41d91c0225ea80c2c84","ffffd9edf8b1c7e9b47fcdbb41b6c41d91c0225ea80c2c84","ffffd9edf8b1c7e9b47fcdbb41b6c41d91c0225ea8253494081d58").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/YlOrBr.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fff7bcfec44fd95f0e","ffffd4fed98efe9929cc4c02","ffffd4fed98efe9929d95f0e993404","ffffd4fee391fec44ffe9929d95f0e993404","ffffd4fee391fec44ffe9929ec7014cc4c028c2d04","ffffe5fff7bcfee391fec44ffe9929ec7014cc4c028c2d04","ffffe5fff7bcfee391fec44ffe9929ec7014cc4c02993404662506").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/YlOrRd.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("ffeda0feb24cf03b20","ffffb2fecc5cfd8d3ce31a1c","ffffb2fecc5cfd8d3cf03b20bd0026","ffffb2fed976feb24cfd8d3cf03b20bd0026","ffffb2fed976feb24cfd8d3cfc4e2ae31a1cb10026","ffffccffeda0fed976feb24cfd8d3cfc4e2ae31a1cb10026","ffffccffeda0fed976feb24cfd8d3cfc4e2ae31a1cbd0026800026").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-multi/cubehelix.js":function(t,e,n){"use strict";var r=n("../../../../d3-color/index.js"),i=n("../../../../d3-interpolate/index.js");e.a=Object(i.interpolateCubehelixLong)(Object(r.b)(300,.5,0),Object(r.b)(-240,.5,1))},"../../../../d3-scale-chromatic/src/sequential-multi/rainbow.js":function(t,e,n){"use strict";n.d(e,"c",function(){return o}),n.d(e,"a",function(){return s});var r=n("../../../../d3-color/index.js"),i=n("../../../../d3-interpolate/index.js"),o=Object(i.interpolateCubehelixLong)(Object(r.b)(-100,.75,.35),Object(r.b)(80,1.5,.8)),s=Object(i.interpolateCubehelixLong)(Object(r.b)(260,.75,.35),Object(r.b)(80,1.5,.8)),a=Object(r.b)();e.b=function(t){(t<0||t>1)&&(t-=Math.floor(t));var e=Math.abs(t-.5);return a.h=360*t-100,a.s=1.5-1.5*e,a.l=.8-.9*e,a+""}},"../../../../d3-scale-chromatic/src/sequential-multi/viridis.js":function(t,e,n){"use strict";n.d(e,"c",function(){return o}),n.d(e,"b",function(){return s}),n.d(e,"d",function(){return a});var r=n("../../../../d3-scale-chromatic/src/colors.js");function i(t){var e=t.length;return function(n){return t[Math.max(0,Math.min(e-1,Math.floor(n*e)))]}}e.a=i(Object(r.a)("44015444025645045745055946075a46085c460a5d460b5e470d60470e6147106347116447136548146748166848176948186a481a6c481b6d481c6e481d6f481f70482071482173482374482475482576482677482878482979472a7a472c7a472d7b472e7c472f7d46307e46327e46337f463480453581453781453882443983443a83443b84433d84433e85423f854240864241864142874144874045884046883f47883f48893e49893e4a893e4c8a3d4d8a3d4e8a3c4f8a3c508b3b518b3b528b3a538b3a548c39558c39568c38588c38598c375a8c375b8d365c8d365d8d355e8d355f8d34608d34618d33628d33638d32648e32658e31668e31678e31688e30698e306a8e2f6b8e2f6c8e2e6d8e2e6e8e2e6f8e2d708e2d718e2c718e2c728e2c738e2b748e2b758e2a768e2a778e2a788e29798e297a8e297b8e287c8e287d8e277e8e277f8e27808e26818e26828e26828e25838e25848e25858e24868e24878e23888e23898e238a8d228b8d228c8d228d8d218e8d218f8d21908d21918c20928c20928c20938c1f948c1f958b1f968b1f978b1f988b1f998a1f9a8a1e9b8a1e9c891e9d891f9e891f9f881fa0881fa1881fa1871fa28720a38620a48621a58521a68522a78522a88423a98324aa8325ab8225ac8226ad8127ad8128ae8029af7f2ab07f2cb17e2db27d2eb37c2fb47c31b57b32b67a34b67935b77937b87838b9773aba763bbb753dbc743fbc7340bd7242be7144bf7046c06f48c16e4ac16d4cc26c4ec36b50c46a52c56954c56856c66758c7655ac8645cc8635ec96260ca6063cb5f65cb5e67cc5c69cd5b6ccd5a6ece5870cf5773d05675d05477d1537ad1517cd2507fd34e81d34d84d44b86d54989d5488bd6468ed64590d74393d74195d84098d83e9bd93c9dd93ba0da39a2da37a5db36a8db34aadc32addc30b0dd2fb2dd2db5de2bb8de29bade28bddf26c0df25c2df23c5e021c8e020cae11fcde11dd0e11cd2e21bd5e21ad8e219dae319dde318dfe318e2e418e5e419e7e419eae51aece51befe51cf1e51df4e61ef6e620f8e621fbe723fde725"));var o=i(Object(r.a)("00000401000501010601010802010902020b02020d03030f03031204041405041606051806051a07061c08071e0907200a08220b09240c09260d0a290e0b2b100b2d110c2f120d31130d34140e36150e38160f3b180f3d19103f1a10421c10441d11471e114920114b21114e22115024125325125527125829115a2a115c2c115f2d11612f116331116533106734106936106b38106c390f6e3b0f703d0f713f0f72400f74420f75440f764510774710784910784a10794c117a4e117b4f127b51127c52137c54137d56147d57157e59157e5a167e5c167f5d177f5f187f601880621980641a80651a80671b80681c816a1c816b1d816d1d816e1e81701f81721f817320817521817621817822817922827b23827c23827e24828025828125818326818426818627818827818928818b29818c29818e2a81902a81912b81932b80942c80962c80982d80992d809b2e7f9c2e7f9e2f7fa02f7fa1307ea3307ea5317ea6317da8327daa337dab337cad347cae347bb0357bb2357bb3367ab5367ab73779b83779ba3878bc3978bd3977bf3a77c03a76c23b75c43c75c53c74c73d73c83e73ca3e72cc3f71cd4071cf4070d0416fd2426fd3436ed5446dd6456cd8456cd9466bdb476adc4869de4968df4a68e04c67e24d66e34e65e44f64e55064e75263e85362e95462ea5661eb5760ec5860ed5a5fee5b5eef5d5ef05f5ef1605df2625df2645cf3655cf4675cf4695cf56b5cf66c5cf66e5cf7705cf7725cf8745cf8765cf9785df9795df97b5dfa7d5efa7f5efa815ffb835ffb8560fb8761fc8961fc8a62fc8c63fc8e64fc9065fd9266fd9467fd9668fd9869fd9a6afd9b6bfe9d6cfe9f6dfea16efea36ffea571fea772fea973feaa74feac76feae77feb078feb27afeb47bfeb67cfeb77efeb97ffebb81febd82febf84fec185fec287fec488fec68afec88cfeca8dfecc8ffecd90fecf92fed194fed395fed597fed799fed89afdda9cfddc9efddea0fde0a1fde2a3fde3a5fde5a7fde7a9fde9aafdebacfcecaefceeb0fcf0b2fcf2b4fcf4b6fcf6b8fcf7b9fcf9bbfcfbbdfcfdbf")),s=i(Object(r.a)("00000401000501010601010802010a02020c02020e03021004031204031405041706041907051b08051d09061f0a07220b07240c08260d08290e092b10092d110a30120a32140b34150b37160b39180c3c190c3e1b0c411c0c431e0c451f0c48210c4a230c4c240c4f260c51280b53290b552b0b572d0b592f0a5b310a5c320a5e340a5f3609613809623909633b09643d09653e0966400a67420a68440a68450a69470b6a490b6a4a0c6b4c0c6b4d0d6c4f0d6c510e6c520e6d540f6d550f6d57106e59106e5a116e5c126e5d126e5f136e61136e62146e64156e65156e67166e69166e6a176e6c186e6d186e6f196e71196e721a6e741a6e751b6e771c6d781c6d7a1d6d7c1d6d7d1e6d7f1e6c801f6c82206c84206b85216b87216b88226a8a226a8c23698d23698f24699025689225689326679526679727669827669a28659b29649d29649f2a63a02a63a22b62a32c61a52c60a62d60a82e5fa92e5eab2f5ead305dae305cb0315bb1325ab3325ab43359b63458b73557b93556ba3655bc3754bd3853bf3952c03a51c13a50c33b4fc43c4ec63d4dc73e4cc83f4bca404acb4149cc4248ce4347cf4446d04545d24644d34743d44842d54a41d74b3fd84c3ed94d3dda4e3cdb503bdd513ade5238df5337e05536e15635e25734e35933e45a31e55c30e65d2fe75e2ee8602de9612bea632aeb6429eb6628ec6726ed6925ee6a24ef6c23ef6e21f06f20f1711ff1731df2741cf3761bf37819f47918f57b17f57d15f67e14f68013f78212f78410f8850ff8870ef8890cf98b0bf98c0af98e09fa9008fa9207fa9407fb9606fb9706fb9906fb9b06fb9d07fc9f07fca108fca309fca50afca60cfca80dfcaa0ffcac11fcae12fcb014fcb216fcb418fbb61afbb81dfbba1ffbbc21fbbe23fac026fac228fac42afac62df9c72ff9c932f9cb35f8cd37f8cf3af7d13df7d340f6d543f6d746f5d949f5db4cf4dd4ff4df53f4e156f3e35af3e55df2e661f2e865f2ea69f1ec6df1ed71f1ef75f1f179f2f27df2f482f3f586f3f68af4f88ef5f992f6fa96f8fb9af9fc9dfafda1fcffa4")),a=i(Object(r.a)("0d088710078813078916078a19068c1b068d1d068e20068f2206902406912605912805922a05932c05942e05952f059631059733059735049837049938049a3a049a3c049b3e049c3f049c41049d43039e44039e46039f48039f4903a04b03a14c02a14e02a25002a25102a35302a35502a45601a45801a45901a55b01a55c01a65e01a66001a66100a76300a76400a76600a76700a86900a86a00a86c00a86e00a86f00a87100a87201a87401a87501a87701a87801a87a02a87b02a87d03a87e03a88004a88104a78305a78405a78606a68707a68808a68a09a58b0aa58d0ba58e0ca48f0da4910ea3920fa39410a29511a19613a19814a099159f9a169f9c179e9d189d9e199da01a9ca11b9ba21d9aa31e9aa51f99a62098a72197a82296aa2395ab2494ac2694ad2793ae2892b02991b12a90b22b8fb32c8eb42e8db52f8cb6308bb7318ab83289ba3388bb3488bc3587bd3786be3885bf3984c03a83c13b82c23c81c33d80c43e7fc5407ec6417dc7427cc8437bc9447aca457acb4679cc4778cc4977cd4a76ce4b75cf4c74d04d73d14e72d24f71d35171d45270d5536fd5546ed6556dd7566cd8576bd9586ada5a6ada5b69db5c68dc5d67dd5e66de5f65de6164df6263e06363e16462e26561e26660e3685fe4695ee56a5de56b5de66c5ce76e5be76f5ae87059e97158e97257ea7457eb7556eb7655ec7754ed7953ed7a52ee7b51ef7c51ef7e50f07f4ff0804ef1814df1834cf2844bf3854bf3874af48849f48948f58b47f58c46f68d45f68f44f79044f79143f79342f89441f89540f9973ff9983ef99a3efa9b3dfa9c3cfa9e3bfb9f3afba139fba238fca338fca537fca636fca835fca934fdab33fdac33fdae32fdaf31fdb130fdb22ffdb42ffdb52efeb72dfeb82cfeba2cfebb2bfebd2afebe2afec029fdc229fdc328fdc527fdc627fdc827fdca26fdcb26fccd25fcce25fcd025fcd225fbd324fbd524fbd724fad824fada24f9dc24f9dd25f8df25f8e125f7e225f7e425f6e626f6e826f5e926f5eb27f4ed27f3ee27f3f027f2f227f1f426f1f525f0f724f0f921"))},"../../../../d3-scale-chromatic/src/sequential-single/Blues.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("deebf79ecae13182bd","eff3ffbdd7e76baed62171b5","eff3ffbdd7e76baed63182bd08519c","eff3ffc6dbef9ecae16baed63182bd08519c","eff3ffc6dbef9ecae16baed64292c62171b5084594","f7fbffdeebf7c6dbef9ecae16baed64292c62171b5084594","f7fbffdeebf7c6dbef9ecae16baed64292c62171b508519c08306b").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-single/Greens.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("e5f5e0a1d99b31a354","edf8e9bae4b374c476238b45","edf8e9bae4b374c47631a354006d2c","edf8e9c7e9c0a1d99b74c47631a354006d2c","edf8e9c7e9c0a1d99b74c47641ab5d238b45005a32","f7fcf5e5f5e0c7e9c0a1d99b74c47641ab5d238b45005a32","f7fcf5e5f5e0c7e9c0a1d99b74c47641ab5d238b45006d2c00441b").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-single/Greys.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("f0f0f0bdbdbd636363","f7f7f7cccccc969696525252","f7f7f7cccccc969696636363252525","f7f7f7d9d9d9bdbdbd969696636363252525","f7f7f7d9d9d9bdbdbd969696737373525252252525","fffffff0f0f0d9d9d9bdbdbd969696737373525252252525","fffffff0f0f0d9d9d9bdbdbd969696737373525252252525000000").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-single/Oranges.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fee6cefdae6be6550d","feeddefdbe85fd8d3cd94701","feeddefdbe85fd8d3ce6550da63603","feeddefdd0a2fdae6bfd8d3ce6550da63603","feeddefdd0a2fdae6bfd8d3cf16913d948018c2d04","fff5ebfee6cefdd0a2fdae6bfd8d3cf16913d948018c2d04","fff5ebfee6cefdd0a2fdae6bfd8d3cf16913d94801a636037f2704").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-single/Purples.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("efedf5bcbddc756bb1","f2f0f7cbc9e29e9ac86a51a3","f2f0f7cbc9e29e9ac8756bb154278f","f2f0f7dadaebbcbddc9e9ac8756bb154278f","f2f0f7dadaebbcbddc9e9ac8807dba6a51a34a1486","fcfbfdefedf5dadaebbcbddc9e9ac8807dba6a51a34a1486","fcfbfdefedf5dadaebbcbddc9e9ac8807dba6a51a354278f3f007d").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale-chromatic/src/sequential-single/Reds.js":function(t,e,n){"use strict";n.d(e,"b",function(){return o});var r=n("../../../../d3-scale-chromatic/src/colors.js"),i=n("../../../../d3-scale-chromatic/src/ramp.js"),o=new Array(3).concat("fee0d2fc9272de2d26","fee5d9fcae91fb6a4acb181d","fee5d9fcae91fb6a4ade2d26a50f15","fee5d9fcbba1fc9272fb6a4ade2d26a50f15","fee5d9fcbba1fc9272fb6a4aef3b2ccb181d99000d","fff5f0fee0d2fcbba1fc9272fb6a4aef3b2ccb181d99000d","fff5f0fee0d2fcbba1fc9272fb6a4aef3b2ccb181da50f1567000d").map(r.a);e.a=Object(i.a)(o)},"../../../../d3-scale/index.js":function(t,e,n){"use strict";n("../../../../d3-scale/src/band.js");var r=n("../../../../d3-scale/src/identity.js");n.d(e,"a",function(){return r.a});var i=n("../../../../d3-scale/src/linear.js");n.d(e,"b",function(){return i.a});var o=n("../../../../d3-scale/src/log.js");n.d(e,"c",function(){return o.a});var s=n("../../../../d3-scale/src/ordinal.js");n.d(e,"d",function(){return s.a});var a=n("../../../../d3-scale/src/pow.js");n.d(e,"e",function(){return a.a}),n.d(e,"h",function(){return a.b});var c=n("../../../../d3-scale/src/quantile.js");n.d(e,"f",function(){return c.a});var u=n("../../../../d3-scale/src/quantize.js");n.d(e,"g",function(){return u.a});var l=n("../../../../d3-scale/src/threshold.js");n.d(e,"i",function(){return l.a});var p=n("../../../../d3-scale/src/time.js");n.d(e,"j",function(){return p.b});var d=n("../../../../d3-scale/src/utcTime.js");n.d(e,"k",function(){return d.a});n("../../../../d3-scale/src/sequential.js")},"../../../../d3-scale/src/array.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i}),n.d(e,"b",function(){return o});var r=Array.prototype,i=r.map,o=r.slice},"../../../../d3-scale/src/band.js":function(t,e,n){"use strict";n("../../../../d3-array/index.js"),n("../../../../d3-scale/src/ordinal.js")},"../../../../d3-scale/src/constant.js":function(t,e,n){"use strict";e.a=function(t){return function(){return t}}},"../../../../d3-scale/src/continuous.js":function(t,e,n){"use strict";e.c=u,e.a=function(t,e){return e.domain(t.domain()).range(t.range()).interpolate(t.interpolate()).clamp(t.clamp())},e.b=function(t,e){var n,r,s,d=c,h=c,f=i.interpolate,m=!1;function g(){return n=Math.min(d.length,h.length)>2?p:l,r=s=null,v}function v(e){return(r||(r=n(d,h,m?(i=t,function(t,e){var n=i(t=+t,e=+e);return function(r){return r<=t?0:r>=e?1:n(r)}}):t,f)))(+e);var i}return v.invert=function(t){return(s||(s=n(h,d,u,m?(r=e,function(t,e){var n=r(t=+t,e=+e);return function(r){return r<=0?t:r>=1?e:n(r)}}):e)))(+t);var r},v.domain=function(t){return arguments.length?(d=o.a.call(t,a.a),g()):d.slice()},v.range=function(t){return arguments.length?(h=o.b.call(t),g()):h.slice()},v.rangeRound=function(t){return h=o.b.call(t),f=i.interpolateRound,g()},v.clamp=function(t){return arguments.length?(m=!!t,g()):m},v.interpolate=function(t){return arguments.length?(f=t,g()):f},g()};var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-interpolate/index.js"),o=n("../../../../d3-scale/src/array.js"),s=n("../../../../d3-scale/src/constant.js"),a=n("../../../../d3-scale/src/number.js"),c=[0,1];function u(t,e){return(e-=t=+t)?function(n){return(n-t)/e}:Object(s.a)(e)}function l(t,e,n,r){var i=t[0],o=t[1],s=e[0],a=e[1];return o<i?(i=n(o,i),s=r(a,s)):(i=n(i,o),s=r(s,a)),function(t){return s(i(t))}}function p(t,e,n,i){var o=Math.min(t.length,e.length)-1,s=new Array(o),a=new Array(o),c=-1;for(t[o]<t[0]&&(t=t.slice().reverse(),e=e.slice().reverse());++c<o;)s[c]=n(t[c],t[c+1]),a[c]=i(e[c],e[c+1]);return function(e){var n=Object(r.b)(t,e,1,o)-1;return a[n](s[n](e))}}},"../../../../d3-scale/src/identity.js":function(t,e,n){"use strict";e.a=function t(){var e=[0,1];function n(t){return+t}n.invert=n;n.domain=n.range=function(t){return arguments.length?(e=r.a.call(t,o.a),n):e.slice()};n.copy=function(){return t().domain(e)};return Object(i.b)(n)};var r=n("../../../../d3-scale/src/array.js"),i=n("../../../../d3-scale/src/linear.js"),o=n("../../../../d3-scale/src/number.js")},"../../../../d3-scale/src/linear.js":function(t,e,n){"use strict";e.b=a,e.a=function t(){var e=Object(o.b)(o.c,i.interpolateNumber);e.copy=function(){return Object(o.a)(e,t())};return a(e)};var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-interpolate/index.js"),o=n("../../../../d3-scale/src/continuous.js"),s=n("../../../../d3-scale/src/tickFormat.js");function a(t){var e=t.domain;return t.ticks=function(t){var n=e();return Object(r.s)(n[0],n[n.length-1],null==t?10:t)},t.tickFormat=function(t,n){return Object(s.a)(e(),t,n)},t.nice=function(n){null==n&&(n=10);var i,o=e(),s=0,a=o.length-1,c=o[s],u=o[a];return u<c&&(i=c,c=u,u=i,i=s,s=a,a=i),(i=Object(r.q)(c,u,n))>0?(c=Math.floor(c/i)*i,u=Math.ceil(u/i)*i,i=Object(r.q)(c,u,n)):i<0&&(c=Math.ceil(c*i)/i,u=Math.floor(u*i)/i,i=Object(r.q)(c,u,n)),i>0?(o[s]=Math.floor(c/i)*i,o[a]=Math.ceil(u/i)*i,e(o)):i<0&&(o[s]=Math.ceil(c*i)/i,o[a]=Math.floor(u*i)/i,e(o)),t},t}},"../../../../d3-scale/src/log.js":function(t,e,n){"use strict";e.a=function t(){var e=Object(a.b)(c,u).domain([1,10]),n=e.domain,o=10,l=d(10),f=p(10);function m(){return l=d(o),f=p(o),n()[0]<0&&(l=h(l),f=h(f)),e}e.base=function(t){return arguments.length?(o=+t,m()):o};e.domain=function(t){return arguments.length?(n(t),m()):n()};e.ticks=function(t){var e,i=n(),s=i[0],a=i[i.length-1];(e=a<s)&&(d=s,s=a,a=d);var c,u,p,d=l(s),h=l(a),m=null==t?10:+t,g=[];if(!(o%1)&&h-d<m){if(d=Math.round(d)-1,h=Math.round(h)+1,s>0){for(;d<h;++d)for(u=1,c=f(d);u<o;++u)if(!((p=c*u)<s)){if(p>a)break;g.push(p)}}else for(;d<h;++d)for(u=o-1,c=f(d);u>=1;--u)if(!((p=c*u)<s)){if(p>a)break;g.push(p)}}else g=Object(r.s)(d,h,Math.min(h-d,m)).map(f);return e?g.reverse():g};e.tickFormat=function(t,n){if(null==n&&(n=10===o?".0e":","),"function"!=typeof n&&(n=Object(i.a)(n)),t===1/0)return n;null==t&&(t=10);var r=Math.max(1,o*t/e.ticks().length);return function(t){var e=t/f(Math.round(l(t)));return e*o<o-.5&&(e*=o),e<=r?n(t):""}};e.nice=function(){return n(Object(s.a)(n(),{floor:function(t){return f(Math.floor(l(t)))},ceil:function(t){return f(Math.ceil(l(t)))}}))};e.copy=function(){return Object(a.a)(e,t().base(o))};return e};var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-format/index.js"),o=n("../../../../d3-scale/src/constant.js"),s=n("../../../../d3-scale/src/nice.js"),a=n("../../../../d3-scale/src/continuous.js");function c(t,e){return(e=Math.log(e/t))?function(n){return Math.log(n/t)/e}:Object(o.a)(e)}function u(t,e){return t<0?function(n){return-Math.pow(-e,n)*Math.pow(-t,1-n)}:function(n){return Math.pow(e,n)*Math.pow(t,1-n)}}function l(t){return isFinite(t)?+("1e"+t):t<0?0:t}function p(t){return 10===t?l:t===Math.E?Math.exp:function(e){return Math.pow(t,e)}}function d(t){return t===Math.E?Math.log:10===t&&Math.log10||2===t&&Math.log2||(t=Math.log(t),function(e){return Math.log(e)/t})}function h(t){return function(e){return-t(-e)}}},"../../../../d3-scale/src/nice.js":function(t,e,n){"use strict";e.a=function(t,e){var n,r=0,i=(t=t.slice()).length-1,o=t[r],s=t[i];return s<o&&(n=r,r=i,i=n,n=o,o=s,s=n),t[r]=e.floor(o),t[i]=e.ceil(s),t}},"../../../../d3-scale/src/number.js":function(t,e,n){"use strict";e.a=function(t){return+t}},"../../../../d3-scale/src/ordinal.js":function(t,e,n){"use strict";e.a=function t(e){var n=Object(r.a)(),s=[],a=o;e=null==e?[]:i.b.call(e);function c(t){var r=t+"",i=n.get(r);if(!i){if(a!==o)return a;n.set(r,i=s.push(t))}return e[(i-1)%e.length]}c.domain=function(t){if(!arguments.length)return s.slice();s=[],n=Object(r.a)();for(var e,i,o=-1,a=t.length;++o<a;)n.has(i=(e=t[o])+"")||n.set(i,s.push(e));return c};c.range=function(t){return arguments.length?(e=i.b.call(t),c):e.slice()};c.unknown=function(t){return arguments.length?(a=t,c):a};c.copy=function(){return t().domain(s).range(e).unknown(a)};return c};var r=n("../../../../d3-collection/index.js"),i=n("../../../../d3-scale/src/array.js"),o={name:"implicit"}},"../../../../d3-scale/src/pow.js":function(t,e,n){"use strict";e.a=a,e.b=function(){return a().exponent(.5)};var r=n("../../../../d3-scale/src/constant.js"),i=n("../../../../d3-scale/src/linear.js"),o=n("../../../../d3-scale/src/continuous.js");function s(t,e){return t<0?-Math.pow(-t,e):Math.pow(t,e)}function a(){var t=1,e=Object(o.b)(function(e,n){return(n=s(n,t)-(e=s(e,t)))?function(r){return(s(r,t)-e)/n}:Object(r.a)(n)},function(e,n){return n=s(n,t)-(e=s(e,t)),function(r){return s(e+n*r,1/t)}}),n=e.domain;return e.exponent=function(e){return arguments.length?(t=+e,n(n())):t},e.copy=function(){return Object(o.a)(e,a().exponent(t))},Object(i.b)(e)}},"../../../../d3-scale/src/quantile.js":function(t,e,n){"use strict";e.a=function t(){var e=[],n=[],o=[];function s(){var t=0,i=Math.max(1,n.length);for(o=new Array(i-1);++t<i;)o[t-1]=Object(r.m)(e,t/i);return a}function a(t){if(!isNaN(t=+t))return n[Object(r.b)(o,t)]}a.invertExtent=function(t){var r=n.indexOf(t);return r<0?[NaN,NaN]:[r>0?o[r-1]:e[0],r<o.length?o[r]:e[e.length-1]]};a.domain=function(t){if(!arguments.length)return e.slice();e=[];for(var n,i=0,o=t.length;i<o;++i)n=t[i],null==n||isNaN(n=+n)||e.push(n);return e.sort(r.a),s()};a.range=function(t){return arguments.length?(n=i.b.call(t),s()):n.slice()};a.quantiles=function(){return o.slice()};a.copy=function(){return t().domain(e).range(n)};return a};var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-scale/src/array.js")},"../../../../d3-scale/src/quantize.js":function(t,e,n){"use strict";e.a=function t(){var e=0,n=1,s=1,a=[.5],c=[0,1];function u(t){if(t<=t)return c[Object(r.b)(a,t,0,s)]}function l(){var t=-1;for(a=new Array(s);++t<s;)a[t]=((t+1)*n-(t-s)*e)/(s+1);return u}u.domain=function(t){return arguments.length?(e=+t[0],n=+t[1],l()):[e,n]};u.range=function(t){return arguments.length?(s=(c=i.b.call(t)).length-1,l()):c.slice()};u.invertExtent=function(t){var r=c.indexOf(t);return r<0?[NaN,NaN]:r<1?[e,a[0]]:r>=s?[a[s-1],n]:[a[r-1],a[r]]};u.copy=function(){return t().domain([e,n]).range(c)};return Object(o.b)(u)};var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-scale/src/array.js"),o=n("../../../../d3-scale/src/linear.js")},"../../../../d3-scale/src/sequential.js":function(t,e,n){"use strict";n("../../../../d3-scale/src/linear.js")},"../../../../d3-scale/src/threshold.js":function(t,e,n){"use strict";e.a=function t(){var e=[.5],n=[0,1],o=1;function s(t){if(t<=t)return n[Object(r.b)(e,t,0,o)]}s.domain=function(t){return arguments.length?(e=i.b.call(t),o=Math.min(e.length,n.length-1),s):e.slice()};s.range=function(t){return arguments.length?(n=i.b.call(t),o=Math.min(e.length,n.length-1),s):n.slice()};s.invertExtent=function(t){var r=n.indexOf(t);return[e[r-1],e[r]]};s.copy=function(){return t().domain(e).range(n)};return s};var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-scale/src/array.js")},"../../../../d3-scale/src/tickFormat.js":function(t,e,n){"use strict";var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-format/index.js");e.a=function(t,e,n){var o,s=t[0],a=t[t.length-1],c=Object(r.r)(s,a,null==e?10:e);switch((n=Object(i.c)(null==n?",f":n)).type){case"s":var u=Math.max(Math.abs(s),Math.abs(a));return null!=n.precision||isNaN(o=Object(i.e)(c,u))||(n.precision=o),Object(i.b)(n,u);case"":case"e":case"g":case"p":case"r":null!=n.precision||isNaN(o=Object(i.f)(c,Math.max(Math.abs(s),Math.abs(a))))||(n.precision=o-("e"===n.type));break;case"f":case"%":null!=n.precision||isNaN(o=Object(i.d)(c))||(n.precision=o-2*("%"===n.type))}return Object(i.a)(n)}},"../../../../d3-scale/src/time.js":function(t,e,n){"use strict";e.a=b;var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-interpolate/index.js"),o=n("../../../../d3-time/index.js"),s=n("../../../../d3-time-format/index.js"),a=n("../../../../d3-scale/src/array.js"),c=n("../../../../d3-scale/src/continuous.js"),u=n("../../../../d3-scale/src/nice.js"),l=1e3,p=60*l,d=60*p,h=24*d,f=7*h,m=30*h,g=365*h;function v(t){return new Date(t)}function y(t){return t instanceof Date?+t:+new Date(+t)}function b(t,e,n,o,s,_,j,x,w){var O=Object(c.b)(c.c,i.interpolateNumber),C=O.invert,k=O.domain,E=w(".%L"),S=w(":%S"),P=w("%I:%M"),M=w("%I %p"),A=w("%a %d"),T=w("%b %d"),D=w("%B"),I=w("%Y"),R=[[j,1,l],[j,5,5*l],[j,15,15*l],[j,30,30*l],[_,1,p],[_,5,5*p],[_,15,15*p],[_,30,30*p],[s,1,d],[s,3,3*d],[s,6,6*d],[s,12,12*d],[o,1,h],[o,2,2*h],[n,1,f],[e,1,m],[e,3,3*m],[t,1,g]];function N(r){return(j(r)<r?E:_(r)<r?S:s(r)<r?P:o(r)<r?M:e(r)<r?n(r)<r?A:T:t(r)<r?D:I)(r)}function L(e,n,i,o){if(null==e&&(e=10),"number"==typeof e){var s=Math.abs(i-n)/e,a=Object(r.e)(function(t){return t[2]}).right(R,s);a===R.length?(o=Object(r.r)(n/g,i/g,e),e=t):a?(o=(a=R[s/R[a-1][2]<R[a][2]/s?a-1:a])[1],e=a[0]):(o=Math.max(Object(r.r)(n,i,e),1),e=x)}return null==o?e:e.every(o)}return O.invert=function(t){return new Date(C(t))},O.domain=function(t){return arguments.length?k(a.a.call(t,y)):k().map(v)},O.ticks=function(t,e){var n,r=k(),i=r[0],o=r[r.length-1],s=o<i;return s&&(n=i,i=o,o=n),n=(n=L(t,i,o,e))?n.range(i,o+1):[],s?n.reverse():n},O.tickFormat=function(t,e){return null==e?N:w(e)},O.nice=function(t,e){var n=k();return(t=L(t,n[0],n[n.length-1],e))?k(Object(u.a)(n,t)):O},O.copy=function(){return Object(c.a)(O,b(t,e,n,o,s,_,j,x,w))},O}e.b=function(){return b(o.k,o.f,o.j,o.a,o.b,o.d,o.g,o.c,s.a).domain([new Date(2e3,0,1),new Date(2e3,0,2)])}},"../../../../d3-scale/src/utcTime.js":function(t,e,n){"use strict";var r=n("../../../../d3-scale/src/time.js"),i=n("../../../../d3-time-format/index.js"),o=n("../../../../d3-time/index.js");e.a=function(){return Object(r.a)(o.v,o.q,o.u,o.l,o.m,o.o,o.r,o.n,i.c).domain([Date.UTC(2e3,0,1),Date.UTC(2e3,0,2)])}},"../../../../d3-shape/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-shape/src/arc.js");n.d(e,"a",function(){return r.a});var i=n("../../../../d3-shape/src/area.js");n.d(e,"b",function(){return i.a});var o=n("../../../../d3-shape/src/line.js");n.d(e,"u",function(){return o.a});n("../../../../d3-shape/src/pie.js"),n("../../../../d3-shape/src/areaRadial.js"),n("../../../../d3-shape/src/lineRadial.js"),n("../../../../d3-shape/src/pointRadial.js"),n("../../../../d3-shape/src/link/index.js");var s=n("../../../../d3-shape/src/symbol.js");n.d(e,"v",function(){return s.a});n("../../../../d3-shape/src/symbol/circle.js"),n("../../../../d3-shape/src/symbol/cross.js"),n("../../../../d3-shape/src/symbol/diamond.js"),n("../../../../d3-shape/src/symbol/square.js"),n("../../../../d3-shape/src/symbol/star.js"),n("../../../../d3-shape/src/symbol/triangle.js"),n("../../../../d3-shape/src/symbol/wye.js");var a=n("../../../../d3-shape/src/curve/basisClosed.js");n.d(e,"d",function(){return a.a});var c=n("../../../../d3-shape/src/curve/basisOpen.js");n.d(e,"e",function(){return c.a});var u=n("../../../../d3-shape/src/curve/basis.js");n.d(e,"c",function(){return u.b});var l=n("../../../../d3-shape/src/curve/bundle.js");n.d(e,"f",function(){return l.a});var p=n("../../../../d3-shape/src/curve/cardinalClosed.js");n.d(e,"h",function(){return p.b});var d=n("../../../../d3-shape/src/curve/cardinalOpen.js");n.d(e,"i",function(){return d.b});var h=n("../../../../d3-shape/src/curve/cardinal.js");n.d(e,"g",function(){return h.b});var f=n("../../../../d3-shape/src/curve/catmullRomClosed.js");n.d(e,"k",function(){return f.a});var m=n("../../../../d3-shape/src/curve/catmullRomOpen.js");n.d(e,"l",function(){return m.a});var g=n("../../../../d3-shape/src/curve/catmullRom.js");n.d(e,"j",function(){return g.a});var v=n("../../../../d3-shape/src/curve/linearClosed.js");n.d(e,"n",function(){return v.a});var y=n("../../../../d3-shape/src/curve/linear.js");n.d(e,"m",function(){return y.a});var b=n("../../../../d3-shape/src/curve/monotone.js");n.d(e,"o",function(){return b.a}),n.d(e,"p",function(){return b.b});var _=n("../../../../d3-shape/src/curve/natural.js");n.d(e,"q",function(){return _.a});var j=n("../../../../d3-shape/src/curve/step.js");n.d(e,"r",function(){return j.a}),n.d(e,"s",function(){return j.b}),n.d(e,"t",function(){return j.c});n("../../../../d3-shape/src/stack.js"),n("../../../../d3-shape/src/offset/expand.js"),n("../../../../d3-shape/src/offset/diverging.js"),n("../../../../d3-shape/src/offset/none.js"),n("../../../../d3-shape/src/offset/silhouette.js"),n("../../../../d3-shape/src/offset/wiggle.js"),n("../../../../d3-shape/src/order/ascending.js"),n("../../../../d3-shape/src/order/descending.js"),n("../../../../d3-shape/src/order/insideOut.js"),n("../../../../d3-shape/src/order/none.js"),n("../../../../d3-shape/src/order/reverse.js")},"../../../../d3-shape/src/arc.js":function(t,e,n){"use strict";var r=n("../../../../d3-path/index.js"),i=n("../../../../d3-shape/src/constant.js"),o=n("../../../../d3-shape/src/math.js");function s(t){return t.innerRadius}function a(t){return t.outerRadius}function c(t){return t.startAngle}function u(t){return t.endAngle}function l(t){return t&&t.padAngle}function p(t,e,n,r,i,s,a){var c=t-n,u=e-r,l=(a?s:-s)/Object(o.l)(c*c+u*u),p=l*u,d=-l*c,h=t+p,f=e+d,m=n+p,g=r+d,v=(h+m)/2,y=(f+g)/2,b=m-h,_=g-f,j=b*b+_*_,x=i-s,w=h*g-m*f,O=(_<0?-1:1)*Object(o.l)(Object(o.h)(0,x*x*j-w*w)),C=(w*_-b*O)/j,k=(-w*b-_*O)/j,E=(w*_+b*O)/j,S=(-w*b+_*O)/j,P=C-v,M=k-y,A=E-v,T=S-y;return P*P+M*M>A*A+T*T&&(C=E,k=S),{cx:C,cy:k,x01:-p,y01:-d,x11:C*(i/x-1),y11:k*(i/x-1)}}e.a=function(){var t=s,e=a,n=Object(i.a)(0),d=null,h=c,f=u,m=l,g=null;function v(){var i,s,a,c,u,l,v,y,b,_,j,x=+t.apply(this,arguments),w=+e.apply(this,arguments),O=h.apply(this,arguments)-o.g,C=f.apply(this,arguments)-o.g,k=Object(o.a)(C-O),E=C>O;if(g||(g=i=Object(r.a)()),w<x&&(s=w,w=x,x=s),w>o.f)if(k>o.m-o.f)g.moveTo(w*Object(o.e)(O),w*Object(o.k)(O)),g.arc(0,0,w,O,C,!E),x>o.f&&(g.moveTo(x*Object(o.e)(C),x*Object(o.k)(C)),g.arc(0,0,x,C,O,E));else{var S,P,M=O,A=C,T=O,D=C,I=k,R=k,N=m.apply(this,arguments)/2,L=N>o.f&&(d?+d.apply(this,arguments):Object(o.l)(x*x+w*w)),F=Object(o.i)(Object(o.a)(w-x)/2,+n.apply(this,arguments)),V=F,z=F;if(L>o.f){var B=Object(o.c)(L/x*Object(o.k)(N)),q=Object(o.c)(L/w*Object(o.k)(N));(I-=2*B)>o.f?(T+=B*=E?1:-1,D-=B):(I=0,T=D=(O+C)/2),(R-=2*q)>o.f?(M+=q*=E?1:-1,A-=q):(R=0,M=A=(O+C)/2)}var H=w*Object(o.e)(M),U=w*Object(o.k)(M),W=x*Object(o.e)(D),G=x*Object(o.k)(D);if(F>o.f){var X=w*Object(o.e)(A),Y=w*Object(o.k)(A),K=x*Object(o.e)(T),$=x*Object(o.k)(T);if(k<o.j){var Q=I>o.f?[(a=H)+(j=((b=W-(u=X))*((c=U)-(l=Y))-(_=G-l)*(a-u))/(_*(v=K-a)-b*(y=$-c)))*v,c+j*y]:[W,G],Z=H-Q[0],J=U-Q[1],tt=X-Q[0],et=Y-Q[1],nt=1/Object(o.k)(Object(o.b)((Z*tt+J*et)/(Object(o.l)(Z*Z+J*J)*Object(o.l)(tt*tt+et*et)))/2),rt=Object(o.l)(Q[0]*Q[0]+Q[1]*Q[1]);V=Object(o.i)(F,(x-rt)/(nt-1)),z=Object(o.i)(F,(w-rt)/(nt+1))}}R>o.f?z>o.f?(S=p(K,$,H,U,w,z,E),P=p(X,Y,W,G,w,z,E),g.moveTo(S.cx+S.x01,S.cy+S.y01),z<F?g.arc(S.cx,S.cy,z,Object(o.d)(S.y01,S.x01),Object(o.d)(P.y01,P.x01),!E):(g.arc(S.cx,S.cy,z,Object(o.d)(S.y01,S.x01),Object(o.d)(S.y11,S.x11),!E),g.arc(0,0,w,Object(o.d)(S.cy+S.y11,S.cx+S.x11),Object(o.d)(P.cy+P.y11,P.cx+P.x11),!E),g.arc(P.cx,P.cy,z,Object(o.d)(P.y11,P.x11),Object(o.d)(P.y01,P.x01),!E))):(g.moveTo(H,U),g.arc(0,0,w,M,A,!E)):g.moveTo(H,U),x>o.f&&I>o.f?V>o.f?(S=p(W,G,X,Y,x,-V,E),P=p(H,U,K,$,x,-V,E),g.lineTo(S.cx+S.x01,S.cy+S.y01),V<F?g.arc(S.cx,S.cy,V,Object(o.d)(S.y01,S.x01),Object(o.d)(P.y01,P.x01),!E):(g.arc(S.cx,S.cy,V,Object(o.d)(S.y01,S.x01),Object(o.d)(S.y11,S.x11),!E),g.arc(0,0,x,Object(o.d)(S.cy+S.y11,S.cx+S.x11),Object(o.d)(P.cy+P.y11,P.cx+P.x11),E),g.arc(P.cx,P.cy,V,Object(o.d)(P.y11,P.x11),Object(o.d)(P.y01,P.x01),!E))):g.arc(0,0,x,D,T,E):g.lineTo(W,G)}else g.moveTo(0,0);if(g.closePath(),i)return g=null,i+""||null}return v.centroid=function(){var n=(+t.apply(this,arguments)+ +e.apply(this,arguments))/2,r=(+h.apply(this,arguments)+ +f.apply(this,arguments))/2-o.j/2;return[Object(o.e)(r)*n,Object(o.k)(r)*n]},v.innerRadius=function(e){return arguments.length?(t="function"==typeof e?e:Object(i.a)(+e),v):t},v.outerRadius=function(t){return arguments.length?(e="function"==typeof t?t:Object(i.a)(+t),v):e},v.cornerRadius=function(t){return arguments.length?(n="function"==typeof t?t:Object(i.a)(+t),v):n},v.padRadius=function(t){return arguments.length?(d=null==t?null:"function"==typeof t?t:Object(i.a)(+t),v):d},v.startAngle=function(t){return arguments.length?(h="function"==typeof t?t:Object(i.a)(+t),v):h},v.endAngle=function(t){return arguments.length?(f="function"==typeof t?t:Object(i.a)(+t),v):f},v.padAngle=function(t){return arguments.length?(m="function"==typeof t?t:Object(i.a)(+t),v):m},v.context=function(t){return arguments.length?(g=null==t?null:t,v):g},v}},"../../../../d3-shape/src/area.js":function(t,e,n){"use strict";var r=n("../../../../d3-path/index.js"),i=n("../../../../d3-shape/src/constant.js"),o=n("../../../../d3-shape/src/curve/linear.js"),s=n("../../../../d3-shape/src/line.js"),a=n("../../../../d3-shape/src/point.js");e.a=function(){var t=a.a,e=null,n=Object(i.a)(0),c=a.b,u=Object(i.a)(!0),l=null,p=o.a,d=null;function h(i){var o,s,a,h,f,m=i.length,g=!1,v=new Array(m),y=new Array(m);for(null==l&&(d=p(f=Object(r.a)())),o=0;o<=m;++o){if(!(o<m&&u(h=i[o],o,i))===g)if(g=!g)s=o,d.areaStart(),d.lineStart();else{for(d.lineEnd(),d.lineStart(),a=o-1;a>=s;--a)d.point(v[a],y[a]);d.lineEnd(),d.areaEnd()}g&&(v[o]=+t(h,o,i),y[o]=+n(h,o,i),d.point(e?+e(h,o,i):v[o],c?+c(h,o,i):y[o]))}if(f)return d=null,f+""||null}function f(){return Object(s.a)().defined(u).curve(p).context(l)}return h.x=function(n){return arguments.length?(t="function"==typeof n?n:Object(i.a)(+n),e=null,h):t},h.x0=function(e){return arguments.length?(t="function"==typeof e?e:Object(i.a)(+e),h):t},h.x1=function(t){return arguments.length?(e=null==t?null:"function"==typeof t?t:Object(i.a)(+t),h):e},h.y=function(t){return arguments.length?(n="function"==typeof t?t:Object(i.a)(+t),c=null,h):n},h.y0=function(t){return arguments.length?(n="function"==typeof t?t:Object(i.a)(+t),h):n},h.y1=function(t){return arguments.length?(c=null==t?null:"function"==typeof t?t:Object(i.a)(+t),h):c},h.lineX0=h.lineY0=function(){return f().x(t).y(n)},h.lineY1=function(){return f().x(t).y(c)},h.lineX1=function(){return f().x(e).y(n)},h.defined=function(t){return arguments.length?(u="function"==typeof t?t:Object(i.a)(!!t),h):u},h.curve=function(t){return arguments.length?(p=t,null!=l&&(d=p(l)),h):p},h.context=function(t){return arguments.length?(null==t?l=d=null:d=p(l=t),h):l},h}},"../../../../d3-shape/src/areaRadial.js":function(t,e,n){"use strict";n("../../../../d3-shape/src/curve/radial.js"),n("../../../../d3-shape/src/area.js"),n("../../../../d3-shape/src/lineRadial.js")},"../../../../d3-shape/src/array.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=Array.prototype.slice},"../../../../d3-shape/src/constant.js":function(t,e,n){"use strict";e.a=function(t){return function(){return t}}},"../../../../d3-shape/src/curve/basis.js":function(t,e,n){"use strict";function r(t,e,n){t._context.bezierCurveTo((2*t._x0+t._x1)/3,(2*t._y0+t._y1)/3,(t._x0+2*t._x1)/3,(t._y0+2*t._y1)/3,(t._x0+4*t._x1+e)/6,(t._y0+4*t._y1+n)/6)}function i(t){this._context=t}e.c=r,e.a=i,i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){switch(this._point){case 3:r(this,this._x1,this._y1);case 2:this._context.lineTo(this._x1,this._y1)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;break;case 2:this._point=3,this._context.lineTo((5*this._x0+this._x1)/6,(5*this._y0+this._y1)/6);default:r(this,t,e)}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e}},e.b=function(t){return new i(t)}},"../../../../d3-shape/src/curve/basisClosed.js":function(t,e,n){"use strict";var r=n("../../../../d3-shape/src/noop.js"),i=n("../../../../d3-shape/src/curve/basis.js");function o(t){this._context=t}o.prototype={areaStart:r.a,areaEnd:r.a,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._y0=this._y1=this._y2=this._y3=this._y4=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x2,this._y2),this._context.closePath();break;case 2:this._context.moveTo((this._x2+2*this._x3)/3,(this._y2+2*this._y3)/3),this._context.lineTo((this._x3+2*this._x2)/3,(this._y3+2*this._y2)/3),this._context.closePath();break;case 3:this.point(this._x2,this._y2),this.point(this._x3,this._y3),this.point(this._x4,this._y4)}},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._x2=t,this._y2=e;break;case 1:this._point=2,this._x3=t,this._y3=e;break;case 2:this._point=3,this._x4=t,this._y4=e,this._context.moveTo((this._x0+4*this._x1+t)/6,(this._y0+4*this._y1+e)/6);break;default:Object(i.c)(this,t,e)}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e}},e.a=function(t){return new o(t)}},"../../../../d3-shape/src/curve/basisOpen.js":function(t,e,n){"use strict";var r=n("../../../../d3-shape/src/curve/basis.js");function i(t){this._context=t}i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3;var n=(this._x0+4*this._x1+t)/6,i=(this._y0+4*this._y1+e)/6;this._line?this._context.lineTo(n,i):this._context.moveTo(n,i);break;case 3:this._point=4;default:Object(r.c)(this,t,e)}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e}},e.a=function(t){return new i(t)}},"../../../../d3-shape/src/curve/bundle.js":function(t,e,n){"use strict";var r=n("../../../../d3-shape/src/curve/basis.js");function i(t,e){this._basis=new r.a(t),this._beta=e}i.prototype={lineStart:function(){this._x=[],this._y=[],this._basis.lineStart()},lineEnd:function(){var t=this._x,e=this._y,n=t.length-1;if(n>0)for(var r,i=t[0],o=e[0],s=t[n]-i,a=e[n]-o,c=-1;++c<=n;)r=c/n,this._basis.point(this._beta*t[c]+(1-this._beta)*(i+r*s),this._beta*e[c]+(1-this._beta)*(o+r*a));this._x=this._y=null,this._basis.lineEnd()},point:function(t,e){this._x.push(+t),this._y.push(+e)}},e.a=function t(e){function n(t){return 1===e?new r.a(t):new i(t,e)}return n.beta=function(e){return t(+e)},n}(.85)},"../../../../d3-shape/src/curve/cardinal.js":function(t,e,n){"use strict";function r(t,e,n){t._context.bezierCurveTo(t._x1+t._k*(t._x2-t._x0),t._y1+t._k*(t._y2-t._y0),t._x2+t._k*(t._x1-e),t._y2+t._k*(t._y1-n),t._x2,t._y2)}function i(t,e){this._context=t,this._k=(1-e)/6}e.c=r,e.a=i,i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x2,this._y2);break;case 3:r(this,this._x1,this._y1)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2,this._x1=t,this._y1=e;break;case 2:this._point=3;default:r(this,t,e)}this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.b=function t(e){function n(t){return new i(t,e)}return n.tension=function(e){return t(+e)},n}(0)},"../../../../d3-shape/src/curve/cardinalClosed.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../d3-shape/src/noop.js"),i=n("../../../../d3-shape/src/curve/cardinal.js");function o(t,e){this._context=t,this._k=(1-e)/6}o.prototype={areaStart:r.a,areaEnd:r.a,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._x5=this._y0=this._y1=this._y2=this._y3=this._y4=this._y5=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x3,this._y3),this._context.closePath();break;case 2:this._context.lineTo(this._x3,this._y3),this._context.closePath();break;case 3:this.point(this._x3,this._y3),this.point(this._x4,this._y4),this.point(this._x5,this._y5)}},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._x3=t,this._y3=e;break;case 1:this._point=2,this._context.moveTo(this._x4=t,this._y4=e);break;case 2:this._point=3,this._x5=t,this._y5=e;break;default:Object(i.c)(this,t,e)}this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.b=function t(e){function n(t){return new o(t,e)}return n.tension=function(e){return t(+e)},n}(0)},"../../../../d3-shape/src/curve/cardinalOpen.js":function(t,e,n){"use strict";e.a=i;var r=n("../../../../d3-shape/src/curve/cardinal.js");function i(t,e){this._context=t,this._k=(1-e)/6}i.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3,this._line?this._context.lineTo(this._x2,this._y2):this._context.moveTo(this._x2,this._y2);break;case 3:this._point=4;default:Object(r.c)(this,t,e)}this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.b=function t(e){function n(t){return new i(t,e)}return n.tension=function(e){return t(+e)},n}(0)},"../../../../d3-shape/src/curve/catmullRom.js":function(t,e,n){"use strict";e.b=o;var r=n("../../../../d3-shape/src/math.js"),i=n("../../../../d3-shape/src/curve/cardinal.js");function o(t,e,n){var i=t._x1,o=t._y1,s=t._x2,a=t._y2;if(t._l01_a>r.f){var c=2*t._l01_2a+3*t._l01_a*t._l12_a+t._l12_2a,u=3*t._l01_a*(t._l01_a+t._l12_a);i=(i*c-t._x0*t._l12_2a+t._x2*t._l01_2a)/u,o=(o*c-t._y0*t._l12_2a+t._y2*t._l01_2a)/u}if(t._l23_a>r.f){var l=2*t._l23_2a+3*t._l23_a*t._l12_a+t._l12_2a,p=3*t._l23_a*(t._l23_a+t._l12_a);s=(s*l+t._x1*t._l23_2a-e*t._l12_2a)/p,a=(a*l+t._y1*t._l23_2a-n*t._l12_2a)/p}t._context.bezierCurveTo(i,o,s,a,t._x2,t._y2)}function s(t,e){this._context=t,this._alpha=e}s.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x2,this._y2);break;case 3:this.point(this._x2,this._y2)}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){if(t=+t,e=+e,this._point){var n=this._x2-t,r=this._y2-e;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;break;case 2:this._point=3;default:o(this,t,e)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.a=function t(e){function n(t){return e?new s(t,e):new i.a(t,0)}return n.alpha=function(e){return t(+e)},n}(.5)},"../../../../d3-shape/src/curve/catmullRomClosed.js":function(t,e,n){"use strict";var r=n("../../../../d3-shape/src/curve/cardinalClosed.js"),i=n("../../../../d3-shape/src/noop.js"),o=n("../../../../d3-shape/src/curve/catmullRom.js");function s(t,e){this._context=t,this._alpha=e}s.prototype={areaStart:i.a,areaEnd:i.a,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._x5=this._y0=this._y1=this._y2=this._y3=this._y4=this._y5=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){switch(this._point){case 1:this._context.moveTo(this._x3,this._y3),this._context.closePath();break;case 2:this._context.lineTo(this._x3,this._y3),this._context.closePath();break;case 3:this.point(this._x3,this._y3),this.point(this._x4,this._y4),this.point(this._x5,this._y5)}},point:function(t,e){if(t=+t,e=+e,this._point){var n=this._x2-t,r=this._y2-e;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1,this._x3=t,this._y3=e;break;case 1:this._point=2,this._context.moveTo(this._x4=t,this._y4=e);break;case 2:this._point=3,this._x5=t,this._y5=e;break;default:Object(o.b)(this,t,e)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.a=function t(e){function n(t){return e?new s(t,e):new r.a(t,0)}return n.alpha=function(e){return t(+e)},n}(.5)},"../../../../d3-shape/src/curve/catmullRomOpen.js":function(t,e,n){"use strict";var r=n("../../../../d3-shape/src/curve/cardinalOpen.js"),i=n("../../../../d3-shape/src/curve/catmullRom.js");function o(t,e){this._context=t,this._alpha=e}o.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._x2=this._y0=this._y1=this._y2=NaN,this._l01_a=this._l12_a=this._l23_a=this._l01_2a=this._l12_2a=this._l23_2a=this._point=0},lineEnd:function(){(this._line||0!==this._line&&3===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){if(t=+t,e=+e,this._point){var n=this._x2-t,r=this._y2-e;this._l23_a=Math.sqrt(this._l23_2a=Math.pow(n*n+r*r,this._alpha))}switch(this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3,this._line?this._context.lineTo(this._x2,this._y2):this._context.moveTo(this._x2,this._y2);break;case 3:this._point=4;default:Object(i.b)(this,t,e)}this._l01_a=this._l12_a,this._l12_a=this._l23_a,this._l01_2a=this._l12_2a,this._l12_2a=this._l23_2a,this._x0=this._x1,this._x1=this._x2,this._x2=t,this._y0=this._y1,this._y1=this._y2,this._y2=e}},e.a=function t(e){function n(t){return e?new o(t,e):new r.a(t,0)}return n.alpha=function(e){return t(+e)},n}(.5)},"../../../../d3-shape/src/curve/linear.js":function(t,e,n){"use strict";function r(t){this._context=t}r.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;default:this._context.lineTo(t,e)}}},e.a=function(t){return new r(t)}},"../../../../d3-shape/src/curve/linearClosed.js":function(t,e,n){"use strict";var r=n("../../../../d3-shape/src/noop.js");function i(t){this._context=t}i.prototype={areaStart:r.a,areaEnd:r.a,lineStart:function(){this._point=0},lineEnd:function(){this._point&&this._context.closePath()},point:function(t,e){t=+t,e=+e,this._point?this._context.lineTo(t,e):(this._point=1,this._context.moveTo(t,e))}},e.a=function(t){return new i(t)}},"../../../../d3-shape/src/curve/monotone.js":function(t,e,n){"use strict";function r(t){return t<0?-1:1}function i(t,e,n){var i=t._x1-t._x0,o=e-t._x1,s=(t._y1-t._y0)/(i||o<0&&-0),a=(n-t._y1)/(o||i<0&&-0),c=(s*o+a*i)/(i+o);return(r(s)+r(a))*Math.min(Math.abs(s),Math.abs(a),.5*Math.abs(c))||0}function o(t,e){var n=t._x1-t._x0;return n?(3*(t._y1-t._y0)/n-e)/2:e}function s(t,e,n){var r=t._x0,i=t._y0,o=t._x1,s=t._y1,a=(o-r)/3;t._context.bezierCurveTo(r+a,i+a*e,o-a,s-a*n,o,s)}function a(t){this._context=t}function c(t){this._context=new u(t)}function u(t){this._context=t}e.a=function(t){return new a(t)},e.b=function(t){return new c(t)},a.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=this._t0=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x1,this._y1);break;case 3:s(this,this._t0,o(this,this._t0))}(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line=1-this._line},point:function(t,e){var n=NaN;if(e=+e,(t=+t)!==this._x1||e!==this._y1){switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;break;case 2:this._point=3,s(this,o(this,n=i(this,t,e)),n);break;default:s(this,this._t0,n=i(this,t,e))}this._x0=this._x1,this._x1=t,this._y0=this._y1,this._y1=e,this._t0=n}}},(c.prototype=Object.create(a.prototype)).point=function(t,e){a.prototype.point.call(this,e,t)},u.prototype={moveTo:function(t,e){this._context.moveTo(e,t)},closePath:function(){this._context.closePath()},lineTo:function(t,e){this._context.lineTo(e,t)},bezierCurveTo:function(t,e,n,r,i,o){this._context.bezierCurveTo(e,t,r,n,o,i)}}},"../../../../d3-shape/src/curve/natural.js":function(t,e,n){"use strict";function r(t){this._context=t}function i(t){var e,n,r=t.length-1,i=new Array(r),o=new Array(r),s=new Array(r);for(i[0]=0,o[0]=2,s[0]=t[0]+2*t[1],e=1;e<r-1;++e)i[e]=1,o[e]=4,s[e]=4*t[e]+2*t[e+1];for(i[r-1]=2,o[r-1]=7,s[r-1]=8*t[r-1]+t[r],e=1;e<r;++e)n=i[e]/o[e-1],o[e]-=n,s[e]-=n*s[e-1];for(i[r-1]=s[r-1]/o[r-1],e=r-2;e>=0;--e)i[e]=(s[e]-i[e+1])/o[e];for(o[r-1]=(t[r]+i[r-1])/2,e=0;e<r-1;++e)o[e]=2*t[e+1]-i[e+1];return[i,o]}r.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=[],this._y=[]},lineEnd:function(){var t=this._x,e=this._y,n=t.length;if(n)if(this._line?this._context.lineTo(t[0],e[0]):this._context.moveTo(t[0],e[0]),2===n)this._context.lineTo(t[1],e[1]);else for(var r=i(t),o=i(e),s=0,a=1;a<n;++s,++a)this._context.bezierCurveTo(r[0][s],o[0][s],r[1][s],o[1][s],t[a],e[a]);(this._line||0!==this._line&&1===n)&&this._context.closePath(),this._line=1-this._line,this._x=this._y=null},point:function(t,e){this._x.push(+t),this._y.push(+e)}},e.a=function(t){return new r(t)}},"../../../../d3-shape/src/curve/radial.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r}),e.b=o;var r=o(n("../../../../d3-shape/src/curve/linear.js").a);function i(t){this._curve=t}function o(t){function e(e){return new i(t(e))}return e._curve=t,e}i.prototype={areaStart:function(){this._curve.areaStart()},areaEnd:function(){this._curve.areaEnd()},lineStart:function(){this._curve.lineStart()},lineEnd:function(){this._curve.lineEnd()},point:function(t,e){this._curve.point(e*Math.sin(t),e*-Math.cos(t))}}},"../../../../d3-shape/src/curve/step.js":function(t,e,n){"use strict";function r(t,e){this._context=t,this._t=e}e.c=function(t){return new r(t,0)},e.b=function(t){return new r(t,1)},r.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=this._y=NaN,this._point=0},lineEnd:function(){0<this._t&&this._t<1&&2===this._point&&this._context.lineTo(this._x,this._y),(this._line||0!==this._line&&1===this._point)&&this._context.closePath(),this._line>=0&&(this._t=1-this._t,this._line=1-this._line)},point:function(t,e){switch(t=+t,e=+e,this._point){case 0:this._point=1,this._line?this._context.lineTo(t,e):this._context.moveTo(t,e);break;case 1:this._point=2;default:if(this._t<=0)this._context.lineTo(this._x,e),this._context.lineTo(t,e);else{var n=this._x*(1-this._t)+t*this._t;this._context.lineTo(n,this._y),this._context.lineTo(n,e)}}this._x=t,this._y=e}},e.a=function(t){return new r(t,.5)}},"../../../../d3-shape/src/descending.js":function(t,e,n){"use strict";e.a=function(t,e){return e<t?-1:e>t?1:e>=t?0:NaN}},"../../../../d3-shape/src/identity.js":function(t,e,n){"use strict";e.a=function(t){return t}},"../../../../d3-shape/src/line.js":function(t,e,n){"use strict";var r=n("../../../../d3-path/index.js"),i=n("../../../../d3-shape/src/constant.js"),o=n("../../../../d3-shape/src/curve/linear.js"),s=n("../../../../d3-shape/src/point.js");e.a=function(){var t=s.a,e=s.b,n=Object(i.a)(!0),a=null,c=o.a,u=null;function l(i){var o,s,l,p=i.length,d=!1;for(null==a&&(u=c(l=Object(r.a)())),o=0;o<=p;++o)!(o<p&&n(s=i[o],o,i))===d&&((d=!d)?u.lineStart():u.lineEnd()),d&&u.point(+t(s,o,i),+e(s,o,i));if(l)return u=null,l+""||null}return l.x=function(e){return arguments.length?(t="function"==typeof e?e:Object(i.a)(+e),l):t},l.y=function(t){return arguments.length?(e="function"==typeof t?t:Object(i.a)(+t),l):e},l.defined=function(t){return arguments.length?(n="function"==typeof t?t:Object(i.a)(!!t),l):n},l.curve=function(t){return arguments.length?(c=t,null!=a&&(u=c(a)),l):c},l.context=function(t){return arguments.length?(null==t?a=u=null:u=c(a=t),l):a},l}},"../../../../d3-shape/src/lineRadial.js":function(t,e,n){"use strict";e.a=i;var r=n("../../../../d3-shape/src/curve/radial.js");n("../../../../d3-shape/src/line.js");function i(t){var e=t.curve;return t.angle=t.x,delete t.x,t.radius=t.y,delete t.y,t.curve=function(t){return arguments.length?e(Object(r.b)(t)):e()._curve},t}},"../../../../d3-shape/src/link/index.js":function(t,e,n){"use strict";n("../../../../d3-path/index.js"),n("../../../../d3-shape/src/array.js"),n("../../../../d3-shape/src/constant.js"),n("../../../../d3-shape/src/point.js"),n("../../../../d3-shape/src/pointRadial.js")},"../../../../d3-shape/src/math.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r}),n.d(e,"d",function(){return i}),n.d(e,"e",function(){return o}),n.d(e,"h",function(){return s}),n.d(e,"i",function(){return a}),n.d(e,"k",function(){return c}),n.d(e,"l",function(){return u}),n.d(e,"f",function(){return l}),n.d(e,"j",function(){return p}),n.d(e,"g",function(){return d}),n.d(e,"m",function(){return h}),e.b=function(t){return t>1?0:t<-1?p:Math.acos(t)},e.c=function(t){return t>=1?d:t<=-1?-d:Math.asin(t)};var r=Math.abs,i=Math.atan2,o=Math.cos,s=Math.max,a=Math.min,c=Math.sin,u=Math.sqrt,l=1e-12,p=Math.PI,d=p/2,h=2*p},"../../../../d3-shape/src/noop.js":function(t,e,n){"use strict";e.a=function(){}},"../../../../d3-shape/src/offset/diverging.js":function(t,e,n){"use strict"},"../../../../d3-shape/src/offset/expand.js":function(t,e,n){"use strict";n("../../../../d3-shape/src/offset/none.js")},"../../../../d3-shape/src/offset/none.js":function(t,e,n){"use strict";e.a=function(t,e){if((i=t.length)>1)for(var n,r,i,o=1,s=t[e[0]],a=s.length;o<i;++o)for(r=s,s=t[e[o]],n=0;n<a;++n)s[n][1]+=s[n][0]=isNaN(r[n][1])?r[n][0]:r[n][1]}},"../../../../d3-shape/src/offset/silhouette.js":function(t,e,n){"use strict";n("../../../../d3-shape/src/offset/none.js")},"../../../../d3-shape/src/offset/wiggle.js":function(t,e,n){"use strict";n("../../../../d3-shape/src/offset/none.js")},"../../../../d3-shape/src/order/ascending.js":function(t,e,n){"use strict";e.b=i;var r=n("../../../../d3-shape/src/order/none.js");function i(t){for(var e,n=0,r=-1,i=t.length;++r<i;)(e=+t[r][1])&&(n+=e);return n}e.a=function(t){var e=t.map(i);return Object(r.a)(t).sort(function(t,n){return e[t]-e[n]})}},"../../../../d3-shape/src/order/descending.js":function(t,e,n){"use strict";n("../../../../d3-shape/src/order/ascending.js")},"../../../../d3-shape/src/order/insideOut.js":function(t,e,n){"use strict";n("../../../../d3-shape/src/order/none.js"),n("../../../../d3-shape/src/order/ascending.js")},"../../../../d3-shape/src/order/none.js":function(t,e,n){"use strict";e.a=function(t){for(var e=t.length,n=new Array(e);--e>=0;)n[e]=e;return n}},"../../../../d3-shape/src/order/reverse.js":function(t,e,n){"use strict";n("../../../../d3-shape/src/order/none.js")},"../../../../d3-shape/src/pie.js":function(t,e,n){"use strict";n("../../../../d3-shape/src/constant.js"),n("../../../../d3-shape/src/descending.js"),n("../../../../d3-shape/src/identity.js"),n("../../../../d3-shape/src/math.js")},"../../../../d3-shape/src/point.js":function(t,e,n){"use strict";e.a=function(t){return t[0]},e.b=function(t){return t[1]}},"../../../../d3-shape/src/pointRadial.js":function(t,e,n){"use strict";e.a=function(t,e){return[(e=+e)*Math.cos(t-=Math.PI/2),e*Math.sin(t)]}},"../../../../d3-shape/src/stack.js":function(t,e,n){"use strict";n("../../../../d3-shape/src/array.js"),n("../../../../d3-shape/src/constant.js"),n("../../../../d3-shape/src/offset/none.js"),n("../../../../d3-shape/src/order/none.js")},"../../../../d3-shape/src/symbol.js":function(t,e,n){"use strict";var r=n("../../../../d3-path/index.js"),i=n("../../../../d3-shape/src/symbol/circle.js"),o=n("../../../../d3-shape/src/symbol/cross.js"),s=n("../../../../d3-shape/src/symbol/diamond.js"),a=n("../../../../d3-shape/src/symbol/star.js"),c=n("../../../../d3-shape/src/symbol/square.js"),u=n("../../../../d3-shape/src/symbol/triangle.js"),l=n("../../../../d3-shape/src/symbol/wye.js"),p=n("../../../../d3-shape/src/constant.js");i.a,o.a,s.a,c.a,a.a,u.a,l.a;e.a=function(){var t=Object(p.a)(i.a),e=Object(p.a)(64),n=null;function o(){var i;if(n||(n=i=Object(r.a)()),t.apply(this,arguments).draw(n,+e.apply(this,arguments)),i)return n=null,i+""||null}return o.type=function(e){return arguments.length?(t="function"==typeof e?e:Object(p.a)(e),o):t},o.size=function(t){return arguments.length?(e="function"==typeof t?t:Object(p.a)(+t),o):e},o.context=function(t){return arguments.length?(n=null==t?null:t,o):n},o}},"../../../../d3-shape/src/symbol/circle.js":function(t,e,n){"use strict";var r=n("../../../../d3-shape/src/math.js");e.a={draw:function(t,e){var n=Math.sqrt(e/r.j);t.moveTo(n,0),t.arc(0,0,n,0,r.m)}}},"../../../../d3-shape/src/symbol/cross.js":function(t,e,n){"use strict";e.a={draw:function(t,e){var n=Math.sqrt(e/5)/2;t.moveTo(-3*n,-n),t.lineTo(-n,-n),t.lineTo(-n,-3*n),t.lineTo(n,-3*n),t.lineTo(n,-n),t.lineTo(3*n,-n),t.lineTo(3*n,n),t.lineTo(n,n),t.lineTo(n,3*n),t.lineTo(-n,3*n),t.lineTo(-n,n),t.lineTo(-3*n,n),t.closePath()}}},"../../../../d3-shape/src/symbol/diamond.js":function(t,e,n){"use strict";var r=Math.sqrt(1/3),i=2*r;e.a={draw:function(t,e){var n=Math.sqrt(e/i),o=n*r;t.moveTo(0,-n),t.lineTo(o,0),t.lineTo(0,n),t.lineTo(-o,0),t.closePath()}}},"../../../../d3-shape/src/symbol/square.js":function(t,e,n){"use strict";e.a={draw:function(t,e){var n=Math.sqrt(e),r=-n/2;t.rect(r,r,n,n)}}},"../../../../d3-shape/src/symbol/star.js":function(t,e,n){"use strict";var r=n("../../../../d3-shape/src/math.js"),i=Math.sin(r.j/10)/Math.sin(7*r.j/10),o=Math.sin(r.m/10)*i,s=-Math.cos(r.m/10)*i;e.a={draw:function(t,e){var n=Math.sqrt(.8908130915292852*e),i=o*n,a=s*n;t.moveTo(0,-n),t.lineTo(i,a);for(var c=1;c<5;++c){var u=r.m*c/5,l=Math.cos(u),p=Math.sin(u);t.lineTo(p*n,-l*n),t.lineTo(l*i-p*a,p*i+l*a)}t.closePath()}}},"../../../../d3-shape/src/symbol/triangle.js":function(t,e,n){"use strict";var r=Math.sqrt(3);e.a={draw:function(t,e){var n=-Math.sqrt(e/(3*r));t.moveTo(0,2*n),t.lineTo(-r*n,-n),t.lineTo(r*n,-n),t.closePath()}}},"../../../../d3-shape/src/symbol/wye.js":function(t,e,n){"use strict";var r=Math.sqrt(3)/2,i=1/Math.sqrt(12),o=3*(i/2+1);e.a={draw:function(t,e){var n=Math.sqrt(e/o),s=n/2,a=n*i,c=s,u=n*i+n,l=-c,p=u;t.moveTo(s,a),t.lineTo(c,u),t.lineTo(l,p),t.lineTo(-.5*s-r*a,r*s+-.5*a),t.lineTo(-.5*c-r*u,r*c+-.5*u),t.lineTo(-.5*l-r*p,r*l+-.5*p),t.lineTo(-.5*s+r*a,-.5*a-r*s),t.lineTo(-.5*c+r*u,-.5*u-r*c),t.lineTo(-.5*l+r*p,-.5*p-r*l),t.closePath()}}},"../../../../d3-time-format/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-time-format/src/defaultLocale.js");n.d(e,"a",function(){return r.a}),n.d(e,"b",function(){return r.b}),n.d(e,"c",function(){return r.c}),n.d(e,"d",function(){return r.d});n("../../../../d3-time-format/src/locale.js"),n("../../../../d3-time-format/src/isoFormat.js"),n("../../../../d3-time-format/src/isoParse.js")},"../../../../d3-time-format/src/defaultLocale.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i}),n.d(e,"b",function(){return o}),n.d(e,"c",function(){return s}),n.d(e,"d",function(){return a});var r,i,o,s,a,c,u=n("../../../../d3-time-format/src/locale.js");c={dateTime:"%x, %X",date:"%-m/%-d/%Y",time:"%-I:%M:%S %p",periods:["AM","PM"],days:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],shortDays:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],months:["January","February","March","April","May","June","July","August","September","October","November","December"],shortMonths:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"]},r=Object(u.a)(c),i=r.format,o=r.parse,s=r.utcFormat,a=r.utcParse},"../../../../d3-time-format/src/isoFormat.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var r=n("../../../../d3-time-format/src/defaultLocale.js"),i="%Y-%m-%dT%H:%M:%S.%LZ";Date.prototype.toISOString||Object(r.c)(i)},"../../../../d3-time-format/src/isoParse.js":function(t,e,n){"use strict";var r=n("../../../../d3-time-format/src/isoFormat.js"),i=n("../../../../d3-time-format/src/defaultLocale.js");+new Date("2000-01-01T00:00:00.000Z")||Object(i.d)(r.a)},"../../../../d3-time-format/src/locale.js":function(t,e,n){"use strict";e.a=function(t){var e=t.dateTime,n=t.date,c=t.time,u=t.periods,l=t.days,p=t.shortDays,d=t.months,vt=t.shortMonths,yt=h(u),bt=f(u),_t=h(l),jt=f(l),xt=h(p),wt=f(p),Ot=h(d),Ct=f(d),kt=h(vt),Et=f(vt),St={a:function(t){return p[t.getDay()]},A:function(t){return l[t.getDay()]},b:function(t){return vt[t.getMonth()]},B:function(t){return d[t.getMonth()]},c:null,d:I,e:I,f:V,H:R,I:N,j:L,L:F,m:z,M:B,p:function(t){return u[+(t.getHours()>=12)]},Q:mt,s:gt,S:q,u:H,U:U,V:W,w:G,W:X,x:null,X:null,y:Y,Y:K,Z:$,"%":ft},Pt={a:function(t){return p[t.getUTCDay()]},A:function(t){return l[t.getUTCDay()]},b:function(t){return vt[t.getUTCMonth()]},B:function(t){return d[t.getUTCMonth()]},c:null,d:Q,e:Q,f:nt,H:Z,I:J,j:tt,L:et,m:rt,M:it,p:function(t){return u[+(t.getUTCHours()>=12)]},Q:mt,s:gt,S:ot,u:st,U:at,V:ct,w:ut,W:lt,x:null,X:null,y:pt,Y:dt,Z:ht,"%":ft},Mt={a:function(t,e,n){var r=xt.exec(e.slice(n));return r?(t.w=wt[r[0].toLowerCase()],n+r[0].length):-1},A:function(t,e,n){var r=_t.exec(e.slice(n));return r?(t.w=jt[r[0].toLowerCase()],n+r[0].length):-1},b:function(t,e,n){var r=kt.exec(e.slice(n));return r?(t.m=Et[r[0].toLowerCase()],n+r[0].length):-1},B:function(t,e,n){var r=Ot.exec(e.slice(n));return r?(t.m=Ct[r[0].toLowerCase()],n+r[0].length):-1},c:function(t,n,r){return Dt(t,e,n,r)},d:O,e:O,f:M,H:k,I:k,j:C,L:P,m:w,M:E,p:function(t,e,n){var r=yt.exec(e.slice(n));return r?(t.p=bt[r[0].toLowerCase()],n+r[0].length):-1},Q:T,s:D,S:S,u:g,U:v,V:y,w:m,W:b,x:function(t,e,r){return Dt(t,n,e,r)},X:function(t,e,n){return Dt(t,c,e,n)},y:j,Y:_,Z:x,"%":A};function At(t,e){return function(n){var r,i,o,s=[],c=-1,u=0,l=t.length;for(n instanceof Date||(n=new Date(+n));++c<l;)37===t.charCodeAt(c)&&(s.push(t.slice(u,c)),null!=(i=a[r=t.charAt(++c)])?r=t.charAt(++c):i="e"===r?" ":"0",(o=e[r])&&(r=o(n,i)),s.push(r),u=c+1);return s.push(t.slice(u,c)),s.join("")}}function Tt(t,e){return function(n){var i,a,c=s(1900),u=Dt(c,t,n+="",0);if(u!=n.length)return null;if("Q"in c)return new Date(c.Q);if("p"in c&&(c.H=c.H%12+12*c.p),"V"in c){if(c.V<1||c.V>53)return null;"w"in c||(c.w=1),"Z"in c?(i=o(s(c.y)),a=i.getUTCDay(),i=a>4||0===a?r.p.ceil(i):Object(r.p)(i),i=r.l.offset(i,7*(c.V-1)),c.y=i.getUTCFullYear(),c.m=i.getUTCMonth(),c.d=i.getUTCDate()+(c.w+6)%7):(i=e(s(c.y)),a=i.getDay(),i=a>4||0===a?r.e.ceil(i):Object(r.e)(i),i=r.a.offset(i,7*(c.V-1)),c.y=i.getFullYear(),c.m=i.getMonth(),c.d=i.getDate()+(c.w+6)%7)}else("W"in c||"U"in c)&&("w"in c||(c.w="u"in c?c.u%7:"W"in c?1:0),a="Z"in c?o(s(c.y)).getUTCDay():e(s(c.y)).getDay(),c.m=0,c.d="W"in c?(c.w+6)%7+7*c.W-(a+5)%7:c.w+7*c.U-(a+6)%7);return"Z"in c?(c.H+=c.Z/100|0,c.M+=c.Z%100,o(c)):e(c)}}function Dt(t,e,n,r){for(var i,o,s=0,c=e.length,u=n.length;s<c;){if(r>=u)return-1;if(37===(i=e.charCodeAt(s++))){if(i=e.charAt(s++),!(o=Mt[i in a?e.charAt(s++):i])||(r=o(t,n,r))<0)return-1}else if(i!=n.charCodeAt(r++))return-1}return r}return St.x=At(n,St),St.X=At(c,St),St.c=At(e,St),Pt.x=At(n,Pt),Pt.X=At(c,Pt),Pt.c=At(e,Pt),{format:function(t){var e=At(t+="",St);return e.toString=function(){return t},e},parse:function(t){var e=Tt(t+="",i);return e.toString=function(){return t},e},utcFormat:function(t){var e=At(t+="",Pt);return e.toString=function(){return t},e},utcParse:function(t){var e=Tt(t,o);return e.toString=function(){return t},e}}};var r=n("../../../../d3-time/index.js");function i(t){if(0<=t.y&&t.y<100){var e=new Date(-1,t.m,t.d,t.H,t.M,t.S,t.L);return e.setFullYear(t.y),e}return new Date(t.y,t.m,t.d,t.H,t.M,t.S,t.L)}function o(t){if(0<=t.y&&t.y<100){var e=new Date(Date.UTC(-1,t.m,t.d,t.H,t.M,t.S,t.L));return e.setUTCFullYear(t.y),e}return new Date(Date.UTC(t.y,t.m,t.d,t.H,t.M,t.S,t.L))}function s(t){return{y:t,m:0,d:1,H:0,M:0,S:0,L:0}}var a={"-":"",_:" ",0:"0"},c=/^\s*\d+/,u=/^%/,l=/[\\^$*+?|[\]().{}]/g;function p(t,e,n){var r=t<0?"-":"",i=(r?-t:t)+"",o=i.length;return r+(o<n?new Array(n-o+1).join(e)+i:i)}function d(t){return t.replace(l,"\\$&")}function h(t){return new RegExp("^(?:"+t.map(d).join("|")+")","i")}function f(t){for(var e={},n=-1,r=t.length;++n<r;)e[t[n].toLowerCase()]=n;return e}function m(t,e,n){var r=c.exec(e.slice(n,n+1));return r?(t.w=+r[0],n+r[0].length):-1}function g(t,e,n){var r=c.exec(e.slice(n,n+1));return r?(t.u=+r[0],n+r[0].length):-1}function v(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.U=+r[0],n+r[0].length):-1}function y(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.V=+r[0],n+r[0].length):-1}function b(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.W=+r[0],n+r[0].length):-1}function _(t,e,n){var r=c.exec(e.slice(n,n+4));return r?(t.y=+r[0],n+r[0].length):-1}function j(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.y=+r[0]+(+r[0]>68?1900:2e3),n+r[0].length):-1}function x(t,e,n){var r=/^(Z)|([+-]\d\d)(?::?(\d\d))?/.exec(e.slice(n,n+6));return r?(t.Z=r[1]?0:-(r[2]+(r[3]||"00")),n+r[0].length):-1}function w(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.m=r[0]-1,n+r[0].length):-1}function O(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.d=+r[0],n+r[0].length):-1}function C(t,e,n){var r=c.exec(e.slice(n,n+3));return r?(t.m=0,t.d=+r[0],n+r[0].length):-1}function k(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.H=+r[0],n+r[0].length):-1}function E(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.M=+r[0],n+r[0].length):-1}function S(t,e,n){var r=c.exec(e.slice(n,n+2));return r?(t.S=+r[0],n+r[0].length):-1}function P(t,e,n){var r=c.exec(e.slice(n,n+3));return r?(t.L=+r[0],n+r[0].length):-1}function M(t,e,n){var r=c.exec(e.slice(n,n+6));return r?(t.L=Math.floor(r[0]/1e3),n+r[0].length):-1}function A(t,e,n){var r=u.exec(e.slice(n,n+1));return r?n+r[0].length:-1}function T(t,e,n){var r=c.exec(e.slice(n));return r?(t.Q=+r[0],n+r[0].length):-1}function D(t,e,n){var r=c.exec(e.slice(n));return r?(t.Q=1e3*+r[0],n+r[0].length):-1}function I(t,e){return p(t.getDate(),e,2)}function R(t,e){return p(t.getHours(),e,2)}function N(t,e){return p(t.getHours()%12||12,e,2)}function L(t,e){return p(1+r.a.count(Object(r.k)(t),t),e,3)}function F(t,e){return p(t.getMilliseconds(),e,3)}function V(t,e){return F(t,e)+"000"}function z(t,e){return p(t.getMonth()+1,e,2)}function B(t,e){return p(t.getMinutes(),e,2)}function q(t,e){return p(t.getSeconds(),e,2)}function H(t){var e=t.getDay();return 0===e?7:e}function U(t,e){return p(r.h.count(Object(r.k)(t),t),e,2)}function W(t,e){var n=t.getDay();return t=n>=4||0===n?Object(r.i)(t):r.i.ceil(t),p(r.i.count(Object(r.k)(t),t)+(4===Object(r.k)(t).getDay()),e,2)}function G(t){return t.getDay()}function X(t,e){return p(r.e.count(Object(r.k)(t),t),e,2)}function Y(t,e){return p(t.getFullYear()%100,e,2)}function K(t,e){return p(t.getFullYear()%1e4,e,4)}function $(t){var e=t.getTimezoneOffset();return(e>0?"-":(e*=-1,"+"))+p(e/60|0,"0",2)+p(e%60,"0",2)}function Q(t,e){return p(t.getUTCDate(),e,2)}function Z(t,e){return p(t.getUTCHours(),e,2)}function J(t,e){return p(t.getUTCHours()%12||12,e,2)}function tt(t,e){return p(1+r.l.count(Object(r.v)(t),t),e,3)}function et(t,e){return p(t.getUTCMilliseconds(),e,3)}function nt(t,e){return et(t,e)+"000"}function rt(t,e){return p(t.getUTCMonth()+1,e,2)}function it(t,e){return p(t.getUTCMinutes(),e,2)}function ot(t,e){return p(t.getUTCSeconds(),e,2)}function st(t){var e=t.getUTCDay();return 0===e?7:e}function at(t,e){return p(r.s.count(Object(r.v)(t),t),e,2)}function ct(t,e){var n=t.getUTCDay();return t=n>=4||0===n?Object(r.t)(t):r.t.ceil(t),p(r.t.count(Object(r.v)(t),t)+(4===Object(r.v)(t).getUTCDay()),e,2)}function ut(t){return t.getUTCDay()}function lt(t,e){return p(r.p.count(Object(r.v)(t),t),e,2)}function pt(t,e){return p(t.getUTCFullYear()%100,e,2)}function dt(t,e){return p(t.getUTCFullYear()%1e4,e,4)}function ht(){return"+0000"}function ft(){return"%"}function mt(t){return+t}function gt(t){return Math.floor(+t/1e3)}},"../../../../d3-time/index.js":function(t,e,n){"use strict";n("../../../../d3-time/src/interval.js");var r=n("../../../../d3-time/src/millisecond.js");n.d(e,"c",function(){return r.a}),n.d(e,"n",function(){return r.a});var i=n("../../../../d3-time/src/second.js");n.d(e,"g",function(){return i.a}),n.d(e,"r",function(){return i.a});var o=n("../../../../d3-time/src/minute.js");n.d(e,"d",function(){return o.a});var s=n("../../../../d3-time/src/hour.js");n.d(e,"b",function(){return s.a});var a=n("../../../../d3-time/src/day.js");n.d(e,"a",function(){return a.a});var c=n("../../../../d3-time/src/week.js");n.d(e,"j",function(){return c.b}),n.d(e,"h",function(){return c.b}),n.d(e,"e",function(){return c.a}),n.d(e,"i",function(){return c.c});var u=n("../../../../d3-time/src/month.js");n.d(e,"f",function(){return u.a});var l=n("../../../../d3-time/src/year.js");n.d(e,"k",function(){return l.a});var p=n("../../../../d3-time/src/utcMinute.js");n.d(e,"o",function(){return p.a});var d=n("../../../../d3-time/src/utcHour.js");n.d(e,"m",function(){return d.a});var h=n("../../../../d3-time/src/utcDay.js");n.d(e,"l",function(){return h.a});var f=n("../../../../d3-time/src/utcWeek.js");n.d(e,"u",function(){return f.b}),n.d(e,"s",function(){return f.b}),n.d(e,"p",function(){return f.a}),n.d(e,"t",function(){return f.c});var m=n("../../../../d3-time/src/utcMonth.js");n.d(e,"q",function(){return m.a});var g=n("../../../../d3-time/src/utcYear.js");n.d(e,"v",function(){return g.a})},"../../../../d3-time/src/day.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=n("../../../../d3-time/src/duration.js"),o=Object(r.a)(function(t){t.setHours(0,0,0,0)},function(t,e){t.setDate(t.getDate()+e)},function(t,e){return(e-t-(e.getTimezoneOffset()-t.getTimezoneOffset())*i.c)/i.a},function(t){return t.getDate()-1});e.a=o;o.range},"../../../../d3-time/src/duration.js":function(t,e,n){"use strict";n.d(e,"d",function(){return r}),n.d(e,"c",function(){return i}),n.d(e,"b",function(){return o}),n.d(e,"a",function(){return s}),n.d(e,"e",function(){return a});var r=1e3,i=6e4,o=36e5,s=864e5,a=6048e5},"../../../../d3-time/src/hour.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=n("../../../../d3-time/src/duration.js"),o=Object(r.a)(function(t){var e=t.getTimezoneOffset()*i.c%i.b;e<0&&(e+=i.b),t.setTime(Math.floor((+t-e)/i.b)*i.b+e)},function(t,e){t.setTime(+t+e*i.b)},function(t,e){return(e-t)/i.b},function(t){return t.getHours()});e.a=o;o.range},"../../../../d3-time/src/interval.js":function(t,e,n){"use strict";e.a=function t(e,n,o,s){function a(t){return e(t=new Date(+t)),t}a.floor=a;a.ceil=function(t){return e(t=new Date(t-1)),n(t,1),e(t),t};a.round=function(t){var e=a(t),n=a.ceil(t);return t-e<n-t?e:n};a.offset=function(t,e){return n(t=new Date(+t),null==e?1:Math.floor(e)),t};a.range=function(t,r,i){var o,s=[];if(t=a.ceil(t),i=null==i?1:Math.floor(i),!(t<r&&i>0))return s;do{s.push(o=new Date(+t)),n(t,i),e(t)}while(o<t&&t<r);return s};a.filter=function(r){return t(function(t){if(t>=t)for(;e(t),!r(t);)t.setTime(t-1)},function(t,e){if(t>=t)if(e<0)for(;++e<=0;)for(;n(t,-1),!r(t););else for(;--e>=0;)for(;n(t,1),!r(t););})};o&&(a.count=function(t,n){return r.setTime(+t),i.setTime(+n),e(r),e(i),Math.floor(o(r,i))},a.every=function(t){return t=Math.floor(t),isFinite(t)&&t>0?t>1?a.filter(s?function(e){return s(e)%t==0}:function(e){return a.count(0,e)%t==0}):a:null});return a};var r=new Date,i=new Date},"../../../../d3-time/src/millisecond.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=Object(r.a)(function(){},function(t,e){t.setTime(+t+e)},function(t,e){return e-t});i.every=function(t){return t=Math.floor(t),isFinite(t)&&t>0?t>1?Object(r.a)(function(e){e.setTime(Math.floor(e/t)*t)},function(e,n){e.setTime(+e+n*t)},function(e,n){return(n-e)/t}):i:null},e.a=i;i.range},"../../../../d3-time/src/minute.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=n("../../../../d3-time/src/duration.js"),o=Object(r.a)(function(t){t.setTime(Math.floor(t/i.c)*i.c)},function(t,e){t.setTime(+t+e*i.c)},function(t,e){return(e-t)/i.c},function(t){return t.getMinutes()});e.a=o;o.range},"../../../../d3-time/src/month.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=Object(r.a)(function(t){t.setDate(1),t.setHours(0,0,0,0)},function(t,e){t.setMonth(t.getMonth()+e)},function(t,e){return e.getMonth()-t.getMonth()+12*(e.getFullYear()-t.getFullYear())},function(t){return t.getMonth()});e.a=i;i.range},"../../../../d3-time/src/second.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=n("../../../../d3-time/src/duration.js"),o=Object(r.a)(function(t){t.setTime(Math.floor(t/i.d)*i.d)},function(t,e){t.setTime(+t+e*i.d)},function(t,e){return(e-t)/i.d},function(t){return t.getUTCSeconds()});e.a=o;o.range},"../../../../d3-time/src/utcDay.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=n("../../../../d3-time/src/duration.js"),o=Object(r.a)(function(t){t.setUTCHours(0,0,0,0)},function(t,e){t.setUTCDate(t.getUTCDate()+e)},function(t,e){return(e-t)/i.a},function(t){return t.getUTCDate()-1});e.a=o;o.range},"../../../../d3-time/src/utcHour.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=n("../../../../d3-time/src/duration.js"),o=Object(r.a)(function(t){t.setUTCMinutes(0,0,0)},function(t,e){t.setTime(+t+e*i.b)},function(t,e){return(e-t)/i.b},function(t){return t.getUTCHours()});e.a=o;o.range},"../../../../d3-time/src/utcMinute.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=n("../../../../d3-time/src/duration.js"),o=Object(r.a)(function(t){t.setUTCSeconds(0,0)},function(t,e){t.setTime(+t+e*i.c)},function(t,e){return(e-t)/i.c},function(t){return t.getUTCMinutes()});e.a=o;o.range},"../../../../d3-time/src/utcMonth.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=Object(r.a)(function(t){t.setUTCDate(1),t.setUTCHours(0,0,0,0)},function(t,e){t.setUTCMonth(t.getUTCMonth()+e)},function(t,e){return e.getUTCMonth()-t.getUTCMonth()+12*(e.getUTCFullYear()-t.getUTCFullYear())},function(t){return t.getUTCMonth()});e.a=i;i.range},"../../../../d3-time/src/utcWeek.js":function(t,e,n){"use strict";n.d(e,"b",function(){return s}),n.d(e,"a",function(){return a}),n.d(e,"c",function(){return l});var r=n("../../../../d3-time/src/interval.js"),i=n("../../../../d3-time/src/duration.js");function o(t){return Object(r.a)(function(e){e.setUTCDate(e.getUTCDate()-(e.getUTCDay()+7-t)%7),e.setUTCHours(0,0,0,0)},function(t,e){t.setUTCDate(t.getUTCDate()+7*e)},function(t,e){return(e-t)/i.e})}var s=o(0),a=o(1),c=o(2),u=o(3),l=o(4),p=o(5),d=o(6);s.range,a.range,c.range,u.range,l.range,p.range,d.range},"../../../../d3-time/src/utcYear.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=Object(r.a)(function(t){t.setUTCMonth(0,1),t.setUTCHours(0,0,0,0)},function(t,e){t.setUTCFullYear(t.getUTCFullYear()+e)},function(t,e){return e.getUTCFullYear()-t.getUTCFullYear()},function(t){return t.getUTCFullYear()});i.every=function(t){return isFinite(t=Math.floor(t))&&t>0?Object(r.a)(function(e){e.setUTCFullYear(Math.floor(e.getUTCFullYear()/t)*t),e.setUTCMonth(0,1),e.setUTCHours(0,0,0,0)},function(e,n){e.setUTCFullYear(e.getUTCFullYear()+n*t)}):null},e.a=i;i.range},"../../../../d3-time/src/week.js":function(t,e,n){"use strict";n.d(e,"b",function(){return s}),n.d(e,"a",function(){return a}),n.d(e,"c",function(){return l});var r=n("../../../../d3-time/src/interval.js"),i=n("../../../../d3-time/src/duration.js");function o(t){return Object(r.a)(function(e){e.setDate(e.getDate()-(e.getDay()+7-t)%7),e.setHours(0,0,0,0)},function(t,e){t.setDate(t.getDate()+7*e)},function(t,e){return(e-t-(e.getTimezoneOffset()-t.getTimezoneOffset())*i.c)/i.e})}var s=o(0),a=o(1),c=o(2),u=o(3),l=o(4),p=o(5),d=o(6);s.range,a.range,c.range,u.range,l.range,p.range,d.range},"../../../../d3-time/src/year.js":function(t,e,n){"use strict";var r=n("../../../../d3-time/src/interval.js"),i=Object(r.a)(function(t){t.setMonth(0,1),t.setHours(0,0,0,0)},function(t,e){t.setFullYear(t.getFullYear()+e)},function(t,e){return e.getFullYear()-t.getFullYear()},function(t){return t.getFullYear()});i.every=function(t){return isFinite(t=Math.floor(t))&&t>0?Object(r.a)(function(e){e.setFullYear(Math.floor(e.getFullYear()/t)*t),e.setMonth(0,1),e.setHours(0,0,0,0)},function(e,n){e.setFullYear(e.getFullYear()+n*t)}):null},e.a=i;i.range},"../../../../d3-timer/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-timer/src/timer.js");n.d(e,"a",function(){return r.c});n("../../../../d3-timer/src/timeout.js"),n("../../../../d3-timer/src/interval.js")},"../../../../d3-timer/src/interval.js":function(t,e,n){"use strict";n("../../../../d3-timer/src/timer.js")},"../../../../d3-timer/src/timeout.js":function(t,e,n){"use strict";n("../../../../d3-timer/src/timer.js")},"../../../../d3-timer/src/timer.js":function(t,e,n){"use strict";e.b=f,e.a=g,e.c=v;var r,i,o=0,s=0,a=0,c=1e3,u=0,l=0,p=0,d="object"==typeof performance&&performance.now?performance:Date,h="object"==typeof window&&window.requestAnimationFrame?window.requestAnimationFrame.bind(window):function(t){setTimeout(t,17)};function f(){return l||(h(m),l=d.now()+p)}function m(){l=0}function g(){this._call=this._time=this._next=null}function v(t,e,n){var r=new g;return r.restart(t,e,n),r}function y(){l=(u=d.now())+p,o=s=0;try{!function(){f(),++o;for(var t,e=r;e;)(t=l-e._time)>=0&&e._call.call(null,t),e=e._next;--o}()}finally{o=0,function(){var t,e,n=r,o=1/0;for(;n;)n._call?(o>n._time&&(o=n._time),t=n,n=n._next):(e=n._next,n._next=null,n=t?t._next=e:r=e);i=t,_(o)}(),l=0}}function b(){var t=d.now(),e=t-u;e>c&&(p-=e,u=t)}function _(t){o||(s&&(s=clearTimeout(s)),t-l>24?(t<1/0&&(s=setTimeout(y,t-d.now()-p)),a&&(a=clearInterval(a))):(a||(u=d.now(),a=setInterval(b,c)),o=1,h(y)))}g.prototype=v.prototype={constructor:g,restart:function(t,e,n){if("function"!=typeof t)throw new TypeError("callback is not a function");n=(null==n?f():+n)+(null==e?0:+e),this._next||i===this||(i?i._next=this:r=this,i=this),this._call=t,this._time=n,_()},stop:function(){this._call&&(this._call=null,this._time=1/0,_())}}},"../../../../d3-voronoi/index.js":function(t,e,n){"use strict";var r=n("../../../../d3-voronoi/src/voronoi.js");n.d(e,"a",function(){return r.a})},"../../../../d3-voronoi/src/Beach.js":function(t,e,n){"use strict";e.b=function(t){var e=t.circle,n=e.x,r=e.cy,i=[n,r],c=t.P,u=t.N,p=[t];l(t);var d=c;for(;d.circle&&Math.abs(n-d.circle.x)<a.f&&Math.abs(r-d.circle.cy)<a.f;)c=d.P,p.unshift(d),l(d),d=c;p.unshift(d),Object(o.b)(d);var h=u;for(;h.circle&&Math.abs(n-h.circle.x)<a.f&&Math.abs(r-h.circle.cy)<a.f;)u=h.N,p.push(h),l(h),h=u;p.push(h),Object(o.b)(h);var f,m=p.length;for(f=1;f<m;++f)h=p[f],d=p[f-1],Object(s.d)(h.edge,d.site,h.site,i);d=p[0],(h=p[m-1]).edge=Object(s.c)(d.site,h.site,null,i),Object(o.a)(d),Object(o.a)(h)},e.a=function(t){var e,n,r,c,l=t[0],h=t[1],f=a.a._;for(;f;)if((r=p(f,h)-l)>a.f)f=f.L;else{if(!((c=l-d(f,h))>a.f)){r>-a.f?(e=f.P,n=f):c>-a.f?(e=f,n=f.N):e=n=f;break}if(!f.R){e=f;break}f=f.R}Object(i.c)(t);var m=u(t);if(a.a.insert(e,m),!e&&!n)return;if(e===n)return Object(o.b)(e),n=u(e.site),a.a.insert(m,n),m.edge=n.edge=Object(s.c)(e.site,m.site),Object(o.a)(e),void Object(o.a)(n);if(!n)return void(m.edge=Object(s.c)(e.site,m.site));Object(o.b)(e),Object(o.b)(n);var g=e.site,v=g[0],y=g[1],b=t[0]-v,_=t[1]-y,j=n.site,x=j[0]-v,w=j[1]-y,O=2*(b*w-_*x),C=b*b+_*_,k=x*x+w*w,E=[(w*C-_*k)/O+v,(b*k-x*C)/O+y];Object(s.d)(n.edge,g,j,E),m.edge=Object(s.c)(g,t,null,E),n.edge=Object(s.c)(t,j,null,E),Object(o.a)(e),Object(o.a)(n)};var r=n("../../../../d3-voronoi/src/RedBlackTree.js"),i=n("../../../../d3-voronoi/src/Cell.js"),o=n("../../../../d3-voronoi/src/Circle.js"),s=n("../../../../d3-voronoi/src/Edge.js"),a=n("../../../../d3-voronoi/src/Diagram.js"),c=[];function u(t){var e=c.pop()||new function(){Object(r.a)(this),this.edge=this.site=this.circle=null};return e.site=t,e}function l(t){Object(o.b)(t),a.a.remove(t),c.push(t),Object(r.a)(t)}function p(t,e){var n=t.site,r=n[0],i=n[1],o=i-e;if(!o)return r;var s=t.P;if(!s)return-1/0;var a=(n=s.site)[0],c=n[1],u=c-e;if(!u)return a;var l=a-r,p=1/o-1/u,d=l/u;return p?(-d+Math.sqrt(d*d-2*p*(l*l/(-2*u)-c+u/2+i-o/2)))/p+r:(r+a)/2}function d(t,e){var n=t.N;if(n)return p(n,e);var r=t.site;return r[1]===e?r[0]:1/0}},"../../../../d3-voronoi/src/Cell.js":function(t,e,n){"use strict";e.c=function(t){return i.b[t.index]={site:t,halfedges:[]}},e.a=s,e.d=function(){for(var t,e,n,r,s=0,a=i.b.length;s<a;++s)if((t=i.b[s])&&(r=(e=t.halfedges).length)){var c=new Array(r),u=new Array(r);for(n=0;n<r;++n)c[n]=n,u[n]=o(t,i.e[e[n]]);for(c.sort(function(t,e){return u[e]-u[t]}),n=0;n<r;++n)u[n]=e[c[n]];for(n=0;n<r;++n)e[n]=u[n]}},e.b=function(t,e,n,o){var a,c,u,l,p,d,h,f,m,g,v,y,b=i.b.length,_=!0;for(a=0;a<b;++a)if(c=i.b[a]){for(u=c.site,p=c.halfedges,l=p.length;l--;)i.e[p[l]]||p.splice(l,1);for(l=0,d=p.length;l<d;)j=c,x=i.e[p[l]],g=x[+(x.left===j.site)],v=g[0],y=g[1],h=s(c,i.e[p[++l%d]]),f=h[0],m=h[1],(Math.abs(v-f)>i.f||Math.abs(y-m)>i.f)&&(p.splice(l,0,i.e.push(Object(r.b)(u,g,Math.abs(v-t)<i.f&&o-y>i.f?[t,Math.abs(f-t)<i.f?m:o]:Math.abs(y-o)<i.f&&n-v>i.f?[Math.abs(m-o)<i.f?f:n,o]:Math.abs(v-n)<i.f&&y-e>i.f?[n,Math.abs(f-n)<i.f?m:e]:Math.abs(y-e)<i.f&&v-t>i.f?[Math.abs(m-e)<i.f?f:t,e]:null))-1),++d);d&&(_=!1)}var j,x;if(_){var w,O,C,k=1/0;for(a=0,_=null;a<b;++a)(c=i.b[a])&&(u=c.site,w=u[0]-t,O=u[1]-e,(C=w*w+O*O)<k&&(k=C,_=c));if(_){var E=[t,e],S=[t,o],P=[n,o],M=[n,e];_.halfedges.push(i.e.push(Object(r.b)(u=_.site,E,S))-1,i.e.push(Object(r.b)(u,S,P))-1,i.e.push(Object(r.b)(u,P,M))-1,i.e.push(Object(r.b)(u,M,E))-1)}}for(a=0;a<b;++a)(c=i.b[a])&&(c.halfedges.length||delete i.b[a])};var r=n("../../../../d3-voronoi/src/Edge.js"),i=n("../../../../d3-voronoi/src/Diagram.js");function o(t,e){var n=t.site,r=e.left,i=e.right;return n===i&&(i=r,r=n),i?Math.atan2(i[1]-r[1],i[0]-r[0]):(n===r?(r=e[1],i=e[0]):(r=e[0],i=e[1]),Math.atan2(r[0]-i[0],i[1]-r[1]))}function s(t,e){return e[+(e.left!==t.site)]}},"../../../../d3-voronoi/src/Circle.js":function(t,e,n){"use strict";n.d(e,"c",function(){return r}),e.a=function(t){var e=t.P,n=t.N;if(!e||!n)return;var a=e.site,c=t.site,u=n.site;if(a===u)return;var l=c[0],p=c[1],d=a[0]-l,h=a[1]-p,f=u[0]-l,m=u[1]-p,g=2*(d*m-h*f);if(g>=-o.g)return;var v=d*d+h*h,y=f*f+m*m,b=(m*v-h*y)/g,_=(d*y-f*v)/g,j=s.pop()||new function(){Object(i.a)(this),this.x=this.y=this.arc=this.site=this.cy=null};j.arc=t,j.site=c,j.x=b+l,j.y=(j.cy=_+p)+Math.sqrt(b*b+_*_),t.circle=j;var x=null,w=o.c._;for(;w;)if(j.y<w.y||j.y===w.y&&j.x<=w.x){if(!w.L){x=w.P;break}w=w.L}else{if(!w.R){x=w;break}w=w.R}o.c.insert(x,j),x||(r=j)},e.b=function(t){var e=t.circle;e&&(e.P||(r=e.N),o.c.remove(e),s.push(e),Object(i.a)(e),t.circle=null)};var r,i=n("../../../../d3-voronoi/src/RedBlackTree.js"),o=n("../../../../d3-voronoi/src/Diagram.js"),s=[]},"../../../../d3-voronoi/src/Diagram.js":function(t,e,n){"use strict";n.d(e,"f",function(){return d}),n.d(e,"g",function(){return h}),n.d(e,"a",function(){return r}),n.d(e,"b",function(){return i}),n.d(e,"c",function(){return o}),n.d(e,"e",function(){return s}),e.d=m;var r,i,o,s,a=n("../../../../d3-voronoi/src/Beach.js"),c=n("../../../../d3-voronoi/src/Cell.js"),u=n("../../../../d3-voronoi/src/Circle.js"),l=n("../../../../d3-voronoi/src/Edge.js"),p=n("../../../../d3-voronoi/src/RedBlackTree.js"),d=1e-6,h=1e-12;function f(t,e){return e[1]-t[1]||e[0]-t[0]}function m(t,e){var n,d,h,m=t.sort(f).pop();for(s=[],i=new Array(t.length),r=new p.b,o=new p.b;;)if(h=u.c,m&&(!h||m[1]<h.y||m[1]===h.y&&m[0]<h.x))m[0]===n&&m[1]===d||(Object(a.a)(m),n=m[0],d=m[1]),m=t.pop();else{if(!h)break;Object(a.b)(h.arc)}if(Object(c.d)(),e){var g=+e[0][0],v=+e[0][1],y=+e[1][0],b=+e[1][1];Object(l.a)(g,v,y,b),Object(c.b)(g,v,y,b)}this.edges=s,this.cells=i,r=o=s=i=null}m.prototype={constructor:m,polygons:function(){var t=this.edges;return this.cells.map(function(e){var n=e.halfedges.map(function(n){return Object(c.a)(e,t[n])});return n.data=e.site.data,n})},triangles:function(){var t=[],e=this.edges;return this.cells.forEach(function(n,r){if(o=(i=n.halfedges).length)for(var i,o,s,a,c,u,l=n.site,p=-1,d=e[i[o-1]],h=d.left===l?d.right:d.left;++p<o;)s=h,h=(d=e[i[p]]).left===l?d.right:d.left,s&&h&&r<s.index&&r<h.index&&(c=s,u=h,((a=l)[0]-u[0])*(c[1]-a[1])-(a[0]-c[0])*(u[1]-a[1])<0)&&t.push([l.data,s.data,h.data])}),t},links:function(){return this.edges.filter(function(t){return t.right}).map(function(t){return{source:t.left.data,target:t.right.data}})},find:function(t,e,n){for(var r,i,o=this,s=o._found||0,a=o.cells.length;!(i=o.cells[s]);)if(++s>=a)return null;var c=t-i.site[0],u=e-i.site[1],l=c*c+u*u;do{i=o.cells[r=s],s=null,i.halfedges.forEach(function(n){var r=o.edges[n],a=r.left;if(a!==i.site&&a||(a=r.right)){var c=t-a[0],u=e-a[1],p=c*c+u*u;p<l&&(l=p,s=a.index)}})}while(null!==s);return o._found=r,null==n||l<=n*n?i.site:null}}},"../../../../d3-voronoi/src/Edge.js":function(t,e,n){"use strict";e.c=function(t,e,n,o){var s=[null,null],a=r.e.push(s)-1;s.left=t,s.right=e,n&&i(s,t,e,n);o&&i(s,e,t,o);return r.b[t.index].halfedges.push(a),r.b[e.index].halfedges.push(a),s},e.b=function(t,e,n){var r=[e,n];return r.left=t,r},e.d=i,e.a=function(t,e,n,i){var a,c=r.e.length;for(;c--;)s(a=r.e[c],t,e,n,i)&&o(a,t,e,n,i)&&(Math.abs(a[0][0]-a[1][0])>r.f||Math.abs(a[0][1]-a[1][1])>r.f)||delete r.e[c]};var r=n("../../../../d3-voronoi/src/Diagram.js");function i(t,e,n,r){t[0]||t[1]?t.left===n?t[1]=r:t[0]=r:(t[0]=r,t.left=e,t.right=n)}function o(t,e,n,r,i){var o,s=t[0],a=t[1],c=s[0],u=s[1],l=0,p=1,d=a[0]-c,h=a[1]-u;if(o=e-c,d||!(o>0)){if(o/=d,d<0){if(o<l)return;o<p&&(p=o)}else if(d>0){if(o>p)return;o>l&&(l=o)}if(o=r-c,d||!(o<0)){if(o/=d,d<0){if(o>p)return;o>l&&(l=o)}else if(d>0){if(o<l)return;o<p&&(p=o)}if(o=n-u,h||!(o>0)){if(o/=h,h<0){if(o<l)return;o<p&&(p=o)}else if(h>0){if(o>p)return;o>l&&(l=o)}if(o=i-u,h||!(o<0)){if(o/=h,h<0){if(o>p)return;o>l&&(l=o)}else if(h>0){if(o<l)return;o<p&&(p=o)}return!(l>0||p<1)||(l>0&&(t[0]=[c+l*d,u+l*h]),p<1&&(t[1]=[c+p*d,u+p*h]),!0)}}}}}function s(t,e,n,r,i){var o=t[1];if(o)return!0;var s,a,c=t[0],u=t.left,l=t.right,p=u[0],d=u[1],h=l[0],f=l[1],m=(p+h)/2,g=(d+f)/2;if(f===d){if(m<e||m>=r)return;if(p>h){if(c){if(c[1]>=i)return}else c=[m,n];o=[m,i]}else{if(c){if(c[1]<n)return}else c=[m,i];o=[m,n]}}else if(a=g-(s=(p-h)/(f-d))*m,s<-1||s>1)if(p>h){if(c){if(c[1]>=i)return}else c=[(n-a)/s,n];o=[(i-a)/s,i]}else{if(c){if(c[1]<n)return}else c=[(i-a)/s,i];o=[(n-a)/s,n]}else if(d<f){if(c){if(c[0]>=r)return}else c=[e,s*e+a];o=[r,s*r+a]}else{if(c){if(c[0]<e)return}else c=[r,s*r+a];o=[e,s*e+a]}return t[0]=c,t[1]=o,!0}},"../../../../d3-voronoi/src/RedBlackTree.js":function(t,e,n){"use strict";function r(){this._=null}function i(t,e){var n=e,r=e.R,i=n.U;i?i.L===n?i.L=r:i.R=r:t._=r,r.U=i,n.U=r,n.R=r.L,n.R&&(n.R.U=n),r.L=n}function o(t,e){var n=e,r=e.L,i=n.U;i?i.L===n?i.L=r:i.R=r:t._=r,r.U=i,n.U=r,n.L=r.R,n.L&&(n.L.U=n),r.R=n}function s(t){for(;t.L;)t=t.L;return t}e.a=function(t){t.U=t.C=t.L=t.R=t.P=t.N=null},r.prototype={constructor:r,insert:function(t,e){var n,r,a;if(t){if(e.P=t,e.N=t.N,t.N&&(t.N.P=e),t.N=e,t.R){for(t=t.R;t.L;)t=t.L;t.L=e}else t.R=e;n=t}else this._?(t=s(this._),e.P=null,e.N=t,t.P=t.L=e,n=t):(e.P=e.N=null,this._=e,n=null);for(e.L=e.R=null,e.U=n,e.C=!0,t=e;n&&n.C;)n===(r=n.U).L?(a=r.R)&&a.C?(n.C=a.C=!1,r.C=!0,t=r):(t===n.R&&(i(this,n),n=(t=n).U),n.C=!1,r.C=!0,o(this,r)):(a=r.L)&&a.C?(n.C=a.C=!1,r.C=!0,t=r):(t===n.L&&(o(this,n),n=(t=n).U),n.C=!1,r.C=!0,i(this,r)),n=t.U;this._.C=!1},remove:function(t){t.N&&(t.N.P=t.P),t.P&&(t.P.N=t.N),t.N=t.P=null;var e,n,r,a=t.U,c=t.L,u=t.R;if(n=c?u?s(u):c:u,a?a.L===t?a.L=n:a.R=n:this._=n,c&&u?(r=n.C,n.C=t.C,n.L=c,c.U=n,n!==u?(a=n.U,n.U=t.U,t=n.R,a.L=t,n.R=u,u.U=n):(n.U=a,a=n,t=n.R)):(r=t.C,t=n),t&&(t.U=a),!r)if(t&&t.C)t.C=!1;else{do{if(t===this._)break;if(t===a.L){if((e=a.R).C&&(e.C=!1,a.C=!0,i(this,a),e=a.R),e.L&&e.L.C||e.R&&e.R.C){e.R&&e.R.C||(e.L.C=!1,e.C=!0,o(this,e),e=a.R),e.C=a.C,a.C=e.R.C=!1,i(this,a),t=this._;break}}else if((e=a.L).C&&(e.C=!1,a.C=!0,o(this,a),e=a.L),e.L&&e.L.C||e.R&&e.R.C){e.L&&e.L.C||(e.R.C=!1,e.C=!0,i(this,e),e=a.L),e.C=a.C,a.C=e.L.C=!1,o(this,a),t=this._;break}e.C=!0,t=a,a=a.U}while(!t.C);t&&(t.C=!1)}}},e.b=r},"../../../../d3-voronoi/src/constant.js":function(t,e,n){"use strict";e.a=function(t){return function(){return t}}},"../../../../d3-voronoi/src/point.js":function(t,e,n){"use strict";e.a=function(t){return t[0]},e.b=function(t){return t[1]}},"../../../../d3-voronoi/src/voronoi.js":function(t,e,n){"use strict";var r=n("../../../../d3-voronoi/src/constant.js"),i=n("../../../../d3-voronoi/src/point.js"),o=n("../../../../d3-voronoi/src/Diagram.js");e.a=function(){var t=i.a,e=i.b,n=null;function s(r){return new o.d(r.map(function(n,i){var s=[Math.round(t(n,i,r)/o.f)*o.f,Math.round(e(n,i,r)/o.f)*o.f];return s.index=i,s.data=n,s}),n)}return s.polygons=function(t){return s(t).polygons()},s.links=function(t){return s(t).links()},s.triangles=function(t){return s(t).triangles()},s.x=function(e){return arguments.length?(t="function"==typeof e?e:Object(r.a)(+e),s):t},s.y=function(t){return arguments.length?(e="function"==typeof t?t:Object(r.a)(+t),s):e},s.extent=function(t){return arguments.length?(n=null==t?null:[[+t[0][0],+t[0][1]],[+t[1][0],+t[1][1]]],s):n&&[[n[0][0],n[0][1]],[n[1][0],n[1][1]]]},s.size=function(t){return arguments.length?(n=null==t?null:[[0,0],[+t[0],+t[1]]],s):n&&[n[1][0]-n[0][0],n[1][1]-n[0][1]]},s}},"../../../../ieee754/index.js":function(t,e){e.read=function(t,e,n,r,i){var o,s,a=8*i-r-1,c=(1<<a)-1,u=c>>1,l=-7,p=n?i-1:0,d=n?-1:1,h=t[e+p];for(p+=d,o=h&(1<<-l)-1,h>>=-l,l+=a;l>0;o=256*o+t[e+p],p+=d,l-=8);for(s=o&(1<<-l)-1,o>>=-l,l+=r;l>0;s=256*s+t[e+p],p+=d,l-=8);if(0===o)o=1-u;else{if(o===c)return s?NaN:1/0*(h?-1:1);s+=Math.pow(2,r),o-=u}return(h?-1:1)*s*Math.pow(2,o-r)},e.write=function(t,e,n,r,i,o){var s,a,c,u=8*o-i-1,l=(1<<u)-1,p=l>>1,d=23===i?Math.pow(2,-24)-Math.pow(2,-77):0,h=r?0:o-1,f=r?1:-1,m=e<0||0===e&&1/e<0?1:0;for(e=Math.abs(e),isNaN(e)||e===1/0?(a=isNaN(e)?1:0,s=l):(s=Math.floor(Math.log(e)/Math.LN2),e*(c=Math.pow(2,-s))<1&&(s--,c*=2),(e+=s+p>=1?d/c:d*Math.pow(2,1-p))*c>=2&&(s++,c/=2),s+p>=l?(a=0,s=l):s+p>=1?(a=(e*c-1)*Math.pow(2,i),s+=p):(a=e*Math.pow(2,p-1)*Math.pow(2,i),s=0));i>=8;t[n+h]=255&a,h+=f,a/=256,i-=8);for(s=s<<i|a,u+=i;u>0;t[n+h]=255&s,h+=f,s/=256,u-=8);t[n+h-f]|=128*m}},"../../../../isarray/index.js":function(t,e){var n={}.toString;t.exports=Array.isArray||function(t){return"[object Array]"==n.call(t)}},"../../../../rxjs/_esm5/BehaviorSubject.js":function(t,e,n){"use strict";n.d(e,"a",function(){return s});var r=n("../../../../rxjs/_esm5/Subject.js"),i=n("../../../../rxjs/_esm5/util/ObjectUnsubscribedError.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},s=function(t){function e(e){t.call(this),this._value=e}return o(e,t),Object.defineProperty(e.prototype,"value",{get:function(){return this.getValue()},enumerable:!0,configurable:!0}),e.prototype._subscribe=function(e){var n=t.prototype._subscribe.call(this,e);return n&&!n.closed&&e.next(this._value),n},e.prototype.getValue=function(){if(this.hasError)throw this.thrownError;if(this.closed)throw new i.a;return this._value},e.prototype.next=function(e){t.prototype.next.call(this,this._value=e)},e}(r.a)},"../../../../rxjs/_esm5/InnerSubscriber.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(e,n,r){t.call(this),this.parent=e,this.outerValue=n,this.outerIndex=r,this.index=0}return i(e,t),e.prototype._next=function(t){this.parent.notifyNext(this.outerValue,t,this.outerIndex,this.index++,this)},e.prototype._error=function(t){this.parent.notifyError(t,this),this.unsubscribe()},e.prototype._complete=function(){this.parent.notifyComplete(this),this.unsubscribe()},e}(r.a)},"../../../../rxjs/_esm5/Notification.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var r=n("../../../../rxjs/_esm5/Observable.js"),i=function(){function t(t,e,n){this.kind=t,this.value=e,this.error=n,this.hasValue="N"===t}return t.prototype.observe=function(t){switch(this.kind){case"N":return t.next&&t.next(this.value);case"E":return t.error&&t.error(this.error);case"C":return t.complete&&t.complete()}},t.prototype.do=function(t,e,n){switch(this.kind){case"N":return t&&t(this.value);case"E":return e&&e(this.error);case"C":return n&&n()}},t.prototype.accept=function(t,e,n){return t&&"function"==typeof t.next?this.observe(t):this.do(t,e,n)},t.prototype.toObservable=function(){switch(this.kind){case"N":return r.a.of(this.value);case"E":return r.a.throw(this.error);case"C":return r.a.empty()}throw new Error("unexpected notification kind value")},t.createNext=function(e){return void 0!==e?new t("N",e):t.undefinedValueNotification},t.createError=function(e){return new t("E",void 0,e)},t.createComplete=function(){return t.completeNotification},t.completeNotification=new t("C"),t.undefinedValueNotification=new t("N",void 0),t}()},"../../../../rxjs/_esm5/Observable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var r=n("../../../../rxjs/_esm5/util/root.js"),i=n("../../../../rxjs/_esm5/util/toSubscriber.js"),o=n("../../../../rxjs/_esm5/symbol/observable.js"),s=n("../../../../rxjs/_esm5/util/pipe.js"),a=function(){function t(t){this._isScalar=!1,t&&(this._subscribe=t)}return t.prototype.lift=function(e){var n=new t;return n.source=this,n.operator=e,n},t.prototype.subscribe=function(t,e,n){var r=this.operator,o=Object(i.a)(t,e,n);if(r?r.call(o,this.source):o.add(this.source?this._subscribe(o):this._trySubscribe(o)),o.syncErrorThrowable&&(o.syncErrorThrowable=!1,o.syncErrorThrown))throw o.syncErrorValue;return o},t.prototype._trySubscribe=function(t){try{return this._subscribe(t)}catch(e){t.syncErrorThrown=!0,t.syncErrorValue=e,t.error(e)}},t.prototype.forEach=function(t,e){var n=this;if(e||(r.a.Rx&&r.a.Rx.config&&r.a.Rx.config.Promise?e=r.a.Rx.config.Promise:r.a.Promise&&(e=r.a.Promise)),!e)throw new Error("no Promise impl found");return new e(function(e,r){var i;i=n.subscribe(function(e){if(i)try{t(e)}catch(t){r(t),i.unsubscribe()}else t(e)},r,e)})},t.prototype._subscribe=function(t){return this.source.subscribe(t)},t.prototype[o.a]=function(){return this},t.prototype.pipe=function(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];return 0===t.length?this:Object(s.a)(t)(this)},t.prototype.toPromise=function(t){var e=this;if(t||(r.a.Rx&&r.a.Rx.config&&r.a.Rx.config.Promise?t=r.a.Rx.config.Promise:r.a.Promise&&(t=r.a.Promise)),!t)throw new Error("no Promise impl found");return new t(function(t,n){var r;e.subscribe(function(t){return r=t},function(t){return n(t)},function(){return t(r)})})},t.create=function(e){return new t(e)},t}()},"../../../../rxjs/_esm5/Observer.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r={closed:!0,next:function(t){},error:function(t){throw t},complete:function(){}}},"../../../../rxjs/_esm5/OuterSubscriber.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(){t.apply(this,arguments)}return i(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.destination.next(e)},e.prototype.notifyError=function(t,e){this.destination.error(t)},e.prototype.notifyComplete=function(t){this.destination.complete()},e}(r.a)},"../../../../rxjs/_esm5/ReplaySubject.js":function(t,e,n){"use strict";n.d(e,"a",function(){return l});var r=n("../../../../rxjs/_esm5/Subject.js"),i=n("../../../../rxjs/_esm5/scheduler/queue.js"),o=n("../../../../rxjs/_esm5/Subscription.js"),s=n("../../../../rxjs/_esm5/operators/observeOn.js"),a=n("../../../../rxjs/_esm5/util/ObjectUnsubscribedError.js"),c=n("../../../../rxjs/_esm5/SubjectSubscription.js"),u=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},l=function(t){function e(e,n,r){void 0===e&&(e=Number.POSITIVE_INFINITY),void 0===n&&(n=Number.POSITIVE_INFINITY),t.call(this),this.scheduler=r,this._events=[],this._bufferSize=e<1?1:e,this._windowTime=n<1?1:n}return u(e,t),e.prototype.next=function(e){var n=this._getNow();this._events.push(new p(n,e)),this._trimBufferThenGetEvents(),t.prototype.next.call(this,e)},e.prototype._subscribe=function(t){var e,n=this._trimBufferThenGetEvents(),r=this.scheduler;if(this.closed)throw new a.a;this.hasError?e=o.a.EMPTY:this.isStopped?e=o.a.EMPTY:(this.observers.push(t),e=new c.a(this,t)),r&&t.add(t=new s.a(t,r));for(var i=n.length,u=0;u<i&&!t.closed;u++)t.next(n[u].value);return this.hasError?t.error(this.thrownError):this.isStopped&&t.complete(),e},e.prototype._getNow=function(){return(this.scheduler||i.a).now()},e.prototype._trimBufferThenGetEvents=function(){for(var t=this._getNow(),e=this._bufferSize,n=this._windowTime,r=this._events,i=r.length,o=0;o<i&&!(t-r[o].time<n);)o++;return i>e&&(o=Math.max(o,i-e)),o>0&&r.splice(0,o),r},e}(r.a),p=function(){return function(t,e){this.time=t,this.value=e}}()},"../../../../rxjs/_esm5/Scheduler.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=function(){function t(e,n){void 0===n&&(n=t.now),this.SchedulerAction=e,this.now=n}return t.prototype.schedule=function(t,e,n){return void 0===e&&(e=0),new this.SchedulerAction(this,t).schedule(n,e)},t.now=Date.now?Date.now:function(){return+new Date},t}()},"../../../../rxjs/_esm5/Subject.js":function(t,e,n){"use strict";n.d(e,"b",function(){return l}),n.d(e,"a",function(){return p});var r=n("../../../../rxjs/_esm5/Observable.js"),i=n("../../../../rxjs/_esm5/Subscriber.js"),o=n("../../../../rxjs/_esm5/Subscription.js"),s=n("../../../../rxjs/_esm5/util/ObjectUnsubscribedError.js"),a=n("../../../../rxjs/_esm5/SubjectSubscription.js"),c=n("../../../../rxjs/_esm5/symbol/rxSubscriber.js"),u=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},l=function(t){function e(e){t.call(this,e),this.destination=e}return u(e,t),e}(i.a),p=function(t){function e(){t.call(this),this.observers=[],this.closed=!1,this.isStopped=!1,this.hasError=!1,this.thrownError=null}return u(e,t),e.prototype[c.a]=function(){return new l(this)},e.prototype.lift=function(t){var e=new d(this,this);return e.operator=t,e},e.prototype.next=function(t){if(this.closed)throw new s.a;if(!this.isStopped)for(var e=this.observers,n=e.length,r=e.slice(),i=0;i<n;i++)r[i].next(t)},e.prototype.error=function(t){if(this.closed)throw new s.a;this.hasError=!0,this.thrownError=t,this.isStopped=!0;for(var e=this.observers,n=e.length,r=e.slice(),i=0;i<n;i++)r[i].error(t);this.observers.length=0},e.prototype.complete=function(){if(this.closed)throw new s.a;this.isStopped=!0;for(var t=this.observers,e=t.length,n=t.slice(),r=0;r<e;r++)n[r].complete();this.observers.length=0},e.prototype.unsubscribe=function(){this.isStopped=!0,this.closed=!0,this.observers=null},e.prototype._trySubscribe=function(e){if(this.closed)throw new s.a;return t.prototype._trySubscribe.call(this,e)},e.prototype._subscribe=function(t){if(this.closed)throw new s.a;return this.hasError?(t.error(this.thrownError),o.a.EMPTY):this.isStopped?(t.complete(),o.a.EMPTY):(this.observers.push(t),new a.a(this,t))},e.prototype.asObservable=function(){var t=new r.a;return t.source=this,t},e.create=function(t,e){return new d(t,e)},e}(r.a),d=function(t){function e(e,n){t.call(this),this.destination=e,this.source=n}return u(e,t),e.prototype.next=function(t){var e=this.destination;e&&e.next&&e.next(t)},e.prototype.error=function(t){var e=this.destination;e&&e.error&&this.destination.error(t)},e.prototype.complete=function(){var t=this.destination;t&&t.complete&&this.destination.complete()},e.prototype._subscribe=function(t){return this.source?this.source.subscribe(t):o.a.EMPTY},e}(p)},"../../../../rxjs/_esm5/SubjectSubscription.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/Subscription.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(e,n){t.call(this),this.subject=e,this.subscriber=n,this.closed=!1}return i(e,t),e.prototype.unsubscribe=function(){if(!this.closed){this.closed=!0;var t=this.subject,e=t.observers;if(this.subject=null,e&&0!==e.length&&!t.isStopped&&!t.closed){var n=e.indexOf(this.subscriber);-1!==n&&e.splice(n,1)}}},e}(r.a)},"../../../../rxjs/_esm5/Subscriber.js":function(t,e,n){"use strict";n.d(e,"a",function(){return c});var r=n("../../../../rxjs/_esm5/util/isFunction.js"),i=n("../../../../rxjs/_esm5/Subscription.js"),o=n("../../../../rxjs/_esm5/Observer.js"),s=n("../../../../rxjs/_esm5/symbol/rxSubscriber.js"),a=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},c=function(t){function e(n,r,i){switch(t.call(this),this.syncErrorValue=null,this.syncErrorThrown=!1,this.syncErrorThrowable=!1,this.isStopped=!1,arguments.length){case 0:this.destination=o.a;break;case 1:if(!n){this.destination=o.a;break}if("object"==typeof n){n instanceof e?(this.destination=n,this.destination.add(this)):(this.syncErrorThrowable=!0,this.destination=new u(this,n));break}default:this.syncErrorThrowable=!0,this.destination=new u(this,n,r,i)}}return a(e,t),e.prototype[s.a]=function(){return this},e.create=function(t,n,r){var i=new e(t,n,r);return i.syncErrorThrowable=!1,i},e.prototype.next=function(t){this.isStopped||this._next(t)},e.prototype.error=function(t){this.isStopped||(this.isStopped=!0,this._error(t))},e.prototype.complete=function(){this.isStopped||(this.isStopped=!0,this._complete())},e.prototype.unsubscribe=function(){this.closed||(this.isStopped=!0,t.prototype.unsubscribe.call(this))},e.prototype._next=function(t){this.destination.next(t)},e.prototype._error=function(t){this.destination.error(t),this.unsubscribe()},e.prototype._complete=function(){this.destination.complete(),this.unsubscribe()},e.prototype._unsubscribeAndRecycle=function(){var t=this._parent,e=this._parents;return this._parent=null,this._parents=null,this.unsubscribe(),this.closed=!1,this.isStopped=!1,this._parent=t,this._parents=e,this},e}(i.a),u=function(t){function e(e,n,i,s){var a;t.call(this),this._parentSubscriber=e;var c=this;Object(r.a)(n)?a=n:n&&(a=n.next,i=n.error,s=n.complete,n!==o.a&&(c=Object.create(n),Object(r.a)(c.unsubscribe)&&this.add(c.unsubscribe.bind(c)),c.unsubscribe=this.unsubscribe.bind(this))),this._context=c,this._next=a,this._error=i,this._complete=s}return a(e,t),e.prototype.next=function(t){if(!this.isStopped&&this._next){var e=this._parentSubscriber;e.syncErrorThrowable?this.__tryOrSetError(e,this._next,t)&&this.unsubscribe():this.__tryOrUnsub(this._next,t)}},e.prototype.error=function(t){if(!this.isStopped){var e=this._parentSubscriber;if(this._error)e.syncErrorThrowable?(this.__tryOrSetError(e,this._error,t),this.unsubscribe()):(this.__tryOrUnsub(this._error,t),this.unsubscribe());else{if(!e.syncErrorThrowable)throw this.unsubscribe(),t;e.syncErrorValue=t,e.syncErrorThrown=!0,this.unsubscribe()}}},e.prototype.complete=function(){var t=this;if(!this.isStopped){var e=this._parentSubscriber;if(this._complete){var n=function(){return t._complete.call(t._context)};e.syncErrorThrowable?(this.__tryOrSetError(e,n),this.unsubscribe()):(this.__tryOrUnsub(n),this.unsubscribe())}else this.unsubscribe()}},e.prototype.__tryOrUnsub=function(t,e){try{t.call(this._context,e)}catch(t){throw this.unsubscribe(),t}},e.prototype.__tryOrSetError=function(t,e,n){try{e.call(this._context,n)}catch(e){return t.syncErrorValue=e,t.syncErrorThrown=!0,!0}return!1},e.prototype._unsubscribe=function(){var t=this._parentSubscriber;this._context=null,this._parentSubscriber=null,t.unsubscribe()},e}(c)},"../../../../rxjs/_esm5/Subscription.js":function(t,e,n){"use strict";n.d(e,"a",function(){return u});var r=n("../../../../rxjs/_esm5/util/isArray.js"),i=n("../../../../rxjs/_esm5/util/isObject.js"),o=n("../../../../rxjs/_esm5/util/isFunction.js"),s=n("../../../../rxjs/_esm5/util/tryCatch.js"),a=n("../../../../rxjs/_esm5/util/errorObject.js"),c=n("../../../../rxjs/_esm5/util/UnsubscriptionError.js"),u=function(){function t(t){this.closed=!1,this._parent=null,this._parents=null,this._subscriptions=null,t&&(this._unsubscribe=t)}var e;return t.prototype.unsubscribe=function(){var t,e=!1;if(!this.closed){var n=this._parent,u=this._parents,p=this._unsubscribe,d=this._subscriptions;this.closed=!0,this._parent=null,this._parents=null,this._subscriptions=null;for(var h=-1,f=u?u.length:0;n;)n.remove(this),n=++h<f&&u[h]||null;if(Object(o.a)(p))Object(s.a)(p).call(this)===a.a&&(e=!0,t=t||(a.a.e instanceof c.a?l(a.a.e.errors):[a.a.e]));if(Object(r.a)(d))for(h=-1,f=d.length;++h<f;){var m=d[h];if(Object(i.a)(m))if(Object(s.a)(m.unsubscribe).call(m)===a.a){e=!0,t=t||[];var g=a.a.e;g instanceof c.a?t=t.concat(l(g.errors)):t.push(g)}}if(e)throw new c.a(t)}},t.prototype.add=function(e){if(!e||e===t.EMPTY)return t.EMPTY;if(e===this)return this;var n=e;switch(typeof e){case"function":n=new t(e);case"object":if(n.closed||"function"!=typeof n.unsubscribe)return n;if(this.closed)return n.unsubscribe(),n;if("function"!=typeof n._addParent){var r=n;(n=new t)._subscriptions=[r]}break;default:throw new Error("unrecognized teardown "+e+" added to Subscription.")}return(this._subscriptions||(this._subscriptions=[])).push(n),n._addParent(this),n},t.prototype.remove=function(t){var e=this._subscriptions;if(e){var n=e.indexOf(t);-1!==n&&e.splice(n,1)}},t.prototype._addParent=function(t){var e=this._parent,n=this._parents;e&&e!==t?n?-1===n.indexOf(t)&&n.push(t):this._parents=[t]:this._parent=t},t.EMPTY=((e=new t).closed=!0,e),t}();function l(t){return t.reduce(function(t,e){return t.concat(e instanceof c.a?e.errors:e)},[])}},"../../../../rxjs/_esm5/observable/ArrayLikeObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var r=n("../../../../rxjs/_esm5/Observable.js"),i=n("../../../../rxjs/_esm5/observable/ScalarObservable.js"),o=n("../../../../rxjs/_esm5/observable/EmptyObservable.js"),s=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},a=function(t){function e(e,n){t.call(this),this.arrayLike=e,this.scheduler=n,n||1!==e.length||(this._isScalar=!0,this.value=e[0])}return s(e,t),e.create=function(t,n){var r=t.length;return 0===r?new o.a:1===r?new i.a(t[0],n):new e(t,n)},e.dispatch=function(t){var e=t.arrayLike,n=t.index,r=t.length,i=t.subscriber;i.closed||(n>=r?i.complete():(i.next(e[n]),t.index=n+1,this.schedule(t)))},e.prototype._subscribe=function(t){var n=this.arrayLike,r=this.scheduler,i=n.length;if(r)return r.schedule(e.dispatch,0,{arrayLike:n,index:0,length:i,subscriber:t});for(var o=0;o<i&&!t.closed;o++)t.next(n[o]);t.complete()},e}(r.a)},"../../../../rxjs/_esm5/observable/ArrayObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return c});var r=n("../../../../rxjs/_esm5/Observable.js"),i=n("../../../../rxjs/_esm5/observable/ScalarObservable.js"),o=n("../../../../rxjs/_esm5/observable/EmptyObservable.js"),s=n("../../../../rxjs/_esm5/util/isScheduler.js"),a=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},c=function(t){function e(e,n){t.call(this),this.array=e,this.scheduler=n,n||1!==e.length||(this._isScalar=!0,this.value=e[0])}return a(e,t),e.create=function(t,n){return new e(t,n)},e.of=function(){for(var t=[],n=0;n<arguments.length;n++)t[n-0]=arguments[n];var r=t[t.length-1];Object(s.a)(r)?t.pop():r=null;var a=t.length;return a>1?new e(t,r):1===a?new i.a(t[0],r):new o.a(r)},e.dispatch=function(t){var e=t.array,n=t.index,r=t.count,i=t.subscriber;n>=r?i.complete():(i.next(e[n]),i.closed||(t.index=n+1,this.schedule(t)))},e.prototype._subscribe=function(t){var n=this.array,r=n.length,i=this.scheduler;if(i)return i.schedule(e.dispatch,0,{array:n,index:0,count:r,subscriber:t});for(var o=0;o<r&&!t.closed;o++)t.next(n[o]);t.complete()},e}(r.a)},"../../../../rxjs/_esm5/observable/ConnectableObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return l});var r=n("../../../../rxjs/_esm5/Subject.js"),i=n("../../../../rxjs/_esm5/Observable.js"),o=n("../../../../rxjs/_esm5/Subscriber.js"),s=n("../../../../rxjs/_esm5/Subscription.js"),a=n("../../../../rxjs/_esm5/operators/refCount.js"),c=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},u=function(t){function e(e,n){t.call(this),this.source=e,this.subjectFactory=n,this._refCount=0,this._isComplete=!1}return c(e,t),e.prototype._subscribe=function(t){return this.getSubject().subscribe(t)},e.prototype.getSubject=function(){var t=this._subject;return t&&!t.isStopped||(this._subject=this.subjectFactory()),this._subject},e.prototype.connect=function(){var t=this._connection;return t||(this._isComplete=!1,(t=this._connection=new s.a).add(this.source.subscribe(new p(this.getSubject(),this))),t.closed?(this._connection=null,t=s.a.EMPTY):this._connection=t),t},e.prototype.refCount=function(){return Object(a.a)()(this)},e}(i.a).prototype,l={operator:{value:null},_refCount:{value:0,writable:!0},_subject:{value:null,writable:!0},_connection:{value:null,writable:!0},_subscribe:{value:u._subscribe},_isComplete:{value:u._isComplete,writable:!0},getSubject:{value:u.getSubject},connect:{value:u.connect},refCount:{value:u.refCount}},p=function(t){function e(e,n){t.call(this,e),this.connectable=n}return c(e,t),e.prototype._error=function(e){this._unsubscribe(),t.prototype._error.call(this,e)},e.prototype._complete=function(){this.connectable._isComplete=!0,this._unsubscribe(),t.prototype._complete.call(this)},e.prototype._unsubscribe=function(){var t=this.connectable;if(t){this.connectable=null;var e=t._connection;t._refCount=0,t._subject=null,t._connection=null,e&&e.unsubscribe()}},e}(r.b);o.a},"../../../../rxjs/_esm5/observable/DeferObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var r=n("../../../../rxjs/_esm5/Observable.js"),i=n("../../../../rxjs/_esm5/util/subscribeToResult.js"),o=n("../../../../rxjs/_esm5/OuterSubscriber.js"),s=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},a=function(t){function e(e){t.call(this),this.observableFactory=e}return s(e,t),e.create=function(t){return new e(t)},e.prototype._subscribe=function(t){return new c(t,this.observableFactory)},e}(r.a),c=function(t){function e(e,n){t.call(this,e),this.factory=n,this.tryDefer()}return s(e,t),e.prototype.tryDefer=function(){try{this._callFactory()}catch(t){this._error(t)}},e.prototype._callFactory=function(){var t=this.factory();t&&this.add(Object(i.a)(this,t))},e}(o.a)},"../../../../rxjs/_esm5/observable/EmptyObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/Observable.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(e){t.call(this),this.scheduler=e}return i(e,t),e.create=function(t){return new e(t)},e.dispatch=function(t){t.subscriber.complete()},e.prototype._subscribe=function(t){var n=this.scheduler;if(n)return n.schedule(e.dispatch,0,{subscriber:t});t.complete()},e}(r.a)},"../../../../rxjs/_esm5/observable/ErrorObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/Observable.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(e,n){t.call(this),this.error=e,this.scheduler=n}return i(e,t),e.create=function(t,n){return new e(t,n)},e.dispatch=function(t){var e=t.error;t.subscriber.error(e)},e.prototype._subscribe=function(t){var n=this.error,r=this.scheduler;if(t.syncErrorThrowable=!0,r)return r.schedule(e.dispatch,0,{error:n,subscriber:t});t.error(n)},e}(r.a)},"../../../../rxjs/_esm5/observable/ForkJoinObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return u});var r=n("../../../../rxjs/_esm5/Observable.js"),i=n("../../../../rxjs/_esm5/observable/EmptyObservable.js"),o=n("../../../../rxjs/_esm5/util/isArray.js"),s=n("../../../../rxjs/_esm5/util/subscribeToResult.js"),a=n("../../../../rxjs/_esm5/OuterSubscriber.js"),c=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},u=function(t){function e(e,n){t.call(this),this.sources=e,this.resultSelector=n}return c(e,t),e.create=function(){for(var t=[],n=0;n<arguments.length;n++)t[n-0]=arguments[n];if(null===t||0===arguments.length)return new i.a;var r=null;return"function"==typeof t[t.length-1]&&(r=t.pop()),1===t.length&&Object(o.a)(t[0])&&(t=t[0]),0===t.length?new i.a:new e(t,r)},e.prototype._subscribe=function(t){return new l(t,this.sources,this.resultSelector)},e}(r.a),l=function(t){function e(e,n,r){t.call(this,e),this.sources=n,this.resultSelector=r,this.completed=0,this.haveValues=0;var i=n.length;this.total=i,this.values=new Array(i);for(var o=0;o<i;o++){var a=n[o],c=Object(s.a)(this,a,null,o);c&&(c.outerIndex=o,this.add(c))}}return c(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.values[n]=e,i._hasValue||(i._hasValue=!0,this.haveValues++)},e.prototype.notifyComplete=function(t){var e=this.destination,n=this.haveValues,r=this.resultSelector,i=this.values,o=i.length;if(t._hasValue){if(this.completed++,this.completed===o){if(n===o){var s=r?r.apply(this,i):i;e.next(s)}e.complete()}}else e.complete()},e}(a.a)},"../../../../rxjs/_esm5/observable/FromEventObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return l});var r=n("../../../../rxjs/_esm5/Observable.js"),i=n("../../../../rxjs/_esm5/util/tryCatch.js"),o=n("../../../../rxjs/_esm5/util/isFunction.js"),s=n("../../../../rxjs/_esm5/util/errorObject.js"),a=n("../../../../rxjs/_esm5/Subscription.js"),c=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},u=Object.prototype.toString;var l=function(t){function e(e,n,r,i){t.call(this),this.sourceObj=e,this.eventName=n,this.selector=r,this.options=i}return c(e,t),e.create=function(t,n,r,i){return Object(o.a)(r)&&(i=r,r=void 0),new e(t,n,i,r)},e.setupSubscription=function(t,n,r,i,o){var s,c,l,p,d,h;if((h=t)&&"[object NodeList]"===u.call(h)||(d=t)&&"[object HTMLCollection]"===u.call(d))for(var f=0,m=t.length;f<m;f++)e.setupSubscription(t[f],n,r,i,o);else if(p=t,p&&"function"==typeof p.addEventListener&&"function"==typeof p.removeEventListener){var g=t;t.addEventListener(n,r,o),s=function(){return g.removeEventListener(n,r)}}else if(l=t,l&&"function"==typeof l.on&&"function"==typeof l.off){var v=t;t.on(n,r),s=function(){return v.off(n,r)}}else{if(!(c=t)||"function"!=typeof c.addListener||"function"!=typeof c.removeListener)throw new TypeError("Invalid event target");var y=t;t.addListener(n,r),s=function(){return y.removeListener(n,r)}}i.add(new a.a(s))},e.prototype._subscribe=function(t){var n=this.sourceObj,r=this.eventName,o=this.options,a=this.selector,c=a?function(){for(var e=[],n=0;n<arguments.length;n++)e[n-0]=arguments[n];var r=Object(i.a)(a).apply(void 0,e);r===s.a?t.error(s.a.e):t.next(r)}:function(e){return t.next(e)};e.setupSubscription(n,r,c,t,o)},e}(r.a)},"../../../../rxjs/_esm5/observable/FromEventPatternObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var r=n("../../../../rxjs/_esm5/util/isFunction.js"),i=n("../../../../rxjs/_esm5/Observable.js"),o=n("../../../../rxjs/_esm5/Subscription.js"),s=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},a=function(t){function e(e,n,r){t.call(this),this.addHandler=e,this.removeHandler=n,this.selector=r}return s(e,t),e.create=function(t,n,r){return new e(t,n,r)},e.prototype._subscribe=function(t){var e=this,n=this.removeHandler,i=this.selector?function(){for(var n=[],r=0;r<arguments.length;r++)n[r-0]=arguments[r];e._callSelector(t,n)}:function(e){t.next(e)},s=this._callAddHandler(i,t);Object(r.a)(n)&&t.add(new o.a(function(){n(i,s)}))},e.prototype._callSelector=function(t,e){try{var n=this.selector.apply(this,e);t.next(n)}catch(e){t.error(e)}},e.prototype._callAddHandler=function(t,e){try{return this.addHandler(t)||null}catch(t){e.error(t)}},e}(i.a)},"../../../../rxjs/_esm5/observable/FromObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return m});var r=n("../../../../rxjs/_esm5/util/isArray.js"),i=n("../../../../rxjs/_esm5/util/isArrayLike.js"),o=n("../../../../rxjs/_esm5/util/isPromise.js"),s=n("../../../../rxjs/_esm5/observable/PromiseObservable.js"),a=n("../../../../rxjs/_esm5/observable/IteratorObservable.js"),c=n("../../../../rxjs/_esm5/observable/ArrayObservable.js"),u=n("../../../../rxjs/_esm5/observable/ArrayLikeObservable.js"),l=n("../../../../rxjs/_esm5/symbol/iterator.js"),p=n("../../../../rxjs/_esm5/Observable.js"),d=n("../../../../rxjs/_esm5/operators/observeOn.js"),h=n("../../../../rxjs/_esm5/symbol/observable.js"),f=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},m=function(t){function e(e,n){t.call(this,null),this.ish=e,this.scheduler=n}return f(e,t),e.create=function(t,n){if(null!=t){if("function"==typeof t[h.a])return t instanceof p.a&&!n?t:new e(t,n);if(Object(r.a)(t))return new c.a(t,n);if(Object(o.a)(t))return new s.a(t,n);if("function"==typeof t[l.a]||"string"==typeof t)return new a.a(t,n);if(Object(i.a)(t))return new u.a(t,n)}throw new TypeError((null!==t&&typeof t||t)+" is not observable")},e.prototype._subscribe=function(t){var e=this.ish,n=this.scheduler;return null==n?e[h.a]().subscribe(t):e[h.a]().subscribe(new d.a(t,n,0))},e}(p.a)},"../../../../rxjs/_esm5/observable/IteratorObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var r=n("../../../../rxjs/_esm5/util/root.js"),i=n("../../../../rxjs/_esm5/Observable.js"),o=n("../../../../rxjs/_esm5/symbol/iterator.js"),s=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},a=function(t){function e(e,n){if(t.call(this),this.scheduler=n,null==e)throw new Error("iterator cannot be null.");this.iterator=function(t){var e=t[o.a];if(!e&&"string"==typeof t)return new c(t);if(!e&&void 0!==t.length)return new u(t);if(!e)throw new TypeError("object is not iterable");return t[o.a]()}(e)}return s(e,t),e.create=function(t,n){return new e(t,n)},e.dispatch=function(t){var e=t.index,n=t.hasError,r=t.iterator,i=t.subscriber;if(n)i.error(t.error);else{var o=r.next();o.done?i.complete():(i.next(o.value),t.index=e+1,i.closed?"function"==typeof r.return&&r.return():this.schedule(t))}},e.prototype._subscribe=function(t){var n=this.iterator,r=this.scheduler;if(r)return r.schedule(e.dispatch,0,{index:0,iterator:n,subscriber:t});for(;;){var i=n.next();if(i.done){t.complete();break}if(t.next(i.value),t.closed){"function"==typeof n.return&&n.return();break}}},e}(i.a),c=function(){function t(t,e,n){void 0===e&&(e=0),void 0===n&&(n=t.length),this.str=t,this.idx=e,this.len=n}return t.prototype[o.a]=function(){return this},t.prototype.next=function(){return this.idx<this.len?{done:!1,value:this.str.charAt(this.idx++)}:{done:!0,value:void 0}},t}(),u=function(){function t(t,e,n){void 0===e&&(e=0),void 0===n&&(n=function(t){var e=+t.length;if(isNaN(e))return 0;if(0===e||(n=e,"number"!=typeof n||!r.a.isFinite(n)))return e;var n;if(i=e,o=+i,(e=(0===o?o:isNaN(o)?o:o<0?-1:1)*Math.floor(Math.abs(e)))<=0)return 0;var i,o;if(e>l)return l;return e}(t)),this.arr=t,this.idx=e,this.len=n}return t.prototype[o.a]=function(){return this},t.prototype.next=function(){return this.idx<this.len?{done:!1,value:this.arr[this.idx++]}:{done:!0,value:void 0}},t}();var l=Math.pow(2,53)-1},"../../../../rxjs/_esm5/observable/PromiseObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return s});var r=n("../../../../rxjs/_esm5/util/root.js"),i=n("../../../../rxjs/_esm5/Observable.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},s=function(t){function e(e,n){t.call(this),this.promise=e,this.scheduler=n}return o(e,t),e.create=function(t,n){return new e(t,n)},e.prototype._subscribe=function(t){var e=this,n=this.promise,i=this.scheduler;if(null==i)this._isScalar?t.closed||(t.next(this.value),t.complete()):n.then(function(n){e.value=n,e._isScalar=!0,t.closed||(t.next(n),t.complete())},function(e){t.closed||t.error(e)}).then(null,function(t){r.a.setTimeout(function(){throw t})});else if(this._isScalar){if(!t.closed)return i.schedule(a,0,{value:this.value,subscriber:t})}else n.then(function(n){e.value=n,e._isScalar=!0,t.closed||t.add(i.schedule(a,0,{value:n,subscriber:t}))},function(e){t.closed||t.add(i.schedule(c,0,{err:e,subscriber:t}))}).then(null,function(t){r.a.setTimeout(function(){throw t})})},e}(i.a);function a(t){var e=t.value,n=t.subscriber;n.closed||(n.next(e),n.complete())}function c(t){var e=t.err,n=t.subscriber;n.closed||n.error(e)}},"../../../../rxjs/_esm5/observable/ScalarObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/Observable.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(e,n){t.call(this),this.value=e,this.scheduler=n,this._isScalar=!0,n&&(this._isScalar=!1)}return i(e,t),e.create=function(t,n){return new e(t,n)},e.dispatch=function(t){var e=t.done,n=t.value,r=t.subscriber;e?r.complete():(r.next(n),r.closed||(t.done=!0,this.schedule(t)))},e.prototype._subscribe=function(t){var n=this.value,r=this.scheduler;if(r)return r.schedule(e.dispatch,0,{done:!1,value:n,subscriber:t});t.next(n),t.closed||t.complete()},e}(r.a)},"../../../../rxjs/_esm5/observable/TimerObservable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return u});var r=n("../../../../rxjs/_esm5/util/isNumeric.js"),i=n("../../../../rxjs/_esm5/Observable.js"),o=n("../../../../rxjs/_esm5/scheduler/async.js"),s=n("../../../../rxjs/_esm5/util/isScheduler.js"),a=n("../../../../rxjs/_esm5/util/isDate.js"),c=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},u=function(t){function e(e,n,i){void 0===e&&(e=0),t.call(this),this.period=-1,this.dueTime=0,Object(r.a)(n)?this.period=Number(n)<1?1:Number(n):Object(s.a)(n)&&(i=n),Object(s.a)(i)||(i=o.a),this.scheduler=i,this.dueTime=Object(a.a)(e)?+e-this.scheduler.now():e}return c(e,t),e.create=function(t,n,r){return void 0===t&&(t=0),new e(t,n,r)},e.dispatch=function(t){var e=t.index,n=t.period,r=t.subscriber;if(r.next(e),!r.closed){if(-1===n)return r.complete();t.index=e+1,this.schedule(t,n)}},e.prototype._subscribe=function(t){var n=this.period,r=this.dueTime;return this.scheduler.schedule(e.dispatch,r,{index:0,period:n,subscriber:t})},e}(i.a)},"../../../../rxjs/_esm5/observable/combineLatest.js":function(t,e,n){"use strict";e.a=function(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];var n=null,a=null;Object(r.a)(t[t.length-1])&&(a=t.pop());"function"==typeof t[t.length-1]&&(n=t.pop());1===t.length&&Object(i.a)(t[0])&&(t=t[0]);return new o.a(t,a).lift(new s.a(n))};var r=n("../../../../rxjs/_esm5/util/isScheduler.js"),i=n("../../../../rxjs/_esm5/util/isArray.js"),o=n("../../../../rxjs/_esm5/observable/ArrayObservable.js"),s=n("../../../../rxjs/_esm5/operators/combineLatest.js")},"../../../../rxjs/_esm5/observable/concat.js":function(t,e,n){"use strict";e.a=function(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];if(1===t.length||2===t.length&&Object(r.a)(t[1]))return Object(o.a)(t[0]);return Object(s.a)()(i.a.apply(void 0,t))};var r=n("../../../../rxjs/_esm5/util/isScheduler.js"),i=n("../../../../rxjs/_esm5/observable/of.js"),o=n("../../../../rxjs/_esm5/observable/from.js"),s=n("../../../../rxjs/_esm5/operators/concatAll.js")},"../../../../rxjs/_esm5/observable/defer.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/DeferObservable.js").a.create},"../../../../rxjs/_esm5/observable/empty.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/EmptyObservable.js").a.create},"../../../../rxjs/_esm5/observable/forkJoin.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/ForkJoinObservable.js").a.create},"../../../../rxjs/_esm5/observable/from.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/FromObservable.js").a.create},"../../../../rxjs/_esm5/observable/fromEvent.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/FromEventObservable.js").a.create},"../../../../rxjs/_esm5/observable/fromEventPattern.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/FromEventPatternObservable.js").a.create},"../../../../rxjs/_esm5/observable/fromPromise.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/PromiseObservable.js").a.create},"../../../../rxjs/_esm5/observable/merge.js":function(t,e,n){"use strict";e.a=function(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];var n=Number.POSITIVE_INFINITY,a=null,c=t[t.length-1];Object(o.a)(c)?(a=t.pop(),t.length>1&&"number"==typeof t[t.length-1]&&(n=t.pop())):"number"==typeof c&&(n=t.pop());if(null===a&&1===t.length&&t[0]instanceof r.a)return t[0];return Object(s.a)(n)(new i.a(t,a))};var r=n("../../../../rxjs/_esm5/Observable.js"),i=n("../../../../rxjs/_esm5/observable/ArrayObservable.js"),o=n("../../../../rxjs/_esm5/util/isScheduler.js"),s=n("../../../../rxjs/_esm5/operators/mergeAll.js")},"../../../../rxjs/_esm5/observable/of.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/ArrayObservable.js").a.of},"../../../../rxjs/_esm5/observable/throw.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/ErrorObservable.js").a.create},"../../../../rxjs/_esm5/observable/timer.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=n("../../../../rxjs/_esm5/observable/TimerObservable.js").a.create},"../../../../rxjs/_esm5/operator/concatMap.js":function(t,e,n){"use strict";e.a=function(t,e){return Object(r.a)(t,e)(this)};var r=n("../../../../rxjs/_esm5/operators/concatMap.js")},"../../../../rxjs/_esm5/operator/filter.js":function(t,e,n){"use strict";e.a=function(t,e){return Object(r.a)(t,e)(this)};var r=n("../../../../rxjs/_esm5/operators/filter.js")},"../../../../rxjs/_esm5/operator/map.js":function(t,e,n){"use strict";e.a=function(t,e){return Object(r.a)(t,e)(this)};var r=n("../../../../rxjs/_esm5/operators/map.js")},"../../../../rxjs/_esm5/operator/share.js":function(t,e,n){"use strict";e.a=function(){return Object(r.a)()(this)};var r=n("../../../../rxjs/_esm5/operators/share.js")},"../../../../rxjs/_esm5/operators/audit.js":function(t,e,n){"use strict";e.a=function(t){return function(e){return e.lift(new c(t))}};var r=n("../../../../rxjs/_esm5/util/tryCatch.js"),i=n("../../../../rxjs/_esm5/util/errorObject.js"),o=n("../../../../rxjs/_esm5/OuterSubscriber.js"),s=n("../../../../rxjs/_esm5/util/subscribeToResult.js"),a=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var c=function(){function t(t){this.durationSelector=t}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.durationSelector))},t}(),u=function(t){function e(e,n){t.call(this,e),this.durationSelector=n,this.hasValue=!1}return a(e,t),e.prototype._next=function(t){if(this.value=t,this.hasValue=!0,!this.throttled){var e=Object(r.a)(this.durationSelector)(t);if(e===i.a)this.destination.error(i.a.e);else{var n=Object(s.a)(this,e);n.closed?this.clearThrottle():this.add(this.throttled=n)}}},e.prototype.clearThrottle=function(){var t=this.value,e=this.hasValue,n=this.throttled;n&&(this.remove(n),this.throttled=null,n.unsubscribe()),e&&(this.value=null,this.hasValue=!1,this.destination.next(t))},e.prototype.notifyNext=function(t,e,n,r){this.clearThrottle()},e.prototype.notifyComplete=function(){this.clearThrottle()},e}(o.a)},"../../../../rxjs/_esm5/operators/auditTime.js":function(t,e,n){"use strict";e.a=function(t,e){void 0===e&&(e=r.a);return Object(i.a)(function(){return Object(o.a)(t,e)})};var r=n("../../../../rxjs/_esm5/scheduler/async.js"),i=n("../../../../rxjs/_esm5/operators/audit.js"),o=n("../../../../rxjs/_esm5/observable/timer.js")},"../../../../rxjs/_esm5/operators/catchError.js":function(t,e,n){"use strict";e.a=function(t){return function(e){var n=new s(t),r=e.lift(n);return n.caught=r}};var r=n("../../../../rxjs/_esm5/OuterSubscriber.js"),i=n("../../../../rxjs/_esm5/util/subscribeToResult.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var s=function(){function t(t){this.selector=t}return t.prototype.call=function(t,e){return e.subscribe(new a(t,this.selector,this.caught))},t}(),a=function(t){function e(e,n,r){t.call(this,e),this.selector=n,this.caught=r}return o(e,t),e.prototype.error=function(e){if(!this.isStopped){var n=void 0;try{n=this.selector(e,this.caught)}catch(e){return void t.prototype.error.call(this,e)}this._unsubscribeAndRecycle(),this.add(Object(i.a)(this,n))}},e}(r.a)},"../../../../rxjs/_esm5/operators/combineLatest.js":function(t,e,n){"use strict";e.b=function(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];var n=null;"function"==typeof t[t.length-1]&&(n=t.pop());1===t.length&&Object(i.a)(t[0])&&(t=t[0].slice());return function(e){return e.lift.call(new r.a([e].concat(t)),new u(n))}},n.d(e,"a",function(){return u});var r=n("../../../../rxjs/_esm5/observable/ArrayObservable.js"),i=n("../../../../rxjs/_esm5/util/isArray.js"),o=n("../../../../rxjs/_esm5/OuterSubscriber.js"),s=n("../../../../rxjs/_esm5/util/subscribeToResult.js"),a=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},c={};var u=function(){function t(t){this.project=t}return t.prototype.call=function(t,e){return e.subscribe(new l(t,this.project))},t}(),l=function(t){function e(e,n){t.call(this,e),this.project=n,this.active=0,this.values=[],this.observables=[]}return a(e,t),e.prototype._next=function(t){this.values.push(c),this.observables.push(t)},e.prototype._complete=function(){var t=this.observables,e=t.length;if(0===e)this.destination.complete();else{this.active=e,this.toRespond=e;for(var n=0;n<e;n++){var r=t[n];this.add(Object(s.a)(this,r,r,n))}}},e.prototype.notifyComplete=function(t){0==(this.active-=1)&&this.destination.complete()},e.prototype.notifyNext=function(t,e,n,r,i){var o=this.values,s=o[n],a=this.toRespond?s===c?--this.toRespond:this.toRespond:0;o[n]=e,0===a&&(this.project?this._tryProject(o):this.destination.next(o.slice()))},e.prototype._tryProject=function(t){var e;try{e=this.project.apply(this,t)}catch(t){return void this.destination.error(t)}this.destination.next(e)},e}(o.a)},"../../../../rxjs/_esm5/operators/concatAll.js":function(t,e,n){"use strict";e.a=function(){return Object(r.a)(1)};var r=n("../../../../rxjs/_esm5/operators/mergeAll.js")},"../../../../rxjs/_esm5/operators/concatMap.js":function(t,e,n){"use strict";e.a=function(t,e){return Object(r.a)(t,e,1)};var r=n("../../../../rxjs/_esm5/operators/mergeMap.js")},"../../../../rxjs/_esm5/operators/debounceTime.js":function(t,e,n){"use strict";e.a=function(t,e){void 0===e&&(e=i.a);return function(n){return n.lift(new s(t,e))}};var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=n("../../../../rxjs/_esm5/scheduler/async.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var s=function(){function t(t,e){this.dueTime=t,this.scheduler=e}return t.prototype.call=function(t,e){return e.subscribe(new a(t,this.dueTime,this.scheduler))},t}(),a=function(t){function e(e,n,r){t.call(this,e),this.dueTime=n,this.scheduler=r,this.debouncedSubscription=null,this.lastValue=null,this.hasValue=!1}return o(e,t),e.prototype._next=function(t){this.clearDebounce(),this.lastValue=t,this.hasValue=!0,this.add(this.debouncedSubscription=this.scheduler.schedule(c,this.dueTime,this))},e.prototype._complete=function(){this.debouncedNext(),this.destination.complete()},e.prototype.debouncedNext=function(){this.clearDebounce(),this.hasValue&&(this.destination.next(this.lastValue),this.lastValue=null,this.hasValue=!1)},e.prototype.clearDebounce=function(){var t=this.debouncedSubscription;null!==t&&(this.remove(t),t.unsubscribe(),this.debouncedSubscription=null)},e}(r.a);function c(t){t.debouncedNext()}},"../../../../rxjs/_esm5/operators/delay.js":function(t,e,n){"use strict";e.a=function(t,e){void 0===e&&(e=r.a);var n=Object(i.a)(t)?+t-e.now():Math.abs(t);return function(t){return t.lift(new c(n,e))}};var r=n("../../../../rxjs/_esm5/scheduler/async.js"),i=n("../../../../rxjs/_esm5/util/isDate.js"),o=n("../../../../rxjs/_esm5/Subscriber.js"),s=n("../../../../rxjs/_esm5/Notification.js"),a=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var c=function(){function t(t,e){this.delay=t,this.scheduler=e}return t.prototype.call=function(t,e){return e.subscribe(new u(t,this.delay,this.scheduler))},t}(),u=function(t){function e(e,n,r){t.call(this,e),this.delay=n,this.scheduler=r,this.queue=[],this.active=!1,this.errored=!1}return a(e,t),e.dispatch=function(t){for(var e=t.source,n=e.queue,r=t.scheduler,i=t.destination;n.length>0&&n[0].time-r.now()<=0;)n.shift().notification.observe(i);if(n.length>0){var o=Math.max(0,n[0].time-r.now());this.schedule(t,o)}else e.active=!1},e.prototype._schedule=function(t){this.active=!0,this.add(t.schedule(e.dispatch,this.delay,{source:this,destination:this.destination,scheduler:t}))},e.prototype.scheduleNotification=function(t){if(!0!==this.errored){var e=this.scheduler,n=new l(e.now()+this.delay,t);this.queue.push(n),!1===this.active&&this._schedule(e)}},e.prototype._next=function(t){this.scheduleNotification(s.a.createNext(t))},e.prototype._error=function(t){this.errored=!0,this.queue=[],this.destination.error(t)},e.prototype._complete=function(){this.scheduleNotification(s.a.createComplete())},e}(o.a),l=function(){return function(t,e){this.time=t,this.notification=e}}()},"../../../../rxjs/_esm5/operators/filter.js":function(t,e,n){"use strict";e.a=function(t,e){return function(n){return n.lift(new o(t,e))}};var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var o=function(){function t(t,e){this.predicate=t,this.thisArg=e}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.predicate,this.thisArg))},t}(),s=function(t){function e(e,n,r){t.call(this,e),this.predicate=n,this.thisArg=r,this.count=0}return i(e,t),e.prototype._next=function(t){var e;try{e=this.predicate.call(this.thisArg,t,this.count++)}catch(t){return void this.destination.error(t)}e&&this.destination.next(t)},e}(r.a)},"../../../../rxjs/_esm5/operators/finalize.js":function(t,e,n){"use strict";e.a=function(t){return function(e){return e.lift(new s(t))}};var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=n("../../../../rxjs/_esm5/Subscription.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var s=function(){function t(t){this.callback=t}return t.prototype.call=function(t,e){return e.subscribe(new a(t,this.callback))},t}(),a=function(t){function e(e,n){t.call(this,e),this.add(new i.a(n))}return o(e,t),e}(r.a)},"../../../../rxjs/_esm5/operators/map.js":function(t,e,n){"use strict";e.a=function(t,e){return function(n){if("function"!=typeof t)throw new TypeError("argument is not a function. Are you looking for `mapTo()`?");return n.lift(new o(t,e))}};var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var o=function(){function t(t,e){this.project=t,this.thisArg=e}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.project,this.thisArg))},t}(),s=function(t){function e(e,n,r){t.call(this,e),this.project=n,this.count=0,this.thisArg=r||this}return i(e,t),e.prototype._next=function(t){var e;try{e=this.project.call(this.thisArg,t,this.count++)}catch(t){return void this.destination.error(t)}this.destination.next(e)},e}(r.a)},"../../../../rxjs/_esm5/operators/mergeAll.js":function(t,e,n){"use strict";e.a=function(t){void 0===t&&(t=Number.POSITIVE_INFINITY);return Object(r.a)(i.a,null,t)};var r=n("../../../../rxjs/_esm5/operators/mergeMap.js"),i=n("../../../../rxjs/_esm5/util/identity.js")},"../../../../rxjs/_esm5/operators/mergeMap.js":function(t,e,n){"use strict";e.a=function(t,e,n){void 0===n&&(n=Number.POSITIVE_INFINITY);return function(r){return"number"==typeof e&&(n=e,e=null),r.lift(new s(t,e,n))}};var r=n("../../../../rxjs/_esm5/util/subscribeToResult.js"),i=n("../../../../rxjs/_esm5/OuterSubscriber.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var s=function(){function t(t,e,n){void 0===n&&(n=Number.POSITIVE_INFINITY),this.project=t,this.resultSelector=e,this.concurrent=n}return t.prototype.call=function(t,e){return e.subscribe(new a(t,this.project,this.resultSelector,this.concurrent))},t}(),a=function(t){function e(e,n,r,i){void 0===i&&(i=Number.POSITIVE_INFINITY),t.call(this,e),this.project=n,this.resultSelector=r,this.concurrent=i,this.hasCompleted=!1,this.buffer=[],this.active=0,this.index=0}return o(e,t),e.prototype._next=function(t){this.active<this.concurrent?this._tryNext(t):this.buffer.push(t)},e.prototype._tryNext=function(t){var e,n=this.index++;try{e=this.project(t,n)}catch(t){return void this.destination.error(t)}this.active++,this._innerSub(e,t,n)},e.prototype._innerSub=function(t,e,n){this.add(Object(r.a)(this,t,e,n))},e.prototype._complete=function(){this.hasCompleted=!0,0===this.active&&0===this.buffer.length&&this.destination.complete()},e.prototype.notifyNext=function(t,e,n,r,i){this.resultSelector?this._notifyResultSelector(t,e,n,r):this.destination.next(e)},e.prototype._notifyResultSelector=function(t,e,n,r){var i;try{i=this.resultSelector(t,e,n,r)}catch(t){return void this.destination.error(t)}this.destination.next(i)},e.prototype.notifyComplete=function(t){var e=this.buffer;this.remove(t),this.active--,e.length>0?this._next(e.shift()):0===this.active&&this.hasCompleted&&this.destination.complete()},e}(i.a)},"../../../../rxjs/_esm5/operators/multicast.js":function(t,e,n){"use strict";e.a=function(t,e){return function(n){var o;if(o="function"==typeof t?t:function(){return t},"function"==typeof e)return n.lift(new i(o,e));var s=Object.create(n,r.a);return s.source=n,s.subjectFactory=o,s}};var r=n("../../../../rxjs/_esm5/observable/ConnectableObservable.js");var i=function(){function t(t,e){this.subjectFactory=t,this.selector=e}return t.prototype.call=function(t,e){var n=this.selector,r=this.subjectFactory(),i=n(r).subscribe(t);return i.add(e.subscribe(r)),i},t}()},"../../../../rxjs/_esm5/operators/observeOn.js":function(t,e,n){"use strict";n.d(e,"a",function(){return s});var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=n("../../../../rxjs/_esm5/Notification.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var s=function(t){function e(e,n,r){void 0===r&&(r=0),t.call(this,e),this.scheduler=n,this.delay=r}return o(e,t),e.dispatch=function(t){var e=t.notification,n=t.destination;e.observe(n),this.unsubscribe()},e.prototype.scheduleMessage=function(t){this.add(this.scheduler.schedule(e.dispatch,this.delay,new a(t,this.destination)))},e.prototype._next=function(t){this.scheduleMessage(i.a.createNext(t))},e.prototype._error=function(t){this.scheduleMessage(i.a.createError(t))},e.prototype._complete=function(){this.scheduleMessage(i.a.createComplete())},e}(r.a),a=function(){return function(t,e){this.notification=t,this.destination=e}}()},"../../../../rxjs/_esm5/operators/refCount.js":function(t,e,n){"use strict";e.a=function(){return function(t){return t.lift(new o(t))}};var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var o=function(){function t(t){this.connectable=t}return t.prototype.call=function(t,e){var n=this.connectable;n._refCount++;var r=new s(t,n),i=e.subscribe(r);return r.closed||(r.connection=n.connect()),i},t}(),s=function(t){function e(e,n){t.call(this,e),this.connectable=n}return i(e,t),e.prototype._unsubscribe=function(){var t=this.connectable;if(t){this.connectable=null;var e=t._refCount;if(e<=0)this.connection=null;else if(t._refCount=e-1,e>1)this.connection=null;else{var n=this.connection,r=t._connection;this.connection=null,!r||n&&r!==n||r.unsubscribe()}}else this.connection=null},e}(r.a)},"../../../../rxjs/_esm5/operators/share.js":function(t,e,n){"use strict";e.a=function(){return function(t){return Object(i.a)()(Object(r.a)(s)(t))}};var r=n("../../../../rxjs/_esm5/operators/multicast.js"),i=n("../../../../rxjs/_esm5/operators/refCount.js"),o=n("../../../../rxjs/_esm5/Subject.js");function s(){return new o.a}},"../../../../rxjs/_esm5/operators/startWith.js":function(t,e,n){"use strict";e.a=function(){for(var t=[],e=0;e<arguments.length;e++)t[e-0]=arguments[e];return function(e){var n=t[t.length-1];Object(a.a)(n)?t.pop():n=null;var c=t.length;return 1===c?Object(s.a)(new i.a(t[0],n),e):c>1?Object(s.a)(new r.a(t,n),e):Object(s.a)(new o.a(n),e)}};var r=n("../../../../rxjs/_esm5/observable/ArrayObservable.js"),i=n("../../../../rxjs/_esm5/observable/ScalarObservable.js"),o=n("../../../../rxjs/_esm5/observable/EmptyObservable.js"),s=n("../../../../rxjs/_esm5/observable/concat.js"),a=n("../../../../rxjs/_esm5/util/isScheduler.js")},"../../../../rxjs/_esm5/operators/switchMap.js":function(t,e,n){"use strict";e.a=function(t,e){return function(n){return n.lift(new s(t,e))}};var r=n("../../../../rxjs/_esm5/OuterSubscriber.js"),i=n("../../../../rxjs/_esm5/util/subscribeToResult.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var s=function(){function t(t,e){this.project=t,this.resultSelector=e}return t.prototype.call=function(t,e){return e.subscribe(new a(t,this.project,this.resultSelector))},t}(),a=function(t){function e(e,n,r){t.call(this,e),this.project=n,this.resultSelector=r,this.index=0}return o(e,t),e.prototype._next=function(t){var e,n=this.index++;try{e=this.project(t,n)}catch(t){return void this.destination.error(t)}this._innerSub(e,t,n)},e.prototype._innerSub=function(t,e,n){var r=this.innerSubscription;r&&r.unsubscribe(),this.add(this.innerSubscription=Object(i.a)(this,t,e,n))},e.prototype._complete=function(){var e=this.innerSubscription;e&&!e.closed||t.prototype._complete.call(this)},e.prototype._unsubscribe=function(){this.innerSubscription=null},e.prototype.notifyComplete=function(e){this.remove(e),this.innerSubscription=null,this.isStopped&&t.prototype._complete.call(this)},e.prototype.notifyNext=function(t,e,n,r,i){this.resultSelector?this._tryNotifyNext(t,e,n,r):this.destination.next(e)},e.prototype._tryNotifyNext=function(t,e,n,r){var i;try{i=this.resultSelector(t,e,n,r)}catch(t){return void this.destination.error(t)}this.destination.next(i)},e}(r.a)},"../../../../rxjs/_esm5/operators/take.js":function(t,e,n){"use strict";e.a=function(t){return function(e){return 0===t?new o.a:e.lift(new a(t))}};var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=n("../../../../rxjs/_esm5/util/ArgumentOutOfRangeError.js"),o=n("../../../../rxjs/_esm5/observable/EmptyObservable.js"),s=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var a=function(){function t(t){if(this.total=t,this.total<0)throw new i.a}return t.prototype.call=function(t,e){return e.subscribe(new c(t,this.total))},t}(),c=function(t){function e(e,n){t.call(this,e),this.total=n,this.count=0}return s(e,t),e.prototype._next=function(t){var e=this.total,n=++this.count;n<=e&&(this.destination.next(t),n===e&&(this.destination.complete(),this.unsubscribe()))},e}(r.a)},"../../../../rxjs/_esm5/operators/takeUntil.js":function(t,e,n){"use strict";e.a=function(t){return function(e){return e.lift(new s(t))}};var r=n("../../../../rxjs/_esm5/OuterSubscriber.js"),i=n("../../../../rxjs/_esm5/util/subscribeToResult.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var s=function(){function t(t){this.notifier=t}return t.prototype.call=function(t,e){return e.subscribe(new a(t,this.notifier))},t}(),a=function(t){function e(e,n){t.call(this,e),this.notifier=n,this.add(Object(i.a)(this,n))}return o(e,t),e.prototype.notifyNext=function(t,e,n,r,i){this.complete()},e.prototype.notifyComplete=function(){},e}(r.a)},"../../../../rxjs/_esm5/operators/tap.js":function(t,e,n){"use strict";e.a=function(t,e,n){return function(r){return r.lift(new o(t,e,n))}};var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)};var o=function(){function t(t,e,n){this.nextOrObserver=t,this.error=e,this.complete=n}return t.prototype.call=function(t,e){return e.subscribe(new s(t,this.nextOrObserver,this.error,this.complete))},t}(),s=function(t){function e(e,n,i,o){t.call(this,e);var s=new r.a(n,i,o);s.syncErrorThrowable=!0,this.add(s),this.safeSubscriber=s}return i(e,t),e.prototype._next=function(t){var e=this.safeSubscriber;e.next(t),e.syncErrorThrown?this.destination.error(e.syncErrorValue):this.destination.next(t)},e.prototype._error=function(t){var e=this.safeSubscriber;e.error(t),e.syncErrorThrown?this.destination.error(e.syncErrorValue):this.destination.error(t)},e.prototype._complete=function(){var t=this.safeSubscriber;t.complete(),t.syncErrorThrown?this.destination.error(t.syncErrorValue):this.destination.complete()},e}(r.a)},"../../../../rxjs/_esm5/scheduler/Action.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/Subscription.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(e,n){t.call(this)}return i(e,t),e.prototype.schedule=function(t,e){return void 0===e&&(e=0),this},e}(r.a)},"../../../../rxjs/_esm5/scheduler/AsyncAction.js":function(t,e,n){"use strict";n.d(e,"a",function(){return s});var r=n("../../../../rxjs/_esm5/util/root.js"),i=n("../../../../rxjs/_esm5/scheduler/Action.js"),o=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},s=function(t){function e(e,n){t.call(this,e,n),this.scheduler=e,this.work=n,this.pending=!1}return o(e,t),e.prototype.schedule=function(t,e){if(void 0===e&&(e=0),this.closed)return this;this.state=t,this.pending=!0;var n=this.id,r=this.scheduler;return null!=n&&(this.id=this.recycleAsyncId(r,n,e)),this.delay=e,this.id=this.id||this.requestAsyncId(r,this.id,e),this},e.prototype.requestAsyncId=function(t,e,n){return void 0===n&&(n=0),r.a.setInterval(t.flush.bind(t,this),n)},e.prototype.recycleAsyncId=function(t,e,n){if(void 0===n&&(n=0),null!==n&&this.delay===n&&!1===this.pending)return e;r.a.clearInterval(e)},e.prototype.execute=function(t,e){if(this.closed)return new Error("executing a cancelled action");this.pending=!1;var n=this._execute(t,e);if(n)return n;!1===this.pending&&null!=this.id&&(this.id=this.recycleAsyncId(this.scheduler,this.id,null))},e.prototype._execute=function(t,e){var n=!1,r=void 0;try{this.work(t)}catch(t){n=!0,r=!!t&&t||new Error(t)}if(n)return this.unsubscribe(),r},e.prototype._unsubscribe=function(){var t=this.id,e=this.scheduler,n=e.actions,r=n.indexOf(this);this.work=null,this.state=null,this.pending=!1,this.scheduler=null,-1!==r&&n.splice(r,1),null!=t&&(this.id=this.recycleAsyncId(e,t,null)),this.delay=null},e}(i.a)},"../../../../rxjs/_esm5/scheduler/AsyncScheduler.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/Scheduler.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(){t.apply(this,arguments),this.actions=[],this.active=!1,this.scheduled=void 0}return i(e,t),e.prototype.flush=function(t){var e=this.actions;if(this.active)e.push(t);else{var n;this.active=!0;do{if(n=t.execute(t.state,t.delay))break}while(t=e.shift());if(this.active=!1,n){for(;t=e.shift();)t.unsubscribe();throw n}}},e}(r.a)},"../../../../rxjs/_esm5/scheduler/QueueAction.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/scheduler/AsyncAction.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(e,n){t.call(this,e,n),this.scheduler=e,this.work=n}return i(e,t),e.prototype.schedule=function(e,n){return void 0===n&&(n=0),n>0?t.prototype.schedule.call(this,e,n):(this.delay=n,this.state=e,this.scheduler.flush(this),this)},e.prototype.execute=function(e,n){return n>0||this.closed?t.prototype.execute.call(this,e,n):this._execute(e,n)},e.prototype.requestAsyncId=function(e,n,r){return void 0===r&&(r=0),null!==r&&r>0||null===r&&this.delay>0?t.prototype.requestAsyncId.call(this,e,n,r):e.flush(this)},e}(r.a)},"../../../../rxjs/_esm5/scheduler/QueueScheduler.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/scheduler/AsyncScheduler.js"),i=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},o=function(t){function e(){t.apply(this,arguments)}return i(e,t),e}(r.a)},"../../../../rxjs/_esm5/scheduler/async.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var r=n("../../../../rxjs/_esm5/scheduler/AsyncAction.js"),i=new(n("../../../../rxjs/_esm5/scheduler/AsyncScheduler.js").a)(r.a)},"../../../../rxjs/_esm5/scheduler/queue.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var r=n("../../../../rxjs/_esm5/scheduler/QueueAction.js"),i=new(n("../../../../rxjs/_esm5/scheduler/QueueScheduler.js").a)(r.a)},"../../../../rxjs/_esm5/symbol/iterator.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=function(t){var e=t.Symbol;if("function"==typeof e)return e.iterator||(e.iterator=e("iterator polyfill")),e.iterator;var n=t.Set;if(n&&"function"==typeof(new n)["@@iterator"])return"@@iterator";var r=t.Map;if(r)for(var i=Object.getOwnPropertyNames(r.prototype),o=0;o<i.length;++o){var s=i[o];if("entries"!==s&&"size"!==s&&r.prototype[s]===r.prototype.entries)return s}return"@@iterator"}(n("../../../../rxjs/_esm5/util/root.js").a)},"../../../../rxjs/_esm5/symbol/observable.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=function(t){var e,n=t.Symbol;return"function"==typeof n?n.observable?e=n.observable:(e=n("observable"),n.observable=e):e="@@observable",e}(n("../../../../rxjs/_esm5/util/root.js").a)},"../../../../rxjs/_esm5/symbol/rxSubscriber.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var r=n("../../../../rxjs/_esm5/util/root.js").a.Symbol,i="function"==typeof r&&"function"==typeof r.for?r.for("rxSubscriber"):"@@rxSubscriber"},"../../../../rxjs/_esm5/util/ArgumentOutOfRangeError.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var r=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},i=function(t){function e(){var e=t.call(this,"argument out of range");this.name=e.name="ArgumentOutOfRangeError",this.stack=e.stack,this.message=e.message}return r(e,t),e}(Error)},"../../../../rxjs/_esm5/util/ObjectUnsubscribedError.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var r=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},i=function(t){function e(){var e=t.call(this,"object unsubscribed");this.name=e.name="ObjectUnsubscribedError",this.stack=e.stack,this.message=e.message}return r(e,t),e}(Error)},"../../../../rxjs/_esm5/util/UnsubscriptionError.js":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var r=this&&this.__extends||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);function r(){this.constructor=t}t.prototype=null===e?Object.create(e):(r.prototype=e.prototype,new r)},i=function(t){function e(e){t.call(this),this.errors=e;var n=Error.call(this,e?e.length+" errors occurred during unsubscription:\n  "+e.map(function(t,e){return e+1+") "+t.toString()}).join("\n  "):"");this.name=n.name="UnsubscriptionError",this.stack=n.stack,this.message=n.message}return r(e,t),e}(Error)},"../../../../rxjs/_esm5/util/errorObject.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r={e:{}}},"../../../../rxjs/_esm5/util/identity.js":function(t,e,n){"use strict";e.a=function(t){return t}},"../../../../rxjs/_esm5/util/isArray.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=Array.isArray||function(t){return t&&"number"==typeof t.length}},"../../../../rxjs/_esm5/util/isArrayLike.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=function(t){return t&&"number"==typeof t.length}},"../../../../rxjs/_esm5/util/isDate.js":function(t,e,n){"use strict";e.a=function(t){return t instanceof Date&&!isNaN(+t)}},"../../../../rxjs/_esm5/util/isFunction.js":function(t,e,n){"use strict";e.a=function(t){return"function"==typeof t}},"../../../../rxjs/_esm5/util/isNumeric.js":function(t,e,n){"use strict";e.a=function(t){return!Object(r.a)(t)&&t-parseFloat(t)+1>=0};var r=n("../../../../rxjs/_esm5/util/isArray.js")},"../../../../rxjs/_esm5/util/isObject.js":function(t,e,n){"use strict";e.a=function(t){return null!=t&&"object"==typeof t}},"../../../../rxjs/_esm5/util/isPromise.js":function(t,e,n){"use strict";e.a=function(t){return t&&"function"!=typeof t.subscribe&&"function"==typeof t.then}},"../../../../rxjs/_esm5/util/isScheduler.js":function(t,e,n){"use strict";e.a=function(t){return t&&"function"==typeof t.schedule}},"../../../../rxjs/_esm5/util/noop.js":function(t,e,n){"use strict";e.a=function(){}},"../../../../rxjs/_esm5/util/pipe.js":function(t,e,n){"use strict";e.a=i;var r=n("../../../../rxjs/_esm5/util/noop.js");function i(t){return t?1===t.length?t[0]:function(e){return t.reduce(function(t,e){return e(t)},e)}:r.a}},"../../../../rxjs/_esm5/util/root.js":function(t,e,n){"use strict";(function(t){n.d(e,"a",function(){return o});var r="undefined"!=typeof window&&window,i="undefined"!=typeof self&&"undefined"!=typeof WorkerGlobalScope&&self instanceof WorkerGlobalScope&&self,o=r||void 0!==t&&t||i}).call(e,n("../../../../webpack/buildin/global.js"))},"../../../../rxjs/_esm5/util/subscribeToResult.js":function(t,e,n){"use strict";e.a=function(t,e,n,p){var d=new u.a(t,n,p);if(d.closed)return null;if(e instanceof a.a)return e._isScalar?(d.next(e.value),d.complete(),null):(d.syncErrorThrowable=!0,e.subscribe(d));if(Object(i.a)(e)){for(var h=0,f=e.length;h<f&&!d.closed;h++)d.next(e[h]);d.closed||d.complete()}else{if(Object(o.a)(e))return e.then(function(t){d.closed||(d.next(t),d.complete())},function(t){return d.error(t)}).then(null,function(t){r.a.setTimeout(function(){throw t})}),d;if(e&&"function"==typeof e[c.a])for(var m=e[c.a]();;){var g=m.next();if(g.done){d.complete();break}if(d.next(g.value),d.closed)break}else if(e&&"function"==typeof e[l.a]){var v=e[l.a]();if("function"==typeof v.subscribe)return v.subscribe(new u.a(t,n,p));d.error(new TypeError("Provided object does not correctly implement Symbol.observable"))}else{var y=Object(s.a)(e)?"an invalid object":"'"+e+"'",b="You provided "+y+" where a stream was expected. You can provide an Observable, Promise, Array, or Iterable.";d.error(new TypeError(b))}}return null};var r=n("../../../../rxjs/_esm5/util/root.js"),i=n("../../../../rxjs/_esm5/util/isArrayLike.js"),o=n("../../../../rxjs/_esm5/util/isPromise.js"),s=n("../../../../rxjs/_esm5/util/isObject.js"),a=n("../../../../rxjs/_esm5/Observable.js"),c=n("../../../../rxjs/_esm5/symbol/iterator.js"),u=n("../../../../rxjs/_esm5/InnerSubscriber.js"),l=n("../../../../rxjs/_esm5/symbol/observable.js")},"../../../../rxjs/_esm5/util/toSubscriber.js":function(t,e,n){"use strict";e.a=function(t,e,n){if(t){if(t instanceof r.a)return t;if(t[i.a])return t[i.a]()}if(!t&&!e&&!n)return new r.a(o.a);return new r.a(t,e,n)};var r=n("../../../../rxjs/_esm5/Subscriber.js"),i=n("../../../../rxjs/_esm5/symbol/rxSubscriber.js"),o=n("../../../../rxjs/_esm5/Observer.js")},"../../../../rxjs/_esm5/util/tryCatch.js":function(t,e,n){"use strict";e.a=function(t){return r=t,o};var r,i=n("../../../../rxjs/_esm5/util/errorObject.js");function o(){try{return r.apply(this,arguments)}catch(t){return i.a.e=t,i.a}}},"../../../../tslib/tslib.es6.js":function(t,e,n){"use strict";e.b=function(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)},n.d(e,"a",function(){return i});var r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])};var i=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++){e=arguments[n];for(var i in e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i])}return t}},"../../../../vega-canvas/index.js":function(t,e,n){"use strict";e.a=function(t,e){return Object(r.a)(t,e)||Object(i.a)(t,e)||null},e.b=function(){return Object(r.b)()||Object(i.b)()||null};var r=n("../../../../vega-canvas/src/domCanvas.js"),i=n("../../../../vega-canvas/src/nodeCanvas.js")},"../../../../vega-canvas/src/domCanvas.js":function(t,e,n){"use strict";e.a=function(t,e){if("undefined"!=typeof document&&document.createElement){var n=document.createElement("canvas");if(n&&n.getContext)return n.width=t,n.height=e,n}return null},e.b=function(){return"undefined"!=typeof Image?Image:null}},"../../../../vega-canvas/src/nodeCanvas.js":function(t,e,n){"use strict";var r;e.a=function(t,e){if(r)try{return new r(t,e)}catch(t){}return null},e.b=function(){return r&&r.Image||null};try{if(!(r=n(2)))throw 1}catch(t){try{r=n(3)}catch(t){r=null}}},"../../../../vega-crossfilter/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../vega-crossfilter/src/CrossFilter.js");n.d(e,"crossfilter",function(){return r.a});var i=n("../../../../vega-crossfilter/src/ResolveFilter.js");n.d(e,"resolvefilter",function(){return i.a})},"../../../../vega-crossfilter/src/Bitmaps.js":function(t,e,n){"use strict";e.a=function(){var t=8,e=[],n=Object(r.b)(0),o=i(0,t),s=i(0,t);return{data:function(){return e},seen:function(){return t=n,r=e.length,n=t.length>=r?t:((i=i||new t.constructor(r)).set(t),i);var t,r,i},add:function(t){for(var n,r=0,i=e.length,o=t.length;r<o;++r)(n=t[r])._index=i++,e.push(n)},remove:function(t,n){var r,i,a,c=e.length,u=Array(c-t),l=e;for(i=0;!n[i]&&i<c;++i)u[i]=e[i],l[i]=i;for(a=i;i<c;++i)r=e[i],n[i]?l[i]=-1:(l[i]=a,o[a]=o[i],s[a]=s[i],u[a]=r,r._index=a++),o[i]=0;return e=u,l},size:function(){return e.length},curr:function(){return o},prev:function(){return s},reset:function(t){s[t]=o[t]},all:function(){return t<257?255:t<65537?65535:4294967295},set:function(t,e){o[t]|=e},clear:function(t,e){o[t]&=~e},resize:function(e,n){var r=o.length;(e>r||n>t)&&(t=Math.max(n,t),o=i(e,t,o),s=i(e,t))}}};var r=n("../../../../vega-crossfilter/src/arrays.js");function i(t,e,n){var i=(e<257?r.c:e<65537?r.a:r.b)(t);return n&&i.set(n),i}},"../../../../vega-crossfilter/src/CrossFilter.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../vega-crossfilter/src/Bitmaps.js"),i=n("../../../../vega-crossfilter/src/Dimension.js"),o=n("../../../../vega-crossfilter/src/SortedIndex.js"),s=n("../../../../vega-dataflow/index.js"),a=n("../../../../vega-util/index.js");function c(t){s.d.call(this,Object(r.a)(),t),this._indices=null,this._dims=null}c.Definition={type:"CrossFilter",metadata:{},params:[{name:"fields",type:"field",array:!0,required:!0},{name:"query",type:"array",array:!0,required:!0,content:{type:"number",array:!0,length:2}}]};var u=Object(a.s)(c,s.d);u.transform=function(t,e){return this._dims?t.modified("fields")||t.fields.some(function(t){return e.modified(t.fields)})?this.reinit(t,e):this.eval(t,e):this.init(t,e)},u.init=function(t,e){for(var n,r,s=t.fields,a=t.query,c=this._indices={},u=this._dims=[],l=a.length,p=0;p<l;++p)r=c[n=s[p].fname]||(c[n]=Object(o.a)()),u.push(Object(i.a)(r,p,a[p]));return this.eval(t,e)},u.reinit=function(t,e){var n,r,s,a,c,u,l,p,d,h=e.materialize().fork(),f=t.fields,m=t.query,g=this._indices,v=this._dims,y=this.value,b=y.curr(),_=y.prev(),j=y.all(),x=h.rem=h.add,w=h.mod,O=m.length,C={};if(_.set(b),e.rem.length&&(c=this.remove(t,e,h)),e.add.length&&y.add(e.add),e.mod.length)for(u={},l=0,p=(a=e.mod).length;l<p;++l)u[a[l]._index]=1;for(l=0;l<O;++l)d=f[l],(!v[l]||t.modified("fields",l)||e.modified(d.fields))&&((n=C[s=d.fname])||(g[s]=r=Object(o.a)(),C[s]=n=r.insert(d,e.source,0)),v[l]=Object(i.a)(r,l,m[l]).onAdd(n,b));for(l=0,p=y.data().length;l<p;++l)c[l]||(_[l]!==b[l]?x.push(l):u[l]&&b[l]!==j&&w.push(l));return y.mask=(1<<O)-1,h},u.eval=function(t,e){var n=e.materialize().fork(),r=this._dims.length,i=0;return e.rem.length&&(this.remove(t,e,n),i|=(1<<r)-1),t.modified("query")&&!t.modified("fields")&&(i|=this.update(t,e,n)),e.add.length&&(this.insert(t,e,n),i|=(1<<r)-1),e.mod.length&&(this.modify(e,n),i|=(1<<r)-1),this.value.mask=i,n},u.insert=function(t,e,n){var r,i,o,s=e.add,a=this.value,c=this._dims,u=this._indices,l=t.fields,p={},d=n.add,h=a.size(),f=h+s.length,m=c.length;a.resize(f,m),a.add(s);var g=a.curr(),v=a.prev(),y=a.all();for(r=0;r<m;++r)o=p[i=l[r].fname]||(p[i]=u[i].insert(l[r],s,h)),c[r].onAdd(o,g);for(;h<f;++h)v[h]=y,g[h]!==y&&d.push(h)},u.modify=function(t,e){var n,r,i,o=e.mod,s=this.value,a=s.curr(),c=s.all(),u=t.mod;for(n=0,r=u.length;n<r;++n)a[i=u[n]._index]!==c&&o.push(i)},u.remove=function(t,e,n){var r,i,o,s,a=this._indices,c=this.value,u=c.curr(),l=c.prev(),p=c.all(),d={},h=n.rem,f=e.rem;for(r=0,i=f.length;r<i;++r)d[o=f[r]._index]=1,l[o]=s=u[o],u[o]=p,s!==p&&h.push(o);for(o in a)a[o].remove(i,d);return this.reindex(e,i,d),d},u.reindex=function(t,e,n){var r=this._indices,i=this.value;t.runAfter(function(){var t=i.remove(e,n);for(var o in r)r[o].reindex(t)})},u.update=function(t,e,n){var r,i,o=this._dims,s=t.query,a=e.stamp,c=o.length,u=0;for(n.filters=0,i=0;i<c;++i)t.modified("query",i)&&(r=i,++u);if(1===u)u=o[r].one,this.incrementOne(o[r],s[r],n.add,n.rem);else for(i=0,u=0;i<c;++i)t.modified("query",i)&&(u|=o[i].one,this.incrementAll(o[i],s[i],a,n.add),n.rem=n.add);return u},u.incrementAll=function(t,e,n,r){var i,o,s,a=this.value,c=a.seen(),u=a.curr(),l=a.prev(),p=t.index(),d=t.bisect(t.range),h=t.bisect(e),f=h[0],m=h[1],g=d[0],v=d[1],y=t.one;if(f<g)for(i=f,o=Math.min(g,m);i<o;++i)c[s=p[i]]!==n&&(l[s]=u[s],c[s]=n,r.push(s)),u[s]^=y;else if(f>g)for(i=g,o=Math.min(f,v);i<o;++i)c[s=p[i]]!==n&&(l[s]=u[s],c[s]=n,r.push(s)),u[s]^=y;if(m>v)for(i=Math.max(f,v),o=m;i<o;++i)c[s=p[i]]!==n&&(l[s]=u[s],c[s]=n,r.push(s)),u[s]^=y;else if(m<v)for(i=Math.max(g,m),o=v;i<o;++i)c[s=p[i]]!==n&&(l[s]=u[s],c[s]=n,r.push(s)),u[s]^=y;t.range=e.slice()},u.incrementOne=function(t,e,n,r){var i,o,s,a=this.value.curr(),c=t.index(),u=t.bisect(t.range),l=t.bisect(e),p=l[0],d=l[1],h=u[0],f=u[1],m=t.one;if(p<h)for(i=p,o=Math.min(h,d);i<o;++i)a[s=c[i]]^=m,n.push(s);else if(p>h)for(i=h,o=Math.min(p,f);i<o;++i)a[s=c[i]]^=m,r.push(s);if(d>f)for(i=Math.max(p,f),o=d;i<o;++i)a[s=c[i]]^=m,n.push(s);else if(d<f)for(i=Math.max(h,d),o=f;i<o;++i)a[s=c[i]]^=m,r.push(s);t.range=e.slice()}},"../../../../vega-crossfilter/src/Dimension.js":function(t,e,n){"use strict";e.a=function(t,e,n){var r=1<<e;return{one:r,zero:~r,range:n.slice(),bisect:t.bisect,index:t.index,size:t.size,onAdd:function(t,e){var n,i=this.bisect(this.range,t.value),o=t.index,s=i[0],a=i[1],c=o.length;for(n=0;n<s;++n)e[o[n]]|=r;for(n=a;n<c;++n)e[o[n]]|=r;return this}}}},"../../../../vega-crossfilter/src/ResolveFilter.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}o.Definition={type:"ResolveFilter",metadata:{},params:[{name:"ignore",type:"number",required:!0,description:"A bit mask indicating which filters to ignore."},{name:"filter",type:"object",required:!0,description:"Per-tuple filter bitmaps from a CrossFilter transform."}]},Object(i.s)(o,r.d).transform=function(t,e){var n=~(t.ignore||0),r=t.filter,i=r.mask;if(0==(i&n))return e.StopPropagation;var o=e.fork(e.ALL),s=r.data(),a=r.curr(),c=r.prev(),u=function(t){return a[t]&n?null:s[t]};return o.filter(o.MOD,u),i&i-1?(o.filter(o.ADD,function(t){var e=a[t]&n;return!e&&e^c[t]&n?s[t]:null}),o.filter(o.REM,function(t){var e=a[t]&n;return e&&!(e^e^c[t]&n)?s[t]:null})):(o.filter(o.ADD,u),o.filter(o.REM,function(t){return(a[t]&n)===i?s[t]:null})),o.filter(o.SOURCE,function(t){return u(t._index)})}},"../../../../vega-crossfilter/src/SortedIndex.js":function(t,e,n){"use strict";e.a=function(){var t=Object(r.b)(0),e=[],n=0;return{insert:function(o,s,a){if(!s.length)return[];var c,u,l,p=n,d=s.length,h=Array(d),f=Object(r.b)(d);for(l=0;l<d;++l)h[l]=o(s[l]),f[l]=l;if(h=function(t,e){return t.sort.call(e,function(e,n){var r=t[e],i=t[n];return r<i?-1:r>i?1:0}),Object(i.l)(t,e)}(h,f),p)c=e,u=t,e=Array(p+d),t=Object(r.b)(p+d),function(t,e,n,r,i,o,s,a,c){var u,l=0,p=0;for(u=0;l<r&&p<s;++u)e[l]<i[p]?(a[u]=e[l],c[u]=n[l++]):(a[u]=i[p],c[u]=o[p++]+t);for(;l<r;++l,++u)a[u]=e[l],c[u]=n[l];for(;p<s;++p,++u)a[u]=i[p],c[u]=o[p]+t}(a,c,u,p,h,f,d,e,t);else{if(a>0)for(l=0;l<d;++l)f[l]+=a;e=h,t=f}return n=p+d,{index:f,value:h}},remove:function(r,i){var o,s,a,c=n;for(s=0;!i[t[s]]&&s<c;++s);for(a=s;s<c;++s)i[o=t[s]]||(t[a]=o,e[a]=e[s],++a);n=c-r},bisect:function(t,r){var o;r?o=r.length:(r=e,o=n);return[Object(i.c)(r,t[0],0,o),Object(i.d)(r,t[1],0,o)]},reindex:function(e){for(var r=0,i=n;r<i;++r)t[r]=e[t[r]]},index:function(){return t},size:function(){return n}}};var r=n("../../../../vega-crossfilter/src/arrays.js"),i=n("../../../../d3-array/index.js")},"../../../../vega-crossfilter/src/arrays.js":function(t,e,n){"use strict";e.c=function(t){return new Uint8Array(t)},e.a=function(t){return new Uint16Array(t)},e.b=function(t){return new Uint32Array(t)}},"../../../../vega-dataflow/index.js":function(t,e,n){"use strict";n("../../../../vega-dataflow/src/util/UniqueList.js");var r=n("../../../../vega-dataflow/src/ChangeSet.js");n.d(e,"e",function(){return r.a}),n.d(e,"i",function(){return r.b});var i=n("../../../../vega-dataflow/src/dataflow/Dataflow.js");n.d(e,"a",function(){return i.a});var o=n("../../../../vega-dataflow/src/EventStream.js");n.d(e,"b",function(){return o.a});n("../../../../vega-dataflow/src/Parameters.js"),n("../../../../vega-dataflow/src/Pulse.js"),n("../../../../vega-dataflow/src/MultiPulse.js");var s=n("../../../../vega-dataflow/src/Operator.js");n.d(e,"c",function(){return s.a});var a=n("../../../../vega-dataflow/src/Transform.js");n.d(e,"d",function(){return a.a});var c=n("../../../../vega-dataflow/src/Tuple.js");n.d(e,"g",function(){return c.a}),n.d(e,"k",function(){return c.d}),n.d(e,"h",function(){return c.b}),n.d(e,"j",function(){return c.c}),n.d(e,"l",function(){return c.e}),n.d(e,"n",function(){return c.f});var u=n("../../../../vega-dataflow/src/register.js");n.d(e,"f",function(){return u.a}),n.d(e,"m",function(){return u.b})},"../../../../vega-dataflow/src/ChangeSet.js":function(t,e,n){"use strict";e.b=function(t){return t&&t.constructor===o},e.a=o;var r=n("../../../../vega-dataflow/src/Tuple.js"),i=n("../../../../vega-util/index.js");function o(){var t=[],e=[],n=[],s=[],a=[],c=!1;return{constructor:o,insert:function(e){for(var n=Object(i.g)(e),r=0,o=n.length;r<o;++r)t.push(n[r]);return this},remove:function(t){for(var n=Object(i.w)(t)?s:e,r=Object(i.g)(t),o=0,a=r.length;o<a;++o)n.push(r[o]);return this},modify:function(t,e,r){var o={field:e,value:Object(i.i)(r)};return Object(i.w)(t)?(o.filter=t,a.push(o)):(o.tuple=t,n.push(o)),this},encode:function(t,e){return Object(i.w)(t)?a.push({filter:t,field:e}):n.push({tuple:t,field:e}),this},reflow:function(){return c=!0,this},pulse:function(i,o){var u,l,p,d,h,f,m;for(l=0,p=t.length;l<p;++l)i.add.push(Object(r.b)(t[l]));for(u={},l=0,p=e.length;l<p;++l)f=e[l],u[Object(r.f)(f)]=f;for(l=0,p=s.length;l<p;++l)h=s[l],o.forEach(function(t){h(t)&&(u[Object(r.f)(t)]=t)});for(m in u)i.rem.push(u[m]);function g(t,e,n){n?t[e]=n(t):i.encode=e,c||(u[Object(r.f)(t)]=t)}for(u={},l=0,p=n.length;l<p;++l)g((d=n[l]).tuple,d.field,d.value),i.modifies(d.field);for(l=0,p=a.length;l<p;++l)d=a[l],h=d.filter,o.forEach(function(t){h(t)&&g(t,d.field,d.value)}),i.modifies(d.field);if(c)i.mod=e.length||s.length?o.filter(function(t){return u.hasOwnProperty(Object(r.f)(t))}):o.slice();else for(m in u)i.mod.push(u[m]);return i}}}},"../../../../vega-dataflow/src/EventStream.js":function(t,e,n){"use strict";e.a=s,e.b=a;var r=n("../../../../vega-dataflow/src/util/UniqueList.js"),i=n("../../../../vega-util/index.js"),o=0;function s(t,e,n){this.id=++o,this.value=null,n&&(this.receive=n),t&&(this._filter=t),e&&(this._apply=e)}function a(t,e,n){return new s(t,e,n)}var c=s.prototype;c._filter=i.S,c._apply=i.r,c.targets=function(){return this._targets||(this._targets=Object(r.a)(i.q))},c.consume=function(t){return arguments.length?(this._consume=!!t,this):!!this._consume},c.receive=function(t){if(this._filter(t)){for(var e=this.value=this._apply(t),n=this._targets,r=n?n.length:0,i=0;i<r;++i)n[i].receive(e);this._consume&&(t.preventDefault(),t.stopPropagation())}},c.filter=function(t){var e=a(t);return this.targets().add(e),e},c.apply=function(t){var e=a(null,t);return this.targets().add(e),e},c.merge=function(){var t=a();this.targets().add(t);for(var e=0,n=arguments.length;e<n;++e)arguments[e].targets().add(t);return t},c.throttle=function(t){var e=-1;return this.filter(function(){var n=Date.now();return n-e>t?(e=n,1):0})},c.debounce=function(t){var e=a();return this.targets().add(a(null,null,Object(i.j)(t,function(t){var n=t.dataflow;e.receive(t),n&&n.run&&n.run()}))),e},c.between=function(t,e){var n=!1;return t.targets().add(a(null,null,function(){n=!0})),e.targets().add(a(null,null,function(){n=!1})),this.filter(function(){return n})}},"../../../../vega-dataflow/src/MultiPulse.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/src/Pulse.js"),i=n("../../../../vega-util/index.js");function o(t,e,n,r){var i,o,s,a,c,u=0;for(this.dataflow=t,this.stamp=e,this.fields=null,this.encode=r||null,this.pulses=n,s=0,a=n.length;s<a;++s)if((i=n[s]).stamp===e){if(i.fields){o=this.fields||(this.fields={});for(c in i.fields)o[c]=1}i.changed(this.ADD)&&(u|=this.ADD),i.changed(this.REM)&&(u|=this.REM),i.changed(this.MOD)&&(u|=this.MOD)}this.changes=u}var s=Object(i.s)(o,r.b);s.fork=function(t){var e=new r.b(this.dataflow).init(this,t&this.NO_FIELDS);return void 0!==t&&(t&e.ADD&&this.visit(e.ADD,function(t){return e.add.push(t)}),t&e.REM&&this.visit(e.REM,function(t){return e.rem.push(t)}),t&e.MOD&&this.visit(e.MOD,function(t){return e.mod.push(t)})),e},s.changed=function(t){return this.changes&t},s.modified=function(t){var e=this.fields;return e&&this.changes&this.MOD?Object(i.t)(t)?t.some(function(t){return e[t]}):e[t]:0},s.filter=function(){Object(i.k)("MultiPulse does not support filtering.")},s.materialize=function(){Object(i.k)("MultiPulse does not support materialization.")},s.visit=function(t,e){var n=this.pulses,r=n.length,i=0;if(t&this.SOURCE)for(;i<r;++i)n[i].visit(t,e);else for(;i<r;++i)n[i].stamp===this.stamp&&n[i].visit(t,e);return this}},"../../../../vega-dataflow/src/Operator.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../vega-dataflow/src/Parameters.js"),i=n("../../../../vega-dataflow/src/util/UniqueList.js"),o=n("../../../../vega-util/index.js"),s=0,a=new r.a;function c(t,e,n,r){this.id=++s,this.value=t,this.stamp=-1,this.rank=-1,this.qrank=-1,this.flags=0,e&&(this._update=e),n&&this.parameters(n,r)}var u=c.prototype;function l(t){return function(e){var n=this.flags;return 0===arguments.length?!!(n&t):(this.flags=e?n|t:n&~t,this)}}u.targets=function(){return this._targets||(this._targets=Object(i.a)(o.q))},u.set=function(t){return this.value!==t?(this.value=t,1):0},u.skip=l(1),u.modified=l(2),u.parameters=function(t,e){e=!1!==e;var n,i,s,a,u=this,l=u._argval=u._argval||new r.a,p=u._argops=u._argops||[],d=[];function h(t,n,r){r instanceof c?(r!==u&&(e&&r.targets().add(u),d.push(r)),p.push({op:r,name:t,index:n})):l.set(t,n,r)}for(n in t)if(i=t[n],"pulse"===n)Object(o.g)(i).forEach(function(t){t instanceof c?t!==u&&(t.targets().add(u),d.push(t)):Object(o.k)("Pulse parameters must be operator instances.")}),u.source=i;else if(Object(o.t)(i))for(l.set(n,-1,Array(s=i.length)),a=0;a<s;++a)h(n,a,i[a]);else h(n,-1,i);return this.marshall().clear(),d},u.marshall=function(t){var e,n,r,i,o,s=this._argval||a,c=this._argops;if(c&&(r=c.length))for(n=0;n<r;++n)o=(i=(e=c[n]).op).modified()&&i.stamp===t,s.set(e.name,e.index,i.value,o);return s},u.evaluate=function(t){if(this._update){var e=this.marshall(t.stamp),n=this._update(e,t);if(e.clear(),n!==this.value)this.value=n;else if(!this.modified())return t.StopPropagation}},u.run=function(t){return t.stamp<=this.stamp?t.StopPropagation:(this.skip()?(this.skip(!1),e=0):e=this.evaluate(t),this.stamp=t.stamp,this.pulse=e,e||t);var e}},"../../../../vega-dataflow/src/Parameters.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-util/index.js"),i="_:mod:_";function o(){Object.defineProperty(this,i,{writable:!0,value:{}})}var s=o.prototype;s.set=function(t,e,n,o){var s=this[t],a=this[i];return null!=e&&e>=0?(s[e]!==n||o)&&(s[e]=n,a[e+":"+t]=-1,a[t]=-1):(s!==n||o)&&(this[t]=n,a[t]=Object(r.t)(n)?1+n.length:-1),this},s.modified=function(t,e){var n,o=this[i];if(!arguments.length){for(n in o)if(o[n])return!0;return!1}if(Object(r.t)(t)){for(n=0;n<t.length;++n)if(o[t[n]])return!0;return!1}return null!=e&&e>=0?e+1<o[t]||!!o[e+":"+t]:!!o[t]},s.clear=function(){return this[i]={},this}},"../../../../vega-dataflow/src/Pulse.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o}),e.b=s;var r=n("../../../../vega-dataflow/src/Tuple.js"),i=n("../../../../vega-util/index.js"),o={};function s(t,e,n){this.dataflow=t,this.stamp=null==e?-1:e,this.add=[],this.rem=[],this.mod=[],this.fields=null,this.encode=n||null}var a=s.prototype;function c(t,e){return t?function(n,r){return t(n,r)&&e(n,r)}:e}function u(t,e){var n=[];return Object(i.T)(t,e,function(t){n.push(t)}),n}function l(t,e){var n={};return t.visit(e,function(t){n[Object(r.f)(t)]=1}),function(t){return n[Object(r.f)(t)]?null:t}}a.StopPropagation=o,a.ADD=1,a.REM=2,a.MOD=4,a.ADD_REM=3,a.ADD_MOD=5,a.ALL=7,a.REFLOW=8,a.SOURCE=16,a.NO_SOURCE=32,a.NO_FIELDS=64,a.fork=function(t){return new s(this.dataflow).init(this,t)},a.addAll=function(){var t=this;return this.source&&this.source.length!==this.add.length?((t=new s(this.dataflow).init(this)).add=t.source,t):t},a.init=function(t,e){return this.stamp=t.stamp,this.encode=t.encode,!t.fields||64&e||(this.fields=t.fields),1&e?(this.addF=t.addF,this.add=t.add):(this.addF=null,this.add=[]),2&e?(this.remF=t.remF,this.rem=t.rem):(this.remF=null,this.rem=[]),4&e?(this.modF=t.modF,this.mod=t.mod):(this.modF=null,this.mod=[]),32&e?(this.srcF=null,this.source=null):(this.srcF=t.srcF,this.source=t.source),this},a.runAfter=function(t){this.dataflow.runAfter(t)},a.changed=function(t){var e=t||7;return 1&e&&this.add.length||2&e&&this.rem.length||4&e&&this.mod.length},a.reflow=function(t){if(t)return this.fork(7).reflow();var e=this.add.length,n=this.source&&this.source.length;return n&&n!==e&&(this.mod=this.source,e&&this.filter(4,l(this,1))),this},a.modifies=function(t){var e=Object(i.g)(t),n=this.fields||(this.fields={});return e.forEach(function(t){n[t]=!0}),this},a.modified=function(t){var e=this.fields;return!(!this.mod.length||!e)&&(arguments.length?Object(i.t)(t)?t.some(function(t){return e[t]}):e[t]:!!e)},a.filter=function(t,e){return 1&t&&(this.addF=c(this.addF,e)),2&t&&(this.remF=c(this.remF,e)),4&t&&(this.modF=c(this.modF,e)),16&t&&(this.srcF=c(this.srcF,e)),this},a.materialize=function(t){return 1&(t=t||7)&&this.addF&&(this.add=u(this.add,this.addF),this.addF=null),2&t&&this.remF&&(this.rem=u(this.rem,this.remF),this.remF=null),4&t&&this.modF&&(this.mod=u(this.mod,this.modF),this.modF=null),16&t&&this.srcF&&(this.source=this.source.filter(this.srcF),this.srcF=null),this},a.visit=function(t,e){var n,r,o=e;return 16&t?(Object(i.T)(this.source,this.srcF,o),this):(1&t&&Object(i.T)(this.add,this.addF,o),2&t&&Object(i.T)(this.rem,this.remF,o),4&t&&Object(i.T)(this.mod,this.modF,o),8&t&&(n=this.source)&&((r=this.add.length+this.mod.length)===n.length||(r?Object(i.T)(n,l(this,5),o):Object(i.T)(n,this.srcF,o))),this)}},"../../../../vega-dataflow/src/Transform.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/src/Operator.js"),i=n("../../../../vega-util/index.js");function o(t,e){r.a.call(this,t,null,e)}var s=Object(i.s)(o,r.a);s.run=function(t){return t.stamp<=this.stamp?t.StopPropagation:(this.skip()?this.skip(!1):e=this.evaluate(t),(e=e||t)!==t.StopPropagation&&(this.pulse=e),this.stamp=t.stamp,e);var e},s.evaluate=function(t){var e=this.marshall(t.stamp),n=this.transform(e,t);return e.clear(),n},s.transform=function(){}},"../../../../vega-dataflow/src/Tuple.js":function(t,e,n){"use strict";e.c=function(t){return!(!t||!o(t))},e.f=o,e.b=a,e.a=function(t){return c(t,a({}))},e.d=c,e.e=function(t,e){return s(e,o(t))};var r=Symbol("vega_id"),i=1;function o(t){return t[r]}function s(t,e){return t[r]=e,t}function a(t){var e=t===Object(t)?t:{data:t};return o(e)?e:s(e,i++)}function c(t,e){for(var n in t)e[n]=t[n];return e}},"../../../../vega-dataflow/src/dataflow/Dataflow.js":function(t,e,n){"use strict";e.a=g;var r=n("../../../../vega-dataflow/src/dataflow/add.js"),i=n("../../../../vega-dataflow/src/dataflow/connect.js"),o=n("../../../../vega-dataflow/src/dataflow/events.js"),s=n("../../../../vega-dataflow/src/dataflow/load.js"),a=n("../../../../vega-dataflow/src/dataflow/on.js"),c=n("../../../../vega-dataflow/src/dataflow/rank.js"),u=n("../../../../vega-dataflow/src/dataflow/run.js"),l=n("../../../../vega-dataflow/src/dataflow/update.js"),p=n("../../../../vega-dataflow/src/ChangeSet.js"),d=n("../../../../vega-dataflow/src/util/Heap.js"),h=n("../../../../vega-dataflow/src/util/UniqueList.js"),f=n("../../../../vega-loader/index.js"),m=n("../../../../vega-util/index.js");function g(){this._log=Object(m.C)(),this.logLevel(m.b),this._clock=0,this._rank=0;try{this._loader=Object(f.a)()}catch(t){}this._touched=Object(h.a)(m.q),this._pulses={},this._pulse=null,this._heap=new d.a(function(t,e){return t.qrank-e.qrank}),this._postrun=[]}var v=g.prototype;function y(t){return function(){return this._log[t].apply(this,arguments)}}v.stamp=function(){return this._clock},v.loader=function(t){return arguments.length?(this._loader=t,this):this._loader},v.cleanThreshold=1e4,v.add=r.a,v.connect=i.a,v.rank=c.a,v.rerank=c.b,v.pulse=l.a,v.touch=l.b,v.update=l.c,v.changeset=p.a,v.ingest=s.a,v.request=s.b,v.events=o.a,v.on=a.a,v.run=u.c,v.runAsync=u.e,v.runAfter=u.d,v._enqueue=u.a,v._getPulse=u.b,v.error=y("error"),v.warn=y("warn"),v.info=y("info"),v.debug=y("debug"),v.logLevel=y("level")},"../../../../vega-dataflow/src/dataflow/add.js":function(t,e,n){"use strict";var r=n("../../../../vega-dataflow/src/Operator.js"),i=n("../../../../vega-util/index.js");e.a=function(t,e,n,o){var s,a=1;return t instanceof r.a?s=t:t&&t.prototype instanceof r.a?s=new t:Object(i.w)(t)?s=new r.a(null,t):(a=0,s=new r.a(t,e)),this.rank(s),a&&(o=n,n=e),n&&this.connect(s,s.parameters(n,o)),this.touch(s),s}},"../../../../vega-dataflow/src/dataflow/connect.js":function(t,e,n){"use strict";e.a=function(t,e){var n,r,i=t.rank;for(n=0,r=e.length;n<r;++n)if(i<e[n].rank)return void this.rerank(t)}},"../../../../vega-dataflow/src/dataflow/events.js":function(t,e,n){"use strict";var r=n("../../../../vega-dataflow/src/EventStream.js"),i=n("../../../../vega-util/index.js");e.a=function(t,e,n,o){for(var s,a=this,c=Object(r.b)(n,o),u=function(t){t.dataflow=a;try{c.receive(t)}catch(t){a.error(t)}finally{a.run()}},l=0,p=(s="string"==typeof t&&"undefined"!=typeof document?document.querySelectorAll(t):Object(i.g)(t)).length;l<p;++l)s[l].addEventListener(e,u);return c}},"../../../../vega-dataflow/src/dataflow/load.js":function(t,e,n){"use strict";e.a=function(t,e,n){return this.pulse(t,this.changeset().insert(Object(r.b)(e,n)))},e.b=function(t,e,n){var r=this,i=r._pending||(o=r,c=new Promise(function(t,e){s=t,a=e}),c.requests=0,c.done=function(){0==--c.requests&&o.runAfter(function(){o._pending=null;try{o.run(),s(o)}catch(t){a(t)}})},o._pending=c);var o,s,a,c;i.requests+=1,r.loader().load(e,{context:"dataflow"}).then(function(e){r.ingest(t,e,n)},function(t){r.error("Loading failed",e,t)}).catch(function(t){r.error("Data ingestion failed",e,t)}).then(i.done,i.done)};var r=n("../../../../vega-loader/index.js")},"../../../../vega-dataflow/src/dataflow/on.js":function(t,e,n){"use strict";var r=n("../../../../vega-dataflow/src/Operator.js"),i=n("../../../../vega-dataflow/src/ChangeSet.js"),o=n("../../../../vega-util/index.js"),s={skip:!0};function a(t,e,n,a,c,u){var l,p,d=Object(o.l)({},u,s);Object(o.w)(n)||(n=Object(o.i)(n)),void 0===a?l=function(e){t.touch(n(e))}:Object(o.w)(a)?(p=new r.a(null,a,c,!1),l=function(e){var r,o=n(e);p.evaluate(e),Object(i.b)(r=p.value)?t.pulse(o,r,u):t.update(o,r,d)}):l=function(e){t.update(n(e),a,d)},e.apply(l)}function c(t,e,n,i,s,a){var c,u;void 0===i?u=n:(c=Object(o.w)(i)?i:Object(o.i)(i),i=n?function(t,e){var r=c(t,e);return n.skip()?r:n.skip(!0).value=r}:c,(u=new r.a(null,i,s,!1)).modified(a&&a.force),u.rank=0,n&&(u.skip(!0),u.value=n.value,u.targets().add(n))),e.targets().add(u)}e.a=function(t,e,n,i,o){return(t instanceof r.a?c:a)(this,t,e,n,i,o),this}},"../../../../vega-dataflow/src/dataflow/rank.js":function(t,e,n){"use strict";e.a=function(t){t.rank=++this._rank},e.b=function(t){var e,n,i,o=[t];for(;o.length;)if(this.rank(e=o.pop()),n=e._targets)for(i=n.length;--i>=0;)o.push(e=n[i]),e===t&&Object(r.k)("Cycle detected in dataflow graph.")};var r=n("../../../../vega-util/index.js")},"../../../../vega-dataflow/src/dataflow/run.js":function(t,e,n){"use strict";e.c=function(t){var e,n,i,c,u=this,l=0,p=u.logLevel();if(u._pending)return u.info("Awaiting requests, delaying dataflow run."),0;if(u._pulse)return u.error("Dataflow invoked recursively. Use the runAfter method to queue invocation."),0;if(!u._touched.length)return u.info("Dataflow invoked, but nothing to do."),0;u._pulse=new r.b(u,++u._clock,t),p>=s.c&&(i=Date.now(),u.debug("-- START PROPAGATION ("+u._clock+") -----"));u._touched.forEach(function(t){u._enqueue(t,!0)}),u._touched=Object(o.a)(s.q);try{for(;u._heap.size()>0;)(e=u._heap.pop()).rank===e.qrank?(n=e.run(u._getPulse(e,t)),p>=s.a&&u.debug(e.id,n===r.a?"STOP":n,e),n!==r.a&&(u._pulse=n,e._targets&&e._targets.forEach(function(t){u._enqueue(t)})),++l):u._enqueue(e,!0)}catch(t){c=t}u._pulses={},u._pulse=null,p>=s.c&&(i=Date.now()-i,u.info("> Pulse "+u._clock+": "+l+" operators; "+i+"ms"));c&&(u._postrun=[],u.error(c));if(u._onrun)try{u._onrun(u,l,c)}catch(t){u.error(t)}if(u._postrun.length){var d=u._postrun;u._postrun=[],d.sort(function(t,e){return e.priority-t.priority}).forEach(function(t){a(u,t.callback)})}return l},e.e=function(){return this._pending||Promise.resolve(this.run())},e.d=function(t,e,n){this._pulse||e?this._postrun.push({priority:n||0,callback:t}):a(this,t)},e.a=function(t,e){var n=!this._pulses[t.id];n&&(this._pulses[t.id]=this._pulse);(n||e)&&(t.qrank=t.rank,this._heap.push(t))},e.b=function(t,e){var n,o=t.source,a=this._clock;if(o&&Object(s.t)(o))return n=o.map(function(t){return t.pulse}),new i.a(this,a,n,e);n=this._pulses[t.id],o&&((o=o.pulse)&&o!==r.a?o.stamp===a&&n.target!==t?n=o:n.source=o.source:n.source=[]);return n};var r=n("../../../../vega-dataflow/src/Pulse.js"),i=n("../../../../vega-dataflow/src/MultiPulse.js"),o=n("../../../../vega-dataflow/src/util/UniqueList.js"),s=n("../../../../vega-util/index.js");function a(t,e){try{e(t)}catch(e){t.error(e)}}},"../../../../vega-dataflow/src/dataflow/update.js":function(t,e,n){"use strict";e.b=function(t,e){var n=e||i;this._pulse?this._enqueue(t):this._touched.add(t);n.skip&&t.skip(!0);return this},e.c=function(t,e,n){var r=n||i;(t.set(e)||r.force)&&this.touch(t,r);return this},e.a=function(t,e,n){this.touch(t,n||i);var o=new r.b(this,this._clock+(this._pulse?0:1)),s=t.pulse&&t.pulse.source||[];return o.target=t,this._pulses[t.id]=e.pulse(o,s),this};var r=n("../../../../vega-dataflow/src/Pulse.js"),i={skip:!1,force:!1}},"../../../../vega-dataflow/src/register.js":function(t,e,n){"use strict";n.d(e,"b",function(){return r}),e.a=function(t){var e=(n=t,n=n&&n.toLowerCase(),r.hasOwnProperty(n)?r[n]:null);var n;return e&&e.Definition||null};var r={}},"../../../../vega-dataflow/src/util/Heap.js":function(t,e,n){"use strict";function r(t){this.cmp=t,this.nodes=[]}e.a=r;var i=r.prototype;function o(t,e,n,r){var i,o,s;for(i=t[n];n>e&&r(i,o=t[s=n-1>>1])<0;)t[n]=o,n=s;return t[n]=i}function s(t,e,n){for(var r,i=e,s=t.length,a=t[e],c=2*e+1;c<s;)(r=c+1)<s&&n(t[c],t[r])>=0&&(c=r),t[e]=t[c],c=2*(e=c)+1;return t[e]=a,o(t,i,e,n)}i.size=function(){return this.nodes.length},i.clear=function(){return this.nodes=[],this},i.peek=function(){return this.nodes[0]},i.push=function(t){var e=this.nodes;return e.push(t),o(e,0,e.length-1,this.cmp)},i.pop=function(){var t,e=this.nodes,n=e.pop();return e.length?(t=e[0],e[0]=n,s(e,0,this.cmp)):t=n,t},i.replace=function(t){var e=this.nodes,n=e[0];return e[0]=t,s(e,0,this.cmp),n},i.pushpop=function(t){var e=this.nodes,n=e[0];return e.length&&this.cmp(n,t)<0&&(e[0]=t,t=n,s(e,0,this.cmp)),t}},"../../../../vega-dataflow/src/util/UniqueList.js":function(t,e,n){"use strict";e.a=function(t){var e=t||r.r,n=[],i={};return n.add=function(t){var r=e(t);return i[r]||(i[r]=1,n.push(t)),n},n.remove=function(t){var r,o=e(t);return i[o]&&(i[o]=0,(r=n.indexOf(t))>=0&&n.splice(r,1)),n},n};var r=n("../../../../vega-util/index.js")},"../../../../vega-encode/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../vega-encode/src/AxisTicks.js");n.d(e,"axisticks",function(){return r.a});var i=n("../../../../vega-encode/src/DataJoin.js");n.d(e,"datajoin",function(){return i.a});var o=n("../../../../vega-encode/src/Encode.js");n.d(e,"encode",function(){return o.a});var s=n("../../../../vega-encode/src/LegendEntries.js");n.d(e,"legendentries",function(){return s.a});var a=n("../../../../vega-encode/src/LinkPath.js");n.d(e,"linkpath",function(){return a.a});var c=n("../../../../vega-encode/src/Pie.js");n.d(e,"pie",function(){return c.a});var u=n("../../../../vega-encode/src/Scale.js");n.d(e,"scale",function(){return u.a});var l=n("../../../../vega-encode/src/SortItems.js");n.d(e,"sortitems",function(){return l.a});var p=n("../../../../vega-encode/src/Stack.js");n.d(e,"stack",function(){return p.a});var d=n("../../../../vega-encode/src/ticks.js");n.d(e,"validTicks",function(){return d.d})},"../../../../vega-encode/src/AxisTicks.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-encode/src/ticks.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js");function s(t){i.d.call(this,null,t)}Object(o.s)(s,i.d).transform=function(t,e){if(this.value&&!t.modified())return e.StopPropagation;var n=e.fork(e.NO_SOURCE|e.NO_FIELDS),o=this.value,s=t.scale,a=null==t.count?t.values?t.values.length:10:Object(r.a)(s,t.count),c=t.format||Object(r.b)(s,a,t.formatSpecifier),u=t.values?Object(r.d)(s,t.values,a):Object(r.c)(s,a);return o&&(n.rem=o),o=u.map(function(t,e){return Object(i.h)({index:e/(u.length-1),value:t,label:c(t)})}),t.extra&&o.push(Object(i.h)({index:-1,extra:{value:o[0].value},label:""})),n.source=o,n.add=o,this.value=o,n}},"../../../../vega-encode/src/DataJoin.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}function s(){return Object(r.h)({})}function a(t){return t.exit}Object(i.s)(o,r.d).transform=function(t,e){var n=e.dataflow,o=e.fork(e.NO_SOURCE|e.NO_FIELDS),c=t.item||s,u=t.key||r.n,l=this.value;return Object(i.t)(o.encode)&&(o.encode=null),l&&(t.modified("key")||e.modified(u))&&Object(i.k)("DataJoin does not support modified key function or fields."),l||(e=e.addAll(),this.value=l=Object(i.o)().test(a),l.lookup=function(t){return l.get(u(t))}),e.visit(e.ADD,function(t){var e=u(t),n=l.get(e);n?n.exit?(l.empty--,o.add.push(n)):o.mod.push(n):(l.set(e,n=c(t)),o.add.push(n)),n.datum=t,n.exit=!1}),e.visit(e.MOD,function(t){var e=u(t),n=l.get(e);n&&(n.datum=t,o.mod.push(n))}),e.visit(e.REM,function(t){var e=u(t),n=l.get(e);t!==n.datum||n.exit||(o.rem.push(n),n.exit=!0,++l.empty)}),e.changed(e.ADD_MOD)&&o.modifies("datum"),t.clean&&l.empty>n.cleanThreshold&&n.runAfter(l.clean),o}},"../../../../vega-encode/src/Encode.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}Object(i.s)(o,r.d).transform=function(t,e){var n=e.fork(e.ADD_REM),r=t.encoders,o=e.encode;if(Object(i.t)(o)){if(!n.changed()&&!o.every(function(t){return r[t]}))return e.StopPropagation;o=o[0]}var s="enter"===o,a=r.update||i.n,c=r.enter||i.n,u=r.exit||i.n,l=(o&&!s?r[o]:a)||i.n;if(e.changed(e.ADD)&&(e.visit(e.ADD,function(e){c(e,t),a(e,t),l!==i.n&&l!==a&&l(e,t)}),n.modifies(c.output),n.modifies(a.output),l!==i.n&&l!==a&&n.modifies(l.output)),e.changed(e.REM)&&u!==i.n&&(e.visit(e.REM,function(e){u(e,t)}),n.modifies(u.output)),s||l!==i.n){var p=e.MOD|(t.modified()?e.REFLOW:0);s?(e.visit(p,function(e){var r=c(e,t);(l(e,t)||r)&&n.mod.push(e)}),n.mod.length&&n.modifies(c.output)):e.visit(p,function(e){l(e,t)&&n.mod.push(e)}),n.mod.length&&n.modifies(l.output)}return n.changed()?n:e.StopPropagation}},"../../../../vega-encode/src/LegendEntries.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../vega-encode/src/labels.js"),i=n("../../../../vega-encode/src/ticks.js"),o=n("../../../../vega-dataflow/index.js"),s=n("../../../../vega-scale/index.js"),a=n("../../../../vega-util/index.js");function c(t){o.d.call(this,[],t)}Object(a.s)(c,o.d).transform=function(t,e){if(null!=this.value&&!t.modified())return e.StopPropagation;var n=e.fork(e.NO_SOURCE|e.NO_FIELDS),c=0,u=this.value,l="gradient"===t.type,p=t.scale,d=null==t.count?5:Object(i.a)(p,t.count),h=t.format||Object(i.b)(p,d,t.formatSpecifier),f=t.values||Object(r.b)(p,d,l);if(h=Object(r.a)(p,h),u&&(n.rem=u),l)var m=t.values?p.domain():f,g=Object(s.e)(p,m[0],Object(a.J)(m));else{var v,y=t.size;Object(a.w)(y)?(t.values||0!==p(f[0])||(f=f.slice(1)),v=f.reduce(function(e,n){return Math.max(e,y(n,t))},0)):y=Object(a.i)(v=y||8)}return u=f.map(function(e,n){var r=Object(o.h)({index:n,label:h(e,n,f),value:e});return l?r.perc=g(e):(r.offset=v,r.size=y(e,t),r.total=Math.round(c),c+=r.size),r}),n.source=u,n.add=u,this.value=u,n}},"../../../../vega-encode/src/LinkPath.js":function(t,e,n){"use strict";e.a=l;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=Object(i.o)({line:p,"line-radial":function(t,e,n,r){return p(e*Math.cos(t),e*Math.sin(t),r*Math.cos(n),r*Math.sin(n))},arc:d,"arc-radial":function(t,e,n,r){return d(e*Math.cos(t),e*Math.sin(t),r*Math.cos(n),r*Math.sin(n))},curve:h,"curve-radial":function(t,e,n,r){return h(e*Math.cos(t),e*Math.sin(t),r*Math.cos(n),r*Math.sin(n))},"orthogonal-horizontal":function(t,e,n,r){return"M"+t+","+e+"V"+r+"H"+n},"orthogonal-vertical":function(t,e,n,r){return"M"+t+","+e+"H"+n+"V"+r},"orthogonal-radial":function(t,e,n,r){var i=Math.cos(t),o=Math.sin(t),s=Math.cos(n),a=Math.sin(n),c=Math.abs(n-t)>Math.PI?n<=t:n>t;return"M"+e*i+","+e*o+"A"+e+","+e+" 0 0,"+(c?1:0)+" "+e*s+","+e*a+"L"+r*s+","+r*a},"diagonal-horizontal":function(t,e,n,r){var i=(t+n)/2;return"M"+t+","+e+"C"+i+","+e+" "+i+","+r+" "+n+","+r},"diagonal-vertical":function(t,e,n,r){var i=(e+r)/2;return"M"+t+","+e+"C"+t+","+i+" "+n+","+i+" "+n+","+r},"diagonal-radial":function(t,e,n,r){var i=Math.cos(t),o=Math.sin(t),s=Math.cos(n),a=Math.sin(n),c=(e+r)/2;return"M"+e*i+","+e*o+"C"+c*i+","+c*o+" "+c*s+","+c*a+" "+r*s+","+r*a}});function s(t){return t.source.x}function a(t){return t.source.y}function c(t){return t.target.x}function u(t){return t.target.y}function l(t){r.d.call(this,{},t)}function p(t,e,n,r){return"M"+t+","+e+"L"+n+","+r}function d(t,e,n,r){var i=n-t,o=r-e,s=Math.sqrt(i*i+o*o)/2;return"M"+t+","+e+"A"+s+","+s+" "+180*Math.atan2(o,i)/Math.PI+" 0 1 "+n+","+r}function h(t,e,n,r){var i=n-t,o=r-e,s=.2*(i+o),a=.2*(o-i);return"M"+t+","+e+"C"+(t+s)+","+(e+a)+" "+(n+a)+","+(r-s)+" "+n+","+r}l.Definition={type:"LinkPath",metadata:{modifies:!0},params:[{name:"sourceX",type:"field",default:"source.x"},{name:"sourceY",type:"field",default:"source.y"},{name:"targetX",type:"field",default:"target.x"},{name:"targetY",type:"field",default:"target.y"},{name:"orient",type:"enum",default:"vertical",values:["horizontal","vertical","radial"]},{name:"shape",type:"enum",default:"line",values:["line","arc","curve","diagonal","orthogonal"]},{name:"as",type:"string",default:"path"}]},Object(i.s)(l,r.d).transform=function(t,e){var n=t.sourceX||s,r=t.sourceY||a,l=t.targetX||c,p=t.targetY||u,d=t.as||"path",h=t.orient||"vertical",f=t.shape||"line",m=o.get(f+"-"+h)||o.get(f);return m||Object(i.k)("LinkPath unsupported type: "+t.shape+(t.orient?"-"+t.orient:"")),e.visit(e.SOURCE,function(t){t[d]=m(n(t),r(t),l(t),p(t))}),e.reflow(t.modified()).modifies(d)}},"../../../../vega-encode/src/Pie.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-array/index.js");function s(t){r.d.call(this,null,t)}s.Definition={type:"Pie",metadata:{modifies:!0},params:[{name:"field",type:"field"},{name:"startAngle",type:"number",default:0},{name:"endAngle",type:"number",default:6.283185307179586},{name:"sort",type:"boolean",default:!1},{name:"as",type:"string",array:!0,length:2,default:["startAngle","endAngle"]}]},Object(i.s)(s,r.d).transform=function(t,e){var n,r,s,a=t.as||["startAngle","endAngle"],c=a[0],u=a[1],l=t.field||i.E,p=t.startAngle||0,d=null!=t.endAngle?t.endAngle:2*Math.PI,h=e.source,f=h.map(l),m=f.length,g=p,v=(d-p)/Object(o.o)(f),y=Object(o.n)(m);for(t.sort&&y.sort(function(t,e){return f[t]-f[e]}),n=0;n<m;++n)s=f[y[n]],(r=h[y[n]])[c]=g,r[u]=g+=s*v;return this.value=f,e.reflow(t.modified()).modifies(a)}},"../../../../vega-encode/src/Scale.js":function(t,e,n){"use strict";e.a=h;var r=n("../../../../vega-encode/src/ticks.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js"),s=n("../../../../vega-encode/src/scale-types.js"),a=n("../../../../vega-scale/index.js"),c=n("../../../../d3-interpolate/index.js"),u=5,l=Object(o.P)([s.d,s.h,s.l]),p=Object(o.P)([s.d,s.e,s.h,s.l,s.n,s.o]),d=Object(o.P)(["set","modified","clear","type","scheme","schemeExtent","schemeCount","domain","domainMin","domainMid","domainMax","domainRaw","nice","zero","range","rangeStep","round","reverse","interpolate","interpolateGamma"]);function h(t){i.d.call(this,null,t),this.modified(!0)}function f(t,e,n){return Object(o.w)(t)&&(e||n)?Object(a.c)(t,m(e||[0,1],n)):t}function m(t,e){return e?t.slice().reverse():t}Object(o.s)(h,i.d).transform=function(t,e){var n,i=e.dataflow,h=this.value;h&&!t.modified("type")||(this.value=h=Object(a.d)((t.type||s.d).toLowerCase())());for(n in t)if(!d[n]){if("padding"===n&&p[h.type])continue;Object(o.w)(h[n])?h[n](t[n]):i.warn("Unsupported scale property: "+n)}return function(t,e,n){var r=e.round||!1,i=e.range;if(null!=e.rangeStep)i=function(t,e,n){t!==s.a&&t!==s.g&&Object(o.k)("Only band and point scales support rangeStep.");var r=(null!=e.paddingOuter?e.paddingOuter:e.padding)||0,i=t===s.g?1:(null!=e.paddingInner?e.paddingInner:e.padding)||0;return[0,e.rangeStep*Object(a.a)(n,i,r)]}(t.type,e,n);else if(e.scheme){if(i=function(t,e,n){var r,i=e.scheme.toLowerCase(),c=Object(a.f)(i),l=e.schemeExtent;c||Object(o.k)("Unrecognized scheme name: "+e.scheme);return n=t===s.m?n+1:t===s.c?n-1:t===s.i||t===s.j?+e.schemeCount||u:n,t===s.k?f(c,l,e.reverse):!l&&(r=Object(a.f)(i+"-"+n))?r:Object(o.w)(c)?function(t,e){for(var n=new Array(e),r=e-1||1,i=0;i<e;++i)n[i]=t(i/r);return n}(f(c,l),n):t===s.f?c:c.slice(0,n)}(t.type,e,n),Object(o.w)(i))return t.interpolator(i)}else if(i&&t.type===s.k)return t.interpolator(Object(c.interpolateRgbBasis)(m(i,e.reverse)));i&&e.interpolate&&t.interpolate?t.interpolate(Object(a.b)(e.interpolate,e.interpolateGamma)):Object(o.w)(t.round)?t.round(r):Object(o.w)(t.rangeRound)&&t.interpolate(r?c.interpolateRound:c.interpolate);i&&t.range(m(i,e.reverse))}(h,t,function(t,e,n){var i=(a=t,c=e.domainRaw,c?(a.domain(c),c.length):-1);var a,c;if(i>-1)return i;var u,d,h=e.domain,f=t.type,m=e.zero||void 0===e.zero&&l[f];if(!h)return 0;p[f]&&e.padding&&h[0]!==Object(o.J)(h)&&(g=f,v=h,y=e.range,b=e.padding,_=e.exponent,j=Math.abs(Object(o.J)(y)-y[0]),x=j/(j-2*b),w=g===s.e?Object(o.W)(v,null,x):g===s.l?Object(o.X)(v,null,x,.5):g===s.h?Object(o.X)(v,null,x,_):Object(o.V)(v,null,x),(v=v.slice())[0]=w[0],v[v.length-1]=w[1],h=v);var g,v,y,b,_,j,x,w;(m||null!=e.domainMin||null!=e.domainMax||null!=e.domainMid)&&(u=(h=h.slice()).length-1||1,m&&(h[0]>0&&(h[0]=0),h[u]<0&&(h[u]=0)),null!=e.domainMin&&(h[0]=e.domainMin),null!=e.domainMax&&(h[u]=e.domainMax),null!=e.domainMid&&(((d=e.domainMid)<h[0]||d>h[u])&&n.warn("Scale domainMid exceeds domain min or max.",d),h.splice(u,0,d)));t.domain(h),e.nice&&t.nice&&t.nice(!0!==e.nice&&Object(r.a)(t,e.nice)||null);return h.length}(h,t,i)),e.fork(e.NO_SOURCE|e.NO_FIELDS)}},"../../../../vega-encode/src/SortItems.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}Object(i.s)(o,r.d).transform=function(t,e){var n=t.modified("sort")||e.changed(e.ADD)||e.modified(t.sort.fields)||e.modified("datum");return n&&e.source.sort(t.sort),this.modified(n),e}},"../../../../vega-encode/src/Stack.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}function s(t,e,n,r,i){for(var o,s=(e-t.sum)/2,a=t.length,c=0;c<a;++c)(o=t[c])[r]=s,o[i]=s+=Math.abs(n(o))}function a(t,e,n,r,i){for(var o,s=1/t.sum,a=0,c=t.length,u=0,l=0;u<c;++u)(o=t[u])[r]=a,o[i]=a=s*(l+=Math.abs(n(o)))}function c(t,e,n,r,i){for(var o,s,a=0,c=0,u=t.length,l=0;l<u;++l)(o=n(s=t[l]))<0?(s[r]=c,s[i]=c+=o):(s[r]=a,s[i]=a+=o)}o.Definition={type:"Stack",metadata:{modifies:!0},params:[{name:"field",type:"field"},{name:"groupby",type:"field",array:!0},{name:"sort",type:"compare"},{name:"offset",type:"enum",default:"zero",values:["zero","center","normalize"]},{name:"as",type:"string",array:!0,length:2,default:["y0","y1"]}]},Object(i.s)(o,r.d).transform=function(t,e){var n,r,o,u,l=t.as||["y0","y1"],p=l[0],d=l[1],h=t.field||i.E,f="center"===t.offset?s:"normalize"===t.offset?a:c;for(r=0,o=(n=function(t,e,n,r){var i,o,s,a,c,u,l,p,d,h=[],f=function(t){return t(c)};if(null==e)h.push(t.slice());else for(i={},o=0,s=t.length;o<s;++o)c=t[o],u=e.map(f),(l=i[u])||(i[u]=l=[],h.push(l)),l.push(c);for(u=0,d=0,a=h.length;u<a;++u){for(l=h[u],o=0,p=0,s=l.length;o<s;++o)p+=Math.abs(r(l[o]));l.sum=p,p>d&&(d=p),n&&l.sort(n)}return h.max=d,h}(e.source,t.groupby,t.sort,h)).length,u=n.max;r<o;++r)f(n[r],u,h,p,d);return e.reflow(t.modified()).modifies(l)}},"../../../../vega-encode/src/labels.js":function(t,e,n){"use strict";e.b=function(t,e,n){if(n)return t.domain();var r=s[t.type];return r?r(t):Object(i.c)(t,e)},e.a=function(t,e){return s[t.type]?(r=e,function(t,e,n){var i=n[e+1]||n.max||1/0,o=c(t,r),s=c(i,r);return o&&s?o+"–"+s:s?"< "+s:"≥ "+o}):(n=e,function(t){return n(t)});var n;var r};var r=n("../../../../vega-encode/src/scale-types.js"),i=n("../../../../vega-encode/src/ticks.js"),o=n("../../../../vega-util/index.js"),s={};function a(t){var e=t.domain();return e.max=e.pop(),e}function c(t,e){return isFinite(t)?e(t):null}s[r.i]=function(t){var e=[-1/0].concat(t.quantiles());return e.max=1/0,e},s[r.j]=function(t){var e=t.domain(),n=e[0],r=Object(o.J)(e),i=t.range().length,s=new Array(i),a=0;s[0]=-1/0;for(;++a<i;)s[a]=(a*r-(a-i)*n)/i;return s.max=1/0,s},s[r.m]=function(t){var e=[-1/0].concat(t.domain());return e.max=1/0,e},s[r.b]=a,s[r.c]=a},"../../../../vega-encode/src/scale-types.js":function(t,e,n){"use strict";n.d(e,"e",function(){return r}),n.d(e,"h",function(){return i}),n.d(e,"o",function(){return o}),n.d(e,"l",function(){return s}),n.d(e,"a",function(){return a}),n.d(e,"n",function(){return c}),n.d(e,"g",function(){return u}),n.d(e,"d",function(){return l}),n.d(e,"f",function(){return p}),n.d(e,"i",function(){return d}),n.d(e,"j",function(){return h}),n.d(e,"m",function(){return f}),n.d(e,"b",function(){return m}),n.d(e,"c",function(){return g}),n.d(e,"k",function(){return v});var r="log",i="pow",o="utc",s="sqrt",a="band",c="time",u="point",l="linear",p="ordinal",d="quantile",h="quantize",f="threshold",m="bin-linear",g="bin-ordinal",v="sequential"},"../../../../vega-encode/src/ticks.js":function(t,e,n){"use strict";e.a=function(t,e){var n;Object(o.y)(e)&&(n=e.step,e=e.interval);Object(o.A)(e)&&(e="time"===t.type?Object(i.g)(e):"utc"===t.type?Object(i.h)(e):Object(o.k)("Only time and utc scales accept interval strings."),n&&(e=e.every(n)));return e},e.d=function(t,e,n){var r=t.range(),i=r[0],s=Object(o.J)(r);i>s&&(r=s,s=i,i=r);if(e=e.filter(function(e){return!((e=t(e))<i||e>s)}),n>0&&e.length>1){for(var a=[e[0],Object(o.J)(e)];e.length>n&&e.length>=3;)e=e.filter(function(t,e){return!(e%2)});e.length<3&&(e=a)}return e},e.c=function(t,e){return t.ticks?t.ticks(e):t.domain()},e.b=function(t,e,n){var i=t.tickFormat?t.tickFormat(e,n):String;return t.type===r.e?(o=i,a=function(t){var e=Object(s.c)(t||",");if(null==e.precision){switch(e.precision=12,e.type){case"%":e.precision-=2;break;case"e":e.precision-=1}return n=Object(s.a)(e),r=Object(s.a)(".1f")(1)[1],function(t){var e,i,o=n(t),s=o.indexOf(r);if(s<0)return o;for(e=function(t,e){var n,r=t.lastIndexOf("e");if(r>0)return r;for(r=t.length;--r>e;)if((n=t.charCodeAt(r))>=48&&n<=57)return r+1}(o,s),i=e<o.length?o.slice(e):"";--e>s;)if("0"!==o[e]){++e;break}return o.slice(0,e)+i}}return Object(s.a)(e);var n,r}(n),function(t){return o(t)?a(t):""}):i;var o,a};var r=n("../../../../vega-encode/src/scale-types.js"),i=n("../../../../vega-scale/index.js"),o=n("../../../../vega-util/index.js"),s=n("../../../../d3-format/index.js")},"../../../../vega-event-selector/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-event-selector/src/event-selector.js");n.d(e,"a",function(){return r.a})},"../../../../vega-event-selector/src/event-selector.js":function(t,e,n){"use strict";e.a=function(t,e,n){return r=e||o,i=n||m,v(t.trim()).map(y)};var r,i,o="view",s="[",a="]",c="{",u="}",l=":",p=",",d="@",h=">",f=/[[\]{}]/,m={"*":1,arc:1,area:1,group:1,image:1,line:1,path:1,rect:1,rule:1,shape:1,symbol:1,text:1,trail:1};function g(t,e,n,r,i){for(var o,s=0,a=t.length;e<a;++e){if(o=t[e],!s&&o===n)return e;i&&i.indexOf(o)>=0?--s:r&&r.indexOf(o)>=0&&++s}return e}function v(t){for(var e=[],n=0,r=t.length,i=0;i<r;)i=g(t,i,p,s+c,a+u),e.push(t.substring(n,i).trim()),n=++i;if(0===e.length)throw"Empty event selector: "+t;return e}function y(t){return"["===t[0]?function(t){var e,n,r=t.length,i=1;if((i=g(t,i,a,s,a))===r)throw"Empty between selector: "+t;if(2!==(e=v(t.substring(1,i))).length)throw"Between selector must have two elements: "+t;if((t=t.slice(i+1).trim())[0]!==h)throw"Expected '>' after between selector: "+t;if(e=e.map(y),(n=y(t.slice(1).trim())).between)return{between:e,stream:n};n.between=e;return n}(t):function(t){var e,n,o={source:r},h=[],m=[0,0],v=0,y=0,b=t.length,_=0;if(t[b-1]===u){if(!((_=t.lastIndexOf(c))>=0))throw"Unmatched right brace: "+t;try{m=function(t){var e=t.split(p);if(!t.length||e.length>2)throw t;return e.map(function(e){var n=+e;if(n!=n)throw t;return n})}(t.substring(_+1,b-1))}catch(e){throw"Invalid throttle specification: "+t}t=t.slice(0,_).trim(),b=t.length,_=0}if(!b)throw t;t[0]===d&&(v=++_);(e=g(t,_,l))<b&&(h.push(t.substring(y,e).trim()),y=_=++e);if((_=g(t,_,s))===b)h.push(t.substring(y,b).trim());else if(h.push(t.substring(y,_).trim()),n=[],(y=++_)===b)throw"Unmatched left bracket: "+t;for(;_<b;){if((_=g(t,_,a))===b)throw"Unmatched left bracket: "+t;if(n.push(t.substring(y,_).trim()),_<b-1&&t[++_]!==s)throw"Expected left bracket: "+t;y=++_}if(!(b=h.length)||f.test(h[b-1]))throw"Invalid event selector: "+t;b>1?(o.type=h[1],v?o.markname=h[0].slice(1):(j=h[0],i.hasOwnProperty(j)?o.marktype=h[0]:o.source=h[0])):o.type=h[0];var j;"!"===o.type.slice(-1)&&(o.consume=!0,o.type=o.type.slice(0,-1));null!=n&&(o.filter=n);m[0]&&(o.throttle=m[0]);m[1]&&(o.debounce=m[1]);return o}(t)}},"../../../../vega-expression/index.js":function(t,e,n){"use strict";n("../../../../vega-expression/src/ast.js");var r=n("../../../../vega-expression/src/parser.js");n.d(e,"d",function(){return r.a});var i=n("../../../../vega-expression/src/codegen.js");n.d(e,"a",function(){return i.a});var o=n("../../../../vega-expression/src/functions.js");n.d(e,"c",function(){return o.a});var s=n("../../../../vega-expression/src/constants.js");n.d(e,"b",function(){return s.a})},"../../../../vega-expression/src/ast.js":function(t,e,n){"use strict";function r(t){this.type=t}e.a=r,r.prototype.visit=function(t){var e,n,r;if(t(this))return 1;for(n=0,r=(e=function(t){switch(t.type){case"ArrayExpression":return t.elements;case"BinaryExpression":case"LogicalExpression":return[t.left,t.right];case"CallExpression":var e=t.arguments.slice();return e.unshift(t.callee),e;case"ConditionalExpression":return[t.test,t.consequent,t.alternate];case"MemberExpression":return[t.object,t.property];case"ObjectExpression":return t.properties;case"Property":return[t.key,t.value];case"UnaryExpression":return[t.argument];case"Identifier":case"Literal":case"RawCode":default:return[]}}(this)).length;n<r;++n)if(e[n].visit(t))return 1}},"../../../../vega-expression/src/codegen.js":function(t,e,n){"use strict";var r=n("../../../../vega-expression/src/constants.js"),i=n("../../../../vega-expression/src/functions.js"),o=n("../../../../vega-util/index.js");e.a=function(t){var e=(t=t||{}).whitelist?Object(o.P)(t.whitelist):{},n=t.blacklist?Object(o.P)(t.blacklist):{},s=t.constants||r.a,a=(t.functions||i.a)(f),c=t.globalvar,u=t.fieldvar,l={},p={},d=0,h=Object(o.w)(c)?c:function(t){return c+'["'+t+'"]'};function f(t){if(Object(o.A)(t))return t;var e=m[t.type];return null==e&&Object(o.k)("Unsupported type: "+t.type),e(t)}var m={Literal:function(t){return t.raw},Identifier:function(t){var r=t.name;return d>0?r:n.hasOwnProperty(r)?Object(o.k)("Illegal identifier: "+r):s.hasOwnProperty(r)?s[r]:e.hasOwnProperty(r)?r:(l[r]=1,h(r))},MemberExpression:function(t){var e=!t.computed,n=f(t.object);e&&(d+=1);var r=f(t.property);return n===u&&(p[r]=1),e&&(d-=1),n+(e?"."+r:"["+r+"]")},CallExpression:function(t){"Identifier"!==t.callee.type&&Object(o.k)("Illegal callee type: "+t.callee.type);var e=t.callee.name,n=t.arguments,r=a.hasOwnProperty(e)&&a[e];return r||Object(o.k)("Unrecognized function: "+e),Object(o.w)(r)?r(n):r+"("+n.map(f).join(",")+")"},ArrayExpression:function(t){return"["+t.elements.map(f).join(",")+"]"},BinaryExpression:function(t){return"("+f(t.left)+t.operator+f(t.right)+")"},UnaryExpression:function(t){return"("+t.operator+f(t.argument)+")"},ConditionalExpression:function(t){return"("+f(t.test)+"?"+f(t.consequent)+":"+f(t.alternate)+")"},LogicalExpression:function(t){return"("+f(t.left)+t.operator+f(t.right)+")"},ObjectExpression:function(t){return"{"+t.properties.map(f).join(",")+"}"},Property:function(t){d+=1;var e=f(t.key);return d-=1,e+":"+f(t.value)}};function g(t){var e={code:f(t),globals:Object.keys(l),fields:Object.keys(p)};return l={},p={},e}return g.functions=a,g.constants=s,g}},"../../../../vega-expression/src/constants.js":function(t,e,n){"use strict";e.a={NaN:"NaN",E:"Math.E",LN2:"Math.LN2",LN10:"Math.LN10",LOG2E:"Math.LOG2E",LOG10E:"Math.LOG10E",PI:"Math.PI",SQRT1_2:"Math.SQRT1_2",SQRT2:"Math.SQRT2",MIN_VALUE:"Number.MIN_VALUE",MAX_VALUE:"Number.MAX_VALUE"}},"../../../../vega-expression/src/functions.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");e.a=function(t){function e(e,n,r){return function(i){return o=e,a=n,c=r,u=t((s=i)[0]),a&&(u=a+"("+u+")",0===a.lastIndexOf("new ",0)&&(u="("+u+")")),u+"."+o+(c<0?"":0===c?"()":"("+s.slice(1).map(t).join(",")+")");var o,s,a,c,u}}var n="new Date";return{isNaN:"isNaN",isFinite:"isFinite",abs:"Math.abs",acos:"Math.acos",asin:"Math.asin",atan:"Math.atan",atan2:"Math.atan2",ceil:"Math.ceil",cos:"Math.cos",exp:"Math.exp",floor:"Math.floor",log:"Math.log",max:"Math.max",min:"Math.min",pow:"Math.pow",random:"Math.random",round:"Math.round",sin:"Math.sin",sqrt:"Math.sqrt",tan:"Math.tan",clamp:function(e){e.length<3&&Object(r.k)("Missing arguments to clamp function."),e.length>3&&Object(r.k)("Too many arguments to clamp function.");var n=e.map(t);return"Math.max("+n[1]+", Math.min("+n[2]+","+n[0]+"))"},now:"Date.now",utc:"Date.UTC",datetime:n,date:e("getDate",n,0),day:e("getDay",n,0),year:e("getFullYear",n,0),month:e("getMonth",n,0),hours:e("getHours",n,0),minutes:e("getMinutes",n,0),seconds:e("getSeconds",n,0),milliseconds:e("getMilliseconds",n,0),time:e("getTime",n,0),timezoneoffset:e("getTimezoneOffset",n,0),utcdate:e("getUTCDate",n,0),utcday:e("getUTCDay",n,0),utcyear:e("getUTCFullYear",n,0),utcmonth:e("getUTCMonth",n,0),utchours:e("getUTCHours",n,0),utcminutes:e("getUTCMinutes",n,0),utcseconds:e("getUTCSeconds",n,0),utcmilliseconds:e("getUTCMilliseconds",n,0),length:e("length",null,-1),indexof:e("indexOf",null),lastindexof:e("lastIndexOf",null),slice:e("slice",null),parseFloat:"parseFloat",parseInt:"parseInt",upper:e("toUpperCase","String",0),lower:e("toLowerCase","String",0),substring:e("substring","String"),replace:e("replace","String"),regexp:"RegExp",test:e("test","RegExp"),if:function(e){e.length<3&&Object(r.k)("Missing arguments to if function."),e.length>3&&Object(r.k)("Too many arguments to if function.");var n=e.map(t);return"("+n[0]+"?"+n[1]+":"+n[2]+")"}}}},"../../../../vega-expression/src/parser.js":function(t,e,n){"use strict";var r,i,o,s,a,c=n("../../../../vega-expression/src/ast.js"),u=1,l=2,p=3,d=4,h=5,f=6,m=7,g=8;(r={})[u]="Boolean",r[l]="<end>",r[p]="Identifier",r[d]="Keyword",r[h]="Null",r[f]="Numeric",r[m]="Punctuator",r[g]="String",r[9]="RegularExpression";var v="ArrayExpression",y="BinaryExpression",b="CallExpression",_="ConditionalExpression",j="Identifier",x="Literal",w="LogicalExpression",O="MemberExpression",C="ObjectExpression",k="Property",E="UnaryExpression",S="Unexpected token %0",P="Unexpected number",M="Unexpected string",A="Unexpected identifier",T="Unexpected reserved word",D="Unexpected end of input",I="Invalid regular expression",R="Invalid regular expression: missing /",N="Octal literals are not allowed in strict mode.",L="Duplicate data property in object literal not allowed in strict mode",F="ILLEGAL",V="Disabled.",z=new RegExp("[ªµºÀ-ÖØ-öø-ˁˆ-ˑˠ-ˤˬˮͰ-ʹͶͷͺ-ͽͿΆΈ-ΊΌΎ-ΡΣ-ϵϷ-ҁҊ-ԯԱ-Ֆՙա-ևא-תװ-ײؠ-يٮٯٱ-ۓەۥۦۮۯۺ-ۼۿܐܒ-ܯݍ-ޥޱߊ-ߪߴߵߺࠀ-ࠕࠚࠤࠨࡀ-ࡘࢠ-ࢲऄ-हऽॐक़-ॡॱ-ঀঅ-ঌএঐও-নপ-রলশ-হঽৎড়ঢ়য়-ৡৰৱਅ-ਊਏਐਓ-ਨਪ-ਰਲਲ਼ਵਸ਼ਸਹਖ਼-ੜਫ਼ੲ-ੴઅ-ઍએ-ઑઓ-નપ-રલળવ-હઽૐૠૡଅ-ଌଏଐଓ-ନପ-ରଲଳଵ-ହଽଡ଼ଢ଼ୟ-ୡୱஃஅ-ஊஎ-ஐஒ-கஙசஜஞடணதந-பம-ஹௐఅ-ఌఎ-ఐఒ-నప-హఽౘౙౠౡಅ-ಌಎ-ಐಒ-ನಪ-ಳವ-ಹಽೞೠೡೱೲഅ-ഌഎ-ഐഒ-ഺഽൎൠൡൺ-ൿඅ-ඖක-නඳ-රලව-ෆก-ะาำเ-ๆກຂຄງຈຊຍດ-ທນ-ຟມ-ຣລວສຫອ-ະາຳຽເ-ໄໆໜ-ໟༀཀ-ཇཉ-ཬྈ-ྌက-ဪဿၐ-ၕၚ-ၝၡၥၦၮ-ၰၵ-ႁႎႠ-ჅჇჍა-ჺჼ-ቈቊ-ቍቐ-ቖቘቚ-ቝበ-ኈኊ-ኍነ-ኰኲ-ኵኸ-ኾዀዂ-ዅወ-ዖዘ-ጐጒ-ጕጘ-ፚᎀ-ᎏᎠ-Ᏼᐁ-ᙬᙯ-ᙿᚁ-ᚚᚠ-ᛪᛮ-ᛸᜀ-ᜌᜎ-ᜑᜠ-ᜱᝀ-ᝑᝠ-ᝬᝮ-ᝰក-ឳៗៜᠠ-ᡷᢀ-ᢨᢪᢰ-ᣵᤀ-ᤞᥐ-ᥭᥰ-ᥴᦀ-ᦫᧁ-ᧇᨀ-ᨖᨠ-ᩔᪧᬅ-ᬳᭅ-ᭋᮃ-ᮠᮮᮯᮺ-ᯥᰀ-ᰣᱍ-ᱏᱚ-ᱽᳩ-ᳬᳮ-ᳱᳵᳶᴀ-ᶿḀ-ἕἘ-Ἕἠ-ὅὈ-Ὅὐ-ὗὙὛὝὟ-ώᾀ-ᾴᾶ-ᾼιῂ-ῄῆ-ῌῐ-ΐῖ-Ίῠ-Ῥῲ-ῴῶ-ῼⁱⁿₐ-ₜℂℇℊ-ℓℕℙ-ℝℤΩℨK-ℭℯ-ℹℼ-ℿⅅ-ⅉⅎⅠ-ↈⰀ-Ⱞⰰ-ⱞⱠ-ⳤⳫ-ⳮⳲⳳⴀ-ⴥⴧⴭⴰ-ⵧⵯⶀ-ⶖⶠ-ⶦⶨ-ⶮⶰ-ⶶⶸ-ⶾⷀ-ⷆⷈ-ⷎⷐ-ⷖⷘ-ⷞⸯ々-〇〡-〩〱-〵〸-〼ぁ-ゖゝ-ゟァ-ヺー-ヿㄅ-ㄭㄱ-ㆎㆠ-ㆺㇰ-ㇿ㐀-䶵一-鿌ꀀ-ꒌꓐ-ꓽꔀ-ꘌꘐ-ꘟꘪꘫꙀ-ꙮꙿ-ꚝꚠ-ꛯꜗ-ꜟꜢ-ꞈꞋ-ꞎꞐ-ꞭꞰꞱꟷ-ꠁꠃ-ꠅꠇ-ꠊꠌ-ꠢꡀ-ꡳꢂ-ꢳꣲ-ꣷꣻꤊ-ꤥꤰ-ꥆꥠ-ꥼꦄ-ꦲꧏꧠ-ꧤꧦ-ꧯꧺ-ꧾꨀ-ꨨꩀ-ꩂꩄ-ꩋꩠ-ꩶꩺꩾ-ꪯꪱꪵꪶꪹ-ꪽꫀꫂꫛ-ꫝꫠ-ꫪꫲ-ꫴꬁ-ꬆꬉ-ꬎꬑ-ꬖꬠ-ꬦꬨ-ꬮꬰ-ꭚꭜ-ꭟꭤꭥꯀ-ꯢ가-힣ힰ-ퟆퟋ-ퟻ豈-舘並-龎ﬀ-ﬆﬓ-ﬗיִײַ-ﬨשׁ-זּטּ-לּמּנּסּףּפּצּ-ﮱﯓ-ﴽﵐ-ﶏﶒ-ﷇﷰ-ﷻﹰ-ﹴﹶ-ﻼＡ-Ｚａ-ｚｦ-ﾾￂ-ￇￊ-ￏￒ-ￗￚ-ￜ]"),B=new RegExp("[ªµºÀ-ÖØ-öø-ˁˆ-ˑˠ-ˤˬˮ̀-ʹͶͷͺ-ͽͿΆΈ-ΊΌΎ-ΡΣ-ϵϷ-ҁ҃-҇Ҋ-ԯԱ-Ֆՙա-և֑-ׇֽֿׁׂׅׄא-תװ-ײؐ-ؚؠ-٩ٮ-ۓە-ۜ۟-۪ۨ-ۼۿܐ-݊ݍ-ޱ߀-ߵߺࠀ-࠭ࡀ-࡛ࢠ-ࢲࣤ-ॣ०-९ॱ-ঃঅ-ঌএঐও-নপ-রলশ-হ়-ৄেৈো-ৎৗড়ঢ়য়-ৣ০-ৱਁ-ਃਅ-ਊਏਐਓ-ਨਪ-ਰਲਲ਼ਵਸ਼ਸਹ਼ਾ-ੂੇੈੋ-੍ੑਖ਼-ੜਫ਼੦-ੵઁ-ઃઅ-ઍએ-ઑઓ-નપ-રલળવ-હ઼-ૅે-ૉો-્ૐૠ-ૣ૦-૯ଁ-ଃଅ-ଌଏଐଓ-ନପ-ରଲଳଵ-ହ଼-ୄେୈୋ-୍ୖୗଡ଼ଢ଼ୟ-ୣ୦-୯ୱஂஃஅ-ஊஎ-ஐஒ-கஙசஜஞடணதந-பம-ஹா-ூெ-ைொ-்ௐௗ௦-௯ఀ-ఃఅ-ఌఎ-ఐఒ-నప-హఽ-ౄె-ైొ-్ౕౖౘౙౠ-ౣ౦-౯ಁ-ಃಅ-ಌಎ-ಐಒ-ನಪ-ಳವ-ಹ಼-ೄೆ-ೈೊ-್ೕೖೞೠ-ೣ೦-೯ೱೲഁ-ഃഅ-ഌഎ-ഐഒ-ഺഽ-ൄെ-ൈൊ-ൎൗൠ-ൣ൦-൯ൺ-ൿංඃඅ-ඖක-නඳ-රලව-ෆ්ා-ුූෘ-ෟ෦-෯ෲෳก-ฺเ-๎๐-๙ກຂຄງຈຊຍດ-ທນ-ຟມ-ຣລວສຫອ-ູົ-ຽເ-ໄໆ່-ໍ໐-໙ໜ-ໟༀ༘༙༠-༩༹༵༷༾-ཇཉ-ཬཱ-྄྆-ྗྙ-ྼ࿆က-၉ၐ-ႝႠ-ჅჇჍა-ჺჼ-ቈቊ-ቍቐ-ቖቘቚ-ቝበ-ኈኊ-ኍነ-ኰኲ-ኵኸ-ኾዀዂ-ዅወ-ዖዘ-ጐጒ-ጕጘ-ፚ፝-፟ᎀ-ᎏᎠ-Ᏼᐁ-ᙬᙯ-ᙿᚁ-ᚚᚠ-ᛪᛮ-ᛸᜀ-ᜌᜎ-᜔ᜠ-᜴ᝀ-ᝓᝠ-ᝬᝮ-ᝰᝲᝳក-៓ៗៜ៝០-៩᠋-᠍᠐-᠙ᠠ-ᡷᢀ-ᢪᢰ-ᣵᤀ-ᤞᤠ-ᤫᤰ-᤻᥆-ᥭᥰ-ᥴᦀ-ᦫᦰ-ᧉ᧐-᧙ᨀ-ᨛᨠ-ᩞ᩠-᩿᩼-᪉᪐-᪙ᪧ᪰-᪽ᬀ-ᭋ᭐-᭙᭫-᭳ᮀ-᯳ᰀ-᰷᱀-᱉ᱍ-ᱽ᳐-᳔᳒-ᳶ᳸᳹ᴀ-᷵᷼-ἕἘ-Ἕἠ-ὅὈ-Ὅὐ-ὗὙὛὝὟ-ώᾀ-ᾴᾶ-ᾼιῂ-ῄῆ-ῌῐ-ΐῖ-Ίῠ-Ῥῲ-ῴῶ-ῼ‌‍‿⁀⁔ⁱⁿₐ-ₜ⃐-⃥⃜⃡-⃰ℂℇℊ-ℓℕℙ-ℝℤΩℨK-ℭℯ-ℹℼ-ℿⅅ-ⅉⅎⅠ-ↈⰀ-Ⱞⰰ-ⱞⱠ-ⳤⳫ-ⳳⴀ-ⴥⴧⴭⴰ-ⵧⵯ⵿-ⶖⶠ-ⶦⶨ-ⶮⶰ-ⶶⶸ-ⶾⷀ-ⷆⷈ-ⷎⷐ-ⷖⷘ-ⷞⷠ-ⷿⸯ々-〇〡-〯〱-〵〸-〼ぁ-ゖ゙゚ゝ-ゟァ-ヺー-ヿㄅ-ㄭㄱ-ㆎㆠ-ㆺㇰ-ㇿ㐀-䶵一-鿌ꀀ-ꒌꓐ-ꓽꔀ-ꘌꘐ-ꘫꙀ-꙯ꙴ-꙽ꙿ-ꚝꚟ-꛱ꜗ-ꜟꜢ-ꞈꞋ-ꞎꞐ-ꞭꞰꞱꟷ-ꠧꡀ-ꡳꢀ-꣄꣐-꣙꣠-ꣷꣻ꤀-꤭ꤰ-꥓ꥠ-ꥼꦀ-꧀ꧏ-꧙ꧠ-ꧾꨀ-ꨶꩀ-ꩍ꩐-꩙ꩠ-ꩶꩺ-ꫂꫛ-ꫝꫠ-ꫯꫲ-꫶ꬁ-ꬆꬉ-ꬎꬑ-ꬖꬠ-ꬦꬨ-ꬮꬰ-ꭚꭜ-ꭟꭤꭥꯀ-ꯪ꯬꯭꯰-꯹가-힣ힰ-ퟆퟋ-ퟻ豈-舘並-龎ﬀ-ﬆﬓ-ﬗיִ-ﬨשׁ-זּטּ-לּמּנּסּףּפּצּ-ﮱﯓ-ﴽﵐ-ﶏﶒ-ﷇﷰ-ﷻ︀-️︠-︭︳︴﹍-﹏ﹰ-ﹴﹶ-ﻼ０-９Ａ-Ｚ＿ａ-ｚｦ-ﾾￂ-ￇￊ-ￏￒ-ￗￚ-ￜ]");function q(t,e){if(!t)throw new Error("ASSERT: "+e)}function H(t){return t>=48&&t<=57}function U(t){return"0123456789abcdefABCDEF".indexOf(t)>=0}function W(t){return"01234567".indexOf(t)>=0}function G(t){return 10===t||13===t||8232===t||8233===t}function X(t){return 36===t||95===t||t>=65&&t<=90||t>=97&&t<=122||92===t||t>=128&&z.test(String.fromCharCode(t))}function Y(t){return 36===t||95===t||t>=65&&t<=90||t>=97&&t<=122||t>=48&&t<=57||92===t||t>=128&&B.test(String.fromCharCode(t))}var K={if:1,in:1,do:1,var:1,for:1,new:1,try:1,let:1,this:1,else:1,case:1,void:1,with:1,enum:1,while:1,break:1,catch:1,throw:1,const:1,yield:1,class:1,super:1,return:1,typeof:1,delete:1,switch:1,export:1,import:1,public:1,static:1,default:1,finally:1,extends:1,package:1,private:1,function:1,continue:1,debugger:1,interface:1,protected:1,instanceof:1,implements:1};function $(){for(var t,e;o<s&&(t=i.charCodeAt(o),32===(e=t)||9===e||11===e||12===e||160===e||e>=5760&&[5760,6158,8192,8193,8194,8195,8196,8197,8198,8199,8200,8201,8202,8239,8287,12288,65279].indexOf(e)>=0||G(t));)++o}function Q(t){var e,n,r,a=0;for(n="u"===t?4:2,e=0;e<n;++e)o<s&&U(i[o])?(r=i[o++],a=16*a+"0123456789abcdef".indexOf(r.toLowerCase())):dt({},S,F);return String.fromCharCode(a)}function Z(){var t,e,n,r;for(e=0,"}"===(t=i[o])&&dt({},S,F);o<s&&U(t=i[o++]);)e=16*e+"0123456789abcdef".indexOf(t.toLowerCase());return(e>1114111||"}"!==t)&&dt({},S,F),e<=65535?String.fromCharCode(e):(n=55296+(e-65536>>10),r=56320+(e-65536&1023),String.fromCharCode(n,r))}function J(){var t,e;for(t=i.charCodeAt(o++),e=String.fromCharCode(t),92===t&&(117!==i.charCodeAt(o)&&dt({},S,F),++o,(t=Q("u"))&&"\\"!==t&&X(t.charCodeAt(0))||dt({},S,F),e=t);o<s&&Y(t=i.charCodeAt(o));)++o,e+=String.fromCharCode(t),92===t&&(e=e.substr(0,e.length-1),117!==i.charCodeAt(o)&&dt({},S,F),++o,(t=Q("u"))&&"\\"!==t&&Y(t.charCodeAt(0))||dt({},S,F),e+=t);return e}function tt(){var t,e;return t=o,{type:1===(e=92===i.charCodeAt(o)?J():function(){var t,e;for(t=o++;o<s;){if(92===(e=i.charCodeAt(o)))return o=t,J();if(!Y(e))break;++o}return i.slice(t,o)}()).length?p:K.hasOwnProperty(e)?d:"null"===e?h:"true"===e||"false"===e?u:p,value:e,start:t,end:o}}function et(){var t,e,n,r,s=o,a=i.charCodeAt(o),c=i[o];switch(a){case 46:case 40:case 41:case 59:case 44:case 123:case 125:case 91:case 93:case 58:case 63:case 126:return++o,{type:m,value:String.fromCharCode(a),start:s,end:o};default:if(61===(t=i.charCodeAt(o+1)))switch(a){case 43:case 45:case 47:case 60:case 62:case 94:case 124:case 37:case 38:case 42:return o+=2,{type:m,value:String.fromCharCode(a)+String.fromCharCode(t),start:s,end:o};case 33:case 61:return o+=2,61===i.charCodeAt(o)&&++o,{type:m,value:i.slice(s,o),start:s,end:o}}}return">>>="===(r=i.substr(o,4))?{type:m,value:r,start:s,end:o+=4}:">>>"===(n=r.substr(0,3))||"<<="===n||">>="===n?{type:m,value:n,start:s,end:o+=3}:c===(e=n.substr(0,2))[1]&&"+-<>&|".indexOf(c)>=0||"=>"===e?{type:m,value:e,start:s,end:o+=2}:"<>=!+-*%&|^/".indexOf(c)>=0?{type:m,value:c,start:s,end:++o}:void dt({},S,F)}function nt(){var t,e,n;if(q(H((n=i[o]).charCodeAt(0))||"."===n,"Numeric literal must start with a decimal digit or a decimal point"),e=o,t="","."!==n){if(t=i[o++],n=i[o],"0"===t){if("x"===n||"X"===n)return++o,function(t){for(var e="";o<s&&U(i[o]);)e+=i[o++];return 0===e.length&&dt({},S,F),X(i.charCodeAt(o))&&dt({},S,F),{type:f,value:parseInt("0x"+e,16),start:t,end:o}}(e);if(W(n))return function(t){for(var e="0"+i[o++];o<s&&W(i[o]);)e+=i[o++];return(X(i.charCodeAt(o))||H(i.charCodeAt(o)))&&dt({},S,F),{type:f,value:parseInt(e,8),octal:!0,start:t,end:o}}(e);n&&H(n.charCodeAt(0))&&dt({},S,F)}for(;H(i.charCodeAt(o));)t+=i[o++];n=i[o]}if("."===n){for(t+=i[o++];H(i.charCodeAt(o));)t+=i[o++];n=i[o]}if("e"===n||"E"===n)if(t+=i[o++],"+"!==(n=i[o])&&"-"!==n||(t+=i[o++]),H(i.charCodeAt(o)))for(;H(i.charCodeAt(o));)t+=i[o++];else dt({},S,F);return X(i.charCodeAt(o))&&dt({},S,F),{type:f,value:parseFloat(t),start:e,end:o}}function rt(){var t,e,n,r;return a=null,$(),t=o,e=function(){var t,e,n,r;for(q("/"===(t=i[o]),"Regular expression literal must start with a slash"),e=i[o++],n=!1,r=!1;o<s;)if(e+=t=i[o++],"\\"===t)G((t=i[o++]).charCodeAt(0))&&dt({},R),e+=t;else if(G(t.charCodeAt(0)))dt({},R);else if(n)"]"===t&&(n=!1);else{if("/"===t){r=!0;break}"["===t&&(n=!0)}return r||dt({},R),{value:e.substr(1,e.length-2),literal:e}}(),n=function(){var t,e,n;for(e="",n="";o<s&&Y((t=i[o]).charCodeAt(0));)++o,"\\"===t&&o<s?dt({},S,F):(n+=t,e+=t);return n.search(/[^gimuy]/g)>=0&&dt({},I,n),{value:n,literal:e}}(),r=function(t,e){var n=t;e.indexOf("u")>=0&&(n=n.replace(/\\u\{([0-9a-fA-F]+)\}/g,function(t,e){if(parseInt(e,16)<=1114111)return"x";dt({},I)}).replace(/[\uD800-\uDBFF][\uDC00-\uDFFF]/g,"x"));try{new RegExp(n)}catch(t){dt({},I)}try{return new RegExp(t,e)}catch(t){return null}}(e.value,n.value),{literal:e.literal+n.literal,value:r,regex:{pattern:e.value,flags:n.value},start:t,end:o}}function it(){var t;return $(),o>=s?{type:l,start:o,end:o}:X(t=i.charCodeAt(o))?tt():40===t||41===t||59===t?et():39===t||34===t?function(){var t,e,n,r,a="",c=!1;for(q("'"===(t=i[o])||'"'===t,"String literal must starts with a quote"),e=o,++o;o<s;){if((n=i[o++])===t){t="";break}if("\\"===n)if((n=i[o++])&&G(n.charCodeAt(0)))"\r"===n&&"\n"===i[o]&&++o;else switch(n){case"u":case"x":"{"===i[o]?(++o,a+=Z()):a+=Q(n);break;case"n":a+="\n";break;case"r":a+="\r";break;case"t":a+="\t";break;case"b":a+="\b";break;case"f":a+="\f";break;case"v":a+="\v";break;default:W(n)?(0!==(r="01234567".indexOf(n))&&(c=!0),o<s&&W(i[o])&&(c=!0,r=8*r+"01234567".indexOf(i[o++]),"0123".indexOf(n)>=0&&o<s&&W(i[o])&&(r=8*r+"01234567".indexOf(i[o++]))),a+=String.fromCharCode(r)):a+=n}else{if(G(n.charCodeAt(0)))break;a+=n}}return""!==t&&dt({},S,F),{type:g,value:a,octal:c,start:e,end:o}}():46===t?H(i.charCodeAt(o+1))?nt():et():H(t)?nt():et()}function ot(){var t;return o=(t=a).end,a=it(),o=t.end,t}function st(){var t;t=o,a=it(),o=t}function at(t,e,n){var r=new c.a("||"===t||"&&"===t?w:y);return r.operator=t,r.left=e,r.right=n,r}function ct(t){var e=new c.a(j);return e.name=t,e}function ut(t){var e=new c.a(x);return e.value=t.value,e.raw=i.slice(t.start,t.end),t.regex&&("//"===e.raw&&(e.raw="/(?:)/"),e.regex=t.regex),e}function lt(t,e,n){var r=new c.a(O);return r.computed="["===t,r.object=e,r.property=n,r.computed||(n.member=!0),r}function pt(t,e,n){var r=new c.a(k);return r.key=e,r.value=n,r.kind=t,r}function dt(t,e){var n,r=Array.prototype.slice.call(arguments,2),i=e.replace(/%(\d)/g,function(t,e){return q(e<r.length,"Message reference must be in range"),r[e]});throw(n=new Error(i)).index=o,n.description=i,n}function ht(t){t.type===l&&dt(t,D),t.type===f&&dt(t,P),t.type===g&&dt(t,M),t.type===p&&dt(t,A),t.type===d&&dt(t,T),dt(t,S,t.value)}function ft(t){var e=ot();e.type===m&&e.value===t||ht(e)}function mt(t){return a.type===m&&a.value===t}function gt(t){return a.type===d&&a.value===t}function vt(){var t,e,n=[];for(o=a.start,ft("[");!mt("]");)mt(",")?(ot(),n.push(null)):(n.push(Et()),mt("]")||ft(","));return ot(),t=n,(e=new c.a(v)).elements=t,e}function yt(){var t;return o=a.start,(t=ot()).type===g||t.type===f?(t.octal&&dt(t,N),ut(t)):ct(t.value)}function bt(){var t,e,n,r,i,s,u,d=[],h={},f=String;for(o=a.start,ft("{");!mt("}");)n=void 0,r=void 0,i=void 0,o=a.start,e="$"+((t=(n=a).type===p?(i=yt(),ft(":"),pt("init",i,Et())):n.type!==l&&n.type!==m?(r=yt(),ft(":"),pt("init",r,Et())):void ht(n)).key.type===j?t.key.name:f(t.key.value)),Object.prototype.hasOwnProperty.call(h,e)?dt({},L):h[e]=!0,d.push(t),mt("}")||ft(",");return ft("}"),s=d,(u=new c.a(C)).properties=s,u}var _t={if:1,this:1};function jt(){var t,e,n,r;if(mt("("))return ft("("),r=St(),ft(")"),r;if(mt("["))return vt();if(mt("{"))return bt();if(t=a.type,o=a.start,t===p||_t[a.value])n=ct(ot().value);else if(t===g||t===f)a.octal&&dt(a,N),n=ut(ot());else{if(t===d)throw new Error(V);t===u?((e=ot()).value="true"===e.value,n=ut(e)):t===h?((e=ot()).value=null,n=ut(e)):mt("/")||mt("/=")?(n=ut(rt()),st()):ht(ot())}return n}function xt(){var t=[];if(ft("("),!mt(")"))for(;o<s&&(t.push(Et()),!mt(")"));)ft(",");return ft(")"),t}function wt(){var t,e;return o=a.start,t=ot(),(e=t).type!==p&&e.type!==d&&e.type!==u&&e.type!==h&&ht(t),ct(t.value)}function Ot(){var t=function(){var t,e,n,r,i,o;for(t=jt();;)if(mt("."))ft("."),t=lt(".",t,wt());else if(mt("("))e=xt(),r=t,i=e,o=void 0,(o=new c.a(b)).callee=r,o.arguments=i,t=o;else{if(!mt("["))break;n=void 0,ft("["),n=St(),ft("]"),t=lt("[",t,n)}return t}();if(a.type===m&&(mt("++")||mt("--")))throw new Error(V);return t}function Ct(){var t,e,n,r,i;if(a.type!==m&&a.type!==d)e=Ot();else{if(mt("++")||mt("--"))throw new Error(V);if(mt("+")||mt("-")||mt("~")||mt("!"))t=ot(),e=Ct(),n=t.value,r=e,(i=new c.a(E)).operator=n,i.argument=r,i.prefix=!0,e=i;else{if(gt("delete")||gt("void")||gt("typeof"))throw new Error(V);e=Ot()}}return e}function kt(t){var e=0;if(t.type!==m&&t.type!==d)return 0;switch(t.value){case"||":e=1;break;case"&&":e=2;break;case"|":e=3;break;case"^":e=4;break;case"&":e=5;break;case"==":case"!=":case"===":case"!==":e=6;break;case"<":case">":case"<=":case">=":case"instanceof":case"in":e=7;break;case"<<":case">>":case">>>":e=8;break;case"+":case"-":e=9;break;case"*":case"/":case"%":e=11}return e}function Et(){var t,e,n,r,i,o,s;return t=function(){var t,e,n,r,i,o,s,c,u,l;if(t=a,u=Ct(),0===(i=kt(r=a)))return u;for(r.prec=i,ot(),e=[t,a],o=[u,r,s=Ct()];(i=kt(a))>0;){for(;o.length>2&&i<=o[o.length-2].prec;)s=o.pop(),c=o.pop().value,u=o.pop(),e.pop(),n=at(c,u,s),o.push(n);(r=ot()).prec=i,o.push(r),e.push(a),n=Ct(),o.push(n)}for(n=o[l=o.length-1],e.pop();l>1;)e.pop(),n=at(o[l-1].value,o[l-2],n),l-=2;return n}(),mt("?")&&(ot(),e=Et(),ft(":"),n=Et(),r=t,i=e,o=n,(s=new c.a(_)).test=r,s.consequent=i,s.alternate=o,t=s),t}function St(){var t=Et();if(mt(","))throw new Error(V);return t}e.a=function(t){o=0,s=(i=t).length,a=null,st();var e=St();if(a.type!==l)throw new Error("Unexpect token after expression.");return e}},"../../../../vega-force/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../vega-force/src/Force.js");n.d(e,"force",function(){return r.a})},"../../../../vega-force/src/Force.js":function(t,e,n){"use strict";e.a=p;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-force/index.js"),s={center:o.a,collide:o.b,nbody:o.d,link:o.c,x:o.f,y:o.g},a="forces",c=["alpha","alphaMin","alphaTarget","velocityDecay","forces"],u=["static","iterations"],l=["x","y","vx","vy"];function p(t){r.d.call(this,null,t)}p.Definition={type:"Force",metadata:{modifies:!0},params:[{name:"static",type:"boolean",default:!1},{name:"restart",type:"boolean",default:!1},{name:"iterations",type:"number",default:300},{name:"alpha",type:"number",default:1},{name:"alphaMin",type:"number",default:.001},{name:"alphaTarget",type:"number",default:0},{name:"velocityDecay",type:"number",default:.4},{name:"forces",type:"param",array:!0,params:[{key:{force:"center"},params:[{name:"x",type:"number",default:0},{name:"y",type:"number",default:0}]},{key:{force:"collide"},params:[{name:"radius",type:"number",expr:!0},{name:"strength",type:"number",default:.7},{name:"iterations",type:"number",default:1}]},{key:{force:"nbody"},params:[{name:"strength",type:"number",default:-30},{name:"theta",type:"number",default:.9},{name:"distanceMin",type:"number",default:1},{name:"distanceMax",type:"number"}]},{key:{force:"link"},params:[{name:"links",type:"data"},{name:"id",type:"field"},{name:"distance",type:"number",default:30,expr:!0},{name:"strength",type:"number",expr:!0},{name:"iterations",type:"number",default:1}]},{key:{force:"x"},params:[{name:"strength",type:"number",default:.1},{name:"x",type:"field"}]},{key:{force:"y"},params:[{name:"strength",type:"number",default:.1},{name:"y",type:"field"}]}]},{name:"as",type:"string",array:!0,modify:!1,default:l}]};var d=Object(i.s)(p,r.d);function h(t,e,n,r){var o,s,u,l,p=Object(i.g)(e.forces);for(o=0,s=c.length;o<s;++o)(u=c[o])!==a&&e.modified(u)&&t[u](e[u]);for(o=0,s=p.length;o<s;++o)l=a+o,(u=n||e.modified(a,o)?m(p[o]):r&&f(p[o],r)?t.force(l):null)&&t.force(l,u);for(s=t.numForces||0;o<s;++o)t.force(a+o,null);return t.numForces=p.length,t}function f(t,e){var n,r;for(n in t)if(Object(i.w)(r=t[n])&&e.modified(Object(i.e)(r)))return 1;return 0}function m(t){var e,n;s.hasOwnProperty(t.force)||Object(i.k)("Unrecognized force: "+t.force),e=s[t.force]();for(n in t)Object(i.w)(e[n])&&g(e[n],t[n],t);return e}function g(t,e,n){t(Object(i.w)(e)?function(t){return e(t,n)}:e)}d.transform=function(t,e){var n,r,i,s,a,l,p,d,f=this.value,m=e.changed(e.ADD_REM),g=t.modified(c),v=t.iterations||300;if(f?(m&&(e.modifies("index"),f.nodes(e.source)),(g||e.changed(e.MOD))&&h(f,t,0,e)):(this.value=(i=e.source,s=t,a=Object(o.e)(i),l=!1,p=a.stop,d=a.restart,a.stopped=function(){return l},a.restart=function(){return l=!1,d()},a.stop=function(){return l=!0,p()},f=h(a,s,!0).on("end",function(){l=!0})),f.on("tick",(n=e.dataflow,r=this,function(){n.touch(r).run()})),t.static||(m=!0,f.tick()),e.modifies("index")),g||m||t.modified(u)||e.changed()&&t.restart)if(f.alpha(Math.max(f.alpha(),t.alpha||1)).alphaDecay(1-Math.pow(f.alphaMin(),1/v)),t.static)for(f.stop();--v>=0;)f.tick();else if(f.stopped()&&f.restart(),!m)return e.StopPropagation;return this.finish(t,e)},d.finish=function(t,e){for(var n,r=e.dataflow,i=this._argops,o=0,s=i.length;o<s;++o)if((n=i[o]).name===a&&"link"===n.op._argval.force)for(var c,u=n.op._argops,p=0,d=u.length;p<d;++p)if("links"===u[p].name&&(c=u[p].op.source)){r.pulse(c,r.changeset().reflow());break}return e.reflow(t.modified()).modifies(l)}},"../../../../vega-geo/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../vega-geo/src/Contour.js");n.d(e,"contour",function(){return r.a});var i=n("../../../../vega-geo/src/GeoJSON.js");n.d(e,"geojson",function(){return i.a});var o=n("../../../../vega-geo/src/GeoPath.js");n.d(e,"geopath",function(){return o.a});var s=n("../../../../vega-geo/src/GeoPoint.js");n.d(e,"geopoint",function(){return s.a});var a=n("../../../../vega-geo/src/GeoShape.js");n.d(e,"geoshape",function(){return a.a});var c=n("../../../../vega-geo/src/Graticule.js");n.d(e,"graticule",function(){return c.a});var u=n("../../../../vega-geo/src/Projection.js");n.d(e,"projection",function(){return u.a})},"../../../../vega-geo/src/Contour.js":function(t,e,n){"use strict";e.a=u;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-array/index.js"),s=n("../../../../d3-contour/index.js"),a=["values","size"],c=["x","y","size","cellSize","bandwidth"];function u(t){r.d.call(this,null,t)}u.Definition={type:"Contour",metadata:{generates:!0},params:[{name:"size",type:"number",array:!0,length:2,required:!0},{name:"values",type:"number",array:!0},{name:"x",type:"field"},{name:"y",type:"field"},{name:"cellSize",type:"number"},{name:"bandwidth",type:"number"},{name:"count",type:"number"},{name:"nice",type:"number",default:!1},{name:"thresholds",type:"number",array:!0}]},Object(i.s)(u,r.d).transform=function(t,e){if(this.value&&!e.changed()&&!t.modified())return e.StopPropagation;var n,i,u,l,p=e.fork(e.NO_SOURCE|e.NO_FIELDS),d=t.count||10;return t.values?(n=Object(s.b)(),i=a,u=t.values):(n=Object(s.a)(),i=c,u=e.materialize(e.SOURCE).source),n.thresholds(t.thresholds||(t.nice?d:(l=d,function(t){for(var e=Object(o.f)(t),n=e[0],r=e[1]-n,i=[],s=1;s<=l;++s)i.push(n+r*s/(l+1));return i}))),i.forEach(function(e){null!=t[e]&&n[e](t[e])}),this.value&&(p.rem=this.value),u=u&&u.length?n(u).map(r.h):[],this.value=p.source=p.add=u,p}},"../../../../vega-geo/src/GeoJSON.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-geo/src/constants.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js");function s(t){i.d.call(this,null,t)}s.Definition={type:"GeoJSON",metadata:{},params:[{name:"fields",type:"field",array:!0,length:2},{name:"geojson",type:"field"}]},Object(o.s)(s,i.d).transform=function(t,e){var n,i=this._features,s=this._points,a=t.fields,c=a&&a[0],u=a&&a[1],l=t.geojson,p=e.ADD;n=t.modified()||e.changed(e.REM)||e.modified(Object(o.e)(l))||c&&e.modified(Object(o.e)(c))||u&&e.modified(Object(o.e)(u)),this.value&&!n||(p=e.SOURCE,this._features=i=[],this._points=s=[]),l&&e.visit(p,function(t){i.push(l(t))}),c&&u&&(e.visit(p,function(t){var e=c(t),n=u(t);null!=e&&null!=n&&(e=+e)===e&&(n=+n)===n&&s.push([e,n])}),i=i.concat({type:r.a,geometry:{type:r.c,coordinates:s}})),this.value={type:r.b,features:i}}},"../../../../vega-geo/src/GeoPath.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-projection/index.js"),o=n("../../../../vega-util/index.js");function s(t){r.d.call(this,null,t)}s.Definition={type:"GeoPath",metadata:{modifies:!0},params:[{name:"projection",type:"projection"},{name:"field",type:"field"},{name:"pointRadius",type:"number",expr:!0},{name:"as",type:"string",default:"path"}]},Object(o.s)(s,r.d).transform=function(t,e){var n=e.fork(e.ALL),r=this.value,s=t.field||o.r,a=t.as||"path",c=n.SOURCE;!r||t.modified()?(this.value=r=Object(i.a)(t.projection),n.materialize().reflow()):c=s===o.r||e.modified(s.fields)?n.ADD_MOD:n.ADD;var u=function(t,e){var n=t.pointRadius();t.context(null),null!=e&&t.pointRadius(e);return n}(r,t.pointRadius);return n.visit(c,function(t){t[a]=r(s(t))}),r.pointRadius(u),n.modifies(a)}},"../../../../vega-geo/src/GeoPoint.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}o.Definition={type:"GeoPoint",metadata:{modifies:!0},params:[{name:"projection",type:"projection",required:!0},{name:"fields",type:"field",array:!0,required:!0,length:2},{name:"as",type:"string",array:!0,length:2,default:["x","y"]}]},Object(i.s)(o,r.d).transform=function(t,e){var n,r=t.projection,i=t.fields[0],o=t.fields[1],s=t.as||["x","y"],a=s[0],c=s[1];function u(t){var e=r([i(t),o(t)]);e?(t[a]=e[0],t[c]=e[1]):(t[a]=void 0,t[c]=void 0)}return t.modified()?e=e.materialize().reflow(!0).visit(e.SOURCE,u):(n=e.modified(i.fields)||e.modified(o.fields),e.visit(n?e.ADD_MOD:e.ADD,u)),e.modifies(s)}},"../../../../vega-geo/src/GeoShape.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-projection/index.js"),o=n("../../../../vega-util/index.js");function s(t){r.d.call(this,null,t)}s.Definition={type:"GeoShape",metadata:{modifies:!0},params:[{name:"projection",type:"projection"},{name:"field",type:"field",default:"datum"},{name:"pointRadius",type:"number",expr:!0},{name:"as",type:"string",default:"shape"}]},Object(o.s)(s,r.d).transform=function(t,e){var n,r,s,a,c=e.fork(e.ALL),u=this.value,l=t.field||Object(o.p)("datum"),p=t.as||"shape",d=c.ADD_MOD;return u&&!t.modified()||(this.value=(n=Object(i.a)(t.projection),r=l,s=t.pointRadius,(a=null==s?function(t){return n(r(t))}:function(t){var e=n.pointRadius(),i=n.pointRadius(s)(r(t));return n.pointRadius(e),i}).context=function(t){return n.context(t),a},u=a),c.materialize().reflow(),d=c.SOURCE),c.visit(d,function(t){t[p]=u}),c.modifies(p)}},"../../../../vega-geo/src/Graticule.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-geo/index.js");function s(t){r.d.call(this,[],t),this.generator=Object(o.m)()}s.Definition={type:"Graticule",metadata:{source:!0,generates:!0,changes:!0},params:[{name:"extent",type:"array",array:!0,length:2,content:{type:"number",array:!0,length:2}},{name:"extentMajor",type:"array",array:!0,length:2,content:{type:"number",array:!0,length:2}},{name:"extentMinor",type:"array",array:!0,length:2,content:{type:"number",array:!0,length:2}},{name:"step",type:"number",array:!0,length:2},{name:"stepMajor",type:"number",array:!0,length:2,default:[90,360]},{name:"stepMinor",type:"number",array:!0,length:2,default:[10,10]},{name:"precision",type:"number",default:2.5}]},Object(i.s)(s,r.d).transform=function(t,e){var n,o=e.fork(),s=this.value,a=this.generator;if(!s.length||t.modified())for(var c in t)Object(i.w)(a[c])&&a[c](t[c]);return n=a(),s.length?o.mod.push(Object(r.l)(s[0],n)):o.add.push(Object(r.h)(n)),s[0]=n,o.source=s,o}},"../../../../vega-geo/src/Projection.js":function(t,e,n){"use strict";e.a=a;var r=n("../../../../vega-geo/src/constants.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-projection/index.js"),s=n("../../../../vega-util/index.js");function a(t){i.d.call(this,null,t),this.modified(!0)}function c(t,e,n){Object(s.w)(t[e])&&t[e](n)}Object(s.s)(a,i.d).transform=function(t,e){var n,i,a,u,l=this.value;return!l||t.modified("type")?(this.value=l=function(t){var e=Object(o.b)((t||"mercator").toLowerCase());e||Object(s.k)("Unrecognized projection type: "+t);return e()}(t.type),o.c.forEach(function(e){null!=t[e]&&c(l,e,t[e])})):o.c.forEach(function(e){t.modified(e)&&c(l,e,t[e])}),null!=t.pointRadius&&l.path.pointRadius(t.pointRadius),t.fit&&(n=l,u=(i=t).fit,a=1===(u=Object(s.g)(u)).length?u[0]:{type:r.b,features:u.reduce(function(t,e){return e&&e.type===r.b?t.push.apply(t,e.features):Object(s.t)(e)?t.push.apply(t,e):t.push(e),t},[])},i.extent?n.fitExtent(i.extent,a):i.size&&n.fitSize(i.size,a)),e.fork(e.NO_SOURCE|e.NO_FIELDS)}},"../../../../vega-geo/src/constants.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r}),n.d(e,"b",function(){return i}),n.d(e,"c",function(){return o});var r="Feature",i="FeatureCollection",o="MultiPoint"},"../../../../vega-hierarchy/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../vega-hierarchy/src/Nest.js");n.d(e,"nest",function(){return r.a});var i=n("../../../../vega-hierarchy/src/Pack.js");n.d(e,"pack",function(){return i.a});var o=n("../../../../vega-hierarchy/src/Partition.js");n.d(e,"partition",function(){return o.a});var s=n("../../../../vega-hierarchy/src/Stratify.js");n.d(e,"stratify",function(){return s.a});var a=n("../../../../vega-hierarchy/src/Tree.js");n.d(e,"tree",function(){return a.a});var c=n("../../../../vega-hierarchy/src/TreeLinks.js");n.d(e,"treelinks",function(){return c.a});var u=n("../../../../vega-hierarchy/src/Treemap.js");n.d(e,"treemap",function(){return u.a})},"../../../../vega-hierarchy/src/HierarchyLayout.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}Object(i.s)(o,r.d).transform=function(t,e){e.source&&e.source.root||Object(i.k)(this.constructor.name+" transform requires a backing tree data source.");var n=this.layout(t.method),r=this.fields,o=e.source.root,s=t.as||r;t.field&&o.sum(t.field),t.sort&&o.sort(t.sort),function(t,e,n){for(var r,i=0,o=e.length;i<o;++i)(r=e[i])in n&&t[r](n[r])}(n,this.params,t);try{this.value=n(o)}catch(t){Object(i.k)(t)}return o.each(function(t){!function(t,e,n){for(var r=t.data,i=0,o=e.length-1;i<o;++i)r[n[i]]=t[e[i]];r[n[o]]=t.children?t.children.length:0}(t,r,s)}),e.reflow(t.modified()).modifies(s).modifies("leaf")}},"../../../../vega-hierarchy/src/Nest.js":function(t,e,n){"use strict";e.a=a;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-collection/index.js"),s=n("../../../../d3-hierarchy/index.js");function a(t){r.d.call(this,null,t)}function c(t){return t.values}a.Definition={type:"Nest",metadata:{treesource:!0,generates:!0},params:[{name:"keys",type:"field",array:!0},{name:"key",type:"field"},{name:"generate",type:"boolean"}]},Object(i.s)(a,r.d).transform=function(t,e){e.source||Object(i.k)("Nest transform requires an upstream data source.");var n,a,u,l=t.key||r.n,p=t.generate,d=t.modified(),h=p||d?e.fork(e.ALL):e;return(!this.value||d||e.changed())&&(p&&this.value&&(h.materialize(h.REM),this.value.each(function(t){t.children&&h.rem.push(t)})),n=Object(i.g)(t.keys).reduce(function(t,e){return t.key(e),t},Object(o.b)()).entries(h.materialize(h.SOURCE).source),this.value=a=Object(s.b)({values:n},c),p&&(h.materialize(h.ADD),h.source=h.source.slice(),a.each(function(t){t.children&&(t=Object(r.h)(t.data),h.add.push(t),h.source.push(t))})),u=a.lookup={},a.each(function(t){null!=Object(r.n)(t.data)&&(u[l(t.data)]=t)})),h.source.root=this.value,h}},"../../../../vega-hierarchy/src/Pack.js":function(t,e,n){"use strict";e.a=a;var r=n("../../../../vega-hierarchy/src/HierarchyLayout.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-hierarchy/index.js"),s=["x","y","r","depth","children"];function a(t){r.a.call(this,t)}a.Definition={type:"Pack",metadata:{tree:!0,modifies:!0},params:[{name:"field",type:"field"},{name:"sort",type:"compare"},{name:"padding",type:"number",default:0},{name:"radius",type:"field",default:null},{name:"size",type:"number",array:!0,length:2},{name:"as",type:"string",array:!0,length:3,default:s}]};var c=Object(i.s)(a,r.a);c.layout=o.c,c.params=["size","padding"],c.fields=s},"../../../../vega-hierarchy/src/Partition.js":function(t,e,n){"use strict";e.a=a;var r=n("../../../../vega-hierarchy/src/HierarchyLayout.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-hierarchy/index.js"),s=["x0","y0","x1","y1","depth","children"];function a(t){r.a.call(this,t)}a.Definition={type:"Partition",metadata:{tree:!0,modifies:!0},params:[{name:"field",type:"field"},{name:"sort",type:"compare"},{name:"padding",type:"number",default:0},{name:"round",type:"boolean",default:!1},{name:"size",type:"number",array:!0,length:2},{name:"as",type:"string",array:!0,length:4,default:s}]};var c=Object(i.s)(a,r.a);c.layout=o.d,c.params=["size","round","padding"],c.fields=s},"../../../../vega-hierarchy/src/Stratify.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-hierarchy/index.js");function s(t){r.d.call(this,null,t)}s.Definition={type:"Stratify",metadata:{treesource:!0},params:[{name:"key",type:"field",required:!0},{name:"parentKey",type:"field",required:!0}]},Object(i.s)(s,r.d).transform=function(t,e){e.source||Object(i.k)("Stratify transform requires an upstream data source.");var n,r,s=t.modified();return(!this.value||s||e.changed(e.ADD_REM)||e.modified(t.key.fields)||e.modified(t.parentKey.fields))&&(n=Object(o.e)().id(t.key).parentId(t.parentKey)(e.source),r=n.lookup={},n.each(function(e){r[t.key(e.data)]=e}),this.value=n),e.source.root=this.value,s?e.fork(e.ALL):e}},"../../../../vega-hierarchy/src/Tree.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../vega-hierarchy/src/HierarchyLayout.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-hierarchy/index.js"),s={tidy:o.f,cluster:o.a},a=["x","y","depth","children"];function c(t){r.a.call(this,t)}c.Definition={type:"Tree",metadata:{tree:!0,modifies:!0},params:[{name:"field",type:"field"},{name:"sort",type:"compare"},{name:"method",type:"enum",default:"tidy",values:["tidy","cluster"]},{name:"size",type:"number",array:!0,length:2},{name:"nodeSize",type:"number",array:!0,length:2},{name:"as",type:"string",array:!0,length:4,default:a}]};var u=Object(i.s)(c,r.a);u.layout=function(t){var e=t||"tidy";if(s.hasOwnProperty(e))return s[e]();Object(i.k)("Unrecognized Tree layout method: "+e)},u.params=["size","nodeSize","separation"],u.fields=a},"../../../../vega-hierarchy/src/TreeLinks.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,{},t)}o.Definition={type:"TreeLinks",metadata:{tree:!0,generates:!0,changes:!0},params:[{name:"key",type:"field"}]},Object(i.s)(o,r.d).transform=function(t,e){e.source&&e.source.root||Object(i.k)("TreeLinks transform requires a backing tree data source.");var n=e.source.root.lookup,o=this.value,s=t.key||r.n,a={},c=e.fork();function u(t){var e=o[t];e&&(a[t]=1,c.mod.push(e))}return e.visit(e.REM,function(t){var e=s(t),n=o[e];n&&(delete o[e],c.rem.push(n))}),e.visit(e.ADD,function(t){var e,i,u,l=s(t);(e=(i=n[l]).parent&&(u=i.parent.data)&&null!=Object(r.n)(u)&&u)&&(c.add.push(o[l]=Object(r.h)({source:e,target:t})),a[l]=1)}),e.visit(e.MOD,function(t){var e=s(t),r=n[e].children;if(u(e),r)for(var i=0,o=r.length;i<o;++i)a[e=s(r[i].data)]||u(e)}),c}},"../../../../vega-hierarchy/src/Treemap.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../vega-hierarchy/src/HierarchyLayout.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-hierarchy/index.js"),s={binary:o.h,dice:o.i,slice:o.k,slicedice:o.l,squarify:o.m,resquarify:o.j},a=["x0","y0","x1","y1","depth","children"];function c(t){r.a.call(this,t)}c.Definition={type:"Treemap",metadata:{tree:!0,modifies:!0},params:[{name:"field",type:"field"},{name:"sort",type:"compare"},{name:"method",type:"enum",default:"squarify",values:["squarify","resquarify","binary","dice","slice","slicedice"]},{name:"padding",type:"number",default:0},{name:"paddingInner",type:"number",default:0},{name:"paddingOuter",type:"number",default:0},{name:"paddingTop",type:"number",default:0},{name:"paddingRight",type:"number",default:0},{name:"paddingBottom",type:"number",default:0},{name:"paddingLeft",type:"number",default:0},{name:"ratio",type:"number",default:1.618033988749895},{name:"round",type:"boolean",default:!1},{name:"size",type:"number",array:!0,length:2},{name:"as",type:"string",array:!0,length:4,default:a}]};var u=Object(i.s)(c,r.a);u.layout=function(){var t=Object(o.g)();return t.ratio=function(e){var n=t.tile();n.ratio&&t.tile(n.ratio(e))},t.method=function(e){s.hasOwnProperty(e)?t.tile(s[e]):Object(i.k)("Unrecognized Treemap layout method: "+e)},t},u.params=["method","ratio","size","round","padding","paddingInner","paddingOuter","paddingTop","paddingRight","paddingBottom","paddingLeft"],u.fields=a},"../../../../vega-lib/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-transforms/index.js"),s=n("../../../../vega-view-transforms/index.js"),a=n("../../../../vega-encode/index.js"),c=n("../../../../vega-geo/index.js"),u=n("../../../../vega-force/index.js"),l=n("../../../../vega-hierarchy/index.js"),p=n("../../../../vega-voronoi/index.js"),d=n("../../../../vega-wordcloud/index.js"),h=n("../../../../vega-crossfilter/index.js"),f=n("../../../../vega-lib/package.json"),m=(n.n(f),n("../../../../vega-statistics/index.js"),n("../../../../vega-loader/index.js"),n("../../../../vega-scenegraph/index.js"),n("../../../../vega-scale/index.js"),n("../../../../vega-projection/index.js"),n("../../../../vega-view/index.js"));n.d(e,"a",function(){return m.a});var g=n("../../../../vega-parser/index.js");n.d(e,"b",function(){return g.b});n("../../../../vega-runtime/index.js");Object(r.l)(i.m,o,s,a,c,u,l,p,d,h)},"../../../../vega-lib/package.json":function(t,e){t.exports={_args:[["vega-lib@3.1.0","/home/travis/build/opencobra/memote/memote-report-app"]],_from:"vega-lib@3.1.0",_id:"vega-lib@3.1.0",_inBundle:!1,_integrity:"sha512-76M6DTK6ErAEIWrbah2IPy1CYgTfvYIyqrdGid5cNanUWWXAonP2arcmibWIYxLhbAr7o6QWL+GQM1vkqwMxhg==",_location:"/vega-lib",_phantomChildren:{},_requested:{type:"version",registry:!0,raw:"vega-lib@3.1.0",name:"vega-lib",escapedName:"vega-lib",rawSpec:"3.1.0",saveSpec:null,fetchSpec:"3.1.0"},_requiredBy:["/"],_resolved:"https://registry.npmjs.org/vega-lib/-/vega-lib-3.1.0.tgz",_spec:"3.1.0",_where:"/home/travis/build/opencobra/memote/memote-report-app",author:{name:"UW Interactive Data Lab",url:"http://idl.cs.washington.edu"},browser:{buffer:!1,canvas:!1,"canvas-prebuilt":!1,fs:!1},bugs:{url:"https://github.com/vega/vega-lib/issues"},contributors:[{name:"Jeffrey Heer",url:"http://idl.cs.washington.edu"},{name:"Arvind Satyanarayan",url:"http://arvindsatya.com"},{name:"Ryan Russell",url:"https://github.com/RussellSprouts"},{name:"Emily Gu",url:"https://github.com/emilygu"},{name:"Dominik Moritz",url:"https://www.domoritz.de"},{name:"Yuri Astrakhan",url:"https://github.com/nyurik"}],dependencies:{"vega-crossfilter":"2","vega-dataflow":"3","vega-encode":"2","vega-expression":"^2.3","vega-force":"2","vega-geo":"^2.2","vega-hierarchy":"^2.1","vega-loader":"2","vega-parser":"^2.5","vega-projection":"1","vega-runtime":"2","vega-scale":"^2.1","vega-scenegraph":"^2.3","vega-statistics":"^1.2","vega-transforms":"^1.2","vega-util":"^1.7","vega-view":"^2.2","vega-view-transforms":"^1.2","vega-voronoi":"2","vega-wordcloud":"^2.1"},description:"Include Vega in projeccts using minimal dependencies.",devDependencies:{eslint:"4",rollup:"0.43","rollup-plugin-json":"2","rollup-plugin-node-resolve":"3",tape:"4",tv4:"1","uglify-js":"3"},homepage:"https://github.com/vega/vega-lib#readme","jsnext:main":"index",keywords:["vega","visualization","interaction","dataflow","library","data","d3"],license:"BSD-3-Clause",main:"build/vega.js",module:"index",name:"vega-lib",repository:{type:"git",url:"git+https://github.com/vega/vega-lib.git"},scripts:{__postpublish:"git push && git push --tags",build:"npm run rollup && bin/rollup -e",postbuild:"uglifyjs build/vega.js -c -m -o build/vega.min.js && uglifyjs build/vega-core.js -c -m -o build/vega-core.min.js",prebuild:"rm -rf build && mkdir build",prepublish:"npm run test && npm run build",rollup:"bin/rollup && bin/schema > build/vega-schema.json",test:"eslint index.js"},version:"3.1.0"}},"../../../../vega-loader/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-loader/src/loader.js");n.d(e,"a",function(){return r.a});var i=n("../../../../vega-loader/src/read.js");n.d(e,"b",function(){return i.a});n("../../../../vega-loader/src/type.js"),n("../../../../vega-loader/src/formats/index.js")},"../../../../vega-loader/node_modules/topojson-client/index.js":function(t,e,n){"use strict";n("../../../../vega-loader/node_modules/topojson-client/src/bbox.js");var r=n("../../../../vega-loader/node_modules/topojson-client/src/feature.js");n.d(e,"a",function(){return r.a});var i=n("../../../../vega-loader/node_modules/topojson-client/src/mesh.js");n.d(e,"b",function(){return i.a});n("../../../../vega-loader/node_modules/topojson-client/src/merge.js"),n("../../../../vega-loader/node_modules/topojson-client/src/neighbors.js"),n("../../../../vega-loader/node_modules/topojson-client/src/quantize.js"),n("../../../../vega-loader/node_modules/topojson-client/src/transform.js"),n("../../../../vega-loader/node_modules/topojson-client/src/untransform.js")},"../../../../vega-loader/node_modules/topojson-client/src/bbox.js":function(t,e,n){"use strict";var r=n("../../../../vega-loader/node_modules/topojson-client/src/transform.js");e.a=function(t){var e,n=Object(r.a)(t.transform),i=1/0,o=i,s=-i,a=-i;function c(t){(t=n(t))[0]<i&&(i=t[0]),t[0]>s&&(s=t[0]),t[1]<o&&(o=t[1]),t[1]>a&&(a=t[1])}function u(t){switch(t.type){case"GeometryCollection":t.geometries.forEach(u);break;case"Point":c(t.coordinates);break;case"MultiPoint":t.coordinates.forEach(c)}}t.arcs.forEach(function(t){for(var e,r=-1,c=t.length;++r<c;)(e=n(t[r],r))[0]<i&&(i=e[0]),e[0]>s&&(s=e[0]),e[1]<o&&(o=e[1]),e[1]>a&&(a=e[1])});for(e in t.objects)u(t.objects[e]);return[i,o,s,a]}},"../../../../vega-loader/node_modules/topojson-client/src/bisect.js":function(t,e,n){"use strict";e.a=function(t,e){for(var n=0,r=t.length;n<r;){var i=n+r>>>1;t[i]<e?n=i+1:r=i}return n}},"../../../../vega-loader/node_modules/topojson-client/src/feature.js":function(t,e,n){"use strict";e.b=s;var r=n("../../../../vega-loader/node_modules/topojson-client/src/reverse.js"),i=n("../../../../vega-loader/node_modules/topojson-client/src/transform.js");function o(t,e){var n=e.id,r=e.bbox,i=null==e.properties?{}:e.properties,o=s(t,e);return null==n&&null==r?{type:"Feature",properties:i,geometry:o}:null==r?{type:"Feature",id:n,properties:i,geometry:o}:{type:"Feature",id:n,bbox:r,properties:i,geometry:o}}function s(t,e){var n=Object(i.a)(t.transform),o=t.arcs;function s(t,e){e.length&&e.pop();for(var i=o[t<0?~t:t],s=0,a=i.length;s<a;++s)e.push(n(i[s],s));t<0&&Object(r.a)(e,a)}function a(t){return n(t)}function c(t){for(var e=[],n=0,r=t.length;n<r;++n)s(t[n],e);return e.length<2&&e.push(e[0]),e}function u(t){for(var e=c(t);e.length<4;)e.push(e[0]);return e}function l(t){return t.map(u)}return function t(e){var n,r=e.type;switch(r){case"GeometryCollection":return{type:r,geometries:e.geometries.map(t)};case"Point":n=a(e.coordinates);break;case"MultiPoint":n=e.coordinates.map(a);break;case"LineString":n=c(e.arcs);break;case"MultiLineString":n=e.arcs.map(c);break;case"Polygon":n=l(e.arcs);break;case"MultiPolygon":n=e.arcs.map(l);break;default:return null}return{type:r,coordinates:n}}(e)}e.a=function(t,e){return"GeometryCollection"===e.type?{type:"FeatureCollection",features:e.geometries.map(function(e){return o(t,e)})}:o(t,e)}},"../../../../vega-loader/node_modules/topojson-client/src/identity.js":function(t,e,n){"use strict";e.a=function(t){return t}},"../../../../vega-loader/node_modules/topojson-client/src/merge.js":function(t,e,n){"use strict";n("../../../../vega-loader/node_modules/topojson-client/src/feature.js"),n("../../../../vega-loader/node_modules/topojson-client/src/stitch.js")},"../../../../vega-loader/node_modules/topojson-client/src/mesh.js":function(t,e,n){"use strict";var r=n("../../../../vega-loader/node_modules/topojson-client/src/feature.js"),i=n("../../../../vega-loader/node_modules/topojson-client/src/stitch.js");e.a=function(t){return Object(r.b)(t,function(t,e,n){var r,o,s;if(arguments.length>1)r=function(t,e,n){var r,i=[],o=[];function s(t){var e=t<0?~t:t;(o[e]||(o[e]=[])).push({i:t,g:r})}function a(t){t.forEach(s)}function c(t){t.forEach(a)}return function t(e){switch(r=e,e.type){case"GeometryCollection":e.geometries.forEach(t);break;case"LineString":a(e.arcs);break;case"MultiLineString":case"Polygon":c(e.arcs);break;case"MultiPolygon":e.arcs.forEach(c)}}(e),o.forEach(null==n?function(t){i.push(t[0].i)}:function(t){n(t[0].g,t[t.length-1].g)&&i.push(t[0].i)}),i}(0,e,n);else for(o=0,r=new Array(s=t.arcs.length);o<s;++o)r[o]=o;return{type:"MultiLineString",arcs:Object(i.a)(t,r)}}.apply(this,arguments))}},"../../../../vega-loader/node_modules/topojson-client/src/neighbors.js":function(t,e,n){"use strict";n("../../../../vega-loader/node_modules/topojson-client/src/bisect.js")},"../../../../vega-loader/node_modules/topojson-client/src/quantize.js":function(t,e,n){"use strict";n("../../../../vega-loader/node_modules/topojson-client/src/bbox.js"),n("../../../../vega-loader/node_modules/topojson-client/src/untransform.js")},"../../../../vega-loader/node_modules/topojson-client/src/reverse.js":function(t,e,n){"use strict";e.a=function(t,e){for(var n,r=t.length,i=r-e;i<--r;)n=t[i],t[i++]=t[r],t[r]=n}},"../../../../vega-loader/node_modules/topojson-client/src/stitch.js":function(t,e,n){"use strict";e.a=function(t,e){var n={},r={},i={},o=[],s=-1;function a(t,e){for(var r in t){var i=t[r];delete e[i.start],delete i.start,delete i.end,i.forEach(function(t){n[t<0?~t:t]=1}),o.push(i)}}return e.forEach(function(n,r){var i,o=t.arcs[n<0?~n:n];o.length<3&&!o[1][0]&&!o[1][1]&&(i=e[++s],e[s]=n,e[r]=i)}),e.forEach(function(e){var n,o,s=function(e){var n,r=t.arcs[e<0?~e:e],i=r[0];t.transform?(n=[0,0],r.forEach(function(t){n[0]+=t[0],n[1]+=t[1]})):n=r[r.length-1];return e<0?[n,i]:[i,n]}(e),a=s[0],c=s[1];if(n=i[a])if(delete i[n.end],n.push(e),n.end=c,o=r[c]){delete r[o.start];var u=o===n?n:n.concat(o);r[u.start=n.start]=i[u.end=o.end]=u}else r[n.start]=i[n.end]=n;else if(n=r[c])if(delete r[n.start],n.unshift(e),n.start=a,o=i[a]){delete i[o.end];var l=o===n?n:o.concat(n);r[l.start=o.start]=i[l.end=n.end]=l}else r[n.start]=i[n.end]=n;else r[(n=[e]).start=a]=i[n.end=c]=n}),a(i,r),a(r,i),e.forEach(function(t){n[t<0?~t:t]||o.push([t])}),o}},"../../../../vega-loader/node_modules/topojson-client/src/transform.js":function(t,e,n){"use strict";var r=n("../../../../vega-loader/node_modules/topojson-client/src/identity.js");e.a=function(t){if(null==t)return r.a;var e,n,i=t.scale[0],o=t.scale[1],s=t.translate[0],a=t.translate[1];return function(t,r){r||(e=n=0);var c=2,u=t.length,l=new Array(u);for(l[0]=(e+=t[0])*i+s,l[1]=(n+=t[1])*o+a;c<u;)l[c]=t[c],++c;return l}}},"../../../../vega-loader/node_modules/topojson-client/src/untransform.js":function(t,e,n){"use strict";var r=n("../../../../vega-loader/node_modules/topojson-client/src/identity.js");e.a=function(t){if(null==t)return r.a;var e,n,i=t.scale[0],o=t.scale[1],s=t.translate[0],a=t.translate[1];return function(t,r){r||(e=n=0);var c=2,u=t.length,l=new Array(u),p=Math.round((t[0]-s)/i),d=Math.round((t[1]-a)/o);for(l[0]=p-e,e=p,l[1]=d-n,n=d;c<u;)l[c]=t[c],++c;return l}}},"../../../../vega-loader/src/formats/dsv.js":function(t,e,n){"use strict";e.b=function(t){return function(e,n){var r={delimiter:t};return o(e,n?Object(i.l)(n,r):r)}},e.a=o;var r=n("../../../../d3-dsv/index.js"),i=n("../../../../vega-util/index.js");function o(t,e){return e.header&&(t=e.header.map(i.L).join(e.delimiter)+"\n"+t),Object(r.b)(e.delimiter).parse(t+"")}},"../../../../vega-loader/src/formats/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-loader/src/formats/dsv.js"),i=n("../../../../vega-loader/src/formats/json.js"),o=n("../../../../vega-loader/src/formats/topojson.js"),s={dsv:r.a,csv:Object(r.b)(","),tsv:Object(r.b)("\t"),json:i.a,topojson:o.a};e.a=function(t,e){return arguments.length>1?(s[t]=e,this):s.hasOwnProperty(t)?s[t]:null}},"../../../../vega-loader/src/formats/json.js":function(t,e,n){"use strict";(function(t){var r=n("../../../../vega-util/index.js");e.a=function(e,n){var i,o,s,a=n&&n.property?Object(r.p)(n.property):r.r;return!Object(r.y)(e)||(s=e,"function"==typeof t&&Object(r.w)(t.isBuffer)&&t.isBuffer(s))?a(JSON.parse(e)):(i=a(e),o&&o.copy?JSON.parse(JSON.stringify(i)):i)}}).call(e,n("../../../../buffer/index.js").Buffer)},"../../../../vega-loader/src/formats/topojson.js":function(t,e,n){"use strict";var r=n("../../../../vega-loader/src/formats/json.js"),i=n("../../../../vega-loader/node_modules/topojson-client/index.js"),o=n("../../../../vega-util/index.js");e.a=function(t,e){var n,s,a;return t=Object(r.a)(t,e),n=e&&(a=e.feature)?i.a:e&&(a=e.mesh)?i.b:Object(o.k)("Missing TopoJSON feature or mesh parameter."),(s=(s=t.objects[a])?n(t,s):Object(o.k)("Invalid TopoJSON object: "+a))&&s.features||[s]}},"../../../../vega-loader/src/loader.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js"),i=n("../../../../d3-request/index.js"),o=/^([A-Za-z]+:)?\/\//,s="file://",a=["mimeType","responseType","user","password"];function c(t,e){return Object(r.l)({},t.options,e)}function u(t,e){var n=this;return n.sanitize(t,e).then(function(t){var r=t.href;return t.localFile?n.file(r):n.http(r,e)})}function l(t,e){return e=c(this,e),new Promise(function(n,i){var a,c,u,l,p={href:null};null!=t&&"string"==typeof t?(c=o.test(t),(l=e.baseURL)&&!c&&(f(t,"/")||"/"===l[l.length-1]||(t="/"+t),t=l+t),u=(a=f(t,s))||"file"===e.mode||"http"!==e.mode&&!c&&h(),a?t=t.slice(s.length):f(t,"//")&&("file"===e.defaultProtocol?(t=t.slice(2),u=!0):t=(e.defaultProtocol||"http")+":"+t),Object.defineProperty(p,"localFile",{value:!!u}),p.href=t,n(p)):i("Sanitize failure, invalid URI: "+Object(r.L)(t))})}function p(t,e){return e=c(this,e),new Promise(function(n,r){var o,s=Object(i.a)(t);for(o in e.headers)s.header(o,e.headers[o]);a.forEach(function(t){e[t]&&s[t](e[t])}),s.on("error",function(e){r(e||"Error loading URL: "+t)}).on("load",function(t){var e=t&&t.responseText;t&&0!==t.status?n(e):r(e||"Error")}).get()})}function d(t){return new Promise(function(e,n){var r=h();r?r.readFile(t,function(t,r){t?n(t):e(r)}):n("No file system access for "+t)})}function h(){var t=n(1);return t&&Object(r.w)(t.readFile)?t:null}function f(t,e){return null!=t&&0===t.lastIndexOf(e,0)}e.a=function(t){return{options:t||{},sanitize:l,load:u,file:d,http:p}}},"../../../../vega-loader/src/read.js":function(t,e,n){"use strict";var r=n("../../../../vega-loader/src/type.js"),i=n("../../../../vega-loader/src/formats/index.js"),o=n("../../../../vega-util/index.js"),s=n("../../../../d3-time-format/index.js");e.a=function(t,e,n){e=e||{};var a=Object(i.a)(e.type||"json");return a||Object(o.k)("Unknown data format type: "+e.type),t=a(t,e),e.parse&&function(t,e,n){if(!t.length)return;n=n||s.b;var i,o,a,c,u,l,p,d=t.columns||Object.keys(t[0]);"auto"===e&&(e=Object(r.a)(t,d));for(d=Object.keys(e),i=d.map(function(t){var i,o,a=e[t];if(a&&(0===a.indexOf("date:")||0===a.indexOf("utc:")))return("'"===(o=(i=a.split(/:(.+)?/,2))[1])[0]&&"'"===o[o.length-1]||'"'===o[0]&&'"'===o[o.length-1])&&(o=o.slice(1,-1)),"utc"===i[0]?Object(s.d)(o):n(o);if(!r.b[a])throw Error("Illegal format pattern: "+t+":"+a);return r.b[a]}),c=0,l=t.length,p=d.length;c<l;++c)for(o=t[c],u=0;u<p;++u)a=d[u],o[a]=i[u](o[a])}(t,e.parse,n),t.hasOwnProperty("columns")&&delete t.columns,t}},"../../../../vega-loader/src/type.js":function(t,e,n){"use strict";n.d(e,"b",function(){return i}),e.a=function(t,e){return e.reduce(function(e,n){return e[n]=function(t,e){if(!t||!t.length)return"unknown";var n,r,i,a,c=o.slice();for(r=0,i=t.length;r<i;++r){for(n=e?t[r][e]:t[r],a=0;a<c.length;++a)null==(u=n)||u!=u||c[a](n)||(c.splice(a,1),--a);if(0===c.length)return"string"}var u;return s[o.indexOf(c[0])]}(t,n),e},{})};var r=n("../../../../vega-util/index.js"),i={boolean:r.M,integer:r.O,number:r.O,date:r.N,string:r.Q,unknown:r.r},o=[function(t){return"true"===t||"false"===t||!0===t||!1===t},function(t){return a(t)&&(t=+t)==~~t},a,function(t){return!isNaN(Date.parse(t))}],s=["boolean","integer","number","date"];function a(t){return!(isNaN(+t)||t instanceof Date)}},"../../../../vega-parser/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parse.js");n.d(e,"b",function(){return r.a});n("../../../../vega-parser/src/config.js"),n("../../../../vega-parser/src/parsers/signal.js"),n("../../../../vega-parser/src/parsers/signal-updates.js"),n("../../../../vega-parser/src/parsers/stream.js");var i=n("../../../../vega-parser/src/parsers/expression/codegen.js");n.d(e,"a",function(){return i.c});n("../../../../vega-parser/src/parsers/marks/roles.js"),n("../../../../vega-parser/src/Scope.js"),n("../../../../vega-parser/src/DataScope.js"),n("../../../../d3-format/index.js"),n("../../../../d3-time-format/index.js")},"../../../../vega-parser/src/DataScope.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-parser/src/util.js"),i=n("../../../../vega-parser/src/transforms.js"),o=n("../../../../vega-util/index.js");function s(t,e,n,r,i){this.scope=t,this.input=e,this.output=n,this.values=r,this.aggregate=i,this.index={}}s.fromEntries=function(t,e){var n=e.length,i=1,o=e[0],a=e[n-1],c=e[n-2],u=null;for(t.add(e[0]);i<n;++i)e[i].params.pulse=Object(r.k)(e[i-1]),t.add(e[i]),"aggregate"===e[i].type&&(u=e[i]);return new s(t,o,c,a,u)};var a=s.prototype;function c(t){return Object(o.A)(t)?t:null}function u(t,e,n){var i,o=Object(r.c)(n.op,n.field);if(e.ops){for(var s=0,a=e.as.length;s<a;++s)if(e.as[s]===o)return}else e.ops=["count"],e.fields=[null],e.as=["count"];n.op&&(e.ops.push((i=n.op.signal)?t.signalRef(i):n.op),e.fields.push(t.fieldRef(n.field)),e.as.push(o))}function l(t,e,n,i,o,s,a){var u,l,p=e[n]||(e[n]={}),d=Object(r.l)(s),h=c(o);if(null!=h&&(t=e.scope,u=p[h+=d?"|"+d:""]),!u){var f=s?{field:r.h,pulse:e.countsRef(t,o,s)}:{field:t.fieldRef(o),pulse:Object(r.k)(e.output)};d&&(f.sort=t.sortRef(s)),l=t.add(Object(r.e)(i,void 0,f)),a&&(e.index[o]=l),u=Object(r.k)(l),null!=h&&(p[h]=u)}return u}a.countsRef=function(t,e,n){var o,s,a,l=this.counts||(this.counts={}),p=c(e);return null!=p&&(t=this.scope,o=l[p]),o?n&&n.field&&u(t,o.agg.params,n):(a={groupby:t.fieldRef(e,"key"),pulse:Object(r.k)(this.output)},n&&n.field&&u(t,a,n),s=t.add(Object(i.a)(a)),o=t.add(Object(i.d)({pulse:Object(r.k)(s)})),o={agg:s,ref:Object(r.k)(o)},null!=p&&(l[p]=o)),o.ref},a.tuplesRef=function(){return Object(r.k)(this.values)},a.extentRef=function(t,e){return l(t,this,"extent","extent",e,!1)},a.domainRef=function(t,e){return l(t,this,"domain","values",e,!1)},a.valuesRef=function(t,e,n){return l(t,this,"vals","values",e,n||!0)},a.lookupRef=function(t,e){return l(t,this,"lookup","tupleindex",e,!1)},a.indataRef=function(t,e){return l(t,this,"indata","tupleindex",e,!0,!0)}},"../../../../vega-parser/src/Scope.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../vega-parser/src/DataScope.js"),i=n("../../../../vega-parser/src/util.js"),o=n("../../../../vega-parser/src/parsers/expression.js"),s=n("../../../../vega-parser/src/transforms.js"),a=n("../../../../vega-util/index.js");function c(t){this.config=t,this.bindings=[],this.field={},this.signals={},this.lambdas={},this.scales={},this.events={},this.data={},this.streams=[],this.updates=[],this.operators=[],this.background=null,this.eventConfig=null,this._id=0,this._subid=0,this._nextsub=[0],this._parent=[],this._encode=[],this._lookup=[],this._markpath=[]}function u(t){this.config=t.config,this.field=Object.create(t.field),this.signals=Object.create(t.signals),this.lambdas=Object.create(t.lambdas),this.scales=Object.create(t.scales),this.events=Object.create(t.events),this.data=Object.create(t.data),this.streams=[],this.updates=[],this.operators=[],this._id=0,this._subid=++t._nextsub[0],this._nextsub=t._nextsub,this._parent=t._parent.slice(),this._encode=t._encode.slice(),this._lookup=t._lookup.slice(),this._markpath=t._markpath}var l=c.prototype=u.prototype;function p(t){return(Object(a.t)(t)?function(t){for(var e,n="[",r=0,i=t.length;r<i;++r)e=t[r],n+=(r>0?",":"")+(Object(a.y)(e)?e.signal||p(e):Object(a.L)(e));return n+"]"}:function(t){var e,n,r="{",i=0;for(e in t)n=t[e],r+=(++i>1?",":"")+Object(a.L)(e)+":"+(Object(a.y)(n)?n.signal||p(n):Object(a.L)(n));return r+"}"})(t)}l.fork=function(){return new u(this)},l.toRuntime=function(){return this.finish(),{background:this.background,operators:this.operators,streams:this.streams,updates:this.updates,bindings:this.bindings,eventConfig:this.eventConfig}},l.id=function(){return(this._subid?this._subid+":":0)+this._id++},l.add=function(t){return this.operators.push(t),t.id=this.id(),t.refs&&(t.refs.forEach(function(e){e.$ref=t.id}),t.refs=null),t},l.proxy=function(t){var e=t instanceof i.b?Object(i.k)(t):t;return this.add(Object(s.s)({value:e}))},l.addStream=function(t){return this.streams.push(t),t.id=this.id(),t},l.addUpdate=function(t){return this.updates.push(t),t},l.finish=function(){var t,e;this.root&&(this.root.root=!0);for(t in this.signals)this.signals[t].signal=t;for(t in this.scales)this.scales[t].scale=t;function n(t,e,n){var r;t&&((r=t.data||(t.data={}))[e]||(r[e]=[])).push(n)}for(t in this.data){n((e=this.data[t]).input,t,"input"),n(e.output,t,"output"),n(e.values,t,"values");for(var r in e.index)n(e.index[r],t,"index:"+r)}return this},l.pushState=function(t,e,n){this._encode.push(Object(i.k)(this.add(Object(s.w)({pulse:t})))),this._parent.push(e),this._lookup.push(n?Object(i.k)(this.proxy(n)):null),this._markpath.push(-1)},l.popState=function(){this._encode.pop(),this._parent.pop(),this._lookup.pop(),this._markpath.pop()},l.parent=function(){return Object(a.J)(this._parent)},l.encode=function(){return Object(a.J)(this._encode)},l.lookup=function(){return Object(a.J)(this._lookup)},l.markpath=function(){var t=this._markpath;return++t[t.length-1]},l.fieldRef=function(t,e){if(Object(a.A)(t))return Object(i.f)(t,e);t.signal||Object(a.k)("Unsupported field reference: "+Object(a.L)(t));var n,r=t.signal,o=this.field[r];return o||(n={name:this.signalRef(r)},e&&(n.as=e),this.field[r]=o=Object(i.k)(this.add(Object(s.i)(n)))),o},l.compareRef=function(t,e){function n(t){return Object(i.g)(t)?(o=!0,Object(i.k)(r[t.signal])):t}var r=this.signals,o=!1,c=Object(a.g)(t.field).map(n),u=Object(a.g)(t.order).map(n);return e&&c.push(i.m),o?Object(i.k)(this.add(Object(s.e)({fields:c,orders:u}))):Object(i.d)(c,u)},l.keyRef=function(t,e){var n=this.signals,r=!1;return t=Object(a.g)(t).map(function(t){return Object(i.g)(t)?(r=!0,Object(i.k)(n[t.signal])):t}),r?Object(i.k)(this.add(Object(s.j)({fields:t,flat:e}))):Object(i.i)(t,e)},l.sortRef=function(t){if(!t)return t;var e=[Object(i.c)(t.op,t.field),i.m],n=t.order||i.a;return n.signal?Object(i.k)(this.add(Object(s.e)({fields:e,orders:[n=this.signalRef(n.signal),n]}))):Object(i.d)(e,[n,n])},l.event=function(t,e){var n=t+":"+e;if(!this.events[n]){var r=this.id();this.streams.push({id:r,source:t,type:e}),this.events[n]=r}return this.events[n]},l.addSignal=function(t,e){this.signals.hasOwnProperty(t)&&Object(a.k)("Duplicate signal name: "+Object(a.L)(t));var n=e instanceof i.b?e:this.add(Object(i.j)(e));return this.signals[t]=n},l.getSignal=function(t){return this.signals[t]||Object(a.k)("Unrecognized signal name: "+Object(a.L)(t)),this.signals[t]},l.signalRef=function(t){return this.signals[t]?Object(i.k)(this.signals[t]):(this.lambdas.hasOwnProperty(t)||(this.lambdas[t]=this.add(Object(i.j)(null))),Object(i.k)(this.lambdas[t]))},l.parseLambdas=function(){for(var t=Object.keys(this.lambdas),e=0,n=t.length;e<n;++e){var r=t[e],i=Object(o.a)(r,this),s=this.lambdas[r];s.params=i.$params,s.update=i.$expr}},l.property=function(t){return t&&t.signal?this.signalRef(t.signal):t},l.objectProperty=function(t){return t&&Object(a.y)(t)?this.signalRef(t.signal||p(t)):t},l.addBinding=function(t,e){this.bindings||Object(a.k)("Nested signals do not support binding: "+Object(a.L)(t)),this.bindings.push(Object(a.l)({signal:t},e))},l.addScaleProj=function(t,e){this.scales.hasOwnProperty(t)&&Object(a.k)("Duplicate scale or projection name: "+Object(a.L)(t)),this.scales[t]=this.add(e)},l.addScale=function(t,e){this.addScaleProj(t,Object(s.v)(e))},l.addProjection=function(t,e){this.addScaleProj(t,Object(s.r)(e))},l.getScale=function(t){return this.scales[t]||Object(a.k)("Unrecognized scale name: "+Object(a.L)(t)),this.scales[t]},l.projectionRef=l.scaleRef=function(t){return Object(i.k)(this.getScale(t))},l.projectionType=l.scaleType=function(t){return this.getScale(t).params.type},l.addData=function(t,e){return this.data.hasOwnProperty(t)&&Object(a.k)("Duplicate data set name: "+Object(a.L)(t)),this.data[t]=e},l.getData=function(t){return this.data[t]||Object(a.k)("Undefined data set name: "+Object(a.L)(t)),this.data[t]},l.addDataPipeline=function(t,e){return this.data.hasOwnProperty(t)&&Object(a.k)("Duplicate data set name: "+Object(a.L)(t)),this.addData(t,r.a.fromEntries(this,e))}},"../../../../vega-parser/src/config.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");e.a=function(t){var e={padding:0,autosize:"pad",background:null,events:{defaults:{allow:["wheel"]}},group:null,mark:null,arc:{fill:a},area:{fill:a},image:null,line:{stroke:a,strokeWidth:s},path:{stroke:a},rect:{fill:a},rule:{stroke:c},shape:{stroke:a},symbol:{fill:a,size:64},text:{fill:c,font:i,fontSize:11},style:{"guide-label":{fill:c,font:i,fontSize:10},"guide-title":{fill:c,font:i,fontSize:11,fontWeight:"bold"},"group-title":{fill:c,font:i,fontSize:13,fontWeight:"bold"},point:{size:o,strokeWidth:s,shape:"circle"},circle:{size:o,strokeWidth:s},square:{size:o,strokeWidth:s,shape:"square"},cell:{fill:"transparent",stroke:l}},axis:{minExtent:0,maxExtent:200,bandPosition:.5,domain:!0,domainWidth:1,domainColor:u,grid:!1,gridWidth:1,gridColor:l,gridOpacity:1,labels:!0,labelAngle:0,labelLimit:180,labelPadding:2,ticks:!0,tickColor:u,tickOffset:0,tickRound:!0,tickSize:5,tickWidth:1,titleAlign:"center",titlePadding:4},axisBand:{tickOffset:-1},legend:{orient:"right",offset:18,padding:0,entryPadding:5,titlePadding:5,gradientWidth:100,gradientHeight:20,gradientStrokeColor:l,gradientStrokeWidth:0,gradientLabelBaseline:"top",gradientLabelOffset:2,labelAlign:"left",labelBaseline:"middle",labelOffset:8,labelLimit:160,symbolType:"circle",symbolSize:100,symbolFillColor:"transparent",symbolStrokeColor:u,symbolStrokeWidth:1.5,titleAlign:"left",titleBaseline:"top",titleLimit:180},title:{orient:"top",anchor:"middle",offset:4},range:{category:{scheme:"tableau10"},ordinal:{scheme:"blues",extent:[.2,1]},heatmap:{scheme:"viridis"},ramp:{scheme:"blues",extent:[.2,1]},diverging:{scheme:"blueorange"},symbol:["circle","square","triangle-up","cross","diamond","triangle-right","triangle-down","triangle-left"]}};return(t||[]).forEach(function(t){var n,i,o;if(t)for(n in t)if("style"===n){o=e.style||(e.style={});for(n in t.style)o[n]=Object(r.l)(o[n]||{},t.style[n])}else i=t[n],e[n]=Object(r.y)(i)&&!Object(r.t)(i)?Object(r.l)(Object(r.y)(e[n])?e[n]:{},i):i}),e};var i="sans-serif",o=30,s=2,a="#4c78a8",c="#000",u="#888",l="#ddd"},"../../../../vega-parser/src/parse.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/view.js"),i=n("../../../../vega-parser/src/Scope.js"),o=n("../../../../vega-parser/src/config.js"),s=n("../../../../vega-util/index.js");e.a=function(t,e){return Object(s.y)(t)||Object(s.k)("Input Vega specification must be an object."),Object(r.a)(t,new i.a(Object(o.a)([e,t.config]))).toRuntime()}},"../../../../vega-parser/src/parsers/autosize.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");e.a=function(t,e){return t=t||e.autosize,Object(r.y)(t)?t:{type:t=t||"pad"}}},"../../../../vega-parser/src/parsers/axis.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/axis-config.js"),i=n("../../../../vega-parser/src/parsers/guides/axis-domain.js"),o=n("../../../../vega-parser/src/parsers/guides/axis-grid.js"),s=n("../../../../vega-parser/src/parsers/guides/axis-ticks.js"),a=n("../../../../vega-parser/src/parsers/guides/axis-labels.js"),c=n("../../../../vega-parser/src/parsers/guides/axis-title.js"),u=n("../../../../vega-parser/src/parsers/guides/guide-group.js"),l=n("../../../../vega-parser/src/parsers/marks/roles.js"),p=n("../../../../vega-parser/src/parsers/mark.js"),d=n("../../../../vega-parser/src/parsers/encode/encode-util.js"),h=n("../../../../vega-parser/src/parsers/guides/constants.js"),f=n("../../../../vega-parser/src/util.js"),m=n("../../../../vega-parser/src/transforms.js");e.a=function(t,e){var n,g,v,y,b,_,j=Object(r.a)(t,e),x=t.encode||{},w=x.axis||{},O=w.name||void 0,C=w.interactive,k=w.style;return n={orient:t.orient,ticks:!!Object(f.n)(t.ticks,j.ticks),labels:!!Object(f.n)(t.labels,j.labels),grid:!!Object(f.n)(t.grid,j.grid),domain:!!Object(f.n)(t.domain,j.domain),title:!!Object(f.n)(t.title,!1)},g=Object(f.k)(e.add(Object(m.d)({},[n]))),w=Object(d.d)({update:{range:{signal:'abs(span(range("'+t.scale+'")))'},offset:Object(d.b)(Object(f.n)(t.offset,0)),position:Object(d.b)(Object(f.n)(t.position,0)),titlePadding:Object(d.b)(Object(f.n)(t.titlePadding,j.titlePadding)),minExtent:Object(d.b)(Object(f.n)(t.minExtent,j.minExtent)),maxExtent:Object(d.b)(Object(f.n)(t.maxExtent,j.maxExtent))}},x.axis,h.m),v=Object(f.k)(e.add(Object(m.b)({scale:e.scaleRef(t.scale),extra:j.tickExtra,count:e.objectProperty(t.tickCount),values:e.objectProperty(t.values),formatSpecifier:e.property(t.format)}))),_=[],n.grid&&_.push(Object(o.a)(t,j,x.grid,v)),n.ticks&&(y=Object(f.n)(t.tickSize,j.tickSize),_.push(Object(s.a)(t,j,x.ticks,v,y))),n.labels&&(y=n.ticks?y:0,_.push(Object(a.a)(t,j,x.labels,v,y))),n.domain&&_.push(Object(i.a)(t,j,x.domain,g)),n.title&&_.push(Object(c.a)(t,j,x.title,g)),b=Object(u.a)(l.d,k,O,g,C,w,_),t.zindex&&(b.zindex=t.zindex),Object(p.a)(b,e)}},"../../../../vega-parser/src/parsers/data.js":function(t,e,n){"use strict";e.a=function(t,e){var n=[];t.transform&&t.transform.forEach(function(t){n.push(Object(r.a)(t,e))});t.on&&t.on.forEach(function(n){Object(i.a)(n,e,t.name)});e.addDataPipeline(t.name,function(t,e,n){var r,i,u,l,p,d=[],h=null,f=!1,m=!1;t.values?d.push(h=c({$ingest:t.values,$format:t.format})):t.url?d.push(h=c({$request:t.url,$format:t.format})):t.source&&(h=r=Object(a.g)(t.source).map(function(t){return Object(o.k)(e.getData(t).output)}),d.push(null));for(i=0,u=n.length;i<u;++i)l=n[i],p=l.metadata,h||p.source||d.push(h=c()),d.push(l),p.generates&&(m=!0),p.modifies&&!m&&(f=!0),p.source?h=l:p.changes&&(h=null);r&&(u=r.length-1,d[0]=Object(s.t)({derive:f,pulse:u?r:r[0]}),(f||u)&&d.splice(1,0,c()));h||d.push(c());return d.push(Object(s.w)({})),d}(t,e,n))};var r=n("../../../../vega-parser/src/parsers/transform.js"),i=n("../../../../vega-parser/src/parsers/trigger.js"),o=n("../../../../vega-parser/src/util.js"),s=n("../../../../vega-parser/src/transforms.js"),a=n("../../../../vega-util/index.js");function c(t){var e=Object(s.d)({},t);return e.metadata={source:!0},e}},"../../../../vega-parser/src/parsers/encode.js":function(t,e,n){"use strict";e.a=function(t,e,n,c){var u,l,p,d={},h="var o=item,datum=o.datum,$;";for(u in t)l=t[u],Object(a.t)(l)?h+=Object(o.a)(u,l,c,n,d):(p=Object(i.a)(u,l,c,n,d),h+=Object(s.a)("o",u,p));return h+=Object(r.a)(t,e),{$expr:h+="return 1;",$fields:Object.keys(d),$output:Object.keys(t)}};var r=n("../../../../vega-parser/src/parsers/encode/adjust-spatial.js"),i=n("../../../../vega-parser/src/parsers/encode/entry.js"),o=n("../../../../vega-parser/src/parsers/encode/rule.js"),s=n("../../../../vega-parser/src/parsers/encode/set.js"),a=n("../../../../vega-util/index.js")},"../../../../vega-parser/src/parsers/encode/adjust-spatial.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js"),i=Object(r.P)(["rule"]),o=Object(r.P)(["group","image","rect"]);e.a=function(t,e){var n="";return i[e]?n:(t.x2&&(t.x?(o[e]&&(n+="if(o.x>o.x2)$=o.x,o.x=o.x2,o.x2=$;"),n+="o.width=o.x2-o.x;"):n+="o.x=o.x2-(o.width||0);"),t.xc&&(n+="o.x=o.xc-(o.width||0)/2;"),t.y2&&(t.y?(o[e]&&(n+="if(o.y>o.y2)$=o.y,o.y=o.y2,o.y2=$;"),n+="o.height=o.y2-o.y;"):n+="o.y=o.y2-(o.height||0);"),t.yc&&(n+="o.y=o.yc-(o.height||0)/2;"),n)}},"../../../../vega-parser/src/parsers/encode/color.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/encode/entry.js");e.a=function(t,e,n,i){function o(t,o,s,a){return"this."+t+"("+[Object(r.a)(null,o,e,n,i),Object(r.a)(null,s,e,n,i),Object(r.a)(null,a,e,n,i)].join(",")+").toString()"}return t.c?o("hcl",t.h,t.c,t.l):t.h||t.s?o("hsl",t.h,t.s,t.l):t.l||t.a?o("lab",t.l,t.a,t.b):t.r||t.g||t.b?o("rgb",t.r,t.g,t.b):null}},"../../../../vega-parser/src/parsers/encode/encode-util.js":function(t,e,n){"use strict";e.b=function(t){return Object(o.y)(t)?t:{value:t}},e.a=function(t,e,n){return null!=n?(t[e]=Object(o.y)(n)&&!Object(o.t)(n)?n:{value:n},1):0},e.d=function(t,e,n){for(var r in e)n&&n.hasOwnProperty(r)||(t[r]=Object(o.l)(t[r]||{},e[r]));return t},e.c=function(t,e,n,a,c,u){var l,p;(u=u||{}).encoders={$encode:l={}},t=function(t,e,n,r,a){var c,u,l={};"legend"!=n&&0!==String(n).indexOf("axis")||(n=null);u=n===i.g?a.group:n===i.n?Object(o.l)({},a.mark,a[e]):null;for(c in u)s(c,t)||("fill"===c||"stroke"===c)&&(s("fill",t)||s("stroke",t))||(l[c]={value:u[c]});return Object(o.g)(r).forEach(function(e){var n=a.style&&a.style[e];for(var r in n)s(r,t)||(l[r]={value:n[r]})}),(t=Object(o.l)({},t)).enter=Object(o.l)(l,t.enter),t}(t,e,n,a,c.config);for(p in t)l[p]=Object(r.a)(t[p],e,u,c);return u},e.e=s;var r=n("../../../../vega-parser/src/parsers/encode.js"),i=n("../../../../vega-parser/src/parsers/marks/roles.js"),o=n("../../../../vega-util/index.js");function s(t,e){return e&&(e.enter&&e.enter[t]||e.update&&e.update[t])}},"../../../../vega-parser/src/parsers/encode/entry.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/encode/color.js"),i=n("../../../../vega-parser/src/parsers/encode/field.js"),o=n("../../../../vega-parser/src/parsers/encode/scale.js"),s=n("../../../../vega-parser/src/parsers/encode/gradient.js"),a=n("../../../../vega-parser/src/parsers/encode/property.js"),c=n("../../../../vega-parser/src/parsers/encode/expression.js"),u=n("../../../../vega-util/index.js");e.a=function(t,e,n,l,p){if(null!=e.gradient)return Object(s.a)(e,n,l,p);var d=e.signal?Object(c.a)(e.signal,n,l,p):e.color?Object(r.a)(e.color,n,l,p):null!=e.field?Object(i.a)(e.field,n,l,p):void 0!==e.value?Object(u.L)(e.value):void 0;return null!=e.scale&&(d=Object(o.a)(e,d,n,l,p)),void 0===d&&(d=null),null!=e.exponent&&(d="Math.pow("+d+","+Object(a.a)(e.exponent,n,l,p)+")"),null!=e.mult&&(d+="*"+Object(a.a)(e.mult,n,l,p)),null!=e.offset&&(d+="+"+Object(a.a)(e.offset,n,l,p)),e.round&&(d="Math.round("+d+")"),d}},"../../../../vega-parser/src/parsers/encode/expression.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/expression.js"),i=n("../../../../vega-util/index.js");e.a=function(t,e,n,o){var s=Object(r.a)(t,e);return s.$fields.forEach(function(t){o[t]=1}),Object(i.l)(n,s.$params),s.$expr}},"../../../../vega-parser/src/parsers/encode/field.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/encode/expression.js"),i=n("../../../../vega-util/index.js");e.a=function(t,e,n,o){return function t(e,n,o,s){var a,c,u;if(e.signal)a="datum",u=Object(r.a)(e.signal,n,o,s);else if(e.group||e.parent){for(c=Math.max(1,e.level||1),a="item";c-- >0;)a+=".mark.group";e.parent?(u=e.parent,a+=".datum"):u=e.group}else e.datum?(a="datum",u=e.datum):Object(i.k)("Invalid field reference: "+Object(i.L)(e));e.signal||(Object(i.A)(u)?(s[u]=1,u=Object(i.K)(u).map(i.L).join("][")):u=t(u,n,o,s));return a+"["+u+"]"}(Object(i.y)(t)?t:{datum:t},e,n,o)}},"../../../../vega-parser/src/parsers/encode/gradient.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/encode/scale.js"),i=n("../../../../vega-util/index.js");e.a=function(t,e,n,o){return"this.gradient("+Object(r.b)(t.gradient,e,n,o)+","+Object(i.L)(t.start)+","+Object(i.L)(t.stop)+","+Object(i.L)(t.count)+")"}},"../../../../vega-parser/src/parsers/encode/property.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/encode/entry.js"),i=n("../../../../vega-util/index.js");e.a=function(t,e,n,o){return Object(i.y)(t)?"("+Object(r.a)(null,t,e,n,o)+")":t}},"../../../../vega-parser/src/parsers/encode/rule.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/encode/entry.js"),i=n("../../../../vega-parser/src/parsers/encode/set.js"),o=n("../../../../vega-parser/src/parsers/encode/expression.js");e.a=function(t,e,n,s,a){var c="";return e.forEach(function(e){var i=Object(r.a)(t,e,n,s,a);c+=e.test?Object(o.a)(e.test,n,s,a)+"?"+i+":":i}),Object(i.a)("o",t,c)}},"../../../../vega-parser/src/parsers/encode/scale.js":function(t,e,n){"use strict";e.b=a;var r=n("../../../../vega-parser/src/parsers/encode/expression.js"),i=n("../../../../vega-parser/src/parsers/encode/field.js"),o=n("../../../../vega-parser/src/parsers/expression/prefixes.js"),s=n("../../../../vega-util/index.js");function a(t,e,n,a){var c;if(Object(s.A)(t))c=o.c+t,n.hasOwnProperty(c)||(n[c]=e.scaleRef(t)),c=Object(s.L)(c);else{for(c in e.scales)n[o.c+c]=e.scaleRef(c);c=Object(s.L)(o.c)+"+"+(t.signal?"("+Object(r.a)(t.signal,e,n,a)+")":Object(i.a)(t,e,n,a))}return"_["+c+"]"}e.a=function(t,e,n,r,i){var o,c,u,l=a(t.scale,n,r,i);return null!=t.range?(c=l+".range()",e=0===(o=+t.range)?c+"[0]":"($="+c+","+(1===o?"$[$.length-1]":"$[0]+"+o+"*($[$.length-1]-$[0])")+")"):(void 0!==e&&(e=l+"("+e+")"),t.band&&(u=function(t,e){if(!Object(s.A)(t))return-1;var n=e.scaleType(t);return"band"===n||"point"===n?1:0}(t.scale,n))&&(o=(c=l+".bandwidth")+"()"+(1===(o=+t.band)?"":"*"+o),u<0&&(o="("+c+"?"+o+":0)"),e=(e?e+"+":"")+o,t.extra&&(e="(datum.extra?"+l+"(datum.extra.value):"+e+")")),null==e&&(e="0")),e}},"../../../../vega-parser/src/parsers/encode/set.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");e.a=function(t,e,n){return t+"["+Object(r.L)(e)+"]="+n+";"}},"../../../../vega-parser/src/parsers/expression.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/expression/codegen.js"),i=n("../../../../vega-expression/index.js"),o=n("../../../../vega-util/index.js");e.a=function(t,e,n){var s,a,c={};try{s=Object(i.d)(t)}catch(e){Object(o.k)("Expression parse error: "+Object(o.L)(t))}return s.visit(function(t){if("CallExpression"===t.type){var n=t.callee.name,i=r.b.visitors[n];i&&i(n,t.arguments,e,c)}}),(a=Object(r.a)(s)).globals.forEach(function(t){var n="$"+t;!c.hasOwnProperty(n)&&e.getSignal(t)&&(c[n]=e.signalRef(t))}),{$expr:n?n+"return("+a.code+");":a.code,$fields:a.fields,$params:c}}},"../../../../vega-parser/src/parsers/expression/ast.js":function(t,e,n){"use strict";n.d(e,"b",function(){return r}),n.d(e,"a",function(){return i});var r="Literal",i="Identifier"},"../../../../vega-parser/src/parsers/expression/clamp-range.js":function(t,e,n){"use strict";e.a=function(t,e,n){var r,i=t[0],o=t[1];return o<i&&(r=o,o=i,i=r),(r=o-i)>=n-e?[e,n]:[Math.min(Math.max(i,e),n-r),Math.min(Math.max(o,r),n)]}},"../../../../vega-parser/src/parsers/expression/codegen.js":function(t,e,n){"use strict";n.d(e,"c",function(){return S}),n.d(e,"b",function(){return I}),n.d(e,"a",function(){return R});var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-expression/index.js"),o=n("../../../../vega-statistics/index.js"),s=n("../../../../vega-util/index.js"),a=n("../../../../d3-color/index.js"),c=n("../../../../d3-array/index.js"),u=n("../../../../vega-parser/src/parsers/expression/format.js"),l=n("../../../../vega-parser/src/parsers/expression/quarter.js"),p=n("../../../../vega-parser/src/parsers/expression/log.js"),d=n("../../../../vega-parser/src/parsers/expression/inscope.js"),h=n("../../../../vega-parser/src/parsers/expression/clamp-range.js"),f=n("../../../../vega-parser/src/parsers/expression/pinch.js"),m=n("../../../../vega-parser/src/parsers/expression/window.js"),g=n("../../../../vega-parser/src/parsers/expression/flush.js"),v=n("../../../../vega-parser/src/parsers/expression/span.js"),y=n("../../../../vega-parser/src/parsers/expression/scale.js"),b=n("../../../../vega-parser/src/parsers/expression/scale-gradient.js"),_=n("../../../../vega-parser/src/parsers/expression/geo.js"),j=n("../../../../vega-parser/src/parsers/expression/shape.js"),x=n("../../../../vega-parser/src/parsers/expression/data.js"),w=n("../../../../vega-parser/src/parsers/expression/tree.js"),O=n("../../../../vega-parser/src/parsers/expression/inrange.js"),C=n("../../../../vega-parser/src/parsers/expression/encode.js"),k=n("../../../../vega-parser/src/parsers/expression/modify.js"),E=n("../../../../vega-parser/src/parsers/expression/selection.js"),S={random:function(){return Object(o.d)()},isArray:s.t,isBoolean:s.u,isDate:s.v,isNumber:s.x,isObject:s.y,isRegExp:s.z,isString:s.A,isTuple:r.j,toBoolean:s.M,toDate:s.N,toNumber:s.O,toString:s.Q,pad:s.F,peek:s.J,truncate:s.R,rgb:a.f,lab:a.e,hcl:a.c,hsl:a.d,sequence:c.n,format:u.c,utcFormat:u.h,utcParse:u.i,timeFormat:u.f,timeParse:u.g,monthFormat:u.e,monthAbbrevFormat:u.d,dayFormat:u.b,dayAbbrevFormat:u.a,quarter:l.a,utcquarter:l.b,warn:p.c,info:p.b,debug:p.a,inScope:d.a,clampRange:h.a,pinchDistance:f.b,pinchAngle:f.a,screen:m.b,containerSize:m.a,windowSize:m.c,span:v.a,flush:g.a,bandspace:y.a,inrange:O.a,setdata:x.e,pathShape:j.b,panLinear:s.G,panLog:s.H,panPow:s.I,zoomLinear:s.V,zoomLog:s.W,zoomPow:s.X,encode:C.a,modify:k.a},P=["view","item","group","xy","x","y"],M="event.vega.",A="this.",T={};function D(t,e,n){return 1===arguments.length?S[t]:(S[t]=e,n&&(T[t]=n),R&&(R.functions[t]=A+t),this)}D("bandwidth",y.b,y.i),D("copy",y.c,y.i),D("domain",y.d,y.i),D("range",y.g,y.i),D("invert",y.f,y.i),D("scale",y.h,y.i),D("gradient",b.a,y.i),D("geoArea",_.a,y.i),D("geoBounds",_.b,y.i),D("geoCentroid",_.c,y.i),D("geoShape",j.a,y.i),D("indata",x.c,x.d),D("data",x.a,x.b),D("vlSingle",E.d,x.b),D("vlSingleDomain",E.e,x.b),D("vlMulti",E.d,E.c),D("vlMultiDomain",E.e,E.c),D("vlInterval",E.a,x.b),D("vlIntervalDomain",E.b,x.b),D("treePath",w.b,x.b),D("treeAncestors",w.a,x.b);var I={blacklist:["_"],whitelist:["datum","event","item"],fieldvar:"datum",globalvar:function(t){return"_["+Object(s.L)("$"+t)+"]"},functions:function(t){var e=Object(i.c)(t);P.forEach(function(t){e[t]=M+t});for(var n in S)e[n]=A+n;return e},constants:i.b,visitors:T},R=Object(i.a)(I)},"../../../../vega-parser/src/parsers/expression/data.js":function(t,e,n){"use strict";e.a=function(t){var e=this.context.data[t];return e?e.values.value:[]},e.b=function(t,e,n,s){e[0].type!==r.b&&Object(o.k)("First argument to data functions must be a string literal.");var a=e[0].value,c=i.a+a;s.hasOwnProperty(c)||(s[c]=n.getData(a).tuplesRef())},e.c=function(t,e,n){var r=this.context.data[t]["index:"+e],i=r?r.value.get(n):void 0;return i?i.count:i},e.d=function(t,e,n,s){e[0].type!==r.b&&Object(o.k)("First argument to indata must be a string literal.");e[1].type!==r.b&&Object(o.k)("Second argument to indata must be a string literal.");var a=e[0].value,c=e[1].value,u=i.b+c;s.hasOwnProperty(u)||(s[u]=n.getData(a).indataRef(n,c))},e.e=function(t,e){var n=this.context.dataflow,r=this.context.data[t].input;return n.pulse(r,n.changeset().remove(o.S).insert(e)),1};var r=n("../../../../vega-parser/src/parsers/expression/ast.js"),i=n("../../../../vega-parser/src/parsers/expression/prefixes.js"),o=n("../../../../vega-util/index.js")},"../../../../vega-parser/src/parsers/expression/encode.js":function(t,e,n){"use strict";e.a=function(t,e,n){if(t){var r=this.context.dataflow,i=t.mark.source;r.pulse(i,r.changeset().encode(t,e))}return void 0!==n?n:t}},"../../../../vega-parser/src/parsers/expression/flush.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");e.a=function(t,e,n,i,o,s){var a=Math.abs(e-t[0]),c=Math.abs(Object(r.J)(t)-e);return a<c&&a<=n?i:c<=n?o:s}},"../../../../vega-parser/src/parsers/expression/format.js":function(t,e,n){"use strict";e.c=function(t,e){return s("format",r.a,e)(t)},e.f=a,e.h=function(t,e){return s("utcFormat",i.c,e)(t)},e.g=function(t,e){return s("timeParse",i.b,e)(t)},e.i=function(t,e){return s("utcParse",i.d,e)(t)},e.e=function(t){return u(t,1,"%B")},e.d=function(t){return u(t,1,"%b")},e.b=function(t){return u(0,2+t,"%A")},e.a=function(t){return u(0,2+t,"%a")};var r=n("../../../../d3-format/index.js"),i=n("../../../../d3-time-format/index.js"),o={};function s(t,e,n){var r=t+":"+n,i=o[r];return i&&i[0]===e||(o[r]=i=[e,e(n)]),i[1]}function a(t,e){return s("timeFormat",i.a,e)(t)}var c=new Date(2e3,0,1);function u(t,e,n){return c.setMonth(t),c.setDate(e),a(c,n)}},"../../../../vega-parser/src/parsers/expression/geo.js":function(t,e,n){"use strict";n.d(e,"a",function(){return s}),n.d(e,"b",function(){return a}),n.d(e,"c",function(){return c});var r=n("../../../../d3-geo/index.js"),i=n("../../../../vega-parser/src/parsers/expression/scale.js");function o(t,e){return function(n,r,o){if(n){var s=Object(i.e)(n,(o||this).context);return s&&s.path[t](r)}return e(r)}}var s=o("area",r.c),a=o("bounds",r.f),c=o("centroid",r.g)},"../../../../vega-parser/src/parsers/expression/inrange.js":function(t,e,n){"use strict";e.a=function(t,e,n,r){var i,o=e[0],s=e[e.length-1];return o>s&&(i=o,o=s,s=i),n=void 0===n||n,r=void 0===r||r,(n?o<=t:o<t)&&(r?t<=s:t<s)}},"../../../../vega-parser/src/parsers/expression/inscope.js":function(t,e,n){"use strict";e.a=function(t){var e=this.context.group,n=!1;if(e)for(;t;){if(t===e){n=!0;break}t=t.mark.group}return n}},"../../../../vega-parser/src/parsers/expression/log.js":function(t,e,n){"use strict";function r(t,e,n){try{t[e].apply(t,["EXPRESSION"].concat([].slice.call(n)))}catch(e){t.warn(e)}return n[n.length-1]}e.c=function(){return r(this.context.dataflow,"warn",arguments)},e.b=function(){return r(this.context.dataflow,"info",arguments)},e.a=function(){return r(this.context.dataflow,"debug",arguments)}},"../../../../vega-parser/src/parsers/expression/modify.js":function(t,e,n){"use strict";var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t,e){return t===e||t!=t&&e!=e||!(!Object(i.t)(t)||!Object(i.t)(e)||t.length!==e.length)&&function(t,e){for(var n=0,r=t.length;n<r;++n)if(!o(t[n],e[n]))return!1;return!0}(t,e)}function s(t){return function(e){for(var n in t)if(!o(e[n],t[n]))return!1;return!0}}e.a=function(t,e,n,o,a,c){var u,l,p=this.context.dataflow,d=this.context.data[t],h=d.input,f=d.changes,m=p.stamp();if(!1===p._trigger||!(h.value.length||e||o))return 0;if((!f||f.stamp<m)&&(d.changes=f=p.changeset(),f.stamp=m,p.runAfter(function(){d.modified=!0,p.pulse(h,f).run()},!0,1)),n&&(u=!0===n?i.S:Object(i.t)(n)||Object(r.j)(n)?n:s(n),f.remove(u)),e&&f.insert(e),o&&(u=s(o),h.value.some(u)?f.remove(u):f.insert(o)),a)for(l in c)f.modify(a,l,c[l]);return 1}},"../../../../vega-parser/src/parsers/expression/pinch.js":function(t,e,n){"use strict";e.b=function(t){var e=t.touches,n=e[0].clientX-e[1].clientX,r=e[0].clientY-e[1].clientY;return Math.sqrt(n*n+r*r)},e.a=function(t){var e=t.touches;return Math.atan2(e[0].clientY-e[1].clientY,e[0].clientX-e[1].clientX)}},"../../../../vega-parser/src/parsers/expression/prefixes.js":function(t,e,n){"use strict";n.d(e,"b",function(){return r}),n.d(e,"c",function(){return i}),n.d(e,"a",function(){return o});var r="@",i="%",o=":"},"../../../../vega-parser/src/parsers/expression/quarter.js":function(t,e,n){"use strict";e.a=function(t){return 1+~~(new Date(t).getMonth()/3)},e.b=function(t){return 1+~~(new Date(t).getUTCMonth()/3)}},"../../../../vega-parser/src/parsers/expression/scale-gradient.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/expression/scale.js"),i=n("../../../../vega-scenegraph/index.js"),o=n("../../../../vega-scale/index.js");e.a=function(t,e,n,s,a){t=Object(r.e)(t,(a||this).context);var c=Object(i.c)(e,n),u=t.domain(),l=u[0],p=u[u.length-1],d=Object(o.e)(t,l,p);t.ticks&&(l!==(u=t.ticks(+s||15))[0]&&u.unshift(l),p!==u[u.length-1]&&u.push(p));for(var h=0,f=u.length;h<f;++h)c.stop(d(u[h]),t(u[h]));return c}},"../../../../vega-parser/src/parsers/expression/scale.js":function(t,e,n){"use strict";e.e=a,e.i=function(t,e,n,i){if(e[0].type===r.b)c(n,i,e[0].value);else if(e[0].type===r.a)for(t in n.scales)c(n,i,t)},e.g=function(t,e){var n=a(t,(e||this).context);return n&&n.range?n.range():[]},e.d=function(t,e){var n=a(t,(e||this).context);return n?n.domain():[]},e.b=function(t,e){var n=a(t,(e||this).context);return n&&n.bandwidth?n.bandwidth():0},e.a=function(t,e,n){return Object(o.a)(t||0,e||0,n||0)},e.c=function(t,e){var n=a(t,(e||this).context);return n?n.copy():void 0},e.h=function(t,e,n){var r=a(t,(n||this).context);return r?r(e):void 0},e.f=function(t,e,n){var r=a(t,(n||this).context);return r?Object(s.t)(e)?(r.invertRange||r.invert)(e):(r.invert||r.invertExtent)(e):void 0};var r=n("../../../../vega-parser/src/parsers/expression/ast.js"),i=n("../../../../vega-parser/src/parsers/expression/prefixes.js"),o=n("../../../../vega-scale/index.js"),s=n("../../../../vega-util/index.js");function a(t,e){var n;return Object(s.w)(t)?t:Object(s.A)(t)?(n=e.scales[t])&&n.value:void 0}function c(t,e,n){var r=i.c+n;if(!e.hasOwnProperty(r))try{e[r]=t.scaleRef(n)}catch(t){}}},"../../../../vega-parser/src/parsers/expression/selection.js":function(t,e,n){"use strict";e.d=function(t,e,n){return f.call(this,t,e,n,d)},e.a=function(t,e,n){return f.call(this,t,e,n,h)},e.c=function(t,e,n,o){e[0].type!==r.b&&Object(a.k)("First argument to indata must be a string literal.");var c=e[0].value,l=e.length>=2&&e[e.length-1].value,p=s.b+"unit";l!==u||o.hasOwnProperty(p)||(o[p]=n.getData(c).indataRef(n,"unit"));Object(i.b)(t,e,n,o)},e.e=function(t,e,n,r){var i,o,s,a,d,h=this.context.data[t],f=h?h.values.value:[],v=h?h[p]&&h[p].value:void 0,y=f[0],b=0;if(!y)return;for(i=e?y.encodings.length:y.fields.length;b<i;++b)if(e&&y.encodings[b]===e||n&&y.fields[b]===n){o=b,a=y[c+y.fields[b]];break}v&&1===v.size&&(r=l);v&&r===u?(d=f.reduce(function(t,e){var n=t[e.unit]||(t[e.unit]=[]);return n.push({unit:e.unit,value:e.values[o]}),t},{}),s=Object.keys(d).map(function(t){return{unit:t,value:a?g(d[t],l):m(d[t],l)}})):s=f.map(function(t){return{unit:t.unit,value:t.values[o]}});return a?g(s,r):m(s,r)},e.b=function(t,e,n,r){var i,o,s,a,c,u=this.context.data[t],l=u?u.values.value:[],p=l[0],d=0;if(!p)return;for(i=p.intervals.length;d<i;++d)if(o=p.intervals[d],e&&o.encoding===e||n&&o.field===n){if(!o.extent)return;s=d,c=o.extent.length>2;break}return a=l.reduce(function(t,e){var n=e.intervals[s].extent,r=c?n.map(function(t){return{unit:e.unit,value:t}}):{unit:e.unit,value:n};return c?t.push.apply(t,r):t.push(r),t},[]),c?m(a,r):g(a,r)};var r=n("../../../../vega-parser/src/parsers/expression/ast.js"),i=n("../../../../vega-parser/src/parsers/expression/data.js"),o=n("../../../../vega-parser/src/parsers/expression/inrange.js"),s=n("../../../../vega-parser/src/parsers/expression/prefixes.js"),a=n("../../../../vega-util/index.js"),c="bin_",u="intersect",l="union",p="index:unit";function d(t,e){for(var n,r=e.fields,i=e.values,s=e.getter||(e.getter=[]),u=r.length,l=0;l<u;++l)if(s[l]=s[l]||Object(a.p)(r[l]),n=s[l](t),Object(a.v)(n)&&(n=Object(a.O)(n)),Object(a.v)(i[l])&&(i[l]=Object(a.O)(i[l])),e[c+r[l]]){if(Object(a.v)(i[l][0])&&(i[l]=i[l].map(a.O)),!Object(o.a)(n,i[l],!0,!1))return!1}else if(n!==i[l])return!1;return!0}function h(t,e){for(var n,r,i=e.intervals,s=i.length,c=0;c<s;++c){if(n=i[c].extent,r=(i[c].getter||(i[c].getter=Object(a.p)(i[c].field)))(t),!n||n[0]===n[1])return!1;if(Object(a.v)(r)&&(r=Object(a.O)(r)),Object(a.v)(n[0])&&(n=i[c].extent=n.map(a.O)),Object(a.x)(n[0])&&!Object(o.a)(r,n))return!1;if(Object(a.A)(n[0])&&n.indexOf(r)<0)return!1}return!0}function f(t,e,n,r){for(var i,o,s,a,c,l=this.context.data[t],d=l?l.values.value:[],h=l?l[p]&&l[p].value:void 0,f=n===u,m=d.length,g=0;g<m;++g)if(i=d[g],h&&f){if(-1===(s=(o=o||{})[a=i.unit]||0))continue;if(c=r(e,i),o[a]=c?-1:++s,c&&1===h.size)return!0;if(!c&&s===h.get(a).count)return!1}else if(f^(c=r(e,i)))return c;return m&&f}function m(t,e){for(var n,r,i,o,s={},a=0,c={},l=[],p=0,d=t.length;p<d;++p)r=(n=t[p]).unit,o=n.value,s[r]||(s[r]=++a),(i=c[o])||(c[o]=i={value:o,units:{},count:0}),i.units[r]||(i.units[r]=++i.count);for(o in c)i=c[o],e===u&&i.count!==a||l.push(i.value);return l.length?l:void 0}function g(t,e){for(var n,r,i,o,s=e===u?y:v,c=0,l=t.length;c<l;++c)n=t[c].value,Object(a.v)(n[0])&&(n=n.map(a.O)),(i=n[0])>(o=n[1])&&(o=n[0],i=n[1]),r=r?s(r,i,o):[i,o];return r&&r.length&&+r[0]!=+r[1]?r:void 0}function v(t,e,n){return t[0]>e&&(t[0]=e),t[1]<n&&(t[1]=n),t}function y(t,e,n){return n<t[0]||t[1]<e?[]:(t[0]<e&&(t[0]=e),t[1]>n&&(t[1]=n),t)}},"../../../../vega-parser/src/parsers/expression/shape.js":function(t,e,n){"use strict";e.a=function(t,e,n){var i=Object(r.e)(t,(n||this).context);return function(t){return i?i.path.context(t)(e):""}},e.b=function(t){var e=null;return function(n){return n?Object(i.m)(n,e=e||Object(i.l)(t)):t}};var r=n("../../../../vega-parser/src/parsers/expression/scale.js"),i=n("../../../../vega-scenegraph/index.js")},"../../../../vega-parser/src/parsers/expression/span.js":function(t,e,n){"use strict";e.a=function(t){return t[t.length-1]-t[0]||0}},"../../../../vega-parser/src/parsers/expression/tree.js":function(t,e,n){"use strict";e.b=function(t,e,n){var r=s(t,this),i=r[e],a=r[n];return i&&a?i.path(a).map(o):void 0},e.a=function(t,e){var n=s(t,this)[e];return n?n.ancestors().map(o):void 0};var r=n("../../../../vega-parser/src/parsers/expression/data.js"),i={};function o(t){return t.data}function s(t,e){var n=r.a.call(e,t);return n.root&&n.root.lookup||i}},"../../../../vega-parser/src/parsers/expression/window.js":function(t,e,n){"use strict";e.b=function(){return r?r.screen:{}},e.c=function(){return r?[r.innerWidth,r.innerHeight]:[void 0,void 0]},e.a=function(){var t=this.context.dataflow,e=t.container&&t.container();return e?[e.clientWidth,e.clientHeight]:[void 0,void 0]};var r="undefined"!=typeof window&&window||null},"../../../../vega-parser/src/parsers/guides/axis-config.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-util/index.js");e.a=function(t,e){var n=e.config,o=t.orient,s=o===r.n||o===r.a?n.axisX:n.axisY,a=n["axis"+o[0].toUpperCase()+o.slice(1)],c="band"===e.scaleType(t.scale)&&n.axisBand;return s||a||c?Object(i.l)({},n.axis,s,a,c):n.axis}},"../../../../vega-parser/src/parsers/guides/axis-domain.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),s=n("../../../../vega-parser/src/parsers/marks/roles.js"),a=n("../../../../vega-parser/src/parsers/encode/encode-util.js");function c(t,e){return{scale:t.scale,range:e}}e.a=function(t,e,n,u){var l,p,d,h,f,m=t.orient,g={value:0},v={};return v.enter=l={opacity:g},Object(a.a)(l,"stroke",e.domainColor),Object(a.a)(l,"strokeWidth",e.domainWidth),v.exit={opacity:g},v.update=p={opacity:{value:1}},m===r.n||m===r.a?(d="x",f="y"):(d="y",f="x"),h=d+"2",l[f]=g,p[d]=l[d]=c(t,0),p[h]=l[h]=c(t,1),Object(i.a)(o.c,s.a,null,null,u,v,n)}},"../../../../vega-parser/src/parsers/guides/axis-grid.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),s=n("../../../../vega-parser/src/parsers/marks/roles.js"),a=n("../../../../vega-parser/src/parsers/encode/encode-util.js");e.a=function(t,e,n,c){var u,l,p,d,h,f,m,g,v=t.orient,y=t.gridScale,b=v===r.g||v===r.n?1:-1,_=b*t.offset||0,j={value:0},x={};return x.enter=u={opacity:j},Object(a.a)(u,"stroke",e.gridColor),Object(a.a)(u,"strokeWidth",e.gridWidth),Object(a.a)(u,"strokeDash",e.gridDash),x.exit=l={opacity:j},x.update=p={},Object(a.a)(p,"opacity",e.gridOpacity),d={scale:t.scale,field:r.p,band:e.bandPosition,round:e.tickRound,extra:e.tickExtra,offset:e.tickOffset},v===r.n||v===r.a?(h="x",f="y",g="height"):(h="y",f="x",g="width"),m=f+"2",p[h]=u[h]=l[h]=d,y?(u[f]={scale:y,range:0,mult:b,offset:_},p[m]=u[m]={scale:y,range:1,mult:b,offset:_}):(u[f]={value:_},p[m]=u[m]={signal:g,mult:b,offset:_}),Object(i.a)(o.c,s.b,null,r.p,c,x,n)}},"../../../../vega-parser/src/parsers/guides/axis-labels.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),s=n("../../../../vega-parser/src/parsers/marks/roles.js"),a=n("../../../../vega-parser/src/parsers/encode/encode-util.js"),c=n("../../../../vega-parser/src/util.js");function u(t,e,n,r,i){return{signal:'flush(range("'+t+'"), scale("'+t+'", datum.value), '+e+","+n+","+r+","+i+")"}}e.a=function(t,e,n,l,p){var d,h,f,m,g,v=t.orient,y=v===r.g||v===r.n?-1:1,b=t.scale,_=Object(c.n)(t.labelPadding,e.labelPadding),j=Object(c.n)(t.labelBound,e.labelBound),x=Object(c.n)(t.labelFlush,e.labelFlush),w=null!=x&&!1!==x&&(x=+x)===x,O=+Object(c.n)(t.labelFlushOffset,e.labelFlushOffset),C=Object(c.n)(t.labelOverlap,e.labelOverlap),k={value:0},E={};return E.enter=d={opacity:k},Object(a.a)(d,"angle",e.labelAngle),Object(a.a)(d,"fill",e.labelColor),Object(a.a)(d,"font",e.labelFont),Object(a.a)(d,"fontSize",e.labelFontSize),Object(a.a)(d,"fontWeight",e.labelFontWeight),Object(a.a)(d,"limit",e.labelLimit),E.exit=h={opacity:k},E.update=f={opacity:{value:1},text:{field:r.f}},(m=Object(a.b)(p)).mult=y,m.offset=Object(a.b)(_),m.offset.mult=y,g={scale:b,field:r.p,band:.5,offset:e.tickOffset},v===r.n||v===r.a?(f.y=d.y=m,f.x=d.x=h.x=g,Object(a.a)(f,"align",w?u(b,x,'"left"','"right"','"center"'):"center"),w&&O&&Object(a.a)(f,"dx",u(b,x,-O,O,0)),Object(a.a)(f,"baseline",v===r.n?"bottom":"top")):(f.x=d.x=m,f.y=d.y=h.y=g,Object(a.a)(f,"align",v===r.k?"left":"right"),Object(a.a)(f,"baseline",w?u(b,x,'"bottom"','"top"','"middle"'):"middle"),w&&O&&Object(a.a)(f,"dy",u(b,x,O,-O,0))),t=Object(i.a)(o.e,s.c,r.c,r.p,l,E,n),(C||j)&&(t.overlap={method:C,order:"datum.index",bound:j?{scale:b,orient:v,tolerance:+j}:null}),t}},"../../../../vega-parser/src/parsers/guides/axis-ticks.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),s=n("../../../../vega-parser/src/parsers/marks/roles.js"),a=n("../../../../vega-parser/src/parsers/encode/encode-util.js");e.a=function(t,e,n,c,u){var l,p,d,h,f,m=t.orient,g=m===r.g||m===r.n?-1:1,v={value:0},y={};return y.enter=l={opacity:v},Object(a.a)(l,"stroke",e.tickColor),Object(a.a)(l,"strokeWidth",e.tickWidth),y.exit=p={opacity:v},y.update=d={opacity:{value:1}},(h=Object(a.b)(u)).mult=g,f={scale:t.scale,field:r.p,band:e.bandPosition,round:e.tickRound,extra:e.tickExtra,offset:e.tickOffset},m===r.n||m===r.a?(d.y=l.y=v,d.y2=l.y2=h,d.x=l.x=p.x=f):(d.x=l.x=v,d.x2=l.x2=h,d.y=l.y=p.y=f),Object(i.a)(o.c,s.e,null,r.p,c,y,n)}},"../../../../vega-parser/src/parsers/guides/axis-title.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/encode/encode-util.js"),s=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),a=n("../../../../vega-parser/src/parsers/marks/roles.js");e.a=function(t,e,n,c){var u,l,p,d=t.orient,h=t.title,f=d===r.g||d===r.n?-1:1,m=d===r.n||d===r.a,g={};return g.enter=u={opacity:{value:0}},Object(o.a)(u,"align",e.titleAlign),Object(o.a)(u,"fill",e.titleColor),Object(o.a)(u,"font",e.titleFont),Object(o.a)(u,"fontSize",e.titleFontSize),Object(o.a)(u,"fontWeight",e.titleFontWeight),Object(o.a)(u,"limit",e.titleLimit),g.exit={opacity:{value:0}},g.update=l={opacity:{value:1},text:h&&h.signal?{signal:h.signal}:{value:h+""}},p={scale:t.scale,range:.5},m?(l.x=p,l.angle={value:0},l.baseline={value:d===r.n?"bottom":"top"}):(l.y=p,l.angle={value:90*f},l.baseline={value:"bottom"}),Object(o.a)(l,"angle",e.titleAngle),Object(o.a)(l,"baseline",e.titleBaseline),!Object(o.a)(l,"x",e.titleX)&&m&&!Object(o.e)("x",n)&&(g.enter.auto={value:!0}),!Object(o.a)(l,"y",e.titleY)&&!m&&!Object(o.e)("y",n)&&(g.enter.auto={value:!0}),Object(i.a)(s.e,a.f,r.d,null,c,g,n)}},"../../../../vega-parser/src/parsers/guides/constants.js":function(t,e,n){"use strict";n.d(e,"n",function(){return r}),n.d(e,"g",function(){return i}),n.d(e,"k",function(){return o}),n.d(e,"a",function(){return s}),n.d(e,"e",function(){return a}),n.d(e,"f",function(){return c}),n.d(e,"i",function(){return u}),n.d(e,"j",function(){return l}),n.d(e,"l",function(){return p}),n.d(e,"o",function(){return d}),n.d(e,"p",function(){return h}),n.d(e,"c",function(){return f}),n.d(e,"d",function(){return m}),n.d(e,"b",function(){return g}),n.d(e,"h",function(){return v}),n.d(e,"m",function(){return y});var r="top",i="left",o="right",s="bottom",a="index",c="label",u="offset",l="perc",p="size",d="total",h="value",f="guide-label",m="guide-title",g="group-title",v=["shape","size","fill","stroke","strokeDash","opacity"],y={name:1,interactive:1}},"../../../../vega-parser/src/parsers/guides/guide-group.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/marks/marktypes.js");e.a=function(t,e,n,i,o,s,a){return{type:r.a,name:n,role:t,style:e,from:i,interactive:o||!1,encode:s,marks:a}}},"../../../../vega-parser/src/parsers/guides/guide-mark.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/encode/encode-util.js");e.a=function(t,e,n,o,s,a,c){return{type:t,name:c?c.name:void 0,role:e,style:c&&c.style||n,key:o,from:s,interactive:!(!c||!c.interactive),encode:Object(i.d)(a,c,r.m)}}},"../../../../vega-parser/src/parsers/guides/legend-gradient-labels.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),s=n("../../../../vega-parser/src/parsers/marks/roles.js"),a=n("../../../../vega-parser/src/parsers/encode/encode-util.js"),c="datum."+r.j+'<=0?"left":datum.'+r.j+'>=1?"right":"center"';e.a=function(t,e,n,u){var l,p,d={value:0},h={};return h.enter=l={opacity:d},Object(a.a)(l,"fill",e.labelColor),Object(a.a)(l,"font",e.labelFont),Object(a.a)(l,"fontSize",e.labelFontSize),Object(a.a)(l,"fontWeight",e.labelFontWeight),Object(a.a)(l,"baseline",e.gradientLabelBaseline),Object(a.a)(l,"limit",e.gradientLabelLimit),h.exit={opacity:d},h.update=p={opacity:{value:1},text:{field:r.f}},l.x=p.x={field:r.j,mult:e.gradientWidth},l.y=p.y={value:e.gradientHeight,offset:e.gradientLabelOffset},l.align=p.align={signal:c},Object(i.a)(o.e,s.j,r.c,r.j,u,h,n)}},"../../../../vega-parser/src/parsers/guides/legend-gradient.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),i=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),o=n("../../../../vega-parser/src/parsers/marks/roles.js"),s=n("../../../../vega-parser/src/parsers/encode/encode-util.js");e.a=function(t,e,n,a){var c,u,l={value:0},p={};return p.enter=c={opacity:l,x:l,y:l},Object(s.a)(c,"width",n.gradientWidth),Object(s.a)(c,"height",n.gradientHeight),Object(s.a)(c,"stroke",n.gradientStrokeColor),Object(s.a)(c,"strokeWidth",n.gradientStrokeWidth),p.exit={opacity:l},p.update=u={x:l,y:l,fill:{gradient:e},opacity:{value:1}},Object(s.a)(u,"width",n.gradientWidth),Object(s.a)(u,"height",n.gradientHeight),Object(r.a)(i.b,o.i,null,void 0,void 0,p,a)}},"../../../../vega-parser/src/parsers/guides/legend-labels.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),s=n("../../../../vega-parser/src/parsers/marks/roles.js"),a=n("../../../../vega-parser/src/parsers/encode/encode-util.js");e.a=function(t,e,n,c){var u,l,p={value:0},d={};return d.enter=u={opacity:p},Object(a.a)(u,"align",e.labelAlign),Object(a.a)(u,"baseline",e.labelBaseline),Object(a.a)(u,"fill",e.labelColor),Object(a.a)(u,"font",e.labelFont),Object(a.a)(u,"fontSize",e.labelFontSize),Object(a.a)(u,"fontWeight",e.labelFontWeight),Object(a.a)(u,"limit",e.labelLimit),d.exit={opacity:p},d.update=l={opacity:{value:1},text:{field:r.f}},u.x=l.x={field:r.i,offset:e.labelOffset},u.y=l.y={field:r.l,mult:.5,offset:{field:r.o,offset:{field:{group:"entryPadding"},mult:{field:r.e}}}},Object(i.a)(o.e,s.j,r.c,r.p,c,d,n)}},"../../../../vega-parser/src/parsers/guides/legend-symbols.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),s=n("../../../../vega-parser/src/parsers/marks/roles.js"),a=n("../../../../vega-parser/src/parsers/encode/encode-util.js");e.a=function(t,e,n,c){var u,l,p={value:0},d={};return d.enter=u={opacity:p},Object(a.a)(u,"shape",e.symbolType),Object(a.a)(u,"size",e.symbolSize),Object(a.a)(u,"strokeWidth",e.symbolStrokeWidth),t.fill||(Object(a.a)(u,"fill",e.symbolFillColor),Object(a.a)(u,"stroke",e.symbolStrokeColor)),d.exit={opacity:p},d.update=l={opacity:{value:1}},u.x=l.x={field:r.i,mult:.5},u.y=l.y={field:r.l,mult:.5,offset:{field:r.o,offset:{field:{group:"entryPadding"},mult:{field:r.e}}}},r.h.forEach(function(e){t[e]&&(l[e]=u[e]={scale:t[e],field:r.p})}),Object(i.a)(o.d,s.l,null,r.p,c,d,n)}},"../../../../vega-parser/src/parsers/guides/legend-title.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),s=n("../../../../vega-parser/src/parsers/marks/roles.js"),a=n("../../../../vega-parser/src/parsers/encode/encode-util.js");e.a=function(t,e,n,c){var u,l={value:0},p=t.title,d={};return d.enter=u={x:{field:{group:"padding"}},y:{field:{group:"padding"}},opacity:l},Object(a.a)(u,"align",e.titleAlign),Object(a.a)(u,"baseline",e.titleBaseline),Object(a.a)(u,"fill",e.titleColor),Object(a.a)(u,"font",e.titleFont),Object(a.a)(u,"fontSize",e.titleFontSize),Object(a.a)(u,"fontWeight",e.titleFontWeight),Object(a.a)(u,"limit",e.titleLimit),d.exit={opacity:l},d.update={opacity:{value:1},text:p&&p.signal?{signal:p.signal}:{value:p+""}},Object(i.a)(o.e,s.m,r.d,null,c,d,n)}},"../../../../vega-parser/src/parsers/legend.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/legend-gradient.js"),i=n("../../../../vega-parser/src/parsers/guides/legend-gradient-labels.js"),o=n("../../../../vega-parser/src/parsers/guides/legend-labels.js"),s=n("../../../../vega-parser/src/parsers/guides/legend-symbols.js"),a=n("../../../../vega-parser/src/parsers/guides/legend-title.js"),c=n("../../../../vega-parser/src/parsers/guides/guide-group.js"),u=n("../../../../vega-parser/src/parsers/mark.js"),l=n("../../../../vega-parser/src/parsers/marks/roles.js"),p=n("../../../../vega-parser/src/parsers/encode/encode-util.js"),d=n("../../../../vega-parser/src/parsers/guides/constants.js"),h=n("../../../../vega-parser/src/util.js"),f=n("../../../../vega-parser/src/transforms.js"),m=n("../../../../vega-util/index.js");function g(t,e,n,r){var i=e&&(e.update&&e.update[n]||e.enter&&e.enter[n]);return+(i?i.value:r&&(i=t.config.style[r])&&i[n])}e.a=function(t,e){var n,v,y,b,_,j,x,w,O,C,k=t.type||"symbol",E=e.config.legend,S=t.encode||{},P=S.legend||{},M=P.name||void 0,A=P.interactive,T=P.style,D=t.size||t.shape||t.fill||t.stroke||t.strokeDash||t.opacity;return D||Object(m.k)("Missing valid scale for legend."),n={orient:Object(h.n)(t.orient,E.orient),title:null!=t.title},v=Object(h.k)(e.add(Object(f.d)(null,[n]))),P=Object(p.d)({enter:(O=E,C={},Object(p.a)(C,"fill",O.fillColor)+Object(p.a)(C,"stroke",O.strokeColor)+Object(p.a)(C,"strokeWidth",O.strokeWidth)+Object(p.a)(C,"strokeDash",O.strokeDash)+Object(p.a)(C,"cornerRadius",O.cornerRadius)?C:void 0),update:{offset:Object(p.b)(Object(h.n)(t.offset,E.offset)),padding:Object(p.b)(Object(h.n)(t.padding,E.padding)),titlePadding:Object(p.b)(Object(h.n)(t.titlePadding,E.titlePadding))}},P,d.m),j={update:{x:{field:{group:"padding"}},y:{field:{group:"padding"}},entryPadding:Object(p.b)(Object(h.n)(t.entryPadding,E.entryPadding))}},"gradient"===k?(y=Object(h.k)(e.add(Object(f.k)({type:"gradient",scale:e.scaleRef(D),count:e.objectProperty(t.tickCount),values:e.objectProperty(t.values),formatSpecifier:e.property(t.format)}))),w=[Object(r.a)(t,D,E,S.gradient),Object(i.a)(t,E,S.labels,y)]):(y=Object(h.k)(e.add(Object(f.k)(x={scale:e.scaleRef(D),count:e.objectProperty(t.tickCount),values:e.objectProperty(t.values),formatSpecifier:e.property(t.format)}))),w=[Object(s.a)(t,E,S.symbols,y),Object(o.a)(t,E,S.labels,y)],x.size=function(t,e,n){var r=g(e,n[1].encode,"fontSize",d.c);if(t.size)return{$expr:"Math.max(Math.ceil(Math.sqrt(_.scale(datum))),"+r+")"};var i=g(e,n[0].encode,"size");return Math.max(Math.ceil(Math.sqrt(i)),r)}(t,e,w)),w=[Object(c.a)(l.h,null,null,v,A,j,w)],n.title&&(_=Object(a.a)(t,E,S.title,v),j.update.y.offset={field:{group:"titlePadding"},offset:g(e,_.encode,"fontSize",d.d)},w.push(_)),b=Object(c.a)(l.k,T,M,v,A,P,w),t.zindex&&(b.zindex=t.zindex),Object(u.a)(b,e)}},"../../../../vega-parser/src/parsers/mark.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/marks/clip.js"),i=n("../../../../vega-parser/src/parsers/marks/definition.js"),o=n("../../../../vega-parser/src/parsers/marks/interactive.js"),s=n("../../../../vega-parser/src/parsers/marks/data.js"),a=n("../../../../vega-parser/src/parsers/marks/facet.js"),c=n("../../../../vega-parser/src/parsers/marks/subflow.js"),u=n("../../../../vega-parser/src/parsers/marks/role.js"),l=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),p=n("../../../../vega-parser/src/parsers/marks/roles.js"),d=n("../../../../vega-parser/src/parsers/encode/encode-util.js"),h=n("../../../../vega-parser/src/parsers/transform.js"),f=n("../../../../vega-parser/src/parsers/trigger.js"),m=n("../../../../vega-parser/src/parsers/spec.js"),g=n("../../../../vega-parser/src/DataScope.js"),v=n("../../../../vega-parser/src/util.js"),y=n("../../../../vega-util/index.js"),b=n("../../../../vega-parser/src/transforms.js");e.a=function(t,e){var n,_,j,x,w,O,C,k,E,S,P,M,A,T=Object(u.a)(t),D=t.type===l.a,I=t.from&&t.from.facet,R=t.layout||T===p.o||T===p.g,N=T===p.n||R||I,L=t.overlap;j=Object(s.a)(t.from,D,e),_=e.add(Object(b.f)({key:j.key||(t.key?Object(v.f)(t.key):void 0),pulse:j.pulse,clean:!D})),E=Object(v.k)(_),_=x=e.add(Object(b.d)({pulse:E})),_=e.add(Object(b.l)({markdef:Object(i.a)(t),interactive:Object(o.a)(t.interactive,e),clip:Object(r.a)(t.clip,e),context:{$context:!0},groups:e.lookup(),parent:e.signals.parent?e.signalRef("parent"):null,index:e.markpath(),pulse:Object(v.k)(_)})),S=Object(v.k)(_),(_=e.add(Object(b.g)(Object(d.c)(t.encode,t.type,T,t.style,e,{pulse:S})))).params.parent=e.encode(),t.transform&&t.transform.forEach(function(t){var n=Object(h.a)(t,e);(n.metadata.generates||n.metadata.changes)&&Object(y.k)("Mark transforms should not generate new data."),n.params.pulse=Object(v.k)(_),e.add(_=n)}),t.sort&&(_=e.add(Object(b.x)({sort:e.compareRef(t.sort,!0),pulse:Object(v.k)(_)}))),P=Object(v.k)(_),(I||R)&&(R=e.add(Object(b.z)({layout:e.objectProperty(t.layout),legendMargin:e.config.legendMargin,mark:S,pulse:P})),M=Object(v.k)(R)),w=e.add(Object(b.c)({mark:S,pulse:M||P})),A=Object(v.k)(w),D&&(N&&((n=e.operators).pop(),R&&n.pop()),e.pushState(P,M||A,E),I?Object(a.a)(t,e,j):N?Object(c.a)(t,e,j):Object(m.a)(t,e),e.popState(),N&&(R&&n.push(R),n.push(w))),L&&(_={method:!0===L.method?"parity":L.method,pulse:A},L.order&&(_.sort=e.compareRef({field:L.order})),L.bound&&(_.boundScale=e.scaleRef(L.bound.scale),_.boundOrient=L.bound.orient,_.boundTolerance=L.bound.tolerance),A=Object(v.k)(e.add(Object(b.o)(_)))),O=e.add(Object(b.u)({pulse:A})),C=e.add(Object(b.w)({pulse:Object(v.k)(O)},void 0,e.parent())),null!=t.name&&(k=t.name,e.addData(k,new g.a(e,x,O,C)),t.on&&t.on.forEach(function(t){(t.insert||t.remove||t.toggle)&&Object(y.k)("Marks only support modify triggers."),Object(f.a)(t,e,k)}))}},"../../../../vega-parser/src/parsers/marks/clip.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");function i(t){return Object(r.y)(t)&&t.signal?t.signal:Object(r.L)(t)}e.a=function(t,e){var n;return Object(r.y)(t)&&(t.signal?n=t.signal:t.path?n="pathShape("+i(t.path)+")":t.sphere&&(n="geoShape("+i(t.sphere)+', {type: "Sphere"})')),n?e.signalRef(n):!!t}},"../../../../vega-parser/src/parsers/marks/data.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/transform.js"),i=n("../../../../vega-parser/src/util.js"),o=n("../../../../vega-parser/src/transforms.js"),s=n("../../../../vega-util/index.js");e.a=function(t,e,n){var a,c,u,l,p;return t?(a=t.facet)&&(e||Object(s.k)("Only group marks can be faceted."),null!=a.field?l=p=Object(i.k)(n.getData(a.data).output):(t.data?p=Object(i.k)(n.getData(t.data).aggregate):((u=Object(r.a)(Object(s.l)({type:"aggregate",groupby:Object(s.g)(a.groupby)},a.aggregate),n)).params.key=n.keyRef(a.groupby),u.params.pulse=Object(i.k)(n.getData(a.data).output),l=p=Object(i.k)(n.add(u))),c=n.keyRef(a.groupby,!0))):l=Object(i.k)(n.add(Object(o.d)(null,[{}]))),l||(l=t.$ref?t:Object(i.k)(n.getData(t.data).output)),{key:c,pulse:l,parent:p}}},"../../../../vega-parser/src/parsers/marks/definition.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/marks/role.js");e.a=function(t){return{marktype:t.type,name:t.name||void 0,role:t.role||Object(r.a)(t),zindex:+t.zindex||void 0}}},"../../../../vega-parser/src/parsers/marks/facet.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/spec.js"),i=n("../../../../vega-parser/src/DataScope.js"),o=n("../../../../vega-parser/src/util.js"),s=n("../../../../vega-parser/src/transforms.js"),a=n("../../../../vega-util/index.js");e.a=function(t,e,n){var c,u,l,p,d=t.from.facet,h=d.name,f=Object(o.k)(e.getData(d.data).output);d.name||Object(a.k)("Facet must have a name: "+Object(a.L)(d)),d.data||Object(a.k)("Facet must reference a data set: "+Object(a.L)(d)),d.field?p=e.add(Object(s.q)({field:e.fieldRef(d.field),pulse:f})):d.groupby?p=e.add(Object(s.h)({key:e.keyRef(d.groupby),group:Object(o.k)(e.proxy(n.parent)),pulse:f})):Object(a.k)("Facet must specify groupby or field: "+Object(a.L)(d)),u=(c=e.fork()).add(Object(s.d)()),l=c.add(Object(s.w)({pulse:Object(o.k)(u)})),c.addData(h,new i.a(c,u,u,l)),c.addSignal("parent",null),p.params.subflow={$subflow:Object(r.a)(t,c).toRuntime()}}},"../../../../vega-parser/src/parsers/marks/interactive.js":function(t,e,n){"use strict";e.a=function(t,e){return t&&t.signal?e.signalRef(t.signal):!1!==t}},"../../../../vega-parser/src/parsers/marks/marktypes.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r}),n.d(e,"b",function(){return i}),n.d(e,"c",function(){return o}),n.d(e,"d",function(){return s}),n.d(e,"e",function(){return a});var r="group",i="rect",o="rule",s="symbol",a="text"},"../../../../vega-parser/src/parsers/marks/role.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),i=n("../../../../vega-parser/src/parsers/marks/roles.js");e.a=function(t){var e=t.role||"";return e.indexOf("axis")&&e.indexOf("legend")?t.type===r.a?i.o:e||i.n:e}},"../../../../vega-parser/src/parsers/marks/roles.js":function(t,e,n){"use strict";n.d(e,"n",function(){return r}),n.d(e,"g",function(){return i}),n.d(e,"o",function(){return o}),n.d(e,"d",function(){return s}),n.d(e,"a",function(){return a}),n.d(e,"b",function(){return c}),n.d(e,"c",function(){return u}),n.d(e,"e",function(){return l}),n.d(e,"f",function(){return p}),n.d(e,"k",function(){return d}),n.d(e,"h",function(){return h}),n.d(e,"i",function(){return f}),n.d(e,"j",function(){return m}),n.d(e,"l",function(){return g}),n.d(e,"m",function(){return v}),n.d(e,"p",function(){return y});var r="mark",i="frame",o="scope",s="axis",a="axis-domain",c="axis-grid",u="axis-label",l="axis-tick",p="axis-title",d="legend",h="legend-entry",f="legend-gradient",m="legend-label",g="legend-symbol",v="legend-title",y="title"},"../../../../vega-parser/src/parsers/marks/subflow.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/spec.js"),i=n("../../../../vega-parser/src/transforms.js");e.a=function(t,e,n){var o=e.add(Object(i.q)({pulse:n.pulse})),s=e.fork();s.add(Object(i.w)()),s.addSignal("parent",null),o.params.subflow={$subflow:Object(r.a)(t,s).toRuntime()}}},"../../../../vega-parser/src/parsers/padding.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");function i(t){return+t||0}e.a=function(t,e){return t=t||e.padding,Object(r.y)(t)?{top:i(t.top),bottom:i(t.bottom),left:i(t.left),right:i(t.right)}:{top:n=i(t),bottom:n,left:n,right:n};var n}},"../../../../vega-parser/src/parsers/projection.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");function i(t,e){return Object(r.t)(t)?t.map(function(t){return i(t,e)}):Object(r.y)(t)?t.signal?e.signalRef(t.signal):Object(r.k)("Unsupported parameter object: "+Object(r.L)(t)):t}e.a=function(t,e){var n={};for(var r in t)"name"!==r&&(n[r]=i(t[r],e));e.addProjection(t.name,n)}},"../../../../vega-parser/src/parsers/scale.js":function(t,e,n){"use strict";e.a=function(t,e){var n=t.type||"linear";c.hasOwnProperty(n)||Object(o.k)("Unrecognized scale type: "+Object(o.L)(n));e.addScale(t.name,{type:n,domain:void 0})},e.b=function(t,e){var n,r=e.getScale(t.name).params;r.domain=f(t.domain,t,e),null!=t.range&&(r.range=function t(e,n,r){var i=e.range,s=n.config.range;if(i.signal)return n.signalRef(i.signal);if(Object(o.A)(i)){if(s&&s.hasOwnProperty(i))return e=Object(o.l)({},e,{range:s[i]}),t(e,n,r);"width"===i?i=[0,{signal:"width"}]:"height"===i?i=l(e.type)?[0,{signal:"height"}]:[{signal:"height"},0]:Object(o.k)("Unrecognized scale range value: "+Object(o.L)(i))}else{if(i.scheme)return r.scheme=d(i.scheme,n),i.extent&&(r.schemeExtent=(a=i.extent,c=n,a.signal?c.signalRef(a.signal):a.map(function(t){return d(t,c)}))),void(i.count&&(r.schemeCount=d(i.count,n)));if(i.step)return void(r.rangeStep=d(i.step,n));if(l(e.type)&&!Object(o.t)(i))return f(i,e,n);Object(o.t)(i)||Object(o.k)("Unsupported range type: "+Object(o.L)(i))}var a,c;return i.map(function(t){return d(t,n)})}(t,e,r));null!=t.interpolate&&(i=t.interpolate,(s=r).interpolate=d(i.type||i),null!=i.gamma&&(s.interpolateGamma=d(i.gamma)));var i,s;null!=t.nice&&(a=t.nice,r.nice=Object(o.y)(a)?{interval:d(a.interval),step:d(a.step)}:d(a));var a;for(n in t)r.hasOwnProperty(n)||"name"===n||(r[n]=d(t[n],e))};var r=n("../../../../vega-parser/src/util.js"),i=n("../../../../vega-parser/src/transforms.js"),o=n("../../../../vega-util/index.js"),s=0,a=["identity","ordinal","band","point","bin-linear","bin-ordinal","linear","pow","sqrt","log","sequential","time","utc","quantize","quantile","threshold"],c=Object(o.P)(a),u=Object(o.P)(a.slice(1,6));function l(t){return u.hasOwnProperty(t)}function p(t){return"quantile"===t}function d(t,e){return Object(o.y)(t)?t.signal?e.signalRef(t.signal):Object(o.k)("Unsupported object: "+Object(o.L)(t)):t}function h(t){Object(o.k)("Can not find data set: "+Object(o.L)(t))}function f(t,e,n){if(t)return t.signal?n.signalRef(t.signal):(Object(o.t)(t)?function(t,e,n){return t.map(function(t){return d(t,n)})}:t.fields?function(t,e,n){var a=t.data,c=t.fields.reduce(function(t,e){return e=Object(o.A)(e)?{data:a,field:e}:Object(o.t)(e)||e.signal?function(t,e){var n="_:vega:_"+s++,r=Object(i.d)({});if(Object(o.t)(t))r.value={$ingest:t};else if(t.signal){var a="setdata("+Object(o.L)(n)+","+t.signal+")";r.params.input=e.signalRef(a)}return e.addDataPipeline(n,[r,Object(i.w)({})]),{data:n,field:"data"}}(e,n):e,t.push(e),t},[]);return(l(e.type)?function(t,e,n){var o,s,a,c;return o=n.map(function(t){var n=e.getData(t.data);return n||h(t.data),n.countsRef(e,t.field)}),s=e.add(Object(i.a)({groupby:r.h,ops:["sum"],fields:[e.fieldRef("count")],as:["count"],pulse:o})),a=e.add(Object(i.d)({pulse:Object(r.k)(s)})),c=e.add(Object(i.y)({field:r.h,sort:e.sortRef(m(t.sort,!0)),pulse:Object(r.k)(a)})),Object(r.k)(c)}:p(e.type)?function(t,e,n){var o=n.map(function(t){var n=e.getData(t.data);return n||h(t.data),n.domainRef(e,t.field)});return Object(r.k)(e.add(Object(i.n)({values:o})))}:function(t,e,n){var o=n.map(function(t){var n=e.getData(t.data);return n||h(t.data),n.extentRef(e,t.field)});return Object(r.k)(e.add(Object(i.m)({extents:o})))})(t,n,c)}:function(t,e,n){var r=n.getData(t.data);r||h(t.data);return l(e.type)?r.valuesRef(n,t.field,m(t.sort,!1)):p(e.type)?r.domainRef(n,t.field):r.extentRef(n,t.field)})(t,e,n);null==e.domainMin&&null==e.domainMax||Object(o.k)("No scale domain defined for domainMin/domainMax to override.")}function m(t,e){return t&&(t.field||t.op?t.field||"count"===t.op?e&&t.field?Object(o.k)("Multiple domain scales can not sort by field."):e&&t.op&&"count"!==t.op&&Object(o.k)("Multiple domain scales support op count only."):Object(o.k)("No field provided for sort aggregate op: "+t.op):Object(o.y)(t)?t.field="key":t={field:"key"}),t}},"../../../../vega-parser/src/parsers/signal-updates.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/expression.js"),i=n("../../../../vega-parser/src/parsers/update.js");e.a=function(t,e){var n=e.getSignal(t.name);if(t.update){var o=Object(r.a)(t.update,e);n.update=o.$expr,n.params=o.$params}t.on&&t.on.forEach(function(t){Object(i.a)(t,e,n.id)})}},"../../../../vega-parser/src/parsers/signal.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js"),i=["value","update","react","bind"];function o(t,e){Object(r.k)(t+' for "outer" push: '+Object(r.L)(e))}e.a=function(t,e){var n=t.name;if("outer"===t.push)e.signals[n]||o("No prior signal definition",n),i.forEach(function(e){void 0!==t[e]&&o("Invalid property ",e)});else{var r=e.addSignal(n,t.value);!1===t.react&&(r.react=!1),t.bind&&e.addBinding(n,t.bind)}}},"../../../../vega-parser/src/parsers/spec.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/signal-updates.js"),i=n("../../../../vega-parser/src/parsers/scale.js"),o=n("../../../../vega-parser/src/parsers/projection.js"),s=n("../../../../vega-parser/src/parsers/legend.js"),a=n("../../../../vega-parser/src/parsers/signal.js"),c=n("../../../../vega-parser/src/parsers/title.js"),u=n("../../../../vega-parser/src/parsers/data.js"),l=n("../../../../vega-parser/src/parsers/mark.js"),p=n("../../../../vega-parser/src/parsers/axis.js"),d=n("../../../../vega-util/index.js");e.a=function(t,e,n){var h=Object(d.g)(t.signals),f=Object(d.g)(t.scales);return n||h.forEach(function(t){Object(a.a)(t,e)}),Object(d.g)(t.projections).forEach(function(t){Object(o.a)(t,e)}),f.forEach(function(t){Object(i.a)(t,e)}),Object(d.g)(t.data).forEach(function(t){Object(u.a)(t,e)}),f.forEach(function(t){Object(i.b)(t,e)}),h.forEach(function(t){Object(r.a)(t,e)}),Object(d.g)(t.axes).forEach(function(t){Object(p.a)(t,e)}),Object(d.g)(t.marks).forEach(function(t){Object(l.a)(t,e)}),Object(d.g)(t.legends).forEach(function(t){Object(s.a)(t,e)}),t.title&&Object(c.a)(t.title,e),e.parseLambdas(),e}},"../../../../vega-parser/src/parsers/stream.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/expression.js"),i=n("../../../../vega-util/index.js"),o="view",s="scope";function a(t,e){return(t.merge?c:t.stream?u:t.type?l:Object(i.k)("Invalid stream specification: "+Object(i.L)(t)))(t,e)}function c(t,e){var n=p({merge:t.merge.map(function(t){return a(t,e)})},t,e);return e.addStream(n).id}function u(t,e){var n=p({stream:a(t.stream,e)},t,e);return e.addStream(n).id}function l(t,e){var n,r=e.event((n=t.source)===s?o:n||o,t.type),i=p({stream:r},t,e);return 1===Object.keys(i).length?r:e.addStream(i).id}function p(t,e,n){var o,c,u,l=e.between;return l&&(2!==l.length&&Object(i.k)('Stream "between" parameter must have 2 entries: '+Object(i.L)(e)),t.between=[a(l[0],n),a(l[1],n)]),l=e.filter?Object(i.g)(e.filter):[],(e.marktype||e.markname||e.markrole)&&l.push((o=e.marktype,c=e.markname,u=e.markrole,"event.item"+(o&&"*"!==o?"&&event.item.mark.marktype==='"+o+"'":"")+(u?"&&event.item.mark.role==='"+u+"'":"")+(c?"&&event.item.mark.name==='"+c+"'":""))),e.source===s&&l.push("inScope(event.item)"),l.length&&(t.filter=Object(r.a)("("+l.join(")&&(")+")").$expr),null!=(l=e.throttle)&&(t.throttle=+l),null!=(l=e.debounce)&&(t.debounce=+l),e.consume&&(t.consume=!0),t}e.a=function(t,e){return t.signal?e.getSignal(t.signal).id:t.scale?e.getScale(t.scale).id:a(t,e)}},"../../../../vega-parser/src/parsers/title.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/guides/constants.js"),i=n("../../../../vega-parser/src/parsers/guides/guide-mark.js"),o=n("../../../../vega-parser/src/parsers/mark.js"),s=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),a=n("../../../../vega-parser/src/parsers/marks/roles.js"),c=n("../../../../vega-parser/src/parsers/encode/encode-util.js"),u=n("../../../../vega-parser/src/util.js"),l=n("../../../../vega-parser/src/transforms.js"),p=n("../../../../vega-util/index.js");e.a=function(t,e){t=Object(p.A)(t)?{text:t}:t;var n,d,h,f=e.config.title,m=Object(p.l)({},t.encode);return n={orient:null!=t.orient?t.orient:f.orient},d=Object(u.k)(e.add(Object(l.d)(null,[n]))),m.name=t.name,m.interactive=t.interactive,h=function(t,e,n,o){var u,l,d,h,f,m,g=t.text,v=t.orient||e.orient,y=t.anchor||e.anchor,b=v===r.g||v===r.n?-1:1,_=v===r.n||v===r.a,j={group:_?"width":"height"},x={};x.enter=u={opacity:{value:0}},Object(c.a)(u,"fill",e.color),Object(c.a)(u,"font",e.font),Object(c.a)(u,"fontSize",e.fontSize),Object(c.a)(u,"fontWeight",e.fontWeight),x.exit={opacity:{value:0}},x.update=l={opacity:{value:1},text:Object(p.y)(g)?g:{value:g+""},offset:Object(c.b)((null!=t.offset?t.offset:e.offset)||0)},"start"===y?(f=0,m="left"):"end"===y?(f=1,m="right"):(f=.5,m="center");d={field:j,mult:f},h=b<0?{value:0}:_?{field:{group:"height"}}:{field:{group:"width"}},_?(l.x=d,l.y=h,l.angle={value:0},l.baseline={value:v===r.n?"bottom":"top"}):(l.x=h,l.y=d,l.angle={value:90*b},l.baseline={value:"bottom"});return l.align={value:m},l.limit={field:j},Object(c.a)(l,"angle",e.angle),Object(c.a)(l,"baseline",e.baseline),Object(c.a)(l,"limit",e.limit),Object(i.a)(s.e,a.p,t.style||r.b,null,o,x,n)}(t,f,m,d),t.zindex&&(h.zindex=t.zindex),Object(o.a)(h,e)}},"../../../../vega-parser/src/parsers/transform.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/expression.js"),i=n("../../../../vega-parser/src/util.js"),o=n("../../../../vega-parser/src/transforms.js"),s=n("../../../../vega-dataflow/index.js"),a=n("../../../../vega-util/index.js");function c(t,e,n){var r,i,o,s={};for(i=0,o=t.params.length;i<o;++i)s[(r=t.params[i]).name]=u(r,e,n);return s}function u(t,e,n){var r,o,s,c=t.type,u=e[t.name];if("index"===c)return function(t,e,n){Object(a.A)(e.from)||Object(a.k)('Lookup "from" parameter must be a string literal.');return n.getData(e.from).lookupRef(n,e.key)}(0,e,n);{if(void 0!==u)return"param"===c?(o=n,s=e[(r=t).name],r.array?(Object(a.t)(s)||Object(a.k)("Expected an array of sub-parameters. Instead: "+Object(a.L)(s)),s.map(function(t){return p(r,t,o)})):p(r,s,o)):"projection"===c?n.projectionRef(e[t.name]):t.array&&!Object(i.g)(u)?u.map(function(e){return l(t,e,n)}):l(t,u,n);t.required&&Object(a.k)("Missing required "+Object(a.L)(e.type)+" parameter: "+Object(a.L)(t.name))}}function l(t,e,n){var o=t.type;if(Object(i.g)(e))return d(o)?Object(a.k)("Expression references can not be signals."):h(o)?n.fieldRef(e):f(o)?n.compareRef(e):n.signalRef(e.signal);var s,c,u=t.expr||h(o);return u&&((c=e)&&c.expr)?Object(r.a)(e.expr,n):u&&((s=e)&&s.field)?Object(i.f)(e.field):d(o)?Object(r.a)(e,n):"data"===o?Object(i.k)(n.getData(e).values):h(o)?Object(i.f)(e):f(o)?n.compareRef(e):e}function p(t,e,n){var r,s,u,l,p;for(l=0,p=t.params.length;l<p;++l){s=t.params[l];for(u in s.key)if(s.key[u]!==e[u]){s=null;break}if(s)break}return s||Object(a.k)("Unsupported parameter: "+Object(a.L)(e)),r=Object(a.l)(c(s,e,n),s.key),Object(i.k)(n.add(Object(o.p)(r)))}function d(t){return"expr"===t}function h(t){return"field"===t}function f(t){return"compare"===t}e.a=function(t,e){var n=Object(s.f)(t.type);n||Object(a.k)("Unrecognized transform type: "+Object(a.L)(t.type));var r=Object(i.e)(n.type.toLowerCase(),null,c(n,t,e));return t.signal&&e.addSignal(t.signal,e.proxy(r)),r.metadata=n.metadata||{},r}},"../../../../vega-parser/src/parsers/trigger.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/expression.js"),i=n("../../../../vega-parser/src/util.js");e.a=function(t,e,n){var o,s,a=t.remove,c=t.insert,u=t.toggle,l=t.modify,p=t.values,d=e.add(Object(i.j)());o="if("+t.trigger+',modify("'+n+'",'+[c,a,u,l,p].map(function(t){return null==t?"null":t}).join(",")+"),0)",s=Object(r.a)(o,e),d.update=s.$expr,d.params=s.$params}},"../../../../vega-parser/src/parsers/update.js":function(t,e,n){"use strict";var r=n("../../../../vega-parser/src/parsers/expression.js"),i=n("../../../../vega-parser/src/parsers/stream.js"),o=n("../../../../vega-event-selector/index.js"),s=n("../../../../vega-util/index.js"),a="var datum=event.item&&event.item.datum;";e.a=function(t,e,n){var c,u,l=t.events,p=t.update,d=t.encode,h=[];l||Object(s.k)("Signal update missing events specification."),Object(s.A)(l)&&(l=Object(o.a)(l)),(l=Object(s.g)(l).filter(function(t){return t.signal||t.scale?(h.push(t),0):1})).length&&h.push(l.length>1?{merge:l}:l[0]),null!=d&&(p&&Object(s.k)("Signal encode and update are mutually exclusive."),p="encode(item(),"+Object(s.L)(d)+")"),c=Object(s.A)(p)?Object(r.a)(p,e,a):null!=p.expr?Object(r.a)(p.expr,e,a):null!=p.value?p.value:null!=p.signal?{$expr:"_.value",$params:{value:e.signalRef(p.signal)}}:Object(s.k)("Invalid signal update specification."),u={target:n,update:c},t.force&&(u.options={force:!0}),h.forEach(function(t){t={source:Object(i.a)(t,e)},e.addUpdate(Object(s.l)(t,u))})}},"../../../../vega-parser/src/parsers/view.js":function(t,e,n){"use strict";e.a=function(t,e){var n,m,g,v,y,b=e.config;return e.background=t.background||b.background,e.eventConfig=b.events,y=Object(l.k)(e.root=e.add(Object(l.j)())),e.addSignal("width",t.width||0),e.addSignal("height",t.height||0),e.addSignal("padding",Object(i.a)(t.padding,b)),e.addSignal("autosize",Object(r.a)(t.autosize,b)),Object(h.g)(t.signals).forEach(function(t){f[t.name]||Object(o.a)(t,e)}),m=e.add(Object(d.d)()),g=Object(a.d)({enter:{x:{value:0},y:{value:0}},update:{width:{signal:"width"},height:{signal:"height"}}},t.encode),g=e.add(Object(d.g)(Object(a.c)(g,c.a,u.g,t.style,e,{pulse:Object(l.k)(m)}))),v=e.add(Object(d.z)({layout:e.objectProperty(t.layout),legendMargin:b.legendMargin,autosize:e.signalRef("autosize"),mark:y,pulse:Object(l.k)(g)})),e.operators.pop(),e.pushState(Object(l.k)(g),Object(l.k)(v),null),Object(s.a)(t,e,!0),e.operators.push(v),n=e.add(Object(d.c)({mark:y,pulse:Object(l.k)(v)})),n=e.add(Object(d.u)({pulse:Object(l.k)(n)})),n=e.add(Object(d.w)({pulse:Object(l.k)(n)})),e.addData("root",new p.a(e,m,m,n)),e};var r=n("../../../../vega-parser/src/parsers/autosize.js"),i=n("../../../../vega-parser/src/parsers/padding.js"),o=n("../../../../vega-parser/src/parsers/signal.js"),s=n("../../../../vega-parser/src/parsers/spec.js"),a=n("../../../../vega-parser/src/parsers/encode/encode-util.js"),c=n("../../../../vega-parser/src/parsers/marks/marktypes.js"),u=n("../../../../vega-parser/src/parsers/marks/roles.js"),l=n("../../../../vega-parser/src/util.js"),p=n("../../../../vega-parser/src/DataScope.js"),d=n("../../../../vega-parser/src/transforms.js"),h=n("../../../../vega-util/index.js"),f=Object(h.P)(["width","height","padding","autosize"])},"../../../../vega-parser/src/transforms.js":function(t,e,n){"use strict";n.d(e,"a",function(){return o}),n.d(e,"b",function(){return s}),n.d(e,"c",function(){return a}),n.d(e,"d",function(){return c}),n.d(e,"e",function(){return u}),n.d(e,"f",function(){return l}),n.d(e,"g",function(){return p}),n.d(e,"h",function(){return d}),n.d(e,"i",function(){return h}),n.d(e,"j",function(){return f}),n.d(e,"k",function(){return m}),n.d(e,"l",function(){return g}),n.d(e,"m",function(){return v}),n.d(e,"n",function(){return y}),n.d(e,"o",function(){return b}),n.d(e,"p",function(){return _}),n.d(e,"q",function(){return j}),n.d(e,"r",function(){return x}),n.d(e,"s",function(){return w}),n.d(e,"t",function(){return O}),n.d(e,"u",function(){return C}),n.d(e,"v",function(){return k}),n.d(e,"w",function(){return E}),n.d(e,"x",function(){return S}),n.d(e,"z",function(){return P}),n.d(e,"y",function(){return M});var r=n("../../../../vega-parser/src/util.js");function i(t){return function(e,n,i){return Object(r.e)(t,n,e||void 0,i)}}var o=i("aggregate"),s=i("axisticks"),a=i("bound"),c=i("collect"),u=i("compare"),l=i("datajoin"),p=i("encode"),d=(i("extent"),i("facet")),h=i("field"),f=i("key"),m=i("legendentries"),g=i("mark"),v=i("multiextent"),y=i("multivalues"),b=i("overlap"),_=i("params"),j=i("prefacet"),x=i("projection"),w=i("proxy"),O=i("relay"),C=i("render"),k=i("scale"),E=i("sieve"),S=i("sortitems"),P=i("viewlayout"),M=i("values")},"../../../../vega-parser/src/util.js":function(t,e,n){"use strict";e.b=i,e.e=o,e.j=function(t,e){return o("operator",t,e)},e.k=function(t){var e={$ref:t.id};t.id<0&&(t.refs=t.refs||[]).push(e);return e},n.d(e,"m",function(){return s}),e.f=a,n.d(e,"h",function(){return c}),e.d=function(t,e){return{$compare:t,$order:e}},e.i=function(t,e){var n={$key:t};e&&(n.$flat=!0);return n},n.d(e,"a",function(){return u}),e.l=function(t){return Object(r.y)(t)?(t.order===l?"-":"+")+p(t.op,t.field):""},e.c=p,e.g=function(t){return t&&t.signal},e.n=function(t,e){return null!=t?t:e};var r=n("../../../../vega-util/index.js");function i(t,e,n,r){this.id=-1,this.type=t,this.value=e,this.params=n,r&&(this.parent=r)}function o(t,e,n,r){return new i(t,e,n,r)}var s={$tupleid:1,toString:function(){return":_tupleid_:"}};function a(t,e){return e?{$field:t,$name:e}:{$field:t}}var c=a("key");var u="ascending",l="descending";function p(t,e){return(t&&t.signal?"$"+t.signal:t||"")+(t&&e?"_":"")+(e&&e.signal?"$"+e.signal:e||"")}},"../../../../vega-projection/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-projection/src/projections.js");n.d(e,"b",function(){return r.b}),n.d(e,"c",function(){return r.c}),n.d(e,"a",function(){return r.a})},"../../../../vega-projection/src/projections.js":function(t,e,n){"use strict";n.d(e,"c",function(){return o}),e.b=s,e.a=function(t){return t&&t.path||i};var r=n("../../../../d3-geo/index.js"),i=Object(r.p)(),o=["clipAngle","clipExtent","scale","translate","center","rotate","parallels","precision","coefficient","distance","fraction","lobes","parallel","radius","ratio","spacing","tilt"];function s(t,e){if(!t||"string"!=typeof t)throw new Error("Projection type must be a name string.");return t=t.toLowerCase(),arguments.length>1?(a[t]=(n=t,i=e,function t(){var e=i();return e.type=n,e.path=Object(r.p)().projection(e),e.copy=e.copy||function(){var n=t();return o.forEach(function(t){e.hasOwnProperty(t)&&n[t](e[t]())}),n.path.pointRadius(e.path.pointRadius()),n},e}),this):a.hasOwnProperty(t)?a[t]:null;var n,i}var a={albers:r.a,albersusa:r.b,azimuthalequalarea:r.d,azimuthalequidistant:r.e,conicconformal:r.h,conicequalarea:r.i,conicequidistant:r.j,equirectangular:r.k,gnomonic:r.l,mercator:r.n,orthographic:r.o,stereographic:r.q,transversemercator:r.r};for(var c in a)s(c,a[c])},"../../../../vega-runtime/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-runtime/src/dataflow.js");n.d(e,"b",function(){return r.a});var i=n("../../../../vega-runtime/src/context.js");n.d(e,"a",function(){return i.a});n("../../../../vega-runtime/src/expression.js")},"../../../../vega-runtime/src/context.js":function(t,e,n){"use strict";var r=n("../../../../vega-runtime/src/state.js"),i=n("../../../../vega-runtime/src/util.js");function o(t,e,n){this.dataflow=t,this.transforms=e,this.events=t.events.bind(t),this.signals={},this.scales={},this.nodes={},this.data={},this.fn={},n&&(this.functions=Object.create(n),this.functions.context=this)}function s(t){this.dataflow=t.dataflow,this.transforms=t.transforms,this.functions=t.functions,this.events=t.events,this.signals=Object.create(t.signals),this.scales=Object.create(t.scales),this.nodes=Object.create(t.nodes),this.data=Object.create(t.data),this.fn=Object.create(t.fn),t.functions&&(this.functions=Object.create(t.functions),this.functions.context=this)}e.a=function(t,e,n){return new o(t,e,n)},o.prototype=s.prototype={fork:function(){var t=new s(this);return(this.subcontext||(this.subcontext=[])).push(t),t},get:function(t){return this.nodes[t]},set:function(t,e){return this.nodes[t]=e},add:function(t,e){var n,r=this,o=r.dataflow;if(r.set(t.id,e),Object(i.b)(t.type)&&(n=t.value)&&(n.$ingest?o.ingest(e,n.$ingest,n.$format):n.$request?o.request(e,n.$request,n.$format):o.pulse(e,o.changeset().insert(n))),t.root&&(r.root=e),t.parent){var s=r.get(t.parent.$ref);s?(o.connect(s,[e]),e.targets().add(s)):(r.unresolved=r.unresolved||[]).push(function(){s=r.get(t.parent.$ref),o.connect(s,[e]),e.targets().add(s)})}if(t.signal&&(r.signals[t.signal]=e),t.scale&&(r.scales[t.scale]=e),t.data)for(var a in t.data)n=r.data[a]||(r.data[a]={}),t.data[a].forEach(function(t){n[t]=e})},resolve:function(){return(this.unresolved||[]).forEach(function(t){t()}),delete this.unresolved,this},operator:function(t,e,n){this.add(t,this.dataflow.add(t.value,e,n,t.react))},transform:function(t,e,n){this.add(t,this.dataflow.add(this.transforms[Object(i.a)(e)],n))},stream:function(t,e){this.set(t.id,e)},update:function(t,e,n,r,i){this.dataflow.on(e,n,r,i,t.options)},getState:r.a,setState:r.b}},"../../../../vega-runtime/src/dataflow.js":function(t,e,n){"use strict";var r=n("../../../../vega-runtime/src/operator.js"),i=n("../../../../vega-runtime/src/stream.js"),o=n("../../../../vega-runtime/src/update.js");e.a=function(t,e){var n=t.operators||[];return t.background&&(e.background=t.background),t.eventConfig&&(e.eventConfig=t.eventConfig),n.forEach(function(t){Object(r.a)(t,e)}),n.forEach(function(t){Object(r.b)(t,e)}),(t.streams||[]).forEach(function(t){Object(i.a)(t,e)}),(t.updates||[]).forEach(function(t){Object(o.a)(t,e)}),e.resolve()}},"../../../../vega-runtime/src/expression.js":function(t,e,n){"use strict";function r(t,e,n){";"!==e[e.length-1]&&(e="return("+e+");");var r=Function.apply(null,t.concat(e));return n&&n.functions?r.bind(n.functions):r}e.d=function(t,e){return r(["_"],t,e)},e.e=function(t,e){return r(["datum","_"],t,e)},e.b=function(t,e){return r(["event"],t,e)},e.c=function(t,e){return r(["_","event"],t,e)},e.a=function(t,e){return r(["item","_"],t,e)}},"../../../../vega-runtime/src/operator.js":function(t,e,n){"use strict";e.b=function(t,e){var n,r;t.params&&((n=e.get(t.id))||Object(s.k)("Invalid operator id: "+t.id),r=Object(i.a)(t.params,e),e.dataflow.connect(n,n.parameters(r)))};var r=n("../../../../vega-runtime/src/expression.js"),i=n("../../../../vega-runtime/src/parameters.js"),o=n("../../../../vega-runtime/src/util.js"),s=n("../../../../vega-util/index.js");e.a=function(t,e){Object(o.c)(t.type)||!t.type?e.operator(t,t.update?Object(r.d)(t.update,e):null):e.transform(t,t.type)}},"../../../../vega-runtime/src/parameters.js":function(t,e,n){"use strict";e.a=function t(e,n,r){r=r||{};var i,o;for(i in e)(o=e[i])&&o.$expr&&o.$params&&t(o.$params,n,r),r[i]=Object(s.t)(o)?o.map(function(t){return a(t,n)}):a(o,n);return r};var r=n("../../../../vega-runtime/src/dataflow.js"),i=n("../../../../vega-runtime/src/expression.js"),o=n("../../../../vega-dataflow/index.js"),s=n("../../../../vega-util/index.js");function a(t,e){if(!t||!Object(s.y)(t))return t;for(var n,r=0,i=c.length;r<i;++r)if(n=c[r],t.hasOwnProperty(n.key))return n.parse(t,e);return t}var c=[{key:"$ref",parse:function(t,e){return e.get(t.$ref)||Object(s.k)("Operator not defined: "+t.$ref)}},{key:"$key",parse:function(t,e){var n="k:"+t.$key+"_"+!!t.$flat;return e.fn[n]||(e.fn[n]=Object(s.B)(t.$key,t.$flat))}},{key:"$expr",parse:function(t,e){var n="e:"+t.$expr;return e.fn[n]||(e.fn[n]=Object(s.d)(Object(i.e)(t.$expr,e),t.$fields,t.$name))}},{key:"$field",parse:function(t,e){if(!t.$field)return null;var n="f:"+t.$field+"_"+t.$name;return e.fn[n]||(e.fn[n]=Object(s.p)(t.$field,t.$name))}},{key:"$encode",parse:function(t,e){var n,r,o=t.$encode,a={};for(n in o)r=o[n],a[n]=Object(s.d)(Object(i.a)(r.$expr,e),r.$fields),a[n].output=r.$output;return a}},{key:"$compare",parse:function(t,e){var n="c:"+t.$compare+"_"+t.$order,r=Object(s.g)(t.$compare).map(function(t){return t&&t.$tupleid?o.n:t});return e.fn[n]||(e.fn[n]=Object(s.h)(r,t.$order))}},{key:"$context",parse:function(t,e){return e}},{key:"$subflow",parse:function(t,e){var n=t.$subflow;return function(t,i,o){var s=Object(r.a)(n,e.fork()),a=s.get(n.operators[0].id),c=s.signals.parent;return c&&c.set(o),a}}},{key:"$tupleid",parse:function(){return o.n}}]},"../../../../vega-runtime/src/state.js":function(t,e,n){"use strict";e.a=function(t){var e=this,n={};if(t.signals){var r=n.signals={};Object.keys(e.signals).forEach(function(n){var i=e.signals[n];t.signals(n,i)&&(r[n]=i.value)})}if(t.data){var i=n.data={};Object.keys(e.data).forEach(function(n){var r=e.data[n];t.data(n,r)&&(i[n]=r.input.value)})}e.subcontext&&!1!==t.recurse&&(n.subcontext=e.subcontext.map(function(e){return e.getState(t)}));return n},e.b=function(t){var e=this,n=e.dataflow,o=t.data,s=t.signals;Object.keys(s||{}).forEach(function(t){n.update(e.signals[t],s[t],i)}),Object.keys(o||{}).forEach(function(t){n.pulse(e.data[t].input,n.changeset().remove(r.S).insert(o[t]))}),(t.subcontext||[]).forEach(function(t,n){var r=e.subcontext[n];r&&r.setState(t)})};var r=n("../../../../vega-util/index.js"),i={skip:!0}},"../../../../vega-runtime/src/stream.js":function(t,e,n){"use strict";var r=n("../../../../vega-runtime/src/expression.js"),i=n("../../../../vega-util/index.js");e.a=function(t,e){var n,o=null!=t.filter?Object(r.b)(t.filter,e):void 0,s=null!=t.stream?e.get(t.stream):void 0;t.source?s=e.events(t.source,t.type,o):t.merge&&(s=(n=t.merge.map(e.get.bind(e)))[0].merge.apply(n[0],n.slice(1))),t.between&&(n=t.between.map(e.get.bind(e)),s=s.between(n[0],n[1])),t.filter&&(s=s.filter(o)),null!=t.throttle&&(s=s.throttle(+t.throttle)),null!=t.debounce&&(s=s.debounce(+t.debounce)),null==s&&Object(i.k)("Invalid stream definition: "+JSON.stringify(t)),t.consume&&s.consume(!0),e.stream(t,s)}},"../../../../vega-runtime/src/update.js":function(t,e,n){"use strict";var r=n("../../../../vega-runtime/src/expression.js"),i=n("../../../../vega-runtime/src/parameters.js"),o=n("../../../../vega-util/index.js");e.a=function(t,e){var n=e.get(t.source),s=null,a=t.update,c=void 0;n||Object(o.k)("Source not defined: "+t.source),s=t.target&&t.target.$expr?Object(r.b)(t.target.$expr,e):e.get(t.target),a&&a.$expr&&(a.$params&&(c=Object(i.a)(a.$params,e)),a=Object(r.c)(a.$expr,e)),e.update(t,n,s,a,c)}},"../../../../vega-runtime/src/util.js":function(t,e,n){"use strict";function r(t){return(t+"").toLowerCase()}e.a=r,e.c=function(t){return"operator"===r(t)},e.b=function(t){return"collect"===r(t)}},"../../../../vega-scale/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-scale/src/scales.js");n.d(e,"d",function(){return r.a});var i=n("../../../../vega-scale/src/schemes.js");n.d(e,"f",function(){return i.a});var o=n("../../../../vega-scale/src/bandSpace.js");n.d(e,"a",function(){return o.a});var s=n("../../../../vega-scale/src/interpolate.js");n.d(e,"b",function(){return s.a}),n.d(e,"c",function(){return s.b}),n.d(e,"e",function(){return s.c});var a=n("../../../../vega-scale/src/timeInterval.js");n.d(e,"g",function(){return a.a}),n.d(e,"h",function(){return a.b})},"../../../../vega-scale/src/bandSpace.js":function(t,e,n){"use strict";e.a=function(t,e,n){var r=t-e+2*n;return t?r>0?r:1:0}},"../../../../vega-scale/src/interpolate.js":function(t,e,n){"use strict";e.b=function(t,e){var n=e[0],i=Object(r.J)(e)-n;return function(e){return t(n+e*i)}},e.c=function(t,e,n){var i=n-e;return i?"linear"===t.type||"sequential"===t.type?function(t){return(t-e)/i}:t.copy().domain([e,n]).range([0,1]).interpolate(o):Object(r.i)(0)},e.a=function(t,e){var n=i[(r=t,"interpolate"+r.toLowerCase().split("-").map(function(t){return t[0].toUpperCase()+t.slice(1)}).join(""))];var r;return null!=e&&n&&n.gamma?n.gamma(e):n};var r=n("../../../../vega-util/index.js"),i=n("../../../../d3-interpolate/index.js");function o(t,e){var n=e-t;return function(e){return t+e*n}}},"../../../../vega-scale/src/invertRange.js":function(t,e,n){"use strict";e.a=function(t){return function(e){var n,r=e[0],i=e[1];return i<r&&(n=r,r=i,i=n),[t.invert(r),t.invert(i)]}}},"../../../../vega-scale/src/invertRangeExtent.js":function(t,e,n){"use strict";e.a=function(t){return function(e){var n,r,i,o,s=t.range(),a=e[0],c=e[1],u=-1;for(c<a&&(r=a,a=c,c=r),i=0,o=s.length;i<o;++i)s[i]>=a&&s[i]<=c&&(u<0&&(u=i),n=i);if(!(u<0))return a=t.invertExtent(s[u]),c=t.invertExtent(s[n]),[void 0===a[0]?a[1]:a[0],void 0===c[1]?c[0]:c[1]]}}},"../../../../vega-scale/src/palettes.js":function(t,e,n){"use strict";function r(t){for(var e=t.length/6|0,n=new Array(e),r=0;r<e;)n[r]="#"+t.slice(6*r,6*++r);return n}n.d(e,"b",function(){return i}),n.d(e,"c",function(){return o}),n.d(e,"d",function(){return s}),n.d(e,"e",function(){return a}),n.d(e,"f",function(){return c}),n.d(e,"a",function(){return u});var i=r("1f77b4aec7e8ff7f0effbb782ca02c98df8ad62728ff98969467bdc5b0d58c564bc49c94e377c2f7b6d27f7f7fc7c7c7bcbd22dbdb8d17becf9edae5"),o=r("393b795254a36b6ecf9c9ede6379398ca252b5cf6bcedb9c8c6d31bd9e39e7ba52e7cb94843c39ad494ad6616be7969c7b4173a55194ce6dbdde9ed6"),s=r("3182bd6baed69ecae1c6dbefe6550dfd8d3cfdae6bfdd0a231a35474c476a1d99bc7e9c0756bb19e9ac8bcbddcdadaeb636363969696bdbdbdd9d9d9"),a=r("4c78a8f58518e4575672b7b254a24beeca3bb279a2ff9da69d755dbab0ac"),c=r("4c78a89ecae9f58518ffbf7954a24b88d27ab79a20f2cf5b43989483bcb6e45756ff9d9879706ebab0acd67195fcbfd2b279a2d6a5c99e765fd8b5a5"),u=new Array(3).concat("67a9cff7f7f7f1a340","0571b092c5defdb863e66101","0571b092c5def7f7f7fdb863e66101","2166ac67a9cfd1e5f0fee0b6f1a340b35806","2166ac67a9cfd1e5f0f7f7f7fee0b6f1a340b35806","2166ac4393c392c5ded1e5f0fee0b6fdb863e08214b35806","2166ac4393c392c5ded1e5f0f7f7f7fee0b6fdb863e08214b35806","0530612166ac4393c392c5ded1e5f0fee0b6fdb863e08214b358067f3b08","0530612166ac4393c392c5ded1e5f0f7f7f7fee0b6fdb863e08214b358067f3b08").map(r)},"../../../../vega-scale/src/scaleBand.js":function(t,e,n){"use strict";e.a=s,e.b=function(){return function t(e){var n=e.copy;e.padding=e.paddingOuter;delete e.paddingInner;e.copy=function(){return t(n())};return e}(s().paddingInner(1))};var r=n("../../../../vega-scale/src/bandSpace.js"),i=n("../../../../d3-array/index.js"),o=n("../../../../d3-scale/index.js");function s(){var t,e,n=Object(o.d)().unknown(void 0),a=n.domain,c=n.range,u=[0,1],l=!1,p=0,d=0,h=.5;function f(){var n=a().length,o=u[1]<u[0],s=u[o-0],f=u[1-o],m=Object(r.a)(n,p,d);t=(f-s)/(m||1),l&&(t=Math.floor(t)),s+=(f-s-t*(n-p))*h,e=t*(1-p),l&&(s=Math.round(s),e=Math.round(e));var g=Object(i.n)(n).map(function(e){return s+t*e});return c(o?g.reverse():g)}return delete n.unknown,n.domain=function(t){return arguments.length?(a(t),f()):a()},n.range=function(t){return arguments.length?(u=[+t[0],+t[1]],f()):u.slice()},n.rangeRound=function(t){return u=[+t[0],+t[1]],l=!0,f()},n.bandwidth=function(){return e},n.step=function(){return t},n.round=function(t){return arguments.length?(l=!!t,f()):l},n.padding=function(t){return arguments.length?(d=Math.max(0,Math.min(1,t)),p=d,f()):p},n.paddingInner=function(t){return arguments.length?(p=Math.max(0,Math.min(1,t)),f()):p},n.paddingOuter=function(t){return arguments.length?(d=Math.max(0,Math.min(1,t)),f()):d},n.align=function(t){return arguments.length?(h=Math.max(0,Math.min(1,t)),f()):h},n.invertRange=function(t){if(null!=t[0]&&null!=t[1]){var n,r,o,s=+t[0],l=+t[1],p=u[1]<u[0],d=p?c().reverse():c(),h=d.length-1;if(s==s&&l==l&&(l<s&&(o=s,s=l,l=o),!(l<d[0]||s>u[1-p])))return n=Math.max(0,Object(i.d)(d,s)-1),r=s===l?n:Object(i.d)(d,l)-1,s-d[n]>e+1e-10&&++n,p&&(o=n,n=h-r,r=h-o),n>r?void 0:a().slice(n,r+1)}},n.invert=function(t){var e=n.invertRange([t,t]);return e?e[0]:e},n.copy=function(){return s().domain(a()).range(u).round(l).paddingInner(p).paddingOuter(d).align(h)},f()}},"../../../../vega-scale/src/scaleBin.js":function(t,e,n){"use strict";e.a=function t(){var e=Object(i.b)(),n=[];function r(t){return e(t)}r.domain=function(t){return arguments.length?(n=c(t),e.domain([n[0],Object(o.J)(n)]),r):n.slice()};r.range=function(t){return arguments.length?(e.range(t),r):e.range()};r.rangeRound=function(t){return arguments.length?(e.rangeRound(t),r):e.rangeRound()};r.interpolate=function(t){return arguments.length?(e.interpolate(t),r):e.interpolate()};r.invert=function(t){return e.invert(t)};r.ticks=function(t){var e=n.length,i=~~(e/(t||e));return i<2?r.domain():n.filter(function(t,e){return!(e%i)})};r.tickFormat=function(){return e.tickFormat.apply(e,arguments)};r.copy=function(){return t().domain(r.domain()).range(r.range())};return r},e.b=function t(){var e=[],n=[];function i(t){return null==t||t!=t?void 0:n[(Object(r.b)(e,t)-1)%n.length]}i.domain=function(t){return arguments.length?(e=c(t),i):e.slice()};i.range=function(t){return arguments.length?(n=a.call(t),i):n.slice()};i.copy=function(){return t().domain(i.domain()).range(i.range())};return i};var r=n("../../../../d3-array/index.js"),i=n("../../../../d3-scale/index.js"),o=n("../../../../vega-util/index.js"),s=Array.prototype.map,a=Array.prototype.slice;function c(t){return s.call(t,function(t){return+t})}},"../../../../vega-scale/src/scaleSequential.js":function(t,e,n){"use strict";e.a=function t(e){var n=Object(i.b)(),o=0,s=1,a=!1;function c(){var t=n.domain();o=t[0],s=Object(r.J)(t)-o}function u(t){var n=(t-o)/s;return e(a?Math.max(0,Math.min(1,n)):n)}u.clamp=function(t){return arguments.length?(a=!!t,u):a};u.domain=function(t){return arguments.length?(n.domain(t),c(),u):n.domain()};u.interpolator=function(t){return arguments.length?(e=t,u):e};u.copy=function(){return t().domain(n.domain()).clamp(a).interpolator(e)};u.ticks=function(t){return n.ticks(t)};u.tickFormat=function(t,e){return n.tickFormat(t,e)};u.nice=function(t){return n.nice(t),c(),u};return u};var r=n("../../../../vega-util/index.js"),i=n("../../../../d3-scale/index.js")},"../../../../vega-scale/src/scales.js":function(t,e,n){"use strict";e.a=u;var r=n("../../../../vega-scale/src/invertRange.js"),i=n("../../../../vega-scale/src/invertRangeExtent.js"),o=n("../../../../vega-scale/src/scaleBand.js"),s=n("../../../../vega-scale/src/scaleBin.js"),a=n("../../../../vega-scale/src/scaleSequential.js"),c=n("../../../../d3-scale/index.js");function u(t,e){return arguments.length>1?(l[t]=(n=t,o=e,function(){var t=o();return t.invertRange||(t.invertRange=t.invert?Object(r.a)(t):t.invertExtent?Object(i.a)(t):void 0),t.type=n,t}),this):l.hasOwnProperty(t)?l[t]:void 0;var n,o}var l={identity:c.a,linear:c.b,log:c.c,ordinal:c.d,pow:c.e,sqrt:c.h,quantile:c.f,quantize:c.g,threshold:c.i,time:c.j,utc:c.k,band:o.a,point:o.b,sequential:a.a,"bin-linear":s.a,"bin-ordinal":s.b};for(var p in l)u(p,l[p])},"../../../../vega-scale/src/schemes.js":function(t,e,n){"use strict";var r=n("../../../../vega-scale/src/palettes.js"),i=n("../../../../d3-scale-chromatic/index.js"),o=n("../../../../d3-interpolate/index.js"),s=n("../../../../vega-util/index.js"),a={blueorange:r.a},c={category10:i.schemeCategory10,accent:i.schemeAccent,dark2:i.schemeDark2,paired:i.schemePaired,pastel1:i.schemePastel1,pastel2:i.schemePastel2,set1:i.schemeSet1,set2:i.schemeSet2,set3:i.schemeSet3,category20:r.b,category20b:r.c,category20c:r.d,tableau10:r.e,tableau20:r.f,viridis:i.interpolateViridis,magma:i.interpolateMagma,inferno:i.interpolateInferno,plasma:i.interpolatePlasma,blueorange:Object(o.interpolateRgbBasis)(Object(s.J)(r.a))};function u(t,e){c[t]=i["interpolate"+e],a[t]=i["scheme"+e]}u("blues","Blues"),u("greens","Greens"),u("greys","Greys"),u("purples","Purples"),u("reds","Reds"),u("oranges","Oranges"),u("brownbluegreen","BrBG"),u("purplegreen","PRGn"),u("pinkyellowgreen","PiYG"),u("purpleorange","PuOr"),u("redblue","RdBu"),u("redgrey","RdGy"),u("redyellowblue","RdYlBu"),u("redyellowgreen","RdYlGn"),u("spectral","Spectral"),u("bluegreen","BuGn"),u("bluepurple","BuPu"),u("greenblue","GnBu"),u("orangered","OrRd"),u("purplebluegreen","PuBuGn"),u("purpleblue","PuBu"),u("purplered","PuRd"),u("redpurple","RdPu"),u("yellowgreenblue","YlGnBu"),u("yellowgreen","YlGn"),u("yelloworangebrown","YlOrBr"),u("yelloworangered","YlOrRd"),e.a=function(t,e){if(arguments.length>1)return c[t]=e,this;var n=t.split("-");return t=n[0],(n=+n[1]+1)&&a.hasOwnProperty(t)?a[t][n-1]:!n&&c.hasOwnProperty(t)?c[t]:void 0}},"../../../../vega-scale/src/timeInterval.js":function(t,e,n){"use strict";e.a=function(t){return i.hasOwnProperty(t)&&i[t]},e.b=function(t){return o.hasOwnProperty(t)&&o[t]};var r=n("../../../../d3-time/index.js"),i={millisecond:r.c,second:r.g,minute:r.d,hour:r.b,day:r.a,week:r.j,month:r.f,year:r.k},o={millisecond:r.n,second:r.r,minute:r.o,hour:r.m,day:r.l,week:r.u,month:r.q,year:r.v}},"../../../../vega-scenegraph/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/Bounds.js");n.d(e,"a",function(){return r.a});var i=n("../../../../vega-scenegraph/src/Gradient.js");n.d(e,"c",function(){return i.a});var o=n("../../../../vega-scenegraph/src/GroupItem.js");n.d(e,"d",function(){return o.a});n("../../../../vega-scenegraph/src/ResourceLoader.js");var s=n("../../../../vega-scenegraph/src/Item.js");n.d(e,"f",function(){return s.a});var a=n("../../../../vega-scenegraph/src/Scenegraph.js");n.d(e,"i",function(){return a.a});var c=n("../../../../vega-scenegraph/src/Handler.js");n.d(e,"e",function(){return c.a});n("../../../../vega-scenegraph/src/Renderer.js");var u=n("../../../../vega-scenegraph/src/CanvasHandler.js");n.d(e,"b",function(){return u.a});n("../../../../vega-scenegraph/src/CanvasRenderer.js"),n("../../../../vega-scenegraph/src/SVGHandler.js"),n("../../../../vega-scenegraph/src/SVGRenderer.js"),n("../../../../vega-scenegraph/src/SVGStringRenderer.js");var l=n("../../../../vega-scenegraph/src/modules.js");n.d(e,"h",function(){return l.a}),n.d(e,"o",function(){return l.b});var p=n("../../../../vega-scenegraph/src/marks/index.js");n.d(e,"g",function(){return p.a});var d=n("../../../../vega-scenegraph/src/bound/boundClip.js");n.d(e,"j",function(){return d.a});n("../../../../vega-scenegraph/src/bound/boundContext.js");var h=n("../../../../vega-scenegraph/src/bound/boundStroke.js");n.d(e,"k",function(){return h.a});n("../../../../vega-scenegraph/src/bound/boundItem.js"),n("../../../../vega-scenegraph/src/bound/boundMark.js"),n("../../../../vega-scenegraph/src/path/curves.js"),n("../../../../vega-scenegraph/src/path/symbols.js"),n("../../../../vega-scenegraph/src/path/rectangle.js"),n("../../../../vega-scenegraph/src/path/trail.js");var f=n("../../../../vega-scenegraph/src/path/parse.js");n.d(e,"l",function(){return f.a});var m=n("../../../../vega-scenegraph/src/path/render.js");n.d(e,"m",function(){return m.a});var g=n("../../../../vega-scenegraph/src/util/point.js");n.d(e,"n",function(){return g.a});n("../../../../vega-scenegraph/src/util/dom.js"),n("../../../../vega-scenegraph/src/util/tags.js"),n("../../../../vega-scenegraph/src/util/text.js"),n("../../../../vega-scenegraph/src/util/svg/clip.js"),n("../../../../vega-scenegraph/src/util/equal.js"),n("../../../../vega-scenegraph/src/util/serialize.js"),n("../../../../vega-scenegraph/src/util/visit.js")},"../../../../vega-scenegraph/src/Bounds.js":function(t,e,n){"use strict";function r(t){this.clear(),t&&this.union(t)}e.a=r;var i=r.prototype;i.clone=function(){return new r(this)},i.clear=function(){return this.x1=+Number.MAX_VALUE,this.y1=+Number.MAX_VALUE,this.x2=-Number.MAX_VALUE,this.y2=-Number.MAX_VALUE,this},i.empty=function(){return this.x1===+Number.MAX_VALUE&&this.y1===+Number.MAX_VALUE&&this.x2===-Number.MAX_VALUE&&this.y2===-Number.MAX_VALUE},i.set=function(t,e,n,r){return n<t?(this.x2=t,this.x1=n):(this.x1=t,this.x2=n),r<e?(this.y2=e,this.y1=r):(this.y1=e,this.y2=r),this},i.add=function(t,e){return t<this.x1&&(this.x1=t),e<this.y1&&(this.y1=e),t>this.x2&&(this.x2=t),e>this.y2&&(this.y2=e),this},i.expand=function(t){return this.x1-=t,this.y1-=t,this.x2+=t,this.y2+=t,this},i.round=function(){return this.x1=Math.floor(this.x1),this.y1=Math.floor(this.y1),this.x2=Math.ceil(this.x2),this.y2=Math.ceil(this.y2),this},i.translate=function(t,e){return this.x1+=t,this.x2+=t,this.y1+=e,this.y2+=e,this},i.rotate=function(t,e,n){var r=Math.cos(t),i=Math.sin(t),o=e-e*r+n*i,s=n-e*i-n*r,a=this.x1,c=this.x2,u=this.y1,l=this.y2;return this.clear().add(r*a-i*u+o,i*a+r*u+s).add(r*a-i*l+o,i*a+r*l+s).add(r*c-i*u+o,i*c+r*u+s).add(r*c-i*l+o,i*c+r*l+s)},i.union=function(t){return t.x1<this.x1&&(this.x1=t.x1),t.y1<this.y1&&(this.y1=t.y1),t.x2>this.x2&&(this.x2=t.x2),t.y2>this.y2&&(this.y2=t.y2),this},i.intersect=function(t){return t.x1>this.x1&&(this.x1=t.x1),t.y1>this.y1&&(this.y1=t.y1),t.x2<this.x2&&(this.x2=t.x2),t.y2<this.y2&&(this.y2=t.y2),this},i.encloses=function(t){return t&&this.x1<=t.x1&&this.x2>=t.x2&&this.y1<=t.y1&&this.y2>=t.y2},i.alignsWith=function(t){return t&&(this.x1==t.x1||this.x2==t.x2||this.y1==t.y1||this.y2==t.y2)},i.intersects=function(t){return t&&!(this.x2<t.x1||this.x1>t.x2||this.y2<t.y1||this.y1>t.y2)},i.contains=function(t,e){return!(t<this.x1||t>this.x2||e<this.y1||e>this.y2)},i.width=function(){return this.x2-this.x1},i.height=function(){return this.y2-this.y1}},"../../../../vega-scenegraph/src/CanvasHandler.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../vega-scenegraph/src/Handler.js"),i=n("../../../../vega-scenegraph/src/marks/index.js"),o=n("../../../../vega-scenegraph/src/util/point.js"),s=n("../../../../vega-scenegraph/src/util/dom.js"),a=n("../../../../vega-util/index.js");function c(t){r.a.call(this,t),this._down=null,this._touch=null,this._first=!0}var u=Object(a.s)(c,r.a);function l(t,e,n){return function(r){var i=this._active,o=this.pickEvent(r);o===i?this.fire(t,r):(i&&i.exit||this.fire(n,r),this._active=o,this.fire(e,r),this.fire(t,r))}}function p(t){return function(e){this.fire(t,e),this._active=null}}u.initialize=function(t,e,n){var i=this._canvas=t&&Object(s.e)(t,"canvas");if(i){var o=this;this.events.forEach(function(t){i.addEventListener(t,function(e){u[t]?u[t].call(o,e):o.fire(t,e)})})}return r.a.prototype.initialize.call(this,t,e,n)},u.canvas=function(){return this._canvas},u.context=function(){return this._canvas.getContext("2d")},u.events=["keydown","keypress","keyup","dragenter","dragleave","dragover","mousedown","mouseup","mousemove","mouseout","mouseover","click","dblclick","wheel","mousewheel","touchstart","touchmove","touchend"],u.DOMMouseScroll=function(t){this.fire("mousewheel",t)},u.mousemove=l("mousemove","mouseover","mouseout"),u.dragover=l("dragover","dragenter","dragleave"),u.mouseout=p("mouseout"),u.dragleave=p("dragleave"),u.mousedown=function(t){this._down=this._active,this.fire("mousedown",t)},u.click=function(t){this._down===this._active&&(this.fire("click",t),this._down=null)},u.touchstart=function(t){this._touch=this.pickEvent(t.changedTouches[0]),this._first&&(this._active=this._touch,this._first=!1),this.fire("touchstart",t,!0)},u.touchmove=function(t){this.fire("touchmove",t,!0)},u.touchend=function(t){this.fire("touchend",t,!0),this._touch=null},u.fire=function(t,e,n){var r,i,o=n?this._touch:this._active,s=this._handlers[t];if("click"===t&&o&&o.href?this.handleHref(e,o,o.href):("mouseover"===t||"mouseout"===t)&&o&&o.tooltip&&this.handleTooltip(e,o,"mouseover"===t?o.tooltip:null),s)for(e.vegaType=t,r=0,i=s.length;r<i;++r)s[r].handler.call(this._obj,e,o)},u.on=function(t,e){var n=this.eventName(t),r=this._handlers;return this._handlerIndex(r[n],t,e)<0&&(r[n]||(r[n]=[])).push({type:t,handler:e}),this},u.off=function(t,e){var n=this.eventName(t),r=this._handlers[n],i=this._handlerIndex(r,t,e);return i>=0&&r.splice(i,1),this},u.pickEvent=function(t){var e=Object(o.a)(t,this._canvas),n=this._origin;return this.pick(this._scene,e[0],e[1],e[0]-n[0],e[1]-n[1])},u.pick=function(t,e,n,r,o){var s=this.context();return i.a[t.marktype].pick.call(this,s,t,e,n,r,o)}},"../../../../vega-scenegraph/src/CanvasRenderer.js":function(t,e,n){"use strict";e.a=p;var r=n("../../../../vega-scenegraph/src/Renderer.js"),i=n("../../../../vega-scenegraph/src/Bounds.js"),o=n("../../../../vega-scenegraph/src/marks/index.js"),s=n("../../../../vega-scenegraph/src/util/dom.js"),a=n("../../../../vega-scenegraph/src/util/canvas/clip.js"),c=n("../../../../vega-scenegraph/src/util/canvas/resize.js"),u=n("../../../../vega-canvas/index.js"),l=n("../../../../vega-util/index.js");function p(t){r.a.call(this,t),this._redraw=!1,this._dirty=new i.a}var d=Object(l.s)(p,r.a),h=r.a.prototype,f=new i.a;d.initialize=function(t,e,n,r,i){return this._canvas=Object(u.a)(1,1),t&&(Object(s.c)(t,0).appendChild(this._canvas),this._canvas.setAttribute("class","marks")),h.initialize.call(this,t,e,n,r,i)},d.resize=function(t,e,n,r){return h.resize.call(this,t,e,n,r),Object(c.a)(this._canvas,this._width,this._height,this._origin,this._scale),this._redraw=!0,this},d.canvas=function(){return this._canvas},d.context=function(){return this._canvas?this._canvas.getContext("2d"):null},d.dirty=function(t){var e=function(t,e){if(null==e)return t;for(var n=f.clear().union(t);null!=e;e=e.mark.group)n.translate(e.x||0,e.y||0);return n}(t.bounds,t.mark.group);this._dirty.union(e)},d._render=function(t){var e,n,r,i=this.context(),o=this._origin,s=this._width,a=this._height,c=this._dirty;return i.save(),this._redraw||c.empty()?(this._redraw=!1,c=null):(e=i,r=o,(n=c).expand(1).round(),n.translate(-r[0]%1,-r[1]%1),e.beginPath(),e.rect(n.x1,n.y1,n.width(),n.height()),e.clip(),c=n),this.clear(-o[0],-o[1],s,a),this.draw(i,t,c),i.restore(),this._dirty.clear(),this},d.draw=function(t,e,n){var r=o.a[e.marktype];e.clip&&Object(a.a)(t,e),r.draw.call(this,t,e,n),e.clip&&t.restore()},d.clear=function(t,e,n,r){var i=this.context();i.clearRect(t,e,n,r),null!=this._bgcolor&&(i.fillStyle=this._bgcolor,i.fillRect(t,e,n,r))}},"../../../../vega-scenegraph/src/Gradient.js":function(t,e,n){"use strict";var r=0;e.a=function(t,e){var n,i=[];return n={id:"gradient_"+r++,x1:t?t[0]:0,y1:t?t[1]:0,x2:e?e[0]:1,y2:e?e[1]:0,stops:i,stop:function(t,e){return i.push({offset:t,color:e}),n}}}},"../../../../vega-scenegraph/src/GroupItem.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-scenegraph/src/Item.js"),i=n("../../../../vega-util/index.js");function o(t){r.a.call(this,t),this.items=this.items||[]}Object(i.s)(o,r.a)},"../../../../vega-scenegraph/src/Handler.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-scenegraph/src/util/dom.js"),i=n("../../../../vega-loader/index.js");function o(t){this._active=null,this._handlers={},this._loader=t||Object(i.a)()}var s=o.prototype;s.initialize=function(t,e,n){return this._el=t,this._obj=n||null,this.origin(e)},s.element=function(){return this._el},s.origin=function(t){return this._origin=t||[0,0],this},s.scene=function(t){return arguments.length?(this._scene=t,this):this._scene},s.on=function(){},s.off=function(){},s._handlerIndex=function(t,e,n){for(var r=t?t.length:0;--r>=0;)if(t[r].type===e&&!n||t[r].handler===n)return r;return-1},s.handlers=function(){var t,e=this._handlers,n=[];for(t in e)n.push.apply(n,e[t]);return n},s.eventName=function(t){var e=t.indexOf(".");return e<0?t:t.slice(0,e)},s.handleHref=function(t,e,n){this._loader.sanitize(n,{context:"href"}).then(function(e){var n=new MouseEvent(t.type,t),i=Object(r.d)(null,"a");for(var o in e)i.setAttribute(o,e[o]);i.dispatchEvent(n)}).catch(function(){})},s.handleTooltip=function(t,e,n){this._el.setAttribute("title",n||"")}},"../../../../vega-scenegraph/src/Item.js":function(t,e,n){"use strict";e.a=function(t){this.mark=t,this.bounds=this.bounds||new r.a};var r=n("../../../../vega-scenegraph/src/Bounds.js")},"../../../../vega-scenegraph/src/Renderer.js":function(t,e,n){"use strict";e.a=i;var r=n("../../../../vega-scenegraph/src/ResourceLoader.js");function i(t){this._el=null,this._bgcolor=null,this._loader=new r.a(t)}var o=i.prototype;o.initialize=function(t,e,n,r,i){return this._el=t,this.resize(e,n,r,i)},o.element=function(){return this._el},o.scene=function(){return this._el&&this._el.firstChild},o.background=function(t){return 0===arguments.length?this._bgcolor:(this._bgcolor=t,this)},o.resize=function(t,e,n,r){return this._width=t,this._height=e,this._origin=n||[0,0],this._scale=r||1,this},o.dirty=function(){},o.render=function(t){var e=this;return e._call=function(){e._render(t)},e._call(),e._call=null,e},o._render=function(){},o.renderAsync=function(t){var e=this.render(t);return this._ready?this._ready.then(function(){return e}):Promise.resolve(e)},o._load=function(t,e){var n=this,r=n._loader[t](e);if(!n._ready){var i=n._call;n._ready=n._loader.ready().then(function(t){t&&i(),n._ready=null})}return r},o.sanitizeURL=function(t){return this._load("sanitizeURL",t)},o.loadImage=function(t){return this._load("loadImage",t)}},"../../../../vega-scenegraph/src/ResourceLoader.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-canvas/index.js"),i=n("../../../../vega-loader/index.js");function o(t){this._pending=0,this._loader=t||Object(i.a)()}var s=o.prototype;function a(t){t._pending+=1}function c(t){t._pending-=1}s.pending=function(){return this._pending},s.sanitizeURL=function(t){var e=this;return a(e),e._loader.sanitize(t,{context:"href"}).then(function(t){return c(e),t}).catch(function(){return c(e),null})},s.loadImage=function(t){var e=this,n=Object(r.b)();return a(e),e._loader.sanitize(t,{context:"image"}).then(function(t){var r=t.href;if(!r||!n)throw{url:r};var i=new n;return i.onload=function(){c(e),i.loaded=!0},i.onerror=function(){c(e),i.loaded=!1},i.src=r,i}).catch(function(t){return c(e),{loaded:!1,width:0,height:0,src:t&&t.url||""}})},s.ready=function(){var t=this;return new Promise(function(e){!function n(r){t.pending()?setTimeout(function(){n(!0)},10):e(r)}(!1)})}},"../../../../vega-scenegraph/src/SVGHandler.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-scenegraph/src/Handler.js"),i=n("../../../../vega-scenegraph/src/util/dom.js"),o=n("../../../../vega-util/index.js");function s(t){r.a.call(this,t);var e=this;e._hrefHandler=c(e,function(t,n){n&&n.href&&e.handleHref(t,n,n.href)}),e._tooltipHandler=c(e,function(t,n){n&&n.tooltip&&e.handleTooltip(t,n,"mouseover"===t.type?n.tooltip:null)})}var a=Object(o.s)(s,r.a);function c(t,e){return function(n){var r=n.target.__data__;n.vegaType=n.type,r=Array.isArray(r)?r[0]:r,e.call(t._obj,n,r)}}a.initialize=function(t,e,n){var o=this._svg;return o&&(o.removeEventListener("click",this._hrefHandler),o.removeEventListener("mouseover",this._tooltipHandler),o.removeEventListener("mouseout",this._tooltipHandler)),this._svg=o=t&&Object(i.e)(t,"svg"),o&&(o.addEventListener("click",this._hrefHandler),o.addEventListener("mouseover",this._tooltipHandler),o.addEventListener("mouseout",this._tooltipHandler)),r.a.prototype.initialize.call(this,t,e,n)},a.svg=function(){return this._svg},a.on=function(t,e){var n=this.eventName(t),r=this._handlers;if(this._handlerIndex(r[n],t,e)<0){var i={type:t,handler:e,listener:c(this,e)};(r[n]||(r[n]=[])).push(i),this._svg&&this._svg.addEventListener(n,i.listener)}return this},a.off=function(t,e){var n=this.eventName(t),r=this._handlers[n],i=this._handlerIndex(r,t,e);return i>=0&&(this._svg&&this._svg.removeEventListener(n,r[i].listener),r.splice(i,1)),this}},"../../../../vega-scenegraph/src/SVGRenderer.js":function(t,e,n){"use strict";e.a=f;var r=n("../../../../vega-scenegraph/src/Renderer.js"),i=n("../../../../vega-scenegraph/src/marks/index.js"),o=n("../../../../vega-scenegraph/src/util/dom.js"),s=n("../../../../vega-scenegraph/src/util/tags.js"),a=n("../../../../vega-scenegraph/src/util/text.js"),c=n("../../../../vega-scenegraph/src/util/visit.js"),u=n("../../../../vega-scenegraph/src/util/svg/clip.js"),l=n("../../../../vega-scenegraph/src/util/svg/metadata.js"),p=n("../../../../vega-scenegraph/src/util/svg/styles.js"),d=n("../../../../vega-util/index.js"),h=l.a.xmlns;function f(t){r.a.call(this,t),this._dirtyID=1,this._dirty=[],this._svg=null,this._root=null,this._defs=null}var m=Object(d.s)(f,r.a),g=r.a.prototype;function v(t,e,n){var r,i,s;for((t=Object(o.b)(t,n,"linearGradient",h)).setAttribute("id",e.id),t.setAttribute("x1",e.x1),t.setAttribute("x2",e.x2),t.setAttribute("y1",e.y1),t.setAttribute("y2",e.y2),r=0,i=e.stops.length;r<i;++r)(s=Object(o.b)(t,r,"stop",h)).setAttribute("offset",e.stops[r].offset),s.setAttribute("stop-color",e.stops[r].color);Object(o.c)(t,r)}function y(t,e,n){var r;(t=Object(o.b)(t,n,"clipPath",h)).setAttribute("id",e.id),e.path?(r=Object(o.b)(t,0,"path",h)).setAttribute("d",e.path):((r=Object(o.b)(t,0,"rect",h)).setAttribute("x",0),r.setAttribute("y",0),r.setAttribute("width",e.width),r.setAttribute("height",e.height))}function b(t,e){for(;t&&t.dirty!==e;t=t.mark.group){if(t.dirty=e,!t.mark||t.mark.dirty===e)return;t.mark.dirty=e}}function _(t,e,n,r){var i,s=t._svg;if(!s&&(i=e.ownerDocument,s=Object(o.d)(i,r,h),t._svg=s,t.mark&&(s.__data__=t,s.__values__={fill:"default"},"g"===r))){var a=Object(o.d)(i,"path",h);a.setAttribute("class","background"),s.appendChild(a),a.__data__=t;var c=Object(o.d)(i,"g",h);s.appendChild(c),c.__data__=t}return!i&&s.previousSibling===n&&n||e.insertBefore(s,n?n.nextSibling:e.firstChild),s}m.initialize=function(t,e,n,r){return t&&(this._svg=Object(o.b)(t,0,"svg",h),this._svg.setAttribute("class","marks"),Object(o.c)(t,1),this._root=Object(o.b)(this._svg,0,"g",h),Object(o.c)(this._svg,1)),this._defs={gradient:{},clipping:{}},this.background(this._bgcolor),g.initialize.call(this,t,e,n,r)},m.background=function(t){return arguments.length&&this._svg&&this._svg.style.setProperty("background-color",t),g.background.apply(this,arguments)},m.resize=function(t,e,n,r){return g.resize.call(this,t,e,n,r),this._svg&&(this._svg.setAttribute("width",this._width*this._scale),this._svg.setAttribute("height",this._height*this._scale),this._svg.setAttribute("viewBox","0 0 "+this._width+" "+this._height),this._root.setAttribute("transform","translate("+this._origin+")")),this._dirty=[],this},m.svg=function(){if(!this._svg)return null;var t={class:"marks",width:this._width*this._scale,height:this._height*this._scale,viewBox:"0 0 "+this._width+" "+this._height};for(var e in l.a)t[e]=l.a[e];var n=this._bgcolor?Object(s.b)("rect",{width:this._width,height:this._height,style:"fill: "+this._bgcolor+";"})+Object(s.a)("rect"):"";return Object(s.b)("svg",t)+n+this._svg.innerHTML+Object(s.a)("svg")},m._render=function(t){return this._dirtyCheck()&&(this._dirtyAll&&this._resetDefs(),this.draw(this._root,t),Object(o.c)(this._root,1)),this.updateDefs(),this._dirty=[],++this._dirtyID,this},m.updateDefs=function(){var t,e=this._svg,n=this._defs,r=n.el,i=0;for(t in n.gradient)r||(n.el=r=Object(o.b)(e,0,"defs",h)),v(r,n.gradient[t],i++);for(t in n.clipping)r||(n.el=r=Object(o.b)(e,0,"defs",h)),y(r,n.clipping[t],i++);r&&(0===i?(e.removeChild(r),n.el=null):Object(o.c)(r,i))},m._resetDefs=function(){var t=this._defs;t.gradient={},t.clipping={}},m.dirty=function(t){t.dirty!==this._dirtyID&&(t.dirty=this._dirtyID,this._dirty.push(t))},m.isDirty=function(t){return this._dirtyAll||!t._svg||t.dirty===this._dirtyID},m._dirtyCheck=function(){this._dirtyAll=!0;var t=this._dirty;if(!t.length)return!0;var e,n,r,o,s,a,c,u=++this._dirtyID;for(s=0,a=t.length;s<a;++s)(n=(e=t[s]).mark).marktype!==r&&(r=n.marktype,o=i.a[r]),n.zdirty&&n.dirty!==u&&(this._dirtyAll=!1,n.dirty=u,b(n.group,u)),e.exit?(o.nested&&n.items.length?(c=n.items[0])._svg&&this._update(o,c._svg,c):e._svg&&(c=e._svg.parentNode)&&c.removeChild(e._svg),e._svg=null):(e=o.nested?n.items[0]:e)._update!==u&&(e._svg&&e._svg.ownerSVGElement?this._update(o,e._svg,e):(this._dirtyAll=!1,b(e,u)),e._update=u);return!this._dirtyAll},m.draw=function(t,e,n){if(!this.isDirty(e))return e._svg;var r,s=this,a=i.a[e.marktype],l=!1===e.interactive?"none":null,p="g"===a.tag,d=null,h=0;function f(t){var e=s.isDirty(t),n=_(t,r,d,a.tag);e&&(s._update(a,n,t),p&&function(t,e,n){e=e.lastChild;var r,i=0;Object(c.b)(n,function(n){r=t.draw(e,n,r),++i}),Object(o.c)(e,1+i)}(s,n,t)),d=n,++h}return(r=_(e,t,n,"g")).setAttribute("class",Object(o.a)(e)),p||r.style.setProperty("pointer-events",l),e.clip?r.setAttribute("clip-path",Object(u.a)(s,e,e.group)):r.removeAttribute("clip-path"),a.nested?e.items.length&&f(e.items[0]):Object(c.b)(e,f),Object(o.c)(r,h),r};var j=null,x=null,w={group:function(t,e,n){x=e.__values__,j=e.childNodes[1],t.foreground(O,n,this),j=e.childNodes[0],t.background(O,n,this);var r=!1===n.mark.interactive?"none":null;r!==x.events&&(j.style.setProperty("pointer-events",r),x.events=r)},text:function(t,e,n){var r=Object(a.d)(n);r!==x.text&&(e.textContent=r,x.text=r),(r=Object(a.a)(n))!==x.font&&(e.style.setProperty("font",r),x.font=r)}};function O(t,e,n){e!==x[t]&&(null!=e?n?j.setAttributeNS(n,t,e):j.setAttribute(t,e):n?j.removeAttributeNS(n,t):j.removeAttribute(t),x[t]=e)}m._update=function(t,e,n){j=e,x=e.__values__,t.attr(O,n,this);var r=w[t.type];r&&r.call(this,t,e,n),this.style(j,n)},m.style=function(t,e){var n,r,i,o,s,a;if(null!=e)for(n=0,r=p.a.length;n<r;++n)(s=e[i=p.a[n]])!==x[i]&&(o=p.b[i],null==s?"fill"===o?t.style.setProperty(o,"none"):t.style.removeProperty(o):(s.id&&(this._defs.gradient[s.id]=s,s="url("+(a=void 0,"undefined"==typeof window?"":(a=window.location).hash?a.href.slice(0,-a.hash.length):a.href)+"#"+s.id+")"),t.style.setProperty(o,s+"")),x[i]=s)}},"../../../../vega-scenegraph/src/SVGStringRenderer.js":function(t,e,n){"use strict";e.a=h;var r=n("../../../../vega-scenegraph/src/Renderer.js"),i=n("../../../../vega-scenegraph/src/marks/index.js"),o=n("../../../../vega-scenegraph/src/util/dom.js"),s=n("../../../../vega-scenegraph/src/util/tags.js"),a=n("../../../../vega-scenegraph/src/util/text.js"),c=n("../../../../vega-scenegraph/src/util/visit.js"),u=n("../../../../vega-scenegraph/src/util/svg/clip.js"),l=n("../../../../vega-scenegraph/src/util/svg/metadata.js"),p=n("../../../../vega-scenegraph/src/util/svg/styles.js"),d=n("../../../../vega-util/index.js");function h(t){r.a.call(this,t),this._text={head:"",bg:"",root:"",foot:"",defs:"",body:""},this._defs={gradient:{},clipping:{}}}var f,m=Object(d.s)(h,r.a),g=r.a.prototype;function v(t,e,n,r){f[r||t]=e}function y(t,e,n,r){if(null==t)return"";var i,o,s,c,u,l="";for("bgrect"===n&&!1===e.interactive&&(l+="pointer-events: none; "),"text"===n&&(l+="font: "+Object(a.a)(t)+"; "),i=0,o=p.a.length;i<o;++i)s=p.a[i],c=p.b[s],null==(u=t[s])?"fill"===c&&(l+="fill: none; "):"transparent"!==u||"fill"!==c&&"stroke"!==c?(u.id&&(r.gradient[u.id]=u,u="url(#"+u.id+")"),l+=c+": "+u+"; "):l+=c+": none; ";return l?'style="'+l.trim()+'"':null}m.resize=function(t,e,n,r){g.resize.call(this,t,e,n,r);var i=this._origin,o=this._text,a={class:"marks",width:this._width*this._scale,height:this._height*this._scale,viewBox:"0 0 "+this._width+" "+this._height};for(var c in l.a)a[c]=l.a[c];o.head=Object(s.b)("svg",a);var u=this._bgcolor;return"transparent"!==u&&"none"!==u||(u=null),o.bg=u?Object(s.b)("rect",{width:this._width,height:this._height,style:"fill: "+u+";"})+Object(s.a)("rect"):"",o.root=Object(s.b)("g",{transform:"translate("+i+")"}),o.foot=Object(s.a)("g")+Object(s.a)("svg"),this},m.background=function(){var t=g.background.apply(this,arguments);return arguments.length&&this._text.head&&this.resize(this._width,this._height,this._origin,this._scale),t},m.svg=function(){var t=this._text;return t.head+t.bg+t.defs+t.root+t.body+t.foot},m._render=function(t){return this._text.body=this.mark(t),this._text.defs=this.buildDefs(),this},m.buildDefs=function(){var t,e,n,r,i=this._defs,o="";for(e in i.gradient){for(r=(n=i.gradient[e]).stops,o+=Object(s.b)("linearGradient",{id:e,x1:n.x1,x2:n.x2,y1:n.y1,y2:n.y2}),t=0;t<r.length;++t)o+=Object(s.b)("stop",{offset:r[t].offset,"stop-color":r[t].color})+Object(s.a)("stop");o+=Object(s.a)("linearGradient")}for(e in i.clipping)n=i.clipping[e],o+=Object(s.b)("clipPath",{id:e}),n.path?o+=Object(s.b)("path",{d:n.path})+Object(s.a)("path"):o+=Object(s.b)("rect",{x:0,y:0,width:n.width,height:n.height})+Object(s.a)("rect"),o+=Object(s.a)("clipPath");return o.length>0?Object(s.b)("defs")+o+Object(s.a)("defs"):""},m.attributes=function(t,e){return f={},t(v,e,this),f},m.href=function(t){var e,n=this,r=t.href;if(r){if(e=n._hrefs&&n._hrefs[r])return e;n.sanitizeURL(r).then(function(t){t["xlink:href"]=t.href,t.href=null,(n._hrefs||(n._hrefs={}))[r]=t})}return null},m.mark=function(t){var e,n=this,r=i.a[t.marktype],l=r.tag,p=this._defs,d="";function h(i){var o=n.href(i);o&&(d+=Object(s.b)("a",o)),e="g"!==l?y(i,t,l,p):null,d+=Object(s.b)(l,n.attributes(r.attr,i),e),"text"===l?d+=Object(a.d)(i).replace(/&/g,"&amp;").replace(/</g,"&lt;").replace(/>/g,"&gt;"):"g"===l&&(d+=Object(s.b)("path",n.attributes(r.background,i),y(i,t,"bgrect",p))+Object(s.a)("path"),d+=Object(s.b)("g",n.attributes(r.foreground,i))+n.markGroup(i)+Object(s.a)("g")),d+=Object(s.a)(l),o&&(d+=Object(s.a)("a"))}return"g"!==l&&!1===t.interactive&&(e='style="pointer-events: none;"'),d+=Object(s.b)("g",{class:Object(o.a)(t),"clip-path":t.clip?Object(u.a)(n,t,t.group):null},e),r.nested?t.items&&t.items.length&&h(t.items[0]):Object(c.b)(t,h),d+Object(s.a)("g")},m.markGroup=function(t){var e=this,n="";return Object(c.b)(t,function(t){n+=e.mark(t)}),n}},"../../../../vega-scenegraph/src/Scenegraph.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-scenegraph/src/Bounds.js"),i=n("../../../../vega-scenegraph/src/GroupItem.js"),o=n("../../../../vega-scenegraph/src/util/serialize.js");function s(t){arguments.length?this.root=Object(o.a)(t):(this.root=c({marktype:"group",name:"root",role:"frame"}),this.root.items=[new i.a(this.root)])}var a=s.prototype;function c(t,e){return{bounds:new r.a,clip:!!t.clip,group:e,interactive:!1!==t.interactive,items:[],marktype:t.marktype,name:t.name||void 0,role:t.role||void 0,zindex:t.zindex||0}}a.toJSON=function(t){return Object(o.b)(this.root,t||0)},a.mark=function(t,e,n){var r=c(t,e=e||this.root.items[0]);return e.items[n]=r,r.zindex&&(r.group.zdirty=!0),r}},"../../../../vega-scenegraph/src/bound/boundClip.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/Bounds.js"),i=n("../../../../vega-scenegraph/src/bound/boundContext.js"),o=n("../../../../vega-util/index.js"),s=new r.a;e.a=function(t){var e=t.clip;if(Object(o.w)(e))e(Object(i.a)(s.clear()));else{if(!e)return;s.set(0,0,t.group.width,t.group.height)}t.bounds.intersect(s)}},"../../../../vega-scenegraph/src/bound/boundContext.js":function(t,e,n){"use strict";e.a=a;var r,i=2*Math.PI,o=i/4,s=i-1e-8;function a(t){return r=t,a}function c(){}function u(t,e){r.add(t,e)}a.beginPath=c,a.closePath=c,a.moveTo=u,a.lineTo=u,a.rect=function(t,e,n,r){u(t,e),u(t+n,e+r)},a.quadraticCurveTo=function(t,e,n,r){u(t,e),u(n,r)},a.bezierCurveTo=function(t,e,n,r,i,o){u(t,e),u(n,r),u(i,o)},a.arc=function(t,e,n,r,a,c){if(Math.abs(a-r)>s)return u(t-n,e-n),void u(t+n,e+n);var l,p,d,h,f=1/0,m=-1/0,g=1/0,v=-1/0;function y(t){d=n*Math.cos(t),h=n*Math.sin(t),d<f&&(f=d),d>m&&(m=d),h<g&&(g=h),h>v&&(v=h)}if(y(r),y(a),a!==r)if((r%=i)<0&&(r+=i),(a%=i)<0&&(a+=i),a<r&&(c=!c,l=r,r=a,a=l),c)for(a-=i,l=r-r%o,p=0;p<4&&l>a;++p,l-=o)y(l);else for(l=r-r%o+o,p=0;p<4&&l<a;++p,l+=o)y(l);u(t+f,e+g),u(t+m,e+v)}},"../../../../vega-scenegraph/src/bound/boundItem.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/Bounds.js"),i=n("../../../../vega-scenegraph/src/marks/index.js");e.a=function(t,e,n){var o=i.a[t.mark.marktype],s=e||o.bound;return o.nested&&(t=t.mark),s(t.bounds||(t.bounds=new r.a),t,n)}},"../../../../vega-scenegraph/src/bound/boundMark.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/Bounds.js"),i=n("../../../../vega-scenegraph/src/bound/boundItem.js"),o=n("../../../../vega-scenegraph/src/marks/index.js"),s={mark:null};e.a=function(t,e,n){var a,c,u,l,p=o.a[t.marktype],d=p.bound,h=t.items,f=h&&h.length;if(p.nested)return f?u=h[0]:(s.mark=t,u=s),l=Object(i.a)(u,d,n),e=e&&e.union(l)||l;if(e=e||t.bounds&&t.bounds.clear()||new r.a,f)for(a=0,c=h.length;a<c;++a)e.union(Object(i.a)(h[a],d,n));return t.bounds=e}},"../../../../vega-scenegraph/src/bound/boundStroke.js":function(t,e,n){"use strict";e.a=function(t,e){return e.stroke&&0!==e.opacity&&0!==e.strokeOpacity&&t.expand(null!=e.strokeWidth?+e.strokeWidth:1),t}},"../../../../vega-scenegraph/src/marks/arc.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/path/shapes.js"),i=n("../../../../vega-scenegraph/src/marks/markItemPath.js");e.a=Object(i.a)("arc",r.a)},"../../../../vega-scenegraph/src/marks/area.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/path/shapes.js"),i=n("../../../../vega-scenegraph/src/marks/markMultiItemPath.js");e.a=Object(i.a)("area",r.b)},"../../../../vega-scenegraph/src/marks/group.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/path/shapes.js"),i=n("../../../../vega-scenegraph/src/bound/boundStroke.js"),o=n("../../../../vega-scenegraph/src/util/visit.js"),s=n("../../../../vega-scenegraph/src/util/canvas/stroke.js"),a=n("../../../../vega-scenegraph/src/util/canvas/fill.js"),c=n("../../../../vega-scenegraph/src/util/svg/clip.js"),u=n("../../../../vega-scenegraph/src/util/svg/translateItem.js"),l=.5;e.a={type:"group",tag:"g",nested:!1,attr:function(t,e){t("transform",Object(u.a)(e))},bound:function(t,e){if(!e.clip&&e.items)for(var n=e.items,r=0,o=n.length;r<o;++r)t.union(n[r].bounds);return(e.clip||e.width||e.height)&&Object(i.a)(t.add(0,0).add(e.width||0,e.height||0),e),t.translate(e.x||0,e.y||0)},draw:function(t,e,n){var i=this;Object(o.b)(e,function(e){var c,u,p=e.x||0,d=e.y||0,h=e.width||0,f=e.height||0;t.save(),t.translate(p,d),(e.stroke||e.fill)&&(u=null==e.opacity?1:e.opacity)>0&&(t.beginPath(),c=e.stroke?l:0,Object(r.d)(t,e,c,c),e.fill&&Object(a.a)(t,e,u)&&t.fill(),e.stroke&&Object(s.a)(t,e,u)&&t.stroke()),e.clip&&(t.beginPath(),t.rect(0,0,h,f),t.clip()),n&&n.translate(-p,-d),Object(o.b)(e,function(e){i.draw(t,e,n)}),n&&n.translate(p,d),t.restore()})},pick:function(t,e,n,r,i,s){if(e.bounds&&!e.bounds.contains(i,s)||!e.items)return null;var a=this;return Object(o.a)(e,function(c){var u,l,p,d;if(!(d=c.bounds)||d.contains(i,s))return l=c.x||0,p=c.y||0,t.save(),t.translate(l,p),l=i-l,p=s-p,u=Object(o.a)(c,function(t){return i=l,o=p,(!1!==(e=t).interactive||"group"===e.marktype)&&e.bounds&&e.bounds.contains(i,o)?a.pick(t,n,r,l,p):null;var e,i,o}),t.restore(),u||((u=!1!==e.interactive&&(c.fill||c.stroke)&&l>=0&&l<=c.width&&p>=0&&p<=c.height)?c:null)})},background:function(t,e){var n=e.stroke?l:0;t("class","background"),t("d",Object(r.d)(null,e,n,n))},foreground:function(t,e,n){t("clip-path",e.clip?Object(c.a)(n,e,e):null)}}},"../../../../vega-scenegraph/src/marks/image.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/util/visit.js"),i=n("../../../../vega-scenegraph/src/util/canvas/pick.js"),o=n("../../../../vega-scenegraph/src/util/svg/translate.js");function s(t,e){var n=t.image;return n&&n.url===t.url||(n={loaded:!1,width:0,height:0},e.loadImage(t.url).then(function(e){t.image=e,t.image.url=t.url})),n}function a(t,e){return"center"===t?e/2:"right"===t?e:0}function c(t,e){return"middle"===t?e/2:"bottom"===t?e:0}e.a={type:"image",tag:"image",nested:!1,attr:function(t,e,n){var r=s(e,n),i=e.x||0,u=e.y||0,l=(null!=e.width?e.width:r.width)||0,p=(null!=e.height?e.height:r.height)||0,d=!1===e.aspect?"none":"xMidYMid";i-=a(e.align,l),u-=c(e.baseline,p),t("href",r.src||"","http://www.w3.org/1999/xlink","xlink:href"),t("transform",Object(o.a)(i,u)),t("width",l),t("height",p),t("preserveAspectRatio",d)},bound:function(t,e){var n=e.image,r=e.x||0,i=e.y||0,o=(null!=e.width?e.width:n&&n.width)||0,s=(null!=e.height?e.height:n&&n.height)||0;return r-=a(e.align,o),i-=c(e.baseline,s),t.set(r,i,r+o,i+s)},draw:function(t,e,n){var i=this;Object(r.b)(e,function(e){if(!n||n.intersects(e.bounds)){var r,o,u,l,p=s(e,i),d=e.x||0,h=e.y||0,f=(null!=e.width?e.width:p.width)||0,m=(null!=e.height?e.height:p.height)||0;d-=a(e.align,f),h-=c(e.baseline,m),!1!==e.aspect&&(o=p.width/p.height,u=e.width/e.height,o==o&&u==u&&o!==u&&(u<o?(h+=(m-(l=f/o))/2,m=l):(d+=(f-(l=m*o))/2,f=l))),p.loaded&&(t.globalAlpha=null!=(r=e.opacity)?r:1,t.drawImage(p,d,h,f,m))}})},pick:Object(i.b)(),get:s,xOffset:a,yOffset:c}},"../../../../vega-scenegraph/src/marks/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/marks/arc.js"),i=n("../../../../vega-scenegraph/src/marks/area.js"),o=n("../../../../vega-scenegraph/src/marks/group.js"),s=n("../../../../vega-scenegraph/src/marks/image.js"),a=n("../../../../vega-scenegraph/src/marks/line.js"),c=n("../../../../vega-scenegraph/src/marks/path.js"),u=n("../../../../vega-scenegraph/src/marks/rect.js"),l=n("../../../../vega-scenegraph/src/marks/rule.js"),p=n("../../../../vega-scenegraph/src/marks/shape.js"),d=n("../../../../vega-scenegraph/src/marks/symbol.js"),h=n("../../../../vega-scenegraph/src/marks/text.js"),f=n("../../../../vega-scenegraph/src/marks/trail.js");e.a={arc:r.a,area:i.a,group:o.a,image:s.a,line:a.a,path:c.a,rect:u.a,rule:l.a,shape:p.a,symbol:d.a,text:h.a,trail:f.a}},"../../../../vega-scenegraph/src/marks/line.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/path/shapes.js"),i=n("../../../../vega-scenegraph/src/marks/markMultiItemPath.js");e.a=Object(i.a)("line",r.c)},"../../../../vega-scenegraph/src/marks/markItemPath.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/bound/boundStroke.js"),i=n("../../../../vega-scenegraph/src/bound/boundContext.js"),o=n("../../../../vega-scenegraph/src/util/canvas/draw.js"),s=n("../../../../vega-scenegraph/src/util/canvas/pick.js"),a=n("../../../../vega-scenegraph/src/util/svg/translateItem.js");e.a=function(t,e){function n(t,n){var r=n.x||0,i=n.y||0;t.translate(r,i),t.beginPath(),e(t,n),t.translate(-r,-i)}return{type:t,tag:"path",nested:!1,attr:function(t,n){t("transform",Object(a.a)(n)),t("d",e(null,n))},bound:function(t,n){return e(Object(i.a)(t),n),Object(r.a)(t,n).translate(n.x||0,n.y||0)},draw:Object(o.a)(n),pick:Object(s.c)(n)}}},"../../../../vega-scenegraph/src/marks/markMultiItemPath.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/bound/boundStroke.js"),i=n("../../../../vega-scenegraph/src/bound/boundContext.js"),o=n("../../../../vega-scenegraph/src/util/canvas/draw.js"),s=n("../../../../vega-scenegraph/src/util/canvas/pick.js");e.a=function(t,e){function n(t,n){t.beginPath(),e(t,n)}var a=Object(s.a)(n);return{type:t,tag:"path",nested:!0,attr:function(t,n){var r=n.mark.items;r.length&&t("d",e(null,r))},bound:function(t,n){var o=n.items;return 0===o.length?t:(e(Object(i.a)(t),o),Object(r.a)(t,o[0]))},draw:Object(o.b)(n),pick:function(t,e,n,r,i,o){var s=e.items,c=e.bounds;return!s||!s.length||c&&!c.contains(i,o)?null:(t.pixelRatio>1&&(n*=t.pixelRatio,r*=t.pixelRatio),a(t,s,n,r)?s[0]:null)}}}},"../../../../vega-scenegraph/src/marks/path.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/bound/boundStroke.js"),i=n("../../../../vega-scenegraph/src/bound/boundContext.js"),o=n("../../../../vega-scenegraph/src/path/parse.js"),s=n("../../../../vega-scenegraph/src/path/render.js"),a=n("../../../../vega-scenegraph/src/util/canvas/draw.js"),c=n("../../../../vega-scenegraph/src/util/canvas/pick.js"),u=n("../../../../vega-scenegraph/src/util/svg/translateItem.js");function l(t,e){var n=e.path;if(null==n)return!0;var r=e.pathCache;r&&r.path===n||((e.pathCache=r=Object(o.a)(n)).path=n),Object(s.a)(t,r,e.x,e.y)}e.a={type:"path",tag:"path",nested:!1,attr:function(t,e){t("transform",Object(u.a)(e)),t("d",e.path)},bound:function(t,e){return l(Object(i.a)(t),e)?t.set(0,0,0,0):Object(r.a)(t,e)},draw:Object(a.a)(l),pick:Object(c.c)(l)}},"../../../../vega-scenegraph/src/marks/rect.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/bound/boundStroke.js"),i=n("../../../../vega-scenegraph/src/path/shapes.js"),o=n("../../../../vega-scenegraph/src/util/canvas/draw.js"),s=n("../../../../vega-scenegraph/src/util/canvas/pick.js");function a(t,e){t.beginPath(),Object(i.d)(t,e)}e.a={type:"rect",tag:"path",nested:!1,attr:function(t,e){t("d",Object(i.d)(null,e))},bound:function(t,e){var n,i;return Object(r.a)(t.set(n=e.x||0,i=e.y||0,n+e.width||0,i+e.height||0),e)},draw:Object(o.a)(a),pick:Object(s.c)(a)}},"../../../../vega-scenegraph/src/marks/rule.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/bound/boundStroke.js"),i=n("../../../../vega-scenegraph/src/util/visit.js"),o=n("../../../../vega-scenegraph/src/util/canvas/pick.js"),s=n("../../../../vega-scenegraph/src/util/canvas/stroke.js"),a=n("../../../../vega-scenegraph/src/util/svg/translateItem.js");function c(t,e,n){var r,i,o,a;return!(!e.stroke||!Object(s.a)(t,e,n))&&(r=e.x||0,i=e.y||0,o=null!=e.x2?e.x2:r,a=null!=e.y2?e.y2:i,t.beginPath(),t.moveTo(r,i),t.lineTo(o,a),!0)}e.a={type:"rule",tag:"line",nested:!1,attr:function(t,e){t("transform",Object(a.a)(e)),t("x2",null!=e.x2?e.x2-(e.x||0):0),t("y2",null!=e.y2?e.y2-(e.y||0):0)},bound:function(t,e){var n,i;return Object(r.a)(t.set(n=e.x||0,i=e.y||0,null!=e.x2?e.x2:n,null!=e.y2?e.y2:i),e)},draw:function(t,e,n){Object(i.b)(e,function(e){if(!n||n.intersects(e.bounds)){var r=null==e.opacity?1:e.opacity;r&&c(t,e,r)&&t.stroke()}})},pick:Object(o.b)(function(t,e,n,r){return!!t.isPointInStroke&&c(t,e,1)&&t.isPointInStroke(n,r)})}},"../../../../vega-scenegraph/src/marks/shape.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/path/shapes.js"),i=n("../../../../vega-scenegraph/src/marks/markItemPath.js");e.a=Object(i.a)("shape",r.e)},"../../../../vega-scenegraph/src/marks/symbol.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/path/shapes.js"),i=n("../../../../vega-scenegraph/src/marks/markItemPath.js");e.a=Object(i.a)("symbol",r.f)},"../../../../vega-scenegraph/src/marks/text.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/Bounds.js"),i=n("../../../../vega-scenegraph/src/util/text.js"),o=n("../../../../vega-scenegraph/src/util/visit.js"),s=n("../../../../vega-scenegraph/src/util/canvas/fill.js"),a=n("../../../../vega-scenegraph/src/util/canvas/pick.js"),c=n("../../../../vega-scenegraph/src/util/canvas/stroke.js"),u=n("../../../../vega-scenegraph/src/util/svg/translate.js"),l={left:"start",center:"middle",right:"end"},p=new r.a;function d(t,e,n){var r,o,s=i.c.height(e),a=e.align,c=e.radius||0,u=e.x||0,l=e.y||0,p=e.dx||0,d=(e.dy||0)+Object(i.b)(e)-Math.round(.8*s);return c&&(o=(e.theta||0)-Math.PI/2,u+=c*Math.cos(o),l+=c*Math.sin(o)),r=i.c.width(e),"center"===a?p-=r/2:"right"===a&&(p-=r),t.set(p+=u,d+=l,p+r,d+s),e.angle&&!n&&t.rotate(e.angle*Math.PI/180,u,l),t.expand(n||!r?0:1)}e.a={type:"text",tag:"text",nested:!1,attr:function(t,e){var n,r=e.dx||0,o=(e.dy||0)+Object(i.b)(e),s=e.x||0,a=e.y||0,c=e.angle||0,p=e.radius||0;p&&(n=(e.theta||0)-Math.PI/2,s+=p*Math.cos(n),a+=p*Math.sin(n)),t("text-anchor",l[e.align]||"start"),c?(n=Object(u.a)(s,a)+" rotate("+c+")",(r||o)&&(n+=" "+Object(u.a)(r,o))):n=Object(u.a)(s+r,a+o),t("transform",n)},bound:d,draw:function(t,e,n){Object(o.b)(e,function(e){var r,o,a,u,l,p;n&&!n.intersects(e.bounds)||(p=Object(i.d)(e))&&0!==(r=null==e.opacity?1:e.opacity)&&(t.font=Object(i.a)(e),t.textAlign=e.align||"left",o=e.x||0,a=e.y||0,(u=e.radius)&&(l=(e.theta||0)-Math.PI/2,o+=u*Math.cos(l),a+=u*Math.sin(l)),e.angle&&(t.save(),t.translate(o,a),t.rotate(e.angle*Math.PI/180),o=a=0),o+=e.dx||0,a+=(e.dy||0)+Object(i.b)(e),e.fill&&Object(s.a)(t,e,r)&&t.fillText(p,o,a),e.stroke&&Object(c.a)(t,e,r)&&t.strokeText(p,o,a),e.angle&&t.restore())})},pick:Object(a.b)(function(t,e,n,r,i,o){if(e.fontSize<=0)return!1;if(!e.angle)return!0;var s=d(p,e,!0),a=-e.angle*Math.PI/180,c=Math.cos(a),u=Math.sin(a),l=e.x,h=e.y,f=c*i-u*o+(l-l*c+h*u),m=u*i+c*o+(h-l*u-h*c);return s.contains(f,m)})}},"../../../../vega-scenegraph/src/marks/trail.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/path/shapes.js"),i=n("../../../../vega-scenegraph/src/marks/markMultiItemPath.js");e.a=Object(i.a)("trail",r.g)},"../../../../vega-scenegraph/src/modules.js":function(t,e,n){"use strict";n.d(e,"a",function(){return c}),e.b=function(t,e){return t=String(t||"").toLowerCase(),arguments.length>1?(u[t]=e,this):u[t]};var r=n("../../../../vega-scenegraph/src/CanvasHandler.js"),i=n("../../../../vega-scenegraph/src/CanvasRenderer.js"),o=n("../../../../vega-scenegraph/src/SVGHandler.js"),s=n("../../../../vega-scenegraph/src/SVGRenderer.js"),a=n("../../../../vega-scenegraph/src/SVGStringRenderer.js"),c={Canvas:"canvas",PNG:"png",SVG:"svg",None:"none"},u={};u.canvas=u.png={renderer:i.a,headless:i.a,handler:r.a},u.svg={renderer:s.a,headless:a.a,handler:o.a},u.none={}},"../../../../vega-scenegraph/src/path/arc.js":function(t,e,n){"use strict";e.b=function(t,e,n,i,s,a,c,u,l){var p=o.call(arguments);if(r[p])return r[p];var d=c*(Math.PI/180),h=Math.sin(d),f=Math.cos(d);n=Math.abs(n),i=Math.abs(i);var m=f*(u-t)*.5+h*(l-e)*.5,g=f*(l-e)*.5-h*(u-t)*.5,v=m*m/(n*n)+g*g/(i*i);v>1&&(v=Math.sqrt(v),n*=v,i*=v);var y=f/n,b=h/n,_=-h/i,j=f/i,x=y*u+b*l,w=_*u+j*l,O=y*t+b*e,C=_*t+j*e,k=1/((O-x)*(O-x)+(C-w)*(C-w))-.25;k<0&&(k=0);var E=Math.sqrt(k);a==s&&(E=-E);var S=.5*(x+O)-E*(C-w),P=.5*(w+C)+E*(O-x),M=Math.atan2(w-P,x-S),A=Math.atan2(C-P,O-S)-M;A<0&&1===a?A+=2*Math.PI:A>0&&0===a&&(A-=2*Math.PI);for(var T=Math.ceil(Math.abs(A/(.5*Math.PI+.001))),D=[],I=0;I<T;++I){var R=M+I*A/T,N=M+(I+1)*A/T;D[I]=[S,P,R,N,n,i,h,f]}return r[p]=D},e.a=function(t){var e=o.call(t);if(i[e])return i[e];var n=t[0],r=t[1],s=t[2],a=t[3],c=t[4],u=t[5],l=t[6],p=t[7],d=p*c,h=-l*u,f=l*c,m=p*u,g=Math.cos(s),v=Math.sin(s),y=Math.cos(a),b=Math.sin(a),_=.5*(a-s),j=Math.sin(.5*_),x=8/3*j*j/Math.sin(_),w=n+g-x*v,O=r+v+x*g,C=n+y,k=r+b,E=C+x*b,S=k-x*y;return i[e]=[d*w+h*O,f*w+m*O,d*E+h*S,f*E+m*S,d*C+h*k,f*C+m*k]};var r={},i={},o=[].join},"../../../../vega-scenegraph/src/path/curves.js":function(t,e,n){"use strict";e.a=function(t,e,n){var r=i.hasOwnProperty(t)&&i[t],o=null;r&&(o=r.curve||r[e||"vertical"],r.tension&&null!=n&&(o=o[r.tension](n)));return o};var r=n("../../../../d3-shape/index.js"),i={basis:{curve:r.c},"basis-closed":{curve:r.d},"basis-open":{curve:r.e},bundle:{curve:r.f,tension:"beta",value:.85},cardinal:{curve:r.g,tension:"tension",value:0},"cardinal-open":{curve:r.i,tension:"tension",value:0},"cardinal-closed":{curve:r.h,tension:"tension",value:0},"catmull-rom":{curve:r.j,tension:"alpha",value:.5},"catmull-rom-closed":{curve:r.k,tension:"alpha",value:.5},"catmull-rom-open":{curve:r.l,tension:"alpha",value:.5},linear:{curve:r.m},"linear-closed":{curve:r.n},monotone:{horizontal:r.p,vertical:r.o},natural:{curve:r.q},step:{curve:r.r},"step-after":{curve:r.s},"step-before":{curve:r.t}}},"../../../../vega-scenegraph/src/path/parse.js":function(t,e,n){"use strict";var r={m:2,l:2,h:1,v:1,c:6,s:4,q:4,t:2,a:7},i=[/([MLHVCSQTAZmlhvcsqtaz])/g,/###/,/(\d)([-+])/g,/\s|,|###/];e.a=function(t){var e,n,o,s,a,c,u,l,p,d,h,f=[];for(l=0,d=(e=t.slice().replace(i[0],"###$1").split(i[1]).slice(1)).length;l<d;++l){for(o=(n=e[l]).slice(1).trim().replace(i[2],"$1###$2").split(i[3]),s=[c=n.charAt(0)],p=0,h=o.length;p<h;++p)(a=+o[p])===a&&s.push(a);if(u=r[c.toLowerCase()],s.length-1>u)for(p=1,h=s.length;p<h;p+=u)f.push([c].concat(s.slice(p,p+u)));else f.push(s)}return f}},"../../../../vega-scenegraph/src/path/rectangle.js":function(t,e,n){"use strict";var r=n("../../../../d3-path/index.js");function i(t){return t.x}function o(t){return t.y}function s(t){return t.width}function a(t){return t.height}function c(t){return function(){return t}}e.a=function(){var t=i,e=o,n=s,u=a,l=c(0),p=null;function d(i,o,s){var a,c=null!=o?o:+t.call(this,i),d=null!=s?s:+e.call(this,i),h=+n.call(this,i),f=+u.call(this,i),m=+l.call(this,i);if(p||(p=a=Object(r.a)()),m<=0)p.rect(c,d,h,f);else{var g=c+h,v=d+f;p.moveTo(c+m,d),p.lineTo(g-m,d),p.quadraticCurveTo(g,d,g,d+m),p.lineTo(g,v-m),p.quadraticCurveTo(g,v,g-m,v),p.lineTo(c+m,v),p.quadraticCurveTo(c,v,c,v-m),p.lineTo(c,d+m),p.quadraticCurveTo(c,d,c+m,d),p.closePath()}if(a)return p=null,a+""||null}return d.x=function(e){return arguments.length?(t="function"==typeof e?e:c(+e),d):t},d.y=function(t){return arguments.length?(e="function"==typeof t?t:c(+t),d):e},d.width=function(t){return arguments.length?(n="function"==typeof t?t:c(+t),d):n},d.height=function(t){return arguments.length?(u="function"==typeof t?t:c(+t),d):u},d.cornerRadius=function(t){return arguments.length?(l="function"==typeof t?t:c(+t),d):l},d.context=function(t){return arguments.length?(p=null==t?null:t,d):p},d}},"../../../../vega-scenegraph/src/path/render.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/path/arc.js"),i=["l",0,0,0,0,0,0,0];function o(t,e){var n=i[0]=t[0];if("a"===n||"A"===n)i[1]=e*t[1],i[2]=e*t[2],i[6]=e*t[6],i[7]=e*t[7];else for(var r=1,o=t.length;r<o;++r)i[r]=e*t[r];return i}function s(t,e,n,i){for(var o=Object(r.b)(i[5],i[6],i[0],i[1],i[3],i[4],i[2],e,n),s=0;s<o.length;++s){var a=Object(r.a)(o[s]);t.bezierCurveTo(a[0],a[1],a[2],a[3],a[4],a[5])}}e.a=function(t,e,n,r,i){var a,c,u,l,p,d=null,h=0,f=0,m=0,g=0;null==n&&(n=0),null==r&&(r=0),null==i&&(i=1),t.beginPath&&t.beginPath();for(var v=0,y=e.length;v<y;++v){switch(a=e[v],1!==i&&(a=o(a,i)),a[0]){case"l":h+=a[1],f+=a[2],t.lineTo(h+n,f+r);break;case"L":h=a[1],f=a[2],t.lineTo(h+n,f+r);break;case"h":h+=a[1],t.lineTo(h+n,f+r);break;case"H":h=a[1],t.lineTo(h+n,f+r);break;case"v":f+=a[1],t.lineTo(h+n,f+r);break;case"V":f=a[1],t.lineTo(h+n,f+r);break;case"m":h+=a[1],f+=a[2],t.moveTo(h+n,f+r);break;case"M":h=a[1],f=a[2],t.moveTo(h+n,f+r);break;case"c":c=h+a[5],u=f+a[6],m=h+a[3],g=f+a[4],t.bezierCurveTo(h+a[1]+n,f+a[2]+r,m+n,g+r,c+n,u+r),h=c,f=u;break;case"C":h=a[5],f=a[6],m=a[3],g=a[4],t.bezierCurveTo(a[1]+n,a[2]+r,m+n,g+r,h+n,f+r);break;case"s":c=h+a[3],u=f+a[4],m=2*h-m,g=2*f-g,t.bezierCurveTo(m+n,g+r,h+a[1]+n,f+a[2]+r,c+n,u+r),m=h+a[1],g=f+a[2],h=c,f=u;break;case"S":c=a[3],u=a[4],m=2*h-m,g=2*f-g,t.bezierCurveTo(m+n,g+r,a[1]+n,a[2]+r,c+n,u+r),h=c,f=u,m=a[1],g=a[2];break;case"q":c=h+a[3],u=f+a[4],m=h+a[1],g=f+a[2],t.quadraticCurveTo(m+n,g+r,c+n,u+r),h=c,f=u;break;case"Q":c=a[3],u=a[4],t.quadraticCurveTo(a[1]+n,a[2]+r,c+n,u+r),h=c,f=u,m=a[1],g=a[2];break;case"t":c=h+a[1],u=f+a[2],null===d[0].match(/[QqTt]/)?(m=h,g=f):"t"===d[0]?(m=2*h-l,g=2*f-p):"q"===d[0]&&(m=2*h-m,g=2*f-g),l=m,p=g,t.quadraticCurveTo(m+n,g+r,c+n,u+r),f=u,m=(h=c)+a[1],g=f+a[2];break;case"T":c=a[1],u=a[2],m=2*h-m,g=2*f-g,t.quadraticCurveTo(m+n,g+r,c+n,u+r),h=c,f=u;break;case"a":s(t,h+n,f+r,[a[1],a[2],a[3],a[4],a[5],a[6]+h+n,a[7]+f+r]),h+=a[6],f+=a[7];break;case"A":s(t,h+n,f+r,[a[1],a[2],a[3],a[4],a[5],a[6]+n,a[7]+r]),h=a[6],f=a[7];break;case"z":case"Z":t.closePath()}d=a}}},"../../../../vega-scenegraph/src/path/shapes.js":function(t,e,n){"use strict";e.a=function(t,e){return d.context(t)(e)},e.b=function(t,e){var n=e[0],i=n.interpolate||"linear";return("horizontal"===n.orient?f:h).curve(Object(r.a)(i,n.orient,n.tension)).context(t)(e)},e.c=function(t,e){var n=e[0],i=n.interpolate||"linear";return m.curve(Object(r.a)(i,n.orient,n.tension)).context(t)(e)},e.d=function(t,e,n,r){return g.context(t)(e,n,r)},e.e=function(t,e){return(e.mark.shape||e.shape).context(t)(e)},e.f=function(t,e){return v.context(t)(e)},e.g=function(t,e){return y.context(t)(e)};var r=n("../../../../vega-scenegraph/src/path/curves.js"),i=n("../../../../vega-scenegraph/src/path/symbols.js"),o=n("../../../../vega-scenegraph/src/path/rectangle.js"),s=n("../../../../vega-scenegraph/src/path/trail.js"),a=n("../../../../d3-shape/index.js");function c(t){return t.x||0}function u(t){return t.y||0}function l(t){return t.cornerRadius||0}function p(t){return!(!1===t.defined)}var d=Object(a.a)().startAngle(function(t){return t.startAngle||0}).endAngle(function(t){return t.endAngle||0}).padAngle(function(t){return t.padAngle||0}).innerRadius(function(t){return t.innerRadius||0}).outerRadius(function(t){return t.outerRadius||0}).cornerRadius(l),h=Object(a.b)().x(c).y1(u).y0(function(t){return(t.y||0)+(t.height||0)}).defined(p),f=Object(a.b)().y(u).x1(c).x0(function(t){return(t.x||0)+(t.width||0)}).defined(p),m=Object(a.u)().x(c).y(u).defined(p),g=Object(o.a)().x(c).y(u).width(function(t){return t.width||0}).height(function(t){return t.height||0}).cornerRadius(l),v=Object(a.v)().type(function(t){return Object(i.a)(t.shape||"circle")}).size(function(t){return null==t.size?64:t.size}),y=Object(s.a)().x(c).y(u).defined(p).size(function(t){return t.size||1})},"../../../../vega-scenegraph/src/path/symbols.js":function(t,e,n){"use strict";e.a=function(t){return a.hasOwnProperty(t)?a[t]:function(t){if(!c.hasOwnProperty(t)){var e=Object(r.a)(t);c[t]={draw:function(t,n){Object(i.a)(t,e,0,0,Math.sqrt(n)/2)}}}return c[t]}(t)};var r=n("../../../../vega-scenegraph/src/path/parse.js"),i=n("../../../../vega-scenegraph/src/path/render.js"),o=2*Math.PI,s=Math.sqrt(3)/2,a={circle:{draw:function(t,e){var n=Math.sqrt(e)/2;t.moveTo(n,0),t.arc(0,0,n,0,o)}},cross:{draw:function(t,e){var n=Math.sqrt(e)/2,r=n/2.5;t.moveTo(-n,-r),t.lineTo(-n,r),t.lineTo(-r,r),t.lineTo(-r,n),t.lineTo(r,n),t.lineTo(r,r),t.lineTo(n,r),t.lineTo(n,-r),t.lineTo(r,-r),t.lineTo(r,-n),t.lineTo(-r,-n),t.lineTo(-r,-r),t.closePath()}},diamond:{draw:function(t,e){var n=Math.sqrt(e)/2;t.moveTo(-n,0),t.lineTo(0,-n),t.lineTo(n,0),t.lineTo(0,n),t.closePath()}},square:{draw:function(t,e){var n=Math.sqrt(e),r=-n/2;t.rect(r,r,n,n)}},"triangle-up":{draw:function(t,e){var n=Math.sqrt(e)/2,r=s*n;t.moveTo(0,-r),t.lineTo(-n,r),t.lineTo(n,r),t.closePath()}},"triangle-down":{draw:function(t,e){var n=Math.sqrt(e)/2,r=s*n;t.moveTo(0,r),t.lineTo(-n,-r),t.lineTo(n,-r),t.closePath()}},"triangle-right":{draw:function(t,e){var n=Math.sqrt(e)/2,r=s*n;t.moveTo(r,0),t.lineTo(-r,-n),t.lineTo(-r,n),t.closePath()}},"triangle-left":{draw:function(t,e){var n=Math.sqrt(e)/2,r=s*n;t.moveTo(-r,0),t.lineTo(r,-n),t.lineTo(r,n),t.closePath()}}};var c={}},"../../../../vega-scenegraph/src/path/trail.js":function(t,e,n){"use strict";var r=n("../../../../d3-path/index.js"),i=Math.PI;e.a=function(){var t,e,n,o,s,a,c,u,l=null;function p(t,e,n){var r=n/2;if(s){var o=c-e,p=t-a;if(o||p){var d=Math.sqrt(o*o+p*p),h=(o/=d)*u,f=(p/=d)*u,m=Math.atan2(p,o);l.moveTo(a-h,c-f),l.lineTo(t-o*r,e-p*r),l.arc(t,e,r,m-i,m),l.lineTo(a+h,c+f),l.arc(a,c,u,m,m+i)}else l.arc(t,e,r,0,2*i);l.closePath()}else s=1;a=t,c=e,u=r}function d(i){var a,c,u,d=i.length,h=!1;for(null==l&&(l=u=Object(r.a)()),a=0;a<=d;++a)!(a<d&&o(c=i[a],a,i))===h&&(h=!h)&&(s=0),h&&p(+t(c,a,i),+e(c,a,i),+n(c,a,i));if(u)return l=null,u+""||null}return d.x=function(e){return arguments.length?(t=e,d):t},d.y=function(t){return arguments.length?(e=t,d):e},d.size=function(t){return arguments.length?(n=t,d):n},d.defined=function(t){return arguments.length?(o=t,d):o},d.context=function(t){return arguments.length?(l=null==t?null:t,d):l},d}},"../../../../vega-scenegraph/src/util/canvas/clip.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");e.a=function(t,e){var n=e.clip;if(t.save(),t.beginPath(),Object(r.w)(n))n(t);else{var i=e.group;t.rect(0,0,i.width||0,i.height||0)}t.clip()}},"../../../../vega-scenegraph/src/util/canvas/color.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/util/canvas/gradient.js");e.a=function(t,e,n){return n.id?Object(r.a)(t,n,e.bounds):n}},"../../../../vega-scenegraph/src/util/canvas/draw.js":function(t,e,n){"use strict";e.a=function(t){return function(e,n,r){Object(o.b)(n,function(n){r&&!r.intersects(n.bounds)||s(t,e,n,n)})}},e.b=function(t){return function(e,n,r){!n.items.length||r&&!r.intersects(n.bounds)||s(t,e,n.items[0],n.items)}};var r=n("../../../../vega-scenegraph/src/util/canvas/fill.js"),i=n("../../../../vega-scenegraph/src/util/canvas/stroke.js"),o=n("../../../../vega-scenegraph/src/util/visit.js");function s(t,e,n,o){var s=null==n.opacity?1:n.opacity;0!==s&&(t(e,o)||(n.fill&&Object(r.a)(e,n,s)&&e.fill(),n.stroke&&Object(i.a)(e,n,s)&&e.stroke()))}},"../../../../vega-scenegraph/src/util/canvas/fill.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/util/canvas/color.js");e.a=function(t,e,n){return(n*=null==e.fillOpacity?1:e.fillOpacity)>0&&(t.globalAlpha=n,t.fillStyle=Object(r.a)(t,e,e.fill),!0)}},"../../../../vega-scenegraph/src/util/canvas/gradient.js":function(t,e,n){"use strict";e.a=function(t,e,n){for(var r=n.width(),i=n.height(),o=n.x1+e.x1*r,s=n.y1+e.y1*i,a=n.x1+e.x2*r,c=n.y1+e.y2*i,u=e.stops,l=0,p=u.length,d=t.createLinearGradient(o,s,a,c);l<p;++l)d.addColorStop(u[l].offset,u[l].color);return d}},"../../../../vega-scenegraph/src/util/canvas/pick.js":function(t,e,n){"use strict";e.b=o,e.a=s,e.c=function(t){return o(s(t))};var r=n("../../../../vega-scenegraph/src/util/visit.js"),i=function(){return!0};function o(t){return t||(t=i),function(e,n,i,o,s,a){return e.pixelRatio>1&&(i*=e.pixelRatio,o*=e.pixelRatio),Object(r.a)(n,function(n){var r=n.bounds;if((!r||r.contains(s,a))&&r)return t(e,n,i,o,s,a)?n:void 0})}}function s(t,e){return function(n,r,i,o){var s,a,c=Array.isArray(r)?r[0]:r,u=null==e?c.fill:e,l=c.stroke&&n.isPointInStroke;return l&&(s=c.strokeWidth,a=c.strokeCap,n.lineWidth=null!=s?s:1,n.lineCap=null!=a?a:"butt"),!t(n,r)&&(u&&n.isPointInPath(i,o)||l&&n.isPointInStroke(i,o))}}},"../../../../vega-scenegraph/src/util/canvas/resize.js":function(t,e,n){"use strict";var r="undefined"!=typeof window?window.devicePixelRatio||1:1;e.a=function(t,e,n,i,o){var s="undefined"!=typeof HTMLElement&&t instanceof HTMLElement&&null!=t.parentNode,a=t.getContext("2d"),c=s?r:o;return t.width=e*c,t.height=n*c,s&&1!==c&&(t.style.width=e+"px",t.style.height=n+"px"),a.pixelRatio=c,a.setTransform(c,0,0,c,c*i[0],c*i[1]),t}},"../../../../vega-scenegraph/src/util/canvas/stroke.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/util/canvas/color.js"),i=[];e.a=function(t,e,n){var o=null!=(o=e.strokeWidth)?o:1;return!(o<=0)&&((n*=null==e.strokeOpacity?1:e.strokeOpacity)>0&&(t.globalAlpha=n,t.strokeStyle=Object(r.a)(t,e,e.stroke),t.lineWidth=o,t.lineCap=e.strokeCap||"butt",t.lineJoin=e.strokeJoin||"miter",t.miterLimit=e.strokeMiterLimit||10,t.setLineDash&&(t.setLineDash(e.strokeDash||i),t.lineDashOffset=e.strokeDashOffset||0),!0))}},"../../../../vega-scenegraph/src/util/dom.js":function(t,e,n){"use strict";function r(t,e,n){return!t&&"undefined"!=typeof document&&document.createElement&&(t=document),t?n?t.createElementNS(n,e):t.createElement(e):null}e.d=r,e.e=function(t,e){e=e.toLowerCase();for(var n=t.childNodes,r=0,i=n.length;r<i;++r)if(n[r].tagName.toLowerCase()===e)return n[r]},e.b=function(t,e,n,i){var o,s=t.childNodes[e];s&&s.tagName.toLowerCase()===n.toLowerCase()||(o=s||null,s=r(t.ownerDocument,n,i),t.insertBefore(s,o));return s},e.c=function(t,e){var n=t.childNodes,r=n.length;for(;r>e;)t.removeChild(n[--r]);return t},e.a=function(t){return"mark-"+t.marktype+(t.role?" role-"+t.role:"")+(t.name?" "+t.name:"")}},"../../../../vega-scenegraph/src/util/equal.js":function(t,e,n){"use strict";n("../../../../vega-scenegraph/src/path/parse.js"),n("../../../../vega-util/index.js")},"../../../../vega-scenegraph/src/util/point.js":function(t,e,n){"use strict";e.a=function(t,e){var n=e.getBoundingClientRect();return[t.clientX-n.left-(e.clientLeft||0),t.clientY-n.top-(e.clientTop||0)]}},"../../../../vega-scenegraph/src/util/serialize.js":function(t,e,n){"use strict";e.b=function(t,e){return JSON.stringify(t,i,e)},e.a=function(t){return function t(e){var n,i,o,s=e.marktype,a=e.items;if(a)for(i=0,o=a.length;i<o;++i)n=s?"mark":"group",a[i][n]=e,a[i].zindex&&(a[i][n].zdirty=!0),"group"===(s||n)&&t(a[i]);s&&Object(r.a)(e);return e}("string"==typeof t?JSON.parse(t):t)};var r=n("../../../../vega-scenegraph/src/bound/boundMark.js"),i=["marktype","name","role","interactive","clip","items","zindex","x","y","width","height","align","baseline","fill","fillOpacity","opacity","stroke","strokeOpacity","strokeWidth","strokeCap","strokeDash","strokeDashOffset","startAngle","endAngle","innerRadius","outerRadius","cornerRadius","padAngle","interpolate","tension","orient","defined","url","path","x2","y2","size","shape","text","angle","theta","radius","dx","dy","font","fontSize","fontWeight","fontStyle","fontVariant"]},"../../../../vega-scenegraph/src/util/svg/clip.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js"),i=1;e.a=function(t,e,n){var o=e.clip,s=t._defs,a=e.clip_id||(e.clip_id="clip"+i++),c=s.clipping[a]||(s.clipping[a]={id:a});return Object(r.w)(o)?c.path=o(null):(c.width=n.width||0,c.height=n.height||0),"url(#"+a+")"}},"../../../../vega-scenegraph/src/util/svg/metadata.js":function(t,e,n){"use strict";e.a={version:"1.1",xmlns:"http://www.w3.org/2000/svg","xmlns:xlink":"http://www.w3.org/1999/xlink"}},"../../../../vega-scenegraph/src/util/svg/styles.js":function(t,e,n){"use strict";n.d(e,"b",function(){return r}),n.d(e,"a",function(){return i});var r={fill:"fill",fillOpacity:"fill-opacity",stroke:"stroke",strokeOpacity:"stroke-opacity",strokeWidth:"stroke-width",strokeCap:"stroke-linecap",strokeJoin:"stroke-linejoin",strokeDash:"stroke-dasharray",strokeDashOffset:"stroke-dashoffset",strokeMiterLimit:"stroke-miterlimit",opacity:"opacity"},i=Object.keys(r)},"../../../../vega-scenegraph/src/util/svg/translate.js":function(t,e,n){"use strict";e.a=function(t,e){return"translate("+t+","+e+")"}},"../../../../vega-scenegraph/src/util/svg/translateItem.js":function(t,e,n){"use strict";var r=n("../../../../vega-scenegraph/src/util/svg/translate.js");e.a=function(t){return Object(r.a)(t.x||0,t.y||0)}},"../../../../vega-scenegraph/src/util/tags.js":function(t,e,n){"use strict";e.b=function(t,e,n){var r,i,o="<"+t;if(e)for(r in e)null!=(i=e[r])&&(o+=" "+r+'="'+i+'"');n&&(o+=" "+n);return o+">"},e.a=function(t){return"</"+t+">"}},"../../../../vega-scenegraph/src/util/text.js":function(t,e,n){"use strict";n.d(e,"c",function(){return s}),e.d=h,e.a=f,e.b=function(t){var e=t.baseline,n=p(t);return Math.round("top"===e?.79*n:"middle"===e?.3*n:"bottom"===e?-.21*n:0)};var r,i,o=n("../../../../vega-canvas/index.js"),s={height:p,measureWidth:u,estimateWidth:a,width:a,canvas:d};function a(t){return i=p(t),c(h(t))}function c(t){return~~(.8*t.length*i)}function u(t){return r.font=f(t),l(h(t))}function l(t){return r.measureText(t).width}function p(t){return null!=t.fontSize?t.fontSize:11}function d(t){r=t&&(r=Object(o.a)(1,1))?r.getContext("2d"):null,s.width=r?u:a}function h(t){var e=t.text;return null==e?"":t.limit>0?function(t){var e,n=+t.limit,o=t.text+"";r?(r.font=f(t),e=l):(i=p(t),e=c);if(e(o)<n)return o;var s,a=t.ellipsis||"…",u="rtl"===t.dir,d=0,h=o.length;if(n-=e(a),u){for(;d<h;)s=d+h>>>1,e(o.slice(s))>n?d=s+1:h=s;return a+o.slice(d)}for(;d<h;)s=1+(d+h>>>1),e(o.slice(0,s))<n?d=s:h=s-1;return o.slice(0,d)+a}(t):e+""}function f(t,e){var n=t.font;return e&&n&&(n=String(n).replace(/"/g,"'")),(t.fontStyle?t.fontStyle+" ":"")+(t.fontVariant?t.fontVariant+" ":"")+(t.fontWeight?t.fontWeight+" ":"")+p(t)+"px "+(n||"sans-serif")}d(!0)},"../../../../vega-scenegraph/src/util/visit.js":function(t,e,n){"use strict";function r(t,e){return t.zindex-e.zindex||t.index-e.index}function i(t){if(!t.zdirty)return t.zitems;var e,n,i,o=t.items,s=[];for(n=0,i=o.length;n<i;++n)(e=o[n]).index=n,e.zindex&&s.push(e);return t.zdirty=!1,t.zitems=s.sort(r)}e.b=function(t,e){var n,r,o=t.items;if(!o||!o.length)return;var s=i(t);if(s&&s.length){for(n=0,r=o.length;n<r;++n)o[n].zindex||e(o[n]);o=s}for(n=0,r=o.length;n<r;++n)e(o[n])},e.a=function(t,e){var n,r,o=t.items;if(!o||!o.length)return null;var s=i(t);s&&s.length&&(o=s);for(r=o.length;--r>=0;)if(n=e(o[r]))return n;if(o===s)for(o=t.items,r=o.length;--r>=0;)if(!o[r].zindex&&(n=e(o[r])))return n;return null}},"../../../../vega-statistics/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-statistics/src/bin.js");n.d(e,"a",function(){return r.a});var i=n("../../../../vega-statistics/src/bootstrapCI.js");n.d(e,"b",function(){return i.a});var o=n("../../../../vega-statistics/src/quartiles.js");n.d(e,"c",function(){return o.a});var s=n("../../../../vega-statistics/src/random.js");n.d(e,"d",function(){return s.a});n("../../../../vega-statistics/src/integer.js");var a=n("../../../../vega-statistics/src/kde.js");n.d(e,"e",function(){return a.a});var c=n("../../../../vega-statistics/src/mixture.js");n.d(e,"f",function(){return c.a});var u=n("../../../../vega-statistics/src/normal.js");n.d(e,"g",function(){return u.a});var l=n("../../../../vega-statistics/src/uniform.js");n.d(e,"h",function(){return l.a})},"../../../../vega-statistics/src/bin.js":function(t,e,n){"use strict";e.a=function(t){var e,n,r,i,o,s,a,c,u=t.maxbins||20,l=t.base||10,p=Math.log(l),d=t.divide||[5,2],h=t.extent[0],f=t.extent[1],m=f-h;if(t.step)e=t.step;else if(t.steps){for(o=m/u,s=0,a=t.steps.length;s<a&&t.steps[s]<o;++s);e=t.steps[Math.max(0,s-1)]}else{for(n=Math.ceil(Math.log(u)/p),r=t.minstep||0,e=Math.max(r,Math.pow(l,Math.round(Math.log(m)/p)-n));Math.ceil(m/e)>u;)e*=l;for(s=0,a=d.length;s<a;++s)(o=e/d[s])>=r&&m/o<=u&&(e=o)}return i=(o=Math.log(e))>=0?0:1+~~(-o/p),c=Math.pow(l,-i-1),(t.nice||void 0===t.nice)&&(h=h<(o=Math.floor(h/e+c)*e)?o-e:o,f=Math.ceil(f/e)*e),{start:h,stop:f,step:e}}},"../../../../vega-statistics/src/bootstrapCI.js":function(t,e,n){"use strict";var r=n("../../../../vega-statistics/src/numbers.js"),i=n("../../../../vega-statistics/src/random.js"),o=n("../../../../d3-array/index.js");e.a=function(t,e,n,s){var a,c,u,l,p=Object(r.a)(t,s),d=p.length,h=e;for(u=0,l=Array(h);u<h;++u){for(a=0,c=0;c<d;++c)a+=p[~~(Object(i.a)()*d)];l[u]=a/d}return[Object(o.m)(l.sort(o.a),n/2),Object(o.m)(l,1-n/2)]}},"../../../../vega-statistics/src/integer.js":function(t,e,n){"use strict";n("../../../../vega-statistics/src/random.js")},"../../../../vega-statistics/src/kde.js":function(t,e,n){"use strict";var r=n("../../../../vega-statistics/src/normal.js"),i=n("../../../../vega-statistics/src/quartiles.js"),o=n("../../../../vega-statistics/src/random.js"),s=n("../../../../d3-array/index.js");e.a=function(t,e){var n=Object(r.a)(),a={},c=0;return a.data=function(n){return arguments.length?(t=n,c=n?n.length:0,a.bandwidth(e)):t},a.bandwidth=function(n){return arguments.length?(!(e=n)&&t&&(o=(r=t).length,c=Object(i.a)(r),u=(c[2]-c[0])/1.34,e=1.06*Math.min(Math.sqrt(Object(s.t)(r)),u)*Math.pow(o,-.2)),a):e;var r,o,c,u},a.sample=function(){return t[~~(Object(o.a)()*c)]+e*n.sample()},a.pdf=function(r){for(var i=0,o=0;o<c;++o)i+=n.pdf((r-t[o])/e);return i/e/c},a.cdf=function(r){for(var i=0,o=0;o<c;++o)i+=n.cdf((r-t[o])/e);return i/c},a.icdf=function(){throw Error("KDE icdf not supported.")},a.data(t)}},"../../../../vega-statistics/src/mixture.js":function(t,e,n){"use strict";var r=n("../../../../vega-statistics/src/random.js");e.a=function(t,e){var n,i={},o=0;return i.weights=function(t){return arguments.length?(n=function(t){var e,n=[],r=0;for(e=0;e<o;++e)r+=n[e]=null==t[e]?1:+t[e];for(e=0;e<o;++e)n[e]/=r;return n}(e=t||[]),i):e},i.distributions=function(n){return arguments.length?(n?(o=n.length,t=n):(o=0,t=[]),i.weights(e)):t},i.sample=function(){for(var e=Object(r.a)(),i=t[o-1],s=n[0],a=0;a<o-1;s+=n[++a])if(e<s){i=t[a];break}return i.sample()},i.pdf=function(e){for(var r=0,i=0;i<o;++i)r+=n[i]*t[i].pdf(e);return r},i.cdf=function(e){for(var r=0,i=0;i<o;++i)r+=n[i]*t[i].cdf(e);return r},i.icdf=function(){throw Error("Mixture icdf not supported.")},i.distributions(t).weights(e)}},"../../../../vega-statistics/src/normal.js":function(t,e,n){"use strict";var r=n("../../../../vega-statistics/src/random.js");e.a=function(t,e){var n,i,o=NaN,s={};return s.mean=function(t){return arguments.length?(n=t||0,o=NaN,s):n},s.stdev=function(t){return arguments.length?(i=null==t?1:t,o=NaN,s):i},s.sample=function(){var t,e,s=0,a=0;if(o==o)return s=o,o=NaN,s;do{t=(s=2*Object(r.a)()-1)*s+(a=2*Object(r.a)()-1)*a}while(0===t||t>1);return e=Math.sqrt(-2*Math.log(t)/t),o=n+a*e*i,n+s*e*i},s.pdf=function(t){var e=Math.exp(Math.pow(t-n,2)/(-2*Math.pow(i,2)));return 1/(i*Math.sqrt(2*Math.PI))*e},s.cdf=function(t){var e,r=(t-n)/i,o=Math.abs(r);if(o>37)e=0;else{var s=Math.exp(-o*o/2);o<7.07106781186547?(e=s*((((((.0352624965998911*o+.700383064443688)*o+6.37396220353165)*o+33.912866078383)*o+112.079291497871)*o+221.213596169931)*o+220.206867912376),e/=((((((.0883883476483184*o+1.75566716318264)*o+16.064177579207)*o+86.7807322029461)*o+296.564248779674)*o+637.333633378831)*o+793.826512519948)*o+440.413735824752):e=s/(o+1/(o+2/(o+3/(o+4/(o+.65)))))/2.506628274631}return r>0?1-e:e},s.icdf=function(t){if(t<=0||t>=1)return NaN;var e=2*t-1,r=8*(Math.PI-3)/(3*Math.PI*(4-Math.PI)),o=2/(Math.PI*r)+Math.log(1-Math.pow(e,2))/2,s=Math.log(1-e*e)/r,a=(e>0?1:-1)*Math.sqrt(Math.sqrt(o*o-s)-o);return n+i*Math.SQRT2*a},s.mean(t).stdev(e)}},"../../../../vega-statistics/src/numbers.js":function(t,e,n){"use strict";function r(t){return null===t?NaN:+t}e.a=function(t,e){var n,i=[],o=t.length,s=-1;if(null==e)for(;++s<o;)isNaN(n=r(t[s]))||i.push(n);else for(;++s<o;)isNaN(n=r(e(t[s],s,t)))||i.push(n);return i}},"../../../../vega-statistics/src/quartiles.js":function(t,e,n){"use strict";var r=n("../../../../vega-statistics/src/numbers.js"),i=n("../../../../d3-array/index.js");e.a=function(t,e){var n=Object(r.a)(t,e);return[Object(i.m)(n.sort(i.a),.25),Object(i.m)(n,.5),Object(i.m)(n,.75)]}},"../../../../vega-statistics/src/random.js":function(t,e,n){"use strict";n.d(e,"a",function(){return r});var r=Math.random},"../../../../vega-statistics/src/uniform.js":function(t,e,n){"use strict";var r=n("../../../../vega-statistics/src/random.js");e.a=function(t,e){null==e&&(e=null==t?1:t,t=0);var n,i,o,s={};return s.min=function(t){return arguments.length?(o=i-(n=t||0),s):n},s.max=function(t){return arguments.length?(o=(i=t||0)-n,s):i},s.sample=function(){return n+o*Object(r.a)()},s.pdf=function(t){return t>=n&&t<=i?1/o:0},s.cdf=function(t){return t<n?0:t>i?1:(t-n)/o},s.icdf=function(t){return t>=0&&t<=1?n+t*o:NaN},s.min(t).max(e)}},"../../../../vega-transforms/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../vega-transforms/src/Aggregate.js");n.d(e,"aggregate",function(){return r.a});var i=n("../../../../vega-transforms/src/Bin.js");n.d(e,"bin",function(){return i.a});var o=n("../../../../vega-transforms/src/Collect.js");n.d(e,"collect",function(){return o.a});var s=n("../../../../vega-transforms/src/Compare.js");n.d(e,"compare",function(){return s.a});var a=n("../../../../vega-transforms/src/CountPattern.js");n.d(e,"countpattern",function(){return a.a});var c=n("../../../../vega-transforms/src/Cross.js");n.d(e,"cross",function(){return c.a});var u=n("../../../../vega-transforms/src/Density.js");n.d(e,"density",function(){return u.a});var l=n("../../../../vega-transforms/src/Extent.js");n.d(e,"extent",function(){return l.a});var p=n("../../../../vega-transforms/src/Facet.js");n.d(e,"facet",function(){return p.a});var d=n("../../../../vega-transforms/src/Field.js");n.d(e,"field",function(){return d.a});var h=n("../../../../vega-transforms/src/Filter.js");n.d(e,"filter",function(){return h.a});var f=n("../../../../vega-transforms/src/Flatten.js");n.d(e,"flatten",function(){return f.a});var m=n("../../../../vega-transforms/src/Fold.js");n.d(e,"fold",function(){return m.a});var g=n("../../../../vega-transforms/src/Formula.js");n.d(e,"formula",function(){return g.a});var v=n("../../../../vega-transforms/src/Generate.js");n.d(e,"generate",function(){return v.a});var y=n("../../../../vega-transforms/src/Impute.js");n.d(e,"impute",function(){return y.a});var b=n("../../../../vega-transforms/src/JoinAggregate.js");n.d(e,"joinaggregate",function(){return b.a});var _=n("../../../../vega-transforms/src/Key.js");n.d(e,"key",function(){return _.a});var j=n("../../../../vega-transforms/src/Lookup.js");n.d(e,"lookup",function(){return j.a});var x=n("../../../../vega-transforms/src/MultiExtent.js");n.d(e,"multiextent",function(){return x.a});var w=n("../../../../vega-transforms/src/MultiValues.js");n.d(e,"multivalues",function(){return w.a});var O=n("../../../../vega-transforms/src/Params.js");n.d(e,"params",function(){return O.a});var C=n("../../../../vega-transforms/src/PreFacet.js");n.d(e,"prefacet",function(){return C.a});var k=n("../../../../vega-transforms/src/Project.js");n.d(e,"project",function(){return k.a});var E=n("../../../../vega-transforms/src/Proxy.js");n.d(e,"proxy",function(){return E.a});var S=n("../../../../vega-transforms/src/Relay.js");n.d(e,"relay",function(){return S.a});var P=n("../../../../vega-transforms/src/Sample.js");n.d(e,"sample",function(){return P.a});var M=n("../../../../vega-transforms/src/Sequence.js");n.d(e,"sequence",function(){return M.a});var A=n("../../../../vega-transforms/src/Sieve.js");n.d(e,"sieve",function(){return A.a});var T=n("../../../../vega-transforms/src/Subflow.js");n.d(e,"subflow",function(){return T.a});var D=n("../../../../vega-transforms/src/TupleIndex.js");n.d(e,"tupleindex",function(){return D.a});var I=n("../../../../vega-transforms/src/Values.js");n.d(e,"values",function(){return I.a});var R=n("../../../../vega-transforms/src/Window.js");n.d(e,"window",function(){return R.a})},"../../../../vega-transforms/src/Aggregate.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../vega-transforms/src/util/AggregateKeys.js"),i=n("../../../../vega-transforms/src/util/AggregateOps.js"),o=n("../../../../vega-transforms/src/util/TupleStore.js"),s=n("../../../../vega-dataflow/index.js"),a=n("../../../../vega-util/index.js");function c(t){s.d.call(this,null,t),this._adds=[],this._mods=[],this._alen=0,this._mlen=0,this._drop=!0,this._cross=!1,this._dims=[],this._dnames=[],this._measures=[],this._countOnly=!1,this._counts=null,this._prev=null,this._inputs=null,this._outputs=null}c.Definition={type:"Aggregate",metadata:{generates:!0,changes:!0},params:[{name:"groupby",type:"field",array:!0},{name:"ops",type:"enum",array:!0,values:i.a},{name:"fields",type:"field",null:!0,array:!0},{name:"as",type:"string",null:!0,array:!0},{name:"drop",type:"boolean",default:!0},{name:"cross",type:"boolean",default:!1},{name:"key",type:"field"}]};var u=Object(a.s)(c,s.d);u.transform=function(t,e){var n,r=this,i=e.fork(e.NO_SOURCE|e.NO_FIELDS);return this.stamp=i.stamp,this.value&&((n=t.modified())||e.modified(this._inputs))?(this._prev=this.value,this.value=n?this.init(t):{},e.visit(e.SOURCE,function(t){r.add(t)})):(this.value=this.value||this.init(t),e.visit(e.REM,function(t){r.rem(t)}),e.visit(e.ADD,function(t){r.add(t)})),i.modifies(this._outputs),r._drop=!1!==t.drop,t.cross&&r._dims.length>1&&(r._drop=!1,this.cross()),r.changes(i)},u.cross=function(){var t=this,e=t.value,n=t._dnames,r=n.map(function(){return{}}),i=n.length;function o(t){var e,o,s,a;for(e in t)for(s=t[e].tuple,o=0;o<i;++o)r[o][a=s[n[o]]]=a}o(t._prev),o(e),function o(s,a,c){var u,l,p=n[c],d=r[c++];for(u in d)a[p]=d[u],l=s?s+"|"+u:u,c<i?o(l,a,c):e[l]||t.cell(l,a)}("",{},0)},u.init=function(t){var e=this._inputs=[],n=this._outputs=[],o={};function s(t){for(var n,r=Object(a.g)(Object(a.e)(t)),i=0,s=r.length;i<s;++i)o[n=r[i]]||(o[n]=1,e.push(n))}this._dims=Object(a.g)(t.groupby),this._dnames=this._dims.map(function(t){var e=Object(a.f)(t);return s(t),n.push(e),e}),this.cellkey=t.key?t.key:Object(r.a)(this._dims),this._countOnly=!0,this._counts=[],this._measures=[];var c,u,l,p,d,h,f=t.fields||[null],m=t.ops||["count"],g=t.as||[],v=f.length,y={};for(v!==m.length&&Object(a.k)("Unmatched number of fields and aggregate ops."),h=0;h<v;++h)c=f[h],u=m[h],null==c&&"count"!==u&&Object(a.k)("Null aggregate field specified."),p=Object(a.f)(c),d=Object(i.d)(u,p,g[h]),n.push(d),"count"!==u?((l=y[p])||(s(c),(l=y[p]=[]).field=c,this._measures.push(l)),"count"!==u&&(this._countOnly=!1),l.push(Object(i.c)(u,d))):this._counts.push(d);return this._measures=this._measures.map(function(t){return Object(i.b)(t,t.field)}),{}},u.cellkey=Object(r.a)(),u.cell=function(t,e){var n=this.value[t];return n?0===n.num&&this._drop&&n.stamp<this.stamp?(n.stamp=this.stamp,this._adds[this._alen++]=n):n.stamp<this.stamp&&(n.stamp=this.stamp,this._mods[this._mlen++]=n):(n=this.value[t]=this.newcell(t,e),this._adds[this._alen++]=n),n},u.newcell=function(t,e){var n={key:t,num:0,agg:null,tuple:this.newtuple(e,this._prev&&this._prev[t]),stamp:this.stamp,store:!1};if(!this._countOnly){var r,i=this._measures,s=i.length;for(n.agg=Array(s),r=0;r<s;++r)n.agg[r]=new i[r](n)}return n.store&&(n.data=new o.a),n},u.newtuple=function(t,e){var n,r,i=this._dnames,o=this._dims,a={};for(n=0,r=o.length;n<r;++n)a[i[n]]=o[n](t);return e?Object(s.l)(e.tuple,a):Object(s.h)(a)},u.add=function(t){var e,n,r,i=this.cellkey(t),o=this.cell(i,t);if(o.num+=1,!this._countOnly)for(o.store&&o.data.add(t),n=0,r=(e=o.agg).length;n<r;++n)e[n].add(e[n].get(t),t)},u.rem=function(t){var e,n,r,i=this.cellkey(t),o=this.cell(i,t);if(o.num-=1,!this._countOnly)for(o.store&&o.data.rem(t),n=0,r=(e=o.agg).length;n<r;++n)e[n].rem(e[n].get(t),t)},u.celltuple=function(t){var e,n,r,i=t.tuple,o=this._counts;for(t.store&&t.data.values(),n=0,r=o.length;n<r;++n)i[o[n]]=t.num;if(!this._countOnly)for(n=0,r=(e=t.agg).length;n<r;++n)e[n].set(i);return i},u.changes=function(t){var e,n,r,i,o=this._adds,s=this._mods,a=this._prev,c=this._drop,u=t.add,l=t.rem,p=t.mod;if(a)for(n in a)e=a[n],c&&!e.num||l.push(e.tuple);for(r=0,i=this._alen;r<i;++r)u.push(this.celltuple(o[r])),o[r]=null;for(r=0,i=this._mlen;r<i;++r)(0===(e=s[r]).num&&c?l:p).push(this.celltuple(e)),s[r]=null;return this._alen=this._mlen=0,this._prev=null,t}},"../../../../vega-transforms/src/Bin.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-statistics/index.js"),o=n("../../../../vega-util/index.js");function s(t){r.d.call(this,null,t)}s.Definition={type:"Bin",metadata:{modifies:!0},params:[{name:"field",type:"field",required:!0},{name:"anchor",type:"number"},{name:"maxbins",type:"number",default:20},{name:"base",type:"number",default:10},{name:"divide",type:"number",array:!0,default:[5,2]},{name:"extent",type:"number",array:!0,length:2,required:!0},{name:"step",type:"number"},{name:"steps",type:"number",array:!0},{name:"minstep",type:"number",default:0},{name:"nice",type:"boolean",default:!0},{name:"name",type:"string"},{name:"as",type:"string",array:!0,length:2,default:["bin0","bin1"]}]};var a=Object(o.s)(s,r.d);a.transform=function(t,e){var n,r=this._bins(t),i=r.start,s=r.step,a=t.as||["bin0","bin1"],c=a[0],u=a[1];return n=t.modified()?(e=e.reflow(!0)).SOURCE:e.modified(Object(o.e)(t.field))?e.ADD_MOD:e.ADD,e.visit(n,function(t){var e=r(t);t[c]=e,t[u]=null==e?null:i+s*(1+(e-i)/s)}),e.modifies(a)},a._bins=function(t){if(this.value&&!t.modified())return this.value;var e,n,r=t.field,s=Object(i.a)(t),a=s.start,c=s.stop,u=s.step;null!=(e=t.anchor)&&(n=e-(a+u*Math.floor((e-a)/u)),a+=n,c+=n);var l=function(t){var e=r(t);return null==e?null:(e=Math.max(a,Math.min(+e,c-u)),a+u*Math.floor((e-a)/u))};return l.start=a,l.stop=c,l.step=u,this.value=Object(o.d)(l,Object(o.e)(r),t.name||"bin_"+Object(o.f)(r))}},"../../../../vega-transforms/src/Collect.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-transforms/src/util/SortedList.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js");function s(t){i.d.call(this,[],t)}s.Definition={type:"Collect",metadata:{source:!0},params:[{name:"sort",type:"compare"}]},Object(o.s)(s,i.d).transform=function(t,e){var n=e.fork(e.ALL),o=Object(r.a)(i.n,this.value,n.materialize(n.ADD).add),s=t.sort,a=e.changed()||s&&(t.modified("sort")||e.modified(s.fields));return n.visit(n.REM,o.remove),this.modified(a),this.value=n.source=o.data(s,a),e.source&&e.source.root&&(this.value.root=e.source.root),n}},"../../../../vega-transforms/src/Compare.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.c.call(this,null,s,t)}function s(t){return this.value&&!t.modified()?this.value:Object(i.h)(t.fields,t.orders)}Object(i.s)(o,r.c)},"../../../../vega-transforms/src/CountPattern.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}o.Definition={type:"CountPattern",metadata:{generates:!0,changes:!0},params:[{name:"field",type:"field",required:!0},{name:"case",type:"enum",values:["upper","lower","mixed"],default:"mixed"},{name:"pattern",type:"string",default:'[\\w"]+'},{name:"stopwords",type:"string",default:""},{name:"as",type:"string",array:!0,length:2,default:["text","count"]}]};var s=Object(i.s)(o,r.d);s.transform=function(t,e){function n(e){return function(n){for(var r,i=function(t,e,n){switch(e){case"upper":t=t.toUpperCase();break;case"lower":t=t.toLowerCase()}return t.match(n)}(a(n),t.case,o)||[],c=0,u=i.length;c<u;++c)s.test(r=i[c])||e(r)}}var r=this._parameterCheck(t,e),i=this._counts,o=this._match,s=this._stop,a=t.field,c=t.as||["text","count"],u=n(function(t){i[t]=1+(i[t]||0)}),l=n(function(t){i[t]-=1});return r?e.visit(e.SOURCE,u):(e.visit(e.ADD,u),e.visit(e.REM,l)),this._finish(e,c)},s._parameterCheck=function(t,e){var n=!1;return!t.modified("stopwords")&&this._stop||(this._stop=new RegExp("^"+(t.stopwords||"")+"$","i"),n=!0),!t.modified("pattern")&&this._match||(this._match=new RegExp(t.pattern||"[\\w']+","g"),n=!0),(t.modified("field")||e.modified(t.field.fields))&&(n=!0),n&&(this._counts={}),n},s._finish=function(t,e){var n,i,o,s=this._counts,a=this._tuples||(this._tuples={}),c=e[0],u=e[1],l=t.fork(t.NO_SOURCE|t.NO_FIELDS);for(n in s)i=a[n],o=s[n]||0,!i&&o?(a[n]=i=Object(r.h)({}),i[c]=n,i[u]=o,l.add.push(i)):0===o?(i&&l.rem.push(i),s[n]=null,a[n]=null):i[u]!==o&&(i[u]=o,l.mod.push(i));return l.modifies(e)}},"../../../../vega-transforms/src/Cross.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}o.Definition={type:"Cross",metadata:{generates:!0},params:[{name:"filter",type:"expr"},{name:"as",type:"string",array:!0,length:2,default:["a","b"]}]},Object(i.s)(o,r.d).transform=function(t,e){var n=e.fork(e.NO_SOURCE),o=this.value,s=t.as||["a","b"],a=s[0],c=s[1];return!o||e.changed(e.ADD_REM)||t.modified("as")||t.modified("filter")?(o&&(n.rem=o),o=e.materialize(e.SOURCE).source,n.add=this.value=function(t,e,n,i){for(var o,s,a=[],c={},u=t.length,l=0;l<u;++l)for(c[e]=s=t[l],o=0;o<u;++o)c[n]=t[o],i(c)&&(a.push(Object(r.h)(c)),(c={})[e]=s);return a}(o,a,c,t.filter||i.S)):n.mod=o,n.source=this.value,n.modifies(s)}},"../../../../vega-transforms/src/Density.js":function(t,e,n){"use strict";e.a=a;var r=n("../../../../vega-transforms/src/util/Distributions.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js"),s=n("../../../../d3-array/index.js");function a(t){i.d.call(this,null,t)}var c=[{key:{function:"normal"},params:[{name:"mean",type:"number",default:0},{name:"stdev",type:"number",default:1}]},{key:{function:"uniform"},params:[{name:"min",type:"number",default:0},{name:"max",type:"number",default:1}]},{key:{function:"kde"},params:[{name:"field",type:"field",required:!0},{name:"from",type:"data"},{name:"bandwidth",type:"number",default:0}]}],u={key:{function:"mixture"},params:[{name:"distributions",type:"param",array:!0,params:c},{name:"weights",type:"number",array:!0}]};a.Definition={type:"Density",metadata:{generates:!0},params:[{name:"extent",type:"number",array:!0,length:2},{name:"steps",type:"number",default:100},{name:"method",type:"string",default:"pdf",values:["pdf","cdf"]},{name:"distribution",type:"param",params:c.concat(u)},{name:"as",type:"string",array:!0,default:["value","density"]}]},Object(o.s)(a,i.d).transform=function(t,e){var n,a=e.fork(e.NO_SOURCE|e.NO_FIELDS);if(!this.value||e.changed()||t.modified()){var c=Object(r.a)(t.distribution,(n=e,function(){return n.materialize(n.SOURCE).source})),u=t.method||"pdf";"pdf"!==u&&"cdf"!==u&&Object(o.k)("Invalid density method: "+u),t.extent||c.data||Object(o.k)("Missing density extent parameter."),u=c[u];var l=t.as||["value","density"],p=t.extent||Object(s.f)(c.data()),d=(p[1]-p[0])/(t.steps||100),h=Object(s.n)(p[0],p[1]+d/2,d).map(function(t){var e={};return e[l[0]]=t,e[l[1]]=u(t),Object(i.h)(e)});this.value&&(a.rem=this.value),this.value=a.add=a.source=h}return a}},"../../../../vega-transforms/src/Extent.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,[1/0,-1/0],t)}o.Definition={type:"Extent",metadata:{},params:[{name:"field",type:"field",required:!0}]},Object(i.s)(o,r.d).transform=function(t,e){var n=this.value,r=t.field,i=n[0],o=n[1],s=e.ADD;(e.changed()||e.modified(r.fields)||t.modified("field"))&&(s=e.SOURCE,i=1/0,o=-1/0),e.visit(s,function(t){var e=r(t);null!=e&&((e=+e)<i&&(i=e),e>o&&(o=e))}),this.value=[i,o]}},"../../../../vega-transforms/src/Facet.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-transforms/src/Subflow.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js");function s(t){i.d.call(this,{},t),this._keys=Object(o.o)();var e=this._targets=[];e.active=0,e.forEach=function(t){for(var n=0,r=e.active;n<r;++n)t(e[n],n,e)}}var a=Object(o.s)(s,i.d);a.activate=function(t){this._targets[this._targets.active++]=t},a.subflow=function(t,e,n,i){var o,s,a=this.value,c=a.hasOwnProperty(t)&&a[t];return c?c.value.stamp<n.stamp&&(c.init(n),this.activate(c)):(s=i||(s=this._group[t])&&s.tuple,c=(o=n.dataflow).add(new r.a(n.fork(n.NO_SOURCE),this)).connect(e(o,t,s)),a[t]=c,this.activate(c)),c},a.transform=function(t,e){var n=e.dataflow,r=this,o=t.key,s=t.subflow,a=this._keys,c=t.modified("key");function u(t){return r.subflow(t,s,e)}return this._group=t.group||{},this._targets.active=0,e.visit(e.REM,function(t){var e=Object(i.n)(t),n=a.get(e);void 0!==n&&(a.delete(e),u(n).rem(t))}),e.visit(e.ADD,function(t){var e=o(t);a.set(Object(i.n)(t),e),u(e).add(t)}),c||e.modified(o.fields)?e.visit(e.MOD,function(t){var e=Object(i.n)(t),n=a.get(e),r=o(t);n===r?u(r).mod(t):(a.set(e,r),u(n).rem(t),u(r).add(t))}):e.changed(e.MOD)&&e.visit(e.MOD,function(t){u(a.get(Object(i.n)(t))).mod(t)}),c&&e.visit(e.REFLOW,function(t){var e=Object(i.n)(t),n=a.get(e),r=o(t);n!==r&&(a.set(e,r),u(n).rem(t),u(r).add(t))}),a.empty>n.cleanThreshold&&n.runAfter(a.clean),e}},"../../../../vega-transforms/src/Field.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.c.call(this,null,s,t)}function s(t){return this.value&&!t.modified()?this.value:Object(i.t)(t.name)?Object(i.g)(t.name).map(function(t){return Object(i.p)(t)}):Object(i.p)(t.name,t.as)}Object(i.s)(o,r.c)},"../../../../vega-transforms/src/Filter.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,Object(i.o)(),t)}o.Definition={type:"Filter",metadata:{changes:!0},params:[{name:"expr",type:"expr",required:!0}]},Object(i.s)(o,r.d).transform=function(t,e){var n=e.dataflow,i=this.value,o=e.fork(),s=o.add,a=o.rem,c=o.mod,u=t.expr,l=!0;function p(e){var n=Object(r.n)(e),o=u(e,t),p=i.get(n);o&&p?(i.delete(n),s.push(e)):o||p?l&&o&&!p&&c.push(e):(i.set(n,1),a.push(e))}return e.visit(e.REM,function(t){var e=Object(r.n)(t);i.has(e)?i.delete(e):a.push(t)}),e.visit(e.ADD,function(e){u(e,t)?s.push(e):i.set(Object(r.n)(e),1)}),e.visit(e.MOD,p),t.modified()&&(l=!1,e.visit(e.REFLOW,p)),i.empty>n.cleanThreshold&&n.runAfter(i.clean),o}},"../../../../vega-transforms/src/Flatten.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-transforms/src/util/util.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js");function s(t){i.d.call(this,[],t)}s.Definition={type:"Flatten",metadata:{generates:!0,source:!0},params:[{name:"fields",type:"field",array:!0,required:!0},{name:"as",type:"string",array:!0}]},Object(o.s)(s,i.d).transform=function(t,e){var n=e.fork(e.NO_SOURCE),o=t.fields,s=Object(r.a)(o,t.as||[]),a=s.length;return n.rem=this.value,e.visit(e.SOURCE,function(t){for(var e,r,c,u=o.map(function(e){return e(t)}),l=u.reduce(function(t,e){return Math.max(t,e.length)},0),p=0;p<l;++p){for(r=Object(i.g)(t),e=0;e<a;++e)r[s[e]]=null==(c=u[e][p])?null:c;n.add.push(r)}}),this.value=n.source=n.add,n.modifies(s)}},"../../../../vega-transforms/src/Fold.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,[],t)}o.Definition={type:"Fold",metadata:{generates:!0,source:!0},params:[{name:"fields",type:"field",array:!0,required:!0},{name:"as",type:"string",array:!0,length:2,default:["key","value"]}]},Object(i.s)(o,r.d).transform=function(t,e){var n=e.fork(e.NO_SOURCE),o=t.fields,s=o.map(i.f),a=t.as||["key","value"],c=a[0],u=a[1],l=o.length;return n.rem=this.value,e.visit(e.SOURCE,function(t){for(var e,i=0;i<l;++i)(e=Object(r.g)(t))[c]=s[i],e[u]=o[i](t),n.add.push(e)}),this.value=n.source=n.add,n.modifies(a)}},"../../../../vega-transforms/src/Formula.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}o.Definition={type:"Formula",metadata:{modifies:!0},params:[{name:"expr",type:"expr",required:!0},{name:"as",type:"string",required:!0},{name:"initonly",type:"boolean"}]},Object(i.s)(o,r.d).transform=function(t,e){var n=t.expr,r=t.as,i=t.modified(),o=t.initonly?e.ADD:i?e.SOURCE:e.modified(n.fields)?e.ADD_MOD:e.ADD;return i&&(e=e.materialize().reflow(!0)),t.initonly||e.modifies(r),e.visit(o,function(e){e[r]=n(e,t)})}},"../../../../vega-transforms/src/Generate.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,[],t)}Object(i.s)(o,r.d).transform=function(t,e){var n,i,o,s=this.value,a=e.fork(e.ALL),c=t.size-s.length,u=t.generator;if(c>0){for(n=[];--c>=0;)n.push(o=Object(r.h)(u(t))),s.push(o);a.add=a.add.length?a.materialize(a.ADD).add.concat(n):n}else i=s.slice(0,-c),a.rem=a.rem.length?a.materialize(a.REM).rem.concat(i):i,s=s.slice(-c);return a.source=this.value=s,a}},"../../../../vega-transforms/src/Impute.js":function(t,e,n){"use strict";e.a=c;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-array/index.js"),s={value:"value",median:o.i,mean:o.h,min:o.k,max:o.g},a=[];function c(t){r.d.call(this,[],t)}c.Definition={type:"Impute",metadata:{generates:!0,changes:!0},params:[{name:"field",type:"field",required:!0},{name:"key",type:"field",required:!0},{name:"keyvals",array:!0},{name:"groupby",type:"field",array:!0},{name:"method",type:"enum",default:"value",values:["value","mean","median","max","min"]},{name:"value",default:0}]},Object(i.s)(c,r.d).transform=function(t,e){var n,o,c,u,l,p,d,h,f,m,g,v=e.fork(e.ALL),y=function(t){var e,n=t.method||s.value;if(null!=s[n])return n===s.value?(e=void 0!==t.value?t.value:0,function(){return e}):s[n];Object(i.k)("Unrecognized imputation method: "+n)}(t),b=(g=t.field,function(t){return t?g(t):NaN}),_=Object(i.f)(t.field),j=Object(i.f)(t.key),x=(t.groupby||[]).map(i.f),w=function(t,e,n,r){var i,o,s,c,u,l,p,d,h=function(t){return t(d)},f=[],m=r?r.slice():[],g={},v={};for(m.forEach(function(t,e){g[t]=e+1}),c=0,p=t.length;c<p;++c)d=t[c],l=n(d),u=g[l]||(g[l]=m.push(l)),o=(i=e?e.map(h):a)+"",(s=v[o])||(s=v[o]=[],f.push(s),s.values=i),s[u-1]=d;return f.domain=m,f}(e.source,t.groupby,t.key,t.keyvals),O=[],C=this.value,k=w.domain.length;for(l=0,h=w.length;l<h;++l)for(c=(n=w[l]).values,o=NaN,d=0;d<k;++d)if(null==n[d]){for(u=w.domain[d],m={_impute:!0},p=0,f=c.length;p<f;++p)m[x[p]]=c[p];m[j]=u,m[_]=isNaN(o)?o=y(n,b):o,O.push(Object(r.h)(m))}return O.length&&(v.add=v.materialize(v.ADD).add.concat(O)),C.length&&(v.rem=v.materialize(v.REM).rem.concat(C)),this.value=O,v}},"../../../../vega-transforms/src/JoinAggregate.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-transforms/src/Aggregate.js"),i=n("../../../../vega-transforms/src/util/AggregateOps.js"),o=n("../../../../vega-util/index.js");function s(t){r.a.call(this,t)}s.Definition={type:"JoinAggregate",metadata:{modifies:!0},params:[{name:"groupby",type:"field",array:!0},{name:"fields",type:"field",null:!0,array:!0},{name:"ops",type:"enum",array:!0,values:i.a},{name:"as",type:"string",null:!0,array:!0},{name:"key",type:"field"}]};var a=Object(o.s)(s,r.a);a.transform=function(t,e){var n,r=this,i=t.modified();return r.value&&(i||e.modified(r._inputs))?(n=r.value=i?r.init(t):{},e.visit(e.SOURCE,function(t){r.add(t)})):(n=r.value=r.value||this.init(t),e.visit(e.REM,function(t){r.rem(t)}),e.visit(e.ADD,function(t){r.add(t)})),r.changes(),e.visit(e.SOURCE,function(t){Object(o.l)(t,n[r.cellkey(t)].tuple)}),e.reflow(i).modifies(this._outputs)},a.changes=function(){var t,e,n=this._adds,r=this._mods;for(t=0,e=this._alen;t<e;++t)this.celltuple(n[t]),n[t]=null;for(t=0,e=this._mlen;t<e;++t)this.celltuple(r[t]),r[t]=null;this._alen=this._mlen=0}},"../../../../vega-transforms/src/Key.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.c.call(this,null,s,t)}function s(t){return this.value&&!t.modified()?this.value:Object(i.B)(t.fields,t.flat)}Object(i.s)(o,r.c)},"../../../../vega-transforms/src/Lookup.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,{},t)}o.Definition={type:"Lookup",metadata:{modifies:!0},params:[{name:"index",type:"index",params:[{name:"from",type:"data",required:!0},{name:"key",type:"field",required:!0}]},{name:"values",type:"field",array:!0},{name:"fields",type:"field",array:!0,required:!0},{name:"as",type:"string",array:!0},{name:"default",default:null}]},Object(i.s)(o,r.d).transform=function(t,e){var n,r,o=e,s=t.as,a=t.fields,c=t.index,u=t.values,l=null==t.default?null:t.default,p=t.modified(),d=p?e.SOURCE:e.ADD,h=a.length;return u?(r=u.length,h>1&&!s&&Object(i.k)('Multi-field lookup requires explicit "as" parameter.'),s&&s.length!==h*r&&Object(i.k)('The "as" parameter has too few output field names.'),s=s||u.map(i.f),n=function(t){for(var e,n,i=0,o=0;i<h;++i)if(null==(n=c.get(a[i](t))))for(e=0;e<r;++e,++o)t[s[o]]=l;else for(e=0;e<r;++e,++o)t[s[o]]=u[e](n)}):(s||Object(i.k)("Missing output field names."),n=function(t){for(var e,n=0;n<h;++n)e=c.get(a[n](t)),t[s[n]]=null==e?l:e}),p?o=e.reflow(!0):d|=a.some(function(t){return e.modified(t.fields)})?e.MOD:0,e.visit(d,n),o.modifies(s)}},"../../../../vega-transforms/src/MultiExtent.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.c.call(this,null,s,t)}function s(t){if(this.value&&!t.modified())return this.value;var e,n,r,i=1/0,o=-1/0,s=t.extents;for(e=0,n=s.length;e<n;++e)(r=s[e])[0]<i&&(i=r[0]),r[1]>o&&(o=r[1]);return[i,o]}Object(i.s)(o,r.c)},"../../../../vega-transforms/src/MultiValues.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.c.call(this,null,s,t)}function s(t){return this.value&&!t.modified()?this.value:t.values.reduce(function(t,e){return t.concat(e)},[])}Object(i.s)(o,r.c)},"../../../../vega-transforms/src/Params.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}Object(i.s)(o,r.d),o.prototype.transform=function(t,e){return this.modified(t.modified()),this.value=t,e.fork(e.NO_SOURCE|e.NO_FIELDS)}},"../../../../vega-transforms/src/PreFacet.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-transforms/src/Facet.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js");function s(t){r.a.call(this,t)}Object(o.s)(s,r.a).transform=function(t,e){var n=this,r=t.subflow,s=t.field;return(t.modified("field")||s&&e.modified(Object(o.e)(s)))&&Object(o.k)("PreFacet does not support field modification."),this._targets.active=0,e.visit(e.MOD,function(t){var o=n.subflow(Object(i.n)(t),r,e,t);s?s(t).forEach(function(t){o.mod(t)}):o.mod(t)}),e.visit(e.ADD,function(t){var o=n.subflow(Object(i.n)(t),r,e,t);s?s(t).forEach(function(t){o.add(Object(i.h)(t))}):o.add(t)}),e.visit(e.REM,function(t){var o=n.subflow(Object(i.n)(t),r,e,t);s?s(t).forEach(function(t){o.rem(t)}):o.rem(t)}),e}},"../../../../vega-transforms/src/Project.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-transforms/src/util/util.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js");function s(t){i.d.call(this,null,t)}s.Definition={type:"Project",metadata:{generates:!0,changes:!0,modifies:!0},params:[{name:"fields",type:"field",array:!0},{name:"as",type:"string",null:!0,array:!0}]},Object(o.s)(s,i.d).transform=function(t,e){var n,o,s=t.fields,a=Object(r.a)(t.fields,t.as||[]),c=s?function(t,e){return function(t,e,n,r){for(var i=0,o=n.length;i<o;++i)e[r[i]]=n[i](t);return e}(t,e,s,a)}:i.k;return this.value?o=this.value:(e=e.addAll(),o=this.value={}),n=e.fork(e.NO_SOURCE),e.visit(e.REM,function(t){var e=Object(i.n)(t);n.rem.push(o[e]),o[e]=null}),e.visit(e.ADD,function(t){var e=c(t,Object(i.h)({}));o[Object(i.n)(t)]=e,n.add.push(e)}),e.visit(e.MOD,function(t){n.mod.push(c(t,o[Object(i.n)(t)]))}),n}},"../../../../vega-transforms/src/Proxy.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}Object(i.s)(o,r.d).transform=function(t,e){return this.value=t.value,t.modified("value")?e.fork(e.NO_SOURCE|e.NO_FIELDS):e.StopPropagation}},"../../../../vega-transforms/src/Relay.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}Object(i.s)(o,r.d).transform=function(t,e){var n,i;return this.value?i=this.value:(n=e=e.addAll(),i=this.value={}),t.derive&&(n=e.fork(e.NO_SOURCE),e.visit(e.REM,function(t){var e=Object(r.n)(t);n.rem.push(i[e]),i[e]=null}),e.visit(e.ADD,function(t){var e=Object(r.g)(t);i[Object(r.n)(t)]=e,n.add.push(e)}),e.visit(e.MOD,function(t){n.mod.push(Object(r.k)(t,i[Object(r.n)(t)]))})),n}},"../../../../vega-transforms/src/Sample.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-statistics/index.js"),o=n("../../../../vega-util/index.js");function s(t){r.d.call(this,[],t),this.count=0}s.Definition={type:"Sample",metadata:{},params:[{name:"size",type:"number",default:1e3}]},Object(o.s)(s,r.d).transform=function(t,e){var n=e.fork(e.NO_SOURCE),o=t.modified("size"),s=t.size,a=this.value,c=this.count,u=0,l=a.reduce(function(t,e){return t[Object(r.n)(e)]=1,t},{});function p(t){var e,o;a.length<s?a.push(t):(o=~~((c+1)*Object(i.d)()))<a.length&&o>=u&&(e=a[o],l[Object(r.n)(e)]&&n.rem.push(e),a[o]=t),++c}if(e.rem.length&&(e.visit(e.REM,function(t){var e=Object(r.n)(t);l[e]&&(l[e]=-1,n.rem.push(t)),--c}),a=a.filter(function(t){return-1!==l[Object(r.n)(t)]})),(e.rem.length||o)&&a.length<s&&e.source&&(u=c=a.length,e.visit(e.SOURCE,function(t){l[Object(r.n)(t)]||p(t)}),u=-1),o&&a.length>s){for(var d=0,h=a.length-s;d<h;++d)l[Object(r.n)(a[d])]=-1,n.rem.push(a[d]);a=a.slice(h)}return e.mod.length&&e.visit(e.MOD,function(t){l[Object(r.n)(t)]&&n.mod.push(t)}),e.add.length&&e.visit(e.ADD,p),(e.add.length||u<0)&&(n.add=a.filter(function(t){return!l[Object(r.n)(t)]})),this.count=c,this.value=n.source=a,n}},"../../../../vega-transforms/src/Sequence.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-array/index.js");function s(t){r.d.call(this,null,t)}s.Definition={type:"Sequence",metadata:{generates:!0,changes:!0},params:[{name:"start",type:"number",required:!0},{name:"stop",type:"number",required:!0},{name:"step",type:"number",default:1}],output:["value"]},Object(i.s)(s,r.d).transform=function(t,e){if(!this.value||t.modified()){var n=e.materialize().fork(e.MOD);return n.rem=this.value?e.rem.concat(this.value):e.rem,this.value=Object(o.n)(t.start,t.stop,t.step||1).map(r.h),n.add=e.add.concat(this.value),n}}},"../../../../vega-transforms/src/Sieve.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t),this.modified(!0)}Object(i.s)(o,r.d).transform=function(t,e){return this.value=e.source,e.changed()?e.fork(e.NO_SOURCE|e.NO_FIELDS):e.StopPropagation}},"../../../../vega-transforms/src/Subflow.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t,e){r.c.call(this,t),this.parent=e}var s=Object(i.s)(o,r.c);s.connect=function(t){return this.targets().add(t),t.source=this},s.add=function(t){this.value.add.push(t)},s.rem=function(t){this.value.rem.push(t)},s.mod=function(t){this.value.mod.push(t)},s.init=function(t){this.value.init(t,t.NO_SOURCE)},s.evaluate=function(){return this.value}},"../../../../vega-transforms/src/TupleIndex.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,Object(i.o)(),t)}Object(i.s)(o,r.d).transform=function(t,e){var n=e.dataflow,r=t.field,i=this.value,o=!0;function s(t){i.set(r(t),t)}return t.modified("field")||e.modified(r.fields)?(i.clear(),e.visit(e.SOURCE,s)):e.changed()?(e.visit(e.REM,function(t){i.delete(r(t))}),e.visit(e.ADD,s)):o=!1,this.modified(o),i.empty>n.cleanThreshold&&n.runAfter(i.clean),e.fork()}},"../../../../vega-transforms/src/Values.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}Object(i.s)(o,r.d).transform=function(t,e){(!this.value||t.modified("field")||t.modified("sort")||e.changed()||t.sort&&e.modified(t.sort.fields))&&(this.value=(t.sort?e.source.slice().sort(t.sort):e.source).map(t.field))}},"../../../../vega-transforms/src/Window.js":function(t,e,n){"use strict";e.a=p;var r=n("../../../../vega-transforms/src/util/AggregateKeys.js"),i=n("../../../../vega-transforms/src/util/AggregateOps.js"),o=n("../../../../vega-transforms/src/util/SortedList.js"),s=n("../../../../vega-transforms/src/util/WindowOps.js"),a=n("../../../../vega-transforms/src/util/WindowState.js"),c=n("../../../../vega-dataflow/index.js"),u=n("../../../../vega-util/index.js"),l=n("../../../../d3-array/index.js");function p(t){c.d.call(this,{},t),this._mlen=0,this._mods=[]}p.Definition={type:"Window",metadata:{modifies:!0},params:[{name:"sort",type:"compare"},{name:"groupby",type:"field",array:!0},{name:"ops",type:"enum",array:!0,values:s.a.concat(i.a)},{name:"params",type:"number",null:!0,array:!0},{name:"fields",type:"field",null:!0,array:!0},{name:"as",type:"string",null:!0,array:!0},{name:"frame",type:"number",null:!0,array:!0,length:2,default:[null,0]},{name:"ignorePeers",type:"boolean",default:!1}]};var d=Object(u.s)(p,c.d);function h(t,e,n){var r=n.sort,i=r&&!n.ignorePeers,o=n.frame||[null,0],s=t.data(r),a=s.length,c=0,p=i?Object(l.e)(r):null,d={i0:0,i1:0,p0:0,p1:0,index:0,data:s,compare:r||Object(u.i)(-1)};for(e.init();c<a;++c)f(d,o,c,a),i&&m(d,p),e.update(d,s[c])}function f(t,e,n,r){t.p0=t.i0,t.p1=t.i1,t.i0=null==e[0]?0:Math.max(0,n-Math.abs(e[0])),t.i1=null==e[1]?r:Math.min(r,n+Math.abs(e[1])+1),t.index=n}function m(t,e){var n=t.i0,r=t.i1-1,i=t.compare,o=t.data,s=o.length-1;n>0&&!i(o[n],o[n-1])&&(t.i0=e.left(o,o[n])),r<s&&!i(o[r],o[r+1])&&(t.i1=e.right(o,o[r]))}d.transform=function(t,e){var n,i,o=this,s=o.state,c=t.modified();this.stamp=e.stamp,s&&!c||(s=o.state=new a.a(t));var u=Object(r.a)(t.groupby);function l(t){return o.group(u(t))}for(c||e.modified(s.inputs)?(o.value={},e.visit(e.SOURCE,function(t){l(t).add(t)})):(e.visit(e.REM,function(t){l(t).remove(t)}),e.visit(e.ADD,function(t){l(t).add(t)})),n=0,i=o._mlen;n<i;++n)h(o._mods[n],s,t);return o._mlen=0,o._mods=[],e.reflow(c).modifies(s.outputs)},d.group=function(t){var e=this.value[t];return e||((e=this.value[t]=Object(o.a)(c.n)).stamp=-1),e.stamp<this.stamp&&(e.stamp=this.stamp,this._mods[this._mlen++]=e),e}},"../../../../vega-transforms/src/util/AggregateKeys.js":function(t,e,n){"use strict";e.a=function(t){return t&&t.length?1===t.length?t[0]:(e=t,function(t){for(var n=e.length,r=1,i=String(e[0](t));r<n;++r)i+="|"+e[r](t);return i}):function(){return""};var e}},"../../../../vega-transforms/src/util/AggregateOps.js":function(t,e,n){"use strict";e.d=function(t,e,n){return n||t+(e?"_"+e:"")},n.d(e,"a",function(){return o}),e.c=function(t,e){return i[t](e)},e.b=function(t,e){var n=e||r.r,o="var cell = this.cell; this.valid = 0; this.missing = 0;",s="this.cell = cell; this.init();",c="if(v==null){++this.missing; return;} if(v!==v) return; ++this.valid;",u="if(v==null){--this.missing; return;} if(v!==v) return; --this.valid;",l="var cell = this.cell;";return function(t,e){var n,r=t.reduce(function t(n,r){function o(e){n[e]||t(n,n[e]=i[e]())}r.req&&r.req.forEach(o);e&&r.str&&r.str.forEach(o);return n},t.reduce(function(t,e){return t[e.name]=e,t},{})),o=[];for(n in r)o.push(r[n]);return o.sort(a)}(t,!0).forEach(function(t){o+=t.init,c+=t.add,u+=t.rem}),t.slice().sort(a).forEach(function(t){l+="t['"+t.out+"']="+t.set+";"}),l+="return t;",(s=Function("cell",s)).prototype.init=Function(o),s.prototype.add=Function("v","t",c),s.prototype.rem=Function("v","t",u),s.prototype.set=Function("t",l),s.prototype.get=n,s.fields=t.map(function(t){return t.out}),s};var r=n("../../../../vega-util/index.js");var i={values:s({name:"values",init:"cell.store = true;",set:"cell.data.values()",idx:-1}),count:s({name:"count",set:"cell.num"}),missing:s({name:"missing",set:"this.missing"}),valid:s({name:"valid",set:"this.valid"}),sum:s({name:"sum",init:"this.sum = 0;",add:"this.sum += v;",rem:"this.sum -= v;",set:"this.sum"}),mean:s({name:"mean",init:"this.mean = 0;",add:"var d = v - this.mean; this.mean += d / this.valid;",rem:"var d = v - this.mean; this.mean -= this.valid ? d / this.valid : this.mean;",set:"this.mean"}),average:s({name:"average",set:"this.mean",req:["mean"],idx:1}),variance:s({name:"variance",init:"this.dev = 0;",add:"this.dev += d * (v - this.mean);",rem:"this.dev -= d * (v - this.mean);",set:"this.valid > 1 ? this.dev / (this.valid-1) : 0",req:["mean"],idx:1}),variancep:s({name:"variancep",set:"this.valid > 1 ? this.dev / this.valid : 0",req:["variance"],idx:2}),stdev:s({name:"stdev",set:"this.valid > 1 ? Math.sqrt(this.dev / (this.valid-1)) : 0",req:["variance"],idx:2}),stdevp:s({name:"stdevp",set:"this.valid > 1 ? Math.sqrt(this.dev / this.valid) : 0",req:["variance"],idx:2}),stderr:s({name:"stderr",set:"this.valid > 1 ? Math.sqrt(this.dev / (this.valid * (this.valid-1))) : 0",req:["variance"],idx:2}),distinct:s({name:"distinct",set:"cell.data.distinct(this.get)",req:["values"],idx:3}),ci0:s({name:"ci0",set:"cell.data.ci0(this.get)",req:["values"],idx:3}),ci1:s({name:"ci1",set:"cell.data.ci1(this.get)",req:["values"],idx:3}),median:s({name:"median",set:"cell.data.q2(this.get)",req:["values"],idx:3}),q1:s({name:"q1",set:"cell.data.q1(this.get)",req:["values"],idx:3}),q3:s({name:"q3",set:"cell.data.q3(this.get)",req:["values"],idx:3}),argmin:s({name:"argmin",init:"this.argmin = null;",add:"if (v < this.min) this.argmin = t;",rem:"if (v <= this.min) this.argmin = null;",set:"this.argmin || cell.data.argmin(this.get)",req:["min"],str:["values"],idx:3}),argmax:s({name:"argmax",init:"this.argmax = null;",add:"if (v > this.max) this.argmax = t;",rem:"if (v >= this.max) this.argmax = null;",set:"this.argmax || cell.data.argmax(this.get)",req:["max"],str:["values"],idx:3}),min:s({name:"min",init:"this.min = null;",add:"if (v < this.min || this.min === null) this.min = v;",rem:"if (v <= this.min) this.min = NaN;",set:"this.min = (isNaN(this.min) ? cell.data.min(this.get) : this.min)",str:["values"],idx:4}),max:s({name:"max",init:"this.max = null;",add:"if (v > this.max || this.max === null) this.max = v;",rem:"if (v >= this.max) this.max = NaN;",set:"this.max = (isNaN(this.max) ? cell.data.max(this.get) : this.max)",str:["values"],idx:4})},o=Object.keys(i);function s(t){return function(e){var n=Object(r.l)({init:"",add:"",rem:"",idx:0},t);return n.out=e||t.name,n}}function a(t,e){return t.idx-e.idx}},"../../../../vega-transforms/src/util/Distributions.js":function(t,e,n){"use strict";e.a=function t(e,n){var r=e[a];o.hasOwnProperty(r)||Object(i.k)("Unknown distribution function: "+r);var u=o[r]();for(var l in e)l===c?u.data((e.from||n()).map(e[l])):l===s?u[l](e[l].map(function(e){return t(e,n)})):typeof u[l]===a&&u[l](e[l]);return u};var r=n("../../../../vega-statistics/index.js"),i=n("../../../../vega-util/index.js"),o={kde:r.e,mixture:r.f,normal:r.g,uniform:r.h},s="distributions",a="function",c="field"},"../../../../vega-transforms/src/util/SortedList.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");e.a=function(t,e,n){var i=t,o=e||[],s=n||[],a={},c=0;return{add:function(t){s.push(t)},remove:function(t){a[i(t)]=++c},size:function(){return o.length},data:function(t,e){return c&&(o=o.filter(function(t){return!a[i(t)]}),a={},c=0),e&&t&&o.sort(t),s.length&&(o=t?Object(r.D)(t,o,s.sort(t)):o.concat(s),s=[]),o}}}},"../../../../vega-transforms/src/util/TupleStore.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-statistics/index.js"),o=n("../../../../vega-util/index.js");function s(t){this._key=t?Object(o.p)(t):r.n,this.reset()}var a=s.prototype;a.reset=function(){this._add=[],this._rem=[],this._ext=null,this._get=null,this._q=null},a.add=function(t){this._add.push(t)},a.rem=function(t){this._rem.push(t)},a.values=function(){if(this._get=null,0===this._rem.length)return this._add;var t,e,n,r=this._add,i=this._rem,o=this._key,s=r.length,a=i.length,c=Array(s-a),u={};for(t=0;t<a;++t)u[o(i[t])]=1;for(t=0,e=0;t<s;++t)u[o(n=r[t])]?u[o(n)]=0:c[e++]=n;return this._rem=[],this._add=c},a.distinct=function(t){for(var e,n=this.values(),r=n.length,i={},o=0;--r>=0;)e=t(n[r])+"",i.hasOwnProperty(e)||(i[e]=1,++o);return o},a.extent=function(t){if(this._get!==t||!this._ext){var e=this.values(),n=Object(o.m)(e,t);this._ext=[e[n[0]],e[n[1]]],this._get=t}return this._ext},a.argmin=function(t){return this.extent(t)[0]||{}},a.argmax=function(t){return this.extent(t)[1]||{}},a.min=function(t){var e=this.extent(t)[0];return null!=e?t(e):1/0},a.max=function(t){var e=this.extent(t)[1];return null!=e?t(e):-1/0},a.quartile=function(t){return this._get===t&&this._q||(this._q=Object(i.c)(this.values(),t),this._get=t),this._q},a.q1=function(t){return this.quartile(t)[0]},a.q2=function(t){return this.quartile(t)[1]},a.q3=function(t){return this.quartile(t)[2]},a.ci=function(t){return this._get===t&&this._ci||(this._ci=Object(i.b)(this.values(),1e3,.05,t),this._get=t),this._ci},a.ci0=function(t){return this.ci(t)[0]},a.ci1=function(t){return this.ci(t)[1]}},"../../../../vega-transforms/src/util/WindowOps.js":function(t,e,n){"use strict";e.b=function(t,e,n,o){var s=i[t](e,n);return{init:s.init||r.U,update:function(t,e){e[o]=s.next(t)}}},n.d(e,"c",function(){return i}),n.d(e,"a",function(){return o});var r=n("../../../../vega-util/index.js");var i={row_number:function(){return{next:function(t){return t.index+1}}},rank:function(){var t;return{init:function(){t=1},next:function(e){var n=e.index,r=e.data;return n&&e.compare(r[n-1],r[n])?t=n+1:t}}},dense_rank:function(){var t;return{init:function(){t=1},next:function(e){var n=e.index,r=e.data;return n&&e.compare(r[n-1],r[n])?++t:t}}},percent_rank:function(){var t=i.rank(),e=t.next;return{init:t.init,next:function(t){return(e(t)-1)/(t.data.length-1)}}},cume_dist:function(){var t;return{init:function(){t=0},next:function(e){var n=e.index,r=e.data,i=e.compare;if(t<n){for(;n+1<r.length&&!i(r[n],r[n+1]);)++n;t=n}return(1+t)/r.length}}},ntile:function(t,e){(e=+e)>0||Object(r.k)("ntile num must be greater than zero.");var n=i.cume_dist(),o=n.next;return{init:n.init,next:function(t){return Math.ceil(e*o(t))}}},lag:function(t,e){return e=+e||1,{next:function(n){var r=n.index-e;return r>=0?t(n.data[r]):null}}},lead:function(t,e){return e=+e||1,{next:function(n){var r=n.index+e,i=n.data;return r<i.length?t(i[r]):null}}},first_value:function(t){return{next:function(e){return t(e.data[e.i0])}}},last_value:function(t){return{next:function(e){return t(e.data[e.i1-1])}}},nth_value:function(t,e){return(e=+e)>0||Object(r.k)("nth_value nth must be greater than zero."),{next:function(n){var r=n.i0+(e-1);return r<n.i1?t(n.data[r]):null}}}},o=Object.keys(i)},"../../../../vega-transforms/src/util/WindowState.js":function(t,e,n){"use strict";e.a=a;var r=n("../../../../vega-transforms/src/util/AggregateOps.js"),i=n("../../../../vega-transforms/src/util/TupleStore.js"),o=n("../../../../vega-transforms/src/util/WindowOps.js"),s=n("../../../../vega-util/index.js");function a(t){var e=Object(s.g)(t.ops),n=Object(s.g)(t.fields),a=Object(s.g)(t.params),c=Object(s.g)(t.as),u=this.outputs=[],l=this.windows=[],p={},d={},h=!0,f=[],m=[];function g(t){Object(s.g)(Object(s.e)(t)).forEach(function(t){p[t]=1})}g(t.sort),e.forEach(function(t,e){var i=n[e],p=Object(s.f)(i),v=Object(r.d)(t,p,c[e]);if(g(i),u.push(v),o.c.hasOwnProperty(t))l.push(Object(o.b)(t,n[e],a[e],v));else{if(null==i&&"count"!==t&&Object(s.k)("Null aggregate field specified."),"count"===t)return void f.push(v);h=!1;var y=d[p];y||((y=d[p]=[]).field=i,m.push(y)),y.push(Object(r.c)(t,v))}}),(f.length||m.length)&&(this.cell=function(t,e,n){t=t.map(function(t){return Object(r.b)(t,t.field)});var o={num:0,agg:null,store:!1,count:e};if(!n)for(var s=t.length,a=o.agg=Array(s),c=0;c<s;++c)a[c]=new t[c](o);if(o.store)var u=o.data=new i.a;return o.add=function(t){if(o.num+=1,!n){u&&u.add(t);for(var e=0;e<s;++e)a[e].add(a[e].get(t),t)}},o.rem=function(t){if(o.num-=1,!n){u&&u.rem(t);for(var e=0;e<s;++e)a[e].rem(a[e].get(t),t)}},o.set=function(t){var r,i;for(u&&u.values(),r=0,i=e.length;r<i;++r)t[e[r]]=o.num;if(!n)for(r=0,i=a.length;r<i;++r)a[r].set(t)},o.init=function(){o.num=0,u&&u.reset();for(var t=0;t<s;++t)a[t].init()},o}(m,f,h)),this.inputs=Object.keys(p)}var c=a.prototype;c.init=function(){this.windows.forEach(function(t){t.init()}),this.cell&&this.cell.init()},c.update=function(t,e){var n,r=this.cell,i=this.windows,o=t.data,s=i&&i.length;if(r){for(n=t.p0;n<t.i0;++n)r.rem(o[n]);for(n=t.p1;n<t.i1;++n)r.add(o[n]);r.set(e)}for(n=0;n<s;++n)i[n].update(t,e)}},"../../../../vega-transforms/src/util/util.js":function(t,e,n){"use strict";e.a=function(t,e){return t?t.map(function(t,n){return e[n]||Object(r.f)(t)}):null};var r=n("../../../../vega-util/index.js")},"../../../../vega-util/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/src/accessor.js");n.d(e,"d",function(){return r.c}),n.d(e,"f",function(){return r.b}),n.d(e,"e",function(){return r.a});var i=n("../../../../vega-util/src/accessors.js");n.d(e,"q",function(){return i.b}),n.d(e,"r",function(){return i.c}),n.d(e,"U",function(){return i.f}),n.d(e,"E",function(){return i.d}),n.d(e,"S",function(){return i.e}),n.d(e,"n",function(){return i.a});var o=n("../../../../vega-util/src/logger.js");n.d(e,"C",function(){return o.d}),n.d(e,"b",function(){return o.b}),n.d(e,"c",function(){return o.c}),n.d(e,"a",function(){return o.a});var s=n("../../../../vega-util/src/transform.js");n.d(e,"G",function(){return s.a}),n.d(e,"H",function(){return s.b}),n.d(e,"I",function(){return s.c}),n.d(e,"V",function(){return s.d}),n.d(e,"W",function(){return s.e}),n.d(e,"X",function(){return s.f});var a=n("../../../../vega-util/src/array.js");n.d(e,"g",function(){return a.a});var c=n("../../../../vega-util/src/compare.js");n.d(e,"h",function(){return c.a});var u=n("../../../../vega-util/src/constant.js");n.d(e,"i",function(){return u.a});var l=n("../../../../vega-util/src/debounce.js");n.d(e,"j",function(){return l.a});var p=n("../../../../vega-util/src/error.js");n.d(e,"k",function(){return p.a});var d=n("../../../../vega-util/src/extend.js");n.d(e,"l",function(){return d.a});var h=n("../../../../vega-util/src/extentIndex.js");n.d(e,"m",function(){return h.a});var f=n("../../../../vega-util/src/fastmap.js");n.d(e,"o",function(){return f.a});var m=n("../../../../vega-util/src/field.js");n.d(e,"p",function(){return m.a});var g=n("../../../../vega-util/src/inherits.js");n.d(e,"s",function(){return g.a});var v=n("../../../../vega-util/src/isArray.js");n.d(e,"t",function(){return v.a});var y=n("../../../../vega-util/src/isBoolean.js");n.d(e,"u",function(){return y.a});var b=n("../../../../vega-util/src/isDate.js");n.d(e,"v",function(){return b.a});var _=n("../../../../vega-util/src/isFunction.js");n.d(e,"w",function(){return _.a});var j=n("../../../../vega-util/src/isNumber.js");n.d(e,"x",function(){return j.a});var x=n("../../../../vega-util/src/isObject.js");n.d(e,"y",function(){return x.a});var w=n("../../../../vega-util/src/isRegExp.js");n.d(e,"z",function(){return w.a});var O=n("../../../../vega-util/src/isString.js");n.d(e,"A",function(){return O.a});var C=n("../../../../vega-util/src/key.js");n.d(e,"B",function(){return C.a});var k=n("../../../../vega-util/src/merge.js");n.d(e,"D",function(){return k.a});var E=n("../../../../vega-util/src/pad.js");n.d(e,"F",function(){return E.a});var S=n("../../../../vega-util/src/peek.js");n.d(e,"J",function(){return S.a});n("../../../../vega-util/src/repeat.js");var P=n("../../../../vega-util/src/splitAccessPath.js");n.d(e,"K",function(){return P.a});var M=n("../../../../vega-util/src/stringValue.js");n.d(e,"L",function(){return M.a});var A=n("../../../../vega-util/src/toBoolean.js");n.d(e,"M",function(){return A.a});var T=n("../../../../vega-util/src/toDate.js");n.d(e,"N",function(){return T.a});var D=n("../../../../vega-util/src/toNumber.js");n.d(e,"O",function(){return D.a});var I=n("../../../../vega-util/src/toString.js");n.d(e,"Q",function(){return I.a});var R=n("../../../../vega-util/src/toSet.js");n.d(e,"P",function(){return R.a});var N=n("../../../../vega-util/src/truncate.js");n.d(e,"R",function(){return N.a});var L=n("../../../../vega-util/src/visitArray.js");n.d(e,"T",function(){return L.a})},"../../../../vega-util/src/accessor.js":function(t,e,n){"use strict";e.b=function(t){return null==t?null:t.fname},e.a=function(t){return null==t?null:t.fields},e.c=function(t,e,n){return t.fields=e||[],t.fname=n,t}},"../../../../vega-util/src/accessors.js":function(t,e,n){"use strict";n.d(e,"b",function(){return s}),n.d(e,"c",function(){return a}),n.d(e,"f",function(){return c}),n.d(e,"d",function(){return u}),n.d(e,"e",function(){return l}),n.d(e,"a",function(){return p});var r=n("../../../../vega-util/src/accessor.js"),i=n("../../../../vega-util/src/field.js"),o=[],s=Object(i.a)("id"),a=Object(r.c)(function(t){return t},o,"identity"),c=Object(r.c)(function(){return 0},o,"zero"),u=Object(r.c)(function(){return 1},o,"one"),l=Object(r.c)(function(){return!0},o,"true"),p=Object(r.c)(function(){return!1},o,"false")},"../../../../vega-util/src/array.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/src/isArray.js");e.a=function(t){return null!=t?Object(r.a)(t)?t:[t]:[]}},"../../../../vega-util/src/compare.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/src/accessor.js"),i=n("../../../../vega-util/src/array.js"),o=n("../../../../vega-util/src/isFunction.js"),s=n("../../../../vega-util/src/splitAccessPath.js"),a=n("../../../../vega-util/src/stringValue.js");e.a=function(t,e){var n,c,u,l,p,d,h,f,m,g=[],v=(t=Object(i.a)(t)).map(function(t,e){return null==t?null:(g.push(e),Object(o.a)(t)?t:Object(s.a)(t).map(a.a).join("]["))}),y=g.length-1,b=Object(i.a)(e),_="var u,v;return ";if(y<0)return null;for(c=0;c<=y;++c)u=v[n=g[c]],Object(o.a)(u)?(l="(u=this."+(d="f"+n)+"(a))",p="(v=this."+d+"(b))",(h=h||{})[d]=u):(l="(u=a["+u+"])",p="(v=b["+u+"])"),d="((v=v instanceof Date?+v:v),(u=u instanceof Date?+u:u))","descending"!==b[n]?(m=1,f=-1):(m=-1,f=1),_+="("+l+"<"+p+"||u==null)&&v!=null?"+f+":(u>v||v==null)&&u!=null?"+m+":"+d+"!==u&&v===v?"+f+":v!==v&&u===u?"+m+(n<y?":":":0");return u=Function("a","b",_+";"),h&&(u=u.bind(h)),t=t.reduce(function(t,e){return Object(o.a)(e)?(Object(r.a)(e)||[]).forEach(function(e){t[e]=1}):null!=e&&(t[e+""]=1),t},{}),Object(r.c)(u,Object.keys(t))}},"../../../../vega-util/src/constant.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/src/isFunction.js");e.a=function(t){return Object(r.a)(t)?t:function(){return t}}},"../../../../vega-util/src/debounce.js":function(t,e,n){"use strict";e.a=function(t,e){var n,r;function i(){e(r),n=r=null}return function(e){r=e,n&&clearTimeout(n),n=setTimeout(i,t)}}},"../../../../vega-util/src/error.js":function(t,e,n){"use strict";e.a=function(t){throw Error(t)}},"../../../../vega-util/src/extend.js":function(t,e,n){"use strict";e.a=function(t){for(var e,n,r=1,i=arguments.length;r<i;++r){e=arguments[r];for(n in e)t[n]=e[n]}return t}},"../../../../vega-util/src/extentIndex.js":function(t,e,n){"use strict";e.a=function(t,e){var n,r,i,o,s,a=-1,c=t.length;if(null==e){for(;++a<c;)if(null!=(r=t[a])&&r>=r){n=i=r;break}for(o=s=a;++a<c;)null!=(r=t[a])&&(n>r&&(n=r,o=a),i<r&&(i=r,s=a))}else{for(;++a<c;)if(null!=(r=e(t[a],a,t))&&r>=r){n=i=r;break}for(o=s=a;++a<c;)null!=(r=e(t[a],a,t))&&(n>r&&(n=r,o=a),i<r&&(i=r,s=a))}return[o,s]}},"../../../../vega-util/src/fastmap.js":function(t,e,n){"use strict";var r={};e.a=function(t){var e,n,i={};function o(t){return i.hasOwnProperty(t)&&i[t]!==r}return e={size:0,empty:0,object:i,has:o,get:function(t){return o(t)?i[t]:void 0},set:function(t,n){return o(t)||(++e.size,i[t]===r&&--e.empty),i[t]=n,this},delete:function(t){return o(t)&&(--e.size,++e.empty,i[t]=r),this},clear:function(){e.size=e.empty=0,e.object=i={}},test:function(t){return arguments.length?(n=t,e):n},clean:function(){var t,o,s={},a=0;for(t in i)(o=i[t])===r||n&&n(o)||(s[t]=o,++a);e.size=a,e.empty=0,e.object=i=s}},t&&Object.keys(t).forEach(function(n){e.set(n,t[n])}),e}},"../../../../vega-util/src/field.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/src/accessor.js"),i=n("../../../../vega-util/src/splitAccessPath.js"),o=n("../../../../vega-util/src/stringValue.js");e.a=function(t,e){var n=Object(i.a)(t),s="return _["+n.map(o.a).join("][")+"];";return Object(r.c)(Function("_",s),[t=1===n.length?n[0]:t],e||t)}},"../../../../vega-util/src/inherits.js":function(t,e,n){"use strict";e.a=function(t,e){var n=t.prototype=Object.create(e.prototype);return n.constructor=t,n}},"../../../../vega-util/src/isArray.js":function(t,e,n){"use strict";e.a=Array.isArray},"../../../../vega-util/src/isBoolean.js":function(t,e,n){"use strict";e.a=function(t){return"boolean"==typeof t}},"../../../../vega-util/src/isDate.js":function(t,e,n){"use strict";e.a=function(t){return"[object Date]"===Object.prototype.toString.call(t)}},"../../../../vega-util/src/isFunction.js":function(t,e,n){"use strict";e.a=function(t){return"function"==typeof t}},"../../../../vega-util/src/isNumber.js":function(t,e,n){"use strict";e.a=function(t){return"number"==typeof t}},"../../../../vega-util/src/isObject.js":function(t,e,n){"use strict";e.a=function(t){return t===Object(t)}},"../../../../vega-util/src/isRegExp.js":function(t,e,n){"use strict";e.a=function(t){return"[object RegExp]"===Object.prototype.toString.call(t)}},"../../../../vega-util/src/isString.js":function(t,e,n){"use strict";e.a=function(t){return"string"==typeof t}},"../../../../vega-util/src/key.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/src/accessor.js"),i=n("../../../../vega-util/src/array.js"),o=n("../../../../vega-util/src/splitAccessPath.js"),s=n("../../../../vega-util/src/stringValue.js");e.a=function(t,e){t&&(t=e?Object(i.a)(t).map(function(t){return t.replace(/\\(.)/g,"$1")}):Object(i.a)(t));var n=t&&t.length?Function("_","return ''+"+t.map(function(t){return"_["+(e?Object(s.a)(t):Object(o.a)(t).map(s.a).join("]["))+"]"}).join("+'|'+")+";"):function(){return""};return Object(r.c)(n,t,"key")}},"../../../../vega-util/src/logger.js":function(t,e,n){"use strict";function r(t,e,n){var r=[e].concat([].slice.call(n));console[t].apply(console,r)}n.d(e,"b",function(){return i}),n.d(e,"c",function(){return o}),n.d(e,"a",function(){return s});var i=1,o=3,s=4;e.d=function(t){var e=t||0;return{level:function(t){return arguments.length?(e=+t,this):e},error:function(){return e>=i&&r("error","ERROR",arguments),this},warn:function(){return e>=2&&r("warn","WARN",arguments),this},info:function(){return e>=o&&r("log","INFO",arguments),this},debug:function(){return e>=s&&r("log","DEBUG",arguments),this}}}},"../../../../vega-util/src/merge.js":function(t,e,n){"use strict";e.a=function(t,e,n,r){var i=e.length,o=n.length;if(!o)return e;if(!i)return n;for(var s=r||new e.constructor(i+o),a=0,c=0,u=0;a<i&&c<o;++u)s[u]=t(e[a],n[c])>0?n[c++]:e[a++];for(;a<i;++a,++u)s[u]=e[a];for(;c<o;++c,++u)s[u]=n[c];return s}},"../../../../vega-util/src/pad.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/src/repeat.js");e.a=function(t,e,n,i){var o=n||" ",s=t+"",a=e-s.length;return a<=0?s:"left"===i?Object(r.a)(o,a)+s:"center"===i?Object(r.a)(o,~~(a/2))+s+Object(r.a)(o,Math.ceil(a/2)):s+Object(r.a)(o,a)}},"../../../../vega-util/src/peek.js":function(t,e,n){"use strict";e.a=function(t){return t[t.length-1]}},"../../../../vega-util/src/repeat.js":function(t,e,n){"use strict";e.a=function(t,e){for(var n="";--e>=0;)n+=t;return n}},"../../../../vega-util/src/splitAccessPath.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/src/error.js");e.a=function(t){var e,n,i,o=[],s=null,a=0,c=t.length,u="";function l(){o.push(u+t.substring(e,n)),u="",e=n+1}for(t+="",e=n=0;n<c;++n)if("\\"===(i=t[n]))u+=t.substring(e,n),e=++n;else if(i===s)l(),s=null,a=-1;else{if(s)continue;e===a&&'"'===i?(e=n+1,s=i):e===a&&"'"===i?(e=n+1,s=i):"."!==i||a?"["===i?(n>e&&l(),a=e=n+1):"]"===i&&(a||Object(r.a)("Access path missing open bracket: "+t),a>0&&l(),a=0,e=n+1):n>e?l():e=n+1}return a&&Object(r.a)("Access path missing closing bracket: "+t),s&&Object(r.a)("Access path missing closing quote: "+t),n>e&&(n++,l()),o}},"../../../../vega-util/src/stringValue.js":function(t,e,n){"use strict";e.a=function t(e){return Object(r.a)(e)?"["+e.map(t)+"]":Object(i.a)(e)||Object(o.a)(e)?JSON.stringify(e).replace("\u2028","\\u2028").replace("\u2029","\\u2029"):e};var r=n("../../../../vega-util/src/isArray.js"),i=n("../../../../vega-util/src/isObject.js"),o=n("../../../../vega-util/src/isString.js")},"../../../../vega-util/src/toBoolean.js":function(t,e,n){"use strict";e.a=function(t){return null==t||""===t?null:!(!t||"false"===t||"0"===t)&&!!t}},"../../../../vega-util/src/toDate.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/src/isDate.js"),i=n("../../../../vega-util/src/isNumber.js");function o(t){return Object(i.a)(t)?t:Object(r.a)(t)?t:Date.parse(t)}e.a=function(t,e){return e=e||o,null==t||""===t?null:e(t)}},"../../../../vega-util/src/toNumber.js":function(t,e,n){"use strict";e.a=function(t){return null==t||""===t?null:+t}},"../../../../vega-util/src/toSet.js":function(t,e,n){"use strict";e.a=function(t){for(var e={},n=0,r=t.length;n<r;++n)e[t[n]]=!0;return e}},"../../../../vega-util/src/toString.js":function(t,e,n){"use strict";e.a=function(t){return null==t||""===t?null:t+""}},"../../../../vega-util/src/transform.js":function(t,e,n){"use strict";e.a=function(t,e){return u(t,e,o.a,r.c)},e.b=function(t,e){var n=Math.sign(t[0]);return u(t,e,a(n),s(n))},e.c=function(t,e,n){return u(t,e,c(n),c(1/n))},e.d=function(t,e,n){return l(t,e,n,o.a,r.c)},e.e=function(t,e,n){var r=Math.sign(t[0]);return l(t,e,n,a(r),s(r))},e.f=function(t,e,n,r){return l(t,e,n,c(r),c(1/r))};var r=n("../../../../vega-util/src/accessors.js"),i=n("../../../../vega-util/src/peek.js"),o=n("../../../../vega-util/src/toNumber.js");function s(t){return function(e){return t*Math.exp(e)}}function a(t){return function(e){return Math.log(t*e)}}function c(t){return function(e){return e<0?-Math.pow(-e,t):Math.pow(e,t)}}function u(t,e,n,r){var o=n(t[0]),s=n(Object(i.a)(t)),a=(s-o)*e;return[r(o-a),r(s-a)]}function l(t,e,n,r,o){var s=r(t[0]),a=r(Object(i.a)(t)),c=null!=e?r(e):(s+a)/2;return[o(c+(s-c)*n),o(c+(a-c)*n)]}},"../../../../vega-util/src/truncate.js":function(t,e,n){"use strict";e.a=function(t,e,n,r){var i=null!=r?r:"…",o=t+"",s=o.length,a=Math.max(0,e-i.length);return s<=e?o:"left"===n?i+o.slice(s-a):"center"===n?o.slice(0,Math.ceil(a/2))+i+o.slice(s-~~(a/2)):o.slice(0,a)+i}},"../../../../vega-util/src/visitArray.js":function(t,e,n){"use strict";e.a=function(t,e,n){if(t){var r,i=0,o=t.length;if(e)for(;i<o;++i)(r=e(t[i]))&&n(r,i,t);else t.forEach(n)}}},"../../../../vega-view-transforms/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../vega-view-transforms/src/Bound.js");n.d(e,"bound",function(){return r.a});var i=n("../../../../vega-view-transforms/src/Identifier.js");n.d(e,"identifier",function(){return i.a});var o=n("../../../../vega-view-transforms/src/Mark.js");n.d(e,"mark",function(){return o.a});var s=n("../../../../vega-view-transforms/src/Overlap.js");n.d(e,"overlap",function(){return s.a});var a=n("../../../../vega-view-transforms/src/Render.js");n.d(e,"render",function(){return a.a});var c=n("../../../../vega-view-transforms/src/ViewLayout.js");n.d(e,"viewlayout",function(){return c.a})},"../../../../vega-view-transforms/src/Bound.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-scenegraph/index.js"),o=n("../../../../vega-util/index.js");function s(t){r.d.call(this,null,t)}function a(t,e,n){return e(t.bounds.clear(),t,n)}Object(o.s)(s,r.d).transform=function(t,e){var n,r=e.dataflow,o=t.mark,s=o.marktype,c=i.g[s],u=c.bound,l=o.bounds;return c.nested?(o.items.length&&r.dirty(o.items[0]),l=a(o,u),o.items.forEach(function(t){t.bounds.clear().union(l)})):"group"===s||t.modified()?(e.visit(e.MOD,function(t){r.dirty(t)}),l.clear(),o.items.forEach(function(t){l.union(a(t,u))})):(n=e.changed(e.REM),e.visit(e.ADD,function(t){l.union(a(t,u))}),e.visit(e.MOD,function(t){n=n||l.alignsWith(t.bounds),r.dirty(t),l.union(a(t,u))}),n&&(l.clear(),o.items.forEach(function(t){l.union(t.bounds)}))),Object(i.j)(o),e.modifies("bounds")}},"../../../../vega-view-transforms/src/Identifier.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=":vega_identifier:";function s(t){r.d.call(this,0,t)}s.Definition={type:"Identifier",metadata:{modifies:!0},params:[{name:"as",type:"string",required:!0}]},Object(i.s)(s,r.d).transform=function(t,e){var n=function(t){var e=t._signals[o];e||(t._signals[o]=e=t.add(0));return e}(e.dataflow),r=n.value,i=t.as;return e.visit(e.ADD,function(t){t[i]||(t[i]=++r)}),n.set(this.value=r),e}},"../../../../vega-view-transforms/src/Mark.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-scenegraph/index.js"),o=n("../../../../vega-util/index.js");function s(t){r.d.call(this,null,t)}Object(o.s)(s,r.d).transform=function(t,e){var n,r,o,s=this.value;s||((s=e.dataflow.scenegraph().mark(t.markdef,(r=(n=t).groups,o=n.parent,r&&1===r.size?r.get(Object.keys(r.object)[0]):r&&o?r.lookup(o):null),t.index)).group.context=t.context,t.context.group||(t.context.group=s.group),s.source=this,s.clip=t.clip,s.interactive=t.interactive,this.value=s);var a="group"===s.marktype?i.d:i.f;return e.visit(e.ADD,function(t){a.call(t,s)}),(t.modified("clip")||t.modified("interactive"))&&(s.clip=t.clip,s.interactive=!!t.interactive,s.zdirty=!0,e.reflow()),s.items=e.source,e}},"../../../../vega-view-transforms/src/Overlap.js":function(t,e,n){"use strict";e.a=a;var r=n("../../../../vega-view-transforms/src/orient.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-scenegraph/index.js"),s=n("../../../../vega-util/index.js");function a(t){i.d.call(this,null,t)}var c={parity:function(t){return t.filter(function(t,e){return e%2?t.opacity=0:1})},greedy:function(t){var e;return t.filter(function(t,n){return n&&u(e.bounds,t.bounds)?t.opacity=0:(e=t,1)})}};function u(t,e){return!(t.x2-1<e.x1||t.x1+1>e.x2||t.y2-1<e.y1||t.y1+1>e.y2)}function l(t){for(var e,n=1,r=t.length,i=t[0].bounds;n<r;i=e,++n)if(u(i,e=t[n].bounds))return!0}function p(t){var e=t.bounds;return e.width()>1&&e.height()>1}Object(s.s)(a,i.d).transform=function(t,e){var n=c[t.method]||c.parity,i=e.materialize(e.SOURCE).source;if(i){t.sort&&(i=i.slice().sort(t.sort)),"greedy"===t.method&&(i=i.filter(p)),i.forEach(function(t){t.opacity=1});var a,u,d,h,f,m=i;if(m.length>=3&&l(m)){e=e.reflow(t.modified()).modifies("opacity");do{m=n(m)}while(m.length>=3&&l(m));m.length<3&&!Object(s.J)(i).opacity&&(m.length>1&&(Object(s.J)(m).opacity=0),Object(s.J)(i).opacity=1)}if(t.boundScale){var g=(a=t.boundScale,u=t.boundOrient,d=t.boundTolerance,h=a.range(),f=new o.a,u===r.d||u===r.a?f.set(h[0],-1/0,h[1],1/0):f.set(-1/0,h[0],1/0,h[1]),f.expand(d||1),function(t){return f.encloses(t.bounds)});i.forEach(function(t){g(t)||(t.opacity=0)})}return e}}},"../../../../vega-view-transforms/src/Render.js":function(t,e,n){"use strict";e.a=o;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t){r.d.call(this,null,t)}Object(i.s)(o,r.d).transform=function(t,e){var n=e.dataflow;if(e.visit(e.ALL,function(t){n.dirty(t)}),e.fields&&e.fields.zindex){var r=e.source&&e.source[0];r&&(r.mark.zdirty=!0)}}},"../../../../vega-view-transforms/src/ViewLayout.js":function(t,e,n){"use strict";e.a=C;var r=n("../../../../vega-view-transforms/src/gridLayout.js"),i=n("../../../../vega-view-transforms/src/orient.js"),o=n("../../../../vega-dataflow/index.js"),s=n("../../../../vega-scenegraph/index.js"),a=n("../../../../vega-util/index.js"),c="fit",u="fit-x",l="fit-y",p="pad",d="none",h="padding",f="axis",m="title",g="frame",v="legend",y="scope",b="row-header",_="row-footer",j="column-header",x="column-footer",w=.5,O=new s.a;function C(t){o.d.call(this,null,t)}function k(t,e,n){return t[e]===n?0:(t[e]=n,1)}function E(t){var e=t.items[0].datum.orient;return e===i.b||e===i.c}function S(t,e,n,r){var o,a,c,u,l=e.items[0],p=l.datum,d=p.orient,h=(u=+(c=p).grid,[c.ticks?u++:-1,c.labels?u++:-1,u+ +c.domain]),f=l.range,m=l.offset,g=l.position,v=l.minExtent,y=l.maxExtent,b=p.title&&l.items[h[2]].items[0],_=l.titlePadding,j=l.bounds,x=0,C=0;switch(O.clear().union(j),j.clear(),(o=h[0])>-1&&j.union(l.items[o].bounds),(o=h[1])>-1&&j.union(l.items[o].bounds),d){case i.d:x=g||0,C=-m,a=Math.max(v,Math.min(y,-j.y1)),b&&(b.auto?(a+=_,b.y=-a,a+=b.bounds.height(),j.add(b.bounds.x1,0).add(b.bounds.x2,0)):j.union(b.bounds)),j.add(0,-a).add(f,0);break;case i.b:x=-m,C=g||0,a=Math.max(v,Math.min(y,-j.x1)),b&&(b.auto?(a+=_,b.x=-a,a+=b.bounds.width(),j.add(0,b.bounds.y1).add(0,b.bounds.y2)):j.union(b.bounds)),j.add(-a,0).add(0,f);break;case i.c:x=n+m,C=g||0,a=Math.max(v,Math.min(y,j.x2)),b&&(b.auto?(a+=_,b.x=a,a+=b.bounds.width(),j.add(0,b.bounds.y1).add(0,b.bounds.y2)):j.union(b.bounds)),j.add(0,0).add(a,f);break;case i.a:x=g||0,C=r+m,a=Math.max(v,Math.min(y,j.y2)),b&&(b.auto?(a+=_,b.y=a,a+=b.bounds.height(),j.add(b.bounds.x1,0).add(b.bounds.x2,0)):j.union(b.bounds)),j.add(0,0).add(f,a);break;default:x=l.x,C=l.y}return Object(s.k)(j.translate(x,C),l),k(l,"x",x+w)|k(l,"y",C+w)&&(l.bounds=O,t.dirty(l),l.bounds=j,t.dirty(l)),l.mark.bounds.clear().union(j)}function P(t,e,n,r,o,a,c){var u,l,p,d=e.items[0],h=d.datum.orient,f=d.offset,m=d.bounds,g=0,v=0;switch(h===i.d||h===i.a?(p=o,g=n[h]):h!==i.b&&h!==i.c||(p=r,v=n[h]),O.clear().union(m),m.clear(),d.items.forEach(function(t){m.union(t.bounds)}),u=Math.round(m.width())+2*d.padding-1,l=Math.round(m.height())+2*d.padding-1,h){case i.b:g-=u+f-Math.floor(p.x1),n.left+=l+n.margin;break;case i.c:g+=f+Math.ceil(p.x2),n.right+=l+n.margin;break;case i.d:v-=l+f-Math.floor(p.y1),n.top+=u+n.margin;break;case i.a:v+=f+Math.ceil(p.y2),n.bottom+=u+n.margin;break;case"top-left":g+=f,v+=f;break;case"top-right":g+=a-u-f,v+=f;break;case"bottom-left":g+=f,v+=c-l-f;break;case"bottom-right":g+=a-u-f,v+=c-l-f;break;default:g=d.x,v=d.y}return Object(s.k)(m.set(g,v,g+u,v+l),d),k(d,"x",g)|k(d,"width",u)|k(d,"y",v)|k(d,"height",l)&&(d.bounds=O,t.dirty(d),d.bounds=m,t.dirty(d)),d.mark.bounds.clear().union(m)}Object(a.s)(C,o.d).transform=function(t,e){var n=e.dataflow;return t.mark.items.forEach(function(e){t.layout&&Object(r.a)(n,e,t.layout),function(t,e,n){var r,o,a,w,C,M,A=e.items,T=Math.max(0,e.width||0),D=Math.max(0,e.height||0),I=(new s.a).set(0,0,T,D),R=I.clone(),N=I.clone(),L=I.clone(),F=[];for(C=0,M=A.length;C<M;++C)switch((o=A[C]).role){case f:R.union(w=S(t,o,T,D)),(E(o)?N:L).union(w);break;case m:r=o;break;case v:F.push(o);break;case g:case y:case b:case _:case j:case x:N.union(o.bounds),L.union(o.bounds);break;default:I.union(o.bounds)}r&&(R.union(w=function(t,e,n){var r=e.items[0],o=r.datum.orient,s=r.offset,a=r.bounds,c=0,u=0;switch(O.clear().union(a),o){case i.d:c=r.x,u=n.y1-s;break;case i.b:c=n.x1-s,u=r.y;break;case i.c:c=n.x2+s,u=r.y;break;case i.a:c=r.x,u=n.y2+s;break;default:c=r.x,u=r.y}a.translate(c-r.x,u-r.y),k(r,"x",c)|k(r,"y",u)&&(r.bounds=O,t.dirty(r),r.bounds=a,t.dirty(r));return e.bounds.clear().union(a)}(t,r,R)),(E(r)?N:L).union(w));if(F.length)for(a={left:0,right:0,top:0,bottom:0,margin:n.legendMargin||8},C=0,M=F.length;C<M;++C)if(w=P(t,F[C],a,N,L,T,D),n.autosize&&n.autosize.type===c){var V=F[C].items[0].datum.orient;V===i.b||V===i.c?I.add(w.x1,0).add(w.x2,0):V!==i.d&&V!==i.a||I.add(0,w.y1).add(0,w.y2)}else I.union(w);I.union(N).union(L).union(R),function(t,e,n,r){var i=r.autosize||{},o=i.type,s=t._width,a=t._height,f=t.padding();if(t._autosize<1||!o)return;var m=Math.max(0,e.width||0),g=Math.max(0,Math.ceil(-n.x1)),v=Math.max(0,Math.ceil(n.x2-m)),y=Math.max(0,e.height||0),b=Math.max(0,Math.ceil(-n.y1)),_=Math.max(0,Math.ceil(n.y2-y));i.contains===h&&(s-=f.left+f.right,a-=f.top+f.bottom);o===d?(g=0,b=0,m=s,y=a):o===c?(m=Math.max(0,s-g-v),y=Math.max(0,a-b-_)):o===u?(m=Math.max(0,s-g-v),a=y+b+_):o===l?(s=m+g+v,y=Math.max(0,a-b-_)):o===p&&(s=m+g+v,a=y+b+_);t._resizeView(s,a,m,y,[g,b],i.resize)}(t,e,I,n)}(n,e,t)}),e}},"../../../../vega-view-transforms/src/gridLayout.js":function(t,e,n){"use strict";e.a=function(t,e,n){var r,x,w,O,C,k,E,S,P,M,A,T,D,I=function(t){for(var e,n,r=t.items,i=r.length,f=0,m={marks:[],rowheaders:[],rowfooters:[],colheaders:[],colfooters:[],rowtitle:null,coltitle:null};f<i;++f)if(e=r[f],n=e.items,"group"===e.marktype)switch(e.role){case o:case s:break;case a:h(n,m.rowheaders);break;case c:h(n,m.rowfooters);break;case l:h(n,m.colheaders);break;case p:h(n,m.colfooters);break;case u:m.rowtitle=n[0];break;case d:m.coltitle=n[0];break;default:h(n,m.marks)}return m}(e),R=I.marks,N="flush"===n.bounds,L=N?f:m,F=new i.a(0,0,0,0),V=y(n.align,"column"),z=y(n.align,"row"),B=y(n.padding,"column"),q=y(n.padding,"row"),H=n.offset,U=e.columns||n.columns||R.length,W=U<0?1:Math.ceil(R.length/U),G=W*U,X=[],Y=[],K=[],$=[],Q=R.length;for(x=0;x<U;++x)Y[x]=0;for(x=0;x<W;++x)$[x]=0;for(x=0;x<Q;++x)C=L(R[x]),O=~~(x/U),E=(w=x%U)?Math.ceil(L(R[x-1]).x2):0,S=O?Math.ceil(L(R[x-U]).y2):0,Y[w]=Math.max(Y[w],E),$[O]=Math.max($[O],S),X.push(B+b(C.x1)),K.push(q+b(C.y1)),t.dirty(R[x]);for(x=0;x<Q;++x)x%U==0&&(X[x]=0),x<U&&(K[x]=0);if("each"===V)for(w=1;w<U;++w){for(D=0,x=w;x<Q;x+=U)D<X[x]&&(D=X[x]);for(x=w;x<Q;x+=U)X[x]=D+Y[w]}else if("all"===V){for(T=0,w=1;w<U;++w)T<Y[w]&&(T=Y[w]);for(D=0,x=0;x<Q;++x)x%U&&D<X[x]&&(D=X[x]);for(x=0;x<Q;++x)x%U&&(X[x]=D+T)}else for(w=1;w<U;++w)for(x=w;x<Q;x+=U)X[x]+=Y[w];if("each"===z)for(O=1;O<W;++O){for(D=0,r=(x=O*U)+U;x<r;++x)D<K[x]&&(D=K[x]);for(x=O*U;x<r;++x)K[x]=D+$[O]}else if("all"===z){for(T=0,O=1;O<W;++O)T<$[O]&&(T=$[O]);for(D=0,x=U;x<Q;++x)D<K[x]&&(D=K[x]);for(x=U;x<Q;++x)K[x]=D+T}else for(O=1;O<W;++O)for(r=(x=O*U)+U;x<r;++x)K[x]+=$[O];for(P=0,x=0;x<Q;++x)k=R[x],E=k.x||0,k.x=P=X[x]+(x%U?P:0),k.bounds.translate(P-E,0);for(w=0;w<U;++w)for(M=0,x=w;x<Q;x+=U)k=R[x],S=k.y||0,k.y=M+=K[x],k.bounds.translate(0,M-S);for(x=0;x<Q;++x)R[x].mark.bounds.clear();for(x=0;x<Q;++x)k=R[x],t.dirty(k),F.union(k.mark.bounds.union(k.bounds));function Z(t,e){return Math.floor(Math.min(t,e))}function J(t,e){return Math.ceil(Math.max(t,e))}L=N?g:v,A=y(n.headerBand,"row",null),P=_(t,I.rowheaders,R,U,W,-y(H,"rowHeader"),Z,0,L,"x1",0,U,1,A),A=y(n.headerBand,"column",null),M=_(t,I.colheaders,R,U,U,-y(H,"columnHeader"),Z,1,L,"y1",0,1,U,A),A=y(n.footerBand,"row",null),_(t,I.rowfooters,R,U,W,y(H,"rowFooter"),J,0,L,"x2",U-1,U,1,A),A=y(n.footerBand,"column",null),_(t,I.colfooters,R,U,U,y(H,"columnFooter"),J,1,L,"y2",G-U,1,U,A),I.rowtitle&&(D=P-y(H,"rowTitle"),A=y(n.titleBand,"row",.5),j(t,I.rowtitle,D,0,F,A));I.coltitle&&(D=M-y(H,"columnTitle"),A=y(n.titleBand,"column",.5),j(t,I.coltitle,D,1,F,A))};var r=n("../../../../vega-util/index.js"),i=n("../../../../vega-scenegraph/index.js"),o="axis",s="legend",a="row-header",c="row-footer",u="row-title",l="column-header",p="column-footer",d="column-title";function h(t,e){for(var n=0,r=t.length;n<r;++n)e.push(t[n])}function f(t){return{x1:0,y1:0,x2:t.width||0,y2:t.height||0}}function m(t){var e=t.bounds.clone();return e.empty()?e.set(0,0,0,0):e.translate(-(t.x||0),-(t.y||0))}function g(t,e){return"x1"===e?t.x||0:"y1"===e?t.y||0:"x2"===e?(t.x||0)+(t.width||0):"y2"===e?(t.y||0)+(t.height||0):void 0}function v(t,e){return t.bounds[e]}function y(t,e,n){var i=Object(r.y)(t)?t[e]:t;return null!=i?i:void 0!==n?n:0}function b(t){return t<0?Math.ceil(-t):0}function _(t,e,n,r,i,o,s,a,c,u,l,p,d,h){var f,m,g,v,y,b,_,j,x,w=n.length,O=0,C=0;if(!w)return O;for(f=l;f<w;f+=p)n[f]&&(O=s(O,c(n[f],u)));if(!e.length)return O;for(e.length>i&&(t.warn("Grid headers exceed limit: "+i),e=e.slice(0,i)),O+=o,m=0,v=e.length;m<v;++m)t.dirty(e[m]),e[m].mark.bounds.clear();for(f=l,m=0,v=e.length;m<v;++m,f+=p){for(y=(b=e[m]).mark.bounds,g=f;g>=0&&null==(_=n[g]);g-=d);a?(j=null==h?_.x:Math.round(_.bounds.x1+h*_.bounds.width()),x=O):(j=O,x=null==h?_.y:Math.round(_.bounds.y1+h*_.bounds.height())),y.union(b.bounds.translate(j-(b.x||0),x-(b.y||0))),b.x=j,b.y=x,t.dirty(b),C=s(C,y[u])}return C}function j(t,e,n,r,i,o){if(e){t.dirty(e);var s=n,a=n;r?s=Math.round(i.x1+o*i.width()):a=Math.round(i.y1+o*i.height()),e.bounds.translate(s-(e.x||0),a-(e.y||0)),e.mark.bounds.clear().union(e.bounds),e.x=s,e.y=a,t.dirty(e)}}},"../../../../vega-view-transforms/src/orient.js":function(t,e,n){"use strict";n.d(e,"d",function(){return r}),n.d(e,"b",function(){return i}),n.d(e,"c",function(){return o}),n.d(e,"a",function(){return s});var r="top",i="left",o="right",s="bottom"},"../../../../vega-view/index.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/View.js");n.d(e,"a",function(){return r.a})},"../../../../vega-view/src/View.js":function(t,e,n){"use strict";e.a=b;var r=n("../../../../vega-view/src/cursor.js"),i=n("../../../../vega-view/src/data.js"),o=n("../../../../vega-view/src/events.js"),s=n("../../../../vega-view/src/hover.js"),a=n("../../../../vega-view/src/finalize.js"),c=n("../../../../vega-view/src/initialize.js"),u=n("../../../../vega-view/src/render-to-image-url.js"),l=n("../../../../vega-view/src/render-to-canvas.js"),p=n("../../../../vega-view/src/render-to-svg.js"),d=n("../../../../vega-view/src/render-size.js"),h=n("../../../../vega-view/src/runtime.js"),f=n("../../../../vega-view/src/size.js"),m=n("../../../../vega-view/src/state.js"),g=n("../../../../vega-dataflow/index.js"),v=n("../../../../vega-util/index.js"),y=n("../../../../vega-scenegraph/index.js");function b(t,e){e=e||{},g.a.call(this),this.loader(e.loader||this._loader),this.logLevel(e.logLevel||0),this._el=null,this._renderType=e.renderer||y.h.Canvas,this._scenegraph=new y.i;var n=this._scenegraph.root;this._renderer=null,this._redraw=!0,this._handler=(new y.b).scene(n),this._preventDefault=!1,this._eventListeners=[],this._resizeListeners=[];var i=Object(h.a)(this,t,e.functions);this._runtime=i,this._signals=i.signals,this._bind=(t.bindings||[]).map(function(t){return{state:null,param:Object(v.l)({},t)}}),i.root&&i.root.set(n),n.source=i.data.root.input,this.pulse(i.data.root.input,this.changeset().insert(n.items)),this._background=i.background||null,this._eventConfig=Object(o.b)(i.eventConfig),this._width=this.width(),this._height=this.height(),this._viewWidth=Object(f.d)(this,this._width),this._viewHeight=Object(f.c)(this,this._height),this._origin=[0,0],this._resize=0,this._autosize=1,Object(f.a)(this),Object(r.a)(this)}var _=Object(v.s)(b,g.a);function j(t,e){return t._signals.hasOwnProperty(e)?t._signals[e]:Object(v.k)("Unrecognized signal name: "+Object(v.L)(e))}function x(t,e){var n=(t._targets||[]).filter(function(t){var n=t._update;return n&&n.handler===e});return n.length?n[0]:null}_.run=function(t){if(g.a.prototype.run.call(this,t),this._redraw||this._resize)try{this.render()}catch(t){this.error(t)}return this},_.render=function(){return this._renderer&&(this._resize&&(this._resize=0,Object(d.c)(this)),this._renderer.render(this._scenegraph.root)),this._redraw=!1,this},_.dirty=function(t){this._redraw=!0,this._renderer&&this._renderer.dirty(t)},_.container=function(){return this._el},_.scenegraph=function(){return this._scenegraph},_.signal=function(t,e,n){var r=j(this,t);return 1===arguments.length?r.value:this.update(r,e,n)},_.background=function(t){return arguments.length?(this._background=t,this._resize=1,this):this._background},_.width=function(t){return arguments.length?this.signal("width",t):this.signal("width")},_.height=function(t){return arguments.length?this.signal("height",t):this.signal("height")},_.padding=function(t){return arguments.length?this.signal("padding",t):this.signal("padding")},_.autosize=function(t){return arguments.length?this.signal("autosize",t):this.signal("autosize")},_.renderer=function(t){return arguments.length?(Object(y.o)(t)||Object(v.k)("Unrecognized renderer type: "+t),t!==this._renderType&&(this._renderType=t,this._renderer&&(this._renderer=null,this.initialize(this._el))),this):this._renderType},_.loader=function(t){return arguments.length?(t!==this._loader&&(g.a.prototype.loader.call(this,t),this._renderer&&(this._renderer=null,this.initialize(this._el))),this):this._loader},_.resize=function(){return this._autosize=1,this},_._resizeView=f.b,_.addEventListener=function(t,e){return this._handler.on(t,e),this},_.removeEventListener=function(t,e){return this._handler.off(t,e),this},_.addResizeListener=function(t){var e=this._resizeListeners;return e.indexOf(t)<0&&e.push(t),this},_.removeResizeListener=function(t){var e=this._resizeListeners,n=e.indexOf(t);return n>=0&&e.splice(n,1),this},_.addSignalListener=function(t,e){var n=j(this,t),r=x(n,e);return r||((r=function(){e(t,n.value)}).handler=e,this.on(n,null,r)),this},_.removeSignalListener=function(t,e){var n=j(this,t),r=x(n,e);return r&&n._targets.remove(r),this},_.preventDefault=function(t){return arguments.length?(this._preventDefault=t,this):this._preventDefault},_.tooltipHandler=function(t){var e=this._handler;return arguments.length?(e.handleTooltip=t||y.e.prototype.handleTooltip,this):e.handleTooltip},_.events=o.a,_.finalize=a.a,_.hover=s.a,_.data=i.b,_.change=i.a,_.insert=i.c,_.remove=i.d,_.initialize=c.a,_.toImageURL=u.a,_.toCanvas=l.a,_.toSVG=p.a,_.getState=m.a,_.setState=m.b},"../../../../vega-view/src/bind.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/element.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-array/index.js"),s="vega-bind",a="vega-bind-name",c="vega-bind-radio",u="vega-option-";function l(t,e,n,i){var o=Object(r.a)("input");for(var s in n)"signal"!==s&&"element"!==s&&o.setAttribute("input"===s?"type":s,n[s]);o.setAttribute("name",n.signal),o.value=i,e.appendChild(o),o.addEventListener("input",function(){t.update(o.value)}),t.elements=[o],t.set=function(t){o.value=t}}function p(t,e,n,i){var o={type:"checkbox",name:n.signal};i&&(o.checked=!0);var s=Object(r.a)("input",o);e.appendChild(s),s.addEventListener("change",function(){t.update(s.checked)}),t.elements=[s],t.set=function(t){s.checked=!!t||null}}function d(t,e,n,i){var o=Object(r.a)("select",{name:n.signal});n.options.forEach(function(t){var e={value:t};m(t,i)&&(e.selected=!0),o.appendChild(Object(r.a)("option",e,t+""))}),e.appendChild(o),o.addEventListener("change",function(){t.update(n.options[o.selectedIndex])}),t.elements=[o],t.set=function(t){for(var e=0,r=n.options.length;e<r;++e)if(m(n.options[e],t))return void(o.selectedIndex=e)}}function h(t,e,n,i){var o=Object(r.a)("span",{class:c});e.appendChild(o),t.elements=n.options.map(function(e){var s=u+n.signal+"-"+e,a={id:s,type:"radio",name:n.signal,value:e};m(e,i)&&(a.checked=!0);var c=Object(r.a)("input",a);return c.addEventListener("change",function(){t.update(e)}),o.appendChild(c),o.appendChild(Object(r.a)("label",{for:s},e+"")),c}),t.set=function(e){for(var n=t.elements,r=0,i=n.length;r<i;++r)m(n[r].value,e)&&(n[r].checked=!0)}}function f(t,e,n,i){i=void 0!==i?i:(+n.max+ +n.min)/2;var s=n.min||Math.min(0,+i)||0,a=n.max||Math.max(100,+i)||100,c=n.step||Object(o.r)(s,a,100),u=Object(r.a)("input",{type:"range",name:n.signal,min:s,max:a,step:c});u.value=i;var l=Object(r.a)("label",{},+i);function p(){l.textContent=u.value,t.update(+u.value)}e.appendChild(u),e.appendChild(l),u.addEventListener("input",p),u.addEventListener("change",p),t.elements=[u],t.set=function(t){u.value=t,l.textContent=t}}function m(t,e){return t===e||t+""==e+""}e.a=function(t,e,n){if(e){var o=n.param,c=n.state;return c||(c=n.state={elements:null,active:!1,set:null,update:function(e){c.source=!0,t.signal(o.signal,e).run()}},o.debounce&&(c.update=Object(i.j)(o.debounce,c.update))),function(t,e,n,i){var o=Object(r.a)("div",{class:s});o.appendChild(Object(r.a)("span",{class:a},n.name||n.signal)),e.appendChild(o);var c=l;switch(n.input){case"checkbox":c=p;break;case"select":c=d;break;case"radio":c=h;break;case"range":c=f}c(t,o,n,i)}(c,e,o,t.signal(o.signal)),c.active||(t.on(t._signals[o.signal],null,function(){c.source?c.source=!1:c.set(t.signal(o.signal))}),c.active=!0),c}}},"../../../../vega-view/src/cursor.js":function(t,e,n){"use strict";var r=n("../../../../vega-util/index.js");e.a=function(t){var e=t._signals.cursor;e||(t._signals.cursor=e=t.add({user:"default",item:null})),t.on(t.events("view","mousemove"),e,function(t,n){var i=e.value,o=i?Object(r.A)(i)?i:i.user:"default",s=n.item&&n.item.cursor||null;return i&&o===i.user&&s==i.item?i:{user:o,item:s}}),t.add(null,function(t){var e,n=t.cursor,i=this.value;return Object(r.A)(n)||(i=n.item,n=n.user),e=n&&"default"!==n?n:i||n,"undefined"!=typeof document&&document.body&&(document.body.style.cursor=e),i},{cursor:e})}},"../../../../vega-view/src/data.js":function(t,e,n){"use strict";e.b=function(t){return o(this,t).values.value},e.a=s,e.c=function(t,e){return s.call(this,t,Object(r.e)().insert(e))},e.d=function(t,e){return s.call(this,t,Object(r.e)().remove(e))};var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t,e){var n=t._runtime.data;return n.hasOwnProperty(e)||Object(i.k)("Unrecognized data set: "+e),n[e]}function s(t,e){Object(r.i)(e)||Object(i.k)("Second argument to changes must be a changeset.");var n=o(this,t);return n.modified=!0,this.pulse(n.input,e)}},"../../../../vega-view/src/element.js":function(t,e,n){"use strict";e.a=function(t,e,n){var r=document.createElement(t);for(var i in e)r.setAttribute(i,e[i]);return null!=n&&(r.textContent=n),r}},"../../../../vega-view/src/events-extend.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/render-size.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../vega-scenegraph/index.js");e.a=function(t,e,n){var s,a,c,u=t._renderer.scene();return u&&(c=Object(r.b)(t),a=e.changedTouches?e.changedTouches[0]:e,(s=Object(o.n)(a,u))[0]-=c[0],s[1]-=c[1]),e.dataflow=t,e.vega=function(t,e,n){var r=e?"group"===e.mark.marktype?e:e.mark.group:null;function o(t){var n,i=r;if(t)for(n=e;n;n=n.mark.group)if(n.mark.name===t){i=n;break}return i&&i.mark&&i.mark.interactive?i:{}}function s(t){if(!t)return n;Object(i.A)(t)&&(t=o(t));for(var e=n.slice();t;)e[0]-=t.x||0,e[1]-=t.y||0,t=t.mark&&t.mark.group;return e}return{view:Object(i.i)(t),item:Object(i.i)(e||{}),group:o,xy:s,x:function(t){return s(t)[0]},y:function(t){return s(t)[1]}}}(t,n,s),e.item=n,e}},"../../../../vega-view/src/events.js":function(t,e,n){"use strict";e.b=function(t){var e=(t=Object(o.l)({},t)).defaults;e&&(Object(o.t)(e.prevent)&&(e.prevent=Object(o.P)(e.prevent)),Object(o.t)(e.allow)&&(e.allow=Object(o.P)(e.allow)));return t},e.a=function(t,e,n){var o,c=this,u=new i.b(n),l=function(n,i){var o,a,l,p,d;t!==s||(a=e,l=(o=c)._eventConfig.defaults,p=l&&l.prevent,d=l&&l.allow,!1===p||!0===d||!0!==p&&!1!==d&&(p?!p[a]:d?d[a]:!o.preventDefault()))||n.preventDefault();try{u.receive(Object(r.a)(c,n,i))}catch(t){c.error(t)}finally{c.run()}};if(t===s)return c.addEventListener(e,l),u;t===a?"undefined"!=typeof window&&(o=[window]):"undefined"!=typeof document&&(o=document.querySelectorAll(t));if(!o)return c.warn("Can not resolve event source: "+t),u;for(var p=0,d=o.length;p<d;++p)o[p].addEventListener(e,l);return c._eventListeners.push({type:e,sources:o,handler:l}),u};var r=n("../../../../vega-view/src/events-extend.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js"),s="view",a="window"},"../../../../vega-view/src/finalize.js":function(t,e,n){"use strict";e.a=function(){for(var t,e,n=this._eventListeners,r=n.length;--r>=0;)for(t=(e=n[r]).sources.length;--t>=0;)e.sources[t].removeEventListener(e.type,e.handler)}},"../../../../vega-view/src/hover.js":function(t,e,n){"use strict";function r(t){return t.item}function i(t){var e=t.item.mark.source;return e.source||e}function o(t){return function(e,n){return n.vega.view().changeset().encode(n.item,t)}}e.a=function(t,e){return t=[t||"hover"],e=[e||"update",t],this.on(this.events("view","mouseover",r),i,o(t)),this.on(this.events("view","mouseout",r),i,o(e)),this}},"../../../../vega-view/src/initialize-handler.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/render-size.js");e.a=function(t,e,n,i){var o=(new i).scene(t.scenegraph().root).initialize(n,Object(r.b)(t),t);return e&&(o.handleTooltip=e.handleTooltip,e.handlers().forEach(function(t){o.on(t.type,t.handler)})),o}},"../../../../vega-view/src/initialize-renderer.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/render-size.js");e.a=function(t,e,n,i,o){return(e=e||new i(t.loader())).initialize(n,Object(r.d)(t),Object(r.a)(t),Object(r.b)(t),o).background(t._background)}},"../../../../vega-view/src/initialize.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/bind.js"),i=n("../../../../vega-view/src/element.js"),o=n("../../../../vega-view/src/initialize-renderer.js"),s=n("../../../../vega-view/src/initialize-handler.js"),a=n("../../../../vega-scenegraph/index.js");function c(t,e){if("string"==typeof e){if("undefined"==typeof document)return t.error("DOM document instance not found."),null;if(!(e=document.querySelector(e)))return t.error("Signal bind element not found: "+e),null}if(e)try{e.innerHTML=""}catch(n){e=null,t.error(n)}return e}e.a=function(t,e){var n,u,l=this,p=l._renderType,d=Object(a.o)(p);return t=l._el=t?c(l,t):null,d||l.error("Unrecognized renderer type: "+p),n=d.handler||a.b,u=t?d.renderer:d.headless,l._renderer=u?Object(o.a)(l,l._renderer,t,u):null,l._handler=Object(s.a)(l,l._handler,t,n),l._redraw=!0,t&&(e=e?c(l,e):t.appendChild(Object(i.a)("div",{class:"vega-bindings"})),l._bind.forEach(function(t){t.param.element&&(t.element=c(l,t.param.element))}),l._bind.forEach(function(t){Object(r.a)(l,t.element||e,t)})),l}},"../../../../vega-view/src/render-headless.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/initialize-renderer.js"),i=n("../../../../vega-scenegraph/index.js");e.a=function(t,e,n){var o=Object(i.o)(e),s=o&&o.headless;return s?t.runAsync().then(function(){return Object(r.a)(t,null,null,s,n).renderAsync(t._scenegraph.root)}):Promise.reject("Unrecognized renderer type: "+e)}},"../../../../vega-view/src/render-size.js":function(t,e,n){"use strict";function r(t){var e=t.padding();return Math.max(0,t._viewWidth+e.left+e.right)}function i(t){var e=t.padding();return Math.max(0,t._viewHeight+e.top+e.bottom)}function o(t){var e=t.padding(),n=t._origin;return[e.left+n[0],e.top+n[1]]}e.d=r,e.a=i,e.b=o,e.c=function(t){var e=o(t),n=r(t),s=i(t);t._renderer.background(t._background),t._renderer.resize(n,s,e),t._handler.origin(e),t._resizeListeners.forEach(function(t){t(n,s)})}},"../../../../vega-view/src/render-to-canvas.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/render-headless.js"),i=n("../../../../vega-scenegraph/index.js");e.a=function(t){return Object(r.a)(this,i.h.Canvas,t).then(function(t){return t.canvas()})}},"../../../../vega-view/src/render-to-image-url.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/render-headless.js"),i=n("../../../../vega-scenegraph/index.js");e.a=function(t,e){return t!==i.h.Canvas&&t!==i.h.SVG&&t!==i.h.PNG?Promise.reject("Unrecognized image type: "+t):Object(r.a)(this,t,e).then(function(e){return t===i.h.SVG?(n=e.svg(),r=new Blob([n],{type:"image/svg+xml"}),window.URL.createObjectURL(r)):e.canvas().toDataURL("image/png");var n,r})}},"../../../../vega-view/src/render-to-svg.js":function(t,e,n){"use strict";var r=n("../../../../vega-view/src/render-headless.js"),i=n("../../../../vega-scenegraph/index.js");e.a=function(t){return Object(r.a)(this,i.h.SVG,t).then(function(t){return t.svg()})}},"../../../../vega-view/src/runtime.js":function(t,e,n){"use strict";var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-parser/index.js"),o=n("../../../../vega-runtime/index.js");e.a=function(t,e,n){var s=n||i.a;return Object(o.b)(e,Object(o.a)(t,r.m,s))}},"../../../../vega-view/src/size.js":function(t,e,n){"use strict";e.d=i,e.c=o,e.a=function(t){var e=t._signals,n=e.width,r=e.height,s=e.padding;function a(){t._autosize=t._resize=1}t._resizeWidth=t.add(null,function(e){t._width=e.size,t._viewWidth=i(t,e.size),a()},{size:n}),t._resizeHeight=t.add(null,function(e){t._height=e.size,t._viewHeight=o(t,e.size),a()},{size:r});var c=t.add(null,a,{pad:s});t._resizeWidth.rank=n.rank+1,t._resizeHeight.rank=r.rank+1,c.rank=s.rank+1},e.b=function(t,e,n,r,i,o){this.runAfter(function(s){var a=0;s._autosize=0,s.width()!==n&&(a=1,s.width(n),s._resizeWidth.skip(!0)),s.height()!==r&&(a=1,s.height(r),s._resizeHeight.skip(!0)),s._viewWidth!==t&&(s._resize=1,s._viewWidth=t),s._viewHeight!==e&&(s._resize=1,s._viewHeight=e),s._origin[0]===i[0]&&s._origin[1]===i[1]||(s._resize=1,s._origin=i),a&&s.run("enter"),o&&s.runAfter(function(){s.resize()})},!1,1)};var r="padding";function i(t,e){var n=t.autosize(),i=t.padding();return e-(n&&n.contains===r?i.left+i.right:0)}function o(t,e){var n=t.autosize(),i=t.padding();return e-(n&&n.contains===r?i.top+i.bottom:0)}},"../../../../vega-view/src/state.js":function(t,e,n){"use strict";e.a=function(t){return this._runtime.getState(t||{data:o,signals:s,recurse:!0})},e.b=function(t){var e=this;return e.runAfter(function(){e._trigger=!1,e._runtime.setState(t),e.run().runAfter(function(){e._trigger=!0})}),this};var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js");function o(t,e){return e.modified&&Object(i.t)(e.input.value)&&t.indexOf("_:vega:_")}function s(t,e){return!("parent"===t||e instanceof r.m.proxy)}},"../../../../vega-voronoi/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../vega-voronoi/src/Voronoi.js");n.d(e,"voronoi",function(){return r.a})},"../../../../vega-voronoi/src/Voronoi.js":function(t,e,n){"use strict";e.a=s;var r=n("../../../../vega-dataflow/index.js"),i=n("../../../../vega-util/index.js"),o=n("../../../../d3-voronoi/index.js");function s(t){r.d.call(this,null,t)}s.Definition={type:"Voronoi",metadata:{modifies:!0},params:[{name:"x",type:"field",required:!0},{name:"y",type:"field",required:!0},{name:"size",type:"number",array:!0,length:2},{name:"extent",type:"array",array:!0,length:2,default:[[-1e5,-1e5],[1e5,1e5]],content:{type:"number",array:!0,length:2}},{name:"as",type:"string",default:"path"}]};var a=[[-1e5,-1e5],[1e5,1e5]];Object(i.s)(s,r.d).transform=function(t,e){var n,r,i,s,c=t.as||"path",u=e.source;for(n=Object(o.a)().x(t.x).y(t.y),t.size?n.size(t.size):n.extent(t.extent||a),this.value=n=n(u),r=n.polygons(),i=0,s=u.length;i<s;++i)u[i][c]=r[i]?"M"+r[i].join("L")+"Z":null;return e.reflow(t.modified()).modifies(c)}},"../../../../vega-wordcloud/index.js":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var r=n("../../../../vega-wordcloud/src/Wordcloud.js");n.d(e,"wordcloud",function(){return r.a})},"../../../../vega-wordcloud/src/CloudLayout.js":function(t,e,n){"use strict";var r=n("../../../../vega-canvas/index.js"),i=Math.PI/180,o=64,s=2048;function a(t,e,n,r){if(!e.sprite){var a=t.context,c=t.ratio;a.clearRect(0,0,(o<<5)/c,s/c);var u,l,p,d,h,f=0,m=0,g=0,v=n.length;for(--r;++r<v;){if(e=n[r],a.save(),a.font=e.style+" "+e.weight+" "+~~((e.size+1)/c)+"px "+e.font,u=a.measureText(e.text+"m").width*c,p=e.size<<1,e.rotate){var y=Math.sin(e.rotate*i),b=Math.cos(e.rotate*i),_=u*b,j=u*y,x=p*b,w=p*y;u=Math.max(Math.abs(_+w),Math.abs(_-w))+31>>5<<5,p=~~Math.max(Math.abs(j+x),Math.abs(j-x))}else u=u+31>>5<<5;if(p>g&&(g=p),f+u>=o<<5&&(f=0,m+=g,g=0),m+p>=s)break;a.translate((f+(u>>1))/c,(m+(p>>1))/c),e.rotate&&a.rotate(e.rotate*i),a.fillText(e.text,0,0),e.padding&&(a.lineWidth=2*e.padding,a.strokeText(e.text,0,0)),a.restore(),e.width=u,e.height=p,e.xoff=f,e.yoff=m,e.x1=u>>1,e.y1=p>>1,e.x0=-e.x1,e.y0=-e.y1,e.hasText=!0,f+=u}for(var O=a.getImageData(0,0,(o<<5)/c,s/c).data,C=[];--r>=0;)if((e=n[r]).hasText){for(l=(u=e.width)>>5,p=e.y1-e.y0,d=0;d<p*l;d++)C[d]=0;if(null==(f=e.xoff))return;m=e.yoff;var k=0,E=-1;for(h=0;h<p;h++){for(d=0;d<u;d++){var S=l*h+(d>>5),P=O[(m+h)*(o<<5)+(f+d)<<2]?1<<31-d%32:0;C[S]|=P,k|=P}k?E=h:(e.y0++,p--,h--,m++)}e.y1=e.y0+E,e.sprite=C.slice(0,(e.y1-e.y0)*l)}}}function c(t,e,n){n>>=5;for(var r,i=t.sprite,o=t.width>>5,s=t.x-(o<<4),a=127&s,c=32-a,u=t.y1-t.y0,l=(t.y+t.y0)*n+(s>>5),p=0;p<u;p++){r=0;for(var d=0;d<=o;d++)if((r<<c|(d<o?(r=i[p*o+d])>>>a:0))&e[l+d])return!0;l+=n}return!1}function u(t,e){var n=t[0],r=t[1];e.x+e.x0<n.x&&(n.x=e.x+e.x0),e.y+e.y0<n.y&&(n.y=e.y+e.y0),e.x+e.x1>r.x&&(r.x=e.x+e.x1),e.y+e.y1>r.y&&(r.y=e.y+e.y1)}function l(t){var e=t[0]/t[1];return function(t){return[e*(t*=.1)*Math.cos(t),t*Math.sin(t)]}}function p(t){return"function"==typeof t?t:function(){return t}}e.a=function(){var t,e,n,i,h,f,m,g=[256,256],v=l,y=[],b=Math.random,_={};function j(t,e,n){for(var r,i,o,s,a,u=e.x,l=e.y,p=Math.sqrt(g[0]*g[0]+g[1]*g[1]),d=v(g),h=b()<.5?1:-1,f=-h;(r=d(f+=h))&&(i=~~r[0],o=~~r[1],!(Math.min(Math.abs(i),Math.abs(o))>=p));)if(e.x=u+i,e.y=l+o,!(e.x+e.x0<0||e.y+e.y0<0||e.x+e.x1>g[0]||e.y+e.y1>g[1]||n&&c(e,t,g[0])||n&&(a=n,!((s=e).x+s.x1>a[0].x&&s.x+s.x0<a[1].x&&s.y+s.y1>a[0].y&&s.y+s.y0<a[1].y)))){for(var m,y=e.sprite,_=e.width>>5,j=g[0]>>5,x=e.x-(_<<4),w=127&x,O=32-w,C=e.y1-e.y0,k=(e.y+e.y0)*j+(x>>5),E=0;E<C;E++){m=0;for(var S=0;S<=_;S++)t[k+S]|=m<<O|(S<_?(m=y[E*_+S])>>>w:0);k+=j}return e.sprite=null,!0}return!1}return _.layout=function(){for(var c=function(t){t.width=t.height=1;var e=Math.sqrt(t.getContext("2d").getImageData(0,0,1,1).data.length>>2);t.width=(o<<5)/e,t.height=s/e;var n=t.getContext("2d");return n.fillStyle=n.strokeStyle="red",n.textAlign="center",{context:n,ratio:e}}(Object(r.a)()),l=function(t){var e=[],n=-1;for(;++n<t;)e[n]=0;return e}((g[0]>>5)*g[1]),p=null,d=y.length,v=-1,_=[],x=y.map(function(r){return{text:t(r),font:e(r),style:i(r),weight:h(r),rotate:f(r),size:~~n(r),padding:m(r),xoff:0,yoff:0,x1:0,y1:0,x0:0,y0:0,hasText:!1,sprite:null,datum:r}}).sort(function(t,e){return e.size-t.size});++v<d;){var w=x[v];w.x=g[0]*(b()+.5)>>1,w.y=g[1]*(b()+.5)>>1,a(c,w,x,v),w.hasText&&j(l,w,p)&&(_.push(w),p?u(p,w):p=[{x:w.x+w.x0,y:w.y+w.y0},{x:w.x+w.x1,y:w.y+w.y1}],w.x-=g[0]>>1,w.y-=g[1]>>1)}return _},_.words=function(t){return arguments.length?(y=t,_):y},_.size=function(t){return arguments.length?(g=[+t[0],+t[1]],_):g},_.font=function(t){return arguments.length?(e=p(t),_):e},_.fontStyle=function(t){return arguments.length?(i=p(t),_):i},_.fontWeight=function(t){return arguments.length?(h=p(t),_):h},_.rotate=function(t){return arguments.length?(f=p(t),_):f},_.text=function(e){return arguments.length?(t=p(e),_):t},_.spiral=function(t){return arguments.length?(v=d[t]||t,_):v},_.fontSize=function(t){return arguments.length?(n=p(t),_):n},_.padding=function(t){return arguments.length?(m=p(t),_):m},_.random=function(t){return arguments.length?(b=t,_):b},_};var d={archimedean:l,rectangular:function(t){var e=4*t[0]/t[1],n=0,r=0;return function(t){var i=t<0?-1:1;switch(Math.sqrt(1+4*i*t)-i&3){case 0:n+=e;break;case 1:r+=4;break;case 2:n-=e;break;default:r-=4}return[n,r]}}}},"../../../../vega-wordcloud/src/Wordcloud.js":function(t,e,n){"use strict";e.a=l;var r=n("../../../../vega-wordcloud/src/CloudLayout.js"),i=n("../../../../vega-dataflow/index.js"),o=n("../../../../vega-util/index.js"),s=n("../../../../vega-scale/index.js"),a=n("../../../../vega-statistics/index.js"),c=["x","y","font","fontSize","fontStyle","fontWeight","angle"],u=["text","font","rotate","fontSize","fontStyle","fontWeight"];function l(t){i.d.call(this,Object(r.a)(),t)}l.Definition={type:"Wordcloud",metadata:{modifies:!0},params:[{name:"size",type:"number",array:!0,length:2},{name:"font",type:"string",expr:!0,default:"sans-serif"},{name:"fontStyle",type:"string",expr:!0,default:"normal"},{name:"fontWeight",type:"string",expr:!0,default:"normal"},{name:"fontSize",type:"number",expr:!0,default:14},{name:"fontSizeRange",type:"number",array:"nullable",default:[10,50]},{name:"rotate",type:"number",expr:!0,default:0},{name:"text",type:"field"},{name:"spiral",type:"string",values:["archimedean","rectangular"]},{name:"padding",type:"number",expr:!0},{name:"as",type:"string",array:!0,length:7,default:c}]},Object(o.s)(l,i.d).transform=function(t,e){var n=t.modified();if(n||e.changed(e.ADD_REM)||u.some(function(n){var r=t[n];return Object(o.w)(r)&&e.modified(r.fields)})){var r,i=e.materialize(e.SOURCE).source,l=this.value,p=t.as||c,d=t.fontSize||14;if(Object(o.w)(d)?r=t.fontSizeRange:d=Object(o.i)(d),r){var h=d,f=Object(s.d)("sqrt")().domain(function(t,e){for(var n,r=1/0,i=-1/0,o=0,s=e.length;o<s;++o)(n=t(e[o]))<r&&(r=n),n>i&&(i=n);return[r,i]}(h,i)).range(r);d=function(t){return f(h(t))}}i.forEach(function(t){t[p[0]]=NaN,t[p[1]]=NaN,t[p[3]]=0});for(var m,g,v=l.words(i).text(t.text).size(t.size||[500,500]).padding(t.padding||1).spiral(t.spiral||"archimedean").rotate(t.rotate||0).font(t.font||"sans-serif").fontStyle(t.fontStyle||"normal").fontWeight(t.fontWeight||"normal").fontSize(d).random(a.d).layout(),y=l.size(),b=y[0]>>1,_=y[1]>>1,j=0,x=v.length;j<x;++j)(g=(m=v[j]).datum)[p[0]]=m.x+b,g[p[1]]=m.y+_,g[p[2]]=m.font,g[p[3]]=m.size,g[p[4]]=m.style,g[p[5]]=m.weight,g[p[6]]=m.rotate;return e.reflow(n).modifies(p)}}},"../../../../webpack/buildin/global.js":function(t,e){var n;n=function(){return this}();try{n=n||Function("return this")()||(0,eval)("this")}catch(t){"object"==typeof window&&(n=window)}t.exports=n},"../../../animations/esm5/animations.js":function(t,e,n){"use strict";n.d(e,"b",function(){return r}),n.d(e,"c",function(){return i}),n.d(e,"a",function(){return o}),n.d(e,"e",function(){return a}),n.d(e,"f",function(){return p}),n.d(e,"g",function(){return c}),n.d(e,"h",function(){return l}),n.d(e,"i",function(){return u}),n.d(e,"j",function(){return d}),n.d(e,"k",function(){return s}),n.d(e,"d",function(){return f}),n.d(e,"l",function(){return m}),n.d(e,"m",function(){return g});var r=function(){return function(){}}(),i=function(){return function(){}}(),o="*";function s(t,e){return{type:7,name:t,definitions:e,options:{}}}function a(t,e){return void 0===e&&(e=null),{type:4,styles:e,timings:t}}function c(t,e){return void 0===e&&(e=null),{type:2,steps:t,options:e}}function u(t){return{type:6,styles:t,offset:null}}function l(t,e,n){return{type:0,name:t,styles:e,options:n}}function p(t){return{type:5,steps:t}}function d(t,e,n){return void 0===n&&(n=null),{type:1,expr:t,animation:e,options:n}}function h(t){Promise.resolve(null).then(t)}var f=function(){function t(){this._onDoneFns=[],this._onStartFns=[],this._onDestroyFns=[],this._started=!1,this._destroyed=!1,this._finished=!1,this.parentPlayer=null,this.totalTime=0}return t.prototype._onFinish=function(){this._finished||(this._finished=!0,this._onDoneFns.forEach(function(t){return t()}),this._onDoneFns=[])},t.prototype.onStart=function(t){this._onStartFns.push(t)},t.prototype.onDone=function(t){this._onDoneFns.push(t)},t.prototype.onDestroy=function(t){this._onDestroyFns.push(t)},t.prototype.hasStarted=function(){return this._started},t.prototype.init=function(){},t.prototype.play=function(){this.hasStarted()||(this._onStart(),this.triggerMicrotask()),this._started=!0},t.prototype.triggerMicrotask=function(){var t=this;h(function(){return t._onFinish()})},t.prototype._onStart=function(){this._onStartFns.forEach(function(t){return t()}),this._onStartFns=[]},t.prototype.pause=function(){},t.prototype.restart=function(){},t.prototype.finish=function(){this._onFinish()},t.prototype.destroy=function(){this._destroyed||(this._destroyed=!0,this.hasStarted()||this._onStart(),this.finish(),this._onDestroyFns.forEach(function(t){return t()}),this._onDestroyFns=[])},t.prototype.reset=function(){},t.prototype.setPosition=function(t){},t.prototype.getPosition=function(){return 0},t.prototype.triggerCallback=function(t){var e="start"==t?this._onStartFns:this._onDoneFns;e.forEach(function(t){return t()}),e.length=0},t}(),m=function(){function t(t){var e=this;this._onDoneFns=[],this._onStartFns=[],this._finished=!1,this._started=!1,this._destroyed=!1,this._onDestroyFns=[],this.parentPlayer=null,this.totalTime=0,this.players=t;var n=0,r=0,i=0,o=this.players.length;0==o?h(function(){return e._onFinish()}):this.players.forEach(function(t){t.onDone(function(){++n>=o&&e._onFinish()}),t.onDestroy(function(){++r>=o&&e._onDestroy()}),t.onStart(function(){++i>=o&&e._onStart()})}),this.totalTime=this.players.reduce(function(t,e){return Math.max(t,e.totalTime)},0)}return t.prototype._onFinish=function(){this._finished||(this._finished=!0,this._onDoneFns.forEach(function(t){return t()}),this._onDoneFns=[])},t.prototype.init=function(){this.players.forEach(function(t){return t.init()})},t.prototype.onStart=function(t){this._onStartFns.push(t)},t.prototype._onStart=function(){this.hasStarted()||(this._onStartFns.forEach(function(t){return t()}),this._onStartFns=[],this._started=!0)},t.prototype.onDone=function(t){this._onDoneFns.push(t)},t.prototype.onDestroy=function(t){this._onDestroyFns.push(t)},t.prototype.hasStarted=function(){return this._started},t.prototype.play=function(){this.parentPlayer||this.init(),this._onStart(),this.players.forEach(function(t){return t.play()})},t.prototype.pause=function(){this.players.forEach(function(t){return t.pause()})},t.prototype.restart=function(){this.players.forEach(function(t){return t.restart()})},t.prototype.finish=function(){this._onFinish(),this.players.forEach(function(t){return t.finish()})},t.prototype.destroy=function(){this._onDestroy()},t.prototype._onDestroy=function(){this._destroyed||(this._destroyed=!0,this._onFinish(),this.players.forEach(function(t){return t.destroy()}),this._onDestroyFns.forEach(function(t){return t()}),this._onDestroyFns=[])},t.prototype.reset=function(){this.players.forEach(function(t){return t.reset()}),this._destroyed=!1,this._finished=!1,this._started=!1},t.prototype.setPosition=function(t){var e=t*this.totalTime;this.players.forEach(function(t){var n=t.totalTime?Math.min(1,e/t.totalTime):1;t.setPosition(n)})},t.prototype.getPosition=function(){var t=0;return this.players.forEach(function(e){var n=e.getPosition();t=Math.min(n,t)}),t},t.prototype.beforeDestroy=function(){this.players.forEach(function(t){t.beforeDestroy&&t.beforeDestroy()})},t.prototype.triggerCallback=function(t){var e="start"==t?this._onStartFns:this._onDoneFns;e.forEach(function(t){return t()}),e.length=0},t}(),g="!"},"../../../animations/esm5/browser.js":function(t,e,n){"use strict";n.d(e,"a",function(){return O}),n.d(e,"c",function(){return vt}),n.d(e,"f",function(){return yt}),n.d(e,"d",function(){return w}),n.d(e,"b",function(){return Kt}),n.d(e,"e",function(){return Zt}),n.d(e,"g",function(){return Jt});var r=n("../../../animations/esm5/animations.js"),i=n("../../../../tslib/tslib.es6.js");function o(t){switch(t.length){case 0:return new r.d;case 1:return t[0];default:return new r.l(t)}}function s(t,e,n,i,o,s){void 0===o&&(o={}),void 0===s&&(s={});var a=[],c=[],u=-1,l=null;if(i.forEach(function(t){var n=t.offset,i=n==u,p=i&&l||{};Object.keys(t).forEach(function(n){var i=n,c=t[n];if("offset"!==n)switch(i=e.normalizePropertyName(i,a),c){case r.m:c=o[n];break;case r.a:c=s[n];break;default:c=e.normalizeStyleValue(n,i,c,a)}p[i]=c}),i||c.push(p),l=p,u=n}),a.length){throw new Error("Unable to animate due to the following errors:\n - "+a.join("\n - "))}return c}function a(t,e,n,r){switch(e){case"start":t.onStart(function(){return r(n&&c(n,"start",t.totalTime))});break;case"done":t.onDone(function(){return r(n&&c(n,"done",t.totalTime))});break;case"destroy":t.onDestroy(function(){return r(n&&c(n,"destroy",t.totalTime))})}}function c(t,e,n){var r=u(t.element,t.triggerName,t.fromState,t.toState,e||t.phaseName,void 0==n?t.totalTime:n),i=t._data;return null!=i&&(r._data=i),r}function u(t,e,n,r,i,o){return void 0===i&&(i=""),void 0===o&&(o=0),{element:t,triggerName:e,fromState:n,toState:r,phaseName:i,totalTime:o}}function l(t,e,n){var r;return t instanceof Map?(r=t.get(e))||t.set(e,r=n):(r=t[e])||(r=t[e]=n),r}function p(t){var e=t.indexOf(":");return[t.substring(1,e),t.substr(e+1)]}var d=function(t,e){return!1},h=function(t,e){return!1},f=function(t,e,n){return[]};if("undefined"!=typeof Element){if(d=function(t,e){return t.contains(e)},Element.prototype.matches)h=function(t,e){return t.matches(e)};else{var m=Element.prototype,g=m.matchesSelector||m.mozMatchesSelector||m.msMatchesSelector||m.oMatchesSelector||m.webkitMatchesSelector;g&&(h=function(t,e){return g.apply(t,[e])})}f=function(t,e,n){var r=[];if(n)r.push.apply(r,t.querySelectorAll(e));else{var i=t.querySelector(e);i&&r.push(i)}return r}}var v=null;function y(t){return v||(v=b()||{}),!v.style||t in v.style}function b(){return"undefined"!=typeof document?document.body:null}var _=h,j=d,x=f,w=function(){function t(){}return t.prototype.validateStyleProperty=function(t){return y(t)},t.prototype.matchesElement=function(t,e){return _(t,e)},t.prototype.containsElement=function(t,e){return j(t,e)},t.prototype.query=function(t,e,n){return x(t,e,n)},t.prototype.computeStyle=function(t,e,n){return n||""},t.prototype.animate=function(t,e,n,i,o,s){return void 0===s&&(s=[]),new r.d},t}(),O=function(){function t(){}return t.NOOP=new w,t}(),C=1e3,k="ng-trigger",E=".ng-trigger",S=".ng-animating";function P(t){if("number"==typeof t)return t;var e=t.match(/^(-?[\.\d]+)(m?s)/);return!e||e.length<2?0:M(parseFloat(e[1]),e[2])}function M(t,e){switch(e){case"s":return t*C;default:return t}}function A(t,e,n){return t.hasOwnProperty("duration")?t:function(t,e,n){var r,i=0,o="";if("string"==typeof t){var s=t.match(/^(-?[\.\d]+)(m?s)(?:\s+(-?[\.\d]+)(m?s))?(?:\s+([-a-z]+(?:\(.+?\))?))?$/i);if(null===s)return e.push('The provided timing value "'+t+'" is invalid.'),{duration:0,delay:0,easing:""};r=M(parseFloat(s[1]),s[2]);var a=s[3];null!=a&&(i=M(Math.floor(parseFloat(a)),s[4]));var c=s[5];c&&(o=c)}else r=t;if(!n){var u=!1,l=e.length;r<0&&(e.push("Duration values below 0 are not allowed for this animation step."),u=!0),i<0&&(e.push("Delay values below 0 are not allowed for this animation step."),u=!0),u&&e.splice(l,0,'The provided timing value "'+t+'" is invalid.')}return{duration:r,delay:i,easing:o}}(t,e,n)}function T(t,e){return void 0===e&&(e={}),Object.keys(t).forEach(function(n){e[n]=t[n]}),e}function D(t){var e={};return Array.isArray(t)?t.forEach(function(t){return I(t,!1,e)}):I(t,!1,e),e}function I(t,e,n){if(void 0===n&&(n={}),e)for(var r in t)n[r]=t[r];else T(t,n);return n}function R(t,e){t.style&&Object.keys(e).forEach(function(n){var r=H(n);t.style[r]=e[n]})}function N(t,e){t.style&&Object.keys(e).forEach(function(e){var n=H(e);t.style[n]=""})}function L(t){return Array.isArray(t)?1==t.length?t[0]:Object(r.g)(t):t}var F=new RegExp("{{\\s*(.+?)\\s*}}","g");function V(t){var e=[];if("string"==typeof t){for(var n=t.toString(),r=void 0;r=F.exec(n);)e.push(r[1]);F.lastIndex=0}return e}function z(t,e,n){var r=t.toString(),i=r.replace(F,function(t,r){var i=e[r];return e.hasOwnProperty(r)||(n.push("Please provide a value for the animation param "+r),i=""),i.toString()});return i==r?t:i}function B(t){for(var e=[],n=t.next();!n.done;)e.push(n.value),n=t.next();return e}var q=/-+([a-z0-9])/g;function H(t){return t.replace(q,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return t[1].toUpperCase()})}function U(t,e,n){switch(e.type){case 7:return t.visitTrigger(e,n);case 0:return t.visitState(e,n);case 1:return t.visitTransition(e,n);case 2:return t.visitSequence(e,n);case 3:return t.visitGroup(e,n);case 4:return t.visitAnimate(e,n);case 5:return t.visitKeyframes(e,n);case 6:return t.visitStyle(e,n);case 8:return t.visitReference(e,n);case 9:return t.visitAnimateChild(e,n);case 10:return t.visitAnimateRef(e,n);case 11:return t.visitQuery(e,n);case 12:return t.visitStagger(e,n);default:throw new Error("Unable to resolve animation metadata node #"+e.type)}}var W="*";function G(t,e){var n=[];return"string"==typeof t?t.split(/\s*,\s*/).forEach(function(t){return function(t,e,n){if(":"==t[0]){var r=function(t,e){switch(t){case":enter":return"void => *";case":leave":return"* => void";case":increment":return function(t,e){return parseFloat(e)>parseFloat(t)};case":decrement":return function(t,e){return parseFloat(e)<parseFloat(t)};default:return e.push('The transition alias value "'+t+'" is not supported'),"* => *"}}(t,n);if("function"==typeof r)return void e.push(r);t=r}var i=t.match(/^(\*|[-\w]+)\s*(<?[=-]>)\s*(\*|[-\w]+)$/);if(null==i||i.length<4)return n.push('The provided transition expression "'+t+'" is not supported'),e;var o=i[1],s=i[2],a=i[3];e.push(K(o,a));var c=o==W&&a==W;"<"!=s[0]||c||e.push(K(a,o))}(t,n,e)}):n.push(t),n}var X=new Set;X.add("true"),X.add("1");var Y=new Set;function K(t,e){var n=X.has(t)||Y.has(t),r=X.has(e)||Y.has(e);return function(i,o){var s=t==W||t==i,a=e==W||e==o;return!s&&n&&"boolean"==typeof i&&(s=i?X.has(t):Y.has(t)),!a&&r&&"boolean"==typeof o&&(a=o?X.has(e):Y.has(e)),s&&a}}Y.add("false"),Y.add("0");var $=":self",Q=new RegExp("s*"+$+"s*,?","g");function Z(t,e,n){return new J(t).build(e,n)}var J=function(){function t(t){this._driver=t}return t.prototype.build=function(t,e){var n=new tt(e);return this._resetContextStyleTimingState(n),U(this,L(t),n)},t.prototype._resetContextStyleTimingState=function(t){t.currentQuerySelector="",t.collectedStyles={},t.collectedStyles[""]={},t.currentTime=0},t.prototype.visitTrigger=function(t,e){var n=this,r=e.queryCount=0,i=e.depCount=0,o=[],s=[];return"@"==t.name.charAt(0)&&e.errors.push("animation triggers cannot be prefixed with an `@` sign (e.g. trigger('@foo', [...]))"),t.definitions.forEach(function(t){if(n._resetContextStyleTimingState(e),0==t.type){var a=t,c=a.name;c.split(/\s*,\s*/).forEach(function(t){a.name=t,o.push(n.visitState(a,e))}),a.name=c}else if(1==t.type){var u=n.visitTransition(t,e);r+=u.queryCount,i+=u.depCount,s.push(u)}else e.errors.push("only state() and transition() definitions can sit inside of a trigger()")}),{type:7,name:t.name,states:o,transitions:s,queryCount:r,depCount:i,options:null}},t.prototype.visitState=function(t,e){var n=this.visitStyle(t.styles,e),r=t.options&&t.options.params||null;if(n.containsDynamicStyles){var i=new Set,o=r||{};if(n.styles.forEach(function(t){if(et(t)){var e=t;Object.keys(e).forEach(function(t){V(e[t]).forEach(function(t){o.hasOwnProperty(t)||i.add(t)})})}}),i.size){var s=B(i.values());e.errors.push('state("'+t.name+'", ...) must define default values for all the following style substitutions: '+s.join(", "))}}return{type:0,name:t.name,style:n,options:r?{params:r}:null}},t.prototype.visitTransition=function(t,e){e.queryCount=0,e.depCount=0;var n=U(this,L(t.animation),e);return{type:1,matchers:G(t.expr,e.errors),animation:n,queryCount:e.queryCount,depCount:e.depCount,options:nt(t.options)}},t.prototype.visitSequence=function(t,e){var n=this;return{type:2,steps:t.steps.map(function(t){return U(n,t,e)}),options:nt(t.options)}},t.prototype.visitGroup=function(t,e){var n=this,r=e.currentTime,i=0,o=t.steps.map(function(t){e.currentTime=r;var o=U(n,t,e);return i=Math.max(i,e.currentTime),o});return e.currentTime=i,{type:3,steps:o,options:nt(t.options)}},t.prototype.visitAnimate=function(t,e){var n,i=function(t,e){var n=null;if(t.hasOwnProperty("duration"))n=t;else if("number"==typeof t){var r=A(t,e).duration;return rt(r,0,"")}var i=t;if(i.split(/\s+/).some(function(t){return"{"==t.charAt(0)&&"{"==t.charAt(1)})){var o=rt(0,0,"");return o.dynamic=!0,o.strValue=i,o}return rt((n=n||A(i,e)).duration,n.delay,n.easing)}(t.timings,e.errors);e.currentAnimateTimings=i;var o=t.styles?t.styles:Object(r.i)({});if(5==o.type)n=this.visitKeyframes(o,e);else{var s=t.styles,a=!1;if(!s){a=!0;var c={};i.easing&&(c.easing=i.easing),s=Object(r.i)(c)}e.currentTime+=i.duration+i.delay;var u=this.visitStyle(s,e);u.isEmptyStep=a,n=u}return e.currentAnimateTimings=null,{type:4,timings:i,style:n,options:null}},t.prototype.visitStyle=function(t,e){var n=this._makeStyleAst(t,e);return this._validateStyleAst(n,e),n},t.prototype._makeStyleAst=function(t,e){var n=[];Array.isArray(t.styles)?t.styles.forEach(function(t){"string"==typeof t?t==r.a?n.push(t):e.errors.push("The provided style string value "+t+" is not allowed."):n.push(t)}):n.push(t.styles);var i=!1,o=null;return n.forEach(function(t){if(et(t)){var e=t,n=e.easing;if(n&&(o=n,delete e.easing),!i)for(var r in e){if(e[r].toString().indexOf("{{")>=0){i=!0;break}}}}),{type:6,styles:n,easing:o,offset:t.offset,containsDynamicStyles:i,options:null}},t.prototype._validateStyleAst=function(t,e){var n=this,r=e.currentAnimateTimings,i=e.currentTime,o=e.currentTime;r&&o>0&&(o-=r.duration+r.delay),t.styles.forEach(function(t){"string"!=typeof t&&Object.keys(t).forEach(function(r){if(n._driver.validateStyleProperty(r)){var s,a,c,u,l,p=e.collectedStyles[e.currentQuerySelector],d=p[r],h=!0;d&&(o!=i&&o>=d.startTime&&i<=d.endTime&&(e.errors.push('The CSS property "'+r+'" that exists between the times of "'+d.startTime+'ms" and "'+d.endTime+'ms" is also being animated in a parallel animation between the times of "'+o+'ms" and "'+i+'ms"'),h=!1),o=d.startTime),h&&(p[r]={startTime:o,endTime:i}),e.options&&(s=t[r],a=e.options,c=e.errors,u=a.params||{},(l=V(s)).length&&l.forEach(function(t){u.hasOwnProperty(t)||c.push("Unable to resolve the local animation param "+t+" in the given list of values")}))}else e.errors.push('The provided animation property "'+r+'" is not a supported CSS property for animations')})})},t.prototype.visitKeyframes=function(t,e){var n=this,r={type:5,styles:[],options:null};if(!e.currentAnimateTimings)return e.errors.push("keyframes() must be placed inside of a call to animate()"),r;var i=0,o=[],s=!1,a=!1,c=0,u=t.steps.map(function(t){var r=n._makeStyleAst(t,e),u=null!=r.offset?r.offset:function(t){if("string"==typeof t)return null;var e=null;if(Array.isArray(t))t.forEach(function(t){if(et(t)&&t.hasOwnProperty("offset")){var n=t;e=parseFloat(n.offset),delete n.offset}});else if(et(t)&&t.hasOwnProperty("offset")){var n=t;e=parseFloat(n.offset),delete n.offset}return e}(r.styles),l=0;return null!=u&&(i++,l=r.offset=u),a=a||l<0||l>1,s=s||l<c,c=l,o.push(l),r});a&&e.errors.push("Please ensure that all keyframe offsets are between 0 and 1"),s&&e.errors.push("Please ensure that all keyframe offsets are in order");var l=t.steps.length,p=0;i>0&&i<l?e.errors.push("Not all style() steps within the declared keyframes() contain offsets"):0==i&&(p=1/(l-1));var d=l-1,h=e.currentTime,f=e.currentAnimateTimings,m=f.duration;return u.forEach(function(t,i){var s=p>0?i==d?1:p*i:o[i],a=s*m;e.currentTime=h+f.delay+a,f.duration=a,n._validateStyleAst(t,e),t.offset=s,r.styles.push(t)}),r},t.prototype.visitReference=function(t,e){return{type:8,animation:U(this,L(t.animation),e),options:nt(t.options)}},t.prototype.visitAnimateChild=function(t,e){return e.depCount++,{type:9,options:nt(t.options)}},t.prototype.visitAnimateRef=function(t,e){return{type:10,animation:this.visitReference(t.animation,e),options:nt(t.options)}},t.prototype.visitQuery=function(t,e){var n=e.currentQuerySelector,r=t.options||{};e.queryCount++,e.currentQuery=t;var i=function(t){var e=!!t.split(/\s*,\s*/).find(function(t){return t==$});e&&(t=t.replace(Q,""));return[t=t.replace(/@\*/g,E).replace(/@\w+/g,function(t){return E+"-"+t.substr(1)}).replace(/:animating/g,S),e]}(t.selector),o=i[0],s=i[1];e.currentQuerySelector=n.length?n+" "+o:o,l(e.collectedStyles,e.currentQuerySelector,{});var a=U(this,L(t.animation),e);return e.currentQuery=null,e.currentQuerySelector=n,{type:11,selector:o,limit:r.limit||0,optional:!!r.optional,includeSelf:s,animation:a,originalSelector:t.selector,options:nt(t.options)}},t.prototype.visitStagger=function(t,e){e.currentQuery||e.errors.push("stagger() can only be used inside of query()");var n="full"===t.timings?{duration:0,delay:0,easing:"full"}:A(t.timings,e.errors,!0);return{type:12,animation:U(this,L(t.animation),e),timings:n,options:null}},t}();var tt=function(){return function(t){this.errors=t,this.queryCount=0,this.depCount=0,this.currentTransition=null,this.currentQuery=null,this.currentQuerySelector=null,this.currentAnimateTimings=null,this.currentTime=0,this.collectedStyles={},this.options=null}}();function et(t){return!Array.isArray(t)&&"object"==typeof t}function nt(t){var e;return t?(t=T(t)).params&&(t.params=(e=t.params)?T(e):null):t={},t}function rt(t,e,n){return{duration:t,delay:e,easing:n}}function it(t,e,n,r,i,o,s,a){return void 0===s&&(s=null),void 0===a&&(a=!1),{type:1,element:t,keyframes:e,preStyleProps:n,postStyleProps:r,duration:i,delay:o,totalTime:i+o,easing:s,subTimeline:a}}var ot=function(){function t(){this._map=new Map}return t.prototype.consume=function(t){var e=this._map.get(t);return e?this._map.delete(t):e=[],e},t.prototype.append=function(t,e){var n=this._map.get(t);n||this._map.set(t,n=[]),n.push.apply(n,e)},t.prototype.has=function(t){return this._map.has(t)},t.prototype.clear=function(){this._map.clear()},t}(),st=new RegExp(":enter","g"),at=new RegExp(":leave","g");function ct(t,e,n,r,i,o,s,a,c,u){return void 0===o&&(o={}),void 0===s&&(s={}),void 0===u&&(u=[]),(new ut).buildKeyframes(t,e,n,r,i,o,s,a,c,u)}var ut=function(){function t(){}return t.prototype.buildKeyframes=function(t,e,n,r,i,o,s,a,c,u){void 0===u&&(u=[]),c=c||new ot;var l=new pt(t,e,c,r,i,u,[]);l.options=a,l.currentTimeline.setStyles([o],null,l.errors,a),U(this,n,l);var p=l.timelines.filter(function(t){return t.containsAnimation()});if(p.length&&Object.keys(s).length){var d=p[p.length-1];d.allowOnlyTimelineStyles()||d.setStyles([s],null,l.errors,a)}return p.length?p.map(function(t){return t.buildKeyframes()}):[it(e,[],[],[],0,0,"",!1)]},t.prototype.visitTrigger=function(t,e){},t.prototype.visitState=function(t,e){},t.prototype.visitTransition=function(t,e){},t.prototype.visitAnimateChild=function(t,e){var n=e.subInstructions.consume(e.element);if(n){var r=e.createSubContext(t.options),i=e.currentTimeline.currentTime,o=this._visitSubInstructions(n,r,r.options);i!=o&&e.transformIntoNewTimeline(o)}e.previousNode=t},t.prototype.visitAnimateRef=function(t,e){var n=e.createSubContext(t.options);n.transformIntoNewTimeline(),this.visitReference(t.animation,n),e.transformIntoNewTimeline(n.currentTimeline.currentTime),e.previousNode=t},t.prototype._visitSubInstructions=function(t,e,n){var r=e.currentTimeline.currentTime,i=null!=n.duration?P(n.duration):null,o=null!=n.delay?P(n.delay):null;return 0!==i&&t.forEach(function(t){var n=e.appendInstructionToTimeline(t,i,o);r=Math.max(r,n.duration+n.delay)}),r},t.prototype.visitReference=function(t,e){e.updateOptions(t.options,!0),U(this,t.animation,e),e.previousNode=t},t.prototype.visitSequence=function(t,e){var n=this,r=e.subContextCount,i=e,o=t.options;if(o&&(o.params||o.delay)&&((i=e.createSubContext(o)).transformIntoNewTimeline(),null!=o.delay)){6==i.previousNode.type&&(i.currentTimeline.snapshotCurrentStyles(),i.previousNode=lt);var s=P(o.delay);i.delayNextStep(s)}t.steps.length&&(t.steps.forEach(function(t){return U(n,t,i)}),i.currentTimeline.applyStylesToKeyframe(),i.subContextCount>r&&i.transformIntoNewTimeline()),e.previousNode=t},t.prototype.visitGroup=function(t,e){var n=this,r=[],i=e.currentTimeline.currentTime,o=t.options&&t.options.delay?P(t.options.delay):0;t.steps.forEach(function(s){var a=e.createSubContext(t.options);o&&a.delayNextStep(o),U(n,s,a),i=Math.max(i,a.currentTimeline.currentTime),r.push(a.currentTimeline)}),r.forEach(function(t){return e.currentTimeline.mergeTimelineCollectedStyles(t)}),e.transformIntoNewTimeline(i),e.previousNode=t},t.prototype._visitTiming=function(t,e){if(t.dynamic){var n=t.strValue;return A(e.params?z(n,e.params,e.errors):n,e.errors)}return{duration:t.duration,delay:t.delay,easing:t.easing}},t.prototype.visitAnimate=function(t,e){var n=e.currentAnimateTimings=this._visitTiming(t.timings,e),r=e.currentTimeline;n.delay&&(e.incrementTime(n.delay),r.snapshotCurrentStyles());var i=t.style;5==i.type?this.visitKeyframes(i,e):(e.incrementTime(n.duration),this.visitStyle(i,e),r.applyStylesToKeyframe()),e.currentAnimateTimings=null,e.previousNode=t},t.prototype.visitStyle=function(t,e){var n=e.currentTimeline,r=e.currentAnimateTimings;!r&&n.getCurrentStyleProperties().length&&n.forwardFrame();var i=r&&r.easing||t.easing;t.isEmptyStep?n.applyEmptyStep(i):n.setStyles(t.styles,i,e.errors,e.options),e.previousNode=t},t.prototype.visitKeyframes=function(t,e){var n=e.currentAnimateTimings,r=e.currentTimeline.duration,i=n.duration,o=e.createSubContext().currentTimeline;o.easing=n.easing,t.styles.forEach(function(t){var n=t.offset||0;o.forwardTime(n*i),o.setStyles(t.styles,t.easing,e.errors,e.options),o.applyStylesToKeyframe()}),e.currentTimeline.mergeTimelineCollectedStyles(o),e.transformIntoNewTimeline(r+i),e.previousNode=t},t.prototype.visitQuery=function(t,e){var n=this,r=e.currentTimeline.currentTime,i=t.options||{},o=i.delay?P(i.delay):0;o&&(6===e.previousNode.type||0==r&&e.currentTimeline.getCurrentStyleProperties().length)&&(e.currentTimeline.snapshotCurrentStyles(),e.previousNode=lt);var s=r,a=e.invokeQuery(t.selector,t.originalSelector,t.limit,t.includeSelf,!!i.optional,e.errors);e.currentQueryTotal=a.length;var c=null;a.forEach(function(r,i){e.currentQueryIndex=i;var a=e.createSubContext(t.options,r);o&&a.delayNextStep(o),r===e.element&&(c=a.currentTimeline),U(n,t.animation,a),a.currentTimeline.applyStylesToKeyframe();var u=a.currentTimeline.currentTime;s=Math.max(s,u)}),e.currentQueryIndex=0,e.currentQueryTotal=0,e.transformIntoNewTimeline(s),c&&(e.currentTimeline.mergeTimelineCollectedStyles(c),e.currentTimeline.snapshotCurrentStyles()),e.previousNode=t},t.prototype.visitStagger=function(t,e){var n=e.parentContext,r=e.currentTimeline,i=t.timings,o=Math.abs(i.duration),s=o*(e.currentQueryTotal-1),a=o*e.currentQueryIndex;switch(i.duration<0?"reverse":i.easing){case"reverse":a=s-a;break;case"full":a=n.currentStaggerTime}var c=e.currentTimeline;a&&c.delayNextStep(a);var u=c.currentTime;U(this,t.animation,e),e.previousNode=t,n.currentStaggerTime=r.currentTime-u+(r.startTime-n.currentTimeline.startTime)},t}(),lt={},pt=function(){function t(t,e,n,r,i,o,s,a){this._driver=t,this.element=e,this.subInstructions=n,this._enterClassName=r,this._leaveClassName=i,this.errors=o,this.timelines=s,this.parentContext=null,this.currentAnimateTimings=null,this.previousNode=lt,this.subContextCount=0,this.options={},this.currentQueryIndex=0,this.currentQueryTotal=0,this.currentStaggerTime=0,this.currentTimeline=a||new dt(this._driver,e,0),s.push(this.currentTimeline)}return Object.defineProperty(t.prototype,"params",{get:function(){return this.options.params},enumerable:!0,configurable:!0}),t.prototype.updateOptions=function(t,e){var n=this;if(t){var r=t,i=this.options;null!=r.duration&&(i.duration=P(r.duration)),null!=r.delay&&(i.delay=P(r.delay));var o=r.params;if(o){var s=i.params;s||(s=this.options.params={}),Object.keys(o).forEach(function(t){e&&s.hasOwnProperty(t)||(s[t]=z(o[t],s,n.errors))})}}},t.prototype._copyOptions=function(){var t={};if(this.options){var e=this.options.params;if(e){var n=t.params={};Object.keys(e).forEach(function(t){n[t]=e[t]})}}return t},t.prototype.createSubContext=function(e,n,r){void 0===e&&(e=null);var i=n||this.element,o=new t(this._driver,i,this.subInstructions,this._enterClassName,this._leaveClassName,this.errors,this.timelines,this.currentTimeline.fork(i,r||0));return o.previousNode=this.previousNode,o.currentAnimateTimings=this.currentAnimateTimings,o.options=this._copyOptions(),o.updateOptions(e),o.currentQueryIndex=this.currentQueryIndex,o.currentQueryTotal=this.currentQueryTotal,o.parentContext=this,this.subContextCount++,o},t.prototype.transformIntoNewTimeline=function(t){return this.previousNode=lt,this.currentTimeline=this.currentTimeline.fork(this.element,t),this.timelines.push(this.currentTimeline),this.currentTimeline},t.prototype.appendInstructionToTimeline=function(t,e,n){var r={duration:null!=e?e:t.duration,delay:this.currentTimeline.currentTime+(null!=n?n:0)+t.delay,easing:""},i=new ht(this._driver,t.element,t.keyframes,t.preStyleProps,t.postStyleProps,r,t.stretchStartingKeyframe);return this.timelines.push(i),r},t.prototype.incrementTime=function(t){this.currentTimeline.forwardTime(this.currentTimeline.duration+t)},t.prototype.delayNextStep=function(t){t>0&&this.currentTimeline.delayNextStep(t)},t.prototype.invokeQuery=function(t,e,n,r,i,o){var s=[];if(r&&s.push(this.element),t.length>0){t=(t=t.replace(st,"."+this._enterClassName)).replace(at,"."+this._leaveClassName);var a=1!=n,c=this._driver.query(this.element,t,a);0!==n&&(c=n<0?c.slice(c.length+n,c.length):c.slice(0,n)),s.push.apply(s,c)}return i||0!=s.length||o.push('`query("'+e+'")` returned zero elements. (Use `query("'+e+'", { optional: true })` if you wish to allow this.)'),s},t}(),dt=function(){function t(t,e,n,r){this._driver=t,this.element=e,this.startTime=n,this._elementTimelineStylesLookup=r,this.duration=0,this._previousKeyframe={},this._currentKeyframe={},this._keyframes=new Map,this._styleSummary={},this._pendingStyles={},this._backFill={},this._currentEmptyStepKeyframe=null,this._elementTimelineStylesLookup||(this._elementTimelineStylesLookup=new Map),this._localTimelineStyles=Object.create(this._backFill,{}),this._globalTimelineStyles=this._elementTimelineStylesLookup.get(e),this._globalTimelineStyles||(this._globalTimelineStyles=this._localTimelineStyles,this._elementTimelineStylesLookup.set(e,this._localTimelineStyles)),this._loadKeyframe()}return t.prototype.containsAnimation=function(){switch(this._keyframes.size){case 0:return!1;case 1:return this.getCurrentStyleProperties().length>0;default:return!0}},t.prototype.getCurrentStyleProperties=function(){return Object.keys(this._currentKeyframe)},Object.defineProperty(t.prototype,"currentTime",{get:function(){return this.startTime+this.duration},enumerable:!0,configurable:!0}),t.prototype.delayNextStep=function(t){var e=1==this._keyframes.size&&Object.keys(this._pendingStyles).length;this.duration||e?(this.forwardTime(this.currentTime+t),e&&this.snapshotCurrentStyles()):this.startTime+=t},t.prototype.fork=function(e,n){return this.applyStylesToKeyframe(),new t(this._driver,e,n||this.currentTime,this._elementTimelineStylesLookup)},t.prototype._loadKeyframe=function(){this._currentKeyframe&&(this._previousKeyframe=this._currentKeyframe),this._currentKeyframe=this._keyframes.get(this.duration),this._currentKeyframe||(this._currentKeyframe=Object.create(this._backFill,{}),this._keyframes.set(this.duration,this._currentKeyframe))},t.prototype.forwardFrame=function(){this.duration+=1,this._loadKeyframe()},t.prototype.forwardTime=function(t){this.applyStylesToKeyframe(),this.duration=t,this._loadKeyframe()},t.prototype._updateStyle=function(t,e){this._localTimelineStyles[t]=e,this._globalTimelineStyles[t]=e,this._styleSummary[t]={time:this.currentTime,value:e}},t.prototype.allowOnlyTimelineStyles=function(){return this._currentEmptyStepKeyframe!==this._currentKeyframe},t.prototype.applyEmptyStep=function(t){var e=this;t&&(this._previousKeyframe.easing=t),Object.keys(this._globalTimelineStyles).forEach(function(t){e._backFill[t]=e._globalTimelineStyles[t]||r.a,e._currentKeyframe[t]=r.a}),this._currentEmptyStepKeyframe=this._currentKeyframe},t.prototype.setStyles=function(t,e,n,i){var o=this;e&&(this._previousKeyframe.easing=e);var s,a,c,u,l=i&&i.params||{},p=(s=t,a=this._globalTimelineStyles,u={},s.forEach(function(t){"*"===t?(c=c||Object.keys(a)).forEach(function(t){u[t]=r.a}):I(t,!1,u)}),u);Object.keys(p).forEach(function(t){var e=z(p[t],l,n);o._pendingStyles[t]=e,o._localTimelineStyles.hasOwnProperty(t)||(o._backFill[t]=o._globalTimelineStyles.hasOwnProperty(t)?o._globalTimelineStyles[t]:r.a),o._updateStyle(t,e)})},t.prototype.applyStylesToKeyframe=function(){var t=this,e=this._pendingStyles,n=Object.keys(e);0!=n.length&&(this._pendingStyles={},n.forEach(function(n){var r=e[n];t._currentKeyframe[n]=r}),Object.keys(this._localTimelineStyles).forEach(function(e){t._currentKeyframe.hasOwnProperty(e)||(t._currentKeyframe[e]=t._localTimelineStyles[e])}))},t.prototype.snapshotCurrentStyles=function(){var t=this;Object.keys(this._localTimelineStyles).forEach(function(e){var n=t._localTimelineStyles[e];t._pendingStyles[e]=n,t._updateStyle(e,n)})},t.prototype.getFinalKeyframe=function(){return this._keyframes.get(this.duration)},Object.defineProperty(t.prototype,"properties",{get:function(){var t=[];for(var e in this._currentKeyframe)t.push(e);return t},enumerable:!0,configurable:!0}),t.prototype.mergeTimelineCollectedStyles=function(t){var e=this;Object.keys(t._styleSummary).forEach(function(n){var r=e._styleSummary[n],i=t._styleSummary[n];(!r||i.time>r.time)&&e._updateStyle(n,i.value)})},t.prototype.buildKeyframes=function(){var t=this;this.applyStylesToKeyframe();var e=new Set,n=new Set,i=1===this._keyframes.size&&0===this.duration,o=[];this._keyframes.forEach(function(s,a){var c=I(s,!0);Object.keys(c).forEach(function(t){var i=c[t];i==r.m?e.add(t):i==r.a&&n.add(t)}),i||(c.offset=a/t.duration),o.push(c)});var s=e.size?B(e.values()):[],a=n.size?B(n.values()):[];if(i){var c=o[0],u=T(c);c.offset=0,u.offset=1,o=[c,u]}return it(this.element,o,s,a,this.duration,this.startTime,this.easing,!1)},t}(),ht=function(t){function e(e,n,r,i,o,s,a){void 0===a&&(a=!1);var c=t.call(this,e,n,s.delay)||this;return c.element=n,c.keyframes=r,c.preStyleProps=i,c.postStyleProps=o,c._stretchStartingKeyframe=a,c.timings={duration:s.duration,delay:s.delay,easing:s.easing},c}return Object(i.b)(e,t),e.prototype.containsAnimation=function(){return this.keyframes.length>1},e.prototype.buildKeyframes=function(){var t=this.keyframes,e=this.timings,n=e.delay,r=e.duration,i=e.easing;if(this._stretchStartingKeyframe&&n){var o=[],s=r+n,a=n/s,c=I(t[0],!1);c.offset=0,o.push(c);var u=I(t[0],!1);u.offset=ft(a),o.push(u);for(var l=t.length-1,p=1;p<=l;p++){var d=I(t[p],!1),h=n+d.offset*r;d.offset=ft(h/s),o.push(d)}r=s,n=0,i="",t=o}return it(this.element,t,this.preStyleProps,this.postStyleProps,r,n,i,!0)},e}(dt);function ft(t,e){void 0===e&&(e=3);var n=Math.pow(10,e-1);return Math.round(t*n)/n}!function(){function t(t,e){this._driver=t;var n=[],r=Z(t,e,n);if(n.length){var i="animation validation failed:\n"+n.join("\n");throw new Error(i)}this._animationAst=r}t.prototype.buildTimelines=function(t,e,n,r,i){var o=Array.isArray(e)?D(e):e,s=Array.isArray(n)?D(n):n,a=[];i=i||new ot;var c=ct(this._driver,t,this._animationAst,"ng-enter","ng-leave",o,s,r,i,a);if(a.length){var u="animation building failed:\n"+a.join("\n");throw new Error(u)}return c}}();var mt,gt,vt=function(){return function(){}}(),yt=(function(){function t(){}t.prototype.normalizePropertyName=function(t,e){return t},t.prototype.normalizeStyleValue=function(t,e,n,r){return n}}(),function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.b)(e,t),e.prototype.normalizePropertyName=function(t,e){return H(t)},e.prototype.normalizeStyleValue=function(t,e,n,r){var i="",o=n.toString().trim();if(bt[e]&&0!==n&&"0"!==n)if("number"==typeof n)i="px";else{var s=n.match(/^[+-]?[\d\.]+([a-z]*)$/);s&&0==s[1].length&&r.push("Please provide a CSS unit value for "+t+":"+n)}return o+i},e}(vt)),bt=(mt="width,height,minWidth,minHeight,maxWidth,maxHeight,left,top,bottom,right,fontSize,outlineWidth,outlineOffset,paddingTop,paddingLeft,paddingBottom,paddingRight,marginTop,marginLeft,marginBottom,marginRight,borderRadius,borderWidth,borderTopWidth,borderLeftWidth,borderRightWidth,borderBottomWidth,textIndent,perspective".split(","),gt={},mt.forEach(function(t){return gt[t]=!0}),gt);function _t(t,e,n,r,i,o,s,a,c,u,l,p){return{type:0,element:t,triggerName:e,isRemovalTransition:i,fromState:n,fromStyles:o,toState:r,toStyles:s,timelines:a,queriedElements:c,preStyleProps:u,postStyleProps:l,errors:p}}var jt={},xt=function(){function t(t,e,n){this._triggerName=t,this.ast=e,this._stateStyles=n}return t.prototype.match=function(t,e){return n=this.ast.matchers,r=t,i=e,n.some(function(t){return t(r,i)});var n,r,i},t.prototype.buildStyles=function(t,e,n){var r=this._stateStyles["*"],i=this._stateStyles[t],o=r?r.buildStyles(e,n):{};return i?i.buildStyles(e,n):o},t.prototype.build=function(t,e,n,r,o,s,a,c,u){var p=[],d=this.ast.options&&this.ast.options.params||jt,h=a&&a.params||jt,f=this.buildStyles(n,h,p),m=c&&c.params||jt,g=this.buildStyles(r,m,p),v=new Set,y=new Map,b=new Map,_="void"===r,j={params:Object(i.a)({},d,m)},x=ct(t,e,this.ast.animation,o,s,f,g,j,u,p);if(p.length)return _t(e,this._triggerName,n,r,_,f,g,[],[],y,b,p);x.forEach(function(t){var n=t.element,r=l(y,n,{});t.preStyleProps.forEach(function(t){return r[t]=!0});var i=l(b,n,{});t.postStyleProps.forEach(function(t){return i[t]=!0}),n!==e&&v.add(n)});var w=B(v.values());return _t(e,this._triggerName,n,r,_,f,g,x,w,y,b)},t}();var wt=function(){function t(t,e){this.styles=t,this.defaultParams=e}return t.prototype.buildStyles=function(t,e){var n={},r=T(this.defaultParams);return Object.keys(t).forEach(function(e){var n=t[e];null!=n&&(r[e]=n)}),this.styles.styles.forEach(function(t){if("string"!=typeof t){var i=t;Object.keys(i).forEach(function(t){var o=i[t];o.length>1&&(o=z(o,r,e)),n[t]=o})}}),n},t}();var Ot=function(){function t(t,e){var n,r,i=this;this.name=t,this.ast=e,this.transitionFactories=[],this.states={},e.states.forEach(function(t){var e=t.options&&t.options.params||{};i.states[t.name]=new wt(t.style,e)}),Ct(this.states,"true","1"),Ct(this.states,"false","0"),e.transitions.forEach(function(e){i.transitionFactories.push(new xt(t,e,i.states))}),this.fallbackTransition=(n=t,r=this.states,new xt(n,{type:1,animation:{type:2,steps:[],options:null},matchers:[function(t,e){return!0}],options:null,queryCount:0,depCount:0},r))}return Object.defineProperty(t.prototype,"containsQueries",{get:function(){return this.ast.queryCount>0},enumerable:!0,configurable:!0}),t.prototype.matchTransition=function(t,e){return this.transitionFactories.find(function(n){return n.match(t,e)})||null},t.prototype.matchStyles=function(t,e,n){return this.fallbackTransition.buildStyles(t,e,n)},t}();function Ct(t,e,n){t.hasOwnProperty(e)?t.hasOwnProperty(n)||(t[n]=t[e]):t.hasOwnProperty(n)&&(t[e]=t[n])}var kt=new ot,Et=function(){function t(t,e){this._driver=t,this._normalizer=e,this._animations={},this._playersById={},this.players=[]}return t.prototype.register=function(t,e){var n=[],r=Z(this._driver,e,n);if(n.length)throw new Error("Unable to build the animation due to the following errors: "+n.join("\n"));this._animations[t]=r},t.prototype._buildPlayer=function(t,e,n){var r=t.element,i=s(this._driver,this._normalizer,0,t.keyframes,e,n);return this._driver.animate(r,i,t.duration,t.delay,t.easing,[])},t.prototype.create=function(t,e,n){var i=this;void 0===n&&(n={});var s,a=[],c=this._animations[t],u=new Map;if(c?(s=ct(this._driver,e,c,"ng-enter","ng-leave",{},{},n,kt,a)).forEach(function(t){var e=l(u,t.element,{});t.postStyleProps.forEach(function(t){return e[t]=null})}):(a.push("The requested animation doesn't exist or has already been destroyed"),s=[]),a.length)throw new Error("Unable to create the animation due to the following errors: "+a.join("\n"));u.forEach(function(t,e){Object.keys(t).forEach(function(n){t[n]=i._driver.computeStyle(e,n,r.a)})});var p=o(s.map(function(t){var e=u.get(t.element);return i._buildPlayer(t,{},e)}));return this._playersById[t]=p,p.onDestroy(function(){return i.destroy(t)}),this.players.push(p),p},t.prototype.destroy=function(t){var e=this._getPlayer(t);e.destroy(),delete this._playersById[t];var n=this.players.indexOf(e);n>=0&&this.players.splice(n,1)},t.prototype._getPlayer=function(t){var e=this._playersById[t];if(!e)throw new Error("Unable to find the timeline player referenced by "+t);return e},t.prototype.listen=function(t,e,n,r){var i=u(e,"","","");return a(this._getPlayer(t),n,i,r),function(){}},t.prototype.command=function(t,e,n,r){if("register"!=n)if("create"!=n){var i=this._getPlayer(t);switch(n){case"play":i.play();break;case"pause":i.pause();break;case"reset":i.reset();break;case"restart":i.restart();break;case"finish":i.finish();break;case"init":i.init();break;case"setPosition":i.setPosition(parseFloat(r[0]));break;case"destroy":this.destroy(t)}}else{var o=r[0]||{};this.create(t,e,o)}else this.register(t,r[0])},t}(),St="ng-animate-disabled",Pt=".ng-animate-disabled",Mt=[],At={namespaceId:"",setForRemoval:null,hasAnimation:!1,removedBeforeQueried:!1},Tt={namespaceId:"",setForRemoval:null,hasAnimation:!1,removedBeforeQueried:!0},Dt="__ng_removed",It=function(){function t(t,e){void 0===e&&(e=""),this.namespaceId=e;var n,r=t&&t.hasOwnProperty("value"),i=r?t.value:t;if(this.value=null!=(n=i)?n:null,r){var o=T(t);delete o.value,this.options=o}else this.options={};this.options.params||(this.options.params={})}return Object.defineProperty(t.prototype,"params",{get:function(){return this.options.params},enumerable:!0,configurable:!0}),t.prototype.absorbOptions=function(t){var e=t.params;if(e){var n=this.options.params;Object.keys(e).forEach(function(t){null==n[t]&&(n[t]=e[t])})}},t}(),Rt=new It("void"),Nt=new It("DELETED"),Lt=function(){function t(t,e,n){this.id=t,this.hostElement=e,this._engine=n,this.players=[],this._triggers={},this._queue=[],this._elementListeners=new Map,this._hostClassName="ng-tns-"+t,Wt(e,this._hostClassName)}return t.prototype.listen=function(t,e,n,r){var i,o=this;if(!this._triggers.hasOwnProperty(e))throw new Error('Unable to listen on the animation trigger event "'+n+'" because the animation trigger "'+e+"\" doesn't exist!");if(null==n||0==n.length)throw new Error('Unable to listen on the animation trigger "'+e+'" because the provided event is undefined!');if("start"!=(i=n)&&"done"!=i)throw new Error('The provided animation trigger event "'+n+'" for the animation trigger "'+e+'" is not supported!');var s=l(this._elementListeners,t,[]),a={name:e,phase:n,callback:r};s.push(a);var c=l(this._engine.statesByElement,t,{});return c.hasOwnProperty(e)||(Wt(t,k),Wt(t,k+"-"+e),c[e]=Rt),function(){o._engine.afterFlush(function(){var t=s.indexOf(a);t>=0&&s.splice(t,1),o._triggers[e]||delete c[e]})}},t.prototype.register=function(t,e){return!this._triggers[t]&&(this._triggers[t]=e,!0)},t.prototype._getTrigger=function(t){var e=this._triggers[t];if(!e)throw new Error('The provided animation trigger "'+t+'" has not been registered!');return e},t.prototype.trigger=function(t,e,n,r){var i=this;void 0===r&&(r=!0);var o=this._getTrigger(e),s=new Vt(this.id,e,t),a=this._engine.statesByElement.get(t);a||(Wt(t,k),Wt(t,k+"-"+e),this._engine.statesByElement.set(t,a={}));var c=a[e],u=new It(n,this.id);if(!(n&&n.hasOwnProperty("value"))&&c&&u.absorbOptions(c.options),a[e]=u,c){if(c===Nt)return s}else c=Rt;if("void"===u.value||c.value!==u.value){var p=l(this._engine.playersByElement,t,[]);p.forEach(function(t){t.namespaceId==i.id&&t.triggerName==e&&t.queued&&t.destroy()});var d=o.matchTransition(c.value,u.value),h=!1;if(!d){if(!r)return;d=o.fallbackTransition,h=!0}return this._engine.totalQueuedPlayers++,this._queue.push({element:t,triggerName:e,transition:d,fromState:c,toState:u,player:s,isFallbackTransition:h}),h||(Wt(t,"ng-animate-queued"),s.onStart(function(){Gt(t,"ng-animate-queued")})),s.onDone(function(){var e=i.players.indexOf(s);e>=0&&i.players.splice(e,1);var n=i._engine.playersByElement.get(t);if(n){var r=n.indexOf(s);r>=0&&n.splice(r,1)}}),this.players.push(s),p.push(s),s}if(!function(t,e){var n=Object.keys(t),r=Object.keys(e);if(n.length!=r.length)return!1;for(var i=0;i<n.length;i++){var o=n[i];if(!e.hasOwnProperty(o)||t[o]!==e[o])return!1}return!0}(c.params,u.params)){var f=[],m=o.matchStyles(c.value,c.params,f),g=o.matchStyles(u.value,u.params,f);f.length?this._engine.reportError(f):this._engine.afterFlush(function(){N(t,m),R(t,g)})}},t.prototype.deregister=function(t){var e=this;delete this._triggers[t],this._engine.statesByElement.forEach(function(e,n){delete e[t]}),this._elementListeners.forEach(function(n,r){e._elementListeners.set(r,n.filter(function(e){return e.name!=t}))})},t.prototype.clearElementCache=function(t){this._engine.statesByElement.delete(t),this._elementListeners.delete(t);var e=this._engine.playersByElement.get(t);e&&(e.forEach(function(t){return t.destroy()}),this._engine.playersByElement.delete(t))},t.prototype._signalRemovalForInnerTriggers=function(t,e,n){var r=this;void 0===n&&(n=!1),this._engine.driver.query(t,E,!0).forEach(function(t){var n=r._engine.fetchNamespacesByElement(t);n.size?n.forEach(function(n){n.triggerLeaveAnimation(t,e,!1,!0)}):r.clearElementCache(t)})},t.prototype.triggerLeaveAnimation=function(t,e,n,r){var i=this,s=this._engine.statesByElement.get(t);if(s){var a=[];if(Object.keys(s).forEach(function(e){if(i._triggers[e]){var n=i.trigger(t,e,"void",r);n&&a.push(n)}}),a.length)return this._engine.markElementAsRemoved(this.id,t,!0,e),n&&o(a).onDone(function(){return i._engine.processLeaveNode(t)}),!0}return!1},t.prototype.prepareLeaveAnimationListeners=function(t){var e=this,n=this._elementListeners.get(t);if(n){var r=new Set;n.forEach(function(n){var i=n.name;if(!r.has(i)){r.add(i);var o=e._triggers[i].fallbackTransition,s=e._engine.statesByElement.get(t)[i]||Rt,a=new It("void"),c=new Vt(e.id,i,t);e._engine.totalQueuedPlayers++,e._queue.push({element:t,triggerName:i,transition:o,fromState:s,toState:a,player:c,isFallbackTransition:!0})}})}},t.prototype.removeNode=function(t,e){var n=this,r=this._engine;if(t.childElementCount&&this._signalRemovalForInnerTriggers(t,e,!0),!this.triggerLeaveAnimation(t,e,!0)){var i=!1;if(r.totalAnimations){var o=r.players.length?r.playersByQueriedElement.get(t):[];if(o&&o.length)i=!0;else for(var s=t;s=s.parentNode;){if(r.statesByElement.get(s)){i=!0;break}}}this.prepareLeaveAnimationListeners(t),i?r.markElementAsRemoved(this.id,t,!1,e):(r.afterFlush(function(){return n.clearElementCache(t)}),r.destroyInnerAnimations(t),r._onRemovalComplete(t,e))}},t.prototype.insertNode=function(t,e){Wt(t,this._hostClassName)},t.prototype.drainQueuedTransitions=function(t){var e=this,n=[];return this._queue.forEach(function(r){var i=r.player;if(!i.destroyed){var o=r.element,s=e._elementListeners.get(o);s&&s.forEach(function(e){if(e.name==r.triggerName){var n=u(o,r.triggerName,r.fromState.value,r.toState.value);n._data=t,a(r.player,e.phase,n,e.callback)}}),i.markedForDestroy?e._engine.afterFlush(function(){i.destroy()}):n.push(r)}}),this._queue=[],n.sort(function(t,n){var r=t.transition.ast.depCount,i=n.transition.ast.depCount;return 0==r||0==i?r-i:e._engine.driver.containsElement(t.element,n.element)?1:-1})},t.prototype.destroy=function(t){this.players.forEach(function(t){return t.destroy()}),this._signalRemovalForInnerTriggers(this.hostElement,t)},t.prototype.elementContainsData=function(t){var e=!1;return this._elementListeners.has(t)&&(e=!0),e=!!this._queue.find(function(e){return e.element===t})||e},t}(),Ft=function(){function t(t,e){this.driver=t,this._normalizer=e,this.players=[],this.newHostElements=new Map,this.playersByElement=new Map,this.playersByQueriedElement=new Map,this.statesByElement=new Map,this.disabledNodes=new Set,this.totalAnimations=0,this.totalQueuedPlayers=0,this._namespaceLookup={},this._namespaceList=[],this._flushFns=[],this._whenQuietFns=[],this.namespacesByHostElement=new Map,this.collectedEnterElements=[],this.collectedLeaveElements=[],this.onRemovalComplete=function(t,e){}}return t.prototype._onRemovalComplete=function(t,e){this.onRemovalComplete(t,e)},Object.defineProperty(t.prototype,"queuedPlayers",{get:function(){var t=[];return this._namespaceList.forEach(function(e){e.players.forEach(function(e){e.queued&&t.push(e)})}),t},enumerable:!0,configurable:!0}),t.prototype.createNamespace=function(t,e){var n=new Lt(t,e,this);return e.parentNode?this._balanceNamespaceList(n,e):(this.newHostElements.set(e,n),this.collectEnterElement(e)),this._namespaceLookup[t]=n},t.prototype._balanceNamespaceList=function(t,e){var n=this._namespaceList.length-1;if(n>=0){for(var r=!1,i=n;i>=0;i--){var o=this._namespaceList[i];if(this.driver.containsElement(o.hostElement,e)){this._namespaceList.splice(i+1,0,t),r=!0;break}}r||this._namespaceList.splice(0,0,t)}else this._namespaceList.push(t);return this.namespacesByHostElement.set(e,t),t},t.prototype.register=function(t,e){var n=this._namespaceLookup[t];return n||(n=this.createNamespace(t,e)),n},t.prototype.registerTrigger=function(t,e,n){var r=this._namespaceLookup[t];r&&r.register(e,n)&&this.totalAnimations++},t.prototype.destroy=function(t,e){var n=this;if(t){var r=this._fetchNamespace(t);this.afterFlush(function(){n.namespacesByHostElement.delete(r.hostElement),delete n._namespaceLookup[t];var e=n._namespaceList.indexOf(r);e>=0&&n._namespaceList.splice(e,1)}),this.afterFlushAnimationsDone(function(){return r.destroy(e)})}},t.prototype._fetchNamespace=function(t){return this._namespaceLookup[t]},t.prototype.fetchNamespacesByElement=function(t){var e=new Set,n=this.statesByElement.get(t);if(n)for(var r=Object.keys(n),i=0;i<r.length;i++){var o=n[r[i]].namespaceId;if(o){var s=this._fetchNamespace(o);s&&e.add(s)}}return e},t.prototype.trigger=function(t,e,n,r){return!!zt(e)&&(this._fetchNamespace(t).trigger(e,n,r),!0)},t.prototype.insertNode=function(t,e,n,r){if(zt(e)){var i=e[Dt];i&&i.setForRemoval&&(i.setForRemoval=!1),t&&this._fetchNamespace(t).insertNode(e,n),r&&this.collectEnterElement(e)}},t.prototype.collectEnterElement=function(t){this.collectedEnterElements.push(t)},t.prototype.markElementAsDisabled=function(t,e){e?this.disabledNodes.has(t)||(this.disabledNodes.add(t),Wt(t,St)):this.disabledNodes.has(t)&&(this.disabledNodes.delete(t),Gt(t,St))},t.prototype.removeNode=function(t,e,n){if(zt(e)){var r=t?this._fetchNamespace(t):null;r?r.removeNode(e,n):this.markElementAsRemoved(t,e,!1,n)}else this._onRemovalComplete(e,n)},t.prototype.markElementAsRemoved=function(t,e,n,r){this.collectedLeaveElements.push(e),e[Dt]={namespaceId:t,setForRemoval:r,hasAnimation:n,removedBeforeQueried:!1}},t.prototype.listen=function(t,e,n,r,i){return zt(e)?this._fetchNamespace(t).listen(e,n,r,i):function(){}},t.prototype._buildInstruction=function(t,e,n,r){return t.transition.build(this.driver,t.element,t.fromState.value,t.toState.value,n,r,t.fromState.options,t.toState.options,e)},t.prototype.destroyInnerAnimations=function(t){var e=this,n=this.driver.query(t,E,!0);n.forEach(function(t){return e.destroyActiveAnimationsForElement(t)}),0!=this.playersByQueriedElement.size&&(n=this.driver.query(t,S,!0)).forEach(function(t){return e.finishActiveQueriedAnimationOnElement(t)})},t.prototype.destroyActiveAnimationsForElement=function(t){var e=this.playersByElement.get(t);e&&e.forEach(function(t){t.queued?t.markedForDestroy=!0:t.destroy()});var n=this.statesByElement.get(t);n&&Object.keys(n).forEach(function(t){return n[t]=Nt})},t.prototype.finishActiveQueriedAnimationOnElement=function(t){var e=this.playersByQueriedElement.get(t);e&&e.forEach(function(t){return t.finish()})},t.prototype.whenRenderingDone=function(){var t=this;return new Promise(function(e){if(t.players.length)return o(t.players).onDone(function(){return e()});e()})},t.prototype.processLeaveNode=function(t){var e=this,n=t[Dt];if(n&&n.setForRemoval){if(t[Dt]=At,n.namespaceId){this.destroyInnerAnimations(t);var r=this._fetchNamespace(n.namespaceId);r&&r.clearElementCache(t)}this._onRemovalComplete(t,n.setForRemoval)}this.driver.matchesElement(t,Pt)&&this.markElementAsDisabled(t,!1),this.driver.query(t,Pt,!0).forEach(function(n){e.markElementAsDisabled(t,!1)})},t.prototype.flush=function(t){var e=this;void 0===t&&(t=-1);var n=[];if(this.newHostElements.size&&(this.newHostElements.forEach(function(t,n){return e._balanceNamespaceList(t,n)}),this.newHostElements.clear()),this.totalAnimations&&this.collectedEnterElements.length)for(var r=0;r<this.collectedEnterElements.length;r++){Wt(this.collectedEnterElements[r],"ng-star-inserted")}if(this._namespaceList.length&&(this.totalQueuedPlayers||this.collectedLeaveElements.length)){var i=[];try{n=this._flushAnimations(i,t)}finally{for(r=0;r<i.length;r++)i[r]()}}else for(r=0;r<this.collectedLeaveElements.length;r++){var s=this.collectedLeaveElements[r];this.processLeaveNode(s)}if(this.totalQueuedPlayers=0,this.collectedEnterElements.length=0,this.collectedLeaveElements.length=0,this._flushFns.forEach(function(t){return t()}),this._flushFns=[],this._whenQuietFns.length){var a=this._whenQuietFns;this._whenQuietFns=[],n.length?o(n).onDone(function(){a.forEach(function(t){return t()})}):a.forEach(function(t){return t()})}},t.prototype.reportError=function(t){throw new Error("Unable to process animations due to the following failed trigger transitions\n "+t.join("\n"))},t.prototype._flushAnimations=function(t,e){var n=this,s=new ot,a=[],c=new Map,u=[],p=new Map,d=new Map,h=new Map,f=new Set;this.disabledNodes.forEach(function(t){f.add(t);for(var e=n.driver.query(t,".ng-animate-queued",!0),r=0;r<e.length;r++)f.add(e[r])});var m=b(),g=Array.from(this.statesByElement.keys()),v=Ht(g,this.collectedEnterElements),y=new Map,_=0;v.forEach(function(t,e){var n="ng-enter"+_++;y.set(e,n),t.forEach(function(t){return Wt(t,n)})});for(var j=[],x=new Set,w=new Set,O=0;O<this.collectedLeaveElements.length;O++){(U=(H=this.collectedLeaveElements[O])[Dt])&&U.setForRemoval&&(j.push(H),x.add(H),U.hasAnimation?this.driver.query(H,".ng-star-inserted",!0).forEach(function(t){return x.add(t)}):w.add(H))}var C=new Map,k=Ht(g,Array.from(x));k.forEach(function(t,e){var n="ng-leave"+_++;C.set(e,n),t.forEach(function(t){return Wt(t,n)})}),t.push(function(){v.forEach(function(t,e){var n=y.get(e);t.forEach(function(t){return Gt(t,n)})}),k.forEach(function(t,e){var n=C.get(e);t.forEach(function(t){return Gt(t,n)})}),j.forEach(function(t){n.processLeaveNode(t)})});for(var E=[],P=[],M=this._namespaceList.length-1;M>=0;M--){this._namespaceList[M].drainQueuedTransitions(e).forEach(function(t){var e=t.player;E.push(e);var r=t.element;if(m&&n.driver.containsElement(m,r)){var i=C.get(r),o=y.get(r),c=n._buildInstruction(t,s,o,i);if(c.errors&&c.errors.length)P.push(c);else{if(t.isFallbackTransition)return e.onStart(function(){return N(r,c.fromStyles)}),e.onDestroy(function(){return R(r,c.toStyles)}),void a.push(e);c.timelines.forEach(function(t){return t.stretchStartingKeyframe=!0}),s.append(r,c.timelines);var f={instruction:c,player:e,element:r};u.push(f),c.queriedElements.forEach(function(t){return l(p,t,[]).push(e)}),c.preStyleProps.forEach(function(t,e){var n=Object.keys(t);if(n.length){var r=d.get(e);r||d.set(e,r=new Set),n.forEach(function(t){return r.add(t)})}}),c.postStyleProps.forEach(function(t,e){var n=Object.keys(t),r=h.get(e);r||h.set(e,r=new Set),n.forEach(function(t){return r.add(t)})})}}else e.destroy()})}if(P.length){var A=[];P.forEach(function(t){A.push("@"+t.triggerName+" has failed due to:\n"),t.errors.forEach(function(t){return A.push("- "+t+"\n")})}),E.forEach(function(t){return t.destroy()}),this.reportError(A)}var T=new Map,D=new Map;u.forEach(function(t){var e=t.element;s.has(e)&&(D.set(e,e),n._beforeAnimationBuild(t.player.namespaceId,t.instruction,T))}),a.forEach(function(t){var e=t.element;n._getPreviousPlayers(e,!1,t.namespaceId,t.triggerName,null).forEach(function(t){l(T,e,[]).push(t),t.destroy()})});var I=j.filter(function(t){return Yt(t,d,h)}),L=new Map;qt(L,this.driver,w,h,r.a).forEach(function(t){Yt(t,d,h)&&I.push(t)});var F=new Map;v.forEach(function(t,e){qt(F,n.driver,new Set(t),d,r.m)}),I.forEach(function(t){var e=L.get(t),n=F.get(t);L.set(t,Object(i.a)({},e,n))});var V=[],z=[],B={};u.forEach(function(t){var e=t.element,r=t.player,i=t.instruction;if(s.has(e)){if(f.has(e))return r.onDestroy(function(){return R(e,i.toStyles)}),void a.push(r);var u=B;if(D.size>1){for(var l=e,p=[];l=l.parentNode;){var d=D.get(l);if(d){u=d;break}p.push(l)}p.forEach(function(t){return D.set(t,u)})}var h=n._buildAnimation(r.namespaceId,i,T,c,F,L);if(r.setRealPlayer(h),u===B)V.push(r);else{var m=n.playersByElement.get(u);m&&m.length&&(r.parentPlayer=o(m)),a.push(r)}}else N(e,i.fromStyles),r.onDestroy(function(){return R(e,i.toStyles)}),z.push(r),f.has(e)&&a.push(r)}),z.forEach(function(t){var e=c.get(t.element);if(e&&e.length){var n=o(e);t.setRealPlayer(n)}}),a.forEach(function(t){t.parentPlayer?t.syncPlayerEvents(t.parentPlayer):t.destroy()});for(var q=0;q<j.length;q++){var H,U=(H=j[q])[Dt];if(Gt(H,"ng-leave"),!U||!U.hasAnimation){var W=[];if(p.size){var G=p.get(H);G&&G.length&&W.push.apply(W,G);for(var X=this.driver.query(H,S,!0),Y=0;Y<X.length;Y++){var K=p.get(X[Y]);K&&K.length&&W.push.apply(W,K)}}var $=W.filter(function(t){return!t.destroyed});$.length?Xt(this,H,$):this.processLeaveNode(H)}}return j.length=0,V.forEach(function(t){n.players.push(t),t.onDone(function(){t.destroy();var e=n.players.indexOf(t);n.players.splice(e,1)}),t.play()}),V},t.prototype.elementContainsData=function(t,e){var n=!1,r=e[Dt];return r&&r.setForRemoval&&(n=!0),this.playersByElement.has(e)&&(n=!0),this.playersByQueriedElement.has(e)&&(n=!0),this.statesByElement.has(e)&&(n=!0),this._fetchNamespace(t).elementContainsData(e)||n},t.prototype.afterFlush=function(t){this._flushFns.push(t)},t.prototype.afterFlushAnimationsDone=function(t){this._whenQuietFns.push(t)},t.prototype._getPreviousPlayers=function(t,e,n,r,i){var o=[];if(e){var s=this.playersByQueriedElement.get(t);s&&(o=s)}else{var a=this.playersByElement.get(t);if(a){var c=!i||"void"==i;a.forEach(function(t){t.queued||(c||t.triggerName==r)&&o.push(t)})}}return(n||r)&&(o=o.filter(function(t){return(!n||n==t.namespaceId)&&(!r||r==t.triggerName)})),o},t.prototype._beforeAnimationBuild=function(t,e,n){for(var r=e.triggerName,i=e.element,o=e.isRemovalTransition?void 0:t,s=e.isRemovalTransition?void 0:r,a=function(t){var r=t.element,a=r!==i,u=l(n,r,[]);c._getPreviousPlayers(r,a,o,s,e.toState).forEach(function(t){var e=t.getRealPlayer();e.beforeDestroy&&e.beforeDestroy(),t.destroy(),u.push(t)})},c=this,u=0,p=e.timelines;u<p.length;u++){a(p[u])}N(i,e.fromStyles)},t.prototype._buildAnimation=function(t,e,n,i,a,c){var u=this,p=e.triggerName,d=e.element,h=[],f=new Set,m=new Set,g=e.timelines.map(function(e){var o=e.element;f.add(o);var l=o[Dt];if(l&&l.removedBeforeQueried)return new r.d;var g,v,y=o!==d,b=(g=(n.get(o)||Mt).map(function(t){return t.getRealPlayer()}),v=[],function t(e,n){for(var i=0;i<e.length;i++){var o=e[i];o instanceof r.l?t(o.players,n):n.push(o)}}(g,v),v).filter(function(t){var e=t;return!!e.element&&e.element===o}),_=a.get(o),j=c.get(o),x=s(u.driver,u._normalizer,0,e.keyframes,_,j),w=u._buildPlayer(e,x,b);if(e.subTimeline&&i&&m.add(o),y){var O=new Vt(t,p,o);O.setRealPlayer(w),h.push(O)}return w});h.forEach(function(t){l(u.playersByQueriedElement,t.element,[]).push(t),t.onDone(function(){return function(t,e,n){var r;if(t instanceof Map){if(r=t.get(e)){if(r.length){var i=r.indexOf(n);r.splice(i,1)}0==r.length&&t.delete(e)}}else if(r=t[e]){if(r.length){var i=r.indexOf(n);r.splice(i,1)}0==r.length&&delete t[e]}return r}(u.playersByQueriedElement,t.element,t)})}),f.forEach(function(t){return Wt(t,"ng-animating")});var v=o(g);return v.onDestroy(function(){f.forEach(function(t){return Gt(t,"ng-animating")}),R(d,e.toStyles)}),m.forEach(function(t){l(i,t,[]).push(v)}),v},t.prototype._buildPlayer=function(t,e,n){return e.length>0?this.driver.animate(t.element,e,t.duration,t.delay,t.easing,n):new r.d},t}(),Vt=function(){function t(t,e,n){this.namespaceId=t,this.triggerName=e,this.element=n,this._player=new r.d,this._containsRealPlayer=!1,this._queuedCallbacks={},this.destroyed=!1,this.markedForDestroy=!1}return Object.defineProperty(t.prototype,"queued",{get:function(){return 0==this._containsRealPlayer},enumerable:!0,configurable:!0}),t.prototype.setRealPlayer=function(t){var e=this;this._containsRealPlayer||(this._player=t,Object.keys(this._queuedCallbacks).forEach(function(n){e._queuedCallbacks[n].forEach(function(e){return a(t,n,void 0,e)})}),this._queuedCallbacks={},this._containsRealPlayer=!0)},t.prototype.getRealPlayer=function(){return this._player},t.prototype.syncPlayerEvents=function(t){var e=this,n=this._player;n.triggerCallback&&t.onStart(function(){return n.triggerCallback("start")}),t.onDone(function(){return e.finish()}),t.onDestroy(function(){return e.destroy()})},t.prototype._queueEvent=function(t,e){l(this._queuedCallbacks,t,[]).push(e)},t.prototype.onDone=function(t){this.queued&&this._queueEvent("done",t),this._player.onDone(t)},t.prototype.onStart=function(t){this.queued&&this._queueEvent("start",t),this._player.onStart(t)},t.prototype.onDestroy=function(t){this.queued&&this._queueEvent("destroy",t),this._player.onDestroy(t)},t.prototype.init=function(){this._player.init()},t.prototype.hasStarted=function(){return!this.queued&&this._player.hasStarted()},t.prototype.play=function(){!this.queued&&this._player.play()},t.prototype.pause=function(){!this.queued&&this._player.pause()},t.prototype.restart=function(){!this.queued&&this._player.restart()},t.prototype.finish=function(){this._player.finish()},t.prototype.destroy=function(){this.destroyed=!0,this._player.destroy()},t.prototype.reset=function(){!this.queued&&this._player.reset()},t.prototype.setPosition=function(t){this.queued||this._player.setPosition(t)},t.prototype.getPosition=function(){return this.queued?0:this._player.getPosition()},Object.defineProperty(t.prototype,"totalTime",{get:function(){return this._player.totalTime},enumerable:!0,configurable:!0}),t.prototype.triggerCallback=function(t){var e=this._player;e.triggerCallback&&e.triggerCallback(t)},t}();function zt(t){return t&&1===t.nodeType}function Bt(t,e){var n=t.style.display;return t.style.display=null!=e?e:"none",n}function qt(t,e,n,r,i){var o=[];n.forEach(function(t){return o.push(Bt(t))});var s=[];r.forEach(function(n,r){var o={};n.forEach(function(t){var n=o[t]=e.computeStyle(r,t,i);n&&0!=n.length||(r[Dt]=Tt,s.push(r))}),t.set(r,o)});var a=0;return n.forEach(function(t){return Bt(t,o[a++])}),s}function Ht(t,e){var n=new Map;if(t.forEach(function(t){return n.set(t,[])}),0==e.length)return n;var r=1,i=new Set(e),o=new Map;return e.forEach(function(t){var e=function t(e){if(!e)return r;var s=o.get(e);if(s)return s;var a=e.parentNode;return s=n.has(a)?a:i.has(a)?r:t(a),o.set(e,s),s}(t);e!==r&&n.get(e).push(t)}),n}var Ut="$$classes";function Wt(t,e){if(t.classList)t.classList.add(e);else{var n=t[Ut];n||(n=t[Ut]={}),n[e]=!0}}function Gt(t,e){if(t.classList)t.classList.remove(e);else{var n=t[Ut];n&&delete n[e]}}function Xt(t,e,n){o(n).onDone(function(){return t.processLeaveNode(e)})}function Yt(t,e,n){var r=n.get(t);if(!r)return!1;var i=e.get(t);return i?r.forEach(function(t){return i.add(t)}):e.set(t,r),n.delete(t),!0}var Kt=function(){function t(t,e){var n=this;this._driver=t,this._triggerCache={},this.onRemovalComplete=function(t,e){},this._transitionEngine=new Ft(t,e),this._timelineEngine=new Et(t,e),this._transitionEngine.onRemovalComplete=function(t,e){return n.onRemovalComplete(t,e)}}return t.prototype.registerTrigger=function(t,e,n,r,i){var o=t+"-"+r,s=this._triggerCache[o];if(!s){var a=[],c=Z(this._driver,i,a);if(a.length)throw new Error('The animation trigger "'+r+'" has failed to build due to the following errors:\n - '+a.join("\n - "));s=new Ot(r,c),this._triggerCache[o]=s}this._transitionEngine.registerTrigger(e,r,s)},t.prototype.register=function(t,e){this._transitionEngine.register(t,e)},t.prototype.destroy=function(t,e){this._transitionEngine.destroy(t,e)},t.prototype.onInsert=function(t,e,n,r){this._transitionEngine.insertNode(t,e,n,r)},t.prototype.onRemove=function(t,e,n){this._transitionEngine.removeNode(t,e,n)},t.prototype.disableAnimations=function(t,e){this._transitionEngine.markElementAsDisabled(t,e)},t.prototype.process=function(t,e,n,r){if("@"==n.charAt(0)){var i=p(n),o=i[0],s=i[1],a=r;this._timelineEngine.command(o,e,s,a)}else this._transitionEngine.trigger(t,e,n,r)},t.prototype.listen=function(t,e,n,r,i){if("@"==n.charAt(0)){var o=p(n),s=o[0],a=o[1];return this._timelineEngine.listen(s,e,a,i)}return this._transitionEngine.listen(t,e,n,r,i)},t.prototype.flush=function(t){void 0===t&&(t=-1),this._transitionEngine.flush(t)},Object.defineProperty(t.prototype,"players",{get:function(){return this._transitionEngine.players.concat(this._timelineEngine.players)},enumerable:!0,configurable:!0}),t.prototype.whenRenderingDone=function(){return this._transitionEngine.whenRenderingDone()},t}(),$t=function(){function t(t,e,n,r){void 0===r&&(r=[]);var i,o,s=this;this.element=t,this.keyframes=e,this.options=n,this.previousPlayers=r,this._onDoneFns=[],this._onStartFns=[],this._onDestroyFns=[],this._initialized=!1,this._finished=!1,this._started=!1,this._destroyed=!1,this.time=0,this.parentPlayer=null,this.previousStyles={},this.currentSnapshot={},this._duration=n.duration,this._delay=n.delay||0,this.time=this._duration+this._delay,i=this._duration,o=this._delay,(0===i||0===o)&&r.forEach(function(t){var e=t.currentSnapshot;Object.keys(e).forEach(function(t){return s.previousStyles[t]=e[t]})})}return t.prototype._onFinish=function(){this._finished||(this._finished=!0,this._onDoneFns.forEach(function(t){return t()}),this._onDoneFns=[])},t.prototype.init=function(){this._buildPlayer(),this._preparePlayerBeforeStart()},t.prototype._buildPlayer=function(){var t=this;if(!this._initialized){this._initialized=!0;var e=this.keyframes.map(function(t){return I(t,!1)}),n=Object.keys(this.previousStyles);if(n.length){var r=e[0],i=[];if(n.forEach(function(e){r.hasOwnProperty(e)||i.push(e),r[e]=t.previousStyles[e]}),i.length)for(var o=this,s=function(){var t=e[a];i.forEach(function(e){t[e]=Qt(o.element,e)})},a=1;a<e.length;a++)s()}this.domPlayer=this._triggerWebAnimation(this.element,e,this.options),this._finalKeyframe=e.length?e[e.length-1]:{},this.domPlayer.addEventListener("finish",function(){return t._onFinish()})}},t.prototype._preparePlayerBeforeStart=function(){this._delay?this._resetDomPlayerState():this.domPlayer.pause()},t.prototype._triggerWebAnimation=function(t,e,n){return t.animate(e,n)},t.prototype.onStart=function(t){this._onStartFns.push(t)},t.prototype.onDone=function(t){this._onDoneFns.push(t)},t.prototype.onDestroy=function(t){this._onDestroyFns.push(t)},t.prototype.play=function(){this._buildPlayer(),this.hasStarted()||(this._onStartFns.forEach(function(t){return t()}),this._onStartFns=[],this._started=!0),this.domPlayer.play()},t.prototype.pause=function(){this.init(),this.domPlayer.pause()},t.prototype.finish=function(){this.init(),this._onFinish(),this.domPlayer.finish()},t.prototype.reset=function(){this._resetDomPlayerState(),this._destroyed=!1,this._finished=!1,this._started=!1},t.prototype._resetDomPlayerState=function(){this.domPlayer&&this.domPlayer.cancel()},t.prototype.restart=function(){this.reset(),this.play()},t.prototype.hasStarted=function(){return this._started},t.prototype.destroy=function(){this._destroyed||(this._destroyed=!0,this._resetDomPlayerState(),this._onFinish(),this._onDestroyFns.forEach(function(t){return t()}),this._onDestroyFns=[])},t.prototype.setPosition=function(t){this.domPlayer.currentTime=t*this.time},t.prototype.getPosition=function(){return this.domPlayer.currentTime/this.time},Object.defineProperty(t.prototype,"totalTime",{get:function(){return this._delay+this._duration},enumerable:!0,configurable:!0}),t.prototype.beforeDestroy=function(){var t=this,e={};this.hasStarted()&&Object.keys(this._finalKeyframe).forEach(function(n){"offset"!=n&&(e[n]=t._finished?t._finalKeyframe[n]:Qt(t.element,n))}),this.currentSnapshot=e},t.prototype.triggerCallback=function(t){var e="start"==t?this._onStartFns:this._onDoneFns;e.forEach(function(t){return t()}),e.length=0},t}();function Qt(t,e){return window.getComputedStyle(t)[e]}var Zt=function(){function t(){}return t.prototype.validateStyleProperty=function(t){return y(t)},t.prototype.matchesElement=function(t,e){return _(t,e)},t.prototype.containsElement=function(t,e){return j(t,e)},t.prototype.query=function(t,e,n){return x(t,e,n)},t.prototype.computeStyle=function(t,e,n){return window.getComputedStyle(t)[e]},t.prototype.animate=function(t,e,n,r,i,o){void 0===o&&(o=[]);var s={duration:n,delay:r,fill:0==r?"both":"forwards"};i&&(s.easing=i);var a=o.filter(function(t){return t instanceof $t});return new $t(t,e,s,a)},t}();function Jt(){return"undefined"!=typeof Element&&"function"==typeof Element.prototype.animate}},"../../../cdk/esm5/a11y.es5.js":function(t,e,n){"use strict";n.d(e,"c",function(){return C}),n.d(e,"d",function(){return D}),n.d(e,"b",function(){return I}),n.d(e,"j",function(){return R}),n.d(e,"e",function(){return N}),n.d(e,"g",function(){return j}),n.d(e,"i",function(){return F}),n.d(e,"h",function(){return V}),n.d(e,"f",function(){return z}),n.d(e,"a",function(){return H});var r=n("../../../core/esm5/core.js"),i=n("../../../cdk/esm5/coercion.es5.js"),o=n("../../../../rxjs/_esm5/operators/take.js"),s=n("../../../cdk/esm5/platform.es5.js"),a=n("../../../common/esm5/common.js"),c=n("../../../../tslib/tslib.es6.js"),u=n("../../../../rxjs/_esm5/Subject.js"),l=n("../../../../rxjs/_esm5/Subscription.js"),p=n("../../../cdk/esm5/keycodes.es5.js"),d=n("../../../../rxjs/_esm5/operators/debounceTime.js"),h=n("../../../../rxjs/_esm5/operators/filter.js"),f=n("../../../../rxjs/_esm5/operators/map.js"),m=n("../../../../rxjs/_esm5/operators/tap.js"),g=n("../../../../rxjs/_esm5/observable/of.js"),v=function(){function t(t){this._platform=t}return t.prototype.isDisabled=function(t){return t.hasAttribute("disabled")},t.prototype.isVisible=function(t){return!!((e=t).offsetWidth||e.offsetHeight||"function"==typeof e.getClientRects&&e.getClientRects().length)&&"visible"===getComputedStyle(t).visibility;var e},t.prototype.isTabbable=function(t){if(!this._platform.isBrowser)return!1;var e,n=(e=t,e.ownerDocument.defaultView||window).frameElement;if(n){var r=n&&n.nodeName.toLowerCase();if(-1===b(n))return!1;if((this._platform.BLINK||this._platform.WEBKIT)&&"object"===r)return!1;if((this._platform.BLINK||this._platform.WEBKIT)&&!this.isVisible(n))return!1}var i,o,s,a=t.nodeName.toLowerCase(),c=b(t);if(t.hasAttribute("contenteditable"))return-1!==c;if("iframe"===a)return!1;if("audio"===a){if(!t.hasAttribute("controls"))return!1;if(this._platform.BLINK)return!0}if("video"===a){if(!t.hasAttribute("controls")&&this._platform.TRIDENT)return!1;if(this._platform.BLINK||this._platform.FIREFOX)return!0}return("object"!==a||!this._platform.BLINK&&!this._platform.WEBKIT)&&((!this._platform.WEBKIT||!this._platform.IOS||(o=(i=t).nodeName.toLowerCase(),"text"===(s="input"===o&&i.type)||"password"===s||"select"===o||"textarea"===o))&&t.tabIndex>=0)},t.prototype.isFocusable=function(t){return function(t){if(e=t,n=e,"input"==n.nodeName.toLowerCase()&&"hidden"==e.type)return!1;var e,n;return o=t,s=o.nodeName.toLowerCase(),"input"===s||"select"===s||"button"===s||"textarea"===s||(r=t,i=r,"a"==i.nodeName.toLowerCase()&&r.hasAttribute("href"))||t.hasAttribute("contenteditable")||y(t);var r,i;var o,s}(t)&&!this.isDisabled(t)&&this.isVisible(t)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:s.a}]},t}();function y(t){if(!t.hasAttribute("tabindex")||void 0===t.tabIndex)return!1;var e=t.getAttribute("tabindex");return"-32768"!=e&&!(!e||isNaN(parseInt(e,10)))}function b(t){if(!y(t))return null;var e=parseInt(t.getAttribute("tabindex")||"",10);return isNaN(e)?-1:e}var _=function(){function t(t,e,n,r,i){void 0===i&&(i=!1),this._element=t,this._checker=e,this._ngZone=n,this._document=r,this._enabled=!0,i||this.attachAnchors()}return Object.defineProperty(t.prototype,"enabled",{get:function(){return this._enabled},set:function(t){this._enabled=t,this._startAnchor&&this._endAnchor&&(this._startAnchor.tabIndex=this._endAnchor.tabIndex=this._enabled?0:-1)},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){this._startAnchor&&this._startAnchor.parentNode&&this._startAnchor.parentNode.removeChild(this._startAnchor),this._endAnchor&&this._endAnchor.parentNode&&this._endAnchor.parentNode.removeChild(this._endAnchor),this._startAnchor=this._endAnchor=null},t.prototype.attachAnchors=function(){var t=this;this._startAnchor||(this._startAnchor=this._createAnchor()),this._endAnchor||(this._endAnchor=this._createAnchor()),this._ngZone.runOutsideAngular(function(){t._startAnchor.addEventListener("focus",function(){t.focusLastTabbableElement()}),t._endAnchor.addEventListener("focus",function(){t.focusFirstTabbableElement()}),t._element.parentNode&&(t._element.parentNode.insertBefore(t._startAnchor,t._element),t._element.parentNode.insertBefore(t._endAnchor,t._element.nextSibling))})},t.prototype.focusInitialElementWhenReady=function(){var t=this;return new Promise(function(e){t._executeOnStable(function(){return e(t.focusInitialElement())})})},t.prototype.focusFirstTabbableElementWhenReady=function(){var t=this;return new Promise(function(e){t._executeOnStable(function(){return e(t.focusFirstTabbableElement())})})},t.prototype.focusLastTabbableElementWhenReady=function(){var t=this;return new Promise(function(e){t._executeOnStable(function(){return e(t.focusLastTabbableElement())})})},t.prototype._getRegionBoundary=function(t){for(var e=this._element.querySelectorAll("[cdk-focus-region-"+t+"], [cdkFocusRegion"+t+"], [cdk-focus-"+t+"]"),n=0;n<e.length;n++)e[n].hasAttribute("cdk-focus-"+t)?console.warn("Found use of deprecated attribute 'cdk-focus-"+t+"', use 'cdkFocusRegion"+t+"' instead.",e[n]):e[n].hasAttribute("cdk-focus-region-"+t)&&console.warn("Found use of deprecated attribute 'cdk-focus-region-"+t+"', use 'cdkFocusRegion"+t+"' instead.",e[n]);return"start"==t?e.length?e[0]:this._getFirstTabbableElement(this._element):e.length?e[e.length-1]:this._getLastTabbableElement(this._element)},t.prototype.focusInitialElement=function(){var t=this._element.querySelector("[cdk-focus-initial], [cdkFocusInitial]");return this._element.hasAttribute("cdk-focus-initial")&&console.warn("Found use of deprecated attribute 'cdk-focus-initial', use 'cdkFocusInitial' instead.",this._element),t?(t.focus(),!0):this.focusFirstTabbableElement()},t.prototype.focusFirstTabbableElement=function(){var t=this._getRegionBoundary("start");return t&&t.focus(),!!t},t.prototype.focusLastTabbableElement=function(){var t=this._getRegionBoundary("end");return t&&t.focus(),!!t},t.prototype._getFirstTabbableElement=function(t){if(this._checker.isFocusable(t)&&this._checker.isTabbable(t))return t;for(var e=t.children||t.childNodes,n=0;n<e.length;n++){var r=e[n].nodeType===Node.ELEMENT_NODE?this._getFirstTabbableElement(e[n]):null;if(r)return r}return null},t.prototype._getLastTabbableElement=function(t){if(this._checker.isFocusable(t)&&this._checker.isTabbable(t))return t;for(var e=t.children||t.childNodes,n=e.length-1;n>=0;n--){var r=e[n].nodeType===Node.ELEMENT_NODE?this._getLastTabbableElement(e[n]):null;if(r)return r}return null},t.prototype._createAnchor=function(){var t=this._document.createElement("div");return t.tabIndex=this._enabled?0:-1,t.classList.add("cdk-visually-hidden"),t.classList.add("cdk-focus-trap-anchor"),t},t.prototype._executeOnStable=function(t){this._ngZone.isStable?t():this._ngZone.onStable.asObservable().pipe(Object(o.a)(1)).subscribe(t)},t}(),j=function(){function t(t,e,n){this._checker=t,this._ngZone=e,this._document=n}return t.prototype.create=function(t,e){return void 0===e&&(e=!1),new _(t,this._checker,this._ngZone,this._document,e)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:v},{type:r.L},{type:void 0,decorators:[{type:r.y,args:[a.b]}]}]},t}(),x=function(){function t(t,e){this._elementRef=t,this._focusTrapFactory=e,this.focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement,!0)}return Object.defineProperty(t.prototype,"disabled",{get:function(){return!this.focusTrap.enabled},set:function(t){this.focusTrap.enabled=!Object(i.b)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this.focusTrap.destroy()},t.prototype.ngAfterContentInit=function(){this.focusTrap.attachAnchors()},t.decorators=[{type:r.s,args:[{selector:"cdk-focus-trap"}]}],t.ctorParameters=function(){return[{type:r.t},{type:j}]},t.propDecorators={disabled:[{type:r.C}]},t}(),w=function(){function t(t,e,n){this._elementRef=t,this._focusTrapFactory=e,this._previouslyFocusedElement=null,this._document=n,this.focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement,!0)}return Object.defineProperty(t.prototype,"enabled",{get:function(){return this.focusTrap.enabled},set:function(t){this.focusTrap.enabled=Object(i.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"autoCapture",{get:function(){return this._autoCapture},set:function(t){this._autoCapture=Object(i.b)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this.focusTrap.destroy(),this._previouslyFocusedElement&&(this._previouslyFocusedElement.focus(),this._previouslyFocusedElement=null)},t.prototype.ngAfterContentInit=function(){this.focusTrap.attachAnchors(),this.autoCapture&&(this._previouslyFocusedElement=this._document.activeElement,this.focusTrap.focusInitialElementWhenReady())},t.decorators=[{type:r.s,args:[{selector:"[cdkTrapFocus]",exportAs:"cdkTrapFocus"}]}],t.ctorParameters=function(){return[{type:r.t},{type:j},{type:void 0,decorators:[{type:r.y,args:[a.b]}]}]},t.propDecorators={enabled:[{type:r.C,args:["cdkTrapFocus"]}],autoCapture:[{type:r.C,args:["cdkTrapFocusAutoCapture"]}]},t}(),O=function(){function t(t){this._items=t,this._activeItemIndex=-1,this._wrap=!1,this._letterKeyStream=new u.a,this._typeaheadSubscription=l.a.EMPTY,this._pressedLetters=[],this.tabOut=new u.a,this.change=new u.a}return t.prototype.withWrap=function(){return this._wrap=!0,this},t.prototype.withTypeAhead=function(t){var e=this;if(void 0===t&&(t=200),this._items.length&&this._items.some(function(t){return"function"!=typeof t.getLabel}))throw Error("ListKeyManager items in typeahead mode must implement the `getLabel` method.");return this._typeaheadSubscription.unsubscribe(),this._typeaheadSubscription=this._letterKeyStream.pipe(Object(m.a)(function(t){return e._pressedLetters.push(t)}),Object(d.a)(t),Object(h.a)(function(){return e._pressedLetters.length>0}),Object(f.a)(function(){return e._pressedLetters.join("")})).subscribe(function(t){for(var n=e._items.toArray(),r=1;r<n.length+1;r++){var i=(e._activeItemIndex+r)%n.length,o=n[i];if(!o.disabled&&0===o.getLabel().toUpperCase().trim().indexOf(t)){e.setActiveItem(i);break}}e._pressedLetters=[]}),this},t.prototype.setActiveItem=function(t){var e=this._activeItemIndex;this._activeItemIndex=t,this._activeItem=this._items.toArray()[t],this._activeItemIndex!==e&&this.change.next(t)},t.prototype.onKeydown=function(t){switch(t.keyCode){case p.d:this.setNextItemActive();break;case p.p:this.setPreviousItemActive();break;case p.o:return void this.tabOut.next();default:var e=t.keyCode;return void(t.key&&1===t.key.length?this._letterKeyStream.next(t.key.toLocaleUpperCase()):(e>=p.a&&e<=p.q||e>=p.r&&e<=p.j)&&this._letterKeyStream.next(String.fromCharCode(e)))}this._pressedLetters=[],t.preventDefault()},Object.defineProperty(t.prototype,"activeItemIndex",{get:function(){return this._activeItemIndex},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activeItem",{get:function(){return this._activeItem},enumerable:!0,configurable:!0}),t.prototype.setFirstItemActive=function(){this._setActiveItemByIndex(0,1)},t.prototype.setLastItemActive=function(){this._setActiveItemByIndex(this._items.length-1,-1)},t.prototype.setNextItemActive=function(){this._activeItemIndex<0?this.setFirstItemActive():this._setActiveItemByDelta(1)},t.prototype.setPreviousItemActive=function(){this._activeItemIndex<0&&this._wrap?this.setLastItemActive():this._setActiveItemByDelta(-1)},t.prototype.updateActiveItemIndex=function(t){this._activeItemIndex=t},t.prototype._setActiveItemByDelta=function(t,e){void 0===e&&(e=this._items.toArray()),this._wrap?this._setActiveInWrapMode(t,e):this._setActiveInDefaultMode(t,e)},t.prototype._setActiveInWrapMode=function(t,e){this._activeItemIndex=(this._activeItemIndex+t+e.length)%e.length,e[this._activeItemIndex].disabled?this._setActiveInWrapMode(t,e):this.setActiveItem(this._activeItemIndex)},t.prototype._setActiveInDefaultMode=function(t,e){this._setActiveItemByIndex(this._activeItemIndex+t,t,e)},t.prototype._setActiveItemByIndex=function(t,e,n){if(void 0===n&&(n=this._items.toArray()),n[t]){for(;n[t].disabled;)if(!n[t+=e])return;this.setActiveItem(t)}},t}(),C=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(c.b)(e,t),e.prototype.setActiveItem=function(e){this.activeItem&&this.activeItem.setInactiveStyles(),t.prototype.setActiveItem.call(this,e),this.activeItem&&this.activeItem.setActiveStyles()},e}(O),k=" ";function E(t,e){return(t.getAttribute(e)||"").match(/\S+/g)||[]}var S="cdk-describedby-message",P="cdk-describedby-host",M=0,A=new Map,T=null,D=function(){function t(t){this._document=t}return t.prototype.describe=function(t,e){e.trim()&&(A.has(e)||this._createMessageElement(e),this._isElementDescribedByMessage(t,e)||this._addMessageReference(t,e))},t.prototype.removeDescription=function(t,e){if(e.trim()){this._isElementDescribedByMessage(t,e)&&this._removeMessageReference(t,e);var n=A.get(e);n&&0===n.referenceCount&&this._deleteMessageElement(e),T&&0===T.childNodes.length&&this._deleteMessagesContainer()}},t.prototype.ngOnDestroy=function(){for(var t=this._document.querySelectorAll("["+P+"]"),e=0;e<t.length;e++)this._removeCdkDescribedByReferenceIds(t[e]),t[e].removeAttribute(P);T&&this._deleteMessagesContainer(),A.clear()},t.prototype._createMessageElement=function(t){var e=this._document.createElement("div");e.setAttribute("id",S+"-"+M++),e.appendChild(this._document.createTextNode(t)),T||this._createMessagesContainer(),T.appendChild(e),A.set(t,{messageElement:e,referenceCount:0})},t.prototype._deleteMessageElement=function(t){var e=A.get(t),n=e&&e.messageElement;T&&n&&T.removeChild(n),A.delete(t)},t.prototype._createMessagesContainer=function(){(T=this._document.createElement("div")).setAttribute("id","cdk-describedby-message-container"),T.setAttribute("aria-hidden","true"),T.style.display="none",this._document.body.appendChild(T)},t.prototype._deleteMessagesContainer=function(){this._document.body.removeChild(T),T=null},t.prototype._removeCdkDescribedByReferenceIds=function(t){var e=E(t,"aria-describedby").filter(function(t){return 0!=t.indexOf(S)});t.setAttribute("aria-describedby",e.join(" "))},t.prototype._addMessageReference=function(t,e){var n,r,i,o,s=A.get(e);n=t,r="aria-describedby",i=s.messageElement.id,(o=E(n,r)).some(function(t){return t.trim()==i.trim()})||(o.push(i.trim()),n.setAttribute(r,o.join(k))),t.setAttribute(P,""),s.referenceCount++},t.prototype._removeMessageReference=function(t,e){var n,r,i,o,s=A.get(e);s.referenceCount--,n=t,r="aria-describedby",i=s.messageElement.id,o=E(n,r).filter(function(t){return t!=i.trim()}),n.setAttribute(r,o.join(k)),t.removeAttribute(P)},t.prototype._isElementDescribedByMessage=function(t,e){var n=E(t,"aria-describedby"),r=A.get(e),i=r&&r.messageElement.id;return!!i&&-1!=n.indexOf(i)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[a.b]}]}]},t}();var I={provide:D,deps:[[new r.M,new r._1,D],a.b],useFactory:function(t,e){return t||new D(e)}};function R(t){return 0===t.buttons}var N=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(c.b)(e,t),e.prototype.setActiveItem=function(e){t.prototype.setActiveItem.call(this,e),this.activeItem&&this.activeItem.focus()},e}(O),L=new r.A("liveAnnouncerElement"),F=function(){function t(t,e){this._document=e,this._liveElement=t||this._createLiveElement()}return t.prototype.announce=function(t,e){var n=this;void 0===e&&(e="polite"),this._liveElement.textContent="",this._liveElement.setAttribute("aria-live",e),setTimeout(function(){return n._liveElement.textContent=t},100)},t.prototype.ngOnDestroy=function(){this._liveElement&&this._liveElement.parentNode&&this._liveElement.parentNode.removeChild(this._liveElement)},t.prototype._createLiveElement=function(){var t=this._document.createElement("div");return t.classList.add("cdk-visually-hidden"),t.setAttribute("aria-atomic","true"),t.setAttribute("aria-live","polite"),this._document.body.appendChild(t),t},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.M},{type:r.y,args:[L]}]},{type:void 0,decorators:[{type:r.y,args:[a.b]}]}]},t}();var V={provide:F,deps:[[new r.M,new r._1,F],[new r.M,new r.y(L)],a.b],useFactory:function(t,e,n){return t||new F(e,n)}},z=function(){function t(t,e){this._ngZone=t,this._platform=e,this._origin=null,this._windowFocused=!1,this._elementInfo=new WeakMap,this._unregisterGlobalListeners=function(){},this._monitoredElementCount=0}return t.prototype.monitor=function(t,e,n){var i=this;if(e instanceof r.U||(n=e),n=!!n,!this._platform.isBrowser)return Object(g.a)(null);if(this._elementInfo.has(t)){var o=this._elementInfo.get(t);return o.checkChildren=n,o.subject.asObservable()}var s={unlisten:function(){},checkChildren:n,subject:new u.a};this._elementInfo.set(t,s),this._incrementMonitoredElementCount();var a=function(e){return i._onFocus(e,t)},c=function(e){return i._onBlur(e,t)};return this._ngZone.runOutsideAngular(function(){t.addEventListener("focus",a,!0),t.addEventListener("blur",c,!0)}),s.unlisten=function(){t.removeEventListener("focus",a,!0),t.removeEventListener("blur",c,!0)},s.subject.asObservable()},t.prototype.stopMonitoring=function(t){var e=this._elementInfo.get(t);e&&(e.unlisten(),e.subject.complete(),this._setClasses(t),this._elementInfo.delete(t),this._decrementMonitoredElementCount())},t.prototype.focusVia=function(t,e){this._setOriginForCurrentEventQueue(e),t.focus()},t.prototype._registerGlobalListeners=function(){var t=this;if(this._platform.isBrowser){var e=function(){t._lastTouchTarget=null,t._setOriginForCurrentEventQueue("keyboard")},n=function(){t._lastTouchTarget||t._setOriginForCurrentEventQueue("mouse")},r=function(e){null!=t._touchTimeout&&clearTimeout(t._touchTimeout),t._lastTouchTarget=e.target,t._touchTimeout=setTimeout(function(){return t._lastTouchTarget=null},650)},i=function(){t._windowFocused=!0,setTimeout(function(){return t._windowFocused=!1},0)};this._ngZone.runOutsideAngular(function(){document.addEventListener("keydown",e,!0),document.addEventListener("mousedown",n,!0),document.addEventListener("touchstart",r,!Object(s.d)()||{passive:!0,capture:!0}),window.addEventListener("focus",i)}),this._unregisterGlobalListeners=function(){document.removeEventListener("keydown",e,!0),document.removeEventListener("mousedown",n,!0),document.removeEventListener("touchstart",r,!Object(s.d)()||{passive:!0,capture:!0}),window.removeEventListener("focus",i)}}},t.prototype._toggleClass=function(t,e,n){n?t.classList.add(e):t.classList.remove(e)},t.prototype._setClasses=function(t,e){this._elementInfo.get(t)&&(this._toggleClass(t,"cdk-focused",!!e),this._toggleClass(t,"cdk-touch-focused","touch"===e),this._toggleClass(t,"cdk-keyboard-focused","keyboard"===e),this._toggleClass(t,"cdk-mouse-focused","mouse"===e),this._toggleClass(t,"cdk-program-focused","program"===e))},t.prototype._setOriginForCurrentEventQueue=function(t){var e=this;this._origin=t,setTimeout(function(){return e._origin=null},0)},t.prototype._wasCausedByTouch=function(t){var e=t.target;return this._lastTouchTarget instanceof Node&&e instanceof Node&&(e===this._lastTouchTarget||e.contains(this._lastTouchTarget))},t.prototype._onFocus=function(t,e){var n=this._elementInfo.get(e);n&&(n.checkChildren||e===t.target)&&(this._origin||(this._windowFocused&&this._lastFocusOrigin?this._origin=this._lastFocusOrigin:this._wasCausedByTouch(t)?this._origin="touch":this._origin="program"),this._setClasses(e,this._origin),n.subject.next(this._origin),this._lastFocusOrigin=this._origin,this._origin=null)},t.prototype._onBlur=function(t,e){var n=this._elementInfo.get(e);!n||n.checkChildren&&t.relatedTarget instanceof Node&&e.contains(t.relatedTarget)||(this._setClasses(e),n.subject.next(null))},t.prototype._incrementMonitoredElementCount=function(){1==++this._monitoredElementCount&&this._registerGlobalListeners()},t.prototype._decrementMonitoredElementCount=function(){--this._monitoredElementCount||(this._unregisterGlobalListeners(),this._unregisterGlobalListeners=function(){})},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:r.L},{type:s.a}]},t}(),B=function(){function t(t,e){var n=this;this._elementRef=t,this._focusMonitor=e,this.cdkFocusChange=new r.w,this._monitorSubscription=this._focusMonitor.monitor(this._elementRef.nativeElement,this._elementRef.nativeElement.hasAttribute("cdkMonitorSubtreeFocus")).subscribe(function(t){return n.cdkFocusChange.emit(t)})}return t.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef.nativeElement),this._monitorSubscription.unsubscribe()},t.decorators=[{type:r.s,args:[{selector:"[cdkMonitorElementFocus], [cdkMonitorSubtreeFocus]"}]}],t.ctorParameters=function(){return[{type:r.t},{type:z}]},t.propDecorators={cdkFocusChange:[{type:r.N}]},t}();var q={provide:z,deps:[[new r.M,new r._1,z],r.L,s.a],useFactory:function(t,e,n){return t||new z(e,n)}},H=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[a.a,s.b],declarations:[w,x,B],exports:[w,x,B],providers:[v,j,D,V,I,q]}]}],t.ctorParameters=function(){return[]},t}()},"../../../cdk/esm5/accordion.es5.js":function(t,e,n){"use strict";n.d(e,"b",function(){return u}),n.d(e,"a",function(){return a}),n.d(e,"c",function(){return l});var r=n("../../../core/esm5/core.js"),i=n("../../../cdk/esm5/collections.es5.js"),o=n("../../../cdk/esm5/coercion.es5.js"),s=0,a=function(){function t(){this.id="cdk-accordion-"+s++,this._multi=!1}return Object.defineProperty(t.prototype,"multi",{get:function(){return this._multi},set:function(t){this._multi=Object(o.b)(t)},enumerable:!0,configurable:!0}),t.decorators=[{type:r.s,args:[{selector:"cdk-accordion, [cdkAccordion]",exportAs:"cdkAccordion"}]}],t.ctorParameters=function(){return[]},t.propDecorators={multi:[{type:r.C}]},t}(),c=0,u=function(){function t(t,e,n){var i=this;this.accordion=t,this._changeDetectorRef=e,this._expansionDispatcher=n,this.closed=new r.w,this.opened=new r.w,this.destroyed=new r.w,this.id="cdk-accordion-child-"+c++,this._removeUniqueSelectionListener=function(){},this._removeUniqueSelectionListener=n.listen(function(t,e){i.accordion&&!i.accordion.multi&&i.accordion.id===e&&i.id!==t&&(i.expanded=!1)})}return Object.defineProperty(t.prototype,"expanded",{get:function(){return this._expanded},set:function(t){if(t=Object(o.b)(t),this._expanded!==t){if(this._expanded=t,t){this.opened.emit();var e=this.accordion?this.accordion.id:this.id;this._expansionDispatcher.notify(this.id,e)}else this.closed.emit();this._changeDetectorRef.markForCheck()}},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this.destroyed.emit(),this._removeUniqueSelectionListener()},t.prototype.toggle=function(){this.expanded=!this.expanded},t.prototype.close=function(){this.expanded=!1},t.prototype.open=function(){this.expanded=!0},t.decorators=[{type:r.s,args:[{selector:"cdk-accordion-item",exportAs:"cdkAccordionItem"}]}],t.ctorParameters=function(){return[{type:a,decorators:[{type:r.M}]},{type:r.j},{type:i.c}]},t.propDecorators={closed:[{type:r.N}],opened:[{type:r.N}],destroyed:[{type:r.N}],expanded:[{type:r.C}]},t}(),l=function(){function t(){}return t.decorators=[{type:r.H,args:[{exports:[a,u],declarations:[a,u],providers:[i.b]}]}],t.ctorParameters=function(){return[]},t}()},"../../../cdk/esm5/bidi.es5.js":function(t,e,n){"use strict";n.d(e,"b",function(){return s}),n.d(e,"a",function(){return c});var r=n("../../../core/esm5/core.js"),i=n("../../../common/esm5/common.js"),o=new r.A("cdk-dir-doc"),s=function(){function t(t){if(this.value="ltr",this.change=new r.w,t){var e=t.body?t.body.dir:null,n=t.documentElement?t.documentElement.dir:null;this.value=e||n||"ltr"}}return t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.M},{type:r.y,args:[o]}]}]},t}(),a=function(){function t(){this._dir="ltr",this._isInitialized=!1,this.change=new r.w}return Object.defineProperty(t.prototype,"dir",{get:function(){return this._dir},set:function(t){var e=this._dir;this._dir=t,e!==this._dir&&this._isInitialized&&this.change.emit(this._dir)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this.dir},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._isInitialized=!0},t.decorators=[{type:r.s,args:[{selector:"[dir]",providers:[{provide:s,useExisting:t}],host:{"[dir]":"dir"},exportAs:"dir"}]}],t.ctorParameters=function(){return[]},t.propDecorators={change:[{type:r.N,args:["dirChange"]}],dir:[{type:r.C,args:["dir"]}]},t}(),c=function(){function t(){}return t.decorators=[{type:r.H,args:[{exports:[a],declarations:[a],providers:[{provide:o,useExisting:i.b},s]}]}],t.ctorParameters=function(){return[]},t}()},"../../../cdk/esm5/coercion.es5.js":function(t,e,n){"use strict";function r(t){return null!=t&&""+t!="false"}function i(t,e){return void 0===e&&(e=0),isNaN(parseFloat(t))||isNaN(Number(t))?e:Number(t)}function o(t){return Array.isArray(t)?t:[t]}n.d(e,"b",function(){return r}),n.d(e,"c",function(){return i}),n.d(e,"a",function(){return o})},"../../../cdk/esm5/collections.es5.js":function(t,e,n){"use strict";n.d(e,"c",function(){return a}),n.d(e,"b",function(){return c}),n.d(e,"a",function(){return o});var r=n("../../../../rxjs/_esm5/Subject.js"),i=n("../../../core/esm5/core.js"),o=(function(){}(),function(){function t(t,e,n){void 0===t&&(t=!1),void 0===n&&(n=!0);var i=this;this._isMulti=t,this._emitChanges=n,this._selection=new Set,this._deselectedToEmit=[],this._selectedToEmit=[],this.onChange=this._emitChanges?new r.a:null,e&&(t?e.forEach(function(t){return i._markSelected(t)}):this._markSelected(e[0]),this._selectedToEmit.length=0)}return Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected||(this._selected=Array.from(this._selection.values())),this._selected},enumerable:!0,configurable:!0}),t.prototype.select=function(){for(var t=this,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];this._verifyValueAssignment(e),e.forEach(function(e){return t._markSelected(e)}),this._emitChangeEvent()},t.prototype.deselect=function(){for(var t=this,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];this._verifyValueAssignment(e),e.forEach(function(e){return t._unmarkSelected(e)}),this._emitChangeEvent()},t.prototype.toggle=function(t){this.isSelected(t)?this.deselect(t):this.select(t)},t.prototype.clear=function(){this._unmarkAll(),this._emitChangeEvent()},t.prototype.isSelected=function(t){return this._selection.has(t)},t.prototype.isEmpty=function(){return 0===this._selection.size},t.prototype.hasValue=function(){return!this.isEmpty()},t.prototype.sort=function(t){this._isMulti&&this._selected&&this._selected.sort(t)},t.prototype._emitChangeEvent=function(){if(this._selected=null,this._selectedToEmit.length||this._deselectedToEmit.length){var t=new s(this._selectedToEmit,this._deselectedToEmit);this.onChange&&this.onChange.next(t),this._deselectedToEmit=[],this._selectedToEmit=[]}},t.prototype._markSelected=function(t){this.isSelected(t)||(this._isMulti||this._unmarkAll(),this._selection.add(t),this._emitChanges&&this._selectedToEmit.push(t))},t.prototype._unmarkSelected=function(t){this.isSelected(t)&&(this._selection.delete(t),this._emitChanges&&this._deselectedToEmit.push(t))},t.prototype._unmarkAll=function(){var t=this;this.isEmpty()||this._selection.forEach(function(e){return t._unmarkSelected(e)})},t.prototype._verifyValueAssignment=function(t){if(t.length>1&&!this._isMulti)throw Error("Cannot pass multiple values into SelectionModel with single-value mode.")},t}()),s=function(){return function(t,e){this.added=t,this.removed=e}}();var a=function(){function t(){this._listeners=[]}return t.prototype.notify=function(t,e){for(var n=0,r=this._listeners;n<r.length;n++){(0,r[n])(t,e)}},t.prototype.listen=function(t){var e=this;return this._listeners.push(t),function(){e._listeners=e._listeners.filter(function(e){return t!==e})}},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[]},t}();var c={provide:a,deps:[[new i.M,new i._1,a]],useFactory:function(t){return t||new a}}},"../../../cdk/esm5/keycodes.es5.js":function(t,e,n){"use strict";n.d(e,"p",function(){return r}),n.d(e,"d",function(){return i}),n.d(e,"m",function(){return o}),n.d(e,"i",function(){return s}),n.d(e,"l",function(){return a}),n.d(e,"k",function(){return c}),n.d(e,"h",function(){return u}),n.d(e,"e",function(){return l}),n.d(e,"f",function(){return p}),n.d(e,"n",function(){return d}),n.d(e,"o",function(){return h}),n.d(e,"g",function(){return f}),n.d(e,"b",function(){return m}),n.d(e,"c",function(){return g}),n.d(e,"a",function(){return v}),n.d(e,"q",function(){return y}),n.d(e,"r",function(){return b}),n.d(e,"j",function(){return _});var r=38,i=40,o=39,s=37,a=33,c=34,u=36,l=35,p=13,d=32,h=9,f=27,m=8,g=46,v=65,y=90,b=48,_=91},"../../../cdk/esm5/layout.es5.js":function(t,e,n){"use strict";n.d(e,"c",function(){return v}),n.d(e,"a",function(){return m}),n.d(e,"b",function(){return g});var r=n("../../../core/esm5/core.js"),i=n("../../../cdk/esm5/platform.es5.js"),o=n("../../../../rxjs/_esm5/Subject.js"),s=n("../../../../rxjs/_esm5/operators/map.js"),a=n("../../../../rxjs/_esm5/operators/startWith.js"),c=n("../../../../rxjs/_esm5/operators/takeUntil.js"),u=n("../../../cdk/esm5/coercion.es5.js"),l=n("../../../../rxjs/_esm5/observable/combineLatest.js"),p=n("../../../../rxjs/_esm5/observable/fromEventPattern.js"),d=new Map,h=function(){function t(t){this.platform=t,this._matchMedia=this.platform.isBrowser?window.matchMedia.bind(window):f}return t.prototype.matchMedia=function(t){return this.platform.WEBKIT&&function(t){if(!d.has(t))try{var e=document.createElement("style");if(e.setAttribute("type","text/css"),!e.sheet){var n="@media "+t+" {.fx-query-test{ }}";e.appendChild(document.createTextNode(n))}document.getElementsByTagName("head")[0].appendChild(e),d.set(t,e)}catch(t){console.error(t)}}(t),this._matchMedia(t)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:i.a}]},t}();function f(t){return{matches:"all"===t||""===t,media:t,addListener:function(){},removeListener:function(){}}}var m=function(){function t(t,e){this.mediaMatcher=t,this.zone=e,this._queries=new Map,this._destroySubject=new o.a}return t.prototype.ngOnDestroy=function(){this._destroySubject.next(),this._destroySubject.complete()},t.prototype.isMatched=function(t){var e=this;return Object(u.a)(t).some(function(t){return e._registerQuery(t).mql.matches})},t.prototype.observe=function(t){var e=this,n=Object(u.a)(t).map(function(t){return e._registerQuery(t).observable});return Object(l.a)(n,function(t,e){return{matches:!!(t&&t.matches||e&&e.matches)}})},t.prototype._registerQuery=function(t){var e=this;if(this._queries.has(t))return this._queries.get(t);var n=this.mediaMatcher.matchMedia(t),r={observable:Object(p.a)(function(t){n.addListener(function(n){return e.zone.run(function(){return t(n)})})},function(t){n.removeListener(function(n){return e.zone.run(function(){return t(n)})})}).pipe(Object(c.a)(this._destroySubject),Object(a.a)(n),Object(s.a)(function(t){return{matches:t.matches}})),mql:n};return this._queries.set(t,r),r},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:h},{type:r.L}]},t}(),g={Handset:"(max-width: 599px) and (orientation: portrait), (max-width: 959px) and (orientation: landscape)",Tablet:"(min-width: 600px) and (max-width: 839px) and (orientation: portrait), (min-width: 960px) and (max-width: 1279px) and (orientation: landscape)",Web:"(min-width: 840px) and (orientation: portrait), (min-width: 1280px) and (orientation: landscape)",HandsetPortrait:"(max-width: 599px) and (orientation: portrait)",TabletPortrait:"(min-width: 600px) and (max-width: 839px) and (orientation: portrait)",WebPortrait:"(min-width: 840px) and (orientation: portrait)",HandsetLandscape:"(max-width: 959px) and (orientation: landscape)",TabletLandscape:"(min-width: 960px) and (max-width: 1279px) and (orientation: landscape)",WebLandscape:"(min-width: 1280px) and (orientation: landscape)"},v=function(){function t(){}return t.decorators=[{type:r.H,args:[{providers:[m,h],imports:[i.b]}]}],t.ctorParameters=function(){return[]},t}()},"../../../cdk/esm5/observers.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return c});var r=n("../../../core/esm5/core.js"),i=n("../../../../rxjs/_esm5/Subject.js"),o=n("../../../../rxjs/_esm5/operators/debounceTime.js"),s=function(){function t(){}return t.prototype.create=function(t){return"undefined"==typeof MutationObserver?null:new MutationObserver(t)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[]},t}(),a=function(){function t(t,e,n){this._mutationObserverFactory=t,this._elementRef=e,this._ngZone=n,this.event=new r.w,this._debouncer=new i.a}return t.prototype.ngAfterContentInit=function(){var t=this;this.debounce>0?this._ngZone.runOutsideAngular(function(){t._debouncer.pipe(Object(o.a)(t.debounce)).subscribe(function(e){return t.event.emit(e)})}):this._debouncer.subscribe(function(e){return t.event.emit(e)}),this._observer=this._ngZone.runOutsideAngular(function(){return t._mutationObserverFactory.create(function(e){t._debouncer.next(e)})}),this._observer&&this._observer.observe(this._elementRef.nativeElement,{characterData:!0,childList:!0,subtree:!0})},t.prototype.ngOnDestroy=function(){this._observer&&this._observer.disconnect(),this._debouncer.complete()},t.decorators=[{type:r.s,args:[{selector:"[cdkObserveContent]",exportAs:"cdkObserveContent"}]}],t.ctorParameters=function(){return[{type:s},{type:r.t},{type:r.L}]},t.propDecorators={event:[{type:r.N,args:["cdkObserveContent"]}],debounce:[{type:r.C}]},t}(),c=function(){function t(){}return t.decorators=[{type:r.H,args:[{exports:[a],declarations:[a],providers:[s]}]}],t.ctorParameters=function(){return[]},t}()},"../../../cdk/esm5/overlay.es5.js":function(t,e,n){"use strict";n.d(e,"b",function(){return V}),n.d(e,"a",function(){return U}),n.d(e,"c",function(){return v}),n.d(e,"d",function(){return G});var r=n("../../../core/esm5/core.js"),i=n("../../../cdk/esm5/scrolling.es5.js"),o=n("../../../cdk/esm5/bidi.es5.js"),s=n("../../../cdk/esm5/portal.es5.js"),a=n("../../../../tslib/tslib.es6.js"),c=n("../../../../rxjs/_esm5/operators/take.js"),u=n("../../../../rxjs/_esm5/Subject.js"),l=n("../../../../rxjs/_esm5/Subscription.js"),p=n("../../../common/esm5/common.js"),d=n("../../../../rxjs/_esm5/operators/filter.js"),h=n("../../../../rxjs/_esm5/observable/fromEvent.js"),f=n("../../../cdk/esm5/coercion.es5.js"),m=n("../../../cdk/esm5/keycodes.es5.js");n.d(e,"f",function(){return i.d}),n.d(e,"e",function(){return i.c});var g=function(){function t(){}return t.prototype.enable=function(){},t.prototype.disable=function(){},t.prototype.attach=function(){},t}(),v=function(){return function(t){var e=this;this.scrollStrategy=new g,this.panelClass="",this.hasBackdrop=!1,this.backdropClass="cdk-overlay-dark-backdrop",this.direction="ltr",t&&Object.keys(t).forEach(function(n){return e[n]=t[n]})}}(),y=function(){return function(t,e,n,r){this.offsetX=n,this.offsetY=r,this.originX=t.originX,this.originY=t.originY,this.overlayX=e.overlayX,this.overlayY=e.overlayY}}(),b=function(){return function(){}}(),_=function(){function t(t,e){this.connectionPair=t,this.scrollableViewProperties=e}return t.ctorParameters=function(){return[{type:y},{type:b,decorators:[{type:r.M}]}]},t}();function j(){return Error("Scroll strategy has already been attached.")}var x=function(){function t(t,e){this._scrollDispatcher=t,this._ngZone=e,this._scrollSubscription=null}return t.prototype.attach=function(t){if(this._overlayRef)throw j();this._overlayRef=t},t.prototype.enable=function(){var t=this;this._scrollSubscription||(this._scrollSubscription=this._scrollDispatcher.scrolled(0).subscribe(function(){t._ngZone.run(function(){t.disable(),t._overlayRef.hasAttached()&&t._overlayRef.detach()})}))},t.prototype.disable=function(){this._scrollSubscription&&(this._scrollSubscription.unsubscribe(),this._scrollSubscription=null)},t}(),w=function(){function t(t){this._viewportRuler=t,this._previousHTMLStyles={top:"",left:""},this._isEnabled=!1}return t.prototype.attach=function(){},t.prototype.enable=function(){if(this._canBeEnabled()){var t=document.documentElement;this._previousScrollPosition=this._viewportRuler.getViewportScrollPosition(),this._previousHTMLStyles.left=t.style.left||"",this._previousHTMLStyles.top=t.style.top||"",t.style.left=-this._previousScrollPosition.left+"px",t.style.top=-this._previousScrollPosition.top+"px",t.classList.add("cdk-global-scrollblock"),this._isEnabled=!0}},t.prototype.disable=function(){if(this._isEnabled){var t=document.documentElement,e=document.body,n=t.style.scrollBehavior||"",r=e.style.scrollBehavior||"";this._isEnabled=!1,t.style.left=this._previousHTMLStyles.left,t.style.top=this._previousHTMLStyles.top,t.classList.remove("cdk-global-scrollblock"),t.style.scrollBehavior=e.style.scrollBehavior="auto",window.scroll(this._previousScrollPosition.left,this._previousScrollPosition.top),t.style.scrollBehavior=n,e.style.scrollBehavior=r}},t.prototype._canBeEnabled=function(){if(document.documentElement.classList.contains("cdk-global-scrollblock")||this._isEnabled)return!1;var t=document.body,e=this._viewportRuler.getViewportSize();return t.scrollHeight>e.height||t.scrollWidth>e.width},t}();function O(t,e){return e.some(function(e){var n=t.bottom<e.top,r=t.top>e.bottom,i=t.right<e.left,o=t.left>e.right;return n||r||i||o})}function C(t,e){return e.some(function(e){var n=t.top<e.top,r=t.bottom>e.bottom,i=t.left<e.left,o=t.right>e.right;return n||r||i||o})}var k=function(){function t(t,e,n,r){this._scrollDispatcher=t,this._viewportRuler=e,this._ngZone=n,this._config=r,this._scrollSubscription=null}return t.prototype.attach=function(t){if(this._overlayRef)throw j();this._overlayRef=t},t.prototype.enable=function(){var t=this;if(!this._scrollSubscription){var e=this._config?this._config.scrollThrottle:0;this._scrollSubscription=this._scrollDispatcher.scrolled(e).subscribe(function(){if(t._overlayRef.updatePosition(),t._config&&t._config.autoClose){var e=t._overlayRef.overlayElement.getBoundingClientRect(),n=t._viewportRuler.getViewportSize(),r=n.width,i=n.height;O(e,[{width:r,height:i,bottom:i,right:r,top:0,left:0}])&&(t.disable(),t._ngZone.run(function(){return t._overlayRef.detach()}))}})}},t.prototype.disable=function(){this._scrollSubscription&&(this._scrollSubscription.unsubscribe(),this._scrollSubscription=null)},t}(),E=function(){function t(t,e,n){var r=this;this._scrollDispatcher=t,this._viewportRuler=e,this._ngZone=n,this.noop=function(){return new g},this.close=function(){return new x(r._scrollDispatcher,r._ngZone)},this.block=function(){return new w(r._viewportRuler)},this.reposition=function(t){return new k(r._scrollDispatcher,r._viewportRuler,r._ngZone,t)}}return t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:i.b},{type:i.d},{type:r.L}]},t}(),S=function(){function t(t,e,n,r,i){this._portalOutlet=t,this._pane=e,this._config=n,this._ngZone=r,this._keyboardDispatcher=i,this._backdropElement=null,this._backdropClick=new u.a,this._attachments=new u.a,this._detachments=new u.a,this._keydownEvents=new u.a,n.scrollStrategy&&n.scrollStrategy.attach(this)}return Object.defineProperty(t.prototype,"overlayElement",{get:function(){return this._pane},enumerable:!0,configurable:!0}),t.prototype.attach=function(t){var e=this,n=this._portalOutlet.attach(t);return this._config.positionStrategy&&this._config.positionStrategy.attach(this),this._updateStackingOrder(),this._updateElementSize(),this._updateElementDirection(),this._config.scrollStrategy&&this._config.scrollStrategy.enable(),this._ngZone.onStable.asObservable().pipe(Object(c.a)(1)).subscribe(function(){e.updatePosition()}),this._togglePointerEvents(!0),this._config.hasBackdrop&&this._attachBackdrop(),this._config.panelClass&&(Array.isArray(this._config.panelClass)?this._config.panelClass.forEach(function(t){return e._pane.classList.add(t)}):this._pane.classList.add(this._config.panelClass)),this._attachments.next(),this._keyboardDispatcher.add(this),n},t.prototype.detach=function(){if(this.hasAttached()){this.detachBackdrop(),this._togglePointerEvents(!1),this._config.positionStrategy&&this._config.positionStrategy.detach&&this._config.positionStrategy.detach(),this._config.scrollStrategy&&this._config.scrollStrategy.disable();var t=this._portalOutlet.detach();return this._detachments.next(),this._keyboardDispatcher.remove(this),t}},t.prototype.dispose=function(){var t=this.hasAttached();this._config.positionStrategy&&this._config.positionStrategy.dispose(),this._config.scrollStrategy&&this._config.scrollStrategy.disable(),this.detachBackdrop(),this._keyboardDispatcher.remove(this),this._portalOutlet.dispose(),this._attachments.complete(),this._backdropClick.complete(),this._keydownEvents.complete(),t&&this._detachments.next(),this._detachments.complete()},t.prototype.hasAttached=function(){return this._portalOutlet.hasAttached()},t.prototype.backdropClick=function(){return this._backdropClick.asObservable()},t.prototype.attachments=function(){return this._attachments.asObservable()},t.prototype.detachments=function(){return this._detachments.asObservable()},t.prototype.keydownEvents=function(){return this._keydownEvents.asObservable()},t.prototype.getConfig=function(){return this._config},t.prototype.updatePosition=function(){this._config.positionStrategy&&this._config.positionStrategy.apply()},t.prototype.updateSize=function(t){this._config=Object(a.a)({},this._config,t),this._updateElementSize()},t.prototype.setDirection=function(t){this._config=Object(a.a)({},this._config,{direction:t}),this._updateElementDirection()},t.prototype._updateElementDirection=function(){this._pane.setAttribute("dir",this._config.direction)},t.prototype._updateElementSize=function(){(this._config.width||0===this._config.width)&&(this._pane.style.width=P(this._config.width)),(this._config.height||0===this._config.height)&&(this._pane.style.height=P(this._config.height)),(this._config.minWidth||0===this._config.minWidth)&&(this._pane.style.minWidth=P(this._config.minWidth)),(this._config.minHeight||0===this._config.minHeight)&&(this._pane.style.minHeight=P(this._config.minHeight)),(this._config.maxWidth||0===this._config.maxWidth)&&(this._pane.style.maxWidth=P(this._config.maxWidth)),(this._config.maxHeight||0===this._config.maxHeight)&&(this._pane.style.maxHeight=P(this._config.maxHeight))},t.prototype._togglePointerEvents=function(t){this._pane.style.pointerEvents=t?"auto":"none"},t.prototype._attachBackdrop=function(){var t=this;this._backdropElement=document.createElement("div"),this._backdropElement.classList.add("cdk-overlay-backdrop"),this._config.backdropClass&&this._backdropElement.classList.add(this._config.backdropClass),this._pane.parentElement.insertBefore(this._backdropElement,this._pane),this._backdropElement.addEventListener("click",function(){return t._backdropClick.next(null)}),this._ngZone.runOutsideAngular(function(){requestAnimationFrame(function(){t._backdropElement&&t._backdropElement.classList.add("cdk-overlay-backdrop-showing")})})},t.prototype._updateStackingOrder=function(){this._pane.nextSibling&&this._pane.parentNode.appendChild(this._pane)},t.prototype.detachBackdrop=function(){var t=this,e=this._backdropElement;if(e){var n=function(){e&&e.parentNode&&e.parentNode.removeChild(e),t._backdropElement==e&&(t._backdropElement=null)};e.classList.remove("cdk-overlay-backdrop-showing"),this._config.backdropClass&&e.classList.remove(this._config.backdropClass),e.addEventListener("transitionend",n),e.style.pointerEvents="none",this._ngZone.runOutsideAngular(function(){setTimeout(n,500)})}},t}();function P(t){return"string"==typeof t?t:t+"px"}var M=function(){function t(t,e,n,r,i){this._connectedTo=n,this._viewportRuler=r,this._document=i,this._dir="ltr",this._offsetX=0,this._offsetY=0,this.scrollables=[],this._resizeSubscription=l.a.EMPTY,this._preferredPositions=[],this._applied=!1,this._positionLocked=!1,this._onPositionChange=new u.a,this._origin=this._connectedTo.nativeElement,this.withFallbackPosition(t,e)}return Object.defineProperty(t.prototype,"_isRtl",{get:function(){return"rtl"===this._dir},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"onPositionChange",{get:function(){return this._onPositionChange.asObservable()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"positions",{get:function(){return this._preferredPositions},enumerable:!0,configurable:!0}),t.prototype.attach=function(t){var e=this;this._overlayRef=t,this._pane=t.overlayElement,this._resizeSubscription.unsubscribe(),this._resizeSubscription=this._viewportRuler.change().subscribe(function(){return e.apply()})},t.prototype.dispose=function(){this._applied=!1,this._resizeSubscription.unsubscribe()},t.prototype.detach=function(){this._applied=!1,this._resizeSubscription.unsubscribe()},t.prototype.apply=function(){if(this._applied&&this._positionLocked&&this._lastConnectedPosition)this.recalculateLastPosition();else{this._applied=!0;for(var t,e,n=this._pane,r=this._origin.getBoundingClientRect(),i=n.getBoundingClientRect(),o=this._viewportRuler.getViewportSize(),s=0,a=this._preferredPositions;s<a.length;s++){var c=a[s],u=this._getOriginConnectionPoint(r,c),l=this._getOverlayPoint(u,i,o,c);if(l.fitsInViewport)return this._setElementPosition(n,i,l,c),void(this._lastConnectedPosition=c);(!t||t.visibleArea<l.visibleArea)&&(t=l,e=c)}this._setElementPosition(n,i,t,e)}},t.prototype.recalculateLastPosition=function(){if(this._lastConnectedPosition){var t=this._origin.getBoundingClientRect(),e=this._pane.getBoundingClientRect(),n=this._viewportRuler.getViewportSize(),r=this._lastConnectedPosition||this._preferredPositions[0],i=this._getOriginConnectionPoint(t,r),o=this._getOverlayPoint(i,e,n,r);this._setElementPosition(this._pane,e,o,r)}},t.prototype.withScrollableContainers=function(t){this.scrollables=t},t.prototype.withFallbackPosition=function(t,e,n,r){var i=new y(t,e,n,r);return this._preferredPositions.push(i),this},t.prototype.withDirection=function(t){return this._dir=t,this},t.prototype.withOffsetX=function(t){return this._offsetX=t,this},t.prototype.withOffsetY=function(t){return this._offsetY=t,this},t.prototype.withLockedPosition=function(t){return this._positionLocked=t,this},t.prototype._getStartX=function(t){return this._isRtl?t.right:t.left},t.prototype._getEndX=function(t){return this._isRtl?t.left:t.right},t.prototype._getOriginConnectionPoint=function(t,e){var n=this._getStartX(t),r=this._getEndX(t);return{x:"center"==e.originX?n+t.width/2:"start"==e.originX?n:r,y:"center"==e.originY?t.top+t.height/2:"top"==e.originY?t.top:t.bottom}},t.prototype._getOverlayPoint=function(t,e,n,r){var i,o;i="center"==r.overlayX?-e.width/2:"start"===r.overlayX?this._isRtl?-e.width:0:this._isRtl?0:-e.width,o="center"==r.overlayY?-e.height/2:"top"==r.overlayY?0:-e.height;var s=void 0===r.offsetX?this._offsetX:r.offsetX,a=void 0===r.offsetY?this._offsetY:r.offsetY,c=t.x+i+s,u=t.y+o+a,l=0-c,p=c+e.width-n.width,d=0-u,h=u+e.height-n.height,f=this._subtractOverflows(e.width,l,p)*this._subtractOverflows(e.height,d,h);return{x:c,y:u,fitsInViewport:e.width*e.height===f,visibleArea:f}},t.prototype._getScrollVisibility=function(t){var e=this._origin.getBoundingClientRect(),n=t.getBoundingClientRect(),r=this.scrollables.map(function(t){return t.getElementRef().nativeElement.getBoundingClientRect()});return{isOriginClipped:C(e,r),isOriginOutsideView:O(e,r),isOverlayClipped:C(n,r),isOverlayOutsideView:O(n,r)}},t.prototype._setElementPosition=function(t,e,n,r){var i,o="bottom"===r.overlayY?"bottom":"top",s="top"===o?n.y:this._document.documentElement.clientHeight-(n.y+e.height),a="left"===(i="rtl"===this._dir?"end"===r.overlayX?"left":"right":"end"===r.overlayX?"right":"left")?n.x:this._document.documentElement.clientWidth-(n.x+e.width);["top","bottom","left","right"].forEach(function(e){return t.style[e]=null}),t.style[o]=s+"px",t.style[i]=a+"px";var c=this._getScrollVisibility(t),u=new _(r,c);this._onPositionChange.next(u)},t.prototype._subtractOverflows=function(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return e.reduce(function(t,e){return t-Math.max(e,0)},t)},t}(),A=function(){function t(t){this._document=t,this._cssPosition="static",this._topOffset="",this._bottomOffset="",this._leftOffset="",this._rightOffset="",this._alignItems="",this._justifyContent="",this._width="",this._height="",this._wrapper=null}return t.prototype.attach=function(t){this._overlayRef=t},t.prototype.top=function(t){return void 0===t&&(t=""),this._bottomOffset="",this._topOffset=t,this._alignItems="flex-start",this},t.prototype.left=function(t){return void 0===t&&(t=""),this._rightOffset="",this._leftOffset=t,this._justifyContent="flex-start",this},t.prototype.bottom=function(t){return void 0===t&&(t=""),this._topOffset="",this._bottomOffset=t,this._alignItems="flex-end",this},t.prototype.right=function(t){return void 0===t&&(t=""),this._leftOffset="",this._rightOffset=t,this._justifyContent="flex-end",this},t.prototype.width=function(t){return void 0===t&&(t=""),this._width=t,"100%"===t&&this.left("0px"),this},t.prototype.height=function(t){return void 0===t&&(t=""),this._height=t,"100%"===t&&this.top("0px"),this},t.prototype.centerHorizontally=function(t){return void 0===t&&(t=""),this.left(t),this._justifyContent="center",this},t.prototype.centerVertically=function(t){return void 0===t&&(t=""),this.top(t),this._alignItems="center",this},t.prototype.apply=function(){var t=this._overlayRef.overlayElement;!this._wrapper&&t.parentNode&&(this._wrapper=this._document.createElement("div"),this._wrapper.classList.add("cdk-global-overlay-wrapper"),t.parentNode.insertBefore(this._wrapper,t),this._wrapper.appendChild(t));var e=t.style,n=t.parentNode.style;e.position=this._cssPosition,e.marginTop=this._topOffset,e.marginLeft=this._leftOffset,e.marginBottom=this._bottomOffset,e.marginRight=this._rightOffset,e.width=this._width,e.height=this._height,n.justifyContent=this._justifyContent,n.alignItems=this._alignItems},t.prototype.dispose=function(){this._wrapper&&this._wrapper.parentNode&&(this._wrapper.parentNode.removeChild(this._wrapper),this._wrapper=null)},t}(),T=function(){function t(t,e){this._viewportRuler=t,this._document=e}return t.prototype.global=function(){return new A(this._document)},t.prototype.connectedTo=function(t,e,n){return new M(e,n,t,this._viewportRuler,this._document)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:i.d},{type:void 0,decorators:[{type:r.y,args:[p.b]}]}]},t}(),D=function(){function t(t){this._document=t,this._attachedOverlays=[]}return t.prototype.ngOnDestroy=function(){this._unsubscribeFromKeydownEvents()},t.prototype.add=function(t){this._keydownEventSubscription||this._subscribeToKeydownEvents(),this._attachedOverlays.push(t)},t.prototype.remove=function(t){var e=this._attachedOverlays.indexOf(t);e>-1&&this._attachedOverlays.splice(e,1),0===this._attachedOverlays.length&&this._unsubscribeFromKeydownEvents()},t.prototype._subscribeToKeydownEvents=function(){var t=this,e=Object(h.a)(this._document.body,"keydown");this._keydownEventSubscription=e.pipe(Object(d.a)(function(){return!!t._attachedOverlays.length})).subscribe(function(e){t._selectOverlayFromEvent(e)._keydownEvents.next(e)})},t.prototype._unsubscribeFromKeydownEvents=function(){this._keydownEventSubscription&&(this._keydownEventSubscription.unsubscribe(),this._keydownEventSubscription=null)},t.prototype._selectOverlayFromEvent=function(t){return this._attachedOverlays.find(function(e){return e.overlayElement===t.target||e.overlayElement.contains(t.target)})||this._attachedOverlays[this._attachedOverlays.length-1]},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[p.b]}]}]},t}();var I={provide:D,deps:[[new r.M,new r._1,D],p.b],useFactory:function(t,e){return t||new D(e)}},R=function(){function t(t){this._document=t}return t.prototype.ngOnDestroy=function(){this._containerElement&&this._containerElement.parentNode&&this._containerElement.parentNode.removeChild(this._containerElement)},t.prototype.getContainerElement=function(){return this._containerElement||this._createContainer(),this._containerElement},t.prototype._createContainer=function(){var t=this._document.createElement("div");t.classList.add("cdk-overlay-container"),this._document.body.appendChild(t),this._containerElement=t},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[p.b]}]}]},t}();var N={provide:R,deps:[[new r.M,new r._1,R],p.b],useFactory:function(t,e){return t||new R(e)}},L=0,F=new v,V=function(){function t(t,e,n,r,i,o,s,a,c){this.scrollStrategies=t,this._overlayContainer=e,this._componentFactoryResolver=n,this._positionBuilder=r,this._keyboardDispatcher=i,this._appRef=o,this._injector=s,this._ngZone=a,this._document=c}return t.prototype.create=function(t){void 0===t&&(t=F);var e=this._createPaneElement(),n=this._createPortalOutlet(e);return new S(n,e,t,this._ngZone,this._keyboardDispatcher)},t.prototype.position=function(){return this._positionBuilder},t.prototype._createPaneElement=function(){var t=this._document.createElement("div");return t.id="cdk-overlay-"+L++,t.classList.add("cdk-overlay-pane"),this._overlayContainer.getContainerElement().appendChild(t),t},t.prototype._createPortalOutlet=function(t){return new s.e(t,this._componentFactoryResolver,this._appRef,this._injector)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:E},{type:R},{type:r.o},{type:T},{type:D},{type:r.f},{type:r.B},{type:r.L},{type:void 0,decorators:[{type:r.y,args:[p.b]}]}]},t}(),z=[new y({originX:"start",originY:"bottom"},{overlayX:"start",overlayY:"top"}),new y({originX:"start",originY:"top"},{overlayX:"start",overlayY:"bottom"}),new y({originX:"end",originY:"top"},{overlayX:"end",overlayY:"bottom"}),new y({originX:"end",originY:"bottom"},{overlayX:"end",overlayY:"top"})],B=new r.A("cdk-connected-overlay-scroll-strategy");var q={provide:B,deps:[V],useFactory:function(t){return function(){return t.scrollStrategies.reposition()}}},H=function(){function t(t){this.elementRef=t}return t.decorators=[{type:r.s,args:[{selector:"[cdk-overlay-origin], [overlay-origin], [cdkOverlayOrigin]",exportAs:"cdkOverlayOrigin"}]}],t.ctorParameters=function(){return[{type:r.t}]},t}(),U=function(){function t(t,e,n,i,o,a){var c=this;this._overlay=t,this._scrollStrategy=i,this._dir=o,this._document=a,this._hasBackdrop=!1,this._backdropSubscription=l.a.EMPTY,this._positionSubscription=l.a.EMPTY,this._offsetX=0,this._offsetY=0,this.scrollStrategy=this._scrollStrategy(),this.open=!1,this.backdropClick=new r.w,this.positionChange=new r.w,this.attach=new r.w,this.detach=new r.w,this._escapeListener=function(t){t.keyCode===m.g&&c._detachOverlay()},this._templatePortal=new s.h(e,n)}return Object.defineProperty(t.prototype,"offsetX",{get:function(){return this._offsetX},set:function(t){this._offsetX=t,this._position&&this._position.withOffsetX(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"offsetY",{get:function(){return this._offsetY},set:function(t){this._offsetY=t,this._position&&this._position.withOffsetY(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hasBackdrop",{get:function(){return this._hasBackdrop},set:function(t){this._hasBackdrop=Object(f.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedOrigin",{get:function(){return this.origin},set:function(t){this.origin=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedPositions",{get:function(){return this.positions},set:function(t){this.positions=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedOffsetX",{get:function(){return this.offsetX},set:function(t){this.offsetX=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedOffsetY",{get:function(){return this.offsetY},set:function(t){this.offsetY=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedWidth",{get:function(){return this.width},set:function(t){this.width=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedHeight",{get:function(){return this.height},set:function(t){this.height=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedMinWidth",{get:function(){return this.minWidth},set:function(t){this.minWidth=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedMinHeight",{get:function(){return this.minHeight},set:function(t){this.minHeight=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedBackdropClass",{get:function(){return this.backdropClass},set:function(t){this.backdropClass=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedScrollStrategy",{get:function(){return this.scrollStrategy},set:function(t){this.scrollStrategy=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedOpen",{get:function(){return this.open},set:function(t){this.open=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_deprecatedHasBackdrop",{get:function(){return this.hasBackdrop},set:function(t){this.hasBackdrop=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"overlayRef",{get:function(){return this._overlayRef},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dir",{get:function(){return this._dir?this._dir.value:"ltr"},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this._destroyOverlay()},t.prototype.ngOnChanges=function(t){(t.open||t._deprecatedOpen)&&(this.open?this._attachOverlay():this._detachOverlay())},t.prototype._createOverlay=function(){this.positions&&this.positions.length||(this.positions=z),this._overlayRef=this._overlay.create(this._buildConfig())},t.prototype._buildConfig=function(){var t=this._position=this._createPositionStrategy(),e=new v({positionStrategy:t,scrollStrategy:this.scrollStrategy,hasBackdrop:this.hasBackdrop});return(this.width||0===this.width)&&(e.width=this.width),(this.height||0===this.height)&&(e.height=this.height),(this.minWidth||0===this.minWidth)&&(e.minWidth=this.minWidth),(this.minHeight||0===this.minHeight)&&(e.minHeight=this.minHeight),this.backdropClass&&(e.backdropClass=this.backdropClass),e},t.prototype._createPositionStrategy=function(){var t=this.positions[0],e={originX:t.originX,originY:t.originY},n={overlayX:t.overlayX,overlayY:t.overlayY},r=this._overlay.position().connectedTo(this.origin.elementRef,e,n).withOffsetX(this.offsetX).withOffsetY(this.offsetY);return this._handlePositionChanges(r),r},t.prototype._handlePositionChanges=function(t){for(var e=this,n=1;n<this.positions.length;n++)t.withFallbackPosition({originX:this.positions[n].originX,originY:this.positions[n].originY},{overlayX:this.positions[n].overlayX,overlayY:this.positions[n].overlayY});this._positionSubscription=t.onPositionChange.subscribe(function(t){return e.positionChange.emit(t)})},t.prototype._attachOverlay=function(){var t=this;this._overlayRef||this._createOverlay(),this._position.withDirection(this.dir),this._overlayRef.setDirection(this.dir),this._document.addEventListener("keydown",this._escapeListener),this._overlayRef.hasAttached()||(this._overlayRef.attach(this._templatePortal),this.attach.emit()),this.hasBackdrop&&(this._backdropSubscription=this._overlayRef.backdropClick().subscribe(function(){t.backdropClick.emit()}))},t.prototype._detachOverlay=function(){this._overlayRef&&(this._overlayRef.detach(),this.detach.emit()),this._backdropSubscription.unsubscribe(),this._document.removeEventListener("keydown",this._escapeListener)},t.prototype._destroyOverlay=function(){this._overlayRef&&this._overlayRef.dispose(),this._backdropSubscription.unsubscribe(),this._positionSubscription.unsubscribe(),this._document.removeEventListener("keydown",this._escapeListener)},t.decorators=[{type:r.s,args:[{selector:"[cdk-connected-overlay], [connected-overlay], [cdkConnectedOverlay]",exportAs:"cdkConnectedOverlay"}]}],t.ctorParameters=function(){return[{type:V},{type:r._4},{type:r._9},{type:void 0,decorators:[{type:r.y,args:[B]}]},{type:o.b,decorators:[{type:r.M}]},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[p.b]}]}]},t.propDecorators={origin:[{type:r.C,args:["cdkConnectedOverlayOrigin"]}],positions:[{type:r.C,args:["cdkConnectedOverlayPositions"]}],offsetX:[{type:r.C,args:["cdkConnectedOverlayOffsetX"]}],offsetY:[{type:r.C,args:["cdkConnectedOverlayOffsetY"]}],width:[{type:r.C,args:["cdkConnectedOverlayWidth"]}],height:[{type:r.C,args:["cdkConnectedOverlayHeight"]}],minWidth:[{type:r.C,args:["cdkConnectedOverlayMinWidth"]}],minHeight:[{type:r.C,args:["cdkConnectedOverlayMinHeight"]}],backdropClass:[{type:r.C,args:["cdkConnectedOverlayBackdropClass"]}],scrollStrategy:[{type:r.C,args:["cdkConnectedOverlayScrollStrategy"]}],open:[{type:r.C,args:["cdkConnectedOverlayOpen"]}],hasBackdrop:[{type:r.C,args:["cdkConnectedOverlayHasBackdrop"]}],_deprecatedOrigin:[{type:r.C,args:["origin"]}],_deprecatedPositions:[{type:r.C,args:["positions"]}],_deprecatedOffsetX:[{type:r.C,args:["offsetX"]}],_deprecatedOffsetY:[{type:r.C,args:["offsetY"]}],_deprecatedWidth:[{type:r.C,args:["width"]}],_deprecatedHeight:[{type:r.C,args:["height"]}],_deprecatedMinWidth:[{type:r.C,args:["minWidth"]}],_deprecatedMinHeight:[{type:r.C,args:["minHeight"]}],_deprecatedBackdropClass:[{type:r.C,args:["backdropClass"]}],_deprecatedScrollStrategy:[{type:r.C,args:["scrollStrategy"]}],_deprecatedOpen:[{type:r.C,args:["open"]}],_deprecatedHasBackdrop:[{type:r.C,args:["hasBackdrop"]}],backdropClick:[{type:r.N}],positionChange:[{type:r.N}],attach:[{type:r.N}],detach:[{type:r.N}]},t}(),W=[V,T,I,i.c,N,q],G=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[o.a,s.g,i.a],exports:[U,H,i.a],declarations:[U,H],providers:[W,E]}]}],t.ctorParameters=function(){return[]},t}();!function(t){function e(){return null!==t&&t.apply(this,arguments)||this}Object(a.b)(e,t),e.prototype._createContainer=function(){var e=this;t.prototype._createContainer.call(this),this._adjustParentForFullscreenChange(),this._addFullscreenChangeListener(function(){return e._adjustParentForFullscreenChange()})},e.prototype._adjustParentForFullscreenChange=function(){this._containerElement&&(this.getFullscreenElement()||document.body).appendChild(this._containerElement)},e.prototype._addFullscreenChangeListener=function(t){document.fullscreenEnabled?document.addEventListener("fullscreenchange",t):document.webkitFullscreenEnabled?document.addEventListener("webkitfullscreenchange",t):document.mozFullScreenEnabled?document.addEventListener("mozfullscreenchange",t):document.msFullscreenEnabled&&document.addEventListener("MSFullscreenChange",t)},e.prototype.getFullscreenElement=function(){return document.fullscreenElement||document.webkitFullscreenElement||document.mozFullScreenElement||document.msFullscreenElement||null},e.decorators=[{type:r.z}],e.ctorParameters=function(){return[]}}(R)},"../../../cdk/esm5/platform.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return a}),n.d(e,"d",function(){return c}),n.d(e,"c",function(){return l}),n.d(e,"b",function(){return p});var r,i,o=n("../../../core/esm5/core.js"),s="undefined"!=typeof Intl&&Intl.v8BreakIterator,a=function(){function t(){this.isBrowser="object"==typeof document&&!!document,this.EDGE=this.isBrowser&&/(edge)/i.test(navigator.userAgent),this.TRIDENT=this.isBrowser&&/(msie|trident)/i.test(navigator.userAgent),this.BLINK=this.isBrowser&&!(!window.chrome&&!s)&&!!CSS&&!this.EDGE&&!this.TRIDENT,this.WEBKIT=this.isBrowser&&/AppleWebKit/i.test(navigator.userAgent)&&!this.BLINK&&!this.EDGE&&!this.TRIDENT,this.IOS=this.isBrowser&&/iPad|iPhone|iPod/.test(navigator.userAgent)&&!window.MSStream,this.FIREFOX=this.isBrowser&&/(firefox|minefield)/i.test(navigator.userAgent),this.ANDROID=this.isBrowser&&/android/i.test(navigator.userAgent)&&!this.TRIDENT,this.SAFARI=this.isBrowser&&/safari/i.test(navigator.userAgent)&&this.WEBKIT}return t.decorators=[{type:o.z}],t.ctorParameters=function(){return[]},t}();function c(){if(null==r)try{window.addEventListener("test",null,Object.defineProperty({},"passive",{get:function(){return r=!0}}))}finally{r=r||!1}return r}var u=["color","button","checkbox","date","datetime-local","email","file","hidden","image","month","number","password","radio","range","reset","search","submit","tel","text","time","url","week"];function l(){if(i)return i;if("object"!=typeof document||!document)return i=new Set(u);var t=document.createElement("input");return i=new Set(u.filter(function(e){return t.setAttribute("type",e),t.type===e}))}var p=function(){function t(){}return t.decorators=[{type:o.H,args:[{providers:[a]}]}],t.ctorParameters=function(){return[]},t}()},"../../../cdk/esm5/portal.es5.js":function(t,e,n){"use strict";n.d(e,"d",function(){return a}),n.d(e,"h",function(){return c}),n.d(e,"a",function(){return u}),n.d(e,"e",function(){return l}),n.d(e,"b",function(){return p}),n.d(e,"c",function(){return d}),n.d(e,"g",function(){return h}),n.d(e,"f",function(){return f});var r=n("../../../../tslib/tslib.es6.js"),i=n("../../../core/esm5/core.js");function o(){throw Error("Host already has a portal attached")}var s=function(){function t(){}return t.prototype.attach=function(t){return null==t&&function(){throw Error("Attempting to attach a portal to a null PortalOutlet")}(),t.hasAttached()&&o(),this._attachedHost=t,t.attach(this)},t.prototype.detach=function(){var t=this._attachedHost;null==t?function(){throw Error("Attempting to detach a portal that is not attached to a host")}():(this._attachedHost=null,t.detach())},Object.defineProperty(t.prototype,"isAttached",{get:function(){return null!=this._attachedHost},enumerable:!0,configurable:!0}),t.prototype.setAttachedHost=function(t){this._attachedHost=t},t}(),a=function(t){function e(e,n,r){var i=t.call(this)||this;return i.component=e,i.viewContainerRef=n,i.injector=r,i}return Object(r.b)(e,t),e}(s),c=function(t){function e(e,n,r){var i=t.call(this)||this;return i.templateRef=e,i.viewContainerRef=n,r&&(i.context=r),i}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"origin",{get:function(){return this.templateRef.elementRef},enumerable:!0,configurable:!0}),e.prototype.attach=function(e,n){return void 0===n&&(n=this.context),this.context=n,t.prototype.attach.call(this,e)},e.prototype.detach=function(){return this.context=void 0,t.prototype.detach.call(this)},e}(s),u=function(){function t(){this._isDisposed=!1}return t.prototype.hasAttached=function(){return!!this._attachedPortal},t.prototype.attach=function(t){return t||function(){throw Error("Must provide a portal to attach")}(),this.hasAttached()&&o(),this._isDisposed&&function(){throw Error("This PortalOutlet has already been disposed")}(),t instanceof a?(this._attachedPortal=t,this.attachComponentPortal(t)):t instanceof c?(this._attachedPortal=t,this.attachTemplatePortal(t)):void function(){throw Error("Attempting to attach an unknown Portal type. BasePortalOutlet accepts either a ComponentPortal or a TemplatePortal.")}()},t.prototype.detach=function(){this._attachedPortal&&(this._attachedPortal.setAttachedHost(null),this._attachedPortal=null),this._invokeDisposeFn()},t.prototype.dispose=function(){this.hasAttached()&&this.detach(),this._invokeDisposeFn(),this._isDisposed=!0},t.prototype.setDisposeFn=function(t){this._disposeFn=t},t.prototype._invokeDisposeFn=function(){this._disposeFn&&(this._disposeFn(),this._disposeFn=null)},t}(),l=function(t){function e(e,n,r,i){var o=t.call(this)||this;return o._hostDomElement=e,o._componentFactoryResolver=n,o._appRef=r,o._defaultInjector=i,o}return Object(r.b)(e,t),e.prototype.attachComponentPortal=function(t){var e,n=this,r=this._componentFactoryResolver.resolveComponentFactory(t.component);return t.viewContainerRef?(e=t.viewContainerRef.createComponent(r,t.viewContainerRef.length,t.injector||t.viewContainerRef.parentInjector),this.setDisposeFn(function(){return e.destroy()})):(e=r.create(t.injector||this._defaultInjector),this._appRef.attachView(e.hostView),this.setDisposeFn(function(){n._appRef.detachView(e.hostView),e.destroy()})),this._hostDomElement.appendChild(this._getComponentRootNode(e)),e},e.prototype.attachTemplatePortal=function(t){var e=this,n=t.viewContainerRef,r=n.createEmbeddedView(t.templateRef,t.context);return r.detectChanges(),r.rootNodes.forEach(function(t){return e._hostDomElement.appendChild(t)}),this.setDisposeFn(function(){var t=n.indexOf(r);-1!==t&&n.remove(t)}),r},e.prototype.dispose=function(){t.prototype.dispose.call(this),null!=this._hostDomElement.parentNode&&this._hostDomElement.parentNode.removeChild(this._hostDomElement)},e.prototype._getComponentRootNode=function(t){return t.hostView.rootNodes[0]},e}(u),p=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.b)(e,t),e.decorators=[{type:i.s,args:[{selector:"[cdk-portal], [cdkPortal], [portal]",exportAs:"cdkPortal"}]}],e.ctorParameters=function(){return[{type:i._4},{type:i._9}]},e}(c),d=function(t){function e(e,n){var r=t.call(this)||this;return r._componentFactoryResolver=e,r._viewContainerRef=n,r._portal=null,r}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"_deprecatedPortal",{get:function(){return this.portal},set:function(t){this.portal=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_deprecatedPortalHost",{get:function(){return this.portal},set:function(t){this.portal=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"portal",{get:function(){return this._portal},set:function(e){this.hasAttached()&&t.prototype.detach.call(this),e&&t.prototype.attach.call(this,e),this._portal=e},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){t.prototype.dispose.call(this),this._portal=null},e.prototype.attachComponentPortal=function(e){e.setAttachedHost(this);var n=null!=e.viewContainerRef?e.viewContainerRef:this._viewContainerRef,r=this._componentFactoryResolver.resolveComponentFactory(e.component),i=n.createComponent(r,n.length,e.injector||n.parentInjector);return t.prototype.setDisposeFn.call(this,function(){return i.destroy()}),this._portal=e,i},e.prototype.attachTemplatePortal=function(e){var n=this;e.setAttachedHost(this);var r=this._viewContainerRef.createEmbeddedView(e.templateRef,e.context);return t.prototype.setDisposeFn.call(this,function(){return n._viewContainerRef.clear()}),this._portal=e,r},e.decorators=[{type:i.s,args:[{selector:"[cdkPortalOutlet], [cdkPortalHost], [portalHost]",exportAs:"cdkPortalOutlet, cdkPortalHost",inputs:["portal: cdkPortalOutlet"]}]}],e.ctorParameters=function(){return[{type:i.o},{type:i._9}]},e.propDecorators={_deprecatedPortal:[{type:i.C,args:["portalHost"]}],_deprecatedPortalHost:[{type:i.C,args:["cdkPortalHost"]}]},e}(u),h=function(){function t(){}return t.decorators=[{type:i.H,args:[{exports:[p,d],declarations:[p,d]}]}],t.ctorParameters=function(){return[]},t}(),f=function(){function t(t,e){this._parentInjector=t,this._customTokens=e}return t.prototype.get=function(t,e){var n=this._customTokens.get(t);return void 0!==n?n:this._parentInjector.get(t,e)},t}()},"../../../cdk/esm5/scrolling.es5.js":function(t,e,n){"use strict";n.d(e,"b",function(){return d}),n.d(e,"d",function(){return m}),n.d(e,"c",function(){return g}),n.d(e,"a",function(){return v});var r=n("../../../core/esm5/core.js"),i=n("../../../cdk/esm5/platform.es5.js"),o=n("../../../../rxjs/_esm5/Subject.js"),s=n("../../../../rxjs/_esm5/Observable.js"),a=n("../../../../rxjs/_esm5/observable/of.js"),c=n("../../../../rxjs/_esm5/observable/fromEvent.js"),u=n("../../../../rxjs/_esm5/operators/auditTime.js"),l=n("../../../../rxjs/_esm5/operators/filter.js"),p=n("../../../../rxjs/_esm5/observable/merge.js"),d=function(){function t(t,e){this._ngZone=t,this._platform=e,this._scrolled=new o.a,this._globalSubscription=null,this._scrolledCount=0,this.scrollContainers=new Map}return t.prototype.register=function(t){var e=this,n=t.elementScrolled().subscribe(function(){return e._scrolled.next(t)});this.scrollContainers.set(t,n)},t.prototype.deregister=function(t){var e=this.scrollContainers.get(t);e&&(e.unsubscribe(),this.scrollContainers.delete(t))},t.prototype.scrolled=function(t){var e=this;return void 0===t&&(t=20),this._platform.isBrowser?s.a.create(function(n){e._globalSubscription||e._addGlobalListener();var r=t>0?e._scrolled.pipe(Object(u.a)(t)).subscribe(n):e._scrolled.subscribe(n);return e._scrolledCount++,function(){r.unsubscribe(),e._scrolledCount--,e._globalSubscription&&!e._scrolledCount&&(e._globalSubscription.unsubscribe(),e._globalSubscription=null)}}):Object(a.a)()},t.prototype.ancestorScrolled=function(t,e){var n=this.getAncestorScrollContainers(t);return this.scrolled(e).pipe(Object(l.a)(function(t){return!t||n.indexOf(t)>-1}))},t.prototype.getAncestorScrollContainers=function(t){var e=this,n=[];return this.scrollContainers.forEach(function(r,i){e._scrollableContainsElement(i,t)&&n.push(i)}),n},t.prototype._scrollableContainsElement=function(t,e){var n=e.nativeElement,r=t.getElementRef().nativeElement;do{if(n==r)return!0}while(n=n.parentElement);return!1},t.prototype._addGlobalListener=function(){var t=this;this._globalSubscription=this._ngZone.runOutsideAngular(function(){return Object(c.a)(window.document,"scroll").subscribe(function(){return t._scrolled.next()})})},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:r.L},{type:i.a}]},t}();var h={provide:d,deps:[[new r.M,new r._1,d],r.L,i.a],useFactory:function(t,e,n){return t||new d(e,n)}},f=function(){function t(t,e,n){var r=this;this._elementRef=t,this._scroll=e,this._ngZone=n,this._elementScrolled=new o.a,this._scrollListener=function(t){return r._elementScrolled.next(t)}}return t.prototype.ngOnInit=function(){var t=this;this._ngZone.runOutsideAngular(function(){t.getElementRef().nativeElement.addEventListener("scroll",t._scrollListener)}),this._scroll.register(this)},t.prototype.ngOnDestroy=function(){this._scroll.deregister(this),this._scrollListener&&this.getElementRef().nativeElement.removeEventListener("scroll",this._scrollListener)},t.prototype.elementScrolled=function(){return this._elementScrolled.asObservable()},t.prototype.getElementRef=function(){return this._elementRef},t.decorators=[{type:r.s,args:[{selector:"[cdk-scrollable], [cdkScrollable]"}]}],t.ctorParameters=function(){return[{type:r.t},{type:d},{type:r.L}]},t}(),m=function(){function t(t,e){var n=this;this._change=t.isBrowser?e.runOutsideAngular(function(){return Object(p.a)(Object(c.a)(window,"resize"),Object(c.a)(window,"orientationchange"))}):Object(a.a)(),this._invalidateCache=this.change().subscribe(function(){return n._updateViewportSize()})}return t.prototype.ngOnDestroy=function(){this._invalidateCache.unsubscribe()},t.prototype.getViewportSize=function(){return this._viewportSize||this._updateViewportSize(),{width:this._viewportSize.width,height:this._viewportSize.height}},t.prototype.getViewportRect=function(){var t=this.getViewportScrollPosition(),e=this.getViewportSize(),n=e.width,r=e.height;return{top:t.top,left:t.left,bottom:t.top+r,right:t.left+n,height:r,width:n}},t.prototype.getViewportScrollPosition=function(){var t=document.documentElement.getBoundingClientRect();return{top:-t.top||document.body.scrollTop||window.scrollY||document.documentElement.scrollTop||0,left:-t.left||document.body.scrollLeft||window.scrollX||document.documentElement.scrollLeft||0}},t.prototype.change=function(t){return void 0===t&&(t=20),t>0?this._change.pipe(Object(u.a)(t)):this._change},t.prototype._updateViewportSize=function(){this._viewportSize={width:window.innerWidth,height:window.innerHeight}},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:i.a},{type:r.L}]},t}();var g={provide:m,deps:[[new r.M,new r._1,m],i.a,r.L],useFactory:function(t,e,n){return t||new m(e,n)}},v=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.b],exports:[f],declarations:[f],providers:[h]}]}],t.ctorParameters=function(){return[]},t}()},"../../../cdk/esm5/stepper.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return p}),n.d(e,"c",function(){return d}),n.d(e,"b",function(){return u}),n.d(e,"e",function(){return h}),n.d(e,"f",function(){return f}),n.d(e,"d",function(){return m});var r=n("../../../core/esm5/core.js"),i=n("../../../cdk/esm5/keycodes.es5.js"),o=n("../../../cdk/esm5/coercion.es5.js"),s=(n("../../../forms/esm5/forms.js"),n("../../../cdk/esm5/bidi.es5.js")),a=n("../../../../rxjs/_esm5/Subject.js"),c=n("../../../common/esm5/common.js"),u=function(){function t(t){this.template=t}return t.decorators=[{type:r.s,args:[{selector:"[cdkStepLabel]"}]}],t.ctorParameters=function(){return[{type:r._4}]},t}(),l=0,p=(function(){}(),function(){function t(t){this._stepper=t,this.interacted=!1,this._editable=!0,this._optional=!1,this._customCompleted=null}return Object.defineProperty(t.prototype,"editable",{get:function(){return this._editable},set:function(t){this._editable=Object(o.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"optional",{get:function(){return this._optional},set:function(t){this._optional=Object(o.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"completed",{get:function(){return null==this._customCompleted?this._defaultCompleted:this._customCompleted},set:function(t){this._customCompleted=Object(o.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_defaultCompleted",{get:function(){return this.stepControl?this.stepControl.valid&&this.interacted:this.interacted},enumerable:!0,configurable:!0}),t.prototype.select=function(){this._stepper.selected=this},t.prototype.ngOnChanges=function(){this._stepper._stateChanged()},t.decorators=[{type:r.m,args:[{selector:"cdk-step",exportAs:"cdkStep",template:"<ng-template><ng-content></ng-content></ng-template>",encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush}]}],t.ctorParameters=function(){return[{type:d,decorators:[{type:r.y,args:[Object(r._14)(function(){return d})]}]}]},t.propDecorators={stepLabel:[{type:r.q,args:[u]}],content:[{type:r._7,args:[r._4]}],stepControl:[{type:r.C}],label:[{type:r.C}],editable:[{type:r.C}],optional:[{type:r.C}],completed:[{type:r.C}]},t}()),d=function(){function t(t,e){this._dir=t,this._changeDetectorRef=e,this._destroyed=new a.a,this._linear=!1,this._selectedIndex=0,this.selectionChange=new r.w,this._focusIndex=0,this._groupId=l++}return Object.defineProperty(t.prototype,"linear",{get:function(){return this._linear},set:function(t){this._linear=Object(o.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selectedIndex",{get:function(){return this._selectedIndex},set:function(t){this._steps?this._anyControlsInvalidOrPending(t)||t<this._selectedIndex&&!this._steps.toArray()[t].editable?this._stepHeader.toArray()[t].nativeElement.blur():this._selectedIndex!=t&&(this._emitStepperSelectionEvent(t),this._focusIndex=this._selectedIndex):this._selectedIndex=this._focusIndex=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._steps.toArray()[this.selectedIndex]},set:function(t){this.selectedIndex=this._steps.toArray().indexOf(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this._destroyed.next(),this._destroyed.complete()},t.prototype.next=function(){this.selectedIndex=Math.min(this._selectedIndex+1,this._steps.length-1)},t.prototype.previous=function(){this.selectedIndex=Math.max(this._selectedIndex-1,0)},t.prototype._getStepLabelId=function(t){return"cdk-step-label-"+this._groupId+"-"+t},t.prototype._getStepContentId=function(t){return"cdk-step-content-"+this._groupId+"-"+t},t.prototype._stateChanged=function(){this._changeDetectorRef.markForCheck()},t.prototype._getAnimationDirection=function(t){var e=t-this._selectedIndex;return e<0?"rtl"===this._layoutDirection()?"next":"previous":e>0?"rtl"===this._layoutDirection()?"previous":"next":"current"},t.prototype._getIndicatorType=function(t){var e=this._steps.toArray()[t];return e.completed&&this._selectedIndex!=t?e.editable?"edit":"done":"number"},t.prototype._emitStepperSelectionEvent=function(t){var e=this._steps.toArray();this.selectionChange.emit({selectedIndex:t,previouslySelectedIndex:this._selectedIndex,selectedStep:e[t],previouslySelectedStep:e[this._selectedIndex]}),this._selectedIndex=t,this._stateChanged()},t.prototype._onKeydown=function(t){switch(t.keyCode){case i.m:"rtl"===this._layoutDirection()?this._focusPreviousStep():this._focusNextStep();break;case i.i:"rtl"===this._layoutDirection()?this._focusNextStep():this._focusPreviousStep();break;case i.n:case i.f:this.selectedIndex=this._focusIndex;break;default:return}t.preventDefault()},t.prototype._focusNextStep=function(){this._focusStep((this._focusIndex+1)%this._steps.length)},t.prototype._focusPreviousStep=function(){this._focusStep((this._focusIndex+this._steps.length-1)%this._steps.length)},t.prototype._focusStep=function(t){this._focusIndex=t,this._stepHeader.toArray()[this._focusIndex].nativeElement.focus()},t.prototype._anyControlsInvalidOrPending=function(t){var e=this._steps.toArray();return e[this._selectedIndex].interacted=!0,!!(this._linear&&t>=0)&&e.slice(0,t).some(function(t){return t.stepControl&&(t.stepControl.invalid||t.stepControl.pending)})},t.prototype._layoutDirection=function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},t.decorators=[{type:r.s,args:[{selector:"[cdkStepper]",exportAs:"cdkStepper"}]}],t.ctorParameters=function(){return[{type:s.b,decorators:[{type:r.M}]},{type:r.j}]},t.propDecorators={_steps:[{type:r.r,args:[p]}],linear:[{type:r.C}],selectedIndex:[{type:r.C}],selected:[{type:r.C}],selectionChange:[{type:r.N}]},t}(),h=function(){function t(t){this._stepper=t}return t.decorators=[{type:r.s,args:[{selector:"button[cdkStepperNext]",host:{"(click)":"_stepper.next()"}}]}],t.ctorParameters=function(){return[{type:d}]},t}(),f=function(){function t(t){this._stepper=t}return t.decorators=[{type:r.s,args:[{selector:"button[cdkStepperPrevious]",host:{"(click)":"_stepper.previous()"}}]}],t.ctorParameters=function(){return[{type:d}]},t}(),m=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[s.a,c.a],exports:[p,d,u,h,f],declarations:[p,d,u,h,f]}]}],t.ctorParameters=function(){return[]},t}()},"../../../cdk/esm5/table.es5.js":function(t,e,n){"use strict";n.d(e,"b",function(){return O}),n.d(e,"l",function(){return C}),n.d(e,"d",function(){return g}),n.d(e,"g",function(){return v}),n.d(e,"e",function(){return y}),n.d(e,"f",function(){return b}),n.d(e,"c",function(){return _}),n.d(e,"a",function(){return u}),n.d(e,"i",function(){return p}),n.d(e,"k",function(){return d}),n.d(e,"h",function(){return f}),n.d(e,"j",function(){return m}),n.d(e,"m",function(){return E});var r=n("../../../../tslib/tslib.es6.js"),i=n("../../../core/esm5/core.js"),o=(n("../../../cdk/esm5/collections.es5.js"),n("../../../../rxjs/_esm5/operators/takeUntil.js")),s=n("../../../../rxjs/_esm5/BehaviorSubject.js"),a=n("../../../../rxjs/_esm5/Subject.js"),c=n("../../../common/esm5/common.js"),u="<ng-container cdkCellOutlet></ng-container>",l=function(){function t(t,e){this.template=t,this._differs=e}return t.prototype.ngOnChanges=function(t){var e=t.columns.currentValue||[];this._columnsDiffer||(this._columnsDiffer=this._differs.find(e).create(),this._columnsDiffer.diff(e))},t.prototype.getColumnsDiff=function(){return this._columnsDiffer.diff(this.columns)},t}(),p=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.b)(e,t),e.decorators=[{type:i.s,args:[{selector:"[cdkHeaderRowDef]",inputs:["columns: cdkHeaderRowDef"]}]}],e.ctorParameters=function(){return[{type:i._4},{type:i.D}]},e}(l),d=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.b)(e,t),e.decorators=[{type:i.s,args:[{selector:"[cdkRowDef]",inputs:["columns: cdkRowDefColumns","when: cdkRowDefWhen"]}]}],e.ctorParameters=function(){return[{type:i._4},{type:i.D}]},e}(l),h=function(){function t(e){this._viewContainer=e,t.mostRecentCellOutlet=this}return t.mostRecentCellOutlet=null,t.decorators=[{type:i.s,args:[{selector:"[cdkCellOutlet]"}]}],t.ctorParameters=function(){return[{type:i._9}]},t}(),f=function(){function t(){}return t.decorators=[{type:i.m,args:[{selector:"cdk-header-row",template:u,host:{class:"cdk-header-row",role:"row"},changeDetection:i.i.OnPush,encapsulation:i._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[]},t}(),m=function(){function t(){}return t.decorators=[{type:i.m,args:[{selector:"cdk-row",template:u,host:{class:"cdk-row",role:"row"},changeDetection:i.i.OnPush,encapsulation:i._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[]},t}(),g=function(){function t(t){this.template=t}return t.decorators=[{type:i.s,args:[{selector:"[cdkCellDef]"}]}],t.ctorParameters=function(){return[{type:i._4}]},t}(),v=function(){function t(t){this.template=t}return t.decorators=[{type:i.s,args:[{selector:"[cdkHeaderCellDef]"}]}],t.ctorParameters=function(){return[{type:i._4}]},t}(),y=function(){function t(){}return Object.defineProperty(t.prototype,"name",{get:function(){return this._name},set:function(t){this._name=t,this.cssClassFriendlyName=t.replace(/[^a-z0-9_-]/gi,"-")},enumerable:!0,configurable:!0}),t.decorators=[{type:i.s,args:[{selector:"[cdkColumnDef]"}]}],t.ctorParameters=function(){return[]},t.propDecorators={name:[{type:i.C,args:["cdkColumnDef"]}],cell:[{type:i.q,args:[g]}],headerCell:[{type:i.q,args:[v]}]},t}(),b=function(){function t(t,e){e.nativeElement.classList.add("cdk-column-"+t.cssClassFriendlyName)}return t.decorators=[{type:i.s,args:[{selector:"cdk-header-cell",host:{class:"cdk-header-cell",role:"columnheader"}}]}],t.ctorParameters=function(){return[{type:y},{type:i.t}]},t}(),_=function(){function t(t,e){e.nativeElement.classList.add("cdk-column-"+t.cssClassFriendlyName)}return t.decorators=[{type:i.s,args:[{selector:"cdk-cell",host:{class:"cdk-cell",role:"gridcell"}}]}],t.ctorParameters=function(){return[{type:y},{type:i.t}]},t}();function j(t){return Error('Could not find column with id "'+t+'".')}var x=function(){function t(t){this.viewContainer=t}return t.decorators=[{type:i.s,args:[{selector:"[rowPlaceholder]"}]}],t.ctorParameters=function(){return[{type:i._9}]},t}(),w=function(){function t(t){this.viewContainer=t}return t.decorators=[{type:i.s,args:[{selector:"[headerRowPlaceholder]"}]}],t.ctorParameters=function(){return[{type:i._9}]},t}(),O="\n  <ng-container headerRowPlaceholder></ng-container>\n  <ng-container rowPlaceholder></ng-container>",C=(function(t){function e(){return null!==t&&t.apply(this,arguments)||this}Object(r.b)(e,t)}(i.u),function(){function t(t,e,n,r){this._differs=t,this._changeDetectorRef=e,this._onDestroy=new a.a,this._data=[],this._columnDefsByName=new Map,this.viewChange=new s.a({start:0,end:Number.MAX_VALUE}),r||n.nativeElement.setAttribute("role","grid")}return Object.defineProperty(t.prototype,"trackBy",{get:function(){return this._trackByFn},set:function(t){Object(i._16)()&&null!=t&&"function"!=typeof t&&console&&console.warn&&console.warn("trackBy must be a function, but received "+JSON.stringify(t)+"."),this._trackByFn=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dataSource",{get:function(){return this._dataSource},set:function(t){this._dataSource!==t&&this._switchDataSource(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this._dataDiffer=this._differs.find([]).create(this._trackByFn)},t.prototype.ngAfterContentInit=function(){var t=this;if(!this._headerDef&&!this._rowDefs.length)throw Error("Missing definitions for header and row, cannot determine which columns should be rendered.");this._cacheColumnDefsByName(),this._columnDefs.changes.subscribe(function(){return t._cacheColumnDefsByName()}),this._renderHeaderRow()},t.prototype.ngAfterContentChecked=function(){this._renderUpdatedColumns();var t=this._rowDefs.filter(function(t){return!t.when});if(t.length>1)throw Error("There can only be one default row without a when predicate function.");this._defaultRowDef=t[0],this.dataSource&&!this._renderChangeSubscription&&this._observeRenderChanges()},t.prototype.ngOnDestroy=function(){this._rowPlaceholder.viewContainer.clear(),this._headerRowPlaceholder.viewContainer.clear(),this._onDestroy.next(),this._onDestroy.complete(),this.dataSource&&this.dataSource.disconnect(this)},t.prototype._cacheColumnDefsByName=function(){var t=this;this._columnDefsByName.clear(),this._columnDefs.forEach(function(e){if(t._columnDefsByName.has(e.name))throw n=e.name,Error('Duplicate column definition name provided: "'+n+'".');var n;t._columnDefsByName.set(e.name,e)})},t.prototype._renderUpdatedColumns=function(){var t=this;this._rowDefs.forEach(function(e){e.getColumnsDiff()&&(t._dataDiffer.diff([]),t._rowPlaceholder.viewContainer.clear(),t._renderRowChanges())}),this._headerDef.getColumnsDiff()&&(this._headerRowPlaceholder.viewContainer.clear(),this._renderHeaderRow())},t.prototype._switchDataSource=function(t){this._data=[],this.dataSource&&this.dataSource.disconnect(this),this._renderChangeSubscription&&(this._renderChangeSubscription.unsubscribe(),this._renderChangeSubscription=null),t||this._rowPlaceholder.viewContainer.clear(),this._dataSource=t},t.prototype._observeRenderChanges=function(){var t=this;this._renderChangeSubscription=this.dataSource.connect(this).pipe(Object(o.a)(this._onDestroy)).subscribe(function(e){t._data=e,t._renderRowChanges()})},t.prototype._renderHeaderRow=function(){var t=this._getHeaderCellTemplatesForRow(this._headerDef);t.length&&(this._headerRowPlaceholder.viewContainer.createEmbeddedView(this._headerDef.template,{cells:t}),t.forEach(function(t){h.mostRecentCellOutlet&&h.mostRecentCellOutlet._viewContainer.createEmbeddedView(t.template,{})}),this._changeDetectorRef.markForCheck())},t.prototype._renderRowChanges=function(){var t=this,e=this._dataDiffer.diff(this._data);if(e){var n=this._rowPlaceholder.viewContainer;e.forEachOperation(function(e,r,i){if(null==e.previousIndex)t._insertRow(e.item,i);else if(null==i)n.remove(r);else{var o=n.get(r);n.move(o,i)}}),this._updateRowIndexContext(),e.forEachIdentityChange(function(t){n.get(t.currentIndex).context.$implicit=t.item})}},t.prototype._getRowDef=function(t,e){if(1==this._rowDefs.length)return this._rowDefs.first;var n=this._rowDefs.find(function(n){return n.when&&n.when(e,t)})||this._defaultRowDef;if(!n)throw Error("Could not find a matching row definition for the provided row data.");return n},t.prototype._insertRow=function(t,e){var n=this._getRowDef(t,e),r={$implicit:t};this._rowPlaceholder.viewContainer.createEmbeddedView(n.template,r,e),this._getCellTemplatesForRow(n).forEach(function(t){h.mostRecentCellOutlet&&h.mostRecentCellOutlet._viewContainer.createEmbeddedView(t.template,r)}),this._changeDetectorRef.markForCheck()},t.prototype._updateRowIndexContext=function(){for(var t=this._rowPlaceholder.viewContainer,e=0,n=t.length;e<n;e++){var r=t.get(e);r.context.index=e,r.context.count=n,r.context.first=0===e,r.context.last=e===n-1,r.context.even=e%2==0,r.context.odd=!r.context.even}},t.prototype._getHeaderCellTemplatesForRow=function(t){var e=this;return t.columns?t.columns.map(function(t){var n=e._columnDefsByName.get(t);if(!n)throw j(t);return n.headerCell}):[]},t.prototype._getCellTemplatesForRow=function(t){var e=this;return t.columns?t.columns.map(function(t){var n=e._columnDefsByName.get(t);if(!n)throw j(t);return n.cell}):[]},t.decorators=[{type:i.m,args:[{selector:"cdk-table",exportAs:"cdkTable",template:O,host:{class:"cdk-table"},encapsulation:i._10.None,preserveWhitespaces:!1,changeDetection:i.i.OnPush}]}],t.ctorParameters=function(){return[{type:i.D},{type:i.j},{type:i.t},{type:void 0,decorators:[{type:i.g,args:["role"]}]}]},t.propDecorators={trackBy:[{type:i.C}],dataSource:[{type:i.C}],_rowPlaceholder:[{type:i._7,args:[x]}],_headerRowPlaceholder:[{type:i._7,args:[w]}],_columnDefs:[{type:i.r,args:[y]}],_headerDef:[{type:i.q,args:[p]}],_rowDefs:[{type:i.r,args:[d]}]},t}()),k=[C,d,g,h,v,y,_,m,b,f,p,x,w],E=function(){function t(){}return t.decorators=[{type:i.H,args:[{imports:[c.a],exports:[k],declarations:[k]}]}],t.ctorParameters=function(){return[]},t}()},"../../../common/esm5/common.js":function(t,e,n){"use strict";n.d(e,"h",function(){return T}),n.d(e,"a",function(){return ue}),n.d(e,"d",function(){return D}),n.d(e,"e",function(){return G}),n.d(e,"b",function(){return pe}),n.d(e,"g",function(){return de}),n.d(e,"f",function(){return o}),n.d(e,"c",function(){return c});var r=n("../../../core/esm5/core.js"),i=n("../../../../tslib/tslib.es6.js"),o=function(){return function(){}}(),s=(new r.A("Location Initialized"),function(){return function(){}}()),a=new r.A("appBaseHref"),c=function(){function t(e){var n=this;this._subject=new r.w,this._platformStrategy=e;var i=this._platformStrategy.getBaseHref();this._baseHref=t.stripTrailingSlash(u(i)),this._platformStrategy.onPopState(function(t){n._subject.emit({url:n.path(!0),pop:!0,type:t.type})})}return t.prototype.path=function(t){return void 0===t&&(t=!1),this.normalize(this._platformStrategy.path(t))},t.prototype.isCurrentPathEqualTo=function(e,n){return void 0===n&&(n=""),this.path()==this.normalize(e+t.normalizeQueryParams(n))},t.prototype.normalize=function(e){return t.stripTrailingSlash((n=this._baseHref,r=u(e),n&&r.startsWith(n)?r.substring(n.length):r));var n,r},t.prototype.prepareExternalUrl=function(t){return t&&"/"!==t[0]&&(t="/"+t),this._platformStrategy.prepareExternalUrl(t)},t.prototype.go=function(t,e){void 0===e&&(e=""),this._platformStrategy.pushState(null,"",t,e)},t.prototype.replaceState=function(t,e){void 0===e&&(e=""),this._platformStrategy.replaceState(null,"",t,e)},t.prototype.forward=function(){this._platformStrategy.forward()},t.prototype.back=function(){this._platformStrategy.back()},t.prototype.subscribe=function(t,e,n){return this._subject.subscribe({next:t,error:e,complete:n})},t.normalizeQueryParams=function(t){return t&&"?"!==t[0]?"?"+t:t},t.joinWithSlash=function(t,e){if(0==t.length)return e;if(0==e.length)return t;var n=0;return t.endsWith("/")&&n++,e.startsWith("/")&&n++,2==n?t+e.substring(1):1==n?t+e:t+"/"+e},t.stripTrailingSlash=function(t){var e=t.match(/#|\?|$/),n=e&&e.index||t.length,r=n-("/"===t[n-1]?1:0);return t.slice(0,r)+t.slice(n)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:s}]},t}();function u(t){return t.replace(/\/index.html$/,"")}(function(t){function e(e,n){var r=t.call(this)||this;return r._platformLocation=e,r._baseHref="",null!=n&&(r._baseHref=n),r}Object(i.b)(e,t),e.prototype.onPopState=function(t){this._platformLocation.onPopState(t),this._platformLocation.onHashChange(t)},e.prototype.getBaseHref=function(){return this._baseHref},e.prototype.path=function(t){void 0===t&&(t=!1);var e=this._platformLocation.hash;return null==e&&(e="#"),e.length>0?e.substring(1):e},e.prototype.prepareExternalUrl=function(t){var e=c.joinWithSlash(this._baseHref,t);return e.length>0?"#"+e:e},e.prototype.pushState=function(t,e,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));0==i.length&&(i=this._platformLocation.pathname),this._platformLocation.pushState(t,e,i)},e.prototype.replaceState=function(t,e,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));0==i.length&&(i=this._platformLocation.pathname),this._platformLocation.replaceState(t,e,i)},e.prototype.forward=function(){this._platformLocation.forward()},e.prototype.back=function(){this._platformLocation.back()},e.decorators=[{type:r.z}],e.ctorParameters=function(){return[{type:o},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[a]}]}]}})(s),function(t){function e(e,n){var r=t.call(this)||this;if(r._platformLocation=e,null==n&&(n=r._platformLocation.getBaseHrefFromDOM()),null==n)throw new Error("No base href set. Please provide a value for the APP_BASE_HREF token or add a base element to the document.");return r._baseHref=n,r}Object(i.b)(e,t),e.prototype.onPopState=function(t){this._platformLocation.onPopState(t),this._platformLocation.onHashChange(t)},e.prototype.getBaseHref=function(){return this._baseHref},e.prototype.prepareExternalUrl=function(t){return c.joinWithSlash(this._baseHref,t)},e.prototype.path=function(t){void 0===t&&(t=!1);var e=this._platformLocation.pathname+c.normalizeQueryParams(this._platformLocation.search),n=this._platformLocation.hash;return n&&t?""+e+n:e},e.prototype.pushState=function(t,e,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));this._platformLocation.pushState(t,e,i)},e.prototype.replaceState=function(t,e,n,r){var i=this.prepareExternalUrl(n+c.normalizeQueryParams(r));this._platformLocation.replaceState(t,e,i)},e.prototype.forward=function(){this._platformLocation.forward()},e.prototype.back=function(){this._platformLocation.back()},e.decorators=[{type:r.z}],e.ctorParameters=function(){return[{type:o},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[a]}]}]}}(s);var l={AOA:[,"Kz"],ARS:[,"$"],AUD:["A$","$"],BAM:[,"KM"],BBD:[,"$"],BDT:[,"৳"],BMD:[,"$"],BND:[,"$"],BOB:[,"Bs"],BRL:["R$"],BSD:[,"$"],BWP:[,"P"],BYN:[,"р."],BZD:[,"$"],CAD:["CA$","$"],CLP:[,"$"],CNY:["CN¥","¥"],COP:[,"$"],CRC:[,"₡"],CUC:[,"$"],CUP:[,"$"],CZK:[,"Kč"],DKK:[,"kr"],DOP:[,"$"],EGP:[,"E£"],ESP:[,"₧"],EUR:["€"],FJD:[,"$"],FKP:[,"£"],GBP:["£"],GEL:[,"₾"],GIP:[,"£"],GNF:[,"FG"],GTQ:[,"Q"],GYD:[,"$"],HKD:["HK$","$"],HNL:[,"L"],HRK:[,"kn"],HUF:[,"Ft"],IDR:[,"Rp"],ILS:["₪"],INR:["₹"],ISK:[,"kr"],JMD:[,"$"],JPY:["¥"],KHR:[,"៛"],KMF:[,"CF"],KPW:[,"₩"],KRW:["₩"],KYD:[,"$"],KZT:[,"₸"],LAK:[,"₭"],LBP:[,"L£"],LKR:[,"Rs"],LRD:[,"$"],LTL:[,"Lt"],LVL:[,"Ls"],MGA:[,"Ar"],MMK:[,"K"],MNT:[,"₮"],MUR:[,"Rs"],MXN:["MX$","$"],MYR:[,"RM"],NAD:[,"$"],NGN:[,"₦"],NIO:[,"C$"],NOK:[,"kr"],NPR:[,"Rs"],NZD:["NZ$","$"],PHP:[,"₱"],PKR:[,"Rs"],PLN:[,"zł"],PYG:[,"₲"],RON:[,"lei"],RUB:[,"₽"],RUR:[,"р."],RWF:[,"RF"],SBD:[,"$"],SEK:[,"kr"],SGD:[,"$"],SHP:[,"£"],SRD:[,"$"],SSP:[,"£"],STD:[,"Db"],SYP:[,"£"],THB:[,"฿"],TOP:[,"T$"],TRY:[,"₺"],TTD:[,"$"],TWD:["NT$","$"],UAH:[,"₴"],USD:["$"],UYU:[,"$"],VEF:[,"Bs"],VND:["₫"],XAF:["FCFA"],XCD:["EC$","$"],XOF:["CFA"],XPF:["CFPF"],ZAR:[,"R"],ZMW:[,"ZK"]},p=["en",[["a","p"],["AM","PM"]],[["AM","PM"],,],[["S","M","T","W","T","F","S"],["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],["Su","Mo","Tu","We","Th","Fr","Sa"]],,[["J","F","M","A","M","J","J","A","S","O","N","D"],["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],["January","February","March","April","May","June","July","August","September","October","November","December"]],,[["B","A"],["BC","AD"],["Before Christ","Anno Domini"]],0,[6,0],["M/d/yy","MMM d, y","MMMM d, y","EEEE, MMMM d, y"],["h:mm a","h:mm:ss a","h:mm:ss a z","h:mm:ss a zzzz"],["{1}, {0}",,"{1} 'at' {0}"],[".",",",";","%","+","-","E","×","‰","∞","NaN",":"],["#,##0.###","#,##0%","¤#,##0.00","#E0"],"$","US Dollar",function(t){var e=Math.floor(Math.abs(t)),n=t.toString().replace(/^[^.]*\.?/,"").length;return 1===e&&0===n?1:5}],d={};var h={Decimal:0,Percent:1,Currency:2,Scientific:3};h[h.Decimal]="Decimal",h[h.Percent]="Percent",h[h.Currency]="Currency",h[h.Scientific]="Scientific";var f={Zero:0,One:1,Two:2,Few:3,Many:4,Other:5};f[f.Zero]="Zero",f[f.One]="One",f[f.Two]="Two",f[f.Few]="Few",f[f.Many]="Many",f[f.Other]="Other";var m={Format:0,Standalone:1};m[m.Format]="Format",m[m.Standalone]="Standalone";var g={Narrow:0,Abbreviated:1,Wide:2,Short:3};g[g.Narrow]="Narrow",g[g.Abbreviated]="Abbreviated",g[g.Wide]="Wide",g[g.Short]="Short";var v={Short:0,Medium:1,Long:2,Full:3};v[v.Short]="Short",v[v.Medium]="Medium",v[v.Long]="Long",v[v.Full]="Full";var y={Decimal:0,Group:1,List:2,PercentSign:3,PlusSign:4,MinusSign:5,Exponential:6,SuperscriptingExponent:7,PerMille:8,Infinity:9,NaN:10,TimeSeparator:11,CurrencyDecimal:12,CurrencyGroup:13};y[y.Decimal]="Decimal",y[y.Group]="Group",y[y.List]="List",y[y.PercentSign]="PercentSign",y[y.PlusSign]="PlusSign",y[y.MinusSign]="MinusSign",y[y.Exponential]="Exponential",y[y.SuperscriptingExponent]="SuperscriptingExponent",y[y.PerMille]="PerMille",y[y.Infinity]="Infinity",y[y.NaN]="NaN",y[y.TimeSeparator]="TimeSeparator",y[y.CurrencyDecimal]="CurrencyDecimal",y[y.CurrencyGroup]="CurrencyGroup";var b={Sunday:0,Monday:1,Tuesday:2,Wednesday:3,Thursday:4,Friday:5,Saturday:6};function _(t,e){return E(t)[10][e]}function j(t,e){return E(t)[11][e]}function x(t,e){return C(E(t)[12],e)}function w(t,e){var n=E(t),r=n[13][e];if(void 0===r){if(e===y.CurrencyDecimal)return n[13][y.Decimal];if(e===y.CurrencyGroup)return n[13][y.Group]}return r}function O(t){if(!t[18])throw new Error('Missing extra locale data for the locale "'+t[0]+'". Use "registerLocaleData" to load new data. See the "I18n guide" on angular.io to know more.')}function C(t,e){for(var n=e;n>-1;n--)if(void 0!==t[n])return t[n];throw new Error("Locale data API: locale data undefined")}function k(t){var e=t.split(":");return{hours:+e[0],minutes:+e[1]}}function E(t){var e=t.toLowerCase().replace(/_/g,"-"),n=d[e];if(n)return n;var r=e.split("-")[0];if(n=d[r])return n;if("en"===r)return p;throw new Error('Missing locale data for the locale "'+t+'".')}b[b.Sunday]="Sunday",b[b.Monday]="Monday",b[b.Tuesday]="Tuesday",b[b.Wednesday]="Wednesday",b[b.Thursday]="Thursday",b[b.Friday]="Friday",b[b.Saturday]="Saturday";var S=new r.A("UseV4Plurals"),P=function(){return function(){}}();function M(t,e,n,r){var i="="+t;if(e.indexOf(i)>-1)return i;if(i=n.getPluralCategory(t,r),e.indexOf(i)>-1)return i;if(e.indexOf("other")>-1)return"other";throw new Error('No plural message found for value "'+t+'"')}var A=function(t){function e(e,n){var r=t.call(this)||this;return r.locale=e,r.deprecatedPluralFn=n,r}return Object(i.b)(e,t),e.prototype.getPluralCategory=function(t,e){switch(this.deprecatedPluralFn?this.deprecatedPluralFn(e||this.locale,t):E(e||this.locale)[17](t)){case f.Zero:return"zero";case f.One:return"one";case f.Two:return"two";case f.Few:return"few";case f.Many:return"many";default:return"other"}},e.decorators=[{type:r.z}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[r.F]}]},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[S]}]}]},e}(P);function T(t,e){e=encodeURIComponent(e);for(var n=0,r=t.split(";");n<r.length;n++){var i=r[n],o=i.indexOf("="),s=-1==o?[i,""]:[i.slice(0,o),i.slice(o+1)],a=s[1];if(s[0].trim()===e)return decodeURIComponent(a)}return null}var D=function(){function t(t,e,n,r){this._iterableDiffers=t,this._keyValueDiffers=e,this._ngEl=n,this._renderer=r,this._initialClasses=[]}return Object.defineProperty(t.prototype,"klass",{set:function(t){this._applyInitialClasses(!0),this._initialClasses="string"==typeof t?t.split(/\s+/):[],this._applyInitialClasses(!1),this._applyClasses(this._rawClass,!1)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClass",{set:function(t){this._cleanupClasses(this._rawClass),this._iterableDiffer=null,this._keyValueDiffer=null,this._rawClass="string"==typeof t?t.split(/\s+/):t,this._rawClass&&(Object(r._33)(this._rawClass)?this._iterableDiffer=this._iterableDiffers.find(this._rawClass).create():this._keyValueDiffer=this._keyValueDiffers.find(this._rawClass).create())},enumerable:!0,configurable:!0}),t.prototype.ngDoCheck=function(){if(this._iterableDiffer){var t=this._iterableDiffer.diff(this._rawClass);t&&this._applyIterableChanges(t)}else if(this._keyValueDiffer){var e=this._keyValueDiffer.diff(this._rawClass);e&&this._applyKeyValueChanges(e)}},t.prototype._cleanupClasses=function(t){this._applyClasses(t,!0),this._applyInitialClasses(!1)},t.prototype._applyKeyValueChanges=function(t){var e=this;t.forEachAddedItem(function(t){return e._toggleClass(t.key,t.currentValue)}),t.forEachChangedItem(function(t){return e._toggleClass(t.key,t.currentValue)}),t.forEachRemovedItem(function(t){t.previousValue&&e._toggleClass(t.key,!1)})},t.prototype._applyIterableChanges=function(t){var e=this;t.forEachAddedItem(function(t){if("string"!=typeof t.item)throw new Error("NgClass can only toggle CSS classes expressed as strings, got "+Object(r._48)(t.item));e._toggleClass(t.item,!0)}),t.forEachRemovedItem(function(t){return e._toggleClass(t.item,!1)})},t.prototype._applyInitialClasses=function(t){var e=this;this._initialClasses.forEach(function(n){return e._toggleClass(n,!t)})},t.prototype._applyClasses=function(t,e){var n=this;t&&(Array.isArray(t)||t instanceof Set?t.forEach(function(t){return n._toggleClass(t,!e)}):Object.keys(t).forEach(function(r){null!=t[r]&&n._toggleClass(r,!e)}))},t.prototype._toggleClass=function(t,e){var n=this;(t=t.trim())&&t.split(/\s+/g).forEach(function(t){e?n._renderer.addClass(n._ngEl.nativeElement,t):n._renderer.removeClass(n._ngEl.nativeElement,t)})},t.decorators=[{type:r.s,args:[{selector:"[ngClass]"}]}],t.ctorParameters=function(){return[{type:r.D},{type:r.E},{type:r.t},{type:r.U}]},t.propDecorators={klass:[{type:r.C,args:["class"]}],ngClass:[{type:r.C}]},t}(),I=function(){function t(t){this._viewContainerRef=t,this._componentRef=null,this._moduleRef=null}return t.prototype.ngOnChanges=function(t){if(this._viewContainerRef.clear(),this._componentRef=null,this.ngComponentOutlet){var e=this.ngComponentOutletInjector||this._viewContainerRef.parentInjector;if(t.ngComponentOutletNgModuleFactory)if(this._moduleRef&&this._moduleRef.destroy(),this.ngComponentOutletNgModuleFactory){var n=e.get(r.J);this._moduleRef=this.ngComponentOutletNgModuleFactory.create(n.injector)}else this._moduleRef=null;var i=(this._moduleRef?this._moduleRef.componentFactoryResolver:e.get(r.o)).resolveComponentFactory(this.ngComponentOutlet);this._componentRef=this._viewContainerRef.createComponent(i,this._viewContainerRef.length,e,this.ngComponentOutletContent)}},t.prototype.ngOnDestroy=function(){this._moduleRef&&this._moduleRef.destroy()},t.decorators=[{type:r.s,args:[{selector:"[ngComponentOutlet]"}]}],t.ctorParameters=function(){return[{type:r._9}]},t.propDecorators={ngComponentOutlet:[{type:r.C}],ngComponentOutletInjector:[{type:r.C}],ngComponentOutletContent:[{type:r.C}],ngComponentOutletNgModuleFactory:[{type:r.C}]},t}(),R=function(){function t(t,e,n,r){this.$implicit=t,this.ngForOf=e,this.index=n,this.count=r}return Object.defineProperty(t.prototype,"first",{get:function(){return 0===this.index},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"last",{get:function(){return this.index===this.count-1},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"even",{get:function(){return this.index%2==0},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"odd",{get:function(){return!this.even},enumerable:!0,configurable:!0}),t}(),N=function(){function t(t,e,n){this._viewContainer=t,this._template=e,this._differs=n,this._differ=null}return Object.defineProperty(t.prototype,"ngForTrackBy",{get:function(){return this._trackByFn},set:function(t){Object(r._16)()&&null!=t&&"function"!=typeof t&&console&&console.warn&&console.warn("trackBy must be a function, but received "+JSON.stringify(t)+". See https://angular.io/docs/ts/latest/api/common/index/NgFor-directive.html#!#change-propagation for more information."),this._trackByFn=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngForTemplate",{set:function(t){t&&(this._template=t)},enumerable:!0,configurable:!0}),t.prototype.ngOnChanges=function(t){if("ngForOf"in t){var e=t.ngForOf.currentValue;if(!this._differ&&e)try{this._differ=this._differs.find(e).create(this.ngForTrackBy)}catch(t){throw new Error("Cannot find a differ supporting object '"+e+"' of type '"+((n=e).name||typeof n)+"'. NgFor only supports binding to Iterables such as Arrays.")}}var n},t.prototype.ngDoCheck=function(){if(this._differ){var t=this._differ.diff(this.ngForOf);t&&this._applyChanges(t)}},t.prototype._applyChanges=function(t){var e=this,n=[];t.forEachOperation(function(t,r,i){if(null==t.previousIndex){var o=e._viewContainer.createEmbeddedView(e._template,new R(null,e.ngForOf,-1,-1),i),s=new L(t,o);n.push(s)}else if(null==i)e._viewContainer.remove(r);else{o=e._viewContainer.get(r);e._viewContainer.move(o,i);s=new L(t,o);n.push(s)}});for(var r=0;r<n.length;r++)this._perViewChange(n[r].view,n[r].record);r=0;for(var i=this._viewContainer.length;r<i;r++){var o=this._viewContainer.get(r);o.context.index=r,o.context.count=i}t.forEachIdentityChange(function(t){e._viewContainer.get(t.currentIndex).context.$implicit=t.item})},t.prototype._perViewChange=function(t,e){t.context.$implicit=e.item},t.decorators=[{type:r.s,args:[{selector:"[ngFor][ngForOf]"}]}],t.ctorParameters=function(){return[{type:r._9},{type:r._4},{type:r.D}]},t.propDecorators={ngForOf:[{type:r.C}],ngForTrackBy:[{type:r.C}],ngForTemplate:[{type:r.C}]},t}(),L=function(){return function(t,e){this.record=t,this.view=e}}();var F=function(){function t(t,e){this._viewContainer=t,this._context=new V,this._thenTemplateRef=null,this._elseTemplateRef=null,this._thenViewRef=null,this._elseViewRef=null,this._thenTemplateRef=e}return Object.defineProperty(t.prototype,"ngIf",{set:function(t){this._context.$implicit=this._context.ngIf=t,this._updateView()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngIfThen",{set:function(t){this._thenTemplateRef=t,this._thenViewRef=null,this._updateView()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngIfElse",{set:function(t){this._elseTemplateRef=t,this._elseViewRef=null,this._updateView()},enumerable:!0,configurable:!0}),t.prototype._updateView=function(){this._context.$implicit?this._thenViewRef||(this._viewContainer.clear(),this._elseViewRef=null,this._thenTemplateRef&&(this._thenViewRef=this._viewContainer.createEmbeddedView(this._thenTemplateRef,this._context))):this._elseViewRef||(this._viewContainer.clear(),this._thenViewRef=null,this._elseTemplateRef&&(this._elseViewRef=this._viewContainer.createEmbeddedView(this._elseTemplateRef,this._context)))},t.decorators=[{type:r.s,args:[{selector:"[ngIf]"}]}],t.ctorParameters=function(){return[{type:r._9},{type:r._4}]},t.propDecorators={ngIf:[{type:r.C}],ngIfThen:[{type:r.C}],ngIfElse:[{type:r.C}]},t}(),V=function(){return function(){this.$implicit=null,this.ngIf=null}}(),z=function(){function t(t,e){this._viewContainerRef=t,this._templateRef=e,this._created=!1}return t.prototype.create=function(){this._created=!0,this._viewContainerRef.createEmbeddedView(this._templateRef)},t.prototype.destroy=function(){this._created=!1,this._viewContainerRef.clear()},t.prototype.enforceState=function(t){t&&!this._created?this.create():!t&&this._created&&this.destroy()},t}(),B=function(){function t(){this._defaultUsed=!1,this._caseCount=0,this._lastCaseCheckIndex=0,this._lastCasesMatched=!1}return Object.defineProperty(t.prototype,"ngSwitch",{set:function(t){this._ngSwitch=t,0===this._caseCount&&this._updateDefaultCases(!0)},enumerable:!0,configurable:!0}),t.prototype._addCase=function(){return this._caseCount++},t.prototype._addDefault=function(t){this._defaultViews||(this._defaultViews=[]),this._defaultViews.push(t)},t.prototype._matchCase=function(t){var e=t==this._ngSwitch;return this._lastCasesMatched=this._lastCasesMatched||e,this._lastCaseCheckIndex++,this._lastCaseCheckIndex===this._caseCount&&(this._updateDefaultCases(!this._lastCasesMatched),this._lastCaseCheckIndex=0,this._lastCasesMatched=!1),e},t.prototype._updateDefaultCases=function(t){if(this._defaultViews&&t!==this._defaultUsed){this._defaultUsed=t;for(var e=0;e<this._defaultViews.length;e++){this._defaultViews[e].enforceState(t)}}},t.decorators=[{type:r.s,args:[{selector:"[ngSwitch]"}]}],t.ctorParameters=function(){return[]},t.propDecorators={ngSwitch:[{type:r.C}]},t}(),q=function(){function t(t,e,n){this.ngSwitch=n,n._addCase(),this._view=new z(t,e)}return t.prototype.ngDoCheck=function(){this._view.enforceState(this.ngSwitch._matchCase(this.ngSwitchCase))},t.decorators=[{type:r.s,args:[{selector:"[ngSwitchCase]"}]}],t.ctorParameters=function(){return[{type:r._9},{type:r._4},{type:B,decorators:[{type:r.x}]}]},t.propDecorators={ngSwitchCase:[{type:r.C}]},t}(),H=function(){function t(t,e,n){n._addDefault(new z(t,e))}return t.decorators=[{type:r.s,args:[{selector:"[ngSwitchDefault]"}]}],t.ctorParameters=function(){return[{type:r._9},{type:r._4},{type:B,decorators:[{type:r.x}]}]},t}(),U=function(){function t(t){this._localization=t,this._caseViews={}}return Object.defineProperty(t.prototype,"ngPlural",{set:function(t){this._switchValue=t,this._updateView()},enumerable:!0,configurable:!0}),t.prototype.addCase=function(t,e){this._caseViews[t]=e},t.prototype._updateView=function(){this._clearViews();var t=Object.keys(this._caseViews),e=M(this._switchValue,t,this._localization);this._activateView(this._caseViews[e])},t.prototype._clearViews=function(){this._activeView&&this._activeView.destroy()},t.prototype._activateView=function(t){t&&(this._activeView=t,this._activeView.create())},t.decorators=[{type:r.s,args:[{selector:"[ngPlural]"}]}],t.ctorParameters=function(){return[{type:P}]},t.propDecorators={ngPlural:[{type:r.C}]},t}(),W=function(){function t(t,e,n,r){this.value=t;var i=!isNaN(Number(t));r.addCase(i?"="+t:t,new z(n,e))}return t.decorators=[{type:r.s,args:[{selector:"[ngPluralCase]"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.g,args:["ngPluralCase"]}]},{type:r._4},{type:r._9},{type:U,decorators:[{type:r.x}]}]},t}(),G=function(){function t(t,e,n){this._differs=t,this._ngEl=e,this._renderer=n}return Object.defineProperty(t.prototype,"ngStyle",{set:function(t){this._ngStyle=t,!this._differ&&t&&(this._differ=this._differs.find(t).create())},enumerable:!0,configurable:!0}),t.prototype.ngDoCheck=function(){if(this._differ){var t=this._differ.diff(this._ngStyle);t&&this._applyChanges(t)}},t.prototype._applyChanges=function(t){var e=this;t.forEachRemovedItem(function(t){return e._setStyle(t.key,null)}),t.forEachAddedItem(function(t){return e._setStyle(t.key,t.currentValue)}),t.forEachChangedItem(function(t){return e._setStyle(t.key,t.currentValue)})},t.prototype._setStyle=function(t,e){var n=t.split("."),r=n[0],i=n[1];e=null!=e&&i?""+e+i:e,this._renderer.setStyle(this._ngEl.nativeElement,r,e)},t.decorators=[{type:r.s,args:[{selector:"[ngStyle]"}]}],t.ctorParameters=function(){return[{type:r.E},{type:r.t},{type:r.U}]},t.propDecorators={ngStyle:[{type:r.C}]},t}(),X=function(){function t(t){this._viewContainerRef=t}return t.prototype.ngOnChanges=function(t){this._shouldRecreateView(t)?(this._viewRef&&this._viewContainerRef.remove(this._viewContainerRef.indexOf(this._viewRef)),this.ngTemplateOutlet&&(this._viewRef=this._viewContainerRef.createEmbeddedView(this.ngTemplateOutlet,this.ngTemplateOutletContext))):this._viewRef&&this.ngTemplateOutletContext&&this._updateExistingContext(this.ngTemplateOutletContext)},t.prototype._shouldRecreateView=function(t){var e=t.ngTemplateOutletContext;return!!t.ngTemplateOutlet||e&&this._hasContextShapeChanged(e)},t.prototype._hasContextShapeChanged=function(t){var e=Object.keys(t.previousValue||{}),n=Object.keys(t.currentValue||{});if(e.length===n.length){for(var r=0,i=n;r<i.length;r++){var o=i[r];if(-1===e.indexOf(o))return!0}return!1}return!0},t.prototype._updateExistingContext=function(t){for(var e=0,n=Object.keys(t);e<n.length;e++){var r=n[e];this._viewRef.context[r]=this.ngTemplateOutletContext[r]}},t.decorators=[{type:r.s,args:[{selector:"[ngTemplateOutlet]"}]}],t.ctorParameters=function(){return[{type:r._9}]},t.propDecorators={ngTemplateOutletContext:[{type:r.C}],ngTemplateOutlet:[{type:r.C}]},t}(),Y=[D,I,N,F,X,G,B,q,H,U,W],K={},$=/((?:[^GyMLwWdEabBhHmsSzZO']+)|(?:'(?:[^']|'')*')|(?:G{1,5}|y{1,4}|M{1,5}|L{1,5}|w{1,2}|W{1}|d{1,2}|E{1,6}|a{1,5}|b{1,5}|B{1,5}|h{1,2}|H{1,2}|m{1,2}|s{1,2}|S{1,3}|z{1,4}|Z{1,5}|O{1,4}))([\s\S]*)/,Q={Short:0,ShortGMT:1,Long:2,Extended:3};Q[Q.Short]="Short",Q[Q.ShortGMT]="ShortGMT",Q[Q.Long]="Long",Q[Q.Extended]="Extended";var Z={FullYear:0,Month:1,Date:2,Hours:3,Minutes:4,Seconds:5,Milliseconds:6,Day:7};Z[Z.FullYear]="FullYear",Z[Z.Month]="Month",Z[Z.Date]="Date",Z[Z.Hours]="Hours",Z[Z.Minutes]="Minutes",Z[Z.Seconds]="Seconds",Z[Z.Milliseconds]="Milliseconds",Z[Z.Day]="Day";var J={DayPeriods:0,Days:1,Months:2,Eras:3};function tt(t,e,n,r){e=function t(e,n){var r=(i=e,E(i)[0]);var i;K[r]=K[r]||{};if(K[r][n])return K[r][n];var o="";switch(n){case"shortDate":o=_(e,v.Short);break;case"mediumDate":o=_(e,v.Medium);break;case"longDate":o=_(e,v.Long);break;case"fullDate":o=_(e,v.Full);break;case"shortTime":o=j(e,v.Short);break;case"mediumTime":o=j(e,v.Medium);break;case"longTime":o=j(e,v.Long);break;case"fullTime":o=j(e,v.Full);break;case"short":var s=t(e,"shortTime"),a=t(e,"shortDate");o=et(x(e,v.Short),[s,a]);break;case"medium":var c=t(e,"mediumTime"),u=t(e,"mediumDate");o=et(x(e,v.Medium),[c,u]);break;case"long":var l=t(e,"longTime"),p=t(e,"longDate");o=et(x(e,v.Long),[l,p]);break;case"full":var d=t(e,"fullTime"),h=t(e,"fullDate");o=et(x(e,v.Full),[d,h])}o&&(K[r][n]=o);return o}(n,e)||e;for(var i,o=[];e;){if(!(i=$.exec(e))){o.push(e);break}var s=(o=o.concat(i.slice(1))).pop();if(!s)break;e=s}var a,c,u,l,p,d,h,f=t.getTimezoneOffset();r&&(f=lt(r,f),a=t,c=r,u=!0?-1:1,l=a.getTimezoneOffset(),p=lt(c,l),d=a,h=u*(p-l),(d=new Date(d.getTime())).setMinutes(d.getMinutes()+h),t=d);var y="";return o.forEach(function(e){var r=function(t){if(ut[t])return ut[t];var e;switch(t){case"G":case"GG":case"GGG":e=it(J.Eras,g.Abbreviated);break;case"GGGG":e=it(J.Eras,g.Wide);break;case"GGGGG":e=it(J.Eras,g.Narrow);break;case"y":e=rt(Z.FullYear,1,0,!1,!0);break;case"yy":e=rt(Z.FullYear,2,0,!0,!0);break;case"yyy":e=rt(Z.FullYear,3,0,!1,!0);break;case"yyyy":e=rt(Z.FullYear,4,0,!1,!0);break;case"M":case"L":e=rt(Z.Month,1,1);break;case"MM":case"LL":e=rt(Z.Month,2,1);break;case"MMM":e=it(J.Months,g.Abbreviated);break;case"MMMM":e=it(J.Months,g.Wide);break;case"MMMMM":e=it(J.Months,g.Narrow);break;case"LLL":e=it(J.Months,g.Abbreviated,m.Standalone);break;case"LLLL":e=it(J.Months,g.Wide,m.Standalone);break;case"LLLLL":e=it(J.Months,g.Narrow,m.Standalone);break;case"w":e=ct(1);break;case"ww":e=ct(2);break;case"W":e=ct(1,!0);break;case"d":e=rt(Z.Date,1);break;case"dd":e=rt(Z.Date,2);break;case"E":case"EE":case"EEE":e=it(J.Days,g.Abbreviated);break;case"EEEE":e=it(J.Days,g.Wide);break;case"EEEEE":e=it(J.Days,g.Narrow);break;case"EEEEEE":e=it(J.Days,g.Short);break;case"a":case"aa":case"aaa":e=it(J.DayPeriods,g.Abbreviated);break;case"aaaa":e=it(J.DayPeriods,g.Wide);break;case"aaaaa":e=it(J.DayPeriods,g.Narrow);break;case"b":case"bb":case"bbb":e=it(J.DayPeriods,g.Abbreviated,m.Standalone,!0);break;case"bbbb":e=it(J.DayPeriods,g.Wide,m.Standalone,!0);break;case"bbbbb":e=it(J.DayPeriods,g.Narrow,m.Standalone,!0);break;case"B":case"BB":case"BBB":e=it(J.DayPeriods,g.Abbreviated,m.Format,!0);break;case"BBBB":e=it(J.DayPeriods,g.Wide,m.Format,!0);break;case"BBBBB":e=it(J.DayPeriods,g.Narrow,m.Format,!0);break;case"h":e=rt(Z.Hours,1,-12);break;case"hh":e=rt(Z.Hours,2,-12);break;case"H":e=rt(Z.Hours,1);break;case"HH":e=rt(Z.Hours,2);break;case"m":e=rt(Z.Minutes,1);break;case"mm":e=rt(Z.Minutes,2);break;case"s":e=rt(Z.Seconds,1);break;case"ss":e=rt(Z.Seconds,2);break;case"S":e=rt(Z.Milliseconds,1);break;case"SS":e=rt(Z.Milliseconds,2);break;case"SSS":e=rt(Z.Milliseconds,3);break;case"Z":case"ZZ":case"ZZZ":e=ot(Q.Short);break;case"ZZZZZ":e=ot(Q.Extended);break;case"O":case"OO":case"OOO":case"z":case"zz":case"zzz":e=ot(Q.ShortGMT);break;case"OOOO":case"ZZZZ":case"zzzz":e=ot(Q.Long);break;default:return null}return ut[t]=e,e}(e);y+=r?r(t,n,f):"''"===e?"'":e.replace(/(^'|'$)/g,"").replace(/''/g,"'")}),y}function et(t,e){return e&&(t=t.replace(/\{([^}]+)}/g,function(t,n){return null!=e&&n in e?e[n]:t})),t}function nt(t,e,n,r,i){void 0===n&&(n="-");var o="";(t<0||i&&t<=0)&&(i?t=1-t:(t=-t,o=n));for(var s=""+t;s.length<e;)s="0"+s;return r&&(s=s.substr(s.length-e)),o+s}function rt(t,e,n,r,i){return void 0===n&&(n=0),void 0===r&&(r=!1),void 0===i&&(i=!1),function(o,s){var a=function(t,e,n){switch(t){case Z.FullYear:return e.getFullYear();case Z.Month:return e.getMonth();case Z.Date:return e.getDate();case Z.Hours:return e.getHours();case Z.Minutes:return e.getMinutes();case Z.Seconds:return e.getSeconds();case Z.Milliseconds:var r=1===n?100:2===n?10:1;return Math.round(e.getMilliseconds()/r);case Z.Day:return e.getDay();default:throw new Error('Unknown DateType value "'+t+'".')}}(t,o,e);return(n>0||a>-n)&&(a+=n),t===Z.Hours&&0===a&&-12===n&&(a=12),nt(a,e,w(s,y.MinusSign),r,i)}}function it(t,e,n,r){return void 0===n&&(n=m.Format),void 0===r&&(r=!1),function(i,o){return function(t,e,n,r,i,o){switch(n){case J.Months:return(P=e,M=i,A=r,T=E(P),C(C([T[5],T[6]],M),A))[t.getMonth()];case J.Days:return(j=e,x=i,w=r,S=E(j),C(C([S[3],S[4]],x),w))[t.getDay()];case J.DayPeriods:var s=t.getHours(),a=t.getMinutes();if(o){var c,u=(O(_=E(e)),(_[18][2]||[]).map(function(t){return"string"==typeof t?k(t):[k(t[0]),k(t[1])]})),l=(v=i,y=r,O(b=E(e)),C(C([b[18][0],b[18][1]],v)||[],y)||[]);if(u.forEach(function(t,e){if(Array.isArray(t)){var n=t[0],r=n.hours,i=n.minutes,o=t[1],u=o.hours,p=o.minutes;s>=r&&a>=i&&(s<u||s===u&&a<p)&&(c=l[e])}else{var d=t.hours,h=t.minutes;d===s&&h===a&&(c=l[e])}}),c)return c}return(h=e,f=i,m=r,g=E(h),C(C([g[1],g[2]],f),m))[s<12?0:1];case J.Eras:return(p=e,d=r,C(E(p)[7],d))[t.getFullYear()<=0?0:1]}var p,d;var h,f,m,g;var v,y,b;var _;var j,x,w,S;var P,M,A,T}(i,o,t,e,n,r)}}function ot(t){return function(e,n,r){var i=-1*r,o=w(n,y.MinusSign),s=i>0?Math.floor(i/60):Math.ceil(i/60);switch(t){case Q.Short:return(i>=0?"+":"")+nt(s,2,o)+nt(Math.abs(i%60),2,o);case Q.ShortGMT:return"GMT"+(i>=0?"+":"")+nt(s,1,o);case Q.Long:return"GMT"+(i>=0?"+":"")+nt(s,2,o)+":"+nt(Math.abs(i%60),2,o);case Q.Extended:return 0===r?"Z":(i>=0?"+":"")+nt(s,2,o)+":"+nt(Math.abs(i%60),2,o);default:throw new Error('Unknown zone width "'+t+'"')}}}J[J.DayPeriods]="DayPeriods",J[J.Days]="Days",J[J.Months]="Months",J[J.Eras]="Eras";var st=0,at=4;function ct(t,e){return void 0===e&&(e=!1),function(n,r){var i,o,s,a;if(e){var c=new Date(n.getFullYear(),n.getMonth(),1).getDay()-1,u=n.getDate();i=1+Math.floor((u+c)/7)}else{var l=(s=n.getFullYear(),a=new Date(s,st,1).getDay(),new Date(s,0,1+(a<=at?at:at+7)-a)),p=(o=n,new Date(o.getFullYear(),o.getMonth(),o.getDate()+(at-o.getDay()))).getTime()-l.getTime();i=1+Math.round(p/6048e5)}return nt(i,t,w(r,y.MinusSign))}}var ut={};function lt(t,e){t=t.replace(/:/g,"");var n=Date.parse("Jan 01, 1970 00:00:00 "+t)/6e4;return isNaN(n)?e:n}function pt(t,e){return Error("InvalidPipeArgument: '"+e+"' for pipe '"+Object(r._48)(t)+"'")}var dt=/^(\d{4})-?(\d\d)-?(\d\d)(?:T(\d\d)(?::?(\d\d)(?::?(\d\d)(?:\.(\d+))?)?)?(Z|([+-])(\d\d):?(\d\d))?)?$/,ht=function(){function t(t){this.locale=t}return t.prototype.transform=function(e,n,r,i){if(void 0===n&&(n="mediumDate"),null==e||""===e||e!=e)return null;var o;if("string"==typeof e&&(e=e.trim()),mt(e))o=e;else if(isNaN(e-parseFloat(e)))if("string"==typeof e&&/^(\d{4}-\d{1,2}-\d{1,2})$/.test(e)){var s=e.split("-").map(function(t){return+t}),a=s[0],c=s[1],u=s[2];o=new Date(a,c-1,u)}else o=new Date(e);else o=new Date(parseFloat(e));if(!mt(o)){var l=void 0;if("string"!=typeof e||!(l=e.match(dt)))throw pt(t,e);o=ft(l)}return tt(o,n,i||this.locale,r)},t.decorators=[{type:r.R,args:[{name:"date",pure:!0}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[r.F]}]}]},t}();function ft(t){var e=new Date(0),n=0,r=0,i=t[8]?e.setUTCFullYear:e.setFullYear,o=t[8]?e.setUTCHours:e.setHours;t[9]&&(n=+(t[9]+t[10]),r=+(t[9]+t[11])),i.call(e,+t[1],+t[2]-1,+t[3]);var s=+(t[4]||"0")-n,a=+(t[5]||"0")-r,c=+(t[6]||"0"),u=Math.round(1e3*parseFloat("0."+(t[7]||0)));return o.call(e,s,a,c,u),e}function mt(t){return t instanceof Date&&!isNaN(t.valueOf())}var gt,vt=function(){function t(){}return t.format=function(t,e,n,r){void 0===r&&(r={});var i=r.minimumIntegerDigits,o=r.minimumFractionDigits,s=r.maximumFractionDigits,a=r.currency,c=r.currencyAsSymbol,u=void 0!==c&&c,l={minimumIntegerDigits:i,minimumFractionDigits:o,maximumFractionDigits:s,style:h[n].toLowerCase()};return n==h.Currency&&(l.currency="string"==typeof a?a:void 0,l.currencyDisplay=u?"symbol":"code"),new Intl.NumberFormat(e,l).format(t)},t}(),yt=/((?:[^yMLdHhmsazZEwGjJ']+)|(?:'(?:[^']|'')*')|(?:E+|y+|M+|L+|d+|H+|h+|J+|j+|m+|s+|a|z|Z|G+|w+))(.*)/,bt={yMMMdjms:Pt(St([kt("year",1),Et("month",3),kt("day",1),kt("hour",1),kt("minute",1),kt("second",1)])),yMdjm:Pt(St([kt("year",1),kt("month",1),kt("day",1),kt("hour",1),kt("minute",1)])),yMMMMEEEEd:Pt(St([kt("year",1),Et("month",4),Et("weekday",4),kt("day",1)])),yMMMMd:Pt(St([kt("year",1),Et("month",4),kt("day",1)])),yMMMd:Pt(St([kt("year",1),Et("month",3),kt("day",1)])),yMd:Pt(St([kt("year",1),kt("month",1),kt("day",1)])),jms:Pt(St([kt("hour",1),kt("second",1),kt("minute",1)])),jm:Pt(St([kt("hour",1),kt("minute",1)]))},_t={yyyy:Pt(kt("year",4)),yy:Pt(kt("year",2)),y:Pt(kt("year",1)),MMMM:Pt(Et("month",4)),MMM:Pt(Et("month",3)),MM:Pt(kt("month",2)),M:Pt(kt("month",1)),LLLL:Pt(Et("month",4)),L:Pt(Et("month",1)),dd:Pt(kt("day",2)),d:Pt(kt("day",1)),HH:jt(xt(Pt(Ct(kt("hour",2),!1)))),H:xt(Pt(Ct(kt("hour",1),!1))),hh:jt(xt(Pt(Ct(kt("hour",2),!0)))),h:xt(Pt(Ct(kt("hour",1),!0))),jj:Pt(kt("hour",2)),j:Pt(kt("hour",1)),mm:jt(Pt(kt("minute",2))),m:Pt(kt("minute",1)),ss:jt(Pt(kt("second",2))),s:Pt(kt("second",1)),sss:Pt(kt("second",3)),EEEE:Pt(Et("weekday",4)),EEE:Pt(Et("weekday",3)),EE:Pt(Et("weekday",2)),E:Pt(Et("weekday",1)),a:(gt=Pt(Ct(kt("hour",1),!0)),function(t,e){return gt(t,e).split(" ")[1]}),Z:Ot("short"),z:Ot("long"),ww:Pt({}),w:Pt({}),G:Pt(Et("era",1)),GG:Pt(Et("era",2)),GGG:Pt(Et("era",3)),GGGG:Pt(Et("era",4))};function jt(t){return function(e,n){var r=t(e,n);return 1==r.length?"0"+r:r}}function xt(t){return function(e,n){return t(e,n).split(" ")[0]}}function wt(t,e,n){return new Intl.DateTimeFormat(e,n).format(t).replace(/[\u200e\u200f]/g,"")}function Ot(t){var e={hour:"2-digit",hour12:!1,timeZoneName:t};return function(t,n){var r=wt(t,n,e);return r?r.substring(3):""}}function Ct(t,e){return t.hour12=e,t}function kt(t,e){var n={};return n[t]=2===e?"2-digit":"numeric",n}function Et(t,e){var n={};return n[t]=e<4?e>1?"short":"narrow":"long",n}function St(t){return t.reduce(function(t,e){return Object(i.a)({},t,e)},{})}function Pt(t){return function(e,n){return wt(e,n,t)}}var Mt=new Map;var At=function(){function t(){}return t.format=function(t,e,n){return function(t,e,n){var r=bt[t];if(r)return r(e,n);var i=t,o=Mt.get(i);if(!o){o=[];var s=void 0;yt.exec(t);for(var a=t;a;)(s=yt.exec(a))?a=(o=o.concat(s.slice(1))).pop():(o.push(a),a=null);Mt.set(i,o)}return o.reduce(function(t,r){var i,o=_t[r];return t+(o?o(e,n):"''"===(i=r)?"'":i.replace(/(^'|'$)/g,"").replace(/''/g,"'"))},"")}(n,t,e)},t}(),Tt=function(){function t(t){this._locale=t}return t.prototype.transform=function(e,n){if(void 0===n&&(n="mediumDate"),null==e||""===e||e!=e)return null;var r;if("string"==typeof e&&(e=e.trim()),Dt(e))r=e;else if(isNaN(e-parseFloat(e)))if("string"==typeof e&&/^(\d{4}-\d{1,2}-\d{1,2})$/.test(e)){var i=e.split("-").map(function(t){return parseInt(t,10)}),o=i[0],s=i[1],a=i[2];r=new Date(o,s-1,a)}else r=new Date(e);else r=new Date(parseFloat(e));if(!Dt(r)){var c=void 0;if("string"!=typeof e||!(c=e.match(dt)))throw pt(t,e);r=ft(c)}return At.format(r,this._locale,t._ALIASES[n]||n)},t._ALIASES={medium:"yMMMdjms",short:"yMdjm",fullDate:"yMMMMEEEEd",longDate:"yMMMMd",mediumDate:"yMMMd",shortDate:"yMd",mediumTime:"jms",shortTime:"jm"},t.decorators=[{type:r.R,args:[{name:"date",pure:!0}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[r.F]}]}]},t}();function Dt(t){return t instanceof Date&&!isNaN(t.valueOf())}var It=/^(\d+)?\.((\d+)(-(\d+))?)?$/,Rt=22,Nt=".",Lt="0",Ft=";",Vt=",",zt="#",Bt="¤",qt="%";function Ht(t,e,n,r,i){void 0===i&&(i=null);var o,s,a={str:null},c=(o=n,E(e)[14][o]);if("string"!=typeof t||isNaN(+t-parseFloat(t))){if("number"!=typeof t)return a.error=t+" is not a number",a;s=t}else s=+t;n===h.Percent&&(s*=100);var u=Math.abs(s)+"",l=function(t,e){void 0===e&&(e="-");var n={minInt:1,minFrac:0,maxFrac:0,posPre:"",posSuf:"",negPre:"",negSuf:"",gSize:0,lgSize:0},r=t.split(Ft),i=r[0],o=r[1],s=-1!==i.indexOf(Nt)?i.split(Nt):[i.substring(0,i.lastIndexOf(Lt)+1),i.substring(i.lastIndexOf(Lt)+1)],a=s[0],c=s[1]||"";n.posPre=a.substr(0,a.indexOf(zt));for(var u=0;u<c.length;u++){var l=c.charAt(u);l===Lt?n.minFrac=n.maxFrac=u+1:l===zt?n.maxFrac=u+1:n.posSuf+=l}var p=a.split(Vt);if(n.gSize=p[1]?p[1].length:0,n.lgSize=p[2]||p[1]?(p[2]||p[1]).length:0,o){var d=i.length-n.posPre.length-n.posSuf.length,h=o.indexOf(zt);n.negPre=o.substr(0,h).replace(/'/g,""),n.negSuf=o.substr(h+d).replace(/'/g,"")}else n.negPre=e+n.posPre,n.negSuf=n.posSuf;return n}(c,w(e,y.MinusSign)),p="",d=!1;if(isFinite(s)){var f=function(t){var e,n,r,i,o,s=0;(n=t.indexOf(Nt))>-1&&(t=t.replace(Nt,""));(r=t.search(/e/i))>0?(n<0&&(n=r),n+=+t.slice(r+1),t=t.substring(0,r)):n<0&&(n=t.length);for(r=0;t.charAt(r)===Lt;r++);if(r===(o=t.length))e=[0],n=1;else{for(o--;t.charAt(o)===Lt;)o--;for(n-=r,e=[],i=0;r<=o;r++,i++)e[i]=+t.charAt(r)}n>Rt&&(e=e.splice(0,Rt-1),s=n-1,n=1);return{digits:e,exponent:s,integerLen:n}}(u),m=l.minInt,g=l.minFrac,v=l.maxFrac;if(r){var b=r.match(It);if(null===b)return a.error=r+" is not a valid digit info",a;var _=b[1],j=b[3],x=b[5];null!=_&&(m=Ut(_)),null!=j&&(g=Ut(j)),null!=x?v=Ut(x):null!=j&&g>v&&(v=g)}!function(t,e,n){if(e>n)throw new Error("The minimum number of digits after fraction ("+e+") is higher than the maximum ("+n+").");var r=t.digits,i=r.length-t.integerLen,o=Math.min(Math.max(e,i),n),s=o+t.integerLen,a=r[s];if(s>0){r.splice(Math.max(t.integerLen,s));for(var c=s;c<r.length;c++)r[c]=0}else{i=Math.max(0,i),t.integerLen=1,r.length=Math.max(1,s=o+1),r[0]=0;for(var u=1;u<s;u++)r[u]=0}if(a>=5)if(s-1<0){for(var l=0;l>s;l--)r.unshift(0),t.integerLen++;r.unshift(1),t.integerLen++}else r[s-1]++;for(;i<Math.max(0,o);i++)r.push(0);var p=r.reduceRight(function(t,e,n,r){return e+=t,r[n]=e%10,Math.floor(e/10)},0);p&&(r.unshift(p),t.integerLen++)}(f,g,v);var O=f.digits,C=f.integerLen,k=f.exponent,S=[];for(d=O.every(function(t){return!t});C<m;C++)O.unshift(0);for(;C<0;C++)O.unshift(0);C>0?S=O.splice(C,O.length):(S=O,O=[0]);var P=[];for(O.length>=l.lgSize&&P.unshift(O.splice(-l.lgSize,O.length).join(""));O.length>l.gSize;)P.unshift(O.splice(-l.gSize,O.length).join(""));O.length&&P.unshift(O.join(""));var M=i?y.CurrencyGroup:y.Group;if(p=P.join(w(e,M)),S.length)p+=w(e,i?y.CurrencyDecimal:y.Decimal)+S.join("");k&&(p+=w(e,y.Exponential)+"+"+k)}else p=w(e,y.Infinity);return p=s<0&&!d?l.negPre+p+l.negSuf:l.posPre+p+l.posSuf,n===h.Currency&&null!==i?(a.str=p.replace(Bt,i).replace(Bt,""),a):n===h.Percent?(a.str=p.replace(new RegExp(qt,"g"),w(e,y.PercentSign)),a):(a.str=p,a)}function Ut(t){var e=parseInt(t);if(isNaN(e))throw new Error("Invalid integer literal when parsing "+t);return e}function Wt(t,e,n,r,i,o,s){if(void 0===o&&(o=null),void 0===s&&(s=!1),null==n)return null;if("number"!=typeof(n="string"!=typeof n||isNaN(+n-parseFloat(n))?n:+n))throw pt(t,n);var a,c,u;if(r!==h.Currency&&(a=1,c=0,u=3),i){var l=i.match(It);if(null===l)throw new Error(i+" is not a valid digit info for number pipes");null!=l[1]&&(a=Ut(l[1])),null!=l[3]&&(c=Ut(l[3])),null!=l[5]&&(u=Ut(l[5]))}return vt.format(n,e,r,{minimumIntegerDigits:a,minimumFractionDigits:c,maximumFractionDigits:u,currency:o,currencyAsSymbol:s})}var Gt=[function(){function t(t){this._locale=t}return t.prototype.transform=function(e,n){return Wt(t,this._locale,e,h.Decimal,n)},t.decorators=[{type:r.R,args:[{name:"number"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[r.F]}]}]},t}(),function(){function t(t){this._locale=t}return t.prototype.transform=function(e,n){return Wt(t,this._locale,e,h.Percent,n)},t.decorators=[{type:r.R,args:[{name:"percent"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[r.F]}]}]},t}(),function(){function t(t){this._locale=t}return t.prototype.transform=function(e,n,r,i){return void 0===n&&(n="USD"),void 0===r&&(r=!1),Wt(t,this._locale,e,h.Currency,i,n,r)},t.decorators=[{type:r.R,args:[{name:"currency"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[r.F]}]}]},t}(),Tt],Xt=function(){function t(){}return t.prototype.createSubscription=function(t,e){return t.subscribe({next:e,error:function(t){throw t}})},t.prototype.dispose=function(t){t.unsubscribe()},t.prototype.onDestroy=function(t){t.unsubscribe()},t}(),Yt=new(function(){function t(){}return t.prototype.createSubscription=function(t,e){return t.then(e,function(t){throw t})},t.prototype.dispose=function(t){},t.prototype.onDestroy=function(t){},t}()),Kt=new Xt,$t=function(){function t(t){this._ref=t,this._latestValue=null,this._latestReturnedValue=null,this._subscription=null,this._obj=null,this._strategy=null}return t.prototype.ngOnDestroy=function(){this._subscription&&this._dispose()},t.prototype.transform=function(t){return this._obj?t!==this._obj?(this._dispose(),this.transform(t)):this._latestValue===this._latestReturnedValue?this._latestReturnedValue:(this._latestReturnedValue=this._latestValue,r._11.wrap(this._latestValue)):(t&&this._subscribe(t),this._latestReturnedValue=this._latestValue,this._latestValue)},t.prototype._subscribe=function(t){var e=this;this._obj=t,this._strategy=this._selectStrategy(t),this._subscription=this._strategy.createSubscription(t,function(n){return e._updateLatestValue(t,n)})},t.prototype._selectStrategy=function(e){if(Object(r._35)(e))return Yt;if(Object(r._34)(e))return Kt;throw pt(t,e)},t.prototype._dispose=function(){this._strategy.dispose(this._subscription),this._latestValue=null,this._latestReturnedValue=null,this._subscription=null,this._obj=null},t.prototype._updateLatestValue=function(t,e){t===this._obj&&(this._latestValue=e,this._ref.markForCheck())},t.decorators=[{type:r.R,args:[{name:"async",pure:!1}]}],t.ctorParameters=function(){return[{type:r.j}]},t}(),Qt=function(){function t(){}return t.prototype.transform=function(e){if(!e)return e;if("string"!=typeof e)throw pt(t,e);return e.toLowerCase()},t.decorators=[{type:r.R,args:[{name:"lowercase"}]}],t.ctorParameters=function(){return[]},t}();var Zt=function(){function t(){}return t.prototype.transform=function(e){if(!e)return e;if("string"!=typeof e)throw pt(t,e);return e.split(/\b/g).map(function(t){return(e=t)?e[0].toUpperCase()+e.substr(1).toLowerCase():e;var e}).join("")},t.decorators=[{type:r.R,args:[{name:"titlecase"}]}],t.ctorParameters=function(){return[]},t}(),Jt=function(){function t(){}return t.prototype.transform=function(e){if(!e)return e;if("string"!=typeof e)throw pt(t,e);return e.toUpperCase()},t.decorators=[{type:r.R,args:[{name:"uppercase"}]}],t.ctorParameters=function(){return[]},t}(),te=/#/g,ee=function(){function t(t){this._localization=t}return t.prototype.transform=function(e,n,r){if(null==e)return"";if("object"!=typeof n||null===n)throw pt(t,n);return n[M(e,Object.keys(n),this._localization,r)].replace(te,e.toString())},t.decorators=[{type:r.R,args:[{name:"i18nPlural",pure:!0}]}],t.ctorParameters=function(){return[{type:P}]},t}(),ne=function(){function t(){}return t.prototype.transform=function(e,n){if(null==e)return"";if("object"!=typeof n||"string"!=typeof e)throw pt(t,n);return n.hasOwnProperty(e)?n[e]:n.hasOwnProperty("other")?n.other:""},t.decorators=[{type:r.R,args:[{name:"i18nSelect",pure:!0}]}],t.ctorParameters=function(){return[]},t}(),re=function(){function t(){}return t.prototype.transform=function(t){return JSON.stringify(t,null,2)},t.decorators=[{type:r.R,args:[{name:"json",pure:!1}]}],t.ctorParameters=function(){return[]},t}(),ie=function(){function t(t){this._locale=t}return t.prototype.transform=function(e,n,r){if(ae(e))return null;var i=Ht(e,r=r||this._locale,h.Decimal,n),o=i.str,s=i.error;if(s)throw pt(t,s);return o},t.decorators=[{type:r.R,args:[{name:"number"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[r.F]}]}]},t}(),oe=function(){function t(t){this._locale=t}return t.prototype.transform=function(e,n,r){if(ae(e))return null;var i=Ht(e,r=r||this._locale,h.Percent,n),o=i.str,s=i.error;if(s)throw pt(t,s);return o},t.decorators=[{type:r.R,args:[{name:"percent"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[r.F]}]}]},t}(),se=function(){function t(t){this._locale=t}return t.prototype.transform=function(e,n,r,i,o){if(void 0===r&&(r="symbol"),ae(e))return null;o=o||this._locale,"boolean"==typeof r&&(console&&console.warn&&console.warn('Warning: the currency pipe has been changed in Angular v5. The symbolDisplay option (third parameter) is now a string instead of a boolean. The accepted values are "code", "symbol" or "symbol-narrow".'),r=r?"symbol":"code");var s,a,c,u,p=n||"USD";"code"!==r&&(s=p,a="symbol"===r?"wide":"narrow",c=l[s]||{},u=c[0]||s,p="wide"===a?u:c[1]||u);var d=Ht(e,o,h.Currency,i,p),f=d.str,m=d.error;if(m)throw pt(t,m);return f},t.decorators=[{type:r.R,args:[{name:"currency"}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[r.F]}]}]},t}();function ae(t){return null==t||""===t||t!=t}var ce=[$t,Jt,Qt,re,function(){function t(){}return t.prototype.transform=function(e,n,r){if(null==e)return e;if(!this.supports(e))throw pt(t,e);return e.slice(n,r)},t.prototype.supports=function(t){return"string"==typeof t||Array.isArray(t)},t.decorators=[{type:r.R,args:[{name:"slice",pure:!1}]}],t.ctorParameters=function(){return[]},t}(),ie,oe,Zt,se,ht,ee,ne],ue=function(){function t(){}return t.decorators=[{type:r.H,args:[{declarations:[Y,ce],exports:[Y,ce],providers:[{provide:P,useClass:A}]}]}],t.ctorParameters=function(){return[]},t}(),le=function(t,e){"string"==typeof e&&(e=parseInt(e,10));var n=e,r=n.toString().replace(/^[^.]*\.?/,""),i=Math.floor(Math.abs(n)),o=r.length,s=parseInt(r,10),a=parseInt(n.toString().replace(/^[^.]*\.?|0+$/g,""),10)||0;switch(t.split("-")[0].toLowerCase()){case"af":case"asa":case"az":case"bem":case"bez":case"bg":case"brx":case"ce":case"cgg":case"chr":case"ckb":case"ee":case"el":case"eo":case"es":case"eu":case"fo":case"fur":case"gsw":case"ha":case"haw":case"hu":case"jgo":case"jmc":case"ka":case"kk":case"kkj":case"kl":case"ks":case"ksb":case"ky":case"lb":case"lg":case"mas":case"mgo":case"ml":case"mn":case"nb":case"nd":case"ne":case"nn":case"nnh":case"nyn":case"om":case"or":case"os":case"ps":case"rm":case"rof":case"rwk":case"saq":case"seh":case"sn":case"so":case"sq":case"ta":case"te":case"teo":case"tk":case"tr":case"ug":case"uz":case"vo":case"vun":case"wae":case"xog":return 1===n?f.One:f.Other;case"ak":case"ln":case"mg":case"pa":case"ti":return n===Math.floor(n)&&n>=0&&n<=1?f.One:f.Other;case"am":case"as":case"bn":case"fa":case"gu":case"hi":case"kn":case"mr":case"zu":return 0===i||1===n?f.One:f.Other;case"ar":return 0===n?f.Zero:1===n?f.One:2===n?f.Two:n%100===Math.floor(n%100)&&n%100>=3&&n%100<=10?f.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=99?f.Many:f.Other;case"ast":case"ca":case"de":case"en":case"et":case"fi":case"fy":case"gl":case"it":case"nl":case"sv":case"sw":case"ur":case"yi":return 1===i&&0===o?f.One:f.Other;case"be":return n%10==1&&n%100!=11?f.One:n%10===Math.floor(n%10)&&n%10>=2&&n%10<=4&&!(n%100>=12&&n%100<=14)?f.Few:n%10==0||n%10===Math.floor(n%10)&&n%10>=5&&n%10<=9||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=14?f.Many:f.Other;case"br":return n%10==1&&n%100!=11&&n%100!=71&&n%100!=91?f.One:n%10==2&&n%100!=12&&n%100!=72&&n%100!=92?f.Two:n%10===Math.floor(n%10)&&(n%10>=3&&n%10<=4||n%10==9)&&!(n%100>=10&&n%100<=19||n%100>=70&&n%100<=79||n%100>=90&&n%100<=99)?f.Few:0!==n&&n%1e6==0?f.Many:f.Other;case"bs":case"hr":case"sr":return 0===o&&i%10==1&&i%100!=11||s%10==1&&s%100!=11?f.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)||s%10===Math.floor(s%10)&&s%10>=2&&s%10<=4&&!(s%100>=12&&s%100<=14)?f.Few:f.Other;case"cs":case"sk":return 1===i&&0===o?f.One:i===Math.floor(i)&&i>=2&&i<=4&&0===o?f.Few:0!==o?f.Many:f.Other;case"cy":return 0===n?f.Zero:1===n?f.One:2===n?f.Two:3===n?f.Few:6===n?f.Many:f.Other;case"da":return 1===n||0!==a&&(0===i||1===i)?f.One:f.Other;case"dsb":case"hsb":return 0===o&&i%100==1||s%100==1?f.One:0===o&&i%100==2||s%100==2?f.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||s%100===Math.floor(s%100)&&s%100>=3&&s%100<=4?f.Few:f.Other;case"ff":case"fr":case"hy":case"kab":return 0===i||1===i?f.One:f.Other;case"fil":return 0===o&&(1===i||2===i||3===i)||0===o&&i%10!=4&&i%10!=6&&i%10!=9||0!==o&&s%10!=4&&s%10!=6&&s%10!=9?f.One:f.Other;case"ga":return 1===n?f.One:2===n?f.Two:n===Math.floor(n)&&n>=3&&n<=6?f.Few:n===Math.floor(n)&&n>=7&&n<=10?f.Many:f.Other;case"gd":return 1===n||11===n?f.One:2===n||12===n?f.Two:n===Math.floor(n)&&(n>=3&&n<=10||n>=13&&n<=19)?f.Few:f.Other;case"gv":return 0===o&&i%10==1?f.One:0===o&&i%10==2?f.Two:0!==o||i%100!=0&&i%100!=20&&i%100!=40&&i%100!=60&&i%100!=80?0!==o?f.Many:f.Other:f.Few;case"he":return 1===i&&0===o?f.One:2===i&&0===o?f.Two:0!==o||n>=0&&n<=10||n%10!=0?f.Other:f.Many;case"is":return 0===a&&i%10==1&&i%100!=11||0!==a?f.One:f.Other;case"ksh":return 0===n?f.Zero:1===n?f.One:f.Other;case"kw":case"naq":case"se":case"smn":return 1===n?f.One:2===n?f.Two:f.Other;case"lag":return 0===n?f.Zero:0!==i&&1!==i||0===n?f.Other:f.One;case"lt":return n%10!=1||n%100>=11&&n%100<=19?n%10===Math.floor(n%10)&&n%10>=2&&n%10<=9&&!(n%100>=11&&n%100<=19)?f.Few:0!==s?f.Many:f.Other:f.One;case"lv":case"prg":return n%10==0||n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19||2===o&&s%100===Math.floor(s%100)&&s%100>=11&&s%100<=19?f.Zero:n%10==1&&n%100!=11||2===o&&s%10==1&&s%100!=11||2!==o&&s%10==1?f.One:f.Other;case"mk":return 0===o&&i%10==1||s%10==1?f.One:f.Other;case"mt":return 1===n?f.One:0===n||n%100===Math.floor(n%100)&&n%100>=2&&n%100<=10?f.Few:n%100===Math.floor(n%100)&&n%100>=11&&n%100<=19?f.Many:f.Other;case"pl":return 1===i&&0===o?f.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?f.Few:0===o&&1!==i&&i%10===Math.floor(i%10)&&i%10>=0&&i%10<=1||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=12&&i%100<=14?f.Many:f.Other;case"pt":return n===Math.floor(n)&&n>=0&&n<=2&&2!==n?f.One:f.Other;case"ro":return 1===i&&0===o?f.One:0!==o||0===n||1!==n&&n%100===Math.floor(n%100)&&n%100>=1&&n%100<=19?f.Few:f.Other;case"ru":case"uk":return 0===o&&i%10==1&&i%100!=11?f.One:0===o&&i%10===Math.floor(i%10)&&i%10>=2&&i%10<=4&&!(i%100>=12&&i%100<=14)?f.Few:0===o&&i%10==0||0===o&&i%10===Math.floor(i%10)&&i%10>=5&&i%10<=9||0===o&&i%100===Math.floor(i%100)&&i%100>=11&&i%100<=14?f.Many:f.Other;case"shi":return 0===i||1===n?f.One:n===Math.floor(n)&&n>=2&&n<=10?f.Few:f.Other;case"si":return 0===n||1===n||0===i&&1===s?f.One:f.Other;case"sl":return 0===o&&i%100==1?f.One:0===o&&i%100==2?f.Two:0===o&&i%100===Math.floor(i%100)&&i%100>=3&&i%100<=4||0!==o?f.Few:f.Other;case"tzm":return n===Math.floor(n)&&n>=0&&n<=1||n===Math.floor(n)&&n>=11&&n<=99?f.One:f.Other;default:return f.Other}},pe=(function(){function t(){}t.decorators=[{type:r.H,args:[{declarations:[Gt],exports:[Gt],providers:[{provide:S,useValue:le}]}]}],t.ctorParameters=function(){return[]}}(),new r.A("DocumentToken")),de="browser";new r._6("5.1.0")},"../../../common/esm5/http.js":function(t,e,n){"use strict";n.d(e,"a",function(){return E}),n.d(e,"b",function(){return X});var r=n("../../../core/esm5/core.js"),i=n("../../../../rxjs/_esm5/observable/of.js"),o=n("../../../../rxjs/_esm5/operator/concatMap.js"),s=n("../../../../rxjs/_esm5/operator/filter.js"),a=n("../../../../rxjs/_esm5/operator/map.js"),c=n("../../../../tslib/tslib.es6.js"),u=n("../../../common/esm5/common.js"),l=n("../../../../rxjs/_esm5/Observable.js"),p=function(){return function(){}}(),d=function(){return function(){}}(),h=function(){function t(t){var e=this;this.normalizedNames=new Map,this.lazyUpdate=null,t?this.lazyInit="string"==typeof t?function(){e.headers=new Map,t.split("\n").forEach(function(t){var n=t.indexOf(":");if(n>0){var r=t.slice(0,n),i=r.toLowerCase(),o=t.slice(n+1).trim();e.maybeSetNormalizedName(r,i),e.headers.has(i)?e.headers.get(i).push(o):e.headers.set(i,[o])}})}:function(){e.headers=new Map,Object.keys(t).forEach(function(n){var r=t[n],i=n.toLowerCase();"string"==typeof r&&(r=[r]),r.length>0&&(e.headers.set(i,r),e.maybeSetNormalizedName(n,i))})}:this.headers=new Map}return t.prototype.has=function(t){return this.init(),this.headers.has(t.toLowerCase())},t.prototype.get=function(t){this.init();var e=this.headers.get(t.toLowerCase());return e&&e.length>0?e[0]:null},t.prototype.keys=function(){return this.init(),Array.from(this.normalizedNames.values())},t.prototype.getAll=function(t){return this.init(),this.headers.get(t.toLowerCase())||null},t.prototype.append=function(t,e){return this.clone({name:t,value:e,op:"a"})},t.prototype.set=function(t,e){return this.clone({name:t,value:e,op:"s"})},t.prototype.delete=function(t,e){return this.clone({name:t,value:e,op:"d"})},t.prototype.maybeSetNormalizedName=function(t,e){this.normalizedNames.has(e)||this.normalizedNames.set(e,t)},t.prototype.init=function(){var e=this;this.lazyInit&&(this.lazyInit instanceof t?this.copyFrom(this.lazyInit):this.lazyInit(),this.lazyInit=null,this.lazyUpdate&&(this.lazyUpdate.forEach(function(t){return e.applyUpdate(t)}),this.lazyUpdate=null))},t.prototype.copyFrom=function(t){var e=this;t.init(),Array.from(t.headers.keys()).forEach(function(n){e.headers.set(n,t.headers.get(n)),e.normalizedNames.set(n,t.normalizedNames.get(n))})},t.prototype.clone=function(e){var n=new t;return n.lazyInit=this.lazyInit&&this.lazyInit instanceof t?this.lazyInit:this,n.lazyUpdate=(this.lazyUpdate||[]).concat([e]),n},t.prototype.applyUpdate=function(t){var e=t.name.toLowerCase();switch(t.op){case"a":case"s":var n=t.value;if("string"==typeof n&&(n=[n]),0===n.length)return;this.maybeSetNormalizedName(t.name,e);var r=("a"===t.op?this.headers.get(e):void 0)||[];r.push.apply(r,n),this.headers.set(e,r);break;case"d":var i=t.value;if(i){var o=this.headers.get(e);if(!o)return;0===(o=o.filter(function(t){return-1===i.indexOf(t)})).length?(this.headers.delete(e),this.normalizedNames.delete(e)):this.headers.set(e,o)}else this.headers.delete(e),this.normalizedNames.delete(e)}},t.prototype.forEach=function(t){var e=this;this.init(),Array.from(this.normalizedNames.keys()).forEach(function(n){return t(e.normalizedNames.get(n),e.headers.get(n))})},t}(),f=function(){function t(){}return t.prototype.encodeKey=function(t){return m(t)},t.prototype.encodeValue=function(t){return m(t)},t.prototype.decodeKey=function(t){return decodeURIComponent(t)},t.prototype.decodeValue=function(t){return decodeURIComponent(t)},t}();function m(t){return encodeURIComponent(t).replace(/%40/gi,"@").replace(/%3A/gi,":").replace(/%24/gi,"$").replace(/%2C/gi,",").replace(/%3B/gi,";").replace(/%2B/gi,"+").replace(/%3D/gi,"=").replace(/%3F/gi,"?").replace(/%2F/gi,"/")}var g=function(){function t(t){void 0===t&&(t={});var e,n,r,i=this;if(this.updates=null,this.cloneFrom=null,this.encoder=t.encoder||new f,t.fromString){if(t.fromObject)throw new Error("Cannot specify both fromString and fromObject.");this.map=(e=t.fromString,n=this.encoder,r=new Map,e.length>0&&e.split("&").forEach(function(t){var e=t.indexOf("="),i=-1==e?[n.decodeKey(t),""]:[n.decodeKey(t.slice(0,e)),n.decodeValue(t.slice(e+1))],o=i[0],s=i[1],a=r.get(o)||[];a.push(s),r.set(o,a)}),r)}else t.fromObject?(this.map=new Map,Object.keys(t.fromObject).forEach(function(e){var n=t.fromObject[e];i.map.set(e,Array.isArray(n)?n:[n])})):this.map=null}return t.prototype.has=function(t){return this.init(),this.map.has(t)},t.prototype.get=function(t){this.init();var e=this.map.get(t);return e?e[0]:null},t.prototype.getAll=function(t){return this.init(),this.map.get(t)||null},t.prototype.keys=function(){return this.init(),Array.from(this.map.keys())},t.prototype.append=function(t,e){return this.clone({param:t,value:e,op:"a"})},t.prototype.set=function(t,e){return this.clone({param:t,value:e,op:"s"})},t.prototype.delete=function(t,e){return this.clone({param:t,value:e,op:"d"})},t.prototype.toString=function(){var t=this;return this.init(),this.keys().map(function(e){var n=t.encoder.encodeKey(e);return t.map.get(e).map(function(e){return n+"="+t.encoder.encodeValue(e)}).join("&")}).join("&")},t.prototype.clone=function(e){var n=new t({encoder:this.encoder});return n.cloneFrom=this.cloneFrom||this,n.updates=(this.updates||[]).concat([e]),n},t.prototype.init=function(){var t=this;null===this.map&&(this.map=new Map),null!==this.cloneFrom&&(this.cloneFrom.init(),this.cloneFrom.keys().forEach(function(e){return t.map.set(e,t.cloneFrom.map.get(e))}),this.updates.forEach(function(e){switch(e.op){case"a":case"s":var n=("a"===e.op?t.map.get(e.param):void 0)||[];n.push(e.value),t.map.set(e.param,n);break;case"d":if(void 0===e.value){t.map.delete(e.param);break}var r=t.map.get(e.param)||[],i=r.indexOf(e.value);-1!==i&&r.splice(i,1),r.length>0?t.map.set(e.param,r):t.map.delete(e.param)}}),this.cloneFrom=null)},t}();function v(t){return"undefined"!=typeof ArrayBuffer&&t instanceof ArrayBuffer}function y(t){return"undefined"!=typeof Blob&&t instanceof Blob}function b(t){return"undefined"!=typeof FormData&&t instanceof FormData}var _=function(){function t(t,e,n,r){var i;if(this.url=e,this.body=null,this.reportProgress=!1,this.withCredentials=!1,this.responseType="json",this.method=t.toUpperCase(),function(t){switch(t){case"DELETE":case"GET":case"HEAD":case"OPTIONS":case"JSONP":return!1;default:return!0}}(this.method)||r?(this.body=void 0!==n?n:null,i=r):i=n,i&&(this.reportProgress=!!i.reportProgress,this.withCredentials=!!i.withCredentials,i.responseType&&(this.responseType=i.responseType),i.headers&&(this.headers=i.headers),i.params&&(this.params=i.params)),this.headers||(this.headers=new h),this.params){var o=this.params.toString();if(0===o.length)this.urlWithParams=e;else{var s=e.indexOf("?"),a=-1===s?"?":s<e.length-1?"&":"";this.urlWithParams=e+a+o}}else this.params=new g,this.urlWithParams=e}return t.prototype.serializeBody=function(){return null===this.body?null:v(this.body)||y(this.body)||b(this.body)||"string"==typeof this.body?this.body:this.body instanceof g?this.body.toString():"object"==typeof this.body||"boolean"==typeof this.body||Array.isArray(this.body)?JSON.stringify(this.body):this.body.toString()},t.prototype.detectContentTypeHeader=function(){return null===this.body?null:b(this.body)?null:y(this.body)?this.body.type||null:v(this.body)?null:"string"==typeof this.body?"text/plain":this.body instanceof g?"application/x-www-form-urlencoded;charset=UTF-8":"object"==typeof this.body||"number"==typeof this.body||Array.isArray(this.body)?"application/json":null},t.prototype.clone=function(e){void 0===e&&(e={});var n=e.method||this.method,r=e.url||this.url,i=e.responseType||this.responseType,o=void 0!==e.body?e.body:this.body,s=void 0!==e.withCredentials?e.withCredentials:this.withCredentials,a=void 0!==e.reportProgress?e.reportProgress:this.reportProgress,c=e.headers||this.headers,u=e.params||this.params;return void 0!==e.setHeaders&&(c=Object.keys(e.setHeaders).reduce(function(t,n){return t.set(n,e.setHeaders[n])},c)),e.setParams&&(u=Object.keys(e.setParams).reduce(function(t,n){return t.set(n,e.setParams[n])},u)),new t(n,r,o,{params:u,headers:c,reportProgress:a,responseType:i,withCredentials:s})},t}(),j={Sent:0,UploadProgress:1,ResponseHeader:2,DownloadProgress:3,Response:4,User:5};j[j.Sent]="Sent",j[j.UploadProgress]="UploadProgress",j[j.ResponseHeader]="ResponseHeader",j[j.DownloadProgress]="DownloadProgress",j[j.Response]="Response",j[j.User]="User";var x=function(){return function(t,e,n){void 0===e&&(e=200),void 0===n&&(n="OK"),this.headers=t.headers||new h,this.status=void 0!==t.status?t.status:e,this.statusText=t.statusText||n,this.url=t.url||null,this.ok=this.status>=200&&this.status<300}}(),w=function(t){function e(e){void 0===e&&(e={});var n=t.call(this,e)||this;return n.type=j.ResponseHeader,n}return Object(c.b)(e,t),e.prototype.clone=function(t){return void 0===t&&(t={}),new e({headers:t.headers||this.headers,status:void 0!==t.status?t.status:this.status,statusText:t.statusText||this.statusText,url:t.url||this.url||void 0})},e}(x),O=function(t){function e(e){void 0===e&&(e={});var n=t.call(this,e)||this;return n.type=j.Response,n.body=void 0!==e.body?e.body:null,n}return Object(c.b)(e,t),e.prototype.clone=function(t){return void 0===t&&(t={}),new e({body:void 0!==t.body?t.body:this.body,headers:t.headers||this.headers,status:void 0!==t.status?t.status:this.status,statusText:t.statusText||this.statusText,url:t.url||this.url||void 0})},e}(x),C=function(t){function e(e){var n=t.call(this,e,0,"Unknown Error")||this;return n.name="HttpErrorResponse",n.ok=!1,n.status>=200&&n.status<300?n.message="Http failure during parsing for "+(e.url||"(unknown url)"):n.message="Http failure response for "+(e.url||"(unknown url)")+": "+e.status+" "+e.statusText,n.error=e.error||null,n}return Object(c.b)(e,t),e}(x);function k(t,e){return{body:e,headers:t.headers,observe:t.observe,params:t.params,reportProgress:t.reportProgress,responseType:t.responseType,withCredentials:t.withCredentials}}var E=function(){function t(t){this.handler=t}return t.prototype.request=function(t,e,n){var r,c=this;if(void 0===n&&(n={}),t instanceof _)r=t;else{var u=void 0;u=n.headers instanceof h?n.headers:new h(n.headers);var l=void 0;n.params&&(l=n.params instanceof g?n.params:new g({fromObject:n.params})),r=new _(t,e,void 0!==n.body?n.body:null,{headers:u,params:l,reportProgress:n.reportProgress,responseType:n.responseType||"json",withCredentials:n.withCredentials})}var p=o.a.call(Object(i.a)(r),function(t){return c.handler.handle(t)});if(t instanceof _||"events"===n.observe)return p;var d=s.a.call(p,function(t){return t instanceof O});switch(n.observe||"body"){case"body":switch(r.responseType){case"arraybuffer":return a.a.call(d,function(t){if(null!==t.body&&!(t.body instanceof ArrayBuffer))throw new Error("Response is not an ArrayBuffer.");return t.body});case"blob":return a.a.call(d,function(t){if(null!==t.body&&!(t.body instanceof Blob))throw new Error("Response is not a Blob.");return t.body});case"text":return a.a.call(d,function(t){if(null!==t.body&&"string"!=typeof t.body)throw new Error("Response is not a string.");return t.body});case"json":default:return a.a.call(d,function(t){return t.body})}case"response":return d;default:throw new Error("Unreachable: unhandled observe type "+n.observe+"}")}},t.prototype.delete=function(t,e){return void 0===e&&(e={}),this.request("DELETE",t,e)},t.prototype.get=function(t,e){return void 0===e&&(e={}),this.request("GET",t,e)},t.prototype.head=function(t,e){return void 0===e&&(e={}),this.request("HEAD",t,e)},t.prototype.jsonp=function(t,e){return this.request("JSONP",t,{params:(new g).append(e,"JSONP_CALLBACK"),observe:"body",responseType:"json"})},t.prototype.options=function(t,e){return void 0===e&&(e={}),this.request("OPTIONS",t,e)},t.prototype.patch=function(t,e,n){return void 0===n&&(n={}),this.request("PATCH",t,k(n,e))},t.prototype.post=function(t,e,n){return void 0===n&&(n={}),this.request("POST",t,k(n,e))},t.prototype.put=function(t,e,n){return void 0===n&&(n={}),this.request("PUT",t,k(n,e))},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:p}]},t}(),S=function(){function t(t,e){this.next=t,this.interceptor=e}return t.prototype.handle=function(t){return this.interceptor.intercept(t,this.next)},t}(),P=new r.A("HTTP_INTERCEPTORS"),M=function(){function t(){}return t.prototype.intercept=function(t,e){return e.handle(t)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[]},t}(),A=0,T=function(){return function(){}}(),D=function(){function t(t,e){this.callbackMap=t,this.document=e}return t.prototype.nextCallback=function(){return"ng_jsonp_callback_"+A++},t.prototype.handle=function(t){var e=this;if("JSONP"!==t.method)throw new Error("JSONP requests must use JSONP request method.");if("json"!==t.responseType)throw new Error("JSONP requests must use Json response type.");return new l.a(function(n){var r=e.nextCallback(),i=t.urlWithParams.replace(/=JSONP_CALLBACK(&|$)/,"="+r+"$1"),o=e.document.createElement("script");o.src=i;var s=null,a=!1,c=!1;e.callbackMap[r]=function(t){delete e.callbackMap[r],c||(s=t,a=!0)};var u=function(){o.parentNode&&o.parentNode.removeChild(o),delete e.callbackMap[r]},l=function(t){c||(u(),a?(n.next(new O({body:s,status:200,statusText:"OK",url:i})),n.complete()):n.error(new C({url:i,status:0,statusText:"JSONP Error",error:new Error("JSONP injected script did not invoke callback.")})))},p=function(t){c||(u(),n.error(new C({error:t,status:0,statusText:"JSONP Error",url:i})))};return o.addEventListener("load",l),o.addEventListener("error",p),e.document.body.appendChild(o),n.next({type:j.Sent}),function(){c=!0,o.removeEventListener("load",l),o.removeEventListener("error",p),u()}})},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:T},{type:void 0,decorators:[{type:r.y,args:[u.b]}]}]},t}(),I=function(){function t(t){this.jsonp=t}return t.prototype.intercept=function(t,e){return"JSONP"===t.method?this.jsonp.handle(t):e.handle(t)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:D}]},t}(),R=/^\)\]\}',?\n/;var N=function(){return function(){}}(),L=function(){function t(){}return t.prototype.build=function(){return new XMLHttpRequest},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[]},t}(),F=function(){function t(t){this.xhrFactory=t}return t.prototype.handle=function(t){var e=this;if("JSONP"===t.method)throw new Error("Attempted to construct Jsonp request without JsonpClientModule installed.");return new l.a(function(n){var r=e.xhrFactory.build();if(r.open(t.method,t.urlWithParams),t.withCredentials&&(r.withCredentials=!0),t.headers.forEach(function(t,e){return r.setRequestHeader(t,e.join(","))}),t.headers.has("Accept")||r.setRequestHeader("Accept","application/json, text/plain, */*"),!t.headers.has("Content-Type")){var i=t.detectContentTypeHeader();null!==i&&r.setRequestHeader("Content-Type",i)}if(t.responseType){var o=t.responseType.toLowerCase();r.responseType="json"!==o?o:"text"}var s=t.serializeBody(),a=null,c=function(){if(null!==a)return a;var e,n=1223===r.status?204:r.status,i=r.statusText||"OK",o=new h(r.getAllResponseHeaders()),s=("responseURL"in(e=r)&&e.responseURL?e.responseURL:/^X-Request-URL:/m.test(e.getAllResponseHeaders())?e.getResponseHeader("X-Request-URL"):null)||t.url;return a=new w({headers:o,status:n,statusText:i,url:s})},u=function(){var e=c(),i=e.headers,o=e.status,s=e.statusText,a=e.url,u=null;204!==o&&(u=void 0===r.response?r.responseText:r.response),0===o&&(o=u?200:0);var l=o>=200&&o<300;if("json"===t.responseType&&"string"==typeof u){var p=u;u=u.replace(R,"");try{u=""!==u?JSON.parse(u):null}catch(t){u=p,l&&(l=!1,u={error:t,text:u})}}l?(n.next(new O({body:u,headers:i,status:o,statusText:s,url:a||void 0})),n.complete()):n.error(new C({error:u,headers:i,status:o,statusText:s,url:a||void 0}))},l=function(t){var e=new C({error:t,status:r.status||0,statusText:r.statusText||"Unknown Error"});n.error(e)},p=!1,d=function(e){p||(n.next(c()),p=!0);var i={type:j.DownloadProgress,loaded:e.loaded};e.lengthComputable&&(i.total=e.total),"text"===t.responseType&&r.responseText&&(i.partialText=r.responseText),n.next(i)},f=function(t){var e={type:j.UploadProgress,loaded:t.loaded};t.lengthComputable&&(e.total=t.total),n.next(e)};return r.addEventListener("load",u),r.addEventListener("error",l),t.reportProgress&&(r.addEventListener("progress",d),null!==s&&r.upload&&r.upload.addEventListener("progress",f)),r.send(s),n.next({type:j.Sent}),function(){r.removeEventListener("error",l),r.removeEventListener("load",u),t.reportProgress&&(r.removeEventListener("progress",d),null!==s&&r.upload&&r.upload.removeEventListener("progress",f)),r.abort()}})},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:N}]},t}(),V=new r.A("XSRF_COOKIE_NAME"),z=new r.A("XSRF_HEADER_NAME"),B=function(){return function(){}}(),q=function(){function t(t,e,n){this.doc=t,this.platform=e,this.cookieName=n,this.lastCookieString="",this.lastToken=null,this.parseCount=0}return t.prototype.getToken=function(){if("server"===this.platform)return null;var t=this.doc.cookie||"";return t!==this.lastCookieString&&(this.parseCount++,this.lastToken=Object(u.h)(t,this.cookieName),this.lastCookieString=t),this.lastToken},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.y,args:[u.b]}]},{type:void 0,decorators:[{type:r.y,args:[r.P]}]},{type:void 0,decorators:[{type:r.y,args:[V]}]}]},t}(),H=function(){function t(t,e){this.tokenService=t,this.headerName=e}return t.prototype.intercept=function(t,e){var n=t.url.toLowerCase();if("GET"===t.method||"HEAD"===t.method||n.startsWith("http://")||n.startsWith("https://"))return e.handle(t);var r=this.tokenService.getToken();return null===r||t.headers.has(this.headerName)||(t=t.clone({headers:t.headers.set(this.headerName,r)})),e.handle(t)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:B},{type:void 0,decorators:[{type:r.y,args:[z]}]}]},t}();function U(t,e){return void 0===e&&(e=[]),e?e.reduceRight(function(t,e){return new S(t,e)},t):t}function W(){return"object"==typeof window?window:{}}var G=function(){function t(){}return t.disable=function(){return{ngModule:t,providers:[{provide:H,useClass:M}]}},t.withOptions=function(e){return void 0===e&&(e={}),{ngModule:t,providers:[e.cookieName?{provide:V,useValue:e.cookieName}:[],e.headerName?{provide:z,useValue:e.headerName}:[]]}},t.decorators=[{type:r.H,args:[{providers:[H,{provide:P,useExisting:H,multi:!0},{provide:B,useClass:q},{provide:V,useValue:"XSRF-TOKEN"},{provide:z,useValue:"X-XSRF-TOKEN"}]}]}],t.ctorParameters=function(){return[]},t}(),X=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[G.withOptions({cookieName:"XSRF-TOKEN",headerName:"X-XSRF-TOKEN"})],providers:[E,{provide:p,useFactory:U,deps:[d,[new r.M,new r.y(P)]]},F,{provide:d,useExisting:F},L,{provide:N,useExisting:L}]}]}],t.ctorParameters=function(){return[]},t}();!function(){function t(){}t.decorators=[{type:r.H,args:[{providers:[D,{provide:T,useFactory:W},{provide:P,useClass:I,multi:!0}]}]}],t.ctorParameters=function(){return[]}}()},"../../../compiler/esm5/compiler.js":function(t,e,n){"use strict";n.d(e,"c",function(){return ht}),n.d(e,"j",function(){return fo}),n.d(e,"k",function(){return Bl}),n.d(e,"e",function(){return Ce}),n.d(e,"q",function(){return Ea}),n.d(e,"o",function(){return pa}),n.d(e,"n",function(){return la}),n.d(e,"y",function(){return bu}),n.d(e,"A",function(){return q}),n.d(e,"r",function(){return Rt}),n.d(e,"t",function(){return gt}),n.d(e,"v",function(){return Ml}),n.d(e,"l",function(){return Al}),n.d(e,"b",function(){return Wl}),n.d(e,"x",function(){return Gl}),n.d(e,"z",function(){return Xl}),n.d(e,"s",function(){return Zl}),n.d(e,"g",function(){return Ma}),n.d(e,"i",function(){return to}),n.d(e,"d",function(){return xe}),n.d(e,"m",function(){return fn}),n.d(e,"p",function(){return tr}),n.d(e,"a",function(){return ko}),n.d(e,"h",function(){return Qi}),n.d(e,"f",function(){return La}),n.d(e,"u",function(){return hc}),n.d(e,"w",function(){return qc});var r=n("../../../../tslib/tslib.es6.js");var i=A("Inject",function(t){return{token:t}}),o=A("InjectionToken",function(t){return{_desc:t}});var s=A("Attribute",function(t){return{attributeName:t}});var a=A("ContentChildren",function(t,e){return void 0===e&&(e={}),Object(r.a)({selector:t,first:!1,isViewQuery:!1,descendants:!1},e)}),c=A("ContentChild",function(t,e){return void 0===e&&(e={}),Object(r.a)({selector:t,first:!0,isViewQuery:!1,descendants:!0},e)}),u=A("ViewChildren",function(t,e){return void 0===e&&(e={}),Object(r.a)({selector:t,first:!1,isViewQuery:!0,descendants:!0},e)}),l=A("ViewChild",function(t,e){return Object(r.a)({selector:t,first:!0,isViewQuery:!0,descendants:!0},e)});var p=A("Directive",function(t){return void 0===t&&(t={}),t});var d={Emulated:0,Native:1,None:2};d[d.Emulated]="Emulated",d[d.Native]="Native",d[d.None]="None";var h={OnPush:0,Default:1};h[h.OnPush]="OnPush",h[h.Default]="Default";var f=A("Component",function(t){return void 0===t&&(t={}),Object(r.a)({changeDetection:h.Default},t)});var m=A("Pipe",function(t){return Object(r.a)({pure:!0},t)});var g=A("Input",function(t){return{bindingPropertyName:t}});var v=A("Output",function(t){return{bindingPropertyName:t}});var y=A("HostBinding",function(t){return{hostPropertyName:t}});var b=A("HostListener",function(t,e){return{eventName:t,args:e}});var _=A("NgModule",function(t){return t});var j={name:"custom-elements"},x={name:"no-errors-schema"},w=A("Optional"),O=A("Injectable"),C=A("Self"),k=A("SkipSelf"),E=A("Host"),S=Function,P={NONE:0,HTML:1,STYLE:2,SCRIPT:3,URL:4,RESOURCE_URL:5};P[P.NONE]="NONE",P[P.HTML]="HTML",P[P.STYLE]="STYLE",P[P.SCRIPT]="SCRIPT",P[P.URL]="URL",P[P.RESOURCE_URL]="RESOURCE_URL";var M={Error:0,Warning:1,Ignore:2};function A(t,e){var n=function(){for(var n=[],i=0;i<arguments.length;i++)n[i]=arguments[i];var o=e?e.apply(void 0,n):{};return Object(r.a)({ngMetadataName:t},o)};return n.isTypeOf=function(e){return e&&e.ngMetadataName===t},n.ngMetadataName=t,n}M[M.Error]="Error",M[M.Warning]="Warning",M[M.Ignore]="Ignore";Object.freeze({Inject:function(){},createInject:i,createInjectionToken:o,Attribute:function(){},createAttribute:s,Query:function(){},createContentChildren:a,createContentChild:c,createViewChildren:u,createViewChild:l,Directive:function(){},createDirective:p,Component:function(){},ViewEncapsulation:d,ChangeDetectionStrategy:h,createComponent:f,Pipe:function(){},createPipe:m,Input:function(){},createInput:g,Output:function(){},createOutput:v,HostBinding:function(){},createHostBinding:y,HostListener:function(){},createHostListener:b,NgModule:function(){},createNgModule:_,ModuleWithProviders:function(){},SchemaMetadata:function(){},CUSTOM_ELEMENTS_SCHEMA:j,NO_ERRORS_SCHEMA:x,createOptional:w,createInjectable:O,createSelf:C,createSkipSelf:k,createHost:E,Type:S,SecurityContext:P,NodeFlags:{None:0,TypeElement:1,TypeText:2,ProjectedTemplate:4,CatRenderNode:3,TypeNgContent:8,TypePipe:16,TypePureArray:32,TypePureObject:64,TypePurePipe:128,CatPureExpression:224,TypeValueProvider:256,TypeClassProvider:512,TypeFactoryProvider:1024,TypeUseExistingProvider:2048,LazyProvider:4096,PrivateProvider:8192,TypeDirective:16384,Component:32768,CatProviderNoDirective:3840,CatProvider:20224,OnInit:65536,OnDestroy:131072,DoCheck:262144,OnChanges:524288,AfterContentInit:1048576,AfterContentChecked:2097152,AfterViewInit:4194304,AfterViewChecked:8388608,EmbeddedViews:16777216,ComponentView:33554432,TypeContentQuery:67108864,TypeViewQuery:134217728,StaticQuery:268435456,DynamicQuery:536870912,CatQuery:201326592,Types:201347067},DepFlags:{None:0,SkipSelf:1,Optional:2,Value:8},ArgumentType:{Inline:0,Dynamic:1},BindingFlags:{TypeElementAttribute:1,TypeElementClass:2,TypeElementStyle:4,TypeProperty:8,SyntheticProperty:16,SyntheticHostProperty:32,CatSyntheticProperty:48,Types:15},QueryBindingType:{First:0,All:1},QueryValueType:{ElementRef:0,RenderElement:1,TemplateRef:2,ViewContainerRef:3,Provider:4},ViewFlags:{None:0,OnPush:2},MissingTranslationStrategy:M,MetadataFactory:function(){},Route:function(){}});var T=/-+([a-z0-9])/g;function D(t,e){return I(t,":",e)}function I(t,e,n){var r=t.indexOf(e);return-1==r?n:[t.slice(0,r).trim(),t.slice(r+1).trim()]}function R(t,e,n){return Array.isArray(t)?e.visitArray(t,n):"object"==typeof(r=t)&&null!==r&&Object.getPrototypeOf(r)===G?e.visitStringMap(t,n):null==t||"string"==typeof t||"number"==typeof t||"boolean"==typeof t?e.visitPrimitive(t,n):e.visitOther(t,n);var r}function N(t){return null!==t&&void 0!==t}function L(t){return void 0===t?null:t}var F=function(){function t(){}return t.prototype.visitArray=function(t,e){var n=this;return t.map(function(t){return R(t,n,e)})},t.prototype.visitStringMap=function(t,e){var n=this,r={};return Object.keys(t).forEach(function(i){r[i]=R(t[i],n,e)}),r},t.prototype.visitPrimitive=function(t,e){return t},t.prototype.visitOther=function(t,e){return t},t}(),V=function(t){if($(t))throw new Error("Illegal state: value cannot be a promise");return t},z=function(t,e){return $(t)?t.then(e):e(t)},B=function(t){return t.some($)?Promise.all(t):t};function q(t,e){var n=Error(t);return n[H]=!0,e&&(n[U]=e),n}var H="ngSyntaxError",U="ngParseErrors";function W(t){return t.replace(/([.*+?^=!:${}()|[\]\/\\])/g,"\\$1")}var G=Object.getPrototypeOf({});function X(t){for(var e="",n=0;n<t.length;n++){var r=t.charCodeAt(n);if(r>=55296&&r<=56319&&t.length>n+1){var i=t.charCodeAt(n+1);i>=56320&&i<=57343&&(n++,r=(r-55296<<10)+i-56320+65536)}r<=127?e+=String.fromCharCode(r):r<=2047?e+=String.fromCharCode(r>>6&31|192,63&r|128):r<=65535?e+=String.fromCharCode(r>>12|224,r>>6&63|128,63&r|128):r<=2097151&&(e+=String.fromCharCode(r>>18&7|240,r>>12&63|128,r>>6&63|128,63&r|128))}return e}function Y(t){if("string"==typeof t)return t;if(t instanceof Array)return"["+t.map(Y).join(", ")+"]";if(null==t)return""+t;if(t.overriddenName)return""+t.overriddenName;if(t.name)return""+t.name;var e=t.toString();if(null==e)return""+e;var n=e.indexOf("\n");return-1===n?e:e.substring(0,n)}function K(t){return"function"==typeof t&&t.hasOwnProperty("__forward_ref__")?t():t}function $(t){return!!t&&"function"==typeof t.then}new(function(){return function(t){this.full=t;var e=t.split(".");this.major=e[0],this.minor=e[1],this.patch=e.slice(2).join(".")}}())("5.1.0");var Q=function(){function t(t,e,n){this.value=t,this.ngContentIndex=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitText(this,e)},t}(),Z=function(){function t(t,e,n){this.value=t,this.ngContentIndex=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitBoundText(this,e)},t}(),J=function(){function t(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitAttr(this,e)},t}(),tt=function(){function t(t,e,n,r,i,o){this.name=t,this.type=e,this.securityContext=n,this.value=r,this.unit=i,this.sourceSpan=o,this.isAnimation=this.type===pt.Animation}return t.prototype.visit=function(t,e){return t.visitElementProperty(this,e)},t}(),et=function(){function t(e,n,r,i,o){this.name=e,this.target=n,this.phase=r,this.handler=i,this.sourceSpan=o,this.fullName=t.calcFullName(this.name,this.target,this.phase),this.isAnimation=!!this.phase}return t.calcFullName=function(t,e,n){return e?e+":"+t:n?"@"+t+"."+n:t},t.prototype.visit=function(t,e){return t.visitEvent(this,e)},t}(),nt=function(){function t(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitReference(this,e)},t}(),rt=function(){function t(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitVariable(this,e)},t}(),it=function(){function t(t,e,n,r,i,o,s,a,c,u,l,p,d){this.name=t,this.attrs=e,this.inputs=n,this.outputs=r,this.references=i,this.directives=o,this.providers=s,this.hasViewContainer=a,this.queryMatches=c,this.children=u,this.ngContentIndex=l,this.sourceSpan=p,this.endSourceSpan=d}return t.prototype.visit=function(t,e){return t.visitElement(this,e)},t}(),ot=function(){function t(t,e,n,r,i,o,s,a,c,u,l){this.attrs=t,this.outputs=e,this.references=n,this.variables=r,this.directives=i,this.providers=o,this.hasViewContainer=s,this.queryMatches=a,this.children=c,this.ngContentIndex=u,this.sourceSpan=l}return t.prototype.visit=function(t,e){return t.visitEmbeddedTemplate(this,e)},t}(),st=function(){function t(t,e,n,r){this.directiveName=t,this.templateName=e,this.value=n,this.sourceSpan=r}return t.prototype.visit=function(t,e){return t.visitDirectiveProperty(this,e)},t}(),at=function(){function t(t,e,n,r,i,o){this.directive=t,this.inputs=e,this.hostProperties=n,this.hostEvents=r,this.contentQueryStartId=i,this.sourceSpan=o}return t.prototype.visit=function(t,e){return t.visitDirective(this,e)},t}(),ct=function(){function t(t,e,n,r,i,o,s){this.token=t,this.multiProvider=e,this.eager=n,this.providers=r,this.providerType=i,this.lifecycleHooks=o,this.sourceSpan=s}return t.prototype.visit=function(t,e){return null},t}(),ut={PublicService:0,PrivateService:1,Component:2,Directive:3,Builtin:4};ut[ut.PublicService]="PublicService",ut[ut.PrivateService]="PrivateService",ut[ut.Component]="Component",ut[ut.Directive]="Directive",ut[ut.Builtin]="Builtin";var lt=function(){function t(t,e,n){this.index=t,this.ngContentIndex=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitNgContent(this,e)},t}(),pt={Property:0,Attribute:1,Class:2,Style:3,Animation:4};pt[pt.Property]="Property",pt[pt.Attribute]="Attribute",pt[pt.Class]="Class",pt[pt.Style]="Style",pt[pt.Animation]="Animation";!function(t){function e(){return t.call(this)||this}Object(r.b)(e,t),e.prototype.visitEmbeddedTemplate=function(t,e){return this.visitChildren(e,function(e){e(t.attrs),e(t.references),e(t.variables),e(t.directives),e(t.providers),e(t.children)})},e.prototype.visitElement=function(t,e){return this.visitChildren(e,function(e){e(t.attrs),e(t.inputs),e(t.outputs),e(t.references),e(t.directives),e(t.providers),e(t.children)})},e.prototype.visitDirective=function(t,e){return this.visitChildren(e,function(e){e(t.inputs),e(t.hostProperties),e(t.hostEvents)})},e.prototype.visitChildren=function(t,e){var n=[],r=this;return e(function(e){e&&e.length&&n.push(dt(r,e,t))}),[].concat.apply([],n)}}(function(){function t(){}return t.prototype.visitNgContent=function(t,e){},t.prototype.visitEmbeddedTemplate=function(t,e){},t.prototype.visitElement=function(t,e){},t.prototype.visitReference=function(t,e){},t.prototype.visitVariable=function(t,e){},t.prototype.visitEvent=function(t,e){},t.prototype.visitElementProperty=function(t,e){},t.prototype.visitAttr=function(t,e){},t.prototype.visitBoundText=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitDirective=function(t,e){},t.prototype.visitDirectiveProperty=function(t,e){},t}());function dt(t,e,n){void 0===n&&(n=null);var r=[],i=t.visit?function(e){return t.visit(e,n)||e.visit(t,n)}:function(e){return e.visit(t,n)};return e.forEach(function(t){var e=i(t);e&&r.push(e)}),r}var ht=function(){return function(t){var e=void 0===t?{}:t,n=e.defaultEncapsulation,r=void 0===n?d.Emulated:n,i=e.useJit,o=void 0===i||i,s=e.jitDevMode,a=void 0!==s&&s,c=e.missingTranslation,u=void 0===c?null:c,l=e.enableLegacyTemplate,p=e.preserveWhitespaces,h=e.strictInjectionParameters;this.defaultEncapsulation=r,this.useJit=!!o,this.jitDevMode=!!a,this.missingTranslation=u,this.enableLegacyTemplate=!0===l,this.preserveWhitespaces=ft(L(p)),this.strictInjectionParameters=!0===h}}();function ft(t,e){return void 0===e&&(e=!0),null===t?e:t}var mt=function(){function t(t,e,n){this.filePath=t,this.name=e,this.members=n}return t.prototype.assertNoMembers=function(){if(this.members.length)throw new Error("Illegal state: symbol without members expected, but got "+JSON.stringify(this)+".")},t}(),gt=function(){function t(){this.cache=new Map}return t.prototype.get=function(t,e,n){var r='"'+t+'".'+e+((n=n||[]).length?"."+n.join("."):""),i=this.cache.get(r);return i||(i=new mt(t,e,n),this.cache.set(r,i)),i},t}(),vt=/^(?:(?:\[([^\]]+)\])|(?:\(([^\)]+)\)))|(\@[-\w]+)$/;function yt(t){return t.replace(/\W/g,"_")}var bt=0;function _t(t){if(!t||!t.reference)return null;var e=t.reference;if(e instanceof mt)return e.name;if(e.__anonymousType)return e.__anonymousType;var n=Y(e);return n.indexOf("(")>=0?(n="anonymous_"+bt++,e.__anonymousType=n):n=yt(n),n}function jt(t){var e=t.reference;return e instanceof mt?e.filePath:"./"+Y(e)}function xt(t,e){return"View_"+_t({reference:t})+"_"+e}function wt(t){return"RenderType_"+_t({reference:t})}function Ot(t){return _t({reference:t})+"NgFactory"}var Ct={Pipe:0,Directive:1,NgModule:2,Injectable:3};function kt(t){return null!=t.value?yt(t.value):_t(t.identifier)}function Et(t){return null!=t.identifier?t.identifier.reference:t.value}Ct[Ct.Pipe]="Pipe",Ct[Ct.Directive]="Directive",Ct[Ct.NgModule]="NgModule",Ct[Ct.Injectable]="Injectable";var St=function(){return function(t){var e=void 0===t?{}:t,n=e.moduleUrl,r=e.styles,i=e.styleUrls;this.moduleUrl=n||null,this.styles=It(r),this.styleUrls=It(i)}}(),Pt=function(){function t(t){var e=t.encapsulation,n=t.template,r=t.templateUrl,i=t.htmlAst,o=t.styles,s=t.styleUrls,a=t.externalStylesheets,c=t.animations,u=t.ngContentSelectors,l=t.interpolation,p=t.isInline,d=t.preserveWhitespaces;if(this.encapsulation=e,this.template=n,this.templateUrl=r,this.htmlAst=i,this.styles=It(o),this.styleUrls=It(s),this.externalStylesheets=It(a),this.animations=c?Nt(c):[],this.ngContentSelectors=u||[],l&&2!=l.length)throw new Error("'interpolation' should have a start and an end symbol.");this.interpolation=l,this.isInline=p,this.preserveWhitespaces=d}return t.prototype.toSummary=function(){return{ngContentSelectors:this.ngContentSelectors,encapsulation:this.encapsulation}},t}(),Mt=function(){function t(t){var e=t.isHost,n=t.type,r=t.isComponent,i=t.selector,o=t.exportAs,s=t.changeDetection,a=t.inputs,c=t.outputs,u=t.hostListeners,l=t.hostProperties,p=t.hostAttributes,d=t.providers,h=t.viewProviders,f=t.queries,m=t.viewQueries,g=t.entryComponents,v=t.template,y=t.componentViewType,b=t.rendererType,_=t.componentFactory;this.isHost=!!e,this.type=n,this.isComponent=r,this.selector=i,this.exportAs=o,this.changeDetection=s,this.inputs=a,this.outputs=c,this.hostListeners=u,this.hostProperties=l,this.hostAttributes=p,this.providers=It(d),this.viewProviders=It(h),this.queries=It(f),this.viewQueries=It(m),this.entryComponents=It(g),this.template=v,this.componentViewType=y,this.rendererType=b,this.componentFactory=_}return t.create=function(e){var n=e.isHost,r=e.type,i=e.isComponent,o=e.selector,s=e.exportAs,a=e.changeDetection,c=e.inputs,u=e.outputs,l=e.host,p=e.providers,d=e.viewProviders,h=e.queries,f=e.viewQueries,m=e.entryComponents,g=e.template,v=e.componentViewType,y=e.rendererType,b=e.componentFactory,_={},j={},x={};null!=l&&Object.keys(l).forEach(function(t){var e=l[t],n=t.match(vt);null===n?x[t]=e:null!=n[1]?j[n[1]]=e:null!=n[2]&&(_[n[2]]=e)});var w={};null!=c&&c.forEach(function(t){var e=D(t,[t,t]);w[e[0]]=e[1]});var O={};return null!=u&&u.forEach(function(t){var e=D(t,[t,t]);O[e[0]]=e[1]}),new t({isHost:n,type:r,isComponent:!!i,selector:o,exportAs:s,changeDetection:a,inputs:w,outputs:O,hostListeners:_,hostProperties:j,hostAttributes:x,providers:p,viewProviders:d,queries:h,viewQueries:f,entryComponents:m,template:g,componentViewType:v,rendererType:y,componentFactory:b})},t.prototype.toSummary=function(){return{summaryKind:Ct.Directive,type:this.type,isComponent:this.isComponent,selector:this.selector,exportAs:this.exportAs,inputs:this.inputs,outputs:this.outputs,hostListeners:this.hostListeners,hostProperties:this.hostProperties,hostAttributes:this.hostAttributes,providers:this.providers,viewProviders:this.viewProviders,queries:this.queries,viewQueries:this.viewQueries,entryComponents:this.entryComponents,changeDetection:this.changeDetection,template:this.template&&this.template.toSummary(),componentViewType:this.componentViewType,rendererType:this.rendererType,componentFactory:this.componentFactory}},t}(),At=function(){function t(t){var e=t.type,n=t.name,r=t.pure;this.type=e,this.name=n,this.pure=!!r}return t.prototype.toSummary=function(){return{summaryKind:Ct.Pipe,type:this.type,name:this.name,pure:this.pure}},t}(),Tt=function(){function t(t){var e=t.type,n=t.providers,r=t.declaredDirectives,i=t.exportedDirectives,o=t.declaredPipes,s=t.exportedPipes,a=t.entryComponents,c=t.bootstrapComponents,u=t.importedModules,l=t.exportedModules,p=t.schemas,d=t.transitiveModule,h=t.id;this.type=e||null,this.declaredDirectives=It(r),this.exportedDirectives=It(i),this.declaredPipes=It(o),this.exportedPipes=It(s),this.providers=It(n),this.entryComponents=It(a),this.bootstrapComponents=It(c),this.importedModules=It(u),this.exportedModules=It(l),this.schemas=It(p),this.id=h||null,this.transitiveModule=d||null}return t.prototype.toSummary=function(){var t=this.transitiveModule;return{summaryKind:Ct.NgModule,type:this.type,entryComponents:t.entryComponents,providers:t.providers,modules:t.modules,exportedDirectives:t.exportedDirectives,exportedPipes:t.exportedPipes}},t}(),Dt=function(){function t(){this.directivesSet=new Set,this.directives=[],this.exportedDirectivesSet=new Set,this.exportedDirectives=[],this.pipesSet=new Set,this.pipes=[],this.exportedPipesSet=new Set,this.exportedPipes=[],this.modulesSet=new Set,this.modules=[],this.entryComponentsSet=new Set,this.entryComponents=[],this.providers=[]}return t.prototype.addProvider=function(t,e){this.providers.push({provider:t,module:e})},t.prototype.addDirective=function(t){this.directivesSet.has(t.reference)||(this.directivesSet.add(t.reference),this.directives.push(t))},t.prototype.addExportedDirective=function(t){this.exportedDirectivesSet.has(t.reference)||(this.exportedDirectivesSet.add(t.reference),this.exportedDirectives.push(t))},t.prototype.addPipe=function(t){this.pipesSet.has(t.reference)||(this.pipesSet.add(t.reference),this.pipes.push(t))},t.prototype.addExportedPipe=function(t){this.exportedPipesSet.has(t.reference)||(this.exportedPipesSet.add(t.reference),this.exportedPipes.push(t))},t.prototype.addModule=function(t){this.modulesSet.has(t.reference)||(this.modulesSet.add(t.reference),this.modules.push(t))},t.prototype.addEntryComponent=function(t){this.entryComponentsSet.has(t.componentType)||(this.entryComponentsSet.add(t.componentType),this.entryComponents.push(t))},t}();function It(t){return t||[]}var Rt=function(){return function(t,e){var n=e.useClass,r=e.useValue,i=e.useExisting,o=e.useFactory,s=e.deps,a=e.multi;this.token=t,this.useClass=n||null,this.useValue=r,this.useExisting=i,this.useFactory=o||null,this.dependencies=s||null,this.multi=!!a}}();function Nt(t){return t.reduce(function(t,e){var n=Array.isArray(e)?Nt(e):e;return t.concat(n)},[])}function Lt(t){return t.replace(/(\w+:\/\/[\w:-]+)?(\/+)?/,"ng:///")}function Ft(t,e,n){var r;return r=n.isInline?e.type.reference instanceof mt?e.type.reference.filePath+"."+e.type.reference.name+".html":_t(t)+"/"+_t(e.type)+".html":n.templateUrl,e.type.reference instanceof mt?r:Lt(r)}!function(){function t(t,e){void 0===e&&(e=-1),this.path=t,this.position=e}Object.defineProperty(t.prototype,"empty",{get:function(){return!this.path||!this.path.length},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"head",{get:function(){return this.path[0]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"tail",{get:function(){return this.path[this.path.length-1]},enumerable:!0,configurable:!0}),t.prototype.parentOf=function(t){return t&&this.path[this.path.indexOf(t)-1]},t.prototype.childOf=function(t){return this.path[this.path.indexOf(t)+1]},t.prototype.first=function(t){for(var e=this.path.length-1;e>=0;e--){var n=this.path[e];if(n instanceof t)return n}},t.prototype.push=function(t){this.path.push(t)},t.prototype.pop=function(){return this.path.pop()}}();var Vt=function(){function t(t,e){this.value=t,this.sourceSpan=e}return t.prototype.visit=function(t,e){return t.visitText(this,e)},t}(),zt=function(){function t(t,e,n,r,i){this.switchValue=t,this.type=e,this.cases=n,this.sourceSpan=r,this.switchValueSourceSpan=i}return t.prototype.visit=function(t,e){return t.visitExpansion(this,e)},t}(),Bt=function(){function t(t,e,n,r,i){this.value=t,this.expression=e,this.sourceSpan=n,this.valueSourceSpan=r,this.expSourceSpan=i}return t.prototype.visit=function(t,e){return t.visitExpansionCase(this,e)},t}(),qt=function(){function t(t,e,n,r){this.name=t,this.value=e,this.sourceSpan=n,this.valueSpan=r}return t.prototype.visit=function(t,e){return t.visitAttribute(this,e)},t}(),Ht=function(){function t(t,e,n,r,i,o){void 0===i&&(i=null),void 0===o&&(o=null),this.name=t,this.attrs=e,this.children=n,this.sourceSpan=r,this.startSourceSpan=i,this.endSourceSpan=o}return t.prototype.visit=function(t,e){return t.visitElement(this,e)},t}(),Ut=function(){function t(t,e){this.value=t,this.sourceSpan=e}return t.prototype.visit=function(t,e){return t.visitComment(this,e)},t}();function Wt(t,e,n){void 0===n&&(n=null);var r=[],i=t.visit?function(e){return t.visit(e,n)||e.visit(t,n)}:function(e){return e.visit(t,n)};return e.forEach(function(t){var e=i(t);e&&r.push(e)}),r}!function(){function t(){}t.prototype.visitElement=function(t,e){this.visitChildren(e,function(e){e(t.attrs),e(t.children)})},t.prototype.visitAttribute=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitComment=function(t,e){},t.prototype.visitExpansion=function(t,e){return this.visitChildren(e,function(e){e(t.cases)})},t.prototype.visitExpansionCase=function(t,e){},t.prototype.visitChildren=function(t,e){var n=[],r=this;return e(function(e){e&&n.push(Wt(r,e,t))}),[].concat.apply([],n)}}();function Gt(t,e){if(null!=e){if(!Array.isArray(e))throw new Error("Expected '"+t+"' to be an array of strings.");for(var n=0;n<e.length;n+=1)if("string"!=typeof e[n])throw new Error("Expected '"+t+"' to be an array of strings.")}}var Xt=[/^\s*$/,/[<>]/,/^[{}]$/,/&(#|[a-z])/i,/^\/\//];function Yt(t,e){if(!(null==e||Array.isArray(e)&&2==e.length))throw new Error("Expected '"+t+"' to be an array, [start, end].");if(null!=e){var n=e[0],r=e[1];Xt.forEach(function(t){if(t.test(n)||t.test(r))throw new Error("['"+n+"', '"+r+"'] contains unusable interpolation symbol.")})}}var Kt=function(){function t(t,e){this.start=t,this.end=e}return t.fromArray=function(e){return e?(Yt("interpolation",e),new t(e[0],e[1])):$t},t}(),$t=new Kt("{{","}}"),Qt=function(){return function(t,e){this.style=t,this.styleUrls=e}}();function Zt(t){if(null==t||0===t.length||"/"==t[0])return!1;var e=t.match(ee);return null===e||"package"==e[1]||"asset"==e[1]}var Jt=/@import\s+(?:url\()?\s*(?:(?:['"]([^'"]*))|([^;\)\s]*))[^;]*;?/g,te=/\/\*(?!#\s*(?:sourceURL|sourceMappingURL)=)[\s\S]+?\*\//g,ee=/^([^:/?#]+):/,ne={RAW_TEXT:0,ESCAPABLE_RAW_TEXT:1,PARSABLE_DATA:2};function re(t){if(":"!=t[0])return[null,t];var e=t.indexOf(":",1);if(-1==e)throw new Error('Unsupported format "'+t+'" expecting ":namespace:name"');return[t.slice(1,e),t.slice(e+1)]}function ie(t){return"ng-container"===re(t)[1]}function oe(t){return"ng-content"===re(t)[1]}function se(t){return null===t?null:re(t)[0]}function ae(t,e){return t?":"+t+":"+e:e}ne[ne.RAW_TEXT]="RAW_TEXT",ne[ne.ESCAPABLE_RAW_TEXT]="ESCAPABLE_RAW_TEXT",ne[ne.PARSABLE_DATA]="PARSABLE_DATA";var ce={Aacute:"Á",aacute:"á",Acirc:"Â",acirc:"â",acute:"´",AElig:"Æ",aelig:"æ",Agrave:"À",agrave:"à",alefsym:"ℵ",Alpha:"Α",alpha:"α",amp:"&",and:"∧",ang:"∠",apos:"'",Aring:"Å",aring:"å",asymp:"≈",Atilde:"Ã",atilde:"ã",Auml:"Ä",auml:"ä",bdquo:"„",Beta:"Β",beta:"β",brvbar:"¦",bull:"•",cap:"∩",Ccedil:"Ç",ccedil:"ç",cedil:"¸",cent:"¢",Chi:"Χ",chi:"χ",circ:"ˆ",clubs:"♣",cong:"≅",copy:"©",crarr:"↵",cup:"∪",curren:"¤",dagger:"†",Dagger:"‡",darr:"↓",dArr:"⇓",deg:"°",Delta:"Δ",delta:"δ",diams:"♦",divide:"÷",Eacute:"É",eacute:"é",Ecirc:"Ê",ecirc:"ê",Egrave:"È",egrave:"è",empty:"∅",emsp:" ",ensp:" ",Epsilon:"Ε",epsilon:"ε",equiv:"≡",Eta:"Η",eta:"η",ETH:"Ð",eth:"ð",Euml:"Ë",euml:"ë",euro:"€",exist:"∃",fnof:"ƒ",forall:"∀",frac12:"½",frac14:"¼",frac34:"¾",frasl:"⁄",Gamma:"Γ",gamma:"γ",ge:"≥",gt:">",harr:"↔",hArr:"⇔",hearts:"♥",hellip:"…",Iacute:"Í",iacute:"í",Icirc:"Î",icirc:"î",iexcl:"¡",Igrave:"Ì",igrave:"ì",image:"ℑ",infin:"∞",int:"∫",Iota:"Ι",iota:"ι",iquest:"¿",isin:"∈",Iuml:"Ï",iuml:"ï",Kappa:"Κ",kappa:"κ",Lambda:"Λ",lambda:"λ",lang:"⟨",laquo:"«",larr:"←",lArr:"⇐",lceil:"⌈",ldquo:"“",le:"≤",lfloor:"⌊",lowast:"∗",loz:"◊",lrm:"‎",lsaquo:"‹",lsquo:"‘",lt:"<",macr:"¯",mdash:"—",micro:"µ",middot:"·",minus:"−",Mu:"Μ",mu:"μ",nabla:"∇",nbsp:" ",ndash:"–",ne:"≠",ni:"∋",not:"¬",notin:"∉",nsub:"⊄",Ntilde:"Ñ",ntilde:"ñ",Nu:"Ν",nu:"ν",Oacute:"Ó",oacute:"ó",Ocirc:"Ô",ocirc:"ô",OElig:"Œ",oelig:"œ",Ograve:"Ò",ograve:"ò",oline:"‾",Omega:"Ω",omega:"ω",Omicron:"Ο",omicron:"ο",oplus:"⊕",or:"∨",ordf:"ª",ordm:"º",Oslash:"Ø",oslash:"ø",Otilde:"Õ",otilde:"õ",otimes:"⊗",Ouml:"Ö",ouml:"ö",para:"¶",permil:"‰",perp:"⊥",Phi:"Φ",phi:"φ",Pi:"Π",pi:"π",piv:"ϖ",plusmn:"±",pound:"£",prime:"′",Prime:"″",prod:"∏",prop:"∝",Psi:"Ψ",psi:"ψ",quot:'"',radic:"√",rang:"⟩",raquo:"»",rarr:"→",rArr:"⇒",rceil:"⌉",rdquo:"”",real:"ℜ",reg:"®",rfloor:"⌋",Rho:"Ρ",rho:"ρ",rlm:"‏",rsaquo:"›",rsquo:"’",sbquo:"‚",Scaron:"Š",scaron:"š",sdot:"⋅",sect:"§",shy:"­",Sigma:"Σ",sigma:"σ",sigmaf:"ς",sim:"∼",spades:"♠",sub:"⊂",sube:"⊆",sum:"∑",sup:"⊃",sup1:"¹",sup2:"²",sup3:"³",supe:"⊇",szlig:"ß",Tau:"Τ",tau:"τ",there4:"∴",Theta:"Θ",theta:"θ",thetasym:"ϑ",thinsp:" ",THORN:"Þ",thorn:"þ",tilde:"˜",times:"×",trade:"™",Uacute:"Ú",uacute:"ú",uarr:"↑",uArr:"⇑",Ucirc:"Û",ucirc:"û",Ugrave:"Ù",ugrave:"ù",uml:"¨",upsih:"ϒ",Upsilon:"Υ",upsilon:"υ",Uuml:"Ü",uuml:"ü",weierp:"℘",Xi:"Ξ",xi:"ξ",Yacute:"Ý",yacute:"ý",yen:"¥",yuml:"ÿ",Yuml:"Ÿ",Zeta:"Ζ",zeta:"ζ",zwj:"‍",zwnj:"‌"},ue="";ce.ngsp=ue;var le="select",pe="link",de="rel",he="href",fe="stylesheet",me="style",ge="script",ve="ngNonBindable",ye="ngProjectAs";function be(t){var e=null,n=null,r=null,i=!1,o=null;t.attrs.forEach(function(t){var s=t.name.toLowerCase();s==le?e=t.value:s==he?n=t.value:s==de?r=t.value:t.name==ve?i=!0:t.name==ye&&t.value.length>0&&(o=t.value)}),e=function(t){if(null===t||0===t.length)return"*";return t}(e);var s=t.name.toLowerCase(),a=_e.OTHER;return oe(s)?a=_e.NG_CONTENT:s==me?a=_e.STYLE:s==ge?a=_e.SCRIPT:s==pe&&r==fe&&(a=_e.STYLESHEET),new je(a,e,n,i,o)}var _e={NG_CONTENT:0,STYLE:1,STYLESHEET:2,SCRIPT:3,OTHER:4};_e[_e.NG_CONTENT]="NG_CONTENT",_e[_e.STYLE]="STYLE",_e[_e.STYLESHEET]="STYLESHEET",_e[_e.SCRIPT]="SCRIPT",_e[_e.OTHER]="OTHER";var je=function(){return function(t,e,n,r,i){this.type=t,this.selectAttr=e,this.hrefAttr=n,this.nonBindable=r,this.projectAs=i}}();var xe=function(){function t(t,e,n,r){this._resourceLoader=t,this._urlResolver=e,this._htmlParser=n,this._config=r,this._resourceLoaderCache=new Map}return t.prototype.clearCache=function(){this._resourceLoaderCache.clear()},t.prototype.clearCacheFor=function(t){var e=this;if(t.isComponent){var n=t.template;this._resourceLoaderCache.delete(n.templateUrl),n.externalStylesheets.forEach(function(t){e._resourceLoaderCache.delete(t.moduleUrl)})}},t.prototype._fetch=function(t){var e=this._resourceLoaderCache.get(t);return e||(e=this._resourceLoader.get(t),this._resourceLoaderCache.set(t,e)),e},t.prototype.normalizeTemplate=function(t){var e=this;if(N(t.template)){if(N(t.templateUrl))throw q("'"+Y(t.componentType)+"' component cannot define both template and templateUrl");if("string"!=typeof t.template)throw q("The template specified for component "+Y(t.componentType)+" is not a string")}else{if(!N(t.templateUrl))throw q("No template specified for component "+Y(t.componentType));if("string"!=typeof t.templateUrl)throw q("The templateUrl specified for component "+Y(t.componentType)+" is not a string")}if(N(t.preserveWhitespaces)&&"boolean"!=typeof t.preserveWhitespaces)throw q("The preserveWhitespaces option for component "+Y(t.componentType)+" must be a boolean");return z(this._preParseTemplate(t),function(n){return e._normalizeTemplateMetadata(t,n)})},t.prototype._preParseTemplate=function(t){var e,n,r=this;return null!=t.template?(e=t.template,n=t.moduleUrl):(n=this._urlResolver.resolve(t.moduleUrl,t.templateUrl),e=this._fetch(n)),z(e,function(e){return r._preparseLoadedTemplate(t,e,n)})},t.prototype._preparseLoadedTemplate=function(t,e,n){var r=!!t.template,i=Kt.fromArray(t.interpolation),o=this._htmlParser.parse(e,Ft({reference:t.ngModuleType},{type:{reference:t.componentType}},{isInline:r,templateUrl:n}),!0,i);if(o.errors.length>0)throw q("Template parse errors:\n"+o.errors.join("\n"));var s=this._normalizeStylesheet(new St({styles:t.styles,moduleUrl:t.moduleUrl})),a=new we;Wt(a,o.rootNodes);var c=this._normalizeStylesheet(new St({styles:a.styles,styleUrls:a.styleUrls,moduleUrl:n}));return{template:e,templateUrl:n,isInline:r,htmlAst:o,styles:s.styles.concat(c.styles),inlineStyleUrls:s.styleUrls.concat(c.styleUrls),styleUrls:this._normalizeStylesheet(new St({styleUrls:t.styleUrls,moduleUrl:t.moduleUrl})).styleUrls,ngContentSelectors:a.ngContentSelectors}},t.prototype._normalizeTemplateMetadata=function(t,e){var n=this;return z(this._loadMissingExternalStylesheets(e.styleUrls.concat(e.inlineStyleUrls)),function(r){return n._normalizeLoadedTemplateMetadata(t,e,r)})},t.prototype._normalizeLoadedTemplateMetadata=function(t,e,n){var r=this,i=e.styles.slice();this._inlineStyles(e.inlineStyleUrls,n,i);var o=e.styleUrls,s=o.map(function(t){var e=n.get(t),i=e.styles.slice();return r._inlineStyles(e.styleUrls,n,i),new St({moduleUrl:t,styles:i})}),a=t.encapsulation;return null==a&&(a=this._config.defaultEncapsulation),a===d.Emulated&&0===i.length&&0===o.length&&(a=d.None),new Pt({encapsulation:a,template:e.template,templateUrl:e.templateUrl,htmlAst:e.htmlAst,styles:i,styleUrls:o,ngContentSelectors:e.ngContentSelectors,animations:t.animations,interpolation:t.interpolation,isInline:e.isInline,externalStylesheets:s,preserveWhitespaces:ft(t.preserveWhitespaces,this._config.preserveWhitespaces)})},t.prototype._inlineStyles=function(t,e,n){var r=this;t.forEach(function(t){var i=e.get(t);i.styles.forEach(function(t){return n.push(t)}),r._inlineStyles(i.styleUrls,e,n)})},t.prototype._loadMissingExternalStylesheets=function(t,e){var n=this;return void 0===e&&(e=new Map),z(B(t.filter(function(t){return!e.has(t)}).map(function(t){return z(n._fetch(t),function(r){var i=n._normalizeStylesheet(new St({styles:[r],moduleUrl:t}));return e.set(t,i),n._loadMissingExternalStylesheets(i.styleUrls,e)})})),function(t){return e})},t.prototype._normalizeStylesheet=function(t){var e=this,n=t.moduleUrl,r=t.styleUrls.filter(Zt).map(function(t){return e._urlResolver.resolve(n,t)}),i=t.styles.map(function(t){var i,o,s,a,c=(i=e._urlResolver,o=n,s=[],a=t.replace(te,"").replace(Jt,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t[1]||t[2];return Zt(n)?(s.push(i.resolve(o,n)),""):t[0]}),new Qt(a,s));return r.push.apply(r,c.styleUrls),c.style});return new St({styles:i,styleUrls:r,moduleUrl:n})},t}(),we=function(){function t(){this.ngContentSelectors=[],this.styles=[],this.styleUrls=[],this.ngNonBindableStackCount=0}return t.prototype.visitElement=function(t,e){var n=be(t);switch(n.type){case _e.NG_CONTENT:0===this.ngNonBindableStackCount&&this.ngContentSelectors.push(n.selectAttr);break;case _e.STYLE:var r="";t.children.forEach(function(t){t instanceof Vt&&(r+=t.value)}),this.styles.push(r);break;case _e.STYLESHEET:this.styleUrls.push(n.hrefAttr)}return n.nonBindable&&this.ngNonBindableStackCount++,Wt(this,t.children),n.nonBindable&&this.ngNonBindableStackCount--,null},t.prototype.visitExpansion=function(t,e){Wt(this,t.cases)},t.prototype.visitExpansionCase=function(t,e){Wt(this,t.expression)},t.prototype.visitComment=function(t,e){return null},t.prototype.visitAttribute=function(t,e){return null},t.prototype.visitText=function(t,e){return null},t}(),Oe=[l,u,c,a],Ce=function(){function t(t){this._reflector=t}return t.prototype.isDirective=function(t){var e=this._reflector.annotations(K(t));return e&&e.some(ke)},t.prototype.resolve=function(t,e){void 0===e&&(e=!0);var n=this._reflector.annotations(K(t));if(n){var r=Ee(n,ke);if(r){var i=this._reflector.propMetadata(t);return this._mergeWithPropertyMetadata(r,i,t)}}if(e)throw new Error("No Directive annotation found on "+Y(t));return null},t.prototype._mergeWithPropertyMetadata=function(t,e,n){var r=[],i=[],o={},s={};return Object.keys(e).forEach(function(t){var n=Ee(e[t],function(t){return g.isTypeOf(t)});n&&(n.bindingPropertyName?r.push(t+": "+n.bindingPropertyName):r.push(t));var a=Ee(e[t],function(t){return v.isTypeOf(t)});a&&(a.bindingPropertyName?i.push(t+": "+a.bindingPropertyName):i.push(t)),e[t].filter(function(t){return y.isTypeOf(t)}).forEach(function(e){if(e.hostPropertyName){var n=e.hostPropertyName[0];if("("===n)throw new Error("@HostBinding can not bind to events. Use @HostListener instead.");if("["===n)throw new Error("@HostBinding parameter should be a property name, 'class.<name>', or 'attr.<name>'.");o["["+e.hostPropertyName+"]"]=t}else o["["+t+"]"]=t}),e[t].filter(function(t){return b.isTypeOf(t)}).forEach(function(e){var n=e.args||[];o["("+e.eventName+")"]=t+"("+n.join(",")+")"});var c=Ee(e[t],function(t){return Oe.some(function(e){return e.isTypeOf(t)})});c&&(s[t]=c)}),this._merge(t,r,i,o,s,n)},t.prototype._extractPublicName=function(t){return D(t,[null,t])[1].trim()},t.prototype._dedupeBindings=function(t){for(var e=new Set,n=[],r=t.length-1;r>=0;r--){var i=t[r],o=this._extractPublicName(i);e.has(o)||(e.add(o),n.push(i))}return n.reverse()},t.prototype._merge=function(t,e,n,i,o,s){var a=this._dedupeBindings(t.inputs?t.inputs.concat(e):e),c=this._dedupeBindings(t.outputs?t.outputs.concat(n):n),u=t.host?Object(r.a)({},t.host,i):i,l=t.queries?Object(r.a)({},t.queries,o):o;if(f.isTypeOf(t)){var d=t;return f({selector:d.selector,inputs:a,outputs:c,host:u,exportAs:d.exportAs,moduleId:d.moduleId,queries:l,changeDetection:d.changeDetection,providers:d.providers,viewProviders:d.viewProviders,entryComponents:d.entryComponents,template:d.template,templateUrl:d.templateUrl,styles:d.styles,styleUrls:d.styleUrls,encapsulation:d.encapsulation,animations:d.animations,interpolation:d.interpolation,preserveWhitespaces:t.preserveWhitespaces})}return p({selector:t.selector,inputs:a,outputs:c,host:u,exportAs:t.exportAs,queries:l,providers:t.providers})},t}();function ke(t){return p.isTypeOf(t)||f.isTypeOf(t)}function Ee(t,e){for(var n=t.length-1;n>=0;n--)if(e(t[n]))return t[n];return null}var Se=0,Pe=9,Me=10,Ae=11,Te=12,De=13,Ie=32,Re=34,Ne=36,Le=39,Fe=43,Ve=45,ze=47,Be=59,qe=61,He=62,Ue=48,We=57,Ge=65,Xe=69,Ye=70,Ke=90,$e=95,Qe=97,Ze=101,Je=102,tn=110,en=114,nn=116,rn=118,on=122,sn=123,an=160,cn=96;function un(t){return t>=Pe&&t<=Ie||t==an}function ln(t){return Ue<=t&&t<=We}function pn(t){return t>=Qe&&t<=on||t>=Ge&&t<=Ke}var dn={Character:0,Identifier:1,Keyword:2,String:3,Operator:4,Number:5,Error:6};dn[dn.Character]="Character",dn[dn.Identifier]="Identifier",dn[dn.Keyword]="Keyword",dn[dn.String]="String",dn[dn.Operator]="Operator",dn[dn.Number]="Number",dn[dn.Error]="Error";var hn=["var","let","as","null","undefined","true","false","if","else","this"],fn=function(){function t(){}return t.prototype.tokenize=function(t){for(var e=new bn(t),n=[],r=e.scanToken();null!=r;)n.push(r),r=e.scanToken();return n},t}(),mn=function(){function t(t,e,n,r){this.index=t,this.type=e,this.numValue=n,this.strValue=r}return t.prototype.isCharacter=function(t){return this.type==dn.Character&&this.numValue==t},t.prototype.isNumber=function(){return this.type==dn.Number},t.prototype.isString=function(){return this.type==dn.String},t.prototype.isOperator=function(t){return this.type==dn.Operator&&this.strValue==t},t.prototype.isIdentifier=function(){return this.type==dn.Identifier},t.prototype.isKeyword=function(){return this.type==dn.Keyword},t.prototype.isKeywordLet=function(){return this.type==dn.Keyword&&"let"==this.strValue},t.prototype.isKeywordAs=function(){return this.type==dn.Keyword&&"as"==this.strValue},t.prototype.isKeywordNull=function(){return this.type==dn.Keyword&&"null"==this.strValue},t.prototype.isKeywordUndefined=function(){return this.type==dn.Keyword&&"undefined"==this.strValue},t.prototype.isKeywordTrue=function(){return this.type==dn.Keyword&&"true"==this.strValue},t.prototype.isKeywordFalse=function(){return this.type==dn.Keyword&&"false"==this.strValue},t.prototype.isKeywordThis=function(){return this.type==dn.Keyword&&"this"==this.strValue},t.prototype.isError=function(){return this.type==dn.Error},t.prototype.toNumber=function(){return this.type==dn.Number?this.numValue:-1},t.prototype.toString=function(){switch(this.type){case dn.Character:case dn.Identifier:case dn.Keyword:case dn.Operator:case dn.String:case dn.Error:return this.strValue;case dn.Number:return this.numValue.toString();default:return null}},t}();function gn(t,e){return new mn(t,dn.Character,e,String.fromCharCode(e))}function vn(t,e){return new mn(t,dn.Operator,0,e)}var yn=new mn(-1,dn.Character,0,""),bn=function(){function t(t){this.input=t,this.peek=0,this.index=-1,this.length=t.length,this.advance()}return t.prototype.advance=function(){this.peek=++this.index>=this.length?Se:this.input.charCodeAt(this.index)},t.prototype.scanToken=function(){for(var t=this.input,e=this.length,n=this.peek,r=this.index;n<=Ie;){if(++r>=e){n=Se;break}n=t.charCodeAt(r)}if(this.peek=n,this.index=r,r>=e)return null;if(_n(n))return this.scanIdentifier();if(ln(n))return this.scanNumber(r);var i=r;switch(n){case 46:return this.advance(),ln(this.peek)?this.scanNumber(i):gn(i,46);case 40:case 41:case sn:case 125:case 91:case 93:case 44:case 58:case Be:return this.scanCharacter(i,n);case Le:case Re:return this.scanString();case 35:case Fe:case Ve:case 42:case ze:case 37:case 94:return this.scanOperator(i,String.fromCharCode(n));case 63:return this.scanComplexOperator(i,"?",46,".");case 60:case He:return this.scanComplexOperator(i,String.fromCharCode(n),qe,"=");case 33:case qe:return this.scanComplexOperator(i,String.fromCharCode(n),qe,"=",qe,"=");case 38:return this.scanComplexOperator(i,"&",38,"&");case 124:return this.scanComplexOperator(i,"|",124,"|");case an:for(;un(this.peek);)this.advance();return this.scanToken()}return this.advance(),this.error("Unexpected character ["+String.fromCharCode(n)+"]",0)},t.prototype.scanCharacter=function(t,e){return this.advance(),gn(t,e)},t.prototype.scanOperator=function(t,e){return this.advance(),vn(t,e)},t.prototype.scanComplexOperator=function(t,e,n,r,i,o){this.advance();var s=e;return this.peek==n&&(this.advance(),s+=r),null!=i&&this.peek==i&&(this.advance(),s+=o),vn(t,s)},t.prototype.scanIdentifier=function(){var t=this.index;for(this.advance();jn(this.peek);)this.advance();var e,n,r=this.input.substring(t,this.index);return hn.indexOf(r)>-1?(n=r,new mn(t,dn.Keyword,0,n)):(e=r,new mn(t,dn.Identifier,0,e))},t.prototype.scanNumber=function(t){var e,n,r=this.index===t;for(this.advance();;){if(ln(this.peek));else if(46==this.peek)r=!1;else{if((n=this.peek)!=Ze&&n!=Xe)break;if(this.advance(),((e=this.peek)==Ve||e==Fe)&&this.advance(),!ln(this.peek))return this.error("Invalid exponent",-1);r=!1}this.advance()}var i,o=this.input.substring(t,this.index),s=r?function(t){var e=parseInt(t);if(isNaN(e))throw new Error("Invalid integer literal when parsing "+t);return e}(o):parseFloat(o);return i=s,new mn(t,dn.Number,i,"")},t.prototype.scanString=function(){var t=this.index,e=this.peek;this.advance();for(var n="",r=this.index,i=this.input;this.peek!=e;)if(92==this.peek){n+=i.substring(r,this.index),this.advance();var o=void 0;if(this.peek=this.peek,117==this.peek){var s=i.substring(this.index+1,this.index+5);if(!/^[0-9a-f]+$/i.test(s))return this.error("Invalid unicode escape [\\u"+s+"]",0);o=parseInt(s,16);for(var a=0;a<5;a++)this.advance()}else o=xn(this.peek),this.advance();n+=String.fromCharCode(o),r=this.index}else{if(this.peek==Se)return this.error("Unterminated quote",0);this.advance()}var c,u=i.substring(r,this.index);return this.advance(),c=n+u,new mn(t,dn.String,0,c)},t.prototype.error=function(t,e){var n,r,i=this.index+e;return n=i,r="Lexer Error: "+t+" at column "+i+" in expression ["+this.input+"]",new mn(n,dn.Error,0,r)},t}();function _n(t){return Qe<=t&&t<=on||Ge<=t&&t<=Ke||t==$e||t==Ne}function jn(t){return pn(t)||ln(t)||t==$e||t==Ne}function xn(t){switch(t){case tn:return Me;case Je:return Te;case en:return De;case nn:return Pe;case rn:return Ae;default:return t}}var wn=function(){return function(t,e,n,r){this.input=e,this.errLocation=n,this.ctxLocation=r,this.message="Parser Error: "+t+" "+n+" ["+e+"] in "+r}}(),On=function(){return function(t,e){this.start=t,this.end=e}}(),Cn=function(){function t(t){this.span=t}return t.prototype.visit=function(t,e){return void 0===e&&(e=null),null},t.prototype.toString=function(){return"AST"},t}(),kn=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.prefix=n,o.uninterpretedExpression=r,o.location=i,o}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitQuote(this,e)},e.prototype.toString=function(){return"Quote"},e}(Cn),En=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.visit=function(t,e){void 0===e&&(e=null)},e}(Cn),Sn=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitImplicitReceiver(this,e)},e}(Cn),Pn=function(t){function e(e,n){var r=t.call(this,e)||this;return r.expressions=n,r}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitChain(this,e)},e}(Cn),Mn=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.condition=n,o.trueExp=r,o.falseExp=i,o}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitConditional(this,e)},e}(Cn),An=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.receiver=n,i.name=r,i}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitPropertyRead(this,e)},e}(Cn),Tn=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.receiver=n,o.name=r,o.value=i,o}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitPropertyWrite(this,e)},e}(Cn),Dn=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.receiver=n,i.name=r,i}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitSafePropertyRead(this,e)},e}(Cn),In=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.obj=n,i.key=r,i}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitKeyedRead(this,e)},e}(Cn),Rn=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.obj=n,o.key=r,o.value=i,o}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitKeyedWrite(this,e)},e}(Cn),Nn=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.exp=n,o.name=r,o.args=i,o}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitPipe(this,e)},e}(Cn),Ln=function(t){function e(e,n){var r=t.call(this,e)||this;return r.value=n,r}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitLiteralPrimitive(this,e)},e}(Cn),Fn=function(t){function e(e,n){var r=t.call(this,e)||this;return r.expressions=n,r}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitLiteralArray(this,e)},e}(Cn),Vn=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.keys=n,i.values=r,i}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitLiteralMap(this,e)},e}(Cn),zn=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.strings=n,i.expressions=r,i}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitInterpolation(this,e)},e}(Cn),Bn=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.operation=n,o.left=r,o.right=i,o}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitBinary(this,e)},e}(Cn),qn=function(t){function e(e,n){var r=t.call(this,e)||this;return r.expression=n,r}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitPrefixNot(this,e)},e}(Cn),Hn=function(t){function e(e,n){var r=t.call(this,e)||this;return r.expression=n,r}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitNonNullAssert(this,e)},e}(Cn),Un=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.receiver=n,o.name=r,o.args=i,o}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitMethodCall(this,e)},e}(Cn),Wn=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;return o.receiver=n,o.name=r,o.args=i,o}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitSafeMethodCall(this,e)},e}(Cn),Gn=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i.target=n,i.args=r,i}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),t.visitFunctionCall(this,e)},e}(Cn),Xn=function(t){function e(e,n,r,i){var o=t.call(this,new On(0,null==n?0:n.length))||this;return o.ast=e,o.source=n,o.location=r,o.errors=i,o}return Object(r.b)(e,t),e.prototype.visit=function(t,e){return void 0===e&&(e=null),this.ast.visit(t,e)},e.prototype.toString=function(){return this.source+" in "+this.location},e}(Cn),Yn=function(){return function(t,e,n,r,i){this.span=t,this.key=e,this.keyIsVar=n,this.name=r,this.expression=i}}(),Kn=(function(){function t(){}t.prototype.visitBinary=function(t,e){},t.prototype.visitChain=function(t,e){},t.prototype.visitConditional=function(t,e){},t.prototype.visitFunctionCall=function(t,e){},t.prototype.visitImplicitReceiver=function(t,e){},t.prototype.visitInterpolation=function(t,e){},t.prototype.visitKeyedRead=function(t,e){},t.prototype.visitKeyedWrite=function(t,e){},t.prototype.visitLiteralArray=function(t,e){},t.prototype.visitLiteralMap=function(t,e){},t.prototype.visitLiteralPrimitive=function(t,e){},t.prototype.visitMethodCall=function(t,e){},t.prototype.visitPipe=function(t,e){},t.prototype.visitPrefixNot=function(t,e){},t.prototype.visitNonNullAssert=function(t,e){},t.prototype.visitPropertyRead=function(t,e){},t.prototype.visitPropertyWrite=function(t,e){},t.prototype.visitQuote=function(t,e){},t.prototype.visitSafeMethodCall=function(t,e){},t.prototype.visitSafePropertyRead=function(t,e){}}(),function(){function t(){}return t.prototype.visitBinary=function(t,e){return t.left.visit(this),t.right.visit(this),null},t.prototype.visitChain=function(t,e){return this.visitAll(t.expressions,e)},t.prototype.visitConditional=function(t,e){return t.condition.visit(this),t.trueExp.visit(this),t.falseExp.visit(this),null},t.prototype.visitPipe=function(t,e){return t.exp.visit(this),this.visitAll(t.args,e),null},t.prototype.visitFunctionCall=function(t,e){return t.target.visit(this),this.visitAll(t.args,e),null},t.prototype.visitImplicitReceiver=function(t,e){return null},t.prototype.visitInterpolation=function(t,e){return this.visitAll(t.expressions,e)},t.prototype.visitKeyedRead=function(t,e){return t.obj.visit(this),t.key.visit(this),null},t.prototype.visitKeyedWrite=function(t,e){return t.obj.visit(this),t.key.visit(this),t.value.visit(this),null},t.prototype.visitLiteralArray=function(t,e){return this.visitAll(t.expressions,e)},t.prototype.visitLiteralMap=function(t,e){return this.visitAll(t.values,e)},t.prototype.visitLiteralPrimitive=function(t,e){return null},t.prototype.visitMethodCall=function(t,e){return t.receiver.visit(this),this.visitAll(t.args,e)},t.prototype.visitPrefixNot=function(t,e){return t.expression.visit(this),null},t.prototype.visitNonNullAssert=function(t,e){return t.expression.visit(this),null},t.prototype.visitPropertyRead=function(t,e){return t.receiver.visit(this),null},t.prototype.visitPropertyWrite=function(t,e){return t.receiver.visit(this),t.value.visit(this),null},t.prototype.visitSafePropertyRead=function(t,e){return t.receiver.visit(this),null},t.prototype.visitSafeMethodCall=function(t,e){return t.receiver.visit(this),this.visitAll(t.args,e)},t.prototype.visitAll=function(t,e){var n=this;return t.forEach(function(t){return t.visit(n,e)}),null},t.prototype.visitQuote=function(t,e){return null},t}()),$n=function(){function t(){}return t.prototype.visitImplicitReceiver=function(t,e){return t},t.prototype.visitInterpolation=function(t,e){return new zn(t.span,t.strings,this.visitAll(t.expressions))},t.prototype.visitLiteralPrimitive=function(t,e){return new Ln(t.span,t.value)},t.prototype.visitPropertyRead=function(t,e){return new An(t.span,t.receiver.visit(this),t.name)},t.prototype.visitPropertyWrite=function(t,e){return new Tn(t.span,t.receiver.visit(this),t.name,t.value.visit(this))},t.prototype.visitSafePropertyRead=function(t,e){return new Dn(t.span,t.receiver.visit(this),t.name)},t.prototype.visitMethodCall=function(t,e){return new Un(t.span,t.receiver.visit(this),t.name,this.visitAll(t.args))},t.prototype.visitSafeMethodCall=function(t,e){return new Wn(t.span,t.receiver.visit(this),t.name,this.visitAll(t.args))},t.prototype.visitFunctionCall=function(t,e){return new Gn(t.span,t.target.visit(this),this.visitAll(t.args))},t.prototype.visitLiteralArray=function(t,e){return new Fn(t.span,this.visitAll(t.expressions))},t.prototype.visitLiteralMap=function(t,e){return new Vn(t.span,t.keys,this.visitAll(t.values))},t.prototype.visitBinary=function(t,e){return new Bn(t.span,t.operation,t.left.visit(this),t.right.visit(this))},t.prototype.visitPrefixNot=function(t,e){return new qn(t.span,t.expression.visit(this))},t.prototype.visitNonNullAssert=function(t,e){return new Hn(t.span,t.expression.visit(this))},t.prototype.visitConditional=function(t,e){return new Mn(t.span,t.condition.visit(this),t.trueExp.visit(this),t.falseExp.visit(this))},t.prototype.visitPipe=function(t,e){return new Nn(t.span,t.exp.visit(this),t.name,this.visitAll(t.args))},t.prototype.visitKeyedRead=function(t,e){return new In(t.span,t.obj.visit(this),t.key.visit(this))},t.prototype.visitKeyedWrite=function(t,e){return new Rn(t.span,t.obj.visit(this),t.key.visit(this),t.value.visit(this))},t.prototype.visitAll=function(t){for(var e=new Array(t.length),n=0;n<t.length;++n)e[n]=t[n].visit(this);return e},t.prototype.visitChain=function(t,e){return new Pn(t.span,this.visitAll(t.expressions))},t.prototype.visitQuote=function(t,e){return new kn(t.span,t.prefix,t.uninterpretedExpression,t.location)},t}();var Qn=function(){return function(t,e,n){this.strings=t,this.expressions=e,this.offsets=n}}(),Zn=function(){return function(t,e,n){this.templateBindings=t,this.warnings=e,this.errors=n}}();function Jn(t){var e=W(t.start)+"([\\s\\S]*?)"+W(t.end);return new RegExp(e,"g")}var tr=function(){function t(t){this._lexer=t,this.errors=[]}return t.prototype.parseAction=function(t,e,n){void 0===n&&(n=$t),this._checkNoInterpolation(t,e,n);var r=this._stripComments(t),i=this._lexer.tokenize(this._stripComments(t)),o=new er(t,e,i,r.length,!0,this.errors,t.length-r.length).parseChain();return new Xn(o,t,e,this.errors)},t.prototype.parseBinding=function(t,e,n){void 0===n&&(n=$t);var r=this._parseBindingAst(t,e,n);return new Xn(r,t,e,this.errors)},t.prototype.parseSimpleBinding=function(t,e,n){void 0===n&&(n=$t);var r=this._parseBindingAst(t,e,n),i=nr.check(r);return i.length>0&&this._reportError("Host binding expression cannot contain "+i.join(" "),t,e),new Xn(r,t,e,this.errors)},t.prototype._reportError=function(t,e,n,r){this.errors.push(new wn(t,e,n,r))},t.prototype._parseBindingAst=function(t,e,n){var r=this._parseQuote(t,e);if(null!=r)return r;this._checkNoInterpolation(t,e,n);var i=this._stripComments(t),o=this._lexer.tokenize(i);return new er(t,e,o,i.length,!1,this.errors,t.length-i.length).parseChain()},t.prototype._parseQuote=function(t,e){if(null==t)return null;var n=t.indexOf(":");if(-1==n)return null;var r=t.substring(0,n).trim();if(!function(t){if(0==t.length)return!1;var e=new bn(t);if(!_n(e.peek))return!1;for(e.advance();e.peek!==Se;){if(!jn(e.peek))return!1;e.advance()}return!0}(r))return null;var i=t.substring(n+1);return new kn(new On(0,t.length),r,i,e)},t.prototype.parseTemplateBindings=function(t,e,n){var r=this._lexer.tokenize(e);if(t){var i=this._lexer.tokenize(t).map(function(t){return t.index=0,t});r.unshift.apply(r,i)}return new er(e,n,r,e.length,!1,this.errors,0).parseTemplateBindings()},t.prototype.parseInterpolation=function(t,e,n){void 0===n&&(n=$t);var r=this.splitInterpolation(t,e,n);if(null==r)return null;for(var i=[],o=0;o<r.expressions.length;++o){var s=r.expressions[o],a=this._stripComments(s),c=this._lexer.tokenize(a),u=new er(t,e,c,a.length,!1,this.errors,r.offsets[o]+(s.length-a.length)).parseChain();i.push(u)}return new Xn(new zn(new On(0,null==t?0:t.length),r.strings,i),t,e,this.errors)},t.prototype.splitInterpolation=function(t,e,n){void 0===n&&(n=$t);var r=Jn(n),i=t.split(r);if(i.length<=1)return null;for(var o=[],s=[],a=[],c=0,u=0;u<i.length;u++){var l=i[u];u%2==0?(o.push(l),c+=l.length):l.trim().length>0?(c+=n.start.length,s.push(l),a.push(c),c+=l.length+n.end.length):(this._reportError("Blank expressions are not allowed in interpolated strings",t,"at column "+this._findInterpolationErrorColumn(i,u,n)+" in",e),s.push("$implict"),a.push(c))}return new Qn(o,s,a)},t.prototype.wrapLiteralPrimitive=function(t,e){return new Xn(new Ln(new On(0,null==t?0:t.length),t),t,e,this.errors)},t.prototype._stripComments=function(t){var e=this._commentStart(t);return null!=e?t.substring(0,e).trim():t},t.prototype._commentStart=function(t){for(var e,n=null,r=0;r<t.length-1;r++){var i=t.charCodeAt(r),o=t.charCodeAt(r+1);if(i===ze&&o==ze&&null==n)return r;n===i?n=null:null==n&&((e=i)===Le||e===Re||e===cn)&&(n=i)}return null},t.prototype._checkNoInterpolation=function(t,e,n){var r=Jn(n),i=t.split(r);i.length>1&&this._reportError("Got interpolation ("+n.start+n.end+") where expression was expected",t,"at column "+this._findInterpolationErrorColumn(i,1,n)+" in",e)},t.prototype._findInterpolationErrorColumn=function(t,e,n){for(var r="",i=0;i<e;i++)r+=i%2==0?t[i]:""+n.start+t[i]+n.end;return r.length},t}(),er=function(){function t(t,e,n,r,i,o,s){this.input=t,this.location=e,this.tokens=n,this.inputLength=r,this.parseAction=i,this.errors=o,this.offset=s,this.rparensExpected=0,this.rbracketsExpected=0,this.rbracesExpected=0,this.index=0}return t.prototype.peek=function(t){var e=this.index+t;return e<this.tokens.length?this.tokens[e]:yn},Object.defineProperty(t.prototype,"next",{get:function(){return this.peek(0)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"inputIndex",{get:function(){return this.index<this.tokens.length?this.next.index+this.offset:this.inputLength+this.offset},enumerable:!0,configurable:!0}),t.prototype.span=function(t){return new On(t,this.inputIndex)},t.prototype.advance=function(){this.index++},t.prototype.optionalCharacter=function(t){return!!this.next.isCharacter(t)&&(this.advance(),!0)},t.prototype.peekKeywordLet=function(){return this.next.isKeywordLet()},t.prototype.peekKeywordAs=function(){return this.next.isKeywordAs()},t.prototype.expectCharacter=function(t){this.optionalCharacter(t)||this.error("Missing expected "+String.fromCharCode(t))},t.prototype.optionalOperator=function(t){return!!this.next.isOperator(t)&&(this.advance(),!0)},t.prototype.expectOperator=function(t){this.optionalOperator(t)||this.error("Missing expected operator "+t)},t.prototype.expectIdentifierOrKeyword=function(){var t=this.next;return t.isIdentifier()||t.isKeyword()?(this.advance(),t.toString()):(this.error("Unexpected token "+t+", expected identifier or keyword"),"")},t.prototype.expectIdentifierOrKeywordOrString=function(){var t=this.next;return t.isIdentifier()||t.isKeyword()||t.isString()?(this.advance(),t.toString()):(this.error("Unexpected token "+t+", expected identifier, keyword, or string"),"")},t.prototype.parseChain=function(){for(var t=[],e=this.inputIndex;this.index<this.tokens.length;){var n=this.parsePipe();if(t.push(n),this.optionalCharacter(Be))for(this.parseAction||this.error("Binding expression cannot contain chained expression");this.optionalCharacter(Be););else this.index<this.tokens.length&&this.error("Unexpected token '"+this.next+"'")}return 0==t.length?new En(this.span(e)):1==t.length?t[0]:new Pn(this.span(e),t)},t.prototype.parsePipe=function(){var t=this.parseExpression();if(this.optionalOperator("|")){this.parseAction&&this.error("Cannot have a pipe in an action expression");do{for(var e=this.expectIdentifierOrKeyword(),n=[];this.optionalCharacter(58);)n.push(this.parseExpression());t=new Nn(this.span(t.span.start),t,e,n)}while(this.optionalOperator("|"))}return t},t.prototype.parseExpression=function(){return this.parseConditional()},t.prototype.parseConditional=function(){var t=this.inputIndex,e=this.parseLogicalOr();if(this.optionalOperator("?")){var n=this.parsePipe(),r=void 0;if(this.optionalCharacter(58))r=this.parsePipe();else{var i=this.inputIndex,o=this.input.substring(t,i);this.error("Conditional expression "+o+" requires all 3 expressions"),r=new En(this.span(t))}return new Mn(this.span(t),e,n,r)}return e},t.prototype.parseLogicalOr=function(){for(var t=this.parseLogicalAnd();this.optionalOperator("||");){var e=this.parseLogicalAnd();t=new Bn(this.span(t.span.start),"||",t,e)}return t},t.prototype.parseLogicalAnd=function(){for(var t=this.parseEquality();this.optionalOperator("&&");){var e=this.parseEquality();t=new Bn(this.span(t.span.start),"&&",t,e)}return t},t.prototype.parseEquality=function(){for(var t=this.parseRelational();this.next.type==dn.Operator;){var e=this.next.strValue;switch(e){case"==":case"===":case"!=":case"!==":this.advance();var n=this.parseRelational();t=new Bn(this.span(t.span.start),e,t,n);continue}break}return t},t.prototype.parseRelational=function(){for(var t=this.parseAdditive();this.next.type==dn.Operator;){var e=this.next.strValue;switch(e){case"<":case">":case"<=":case">=":this.advance();var n=this.parseAdditive();t=new Bn(this.span(t.span.start),e,t,n);continue}break}return t},t.prototype.parseAdditive=function(){for(var t=this.parseMultiplicative();this.next.type==dn.Operator;){var e=this.next.strValue;switch(e){case"+":case"-":this.advance();var n=this.parseMultiplicative();t=new Bn(this.span(t.span.start),e,t,n);continue}break}return t},t.prototype.parseMultiplicative=function(){for(var t=this.parsePrefix();this.next.type==dn.Operator;){var e=this.next.strValue;switch(e){case"*":case"%":case"/":this.advance();var n=this.parsePrefix();t=new Bn(this.span(t.span.start),e,t,n);continue}break}return t},t.prototype.parsePrefix=function(){if(this.next.type==dn.Operator){var t=this.inputIndex,e=this.next.strValue,n=void 0;switch(e){case"+":return this.advance(),this.parsePrefix();case"-":return this.advance(),n=this.parsePrefix(),new Bn(this.span(t),e,new Ln(new On(t,t),0),n);case"!":return this.advance(),n=this.parsePrefix(),new qn(this.span(t),n)}}return this.parseCallChain()},t.prototype.parseCallChain=function(){for(var t=this.parsePrimary();;)if(this.optionalCharacter(46))t=this.parseAccessMemberOrMethodCall(t,!1);else if(this.optionalOperator("?."))t=this.parseAccessMemberOrMethodCall(t,!0);else if(this.optionalCharacter(91)){this.rbracketsExpected++;var e=this.parsePipe();if(this.rbracketsExpected--,this.expectCharacter(93),this.optionalOperator("=")){var n=this.parseConditional();t=new Rn(this.span(t.span.start),t,e,n)}else t=new In(this.span(t.span.start),t,e)}else if(this.optionalCharacter(40)){this.rparensExpected++;var r=this.parseCallArguments();this.rparensExpected--,this.expectCharacter(41),t=new Gn(this.span(t.span.start),t,r)}else{if(!this.optionalOperator("!"))return t;t=new Hn(this.span(t.span.start),t)}},t.prototype.parsePrimary=function(){var t=this.inputIndex;if(this.optionalCharacter(40)){this.rparensExpected++;var e=this.parsePipe();return this.rparensExpected--,this.expectCharacter(41),e}if(this.next.isKeywordNull())return this.advance(),new Ln(this.span(t),null);if(this.next.isKeywordUndefined())return this.advance(),new Ln(this.span(t),void 0);if(this.next.isKeywordTrue())return this.advance(),new Ln(this.span(t),!0);if(this.next.isKeywordFalse())return this.advance(),new Ln(this.span(t),!1);if(this.next.isKeywordThis())return this.advance(),new Sn(this.span(t));if(this.optionalCharacter(91)){this.rbracketsExpected++;var n=this.parseExpressionList(93);return this.rbracketsExpected--,this.expectCharacter(93),new Fn(this.span(t),n)}if(this.next.isCharacter(sn))return this.parseLiteralMap();if(this.next.isIdentifier())return this.parseAccessMemberOrMethodCall(new Sn(this.span(t)),!1);if(this.next.isNumber()){var r=this.next.toNumber();return this.advance(),new Ln(this.span(t),r)}if(this.next.isString()){var i=this.next.toString();return this.advance(),new Ln(this.span(t),i)}return this.index>=this.tokens.length?(this.error("Unexpected end of expression: "+this.input),new En(this.span(t))):(this.error("Unexpected token "+this.next),new En(this.span(t)))},t.prototype.parseExpressionList=function(t){var e=[];if(!this.next.isCharacter(t))do{e.push(this.parsePipe())}while(this.optionalCharacter(44));return e},t.prototype.parseLiteralMap=function(){var t=[],e=[],n=this.inputIndex;if(this.expectCharacter(sn),!this.optionalCharacter(125)){this.rbracesExpected++;do{var r=this.next.isString(),i=this.expectIdentifierOrKeywordOrString();t.push({key:i,quoted:r}),this.expectCharacter(58),e.push(this.parsePipe())}while(this.optionalCharacter(44));this.rbracesExpected--,this.expectCharacter(125)}return new Vn(this.span(n),t,e)},t.prototype.parseAccessMemberOrMethodCall=function(t,e){void 0===e&&(e=!1);var n=t.span.start,r=this.expectIdentifierOrKeyword();if(this.optionalCharacter(40)){this.rparensExpected++;var i=this.parseCallArguments();this.expectCharacter(41),this.rparensExpected--;var o=this.span(n);return e?new Wn(o,t,r,i):new Un(o,t,r,i)}if(e)return this.optionalOperator("=")?(this.error("The '?.' operator cannot be used in the assignment"),new En(this.span(n))):new Dn(this.span(n),t,r);if(this.optionalOperator("=")){if(!this.parseAction)return this.error("Bindings cannot contain assignments"),new En(this.span(n));var s=this.parseConditional();return new Tn(this.span(n),t,r,s)}return new An(this.span(n),t,r)},t.prototype.parseCallArguments=function(){if(this.next.isCharacter(41))return[];var t=[];do{t.push(this.parsePipe())}while(this.optionalCharacter(44));return t},t.prototype.expectTemplateBindingKey=function(){var t="",e=!1;do{t+=this.expectIdentifierOrKeywordOrString(),(e=this.optionalOperator("-"))&&(t+="-")}while(e);return t.toString()},t.prototype.parseTemplateBindings=function(){for(var t=[],e=null;this.index<this.tokens.length;){var n=this.inputIndex,r=this.peekKeywordLet();r&&this.advance();var i=this.expectTemplateBindingKey(),o=i;r||(null==e?e=o:o=e+o[0].toUpperCase()+o.substring(1)),this.optionalCharacter(58);var s=null,a=null;if(r)s=this.optionalOperator("=")?this.expectTemplateBindingKey():"$implicit";else if(this.peekKeywordAs()){var c=this.inputIndex;this.advance(),s=i,o=this.expectTemplateBindingKey(),r=!0}else if(this.next!==yn&&!this.peekKeywordLet()){var u=this.inputIndex,l=this.parsePipe(),p=this.input.substring(u-this.offset,this.inputIndex-this.offset);a=new Xn(l,p,this.location,this.errors)}if(t.push(new Yn(this.span(n),o,r,s,a)),this.peekKeywordAs()&&!r){c=this.inputIndex;this.advance();var d=this.expectTemplateBindingKey();t.push(new Yn(this.span(c),d,!0,o,null))}this.optionalCharacter(Be)||this.optionalCharacter(44)}return new Zn(t,[],this.errors)},t.prototype.error=function(t,e){void 0===e&&(e=null),this.errors.push(new wn(t,this.input,this.locationText(e),this.location)),this.skip()},t.prototype.locationText=function(t){return void 0===t&&(t=null),null==t&&(t=this.index),t<this.tokens.length?"at column "+(this.tokens[t].index+1)+" in":"at the end of the expression"},t.prototype.skip=function(){for(var t=this.next;this.index<this.tokens.length&&!t.isCharacter(Be)&&(this.rparensExpected<=0||!t.isCharacter(41))&&(this.rbracesExpected<=0||!t.isCharacter(125))&&(this.rbracketsExpected<=0||!t.isCharacter(93));)this.next.isError()&&this.errors.push(new wn(this.next.toString(),this.input,this.locationText(),this.location)),this.advance(),t=this.next},t}(),nr=function(){function t(){this.errors=[]}return t.check=function(e){var n=new t;return e.visit(n),n.errors},t.prototype.visitImplicitReceiver=function(t,e){},t.prototype.visitInterpolation=function(t,e){},t.prototype.visitLiteralPrimitive=function(t,e){},t.prototype.visitPropertyRead=function(t,e){},t.prototype.visitPropertyWrite=function(t,e){},t.prototype.visitSafePropertyRead=function(t,e){},t.prototype.visitMethodCall=function(t,e){},t.prototype.visitSafeMethodCall=function(t,e){},t.prototype.visitFunctionCall=function(t,e){},t.prototype.visitLiteralArray=function(t,e){this.visitAll(t.expressions)},t.prototype.visitLiteralMap=function(t,e){this.visitAll(t.values)},t.prototype.visitBinary=function(t,e){},t.prototype.visitPrefixNot=function(t,e){},t.prototype.visitNonNullAssert=function(t,e){},t.prototype.visitConditional=function(t,e){},t.prototype.visitPipe=function(t,e){this.errors.push("pipes")},t.prototype.visitKeyedRead=function(t,e){},t.prototype.visitKeyedWrite=function(t,e){},t.prototype.visitAll=function(t){var e=this;return t.map(function(t){return t.visit(e)})},t.prototype.visitChain=function(t,e){},t.prototype.visitQuote=function(t,e){},t}(),rr=function(){function t(t,e,n,r){this.file=t,this.offset=e,this.line=n,this.col=r}return t.prototype.toString=function(){return null!=this.offset?this.file.url+"@"+this.line+":"+this.col:this.file.url},t.prototype.moveBy=function(e){for(var n=this.file.content,r=n.length,i=this.offset,o=this.line,s=this.col;i>0&&e<0;){if(i--,e++,(c=n.charCodeAt(i))==Me){o--;var a=n.substr(0,i-1).lastIndexOf(String.fromCharCode(Me));s=a>0?i-a:i}else s--}for(;i<r&&e>0;){var c=n.charCodeAt(i);i++,e--,c==Me?(o++,s=0):s++}return new t(this.file,i,o,s)},t.prototype.getContext=function(t,e){var n=this.file.content,r=this.offset;if(null!=r){r>n.length-1&&(r=n.length-1);for(var i=r,o=0,s=0;o<t&&r>0&&(o++,"\n"!=n[--r]||++s!=e););for(o=0,s=0;o<t&&i<n.length-1&&(o++,"\n"!=n[++i]||++s!=e););return{before:n.substring(r,this.offset),after:n.substring(this.offset,i+1)}}return null},t}(),ir=function(){return function(t,e){this.content=t,this.url=e}}(),or=function(){function t(t,e,n){void 0===n&&(n=null),this.start=t,this.end=e,this.details=n}return t.prototype.toString=function(){return this.start.file.content.substring(this.start.offset,this.end.offset)},t}(),sr={WARNING:0,ERROR:1};sr[sr.WARNING]="WARNING",sr[sr.ERROR]="ERROR";var ar=function(){function t(t,e,n){void 0===n&&(n=sr.ERROR),this.span=t,this.msg=e,this.level=n}return t.prototype.contextualMessage=function(){var t=this.span.start.getContext(100,3);return t?this.msg+' ("'+t.before+"["+sr[this.level]+" ->]"+t.after+'")':this.msg},t.prototype.toString=function(){var t=this.span.details?", "+this.span.details:"";return this.contextualMessage()+": "+this.span.start+t},t}();var cr={TAG_OPEN_START:0,TAG_OPEN_END:1,TAG_OPEN_END_VOID:2,TAG_CLOSE:3,TEXT:4,ESCAPABLE_RAW_TEXT:5,RAW_TEXT:6,COMMENT_START:7,COMMENT_END:8,CDATA_START:9,CDATA_END:10,ATTR_NAME:11,ATTR_VALUE:12,DOC_TYPE:13,EXPANSION_FORM_START:14,EXPANSION_CASE_VALUE:15,EXPANSION_CASE_EXP_START:16,EXPANSION_CASE_EXP_END:17,EXPANSION_FORM_END:18,EOF:19};cr[cr.TAG_OPEN_START]="TAG_OPEN_START",cr[cr.TAG_OPEN_END]="TAG_OPEN_END",cr[cr.TAG_OPEN_END_VOID]="TAG_OPEN_END_VOID",cr[cr.TAG_CLOSE]="TAG_CLOSE",cr[cr.TEXT]="TEXT",cr[cr.ESCAPABLE_RAW_TEXT]="ESCAPABLE_RAW_TEXT",cr[cr.RAW_TEXT]="RAW_TEXT",cr[cr.COMMENT_START]="COMMENT_START",cr[cr.COMMENT_END]="COMMENT_END",cr[cr.CDATA_START]="CDATA_START",cr[cr.CDATA_END]="CDATA_END",cr[cr.ATTR_NAME]="ATTR_NAME",cr[cr.ATTR_VALUE]="ATTR_VALUE",cr[cr.DOC_TYPE]="DOC_TYPE",cr[cr.EXPANSION_FORM_START]="EXPANSION_FORM_START",cr[cr.EXPANSION_CASE_VALUE]="EXPANSION_CASE_VALUE",cr[cr.EXPANSION_CASE_EXP_START]="EXPANSION_CASE_EXP_START",cr[cr.EXPANSION_CASE_EXP_END]="EXPANSION_CASE_EXP_END",cr[cr.EXPANSION_FORM_END]="EXPANSION_FORM_END",cr[cr.EOF]="EOF";var ur=function(){return function(t,e,n){this.type=t,this.parts=e,this.sourceSpan=n}}(),lr=function(t){function e(e,n,r){var i=t.call(this,r,e)||this;return i.tokenType=n,i}return Object(r.b)(e,t),e}(ar),pr=function(){return function(t,e){this.tokens=t,this.errors=e}}();var dr=/\r\n?/g;function hr(t){return'Unexpected character "'+(t===Se?"EOF":String.fromCharCode(t))+'"'}function fr(t){return'Unknown entity "'+t+'" - use the "&#<decimal>;" or  "&#x<hex>;" syntax'}var mr=function(){return function(t){this.error=t}}(),gr=function(){function t(t,e,n,r){void 0===r&&(r=$t),this._file=t,this._getTagDefinition=e,this._tokenizeIcu=n,this._interpolationConfig=r,this._peek=-1,this._nextPeek=-1,this._index=-1,this._line=0,this._column=-1,this._expansionCaseStack=[],this._inInterpolation=!1,this.tokens=[],this.errors=[],this._input=t.content,this._length=t.content.length,this._advance()}return t.prototype._processCarriageReturns=function(t){return t.replace(dr,"\n")},t.prototype.tokenize=function(){for(;this._peek!==Se;){var t=this._getLocation();try{this._attemptCharCode(60)?this._attemptCharCode(33)?this._attemptCharCode(91)?this._consumeCdata(t):this._attemptCharCode(Ve)?this._consumeComment(t):this._consumeDocType(t):this._attemptCharCode(ze)?this._consumeTagClose(t):this._consumeTagOpen(t):this._tokenizeIcu&&this._tokenizeExpansionForm()||this._consumeText()}catch(t){if(!(t instanceof mr))throw t;this.errors.push(t.error)}}return this._beginToken(cr.EOF),this._endToken([]),new pr(function(t){for(var e=[],n=void 0,r=0;r<t.length;r++){var i=t[r];n&&n.type==cr.TEXT&&i.type==cr.TEXT?(n.parts[0]+=i.parts[0],n.sourceSpan.end=i.sourceSpan.end):(n=i,e.push(n))}return e}(this.tokens),this.errors)},t.prototype._tokenizeExpansionForm=function(){if(jr(this._input,this._index,this._interpolationConfig))return this._consumeExpansionFormStart(),!0;if(((t=this._peek)===qe||pn(t)||ln(t))&&this._isInExpansionForm())return this._consumeExpansionCaseStart(),!0;var t;if(125===this._peek){if(this._isInExpansionCase())return this._consumeExpansionCaseEnd(),!0;if(this._isInExpansionForm())return this._consumeExpansionFormEnd(),!0}return!1},t.prototype._getLocation=function(){return new rr(this._file,this._index,this._line,this._column)},t.prototype._getSpan=function(t,e){return void 0===t&&(t=this._getLocation()),void 0===e&&(e=this._getLocation()),new or(t,e)},t.prototype._beginToken=function(t,e){void 0===e&&(e=this._getLocation()),this._currentTokenStart=e,this._currentTokenType=t},t.prototype._endToken=function(t,e){void 0===e&&(e=this._getLocation());var n=new ur(this._currentTokenType,t,new or(this._currentTokenStart,e));return this.tokens.push(n),this._currentTokenStart=null,this._currentTokenType=null,n},t.prototype._createError=function(t,e){this._isInExpansionForm()&&(t+=' (Do you have an unescaped "{" in your template? Use "{{ \'{\' }}") to escape it.)');var n=new lr(t,this._currentTokenType,e);return this._currentTokenStart=null,this._currentTokenType=null,new mr(n)},t.prototype._advance=function(){if(this._index>=this._length)throw this._createError(hr(Se),this._getSpan());this._peek===Me?(this._line++,this._column=0):this._peek!==Me&&this._peek!==De&&this._column++,this._index++,this._peek=this._index>=this._length?Se:this._input.charCodeAt(this._index),this._nextPeek=this._index+1>=this._length?Se:this._input.charCodeAt(this._index+1)},t.prototype._attemptCharCode=function(t){return this._peek===t&&(this._advance(),!0)},t.prototype._attemptCharCodeCaseInsensitive=function(t){return e=this._peek,n=t,xr(e)==xr(n)&&(this._advance(),!0);var e,n},t.prototype._requireCharCode=function(t){var e=this._getLocation();if(!this._attemptCharCode(t))throw this._createError(hr(this._peek),this._getSpan(e,e))},t.prototype._attemptStr=function(t){var e=t.length;if(this._index+e>this._length)return!1;for(var n=this._savePosition(),r=0;r<e;r++)if(!this._attemptCharCode(t.charCodeAt(r)))return this._restorePosition(n),!1;return!0},t.prototype._attemptStrCaseInsensitive=function(t){for(var e=0;e<t.length;e++)if(!this._attemptCharCodeCaseInsensitive(t.charCodeAt(e)))return!1;return!0},t.prototype._requireStr=function(t){var e=this._getLocation();if(!this._attemptStr(t))throw this._createError(hr(this._peek),this._getSpan(e))},t.prototype._attemptCharCodeUntilFn=function(t){for(;!t(this._peek);)this._advance()},t.prototype._requireCharCodeUntilFn=function(t,e){var n=this._getLocation();if(this._attemptCharCodeUntilFn(t),this._index-n.offset<e)throw this._createError(hr(this._peek),this._getSpan(n,n))},t.prototype._attemptUntilChar=function(t){for(;this._peek!==t;)this._advance()},t.prototype._readChar=function(t){if(t&&38===this._peek)return this._decodeEntity();var e=this._index;return this._advance(),this._input[e]},t.prototype._decodeEntity=function(){var t=this._getLocation();if(this._advance(),!this._attemptCharCode(35)){var e=this._savePosition();if(this._attemptCharCodeUntilFn(_r),this._peek!=Be)return this._restorePosition(e),"&";this._advance();var n=this._input.substring(t.offset+1,this._index-1),r=ce[n];if(!r)throw this._createError(fr(n),this._getSpan(t));return r}var i=this._attemptCharCode(120)||this._attemptCharCode(88),o=this._getLocation().offset;if(this._attemptCharCodeUntilFn(br),this._peek!=Be)throw this._createError(hr(this._peek),this._getSpan());this._advance();var s=this._input.substring(o,this._index-1);try{var a=parseInt(s,i?16:10);return String.fromCharCode(a)}catch(e){var c=this._input.substring(t.offset+1,this._index-1);throw this._createError(fr(c),this._getSpan(t))}},t.prototype._consumeRawText=function(t,e,n){var r,i=this._getLocation();this._beginToken(t?cr.ESCAPABLE_RAW_TEXT:cr.RAW_TEXT,i);for(var o=[];r=this._getLocation(),!this._attemptCharCode(e)||!n();)for(this._index>r.offset&&o.push(this._input.substring(r.offset,this._index));this._peek!==e;)o.push(this._readChar(t));return this._endToken([this._processCarriageReturns(o.join(""))],r)},t.prototype._consumeComment=function(t){var e=this;this._beginToken(cr.COMMENT_START,t),this._requireCharCode(Ve),this._endToken([]);var n=this._consumeRawText(!1,Ve,function(){return e._attemptStr("->")});this._beginToken(cr.COMMENT_END,n.sourceSpan.end),this._endToken([])},t.prototype._consumeCdata=function(t){var e=this;this._beginToken(cr.CDATA_START,t),this._requireStr("CDATA["),this._endToken([]);var n=this._consumeRawText(!1,93,function(){return e._attemptStr("]>")});this._beginToken(cr.CDATA_END,n.sourceSpan.end),this._endToken([])},t.prototype._consumeDocType=function(t){this._beginToken(cr.DOC_TYPE,t),this._attemptUntilChar(He),this._advance(),this._endToken([this._input.substring(t.offset+2,this._index-1)])},t.prototype._consumePrefixAndName=function(){for(var t,e,n=this._index,r=null;58!==this._peek&&!(((t=this._peek)<Qe||on<t)&&(t<Ge||Ke<t)&&(t<Ue||t>We));)this._advance();return 58===this._peek?(this._advance(),r=this._input.substring(n,this._index-1),e=this._index):e=n,this._requireCharCodeUntilFn(yr,this._index===e?1:0),[r,this._input.substring(e,this._index)]},t.prototype._consumeTagOpen=function(t){var e,n,r=this._savePosition();try{if(!pn(this._peek))throw this._createError(hr(this._peek),this._getSpan());var i=this._index;for(this._consumeTagOpenStart(t),n=(e=this._input.substring(i,this._index)).toLowerCase(),this._attemptCharCodeUntilFn(vr);this._peek!==ze&&this._peek!==He;)this._consumeAttributeName(),this._attemptCharCodeUntilFn(vr),this._attemptCharCode(qe)&&(this._attemptCharCodeUntilFn(vr),this._consumeAttributeValue()),this._attemptCharCodeUntilFn(vr);this._consumeTagOpenEnd()}catch(e){if(e instanceof mr)return this._restorePosition(r),this._beginToken(cr.TEXT,t),void this._endToken(["<"]);throw e}var o=this._getTagDefinition(e).contentType;o===ne.RAW_TEXT?this._consumeRawTextWithTagClose(n,!1):o===ne.ESCAPABLE_RAW_TEXT&&this._consumeRawTextWithTagClose(n,!0)},t.prototype._consumeRawTextWithTagClose=function(t,e){var n=this,r=this._consumeRawText(e,60,function(){return!!n._attemptCharCode(ze)&&(n._attemptCharCodeUntilFn(vr),!!n._attemptStrCaseInsensitive(t)&&(n._attemptCharCodeUntilFn(vr),n._attemptCharCode(He)))});this._beginToken(cr.TAG_CLOSE,r.sourceSpan.end),this._endToken([null,t])},t.prototype._consumeTagOpenStart=function(t){this._beginToken(cr.TAG_OPEN_START,t);var e=this._consumePrefixAndName();this._endToken(e)},t.prototype._consumeAttributeName=function(){this._beginToken(cr.ATTR_NAME);var t=this._consumePrefixAndName();this._endToken(t)},t.prototype._consumeAttributeValue=function(){var t;if(this._beginToken(cr.ATTR_VALUE),this._peek===Le||this._peek===Re){var e=this._peek;this._advance();for(var n=[];this._peek!==e;)n.push(this._readChar(!0));t=n.join(""),this._advance()}else{var r=this._index;this._requireCharCodeUntilFn(yr,1),t=this._input.substring(r,this._index)}this._endToken([this._processCarriageReturns(t)])},t.prototype._consumeTagOpenEnd=function(){var t=this._attemptCharCode(ze)?cr.TAG_OPEN_END_VOID:cr.TAG_OPEN_END;this._beginToken(t),this._requireCharCode(He),this._endToken([])},t.prototype._consumeTagClose=function(t){this._beginToken(cr.TAG_CLOSE,t),this._attemptCharCodeUntilFn(vr);var e=this._consumePrefixAndName();this._attemptCharCodeUntilFn(vr),this._requireCharCode(He),this._endToken(e)},t.prototype._consumeExpansionFormStart=function(){this._beginToken(cr.EXPANSION_FORM_START,this._getLocation()),this._requireCharCode(sn),this._endToken([]),this._expansionCaseStack.push(cr.EXPANSION_FORM_START),this._beginToken(cr.RAW_TEXT,this._getLocation());var t=this._readUntil(44);this._endToken([t],this._getLocation()),this._requireCharCode(44),this._attemptCharCodeUntilFn(vr),this._beginToken(cr.RAW_TEXT,this._getLocation());var e=this._readUntil(44);this._endToken([e],this._getLocation()),this._requireCharCode(44),this._attemptCharCodeUntilFn(vr)},t.prototype._consumeExpansionCaseStart=function(){this._beginToken(cr.EXPANSION_CASE_VALUE,this._getLocation());var t=this._readUntil(sn).trim();this._endToken([t],this._getLocation()),this._attemptCharCodeUntilFn(vr),this._beginToken(cr.EXPANSION_CASE_EXP_START,this._getLocation()),this._requireCharCode(sn),this._endToken([],this._getLocation()),this._attemptCharCodeUntilFn(vr),this._expansionCaseStack.push(cr.EXPANSION_CASE_EXP_START)},t.prototype._consumeExpansionCaseEnd=function(){this._beginToken(cr.EXPANSION_CASE_EXP_END,this._getLocation()),this._requireCharCode(125),this._endToken([],this._getLocation()),this._attemptCharCodeUntilFn(vr),this._expansionCaseStack.pop()},t.prototype._consumeExpansionFormEnd=function(){this._beginToken(cr.EXPANSION_FORM_END,this._getLocation()),this._requireCharCode(125),this._endToken([]),this._expansionCaseStack.pop()},t.prototype._consumeText=function(){var t=this._getLocation();this._beginToken(cr.TEXT,t);var e=[];do{this._interpolationConfig&&this._attemptStr(this._interpolationConfig.start)?(e.push(this._interpolationConfig.start),this._inInterpolation=!0):this._interpolationConfig&&this._inInterpolation&&this._attemptStr(this._interpolationConfig.end)?(e.push(this._interpolationConfig.end),this._inInterpolation=!1):e.push(this._readChar(!0))}while(!this._isTextEnd());this._endToken([this._processCarriageReturns(e.join(""))])},t.prototype._isTextEnd=function(){if(60===this._peek||this._peek===Se)return!0;if(this._tokenizeIcu&&!this._inInterpolation){if(jr(this._input,this._index,this._interpolationConfig))return!0;if(125===this._peek&&this._isInExpansionCase())return!0}return!1},t.prototype._savePosition=function(){return[this._peek,this._index,this._column,this._line,this.tokens.length]},t.prototype._readUntil=function(t){var e=this._index;return this._attemptUntilChar(t),this._input.substring(e,this._index)},t.prototype._restorePosition=function(t){this._peek=t[0],this._index=t[1],this._column=t[2],this._line=t[3];var e=t[4];e<this.tokens.length&&(this.tokens=this.tokens.slice(0,e))},t.prototype._isInExpansionCase=function(){return this._expansionCaseStack.length>0&&this._expansionCaseStack[this._expansionCaseStack.length-1]===cr.EXPANSION_CASE_EXP_START},t.prototype._isInExpansionForm=function(){return this._expansionCaseStack.length>0&&this._expansionCaseStack[this._expansionCaseStack.length-1]===cr.EXPANSION_FORM_START},t}();function vr(t){return!un(t)||t===Se}function yr(t){return un(t)||t===He||t===ze||t===Le||t===Re||t===qe}function br(t){return t==Be||t==Se||!((e=t)>=Qe&&e<=Je||e>=Ge&&e<=Ye||ln(e));var e}function _r(t){return t==Be||t==Se||!pn(t)}function jr(t,e,n){var r=!!n&&t.indexOf(n.start,e)==e;return t.charCodeAt(e)==sn&&!r}function xr(t){return t>=Qe&&t<=on?t-Qe+Ge:t}var wr=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.elementName=e,i}return Object(r.b)(e,t),e.create=function(t,n,r){return new e(t,n,r)},e}(ar),Or=function(){return function(t,e){this.rootNodes=t,this.errors=e}}(),Cr=function(){function t(t){this.getTagDefinition=t}return t.prototype.parse=function(t,e,n,r){void 0===n&&(n=!1),void 0===r&&(r=$t);var i,o,s,a,c,u=(i=t,o=e,s=this.getTagDefinition,c=r,void 0===(a=n)&&(a=!1),void 0===c&&(c=$t),new gr(new ir(i,o),s,a,c).tokenize()),l=new kr(u.tokens,this.getTagDefinition).build();return new Or(l.rootNodes,u.errors.concat(l.errors))},t}(),kr=function(){function t(t,e){this.tokens=t,this.getTagDefinition=e,this._index=-1,this._rootNodes=[],this._errors=[],this._elementStack=[],this._advance()}return t.prototype.build=function(){for(;this._peek.type!==cr.EOF;)this._peek.type===cr.TAG_OPEN_START?this._consumeStartTag(this._advance()):this._peek.type===cr.TAG_CLOSE?this._consumeEndTag(this._advance()):this._peek.type===cr.CDATA_START?(this._closeVoidElement(),this._consumeCdata(this._advance())):this._peek.type===cr.COMMENT_START?(this._closeVoidElement(),this._consumeComment(this._advance())):this._peek.type===cr.TEXT||this._peek.type===cr.RAW_TEXT||this._peek.type===cr.ESCAPABLE_RAW_TEXT?(this._closeVoidElement(),this._consumeText(this._advance())):this._peek.type===cr.EXPANSION_FORM_START?this._consumeExpansion(this._advance()):this._advance();return new Or(this._rootNodes,this._errors)},t.prototype._advance=function(){var t=this._peek;return this._index<this.tokens.length-1&&this._index++,this._peek=this.tokens[this._index],t},t.prototype._advanceIf=function(t){return this._peek.type===t?this._advance():null},t.prototype._consumeCdata=function(t){this._consumeText(this._advance()),this._advanceIf(cr.CDATA_END)},t.prototype._consumeComment=function(t){var e=this._advanceIf(cr.RAW_TEXT);this._advanceIf(cr.COMMENT_END);var n=null!=e?e.parts[0].trim():null;this._addToParent(new Ut(n,t.sourceSpan))},t.prototype._consumeExpansion=function(t){for(var e=this._advance(),n=this._advance(),r=[];this._peek.type===cr.EXPANSION_CASE_VALUE;){var i=this._parseExpansionCase();if(!i)return;r.push(i)}if(this._peek.type===cr.EXPANSION_FORM_END){var o=new or(t.sourceSpan.start,this._peek.sourceSpan.end);this._addToParent(new zt(e.parts[0],n.parts[0],r,o,e.sourceSpan)),this._advance()}else this._errors.push(wr.create(null,this._peek.sourceSpan,"Invalid ICU message. Missing '}'."))},t.prototype._parseExpansionCase=function(){var e=this._advance();if(this._peek.type!==cr.EXPANSION_CASE_EXP_START)return this._errors.push(wr.create(null,this._peek.sourceSpan,"Invalid ICU message. Missing '{'.")),null;var n=this._advance(),r=this._collectExpansionExpTokens(n);if(!r)return null;var i=this._advance();r.push(new ur(cr.EOF,[],i.sourceSpan));var o=new t(r,this.getTagDefinition).build();if(o.errors.length>0)return this._errors=this._errors.concat(o.errors),null;var s=new or(e.sourceSpan.start,i.sourceSpan.end),a=new or(n.sourceSpan.start,i.sourceSpan.end);return new Bt(e.parts[0],o.rootNodes,s,e.sourceSpan,a)},t.prototype._collectExpansionExpTokens=function(t){for(var e=[],n=[cr.EXPANSION_CASE_EXP_START];;){if(this._peek.type!==cr.EXPANSION_FORM_START&&this._peek.type!==cr.EXPANSION_CASE_EXP_START||n.push(this._peek.type),this._peek.type===cr.EXPANSION_CASE_EXP_END){if(!Er(n,cr.EXPANSION_CASE_EXP_START))return this._errors.push(wr.create(null,t.sourceSpan,"Invalid ICU message. Missing '}'.")),null;if(n.pop(),0==n.length)return e}if(this._peek.type===cr.EXPANSION_FORM_END){if(!Er(n,cr.EXPANSION_FORM_START))return this._errors.push(wr.create(null,t.sourceSpan,"Invalid ICU message. Missing '}'.")),null;n.pop()}if(this._peek.type===cr.EOF)return this._errors.push(wr.create(null,t.sourceSpan,"Invalid ICU message. Missing '}'.")),null;e.push(this._advance())}},t.prototype._consumeText=function(t){var e=t.parts[0];if(e.length>0&&"\n"==e[0]){var n=this._getParentElement();null!=n&&0==n.children.length&&this.getTagDefinition(n.name).ignoreFirstLf&&(e=e.substring(1))}e.length>0&&this._addToParent(new Vt(e,t.sourceSpan))},t.prototype._closeVoidElement=function(){var t=this._getParentElement();t&&this.getTagDefinition(t.name).isVoid&&this._elementStack.pop()},t.prototype._consumeStartTag=function(t){for(var e=t.parts[0],n=t.parts[1],r=[];this._peek.type===cr.ATTR_NAME;)r.push(this._consumeAttr(this._advance()));var i=this._getElementFullName(e,n,this._getParentElement()),o=!1;if(this._peek.type===cr.TAG_OPEN_END_VOID){this._advance(),o=!0;var s=this.getTagDefinition(i);s.canSelfClose||null!==se(i)||s.isVoid||this._errors.push(wr.create(i,t.sourceSpan,'Only void and foreign elements can be self closed "'+t.parts[1]+'"'))}else this._peek.type===cr.TAG_OPEN_END&&(this._advance(),o=!1);var a=this._peek.sourceSpan.start,c=new or(t.sourceSpan.start,a),u=new Ht(i,r,[],c,c,void 0);this._pushElement(u),o&&(this._popElement(i),u.endSourceSpan=c)},t.prototype._pushElement=function(t){var e=this._getParentElement();e&&this.getTagDefinition(e.name).isClosedByChild(t.name)&&this._elementStack.pop();var n=this.getTagDefinition(t.name),r=this._getParentElementSkippingContainers(),i=r.parent,o=r.container;if(i&&n.requireExtraParent(i.name)){var s=new Ht(n.parentToAdd,[],[],t.sourceSpan,t.startSourceSpan,t.endSourceSpan);this._insertBeforeContainer(i,o,s)}this._addToParent(t),this._elementStack.push(t)},t.prototype._consumeEndTag=function(t){var e=this._getElementFullName(t.parts[0],t.parts[1],this._getParentElement());if(this._getParentElement()&&(this._getParentElement().endSourceSpan=t.sourceSpan),this.getTagDefinition(e).isVoid)this._errors.push(wr.create(e,t.sourceSpan,'Void elements do not have end tags "'+t.parts[1]+'"'));else if(!this._popElement(e)){var n='Unexpected closing tag "'+e+'". It may happen when the tag has already been closed by another tag. For more info see https://www.w3.org/TR/html5/syntax.html#closing-elements-that-have-implied-end-tags';this._errors.push(wr.create(e,t.sourceSpan,n))}},t.prototype._popElement=function(t){for(var e=this._elementStack.length-1;e>=0;e--){var n=this._elementStack[e];if(n.name==t)return this._elementStack.splice(e,this._elementStack.length-e),!0;if(!this.getTagDefinition(n.name).closedByParent)return!1}return!1},t.prototype._consumeAttr=function(t){var e=ae(t.parts[0],t.parts[1]),n=t.sourceSpan.end,r="",i=void 0;if(this._peek.type===cr.ATTR_VALUE){var o=this._advance();r=o.parts[0],n=o.sourceSpan.end,i=o.sourceSpan}return new qt(e,r,new or(t.sourceSpan.start,n),i)},t.prototype._getParentElement=function(){return this._elementStack.length>0?this._elementStack[this._elementStack.length-1]:null},t.prototype._getParentElementSkippingContainers=function(){for(var t=null,e=this._elementStack.length-1;e>=0;e--){if(!ie(this._elementStack[e].name))return{parent:this._elementStack[e],container:t};t=this._elementStack[e]}return{parent:null,container:t}},t.prototype._addToParent=function(t){var e=this._getParentElement();null!=e?e.children.push(t):this._rootNodes.push(t)},t.prototype._insertBeforeContainer=function(t,e,n){if(e){if(t){var r=t.children.indexOf(e);t.children[r]=n}else this._rootNodes.push(n);n.children.push(e),this._elementStack.splice(this._elementStack.indexOf(e),0,n)}else this._addToParent(n),this._elementStack.push(n)},t.prototype._getElementFullName=function(t,e,n){return null==t&&null==(t=this.getTagDefinition(e).implicitNamespacePrefix)&&null!=n&&(t=se(n.name)),ae(t,e)},t}();function Er(t,e){return t.length>0&&t[t.length-1]===e}function Sr(t){return t.id||function(t){var e,n,r=X(t),i=function(t,e){for(var n=Array(t.length+3>>>2),r=0;r<n.length;r++)n[r]=Hr(t,4*r,e);return n}(r,Lr.Big),o=8*r.length,s=new Array(80),a=[1732584193,4023233417,2562383102,271733878,3285377520],c=a[0],u=a[1],l=a[2],p=a[3],d=a[4];i[o>>5]|=128<<24-o%32,i[15+(o+64>>9<<4)]=o;for(var h=0;h<i.length;h+=16){for(var f=[c,u,l,p,d],m=f[0],g=f[1],v=f[2],y=f[3],b=f[4],_=0;_<80;_++){s[_]=_<16?i[h+_]:Br(s[_-3]^s[_-8]^s[_-14]^s[_-16],1);var j=Dr(_,u,l,p),x=j[0],w=j[1],O=[Br(c,5),x,d,w,s[_]].reduce(Fr);e=[p,l,Br(u,30),c,O],d=e[0],p=e[1],l=e[2],u=e[3],c=e[4]}n=[Fr(c,m),Fr(u,g),Fr(l,v),Fr(p,y),Fr(d,b)],c=n[0],u=n[1],l=n[2],p=n[3],d=n[4]}return function(t){for(var e="",n=0;n<t.length;n++){var r=qr(t,n);e+=(r>>>4).toString(16)+(15&r).toString(16)}return e.toLowerCase()}(Ur([c,u,l,p,d]))}((e=t.nodes,e.map(function(t){return t.visit(Ar,null)})).join("")+"["+t.meaning+"]");var e}function Pr(t){if(t.id)return t.id;var e=new Tr;return function(t,e){var n,r=Ir(t),i=r[0],o=r[1];if(e){var s=Ir(e),a=s[0],c=s[1];b=1,_=(y=[i,o])[0],j=y[1],l=[a,c],p=(u=[_<<b|j>>>32-b,j<<b|_>>>32-b])[0],d=u[1],h=l[0],f=l[1],m=Vr(d,f),g=m[0],v=m[1],n=[Fr(Fr(p,h),g),v],i=n[0],o=n[1]}var u,l,p,d,h,f,m,g,v;var y,b,_,j;return function(t){for(var e="",n="1",r=t.length-1;r>=0;r--)e=Wr(e,Gr(qr(t,r),n)),n=Gr(256,n);return e.split("").reverse().join("")}(Ur([2147483647&i,o]))}(t.nodes.map(function(t){return t.visit(e,null)}).join(""),t.meaning)}var Mr=function(){function t(){}return t.prototype.visitText=function(t,e){return t.value},t.prototype.visitContainer=function(t,e){var n=this;return"["+t.children.map(function(t){return t.visit(n)}).join(", ")+"]"},t.prototype.visitIcu=function(t,e){var n=this,r=Object.keys(t.cases).map(function(e){return e+" {"+t.cases[e].visit(n)+"}"});return"{"+t.expression+", "+t.type+", "+r.join(", ")+"}"},t.prototype.visitTagPlaceholder=function(t,e){var n=this;return t.isVoid?'<ph tag name="'+t.startName+'"/>':'<ph tag name="'+t.startName+'">'+t.children.map(function(t){return t.visit(n)}).join(", ")+'</ph name="'+t.closeName+'">'},t.prototype.visitPlaceholder=function(t,e){return t.value?'<ph name="'+t.name+'">'+t.value+"</ph>":'<ph name="'+t.name+'"/>'},t.prototype.visitIcuPlaceholder=function(t,e){return'<ph icu name="'+t.name+'">'+t.value.visit(this)+"</ph>"},t}(),Ar=new Mr;var Tr=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.visitIcu=function(t,e){var n=this,r=Object.keys(t.cases).map(function(e){return e+" {"+t.cases[e].visit(n)+"}"});return"{"+t.type+", "+r.join(", ")+"}"},e}(Mr);function Dr(t,e,n,r){return t<20?[e&n|~e&r,1518500249]:t<40?[e^n^r,1859775393]:t<60?[e&n|e&r|n&r,2400959708]:[e^n^r,3395469782]}function Ir(t){var e=X(t),n=[Rr(e,0),Rr(e,102072)],r=n[0],i=n[1];return 0!=r||0!=i&&1!=i||(r^=319790063,i^=-1801410264),[r,i]}function Rr(t,e){var n,r,i=[2654435769,2654435769],o=i[0],s=i[1],a=t.length;for(n=0;n+12<=a;n+=12)o=(r=Nr([o=Fr(o,Hr(t,n,Lr.Little)),s=Fr(s,Hr(t,n+4,Lr.Little)),e=Fr(e,Hr(t,n+8,Lr.Little))]))[0],s=r[1],e=r[2];return Nr([o=Fr(o,Hr(t,n,Lr.Little)),s=Fr(s,Hr(t,n+4,Lr.Little)),e=Fr(e=Fr(e,a),Hr(t,n+8,Lr.Little)<<8)])[2]}function Nr(t){var e=t[0],n=t[1],r=t[2];return e=zr(e=zr(e,n),r),e^=r>>>13,n=zr(n=zr(n,r),e),n^=e<<8,r=zr(r=zr(r,e),n),r^=n>>>13,e=zr(e=zr(e,n),r),e^=r>>>12,n=zr(n=zr(n,r),e),n^=e<<16,r=zr(r=zr(r,e),n),r^=n>>>5,e=zr(e=zr(e,n),r),e^=r>>>3,n=zr(n=zr(n,r),e),n^=e<<10,r=zr(r=zr(r,e),n),[e,n,r^=n>>>15]}var Lr={Little:0,Big:1};function Fr(t,e){return Vr(t,e)[1]}function Vr(t,e){var n=(65535&t)+(65535&e),r=(t>>>16)+(e>>>16)+(n>>>16);return[r>>>16,r<<16|65535&n]}function zr(t,e){var n=(65535&t)-(65535&e);return(t>>16)-(e>>16)+(n>>16)<<16|65535&n}function Br(t,e){return t<<e|t>>>32-e}function qr(t,e){return e>=t.length?0:255&t.charCodeAt(e)}function Hr(t,e,n){var r=0;if(n===Lr.Big)for(var i=0;i<4;i++)r+=qr(t,e+i)<<24-8*i;else for(i=0;i<4;i++)r+=qr(t,e+i)<<8*i;return r}function Ur(t){return t.reduce(function(t,e){return t+function(t){for(var e="",n=0;n<4;n++)e+=String.fromCharCode(t>>>8*(3-n)&255);return e}(e)},"")}function Wr(t,e){for(var n="",r=Math.max(t.length,e.length),i=0,o=0;i<r||o;i++){var s=o+ +(t[i]||0)+ +(e[i]||0);s>=10?(o=1,n+=s-10):(o=0,n+=s)}return n}function Gr(t,e){for(var n="",r=e;0!==t;t>>>=1)1&t&&(n=Wr(n,r)),r=Wr(r,r);return n}Lr[Lr.Little]="Little",Lr[Lr.Big]="Big";var Xr=function(){return function(t,e,n,r,i,o){this.nodes=t,this.placeholders=e,this.placeholderToMessage=n,this.meaning=r,this.description=i,this.id=o,t.length?this.sources=[{filePath:t[0].sourceSpan.start.file.url,startLine:t[0].sourceSpan.start.line+1,startCol:t[0].sourceSpan.start.col+1,endLine:t[t.length-1].sourceSpan.end.line+1,endCol:t[0].sourceSpan.start.col+1}]:this.sources=[]}}(),Yr=function(){function t(t,e){this.value=t,this.sourceSpan=e}return t.prototype.visit=function(t,e){return t.visitText(this,e)},t}(),Kr=function(){function t(t,e){this.children=t,this.sourceSpan=e}return t.prototype.visit=function(t,e){return t.visitContainer(this,e)},t}(),$r=function(){function t(t,e,n,r){this.expression=t,this.type=e,this.cases=n,this.sourceSpan=r}return t.prototype.visit=function(t,e){return t.visitIcu(this,e)},t}(),Qr=function(){function t(t,e,n,r,i,o,s){this.tag=t,this.attrs=e,this.startName=n,this.closeName=r,this.children=i,this.isVoid=o,this.sourceSpan=s}return t.prototype.visit=function(t,e){return t.visitTagPlaceholder(this,e)},t}(),Zr=function(){function t(t,e,n){this.value=t,this.name=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitPlaceholder(this,e)},t}(),Jr=function(){function t(t,e,n){this.value=t,this.name=e,this.sourceSpan=n}return t.prototype.visit=function(t,e){return t.visitIcuPlaceholder(this,e)},t}(),ti=function(){function t(){}return t.prototype.visitText=function(t,e){return new Yr(t.value,t.sourceSpan)},t.prototype.visitContainer=function(t,e){var n=this,r=t.children.map(function(t){return t.visit(n,e)});return new Kr(r,t.sourceSpan)},t.prototype.visitIcu=function(t,e){var n=this,r={};Object.keys(t.cases).forEach(function(i){return r[i]=t.cases[i].visit(n,e)});var i=new $r(t.expression,t.type,r,t.sourceSpan);return i.expressionPlaceholder=t.expressionPlaceholder,i},t.prototype.visitTagPlaceholder=function(t,e){var n=this,r=t.children.map(function(t){return t.visit(n,e)});return new Qr(t.tag,t.attrs,t.startName,t.closeName,r,t.isVoid,t.sourceSpan)},t.prototype.visitPlaceholder=function(t,e){return new Zr(t.value,t.name,t.sourceSpan)},t.prototype.visitIcuPlaceholder=function(t,e){return new Jr(t.value,t.name,t.sourceSpan)},t}(),ei=function(){function t(){}return t.prototype.visitText=function(t,e){},t.prototype.visitContainer=function(t,e){var n=this;t.children.forEach(function(t){return t.visit(n)})},t.prototype.visitIcu=function(t,e){var n=this;Object.keys(t.cases).forEach(function(e){t.cases[e].visit(n)})},t.prototype.visitTagPlaceholder=function(t,e){var n=this;t.children.forEach(function(t){return t.visit(n)})},t.prototype.visitPlaceholder=function(t,e){},t.prototype.visitIcuPlaceholder=function(t,e){},t}(),ni=function(){function t(t){var e=void 0===t?{}:t,n=e.closedByChildren,r=e.requiredParents,i=e.implicitNamespacePrefix,o=e.contentType,s=void 0===o?ne.PARSABLE_DATA:o,a=e.closedByParent,c=void 0!==a&&a,u=e.isVoid,l=void 0!==u&&u,p=e.ignoreFirstLf,d=void 0!==p&&p,h=this;this.closedByChildren={},this.closedByParent=!1,this.canSelfClose=!1,n&&n.length>0&&n.forEach(function(t){return h.closedByChildren[t]=!0}),this.isVoid=l,this.closedByParent=c||l,r&&r.length>0&&(this.requiredParents={},this.parentToAdd=r[0],r.forEach(function(t){return h.requiredParents[t]=!0})),this.implicitNamespacePrefix=i||null,this.contentType=s,this.ignoreFirstLf=d}return t.prototype.requireExtraParent=function(t){if(!this.requiredParents)return!1;if(!t)return!0;var e=t.toLowerCase();return!("template"===e||"ng-template"===t)&&1!=this.requiredParents[e]},t.prototype.isClosedByChild=function(t){return this.isVoid||t.toLowerCase()in this.closedByChildren},t}(),ri={base:new ni({isVoid:!0}),meta:new ni({isVoid:!0}),area:new ni({isVoid:!0}),embed:new ni({isVoid:!0}),link:new ni({isVoid:!0}),img:new ni({isVoid:!0}),input:new ni({isVoid:!0}),param:new ni({isVoid:!0}),hr:new ni({isVoid:!0}),br:new ni({isVoid:!0}),source:new ni({isVoid:!0}),track:new ni({isVoid:!0}),wbr:new ni({isVoid:!0}),p:new ni({closedByChildren:["address","article","aside","blockquote","div","dl","fieldset","footer","form","h1","h2","h3","h4","h5","h6","header","hgroup","hr","main","nav","ol","p","pre","section","table","ul"],closedByParent:!0}),thead:new ni({closedByChildren:["tbody","tfoot"]}),tbody:new ni({closedByChildren:["tbody","tfoot"],closedByParent:!0}),tfoot:new ni({closedByChildren:["tbody"],closedByParent:!0}),tr:new ni({closedByChildren:["tr"],requiredParents:["tbody","tfoot","thead"],closedByParent:!0}),td:new ni({closedByChildren:["td","th"],closedByParent:!0}),th:new ni({closedByChildren:["td","th"],closedByParent:!0}),col:new ni({requiredParents:["colgroup"],isVoid:!0}),svg:new ni({implicitNamespacePrefix:"svg"}),math:new ni({implicitNamespacePrefix:"math"}),li:new ni({closedByChildren:["li"],closedByParent:!0}),dt:new ni({closedByChildren:["dt","dd"]}),dd:new ni({closedByChildren:["dt","dd"],closedByParent:!0}),rb:new ni({closedByChildren:["rb","rt","rtc","rp"],closedByParent:!0}),rt:new ni({closedByChildren:["rb","rt","rtc","rp"],closedByParent:!0}),rtc:new ni({closedByChildren:["rb","rtc","rp"],closedByParent:!0}),rp:new ni({closedByChildren:["rb","rt","rtc","rp"],closedByParent:!0}),optgroup:new ni({closedByChildren:["optgroup"],closedByParent:!0}),option:new ni({closedByChildren:["option","optgroup"],closedByParent:!0}),pre:new ni({ignoreFirstLf:!0}),listing:new ni({ignoreFirstLf:!0}),style:new ni({contentType:ne.RAW_TEXT}),script:new ni({contentType:ne.RAW_TEXT}),title:new ni({contentType:ne.ESCAPABLE_RAW_TEXT}),textarea:new ni({contentType:ne.ESCAPABLE_RAW_TEXT,ignoreFirstLf:!0})},ii=new ni;function oi(t){return ri[t.toLowerCase()]||ii}var si={A:"LINK",B:"BOLD_TEXT",BR:"LINE_BREAK",EM:"EMPHASISED_TEXT",H1:"HEADING_LEVEL1",H2:"HEADING_LEVEL2",H3:"HEADING_LEVEL3",H4:"HEADING_LEVEL4",H5:"HEADING_LEVEL5",H6:"HEADING_LEVEL6",HR:"HORIZONTAL_RULE",I:"ITALIC_TEXT",LI:"LIST_ITEM",LINK:"MEDIA_LINK",OL:"ORDERED_LIST",P:"PARAGRAPH",Q:"QUOTATION",S:"STRIKETHROUGH_TEXT",SMALL:"SMALL_TEXT",SUB:"SUBSTRIPT",SUP:"SUPERSCRIPT",TBODY:"TABLE_BODY",TD:"TABLE_CELL",TFOOT:"TABLE_FOOTER",TH:"TABLE_HEADER_CELL",THEAD:"TABLE_HEADER",TR:"TABLE_ROW",TT:"MONOSPACED_TEXT",U:"UNDERLINED_TEXT",UL:"UNORDERED_LIST"},ai=function(){function t(){this._placeHolderNameCounts={},this._signatureToName={}}return t.prototype.getStartTagPlaceholderName=function(t,e,n){var r=this._hashTag(t,e,n);if(this._signatureToName[r])return this._signatureToName[r];var i=t.toUpperCase(),o=si[i]||"TAG_"+i,s=this._generateUniqueName(n?o:"START_"+o);return this._signatureToName[r]=s,s},t.prototype.getCloseTagPlaceholderName=function(t){var e=this._hashClosingTag(t);if(this._signatureToName[e])return this._signatureToName[e];var n=t.toUpperCase(),r=si[n]||"TAG_"+n,i=this._generateUniqueName("CLOSE_"+r);return this._signatureToName[e]=i,i},t.prototype.getPlaceholderName=function(t,e){var n=t.toUpperCase(),r="PH: "+n+"="+e;if(this._signatureToName[r])return this._signatureToName[r];var i=this._generateUniqueName(n);return this._signatureToName[r]=i,i},t.prototype.getUniquePlaceholder=function(t){return this._generateUniqueName(t.toUpperCase())},t.prototype._hashTag=function(t,e,n){return"<"+t+Object.keys(e).sort().map(function(t){return" "+t+"="+e[t]}).join("")+(n?"/>":"></"+t+">")},t.prototype._hashClosingTag=function(t){return this._hashTag("/"+t,{},!1)},t.prototype._generateUniqueName=function(t){if(!this._placeHolderNameCounts.hasOwnProperty(t))return this._placeHolderNameCounts[t]=1,t;var e=this._placeHolderNameCounts[t];return this._placeHolderNameCounts[t]=e+1,t+"_"+e},t}(),ci=new tr(new fn);var ui=function(){function t(t,e){this._expressionParser=t,this._interpolationConfig=e}return t.prototype.toI18nMessage=function(t,e,n,r){this._isIcu=1==t.length&&t[0]instanceof zt,this._icuDepth=0,this._placeholderRegistry=new ai,this._placeholderToContent={},this._placeholderToMessage={};var i=Wt(this,t,{});return new Xr(i,this._placeholderToContent,this._placeholderToMessage,e,n,r)},t.prototype.visitElement=function(t,e){var n=Wt(this,t.children),r={};t.attrs.forEach(function(t){r[t.name]=t.value});var i=oi(t.name).isVoid,o=this._placeholderRegistry.getStartTagPlaceholderName(t.name,r,i);this._placeholderToContent[o]=t.sourceSpan.toString();var s="";return i||(s=this._placeholderRegistry.getCloseTagPlaceholderName(t.name),this._placeholderToContent[s]="</"+t.name+">"),new Qr(t.name,r,o,s,n,i,t.sourceSpan)},t.prototype.visitAttribute=function(t,e){return this._visitTextWithInterpolation(t.value,t.sourceSpan)},t.prototype.visitText=function(t,e){return this._visitTextWithInterpolation(t.value,t.sourceSpan)},t.prototype.visitComment=function(t,e){return null},t.prototype.visitExpansion=function(e,n){var r=this;this._icuDepth++;var i={},o=new $r(e.switchValue,e.type,i,e.sourceSpan);if(e.cases.forEach(function(t){i[t.value]=new Kr(t.expression.map(function(t){return t.visit(r,{})}),t.expSourceSpan)}),this._icuDepth--,this._isIcu||this._icuDepth>0){var s=this._placeholderRegistry.getUniquePlaceholder("VAR_"+e.type);return o.expressionPlaceholder=s,this._placeholderToContent[s]=e.switchValue,o}var a=this._placeholderRegistry.getPlaceholderName("ICU",e.sourceSpan.toString()),c=new t(this._expressionParser,this._interpolationConfig);return this._placeholderToMessage[a]=c.toI18nMessage([e],"","",""),new Jr(o,a,e.sourceSpan)},t.prototype.visitExpansionCase=function(t,e){throw new Error("Unreachable code")},t.prototype._visitTextWithInterpolation=function(t,e){var n=this._expressionParser.splitInterpolation(t,e.start.toString(),this._interpolationConfig);if(!n)return new Yr(t,e);for(var r=[],i=new Kr(r,e),o=this._interpolationConfig,s=o.start,a=o.end,c=0;c<n.strings.length-1;c++){var u=n.expressions[c],l=u.split(li)[2]||"INTERPOLATION",p=this._placeholderRegistry.getPlaceholderName(l,u);n.strings[c].length&&r.push(new Yr(n.strings[c],e)),r.push(new Zr(u,p,e)),this._placeholderToContent[p]=s+u+a}var d=n.strings.length-1;return n.strings[d].length&&r.push(new Yr(n.strings[d],e)),i},t}(),li=/\/\/[\s\S]*i18n[\s\S]*\([\s\S]*ph[\s\S]*=[\s\S]*("|')([\s\S]*?)\1[\s\S]*\)/g;var pi=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.b)(e,t),e}(ar),di="i18n",hi=/^i18n:?/,fi="|",mi="@@",gi=!1;var vi=function(){return function(t,e){this.messages=t,this.errors=e}}(),yi={Extract:0,Merge:1};yi[yi.Extract]="Extract",yi[yi.Merge]="Merge";var bi=function(){function t(t,e){this._implicitTags=t,this._implicitAttrs=e}return t.prototype.extract=function(t,e){var n=this;return this._init(yi.Extract,e),t.forEach(function(t){return t.visit(n,null)}),this._inI18nBlock&&this._reportError(t[t.length-1],"Unclosed block"),new vi(this._messages,this._errors)},t.prototype.merge=function(t,e,n){this._init(yi.Merge,n),this._translations=e;var r=new Ht("wrapper",[],t,void 0,void 0,void 0).visit(this,null);return this._inI18nBlock&&this._reportError(t[t.length-1],"Unclosed block"),new Or(r.children,this._errors)},t.prototype.visitExpansionCase=function(t,e){var n=Wt(this,t.expression,e);if(this._mode===yi.Merge)return new Bt(t.value,n,t.sourceSpan,t.valueSourceSpan,t.expSourceSpan)},t.prototype.visitExpansion=function(t,e){this._mayBeAddBlockChildren(t);var n=this._inIcu;this._inIcu||(this._isInTranslatableSection&&this._addMessage([t]),this._inIcu=!0);var r=Wt(this,t.cases,e);return this._mode===yi.Merge&&(t=new zt(t.switchValue,t.type,r,t.sourceSpan,t.switchValueSourceSpan)),this._inIcu=n,t},t.prototype.visitComment=function(t,e){var n,r=!!((n=t)instanceof Ut&&n.value&&n.value.startsWith("i18n"));if(r&&this._isInTranslatableSection)this._reportError(t,"Could not start a block inside a translatable section");else{var i,o=!!((i=t)instanceof Ut&&i.value&&"/i18n"===i.value);if(!o||this._inI18nBlock){if(!this._inI18nNode&&!this._inIcu)if(this._inI18nBlock){if(o){if(this._depth==this._blockStartDepth){this._closeTranslatableSection(t,this._blockChildren),this._inI18nBlock=!1;var s=this._addMessage(this._blockChildren,this._blockMeaningAndDesc);return Wt(this,this._translateMessage(t,s))}return void this._reportError(t,"I18N blocks should not cross element boundaries")}}else if(r){if(!gi&&console&&console.warn){gi=!0;var a=t.sourceSpan.details?", "+t.sourceSpan.details:"";console.warn("I18n comments are deprecated, use an <ng-container> element instead ("+t.sourceSpan.start+a+")")}this._inI18nBlock=!0,this._blockStartDepth=this._depth,this._blockChildren=[],this._blockMeaningAndDesc=t.value.replace(hi,"").trim(),this._openTranslatableSection(t)}}else this._reportError(t,"Trying to close an unopened block")}},t.prototype.visitText=function(t,e){return this._isInTranslatableSection&&this._mayBeAddBlockChildren(t),t},t.prototype.visitElement=function(t,e){var n=this;this._mayBeAddBlockChildren(t),this._depth++;var r=this._inI18nNode,i=this._inImplicitNode,o=[],s=void 0,a=t.attrs.find(function(t){return t.name===di})||null,c=a?a.value:"",u=this._implicitTags.some(function(e){return t.name===e})&&!this._inIcu&&!this._isInTranslatableSection,l=!i&&u;if(this._inImplicitNode=i||u,this._isInTranslatableSection||this._inIcu)(a||l)&&this._reportError(t,"Could not mark an element as translatable inside a translatable section"),this._mode==yi.Extract&&Wt(this,t.children);else{if(a||l){this._inI18nNode=!0;var p=this._addMessage(t.children,c);s=this._translateMessage(t,p)}if(this._mode==yi.Extract){var d=a||l;d&&this._openTranslatableSection(t),Wt(this,t.children),d&&this._closeTranslatableSection(t,t.children)}}this._mode===yi.Merge&&(s||t.children).forEach(function(t){var r=t.visit(n,e);r&&!n._isInTranslatableSection&&(o=o.concat(r))});if(this._visitAttributesOf(t),this._depth--,this._inI18nNode=r,this._inImplicitNode=i,this._mode===yi.Merge){var h=this._translateAttributes(t);return new Ht(t.name,h,o,t.sourceSpan,t.startSourceSpan,t.endSourceSpan)}return null},t.prototype.visitAttribute=function(t,e){throw new Error("unreachable code")},t.prototype._init=function(t,e){var n;this._mode=t,this._inI18nBlock=!1,this._inI18nNode=!1,this._depth=0,this._inIcu=!1,this._msgCountAtSectionStart=void 0,this._errors=[],this._messages=[],this._inImplicitNode=!1,this._createI18nMessage=(n=new ui(ci,e),function(t,e,r,i){return n.toI18nMessage(t,e,r,i)})},t.prototype._visitAttributesOf=function(t){var e=this,n={},r=this._implicitAttrs[t.name]||[];t.attrs.filter(function(t){return t.name.startsWith("i18n-")}).forEach(function(t){return n[t.name.slice("i18n-".length)]=t.value}),t.attrs.forEach(function(t){t.name in n?e._addMessage([t],n[t.name]):r.some(function(e){return t.name===e})&&e._addMessage([t])})},t.prototype._addMessage=function(t,e){if(0==t.length||1==t.length&&t[0]instanceof qt&&!t[0].value)return null;var n=_i(e),r=n.meaning,i=n.description,o=n.id,s=this._createI18nMessage(t,r,i,o);return this._messages.push(s),s},t.prototype._translateMessage=function(t,e){if(e&&this._mode===yi.Merge){var n=this._translations.get(e);if(n)return n;this._reportError(t,'Translation unavailable for message id="'+this._translations.digest(e)+'"')}return[]},t.prototype._translateAttributes=function(t){var e=this,n=t.attrs,r={};n.forEach(function(t){t.name.startsWith("i18n-")&&(r[t.name.slice("i18n-".length)]=_i(t.value))});var i=[];return n.forEach(function(n){if(n.name!==di&&!n.name.startsWith("i18n-"))if(n.value&&""!=n.value&&r.hasOwnProperty(n.name)){var o=r[n.name],s=o.meaning,a=o.description,c=o.id,u=e._createI18nMessage([n],s,a,c),l=e._translations.get(u);if(l)if(0==l.length)i.push(new qt(n.name,"",n.sourceSpan));else if(l[0]instanceof Vt){var p=l[0].value;i.push(new qt(n.name,p,n.sourceSpan))}else e._reportError(t,'Unexpected translation for attribute "'+n.name+'" (id="'+(c||e._translations.digest(u))+'")');else e._reportError(t,'Translation unavailable for attribute "'+n.name+'" (id="'+(c||e._translations.digest(u))+'")')}else i.push(n)}),i},t.prototype._mayBeAddBlockChildren=function(t){this._inI18nBlock&&!this._inIcu&&this._depth==this._blockStartDepth&&this._blockChildren.push(t)},t.prototype._openTranslatableSection=function(t){this._isInTranslatableSection?this._reportError(t,"Unexpected section start"):this._msgCountAtSectionStart=this._messages.length},Object.defineProperty(t.prototype,"_isInTranslatableSection",{get:function(){return void 0!==this._msgCountAtSectionStart},enumerable:!0,configurable:!0}),t.prototype._closeTranslatableSection=function(t,e){if(this._isInTranslatableSection){var n=this._msgCountAtSectionStart;if(1==e.reduce(function(t,e){return t+(e instanceof Ut?0:1)},0))for(var r=this._messages.length-1;r>=n;r--){var i=this._messages[r].nodes;if(!(1==i.length&&i[0]instanceof Yr)){this._messages.splice(r,1);break}}this._msgCountAtSectionStart=void 0}else this._reportError(t,"Unexpected section end")},t.prototype._reportError=function(t,e){this._errors.push(new pi(t.sourceSpan,e))},t}();function _i(t){if(!t)return{meaning:"",description:"",id:""};var e=t.indexOf(mi),n=t.indexOf(fi),r=e>-1?[t.slice(0,e),t.slice(e+2)]:[t,""],i=r[0],o=r[1],s=n>-1?[i.slice(0,n),i.slice(n+1)]:["",i];return{meaning:s[0],description:s[1],id:o}}var ji=new(function(){function t(){this.closedByParent=!1,this.contentType=ne.PARSABLE_DATA,this.isVoid=!1,this.ignoreFirstLf=!1,this.canSelfClose=!0}return t.prototype.requireExtraParent=function(t){return!1},t.prototype.isClosedByChild=function(t){return!1},t}());function xi(t){return ji}var wi=function(t){function e(){return t.call(this,xi)||this}return Object(r.b)(e,t),e.prototype.parse=function(e,n,r){return void 0===r&&(r=!1),t.prototype.parse.call(this,e,n,r)},e}(Cr),Oi=function(){function t(){}return t.prototype.createNameMapper=function(t){return null},t}(),Ci=function(t){function e(e,n){var r=t.call(this)||this;return r.mapName=n,r.internalToPublic={},r.publicToNextId={},r.publicToInternal={},e.nodes.forEach(function(t){return t.visit(r)}),r}return Object(r.b)(e,t),e.prototype.toPublicName=function(t){return this.internalToPublic.hasOwnProperty(t)?this.internalToPublic[t]:null},e.prototype.toInternalName=function(t){return this.publicToInternal.hasOwnProperty(t)?this.publicToInternal[t]:null},e.prototype.visitText=function(t,e){return null},e.prototype.visitTagPlaceholder=function(e,n){this.visitPlaceholderName(e.startName),t.prototype.visitTagPlaceholder.call(this,e,n),this.visitPlaceholderName(e.closeName)},e.prototype.visitPlaceholder=function(t,e){this.visitPlaceholderName(t.name)},e.prototype.visitIcuPlaceholder=function(t,e){this.visitPlaceholderName(t.name)},e.prototype.visitPlaceholderName=function(t){if(t&&!this.internalToPublic.hasOwnProperty(t)){var e=this.mapName(t);if(this.publicToInternal.hasOwnProperty(e)){var n=this.publicToNextId[e];this.publicToNextId[e]=n+1,e=e+"_"+n}else this.publicToNextId[e]=1;this.internalToPublic[t]=e,this.publicToInternal[e]=t}},e}(ei),ki=new(function(){function t(){}return t.prototype.visitTag=function(t){var e=this,n=this._serializeAttributes(t.attrs);if(0==t.children.length)return"<"+t.name+n+"/>";var r=t.children.map(function(t){return t.visit(e)});return"<"+t.name+n+">"+r.join("")+"</"+t.name+">"},t.prototype.visitText=function(t){return t.value},t.prototype.visitDeclaration=function(t){return"<?xml"+this._serializeAttributes(t.attrs)+" ?>"},t.prototype._serializeAttributes=function(t){var e=Object.keys(t).map(function(e){return e+'="'+t[e]+'"'}).join(" ");return e.length>0?" "+e:""},t.prototype.visitDoctype=function(t){return"<!DOCTYPE "+t.rootTag+" [\n"+t.dtd+"\n]>"},t}());function Ei(t){return t.map(function(t){return t.visit(ki)}).join("")}var Si=function(){function t(t){var e=this;this.attrs={},Object.keys(t).forEach(function(n){e.attrs[n]=Ii(t[n])})}return t.prototype.visit=function(t){return t.visitDeclaration(this)},t}(),Pi=function(){function t(t,e){this.rootTag=t,this.dtd=e}return t.prototype.visit=function(t){return t.visitDoctype(this)},t}(),Mi=function(){function t(t,e,n){void 0===e&&(e={}),void 0===n&&(n=[]);var r=this;this.name=t,this.children=n,this.attrs={},Object.keys(e).forEach(function(t){r.attrs[t]=Ii(e[t])})}return t.prototype.visit=function(t){return t.visitTag(this)},t}(),Ai=function(){function t(t){this.value=Ii(t)}return t.prototype.visit=function(t){return t.visitText(this)},t}(),Ti=function(t){function e(e){return void 0===e&&(e=0),t.call(this,"\n"+new Array(e+1).join(" "))||this}return Object(r.b)(e,t),e}(Ai),Di=[[/&/g,"&amp;"],[/"/g,"&quot;"],[/'/g,"&apos;"],[/</g,"&lt;"],[/>/g,"&gt;"]];function Ii(t){return Di.reduce(function(t,e){return t.replace(e[0],e[1])},t)}var Ri=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.write=function(t,e){var n=new Ni,r=[];t.forEach(function(t){var e=[];t.sources.forEach(function(t){var n=new Mi("context-group",{purpose:"location"});n.children.push(new Ti(10),new Mi("context",{"context-type":"sourcefile"},[new Ai(t.filePath)]),new Ti(10),new Mi("context",{"context-type":"linenumber"},[new Ai(""+t.startLine)]),new Ti(8)),e.push(new Ti(8),n)});var i,o=new Mi("trans-unit",{id:t.id,datatype:"html"});(i=o.children).push.apply(i,[new Ti(8),new Mi("source",{},n.serialize(t.nodes))].concat(e)),t.description&&o.children.push(new Ti(8),new Mi("note",{priority:"1",from:"description"},[new Ai(t.description)])),t.meaning&&o.children.push(new Ti(8),new Mi("note",{priority:"1",from:"meaning"},[new Ai(t.meaning)])),o.children.push(new Ti(6)),r.push(new Ti(6),o)});var i=new Mi("body",{},r.concat([new Ti(4)])),o=new Mi("file",{"source-language":e||"en",datatype:"plaintext",original:"ng2.template"},[new Ti(4),i,new Ti(2)]),s=new Mi("xliff",{version:"1.2",xmlns:"urn:oasis:names:tc:xliff:document:1.2"},[new Ti(2),o,new Ti]);return Ei([new Si({version:"1.0",encoding:"UTF-8"}),new Ti,s,new Ti])},e.prototype.load=function(t,e){var n=(new Li).parse(t,e),r=n.locale,i=n.msgIdToHtml,o=n.errors,s={},a=new Fi;if(Object.keys(i).forEach(function(t){var n=a.convert(i[t],e),r=n.i18nNodes,c=n.errors;o.push.apply(o,c),s[t]=r}),o.length)throw new Error("xliff parse errors:\n"+o.join("\n"));return{locale:r,i18nNodesByMsgId:s}},e.prototype.digest=function(t){return Sr(t)},e}(Oi),Ni=function(){function t(){}return t.prototype.visitText=function(t,e){return[new Ai(t.value)]},t.prototype.visitContainer=function(t,e){var n=this,r=[];return t.children.forEach(function(t){return r.push.apply(r,t.visit(n))}),r},t.prototype.visitIcu=function(t,e){var n=this,r=[new Ai("{"+t.expressionPlaceholder+", "+t.type+", ")];return Object.keys(t.cases).forEach(function(e){r.push.apply(r,[new Ai(e+" {")].concat(t.cases[e].visit(n),[new Ai("} ")]))}),r.push(new Ai("}")),r},t.prototype.visitTagPlaceholder=function(t,e){var n=function(t){switch(t.toLowerCase()){case"br":return"lb";case"img":return"image";default:return"x-"+t}}(t.tag);if(t.isVoid)return[new Mi("x",{id:t.startName,ctype:n,"equiv-text":"<"+t.tag+"/>"})];var r=new Mi("x",{id:t.startName,ctype:n,"equiv-text":"<"+t.tag+">"}),i=new Mi("x",{id:t.closeName,ctype:n,"equiv-text":"</"+t.tag+">"});return[r].concat(this.serialize(t.children),[i])},t.prototype.visitPlaceholder=function(t,e){return[new Mi("x",{id:t.name,"equiv-text":"{{"+t.value+"}}"})]},t.prototype.visitIcuPlaceholder=function(t,e){var n="{"+t.value.expression+", "+t.value.type+", "+Object.keys(t.value.cases).map(function(t){return t+" {...}"}).join(" ")+"}";return[new Mi("x",{id:t.name,"equiv-text":n})]},t.prototype.serialize=function(t){var e=this;return[].concat.apply([],t.map(function(t){return t.visit(e)}))},t}(),Li=function(){function t(){this._locale=null}return t.prototype.parse=function(t,e){this._unitMlString=null,this._msgIdToHtml={};var n=(new wi).parse(t,e,!1);return this._errors=n.errors,Wt(this,n.rootNodes,null),{msgIdToHtml:this._msgIdToHtml,errors:this._errors,locale:this._locale}},t.prototype.visitElement=function(t,e){switch(t.name){case"trans-unit":this._unitMlString=null;var n=t.attrs.find(function(t){return"id"===t.name});if(n){var r=n.value;this._msgIdToHtml.hasOwnProperty(r)?this._addError(t,"Duplicated translations for msg "+r):(Wt(this,t.children,null),"string"==typeof this._unitMlString?this._msgIdToHtml[r]=this._unitMlString:this._addError(t,"Message "+r+" misses a translation"))}else this._addError(t,'<trans-unit> misses the "id" attribute');break;case"source":break;case"target":var i=t.startSourceSpan.end.offset,o=t.endSourceSpan.start.offset,s=t.startSourceSpan.start.file.content.slice(i,o);this._unitMlString=s;break;case"file":var a=t.attrs.find(function(t){return"target-language"===t.name});a&&(this._locale=a.value),Wt(this,t.children,null);break;default:Wt(this,t.children,null)}},t.prototype.visitAttribute=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitComment=function(t,e){},t.prototype.visitExpansion=function(t,e){},t.prototype.visitExpansionCase=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new pi(t.sourceSpan,e))},t}(),Fi=function(){function t(){}return t.prototype.convert=function(t,e){var n=(new wi).parse(t,e,!0);return this._errors=n.errors,{i18nNodes:this._errors.length>0||0==n.rootNodes.length?[]:Wt(this,n.rootNodes),errors:this._errors}},t.prototype.visitText=function(t,e){return new Yr(t.value,t.sourceSpan)},t.prototype.visitElement=function(t,e){if("x"===t.name){var n=t.attrs.find(function(t){return"id"===t.name});if(n)return new Zr("",n.value,t.sourceSpan);this._addError(t,'<x> misses the "id" attribute')}else this._addError(t,"Unexpected tag");return null},t.prototype.visitExpansion=function(t,e){var n={};return Wt(this,t.cases).forEach(function(e){n[e.value]=new Kr(e.nodes,t.sourceSpan)}),new $r(t.switchValue,t.type,n,t.sourceSpan)},t.prototype.visitExpansionCase=function(t,e){return{value:t.value,nodes:Wt(this,t.expression)}},t.prototype.visitComment=function(t,e){},t.prototype.visitAttribute=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new pi(t.sourceSpan,e))},t}();var Vi=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.write=function(t,e){var n=new zi,r=[];t.forEach(function(t){var e=new Mi("unit",{id:t.id}),i=new Mi("notes");(t.description||t.meaning)&&(t.description&&i.children.push(new Ti(8),new Mi("note",{category:"description"},[new Ai(t.description)])),t.meaning&&i.children.push(new Ti(8),new Mi("note",{category:"meaning"},[new Ai(t.meaning)]))),t.sources.forEach(function(t){i.children.push(new Ti(8),new Mi("note",{category:"location"},[new Ai(t.filePath+":"+t.startLine+(t.endLine!==t.startLine?","+t.endLine:""))]))}),i.children.push(new Ti(6)),e.children.push(new Ti(6),i);var o=new Mi("segment");o.children.push(new Ti(8),new Mi("source",{},n.serialize(t.nodes)),new Ti(6)),e.children.push(new Ti(6),o,new Ti(4)),r.push(new Ti(4),e)});var i=new Mi("file",{original:"ng.template",id:"ngi18n"},r.concat([new Ti(2)])),o=new Mi("xliff",{version:"2.0",xmlns:"urn:oasis:names:tc:xliff:document:2.0",srcLang:e||"en"},[new Ti(2),i,new Ti]);return Ei([new Si({version:"1.0",encoding:"UTF-8"}),new Ti,o,new Ti])},e.prototype.load=function(t,e){var n=(new Bi).parse(t,e),r=n.locale,i=n.msgIdToHtml,o=n.errors,s={},a=new qi;if(Object.keys(i).forEach(function(t){var n=a.convert(i[t],e),r=n.i18nNodes,c=n.errors;o.push.apply(o,c),s[t]=r}),o.length)throw new Error("xliff2 parse errors:\n"+o.join("\n"));return{locale:r,i18nNodesByMsgId:s}},e.prototype.digest=function(t){return Pr(t)},e}(Oi),zi=function(){function t(){}return t.prototype.visitText=function(t,e){return[new Ai(t.value)]},t.prototype.visitContainer=function(t,e){var n=this,r=[];return t.children.forEach(function(t){return r.push.apply(r,t.visit(n))}),r},t.prototype.visitIcu=function(t,e){var n=this,r=[new Ai("{"+t.expressionPlaceholder+", "+t.type+", ")];return Object.keys(t.cases).forEach(function(e){r.push.apply(r,[new Ai(e+" {")].concat(t.cases[e].visit(n),[new Ai("} ")]))}),r.push(new Ai("}")),r},t.prototype.visitTagPlaceholder=function(t,e){var n=this,r=function(t){switch(t.toLowerCase()){case"br":case"b":case"i":case"u":return"fmt";case"img":return"image";case"a":return"link";default:return"other"}}(t.tag);if(t.isVoid)return[new Mi("ph",{id:(this._nextPlaceholderId++).toString(),equiv:t.startName,type:r,disp:"<"+t.tag+"/>"})];var i=new Mi("pc",{id:(this._nextPlaceholderId++).toString(),equivStart:t.startName,equivEnd:t.closeName,type:r,dispStart:"<"+t.tag+">",dispEnd:"</"+t.tag+">"}),o=[].concat.apply([],t.children.map(function(t){return t.visit(n)}));return o.length?o.forEach(function(t){return i.children.push(t)}):i.children.push(new Ai("")),[i]},t.prototype.visitPlaceholder=function(t,e){var n=(this._nextPlaceholderId++).toString();return[new Mi("ph",{id:n,equiv:t.name,disp:"{{"+t.value+"}}"})]},t.prototype.visitIcuPlaceholder=function(t,e){var n=Object.keys(t.value.cases).map(function(t){return t+" {...}"}).join(" "),r=(this._nextPlaceholderId++).toString();return[new Mi("ph",{id:r,equiv:t.name,disp:"{"+t.value.expression+", "+t.value.type+", "+n+"}"})]},t.prototype.serialize=function(t){var e=this;return this._nextPlaceholderId=0,[].concat.apply([],t.map(function(t){return t.visit(e)}))},t}(),Bi=function(){function t(){this._locale=null}return t.prototype.parse=function(t,e){this._unitMlString=null,this._msgIdToHtml={};var n=(new wi).parse(t,e,!1);return this._errors=n.errors,Wt(this,n.rootNodes,null),{msgIdToHtml:this._msgIdToHtml,errors:this._errors,locale:this._locale}},t.prototype.visitElement=function(t,e){switch(t.name){case"unit":this._unitMlString=null;var n=t.attrs.find(function(t){return"id"===t.name});if(n){var r=n.value;this._msgIdToHtml.hasOwnProperty(r)?this._addError(t,"Duplicated translations for msg "+r):(Wt(this,t.children,null),"string"==typeof this._unitMlString?this._msgIdToHtml[r]=this._unitMlString:this._addError(t,"Message "+r+" misses a translation"))}else this._addError(t,'<unit> misses the "id" attribute');break;case"source":break;case"target":var i=t.startSourceSpan.end.offset,o=t.endSourceSpan.start.offset,s=t.startSourceSpan.start.file.content.slice(i,o);this._unitMlString=s;break;case"xliff":var a=t.attrs.find(function(t){return"trgLang"===t.name});a&&(this._locale=a.value);var c=t.attrs.find(function(t){return"version"===t.name});if(c){var u=c.value;"2.0"!==u?this._addError(t,"The XLIFF file version "+u+" is not compatible with XLIFF 2.0 serializer"):Wt(this,t.children,null)}break;default:Wt(this,t.children,null)}},t.prototype.visitAttribute=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitComment=function(t,e){},t.prototype.visitExpansion=function(t,e){},t.prototype.visitExpansionCase=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new pi(t.sourceSpan,e))},t}(),qi=function(){function t(){}return t.prototype.convert=function(t,e){var n=(new wi).parse(t,e,!0);return this._errors=n.errors,{i18nNodes:this._errors.length>0||0==n.rootNodes.length?[]:[].concat.apply([],Wt(this,n.rootNodes)),errors:this._errors}},t.prototype.visitText=function(t,e){return new Yr(t.value,t.sourceSpan)},t.prototype.visitElement=function(t,e){var n=this;switch(t.name){case"ph":var r=t.attrs.find(function(t){return"equiv"===t.name});if(r)return[new Zr("",r.value,t.sourceSpan)];this._addError(t,'<ph> misses the "equiv" attribute');break;case"pc":var i=t.attrs.find(function(t){return"equivStart"===t.name}),o=t.attrs.find(function(t){return"equivEnd"===t.name});if(i){if(o){var s=i.value,a=o.value,c=[];return c.concat.apply(c,[new Zr("",s,t.sourceSpan)].concat(t.children.map(function(t){return t.visit(n,null)}),[new Zr("",a,t.sourceSpan)]))}this._addError(t,'<ph> misses the "equivEnd" attribute')}else this._addError(t,'<ph> misses the "equivStart" attribute');break;default:this._addError(t,"Unexpected tag")}return null},t.prototype.visitExpansion=function(t,e){var n={};return Wt(this,t.cases).forEach(function(e){n[e.value]=new Kr(e.nodes,t.sourceSpan)}),new $r(t.switchValue,t.type,n,t.sourceSpan)},t.prototype.visitExpansionCase=function(t,e){return{value:t.value,nodes:[].concat.apply([],Wt(this,t.expression))}},t.prototype.visitComment=function(t,e){},t.prototype.visitAttribute=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new pi(t.sourceSpan,e))},t}();var Hi=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.write=function(t,e){var n=new Gi,r=new Ui,i=new Mi("messagebundle");return t.forEach(function(t){var e={id:t.id};t.description&&(e.desc=t.description),t.meaning&&(e.meaning=t.meaning);var n=[];t.sources.forEach(function(t){n.push(new Mi("source",{},[new Ai(t.filePath+":"+t.startLine+(t.endLine!==t.startLine?","+t.endLine:""))]))}),i.children.push(new Ti(2),new Mi("msg",e,n.concat(r.serialize(t.nodes))))}),i.children.push(new Ti),Ei([new Si({version:"1.0",encoding:"UTF-8"}),new Ti,new Pi("messagebundle",'<!ELEMENT messagebundle (msg)*>\n<!ATTLIST messagebundle class CDATA #IMPLIED>\n\n<!ELEMENT msg (#PCDATA|ph|source)*>\n<!ATTLIST msg id CDATA #IMPLIED>\n<!ATTLIST msg seq CDATA #IMPLIED>\n<!ATTLIST msg name CDATA #IMPLIED>\n<!ATTLIST msg desc CDATA #IMPLIED>\n<!ATTLIST msg meaning CDATA #IMPLIED>\n<!ATTLIST msg obsolete (obsolete) #IMPLIED>\n<!ATTLIST msg xml:space (default|preserve) "default">\n<!ATTLIST msg is_hidden CDATA #IMPLIED>\n\n<!ELEMENT source (#PCDATA)>\n\n<!ELEMENT ph (#PCDATA|ex)*>\n<!ATTLIST ph name CDATA #REQUIRED>\n\n<!ELEMENT ex (#PCDATA)>'),new Ti,n.addDefaultExamples(i),new Ti])},e.prototype.load=function(t,e){throw new Error("Unsupported")},e.prototype.digest=function(t){return Wi(t)},e.prototype.createNameMapper=function(t){return new Ci(t,Xi)},e}(Oi),Ui=function(){function t(){}return t.prototype.visitText=function(t,e){return[new Ai(t.value)]},t.prototype.visitContainer=function(t,e){var n=this,r=[];return t.children.forEach(function(t){return r.push.apply(r,t.visit(n))}),r},t.prototype.visitIcu=function(t,e){var n=this,r=[new Ai("{"+t.expressionPlaceholder+", "+t.type+", ")];return Object.keys(t.cases).forEach(function(e){r.push.apply(r,[new Ai(e+" {")].concat(t.cases[e].visit(n),[new Ai("} ")]))}),r.push(new Ai("}")),r},t.prototype.visitTagPlaceholder=function(t,e){var n=new Mi("ex",{},[new Ai("<"+t.tag+">")]),r=new Mi("ph",{name:t.startName},[n]);if(t.isVoid)return[r];var i=new Mi("ex",{},[new Ai("</"+t.tag+">")]),o=new Mi("ph",{name:t.closeName},[i]);return[r].concat(this.serialize(t.children),[o])},t.prototype.visitPlaceholder=function(t,e){var n=new Mi("ex",{},[new Ai("{{"+t.value+"}}")]);return[new Mi("ph",{name:t.name},[n])]},t.prototype.visitIcuPlaceholder=function(t,e){var n=new Mi("ex",{},[new Ai("{"+t.value.expression+", "+t.value.type+", "+Object.keys(t.value.cases).map(function(t){return t+" {...}"}).join(" ")+"}")]);return[new Mi("ph",{name:t.name},[n])]},t.prototype.serialize=function(t){var e=this;return[].concat.apply([],t.map(function(t){return t.visit(e)}))},t}();function Wi(t){return Pr(t)}var Gi=function(){function t(){}return t.prototype.addDefaultExamples=function(t){return t.visit(this),t},t.prototype.visitTag=function(t){var e=this;if("ph"===t.name){if(!t.children||0==t.children.length){var n=new Ai(t.attrs.name||"...");t.children=[new Mi("ex",{},[n])]}}else t.children&&t.children.forEach(function(t){return t.visit(e)})},t.prototype.visitText=function(t){},t.prototype.visitDeclaration=function(t){},t.prototype.visitDoctype=function(t){},t}();function Xi(t){return t.toUpperCase().replace(/[^A-Z0-9_]/g,"_")}var Yi=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.write=function(t,e){throw new Error("Unsupported")},e.prototype.load=function(t,e){var n=(new Ki).parse(t,e),r=n.locale,i=n.msgIdToHtml,o=n.errors,s={},a=new $i;if(Object.keys(i).forEach(function(t){var n,r,o;n=s,r=t,o=function(){var n=a.convert(i[t],e),r=n.i18nNodes,o=n.errors;if(o.length)throw new Error("xtb parse errors:\n"+o.join("\n"));return r},Object.defineProperty(n,r,{configurable:!0,enumerable:!0,get:function(){var t=o();return Object.defineProperty(n,r,{enumerable:!0,value:t}),t},set:function(t){throw new Error("Could not overwrite an XTB translation")}})}),o.length)throw new Error("xtb parse errors:\n"+o.join("\n"));return{locale:r,i18nNodesByMsgId:s}},e.prototype.digest=function(t){return Wi(t)},e.prototype.createNameMapper=function(t){return new Ci(t,Xi)},e}(Oi);var Ki=function(){function t(){this._locale=null}return t.prototype.parse=function(t,e){this._bundleDepth=0,this._msgIdToHtml={};var n=(new wi).parse(t,e,!1);return this._errors=n.errors,Wt(this,n.rootNodes),{msgIdToHtml:this._msgIdToHtml,errors:this._errors,locale:this._locale}},t.prototype.visitElement=function(t,e){switch(t.name){case"translationbundle":this._bundleDepth++,this._bundleDepth>1&&this._addError(t,"<translationbundle> elements can not be nested");var n=t.attrs.find(function(t){return"lang"===t.name});n&&(this._locale=n.value),Wt(this,t.children,null),this._bundleDepth--;break;case"translation":var r=t.attrs.find(function(t){return"id"===t.name});if(r){var i=r.value;if(this._msgIdToHtml.hasOwnProperty(i))this._addError(t,"Duplicated translations for msg "+i);else{var o=t.startSourceSpan.end.offset,s=t.endSourceSpan.start.offset,a=t.startSourceSpan.start.file.content.slice(o,s);this._msgIdToHtml[i]=a}}else this._addError(t,'<translation> misses the "id" attribute');break;default:this._addError(t,"Unexpected tag")}},t.prototype.visitAttribute=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitComment=function(t,e){},t.prototype.visitExpansion=function(t,e){},t.prototype.visitExpansionCase=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new pi(t.sourceSpan,e))},t}(),$i=function(){function t(){}return t.prototype.convert=function(t,e){var n=(new wi).parse(t,e,!0);return this._errors=n.errors,{i18nNodes:this._errors.length>0||0==n.rootNodes.length?[]:Wt(this,n.rootNodes),errors:this._errors}},t.prototype.visitText=function(t,e){return new Yr(t.value,t.sourceSpan)},t.prototype.visitExpansion=function(t,e){var n={};return Wt(this,t.cases).forEach(function(e){n[e.value]=new Kr(e.nodes,t.sourceSpan)}),new $r(t.switchValue,t.type,n,t.sourceSpan)},t.prototype.visitExpansionCase=function(t,e){return{value:t.value,nodes:Wt(this,t.expression)}},t.prototype.visitElement=function(t,e){if("ph"===t.name){var n=t.attrs.find(function(t){return"name"===t.name});if(n)return new Zr("",n.value,t.sourceSpan);this._addError(t,'<ph> misses the "name" attribute')}else this._addError(t,"Unexpected tag");return null},t.prototype.visitComment=function(t,e){},t.prototype.visitAttribute=function(t,e){},t.prototype._addError=function(t,e){this._errors.push(new pi(t.sourceSpan,e))},t}(),Qi=function(t){function e(){return t.call(this,oi)||this}return Object(r.b)(e,t),e.prototype.parse=function(e,n,r,i){return void 0===r&&(r=!1),void 0===i&&(i=$t),t.prototype.parse.call(this,e,n,r,i)},e}(Cr),Zi=function(){function t(t,e,n,r,i,o){void 0===t&&(t={}),void 0===i&&(i=M.Warning),this._i18nNodesByMsgId=t,this.digest=n,this.mapperFactory=r,this._i18nToHtml=new Ji(t,e,n,r,i,o)}return t.load=function(e,n,r,i,o){var s=r.load(e,n),a=s.locale;return new t(s.i18nNodesByMsgId,a,function(t){return r.digest(t)},function(t){return r.createNameMapper(t)},i,o)},t.prototype.get=function(t){var e=this._i18nToHtml.convert(t);if(e.errors.length)throw new Error(e.errors.join("\n"));return e.nodes},t.prototype.has=function(t){return this.digest(t)in this._i18nNodesByMsgId},t}(),Ji=function(){function t(t,e,n,r,i,o){void 0===t&&(t={}),this._i18nNodesByMsgId=t,this._locale=e,this._digest=n,this._mapperFactory=r,this._missingTranslationStrategy=i,this._console=o,this._contextStack=[],this._errors=[]}return t.prototype.convert=function(t){this._contextStack.length=0,this._errors.length=0;var e=this._convertToText(t),n=t.nodes[0].sourceSpan.start.file.url,r=(new Qi).parse(e,n,!0);return{nodes:r.rootNodes,errors:this._errors.concat(r.errors)}},t.prototype.visitText=function(t,e){return t.value},t.prototype.visitContainer=function(t,e){var n=this;return t.children.map(function(t){return t.visit(n)}).join("")},t.prototype.visitIcu=function(t,e){var n=this,r=Object.keys(t.cases).map(function(e){return e+" {"+t.cases[e].visit(n)+"}"});return"{"+(this._srcMsg.placeholders.hasOwnProperty(t.expression)?this._srcMsg.placeholders[t.expression]:t.expression)+", "+t.type+", "+r.join(" ")+"}"},t.prototype.visitPlaceholder=function(t,e){var n=this._mapper(t.name);return this._srcMsg.placeholders.hasOwnProperty(n)?this._srcMsg.placeholders[n]:this._srcMsg.placeholderToMessage.hasOwnProperty(n)?this._convertToText(this._srcMsg.placeholderToMessage[n]):(this._addError(t,'Unknown placeholder "'+t.name+'"'),"")},t.prototype.visitTagPlaceholder=function(t,e){var n=this,r=""+t.tag,i=Object.keys(t.attrs).map(function(e){return e+'="'+t.attrs[e]+'"'}).join(" ");return t.isVoid?"<"+r+" "+i+"/>":"<"+r+" "+i+">"+t.children.map(function(t){return t.visit(n)}).join("")+"</"+r+">"},t.prototype.visitIcuPlaceholder=function(t,e){return this._convertToText(this._srcMsg.placeholderToMessage[t.name])},t.prototype._convertToText=function(t){var e,n=this,r=this._digest(t),i=this._mapperFactory?this._mapperFactory(t):null;if(this._contextStack.push({msg:this._srcMsg,mapper:this._mapper}),this._srcMsg=t,this._i18nNodesByMsgId.hasOwnProperty(r))e=this._i18nNodesByMsgId[r],this._mapper=function(t){return i?i.toInternalName(t):t};else{if(this._missingTranslationStrategy===M.Error){var o=this._locale?' for locale "'+this._locale+'"':"";this._addError(t.nodes[0],'Missing translation for message "'+r+'"'+o)}else if(this._console&&this._missingTranslationStrategy===M.Warning){o=this._locale?' for locale "'+this._locale+'"':"";this._console.warn('Missing translation for message "'+r+'"'+o)}e=t.nodes,this._mapper=function(t){return t}}var s=e.map(function(t){return t.visit(n)}).join(""),a=this._contextStack.pop();return this._srcMsg=a.msg,this._mapper=a.mapper,s},t.prototype._addError=function(t,e){this._errors.push(new pi(t.sourceSpan,e))},t}(),to=function(){function t(t,e,n,r,i){if(void 0===r&&(r=M.Warning),this._htmlParser=t,e){var o=function(t){switch(t=(t||"xlf").toLowerCase()){case"xmb":return new Hi;case"xtb":return new Yi;case"xliff2":case"xlf2":return new Vi;case"xliff":case"xlf":default:return new Ri}}(n);this._translationBundle=Zi.load(e,"i18n",o,r,i)}else this._translationBundle=new Zi({},null,Sr,void 0,r,i)}return t.prototype.parse=function(t,e,n,r){void 0===n&&(n=!1),void 0===r&&(r=$t);var i,o,s,a=this._htmlParser.parse(t,e,n,r);return a.errors.length?new Or(a.rootNodes,a.errors):(i=a.rootNodes,o=this._translationBundle,s=r,new bi([],{}).merge(i,o,s))},t}();var eo=/(\.ts|\.d\.ts|\.js|\.jsx|\.tsx)$/,no=/\.ngfactory\.|\.ngsummary\./,ro=/\.ngsummary\./,io=/NgSummary$/;function oo(t,e){void 0===e&&(e=!1);var n=ao(t,e);return n[0]+".ngfactory"+n[1]}function so(t){return t.replace(no,".")}function ao(t,e){if(void 0===e&&(e=!1),t.endsWith(".d.ts"))return[t.slice(0,-5),e?".ts":".d.ts"];var n=t.lastIndexOf(".");return-1!==n?[t.substring(0,n),t.substring(n)]:[t,""]}function co(t){return t.replace(eo,"")+".ngsummary.json"}function uo(t,e){void 0===e&&(e=!1);var n=ao(so(t),e);return n[0]+".ngsummary"+n[1]}function lo(t){return t+"NgSummary"}var po=/\u0275\d+/;var ho="@angular/core",fo=function(){function t(){}return t.ANALYZE_FOR_ENTRY_COMPONENTS={name:"ANALYZE_FOR_ENTRY_COMPONENTS",moduleName:ho},t.ElementRef={name:"ElementRef",moduleName:ho},t.NgModuleRef={name:"NgModuleRef",moduleName:ho},t.ViewContainerRef={name:"ViewContainerRef",moduleName:ho},t.ChangeDetectorRef={name:"ChangeDetectorRef",moduleName:ho},t.QueryList={name:"QueryList",moduleName:ho},t.TemplateRef={name:"TemplateRef",moduleName:ho},t.CodegenComponentFactoryResolver={name:"ɵCodegenComponentFactoryResolver",moduleName:ho},t.ComponentFactoryResolver={name:"ComponentFactoryResolver",moduleName:ho},t.ComponentFactory={name:"ComponentFactory",moduleName:ho},t.ComponentRef={name:"ComponentRef",moduleName:ho},t.NgModuleFactory={name:"NgModuleFactory",moduleName:ho},t.createModuleFactory={name:"ɵcmf",moduleName:ho},t.moduleDef={name:"ɵmod",moduleName:ho},t.moduleProviderDef={name:"ɵmpd",moduleName:ho},t.RegisterModuleFactoryFn={name:"ɵregisterModuleFactory",moduleName:ho},t.Injector={name:"Injector",moduleName:ho},t.ViewEncapsulation={name:"ViewEncapsulation",moduleName:ho},t.ChangeDetectionStrategy={name:"ChangeDetectionStrategy",moduleName:ho},t.SecurityContext={name:"SecurityContext",moduleName:ho},t.LOCALE_ID={name:"LOCALE_ID",moduleName:ho},t.TRANSLATIONS_FORMAT={name:"TRANSLATIONS_FORMAT",moduleName:ho},t.inlineInterpolate={name:"ɵinlineInterpolate",moduleName:ho},t.interpolate={name:"ɵinterpolate",moduleName:ho},t.EMPTY_ARRAY={name:"ɵEMPTY_ARRAY",moduleName:ho},t.EMPTY_MAP={name:"ɵEMPTY_MAP",moduleName:ho},t.Renderer={name:"Renderer",moduleName:ho},t.viewDef={name:"ɵvid",moduleName:ho},t.elementDef={name:"ɵeld",moduleName:ho},t.anchorDef={name:"ɵand",moduleName:ho},t.textDef={name:"ɵted",moduleName:ho},t.directiveDef={name:"ɵdid",moduleName:ho},t.providerDef={name:"ɵprd",moduleName:ho},t.queryDef={name:"ɵqud",moduleName:ho},t.pureArrayDef={name:"ɵpad",moduleName:ho},t.pureObjectDef={name:"ɵpod",moduleName:ho},t.purePipeDef={name:"ɵppd",moduleName:ho},t.pipeDef={name:"ɵpid",moduleName:ho},t.nodeValue={name:"ɵnov",moduleName:ho},t.ngContentDef={name:"ɵncd",moduleName:ho},t.unwrapValue={name:"ɵunv",moduleName:ho},t.createRendererType2={name:"ɵcrt",moduleName:ho},t.RendererType2={name:"RendererType2",moduleName:ho},t.ViewDefinition={name:"ɵViewDefinition",moduleName:ho},t.createComponentFactory={name:"ɵccf",moduleName:ho},t}();function mo(t){return{identifier:{reference:t}}}function go(t,e){return mo(t.resolveExternalReference(e))}var vo={OnInit:0,OnDestroy:1,DoCheck:2,OnChanges:3,AfterContentInit:4,AfterContentChecked:5,AfterViewInit:6,AfterViewChecked:7};vo[vo.OnInit]="OnInit",vo[vo.OnDestroy]="OnDestroy",vo[vo.DoCheck]="DoCheck",vo[vo.OnChanges]="OnChanges",vo[vo.AfterContentInit]="AfterContentInit",vo[vo.AfterContentChecked]="AfterContentChecked",vo[vo.AfterViewInit]="AfterViewInit",vo[vo.AfterViewChecked]="AfterViewChecked";var yo=[vo.OnInit,vo.OnDestroy,vo.DoCheck,vo.OnChanges,vo.AfterContentInit,vo.AfterContentChecked,vo.AfterViewInit,vo.AfterViewChecked];function bo(t,e,n){return t.hasLifecycleHook(n,function(t){switch(t){case vo.OnInit:return"ngOnInit";case vo.OnDestroy:return"ngOnDestroy";case vo.DoCheck:return"ngDoCheck";case vo.OnChanges:return"ngOnChanges";case vo.AfterContentInit:return"ngAfterContentInit";case vo.AfterContentChecked:return"ngAfterContentChecked";case vo.AfterViewInit:return"ngAfterViewInit";case vo.AfterViewChecked:return"ngAfterViewChecked"}}(e))}var _o=new RegExp("(\\:not\\()|([-\\w]+)|(?:\\.([-\\w]+))|(?:\\[([-.\\w*]+)(?:=([\"']?)([^\\]\"']*)\\5)?\\])|(\\))|(\\s*,\\s*)","g"),jo=function(){function t(){this.element=null,this.classNames=[],this.attrs=[],this.notSelectors=[]}return t.parse=function(e){var n,r=[],i=function(t,e){e.notSelectors.length>0&&!e.element&&0==e.classNames.length&&0==e.attrs.length&&(e.element="*"),t.push(e)},o=new t,s=o,a=!1;for(_o.lastIndex=0;n=_o.exec(e);){if(n[1]){if(a)throw new Error("Nesting :not is not allowed in a selector");a=!0,s=new t,o.notSelectors.push(s)}if(n[2]&&s.setElement(n[2]),n[3]&&s.addClassName(n[3]),n[4]&&s.addAttribute(n[4],n[6]),n[7]&&(a=!1,s=o),n[8]){if(a)throw new Error("Multiple selectors in :not are not supported");i(r,o),o=s=new t}}return i(r,o),r},t.prototype.isElementSelector=function(){return this.hasElementSelector()&&0==this.classNames.length&&0==this.attrs.length&&0===this.notSelectors.length},t.prototype.hasElementSelector=function(){return!!this.element},t.prototype.setElement=function(t){void 0===t&&(t=null),this.element=t},t.prototype.getMatchingElementTemplate=function(){for(var t=this.element||"div",e=this.classNames.length>0?' class="'+this.classNames.join(" ")+'"':"",n="",r=0;r<this.attrs.length;r+=2){n+=" "+this.attrs[r]+(""!==this.attrs[r+1]?'="'+this.attrs[r+1]+'"':"")}return oi(t).isVoid?"<"+t+e+n+"/>":"<"+t+e+n+"></"+t+">"},t.prototype.addAttribute=function(t,e){void 0===e&&(e=""),this.attrs.push(t,e&&e.toLowerCase()||"")},t.prototype.addClassName=function(t){this.classNames.push(t.toLowerCase())},t.prototype.toString=function(){var t=this.element||"";if(this.classNames&&this.classNames.forEach(function(e){return t+="."+e}),this.attrs)for(var e=0;e<this.attrs.length;e+=2){var n=this.attrs[e],r=this.attrs[e+1];t+="["+n+(r?"="+r:"")+"]"}return this.notSelectors.forEach(function(e){return t+=":not("+e+")"}),t},t}(),xo=function(){function t(){this._elementMap=new Map,this._elementPartialMap=new Map,this._classMap=new Map,this._classPartialMap=new Map,this._attrValueMap=new Map,this._attrValuePartialMap=new Map,this._listContexts=[]}return t.createNotMatcher=function(e){var n=new t;return n.addSelectables(e,null),n},t.prototype.addSelectables=function(t,e){var n=null;t.length>1&&(n=new wo(t),this._listContexts.push(n));for(var r=0;r<t.length;r++)this._addSelectable(t[r],e,n)},t.prototype._addSelectable=function(t,e,n){var r=this,i=t.element,o=t.classNames,s=t.attrs,a=new Oo(t,e,n);i&&((u=0===s.length&&0===o.length)?this._addTerminal(r._elementMap,i,a):r=this._addPartial(r._elementPartialMap,i));if(o)for(var c=0;c<o.length;c++){var u=0===s.length&&c===o.length-1,l=o[c];u?this._addTerminal(r._classMap,l,a):r=this._addPartial(r._classPartialMap,l)}if(s)for(c=0;c<s.length;c+=2){u=c===s.length-2;var p=s[c],d=s[c+1];if(u){var h=r._attrValueMap,f=h.get(p);f||(f=new Map,h.set(p,f)),this._addTerminal(f,d,a)}else{var m=r._attrValuePartialMap,g=m.get(p);g||(g=new Map,m.set(p,g)),r=this._addPartial(g,d)}}},t.prototype._addTerminal=function(t,e,n){var r=t.get(e);r||(r=[],t.set(e,r)),r.push(n)},t.prototype._addPartial=function(e,n){var r=e.get(n);return r||(r=new t,e.set(n,r)),r},t.prototype.match=function(t,e){for(var n=!1,r=t.element,i=t.classNames,o=t.attrs,s=0;s<this._listContexts.length;s++)this._listContexts[s].alreadyMatched=!1;if(n=this._matchTerminal(this._elementMap,r,t,e)||n,n=this._matchPartial(this._elementPartialMap,r,t,e)||n,i)for(s=0;s<i.length;s++){var a=i[s];n=this._matchTerminal(this._classMap,a,t,e)||n,n=this._matchPartial(this._classPartialMap,a,t,e)||n}if(o)for(s=0;s<o.length;s+=2){var c=o[s],u=o[s+1],l=this._attrValueMap.get(c);u&&(n=this._matchTerminal(l,"",t,e)||n),n=this._matchTerminal(l,u,t,e)||n;var p=this._attrValuePartialMap.get(c);u&&(n=this._matchPartial(p,"",t,e)||n),n=this._matchPartial(p,u,t,e)||n}return n},t.prototype._matchTerminal=function(t,e,n,r){if(!t||"string"!=typeof e)return!1;var i=t.get(e)||[],o=t.get("*");if(o&&(i=i.concat(o)),0===i.length)return!1;for(var s=!1,a=0;a<i.length;a++)s=i[a].finalize(n,r)||s;return s},t.prototype._matchPartial=function(t,e,n,r){if(!t||"string"!=typeof e)return!1;var i=t.get(e);return!!i&&i.match(n,r)},t}(),wo=function(){return function(t){this.selectors=t,this.alreadyMatched=!1}}(),Oo=function(){function t(t,e,n){this.selector=t,this.cbContext=e,this.listContext=n,this.notSelectors=t.notSelectors}return t.prototype.finalize=function(t,e){var n=!0;!(this.notSelectors.length>0)||this.listContext&&this.listContext.alreadyMatched||(n=!xo.createNotMatcher(this.notSelectors).match(t,null));return!n||!e||this.listContext&&this.listContext.alreadyMatched||(this.listContext&&(this.listContext.alreadyMatched=!0),e(this.selector,this.cbContext)),n},t}(),Co="ngComponentType",ko=function(){function t(t,e,n,r,i,o,s,a,c,u,l,p){this._config=t,this._htmlParser=e,this._ngModuleResolver=n,this._directiveResolver=r,this._pipeResolver=i,this._summaryResolver=o,this._schemaRegistry=s,this._directiveNormalizer=a,this._console=c,this._staticSymbolCache=u,this._reflector=l,this._errorCollector=p,this._nonNormalizedDirectiveCache=new Map,this._directiveCache=new Map,this._summaryCache=new Map,this._pipeCache=new Map,this._ngModuleCache=new Map,this._ngModuleOfTypes=new Map}return t.prototype.getReflector=function(){return this._reflector},t.prototype.clearCacheFor=function(t){var e=this._directiveCache.get(t);this._directiveCache.delete(t),this._nonNormalizedDirectiveCache.delete(t),this._summaryCache.delete(t),this._pipeCache.delete(t),this._ngModuleOfTypes.delete(t),this._ngModuleCache.clear(),e&&this._directiveNormalizer.clearCacheFor(e)},t.prototype.clearCache=function(){this._directiveCache.clear(),this._nonNormalizedDirectiveCache.clear(),this._summaryCache.clear(),this._pipeCache.clear(),this._ngModuleCache.clear(),this._ngModuleOfTypes.clear(),this._directiveNormalizer.clearCache()},t.prototype._createProxyClass=function(t,e){var n=null,r=function(){if(!n)throw new Error("Illegal state: Class "+e+" for type "+Y(t)+" is not compiled yet!");return n.apply(this,arguments)};return r.setDelegate=function(t){n=t,r.prototype=t.prototype},r.overriddenName=e,r},t.prototype.getGeneratedClass=function(t,e){return t instanceof mt?this._staticSymbolCache.get(oo(t.filePath),e):this._createProxyClass(t,e)},t.prototype.getComponentViewClass=function(t){return this.getGeneratedClass(t,xt(t,0))},t.prototype.getHostComponentViewClass=function(t){return this.getGeneratedClass(t,"HostView_"+_t({reference:t}))},t.prototype.getHostComponentType=function(t){var e=_t({reference:t})+"_Host";if(t instanceof mt)return this._staticSymbolCache.get(t.filePath,e);var n=function(){};return n.overriddenName=e,n},t.prototype.getRendererType=function(t){return t instanceof mt?this._staticSymbolCache.get(oo(t.filePath),wt(t)):{}},t.prototype.getComponentFactory=function(t,e,n,r){if(e instanceof mt)return this._staticSymbolCache.get(oo(e.filePath),Ot(e));var i=this.getHostComponentViewClass(e);return this._reflector.resolveExternalReference(fo.createComponentFactory)(t,e,i,n,r,[])},t.prototype.initComponentFactory=function(t,e){var n;t instanceof mt||(n=t.ngContentSelectors).push.apply(n,e)},t.prototype._loadSummary=function(t,e){var n=this._summaryCache.get(t);if(!n){var r=this._summaryResolver.resolveSummary(t);n=r?r.type:null,this._summaryCache.set(t,n||null)}return n&&n.summaryKind===e?n:null},t.prototype.getHostComponentMetadata=function(t,e){var n=this.getHostComponentType(t.type.reference);e||(e=this.getHostComponentViewClass(n));var r=jo.parse(t.selector)[0].getMatchingElementTemplate(),i=this._htmlParser.parse(r,"");return Mt.create({isHost:!0,type:{reference:n,diDeps:[],lifecycleHooks:[]},template:new Pt({encapsulation:d.None,template:r,templateUrl:"",htmlAst:i,styles:[],styleUrls:[],ngContentSelectors:[],animations:[],isInline:!0,externalStylesheets:[],interpolation:null,preserveWhitespaces:!1}),exportAs:null,changeDetection:h.Default,inputs:[],outputs:[],host:{},isComponent:!0,selector:"*",providers:[],viewProviders:[],queries:[],viewQueries:[],componentViewType:e,rendererType:{id:"__Host__",encapsulation:d.None,styles:[],data:{}},entryComponents:[],componentFactory:null})},t.prototype.loadDirectiveMetadata=function(t,e,n){var r=this;if(this._directiveCache.has(e))return null;e=K(e);var i,o,s=this.getNonNormalizedDirectiveMetadata(e),a=s.annotation,c=s.metadata,u=function(t){var n=new Mt({isHost:!1,type:c.type,isComponent:c.isComponent,selector:c.selector,exportAs:c.exportAs,changeDetection:c.changeDetection,inputs:c.inputs,outputs:c.outputs,hostListeners:c.hostListeners,hostProperties:c.hostProperties,hostAttributes:c.hostAttributes,providers:c.providers,viewProviders:c.viewProviders,queries:c.queries,viewQueries:c.viewQueries,entryComponents:c.entryComponents,componentViewType:c.componentViewType,rendererType:c.rendererType,componentFactory:c.componentFactory,template:t});return t&&r.initComponentFactory(c.componentFactory,t.ngContentSelectors),r._directiveCache.set(e,n),r._summaryCache.set(e,n.toSummary()),null};if(c.isComponent){var l=c.template,p=this._directiveNormalizer.normalizeTemplate({ngModuleType:t,componentType:e,moduleUrl:this._reflector.componentModuleUrl(e,a),encapsulation:l.encapsulation,template:l.template,templateUrl:l.templateUrl,styles:l.styles,styleUrls:l.styleUrls,animations:l.animations,interpolation:l.interpolation,preserveWhitespaces:l.preserveWhitespaces});return $(p)&&n?(this._reportError((i=e,(o=Error("Can't compile synchronously as "+Y(i)+" is still being loaded!"))[Co]=i,o),e),null):z(p,u)}return u(null),null},t.prototype.getNonNormalizedDirectiveMetadata=function(t){var e=this;if(!(t=K(t)))return null;var n=this._nonNormalizedDirectiveCache.get(t);if(n)return n;var r=this._directiveResolver.resolve(t,!1);if(!r)return null;var i=void 0;if(f.isTypeOf(r)){Gt("styles",(s=r).styles),Gt("styleUrls",s.styleUrls),Yt("interpolation",s.interpolation);var o=s.animations;i=new Pt({encapsulation:L(s.encapsulation),template:L(s.template),templateUrl:L(s.templateUrl),htmlAst:null,styles:s.styles||[],styleUrls:s.styleUrls||[],animations:o||[],interpolation:L(s.interpolation),isInline:!!s.template,externalStylesheets:[],ngContentSelectors:[],preserveWhitespaces:L(r.preserveWhitespaces)})}var s,a=null,c=[],u=[],l=r.selector;f.isTypeOf(r)?(a=(s=r).changeDetection,s.viewProviders&&(c=this._getProvidersMetadata(s.viewProviders,u,'viewProviders for "'+Mo(t)+'"',[],t)),s.entryComponents&&(u=Eo(s.entryComponents).map(function(t){return e._getEntryComponentMetadata(t)}).concat(u)),l||(l=this._schemaRegistry.getDefaultComponentElementName())):l||(this._reportError(q("Directive "+Mo(t)+" has no selector, please add it!"),t),l="error");var p=[];null!=r.providers&&(p=this._getProvidersMetadata(r.providers,u,'providers for "'+Mo(t)+'"',[],t));var d=[],h=[];null!=r.queries&&(d=this._getQueriesMetadata(r.queries,!1,t),h=this._getQueriesMetadata(r.queries,!0,t));var m=Mt.create({isHost:!1,selector:l,exportAs:L(r.exportAs),isComponent:!!i,type:this._getTypeMetadata(t),template:i,changeDetection:a,inputs:r.inputs||[],outputs:r.outputs||[],host:r.host||{},providers:p||[],viewProviders:c||[],queries:d||[],viewQueries:h||[],entryComponents:u,componentViewType:i?this.getComponentViewClass(t):null,rendererType:i?this.getRendererType(t):null,componentFactory:null});return i&&(m.componentFactory=this.getComponentFactory(l,t,m.inputs,m.outputs)),n={metadata:m,annotation:r},this._nonNormalizedDirectiveCache.set(t,n),n},t.prototype.getDirectiveMetadata=function(t){var e=this._directiveCache.get(t);return e||this._reportError(q("Illegal state: getDirectiveMetadata can only be called after loadNgModuleDirectiveAndPipeMetadata for a module that declares it. Directive "+Mo(t)+"."),t),e},t.prototype.getDirectiveSummary=function(t){var e=this._loadSummary(t,Ct.Directive);return e||this._reportError(q("Illegal state: Could not load the summary for directive "+Mo(t)+"."),t),e},t.prototype.isDirective=function(t){return!!this._loadSummary(t,Ct.Directive)||this._directiveResolver.isDirective(t)},t.prototype.isPipe=function(t){return!!this._loadSummary(t,Ct.Pipe)||this._pipeResolver.isPipe(t)},t.prototype.isNgModule=function(t){return!!this._loadSummary(t,Ct.NgModule)||this._ngModuleResolver.isNgModule(t)},t.prototype.getNgModuleSummary=function(t){var e=this._loadSummary(t,Ct.NgModule);if(!e){var n=this.getNgModuleMetadata(t,!1);(e=n?n.toSummary():null)&&this._summaryCache.set(t,e)}return e},t.prototype.loadNgModuleDirectiveAndPipeMetadata=function(t,e,n){var r=this;void 0===n&&(n=!0);var i=this.getNgModuleMetadata(t,n),o=[];return i&&(i.declaredDirectives.forEach(function(n){var i=r.loadDirectiveMetadata(t,n.reference,e);i&&o.push(i)}),i.declaredPipes.forEach(function(t){return r._loadPipeMetadata(t.reference)})),Promise.all(o)},t.prototype.getNgModuleMetadata=function(t,e){var n=this;void 0===e&&(e=!0),t=K(t);var r=this._ngModuleCache.get(t);if(r)return r;var i=this._ngModuleResolver.resolve(t,e);if(!i)return null;var o=[],s=[],a=[],c=[],u=[],l=[],p=[],d=[],h=[];i.imports&&Eo(i.imports).forEach(function(e){var r=void 0;if(So(e))r=e;else if(e&&e.ngModule){var i=e;r=i.ngModule,i.providers&&l.push.apply(l,n._getProvidersMetadata(i.providers,p,"provider for the NgModule '"+Mo(r)+"'",[],e))}if(r){if(!n._checkSelfImport(t,r)){var o=n.getNgModuleSummary(r);o?c.push(o):n._reportError(q("Unexpected "+n._getTypeDescriptor(e)+" '"+Mo(e)+"' imported by the module '"+Mo(t)+"'. Please add a @NgModule annotation."),t)}}else n._reportError(q("Unexpected value '"+Mo(e)+"' imported by the module '"+Mo(t)+"'"),t)}),i.exports&&Eo(i.exports).forEach(function(e){if(So(e)){var r=n.getNgModuleSummary(e);r?u.push(r):s.push(n._getIdentifierMetadata(e))}else n._reportError(q("Unexpected value '"+Mo(e)+"' exported by the module '"+Mo(t)+"'"),t)});var f=this._getTransitiveNgModuleMetadata(c,u);i.declarations&&Eo(i.declarations).forEach(function(e){if(So(e)){var r=n._getIdentifierMetadata(e);if(n.isDirective(e))f.addDirective(r),o.push(r),n._addTypeToModule(e,t);else{if(!n.isPipe(e))return void n._reportError(q("Unexpected "+n._getTypeDescriptor(e)+" '"+Mo(e)+"' declared by the module '"+Mo(t)+"'. Please add a @Pipe/@Directive/@Component annotation."),t);f.addPipe(r),f.pipes.push(r),a.push(r),n._addTypeToModule(e,t)}}else n._reportError(q("Unexpected value '"+Mo(e)+"' declared by the module '"+Mo(t)+"'"),t)});var m=[],g=[];return s.forEach(function(e){if(f.directivesSet.has(e.reference))m.push(e),f.addExportedDirective(e);else{if(!f.pipesSet.has(e.reference))return void n._reportError(q("Can't export "+n._getTypeDescriptor(e.reference)+" "+Mo(e.reference)+" from "+Mo(t)+" as it was neither declared nor imported!"),t);g.push(e),f.addExportedPipe(e)}}),i.providers&&l.push.apply(l,this._getProvidersMetadata(i.providers,p,"provider for the NgModule '"+Mo(t)+"'",[],t)),i.entryComponents&&p.push.apply(p,Eo(i.entryComponents).map(function(t){return n._getEntryComponentMetadata(t)})),i.bootstrap&&Eo(i.bootstrap).forEach(function(e){So(e)?d.push(n._getIdentifierMetadata(e)):n._reportError(q("Unexpected value '"+Mo(e)+"' used in the bootstrap property of module '"+Mo(t)+"'"),t)}),p.push.apply(p,d.map(function(t){return n._getEntryComponentMetadata(t.reference)})),i.schemas&&h.push.apply(h,Eo(i.schemas)),r=new Tt({type:this._getTypeMetadata(t),providers:l,entryComponents:p,bootstrapComponents:d,schemas:h,declaredDirectives:o,exportedDirectives:m,declaredPipes:a,exportedPipes:g,importedModules:c,exportedModules:u,transitiveModule:f,id:i.id||null}),p.forEach(function(t){return f.addEntryComponent(t)}),l.forEach(function(t){return f.addProvider(t,r.type)}),f.addModule(r.type),this._ngModuleCache.set(t,r),r},t.prototype._checkSelfImport=function(t,e){return t===e&&(this._reportError(q("'"+Mo(t)+"' module can't import itself"),t),!0)},t.prototype._getTypeDescriptor=function(t){if(So(t)){if(this.isDirective(t))return"directive";if(this.isPipe(t))return"pipe";if(this.isNgModule(t))return"module"}return t.provide?"provider":"value"},t.prototype._addTypeToModule=function(t,e){var n=this._ngModuleOfTypes.get(t);n&&n!==e?this._reportError(q("Type "+Mo(t)+" is part of the declarations of 2 modules: "+Mo(n)+" and "+Mo(e)+"! Please consider moving "+Mo(t)+" to a higher module that imports "+Mo(n)+" and "+Mo(e)+". You can also create a new NgModule that exports and includes "+Mo(t)+" then import that NgModule in "+Mo(n)+" and "+Mo(e)+"."),e):this._ngModuleOfTypes.set(t,e)},t.prototype._getTransitiveNgModuleMetadata=function(t,e){var n=new Dt,r=new Map;return t.concat(e).forEach(function(t){t.modules.forEach(function(t){return n.addModule(t)}),t.entryComponents.forEach(function(t){return n.addEntryComponent(t)});var e=new Set;t.providers.forEach(function(t){var i=Et(t.provider.token),o=r.get(i);o||(o=new Set,r.set(i,o));var s=t.module.reference;!e.has(i)&&o.has(s)||(o.add(s),e.add(i),n.addProvider(t.provider,t.module))})}),e.forEach(function(t){t.exportedDirectives.forEach(function(t){return n.addExportedDirective(t)}),t.exportedPipes.forEach(function(t){return n.addExportedPipe(t)})}),t.forEach(function(t){t.exportedDirectives.forEach(function(t){return n.addDirective(t)}),t.exportedPipes.forEach(function(t){return n.addPipe(t)})}),n},t.prototype._getIdentifierMetadata=function(t){return{reference:t=K(t)}},t.prototype.isInjectable=function(t){return this._reflector.annotations(t).some(function(t){return O.isTypeOf(t)})},t.prototype.getInjectableSummary=function(t){return{summaryKind:Ct.Injectable,type:this._getTypeMetadata(t,null,!1)}},t.prototype._getInjectableMetadata=function(t,e){void 0===e&&(e=null);var n=this._loadSummary(t,Ct.Injectable);return n?n.type:this._getTypeMetadata(t,e)},t.prototype._getTypeMetadata=function(t,e,n){void 0===e&&(e=null),void 0===n&&(n=!0);var r,i,o=this._getIdentifierMetadata(t);return{reference:o.reference,diDeps:this._getDependenciesMetadata(o.reference,e,n),lifecycleHooks:(r=this._reflector,i=o.reference,yo.filter(function(t){return bo(r,t,i)}))}},t.prototype._getFactoryMetadata=function(t,e){return void 0===e&&(e=null),{reference:t=K(t),diDeps:this._getDependenciesMetadata(t,e)}},t.prototype.getPipeMetadata=function(t){var e=this._pipeCache.get(t);return e||this._reportError(q("Illegal state: getPipeMetadata can only be called after loadNgModuleDirectiveAndPipeMetadata for a module that declares it. Pipe "+Mo(t)+"."),t),e||null},t.prototype.getPipeSummary=function(t){var e=this._loadSummary(t,Ct.Pipe);return e||this._reportError(q("Illegal state: Could not load the summary for pipe "+Mo(t)+"."),t),e},t.prototype.getOrLoadPipeMetadata=function(t){var e=this._pipeCache.get(t);return e||(e=this._loadPipeMetadata(t)),e},t.prototype._loadPipeMetadata=function(t){t=K(t);var e=this._pipeResolver.resolve(t),n=new At({type:this._getTypeMetadata(t),name:e.name,pure:!!e.pure});return this._pipeCache.set(t,n),this._summaryCache.set(t,n.toSummary()),n},t.prototype._getDependenciesMetadata=function(t,e,n){var r=this;void 0===n&&(n=!0);var a=!1,c=(e||this._reflector.parameters(t)||[]).map(function(t){var e=!1,n=!1,c=!1,u=!1,l=!1,p=null;return Array.isArray(t)?t.forEach(function(t){E.isTypeOf(t)?n=!0:C.isTypeOf(t)?c=!0:k.isTypeOf(t)?u=!0:w.isTypeOf(t)?l=!0:s.isTypeOf(t)?(e=!0,p=t.attributeName):i.isTypeOf(t)?p=t.token:o.isTypeOf(t)||t instanceof mt?p=t:So(t)&&null==p&&(p=t)}):p=t,null==p?(a=!0,null):{isAttribute:e,isHost:n,isSelf:c,isSkipSelf:u,isOptional:l,token:r._getTokenMetadata(p)}});if(a){var u=c.map(function(t){return t?Mo(t.token):"?"}).join(", "),l="Can't resolve all parameters for "+Mo(t)+": ("+u+").";n||this._config.strictInjectionParameters?this._reportError(q(l),t):this._console.warn("Warning: "+l+" This will become an error in Angular v6.x")}return c},t.prototype._getTokenMetadata=function(t){return"string"==typeof(t=K(t))?{value:t}:{identifier:{reference:t}}},t.prototype._getProvidersMetadata=function(t,e,n,r,i){var o=this;return void 0===r&&(r=[]),t.forEach(function(s,a){if(Array.isArray(s))o._getProvidersMetadata(s,e,n,r);else{var c=void 0;if((s=K(s))&&"object"==typeof s&&s.hasOwnProperty("provide"))o._validateProvider(s),c=new Rt(s.provide,s);else{if(!So(s)){if(void 0===s)return void o._reportError(q("Encountered undefined provider! Usually this means you have a circular dependencies (might be caused by using 'barrel' index.ts files."));var u=t.reduce(function(t,e,n){return n<a?t.push(""+Mo(e)):n==a?t.push("?"+Mo(e)+"?"):n==a+1&&t.push("..."),t},[]).join(", ");return void o._reportError(q("Invalid "+(n||"provider")+" - only instances of Provider and Type are allowed, got: ["+u+"]"),i)}c=new Rt(s,{useClass:s})}c.token===o._reflector.resolveExternalReference(fo.ANALYZE_FOR_ENTRY_COMPONENTS)?e.push.apply(e,o._getEntryComponentsFromProvider(c,i)):r.push(o.getProviderMetadata(c))}}),r},t.prototype._validateProvider=function(t){t.hasOwnProperty("useClass")&&null==t.useClass&&this._reportError(q("Invalid provider for "+Mo(t.provide)+". useClass cannot be "+t.useClass+".\n           Usually it happens when:\n           1. There's a circular dependency (might be caused by using index.ts (barrel) files).\n           2. Class was used before it was declared. Use forwardRef in this case."))},t.prototype._getEntryComponentsFromProvider=function(t,e){var n,r,i=this,o=[],s=[];return t.useFactory||t.useExisting||t.useClass?(this._reportError(q("The ANALYZE_FOR_ENTRY_COMPONENTS token only supports useValue!"),e),[]):t.multi?(n=t.useValue,r=s,R(n,new Po,r),s.forEach(function(t){var e=i._getEntryComponentMetadata(t.reference,!1);e&&o.push(e)}),o):(this._reportError(q("The ANALYZE_FOR_ENTRY_COMPONENTS token only supports 'multi = true'!"),e),[])},t.prototype._getEntryComponentMetadata=function(t,e){void 0===e&&(e=!0);var n=this.getNonNormalizedDirectiveMetadata(t);if(n&&n.metadata.isComponent)return{componentType:t,componentFactory:n.metadata.componentFactory};var r=this._loadSummary(t,Ct.Directive);if(r&&r.isComponent)return{componentType:t,componentFactory:r.componentFactory};if(e)throw q(t.name+" cannot be used as an entry component.");return null},t.prototype.getProviderMetadata=function(t){var e=void 0,n=null,r=null,i=this._getTokenMetadata(t.token);return t.useClass?(e=(n=this._getInjectableMetadata(t.useClass,t.dependencies)).diDeps,t.token===t.useClass&&(i={identifier:n})):t.useFactory&&(e=(r=this._getFactoryMetadata(t.useFactory,t.dependencies)).diDeps),{token:i,useClass:n,useValue:t.useValue,useFactory:r,useExisting:t.useExisting?this._getTokenMetadata(t.useExisting):void 0,deps:e,multi:t.multi}},t.prototype._getQueriesMetadata=function(t,e,n){var r=this,i=[];return Object.keys(t).forEach(function(o){var s=t[o];s.isViewQuery===e&&i.push(r._getQueryMetadata(s,o,n))}),i},t.prototype._queryVarBindings=function(t){return t.split(/\s*,\s*/)},t.prototype._getQueryMetadata=function(t,e,n){var r,i=this;return"string"==typeof t.selector?r=this._queryVarBindings(t.selector).map(function(t){return i._getTokenMetadata(t)}):t.selector?r=[this._getTokenMetadata(t.selector)]:(this._reportError(q("Can't construct a query for the property \""+e+'" of "'+Mo(n)+"\" since the query selector wasn't defined."),n),r=[]),{selectors:r,first:t.first,descendants:t.descendants,propertyName:e,read:t.read?this._getTokenMetadata(t.read):null}},t.prototype._reportError=function(t,e,n){if(!this._errorCollector)throw t;this._errorCollector(t,e),n&&this._errorCollector(t,n)},t}();function Eo(t){return(e=function t(e,n){if(void 0===n&&(n=[]),e)for(var r=0;r<e.length;r++){var i=K(e[r]);Array.isArray(i)?t(i,n):n.push(i)}return n}(t))?Array.from(new Set(e)):[];var e}function So(t){return t instanceof mt||t instanceof S}var Po=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.visitOther=function(t,e){e.push({reference:t})},e}(F);function Mo(t){return t instanceof mt?t.name+" in "+t.filePath:Y(t)}var Ao={Const:0};Ao[Ao.Const]="Const";var To=function(){function t(t){void 0===t&&(t=null),this.modifiers=t,t||(this.modifiers=[])}return t.prototype.hasModifier=function(t){return-1!==this.modifiers.indexOf(t)},t}(),Do={Dynamic:0,Bool:1,String:2,Int:3,Number:4,Function:5,Inferred:6};Do[Do.Dynamic]="Dynamic",Do[Do.Bool]="Bool",Do[Do.String]="String",Do[Do.Int]="Int",Do[Do.Number]="Number",Do[Do.Function]="Function",Do[Do.Inferred]="Inferred";var Io=function(t){function e(e,n){void 0===n&&(n=null);var r=t.call(this,n)||this;return r.name=e,r}return Object(r.b)(e,t),e.prototype.visitType=function(t,e){return t.visitBuiltintType(this,e)},e}(To),Ro=function(t){function e(e,n){void 0===n&&(n=null);var r=t.call(this,n)||this;return r.value=e,r}return Object(r.b)(e,t),e.prototype.visitType=function(t,e){return t.visitExpressionType(this,e)},e}(To),No=function(t){function e(e,n){void 0===n&&(n=null);var r=t.call(this,n)||this;return r.of=e,r}return Object(r.b)(e,t),e.prototype.visitType=function(t,e){return t.visitArrayType(this,e)},e}(To),Lo=function(t){function e(e,n){void 0===n&&(n=null);var r=t.call(this,n)||this;return r.valueType=e||null,r}return Object(r.b)(e,t),e.prototype.visitType=function(t,e){return t.visitMapType(this,e)},e}(To),Fo=new Io(Do.Dynamic),Vo=new Io(Do.Inferred),zo=new Io(Do.Bool),Bo=(new Io(Do.Int),new Io(Do.Number),new Io(Do.String),new Io(Do.Function),{Equals:0,NotEquals:1,Identical:2,NotIdentical:3,Minus:4,Plus:5,Divide:6,Multiply:7,Modulo:8,And:9,Or:10,Lower:11,LowerEquals:12,Bigger:13,BiggerEquals:14});function qo(t,e){return null==t||null==e?t==e:t.isEquivalent(e)}function Ho(t,e){var n=t.length;if(n!==e.length)return!1;for(var r=0;r<n;r++)if(!t[r].isEquivalent(e[r]))return!1;return!0}Bo[Bo.Equals]="Equals",Bo[Bo.NotEquals]="NotEquals",Bo[Bo.Identical]="Identical",Bo[Bo.NotIdentical]="NotIdentical",Bo[Bo.Minus]="Minus",Bo[Bo.Plus]="Plus",Bo[Bo.Divide]="Divide",Bo[Bo.Multiply]="Multiply",Bo[Bo.Modulo]="Modulo",Bo[Bo.And]="And",Bo[Bo.Or]="Or",Bo[Bo.Lower]="Lower",Bo[Bo.LowerEquals]="LowerEquals",Bo[Bo.Bigger]="Bigger",Bo[Bo.BiggerEquals]="BiggerEquals";var Uo=function(){function t(t,e){this.type=t||null,this.sourceSpan=e||null}return t.prototype.prop=function(t,e){return new ls(this,t,null,e)},t.prototype.key=function(t,e,n){return new ps(this,t,e,n)},t.prototype.callMethod=function(t,e,n){return new Qo(this,t,e,null,n)},t.prototype.callFn=function(t,e){return new Zo(this,t,null,e)},t.prototype.instantiate=function(t,e,n){return new Jo(this,t,e,n)},t.prototype.conditional=function(t,e,n){return void 0===e&&(e=null),new rs(this,t,e,null,n)},t.prototype.equals=function(t,e){return new us(Bo.Equals,this,t,null,e)},t.prototype.notEquals=function(t,e){return new us(Bo.NotEquals,this,t,null,e)},t.prototype.identical=function(t,e){return new us(Bo.Identical,this,t,null,e)},t.prototype.notIdentical=function(t,e){return new us(Bo.NotIdentical,this,t,null,e)},t.prototype.minus=function(t,e){return new us(Bo.Minus,this,t,null,e)},t.prototype.plus=function(t,e){return new us(Bo.Plus,this,t,null,e)},t.prototype.divide=function(t,e){return new us(Bo.Divide,this,t,null,e)},t.prototype.multiply=function(t,e){return new us(Bo.Multiply,this,t,null,e)},t.prototype.modulo=function(t,e){return new us(Bo.Modulo,this,t,null,e)},t.prototype.and=function(t,e){return new us(Bo.And,this,t,null,e)},t.prototype.or=function(t,e){return new us(Bo.Or,this,t,null,e)},t.prototype.lower=function(t,e){return new us(Bo.Lower,this,t,null,e)},t.prototype.lowerEquals=function(t,e){return new us(Bo.LowerEquals,this,t,null,e)},t.prototype.bigger=function(t,e){return new us(Bo.Bigger,this,t,null,e)},t.prototype.biggerEquals=function(t,e){return new us(Bo.BiggerEquals,this,t,null,e)},t.prototype.isBlank=function(t){return this.equals(vs,t)},t.prototype.cast=function(t,e){return new ss(this,t,e)},t.prototype.toStmt=function(){return new xs(this,null)},t}(),Wo={This:0,Super:1,CatchError:2,CatchStack:3};Wo[Wo.This]="This",Wo[Wo.Super]="Super",Wo[Wo.CatchError]="CatchError",Wo[Wo.CatchStack]="CatchStack";var Go=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return"string"==typeof e?(i.name=e,i.builtin=null):(i.name=null,i.builtin=e),i}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.name===t.name&&this.builtin===t.builtin},e.prototype.visitExpression=function(t,e){return t.visitReadVarExpr(this,e)},e.prototype.set=function(t){if(!this.name)throw new Error("Built in variable "+this.builtin+" can not be assigned to.");return new Xo(this.name,t,null,this.sourceSpan)},e}(Uo),Xo=function(t){function e(e,n,r,i){var o=t.call(this,r||n.type,i)||this;return o.name=e,o.value=n,o}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.name===t.name&&this.value.isEquivalent(t.value)},e.prototype.visitExpression=function(t,e){return t.visitWriteVarExpr(this,e)},e.prototype.toDeclStmt=function(t,e){return new _s(this.name,this.value,t,e,this.sourceSpan)},e}(Uo),Yo=function(t){function e(e,n,r,i,o){var s=t.call(this,i||r.type,o)||this;return s.receiver=e,s.index=n,s.value=r,s}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.index.isEquivalent(t.index)&&this.value.isEquivalent(t.value)},e.prototype.visitExpression=function(t,e){return t.visitWriteKeyExpr(this,e)},e}(Uo),Ko=function(t){function e(e,n,r,i,o){var s=t.call(this,i||r.type,o)||this;return s.receiver=e,s.name=n,s.value=r,s}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.name===t.name&&this.value.isEquivalent(t.value)},e.prototype.visitExpression=function(t,e){return t.visitWritePropExpr(this,e)},e}(Uo),$o={ConcatArray:0,SubscribeObservable:1,Bind:2};$o[$o.ConcatArray]="ConcatArray",$o[$o.SubscribeObservable]="SubscribeObservable",$o[$o.Bind]="Bind";var Qo=function(t){function e(e,n,r,i,o){var s=t.call(this,i,o)||this;return s.receiver=e,s.args=r,"string"==typeof n?(s.name=n,s.builtin=null):(s.name=null,s.builtin=n),s}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.name===t.name&&this.builtin===t.builtin&&Ho(this.args,t.args)},e.prototype.visitExpression=function(t,e){return t.visitInvokeMethodExpr(this,e)},e}(Uo),Zo=function(t){function e(e,n,r,i){var o=t.call(this,r,i)||this;return o.fn=e,o.args=n,o}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.fn.isEquivalent(t.fn)&&Ho(this.args,t.args)},e.prototype.visitExpression=function(t,e){return t.visitInvokeFunctionExpr(this,e)},e}(Uo),Jo=function(t){function e(e,n,r,i){var o=t.call(this,r,i)||this;return o.classExpr=e,o.args=n,o}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.classExpr.isEquivalent(t.classExpr)&&Ho(this.args,t.args)},e.prototype.visitExpression=function(t,e){return t.visitInstantiateExpr(this,e)},e}(Uo),ts=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.value=e,i}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.value===t.value},e.prototype.visitExpression=function(t,e){return t.visitLiteralExpr(this,e)},e}(Uo),es=function(t){function e(e,n,r,i){void 0===r&&(r=null);var o=t.call(this,n,i)||this;return o.value=e,o.typeParams=r,o}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.value.name===t.value.name&&this.value.moduleName===t.value.moduleName&&this.value.runtime===t.value.runtime},e.prototype.visitExpression=function(t,e){return t.visitExternalExpr(this,e)},e}(Uo),ns=function(){return function(t,e,n){this.moduleName=t,this.name=e,this.runtime=n}}(),rs=function(t){function e(e,n,r,i,o){void 0===r&&(r=null);var s=t.call(this,i||n.type,o)||this;return s.condition=e,s.falseCase=r,s.trueCase=n,s}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.condition.isEquivalent(t.condition)&&this.trueCase.isEquivalent(t.trueCase)&&qo(this.falseCase,t.falseCase)},e.prototype.visitExpression=function(t,e){return t.visitConditionalExpr(this,e)},e}(Uo),is=function(t){function e(e,n){var r=t.call(this,zo,n)||this;return r.condition=e,r}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.condition.isEquivalent(t.condition)},e.prototype.visitExpression=function(t,e){return t.visitNotExpr(this,e)},e}(Uo),os=function(t){function e(e,n){var r=t.call(this,e.type,n)||this;return r.condition=e,r}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.condition.isEquivalent(t.condition)},e.prototype.visitExpression=function(t,e){return t.visitAssertNotNullExpr(this,e)},e}(Uo),ss=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.value=e,i}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.value.isEquivalent(t.value)},e.prototype.visitExpression=function(t,e){return t.visitCastExpr(this,e)},e}(Uo),as=function(){function t(t,e){void 0===e&&(e=null),this.name=t,this.type=e}return t.prototype.isEquivalent=function(t){return this.name===t.name},t}(),cs=function(t){function e(e,n,r,i){var o=t.call(this,r,i)||this;return o.params=e,o.statements=n,o}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&Ho(this.params,t.params)&&Ho(this.statements,t.statements)},e.prototype.visitExpression=function(t,e){return t.visitFunctionExpr(this,e)},e.prototype.toDeclStmt=function(t,e){return void 0===e&&(e=null),new js(t,this.params,this.statements,this.type,e,this.sourceSpan)},e}(Uo),us=function(t){function e(e,n,r,i,o){var s=t.call(this,i||n.type,o)||this;return s.operator=e,s.rhs=r,s.lhs=n,s}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.operator===t.operator&&this.lhs.isEquivalent(t.lhs)&&this.rhs.isEquivalent(t.rhs)},e.prototype.visitExpression=function(t,e){return t.visitBinaryOperatorExpr(this,e)},e}(Uo),ls=function(t){function e(e,n,r,i){var o=t.call(this,r,i)||this;return o.receiver=e,o.name=n,o}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.name===t.name},e.prototype.visitExpression=function(t,e){return t.visitReadPropExpr(this,e)},e.prototype.set=function(t){return new Ko(this.receiver,this.name,t,null,this.sourceSpan)},e}(Uo),ps=function(t){function e(e,n,r,i){var o=t.call(this,r,i)||this;return o.receiver=e,o.index=n,o}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.receiver.isEquivalent(t.receiver)&&this.index.isEquivalent(t.index)},e.prototype.visitExpression=function(t,e){return t.visitReadKeyExpr(this,e)},e.prototype.set=function(t){return new Yo(this.receiver,this.index,t,null,this.sourceSpan)},e}(Uo),ds=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.entries=e,i}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&Ho(this.entries,t.entries)},e.prototype.visitExpression=function(t,e){return t.visitLiteralArrayExpr(this,e)},e}(Uo),hs=function(){function t(t,e,n){this.key=t,this.value=e,this.quoted=n}return t.prototype.isEquivalent=function(t){return this.key===t.key&&this.value.isEquivalent(t.value)},t}(),fs=function(t){function e(e,n,r){var i=t.call(this,n,r)||this;return i.entries=e,i.valueType=null,n&&(i.valueType=n.valueType),i}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&Ho(this.entries,t.entries)},e.prototype.visitExpression=function(t,e){return t.visitLiteralMapExpr(this,e)},e}(Uo),ms=function(t){function e(e,n){var r=t.call(this,e[e.length-1].type,n)||this;return r.parts=e,r}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&Ho(this.parts,t.parts)},e.prototype.visitExpression=function(t,e){return t.visitCommaExpr(this,e)},e}(Uo),gs=(new Go(Wo.This,null,null),new Go(Wo.Super,null,null),new Go(Wo.CatchError,null,null),new Go(Wo.CatchStack,null,null),new ts(null,null,null)),vs=new ts(null,Vo,null),ys={Final:0,Private:1,Exported:2};ys[ys.Final]="Final",ys[ys.Private]="Private",ys[ys.Exported]="Exported";var bs=function(){function t(t,e){this.modifiers=t||[],this.sourceSpan=e||null}return t.prototype.hasModifier=function(t){return-1!==this.modifiers.indexOf(t)},t}(),_s=function(t){function e(e,n,r,i,o){void 0===i&&(i=null);var s=t.call(this,i,o)||this;return s.name=e,s.value=n,s.type=r||n.type,s}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.name===t.name&&this.value.isEquivalent(t.value)},e.prototype.visitStatement=function(t,e){return t.visitDeclareVarStmt(this,e)},e}(bs),js=function(t){function e(e,n,r,i,o,s){void 0===o&&(o=null);var a=t.call(this,o,s)||this;return a.name=e,a.params=n,a.statements=r,a.type=i||null,a}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&Ho(this.params,t.params)&&Ho(this.statements,t.statements)},e.prototype.visitStatement=function(t,e){return t.visitDeclareFunctionStmt(this,e)},e}(bs),xs=function(t){function e(e,n){var r=t.call(this,null,n)||this;return r.expr=e,r}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.expr.isEquivalent(t.expr)},e.prototype.visitStatement=function(t,e){return t.visitExpressionStmt(this,e)},e}(bs),ws=function(t){function e(e,n){var r=t.call(this,null,n)||this;return r.value=e,r}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.value.isEquivalent(t.value)},e.prototype.visitStatement=function(t,e){return t.visitReturnStmt(this,e)},e}(bs),Os=function(){function t(t,e){this.modifiers=e,e||(this.modifiers=[]),this.type=t||null}return t.prototype.hasModifier=function(t){return-1!==this.modifiers.indexOf(t)},t}(),Cs=(function(t){function e(e,n,r){void 0===r&&(r=null);var i=t.call(this,n,r)||this;return i.name=e,i}Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return this.name===t.name}}(Os),function(t){function e(e,n,r,i,o){void 0===o&&(o=null);var s=t.call(this,i,o)||this;return s.name=e,s.params=n,s.body=r,s}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return this.name===t.name&&Ho(this.body,t.body)},e}(Os)),ks=function(t){function e(e,n,r,i){void 0===i&&(i=null);var o=t.call(this,r,i)||this;return o.name=e,o.body=n,o}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return this.name===t.name&&Ho(this.body,t.body)},e}(Os),Es=function(t){function e(e,n,r,i,o,s,a,c){void 0===a&&(a=null);var u=t.call(this,a,c)||this;return u.name=e,u.parent=n,u.fields=r,u.getters=i,u.constructorMethod=o,u.methods=s,u}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.name===t.name&&qo(this.parent,t.parent)&&Ho(this.fields,t.fields)&&Ho(this.getters,t.getters)&&this.constructorMethod.isEquivalent(t.constructorMethod)&&Ho(this.methods,t.methods)},e.prototype.visitStatement=function(t,e){return t.visitDeclareClassStmt(this,e)},e}(bs),Ss=function(t){function e(e,n,r,i){void 0===r&&(r=[]);var o=t.call(this,null,i)||this;return o.condition=e,o.trueCase=n,o.falseCase=r,o}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&this.condition.isEquivalent(t.condition)&&Ho(this.trueCase,t.trueCase)&&Ho(this.falseCase,t.falseCase)},e.prototype.visitStatement=function(t,e){return t.visitIfStmt(this,e)},e}(bs),Ps=(function(t){function e(e,n){var r=t.call(this,null,n)||this;return r.comment=e,r}Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e},e.prototype.visitStatement=function(t,e){return t.visitCommentStmt(this,e)}}(bs),function(t){function e(e,n,r){var i=t.call(this,null,r)||this;return i.bodyStmts=e,i.catchStmts=n,i}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof e&&Ho(this.bodyStmts,t.bodyStmts)&&Ho(this.catchStmts,t.catchStmts)},e.prototype.visitStatement=function(t,e){return t.visitTryCatchStmt(this,e)},e}(bs)),Ms=function(t){function e(e,n){var r=t.call(this,null,n)||this;return r.error=e,r}return Object(r.b)(e,t),e.prototype.isEquivalent=function(t){return t instanceof Ps&&this.error.isEquivalent(t.error)},e.prototype.visitStatement=function(t,e){return t.visitThrowStmt(this,e)},e}(bs),As=function(){function t(){}return t.prototype.transformExpr=function(t,e){return t},t.prototype.transformStmt=function(t,e){return t},t.prototype.visitReadVarExpr=function(t,e){return this.transformExpr(t,e)},t.prototype.visitWriteVarExpr=function(t,e){return this.transformExpr(new Xo(t.name,t.value.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitWriteKeyExpr=function(t,e){return this.transformExpr(new Yo(t.receiver.visitExpression(this,e),t.index.visitExpression(this,e),t.value.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitWritePropExpr=function(t,e){return this.transformExpr(new Ko(t.receiver.visitExpression(this,e),t.name,t.value.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitInvokeMethodExpr=function(t,e){var n=t.builtin||t.name;return this.transformExpr(new Qo(t.receiver.visitExpression(this,e),n,this.visitAllExpressions(t.args,e),t.type,t.sourceSpan),e)},t.prototype.visitInvokeFunctionExpr=function(t,e){return this.transformExpr(new Zo(t.fn.visitExpression(this,e),this.visitAllExpressions(t.args,e),t.type,t.sourceSpan),e)},t.prototype.visitInstantiateExpr=function(t,e){return this.transformExpr(new Jo(t.classExpr.visitExpression(this,e),this.visitAllExpressions(t.args,e),t.type,t.sourceSpan),e)},t.prototype.visitLiteralExpr=function(t,e){return this.transformExpr(t,e)},t.prototype.visitExternalExpr=function(t,e){return this.transformExpr(t,e)},t.prototype.visitConditionalExpr=function(t,e){return this.transformExpr(new rs(t.condition.visitExpression(this,e),t.trueCase.visitExpression(this,e),t.falseCase.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitNotExpr=function(t,e){return this.transformExpr(new is(t.condition.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitAssertNotNullExpr=function(t,e){return this.transformExpr(new os(t.condition.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitCastExpr=function(t,e){return this.transformExpr(new ss(t.value.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitFunctionExpr=function(t,e){return this.transformExpr(new cs(t.params,this.visitAllStatements(t.statements,e),t.type,t.sourceSpan),e)},t.prototype.visitBinaryOperatorExpr=function(t,e){return this.transformExpr(new us(t.operator,t.lhs.visitExpression(this,e),t.rhs.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitReadPropExpr=function(t,e){return this.transformExpr(new ls(t.receiver.visitExpression(this,e),t.name,t.type,t.sourceSpan),e)},t.prototype.visitReadKeyExpr=function(t,e){return this.transformExpr(new ps(t.receiver.visitExpression(this,e),t.index.visitExpression(this,e),t.type,t.sourceSpan),e)},t.prototype.visitLiteralArrayExpr=function(t,e){return this.transformExpr(new ds(this.visitAllExpressions(t.entries,e),t.type,t.sourceSpan),e)},t.prototype.visitLiteralMapExpr=function(t,e){var n=this,r=t.entries.map(function(t){return new hs(t.key,t.value.visitExpression(n,e),t.quoted)}),i=new Lo(t.valueType,null);return this.transformExpr(new fs(r,i,t.sourceSpan),e)},t.prototype.visitCommaExpr=function(t,e){return this.transformExpr(new ms(this.visitAllExpressions(t.parts,e),t.sourceSpan),e)},t.prototype.visitAllExpressions=function(t,e){var n=this;return t.map(function(t){return t.visitExpression(n,e)})},t.prototype.visitDeclareVarStmt=function(t,e){return this.transformStmt(new _s(t.name,t.value.visitExpression(this,e),t.type,t.modifiers,t.sourceSpan),e)},t.prototype.visitDeclareFunctionStmt=function(t,e){return this.transformStmt(new js(t.name,t.params,this.visitAllStatements(t.statements,e),t.type,t.modifiers,t.sourceSpan),e)},t.prototype.visitExpressionStmt=function(t,e){return this.transformStmt(new xs(t.expr.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitReturnStmt=function(t,e){return this.transformStmt(new ws(t.value.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitDeclareClassStmt=function(t,e){var n=this,r=t.parent.visitExpression(this,e),i=t.getters.map(function(t){return new ks(t.name,n.visitAllStatements(t.body,e),t.type,t.modifiers)}),o=t.constructorMethod&&new Cs(t.constructorMethod.name,t.constructorMethod.params,this.visitAllStatements(t.constructorMethod.body,e),t.constructorMethod.type,t.constructorMethod.modifiers),s=t.methods.map(function(t){return new Cs(t.name,t.params,n.visitAllStatements(t.body,e),t.type,t.modifiers)});return this.transformStmt(new Es(t.name,r,t.fields,i,o,s,t.modifiers,t.sourceSpan),e)},t.prototype.visitIfStmt=function(t,e){return this.transformStmt(new Ss(t.condition.visitExpression(this,e),this.visitAllStatements(t.trueCase,e),this.visitAllStatements(t.falseCase,e),t.sourceSpan),e)},t.prototype.visitTryCatchStmt=function(t,e){return this.transformStmt(new Ps(this.visitAllStatements(t.bodyStmts,e),this.visitAllStatements(t.catchStmts,e),t.sourceSpan),e)},t.prototype.visitThrowStmt=function(t,e){return this.transformStmt(new Ms(t.error.visitExpression(this,e),t.sourceSpan),e)},t.prototype.visitCommentStmt=function(t,e){return this.transformStmt(t,e)},t.prototype.visitAllStatements=function(t,e){var n=this;return t.map(function(t){return t.visitStatement(n,e)})},t}(),Ts=function(){function t(){}return t.prototype.visitType=function(t,e){return t},t.prototype.visitExpression=function(t,e){return t.type&&t.type.visitType(this,e),t},t.prototype.visitBuiltintType=function(t,e){return this.visitType(t,e)},t.prototype.visitExpressionType=function(t,e){return t.value.visitExpression(this,e),this.visitType(t,e)},t.prototype.visitArrayType=function(t,e){return this.visitType(t,e)},t.prototype.visitMapType=function(t,e){return this.visitType(t,e)},t.prototype.visitReadVarExpr=function(t,e){return this.visitExpression(t,e)},t.prototype.visitWriteVarExpr=function(t,e){return t.value.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitWriteKeyExpr=function(t,e){return t.receiver.visitExpression(this,e),t.index.visitExpression(this,e),t.value.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitWritePropExpr=function(t,e){return t.receiver.visitExpression(this,e),t.value.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitInvokeMethodExpr=function(t,e){return t.receiver.visitExpression(this,e),this.visitAllExpressions(t.args,e),this.visitExpression(t,e)},t.prototype.visitInvokeFunctionExpr=function(t,e){return t.fn.visitExpression(this,e),this.visitAllExpressions(t.args,e),this.visitExpression(t,e)},t.prototype.visitInstantiateExpr=function(t,e){return t.classExpr.visitExpression(this,e),this.visitAllExpressions(t.args,e),this.visitExpression(t,e)},t.prototype.visitLiteralExpr=function(t,e){return this.visitExpression(t,e)},t.prototype.visitExternalExpr=function(t,e){var n=this;return t.typeParams&&t.typeParams.forEach(function(t){return t.visitType(n,e)}),this.visitExpression(t,e)},t.prototype.visitConditionalExpr=function(t,e){return t.condition.visitExpression(this,e),t.trueCase.visitExpression(this,e),t.falseCase.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitNotExpr=function(t,e){return t.condition.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitAssertNotNullExpr=function(t,e){return t.condition.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitCastExpr=function(t,e){return t.value.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitFunctionExpr=function(t,e){return this.visitAllStatements(t.statements,e),this.visitExpression(t,e)},t.prototype.visitBinaryOperatorExpr=function(t,e){return t.lhs.visitExpression(this,e),t.rhs.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitReadPropExpr=function(t,e){return t.receiver.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitReadKeyExpr=function(t,e){return t.receiver.visitExpression(this,e),t.index.visitExpression(this,e),this.visitExpression(t,e)},t.prototype.visitLiteralArrayExpr=function(t,e){return this.visitAllExpressions(t.entries,e),this.visitExpression(t,e)},t.prototype.visitLiteralMapExpr=function(t,e){var n=this;return t.entries.forEach(function(t){return t.value.visitExpression(n,e)}),this.visitExpression(t,e)},t.prototype.visitCommaExpr=function(t,e){return this.visitAllExpressions(t.parts,e),this.visitExpression(t,e)},t.prototype.visitAllExpressions=function(t,e){var n=this;t.forEach(function(t){return t.visitExpression(n,e)})},t.prototype.visitDeclareVarStmt=function(t,e){return t.value.visitExpression(this,e),t.type&&t.type.visitType(this,e),t},t.prototype.visitDeclareFunctionStmt=function(t,e){return this.visitAllStatements(t.statements,e),t.type&&t.type.visitType(this,e),t},t.prototype.visitExpressionStmt=function(t,e){return t.expr.visitExpression(this,e),t},t.prototype.visitReturnStmt=function(t,e){return t.value.visitExpression(this,e),t},t.prototype.visitDeclareClassStmt=function(t,e){var n=this;return t.parent.visitExpression(this,e),t.getters.forEach(function(t){return n.visitAllStatements(t.body,e)}),t.constructorMethod&&this.visitAllStatements(t.constructorMethod.body,e),t.methods.forEach(function(t){return n.visitAllStatements(t.body,e)}),t},t.prototype.visitIfStmt=function(t,e){return t.condition.visitExpression(this,e),this.visitAllStatements(t.trueCase,e),this.visitAllStatements(t.falseCase,e),t},t.prototype.visitTryCatchStmt=function(t,e){return this.visitAllStatements(t.bodyStmts,e),this.visitAllStatements(t.catchStmts,e),t},t.prototype.visitThrowStmt=function(t,e){return t.error.visitExpression(this,e),t},t.prototype.visitCommentStmt=function(t,e){return t},t.prototype.visitAllStatements=function(t,e){var n=this;t.forEach(function(t){return t.visitStatement(n,e)})},t}();function Ds(t){var e=new Is;return e.visitAllStatements(t,null),e.varNames}var Is=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.varNames=new Set,e}return Object(r.b)(e,t),e.prototype.visitDeclareFunctionStmt=function(t,e){return t},e.prototype.visitDeclareClassStmt=function(t,e){return t},e.prototype.visitReadVarExpr=function(t,e){return t.name&&this.varNames.add(t.name),null},e}(Ts);!function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.externalReferences=[],e}Object(r.b)(e,t),e.prototype.visitExternalExpr=function(e,n){return this.externalReferences.push(e.value),t.prototype.visitExternalExpr.call(this,e,n)}}(Ts);function Rs(t,e){if(!e)return t;var n=new Ls(e);return t.visitStatement(n,null)}function Ns(t,e){if(!e)return t;var n=new Ls(e);return t.visitExpression(n,null)}var Ls=function(t){function e(e){var n=t.call(this)||this;return n.sourceSpan=e,n}return Object(r.b)(e,t),e.prototype._clone=function(t){var e=Object.create(t.constructor.prototype);for(var n in t)e[n]=t[n];return e},e.prototype.transformExpr=function(t,e){return t.sourceSpan||((t=this._clone(t)).sourceSpan=this.sourceSpan),t},e.prototype.transformStmt=function(t,e){return t.sourceSpan||((t=this._clone(t)).sourceSpan=this.sourceSpan),t},e}(As);function Fs(t,e,n){return new Go(t,e,n)}function Vs(t,e,n){return void 0===e&&(e=null),new es(t,null,e,n)}function zs(t,e,n){return void 0===e&&(e=null),void 0===n&&(n=null),null!=t?Bs(Vs(t,e,null),n):null}function Bs(t,e){return void 0===e&&(e=null),new Ro(t,e)}function qs(t,e,n){return new ds(t,e,n)}function Hs(t,e){return void 0===e&&(e=null),new fs(t.map(function(t){return new hs(t.key,t.value,t.quoted)}),e,null)}function Us(t,e,n,r){return new cs(t,e,n,r)}function Ws(t,e,n){return new ts(t,e,n)}var Gs=function(t){function e(e,n){return t.call(this,n,e)||this}return Object(r.b)(e,t),e}(ar),Xs=function(){return function(t,e){var n=this;this.reflector=t,this.component=e,this.errors=[],this.viewQueries=(r=e,i=1,o=new Map,r.viewQueries&&r.viewQueries.forEach(function(t){return Js(o,{meta:t,queryId:i++})}),o),this.viewProviders=new Map,e.viewProviders.forEach(function(t){null==n.viewProviders.get(Et(t.token))&&n.viewProviders.set(Et(t.token),!0)});var r,i,o}}(),Ys=function(){function t(t,e,n,r,i,o,s,a,c){var u=this;this.viewContext=t,this._parent=e,this._isViewRoot=n,this._directiveAsts=r,this._sourceSpan=c,this._transformedProviders=new Map,this._seenProviders=new Map,this._queriedTokens=new Map,this.transformedHasViewContainer=!1,this._attrs={},i.forEach(function(t){return u._attrs[t.name]=t.value});var l,p,d,h,f,m,g,v=r.map(function(t){return t.directive});if(this._allProviders=(l=v,p=c,d=t.errors,h=new Map,l.forEach(function(t){Zs([{token:{identifier:t.type},useClass:t.type}],t.isComponent?ut.Component:ut.Directive,!0,p,d,h)}),l.filter(function(t){return t.isComponent}).concat(l.filter(function(t){return!t.isComponent})).forEach(function(t){Zs(t.providers,ut.PublicService,!1,p,d,h),Zs(t.viewProviders,ut.PrivateService,!1,p,d,h)}),h),this._contentQueries=(f=v,m=a,g=new Map,f.forEach(function(t,e){t.queries&&t.queries.forEach(function(t){return Js(g,{meta:t,queryId:m++})})}),g),Array.from(this._allProviders.values()).forEach(function(t){u._addQueryReadsTo(t.token,t.token,u._queriedTokens)}),s){var y=go(this.viewContext.reflector,fo.TemplateRef);this._addQueryReadsTo(y,y,this._queriedTokens)}o.forEach(function(t){var e=t.value||go(u.viewContext.reflector,fo.ElementRef);u._addQueryReadsTo({value:t.name},e,u._queriedTokens)}),this._queriedTokens.get(this.viewContext.reflector.resolveExternalReference(fo.ViewContainerRef))&&(this.transformedHasViewContainer=!0),Array.from(this._allProviders.values()).forEach(function(t){(t.eager||u._queriedTokens.get(Et(t.token)))&&u._getOrCreateLocalProvider(t.providerType,t.token,!0)})}return t.prototype.afterElement=function(){var t=this;Array.from(this._allProviders.values()).forEach(function(e){t._getOrCreateLocalProvider(e.providerType,e.token,!1)})},Object.defineProperty(t.prototype,"transformProviders",{get:function(){var t=[],e=[];return this._transformedProviders.forEach(function(n){n.eager?e.push(n):t.push(n)}),t.concat(e)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"transformedDirectiveAsts",{get:function(){var t=this.transformProviders.map(function(t){return t.token.identifier}),e=this._directiveAsts.slice();return e.sort(function(e,n){return t.indexOf(e.directive.type)-t.indexOf(n.directive.type)}),e},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"queryMatches",{get:function(){var t=[];return this._queriedTokens.forEach(function(e){t.push.apply(t,e)}),t},enumerable:!0,configurable:!0}),t.prototype._addQueryReadsTo=function(t,e,n){this._getQueriesFor(t).forEach(function(t){var r=t.meta.read||e,i=Et(r),o=n.get(i);o||(o=[],n.set(i,o)),o.push({queryId:t.queryId,value:r})})},t.prototype._getQueriesFor=function(t){for(var e,n=[],r=this,i=0;null!==r;)(e=r._contentQueries.get(Et(t)))&&n.push.apply(n,e.filter(function(t){return t.meta.descendants||i<=1})),r._directiveAsts.length>0&&i++,r=r._parent;return(e=this.viewContext.viewQueries.get(Et(t)))&&n.push.apply(n,e),n},t.prototype._getOrCreateLocalProvider=function(t,e,n){var r=this,i=this._allProviders.get(Et(e));if(!i||(t===ut.Directive||t===ut.PublicService)&&i.providerType===ut.PrivateService||(t===ut.PrivateService||t===ut.PublicService)&&i.providerType===ut.Builtin)return null;var o=this._transformedProviders.get(Et(e));if(o)return o;if(null!=this._seenProviders.get(Et(e)))return this.viewContext.errors.push(new Gs("Cannot instantiate cyclic dependency! "+kt(e),this._sourceSpan)),null;this._seenProviders.set(Et(e),!0);var s=i.providers.map(function(t){var e=t.useValue,o=t.useExisting,s=void 0;if(null!=t.useExisting){var a=r._getDependency(i.providerType,{token:t.useExisting},n);null!=a.token?o=a.token:(o=null,e=a.value)}else if(t.useFactory){s=(t.deps||t.useFactory.diDeps).map(function(t){return r._getDependency(i.providerType,t,n)})}else if(t.useClass){s=(t.deps||t.useClass.diDeps).map(function(t){return r._getDependency(i.providerType,t,n)})}return $s(t,{useExisting:o,useValue:e,deps:s})});return o=Qs(i,{eager:n,providers:s}),this._transformedProviders.set(Et(e),o),o},t.prototype._getLocalDependency=function(t,e,n){if(void 0===n&&(n=!1),e.isAttribute){var r=this._attrs[e.token.value];return{isValue:!0,value:null==r?null:r}}if(null!=e.token){if(t===ut.Directive||t===ut.Component){if(Et(e.token)===this.viewContext.reflector.resolveExternalReference(fo.Renderer)||Et(e.token)===this.viewContext.reflector.resolveExternalReference(fo.ElementRef)||Et(e.token)===this.viewContext.reflector.resolveExternalReference(fo.ChangeDetectorRef)||Et(e.token)===this.viewContext.reflector.resolveExternalReference(fo.TemplateRef))return e;Et(e.token)===this.viewContext.reflector.resolveExternalReference(fo.ViewContainerRef)&&(this.transformedHasViewContainer=!0)}if(Et(e.token)===this.viewContext.reflector.resolveExternalReference(fo.Injector))return e;if(null!=this._getOrCreateLocalProvider(t,e.token,n))return e}return null},t.prototype._getDependency=function(t,e,n){void 0===n&&(n=!1);var r=this,i=n,o=null;if(e.isSkipSelf||(o=this._getLocalDependency(t,e,n)),e.isSelf)!o&&e.isOptional&&(o={isValue:!0,value:null});else{for(;!o&&r._parent;){var s=r;r=r._parent,s._isViewRoot&&(i=!1),o=r._getLocalDependency(ut.PublicService,e,i)}o||(o=!e.isHost||this.viewContext.component.isHost||this.viewContext.component.type.reference===Et(e.token)||null!=this.viewContext.viewProviders.get(Et(e.token))?e:e.isOptional?o={isValue:!0,value:null}:null)}return o||this.viewContext.errors.push(new Gs("No provider for "+kt(e.token),this._sourceSpan)),o},t}(),Ks=function(){function t(t,e,n,r){var i=this;this.reflector=t,this._transformedProviders=new Map,this._seenProviders=new Map,this._errors=[],this._allProviders=new Map,e.transitiveModule.modules.forEach(function(t){Zs([{token:{identifier:t},useClass:t}],ut.PublicService,!0,r,i._errors,i._allProviders)}),Zs(e.transitiveModule.providers.map(function(t){return t.provider}).concat(n),ut.PublicService,!1,r,this._errors,this._allProviders)}return t.prototype.parse=function(){var t=this;if(Array.from(this._allProviders.values()).forEach(function(e){t._getOrCreateLocalProvider(e.token,e.eager)}),this._errors.length>0){var e=this._errors.join("\n");throw new Error("Provider parse errors:\n"+e)}var n=[],r=[];return this._transformedProviders.forEach(function(t){t.eager?r.push(t):n.push(t)}),n.concat(r)},t.prototype._getOrCreateLocalProvider=function(t,e){var n=this,r=this._allProviders.get(Et(t));if(!r)return null;var i=this._transformedProviders.get(Et(t));if(i)return i;if(null!=this._seenProviders.get(Et(t)))return this._errors.push(new Gs("Cannot instantiate cyclic dependency! "+kt(t),r.sourceSpan)),null;this._seenProviders.set(Et(t),!0);var o=r.providers.map(function(t){var i=t.useValue,o=t.useExisting,s=void 0;if(null!=t.useExisting){var a=n._getDependency({token:t.useExisting},e,r.sourceSpan);null!=a.token?o=a.token:(o=null,i=a.value)}else if(t.useFactory){s=(t.deps||t.useFactory.diDeps).map(function(t){return n._getDependency(t,e,r.sourceSpan)})}else if(t.useClass){s=(t.deps||t.useClass.diDeps).map(function(t){return n._getDependency(t,e,r.sourceSpan)})}return $s(t,{useExisting:o,useValue:i,deps:s})});return i=Qs(r,{eager:e,providers:o}),this._transformedProviders.set(Et(t),i),i},t.prototype._getDependency=function(t,e,n){void 0===e&&(e=!1);var r=!1;t.isSkipSelf||null==t.token||(Et(t.token)===this.reflector.resolveExternalReference(fo.Injector)||Et(t.token)===this.reflector.resolveExternalReference(fo.ComponentFactoryResolver)?r=!0:null!=this._getOrCreateLocalProvider(t.token,e)&&(r=!0));var i=t;return t.isSelf&&!r&&(t.isOptional?i={isValue:!0,value:null}:this._errors.push(new Gs("No provider for "+kt(t.token),n))),i},t}();function $s(t,e){var n=e.useExisting,r=e.useValue,i=e.deps;return{token:t.token,useClass:t.useClass,useExisting:n,useFactory:t.useFactory,useValue:r,deps:i,multi:t.multi}}function Qs(t,e){var n=e.eager,r=e.providers;return new ct(t.token,t.multiProvider,t.eager||n,r,t.providerType,t.lifecycleHooks,t.sourceSpan)}function Zs(t,e,n,r,i,o){t.forEach(function(t){var s=o.get(Et(t.token));if(null!=s&&!!s.multiProvider!=!!t.multi&&i.push(new Gs("Mixing multi and non multi provider is not possible for token "+kt(s.token),r)),s)t.multi||(s.providers.length=0),s.providers.push(t);else{var a=t.token.identifier&&t.token.identifier.lifecycleHooks?t.token.identifier.lifecycleHooks:[],c=!(t.useClass||t.useExisting||t.useFactory);s=new ct(t.token,!!t.multi,n||c,[t],e,a,r),o.set(Et(t.token),s)}})}function Js(t,e){e.meta.selectors.forEach(function(n){var r=t.get(Et(n));r||(r=[],t.set(Et(n),r)),r.push(e)})}function ta(t,e,n){return void 0===n&&(n=null),R(e,new ea(t),n)}var ea=function(){function t(t){this.ctx=t}return t.prototype.visitArray=function(t,e){var n=this;return qs(t.map(function(t){return R(t,n,null)}),e)},t.prototype.visitStringMap=function(t,e){var n=this,r=[],i=new Set(t&&t.$quoted$);return Object.keys(t).forEach(function(e){r.push(new hs(e,R(t[e],n,null),i.has(e)))}),new fs(r,e)},t.prototype.visitPrimitive=function(t,e){return Ws(t,e)},t.prototype.visitOther=function(t,e){return t instanceof Uo?t:this.ctx.importExpr(t)},t}();function na(t,e){var n=0;e.eager||(n|=4096),e.providerType===ut.PrivateService&&(n|=8192),e.lifecycleHooks.forEach(function(t){t!==vo.OnDestroy&&e.providerType!==ut.Directive&&e.providerType!==ut.Component||(n|=sa(t))});var r=e.multiProvider?function(t,e,n){var r=[],i=[],o=n.map(function(e,n){var r;if(e.useClass){var i=s(n,e.deps||e.useClass.diDeps);r=t.importExpr(e.useClass.reference).instantiate(i)}else if(e.useFactory){var i=s(n,e.deps||e.useFactory.diDeps);r=t.importExpr(e.useFactory.reference).callFn(i)}else if(e.useExisting){var i=s(n,[{token:e.useExisting}]);r=i[0]}else r=ta(t,e.useValue);return r});return{providerExpr:Us(i,[new ws(qs(o))],Vo),flags:1024|e,depsExpr:qs(r)};function s(e,n){return n.map(function(n,o){var s="p"+e+"_"+o;return i.push(new as(s,Fo)),r.push(oa(t,n)),Fs(s)})}}(t,n,e.providers):ra(t,n,e.providerType,e.providers[0]);return{providerExpr:r.providerExpr,flags:r.flags,depsExpr:r.depsExpr,tokenExpr:ia(t,e.token)}}function ra(t,e,n,r){var i,o;return n===ut.Directive||n===ut.Component?(i=t.importExpr(r.useClass.reference),e|=16384,o=r.deps||r.useClass.diDeps):r.useClass?(i=t.importExpr(r.useClass.reference),e|=512,o=r.deps||r.useClass.diDeps):r.useFactory?(i=t.importExpr(r.useFactory.reference),e|=1024,o=r.deps||r.useFactory.diDeps):r.useExisting?(i=gs,e|=2048,o=[{token:r.useExisting}]):(i=ta(t,r.useValue),e|=256,o=[]),{providerExpr:i,flags:e,depsExpr:qs(o.map(function(e){return oa(t,e)}))}}function ia(t,e){return e.identifier?t.importExpr(e.identifier.reference):Ws(e.value)}function oa(t,e){var n=e.isValue?ta(t,e.value):ia(t,e.token),r=0;return e.isSkipSelf&&(r|=1),e.isOptional&&(r|=2),e.isValue&&(r|=8),0===r?n:qs([Ws(r),n])}function sa(t){var e=0;switch(t){case vo.AfterContentChecked:e=2097152;break;case vo.AfterContentInit:e=1048576;break;case vo.AfterViewChecked:e=8388608;break;case vo.AfterViewInit:e=4194304;break;case vo.DoCheck:e=262144;break;case vo.OnChanges:e=524288;break;case vo.OnDestroy:e=131072;break;case vo.OnInit:e=65536}return e}function aa(t,e,n,r){var i=r.map(function(t){return e.importExpr(t.componentFactory)}),o=go(t,fo.ComponentFactoryResolver),s={diDeps:[{isValue:!0,value:qs(i)},{token:o,isSkipSelf:!0,isOptional:!0},{token:go(t,fo.NgModuleRef)}],lifecycleHooks:[],reference:t.resolveExternalReference(fo.CodegenComponentFactoryResolver)},a=ra(e,n,ut.PrivateService,{token:o,multi:!1,useClass:s});return{providerExpr:a.providerExpr,flags:a.flags,depsExpr:a.depsExpr,tokenExpr:ia(e,o)}}var ca=function(){return function(t){this.ngModuleFactoryVar=t}}(),ua=Fs("_l"),la=function(){function t(t){this.reflector=t}return t.prototype.compile=function(t,e,n){var r,i,o,s,a,c=(r="NgModule",i=e.type,o=jt(i),s=null!=o?"in "+r+" "+_t(i)+" in "+o:"in "+r+" "+_t(i),a=new ir("",s),new or(new rr(a,-1,-1,-1),new rr(a,-1,-1,-1))),u=e.transitiveModule.entryComponents,l=e.bootstrapComponents,p=new Ks(this.reflector,e,n,c),d=[aa(this.reflector,t,0,u)].concat(p.parse().map(function(e){return na(t,e)})).map(function(t){var e=t.providerExpr,n=t.depsExpr,r=t.flags,i=t.tokenExpr;return Vs(fo.moduleProviderDef).callFn([Ws(r),i,e,n])}),h=Vs(fo.moduleDef).callFn([qs(d)]),f=Us([new as(ua.name)],[new ws(h)],Vo),m=_t(e.type)+"NgFactory";if(this._createNgModuleFactory(t,e.type.reference,Vs(fo.createModuleFactory).callFn([t.importExpr(e.type.reference),qs(l.map(function(e){return t.importExpr(e.reference)})),f])),e.id){var g=Vs(fo.RegisterModuleFactoryFn).callFn([Ws(e.id),Fs(m)]).toStmt();t.statements.push(g)}return new ca(m)},t.prototype.createStub=function(t,e){this._createNgModuleFactory(t,e,gs)},t.prototype._createNgModuleFactory=function(t,e,n){var r=Fs(_t({reference:e})+"NgFactory").set(n).toDeclStmt(zs(fo.NgModuleFactory,[Bs(t.importExpr(e))],[Ao.Const]),[ys.Final,ys.Exported]);t.statements.push(r)},t}(),pa=function(){function t(t){this._reflector=t}return t.prototype.isNgModule=function(t){return this._reflector.annotations(t).some(_.isTypeOf)},t.prototype.resolve=function(t,e){void 0===e&&(e=!0);var n=Ee(this._reflector.annotations(t),_.isTypeOf);if(n)return n;if(e)throw new Error("No NgModule metadata found for '"+Y(t)+"'.");return null},t}(),da=function(){function t(t){void 0===t&&(t=null),this.file=t,this.sourcesContent=new Map,this.lines=[],this.lastCol0=0,this.hasMappings=!1}return t.prototype.addSource=function(t,e){return void 0===e&&(e=null),this.sourcesContent.has(t)||this.sourcesContent.set(t,e),this},t.prototype.addLine=function(){return this.lines.push([]),this.lastCol0=0,this},t.prototype.addMapping=function(t,e,n,r){if(!this.currentLine)throw new Error("A line must be added before mappings can be added");if(null!=e&&!this.sourcesContent.has(e))throw new Error('Unknown source file "'+e+'"');if(null==t)throw new Error("The column in the generated code must be provided");if(t<this.lastCol0)throw new Error("Mapping should be added in output order");if(e&&(null==n||null==r))throw new Error("The source location must be provided when a source url is provided");return this.hasMappings=!0,this.lastCol0=t,this.currentLine.push({col0:t,sourceUrl:e,sourceLine0:n,sourceCol0:r}),this},Object.defineProperty(t.prototype,"currentLine",{get:function(){return this.lines.slice(-1)[0]},enumerable:!0,configurable:!0}),t.prototype.toJSON=function(){var t=this;if(!this.hasMappings)return null;var e=new Map,n=[],r=[];Array.from(this.sourcesContent.keys()).forEach(function(i,o){e.set(i,o),n.push(i),r.push(t.sourcesContent.get(i)||null)});var i="",o=0,s=0,a=0,c=0;return this.lines.forEach(function(t){o=0,i+=t.map(function(t){var n=ha(t.col0-o);return o=t.col0,null!=t.sourceUrl&&(n+=ha(e.get(t.sourceUrl)-s),s=e.get(t.sourceUrl),n+=ha(t.sourceLine0-a),a=t.sourceLine0,n+=ha(t.sourceCol0-c),c=t.sourceCol0),n}).join(","),i+=";"}),i=i.slice(0,-1),{file:this.file||"",version:3,sourceRoot:"",sources:n,sourcesContent:r,mappings:i}},t.prototype.toJsComment=function(){return this.hasMappings?"//# sourceMappingURL=data:application/json;base64,"+function(t){var e="";t=X(t);for(var n=0;n<t.length;){var r=t.charCodeAt(n++),i=t.charCodeAt(n++),o=t.charCodeAt(n++);e+=ma(r>>2),e+=ma((3&r)<<4|(isNaN(i)?0:i>>4)),e+=isNaN(i)?"=":ma((15&i)<<2|o>>6),e+=isNaN(i)||isNaN(o)?"=":ma(63&o)}return e}(JSON.stringify(this,null,0)):""},t}();function ha(t){t=t<0?1+(-t<<1):t<<1;var e="";do{var n=31&t;(t>>=5)>0&&(n|=32),e+=ma(n)}while(t>0);return e}var fa="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";function ma(t){if(t<0||t>=64)throw new Error("Can only encode value in the range [0, 63]");return fa[t]}var ga=/'|\\|\n|\r|\$/g,va=/^[$A-Z_][0-9A-Z_$]*$/i,ya="  ",ba=Fs("error",null,null),_a=Fs("stack",null,null),ja=function(){return function(t){this.indent=t,this.partsLength=0,this.parts=[],this.srcSpans=[]}}(),xa=function(){function t(t){this._indent=t,this._classes=[],this._preambleLineCount=0,this._lines=[new ja(t)]}return t.createRoot=function(){return new t(0)},Object.defineProperty(t.prototype,"_currentLine",{get:function(){return this._lines[this._lines.length-1]},enumerable:!0,configurable:!0}),t.prototype.println=function(t,e){void 0===e&&(e=""),this.print(t||null,e,!0)},t.prototype.lineIsEmpty=function(){return 0===this._currentLine.parts.length},t.prototype.lineLength=function(){return this._currentLine.indent*ya.length+this._currentLine.partsLength},t.prototype.print=function(t,e,n){void 0===n&&(n=!1),e.length>0&&(this._currentLine.parts.push(e),this._currentLine.partsLength+=e.length,this._currentLine.srcSpans.push(t&&t.sourceSpan||null)),n&&this._lines.push(new ja(this._indent))},t.prototype.removeEmptyLastLine=function(){this.lineIsEmpty()&&this._lines.pop()},t.prototype.incIndent=function(){this._indent++,this.lineIsEmpty()&&(this._currentLine.indent=this._indent)},t.prototype.decIndent=function(){this._indent--,this.lineIsEmpty()&&(this._currentLine.indent=this._indent)},t.prototype.pushClass=function(t){this._classes.push(t)},t.prototype.popClass=function(){return this._classes.pop()},Object.defineProperty(t.prototype,"currentClass",{get:function(){return this._classes.length>0?this._classes[this._classes.length-1]:null},enumerable:!0,configurable:!0}),t.prototype.toSource=function(){return this.sourceLines.map(function(t){return t.parts.length>0?Ca(t.indent)+t.parts.join(""):""}).join("\n")},t.prototype.toSourceMapGenerator=function(t,e){void 0===e&&(e=0);for(var n=new da(t),r=!1,i=function(){r||(n.addSource(t," ").addMapping(0,t,0,0),r=!0)},o=0;o<e;o++)n.addLine(),i();return this.sourceLines.forEach(function(t,e){n.addLine();for(var o=t.srcSpans,s=t.parts,a=t.indent*ya.length,c=0;c<o.length&&!o[c];)a+=s[c].length,c++;for(c<o.length&&0===e&&0===a?r=!0:i();c<o.length;){var u=o[c],l=u.start.file,p=u.start.line,d=u.start.col;for(n.addSource(l.url,l.content).addMapping(a,l.url,p,d),a+=s[c].length,c++;c<o.length&&(u===o[c]||!o[c]);)a+=s[c].length,c++}}),n},t.prototype.setPreambleLineCount=function(t){return this._preambleLineCount=t},t.prototype.spanOf=function(t,e){var n=this._lines[t-this._preambleLineCount];if(n)for(var r=e-Ca(n.indent).length,i=0;i<n.parts.length;i++){var o=n.parts[i];if(o.length>r)return n.srcSpans[i];r-=o.length}return null},Object.defineProperty(t.prototype,"sourceLines",{get:function(){return this._lines.length&&0===this._lines[this._lines.length-1].parts.length?this._lines.slice(0,-1):this._lines},enumerable:!0,configurable:!0}),t}(),wa=function(){function t(t){this._escapeDollarInStrings=t}return t.prototype.visitExpressionStmt=function(t,e){return t.expr.visitExpression(this,e),e.println(t,";"),null},t.prototype.visitReturnStmt=function(t,e){return e.print(t,"return "),t.value.visitExpression(this,e),e.println(t,";"),null},t.prototype.visitIfStmt=function(t,e){e.print(t,"if ("),t.condition.visitExpression(this,e),e.print(t,") {");var n=null!=t.falseCase&&t.falseCase.length>0;return t.trueCase.length<=1&&!n?(e.print(t," "),this.visitAllStatements(t.trueCase,e),e.removeEmptyLastLine(),e.print(t," ")):(e.println(),e.incIndent(),this.visitAllStatements(t.trueCase,e),e.decIndent(),n&&(e.println(t,"} else {"),e.incIndent(),this.visitAllStatements(t.falseCase,e),e.decIndent())),e.println(t,"}"),null},t.prototype.visitThrowStmt=function(t,e){return e.print(t,"throw "),t.error.visitExpression(this,e),e.println(t,";"),null},t.prototype.visitCommentStmt=function(t,e){return t.comment.split("\n").forEach(function(n){e.println(t,"// "+n)}),null},t.prototype.visitWriteVarExpr=function(t,e){var n=e.lineIsEmpty();return n||e.print(t,"("),e.print(t,t.name+" = "),t.value.visitExpression(this,e),n||e.print(t,")"),null},t.prototype.visitWriteKeyExpr=function(t,e){var n=e.lineIsEmpty();return n||e.print(t,"("),t.receiver.visitExpression(this,e),e.print(t,"["),t.index.visitExpression(this,e),e.print(t,"] = "),t.value.visitExpression(this,e),n||e.print(t,")"),null},t.prototype.visitWritePropExpr=function(t,e){var n=e.lineIsEmpty();return n||e.print(t,"("),t.receiver.visitExpression(this,e),e.print(t,"."+t.name+" = "),t.value.visitExpression(this,e),n||e.print(t,")"),null},t.prototype.visitInvokeMethodExpr=function(t,e){t.receiver.visitExpression(this,e);var n=t.name;return null!=t.builtin&&null==(n=this.getBuiltinMethodName(t.builtin))?null:(e.print(t,"."+n+"("),this.visitAllExpressions(t.args,e,","),e.print(t,")"),null)},t.prototype.visitInvokeFunctionExpr=function(t,e){return t.fn.visitExpression(this,e),e.print(t,"("),this.visitAllExpressions(t.args,e,","),e.print(t,")"),null},t.prototype.visitReadVarExpr=function(t,e){var n=t.name;if(null!=t.builtin)switch(t.builtin){case Wo.Super:n="super";break;case Wo.This:n="this";break;case Wo.CatchError:n=ba.name;break;case Wo.CatchStack:n=_a.name;break;default:throw new Error("Unknown builtin variable "+t.builtin)}return e.print(t,n),null},t.prototype.visitInstantiateExpr=function(t,e){return e.print(t,"new "),t.classExpr.visitExpression(this,e),e.print(t,"("),this.visitAllExpressions(t.args,e,","),e.print(t,")"),null},t.prototype.visitLiteralExpr=function(t,e){var n=t.value;return"string"==typeof n?e.print(t,Oa(n,this._escapeDollarInStrings)):e.print(t,""+n),null},t.prototype.visitConditionalExpr=function(t,e){return e.print(t,"("),t.condition.visitExpression(this,e),e.print(t,"? "),t.trueCase.visitExpression(this,e),e.print(t,": "),t.falseCase.visitExpression(this,e),e.print(t,")"),null},t.prototype.visitNotExpr=function(t,e){return e.print(t,"!"),t.condition.visitExpression(this,e),null},t.prototype.visitAssertNotNullExpr=function(t,e){return t.condition.visitExpression(this,e),null},t.prototype.visitBinaryOperatorExpr=function(t,e){var n;switch(t.operator){case Bo.Equals:n="==";break;case Bo.Identical:n="===";break;case Bo.NotEquals:n="!=";break;case Bo.NotIdentical:n="!==";break;case Bo.And:n="&&";break;case Bo.Or:n="||";break;case Bo.Plus:n="+";break;case Bo.Minus:n="-";break;case Bo.Divide:n="/";break;case Bo.Multiply:n="*";break;case Bo.Modulo:n="%";break;case Bo.Lower:n="<";break;case Bo.LowerEquals:n="<=";break;case Bo.Bigger:n=">";break;case Bo.BiggerEquals:n=">=";break;default:throw new Error("Unknown operator "+t.operator)}return e.print(t,"("),t.lhs.visitExpression(this,e),e.print(t," "+n+" "),t.rhs.visitExpression(this,e),e.print(t,")"),null},t.prototype.visitReadPropExpr=function(t,e){return t.receiver.visitExpression(this,e),e.print(t,"."),e.print(t,t.name),null},t.prototype.visitReadKeyExpr=function(t,e){return t.receiver.visitExpression(this,e),e.print(t,"["),t.index.visitExpression(this,e),e.print(t,"]"),null},t.prototype.visitLiteralArrayExpr=function(t,e){return e.print(t,"["),this.visitAllExpressions(t.entries,e,","),e.print(t,"]"),null},t.prototype.visitLiteralMapExpr=function(t,e){var n=this;return e.print(t,"{"),this.visitAllObjects(function(r){e.print(t,Oa(r.key,n._escapeDollarInStrings,r.quoted)+":"),r.value.visitExpression(n,e)},t.entries,e,","),e.print(t,"}"),null},t.prototype.visitCommaExpr=function(t,e){return e.print(t,"("),this.visitAllExpressions(t.parts,e,","),e.print(t,")"),null},t.prototype.visitAllExpressions=function(t,e,n){var r=this;this.visitAllObjects(function(t){return t.visitExpression(r,e)},t,e,n)},t.prototype.visitAllObjects=function(t,e,n,r){for(var i=!1,o=0;o<e.length;o++)o>0&&(n.lineLength()>80?(n.print(null,r,!0),i||(n.incIndent(),n.incIndent(),i=!0)):n.print(null,r,!1)),t(e[o]);i&&(n.decIndent(),n.decIndent())},t.prototype.visitAllStatements=function(t,e){var n=this;t.forEach(function(t){return t.visitStatement(n,e)})},t}();function Oa(t,e,n){if(void 0===n&&(n=!0),null==t)return null;var r=t.replace(ga,function(){for(var t=[],n=0;n<arguments.length;n++)t[n]=arguments[n];return"$"==t[0]?e?"\\$":"$":"\n"==t[0]?"\\n":"\r"==t[0]?"\\r":"\\"+t[0]});return n||!va.test(r)?"'"+r+"'":r}function Ca(t){for(var e="",n=0;n<t;n++)e+=ya;return e}!function(){function t(){}t.prototype.emitStatementsAndContext=function(t,e,n,r,i){void 0===n&&(n=""),void 0===r&&(r=!0);var o=new ka(i),s=xa.createRoot();o.visitAllStatements(e,s);var a=n?n.split("\n"):[];o.reexports.forEach(function(t,e){var n=t.map(function(t){return t.name+" as "+t.as}).join(",");a.push("export {"+n+"} from '"+e+"';")}),o.importsWithPrefixes.forEach(function(t,e){a.push("import * as "+t+" from '"+e+"';")});var c=r?s.toSourceMapGenerator(t,a.length).toJsComment():"",u=a.concat([s.toSource(),c]);return c&&u.push(""),s.setPreambleLineCount(a.length),{sourceText:u.join("\n"),context:s}},t.prototype.emitStatements=function(t,e,n){return void 0===n&&(n=""),this.emitStatementsAndContext(t,e,n).sourceText}}();var ka=function(t){function e(e){var n=t.call(this,!1)||this;return n.referenceFilter=e,n.typeExpression=0,n.importsWithPrefixes=new Map,n.reexports=new Map,n}return Object(r.b)(e,t),e.prototype.visitType=function(t,e,n){void 0===n&&(n="any"),t?(this.typeExpression++,t.visitType(this,e),this.typeExpression--):e.print(null,n)},e.prototype.visitLiteralExpr=function(e,n){var r=e.value;return null==r&&e.type!=Vo?(n.print(e,"("+r+" as any)"),null):t.prototype.visitLiteralExpr.call(this,e,n)},e.prototype.visitLiteralArrayExpr=function(e,n){0===e.entries.length&&n.print(e,"(");var r=t.prototype.visitLiteralArrayExpr.call(this,e,n);return 0===e.entries.length&&n.print(e," as any[])"),r},e.prototype.visitExternalExpr=function(t,e){return this._visitIdentifier(t.value,t.typeParams,e),null},e.prototype.visitAssertNotNullExpr=function(e,n){var r=t.prototype.visitAssertNotNullExpr.call(this,e,n);return n.print(e,"!"),r},e.prototype.visitDeclareVarStmt=function(t,e){if(t.hasModifier(ys.Exported)&&t.value instanceof es&&!t.type){var n=t.value.value,r=n.name,i=n.moduleName;if(i){var o=this.reexports.get(i);return o||(o=[],this.reexports.set(i,o)),o.push({name:r,as:t.name}),null}}return t.hasModifier(ys.Exported)&&e.print(t,"export "),t.hasModifier(ys.Final)?e.print(t,"const"):e.print(t,"var"),e.print(t," "+t.name),this._printColonType(t.type,e),e.print(t," = "),t.value.visitExpression(this,e),e.println(t,";"),null},e.prototype.visitCastExpr=function(t,e){return e.print(t,"(<"),t.type.visitType(this,e),e.print(t,">"),t.value.visitExpression(this,e),e.print(t,")"),null},e.prototype.visitInstantiateExpr=function(t,e){return e.print(t,"new "),this.typeExpression++,t.classExpr.visitExpression(this,e),this.typeExpression--,e.print(t,"("),this.visitAllExpressions(t.args,e,","),e.print(t,")"),null},e.prototype.visitDeclareClassStmt=function(t,e){var n=this;return e.pushClass(t),t.hasModifier(ys.Exported)&&e.print(t,"export "),e.print(t,"class "+t.name),null!=t.parent&&(e.print(t," extends "),this.typeExpression++,t.parent.visitExpression(this,e),this.typeExpression--),e.println(t," {"),e.incIndent(),t.fields.forEach(function(t){return n._visitClassField(t,e)}),null!=t.constructorMethod&&this._visitClassConstructor(t,e),t.getters.forEach(function(t){return n._visitClassGetter(t,e)}),t.methods.forEach(function(t){return n._visitClassMethod(t,e)}),e.decIndent(),e.println(t,"}"),e.popClass(),null},e.prototype._visitClassField=function(t,e){t.hasModifier(ys.Private)&&e.print(null,"/*private*/ "),e.print(null,t.name),this._printColonType(t.type,e),e.println(null,";")},e.prototype._visitClassGetter=function(t,e){t.hasModifier(ys.Private)&&e.print(null,"private "),e.print(null,"get "+t.name+"()"),this._printColonType(t.type,e),e.println(null," {"),e.incIndent(),this.visitAllStatements(t.body,e),e.decIndent(),e.println(null,"}")},e.prototype._visitClassConstructor=function(t,e){e.print(t,"constructor("),this._visitParams(t.constructorMethod.params,e),e.println(t,") {"),e.incIndent(),this.visitAllStatements(t.constructorMethod.body,e),e.decIndent(),e.println(t,"}")},e.prototype._visitClassMethod=function(t,e){t.hasModifier(ys.Private)&&e.print(null,"private "),e.print(null,t.name+"("),this._visitParams(t.params,e),e.print(null,")"),this._printColonType(t.type,e,"void"),e.println(null," {"),e.incIndent(),this.visitAllStatements(t.body,e),e.decIndent(),e.println(null,"}")},e.prototype.visitFunctionExpr=function(t,e){return e.print(t,"("),this._visitParams(t.params,e),e.print(t,")"),this._printColonType(t.type,e,"void"),e.println(t," => {"),e.incIndent(),this.visitAllStatements(t.statements,e),e.decIndent(),e.print(t,"}"),null},e.prototype.visitDeclareFunctionStmt=function(t,e){return t.hasModifier(ys.Exported)&&e.print(t,"export "),e.print(t,"function "+t.name+"("),this._visitParams(t.params,e),e.print(t,")"),this._printColonType(t.type,e,"void"),e.println(t," {"),e.incIndent(),this.visitAllStatements(t.statements,e),e.decIndent(),e.println(t,"}"),null},e.prototype.visitTryCatchStmt=function(t,e){e.println(t,"try {"),e.incIndent(),this.visitAllStatements(t.bodyStmts,e),e.decIndent(),e.println(t,"} catch ("+ba.name+") {"),e.incIndent();var n=[_a.set(ba.prop("stack",null)).toDeclStmt(null,[ys.Final])].concat(t.catchStmts);return this.visitAllStatements(n,e),e.decIndent(),e.println(t,"}"),null},e.prototype.visitBuiltintType=function(t,e){var n;switch(t.name){case Do.Bool:n="boolean";break;case Do.Dynamic:n="any";break;case Do.Function:n="Function";break;case Do.Number:case Do.Int:n="number";break;case Do.String:n="string";break;default:throw new Error("Unsupported builtin type "+t.name)}return e.print(null,n),null},e.prototype.visitExpressionType=function(t,e){return t.value.visitExpression(this,e),null},e.prototype.visitArrayType=function(t,e){return this.visitType(t.of,e),e.print(null,"[]"),null},e.prototype.visitMapType=function(t,e){return e.print(null,"{[key: string]:"),this.visitType(t.valueType,e),e.print(null,"}"),null},e.prototype.getBuiltinMethodName=function(t){var e;switch(t){case $o.ConcatArray:e="concat";break;case $o.SubscribeObservable:e="subscribe";break;case $o.Bind:e="bind";break;default:throw new Error("Unknown builtin method: "+t)}return e},e.prototype._visitParams=function(t,e){var n=this;this.visitAllObjects(function(t){e.print(null,t.name),n._printColonType(t.type,e)},t,e,",")},e.prototype._visitIdentifier=function(t,e,n){var r=this,i=t.name,o=t.moduleName;if(this.referenceFilter&&this.referenceFilter(t))n.print(null,"(null as any)");else{if(o){var s=this.importsWithPrefixes.get(o);null==s&&(s="i"+this.importsWithPrefixes.size,this.importsWithPrefixes.set(o,s)),n.print(null,s+".")}if(n.print(null,i),this.typeExpression>0)(e||[]).length>0&&(n.print(null,"<"),this.visitAllObjects(function(t){return t.visitType(r,n)},e,n,","),n.print(null,">"))}},e.prototype._printColonType=function(t,e,n){t!==Vo&&(e.print(null,":"),this.visitType(t,e,n))},e}(wa),Ea=function(){function t(t){this._reflector=t}return t.prototype.isPipe=function(t){var e=this._reflector.annotations(K(t));return e&&e.some(m.isTypeOf)},t.prototype.resolve=function(t,e){void 0===e&&(e=!0);var n=this._reflector.annotations(K(t));if(n){var r=Ee(n,m.isTypeOf);if(r)return r}if(e)throw new Error("No Pipe decorator found on "+Y(t));return null},t}(),Sa={};function Pa(t,e){for(var n=0,r=e;n<r.length;n++){var i=r[n];Sa[i.toLowerCase()]=t}}Pa(P.HTML,["iframe|srcdoc","*|innerHTML","*|outerHTML"]),Pa(P.STYLE,["*|style"]),Pa(P.URL,["*|formAction","area|href","area|ping","audio|src","a|href","a|ping","blockquote|cite","body|background","del|cite","form|action","img|src","img|srcset","input|src","ins|cite","q|cite","source|src","source|srcset","track|src","video|poster","video|src"]),Pa(P.RESOURCE_URL,["applet|code","applet|codebase","base|href","embed|src","frame|src","head|profile","html|manifest","iframe|src","link|href","media|src","object|codebase","object|data","script|src"]);var Ma=function(){return function(){}}(),Aa="boolean",Ta="number",Da="string",Ia="object",Ra=["[Element]|textContent,%classList,className,id,innerHTML,*beforecopy,*beforecut,*beforepaste,*copy,*cut,*paste,*search,*selectstart,*webkitfullscreenchange,*webkitfullscreenerror,*wheel,outerHTML,#scrollLeft,#scrollTop,slot,*message,*mozfullscreenchange,*mozfullscreenerror,*mozpointerlockchange,*mozpointerlockerror,*webglcontextcreationerror,*webglcontextlost,*webglcontextrestored","[HTMLElement]^[Element]|accessKey,contentEditable,dir,!draggable,!hidden,innerText,lang,*abort,*auxclick,*blur,*cancel,*canplay,*canplaythrough,*change,*click,*close,*contextmenu,*cuechange,*dblclick,*drag,*dragend,*dragenter,*dragleave,*dragover,*dragstart,*drop,*durationchange,*emptied,*ended,*error,*focus,*gotpointercapture,*input,*invalid,*keydown,*keypress,*keyup,*load,*loadeddata,*loadedmetadata,*loadstart,*lostpointercapture,*mousedown,*mouseenter,*mouseleave,*mousemove,*mouseout,*mouseover,*mouseup,*mousewheel,*pause,*play,*playing,*pointercancel,*pointerdown,*pointerenter,*pointerleave,*pointermove,*pointerout,*pointerover,*pointerup,*progress,*ratechange,*reset,*resize,*scroll,*seeked,*seeking,*select,*show,*stalled,*submit,*suspend,*timeupdate,*toggle,*volumechange,*waiting,outerText,!spellcheck,%style,#tabIndex,title,!translate","abbr,address,article,aside,b,bdi,bdo,cite,code,dd,dfn,dt,em,figcaption,figure,footer,header,i,kbd,main,mark,nav,noscript,rb,rp,rt,rtc,ruby,s,samp,section,small,strong,sub,sup,u,var,wbr^[HTMLElement]|accessKey,contentEditable,dir,!draggable,!hidden,innerText,lang,*abort,*auxclick,*blur,*cancel,*canplay,*canplaythrough,*change,*click,*close,*contextmenu,*cuechange,*dblclick,*drag,*dragend,*dragenter,*dragleave,*dragover,*dragstart,*drop,*durationchange,*emptied,*ended,*error,*focus,*gotpointercapture,*input,*invalid,*keydown,*keypress,*keyup,*load,*loadeddata,*loadedmetadata,*loadstart,*lostpointercapture,*mousedown,*mouseenter,*mouseleave,*mousemove,*mouseout,*mouseover,*mouseup,*mousewheel,*pause,*play,*playing,*pointercancel,*pointerdown,*pointerenter,*pointerleave,*pointermove,*pointerout,*pointerover,*pointerup,*progress,*ratechange,*reset,*resize,*scroll,*seeked,*seeking,*select,*show,*stalled,*submit,*suspend,*timeupdate,*toggle,*volumechange,*waiting,outerText,!spellcheck,%style,#tabIndex,title,!translate","media^[HTMLElement]|!autoplay,!controls,%controlsList,%crossOrigin,#currentTime,!defaultMuted,#defaultPlaybackRate,!disableRemotePlayback,!loop,!muted,*encrypted,*waitingforkey,#playbackRate,preload,src,%srcObject,#volume",":svg:^[HTMLElement]|*abort,*auxclick,*blur,*cancel,*canplay,*canplaythrough,*change,*click,*close,*contextmenu,*cuechange,*dblclick,*drag,*dragend,*dragenter,*dragleave,*dragover,*dragstart,*drop,*durationchange,*emptied,*ended,*error,*focus,*gotpointercapture,*input,*invalid,*keydown,*keypress,*keyup,*load,*loadeddata,*loadedmetadata,*loadstart,*lostpointercapture,*mousedown,*mouseenter,*mouseleave,*mousemove,*mouseout,*mouseover,*mouseup,*mousewheel,*pause,*play,*playing,*pointercancel,*pointerdown,*pointerenter,*pointerleave,*pointermove,*pointerout,*pointerover,*pointerup,*progress,*ratechange,*reset,*resize,*scroll,*seeked,*seeking,*select,*show,*stalled,*submit,*suspend,*timeupdate,*toggle,*volumechange,*waiting,%style,#tabIndex",":svg:graphics^:svg:|",":svg:animation^:svg:|*begin,*end,*repeat",":svg:geometry^:svg:|",":svg:componentTransferFunction^:svg:|",":svg:gradient^:svg:|",":svg:textContent^:svg:graphics|",":svg:textPositioning^:svg:textContent|","a^[HTMLElement]|charset,coords,download,hash,host,hostname,href,hreflang,name,password,pathname,ping,port,protocol,referrerPolicy,rel,rev,search,shape,target,text,type,username","area^[HTMLElement]|alt,coords,download,hash,host,hostname,href,!noHref,password,pathname,ping,port,protocol,referrerPolicy,rel,search,shape,target,username","audio^media|","br^[HTMLElement]|clear","base^[HTMLElement]|href,target","body^[HTMLElement]|aLink,background,bgColor,link,*beforeunload,*blur,*error,*focus,*hashchange,*languagechange,*load,*message,*offline,*online,*pagehide,*pageshow,*popstate,*rejectionhandled,*resize,*scroll,*storage,*unhandledrejection,*unload,text,vLink","button^[HTMLElement]|!autofocus,!disabled,formAction,formEnctype,formMethod,!formNoValidate,formTarget,name,type,value","canvas^[HTMLElement]|#height,#width","content^[HTMLElement]|select","dl^[HTMLElement]|!compact","datalist^[HTMLElement]|","details^[HTMLElement]|!open","dialog^[HTMLElement]|!open,returnValue","dir^[HTMLElement]|!compact","div^[HTMLElement]|align","embed^[HTMLElement]|align,height,name,src,type,width","fieldset^[HTMLElement]|!disabled,name","font^[HTMLElement]|color,face,size","form^[HTMLElement]|acceptCharset,action,autocomplete,encoding,enctype,method,name,!noValidate,target","frame^[HTMLElement]|frameBorder,longDesc,marginHeight,marginWidth,name,!noResize,scrolling,src","frameset^[HTMLElement]|cols,*beforeunload,*blur,*error,*focus,*hashchange,*languagechange,*load,*message,*offline,*online,*pagehide,*pageshow,*popstate,*rejectionhandled,*resize,*scroll,*storage,*unhandledrejection,*unload,rows","hr^[HTMLElement]|align,color,!noShade,size,width","head^[HTMLElement]|","h1,h2,h3,h4,h5,h6^[HTMLElement]|align","html^[HTMLElement]|version","iframe^[HTMLElement]|align,!allowFullscreen,frameBorder,height,longDesc,marginHeight,marginWidth,name,referrerPolicy,%sandbox,scrolling,src,srcdoc,width","img^[HTMLElement]|align,alt,border,%crossOrigin,#height,#hspace,!isMap,longDesc,lowsrc,name,referrerPolicy,sizes,src,srcset,useMap,#vspace,#width","input^[HTMLElement]|accept,align,alt,autocapitalize,autocomplete,!autofocus,!checked,!defaultChecked,defaultValue,dirName,!disabled,%files,formAction,formEnctype,formMethod,!formNoValidate,formTarget,#height,!incremental,!indeterminate,max,#maxLength,min,#minLength,!multiple,name,pattern,placeholder,!readOnly,!required,selectionDirection,#selectionEnd,#selectionStart,#size,src,step,type,useMap,value,%valueAsDate,#valueAsNumber,#width","li^[HTMLElement]|type,#value","label^[HTMLElement]|htmlFor","legend^[HTMLElement]|align","link^[HTMLElement]|as,charset,%crossOrigin,!disabled,href,hreflang,integrity,media,referrerPolicy,rel,%relList,rev,%sizes,target,type","map^[HTMLElement]|name","marquee^[HTMLElement]|behavior,bgColor,direction,height,#hspace,#loop,#scrollAmount,#scrollDelay,!trueSpeed,#vspace,width","menu^[HTMLElement]|!compact","meta^[HTMLElement]|content,httpEquiv,name,scheme","meter^[HTMLElement]|#high,#low,#max,#min,#optimum,#value","ins,del^[HTMLElement]|cite,dateTime","ol^[HTMLElement]|!compact,!reversed,#start,type","object^[HTMLElement]|align,archive,border,code,codeBase,codeType,data,!declare,height,#hspace,name,standby,type,useMap,#vspace,width","optgroup^[HTMLElement]|!disabled,label","option^[HTMLElement]|!defaultSelected,!disabled,label,!selected,text,value","output^[HTMLElement]|defaultValue,%htmlFor,name,value","p^[HTMLElement]|align","param^[HTMLElement]|name,type,value,valueType","picture^[HTMLElement]|","pre^[HTMLElement]|#width","progress^[HTMLElement]|#max,#value","q,blockquote,cite^[HTMLElement]|","script^[HTMLElement]|!async,charset,%crossOrigin,!defer,event,htmlFor,integrity,src,text,type","select^[HTMLElement]|!autofocus,!disabled,#length,!multiple,name,!required,#selectedIndex,#size,value","shadow^[HTMLElement]|","slot^[HTMLElement]|name","source^[HTMLElement]|media,sizes,src,srcset,type","span^[HTMLElement]|","style^[HTMLElement]|!disabled,media,type","caption^[HTMLElement]|align","th,td^[HTMLElement]|abbr,align,axis,bgColor,ch,chOff,#colSpan,headers,height,!noWrap,#rowSpan,scope,vAlign,width","col,colgroup^[HTMLElement]|align,ch,chOff,#span,vAlign,width","table^[HTMLElement]|align,bgColor,border,%caption,cellPadding,cellSpacing,frame,rules,summary,%tFoot,%tHead,width","tr^[HTMLElement]|align,bgColor,ch,chOff,vAlign","tfoot,thead,tbody^[HTMLElement]|align,ch,chOff,vAlign","template^[HTMLElement]|","textarea^[HTMLElement]|autocapitalize,!autofocus,#cols,defaultValue,dirName,!disabled,#maxLength,#minLength,name,placeholder,!readOnly,!required,#rows,selectionDirection,#selectionEnd,#selectionStart,value,wrap","title^[HTMLElement]|text","track^[HTMLElement]|!default,kind,label,src,srclang","ul^[HTMLElement]|!compact,type","unknown^[HTMLElement]|","video^media|#height,poster,#width",":svg:a^:svg:graphics|",":svg:animate^:svg:animation|",":svg:animateMotion^:svg:animation|",":svg:animateTransform^:svg:animation|",":svg:circle^:svg:geometry|",":svg:clipPath^:svg:graphics|",":svg:defs^:svg:graphics|",":svg:desc^:svg:|",":svg:discard^:svg:|",":svg:ellipse^:svg:geometry|",":svg:feBlend^:svg:|",":svg:feColorMatrix^:svg:|",":svg:feComponentTransfer^:svg:|",":svg:feComposite^:svg:|",":svg:feConvolveMatrix^:svg:|",":svg:feDiffuseLighting^:svg:|",":svg:feDisplacementMap^:svg:|",":svg:feDistantLight^:svg:|",":svg:feDropShadow^:svg:|",":svg:feFlood^:svg:|",":svg:feFuncA^:svg:componentTransferFunction|",":svg:feFuncB^:svg:componentTransferFunction|",":svg:feFuncG^:svg:componentTransferFunction|",":svg:feFuncR^:svg:componentTransferFunction|",":svg:feGaussianBlur^:svg:|",":svg:feImage^:svg:|",":svg:feMerge^:svg:|",":svg:feMergeNode^:svg:|",":svg:feMorphology^:svg:|",":svg:feOffset^:svg:|",":svg:fePointLight^:svg:|",":svg:feSpecularLighting^:svg:|",":svg:feSpotLight^:svg:|",":svg:feTile^:svg:|",":svg:feTurbulence^:svg:|",":svg:filter^:svg:|",":svg:foreignObject^:svg:graphics|",":svg:g^:svg:graphics|",":svg:image^:svg:graphics|",":svg:line^:svg:geometry|",":svg:linearGradient^:svg:gradient|",":svg:mpath^:svg:|",":svg:marker^:svg:|",":svg:mask^:svg:|",":svg:metadata^:svg:|",":svg:path^:svg:geometry|",":svg:pattern^:svg:|",":svg:polygon^:svg:geometry|",":svg:polyline^:svg:geometry|",":svg:radialGradient^:svg:gradient|",":svg:rect^:svg:geometry|",":svg:svg^:svg:graphics|#currentScale,#zoomAndPan",":svg:script^:svg:|type",":svg:set^:svg:animation|",":svg:stop^:svg:|",":svg:style^:svg:|!disabled,media,title,type",":svg:switch^:svg:graphics|",":svg:symbol^:svg:|",":svg:tspan^:svg:textPositioning|",":svg:text^:svg:textPositioning|",":svg:textPath^:svg:textContent|",":svg:title^:svg:|",":svg:use^:svg:graphics|",":svg:view^:svg:|#zoomAndPan","data^[HTMLElement]|value","keygen^[HTMLElement]|!autofocus,challenge,!disabled,form,keytype,name","menuitem^[HTMLElement]|type,label,icon,!disabled,!checked,radiogroup,!default","summary^[HTMLElement]|","time^[HTMLElement]|dateTime",":svg:cursor^:svg:|"],Na={class:"className",for:"htmlFor",formaction:"formAction",innerHtml:"innerHTML",readonly:"readOnly",tabindex:"tabIndex"},La=function(t){function e(){var e=t.call(this)||this;return e._schema={},Ra.forEach(function(t){var n={},r=t.split("|"),i=r[0],o=r[1].split(","),s=i.split("^"),a=s[0],c=s[1];a.split(",").forEach(function(t){return e._schema[t.toLowerCase()]=n});var u=c&&e._schema[c.toLowerCase()];u&&Object.keys(u).forEach(function(t){n[t]=u[t]}),o.forEach(function(t){if(t.length>0)switch(t[0]){case"*":break;case"!":n[t.substring(1)]=Aa;break;case"#":n[t.substring(1)]=Ta;break;case"%":n[t.substring(1)]=Ia;break;default:n[t]=Da}})}),e}return Object(r.b)(e,t),e.prototype.hasProperty=function(t,e,n){if(n.some(function(t){return t.name===x.name}))return!0;if(t.indexOf("-")>-1){if(ie(t)||oe(t))return!1;if(n.some(function(t){return t.name===j.name}))return!0}return!!(this._schema[t.toLowerCase()]||this._schema.unknown)[e]},e.prototype.hasElement=function(t,e){if(e.some(function(t){return t.name===x.name}))return!0;if(t.indexOf("-")>-1){if(ie(t)||oe(t))return!0;if(e.some(function(t){return t.name===j.name}))return!0}return!!this._schema[t.toLowerCase()]},e.prototype.securityContext=function(t,e,n){n&&(e=this.getMappedPropName(e)),t=t.toLowerCase(),e=e.toLowerCase();var r=Sa[t+"|"+e];return r||((r=Sa["*|"+e])||P.NONE)},e.prototype.getMappedPropName=function(t){return Na[t]||t},e.prototype.getDefaultComponentElementName=function(){return"ng-component"},e.prototype.validateProperty=function(t){return t.toLowerCase().startsWith("on")?{error:!0,msg:"Binding to event property '"+t+"' is disallowed for security reasons, please use ("+t.slice(2)+")=...\nIf '"+t+"' is a directive input, make sure the directive is imported by the current module."}:{error:!1}},e.prototype.validateAttribute=function(t){return t.toLowerCase().startsWith("on")?{error:!0,msg:"Binding to event attribute '"+t+"' is disallowed for security reasons, please use ("+t.slice(2)+")=..."}:{error:!1}},e.prototype.allKnownElementNames=function(){return Object.keys(this._schema)},e.prototype.normalizeAnimationStyleProperty=function(t){return t.replace(T,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return t[1].toUpperCase()})},e.prototype.normalizeAnimationStyleValue=function(t,e,n){var r="",i=n.toString().trim(),o=null;if(function(t){switch(t){case"width":case"height":case"minWidth":case"minHeight":case"maxWidth":case"maxHeight":case"left":case"top":case"bottom":case"right":case"fontSize":case"outlineWidth":case"outlineOffset":case"paddingTop":case"paddingLeft":case"paddingBottom":case"paddingRight":case"marginTop":case"marginLeft":case"marginBottom":case"marginRight":case"borderRadius":case"borderWidth":case"borderTopWidth":case"borderLeftWidth":case"borderRightWidth":case"borderBottomWidth":case"textIndent":return!0;default:return!1}}(t)&&0!==n&&"0"!==n)if("number"==typeof n)r="px";else{var s=n.match(/^[+-]?[\d\.]+([a-z]*)$/);s&&0==s[1].length&&(o="Please provide a CSS unit value for "+e+":"+n)}return{error:o,value:i+r}},e}(Ma);var Fa=function(){function t(){this.strictStyling=!0}return t.prototype.shimCssText=function(t,e,n){void 0===n&&(n="");var r,i=(r=t.match(rc))?r[0]:"";return t=t.replace(nc,""),t=this._insertDirectives(t),this._scopeCssText(t,e,n)+i},t.prototype._insertDirectives=function(t){return t=this._insertPolyfillDirectivesInCssText(t),this._insertPolyfillRulesInCssText(t)},t.prototype._insertPolyfillDirectivesInCssText=function(t){return t.replace(za,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return t[2]+"{"})},t.prototype._insertPolyfillRulesInCssText=function(t){return t.replace(Ba,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t[0].replace(t[1],"").replace(t[2],"");return t[4]+n})},t.prototype._scopeCssText=function(t,e,n){var r=this._extractUnscopedRulesFromCssText(t);return t=this._insertPolyfillHostInCssText(t),t=this._convertColonHost(t),t=this._convertColonHostContext(t),t=this._convertShadowDOMSelectors(t),e&&(t=this._scopeSelectors(t,e,n)),(t=t+"\n"+r).trim()},t.prototype._extractUnscopedRulesFromCssText=function(t){var e,n="";for(qa.lastIndex=0;null!==(e=qa.exec(t));){n+=e[0].replace(e[2],"").replace(e[1],e[4])+"\n\n"}return n},t.prototype._convertColonHost=function(t){return this._convertColonRule(t,Ga,this._colonHostPartReplacer)},t.prototype._convertColonHostContext=function(t){return this._convertColonRule(t,Xa,this._colonHostContextPartReplacer)},t.prototype._convertColonRule=function(t,e,n){return t.replace(e,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];if(t[2]){for(var r=t[2].split(","),i=[],o=0;o<r.length;o++){var s=r[o].trim();if(!s)break;i.push(n(Ya,s,t[3]))}return i.join(",")}return Ya+t[3]})},t.prototype._colonHostContextPartReplacer=function(t,e,n){return e.indexOf(Ha)>-1?this._colonHostPartReplacer(t,e,n):t+e+n+", "+e+" "+t+n},t.prototype._colonHostPartReplacer=function(t,e,n){return t+e.replace(Ha,"")+n},t.prototype._convertShadowDOMSelectors=function(t){return $a.reduce(function(t,e){return t.replace(e," ")},t)},t.prototype._scopeSelectors=function(t,e,n){var r,i,o,s=this;return r=function(t){var r=t.selector,i=t.content;return"@"!=t.selector[0]?r=s._scopeSelector(t.selector,e,n,s.strictStyling):(t.selector.startsWith("@media")||t.selector.startsWith("@supports")||t.selector.startsWith("@page")||t.selector.startsWith("@document"))&&(i=s._scopeSelectors(t.content,e,n)),new uc(r,i)},i=function(t){for(var e=t.split(oc),n=[],r=[],i=0,o=[],s=0;s<e.length;s++){var a=e[s];a==ac&&i--,i>0?o.push(a):(o.length>0&&(r.push(o.join("")),n.push(cc),o=[]),n.push(a)),a==sc&&i++}o.length>0&&(r.push(o.join("")),n.push(cc));return new lc(n.join(""),r)}(t),o=0,i.escapedString.replace(ic,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];var n=t[2],s="",a=t[4],c="";a&&a.startsWith("{"+cc)&&(s=i.blocks[o++],a=a.substring(cc.length+1),c="{");var u=r(new uc(n,s));return""+t[1]+u.selector+t[3]+c+u.content+a})},t.prototype._scopeSelector=function(t,e,n,r){var i=this;return t.split(",").map(function(t){return t.trim().split(Qa)}).map(function(t){var o,s=t[0],a=t.slice(1);return[(o=s,i._selectorNeedsScoping(o,e)?r?i._applyStrictSelectorScope(o,e,n):i._applySelectorScope(o,e,n):o)].concat(a).join(" ")}).join(", ")},t.prototype._selectorNeedsScoping=function(t,e){return!this._makeScopeMatcher(e).test(t)},t.prototype._makeScopeMatcher=function(t){return t=t.replace(/\[/g,"\\[").replace(/\]/g,"\\]"),new RegExp("^("+t+")"+Za,"m")},t.prototype._applySelectorScope=function(t,e,n){return this._applySimpleSelectorScope(t,e,n)},t.prototype._applySimpleSelectorScope=function(t,e,n){if(Ja.lastIndex=0,Ja.test(t)){var r=this.strictStyling?"["+n+"]":e;return t.replace(Ka,function(t,e){return e.replace(/([^:]*)(:*)(.*)/,function(t,e,n,i){return e+r+n+i})}).replace(Ja,r+" ")}return e+" "+t},t.prototype._applyStrictSelectorScope=function(t,e,n){for(var r,i=this,o="["+(e=e.replace(/\[is=([^\]]*)\]/g,function(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return e[0]}))+"]",s=function(t){var r=t.trim();if(!r)return"";if(t.indexOf(Ya)>-1)r=i._applySimpleSelectorScope(t,e,n);else{var s=t.replace(Ja,"");if(s.length>0){var a=s.match(/([^:]*)(:*)(.*)/);a&&(r=a[1]+o+a[2]+a[3])}}return r},a=new Va(t),c="",u=0,l=/( |>|\+|~(?!=))\s*/g,p=!((t=a.content()).indexOf(Ya)>-1);null!==(r=l.exec(t));){var d=r[1],h=t.slice(u,r.index).trim();c+=((p=p||h.indexOf(Ya)>-1)?s(h):h)+" "+d+" ",u=l.lastIndex}var f=t.substring(u);return c+=(p=p||f.indexOf(Ya)>-1)?s(f):f,a.restore(c)},t.prototype._insertPolyfillHostInCssText=function(t){return t.replace(ec,Ua).replace(tc,Ha)},t}(),Va=function(){function t(t){var e=this;this.placeholders=[],this.index=0,t=t.replace(/(\[[^\]]*\])/g,function(t,n){var r="__ph-"+e.index+"__";return e.placeholders.push(n),e.index++,r}),this._content=t.replace(/(:nth-[-\w]+)(\([^)]+\))/g,function(t,n,r){var i="__ph-"+e.index+"__";return e.placeholders.push(r),e.index++,n+i})}return t.prototype.restore=function(t){var e=this;return t.replace(/__ph-(\d+)__/g,function(t,n){return e.placeholders[+n]})},t.prototype.content=function(){return this._content},t}(),za=/polyfill-next-selector[^}]*content:[\s]*?(['"])(.*?)\1[;\s]*}([^{]*?){/gim,Ba=/(polyfill-rule)[^}]*(content:[\s]*(['"])(.*?)\3)[;\s]*[^}]*}/gim,qa=/(polyfill-unscoped-rule)[^}]*(content:[\s]*(['"])(.*?)\3)[;\s]*[^}]*}/gim,Ha="-shadowcsshost",Ua="-shadowcsscontext",Wa=")(?:\\(((?:\\([^)(]*\\)|[^)(]*)+?)\\))?([^,{]*)",Ga=new RegExp("("+Ha+Wa,"gim"),Xa=new RegExp("("+Ua+Wa,"gim"),Ya=Ha+"-no-combinator",Ka=/-shadowcsshost-no-combinator([^\s]*)/,$a=[/::shadow/g,/::content/g,/\/shadow-deep\//g,/\/shadow\//g],Qa=/(?:>>>)|(?:\/deep\/)|(?:::ng-deep)/g,Za="([>\\s~+[.,{:][\\s\\S]*)?$",Ja=/-shadowcsshost/gim,tc=/:host/gim,ec=/:host-context/gim,nc=/\/\*\s*[\s\S]*?\*\//g;var rc=/\/\*\s*#\s*sourceMappingURL=[\s\S]+?\*\//;var ic=/(\s*)([^;\{\}]+?)(\s*)((?:{%BLOCK%}?\s*;?)|(?:\s*;))/g,oc=/([{}])/g,sc="{",ac="}",cc="%BLOCK%",uc=function(){return function(t,e){this.selector=t,this.content=e}}();var lc=function(){return function(t,e){this.escapedString=t,this.blocks=e}}();var pc=function(){return function(t,e,n){this.name=t,this.moduleUrl=e,this.setValue=n}}(),dc=function(){return function(t,e,n,r,i){this.outputCtx=t,this.stylesVar=e,this.dependencies=n,this.isShimmed=r,this.meta=i}}(),hc=function(){function t(t){this._urlResolver=t,this._shadowCss=new Fa}return t.prototype.compileComponent=function(t,e){var n=e.template;return this._compileStyles(t,e,new St({styles:n.styles,styleUrls:n.styleUrls,moduleUrl:jt(e.type)}),this.needsStyleShim(e),!0)},t.prototype.compileStyles=function(t,e,n,r){return void 0===r&&(r=this.needsStyleShim(e)),this._compileStyles(t,e,n,r,!1)},t.prototype.needsStyleShim=function(t){return t.template.encapsulation===d.Emulated},t.prototype._compileStyles=function(t,e,n,r,i){var o=this,s=n.styles.map(function(t){return Ws(o._shimIfNeeded(t,r))}),a=[];n.styleUrls.forEach(function(e){var n=s.length;s.push(null),a.push(new pc(fc(null),e,function(e){return s[n]=t.importExpr(e)}))});var c=fc(i?e:null),u=Fs(c).set(qs(s,new No(Fo,[Ao.Const]))).toDeclStmt(null,i?[ys.Final]:[ys.Final,ys.Exported]);return t.statements.push(u),new dc(t,c,a,r,n)},t.prototype._shimIfNeeded=function(t,e){return e?this._shadowCss.shimCssText(t,"_ngcontent-%COMP%","_nghost-%COMP%"):t},t}();function fc(t){var e="styles";return t&&(e+="_"+_t(t.type)),e}var mc="ngPreserveWhitespaces",gc=new Set(["pre","template","textarea","script","style"]),vc=" \f\n\r\t\v ᠎ - \u2028\u2029  　\ufeff",yc=new RegExp("[^"+vc+"]"),bc=new RegExp("["+vc+"]{2,}","g");function _c(t){return t.replace(new RegExp(ue,"g")," ")}var jc=function(){function t(){}return t.prototype.visitElement=function(t,e){return gc.has(t.name)||t.attrs.some(function(t){return t.name===mc})?new Ht(t.name,Wt(this,t.attrs),t.children,t.sourceSpan,t.startSourceSpan,t.endSourceSpan):new Ht(t.name,t.attrs,Wt(this,t.children),t.sourceSpan,t.startSourceSpan,t.endSourceSpan)},t.prototype.visitAttribute=function(t,e){return t.name!==mc?t:null},t.prototype.visitText=function(t,e){return t.value.match(yc)?new Vt(_c(t.value).replace(bc," "),t.sourceSpan):null},t.prototype.visitComment=function(t,e){return t},t.prototype.visitExpansion=function(t,e){return t},t.prototype.visitExpansionCase=function(t,e){return t},t}();var xc=["zero","one","two","few","many","other"];function wc(t){var e=new kc;return new Oc(Wt(e,t),e.isExpanded,e.errors)}var Oc=function(){return function(t,e,n){this.nodes=t,this.expanded=e,this.errors=n}}(),Cc=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(r.b)(e,t),e}(ar),kc=function(){function t(){this.isExpanded=!1,this.errors=[]}return t.prototype.visitElement=function(t,e){return new Ht(t.name,t.attrs,Wt(this,t.children),t.sourceSpan,t.startSourceSpan,t.endSourceSpan)},t.prototype.visitAttribute=function(t,e){return t},t.prototype.visitText=function(t,e){return t},t.prototype.visitComment=function(t,e){return t},t.prototype.visitExpansion=function(t,e){return this.isExpanded=!0,"plural"==t.type?(s=t,a=this.errors,c=s.cases.map(function(t){-1!=xc.indexOf(t.value)||t.value.match(/^=\d+$/)||a.push(new Cc(t.valueSourceSpan,'Plural cases should be "=<number>" or one of '+xc.join(", ")));var e=wc(t.expression);return a.push.apply(a,e.errors),new Ht("ng-template",[new qt("ngPluralCase",""+t.value,t.valueSourceSpan)],e.nodes,t.sourceSpan,t.sourceSpan,t.sourceSpan)}),u=new qt("[ngPlural]",s.switchValue,s.switchValueSourceSpan),new Ht("ng-container",[u],c,s.sourceSpan,s.sourceSpan,s.sourceSpan)):(n=t,r=this.errors,i=n.cases.map(function(t){var e=wc(t.expression);return r.push.apply(r,e.errors),"other"===t.value?new Ht("ng-template",[new qt("ngSwitchDefault","",t.valueSourceSpan)],e.nodes,t.sourceSpan,t.sourceSpan,t.sourceSpan):new Ht("ng-template",[new qt("ngSwitchCase",""+t.value,t.valueSourceSpan)],e.nodes,t.sourceSpan,t.sourceSpan,t.sourceSpan)}),o=new qt("[ngSwitch]",n.switchValue,n.switchValueSourceSpan),new Ht("ng-container",[o],i,n.sourceSpan,n.sourceSpan,n.sourceSpan));var n,r,i,o,s,a,c,u},t.prototype.visitExpansionCase=function(t,e){throw new Error("Should not be reached")},t}();var Ec={DEFAULT:0,LITERAL_ATTR:1,ANIMATION:2};Ec[Ec.DEFAULT]="DEFAULT",Ec[Ec.LITERAL_ATTR]="LITERAL_ATTR",Ec[Ec.ANIMATION]="ANIMATION";var Sc=function(){return function(t,e,n,r){this.name=t,this.expression=e,this.type=n,this.sourceSpan=r,this.isLiteral=this.type===Ec.LITERAL_ATTR,this.isAnimation=this.type===Ec.ANIMATION}}(),Pc=function(){function t(t,e,n,r,i){var o=this;this._exprParser=t,this._interpolationConfig=e,this._schemaRegistry=n,this._targetErrors=i,this.pipesByName=new Map,this._usedPipes=new Map,r.forEach(function(t){return o.pipesByName.set(t.name,t)})}return t.prototype.getUsedPipes=function(){return Array.from(this._usedPipes.values())},t.prototype.createDirectiveHostPropertyAsts=function(t,e,n){var r=this;if(t.hostProperties){var i=[];return Object.keys(t.hostProperties).forEach(function(e){var o=t.hostProperties[e];"string"==typeof o?r.parsePropertyBinding(e,o,!0,n,[],i):r._reportError('Value of the host property binding "'+e+'" needs to be a string representing an expression but got "'+o+'" ('+typeof o+")",n)}),i.map(function(t){return r.createElementPropertyAst(e,t)})}return null},t.prototype.createDirectiveHostEventAsts=function(t,e){var n=this;if(t.hostListeners){var r=[];return Object.keys(t.hostListeners).forEach(function(i){var o=t.hostListeners[i];"string"==typeof o?n.parseEvent(i,o,e,[],r):n._reportError('Value of the host listener "'+i+'" needs to be a string representing an expression but got "'+o+'" ('+typeof o+")",e)}),r}return null},t.prototype.parseInterpolation=function(t,e){var n=e.start.toString();try{var r=this._exprParser.parseInterpolation(t,n,this._interpolationConfig);return r&&this._reportExpressionParserErrors(r.errors,e),this._checkPipes(r,e),r}catch(t){return this._reportError(""+t,e),this._exprParser.wrapLiteralPrimitive("ERROR",n)}},t.prototype.parseInlineTemplateBinding=function(t,e,n,r,i,o){for(var s=this._parseTemplateBindings(t,e,n),a=0;a<s.length;a++){var c=s[a];c.keyIsVar?o.push(new rt(c.key,c.name,n)):c.expression?this._parsePropertyAst(c.key,c.expression,n,r,i):(r.push([c.key,""]),this.parseLiteralAttr(c.key,null,n,r,i))}},t.prototype._parseTemplateBindings=function(t,e,n){var r=this,i=n.start.toString();try{var o=this._exprParser.parseTemplateBindings(t,e,i);return this._reportExpressionParserErrors(o.errors,n),o.templateBindings.forEach(function(t){t.expression&&r._checkPipes(t.expression,n)}),o.warnings.forEach(function(t){r._reportError(t,n,sr.WARNING)}),o.templateBindings}catch(t){return this._reportError(""+t,n),[]}},t.prototype.parseLiteralAttr=function(t,e,n,r,i){Ac(t)?(t=t.substring(1),e&&this._reportError('Assigning animation triggers via @prop="exp" attributes with an expression is invalid. Use property bindings (e.g. [@prop]="exp") or use an attribute without a value (e.g. @prop) instead.',n,sr.ERROR),this._parseAnimation(t,e,n,r,i)):i.push(new Sc(t,this._exprParser.wrapLiteralPrimitive(e,""),Ec.LITERAL_ATTR,n))},t.prototype.parsePropertyBinding=function(t,e,n,r,i,o){var s=!1;t.startsWith("animate-")?(s=!0,t=t.substring("animate-".length)):Ac(t)&&(s=!0,t=t.substring(1)),s?this._parseAnimation(t,e,r,i,o):this._parsePropertyAst(t,this._parseBinding(e,n,r),r,i,o)},t.prototype.parsePropertyInterpolation=function(t,e,n,r,i){var o=this.parseInterpolation(e,n);return!!o&&(this._parsePropertyAst(t,o,n,r,i),!0)},t.prototype._parsePropertyAst=function(t,e,n,r,i){r.push([t,e.source]),i.push(new Sc(t,e,Ec.DEFAULT,n))},t.prototype._parseAnimation=function(t,e,n,r,i){var o=this._parseBinding(e||"undefined",!1,n);r.push([t,o.source]),i.push(new Sc(t,o,Ec.ANIMATION,n))},t.prototype._parseBinding=function(t,e,n){var r=n.start.toString();try{var i=e?this._exprParser.parseSimpleBinding(t,r,this._interpolationConfig):this._exprParser.parseBinding(t,r,this._interpolationConfig);return i&&this._reportExpressionParserErrors(i.errors,n),this._checkPipes(i,n),i}catch(t){return this._reportError(""+t,n),this._exprParser.wrapLiteralPrimitive("ERROR",r)}},t.prototype.createElementPropertyAst=function(t,e){if(e.isAnimation)return new tt(e.name,pt.Animation,P.NONE,e.expression,null,e.sourceSpan);var n=null,r=void 0,i=null,o=e.name.split("."),s=void 0;if(o.length>1)if("attr"==o[0]){i=o[1],this._validatePropertyOrAttributeName(i,e.sourceSpan,!0),s=Tc(this._schemaRegistry,t,i,!0);var a=i.indexOf(":");if(a>-1)i=ae(i.substring(0,a),i.substring(a+1));r=pt.Attribute}else"class"==o[0]?(i=o[1],r=pt.Class,s=[P.NONE]):"style"==o[0]&&(n=o.length>2?o[2]:null,i=o[1],r=pt.Style,s=[P.STYLE]);return null===i&&(i=this._schemaRegistry.getMappedPropName(e.name),s=Tc(this._schemaRegistry,t,i,!1),r=pt.Property,this._validatePropertyOrAttributeName(i,e.sourceSpan,!1)),new tt(i,r,s[0],e.expression,n,e.sourceSpan)},t.prototype.parseEvent=function(t,e,n,r,i){Ac(t)?(t=t.substr(1),this._parseAnimationEvent(t,e,n,i)):this._parseEvent(t,e,n,r,i)},t.prototype._parseAnimationEvent=function(t,e,n,r){var i=I(t,".",[t,""]),o=i[0],s=i[1].toLowerCase();if(s)switch(s){case"start":case"done":var a=this._parseAction(e,n);r.push(new et(o,null,s,a,n));break;default:this._reportError('The provided animation output phase value "'+s+'" for "@'+o+'" is not supported (use start or done)',n)}else this._reportError("The animation trigger output event (@"+o+") is missing its phase value name (start or done are currently supported)",n)},t.prototype._parseEvent=function(t,e,n,r,i){var o=D(t,[null,t]),s=o[0],a=o[1],c=this._parseAction(e,n);r.push([t,c.source]),i.push(new et(a,s,null,c,n))},t.prototype._parseAction=function(t,e){var n=e.start.toString();try{var r=this._exprParser.parseAction(t,n,this._interpolationConfig);return r&&this._reportExpressionParserErrors(r.errors,e),!r||r.ast instanceof En?(this._reportError("Empty expressions are not allowed",e),this._exprParser.wrapLiteralPrimitive("ERROR",n)):(this._checkPipes(r,e),r)}catch(t){return this._reportError(""+t,e),this._exprParser.wrapLiteralPrimitive("ERROR",n)}},t.prototype._reportError=function(t,e,n){void 0===n&&(n=sr.ERROR),this._targetErrors.push(new ar(e,t,n))},t.prototype._reportExpressionParserErrors=function(t,e){for(var n=0,r=t;n<r.length;n++){var i=r[n];this._reportError(i.message,e)}},t.prototype._checkPipes=function(t,e){var n=this;if(t){var r=new Mc;t.visit(r),r.pipes.forEach(function(t,r){var i=n.pipesByName.get(r);i?n._usedPipes.set(r,i):n._reportError("The pipe '"+r+"' could not be found",new or(e.start.moveBy(t.span.start),e.start.moveBy(t.span.end)))})}},t.prototype._validatePropertyOrAttributeName=function(t,e,n){var r=n?this._schemaRegistry.validateAttribute(t):this._schemaRegistry.validateProperty(t);r.error&&this._reportError(r.msg,e,sr.ERROR)},t}(),Mc=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.pipes=new Map,e}return Object(r.b)(e,t),e.prototype.visitPipe=function(t,e){return this.pipes.set(t.name,t),t.exp.visit(this),this.visitAll(t.args,e),null},e}(Kn);function Ac(t){return"@"==t[0]}function Tc(t,e,n,r){var i=[];return jo.parse(e).forEach(function(e){var o=e.element?[e.element]:t.allKnownElementNames(),s=new Set(e.notSelectors.filter(function(t){return t.isElementSelector()}).map(function(t){return t.element})),a=o.filter(function(t){return!s.has(t)});i.push.apply(i,a.map(function(e){return t.securityContext(e,n,r)}))}),0===i.length?[P.NONE]:Array.from(new Set(i)).sort()}var Dc=/^(?:(?:(?:(bind-)|(let-)|(ref-|#)|(on-)|(bindon-)|(@))(.+))|\[\(([^\)]+)\)\]|\[([^\]]+)\]|\(([^\)]+)\))$/,Ic="template",Rc="class",Nc=jo.parse("*")[0],Lc="The <template> element is deprecated. Use <ng-template> instead",Fc="The template attribute is deprecated. Use an ng-template element instead.",Vc={};var zc=function(t){function e(e,n,r){return t.call(this,n,e,r)||this}return Object(r.b)(e,t),e}(ar),Bc=function(){return function(t,e,n){this.templateAst=t,this.usedPipes=e,this.errors=n}}(),qc=function(){function t(t,e,n,r,i,o,s){this._config=t,this._reflector=e,this._exprParser=n,this._schemaRegistry=r,this._htmlParser=i,this._console=o,this.transforms=s}return t.prototype.parse=function(t,e,n,r,i,o,s){var a,c=this.tryParse(t,e,n,r,i,o,s),u=c.errors.filter(function(t){return t.level===sr.WARNING}).filter((a=[Fc,Lc],function(t){return-1===a.indexOf(t.msg)||(Vc[t.msg]=(Vc[t.msg]||0)+1,Vc[t.msg]<=1)})),l=c.errors.filter(function(t){return t.level===sr.ERROR});if(u.length>0&&this._console.warn("Template parse warnings:\n"+u.join("\n")),l.length>0)throw q("Template parse errors:\n"+l.join("\n"),l);return{template:c.templateAst,pipes:c.usedPipes}},t.prototype.tryParse=function(t,e,n,r,i,o,s){var a,c="string"==typeof e?this._htmlParser.parse(e,o,!0,this.getInterpolationConfig(t)):e;return s||(a=c,c=new Or(Wt(new jc,a.rootNodes),a.errors)),this.tryParseHtml(this.expandHtml(c),t,n,r,i)},t.prototype.tryParseHtml=function(t,e,n,r,i){var o,s=t.errors,a=[];if(t.rootNodes.length>0){var c=Qc(n),u=Qc(r),l=new Xs(this._reflector,e),p=void 0;e.template&&e.template.interpolation&&(p={start:e.template.interpolation[0],end:e.template.interpolation[1]});var d=new Pc(this._exprParser,p,this._schemaRegistry,u,s),h=new Hc(this._reflector,this._config,l,c,d,this._schemaRegistry,i,s);o=Wt(h,t.rootNodes,Yc),s.push.apply(s,l.errors),a.push.apply(a,d.getUsedPipes())}else o=[];return this._assertNoReferenceDuplicationOnTemplate(o,s),s.length>0?new Bc(o,a,s):(this.transforms&&this.transforms.forEach(function(t){o=dt(t,o)}),new Bc(o,a,s))},t.prototype.expandHtml=function(t,e){void 0===e&&(e=!1);var n=t.errors;if(0==n.length||e){var r=wc(t.rootNodes);n.push.apply(n,r.errors),t=new Or(r.nodes,n)}return t},t.prototype.getInterpolationConfig=function(t){if(t.template)return Kt.fromArray(t.template.interpolation)},t.prototype._assertNoReferenceDuplicationOnTemplate=function(t,e){var n=[];t.filter(function(t){return!!t.references}).forEach(function(t){return t.references.forEach(function(t){var r=t.name;if(n.indexOf(r)<0)n.push(r);else{var i=new zc('Reference "#'+r+'" is defined several times',t.sourceSpan,sr.ERROR);e.push(i)}})})},t}(),Hc=function(){function t(t,e,n,r,i,o,s,a){var c=this;this.reflector=t,this.config=e,this.providerViewContext=n,this._bindingParser=i,this._schemaRegistry=o,this._schemas=s,this._targetErrors=a,this.selectorMatcher=new xo,this.directivesIndex=new Map,this.ngContentCount=0,this.contentQueryStartId=n.component.viewQueries.length+1,r.forEach(function(t,e){var n=jo.parse(t.selector);c.selectorMatcher.addSelectables(n,t),c.directivesIndex.set(t,e)})}return t.prototype.visitExpansion=function(t,e){return null},t.prototype.visitExpansionCase=function(t,e){return null},t.prototype.visitText=function(t,e){var n=e.findNgContentIndex(Nc),r=_c(t.value),i=this._bindingParser.parseInterpolation(r,t.sourceSpan);return i?new Z(i,n,t.sourceSpan):new Q(r,n,t.sourceSpan)},t.prototype.visitAttribute=function(t,e){return new J(t.name,t.value,t.sourceSpan)},t.prototype.visitComment=function(t,e){return null},t.prototype.visitElement=function(t,e){var n=this,r=this.contentQueryStartId,i=t.name,o=be(t);if(o.type===_e.SCRIPT||o.type===_e.STYLE)return null;if(o.type===_e.STYLESHEET&&Zt(o.hrefAttr))return null;var s=[],a=[],c=[],u=[],l=[],p=[],d=[],h=[],f=!1,m=[],g=function(t,e,n){if(r=t.name,"ng-template"===re(r)[1])return!0;var r;var i=re(t.name)[1];if(i.toLowerCase()===Ic&&e&&i.toLowerCase()===Ic)return n(Lc,t.sourceSpan),!0;return!1}(t,this.config.enableLegacyTemplate,function(t,e){return n._reportError(t,e,sr.WARNING)});t.attrs.forEach(function(t){var e,r,i=n._parseAttr(g,t,s,a,l,c,u),o=n._normalizeAttributeName(t.name);n.config.enableLegacyTemplate&&"template"==o?(n._reportError(Fc,t.sourceSpan,sr.WARNING),e=t.value):o.startsWith("*")&&(e=t.value,r=o.substring("*".length)+":");var v=null!=e;v&&(f&&n._reportError("Can't have multiple template bindings on one element. Use only one attribute named 'template' or prefixed with *",t.sourceSpan),f=!0,n._bindingParser.parseInlineTemplateBinding(r,e,t.sourceSpan,d,p,h)),i||v||(m.push(n.visitAttribute(t,null)),s.push([t.name,t.value]))});var v=Xc(i,s),y=this._parseDirectives(this.selectorMatcher,v),b=y.directives,_=y.matchElement,j=[],x=new Set,w=this._createDirectiveAsts(g,t.name,b,a,c,t.sourceSpan,j,x),O=this._createElementPropertyAsts(t.name,a,x),C=e.isTemplateElement||f,k=new Ys(this.providerViewContext,e.providerContext,C,w,m,j,g,r,t.sourceSpan),E=Wt(o.nonBindable?Kc:this,t.children,Gc.create(g,w,g?e.providerContext:k));k.afterElement();var S,P=null!=o.projectAs?jo.parse(o.projectAs)[0]:v,M=e.findNgContentIndex(P);if(o.type===_e.NG_CONTENT)t.children&&!t.children.every($c)&&this._reportError("<ng-content> element cannot have content.",t.sourceSpan),S=new lt(this.ngContentCount++,f?null:M,t.sourceSpan);else if(g)this._assertAllEventsPublishedByDirectives(w,l),this._assertNoComponentsNorElementBindingsOnTemplate(w,O,t.sourceSpan),S=new ot(m,l,j,u,k.transformedDirectiveAsts,k.transformProviders,k.transformedHasViewContainer,k.queryMatches,E,f?null:M,t.sourceSpan);else{this._assertElementExists(_,t),this._assertOnlyOneComponent(w,t.sourceSpan);var A=f?null:e.findNgContentIndex(P);S=new it(i,m,O,l,j,k.transformedDirectiveAsts,k.transformProviders,k.transformedHasViewContainer,k.queryMatches,E,f?null:A,t.sourceSpan,t.endSourceSpan||null)}if(f){var T=this.contentQueryStartId,D=Xc(Ic,d),I=this._parseDirectives(this.selectorMatcher,D).directives,R=new Set,N=this._createDirectiveAsts(!0,t.name,I,p,[],t.sourceSpan,[],R),L=this._createElementPropertyAsts(t.name,p,R);this._assertNoComponentsNorElementBindingsOnTemplate(N,L,t.sourceSpan);var F=new Ys(this.providerViewContext,e.providerContext,e.isTemplateElement,N,[],[],!0,T,t.sourceSpan);F.afterElement(),S=new ot([],[],[],h,F.transformedDirectiveAsts,F.transformProviders,F.transformedHasViewContainer,F.queryMatches,[S],M,t.sourceSpan)}return S},t.prototype._parseAttr=function(t,e,n,r,i,o,s){var a=this._normalizeAttributeName(e.name),c=e.value,u=e.sourceSpan,l=a.match(Dc),p=!1;if(null!==l)if(p=!0,null!=l[1])this._bindingParser.parsePropertyBinding(l[7],c,!1,u,n,r);else if(l[2])if(t){var d=l[7];this._parseVariable(d,c,u,s)}else this._reportError('"let-" is only supported on ng-template elements.',u);else if(l[3]){d=l[7];this._parseReference(d,c,u,o)}else l[4]?this._bindingParser.parseEvent(l[7],c,u,n,i):l[5]?(this._bindingParser.parsePropertyBinding(l[7],c,!1,u,n,r),this._parseAssignmentEvent(l[7],c,u,n,i)):l[6]?this._bindingParser.parseLiteralAttr(a,c,u,n,r):l[8]?(this._bindingParser.parsePropertyBinding(l[8],c,!1,u,n,r),this._parseAssignmentEvent(l[8],c,u,n,i)):l[9]?this._bindingParser.parsePropertyBinding(l[9],c,!1,u,n,r):l[10]&&this._bindingParser.parseEvent(l[10],c,u,n,i);else p=this._bindingParser.parsePropertyInterpolation(a,c,u,n,r);return p||this._bindingParser.parseLiteralAttr(a,c,u,n,r),p},t.prototype._normalizeAttributeName=function(t){return/^data-/i.test(t)?t.substring(5):t},t.prototype._parseVariable=function(t,e,n,r){t.indexOf("-")>-1&&this._reportError('"-" is not allowed in variable names',n),r.push(new rt(t,e,n))},t.prototype._parseReference=function(t,e,n,r){t.indexOf("-")>-1&&this._reportError('"-" is not allowed in reference names',n),r.push(new Wc(t,e,n))},t.prototype._parseAssignmentEvent=function(t,e,n,r,i){this._bindingParser.parseEvent(t+"Change",e+"=$event",n,r,i)},t.prototype._parseDirectives=function(t,e){var n=this,r=new Array(this.directivesIndex.size),i=!1;return t.match(e,function(t,e){r[n.directivesIndex.get(e)]=e,i=i||t.hasElementSelector()}),{directives:r.filter(function(t){return!!t}),matchElement:i}},t.prototype._createDirectiveAsts=function(t,e,n,r,i,o,s,a){var c=this,u=new Set,l=null,p=n.map(function(t){var n=new or(o.start,o.end,"Directive "+_t(t.type));t.isComponent&&(l=t);var p=[],d=c._bindingParser.createDirectiveHostPropertyAsts(t,e,n);d=c._checkPropertiesInSchema(e,d);var h=c._bindingParser.createDirectiveHostEventAsts(t,n);c._createDirectivePropertyAsts(t.inputs,r,p,a),i.forEach(function(e){(0===e.value.length&&t.isComponent||e.isReferenceToDirective(t))&&(s.push(new nt(e.name,mo(t.type.reference),e.sourceSpan)),u.add(e.name))});var f=c.contentQueryStartId;return c.contentQueryStartId+=t.queries.length,new at(t,p,d,h,f,n)});return i.forEach(function(e){if(e.value.length>0)u.has(e.name)||c._reportError('There is no directive with "exportAs" set to "'+e.value+'"',e.sourceSpan);else if(!l){var n=null;t&&(n=go(c.reflector,fo.TemplateRef)),s.push(new nt(e.name,n,e.sourceSpan))}}),p},t.prototype._createDirectivePropertyAsts=function(t,e,n,r){if(t){var i=new Map;e.forEach(function(t){var e=i.get(t.name);e&&!e.isLiteral||i.set(t.name,t)}),Object.keys(t).forEach(function(e){var o=t[e],s=i.get(o);s&&(r.add(s.name),Zc(s.expression)||n.push(new st(e,s.name,s.expression,s.sourceSpan)))})}},t.prototype._createElementPropertyAsts=function(t,e,n){var r=this,i=[];return e.forEach(function(e){e.isLiteral||n.has(e.name)||i.push(r._bindingParser.createElementPropertyAst(t,e))}),this._checkPropertiesInSchema(t,i)},t.prototype._findComponentDirectives=function(t){return t.filter(function(t){return t.directive.isComponent})},t.prototype._findComponentDirectiveNames=function(t){return this._findComponentDirectives(t).map(function(t){return _t(t.directive.type)})},t.prototype._assertOnlyOneComponent=function(t,e){var n=this._findComponentDirectiveNames(t);n.length>1&&this._reportError("More than one component matched on this element.\nMake sure that only one component's selector can match a given element.\nConflicting components: "+n.join(","),e)},t.prototype._assertElementExists=function(t,e){var n=e.name.replace(/^:xhtml:/,"");if(!t&&!this._schemaRegistry.hasElement(n,this._schemas)){var r="'"+n+"' is not a known element:\n";r+="1. If '"+n+"' is an Angular component, then verify that it is part of this module.\n",n.indexOf("-")>-1?r+="2. If '"+n+"' is a Web Component then add 'CUSTOM_ELEMENTS_SCHEMA' to the '@NgModule.schemas' of this component to suppress this message.":r+="2. To allow any element add 'NO_ERRORS_SCHEMA' to the '@NgModule.schemas' of this component.",this._reportError(r,e.sourceSpan)}},t.prototype._assertNoComponentsNorElementBindingsOnTemplate=function(t,e,n){var r=this,i=this._findComponentDirectiveNames(t);i.length>0&&this._reportError("Components on an embedded template: "+i.join(","),n),e.forEach(function(t){r._reportError("Property binding "+t.name+' not used by any directive on an embedded template. Make sure that the property name is spelled correctly and all directives are listed in the "@NgModule.declarations".',n)})},t.prototype._assertAllEventsPublishedByDirectives=function(t,e){var n=this,r=new Set;t.forEach(function(t){Object.keys(t.directive.outputs).forEach(function(e){var n=t.directive.outputs[e];r.add(n)})}),e.forEach(function(t){null==t.target&&r.has(t.name)||n._reportError("Event binding "+t.fullName+' not emitted by any directive on an embedded template. Make sure that the event name is spelled correctly and all directives are listed in the "@NgModule.declarations".',t.sourceSpan)})},t.prototype._checkPropertiesInSchema=function(t,e){var n=this;return e.filter(function(e){if(e.type===pt.Property&&!n._schemaRegistry.hasProperty(t,e.name,n._schemas)){var r="Can't bind to '"+e.name+"' since it isn't a known property of '"+t+"'.";t.startsWith("ng-")?r+="\n1. If '"+e.name+"' is an Angular directive, then add 'CommonModule' to the '@NgModule.imports' of this component.\n2. To allow any property add 'NO_ERRORS_SCHEMA' to the '@NgModule.schemas' of this component.":t.indexOf("-")>-1&&(r+="\n1. If '"+t+"' is an Angular component and it has '"+e.name+"' input, then verify that it is part of this module.\n2. If '"+t+"' is a Web Component then add 'CUSTOM_ELEMENTS_SCHEMA' to the '@NgModule.schemas' of this component to suppress this message.\n3. To allow any property add 'NO_ERRORS_SCHEMA' to the '@NgModule.schemas' of this component."),n._reportError(r,e.sourceSpan)}return!Zc(e.value)})},t.prototype._reportError=function(t,e,n){void 0===n&&(n=sr.ERROR),this._targetErrors.push(new ar(e,t,n))},t}(),Uc=function(){function t(){}return t.prototype.visitElement=function(t,e){var n=be(t);if(n.type===_e.SCRIPT||n.type===_e.STYLE||n.type===_e.STYLESHEET)return null;var r=t.attrs.map(function(t){return[t.name,t.value]}),i=Xc(t.name,r),o=e.findNgContentIndex(i),s=Wt(this,t.children,Yc);return new it(t.name,Wt(this,t.attrs),[],[],[],[],[],!1,[],s,o,t.sourceSpan,t.endSourceSpan)},t.prototype.visitComment=function(t,e){return null},t.prototype.visitAttribute=function(t,e){return new J(t.name,t.value,t.sourceSpan)},t.prototype.visitText=function(t,e){var n=e.findNgContentIndex(Nc);return new Q(t.value,n,t.sourceSpan)},t.prototype.visitExpansion=function(t,e){return t},t.prototype.visitExpansionCase=function(t,e){return t},t}(),Wc=function(){function t(t,e,n){this.name=t,this.value=e,this.sourceSpan=n}return t.prototype.isReferenceToDirective=function(t){return-1!==(e=t.exportAs,e?e.split(",").map(function(t){return t.trim()}):[]).indexOf(this.value);var e},t}();var Gc=function(){function t(t,e,n,r){this.isTemplateElement=t,this._ngContentIndexMatcher=e,this._wildcardNgContentIndex=n,this.providerContext=r}return t.create=function(e,n,r){var i=new xo,o=null,s=n.find(function(t){return t.directive.isComponent});if(s)for(var a=s.directive.template.ngContentSelectors,c=0;c<a.length;c++){"*"===a[c]?o=c:i.addSelectables(jo.parse(a[c]),c)}return new t(e,i,o,r)},t.prototype.findNgContentIndex=function(t){var e=[];return this._ngContentIndexMatcher.match(t,function(t,n){e.push(n)}),e.sort(),null!=this._wildcardNgContentIndex&&e.push(this._wildcardNgContentIndex),e.length>0?e[0]:null},t}();function Xc(t,e){var n=new jo,r=re(t)[1];n.setElement(r);for(var i=0;i<e.length;i++){var o=e[i][0],s=re(o)[1],a=e[i][1];if(n.addAttribute(s,a),o.toLowerCase()==Rc)a.trim().split(/\s+/g).forEach(function(t){return n.addClassName(t)})}return n}var Yc=new Gc(!0,new xo,null,null),Kc=new Uc;function $c(t){return t instanceof Vt&&0==t.value.trim().length}function Qc(t){var e=new Map;return t.forEach(function(t){e.get(t.type.reference)||e.set(t.type.reference,t)}),Array.from(e.values())}function Zc(t){return t instanceof Xn&&(t=t.ast),t instanceof En}var Jc=function(){function t(){}return t.event=Fs("$event"),t}(),tu=function(){return function(t,e){this.stmts=t,this.allowDefault=e}}();function eu(t,e,n,r){t||(t=new du);var i=nu({createLiteralArrayConverter:function(t){return function(t){return qs(t)}},createLiteralMapConverter:function(t){return function(e){return Hs(t.map(function(t,n){return{key:t.key,value:e[n],quoted:t.quoted}}))}},createPipeConverter:function(t){throw new Error("Illegal State: Actions are not allowed to contain pipes. Pipe: "+t)}},n),o=new pu(t,e,r),s=[];!function t(e,n){Array.isArray(e)?e.forEach(function(e){return t(e,n)}):n.push(e)}(i.visit(o,au.Statement),s),function(t,e,n){for(var r=t-1;r>=0;r--)n.unshift(su(e,r))}(o.temporaryCount,r,s);var a=s.length-1,c=null;if(a>=0){var u=function(t){if(t instanceof xs)return t.expr;if(t instanceof ws)return t.value;return null}(s[a]);u&&(c=Fs("pd_"+r),s[a]=c.set(u.cast(Fo).notIdentical(Ws(!1))).toDeclStmt(null,[ys.Final]))}return new tu(s,c)}function nu(t,e){return n=e,r=new lu(t),n.visit(r);var n,r}var ru=function(){return function(t,e){this.stmts=t,this.currValExpr=e}}();function iu(t,e,n,r){t||(t=new du);var i=Fs("currVal_"+r),o=[],s=new pu(t,e,r),a=n.visit(s,au.Expression);if(s.temporaryCount)for(var c=0;c<s.temporaryCount;c++)o.push(su(r,c));return o.push(i.set(a).toDeclStmt(null,[ys.Final])),new ru(o,i)}function ou(t,e){return"tmp_"+t+"_"+e}function su(t,e){return new _s(ou(t,e),gs)}var au={Statement:0,Expression:1};function cu(t,e){if(t!==au.Expression)throw new Error("Expected an expression, but saw "+e)}function uu(t,e){return t===au.Statement?e.toStmt():e}au[au.Statement]="Statement",au[au.Expression]="Expression";var lu=function(t){function e(e){var n=t.call(this)||this;return n._converterFactory=e,n}return Object(r.b)(e,t),e.prototype.visitPipe=function(t,e){var n=this,r=[t.exp].concat(t.args).map(function(t){return t.visit(n,e)});return new hu(t.span,r,this._converterFactory.createPipeConverter(t.name,r.length))},e.prototype.visitLiteralArray=function(t,e){var n=this,r=t.expressions.map(function(t){return t.visit(n,e)});return new hu(t.span,r,this._converterFactory.createLiteralArrayConverter(t.expressions.length))},e.prototype.visitLiteralMap=function(t,e){var n=this,r=t.values.map(function(t){return t.visit(n,e)});return new hu(t.span,r,this._converterFactory.createLiteralMapConverter(t.keys))},e}($n),pu=function(){function t(t,e,n){this._localResolver=t,this._implicitReceiver=e,this.bindingId=n,this._nodeMap=new Map,this._resultMap=new Map,this._currentTemporary=0,this.temporaryCount=0}return t.prototype.visitBinary=function(t,e){var n;switch(t.operation){case"+":n=Bo.Plus;break;case"-":n=Bo.Minus;break;case"*":n=Bo.Multiply;break;case"/":n=Bo.Divide;break;case"%":n=Bo.Modulo;break;case"&&":n=Bo.And;break;case"||":n=Bo.Or;break;case"==":n=Bo.Equals;break;case"!=":n=Bo.NotEquals;break;case"===":n=Bo.Identical;break;case"!==":n=Bo.NotIdentical;break;case"<":n=Bo.Lower;break;case">":n=Bo.Bigger;break;case"<=":n=Bo.LowerEquals;break;case">=":n=Bo.BiggerEquals;break;default:throw new Error("Unsupported operation "+t.operation)}return uu(e,new us(n,this._visit(t.left,au.Expression),this._visit(t.right,au.Expression)))},t.prototype.visitChain=function(t,e){return function(t,e){if(t!==au.Statement)throw new Error("Expected a statement, but saw "+e)}(e,t),this.visitAll(t.expressions,e)},t.prototype.visitConditional=function(t,e){return uu(e,this._visit(t.condition,au.Expression).conditional(this._visit(t.trueExp,au.Expression),this._visit(t.falseExp,au.Expression)))},t.prototype.visitPipe=function(t,e){throw new Error("Illegal state: Pipes should have been converted into functions. Pipe: "+t.name)},t.prototype.visitFunctionCall=function(t,e){var n=this.visitAll(t.args,au.Expression);return uu(e,t instanceof hu?t.converter(n):this._visit(t.target,au.Expression).callFn(n))},t.prototype.visitImplicitReceiver=function(t,e){return cu(e,t),this._implicitReceiver},t.prototype.visitInterpolation=function(t,e){cu(e,t);for(var n=[Ws(t.expressions.length)],r=0;r<t.strings.length-1;r++)n.push(Ws(t.strings[r])),n.push(this._visit(t.expressions[r],au.Expression));return n.push(Ws(t.strings[t.strings.length-1])),t.expressions.length<=9?Vs(fo.inlineInterpolate).callFn(n):Vs(fo.interpolate).callFn([n[0],qs(n.slice(1))])},t.prototype.visitKeyedRead=function(t,e){var n=this.leftMostSafeNode(t);return n?this.convertSafeAccess(t,n,e):uu(e,this._visit(t.obj,au.Expression).key(this._visit(t.key,au.Expression)))},t.prototype.visitKeyedWrite=function(t,e){var n=this._visit(t.obj,au.Expression),r=this._visit(t.key,au.Expression),i=this._visit(t.value,au.Expression);return uu(e,n.key(r).set(i))},t.prototype.visitLiteralArray=function(t,e){throw new Error("Illegal State: literal arrays should have been converted into functions")},t.prototype.visitLiteralMap=function(t,e){throw new Error("Illegal State: literal maps should have been converted into functions")},t.prototype.visitLiteralPrimitive=function(t,e){return uu(e,Ws(t.value))},t.prototype._getLocal=function(t){return this._localResolver.getLocal(t)},t.prototype.visitMethodCall=function(t,e){var n=this.leftMostSafeNode(t);if(n)return this.convertSafeAccess(t,n,e);var r=this.visitAll(t.args,au.Expression),i=null,o=this._visit(t.receiver,au.Expression);if(o===this._implicitReceiver){var s=this._getLocal(t.name);s&&(i=s.callFn(r))}return null==i&&(i=o.callMethod(t.name,r)),uu(e,i)},t.prototype.visitPrefixNot=function(t,e){return uu(e,(n=this._visit(t.expression,au.Expression),new is(n,r)));var n,r},t.prototype.visitNonNullAssert=function(t,e){return uu(e,(n=this._visit(t.expression,au.Expression),new os(n,r)));var n,r},t.prototype.visitPropertyRead=function(t,e){var n=this.leftMostSafeNode(t);if(n)return this.convertSafeAccess(t,n,e);var r=null,i=this._visit(t.receiver,au.Expression);return i===this._implicitReceiver&&(r=this._getLocal(t.name)),null==r&&(r=i.prop(t.name)),uu(e,r)},t.prototype.visitPropertyWrite=function(t,e){var n=this._visit(t.receiver,au.Expression);if(n===this._implicitReceiver&&this._getLocal(t.name))throw new Error("Cannot assign to a reference or variable!");return uu(e,n.prop(t.name).set(this._visit(t.value,au.Expression)))},t.prototype.visitSafePropertyRead=function(t,e){return this.convertSafeAccess(t,this.leftMostSafeNode(t),e)},t.prototype.visitSafeMethodCall=function(t,e){return this.convertSafeAccess(t,this.leftMostSafeNode(t),e)},t.prototype.visitAll=function(t,e){var n=this;return t.map(function(t){return n._visit(t,e)})},t.prototype.visitQuote=function(t,e){throw new Error("Quotes are not supported for evaluation!\n        Statement: "+t.uninterpretedExpression+" located at "+t.location)},t.prototype._visit=function(t,e){var n=this._resultMap.get(t);return n||(this._nodeMap.get(t)||t).visit(this,e)},t.prototype.convertSafeAccess=function(t,e,n){var r=this._visit(e.receiver,au.Expression),i=void 0;this.needsTemporary(e.receiver)&&(r=(i=this.allocateTemporary()).set(r),this._resultMap.set(e.receiver,i));var o=r.isBlank();e instanceof Wn?this._nodeMap.set(e,new Un(e.span,e.receiver,e.name,e.args)):this._nodeMap.set(e,new An(e.span,e.receiver,e.name));var s=this._visit(t,au.Expression);return this._nodeMap.delete(e),i&&this.releaseTemporary(i),uu(n,o.conditional(Ws(null),s))},t.prototype.leftMostSafeNode=function(t){var e=this,n=function(t,n){return(e._nodeMap.get(n)||n).visit(t)};return t.visit({visitBinary:function(t){return null},visitChain:function(t){return null},visitConditional:function(t){return null},visitFunctionCall:function(t){return null},visitImplicitReceiver:function(t){return null},visitInterpolation:function(t){return null},visitKeyedRead:function(t){return n(this,t.obj)},visitKeyedWrite:function(t){return null},visitLiteralArray:function(t){return null},visitLiteralMap:function(t){return null},visitLiteralPrimitive:function(t){return null},visitMethodCall:function(t){return n(this,t.receiver)},visitPipe:function(t){return null},visitPrefixNot:function(t){return null},visitNonNullAssert:function(t){return null},visitPropertyRead:function(t){return n(this,t.receiver)},visitPropertyWrite:function(t){return null},visitQuote:function(t){return null},visitSafeMethodCall:function(t){return n(this,t.receiver)||t},visitSafePropertyRead:function(t){return n(this,t.receiver)||t}})},t.prototype.needsTemporary=function(t){var e=this,n=function(t,n){return n&&(e._nodeMap.get(n)||n).visit(t)};return t.visit({visitBinary:function(t){return n(this,t.left)||n(this,t.right)},visitChain:function(t){return!1},visitConditional:function(t){return n(this,t.condition)||n(this,t.trueExp)||n(this,t.falseExp)},visitFunctionCall:function(t){return!0},visitImplicitReceiver:function(t){return!1},visitInterpolation:function(t){return e=this,t.expressions.some(function(t){return n(e,t)});var e},visitKeyedRead:function(t){return!1},visitKeyedWrite:function(t){return!1},visitLiteralArray:function(t){return!0},visitLiteralMap:function(t){return!0},visitLiteralPrimitive:function(t){return!1},visitMethodCall:function(t){return!0},visitPipe:function(t){return!0},visitPrefixNot:function(t){return n(this,t.expression)},visitNonNullAssert:function(t){return n(this,t.expression)},visitPropertyRead:function(t){return!1},visitPropertyWrite:function(t){return!1},visitQuote:function(t){return!1},visitSafeMethodCall:function(t){return!0},visitSafePropertyRead:function(t){return!1}})},t.prototype.allocateTemporary=function(){var t=this._currentTemporary++;return this.temporaryCount=Math.max(this._currentTemporary,this.temporaryCount),new Go(ou(this.bindingId,t))},t.prototype.releaseTemporary=function(t){if(this._currentTemporary--,t.name!=ou(this.bindingId,this._currentTemporary))throw new Error("Temporary "+t.name+" released out of order")},t}();var du=function(){function t(){}return t.prototype.getLocal=function(t){return t===Jc.event.name?Jc.event:null},t}();var hu=function(t){function e(e,n,r){var i=t.call(this,e,null,n)||this;return i.args=n,i.converter=r,i}return Object(r.b)(e,t),e}(Gn),fu=(function(){function t(t,e){this.options=t,this.reflector=e}t.prototype.compileComponent=function(t,e,n,r,i){var o=this,s=new Map;r.forEach(function(t){return s.set(t.name,t.type.reference)});var a=0,c=function(t){var n=a++;return new mu(o.options,o.reflector,i,t,e.type.reference,e.isHost,n,s,c)},u=c(null);return u.visitAll([],n),u.build(t)}}(),new(function(){function t(){}return t.prototype.getLocal=function(t){return t===Jc.event.name?Fs("_any"):null},t}())),mu=function(){function t(t,e,n,r,i,o,s,a,c){this.options=t,this.reflector=e,this.externalReferenceVars=n,this.parent=r,this.component=i,this.isHostComponent=o,this.embeddedViewIndex=s,this.pipes=a,this.viewBuilderFactory=c,this.refOutputVars=new Map,this.variables=[],this.children=[],this.updates=[],this.actions=[]}return t.prototype.getOutputVar=function(t){var e;if(!(e=t===this.component&&this.isHostComponent?"_any":t instanceof mt?this.externalReferenceVars.get(t):"_any"))throw new Error("Illegal State: referring to a type without a variable "+JSON.stringify(t));return e},t.prototype.visitAll=function(t,e){this.variables=t,dt(this,e)},t.prototype.build=function(t,e){var n=this;void 0===e&&(e=[]),this.children.forEach(function(n){return n.build(t,e)});var r=[Fs("_any").set(gs).toDeclStmt(Fo)],i=0;this.updates.forEach(function(t){var e=n.preprocessUpdateExpression(t),o=e.sourceSpan,s=e.context,a=e.value,c=""+i++,u=iu(s===n.component?n:fu,Fs(n.getOutputVar(s)),a,c),l=u.stmts,p=u.currValExpr;l.push(new xs(p)),r.push.apply(r,l.map(function(t){return Rs(t,o)}))}),this.actions.forEach(function(t){var e=t.sourceSpan,o=t.context,s=t.value,a=""+i++,c=eu(o===n.component?n:fu,Fs(n.getOutputVar(o)),s,a).stmts;r.push.apply(r,c.map(function(t){return Rs(t,e)}))});var o="_View_"+t+"_"+this.embeddedViewIndex,s=new js(o,[],r);return e.push(s),e},t.prototype.visitBoundText=function(t,e){var n=this;t.value.ast.expressions.forEach(function(e){return n.updates.push({context:n.component,value:e,sourceSpan:t.sourceSpan})})},t.prototype.visitEmbeddedTemplate=function(t,e){if(this.visitElementOrTemplate(t),this.options.fullTemplateTypeCheck){var n=this.viewBuilderFactory(this);this.children.push(n),n.visitAll(t.variables,t.children)}},t.prototype.visitElement=function(t,e){var n=this;this.visitElementOrTemplate(t);t.inputs.forEach(function(t){n.updates.push({context:n.component,value:t.value,sourceSpan:t.sourceSpan})}),dt(this,t.children)},t.prototype.visitElementOrTemplate=function(t){var e=this;t.directives.forEach(function(t){e.visitDirective(t)}),t.references.forEach(function(t){var n=null;n=t.value&&t.value.identifier&&e.options.fullTemplateTypeCheck?t.value.identifier.reference:Do.Dynamic,e.refOutputVars.set(t.name,n)}),t.outputs.forEach(function(t){e.actions.push({context:e.component,value:t.handler,sourceSpan:t.sourceSpan})})},t.prototype.visitDirective=function(t){var e=this,n=t.directive.type.reference;t.inputs.forEach(function(t){return e.updates.push({context:e.component,value:t.value,sourceSpan:t.sourceSpan})}),this.options.fullTemplateTypeCheck&&(t.hostProperties.forEach(function(t){return e.updates.push({context:n,value:t.value,sourceSpan:t.sourceSpan})}),t.hostEvents.forEach(function(t){return e.actions.push({context:n,value:t.handler,sourceSpan:t.sourceSpan})}))},t.prototype.getLocal=function(t){if(t==Jc.event.name)return Fs(this.getOutputVar(Do.Dynamic));for(var e=this;e;e=e.parent){var n=void 0;if(null==(n=e.refOutputVars.get(t)))e.variables.find(function(e){return e.name===t})&&(n=Do.Dynamic);if(null!=n)return Fs(this.getOutputVar(n))}return null},t.prototype.pipeOutputVar=function(t){var e=this.pipes.get(t);if(!e)throw new Error("Illegal State: Could not find pipe "+t+" in template of "+this.component);return this.getOutputVar(e)},t.prototype.preprocessUpdateExpression=function(t){var e=this;return{sourceSpan:t.sourceSpan,context:t.context,value:nu({createLiteralArrayConverter:function(t){return function(t){var n=qs(t);return e.options.fullTemplateTypeCheck?n:n.cast(Fo)}},createLiteralMapConverter:function(t){return function(n){var r=Hs(t.map(function(t,e){return{key:t.key,value:n[e],quoted:t.quoted}}));return e.options.fullTemplateTypeCheck?r:r.cast(Fo)}},createPipeConverter:function(t,n){return function(n){return(e.options.fullTemplateTypeCheck?Fs(e.pipeOutputVar(t)):Fs(e.getOutputVar(Do.Dynamic))).callMethod("transform",n)}}},t.value)}},t.prototype.visitNgContent=function(t,e){},t.prototype.visitText=function(t,e){},t.prototype.visitDirectiveProperty=function(t,e){},t.prototype.visitReference=function(t,e){},t.prototype.visitVariable=function(t,e){},t.prototype.visitEvent=function(t,e){},t.prototype.visitElementProperty=function(t,e){},t.prototype.visitAttr=function(t,e){},t}(),gu="class",vu="style",yu=function(){return function(t,e){this.viewClassVar=t,this.rendererTypeVar=e}}(),bu=function(){function t(t){this._reflector=t}return t.prototype.compileComponent=function(t,e,n,r,i){var o=this,s=0,a=function t(e,n){void 0===n&&(n=new Map);e.forEach(function(e){var r=new Set,i=new Set,o=void 0;e instanceof it?(t(e.children,n),e.children.forEach(function(t){var e=n.get(t);e.staticQueryIds.forEach(function(t){return r.add(t)}),e.dynamicQueryIds.forEach(function(t){return i.add(t)})}),o=e.queryMatches):e instanceof ot&&(t(e.children,n),e.children.forEach(function(t){var e=n.get(t);e.staticQueryIds.forEach(function(t){return i.add(t)}),e.dynamicQueryIds.forEach(function(t){return i.add(t)})}),o=e.queryMatches),o&&o.forEach(function(t){return r.add(t.queryId)}),i.forEach(function(t){return r.delete(t)}),n.set(e,{staticQueryIds:r,dynamicQueryIds:i})});return n}(n),c=void 0;if(!e.isHost){var u=e.template,l=[];u.animations&&u.animations.length&&l.push(new hs("animation",ta(t,u.animations),!0));var p=Fs(wt(e.type.reference));c=p.name,t.statements.push(p.set(Vs(fo.createRendererType2).callFn([new fs([new hs("encapsulation",Ws(u.encapsulation),!1),new hs("styles",r,!1),new hs("data",new fs(l),!1)])])).toDeclStmt(zs(fo.RendererType2),[ys.Final,ys.Exported]))}var d,h=function(n){var r=s++;return new ku(o._reflector,t,n,e,r,i,a,h)},f=h(null);return f.visitAll([],n),(d=t.statements).push.apply(d,f.build()),new yu(f.viewName,c)},t}(),_u=Fs("_l"),ju=Fs("_v"),xu=Fs("_ck"),wu=Fs("_co"),Ou=Fs("en"),Cu=Fs("ad"),ku=function(){function t(t,e,n,r,i,o,s,a){this.reflector=t,this.outputCtx=e,this.parent=n,this.component=r,this.embeddedViewIndex=i,this.usedPipes=o,this.staticQueryIds=s,this.viewBuilderFactory=a,this.nodes=[],this.purePipeNodeIndices=Object.create(null),this.refNodeIndices=Object.create(null),this.variables=[],this.children=[],this.compType=this.embeddedViewIndex>0?Fo:Bs(e.importExpr(this.component.type.reference)),this.viewName=xt(this.component.type.reference,this.embeddedViewIndex)}return t.prototype.visitAll=function(t,e){var n,r,i,o=this;if(this.variables=t,this.parent||this.usedPipes.forEach(function(t){t.pure&&(o.purePipeNodeIndices[t.name]=o._createPipe(null,t))}),!this.parent){var s=(n=this.staticQueryIds,r=new Set,i=new Set,Array.from(n.values()).forEach(function(t){t.staticQueryIds.forEach(function(t){return r.add(t)}),t.dynamicQueryIds.forEach(function(t){return i.add(t)})}),i.forEach(function(t){return r.delete(t)}),{staticQueryIds:r,dynamicQueryIds:i});this.component.viewQueries.forEach(function(t,e){var n=e+1,r=t.first?0:1,i=134217728|Mu(s,n,t.first);o.nodes.push(function(){return{sourceSpan:null,nodeFlags:i,nodeDef:Vs(fo.queryDef).callFn([Ws(i),Ws(n),new fs([new hs(t.propertyName,Ws(r),!1)])])}})})}dt(this,e),this.parent&&(0===e.length||function t(e){var n=e[e.length-1];if(n instanceof ot)return n.hasViewContainer;if(n instanceof it)return ie(n.name)&&n.children.length?t(n.children):n.hasViewContainer;return n instanceof lt}(e))&&this.nodes.push(function(){return{sourceSpan:null,nodeFlags:1,nodeDef:Vs(fo.anchorDef).callFn([Ws(0),gs,gs,Ws(0)])}})},t.prototype.build=function(t){void 0===t&&(t=[]),this.children.forEach(function(e){return e.build(t)});var e=this._createNodeExpressions(),n=e.updateRendererStmts,r=e.updateDirectivesStmts,i=e.nodeDefExprs,o=this._createUpdateFn(n),s=this._createUpdateFn(r),a=0;this.parent||this.component.changeDetection!==h.OnPush||(a|=2);var c=new js(this.viewName,[new as(_u.name)],[new ws(Vs(fo.viewDef).callFn([Ws(a),qs(i),s,o]))],zs(fo.ViewDefinition),0===this.embeddedViewIndex?[ys.Exported]:[]);return t.push(c),t},t.prototype._createUpdateFn=function(t){var e;if(t.length>0){var n=[];!this.component.isHost&&Ds(t).has(wu.name)&&n.push(wu.set(ju.prop("component")).toDeclStmt(this.compType)),e=Us([new as(xu.name,Vo),new as(ju.name,Vo)],n.concat(t),Vo)}else e=gs;return e},t.prototype.visitNgContent=function(t,e){this.nodes.push(function(){return{sourceSpan:t.sourceSpan,nodeFlags:8,nodeDef:Vs(fo.ngContentDef).callFn([Ws(t.ngContentIndex),Ws(t.index)])}})},t.prototype.visitText=function(t,e){this.nodes.push(function(){return{sourceSpan:t.sourceSpan,nodeFlags:2,nodeDef:Vs(fo.textDef).callFn([Ws(-1),Ws(t.ngContentIndex),qs([Ws(t.value)])])}})},t.prototype.visitBoundText=function(t,e){var n=this,r=this.nodes.length;this.nodes.push(null);var i=t.value.ast,o=i.expressions.map(function(e,i){return n._preprocessUpdateExpression({nodeIndex:r,bindingIndex:i,sourceSpan:t.sourceSpan,context:wu,value:e})}),s=r;this.nodes[r]=function(){return{sourceSpan:t.sourceSpan,nodeFlags:2,nodeDef:Vs(fo.textDef).callFn([Ws(s),Ws(t.ngContentIndex),qs(i.strings.map(function(t){return Ws(t)}))]),updateRenderer:o}}},t.prototype.visitEmbeddedTemplate=function(t,e){var n=this,r=this.nodes.length;this.nodes.push(null);var i=this._visitElementOrTemplate(r,t),o=i.flags,s=i.queryMatchesExpr,a=i.hostEvents,c=this.viewBuilderFactory(this);this.children.push(c),c.visitAll(t.variables,t.children);var u=this.nodes.length-r-1;this.nodes[r]=function(){return{sourceSpan:t.sourceSpan,nodeFlags:1|o,nodeDef:Vs(fo.anchorDef).callFn([Ws(o),s,Ws(t.ngContentIndex),Ws(u),n._createElementHandleEventFn(r,a),Fs(c.viewName)])}}},t.prototype.visitElement=function(t,e){var n=this,r=this.nodes.length;this.nodes.push(null);var i=ie(t.name)?null:t.name,o=this._visitElementOrTemplate(r,t),s=o.flags,a=o.usedEvents,c=o.queryMatchesExpr,u=o.hostBindings,l=o.hostEvents,p=[],d=[],h=[];if(i){var f=t.inputs.map(function(t){return{context:wu,inputAst:t,dirAst:null}}).concat(u);f.length&&(d=f.map(function(t,e){return n._preprocessUpdateExpression({context:t.context,nodeIndex:r,bindingIndex:e,sourceSpan:t.inputAst.sourceSpan,value:t.inputAst.value})}),p=f.map(function(t){return function(t,e){switch(t.type){case pt.Attribute:return qs([Ws(1),Ws(t.name),Ws(t.securityContext)]);case pt.Property:return qs([Ws(8),Ws(t.name),Ws(t.securityContext)]);case pt.Animation:var n=8|(e&&e.directive.isComponent?32:16);return qs([Ws(n),Ws("@"+t.name),Ws(t.securityContext)]);case pt.Class:return qs([Ws(2),Ws(t.name),gs]);case pt.Style:return qs([Ws(4),Ws(t.name),Ws(t.unit)])}}(t.inputAst,t.dirAst)})),h=a.map(function(t){var e=t[0],n=t[1];return qs([Ws(e),Ws(n)])})}dt(this,t.children);var m=this.nodes.length-r-1,g=t.directives.find(function(t){return t.directive.isComponent}),v=gs,y=gs;g&&(y=this.outputCtx.importExpr(g.directive.componentViewType),v=this.outputCtx.importExpr(g.directive.rendererType));var b=r;this.nodes[r]=function(){return{sourceSpan:t.sourceSpan,nodeFlags:1|s,nodeDef:Vs(fo.elementDef).callFn([Ws(b),Ws(s),c,Ws(t.ngContentIndex),Ws(m),Ws(i),i?(e=t,o=Object.create(null),e.attrs.forEach(function(t){o[t.name]=t.value}),e.directives.forEach(function(t){Object.keys(t.directive.hostAttributes).forEach(function(e){var n,r,i=t.directive.hostAttributes[e],s=o[e];o[e]=null!=s?(r=i,(n=e)==gu||n==vu?s+" "+r:r):i})}),qs(Object.keys(o).sort().map(function(t){return qs([Ws(t),Ws(o[t])])}))):gs,p.length?qs(p):gs,h.length?qs(h):gs,n._createElementHandleEventFn(r,l),y,v]),updateRenderer:d};var e,o}},t.prototype._visitElementOrTemplate=function(t,e){var n=this,r=0;e.hasViewContainer&&(r|=16777216);var i=new Map;e.outputs.forEach(function(t){var e=Pu(t,null),n=e.name,r=e.target;i.set(Au(r,n),[r,n])}),e.directives.forEach(function(t){t.hostEvents.forEach(function(e){var n=Pu(e,t),r=n.name,o=n.target;i.set(Au(o,r),[o,r])})});var o=[],s=[];this._visitComponentFactoryResolverProvider(e.directives),e.providers.forEach(function(r,a){var c=void 0,u=void 0;if(e.directives.forEach(function(t,e){t.directive.type.reference===Et(r.token)&&(c=t,u=e)}),c){var l=n._visitDirective(r,c,u,t,e.references,e.queryMatches,i,n.staticQueryIds.get(e)),p=l.hostBindings,d=l.hostEvents;o.push.apply(o,p),s.push.apply(s,d)}else n._visitProvider(r,e.queryMatches)});var a=[];return e.queryMatches.forEach(function(t){var e=void 0;Et(t.value)===n.reflector.resolveExternalReference(fo.ElementRef)?e=0:Et(t.value)===n.reflector.resolveExternalReference(fo.ViewContainerRef)?e=3:Et(t.value)===n.reflector.resolveExternalReference(fo.TemplateRef)&&(e=2),null!=e&&a.push(qs([Ws(t.queryId),Ws(e)]))}),e.references.forEach(function(e){var r=void 0;e.value?Et(e.value)===n.reflector.resolveExternalReference(fo.TemplateRef)&&(r=2):r=1,null!=r&&(n.refNodeIndices[e.name]=t,a.push(qs([Ws(e.name),Ws(r)])))}),e.outputs.forEach(function(t){s.push({context:wu,eventAst:t,dirAst:null})}),{flags:r,usedEvents:Array.from(i.values()),queryMatchesExpr:a.length?qs(a):gs,hostBindings:o,hostEvents:s}},t.prototype._visitDirective=function(t,e,n,r,i,o,s,a){var c=this,u=this.nodes.length;this.nodes.push(null),e.directive.queries.forEach(function(t,n){var r=e.contentQueryStartId+n,i=67108864|Mu(a,r,t.first),o=t.first?0:1;c.nodes.push(function(){return{sourceSpan:e.sourceSpan,nodeFlags:i,nodeDef:Vs(fo.queryDef).callFn([Ws(i),Ws(r),new fs([new hs(t.propertyName,Ws(o),!1)])])}})});var l=this.nodes.length-u-1,p=this._visitProviderOrDirective(t,o),d=p.flags,h=p.queryMatchExprs,f=p.providerExpr,m=p.depsExpr;i.forEach(function(e){e.value&&Et(e.value)===Et(t.token)&&(c.refNodeIndices[e.name]=u,h.push(qs([Ws(e.name),Ws(4)])))}),e.directive.isComponent&&(d|=32768);var g=e.inputs.map(function(t,e){var n=qs([Ws(e),Ws(t.directiveName)]);return new hs(t.directiveName,n,!1)}),v=[],y=e.directive;Object.keys(y.outputs).forEach(function(t){var e=y.outputs[t];s.has(e)&&v.push(new hs(t,Ws(e),!1))});var b=[];(e.inputs.length||(327680&d)>0)&&(b=e.inputs.map(function(t,e){return c._preprocessUpdateExpression({nodeIndex:u,bindingIndex:e,sourceSpan:t.sourceSpan,context:wu,value:t.value})}));var _=Vs(fo.nodeValue).callFn([ju,Ws(u)]),j=e.hostProperties.map(function(t){return{context:_,dirAst:e,inputAst:t}}),x=e.hostEvents.map(function(t){return{context:_,eventAst:t,dirAst:e}}),w=u;return this.nodes[u]=function(){return{sourceSpan:e.sourceSpan,nodeFlags:16384|d,nodeDef:Vs(fo.directiveDef).callFn([Ws(w),Ws(d),h.length?qs(h):gs,Ws(l),f,m,g.length?new fs(g):gs,v.length?new fs(v):gs]),updateDirectives:b,directive:e.directive.type}},{hostBindings:j,hostEvents:x}},t.prototype._visitProvider=function(t,e){this._addProviderNode(this._visitProviderOrDirective(t,e))},t.prototype._visitComponentFactoryResolverProvider=function(t){var e=t.find(function(t){return t.directive.isComponent});if(e&&e.directive.entryComponents.length){var n=aa(this.reflector,this.outputCtx,8192,e.directive.entryComponents),r=n.providerExpr,i=n.depsExpr,o=n.flags,s=n.tokenExpr;this._addProviderNode({providerExpr:r,depsExpr:i,flags:o,tokenExpr:s,queryMatchExprs:[],sourceSpan:e.sourceSpan})}},t.prototype._addProviderNode=function(t){this.nodes.length;this.nodes.push(function(){return{sourceSpan:t.sourceSpan,nodeFlags:t.flags,nodeDef:Vs(fo.providerDef).callFn([Ws(t.flags),t.queryMatchExprs.length?qs(t.queryMatchExprs):gs,t.tokenExpr,t.providerExpr,t.depsExpr])}})},t.prototype._visitProviderOrDirective=function(t,e){var n=[];e.forEach(function(e){Et(e.value)===Et(t.token)&&n.push(qs([Ws(e.queryId),Ws(4)]))});var r=na(this.outputCtx,t),i=r.providerExpr,o=r.depsExpr,s=r.flags,a=r.tokenExpr;return{flags:0|s,queryMatchExprs:n,providerExpr:i,depsExpr:o,tokenExpr:a,sourceSpan:t.sourceSpan}},t.prototype.getLocal=function(t){if(t==Jc.event.name)return Jc.event;for(var e=ju,n=this;n;n=n.parent,e=e.prop("parent").cast(Fo)){var r=n.refNodeIndices[t];if(null!=r)return Vs(fo.nodeValue).callFn([e,Ws(r)]);var i=n.variables.find(function(e){return e.name===t});if(i){var o=i.value||"$implicit";return e.prop("context").prop(o)}}return null},t.prototype._createLiteralArrayConverter=function(t,e){if(0===e){var n=Vs(fo.EMPTY_ARRAY);return function(){return n}}var r=this.nodes.length;return this.nodes.push(function(){return{sourceSpan:t,nodeFlags:32,nodeDef:Vs(fo.pureArrayDef).callFn([Ws(r),Ws(e)])}}),function(t){return Eu(r,t)}},t.prototype._createLiteralMapConverter=function(t,e){if(0===e.length){var n=Vs(fo.EMPTY_MAP);return function(){return n}}var i=Hs(e.map(function(t,e){return Object(r.a)({},t,{value:Ws(e)})})),o=this.nodes.length;return this.nodes.push(function(){return{sourceSpan:t,nodeFlags:64,nodeDef:Vs(fo.pureObjectDef).callFn([Ws(o),i])}}),function(t){return Eu(o,t)}},t.prototype._createPipeConverter=function(t,e,n){var r=this.usedPipes.find(function(t){return t.name===e});if(r.pure){var i=this.nodes.length;this.nodes.push(function(){return{sourceSpan:t.sourceSpan,nodeFlags:128,nodeDef:Vs(fo.purePipeDef).callFn([Ws(i),Ws(n)])}});for(var o=ju,s=this;s.parent;)s=s.parent,o=o.prop("parent").cast(Fo);var a=s.purePipeNodeIndices[e],c=Vs(fo.nodeValue).callFn([o,Ws(a)]);return function(e){return Su(t.nodeIndex,t.bindingIndex,Eu(i,[c].concat(e)))}}var u=this._createPipe(t.sourceSpan,r),l=Vs(fo.nodeValue).callFn([ju,Ws(u)]);return function(e){return Su(t.nodeIndex,t.bindingIndex,l.callMethod("transform",e))}},t.prototype._createPipe=function(t,e){var n=this,r=this.nodes.length,i=0;e.type.lifecycleHooks.forEach(function(t){t===vo.OnDestroy&&(i|=sa(t))});var o=e.type.diDeps.map(function(t){return oa(n.outputCtx,t)});return this.nodes.push(function(){return{sourceSpan:t,nodeFlags:16,nodeDef:Vs(fo.pipeDef).callFn([Ws(i),n.outputCtx.importExpr(e.type.reference),qs(o)])}}),r},t.prototype._preprocessUpdateExpression=function(t){var e=this;return{nodeIndex:t.nodeIndex,bindingIndex:t.bindingIndex,sourceSpan:t.sourceSpan,context:t.context,value:nu({createLiteralArrayConverter:function(n){return e._createLiteralArrayConverter(t.sourceSpan,n)},createLiteralMapConverter:function(n){return e._createLiteralMapConverter(t.sourceSpan,n)},createPipeConverter:function(n,r){return e._createPipeConverter(t,n,r)}},t.value)}},t.prototype._createNodeExpressions=function(){var t=this,e=0,n=[],r=[],i=this.nodes.map(function(t,e){var i=t(),s=i.nodeDef,a=i.nodeFlags,c=i.updateDirectives,u=i.updateRenderer,l=i.sourceSpan;return u&&n.push.apply(n,o(e,l,u,!1)),c&&r.push.apply(r,o(e,l,c,(327680&a)>0)),Ns(3&a?new ms([_u.callFn([]).callFn([]),s]):s,l)});return{updateRendererStmts:n,updateDirectivesStmts:r,nodeDefExprs:i};function o(n,r,i,o){var s=[],a=i.map(function(n){var r=n.sourceSpan,i=n.context,o=n.value,a=""+e++,c=iu(i===wu?t:null,i,o,a),u=c.stmts,l=c.currValExpr;return s.push.apply(s,u.map(function(t){return Rs(t,r)})),Ns(l,r)});return(i.length||o)&&s.push(Rs(Eu(n,a).toStmt(),r)),s}},t.prototype._createElementHandleEventFn=function(t,e){var n,r=this,i=[],o=0;if(e.forEach(function(t){var e=t.context,n=t.eventAst,s=t.dirAst,a=""+o++,c=eu(e===wu?r:null,e,n.handler,a),u=c.stmts,l=c.allowDefault,p=u;l&&p.push(Cu.set(l.and(Cu)).toStmt());var d=Pu(n,s),h=Au(d.target,d.name);i.push(Rs(new Ss(Ws(h).identical(Ou),p),n.sourceSpan))}),i.length>0){var s=[Cu.set(Ws(!0)).toDeclStmt(zo)];!this.component.isHost&&Ds(i).has(wu.name)&&s.push(wu.set(ju.prop("component")).toDeclStmt(this.compType)),n=Us([new as(ju.name,Vo),new as(Ou.name,Vo),new as(Jc.event.name,Vo)],s.concat(i,[new ws(Cu)]),Vo)}else n=gs;return n},t.prototype.visitDirective=function(t,e){},t.prototype.visitDirectiveProperty=function(t,e){},t.prototype.visitReference=function(t,e){},t.prototype.visitVariable=function(t,e){},t.prototype.visitEvent=function(t,e){},t.prototype.visitElementProperty=function(t,e){},t.prototype.visitAttr=function(t,e){},t}();function Eu(t,e){return e.length>10?xu.callFn([ju,Ws(t),Ws(1),qs(e)]):xu.callFn([ju,Ws(t),Ws(0)].concat(e))}function Su(t,e,n){return Vs(fo.unwrapValue).callFn([ju,Ws(t),Ws(e),n])}function Pu(t,e){return t.isAnimation?{name:"@"+t.name+"."+t.phase,target:e&&e.directive.isComponent?"component":null}:t}function Mu(t,e,n){var r=0;return!n||!t.staticQueryIds.has(e)&&t.dynamicQueryIds.has(e)?r|=536870912:r|=268435456,r}function Au(t,e){return t?t+":"+e:e}var Tu=function(){function t(t,e,n,r){void 0===r&&(r=null),this._htmlParser=t,this._implicitTags=e,this._implicitAttrs=n,this._locale=r,this._messages=[]}return t.prototype.updateFromTemplate=function(t,e,n){var r=this._htmlParser.parse(t,e,!0,n);if(r.errors.length)return r.errors;var i,o,s,a,c,u=(i=r.rootNodes,o=n,s=this._implicitTags,a=this._implicitAttrs,new bi(s,a).extract(i,o));return u.errors.length?u.errors:((c=this._messages).push.apply(c,u.messages),[])},t.prototype.getMessages=function(){return this._messages},t.prototype.write=function(t,e){var n={},r=new Du;this._messages.forEach(function(e){var r,i=t.digest(e);n.hasOwnProperty(i)?(r=n[i].sources).push.apply(r,e.sources):n[i]=e});var i=Object.keys(n).map(function(i){var o=t.createNameMapper(n[i]),s=n[i],a=o?r.convert(s.nodes,o):s.nodes,c=new Xr(a,{},{},s.meaning,s.description,i);return c.sources=s.sources,e&&c.sources.forEach(function(t){return t.filePath=e(t.filePath)}),c});return t.write(i,this._locale)},t}(),Du=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.convert=function(t,e){var n=this;return e?t.map(function(t){return t.visit(n,e)}):t},e.prototype.visitTagPlaceholder=function(t,e){var n=this,r=e.toPublicName(t.startName),i=t.closeName?e.toPublicName(t.closeName):t.closeName,o=t.children.map(function(t){return t.visit(n,e)});return new Qr(t.tag,t.attrs,r,i,o,t.isVoid,t.sourceSpan)},e.prototype.visitPlaceholder=function(t,e){return new Zr(t.value,e.toPublicName(t.name),t.sourceSpan)},e.prototype.visitIcuPlaceholder=function(t,e){return new Jr(t.value,e.toPublicName(t.name),t.sourceSpan)},e}(ti),Iu=function(){function t(t,e,n){this.srcFileUrl=t,this.genFileUrl=e,"string"==typeof n?(this.source=n,this.stmts=null):(this.source=null,this.stmts=n)}return t.prototype.isEquivalent=function(t){return this.genFileUrl===t.genFileUrl&&(this.source?this.source===t.source:null!=t.stmts&&Ho(this.stmts,t.stmts))},t}();function Ru(t,e){for(var n=[],r=0,i=t.transitiveModule.providers;r<i.length;r++){var o=i[r],s=o.provider,a=o.module;if(Et(s.token)===e.ROUTES)for(var c=0,u=Nu(s.useValue);c<u.length;c++){var l=u[c];n.push(Lu(l,e,a.reference))}}return n}function Nu(t,e){if(void 0===e&&(e=[]),"string"==typeof t)e.push(t);else if(Array.isArray(t))for(var n=0,r=t;n<r.length;n++){Nu(r[n],e)}else t.loadChildren?Nu(t.loadChildren,e):t.children&&Nu(t.children,e);return e}function Lu(t,e,n){var r=t.split("#"),i=r[0],o=r[1],s=e.resolveExternalReference({moduleName:i,name:o},n?n.filePath:void 0);return{route:t,module:n||s,referencedModule:s}}var Fu=function(){return function(t,e){this.symbol=t,this.metadata=e}}(),Vu=function(){function t(t,e,n,r){this.host=t,this.staticSymbolCache=e,this.summaryResolver=n,this.errorRecorder=r,this.metadataCache=new Map,this.resolvedSymbols=new Map,this.resolvedFilePaths=new Set,this.importAs=new Map,this.symbolResourcePaths=new Map,this.symbolFromFile=new Map,this.knownFileNameToModuleNames=new Map}return t.prototype.resolveSymbol=function(t){if(t.members.length>0)return this._resolveSymbolMembers(t);var e=this._resolveSymbolFromSummary(t);if(e)return e;var n=this.resolvedSymbols.get(t);return n||(this._createSymbolsOf(t.filePath),this.resolvedSymbols.get(t))},t.prototype.getImportAs=function(t){if(t.members.length){var e=this.getStaticSymbol(t.filePath,t.name);return(r=this.getImportAs(e))?this.getStaticSymbol(r.filePath,r.name,t.members):null}var n=t.filePath.replace(ro,".");if(n!==t.filePath){var r,i=t.name.replace(io,"");e=this.getStaticSymbol(n,i,t.members);return(r=this.getImportAs(e))?this.getStaticSymbol(uo(r.filePath),lo(r.name),e.members):null}var o=this.summaryResolver.getImportAs(t);return o||(o=this.importAs.get(t)),o},t.prototype.getResourcePath=function(t){return this.symbolResourcePaths.get(t)||t.filePath},t.prototype.getTypeArity=function(t){if(e=t.filePath,no.test(e))return null;for(var e,n=Bu(this.resolveSymbol(t));n&&n.metadata instanceof mt;)n=Bu(this.resolveSymbol(n.metadata));return n&&n.metadata&&n.metadata.arity||null},t.prototype.getKnownModuleName=function(t){return this.knownFileNameToModuleNames.get(t)||null},t.prototype.recordImportAs=function(t,e){t.assertNoMembers(),e.assertNoMembers(),this.importAs.set(t,e)},t.prototype.recordModuleNameForFileName=function(t,e){this.knownFileNameToModuleNames.set(t,e)},t.prototype.invalidateFile=function(t){this.metadataCache.delete(t),this.resolvedFilePaths.delete(t);var e=this.symbolFromFile.get(t);if(e){this.symbolFromFile.delete(t);for(var n=0,r=e;n<r.length;n++){var i=r[n];this.resolvedSymbols.delete(i),this.importAs.delete(i),this.symbolResourcePaths.delete(i)}}},t.prototype.ignoreErrorsFor=function(t){var e=this.errorRecorder;this.errorRecorder=function(){};try{return t()}finally{this.errorRecorder=e}},t.prototype._resolveSymbolMembers=function(t){var e=t.members,n=this.resolveSymbol(this.getStaticSymbol(t.filePath,t.name));if(!n)return null;var r=Bu(n.metadata);if(r instanceof mt)return new Fu(t,this.getStaticSymbol(r.filePath,r.name,e));if(!r||"class"!==r.__symbolic){for(var i=r,o=0;o<e.length&&i;o++)i=i[e[o]];return new Fu(t,i)}return r.statics&&1===e.length?new Fu(t,r.statics[e[0]]):null},t.prototype._resolveSymbolFromSummary=function(t){var e=this.summaryResolver.resolveSummary(t);return e?new Fu(t,e.metadata):null},t.prototype.getStaticSymbol=function(t,e,n){return this.staticSymbolCache.get(t,e,n)},t.prototype.hasDecorators=function(t){var e=this.getModuleMetadata(t);return!!e.metadata&&Object.keys(e.metadata).some(function(t){var n=e.metadata[t];return n&&"class"===n.__symbolic&&n.decorators})},t.prototype.getSymbolsOf=function(t){var e=this.summaryResolver.getSymbolsOf(t);if(e)return e;this._createSymbolsOf(t);var n=[];return this.resolvedSymbols.forEach(function(e){e.symbol.filePath===t&&n.push(e.symbol)}),n},t.prototype._createSymbolsOf=function(t){var e=this;if(!this.resolvedFilePaths.has(t)){this.resolvedFilePaths.add(t);var n=[],r=this.getModuleMetadata(t);if(r.importAs&&this.knownFileNameToModuleNames.set(t,r.importAs),r.exports)for(var i=function(r){if(r.export)r.export.forEach(function(i){var o,s=o=zu(o="string"==typeof i?i:i.as);"string"!=typeof i&&(s=zu(i.name));var a=e.resolveModule(r.from,t);if(a){var c=e.getStaticSymbol(a,s),u=e.getStaticSymbol(t,o);n.push(e.createExport(u,c))}});else{var i=o.resolveModule(r.from,t);if(i)o.getSymbolsOf(i).forEach(function(r){var i=e.getStaticSymbol(t,r.name);n.push(e.createExport(i,r))})}},o=this,s=0,a=r.exports;s<a.length;s++){i(a[s])}if(r.metadata){var c=new Set(Object.keys(r.metadata).map(zu)),u=r.origins||{};Object.keys(r.metadata).forEach(function(i){var o=r.metadata[i],s=zu(i),a=e.getStaticSymbol(t,s),l=u.hasOwnProperty(i)&&u[i];if(l){var p=e.resolveModule(l,t);p?e.symbolResourcePaths.set(a,p):e.reportError(new Error("Couldn't resolve original symbol for "+l+" from "+t))}n.push(e.createResolvedSymbol(a,t,c,o))})}n.forEach(function(t){return e.resolvedSymbols.set(t.symbol,t)}),this.symbolFromFile.set(t,n.map(function(t){return t.symbol}))}},t.prototype.createResolvedSymbol=function(t,e,n,i){if(this.summaryResolver.isLibraryFile(t.filePath)&&i&&"class"===i.__symbolic){var o={__symbolic:"class",arity:i.arity};return new Fu(t,o)}var s,a=function(){return s||(s=e.replace(/((\.ts)|(\.d\.ts)|)$/,".ts").replace(/^.*node_modules[/\\]/,"")),s},c=this,u=R(i,new(function(i){function o(){return null!==i&&i.apply(this,arguments)||this}return Object(r.b)(o,i),o.prototype.visitStringMap=function(o,s){var u=o.__symbolic;if("function"===u){var l=s.length;s.push.apply(s,o.parameters||[]);var p=i.prototype.visitStringMap.call(this,o,s);return s.length=l,p}if("reference"!==u)return"error"===u?Object(r.a)({},o,{fileName:a()}):i.prototype.visitStringMap.call(this,o,s);var d=o.module,h=o.name?zu(o.name):o.name;if(!h)return null;var f=void 0;return d?(f=c.resolveModule(d,t.filePath))?{__symbolic:"resolved",symbol:c.getStaticSymbol(f,h),line:o.line,character:o.character,fileName:a()}:{__symbolic:"error",message:"Could not resolve "+d+" relative to "+t.filePath+".",line:o.line,character:o.character,fileName:a()}:s.indexOf(h)>=0?{__symbolic:"reference",name:h}:n.has(h)?c.getStaticSymbol(e,h):void 0},o}(F)),[]),l=Bu(u);return l instanceof mt?this.createExport(t,l):new Fu(t,u)},t.prototype.createExport=function(t,e){return t.assertNoMembers(),e.assertNoMembers(),this.summaryResolver.isLibraryFile(t.filePath)&&this.summaryResolver.isLibraryFile(e.filePath)&&this.importAs.set(e,this.getImportAs(t)||t),new Fu(t,e)},t.prototype.reportError=function(t,e,n){if(!this.errorRecorder)throw t;this.errorRecorder(t,e&&e.filePath||n)},t.prototype.getModuleMetadata=function(t){var e=this.metadataCache.get(t);if(!e){var n=this.host.getMetadataFor(t);if(n){var r=-1;n.forEach(function(t){t.version>r&&(r=t.version,e=t)})}if(e||(e={__symbolic:"module",version:4,module:t,metadata:{}}),4!=e.version){var i=2==e.version?"Unsupported metadata version "+e.version+" for module "+t+". This module should be compiled with a newer version of ngc":"Metadata version mismatch for module "+t+", found version "+e.version+", expected 4";this.reportError(new Error(i))}this.metadataCache.set(t,e)}return e},t.prototype.getSymbolByModule=function(t,e,n){var r=this.resolveModule(t,n);return r?this.getStaticSymbol(r,e):(this.reportError(new Error("Could not resolve module "+t+(n?" relative to "+n:""))),this.getStaticSymbol("ERROR:"+t,e))},t.prototype.resolveModule=function(t,e){try{return this.host.moduleNameToFileName(t,e)}catch(n){console.error("Could not resolve module '"+t+"' relative to file "+e),this.reportError(n,void 0,e)}return null},t}();function zu(t){return t.startsWith("___")?t.substr(1):t}function Bu(t){return t&&"resolved"===t.__symbolic?t.symbol:t}function qu(t,e,n){var r=lo(e.name);t.statements.push(Us([],[new ws(n)],new No(Fo)).toDeclStmt(r,[ys.Final,ys.Exported]))}var Hu=function(t){function e(e,n,r){var i=t.call(this)||this;return i.symbolResolver=e,i.summaryResolver=n,i.srcFileName=r,i.symbols=[],i.indexBySymbol=new Map,i.reexportedBy=new Map,i.processedSummaryBySymbol=new Map,i.processedSummaries=[],i.unprocessedSymbolSummariesBySymbol=new Map,i.moduleName=e.getKnownModuleName(r),i}return Object(r.b)(e,t),e.prototype.addSummary=function(t){var e,n,r,i=this,o=this.unprocessedSymbolSummariesBySymbol.get(t.symbol),s=this.processedSummaryBySymbol.get(t.symbol);if(o||(o={symbol:t.symbol,metadata:void 0},this.unprocessedSymbolSummariesBySymbol.set(t.symbol,o),s={symbol:this.processValue(t.symbol,0)},this.processedSummaries.push(s),this.processedSummaryBySymbol.set(t.symbol,s)),!o.metadata&&t.metadata){var a=t.metadata||{};if("class"===a.__symbolic){var c={};Object.keys(a).forEach(function(t){"decorators"!==t&&(c[t]=a[t])}),a=c}else Gu(a)&&(Gu(r=a)&&Bu(r.expression)instanceof mt||Gu(n=a)&&n.expression&&"select"===n.expression.__symbolic&&Bu(n.expression.expression)instanceof mt||(a={__symbolic:"error",message:"Complex function calls are not supported."}));if(o.metadata=a,s.metadata=this.processValue(a,1),a instanceof mt&&this.summaryResolver.isLibraryFile(a.filePath)){var u=this.symbols[this.indexBySymbol.get(a)];e=u.name,po.test(e)||this.reexportedBy.set(u,t.symbol)}}if(!o.type&&t.type&&(o.type=t.type,s.type=this.processValue(t.type,0),t.type.summaryKind===Ct.NgModule)){var l=t.type;l.exportedDirectives.concat(l.exportedPipes).forEach(function(t){var e=t.reference;if(i.summaryResolver.isLibraryFile(e.filePath)&&!i.unprocessedSymbolSummariesBySymbol.has(e)){var n=i.summaryResolver.resolveSummary(e);n&&i.addSummary(n)}})}},e.prototype.serialize=function(){var t=this,e=[];return{json:JSON.stringify({moduleName:this.moduleName,summaries:this.processedSummaries,symbols:this.symbols.map(function(n,r){n.assertNoMembers();var i=void 0;if(t.summaryResolver.isLibraryFile(n.filePath)){var o=t.reexportedBy.get(n);if(o)i=t.indexBySymbol.get(o);else{var s=t.unprocessedSymbolSummariesBySymbol.get(n);s&&s.metadata&&"interface"===s.metadata.__symbolic||(i=n.name+"_"+r,e.push({symbol:n,exportAs:i}))}}return{__symbol:r,name:n.name,filePath:t.summaryResolver.toSummaryFileName(n.filePath,t.srcFileName),importAs:i}})}),exportAs:e}},e.prototype.processValue=function(t,e){return R(t,this,e)},e.prototype.visitOther=function(t,e){if(t instanceof mt){var n=this.symbolResolver.getStaticSymbol(t.filePath,t.name);return{__symbol:this.visitStaticSymbol(n,e),members:t.members}}},e.prototype.visitStaticSymbol=function(t,e){var n=this.indexBySymbol.get(t),r=null;if(1&e&&this.summaryResolver.isLibraryFile(t.filePath)){if(this.unprocessedSymbolSummariesBySymbol.has(t))return n;(r=this.loadSummary(t))&&r.metadata instanceof mt&&(n=this.visitStaticSymbol(r.metadata,e),r=null)}else if(null!=n)return n;return null==n&&(n=this.symbols.length,this.symbols.push(t)),this.indexBySymbol.set(t,n),r&&this.addSummary(r),n},e.prototype.loadSummary=function(t){var e=this.summaryResolver.resolveSummary(t);if(!e){var n=this.symbolResolver.resolveSymbol(t);n&&(e={symbol:n.symbol,metadata:n.metadata})}return e},e}(F),Uu=function(){function t(t,e,n){this.outputCtx=t,this.symbolResolver=e,this.summaryResolver=n,this.data=[]}return t.prototype.addSourceType=function(t,e){this.data.push({summary:t,metadata:e,isLibrary:!1})},t.prototype.addLibType=function(t){this.data.push({summary:t,metadata:null,isLibrary:!0})},t.prototype.serialize=function(t){for(var e=this,n=new Map,r=0,i=t;r<i.length;r++){var o=i[r],s=o.symbol,a=o.exportAs;n.set(s,a)}for(var c=new Set,u=0,l=this.data;u<l.length;u++){var p=l[u],d=p.summary,h=p.metadata,f=p.isLibrary;if(d.summaryKind===Ct.NgModule){c.add(d.type.reference);for(var m=0,g=d.modules;m<g.length;m++){var v=g[m];c.add(v.reference)}}if(!f){lo(d.type.reference.name);qu(this.outputCtx,d.type.reference,this.serializeSummaryWithDeps(d,h))}}c.forEach(function(t){if(e.summaryResolver.isLibraryFile(t.filePath)){var r=lo(n.get(t)||t.name);e.outputCtx.statements.push(Fs(r).set(e.serializeSummaryRef(t)).toDeclStmt(null,[ys.Exported]))}})},t.prototype.serializeSummaryWithDeps=function(t,e){var n=this,r=[this.serializeSummary(t)],i=[];if(e instanceof Tt)r.push.apply(r,e.declaredDirectives.concat(e.declaredPipes).map(function(t){return t.reference}).concat(e.transitiveModule.modules.map(function(t){return t.reference}).filter(function(t){return t!==e.type.reference})).map(function(t){return n.serializeSummaryRef(t)})),i=e.providers;else if(t.summaryKind===Ct.Directive){var o=t;i=o.providers.concat(o.viewProviders)}return r.push.apply(r,i.filter(function(t){return!!t.useClass}).map(function(t){return n.serializeSummary({summaryKind:Ct.Injectable,type:t.useClass})})),qs(r)},t.prototype.serializeSummaryRef=function(t){var e=this.symbolResolver.getStaticSymbol(uo(t.filePath),lo(t.name));return this.outputCtx.importExpr(e)},t.prototype.serializeSummary=function(t){var e=this.outputCtx;return R(t,new(function(){function t(){}return t.prototype.visitArray=function(t,e){var n=this;return qs(t.map(function(t){return R(t,n,e)}))},t.prototype.visitStringMap=function(t,e){var n=this;return new fs(Object.keys(t).map(function(r){return new hs(r,R(t[r],n,e),!1)}))},t.prototype.visitPrimitive=function(t,e){return Ws(t)},t.prototype.visitOther=function(t,n){if(t instanceof mt)return e.importExpr(t);throw new Error("Illegal State: Encountered value "+t)},t}()),null)},t}(),Wu=function(t){function e(e,n){var r=t.call(this)||this;return r.symbolCache=e,r.summaryResolver=n,r}return Object(r.b)(e,t),e.prototype.deserialize=function(t,e){var n=this,r=JSON.parse(e),i=[];this.symbols=r.symbols.map(function(e){return n.symbolCache.get(n.summaryResolver.fromSummaryFileName(e.filePath,t),e.name)}),r.symbols.forEach(function(e,r){var o=n.symbols[r],s=e.importAs;"number"==typeof s?i.push({symbol:o,importAs:n.symbols[s]}):"string"==typeof s&&i.push({symbol:o,importAs:n.symbolCache.get(oo(t),s)})});var o=R(r.summaries,this,null);return{moduleName:r.moduleName,summaries:o,importAs:i}},e.prototype.visitStringMap=function(e,n){if("__symbol"in e){var r=this.symbols[e.__symbol],i=e.members;return i.length?this.symbolCache.get(r.filePath,r.name,i):r}return t.prototype.visitStringMap.call(this,e,n)},e}(F);function Gu(t){return t&&"call"===t.__symbolic}var Xu={Basic:1,TypeCheck:2,All:3};Xu[Xu.Basic]="Basic",Xu[Xu.TypeCheck]="TypeCheck",Xu[Xu.All]="All";!function(){function t(t,e,n,r,i,o,s,a,c,u,l,p,d){this._config=t,this._options=e,this._host=n,this._reflector=r,this._metadataResolver=i,this._templateParser=o,this._styleCompiler=s,this._viewCompiler=a,this._typeCheckCompiler=c,this._ngModuleCompiler=u,this._outputEmitter=l,this._summaryResolver=p,this._symbolResolver=d,this._templateAstCache=new Map,this._analyzedFiles=new Map}t.prototype.clearCache=function(){this._metadataResolver.clearCache()},t.prototype.analyzeModulesSync=function(t){var e=this,n=Zu(t,this._host,this._symbolResolver,this._metadataResolver);return n.ngModules.forEach(function(t){return e._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!0)}),n},t.prototype.analyzeModulesAsync=function(t){var e=this,n=Zu(t,this._host,this._symbolResolver,this._metadataResolver);return Promise.all(n.ngModules.map(function(t){return e._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!1)})).then(function(){return n})},t.prototype._analyzeFile=function(t){var e=this._analyzedFiles.get(t);return e||(e=tl(this._host,this._symbolResolver,this._metadataResolver,t),this._analyzedFiles.set(t,e)),e},t.prototype.findGeneratedFileNames=function(t){var e=this,n=[],r=this._analyzeFile(t);(this._options.allowEmptyCodegenFiles||r.directives.length||r.pipes.length||r.injectables.length||r.ngModules.length||r.exportsNonSourceFiles)&&(n.push(oo(r.fileName,!0)),this._options.enableSummariesForJit&&n.push(uo(r.fileName,!0)));var i=ao(r.fileName,!0)[1];return r.directives.forEach(function(t){var o=e._metadataResolver.getNonNormalizedDirectiveMetadata(t).metadata;o.isComponent&&o.template.styleUrls.forEach(function(t){var s=e._host.resourceNameToFileName(t,r.fileName);if(!s)throw q("Couldn't resolve resource "+t+" relative to "+r.fileName);var a=(o.template.encapsulation||e._config.defaultEncapsulation)===d.Emulated;n.push($u(s,a,i)),e._options.allowEmptyCodegenFiles&&n.push($u(s,!a,i))})}),n},t.prototype.emitBasicStub=function(t,e){var n=this._createOutputContext(t);if(t.endsWith(".ngfactory.ts")){if(!e)throw new Error("Assertion error: require the original file for .ngfactory.ts stubs. File: "+t);var r=this._analyzeFile(e);this._createNgFactoryStub(n,r,Xu.Basic)}else if(t.endsWith(".ngsummary.ts")){if(this._options.enableSummariesForJit){if(!e)throw new Error("Assertion error: require the original file for .ngsummary.ts stubs. File: "+t);r=this._analyzeFile(e);Yu(n),r.ngModules.forEach(function(t){var e,r;e=n,r=t.type.reference,qu(e,r,gs)})}}else t.endsWith(".ngstyle.ts")&&Yu(n);return this._codegenSourceModule("unknown",n)},t.prototype.emitTypeCheckStub=function(t,e){var n=this._analyzeFile(e),r=this._createOutputContext(t);return t.endsWith(".ngfactory.ts")&&this._createNgFactoryStub(r,n,Xu.TypeCheck),r.statements.length>0?this._codegenSourceModule(n.fileName,r):null},t.prototype.loadFilesAsync=function(t){var e=this,n=t.map(function(t){return e._analyzeFile(t)}),r=[];return n.forEach(function(t){return t.ngModules.forEach(function(t){return r.push(e._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!1))})}),Promise.all(r).then(function(t){return nl(n)})},t.prototype.loadFilesSync=function(t){var e=this,n=t.map(function(t){return e._analyzeFile(t)});return n.forEach(function(t){return t.ngModules.forEach(function(t){return e._metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!0)})}),nl(n)},t.prototype._createNgFactoryStub=function(t,e,n){var r=this,i=0;e.ngModules.forEach(function(e,o){r._ngModuleCompiler.createStub(t,e.type.reference);var s=e.transitiveModule.directives.map(function(t){return t.reference}).concat(e.transitiveModule.pipes.map(function(t){return t.reference}),e.importedModules.map(function(t){return t.type.reference}),e.exportedModules.map(function(t){return t.type.reference}),r._externalIdentifierReferences([fo.TemplateRef,fo.ElementRef])),a=new Map;s.forEach(function(t,e){r._host.isSourceFile(t.filePath)&&a.set(t,"_decl"+o+"_"+e)}),a.forEach(function(e,n){t.statements.push(Fs(e).set(gs.cast(Fo)).toDeclStmt(Bs(t.importExpr(n))))}),n&Xu.TypeCheck&&e.declaredDirectives.forEach(function(n){var o=r._metadataResolver.getDirectiveMetadata(n.reference);o.isComponent&&(i++,r._createTypeCheckBlock(t,o.type.reference.name+"_Host_"+i,e,r._metadataResolver.getHostComponentMetadata(o),[o.type],a),r._createTypeCheckBlock(t,o.type.reference.name+"_"+i,e,o,e.transitiveModule.directives,a))})}),0===t.statements.length&&Yu(t)},t.prototype._externalIdentifierReferences=function(t){for(var e=[],n=0,r=t;n<r.length;n++){var i=r[n],o=go(this._reflector,i);o.identifier&&e.push(o.identifier.reference)}return e},t.prototype._createTypeCheckBlock=function(t,e,n,r,i,o){var s,a=this._parseTemplate(r,n,i),c=a.template,u=a.pipes;(s=t.statements).push.apply(s,this._typeCheckCompiler.compileComponent(e,r,c,u,o))},t.prototype.emitMessageBundle=function(t,e){var n=this,r=[],i=new Qi,o=new Tu(i,[],{},e);if(t.files.forEach(function(t){var e=[];t.directives.forEach(function(t){var r=n._metadataResolver.getDirectiveMetadata(t);r&&r.isComponent&&e.push(r)}),e.forEach(function(e){var n=e.template.template,i=Kt.fromArray(e.template.interpolation);r.push.apply(r,o.updateFromTemplate(n,t.fileName,i))})}),r.length)throw new Error(r.map(function(t){return t.toString()}).join("\n"));return o},t.prototype.emitAllImpls=function(t){var e=this,n=t.ngModuleByPipeOrDirective;return Nt(t.files.map(function(t){return e._compileImplFile(t.fileName,n,t.directives,t.pipes,t.ngModules,t.injectables)}))},t.prototype._compileImplFile=function(t,e,n,r,i,o){var s=this,a=ao(t,!0)[1],c=[],u=this._createOutputContext(oo(t,!0));if(c.push.apply(c,this._createSummary(t,n,r,i,o,u)),i.forEach(function(t){return s._compileModule(u,t)}),n.forEach(function(n){var r=s._metadataResolver.getDirectiveMetadata(n);if(r.isComponent){var i=e.get(n);if(!i)throw new Error("Internal Error: cannot determine the module for component "+_t(r.type)+"!");var o=s._styleCompiler.compileComponent(u,r);r.template.externalStylesheets.forEach(function(e){var n=s._styleCompiler.needsStyleShim(r);c.push(s._codegenStyles(t,r,e,n,a)),s._options.allowEmptyCodegenFiles&&c.push(s._codegenStyles(t,r,e,!n,a))});s._compileComponent(u,r,i,i.transitiveModule.directives,o,a);s._compileComponentFactory(u,r,i,a)}}),u.statements.length>0||this._options.allowEmptyCodegenFiles){var l=this._codegenSourceModule(t,u);c.unshift(l)}return c},t.prototype._createSummary=function(t,e,n,r,i,o){var s=this,a=this._symbolResolver.getSymbolsOf(t).map(function(t){return s._symbolResolver.resolveSymbol(t)}),c=r.map(function(t){return{summary:s._metadataResolver.getNgModuleSummary(t.type.reference),metadata:s._metadataResolver.getNgModuleMetadata(t.type.reference)}}).concat(e.map(function(t){return{summary:s._metadataResolver.getDirectiveSummary(t),metadata:s._metadataResolver.getDirectiveMetadata(t)}}),n.map(function(t){return{summary:s._metadataResolver.getPipeSummary(t),metadata:s._metadataResolver.getPipeMetadata(t)}}),i.map(function(t){return{summary:s._metadataResolver.getInjectableSummary(t),metadata:s._metadataResolver.getInjectableSummary(t).type}})),u=this._options.enableSummariesForJit?this._createOutputContext(uo(t,!0)):null,l=function(t,e,n,r,i,o){var s=new Hu(r,n,t);i.forEach(function(t){return s.addSummary({symbol:t.symbol,metadata:t.metadata})}),o.forEach(function(t){var e=t.summary;t.metadata,s.addSummary({symbol:e.type.reference,metadata:void 0,type:e})});var a=s.serialize(),c=a.json,u=a.exportAs;if(e){var l=new Uu(e,r,n);o.forEach(function(t){var e=t.summary,n=t.metadata;l.addSourceType(e,n)}),s.unprocessedSymbolSummariesBySymbol.forEach(function(t){n.isLibraryFile(t.symbol.filePath)&&t.type&&l.addLibType(t.type)}),l.serialize(u)}return{json:c,exportAs:u}}(t,u,this._summaryResolver,this._symbolResolver,a,c),p=l.json;l.exportAs.forEach(function(t){o.statements.push(Fs(t.exportAs).set(o.importExpr(t.symbol)).toDeclStmt(null,[ys.Exported]))});var d=[new Iu(t,co(t),p)];return u&&d.push(this._codegenSourceModule(t,u)),d},t.prototype._compileModule=function(t,e){var n=[];if(this._options.locale){var r=this._options.locale.replace(/_/g,"-");n.push({token:go(this._reflector,fo.LOCALE_ID),useValue:r})}this._options.i18nFormat&&n.push({token:go(this._reflector,fo.TRANSLATIONS_FORMAT),useValue:this._options.i18nFormat}),this._ngModuleCompiler.compile(t,e,n)},t.prototype._compileComponentFactory=function(t,e,n,r){var i=this._metadataResolver.getHostComponentMetadata(e),o=this._compileComponent(t,i,n,[e.type],null,r).viewClassVar,s=Ot(e.type.reference),a=[];for(var c in e.inputs){var u=e.inputs[c];a.push(new hs(c,Ws(u),!1))}var l=[];for(var c in e.outputs){u=e.outputs[c];l.push(new hs(c,Ws(u),!1))}t.statements.push(Fs(s).set(Vs(fo.createComponentFactory).callFn([Ws(e.selector),t.importExpr(e.type.reference),Fs(o),new fs(a),new fs(l),qs(e.template.ngContentSelectors.map(function(t){return Ws(t)}))])).toDeclStmt(zs(fo.ComponentFactory,[Bs(t.importExpr(e.type.reference))],[Ao.Const]),[ys.Final,ys.Exported]))},t.prototype._compileComponent=function(t,e,n,r,i,o){var s=this._parseTemplate(e,n,r),a=s.template,c=s.pipes,u=i?Fs(i.stylesVar):qs([]),l=this._viewCompiler.compileComponent(t,e,a,u,c);return i&&Ku(this._symbolResolver,i,this._styleCompiler.needsStyleShim(e),o),l},t.prototype._parseTemplate=function(t,e,n){var r=this;if(this._templateAstCache.has(t.type.reference))return this._templateAstCache.get(t.type.reference);var i=t.template.preserveWhitespaces,o=n.map(function(t){return r._metadataResolver.getDirectiveSummary(t.reference)}),s=e.transitiveModule.pipes.map(function(t){return r._metadataResolver.getPipeSummary(t.reference)}),a=this._templateParser.parse(t,t.template.htmlAst,o,s,e.schemas,Ft(e.type,t,t.template),i);return this._templateAstCache.set(t.type.reference,a),a},t.prototype._createOutputContext=function(t){var e=this;return{statements:[],genFilePath:t,importExpr:function(n,r){if(void 0===r&&(r=null),!(n instanceof mt))throw new Error("Internal error: unknown identifier "+JSON.stringify(n));var i=e._symbolResolver.getTypeArity(n)||0,o=e._symbolResolver.getImportAs(n)||n,s=o.filePath,a=o.name,c=o.members,u=e._fileNameToModuleName(s,t),l=u===e._fileNameToModuleName(t,t)?null:u,p=r||[],d=i-p.length,h=p.concat(new Array(d).fill(Fo));return c.reduce(function(t,e){return t.prop(e)},Vs(new ns(l,a,null),h))}}},t.prototype._fileNameToModuleName=function(t,e){return this._summaryResolver.getKnownModuleName(t)||this._symbolResolver.getKnownModuleName(t)||this._host.fileNameToModuleName(t,e)},t.prototype._codegenStyles=function(t,e,n,r,i){var o=this._createOutputContext($u(n.moduleUrl,r,i)),s=this._styleCompiler.compileStyles(o,e,n,r);return Ku(this._symbolResolver,s,r,i),this._codegenSourceModule(t,o)},t.prototype._codegenSourceModule=function(t,e){return new Iu(t,e.genFilePath,e.statements)},t.prototype.listLazyRoutes=function(t,e){var n=this;if(t)return function t(e,r,i){void 0===r&&(r=new Set);void 0===i&&(i=[]);if(r.has(e)||!e.name)return i;r.add(e);var o=Ru(n._metadataResolver.getNgModuleMetadata(e,!0),n._reflector);for(var s=0,a=o;s<a.length;s++){var c=a[s];i.push(c),t(c.referencedModule,r,i)}return i}(Lu(t,this._reflector).referencedModule);if(e){for(var r=[],i=0,o=e.ngModules;i<o.length;i++)for(var s=0,a=Ru(o[i],this._reflector);s<a.length;s++){var c=a[s];r.push(c)}return r}throw new Error("Either route or analyzedModules has to be specified!")}}();function Yu(t){t.statements.push(Vs(fo.ComponentFactory).toStmt())}function Ku(t,e,n,r){e.dependencies.forEach(function(e){e.setValue(t.getStaticSymbol($u(e.moduleUrl,n,r),e.name))})}function $u(t,e,n){return t+(e?".shim":"")+".ngstyle"+n}function Qu(t,e,n,r){var i,o,s,a,c,u,l;return el((i=t,o=e,s=n,a=r,c=new Set,u=[],l=function(t){if(c.has(t)||!o.isSourceFile(t))return!1;c.add(t);var e=tl(o,s,a,t);u.push(e),e.ngModules.forEach(function(t){t.transitiveModule.modules.forEach(function(t){return l(t.reference.filePath)})})},i.forEach(function(t){return l(t)}),u))}function Zu(t,e,n,r){return Ju(Qu(t,e,n,r))}function Ju(t){if(t.symbolsMissingModule&&t.symbolsMissingModule.length)throw q(t.symbolsMissingModule.map(function(t){return"Cannot determine the module for class "+t.name+" in "+t.filePath+"! Add "+t.name+" to the NgModule to fix it."}).join("\n"));return t}function tl(t,e,n,r){var i=[],o=[],s=[],a=[],c=e.hasDecorators(r),u=!1;return r.endsWith(".d.ts")&&!c||e.getSymbolsOf(r).forEach(function(r){var c=e.resolveSymbol(r).metadata;if(c&&"error"!==c.__symbolic){var l,p,d,h,f=!1;if("class"===c.__symbolic)if(n.isDirective(r))f=!0,i.push(r);else if(n.isPipe(r))f=!0,o.push(r);else if(n.isNgModule(r)){var m=n.getNgModuleMetadata(r,!1);m&&(f=!0,a.push(m))}else n.isInjectable(r)&&(f=!0,s.push(r));f||(u=u||(l=t,p=c,d=!1,h=function(){function t(){}return t.prototype.visitArray=function(t,e){var n=this;t.forEach(function(t){return R(t,n,e)})},t.prototype.visitStringMap=function(t,e){var n=this;Object.keys(t).forEach(function(r){return R(t[r],n,e)})},t.prototype.visitPrimitive=function(t,e){},t.prototype.visitOther=function(t,e){t instanceof mt&&!l.isSourceFile(t.filePath)&&(d=!0)},t}(),R(p,new h,null),d))}}),{fileName:r,directives:i,pipes:o,ngModules:a,injectables:s,exportsNonSourceFiles:u}}function el(t){var e=[],n=new Map,r=new Set;t.forEach(function(t){t.ngModules.forEach(function(t){e.push(t),t.declaredDirectives.forEach(function(e){return n.set(e.reference,t)}),t.declaredPipes.forEach(function(e){return n.set(e.reference,t)})}),t.directives.forEach(function(t){return r.add(t)}),t.pipes.forEach(function(t){return r.add(t)})});var i=[];return r.forEach(function(t){n.has(t)||i.push(t)}),{ngModules:e,ngModuleByPipeOrDirective:n,symbolsMissingModule:i,files:t}}function nl(t){return Ju(el(t))}var rl="ngFormattedMessage";function il(t,e){if(void 0===e&&(e=0),!t)return"";var n=t.position?t.position.fileName+"("+(t.position.line+1)+","+(t.position.column+1)+")":"",r=n&&0!==e?" at "+n:"",i=""+(n&&0===e?n+": ":"")+t.message+r;return""+function t(e){if(e<=0)return"";if(e<6)return[""," ","  ","   ","    ","     "][e];var n=t(Math.floor(e/2));return n+n+(e%2==1?" ":"")}(e)+i+(t.next&&"\n"+il(t.next,e+2)||"")}var ol="@angular/core",sl=/^\$.*\$$/,al={__symbolic:"ignore"},cl="useValue",ul="provide",ll=new Set([cl,"useFactory","data"]);function pl(t){return t&&"ignore"==t.__symbolic}var dl=function(){function t(t,e,n,r,i){void 0===n&&(n=[]),void 0===r&&(r=[]);var o=this;this.summaryResolver=t,this.symbolResolver=e,this.errorRecorder=i,this.annotationCache=new Map,this.propertyCache=new Map,this.parameterCache=new Map,this.methodCache=new Map,this.conversionMap=new Map,this.annotationForParentClassWithSummaryKind=new Map,this.initializeConversionMap(),n.forEach(function(t){return o._registerDecoratorOrConstructor(o.getStaticSymbol(t.filePath,t.name),t.ctor)}),r.forEach(function(t){return o._registerFunction(o.getStaticSymbol(t.filePath,t.name),t.fn)}),this.annotationForParentClassWithSummaryKind.set(Ct.Directive,[p,f]),this.annotationForParentClassWithSummaryKind.set(Ct.Pipe,[m]),this.annotationForParentClassWithSummaryKind.set(Ct.NgModule,[_]),this.annotationForParentClassWithSummaryKind.set(Ct.Injectable,[O,m,p,f,_])}return t.prototype.componentModuleUrl=function(t){var e=this.findSymbolDeclaration(t);return this.symbolResolver.getResourcePath(e)},t.prototype.resolveExternalReference=function(t,e){var n=this.symbolResolver.getSymbolByModule(t.moduleName,t.name,e),r=this.findSymbolDeclaration(n);return e||(this.symbolResolver.recordModuleNameForFileName(n.filePath,t.moduleName),this.symbolResolver.recordImportAs(r,n)),r},t.prototype.findDeclaration=function(t,e,n){return this.findSymbolDeclaration(this.symbolResolver.getSymbolByModule(t,e,n))},t.prototype.tryFindDeclaration=function(t,e){var n=this;return this.symbolResolver.ignoreErrorsFor(function(){return n.findDeclaration(t,e)})},t.prototype.findSymbolDeclaration=function(t){var e=this.symbolResolver.resolveSymbol(t);if(e){var n=e.metadata;if(n&&"resolved"===n.__symbolic&&(n=n.symbol),n instanceof mt)return this.findSymbolDeclaration(e.metadata)}return t},t.prototype.annotations=function(t){var e=this.annotationCache.get(t);if(!e){e=[];var n=this.getTypeMetadata(t),r=this.findParentType(t,n);if(r){var i=this.annotations(r);e.push.apply(e,i)}var o=[];if(n.decorators&&(o=this.simplify(t,n.decorators),e.push.apply(e,o)),r&&!this.summaryResolver.isLibraryFile(t.filePath)&&this.summaryResolver.isLibraryFile(r.filePath)){var s=this.summaryResolver.resolveSummary(r);if(s&&s.type){var a=this.annotationForParentClassWithSummaryKind.get(s.type.summaryKind);a.some(function(t){return o.some(function(e){return t.isTypeOf(e)})})||this.reportError(El(fl("Class "+t.name+" in "+t.filePath+" extends from a "+Ct[s.type.summaryKind]+" in another compilation unit without duplicating the decorator",void 0,"Please add a "+a.map(function(t){return t.ngMetadataName}).join(" or ")+" decorator to the class"),t),t)}}this.annotationCache.set(t,e.filter(function(t){return!!t}))}return e},t.prototype.propMetadata=function(t){var e=this,n=this.propertyCache.get(t);if(!n){var r=this.getTypeMetadata(t);n={};var i=this.findParentType(t,r);if(i){var o=this.propMetadata(i);Object.keys(o).forEach(function(t){n[t]=o[t]})}var s=r.members||{};Object.keys(s).forEach(function(r){var i=s[r].find(function(t){return"property"==t.__symbolic||"method"==t.__symbolic}),o=[];n[r]&&o.push.apply(o,n[r]),n[r]=o,i&&i.decorators&&o.push.apply(o,e.simplify(t,i.decorators))}),this.propertyCache.set(t,n)}return n},t.prototype.parameters=function(t){var e=this;if(!(t instanceof mt))return this.reportError(new Error("parameters received "+JSON.stringify(t)+" which is not a StaticSymbol"),t),[];try{var n=this.parameterCache.get(t);if(!n){var r=this.getTypeMetadata(t),i=this.findParentType(t,r),o=r?r.members:null,s=o?o.__ctor__:null;if(s){var a=s.find(function(t){return"constructor"==t.__symbolic}),c=a.parameters||[],u=this.simplify(t,a.parameterDecorators||[]);n=[],c.forEach(function(r,i){var o=[],s=e.trySimplify(t,r);s&&o.push(s);var a=u?u[i]:null;a&&o.push.apply(o,a),n.push(o)})}else i&&(n=this.parameters(i));n||(n=[]),this.parameterCache.set(t,n)}return n}catch(e){throw console.error("Failed on type "+JSON.stringify(t)+" with error "+e),e}},t.prototype._methodNames=function(t){var e=this.methodCache.get(t);if(!e){var n=this.getTypeMetadata(t);e={};var r=this.findParentType(t,n);if(r){var i=this._methodNames(r);Object.keys(i).forEach(function(t){e[t]=i[t]})}var o=n.members||{};Object.keys(o).forEach(function(t){var n=o[t].some(function(t){return"method"==t.__symbolic});e[t]=e[t]||n}),this.methodCache.set(t,e)}return e},t.prototype.findParentType=function(t,e){var n=this.trySimplify(t,e.extends);if(n instanceof mt)return n},t.prototype.hasLifecycleHook=function(t,e){t instanceof mt||this.reportError(new Error("hasLifecycleHook received "+JSON.stringify(t)+" which is not a StaticSymbol"),t);try{return!!this._methodNames(t)[e]}catch(e){throw console.error("Failed on type "+JSON.stringify(t)+" with error "+e),e}},t.prototype._registerDecoratorOrConstructor=function(t,e){this.conversionMap.set(t,function(t,n){return new(e.bind.apply(e,[void 0].concat(n)))})},t.prototype._registerFunction=function(t,e){this.conversionMap.set(t,function(t,n){return e.apply(void 0,n)})},t.prototype.initializeConversionMap=function(){this.injectionToken=this.findDeclaration(ol,"InjectionToken"),this.opaqueToken=this.findDeclaration(ol,"OpaqueToken"),this.ROUTES=this.tryFindDeclaration("@angular/router","ROUTES"),this.ANALYZE_FOR_ENTRY_COMPONENTS=this.findDeclaration(ol,"ANALYZE_FOR_ENTRY_COMPONENTS"),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Host"),E),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Injectable"),O),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Self"),C),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"SkipSelf"),k),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Inject"),i),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Optional"),w),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Attribute"),s),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"ContentChild"),c),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"ContentChildren"),a),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"ViewChild"),l),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"ViewChildren"),u),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Input"),g),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Output"),v),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Pipe"),m),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"HostBinding"),y),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"HostListener"),b),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Directive"),p),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Component"),f),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"NgModule"),_),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Host"),E),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Self"),C),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"SkipSelf"),k),this._registerDecoratorOrConstructor(this.findDeclaration(ol,"Optional"),w)},t.prototype.getStaticSymbol=function(t,e,n){return this.symbolResolver.getStaticSymbol(t,e,n)},t.prototype.trySimplify=function(t,e){var n=this.errorRecorder;this.errorRecorder=function(t,e){};var r=this.simplify(t,e);return this.errorRecorder=n,r},t.prototype.simplify=function(t,e){var n,r=this,i=Ol.empty,o=new Map;try{n=function t(e,n,s,a){function c(t){var e=r.symbolResolver.resolveSymbol(t);return e?e.metadata:null}function u(n){return t(e,n,s,0)}function l(n,i){if(n===e)return t(n,i,s+1,a);try{return t(n,i,s+1,a)}catch(t){if(!ml(t))throw t;var o=t.chain?"references '"+t.symbol.name+"'":function(t){if(t.summary)return t.summary;switch(t.message){case gl:if(t.context&&t.context.className)return"references non-exported class "+t.context.className;break;case vl:return"is not initialized";case yl:return"is a destructured variable";case bl:return"could not be resolved";case _l:return t.context&&t.context.name?"calls '"+t.context.name+"'":"calls a function";case jl:return t.context&&t.context.name?"references local variable "+t.context.name:"references a local variable"}return"contains the error"}(t),c={message:"'"+n.name+"' "+o,position:t.position,next:t.chain};r.error({message:t.message,advise:t.advise,context:t.context,chain:c,symbol:n},e)}}function p(n){if(wl(n))return n;if(n instanceof Array){for(var d=[],h=0,f=n;h<f.length;h++){var m=f[h];if(m&&"spread"===m.__symbolic){var g=u(m.expression);if(Array.isArray(g)){for(var v=0,y=g;v<y.length;v++){var b=y[v];d.push(b)}continue}}var _=p(m);pl(_)||d.push(_)}return d}if(n instanceof mt)return n===r.injectionToken||r.conversionMap.has(n)||a>0&&!n.members.length?n:null!=(M=c(j=n))?l(j,M):j;if(n){if(n.__symbolic){var j=void 0;switch(n.__symbolic){case"binop":var x=p(n.left);if(pl(x))return x;var w=p(n.right);if(pl(w))return w;switch(n.operator){case"&&":return x&&w;case"||":return x||w;case"|":return x|w;case"^":return x^w;case"&":return x&w;case"==":return x==w;case"!=":return x!=w;case"===":return x===w;case"!==":return x!==w;case"<":return x<w;case">":return x>w;case"<=":return x<=w;case">=":return x>=w;case"<<":return x<<w;case">>":return x>>w;case"+":return x+w;case"-":return x-w;case"*":return x*w;case"/":return x/w;case"%":return x%w}return null;case"if":return p(p(n.condition)?n.thenExpression:n.elseExpression);case"pre":var O=p(n.operand);if(pl(O))return O;switch(n.operator){case"+":return O;case"-":return-O;case"!":return!O;case"~":return~O}return null;case"index":var C=u(n.expression),k=u(n.index);return C&&wl(k)?C[k]:null;case"select":var E=n.member,S=e,P=p(n.expression);if(P instanceof mt){var M,A=P.members.concat(E);return null!=(M=c(S=r.getStaticSymbol(P.filePath,P.name,A)))?l(S,M):S}return P&&wl(E)?l(S,P[E]):null;case"reference":var T=n.name,D=i.resolve(T);if(D!=Ol.missing)return D;break;case"resolved":try{return p(n.symbol)}catch(t){throw ml(t)&&null!=n.fileName&&null!=n.line&&null!=n.character&&(t.position={fileName:n.fileName,line:n.line,column:n.character}),t}case"class":case"function":return e;case"new":case"call":if((j=t(e,n.expression,s+1,0))instanceof mt){if(j===r.injectionToken||j===r.opaqueToken)return e;var I=n.arguments||[],R=r.conversionMap.get(j);if(R){var N=I.map(function(t){return l(e,t)}).map(function(t){return pl(t)?void 0:t});return R(e,N)}return function(t,n,a,c){if(n&&"function"==n.__symbolic){o.get(t)&&r.error({message:"Recursion is not supported",summary:"called '"+t.name+"' recursively",value:n},t);try{var u=n.value;if(u&&(0!=s||"error"!=u.__symbolic)){var d=n.parameters,h=n.defaults;a=a.map(function(t){return l(e,t)}).map(function(t){return pl(t)?void 0:t}),h&&h.length>a.length&&a.push.apply(a,h.slice(a.length).map(function(t){return p(t)})),o.set(t,!0);for(var f=Ol.build(),m=0;m<d.length;m++)f.define(d[m],a[m]);var g,v=i;try{i=f.done(),g=l(t,u)}finally{i=v}return g}}finally{o.delete(t)}}if(0===s)return al;var y=void 0;if(c&&"resolved"==c.__symbolic){var b=c.line,_=c.character,j=c.fileName;null!=j&&null!=b&&null!=_&&(y={fileName:j,line:b,column:_})}r.error({message:_l,context:t,value:n,position:y},e)}(j,c(j),I,n.expression)}return al;case"error":var L=n.message;return null!=n.line?r.error({message:L,context:n.context,value:n,position:{fileName:n.fileName,line:n.line,column:n.character}},e):r.error({message:L,context:n.context},e),al;case"ignore":return n}return null}return function(t,e){if(!t)return{};var n={};return Object.keys(t).forEach(function(r){var i=e(t[r],r);pl(i)||(sl.test(r)?Object.defineProperty(n,r,{enumerable:!1,configurable:!0,value:i}):n[r]=i)}),n}(n,function(i,o){if(ll.has(o)){if(o===cl&&ul in n){var c=p(n.provide);if(c===r.ROUTES||c==r.ANALYZE_FOR_ENTRY_COMPONENTS)return p(i)}return t(e,i,s,a+1)}return p(i)})}return al}return p(n)}(t,e,0,0)}catch(e){if(!this.errorRecorder)throw El(e,t);this.reportError(e,t)}if(!pl(n))return n},t.prototype.getTypeMetadata=function(t){var e=this.symbolResolver.resolveSymbol(t);return e&&e.metadata?e.metadata:{__symbolic:"class"}},t.prototype.reportError=function(t,e,n){if(!this.errorRecorder)throw t;this.errorRecorder(El(t,e),e&&e.filePath||n)},t.prototype.error=function(t,e){var n=t.message,r=t.summary,i=t.advise,o=t.position,s=t.context,a=(t.value,t.symbol),c=t.chain;this.reportError(fl(n,r,i,o,a,s,c),e)},t}(),hl="ngMetadataError";function fl(t,e,n,r,i,o,s){var a=q(t);return a[hl]=!0,n&&(a.advise=n),r&&(a.position=r),e&&(a.summary=e),o&&(a.context=o),s&&(a.chain=s),i&&(a.symbol=i),a}function ml(t){return!!t[hl]}var gl="Reference to non-exported class",vl="Variable not initialized",yl="Destructuring not supported",bl="Could not resolve type",_l="Function call not supported",jl="Reference to a local symbol",xl="Lambda not supported";function wl(t){return null===t||"function"!=typeof t&&"object"!=typeof t}var Ol=function(){function t(){}return t.build=function(){var e=new Map;return{define:function(t,n){return e.set(t,n),this},done:function(){return e.size>0?new Cl(e):t.empty}}},t.missing={},t.empty={resolve:function(e){return t.missing}},t}(),Cl=function(t){function e(e){var n=t.call(this)||this;return n.bindings=e,n}return Object(r.b)(e,t),e.prototype.resolve=function(t){return this.bindings.has(t)?this.bindings.get(t):Ol.missing},e}(Ol);function kl(t,e){return{message:""+function(t,e){switch(t){case gl:if(e&&e.className)return"References to a non-exported class are not supported in decorators but "+e.className+" was referenced.";break;case vl:return"Only initialized variables and constants can be referenced in decorators because the value of this variable is needed by the template compiler";case yl:return"Referencing an exported destructured variable or constant is not supported in decorators and this value is needed by the template compiler";case bl:if(e&&e.typeName)return"Could not resolve type "+e.typeName;break;case _l:return e&&e.name?"Function calls are not supported in decorators but '"+e.name+"' was called":"Function calls are not supported in decorators";case jl:if(e&&e.name)return"Reference to a local (non-exported) symbols are not supported in decorators but '"+e.name+"' was referenced";break;case xl:return"Function expressions are not supported in decorators"}return t}(t.message,t.context)+(t.symbol?" in '"+t.symbol.name+"'":""),position:t.position,next:t.next?kl(t.next,e):e?{message:e}:void 0}}function El(t,e){if(ml(t)){var n=t.position,r={message:"Error during template compile of '"+e.name+"'",position:n,next:{message:t.message,next:t.chain,context:t.context,symbol:t.symbol}},i=t.advise||function(t,e){switch(t){case gl:if(e&&e.className)return"Consider exporting '"+e.className+"'";break;case yl:return"Consider simplifying to avoid destructuring";case jl:if(e&&e.name)return"Consider exporting '"+e.name+"'";break;case xl:return"Consider changing the function expression into an exported function"}}(t.message,t.context);return o=kl(r,i),(s=q(il(o)+"."))[rl]=!0,s.chain=o,s.position=o.position,s}var o,s;return t}var Sl=function(){function t(t,e){this.host=t,this.staticSymbolCache=e,this.summaryCache=new Map,this.loadedFilePaths=new Map,this.importAs=new Map,this.knownFileNameToModuleNames=new Map}return t.prototype.isLibraryFile=function(t){return!this.host.isSourceFile(so(t))},t.prototype.toSummaryFileName=function(t,e){return this.host.toSummaryFileName(t,e)},t.prototype.fromSummaryFileName=function(t,e){return this.host.fromSummaryFileName(t,e)},t.prototype.resolveSummary=function(t){t.assertNoMembers();var e=this.summaryCache.get(t);return e||(this._loadSummaryFile(t.filePath),e=this.summaryCache.get(t)),e||null},t.prototype.getSymbolsOf=function(t){return this._loadSummaryFile(t)?Array.from(this.summaryCache.keys()).filter(function(e){return e.filePath===t}):null},t.prototype.getImportAs=function(t){return t.assertNoMembers(),this.importAs.get(t)},t.prototype.getKnownModuleName=function(t){return this.knownFileNameToModuleNames.get(t)||null},t.prototype.addSummary=function(t){this.summaryCache.set(t.symbol,t)},t.prototype._loadSummaryFile=function(t){var e=this,n=this.loadedFilePaths.get(t);if(null!=n)return n;var r,i,o,s=null;if(this.isLibraryFile(t)){var a=co(t);try{s=this.host.loadSummary(a)}catch(t){throw console.error("Error loading summary file "+a),t}}if(n=null!=s,this.loadedFilePaths.set(t,n),s){var c=(r=this.staticSymbolCache,i=t,o=s,new Wu(r,this).deserialize(i,o)),u=c.moduleName,l=c.summaries,p=c.importAs;l.forEach(function(t){return e.summaryCache.set(t.symbol,t)}),u&&this.knownFileNameToModuleNames.set(t,u),p.forEach(function(t){e.importAs.set(t.symbol,t.importAs)})}return n},t}();function Pl(t){return{resolve:function(e,n){var r=t.resourceNameToFileName(n,e);if(!r)throw q("Couldn't resolve resource "+n+" from "+e);return r}}}var Ml=function(){return function(){}}(),Al=function(){function t(){this._summaries=new Map}return t.prototype.isLibraryFile=function(){return!1},t.prototype.toSummaryFileName=function(t){return t},t.prototype.fromSummaryFileName=function(t){return t},t.prototype.resolveSummary=function(t){return this._summaries.get(t)||null},t.prototype.getSymbolsOf=function(){return[]},t.prototype.getImportAs=function(t){return t},t.prototype.getKnownModuleName=function(t){return null},t.prototype.addSummary=function(t){this._summaries.set(t.symbol,t)},t}();function Tl(t,e,n,r,i){for(var o=r.createChildWihtLocalVars(),s=0;s<t.length;s++)o.vars.set(t[s],e[s]);var a=i.visitAllStatements(n,o);return a?a.value:null}var Dl=function(){function t(t,e,n,r){this.parent=t,this.instance=e,this.className=n,this.vars=r,this.exports=[]}return t.prototype.createChildWihtLocalVars=function(){return new t(this,this.instance,this.className,new Map)},t}(),Il=function(){return function(t){this.value=t}}();var Rl=function(){function t(t){this.reflector=t}return t.prototype.debugAst=function(t){return e=t,n=new ka,r=xa.createRoot(),(Array.isArray(e)?e:[e]).forEach(function(t){if(t instanceof bs)t.visitStatement(n,r);else if(t instanceof Uo)t.visitExpression(n,r);else{if(!(t instanceof To))throw new Error("Don't know how to print debug info for "+t);t.visitType(n,r)}}),r.toSource();var e,n,r},t.prototype.visitDeclareVarStmt=function(t,e){return e.vars.set(t.name,t.value.visitExpression(this,e)),t.hasModifier(ys.Exported)&&e.exports.push(t.name),null},t.prototype.visitWriteVarExpr=function(t,e){for(var n=t.value.visitExpression(this,e),r=e;null!=r;){if(r.vars.has(t.name))return r.vars.set(t.name,n),n;r=r.parent}throw new Error("Not declared variable "+t.name)},t.prototype.visitReadVarExpr=function(t,e){var n=t.name;if(null!=t.builtin)switch(t.builtin){case Wo.Super:return e.instance.__proto__;case Wo.This:return e.instance;case Wo.CatchError:n=Ll;break;case Wo.CatchStack:n=Fl;break;default:throw new Error("Unknown builtin variable "+t.builtin)}for(var r=e;null!=r;){if(r.vars.has(n))return r.vars.get(n);r=r.parent}throw new Error("Not declared variable "+n)},t.prototype.visitWriteKeyExpr=function(t,e){var n=t.receiver.visitExpression(this,e),r=t.index.visitExpression(this,e),i=t.value.visitExpression(this,e);return n[r]=i,i},t.prototype.visitWritePropExpr=function(t,e){var n=t.receiver.visitExpression(this,e),r=t.value.visitExpression(this,e);return n[t.name]=r,r},t.prototype.visitInvokeMethodExpr=function(t,e){var n,r=t.receiver.visitExpression(this,e),i=this.visitAllExpressions(t.args,e);if(null!=t.builtin)switch(t.builtin){case $o.ConcatArray:n=r.concat.apply(r,i);break;case $o.SubscribeObservable:n=r.subscribe({next:i[0]});break;case $o.Bind:n=r.bind.apply(r,i);break;default:throw new Error("Unknown builtin method "+t.builtin)}else n=r[t.name].apply(r,i);return n},t.prototype.visitInvokeFunctionExpr=function(t,e){var n=this.visitAllExpressions(t.args,e),r=t.fn;return r instanceof Go&&r.builtin===Wo.Super?(e.instance.constructor.prototype.constructor.apply(e.instance,n),null):t.fn.visitExpression(this,e).apply(null,n)},t.prototype.visitReturnStmt=function(t,e){return new Il(t.value.visitExpression(this,e))},t.prototype.visitDeclareClassStmt=function(t,e){var n=function(t,e,n){var r={};t.getters.forEach(function(i){r[i.name]={configurable:!1,get:function(){var r=new Dl(e,this,t.name,e.vars);return Tl([],[],i.body,r,n)}}}),t.methods.forEach(function(i){var o=i.params.map(function(t){return t.name});r[i.name]={writable:!1,configurable:!1,value:function(){for(var r=[],s=0;s<arguments.length;s++)r[s]=arguments[s];var a=new Dl(e,this,t.name,e.vars);return Tl(o,r,i.body,a,n)}}});var i=t.constructorMethod.params.map(function(t){return t.name}),o=function(){for(var r=this,o=[],s=0;s<arguments.length;s++)o[s]=arguments[s];var a=new Dl(e,this,t.name,e.vars);t.fields.forEach(function(t){r[t.name]=void 0}),Tl(i,o,t.constructorMethod.body,a,n)},s=t.parent?t.parent.visitExpression(n,e):Object;return o.prototype=Object.create(s.prototype,r),o}(t,e,this);return e.vars.set(t.name,n),t.hasModifier(ys.Exported)&&e.exports.push(t.name),null},t.prototype.visitExpressionStmt=function(t,e){return t.expr.visitExpression(this,e)},t.prototype.visitIfStmt=function(t,e){return t.condition.visitExpression(this,e)?this.visitAllStatements(t.trueCase,e):null!=t.falseCase?this.visitAllStatements(t.falseCase,e):null},t.prototype.visitTryCatchStmt=function(t,e){try{return this.visitAllStatements(t.bodyStmts,e)}catch(r){var n=e.createChildWihtLocalVars();return n.vars.set(Ll,r),n.vars.set(Fl,r.stack),this.visitAllStatements(t.catchStmts,n)}},t.prototype.visitThrowStmt=function(t,e){throw t.error.visitExpression(this,e)},t.prototype.visitCommentStmt=function(t,e){return null},t.prototype.visitInstantiateExpr=function(t,e){var n=this.visitAllExpressions(t.args,e),r=t.classExpr.visitExpression(this,e);return new(r.bind.apply(r,[void 0].concat(n)))},t.prototype.visitLiteralExpr=function(t,e){return t.value},t.prototype.visitExternalExpr=function(t,e){return this.reflector.resolveExternalReference(t.value)},t.prototype.visitConditionalExpr=function(t,e){return t.condition.visitExpression(this,e)?t.trueCase.visitExpression(this,e):null!=t.falseCase?t.falseCase.visitExpression(this,e):null},t.prototype.visitNotExpr=function(t,e){return!t.condition.visitExpression(this,e)},t.prototype.visitAssertNotNullExpr=function(t,e){return t.condition.visitExpression(this,e)},t.prototype.visitCastExpr=function(t,e){return t.value.visitExpression(this,e)},t.prototype.visitFunctionExpr=function(t,e){return Nl(t.params.map(function(t){return t.name}),t.statements,e,this)},t.prototype.visitDeclareFunctionStmt=function(t,e){var n=t.params.map(function(t){return t.name});return e.vars.set(t.name,Nl(n,t.statements,e,this)),t.hasModifier(ys.Exported)&&e.exports.push(t.name),null},t.prototype.visitBinaryOperatorExpr=function(t,e){var n=this,r=function(){return t.lhs.visitExpression(n,e)},i=function(){return t.rhs.visitExpression(n,e)};switch(t.operator){case Bo.Equals:return r()==i();case Bo.Identical:return r()===i();case Bo.NotEquals:return r()!=i();case Bo.NotIdentical:return r()!==i();case Bo.And:return r()&&i();case Bo.Or:return r()||i();case Bo.Plus:return r()+i();case Bo.Minus:return r()-i();case Bo.Divide:return r()/i();case Bo.Multiply:return r()*i();case Bo.Modulo:return r()%i();case Bo.Lower:return r()<i();case Bo.LowerEquals:return r()<=i();case Bo.Bigger:return r()>i();case Bo.BiggerEquals:return r()>=i();default:throw new Error("Unknown operator "+t.operator)}},t.prototype.visitReadPropExpr=function(t,e){return t.receiver.visitExpression(this,e)[t.name]},t.prototype.visitReadKeyExpr=function(t,e){return t.receiver.visitExpression(this,e)[t.index.visitExpression(this,e)]},t.prototype.visitLiteralArrayExpr=function(t,e){return this.visitAllExpressions(t.entries,e)},t.prototype.visitLiteralMapExpr=function(t,e){var n=this,r={};return t.entries.forEach(function(t){return r[t.key]=t.value.visitExpression(n,e)}),r},t.prototype.visitCommaExpr=function(t,e){var n=this.visitAllExpressions(t.parts,e);return n[n.length-1]},t.prototype.visitAllExpressions=function(t,e){var n=this;return t.map(function(t){return t.visitExpression(n,e)})},t.prototype.visitAllStatements=function(t,e){for(var n=0;n<t.length;n++){var r=t[n].visitStatement(this,e);if(r instanceof Il)return r}return null},t}();function Nl(t,e,n,r){return function(){for(var i=[],o=0;o<arguments.length;o++)i[o]=arguments[o];return Tl(t,i,e,n,r)}}var Ll="error",Fl="stack";function Vl(t,e,n,r){var i=new zl(n),o=xa.createRoot();return i.visitAllStatements(e,o),i.createReturnStmt(o),function(t,e,n,r){var i=e.toSource()+"\n//# sourceURL="+t,o=[],s=[];for(var a in n)o.push(a),s.push(n[a]);if(r){var c=(new(Function.bind.apply(Function,[void 0].concat(o.concat("return null;"))))).toString(),u=c.slice(0,c.indexOf("return null;")).split("\n").length-1;i+="\n"+e.toSourceMapGenerator(t,u).toJsComment()}return(new(Function.bind.apply(Function,[void 0].concat(o.concat(i))))).apply(void 0,s)}(t,o,i.getArgs(),r)}var zl=function(t){function e(e){var n=t.call(this)||this;return n.reflector=e,n._evalArgNames=[],n._evalArgValues=[],n._evalExportedVars=[],n}return Object(r.b)(e,t),e.prototype.createReturnStmt=function(t){new ws(new fs(this._evalExportedVars.map(function(t){return new hs(t,Fs(t),!1)}))).visitStatement(this,t)},e.prototype.getArgs=function(){for(var t={},e=0;e<this._evalArgNames.length;e++)t[this._evalArgNames[e]]=this._evalArgValues[e];return t},e.prototype.visitExternalExpr=function(t,e){var n=this.reflector.resolveExternalReference(t.value),r=this._evalArgValues.indexOf(n);if(-1===r){r=this._evalArgValues.length,this._evalArgValues.push(n);var i=_t({reference:n})||"val";this._evalArgNames.push("jit_"+i+"_"+r)}return e.print(t,this._evalArgNames[r]),null},e.prototype.visitDeclareVarStmt=function(e,n){return e.hasModifier(ys.Exported)&&this._evalExportedVars.push(e.name),t.prototype.visitDeclareVarStmt.call(this,e,n)},e.prototype.visitDeclareFunctionStmt=function(e,n){return e.hasModifier(ys.Exported)&&this._evalExportedVars.push(e.name),t.prototype.visitDeclareFunctionStmt.call(this,e,n)},e.prototype.visitDeclareClassStmt=function(e,n){return e.hasModifier(ys.Exported)&&this._evalExportedVars.push(e.name),t.prototype.visitDeclareClassStmt.call(this,e,n)},e}(function(t){function e(){return t.call(this,!1)||this}return Object(r.b)(e,t),e.prototype.visitDeclareClassStmt=function(t,e){var n=this;return e.pushClass(t),this._visitClassConstructor(t,e),null!=t.parent&&(e.print(t,t.name+".prototype = Object.create("),t.parent.visitExpression(this,e),e.println(t,".prototype);")),t.getters.forEach(function(r){return n._visitClassGetter(t,r,e)}),t.methods.forEach(function(r){return n._visitClassMethod(t,r,e)}),e.popClass(),null},e.prototype._visitClassConstructor=function(t,e){e.print(t,"function "+t.name+"("),null!=t.constructorMethod&&this._visitParams(t.constructorMethod.params,e),e.println(t,") {"),e.incIndent(),null!=t.constructorMethod&&t.constructorMethod.body.length>0&&(e.println(t,"var self = this;"),this.visitAllStatements(t.constructorMethod.body,e)),e.decIndent(),e.println(t,"}")},e.prototype._visitClassGetter=function(t,e,n){n.println(t,"Object.defineProperty("+t.name+".prototype, '"+e.name+"', { get: function() {"),n.incIndent(),e.body.length>0&&(n.println(t,"var self = this;"),this.visitAllStatements(e.body,n)),n.decIndent(),n.println(t,"}});")},e.prototype._visitClassMethod=function(t,e,n){n.print(t,t.name+".prototype."+e.name+" = function("),this._visitParams(e.params,n),n.println(t,") {"),n.incIndent(),e.body.length>0&&(n.println(t,"var self = this;"),this.visitAllStatements(e.body,n)),n.decIndent(),n.println(t,"};")},e.prototype.visitReadVarExpr=function(e,n){if(e.builtin===Wo.This)n.print(e,"self");else{if(e.builtin===Wo.Super)throw new Error("'super' needs to be handled at a parent ast node, not at the variable level!");t.prototype.visitReadVarExpr.call(this,e,n)}return null},e.prototype.visitDeclareVarStmt=function(t,e){return e.print(t,"var "+t.name+" = "),t.value.visitExpression(this,e),e.println(t,";"),null},e.prototype.visitCastExpr=function(t,e){return t.value.visitExpression(this,e),null},e.prototype.visitInvokeFunctionExpr=function(e,n){var r=e.fn;return r instanceof Go&&r.builtin===Wo.Super?(n.currentClass.parent.visitExpression(this,n),n.print(e,".call(this"),e.args.length>0&&(n.print(e,", "),this.visitAllExpressions(e.args,n,",")),n.print(e,")")):t.prototype.visitInvokeFunctionExpr.call(this,e,n),null},e.prototype.visitFunctionExpr=function(t,e){return e.print(t,"function("),this._visitParams(t.params,e),e.println(t,") {"),e.incIndent(),this.visitAllStatements(t.statements,e),e.decIndent(),e.print(t,"}"),null},e.prototype.visitDeclareFunctionStmt=function(t,e){return e.print(t,"function "+t.name+"("),this._visitParams(t.params,e),e.println(t,") {"),e.incIndent(),this.visitAllStatements(t.statements,e),e.decIndent(),e.println(t,"}"),null},e.prototype.visitTryCatchStmt=function(t,e){e.println(t,"try {"),e.incIndent(),this.visitAllStatements(t.bodyStmts,e),e.decIndent(),e.println(t,"} catch ("+ba.name+") {"),e.incIndent();var n=[_a.set(ba.prop("stack")).toDeclStmt(null,[ys.Final])].concat(t.catchStmts);return this.visitAllStatements(n,e),e.decIndent(),e.println(t,"}"),null},e.prototype._visitParams=function(t,e){this.visitAllObjects(function(t){return e.print(null,t.name)},t,e,",")},e.prototype.getBuiltinMethodName=function(t){var e;switch(t){case $o.ConcatArray:e="concat";break;case $o.SubscribeObservable:e="subscribe";break;case $o.Bind:e="bind";break;default:throw new Error("Unknown builtin method: "+t)}return e},e}(wa)),Bl=function(){function t(t,e,n,r,i,o,s,a,c,u){this._metadataResolver=t,this._templateParser=e,this._styleCompiler=n,this._viewCompiler=r,this._ngModuleCompiler=i,this._summaryResolver=o,this._reflector=s,this._compilerConfig=a,this._console=c,this.getExtraNgModuleProviders=u,this._compiledTemplateCache=new Map,this._compiledHostTemplateCache=new Map,this._compiledDirectiveWrapperCache=new Map,this._compiledNgModuleCache=new Map,this._sharedStylesheetCount=0,this._addedAotSummaries=new Set}return t.prototype.compileModuleSync=function(t){return V(this._compileModuleAndComponents(t,!0))},t.prototype.compileModuleAsync=function(t){return Promise.resolve(this._compileModuleAndComponents(t,!1))},t.prototype.compileModuleAndAllComponentsSync=function(t){return V(this._compileModuleAndAllComponents(t,!0))},t.prototype.compileModuleAndAllComponentsAsync=function(t){return Promise.resolve(this._compileModuleAndAllComponents(t,!1))},t.prototype.getComponentFactory=function(t){return this._metadataResolver.getDirectiveSummary(t).componentFactory},t.prototype.loadAotSummaries=function(t){this.clearCache(),this._addAotSummaries(t)},t.prototype._addAotSummaries=function(t){if(!this._addedAotSummaries.has(t)){this._addedAotSummaries.add(t);for(var e=t(),n=0;n<e.length;n++){var r=e[n];if("function"==typeof r)this._addAotSummaries(r);else{var i=r;this._summaryResolver.addSummary({symbol:i.type.reference,metadata:null,type:i})}}}},t.prototype.hasAotSummary=function(t){return!!this._summaryResolver.resolveSummary(t)},t.prototype._filterJitIdentifiers=function(t){var e=this;return t.map(function(t){return t.reference}).filter(function(t){return!e.hasAotSummary(t)})},t.prototype._compileModuleAndComponents=function(t,e){var n=this;return z(this._loadModules(t,e),function(){return n._compileComponents(t,null),n._compileModule(t)})},t.prototype._compileModuleAndAllComponents=function(t,e){var n=this;return z(this._loadModules(t,e),function(){var e=[];return n._compileComponents(t,e),{ngModuleFactory:n._compileModule(t),componentFactories:e}})},t.prototype._loadModules=function(t,e){var n=this,r=[],i=this._metadataResolver.getNgModuleMetadata(t);return this._filterJitIdentifiers(i.transitiveModule.modules).forEach(function(t){var i=n._metadataResolver.getNgModuleMetadata(t);n._filterJitIdentifiers(i.declaredDirectives).forEach(function(t){var o=n._metadataResolver.loadDirectiveMetadata(i.type.reference,t,e);o&&r.push(o)}),n._filterJitIdentifiers(i.declaredPipes).forEach(function(t){return n._metadataResolver.getOrLoadPipeMetadata(t)})}),B(r)},t.prototype._compileModule=function(t){var e,n=this._compiledNgModuleCache.get(t);if(!n){var r=this._metadataResolver.getNgModuleMetadata(t),i=this.getExtraNgModuleProviders(r.type.reference),o=Ul(),s=this._ngModuleCompiler.compile(o,r,i);n=this._interpretOrJit((e=r,Lt(_t(e.type)+"/module.ngfactory.js")),o.statements)[s.ngModuleFactoryVar],this._compiledNgModuleCache.set(r.type.reference,n)}return n},t.prototype._compileComponents=function(t,e){var n=this,r=this._metadataResolver.getNgModuleMetadata(t),i=new Map,o=new Set,s=this._filterJitIdentifiers(r.transitiveModule.modules);s.forEach(function(t){var r=n._metadataResolver.getNgModuleMetadata(t);n._filterJitIdentifiers(r.declaredDirectives).forEach(function(t){i.set(t,r);var s=n._metadataResolver.getDirectiveMetadata(t);if(s.isComponent&&(o.add(n._createCompiledTemplate(s,r)),e)){var a=n._createCompiledHostTemplate(s.type.reference,r);o.add(a),e.push(s.componentFactory)}})}),s.forEach(function(t){var e=n._metadataResolver.getNgModuleMetadata(t);n._filterJitIdentifiers(e.declaredDirectives).forEach(function(t){var e=n._metadataResolver.getDirectiveMetadata(t);e.isComponent&&e.entryComponents.forEach(function(t){var e=i.get(t.componentType);o.add(n._createCompiledHostTemplate(t.componentType,e))})}),e.entryComponents.forEach(function(t){if(!n.hasAotSummary(t.componentType.reference)){var e=i.get(t.componentType);o.add(n._createCompiledHostTemplate(t.componentType,e))}})}),o.forEach(function(t){return n._compileTemplate(t)})},t.prototype.clearCacheFor=function(t){this._compiledNgModuleCache.delete(t),this._metadataResolver.clearCacheFor(t),this._compiledHostTemplateCache.delete(t),this._compiledTemplateCache.get(t)&&this._compiledTemplateCache.delete(t)},t.prototype.clearCache=function(){this._metadataResolver.clearCache(),this._compiledTemplateCache.clear(),this._compiledHostTemplateCache.clear(),this._compiledNgModuleCache.clear()},t.prototype._createCompiledHostTemplate=function(t,e){if(!e)throw new Error("Component "+Y(t)+" is not part of any NgModule or the module has not been imported into your module.");var n=this._compiledHostTemplateCache.get(t);if(!n){var r=this._metadataResolver.getDirectiveMetadata(t);Hl(r);var i=this._metadataResolver.getHostComponentMetadata(r,r.componentFactory.viewDefFactory);n=new ql(!0,r.type,i,e,[r.type]),this._compiledHostTemplateCache.set(t,n)}return n},t.prototype._createCompiledTemplate=function(t,e){var n=this._compiledTemplateCache.get(t.type.reference);return n||(Hl(t),n=new ql(!1,t.type,t,e,e.transitiveModule.directives),this._compiledTemplateCache.set(t.type.reference,n)),n},t.prototype._compileTemplate=function(t){var e=this;if(!t.isCompiled){var n=t.compMeta,r=new Map,i=Ul(),o=this._styleCompiler.compileComponent(i,n);n.template.externalStylesheets.forEach(function(t){var i=e._styleCompiler.compileStyles(Ul(),n,t);r.set(t.moduleUrl,i)}),this._resolveStylesCompileResult(o,r);t.ngModule.transitiveModule.pipes.map(function(t){return e._metadataResolver.getPipeSummary(t.reference)});var s,a,c=this._parseTemplate(n,t.ngModule,t.directives),u=c.template,l=c.pipes,p=this._viewCompiler.compileComponent(i,n,u,Fs(o.stylesVar),l),d=this._interpretOrJit((s=t.ngModule.type,a=t.compMeta,Lt(_t(s)+"/"+_t(a.type)+".ngfactory.js")),i.statements),h=d[p.viewClassVar],f=d[p.rendererTypeVar];t.compiled(h,f)}},t.prototype._parseTemplate=function(t,e,n){var r=this,i=t.template.preserveWhitespaces,o=n.map(function(t){return r._metadataResolver.getDirectiveSummary(t.reference)}),s=e.transitiveModule.pipes.map(function(t){return r._metadataResolver.getPipeSummary(t.reference)});return this._templateParser.parse(t,t.template.htmlAst,o,s,e.schemas,Ft(e.type,t,t.template),i)},t.prototype._resolveStylesCompileResult=function(t,e){var n=this;t.dependencies.forEach(function(t,r){var i=e.get(t.moduleUrl),o=n._resolveAndEvalStylesCompileResult(i,e);t.setValue(o)})},t.prototype._resolveAndEvalStylesCompileResult=function(t,e){return this._resolveStylesCompileResult(t,e),this._interpretOrJit((n=t.meta,r=this._sharedStylesheetCount++,i=n.moduleUrl.split(/\/\\/g),Lt("css/"+r+i[i.length-1]+".ngstyle.js")),t.outputCtx.statements)[t.stylesVar];var n,r,i},t.prototype._interpretOrJit=function(t,e){return this._compilerConfig.useJit?Vl(t,e,this._reflector,this._compilerConfig.jitDevMode):function(t,e){var n=new Dl(null,null,null,new Map);new Rl(e).visitAllStatements(t,n);var r={};return n.exports.forEach(function(t){r[t]=n.vars.get(t)}),r}(e,this._reflector)},t}(),ql=function(){function t(t,e,n,r,i){this.isHost=t,this.compType=e,this.compMeta=n,this.ngModule=r,this.directives=i,this._viewClass=null,this.isCompiled=!1}return t.prototype.compiled=function(t,e){this._viewClass=t,this.compMeta.componentViewType.setDelegate(t);for(var n in e)this.compMeta.rendererType[n]=e[n];this.isCompiled=!0},t}();function Hl(t){if(!t.isComponent)throw new Error("Could not compile '"+_t(t.type)+"' because it is not a component.")}function Ul(){return{statements:[],genFilePath:"",importExpr:function(t){return Vs({name:_t(t),moduleName:null,runtime:t})}}}var Wl=function(){return function(){}}();var Gl=function(){function t(t){void 0===t&&(t=null),this._packagePrefix=t}return t.prototype.resolve=function(t,e){var n=e;null!=t&&t.length>0&&(n=function(t,e){var n=$l(encodeURI(e)),r=$l(t);if(null!=n[Kl.Scheme])return Ql(n);n[Kl.Scheme]=r[Kl.Scheme];for(var i=Kl.Scheme;i<=Kl.Port;i++)null==n[i]&&(n[i]=r[i]);if("/"==n[Kl.Path][0])return Ql(n);var o=r[Kl.Path];null==o&&(o="/");var s=o.lastIndexOf("/");return o=o.substring(0,s+1)+n[Kl.Path],n[Kl.Path]=o,Ql(n)}(t,n));var r=$l(n),i=this._packagePrefix;if(null!=i&&null!=r&&"package"==r[Kl.Scheme]){var o=r[Kl.Path];return(i=i.replace(/\/+$/,""))+"/"+(o=o.replace(/^\/+/,""))}return n},t}();function Xl(t){var e=$l(t);return e&&e[Kl.Scheme]||""}var Yl=new RegExp("^(?:([^:/?#.]+):)?(?://(?:([^/?#]*)@)?([\\w\\d\\-\\u0100-\\uffff.%]*)(?::([0-9]+))?)?([^?#]+)?(?:\\?([^#]*))?(?:#(.*))?$"),Kl={Scheme:1,UserInfo:2,Domain:3,Port:4,Path:5,QueryData:6,Fragment:7};function $l(t){return t.match(Yl)}function Ql(t){var e,n,r,i,o,s,a,c,u=t[Kl.Path];return u=null==u?"":function(t){if("/"==t)return"/";for(var e="/"==t[0]?"/":"",n="/"===t[t.length-1]?"/":"",r=t.split("/"),i=[],o=0,s=0;s<r.length;s++){var a=r[s];switch(a){case"":case".":break;case"..":i.length>0?i.pop():o++;break;default:i.push(a)}}if(""==e){for(;o-- >0;)i.unshift("..");0===i.length&&i.push(".")}return e+i.join("/")+n}(u),t[Kl.Path]=u,e=t[Kl.Scheme],n=t[Kl.UserInfo],r=t[Kl.Domain],i=t[Kl.Port],o=u,s=t[Kl.QueryData],a=t[Kl.Fragment],c=[],null!=e&&c.push(e+":"),null!=r&&(c.push("//"),null!=n&&c.push(n+"@"),c.push(r),null!=i&&c.push(":"+i)),null!=o&&c.push(o),null!=s&&c.push("?"+s),null!=a&&c.push("#"+a),c.join("")}Kl[Kl.Scheme]="Scheme",Kl[Kl.UserInfo]="UserInfo",Kl[Kl.Domain]="Domain",Kl[Kl.Port]="Port",Kl[Kl.Path]="Path",Kl[Kl.QueryData]="QueryData",Kl[Kl.Fragment]="Fragment";var Zl=function(){function t(){}return t.prototype.get=function(t){return""},t}();!function(){function t(t,e,n,r){this.host=t,this.staticSymbolResolver=e,this.messageBundle=n,this.metadataResolver=r}t.prototype.extract=function(t){var e=this,n=Zu(t,this.host,this.staticSymbolResolver,this.metadataResolver),r=n.files,i=n.ngModules;return Promise.all(i.map(function(t){return e.metadataResolver.loadNgModuleDirectiveAndPipeMetadata(t.type.reference,!1)})).then(function(){var t=[];if(r.forEach(function(n){var r=[];n.directives.forEach(function(t){var n=e.metadataResolver.getDirectiveMetadata(t);n&&n.isComponent&&r.push(n)}),r.forEach(function(r){var i=r.template.template,o=Kt.fromArray(r.template.interpolation);t.push.apply(t,e.messageBundle.updateFromTemplate(i,n.fileName,o))})}),t.length)throw new Error(t.map(function(t){return t.toString()}).join("\n"));return e.messageBundle})},t.create=function(e,n){var r=new Qi,i=Pl(e),o=new gt,s=new Sl(e,o),a=new Vu(e,o,s),c=new dl(s,a),u=new ht({defaultEncapsulation:d.Emulated,useJit:!1}),l=new xe({get:function(t){return e.loadResource(t)}},i,r,u),p=new La,h=new ko(u,r,new pa(c),new Ce(c),new Ea(c),s,p,l,console,o,c),f=new Tu(r,[],{},n);return{extractor:new t(e,a,f,h),staticReflector:c}}}()},"../../../core/esm5/core.js":function(t,e,n){"use strict";(function(t){n.d(e,"f",function(){return $e}),n.d(e,"_13",function(){return He}),n.d(e,"_16",function(){return Ue}),n.d(e,"_12",function(){return Ge}),n.d(e,"K",function(){return We}),n.d(e,"b",function(){return Zt}),n.d(e,"O",function(){return ie}),n.d(e,"Q",function(){return ee}),n.d(e,"P",function(){return ne}),n.d(e,"c",function(){return $t}),n.d(e,"d",function(){return Qt}),n.d(e,"_15",function(){return yn}),n.d(e,"_5",function(){return Re}),n.d(e,"_18",function(){return Le}),n.d(e,"_2",function(){return Hn}),n.d(e,"_3",function(){return Un}),n.d(e,"F",function(){return qn}),n.d(e,"G",function(){return Wn}),n.d(e,"e",function(){return Kn}),n.d(e,"w",function(){return Ee}),n.d(e,"v",function(){return jt}),n.d(e,"X",function(){return Qn}),n.d(e,"Y",function(){return $n}),n.d(e,"a",function(){return g}),n.d(e,"g",function(){return v}),n.d(e,"q",function(){return _}),n.d(e,"r",function(){return b}),n.d(e,"_7",function(){return x}),n.d(e,"_8",function(){return j}),n.d(e,"m",function(){return k}),n.d(e,"s",function(){return C}),n.d(e,"C",function(){return S}),n.d(e,"N",function(){return P}),n.d(e,"R",function(){return E}),n.d(e,"H",function(){return M}),n.d(e,"_10",function(){return A}),n.d(e,"_6",function(){return T}),n.d(e,"_14",function(){return Y}),n.d(e,"B",function(){return J}),n.d(e,"A",function(){return c}),n.d(e,"y",function(){return I}),n.d(e,"M",function(){return R}),n.d(e,"z",function(){return N}),n.d(e,"Z",function(){return L}),n.d(e,"_1",function(){return F}),n.d(e,"x",function(){return V}),n.d(e,"L",function(){return Se}),n.d(e,"T",function(){return Ze}),n.d(e,"U",function(){return en}),n.d(e,"V",function(){return Je}),n.d(e,"W",function(){return tn}),n.d(e,"h",function(){return ce}),n.d(e,"k",function(){return ae}),n.d(e,"l",function(){return ue}),n.d(e,"n",function(){return pe}),n.d(e,"p",function(){return le}),n.d(e,"o",function(){return ve}),n.d(e,"t",function(){return nn}),n.d(e,"I",function(){return je}),n.d(e,"J",function(){return _e}),n.d(e,"S",function(){return sn}),n.d(e,"_4",function(){return ln}),n.d(e,"_9",function(){return pn}),n.d(e,"u",function(){return hn}),n.d(e,"i",function(){return w}),n.d(e,"j",function(){return dn}),n.d(e,"D",function(){return Rn}),n.d(e,"E",function(){return Nn}),n.d(e,"_0",function(){return xn}),n.d(e,"_11",function(){return jn}),n.d(e,"_17",function(){return Bn}),n.d(e,"_33",function(){return wn}),n.d(e,"_20",function(){return oe}),n.d(e,"_19",function(){return ye}),n.d(e,"_23",function(){return At}),n.d(e,"_30",function(){return q}),n.d(e,"_36",function(){return G}),n.d(e,"_48",function(){return X}),n.d(e,"_34",function(){return Kt}),n.d(e,"_35",function(){return Yt}),n.d(e,"_47",function(){return on}),n.d(e,"_21",function(){return Ur}),n.d(e,"_22",function(){return Wr}),n.d(e,"_24",function(){return Gr}),n.d(e,"_25",function(){return di}),n.d(e,"_26",function(){return ms}),n.d(e,"_27",function(){return mr}),n.d(e,"_28",function(){return Ai}),n.d(e,"_29",function(){return Xr}),n.d(e,"_31",function(){return qr}),n.d(e,"_32",function(){return Br}),n.d(e,"_37",function(){return ni}),n.d(e,"_38",function(){return ei}),n.d(e,"_39",function(){return Zi}),n.d(e,"_40",function(){return _i}),n.d(e,"_42",function(){return Ti}),n.d(e,"_45",function(){return Di}),n.d(e,"_41",function(){return to}),n.d(e,"_43",function(){return eo}),n.d(e,"_44",function(){return Ji}),n.d(e,"_46",function(){return Xi}),n.d(e,"_49",function(){return ro}),n.d(e,"_50",function(){return dr}),n.d(e,"_51",function(){return so});var r=n("../../../../tslib/tslib.es6.js"),i=n("../../../../rxjs/_esm5/Observable.js"),o=n("../../../../rxjs/_esm5/observable/merge.js"),s=n("../../../../rxjs/_esm5/operator/share.js"),a=n("../../../../rxjs/_esm5/Subject.js"),c=function(){function t(t){this._desc=t,this.ngMetadataName="InjectionToken"}return t.prototype.toString=function(){return"InjectionToken "+this._desc},t}(),u="__annotations__",l="__paramaters__",p="__prop__metadata__";function d(t,e,n,r){var i=h(e);function o(t){if(this instanceof o)return i.call(this,t),this;var e=new o(t),n=function(t){return(t.hasOwnProperty(u)?t[u]:Object.defineProperty(t,u,{value:[]})[u]).push(e),t};return r&&r(n),n}return n&&(o.prototype=Object.create(n.prototype)),o.prototype.ngMetadataName=t,o.annotationCls=o,o}function h(t){return function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];if(t){var r=t.apply(void 0,e);for(var i in r)this[i]=r[i]}}}function f(t,e,n){var r=h(e);function i(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];if(this instanceof i)return r.apply(this,t),this;var n,o=new((n=i).bind.apply(n,[void 0].concat(t)));return s.annotation=o,s;function s(t,e,n){for(var r=t.hasOwnProperty(l)?t[l]:Object.defineProperty(t,l,{value:[]})[l];r.length<=n;)r.push(null);return(r[n]=r[n]||[]).push(o),t}}return n&&(i.prototype=Object.create(n.prototype)),i.prototype.ngMetadataName=t,i.annotationCls=i,i}function m(t,e,n){var r=h(e);function i(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];if(this instanceof i)return r.apply(this,t),this;var n,o=new((n=i).bind.apply(n,[void 0].concat(t)));return function(t,e){var n=t.constructor,r=n.hasOwnProperty(p)?n[p]:Object.defineProperty(n,p,{value:{}})[p];r[e]=r.hasOwnProperty(e)&&r[e]||[],r[e].unshift(o)}}return n&&(i.prototype=Object.create(n.prototype)),i.prototype.ngMetadataName=t,i.annotationCls=i,i}var g=new c("AnalyzeForEntryComponents"),v=f("Attribute",function(t){return{attributeName:t}}),y=function(){return function(){}}(),b=m("ContentChildren",function(t,e){return void 0===e&&(e={}),Object(r.a)({selector:t,first:!1,isViewQuery:!1,descendants:!1},e)},y),_=m("ContentChild",function(t,e){return void 0===e&&(e={}),Object(r.a)({selector:t,first:!0,isViewQuery:!1,descendants:!0},e)},y),j=m("ViewChildren",function(t,e){return void 0===e&&(e={}),Object(r.a)({selector:t,first:!1,isViewQuery:!0,descendants:!0},e)},y),x=m("ViewChild",function(t,e){return Object(r.a)({selector:t,first:!0,isViewQuery:!0,descendants:!0},e)},y),w={OnPush:0,Default:1};w[w.OnPush]="OnPush",w[w.Default]="Default";var O={CheckOnce:0,Checked:1,CheckAlways:2,Detached:3,Errored:4,Destroyed:5};O[O.CheckOnce]="CheckOnce",O[O.Checked]="Checked",O[O.CheckAlways]="CheckAlways",O[O.Detached]="Detached",O[O.Errored]="Errored",O[O.Destroyed]="Destroyed";var C=d("Directive",function(t){return void 0===t&&(t={}),t}),k=d("Component",function(t){return void 0===t&&(t={}),Object(r.a)({changeDetection:w.Default},t)},C),E=d("Pipe",function(t){return Object(r.a)({pure:!0},t)}),S=m("Input",function(t){return{bindingPropertyName:t}}),P=m("Output",function(t){return{bindingPropertyName:t}}),M=(m("HostBinding",function(t){return{hostPropertyName:t}}),m("HostListener",function(t,e){return{eventName:t,args:e}}),d("NgModule",function(t){return t})),A={Emulated:0,Native:1,None:2};A[A.Emulated]="Emulated",A[A.Native]="Native",A[A.None]="None";var T=function(){return function(t){this.full=t,this.major=t.split(".")[0],this.minor=t.split(".")[1],this.patch=t.split(".").slice(2).join(".")}}(),D=new T("5.1.0"),I=f("Inject",function(t){return{token:t}}),R=f("Optional"),N=d("Injectable"),L=f("Self"),F=f("SkipSelf"),V=f("Host"),z="undefined"!=typeof window&&window,B="undefined"!=typeof self&&"undefined"!=typeof WorkerGlobalScope&&self instanceof WorkerGlobalScope&&self,q=z||void 0!==t&&t||B,H=null;function U(){if(!H){var t=q.Symbol;if(t&&t.iterator)H=t.iterator;else for(var e=Object.getOwnPropertyNames(Map.prototype),n=0;n<e.length;++n){var r=e[n];"entries"!==r&&"size"!==r&&Map.prototype[r]===Map.prototype.entries&&(H=r)}}return H}function W(t){Zone.current.scheduleMicroTask("scheduleMicrotask",t)}function G(t,e){return t===e||"number"==typeof t&&"number"==typeof e&&isNaN(t)&&isNaN(e)}function X(t){if("string"==typeof t)return t;if(t instanceof Array)return"["+t.map(X).join(", ")+"]";if(null==t)return""+t;if(t.overriddenName)return""+t.overriddenName;if(t.name)return""+t.name;var e=t.toString();if(null==e)return""+e;var n=e.indexOf("\n");return-1===n?e:e.substring(0,n)}function Y(t){return t.__forward_ref__=Y,t.toString=function(){return X(this())},t}function K(t){return"function"==typeof t&&t.hasOwnProperty("__forward_ref__")&&t.__forward_ref__===Y?t():t}var $=new Object,Q=$,Z=function(){function t(){}return t.prototype.get=function(t,e){if(void 0===e&&(e=$),e===$)throw new Error("NullInjectorError: No provider for "+X(t)+"!");return e},t}(),J=function(){function t(){}return t.create=function(t,e){return new lt(t,e)},t.THROW_IF_NOT_FOUND=$,t.NULL=new Z,t}(),tt=function(t){return t},et=[],nt=tt,rt=function(){return Array.prototype.slice.call(arguments)},it={},ot=function(t){for(var e in t)if(t[e]===it)return e;throw Error("!prop")}({provide:String,useValue:it}),st="ngTempTokenPath",at=J.NULL,ct=/\n/gm,ut="ɵ",lt=function(){function t(t,e){void 0===e&&(e=at),this.parent=e;var n=this._records=new Map;n.set(J,{token:J,fn:tt,deps:et,value:this,useNew:!1}),function t(e,n){if(n)if((n=K(n))instanceof Array)for(var r=0;r<n.length;r++)t(e,n[r]);else{if("function"==typeof n)throw ft("Function/Class not supported",n);if(!n||"object"!=typeof n||!n.provide)throw ft("Unexpected provider",n);var i=K(n.provide),o=function(t){var e=function(t){var e=et,n=t.deps;if(n&&n.length){e=[];for(var r=0;r<n.length;r++){var i=6,o=K(n[r]);if(o instanceof Array)for(var s=0,a=o;s<a.length;s++){var c=a[s];c instanceof R||c==R?i|=1:c instanceof F||c==F?i&=-3:c instanceof L||c==L?i&=-5:o=c instanceof I?c.token:K(c)}e.push({token:o,options:i})}}else if(t.useExisting){var o=K(t.useExisting);e=[{token:o,options:6}]}else if(!(n||ot in t))throw ft("'deps' required",t);return e}(t),n=tt,r=et,i=!1,o=K(t.provide);if(ot in t)r=t.useValue;else if(t.useFactory)n=t.useFactory;else if(t.useExisting);else if(t.useClass)i=!0,n=K(t.useClass);else{if("function"!=typeof o)throw ft("StaticProvider does not have [useValue|useFactory|useExisting|useClass] or [provide] is not newable",t);i=!0,n=o}return{deps:e,fn:n,useNew:i,value:r}}(n);if(!0===n.multi){var s=e.get(i);if(s){if(s.fn!==rt)throw pt(i)}else e.set(i,s={token:n.provide,deps:[],useNew:!1,fn:rt,value:et});i=n,s.deps.push({token:i,options:6})}var a=e.get(i);if(a&&a.fn==rt)throw pt(i);e.set(i,o)}}(n,t)}return t.prototype.get=function(t,e){var n=this._records.get(t);try{return dt(t,n,this._records,this.parent,e)}catch(t){var r=t[st];throw t.message=ht("\n"+t.message,r),t.ngTokenPath=r,t[st]=null,t}},t.prototype.toString=function(){var t=[];return this._records.forEach(function(e,n){return t.push(X(n))}),"StaticInjector["+t.join(", ")+"]"},t}();function pt(t){return ft("Cannot mix multi providers and regular providers",t)}function dt(t,e,n,r,i){try{return function(t,e,n,r,i){var o,s;if(e){if((o=e.value)==nt)throw Error(ut+"Circular dependency");if(o===et){e.value=nt;var a=void 0,c=e.useNew,u=e.fn,l=e.deps,p=et;if(l.length){p=[];for(var d=0;d<l.length;d++){var h=l[d],f=h.options,m=2&f?n.get(h.token):void 0;p.push(dt(h.token,m,n,m||4&f?r:at,1&f?null:J.THROW_IF_NOT_FOUND))}}e.value=o=c?new((s=u).bind.apply(s,[void 0].concat(p))):u.apply(a,p)}}else o=r.get(t,i);return o}(t,e,n,r,i)}catch(n){throw n instanceof Error||(n=new Error(n)),(n[st]=n[st]||[]).unshift(t),e&&e.value==nt&&(e.value=et),n}}function ht(t,e){t=t&&"\n"===t.charAt(0)&&t.charAt(1)==ut?t.substr(2):t;var n=X(e);if(e instanceof Array)n=e.map(X).join(" -> ");else if("object"==typeof e){var r=[];for(var i in e)if(e.hasOwnProperty(i)){var o=e[i];r.push(i+":"+("string"==typeof o?JSON.stringify(o):X(o)))}n="{"+r.join(", ")+"}"}return"StaticInjectorError["+n+"]: "+t.replace(ct,"\n  ")}function ft(t,e){return new Error(ht(t,e))}var mt="ngDebugContext",gt="ngOriginalError",vt="ngErrorLogger";function yt(t){return t[mt]}function bt(t){return t[gt]}function _t(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];t.error.apply(t,e)}var jt=function(){function t(){this._console=console}return t.prototype.handleError=function(t){var e=this._findOriginalError(t),n=this._findContext(t),r=t[vt]||_t;r(this._console,"ERROR",t),e&&r(this._console,"ORIGINAL ERROR",e),n&&r(this._console,"ERROR CONTEXT",n)},t.prototype._findContext=function(t){return t?yt(t)?yt(t):this._findContext(bt(t)):null},t.prototype._findOriginalError=function(t){for(var e=bt(t);e&&bt(e);)e=bt(e);return e},t}();function xt(t){return t.length>1?" ("+function(t){for(var e=[],n=0;n<t.length;++n){if(e.indexOf(t[n])>-1)return e.push(t[n]),e;e.push(t[n])}return e}(t.slice().reverse()).map(function(t){return X(t.token)}).join(" -> ")+")":""}function wt(t,e,n,r){var i,o,s,a=[e],c=n(a),u=r?(o=c+" caused by: "+((i=r)instanceof Error?i.message:i),(s=Error(o))[gt]=i,s):Error(c);return u.addKey=Ot,u.keys=a,u.injectors=[t],u.constructResolvingMessage=n,u[gt]=r,u}function Ot(t,e){this.injectors.push(t),this.keys.push(e),this.message=this.constructResolvingMessage(this.keys)}function Ct(t,e){for(var n=[],r=0,i=e.length;r<i;r++){var o=e[r];o&&0!=o.length?n.push(o.map(X).join(" ")):n.push("?")}return Error("Cannot resolve all parameters for '"+X(t)+"'("+n.join(", ")+"). Make sure that all the parameters are decorated with Inject or have valid type annotations and that '"+X(t)+"' is decorated with Injectable.")}var kt=function(){function t(t,e){if(this.token=t,this.id=e,!t)throw new Error("Token must be defined!");this.displayName=X(this.token)}return t.get=function(t){return Et.get(K(t))},Object.defineProperty(t,"numberOfKeys",{get:function(){return Et.numberOfKeys},enumerable:!0,configurable:!0}),t}(),Et=new(function(){function t(){this._allKeys=new Map}return t.prototype.get=function(t){if(t instanceof kt)return t;if(this._allKeys.has(t))return this._allKeys.get(t);var e=new kt(t,kt.numberOfKeys);return this._allKeys.set(t,e),e},Object.defineProperty(t.prototype,"numberOfKeys",{get:function(){return this._allKeys.size},enumerable:!0,configurable:!0}),t}()),St=Function;function Pt(t){return"function"==typeof t}var Mt=/^function\s+\S+\(\)\s*{[\s\S]+\.apply\(this,\s*arguments\)/,At=function(){function t(t){this._reflect=t||q.Reflect}return t.prototype.isReflectionEnabled=function(){return!0},t.prototype.factory=function(t){return function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return new(t.bind.apply(t,[void 0].concat(e)))}},t.prototype._zipTypesAndAnnotations=function(t,e){var n;n=void 0===t?new Array(e.length):new Array(t.length);for(var r=0;r<n.length;r++)void 0===t?n[r]=[]:t[r]!=Object?n[r]=[t[r]]:n[r]=[],e&&null!=e[r]&&(n[r]=n[r].concat(e[r]));return n},t.prototype._ownParameters=function(t,e){if(Mt.exec(t.toString()))return null;if(t.parameters&&t.parameters!==e.parameters)return t.parameters;var n=t.ctorParameters;if(n&&n!==e.ctorParameters){var r="function"==typeof n?n():n,i=r.map(function(t){return t&&t.type}),o=r.map(function(t){return t&&Tt(t.decorators)});return this._zipTypesAndAnnotations(i,o)}var s=t.hasOwnProperty(l)&&t[l],a=this._reflect&&this._reflect.getOwnMetadata&&this._reflect.getOwnMetadata("design:paramtypes",t);return a||s?this._zipTypesAndAnnotations(a,s):new Array(t.length).fill(void 0)},t.prototype.parameters=function(t){if(!Pt(t))return[];var e=Dt(t),n=this._ownParameters(t,e);return n||e===Object||(n=this.parameters(e)),n||[]},t.prototype._ownAnnotations=function(t,e){if(t.annotations&&t.annotations!==e.annotations){var n=t.annotations;return"function"==typeof n&&n.annotations&&(n=n.annotations),n}return t.decorators&&t.decorators!==e.decorators?Tt(t.decorators):t.hasOwnProperty(u)?t[u]:null},t.prototype.annotations=function(t){if(!Pt(t))return[];var e=Dt(t),n=this._ownAnnotations(t,e)||[];return(e!==Object?this.annotations(e):[]).concat(n)},t.prototype._ownPropMetadata=function(t,e){if(t.propMetadata&&t.propMetadata!==e.propMetadata){var n=t.propMetadata;return"function"==typeof n&&n.propMetadata&&(n=n.propMetadata),n}if(t.propDecorators&&t.propDecorators!==e.propDecorators){var r=t.propDecorators,i={};return Object.keys(r).forEach(function(t){i[t]=Tt(r[t])}),i}return t.hasOwnProperty(p)?t[p]:null},t.prototype.propMetadata=function(t){if(!Pt(t))return{};var e=Dt(t),n={};if(e!==Object){var r=this.propMetadata(e);Object.keys(r).forEach(function(t){n[t]=r[t]})}var i=this._ownPropMetadata(t,e);return i&&Object.keys(i).forEach(function(t){var e=[];n.hasOwnProperty(t)&&e.push.apply(e,n[t]),e.push.apply(e,i[t]),n[t]=e}),n},t.prototype.hasLifecycleHook=function(t,e){return t instanceof St&&e in t.prototype},t.prototype.getter=function(t){return new Function("o","return o."+t+";")},t.prototype.setter=function(t){return new Function("o","v","return o."+t+" = v;")},t.prototype.method=function(t){return new Function("o","args","if (!o."+t+") throw new Error('\""+t+"\" is undefined');\n        return o."+t+".apply(o, args);")},t.prototype.importUri=function(t){return"object"==typeof t&&t.filePath?t.filePath:"./"+X(t)},t.prototype.resourceUri=function(t){return"./"+X(t)},t.prototype.resolveIdentifier=function(t,e,n,r){return r},t.prototype.resolveEnum=function(t,e){return t[e]},t}();function Tt(t){return t?t.map(function(t){var e=t.type.annotationCls,n=t.args?t.args:[];return new(e.bind.apply(e,[void 0].concat(n)))}):[]}function Dt(t){var e=Object.getPrototypeOf(t.prototype);return(e?e.constructor:null)||Object}var It=new(function(){function t(t){this.reflectionCapabilities=t}return t.prototype.updateCapabilities=function(t){this.reflectionCapabilities=t},t.prototype.factory=function(t){return this.reflectionCapabilities.factory(t)},t.prototype.parameters=function(t){return this.reflectionCapabilities.parameters(t)},t.prototype.annotations=function(t){return this.reflectionCapabilities.annotations(t)},t.prototype.propMetadata=function(t){return this.reflectionCapabilities.propMetadata(t)},t.prototype.hasLifecycleHook=function(t,e){return this.reflectionCapabilities.hasLifecycleHook(t,e)},t.prototype.getter=function(t){return this.reflectionCapabilities.getter(t)},t.prototype.setter=function(t){return this.reflectionCapabilities.setter(t)},t.prototype.method=function(t){return this.reflectionCapabilities.method(t)},t.prototype.importUri=function(t){return this.reflectionCapabilities.importUri(t)},t.prototype.resourceUri=function(t){return this.reflectionCapabilities.resourceUri(t)},t.prototype.resolveIdentifier=function(t,e,n,r){return this.reflectionCapabilities.resolveIdentifier(t,e,n,r)},t.prototype.resolveEnum=function(t,e){return this.reflectionCapabilities.resolveEnum(t,e)},t}())(new At),Rt=function(){function t(t,e,n){this.key=t,this.optional=e,this.visibility=n}return t.fromKey=function(e){return new t(e,!1,null)},t}(),Nt=[],Lt=function(){function t(t,e,n){this.key=t,this.resolvedFactories=e,this.multiProvider=n}return Object.defineProperty(t.prototype,"resolvedFactory",{get:function(){return this.resolvedFactories[0]},enumerable:!0,configurable:!0}),t}(),Ft=function(){return function(t,e){this.factory=t,this.dependencies=e}}();function Vt(t){var e,n;if(t.useClass){var r=K(t.useClass);e=It.factory(r),n=qt(r)}else t.useExisting?(e=function(t){return t},n=[Rt.fromKey(kt.get(t.useExisting))]):t.useFactory?(e=t.useFactory,n=function(t,e){if(e){var n=e.map(function(t){return[t]});return e.map(function(e){return Ht(t,e,n)})}return qt(t)}(t.useFactory,t.deps)):(e=function(){return t.useValue},n=Nt);return new Ft(e,n)}function zt(t){return new Lt(kt.get(t.provide),[Vt(t)],t.multi||!1)}function Bt(t){var e=function(t,e){for(var n=0;n<t.length;n++){var r=t[n],i=e.get(r.key.id);if(i){if(r.multiProvider!==i.multiProvider)throw Error("Cannot mix multi providers and regular providers, got: "+i+" "+r);if(r.multiProvider)for(var o=0;o<r.resolvedFactories.length;o++)i.resolvedFactories.push(r.resolvedFactories[o]);else e.set(r.key.id,r)}else{var s=void 0;s=r.multiProvider?new Lt(r.key,r.resolvedFactories.slice(),r.multiProvider):r,e.set(r.key.id,s)}}return e}(function t(e,n){e.forEach(function(e){if(e instanceof St)n.push({provide:e,useClass:e});else if(e&&"object"==typeof e&&void 0!==e.provide)n.push(e);else{if(!(e instanceof Array))throw Error("Invalid provider - only instances of Provider and Type are allowed, got: "+e);t(e,n)}});return n}(t,[]).map(zt),new Map);return Array.from(e.values())}function qt(t){var e=It.parameters(t);if(!e)return[];if(e.some(function(t){return null==t}))throw Ct(t,e);return e.map(function(n){return Ht(t,n,e)})}function Ht(t,e,n){var r=null,i=!1;if(!Array.isArray(e))return Ut(e instanceof I?e.token:e,i,null);for(var o=null,s=0;s<e.length;++s){var a=e[s];a instanceof St?r=a:a instanceof I?r=a.token:a instanceof R?i=!0:a instanceof L||a instanceof F?o=a:a instanceof c&&(r=a)}if(null!=(r=K(r)))return Ut(r,i,o);throw Ct(t,n)}function Ut(t,e,n){return new Rt(kt.get(t),e,n)}var Wt=new Object,Gt=function(){function t(){}return t.resolve=function(t){return Bt(t)},t.resolveAndCreate=function(e,n){var r=t.resolve(e);return t.fromResolvedProviders(r,n)},t.fromResolvedProviders=function(t,e){return new Xt(t,e)},t}(),Xt=function(){function t(t,e){this._constructionCounter=0,this._providers=t,this.parent=e||null;var n=t.length;this.keyIds=new Array(n),this.objs=new Array(n);for(var r=0;r<n;r++)this.keyIds[r]=t[r].key.id,this.objs[r]=Wt}return t.prototype.get=function(t,e){return void 0===e&&(e=Q),this._getByKey(kt.get(t),null,e)},t.prototype.resolveAndCreateChild=function(t){var e=Gt.resolve(t);return this.createChildFromResolved(e)},t.prototype.createChildFromResolved=function(e){var n=new t(e);return n.parent=this,n},t.prototype.resolveAndInstantiate=function(t){return this.instantiateResolved(Gt.resolve([t])[0])},t.prototype.instantiateResolved=function(t){return this._instantiateProvider(t)},t.prototype.getProviderAtIndex=function(t){if(t<0||t>=this._providers.length)throw Error("Index "+t+" is out-of-bounds.");return this._providers[t]},t.prototype._new=function(t){if(this._constructionCounter++>this._getMaxNumberOfObjects())throw e=this,n=t.key,wt(e,n,function(t){return"Cannot instantiate cyclic dependency!"+xt(t)});var e,n;return this._instantiateProvider(t)},t.prototype._getMaxNumberOfObjects=function(){return this.objs.length},t.prototype._instantiateProvider=function(t){if(t.multiProvider){for(var e=new Array(t.resolvedFactories.length),n=0;n<t.resolvedFactories.length;++n)e[n]=this._instantiate(t,t.resolvedFactories[n]);return e}return this._instantiate(t,t.resolvedFactories[0])},t.prototype._instantiate=function(t,e){var n,r,i,o,s,a=this,c=e.factory;try{n=e.dependencies.map(function(t){return a._getByReflectiveDependency(t)})}catch(e){throw e.addKey&&e.addKey(this,t.key),e}try{r=c.apply(void 0,n)}catch(e){throw i=this,o=e,e.stack,s=t.key,wt(i,s,function(t){var e=X(t[0].token);return o.message+": Error during instantiation of "+e+"!"+xt(t)+"."},o)}return r},t.prototype._getByReflectiveDependency=function(t){return this._getByKey(t.key,t.visibility,t.optional?null:Q)},t.prototype._getByKey=function(e,n,r){return e===t.INJECTOR_KEY?this:n instanceof L?this._getByKeySelf(e,r):this._getByKeyDefault(e,r,n)},t.prototype._getObjByKeyId=function(t){for(var e=0;e<this.keyIds.length;e++)if(this.keyIds[e]===t)return this.objs[e]===Wt&&(this.objs[e]=this._new(this._providers[e])),this.objs[e];return Wt},t.prototype._throwOrNull=function(t,e){if(e!==Q)return e;throw wt(this,t,function(t){return"No provider for "+X(t[0].token)+"!"+xt(t)})},t.prototype._getByKeySelf=function(t,e){var n=this._getObjByKeyId(t.id);return n!==Wt?n:this._throwOrNull(t,e)},t.prototype._getByKeyDefault=function(e,n,r){var i;for(i=r instanceof F?this.parent:this;i instanceof t;){var o=i,s=o._getObjByKeyId(e.id);if(s!==Wt)return s;i=o.parent}return null!==i?i.get(e.token,n):this._throwOrNull(e,n)},Object.defineProperty(t.prototype,"displayName",{get:function(){return"ReflectiveInjector(providers: ["+function(t,e){for(var n=new Array(t._providers.length),r=0;r<t._providers.length;++r)n[r]=e(t.getProviderAtIndex(r));return n}(this,function(t){return' "'+t.key.displayName+'" '}).join(", ")+"])"},enumerable:!0,configurable:!0}),t.prototype.toString=function(){return this.displayName},t.INJECTOR_KEY=kt.get(J),t}();function Yt(t){return!!t&&"function"==typeof t.then}function Kt(t){return!!t&&"function"==typeof t.subscribe}var $t=new c("Application Initializer"),Qt=function(){function t(t){var e=this;this.appInits=t,this.initialized=!1,this.done=!1,this.donePromise=new Promise(function(t,n){e.resolve=t,e.reject=n})}return t.prototype.runInitializers=function(){var t=this;if(!this.initialized){var e=[],n=function(){t.done=!0,t.resolve()};if(this.appInits)for(var r=0;r<this.appInits.length;r++){var i=this.appInits[r]();Yt(i)&&e.push(i)}Promise.all(e).then(function(){n()}).catch(function(e){t.reject(e)}),0===e.length&&n(),this.initialized=!0}},t.decorators=[{type:N}],t.ctorParameters=function(){return[{type:Array,decorators:[{type:I,args:[$t]},{type:R}]}]},t}(),Zt=new c("AppId");var Jt={provide:Zt,useFactory:function(){return""+te()+te()+te()},deps:[]};function te(){return String.fromCharCode(97+Math.floor(25*Math.random()))}var ee=new c("Platform Initializer"),ne=new c("Platform ID"),re=new c("appBootstrapListener"),ie=new c("Application Packages Root URL"),oe=function(){function t(){}return t.prototype.log=function(t){console.log(t)},t.prototype.warn=function(t){console.warn(t)},t.decorators=[{type:N}],t.ctorParameters=function(){return[]},t}();function se(){throw new Error("Runtime compiler is not loaded")}var ae=function(){function t(){}return t.prototype.compileModuleSync=function(t){throw se()},t.prototype.compileModuleAsync=function(t){throw se()},t.prototype.compileModuleAndAllComponentsSync=function(t){throw se()},t.prototype.compileModuleAndAllComponentsAsync=function(t){throw se()},t.prototype.clearCache=function(){},t.prototype.clearCacheFor=function(t){},t.decorators=[{type:N}],t.ctorParameters=function(){return[]},t}(),ce=new c("compilerOptions"),ue=function(){return function(){}}(),le=function(){return function(){}}(),pe=function(){return function(){}}();function de(t){var e=Error("No component factory found for "+X(t)+". Did you add it to @NgModule.entryComponents?");return e[me]=t,e}var he,fe,me="ngComponent",ge=function(){function t(){}return t.prototype.resolveComponentFactory=function(t){throw de(t)},t}(),ve=function(){function t(){}return t.NULL=new ge,t}(),ye=function(){function t(t,e,n){this._parent=e,this._ngModule=n,this._factories=new Map;for(var r=0;r<t.length;r++){var i=t[r];this._factories.set(i.componentType,i)}}return t.prototype.resolveComponentFactory=function(t){var e=this._factories.get(t);if(!e&&this._parent&&(e=this._parent.resolveComponentFactory(t)),!e)throw de(t);return new be(e,this._ngModule)},t}(),be=function(t){function e(e,n){var r=t.call(this)||this;return r.factory=e,r.ngModule=n,r}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"selector",{get:function(){return this.factory.selector},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"componentType",{get:function(){return this.factory.componentType},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngContentSelectors",{get:function(){return this.factory.ngContentSelectors},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"inputs",{get:function(){return this.factory.inputs},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"outputs",{get:function(){return this.factory.outputs},enumerable:!0,configurable:!0}),e.prototype.create=function(t,e,n,r){return this.factory.create(t,e,n,r||this.ngModule)},e}(pe),_e=function(){return function(){}}(),je=function(){return function(){}}();var xe,we=!(!(xe=q.wtf)||!(he=xe.trace)||(fe=he.events,0));function Oe(t,e){return null}var Ce=we?function(t,e){return void 0===e&&(e=null),fe.createScope(t,e)}:function(t,e){return Oe},ke=we?function(t,e){return he.leaveScope(t,e),e}:function(t,e){return e},Ee=function(t){function e(e){void 0===e&&(e=!1);var n=t.call(this)||this;return n.__isAsync=e,n}return Object(r.b)(e,t),e.prototype.emit=function(e){t.prototype.next.call(this,e)},e.prototype.subscribe=function(e,n,r){var i,o=function(t){return null},s=function(){return null};return e&&"object"==typeof e?(i=this.__isAsync?function(t){setTimeout(function(){return e.next(t)})}:function(t){e.next(t)},e.error&&(o=this.__isAsync?function(t){setTimeout(function(){return e.error(t)})}:function(t){e.error(t)}),e.complete&&(s=this.__isAsync?function(){setTimeout(function(){return e.complete()})}:function(){e.complete()})):(i=this.__isAsync?function(t){setTimeout(function(){return e(t)})}:function(t){e(t)},n&&(o=this.__isAsync?function(t){setTimeout(function(){return n(t)})}:function(t){n(t)}),r&&(s=this.__isAsync?function(){setTimeout(function(){return r()})}:function(){r()})),t.prototype.subscribe.call(this,i,o,s)},e}(a.a),Se=function(){function t(t){var e=t.enableLongStackTrace,n=void 0!==e&&e;if(this.hasPendingMicrotasks=!1,this.hasPendingMacrotasks=!1,this.isStable=!0,this.onUnstable=new Ee(!1),this.onMicrotaskEmpty=new Ee(!1),this.onStable=new Ee(!1),this.onError=new Ee(!1),"undefined"==typeof Zone)throw new Error("In this configuration Angular requires Zone.js");Zone.assertZonePatched();var r;this._nesting=0,this._outer=this._inner=Zone.current,Zone.wtfZoneSpec&&(this._inner=this._inner.fork(Zone.wtfZoneSpec)),n&&Zone.longStackTraceZoneSpec&&(this._inner=this._inner.fork(Zone.longStackTraceZoneSpec)),(r=this)._inner=r._inner.fork({name:"angular",properties:{isAngularZone:!0},onInvokeTask:function(t,e,n,i,o,s){try{return Te(r),t.invokeTask(n,i,o,s)}finally{De(r)}},onInvoke:function(t,e,n,i,o,s,a){try{return Te(r),t.invoke(n,i,o,s,a)}finally{De(r)}},onHasTask:function(t,e,n,i){t.hasTask(n,i),e===n&&("microTask"==i.change?(r.hasPendingMicrotasks=i.microTask,Ae(r)):"macroTask"==i.change&&(r.hasPendingMacrotasks=i.macroTask))},onHandleError:function(t,e,n,i){return t.handleError(n,i),r.runOutsideAngular(function(){return r.onError.emit(i)}),!1}})}return t.isInAngularZone=function(){return!0===Zone.current.get("isAngularZone")},t.assertInAngularZone=function(){if(!t.isInAngularZone())throw new Error("Expected to be in Angular Zone, but it is not!")},t.assertNotInAngularZone=function(){if(t.isInAngularZone())throw new Error("Expected to not be in Angular Zone, but it is!")},t.prototype.run=function(t,e,n){return this._inner.run(t,e,n)},t.prototype.runTask=function(t,e,n,r){var i=this._inner,o=i.scheduleEventTask("NgZoneEvent: "+r,t,Me,Pe,Pe);try{return i.runTask(o,e,n)}finally{i.cancelTask(o)}},t.prototype.runGuarded=function(t,e,n){return this._inner.runGuarded(t,e,n)},t.prototype.runOutsideAngular=function(t){return this._outer.run(t)},t}();function Pe(){}var Me={};function Ae(t){if(0==t._nesting&&!t.hasPendingMicrotasks&&!t.isStable)try{t._nesting++,t.onMicrotaskEmpty.emit(null)}finally{if(t._nesting--,!t.hasPendingMicrotasks)try{t.runOutsideAngular(function(){return t.onStable.emit(null)})}finally{t.isStable=!0}}}function Te(t){t._nesting++,t.isStable&&(t.isStable=!1,t.onUnstable.emit(null))}function De(t){t._nesting--,Ae(t)}var Ie=function(){function t(){this.hasPendingMicrotasks=!1,this.hasPendingMacrotasks=!1,this.isStable=!0,this.onUnstable=new Ee,this.onMicrotaskEmpty=new Ee,this.onStable=new Ee,this.onError=new Ee}return t.prototype.run=function(t){return t()},t.prototype.runGuarded=function(t){return t()},t.prototype.runOutsideAngular=function(t){return t()},t.prototype.runTask=function(t){return t()},t}(),Re=function(){function t(t){this._ngZone=t,this._pendingCount=0,this._isZoneStable=!0,this._didWork=!1,this._callbacks=[],this._watchAngularEvents()}return t.prototype._watchAngularEvents=function(){var t=this;this._ngZone.onUnstable.subscribe({next:function(){t._didWork=!0,t._isZoneStable=!1}}),this._ngZone.runOutsideAngular(function(){t._ngZone.onStable.subscribe({next:function(){Se.assertNotInAngularZone(),W(function(){t._isZoneStable=!0,t._runCallbacksIfReady()})}})})},t.prototype.increasePendingRequestCount=function(){return this._pendingCount+=1,this._didWork=!0,this._pendingCount},t.prototype.decreasePendingRequestCount=function(){if(this._pendingCount-=1,this._pendingCount<0)throw new Error("pending async requests below zero");return this._runCallbacksIfReady(),this._pendingCount},t.prototype.isStable=function(){return this._isZoneStable&&0==this._pendingCount&&!this._ngZone.hasPendingMacrotasks},t.prototype._runCallbacksIfReady=function(){var t=this;this.isStable()?W(function(){for(;0!==t._callbacks.length;)t._callbacks.pop()(t._didWork);t._didWork=!1}):this._didWork=!0},t.prototype.whenStable=function(t){this._callbacks.push(t),this._runCallbacksIfReady()},t.prototype.getPendingRequestCount=function(){return this._pendingCount},t.prototype.findProviders=function(t,e,n){return[]},t.decorators=[{type:N}],t.ctorParameters=function(){return[{type:Se}]},t}(),Ne=function(){function t(){this._applications=new Map,Ve.addToWindow(this)}return t.prototype.registerApplication=function(t,e){this._applications.set(t,e)},t.prototype.unregisterApplication=function(t){this._applications.delete(t)},t.prototype.unregisterAllApplications=function(){this._applications.clear()},t.prototype.getTestability=function(t){return this._applications.get(t)||null},t.prototype.getAllTestabilities=function(){return Array.from(this._applications.values())},t.prototype.getAllRootElements=function(){return Array.from(this._applications.keys())},t.prototype.findTestabilityInTree=function(t,e){return void 0===e&&(e=!0),Ve.findTestabilityInTree(this,t,e)},t.decorators=[{type:N}],t.ctorParameters=function(){return[]},t}();function Le(t){Ve=t}var Fe,Ve=new(function(){function t(){}return t.prototype.addToWindow=function(t){},t.prototype.findTestabilityInTree=function(t,e,n){return null},t}()),ze=!0,Be=!1,qe=new c("AllowMultipleToken");function He(){if(Be)throw new Error("Cannot enable prod mode after platform setup.");ze=!1}function Ue(){return Be=!0,ze}var We=function(){return function(t,e){this.name=t,this.token=e}}();function Ge(t,e,n){void 0===n&&(n=[]);var r=new c("Platform: "+e);return function(e){void 0===e&&(e=[]);var i=Xe();return i&&!i.injector.get(qe,!1)||(t?t(n.concat(e).concat({provide:r,useValue:!0})):function(t){if(Fe&&!Fe.destroyed&&!Fe.injector.get(qe,!1))throw new Error("There can be only one platform. Destroy the previous one to create a new one.");Fe=t.get(Ye);var e=t.get(ee,null);e&&e.forEach(function(t){return t()})}(J.create(n.concat(e).concat({provide:r,useValue:!0})))),function(t){var e=Xe();if(!e)throw new Error("No platform exists!");if(!e.injector.get(t,null))throw new Error("A platform with a different configuration has been created. Please destroy it first.");return e}(r)}}function Xe(){return Fe&&!Fe.destroyed?Fe:null}var Ye=function(){function t(t){this._injector=t,this._modules=[],this._destroyListeners=[],this._destroyed=!1}return t.prototype.bootstrapModuleFactory=function(t,e){var n=this,r=function(t){var e;e="noop"===t?new Ie:("zone.js"===t?void 0:t)||new Se({enableLongStackTrace:Ue()});return e}(e?e.ngZone:void 0);return r.run(function(){var e=J.create([{provide:Se,useValue:r}],n.injector),i=t.create(e),o=i.injector.get(jt,null);if(!o)throw new Error("No ErrorHandler. Is platform module (BrowserModule) included?");return i.onDestroy(function(){return Qe(n._modules,i)}),r.runOutsideAngular(function(){return r.onError.subscribe({next:function(t){o.handleError(t)}})}),function(t,e,n){try{var r=n();return Yt(r)?r.catch(function(n){throw e.runOutsideAngular(function(){return t.handleError(n)}),n}):r}catch(n){throw e.runOutsideAngular(function(){return t.handleError(n)}),n}}(o,r,function(){var t=i.injector.get(Qt);return t.runInitializers(),t.donePromise.then(function(){return n._moduleDoBootstrap(i),i})})})},t.prototype.bootstrapModule=function(t,e){var n=this;void 0===e&&(e=[]);var r=this.injector.get(ue),i=Ke({},e);return r.createCompiler([i]).compileModuleAsync(t).then(function(t){return n.bootstrapModuleFactory(t,i)})},t.prototype._moduleDoBootstrap=function(t){var e=t.injector.get($e);if(t._bootstrapComponents.length>0)t._bootstrapComponents.forEach(function(t){return e.bootstrap(t)});else{if(!t.instance.ngDoBootstrap)throw new Error("The module "+X(t.instance.constructor)+' was bootstrapped, but it does not declare "@NgModule.bootstrap" components nor a "ngDoBootstrap" method. Please define one of these.');t.instance.ngDoBootstrap(e)}this._modules.push(t)},t.prototype.onDestroy=function(t){this._destroyListeners.push(t)},Object.defineProperty(t.prototype,"injector",{get:function(){return this._injector},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){if(this._destroyed)throw new Error("The platform has already been destroyed!");this._modules.slice().forEach(function(t){return t.destroy()}),this._destroyListeners.forEach(function(t){return t()}),this._destroyed=!0},Object.defineProperty(t.prototype,"destroyed",{get:function(){return this._destroyed},enumerable:!0,configurable:!0}),t.decorators=[{type:N}],t.ctorParameters=function(){return[{type:J}]},t}();function Ke(t,e){return t=Array.isArray(e)?e.reduce(Ke,t):Object(r.a)({},t,e)}var $e=function(){function t(t,e,n,r,a,c){var u=this;this._zone=t,this._console=e,this._injector=n,this._exceptionHandler=r,this._componentFactoryResolver=a,this._initStatus=c,this._bootstrapListeners=[],this._views=[],this._runningTick=!1,this._enforceNoNewChanges=!1,this._stable=!0,this.componentTypes=[],this.components=[],this._enforceNoNewChanges=Ue(),this._zone.onMicrotaskEmpty.subscribe({next:function(){u._zone.run(function(){u.tick()})}});var l=new i.a(function(t){u._stable=u._zone.isStable&&!u._zone.hasPendingMacrotasks&&!u._zone.hasPendingMicrotasks,u._zone.runOutsideAngular(function(){t.next(u._stable),t.complete()})}),p=new i.a(function(t){var e;u._zone.runOutsideAngular(function(){e=u._zone.onStable.subscribe(function(){Se.assertNotInAngularZone(),W(function(){u._stable||u._zone.hasPendingMacrotasks||u._zone.hasPendingMicrotasks||(u._stable=!0,t.next(!0))})})});var n=u._zone.onUnstable.subscribe(function(){Se.assertInAngularZone(),u._stable&&(u._stable=!1,u._zone.runOutsideAngular(function(){t.next(!1)}))});return function(){e.unsubscribe(),n.unsubscribe()}});this.isStable=Object(o.a)(l,s.a.call(p))}return t.prototype.bootstrap=function(t,e){var n,r=this;if(!this._initStatus.done)throw new Error("Cannot bootstrap as there are still asynchronous initializers running. Bootstrap components in the `ngDoBootstrap` method of the root module.");n=t instanceof pe?t:this._componentFactoryResolver.resolveComponentFactory(t),this.componentTypes.push(n.componentType);var i=n instanceof be?null:this._injector.get(_e),o=e||n.selector,s=n.create(J.NULL,[],o,i);s.onDestroy(function(){r._unloadComponent(s)});var a=s.injector.get(Re,null);return a&&s.injector.get(Ne).registerApplication(s.location.nativeElement,a),this._loadComponent(s),Ue()&&this._console.log("Angular is running in the development mode. Call enableProdMode() to enable the production mode."),s},t.prototype.tick=function(){var e=this;if(this._runningTick)throw new Error("ApplicationRef.tick is called recursively");var n=t._tickScope();try{this._runningTick=!0,this._views.forEach(function(t){return t.detectChanges()}),this._enforceNoNewChanges&&this._views.forEach(function(t){return t.checkNoChanges()})}catch(t){this._zone.runOutsideAngular(function(){return e._exceptionHandler.handleError(t)})}finally{this._runningTick=!1,ke(n)}},t.prototype.attachView=function(t){var e=t;this._views.push(e),e.attachToAppRef(this)},t.prototype.detachView=function(t){var e=t;Qe(this._views,e),e.detachFromAppRef()},t.prototype._loadComponent=function(t){this.attachView(t.hostView),this.tick(),this.components.push(t),this._injector.get(re,[]).concat(this._bootstrapListeners).forEach(function(e){return e(t)})},t.prototype._unloadComponent=function(t){this.detachView(t.hostView),Qe(this.components,t)},t.prototype.ngOnDestroy=function(){this._views.slice().forEach(function(t){return t.destroy()})},Object.defineProperty(t.prototype,"viewCount",{get:function(){return this._views.length},enumerable:!0,configurable:!0}),t._tickScope=Ce("ApplicationRef#tick()"),t.decorators=[{type:N}],t.ctorParameters=function(){return[{type:Se},{type:oe},{type:J},{type:jt},{type:ve},{type:Qt}]},t}();function Qe(t,e){var n=t.indexOf(e);n>-1&&t.splice(n,1)}var Ze=function(){return function(){}}(),Je=(new c("Renderer2Interceptor"),function(){}(),function(){return function(){}}()),tn={Important:1,DashCase:2};tn[tn.Important]="Important",tn[tn.DashCase]="DashCase";var en=function(){return function(){}}(),nn=function(){return function(t){this.nativeElement=t}}(),rn=(function(){}(),new Map);function on(t,e){var n=rn.get(t);if(n)throw new Error("Duplicate module registered for "+t+" - "+n.moduleType.name+" vs "+e.moduleType.name);rn.set(t,e)}var sn=function(){function t(){this.dirty=!0,this._results=[],this.changes=new Ee}return Object.defineProperty(t.prototype,"length",{get:function(){return this._results.length},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"first",{get:function(){return this._results[0]},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"last",{get:function(){return this._results[this.length-1]},enumerable:!0,configurable:!0}),t.prototype.map=function(t){return this._results.map(t)},t.prototype.filter=function(t){return this._results.filter(t)},t.prototype.find=function(t){return this._results.find(t)},t.prototype.reduce=function(t,e){return this._results.reduce(t,e)},t.prototype.forEach=function(t){this._results.forEach(t)},t.prototype.some=function(t){return this._results.some(t)},t.prototype.toArray=function(){return this._results.slice()},t.prototype[U()]=function(){return this._results[U()]()},t.prototype.toString=function(){return this._results.toString()},t.prototype.reset=function(t){this._results=function t(e){return e.reduce(function(e,n){var r=Array.isArray(n)?t(n):n;return e.concat(r)},[])}(t),this.dirty=!1},t.prototype.notifyOnChanges=function(){this.changes.emit(this)},t.prototype.setDirty=function(){this.dirty=!0},t.prototype.destroy=function(){this.changes.complete(),this.changes.unsubscribe()},t}();var an=function(){return function(){}}(),cn={factoryPathPrefix:"",factoryPathSuffix:".ngfactory"};!function(){function t(t,e){this._compiler=t,this._config=e||cn}t.prototype.load=function(t){return this._compiler instanceof ae?this.loadFactory(t):this.loadAndCompile(t)},t.prototype.loadAndCompile=function(t){var e=this,r=t.split("#"),i=r[0],o=r[1];return void 0===o&&(o="default"),n("../../../../../src/$$_lazy_route_resource lazy recursive")(i).then(function(t){return t[o]}).then(function(t){return un(t,i,o)}).then(function(t){return e._compiler.compileModuleAsync(t)})},t.prototype.loadFactory=function(t){var e=t.split("#"),r=e[0],i=e[1],o="NgFactory";return void 0===i&&(i="default",o=""),n("../../../../../src/$$_lazy_route_resource lazy recursive")(this._config.factoryPathPrefix+r+this._config.factoryPathSuffix).then(function(t){return t[i+o]}).then(function(t){return un(t,r,i)})},t.decorators=[{type:N}],t.ctorParameters=function(){return[{type:ae},{type:an,decorators:[{type:R}]}]}}();function un(t,e,n){if(!t)throw new Error("Cannot find '"+n+"' in '"+e+"'");return t}var ln=function(){return function(){}}(),pn=function(){return function(){}}(),dn=function(){return function(){}}(),hn=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e}(function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e}(dn)),fn=function(){return function(t,e){this.name=t,this.callback=e}}(),mn=function(){function t(t,e,n){this._debugContext=n,this.nativeNode=t,e&&e instanceof gn?e.addChild(this):this.parent=null,this.listeners=[]}return Object.defineProperty(t.prototype,"injector",{get:function(){return this._debugContext.injector},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"componentInstance",{get:function(){return this._debugContext.component},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"context",{get:function(){return this._debugContext.context},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"references",{get:function(){return this._debugContext.references},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"providerTokens",{get:function(){return this._debugContext.providerTokens},enumerable:!0,configurable:!0}),t}(),gn=function(t){function e(e,n,r){var i=t.call(this,e,n,r)||this;return i.properties={},i.attributes={},i.classes={},i.styles={},i.childNodes=[],i.nativeElement=e,i}return Object(r.b)(e,t),e.prototype.addChild=function(t){t&&(this.childNodes.push(t),t.parent=this)},e.prototype.removeChild=function(t){var e=this.childNodes.indexOf(t);-1!==e&&(t.parent=null,this.childNodes.splice(e,1))},e.prototype.insertChildrenAfter=function(t,e){var n,r=this,i=this.childNodes.indexOf(t);-1!==i&&((n=this.childNodes).splice.apply(n,[i+1,0].concat(e)),e.forEach(function(t){t.parent&&t.parent.removeChild(t),t.parent=r}))},e.prototype.insertBefore=function(t,e){var n=this.childNodes.indexOf(t);-1===n?this.addChild(e):(e.parent&&e.parent.removeChild(e),e.parent=this,this.childNodes.splice(n,0,e))},e.prototype.query=function(t){return this.queryAll(t)[0]||null},e.prototype.queryAll=function(t){var e=[];return function t(e,n,r){e.childNodes.forEach(function(e){e instanceof gn&&(n(e)&&r.push(e),t(e,n,r))})}(this,t,e),e},e.prototype.queryAllNodes=function(t){var e=[];return function t(e,n,r){e instanceof gn&&e.childNodes.forEach(function(e){n(e)&&r.push(e),e instanceof gn&&t(e,n,r)})}(this,t,e),e},Object.defineProperty(e.prototype,"children",{get:function(){return this.childNodes.filter(function(t){return t instanceof e})},enumerable:!0,configurable:!0}),e.prototype.triggerEventHandler=function(t,e){this.listeners.forEach(function(n){n.name==t&&n.callback(e)})},e}(mn);var vn=new Map;function yn(t){return vn.get(t)||null}function bn(t){vn.set(t.nativeNode,t)}function _n(t,e){var n=wn(t),r=wn(e);return n&&r?function(t,e,n){var r=t[U()](),i=e[U()]();for(;;){var o=r.next(),s=i.next();if(o.done&&s.done)return!0;if(o.done||s.done)return!1;if(!n(o.value,s.value))return!1}}(t,e,_n):!(n||!(t&&("object"==typeof t||"function"==typeof t))||r||!(e&&("object"==typeof e||"function"==typeof e)))||G(t,e)}var jn=function(){function t(t){this.wrapped=t}return t.wrap=function(e){return new t(e)},t}(),xn=(function(){function t(){this.hasWrappedValue=!1}t.prototype.unwrap=function(t){return t instanceof jn?(this.hasWrappedValue=!0,t.wrapped):t},t.prototype.reset=function(){this.hasWrappedValue=!1}}(),function(){function t(t,e,n){this.previousValue=t,this.currentValue=e,this.firstChange=n}return t.prototype.isFirstChange=function(){return this.firstChange},t}());function wn(t){return!!On(t)&&(Array.isArray(t)||!(t instanceof Map)&&U()in t)}function On(t){return null!==t&&("function"==typeof t||"object"==typeof t)}var Cn=function(){function t(){}return t.prototype.supports=function(t){return wn(t)},t.prototype.create=function(t){return new En(t)},t}(),kn=function(t,e){return e},En=function(){function t(t){this.length=0,this._linkedRecords=null,this._unlinkedRecords=null,this._previousItHead=null,this._itHead=null,this._itTail=null,this._additionsHead=null,this._additionsTail=null,this._movesHead=null,this._movesTail=null,this._removalsHead=null,this._removalsTail=null,this._identityChangesHead=null,this._identityChangesTail=null,this._trackByFn=t||kn}return t.prototype.forEachItem=function(t){var e;for(e=this._itHead;null!==e;e=e._next)t(e)},t.prototype.forEachOperation=function(t){for(var e=this._itHead,n=this._removalsHead,r=0,i=null;e||n;){var o=!n||e&&e.currentIndex<An(n,r,i)?e:n,s=An(o,r,i),a=o.currentIndex;if(o===n)r--,n=n._nextRemoved;else if(e=e._next,null==o.previousIndex)r++;else{i||(i=[]);var c=s-r,u=a-r;if(c!=u){for(var l=0;l<c;l++){var p=l<i.length?i[l]:i[l]=0,d=p+l;u<=d&&d<c&&(i[l]=p+1)}i[o.previousIndex]=u-c}}s!==a&&t(o,s,a)}},t.prototype.forEachPreviousItem=function(t){var e;for(e=this._previousItHead;null!==e;e=e._nextPrevious)t(e)},t.prototype.forEachAddedItem=function(t){var e;for(e=this._additionsHead;null!==e;e=e._nextAdded)t(e)},t.prototype.forEachMovedItem=function(t){var e;for(e=this._movesHead;null!==e;e=e._nextMoved)t(e)},t.prototype.forEachRemovedItem=function(t){var e;for(e=this._removalsHead;null!==e;e=e._nextRemoved)t(e)},t.prototype.forEachIdentityChange=function(t){var e;for(e=this._identityChangesHead;null!==e;e=e._nextIdentityChange)t(e)},t.prototype.diff=function(t){if(null==t&&(t=[]),!wn(t))throw new Error("Error trying to diff '"+X(t)+"'. Only arrays and iterables are allowed");return this.check(t)?this:null},t.prototype.onDestroy=function(){},t.prototype.check=function(t){var e=this;this._reset();var n,r,i,o=this._itHead,s=!1;if(Array.isArray(t)){this.length=t.length;for(var a=0;a<this.length;a++)r=t[a],i=this._trackByFn(a,r),null!==o&&G(o.trackById,i)?(s&&(o=this._verifyReinsertion(o,r,i,a)),G(o.item,r)||this._addIdentityChange(o,r)):(o=this._mismatch(o,r,i,a),s=!0),o=o._next}else n=0,function(t,e){if(Array.isArray(t))for(var n=0;n<t.length;n++)e(t[n]);else for(var r=t[U()](),i=void 0;!(i=r.next()).done;)e(i.value)}(t,function(t){i=e._trackByFn(n,t),null!==o&&G(o.trackById,i)?(s&&(o=e._verifyReinsertion(o,t,i,n)),G(o.item,t)||e._addIdentityChange(o,t)):(o=e._mismatch(o,t,i,n),s=!0),o=o._next,n++}),this.length=n;return this._truncate(o),this.collection=t,this.isDirty},Object.defineProperty(t.prototype,"isDirty",{get:function(){return null!==this._additionsHead||null!==this._movesHead||null!==this._removalsHead||null!==this._identityChangesHead},enumerable:!0,configurable:!0}),t.prototype._reset=function(){if(this.isDirty){var t=void 0,e=void 0;for(t=this._previousItHead=this._itHead;null!==t;t=t._next)t._nextPrevious=t._next;for(t=this._additionsHead;null!==t;t=t._nextAdded)t.previousIndex=t.currentIndex;for(this._additionsHead=this._additionsTail=null,t=this._movesHead;null!==t;t=e)t.previousIndex=t.currentIndex,e=t._nextMoved;this._movesHead=this._movesTail=null,this._removalsHead=this._removalsTail=null,this._identityChangesHead=this._identityChangesTail=null}},t.prototype._mismatch=function(t,e,n,r){var i;return null===t?i=this._itTail:(i=t._prev,this._remove(t)),null!==(t=null===this._linkedRecords?null:this._linkedRecords.get(n,r))?(G(t.item,e)||this._addIdentityChange(t,e),this._moveAfter(t,i,r)):null!==(t=null===this._unlinkedRecords?null:this._unlinkedRecords.get(n,null))?(G(t.item,e)||this._addIdentityChange(t,e),this._reinsertAfter(t,i,r)):t=this._addAfter(new Sn(e,n),i,r),t},t.prototype._verifyReinsertion=function(t,e,n,r){var i=null===this._unlinkedRecords?null:this._unlinkedRecords.get(n,null);return null!==i?t=this._reinsertAfter(i,t._prev,r):t.currentIndex!=r&&(t.currentIndex=r,this._addToMoves(t,r)),t},t.prototype._truncate=function(t){for(;null!==t;){var e=t._next;this._addToRemovals(this._unlink(t)),t=e}null!==this._unlinkedRecords&&this._unlinkedRecords.clear(),null!==this._additionsTail&&(this._additionsTail._nextAdded=null),null!==this._movesTail&&(this._movesTail._nextMoved=null),null!==this._itTail&&(this._itTail._next=null),null!==this._removalsTail&&(this._removalsTail._nextRemoved=null),null!==this._identityChangesTail&&(this._identityChangesTail._nextIdentityChange=null)},t.prototype._reinsertAfter=function(t,e,n){null!==this._unlinkedRecords&&this._unlinkedRecords.remove(t);var r=t._prevRemoved,i=t._nextRemoved;return null===r?this._removalsHead=i:r._nextRemoved=i,null===i?this._removalsTail=r:i._prevRemoved=r,this._insertAfter(t,e,n),this._addToMoves(t,n),t},t.prototype._moveAfter=function(t,e,n){return this._unlink(t),this._insertAfter(t,e,n),this._addToMoves(t,n),t},t.prototype._addAfter=function(t,e,n){return this._insertAfter(t,e,n),null===this._additionsTail?this._additionsTail=this._additionsHead=t:this._additionsTail=this._additionsTail._nextAdded=t,t},t.prototype._insertAfter=function(t,e,n){var r=null===e?this._itHead:e._next;return t._next=r,t._prev=e,null===r?this._itTail=t:r._prev=t,null===e?this._itHead=t:e._next=t,null===this._linkedRecords&&(this._linkedRecords=new Mn),this._linkedRecords.put(t),t.currentIndex=n,t},t.prototype._remove=function(t){return this._addToRemovals(this._unlink(t))},t.prototype._unlink=function(t){null!==this._linkedRecords&&this._linkedRecords.remove(t);var e=t._prev,n=t._next;return null===e?this._itHead=n:e._next=n,null===n?this._itTail=e:n._prev=e,t},t.prototype._addToMoves=function(t,e){return t.previousIndex===e?t:(null===this._movesTail?this._movesTail=this._movesHead=t:this._movesTail=this._movesTail._nextMoved=t,t)},t.prototype._addToRemovals=function(t){return null===this._unlinkedRecords&&(this._unlinkedRecords=new Mn),this._unlinkedRecords.put(t),t.currentIndex=null,t._nextRemoved=null,null===this._removalsTail?(this._removalsTail=this._removalsHead=t,t._prevRemoved=null):(t._prevRemoved=this._removalsTail,this._removalsTail=this._removalsTail._nextRemoved=t),t},t.prototype._addIdentityChange=function(t,e){return t.item=e,null===this._identityChangesTail?this._identityChangesTail=this._identityChangesHead=t:this._identityChangesTail=this._identityChangesTail._nextIdentityChange=t,t},t}(),Sn=function(){return function(t,e){this.item=t,this.trackById=e,this.currentIndex=null,this.previousIndex=null,this._nextPrevious=null,this._prev=null,this._next=null,this._prevDup=null,this._nextDup=null,this._prevRemoved=null,this._nextRemoved=null,this._nextAdded=null,this._nextMoved=null,this._nextIdentityChange=null}}(),Pn=function(){function t(){this._head=null,this._tail=null}return t.prototype.add=function(t){null===this._head?(this._head=this._tail=t,t._nextDup=null,t._prevDup=null):(this._tail._nextDup=t,t._prevDup=this._tail,t._nextDup=null,this._tail=t)},t.prototype.get=function(t,e){var n;for(n=this._head;null!==n;n=n._nextDup)if((null===e||e<=n.currentIndex)&&G(n.trackById,t))return n;return null},t.prototype.remove=function(t){var e=t._prevDup,n=t._nextDup;return null===e?this._head=n:e._nextDup=n,null===n?this._tail=e:n._prevDup=e,null===this._head},t}(),Mn=function(){function t(){this.map=new Map}return t.prototype.put=function(t){var e=t.trackById,n=this.map.get(e);n||(n=new Pn,this.map.set(e,n)),n.add(t)},t.prototype.get=function(t,e){var n=t,r=this.map.get(n);return r?r.get(t,e):null},t.prototype.remove=function(t){var e=t.trackById;return this.map.get(e).remove(t)&&this.map.delete(e),t},Object.defineProperty(t.prototype,"isEmpty",{get:function(){return 0===this.map.size},enumerable:!0,configurable:!0}),t.prototype.clear=function(){this.map.clear()},t}();function An(t,e,n){var r=t.previousIndex;if(null===r)return r;var i=0;return n&&r<n.length&&(i=n[r]),r+e+i}var Tn=function(){function t(){}return t.prototype.supports=function(t){return t instanceof Map||On(t)},t.prototype.create=function(){return new Dn},t}(),Dn=function(){function t(){this._records=new Map,this._mapHead=null,this._appendAfter=null,this._previousMapHead=null,this._changesHead=null,this._changesTail=null,this._additionsHead=null,this._additionsTail=null,this._removalsHead=null,this._removalsTail=null}return Object.defineProperty(t.prototype,"isDirty",{get:function(){return null!==this._additionsHead||null!==this._changesHead||null!==this._removalsHead},enumerable:!0,configurable:!0}),t.prototype.forEachItem=function(t){var e;for(e=this._mapHead;null!==e;e=e._next)t(e)},t.prototype.forEachPreviousItem=function(t){var e;for(e=this._previousMapHead;null!==e;e=e._nextPrevious)t(e)},t.prototype.forEachChangedItem=function(t){var e;for(e=this._changesHead;null!==e;e=e._nextChanged)t(e)},t.prototype.forEachAddedItem=function(t){var e;for(e=this._additionsHead;null!==e;e=e._nextAdded)t(e)},t.prototype.forEachRemovedItem=function(t){var e;for(e=this._removalsHead;null!==e;e=e._nextRemoved)t(e)},t.prototype.diff=function(t){if(t){if(!(t instanceof Map||On(t)))throw new Error("Error trying to diff '"+X(t)+"'. Only maps and objects are allowed")}else t=new Map;return this.check(t)?this:null},t.prototype.onDestroy=function(){},t.prototype.check=function(t){var e=this;this._reset();var n=this._mapHead;if(this._appendAfter=null,this._forEach(t,function(t,r){if(n&&n.key===r)e._maybeAddToChanges(n,t),e._appendAfter=n,n=n._next;else{var i=e._getOrCreateRecordForKey(r,t);n=e._insertBeforeOrAppend(n,i)}}),n){n._prev&&(n._prev._next=null),this._removalsHead=n;for(var r=n;null!==r;r=r._nextRemoved)r===this._mapHead&&(this._mapHead=null),this._records.delete(r.key),r._nextRemoved=r._next,r.previousValue=r.currentValue,r.currentValue=null,r._prev=null,r._next=null}return this._changesTail&&(this._changesTail._nextChanged=null),this._additionsTail&&(this._additionsTail._nextAdded=null),this.isDirty},t.prototype._insertBeforeOrAppend=function(t,e){if(t){var n=t._prev;return e._next=t,e._prev=n,t._prev=e,n&&(n._next=e),t===this._mapHead&&(this._mapHead=e),this._appendAfter=t,t}return this._appendAfter?(this._appendAfter._next=e,e._prev=this._appendAfter):this._mapHead=e,this._appendAfter=e,null},t.prototype._getOrCreateRecordForKey=function(t,e){if(this._records.has(t)){var n=this._records.get(t);this._maybeAddToChanges(n,e);var r=n._prev,i=n._next;return r&&(r._next=i),i&&(i._prev=r),n._next=null,n._prev=null,n}var o=new In(t);return this._records.set(t,o),o.currentValue=e,this._addToAdditions(o),o},t.prototype._reset=function(){if(this.isDirty){var t=void 0;for(this._previousMapHead=this._mapHead,t=this._previousMapHead;null!==t;t=t._next)t._nextPrevious=t._next;for(t=this._changesHead;null!==t;t=t._nextChanged)t.previousValue=t.currentValue;for(t=this._additionsHead;null!=t;t=t._nextAdded)t.previousValue=t.currentValue;this._changesHead=this._changesTail=null,this._additionsHead=this._additionsTail=null,this._removalsHead=null}},t.prototype._maybeAddToChanges=function(t,e){G(e,t.currentValue)||(t.previousValue=t.currentValue,t.currentValue=e,this._addToChanges(t))},t.prototype._addToAdditions=function(t){null===this._additionsHead?this._additionsHead=this._additionsTail=t:(this._additionsTail._nextAdded=t,this._additionsTail=t)},t.prototype._addToChanges=function(t){null===this._changesHead?this._changesHead=this._changesTail=t:(this._changesTail._nextChanged=t,this._changesTail=t)},t.prototype._forEach=function(t,e){t instanceof Map?t.forEach(e):Object.keys(t).forEach(function(n){return e(t[n],n)})},t}(),In=function(){return function(t){this.key=t,this.previousValue=null,this.currentValue=null,this._nextPrevious=null,this._next=null,this._prev=null,this._nextAdded=null,this._nextRemoved=null,this._nextChanged=null}}(),Rn=function(){function t(t){this.factories=t}return t.create=function(e,n){if(null!=n){var r=n.factories.slice();return new t(e=e.concat(r))}return new t(e)},t.extend=function(e){return{provide:t,useFactory:function(n){if(!n)throw new Error("Cannot extend IterableDiffers without a parent injector");return t.create(e,n)},deps:[[t,new F,new R]]}},t.prototype.find=function(t){var e,n=this.factories.find(function(e){return e.supports(t)});if(null!=n)return n;throw new Error("Cannot find a differ supporting object '"+t+"' of type '"+((e=t).name||typeof e)+"'")},t}();var Nn=function(){function t(t){this.factories=t}return t.create=function(e,n){if(n){var r=n.factories.slice();e=e.concat(r)}return new t(e)},t.extend=function(e){return{provide:t,useFactory:function(n){if(!n)throw new Error("Cannot extend KeyValueDiffers without a parent injector");return t.create(e,n)},deps:[[t,new F,new R]]}},t.prototype.find=function(t){var e=this.factories.find(function(e){return e.supports(t)});if(e)return e;throw new Error("Cannot find a differ supporting object '"+t+"'")},t}(),Ln=[new Tn],Fn=[new Cn],Vn=new Rn(Fn),zn=new Nn(Ln),Bn=Ge(null,"core",[{provide:ne,useValue:"unknown"},{provide:Ye,deps:[J]},{provide:Ne,deps:[]},{provide:oe,deps:[]}]),qn=new c("LocaleId"),Hn=new c("Translations"),Un=new c("TranslationsFormat"),Wn={Error:0,Warning:1,Ignore:2};function Gn(){return Vn}function Xn(){return zn}function Yn(t){return t||"en-US"}Wn[Wn.Error]="Error",Wn[Wn.Warning]="Warning",Wn[Wn.Ignore]="Ignore";var Kn=function(){function t(t){}return t.decorators=[{type:M,args:[{providers:[$e,Qt,ae,Jt,{provide:Rn,useFactory:Gn},{provide:Nn,useFactory:Xn},{provide:qn,useFactory:Yn,deps:[[new I(qn),new R,new F]]}]}]}],t.ctorParameters=function(){return[{type:$e}]},t}(),$n={NONE:0,HTML:1,STYLE:2,SCRIPT:3,URL:4,RESOURCE_URL:5};$n[$n.NONE]="NONE",$n[$n.HTML]="HTML",$n[$n.STYLE]="STYLE",$n[$n.SCRIPT]="SCRIPT",$n[$n.URL]="URL",$n[$n.RESOURCE_URL]="RESOURCE_URL";var Qn=function(){return function(){}}();function Zn(t,e,n){var r=t.state,i=1792&r;return i===e?(t.state=-1793&r|n,t.initIndex=-1,!0):i===n}function Jn(t,e,n){return(1792&t.state)===e&&t.initIndex<=n&&(t.initIndex=n+1,!0)}function tr(t,e){return t.nodes[e]}function er(t,e){return t.nodes[e]}function nr(t,e){return t.nodes[e]}function rr(t,e){return t.nodes[e]}function ir(t,e){return t.nodes[e]}var or={setCurrentNode:void 0,createRootView:void 0,createEmbeddedView:void 0,createComponentView:void 0,createNgModuleRef:void 0,overrideProvider:void 0,overrideComponentView:void 0,clearOverrides:void 0,checkAndUpdateView:void 0,checkNoChangesView:void 0,destroyView:void 0,resolveDep:void 0,createDebugContext:void 0,handleEvent:void 0,updateDirectives:void 0,updateRenderer:void 0,dirtyParentQueries:void 0};function sr(t,e,n,r){var i,o,s="ExpressionChangedAfterItHasBeenCheckedError: Expression has changed after it was checked. Previous value: '"+e+"'. Current value: '"+n+"'.";return r&&(s+=" It seems like the view has been created after its parent and its children have been dirty checked. Has it been created in a change detection hook ?"),i=t,ar(o=new Error(s),i),o}function ar(t,e){t[mt]=e,t[vt]=e.logError.bind(e)}function cr(t){return new Error("ViewDestroyedError: Attempt to use a destroyed view: "+t)}var ur=function(){},lr=new Map;function pr(t){var e=lr.get(t);return e||(e=X(t)+"_"+lr.size,lr.set(t,e)),e}function dr(t,e,n,r){if(r instanceof jn){r=r.wrapped;var i=t.def.nodes[e].bindingIndex+n,o=t.oldValues[i];o instanceof jn&&(o=o.wrapped),t.oldValues[i]=new jn(o)}return r}var hr="$$undefined",fr="$$empty";function mr(t){return{id:hr,styles:t.styles,encapsulation:t.encapsulation,data:t.data}}var gr=0;function vr(t,e,n,r){var i=t.oldValues;return!(!(2&t.state)&&G(i[e.bindingIndex+n],r))}function yr(t,e,n,r){return!!vr(t,e,n,r)&&(t.oldValues[e.bindingIndex+n]=r,!0)}function br(t,e,n,r){var i=t.oldValues[e.bindingIndex+n];if(1&t.state||!_n(i,r))throw sr(or.createDebugContext(t,e.nodeIndex),i,r,0!=(1&t.state))}function _r(t){for(var e=t;e;)2&e.def.flags&&(e.state|=8),e=e.viewContainerParent||e.parent}function jr(t,e){for(var n=t;n&&n!==e;)n.state|=64,n=n.viewContainerParent||n.parent}function xr(t,e,n,r){try{return _r(33554432&t.def.nodes[e].flags?er(t,e).componentView:t),or.handleEvent(t,e,n,r)}catch(e){t.root.errorHandler.handleError(e)}}function wr(t){return t.parent?er(t.parent,t.parentNodeDef.nodeIndex):null}function Or(t){return t.parent?t.parentNodeDef.parent:null}function Cr(t,e){switch(201347067&e.flags){case 1:return er(t,e.nodeIndex).renderElement;case 2:return tr(t,e.nodeIndex).renderText}}function kr(t){return!!t.parent&&!!(32768&t.parentNodeDef.flags)}function Er(t){return 1<<t%32}function Sr(t){var e={},n=0,r={};return t&&t.forEach(function(t){var i=t[0],o=t[1];"number"==typeof i?(e[i]=o,n|=Er(i)):r[i]=o}),{matchedQueries:e,references:r,matchedQueryIds:n}}function Pr(t){return t.map(function(t){var e,n;return Array.isArray(t)?(n=t[0],e=t[1]):(n=0,e=t),{flags:n,token:e,tokenKey:pr(e)}})}function Mr(t,e,n){var r=n.renderParent;return r?0==(1&r.flags)||0==(33554432&r.flags)||r.element.componentRendererType&&r.element.componentRendererType.encapsulation===A.Native?er(t,n.renderParent.nodeIndex).renderElement:void 0:e}var Ar=new WeakMap;function Tr(t){var e=Ar.get(t);return e||((e=t(function(){return ur})).factory=t,Ar.set(t,e)),e}function Dr(t,e,n,r,i){3===e&&(n=t.renderer.parentNode(Cr(t,t.def.lastRenderRootNode))),Ir(t,e,0,t.def.nodes.length-1,n,r,i)}function Ir(t,e,n,r,i,o,s){for(var a=n;a<=r;a++){var c=t.def.nodes[a];11&c.flags&&Nr(t,c,e,i,o,s),a+=c.childCount}}function Rr(t,e,n,r,i,o){for(var s=t;s&&!kr(s);)s=s.parent;for(var a=s.parent,c=Or(s),u=c.nodeIndex+1,l=c.nodeIndex+c.childCount,p=u;p<=l;p++){var d=a.def.nodes[p];d.ngContentIndex===e&&Nr(a,d,n,r,i,o),p+=d.childCount}if(!a.parent){var h=t.root.projectableNodes[e];if(h)for(p=0;p<h.length;p++)Lr(t,h[p],n,r,i,o)}}function Nr(t,e,n,r,i,o){if(8&e.flags)Rr(t,e.ngContent.index,n,r,i,o);else{var s=Cr(t,e);if(3===n&&33554432&e.flags&&48&e.bindingFlags){if(16&e.bindingFlags&&Lr(t,s,n,r,i,o),32&e.bindingFlags)Lr(er(t,e.nodeIndex).componentView,s,n,r,i,o)}else Lr(t,s,n,r,i,o);if(16777216&e.flags)for(var a=er(t,e.nodeIndex).viewContainer._embeddedViews,c=0;c<a.length;c++)Dr(a[c],n,r,i,o);1&e.flags&&!e.element.name&&Ir(t,n,e.nodeIndex+1,e.nodeIndex+e.childCount,r,i,o)}}function Lr(t,e,n,r,i,o){var s=t.renderer;switch(n){case 1:s.appendChild(r,e);break;case 2:s.insertBefore(r,e,i);break;case 3:s.removeChild(r,e);break;case 0:o.push(e)}}var Fr=/^:([^:]+):(.+)$/;function Vr(t){if(":"===t[0]){var e=t.match(Fr);return[e[1],e[2]]}return["",t]}function zr(t){for(var e=0,n=0;n<t.length;n++)e|=t[n].flags;return e}function Br(t,e){for(var n="",r=0;r<2*t;r+=2)n=n+e[r]+Hr(e[r+1]);return n+e[2*t]}function qr(t,e,n,r,i,o,s,a,c,u,l,p,d,h,f,m,g,v,y,b){switch(t){case 1:return e+Hr(n)+r;case 2:return e+Hr(n)+r+Hr(i)+o;case 3:return e+Hr(n)+r+Hr(i)+o+Hr(s)+a;case 4:return e+Hr(n)+r+Hr(i)+o+Hr(s)+a+Hr(c)+u;case 5:return e+Hr(n)+r+Hr(i)+o+Hr(s)+a+Hr(c)+u+Hr(l)+p;case 6:return e+Hr(n)+r+Hr(i)+o+Hr(s)+a+Hr(c)+u+Hr(l)+p+Hr(d)+h;case 7:return e+Hr(n)+r+Hr(i)+o+Hr(s)+a+Hr(c)+u+Hr(l)+p+Hr(d)+h+Hr(f)+m;case 8:return e+Hr(n)+r+Hr(i)+o+Hr(s)+a+Hr(c)+u+Hr(l)+p+Hr(d)+h+Hr(f)+m+Hr(g)+v;case 9:return e+Hr(n)+r+Hr(i)+o+Hr(s)+a+Hr(c)+u+Hr(l)+p+Hr(d)+h+Hr(f)+m+Hr(g)+v+Hr(y)+b;default:throw new Error("Does not support more than 9 expressions")}}function Hr(t){return null!=t?t.toString():""}var Ur=[],Wr={};function Gr(t,e,n,r,i,o){t|=1;var s=Sr(e),a=s.matchedQueries,c=s.references;return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,flags:t,checkIndex:-1,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:a,matchedQueryIds:s.matchedQueryIds,references:c,ngContentIndex:n,childCount:r,bindings:[],bindingFlags:0,outputs:[],element:{ns:null,name:null,attrs:null,template:o?Tr(o):null,componentProvider:null,componentView:null,componentRendererType:null,publicProviders:null,allProviders:null,handleEvent:i||ur},provider:null,text:null,query:null,ngContent:null}}function Xr(t,e,n,r,i,o,s,a,c,u,l,p){void 0===s&&(s=[]),u||(u=ur);var d=Sr(n),h=d.matchedQueries,f=d.references,m=d.matchedQueryIds,g=null,v=null;o&&(g=(I=Vr(o))[0],v=I[1]),a=a||[];for(var y=new Array(a.length),b=0;b<a.length;b++){var _=a[b],j=_[0],x=_[1],w=_[2],O=Vr(x),C=O[0],k=O[1],E=void 0,S=void 0;switch(15&j){case 4:S=w;break;case 1:case 8:E=w}y[b]={flags:j,ns:C,name:k,nonMinifiedName:k,securityContext:E,suffix:S}}c=c||[];var P=new Array(c.length);for(b=0;b<c.length;b++){var M=c[b],T=M[0],D=M[1];P[b]={type:0,target:T,eventName:D,propName:null}}var I,R=(s=s||[]).map(function(t){var e=t[0],n=t[1],r=Vr(e);return[r[0],r[1],n]});return p=function(t){if(t&&t.id===hr){var e=null!=t.encapsulation&&t.encapsulation!==A.None||t.styles.length||Object.keys(t.data).length;t.id=e?"c"+gr++:fr}return t&&t.id===fr&&(t=null),t||null}(p),l&&(e|=33554432),{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:t,flags:e|=1,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:h,matchedQueryIds:m,references:f,ngContentIndex:r,childCount:i,bindings:y,bindingFlags:zr(y),outputs:P,element:{ns:g,name:v,attrs:R,template:null,componentProvider:null,componentView:l||null,componentRendererType:p,publicProviders:null,allProviders:null,handleEvent:u||ur},provider:null,text:null,query:null,ngContent:null}}function Yr(t,e,n){var r,i=n.element,o=t.root.selectorOrNode,s=t.renderer;if(t.parent||!o){r=i.name?s.createElement(i.name,i.ns):s.createComment("");var a=Mr(t,e,n);a&&s.appendChild(a,r)}else r=s.selectRootElement(o);if(i.attrs)for(var c=0;c<i.attrs.length;c++){var u=i.attrs[c],l=u[0],p=u[1],d=u[2];s.setAttribute(r,p,d,l)}return r}function Kr(t,e,n,r){for(var i=0;i<n.outputs.length;i++){var o=n.outputs[i],s=$r(t,n.nodeIndex,(l=o.target,p=o.eventName,l?l+":"+p:p)),a=o.target,c=t;"component"===o.target&&(a=null,c=e);var u=c.renderer.listen(a||r,o.eventName,s);t.disposables[n.outputIndex+i]=u}var l,p}function $r(t,e,n){return function(r){return xr(t,e,n,r)}}function Qr(t,e,n,r){if(!yr(t,e,n,r))return!1;var i,o,s,a,c,u,l,p,d,h,f=e.bindings[n],m=er(t,e.nodeIndex),g=m.renderElement,v=f.name;switch(15&f.flags){case 1:!function(t,e,n,r,i,o){var s=e.securityContext,a=s?t.root.sanitizer.sanitize(s,o):o;a=null!=a?a.toString():null;var c=t.renderer;null!=o?c.setAttribute(n,i,a,r):c.removeAttribute(n,i,r)}(t,f,g,f.ns,v,r);break;case 2:l=g,p=v,d=r,h=t.renderer,d?h.addClass(l,p):h.removeClass(l,p);break;case 4:!function(t,e,n,r,i){var o=t.root.sanitizer.sanitize($n.STYLE,i);if(null!=o){o=o.toString();var s=e.suffix;null!=s&&(o+=s)}else o=null;var a=t.renderer;null!=o?a.setStyle(n,r,o):a.removeStyle(n,r)}(t,f,g,v,r);break;case 8:var y=33554432&e.flags&&32&f.flags?m.componentView:t;i=y,o=g,s=v,a=r,c=f.securityContext,u=c?i.root.sanitizer.sanitize(c,a):a,i.renderer.setProperty(o,s,u)}return!0}var Zr=new Object,Jr=pr(J),ti=pr(_e);function ei(t,e,n,r){return n=K(n),{index:-1,deps:Pr(r),flags:t,token:e,value:n}}function ni(t){for(var e={},n=0;n<t.length;n++){var r=t[n];r.index=n,e[pr(r.token)]=r}return{factory:null,providersByKey:e,providers:t}}function ri(t,e,n){if(void 0===n&&(n=J.THROW_IF_NOT_FOUND),8&e.flags)return e.token;if(2&e.flags&&(n=null),1&e.flags)return t._parent.get(e.token,n);var r=e.tokenKey;switch(r){case Jr:case ti:return t}var i=t._def.providersByKey[r];if(i){var o=t._providers[i.index];return void 0===o&&(o=t._providers[i.index]=ii(t,i)),o===Zr?void 0:o}return t._parent.get(e.token,n)}function ii(t,e){var n;switch(201347067&e.flags){case 512:n=function(t,e,n){var r=n.length;switch(r){case 0:return new e;case 1:return new e(ri(t,n[0]));case 2:return new e(ri(t,n[0]),ri(t,n[1]));case 3:return new e(ri(t,n[0]),ri(t,n[1]),ri(t,n[2]));default:for(var i=new Array(r),o=0;o<r;o++)i[o]=ri(t,n[o]);return new(e.bind.apply(e,[void 0].concat(i)))}}(t,e.value,e.deps);break;case 1024:n=function(t,e,n){var r=n.length;switch(r){case 0:return e();case 1:return e(ri(t,n[0]));case 2:return e(ri(t,n[0]),ri(t,n[1]));case 3:return e(ri(t,n[0]),ri(t,n[1]),ri(t,n[2]));default:for(var i=Array(r),o=0;o<r;o++)i[o]=ri(t,n[o]);return e.apply(void 0,i)}}(t,e.value,e.deps);break;case 2048:n=ri(t,e.deps[0]);break;case 256:n=e.value}return void 0===n?Zr:n}function oi(t,e,n,r){var i=e.viewContainer._embeddedViews;null!==n&&void 0!==n||(n=i.length),r.viewContainerParent=t,ui(i,n,r),function(t,e){var n=wr(e);if(!n||n===t||16&e.state)return;e.state|=16;var r=n.template._projectedViews;r||(r=n.template._projectedViews=[]);r.push(e),function(t,e){if(4&e.flags)return;t.nodeFlags|=4,e.flags|=4;var n=e.parent;for(;n;)n.childFlags|=4,n=n.parent}(e.parent.def,e.parentNodeDef)}(e,r),or.dirtyParentQueries(r),ai(e,n>0?i[n-1]:null,r)}function si(t,e){var n=t.viewContainer._embeddedViews;if((null==e||e>=n.length)&&(e=n.length-1),e<0)return null;var r=n[e];return r.viewContainerParent=null,li(n,e),or.dirtyParentQueries(r),ci(r),r}function ai(t,e,n){var r=e?Cr(e,e.def.lastRenderRootNode):t.renderElement;Dr(n,2,n.renderer.parentNode(r),n.renderer.nextSibling(r),void 0)}function ci(t){Dr(t,3,null,null,void 0)}function ui(t,e,n){e>=t.length?t.push(n):t.splice(e,0,n)}function li(t,e){e>=t.length-1?t.pop():t.splice(e,1)}var pi=new Object;function di(t,e,n,r,i,o){return new hi(t,e,n,r,i,o)}var hi=function(t){function e(e,n,r,i,o,s){var a=t.call(this)||this;return a.selector=e,a.componentType=n,a._inputs=i,a._outputs=o,a.ngContentSelectors=s,a.viewDefFactory=r,a}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"inputs",{get:function(){var t=[],e=this._inputs;for(var n in e){var r=e[n];t.push({propName:n,templateName:r})}return t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"outputs",{get:function(){var t=[];for(var e in this._outputs){var n=this._outputs[e];t.push({propName:e,templateName:n})}return t},enumerable:!0,configurable:!0}),e.prototype.create=function(t,e,n,r){if(!r)throw new Error("ngModule should be provided");var i=Tr(this.viewDefFactory),o=i.nodes[0].element.componentProvider.nodeIndex,s=or.createRootView(t,e||[],n,i,r,pi),a=nr(s,o).instance;return n&&s.renderer.setAttribute(er(s,0).renderElement,"ng-version",D.full),new fi(s,new gi(s),a)},e}(pe),fi=function(t){function e(e,n,r){var i=t.call(this)||this;return i._view=e,i._viewRef=n,i._component=r,i._elDef=i._view.def.nodes[0],i.hostView=n,i.changeDetectorRef=n,i.instance=r,i}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"location",{get:function(){return new nn(er(this._view,this._elDef.nodeIndex).renderElement)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"injector",{get:function(){return new bi(this._view,this._elDef)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"componentType",{get:function(){return this._component.constructor},enumerable:!0,configurable:!0}),e.prototype.destroy=function(){this._viewRef.destroy()},e.prototype.onDestroy=function(t){this._viewRef.onDestroy(t)},e}(le);var mi=function(){function t(t,e,n){this._view=t,this._elDef=e,this._data=n,this._embeddedViews=[]}return Object.defineProperty(t.prototype,"element",{get:function(){return new nn(this._data.renderElement)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"injector",{get:function(){return new bi(this._view,this._elDef)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"parentInjector",{get:function(){for(var t=this._view,e=this._elDef.parent;!e&&t;)e=Or(t),t=t.parent;return t?new bi(t,e):new bi(this._view,null)},enumerable:!0,configurable:!0}),t.prototype.clear=function(){for(var t=this._embeddedViews.length-1;t>=0;t--){var e=si(this._data,t);or.destroyView(e)}},t.prototype.get=function(t){var e=this._embeddedViews[t];if(e){var n=new gi(e);return n.attachToViewContainerRef(this),n}return null},Object.defineProperty(t.prototype,"length",{get:function(){return this._embeddedViews.length},enumerable:!0,configurable:!0}),t.prototype.createEmbeddedView=function(t,e,n){var r=t.createEmbeddedView(e||{});return this.insert(r,n),r},t.prototype.createComponent=function(t,e,n,r,i){var o=n||this.parentInjector;i||t instanceof be||(i=o.get(_e));var s=t.create(o,r,void 0,i);return this.insert(s.hostView,e),s},t.prototype.insert=function(t,e){if(t.destroyed)throw new Error("Cannot insert a destroyed View in a ViewContainer!");var n=t,r=n._view;return oi(this._view,this._data,e,r),n.attachToViewContainerRef(this),t},t.prototype.move=function(t,e){if(t.destroyed)throw new Error("Cannot move a destroyed View in a ViewContainer!");var n,r,i,o,s,a=this._embeddedViews.indexOf(t._view);return n=this._data,r=a,i=e,o=n.viewContainer._embeddedViews,s=o[r],li(o,r),null==i&&(i=o.length),ui(o,i,s),or.dirtyParentQueries(s),ci(s),ai(n,i>0?o[i-1]:null,s),t},t.prototype.indexOf=function(t){return this._embeddedViews.indexOf(t._view)},t.prototype.remove=function(t){var e=si(this._data,t);e&&or.destroyView(e)},t.prototype.detach=function(t){var e=si(this._data,t);return e?new gi(e):null},t}();var gi=function(){function t(t){this._view=t,this._viewContainerRef=null,this._appRef=null}return Object.defineProperty(t.prototype,"rootNodes",{get:function(){return Dr(this._view,0,void 0,void 0,t=[]),t;var t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"context",{get:function(){return this._view.context},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"destroyed",{get:function(){return 0!=(128&this._view.state)},enumerable:!0,configurable:!0}),t.prototype.markForCheck=function(){_r(this._view)},t.prototype.detach=function(){this._view.state&=-5},t.prototype.detectChanges=function(){var t=this._view.root.rendererFactory;t.begin&&t.begin(),or.checkAndUpdateView(this._view),t.end&&t.end()},t.prototype.checkNoChanges=function(){or.checkNoChangesView(this._view)},t.prototype.reattach=function(){this._view.state|=4},t.prototype.onDestroy=function(t){this._view.disposables||(this._view.disposables=[]),this._view.disposables.push(t)},t.prototype.destroy=function(){this._appRef?this._appRef.detachView(this):this._viewContainerRef&&this._viewContainerRef.detach(this._viewContainerRef.indexOf(this)),or.destroyView(this._view)},t.prototype.detachFromAppRef=function(){this._appRef=null,ci(this._view),or.dirtyParentQueries(this._view)},t.prototype.attachToAppRef=function(t){if(this._viewContainerRef)throw new Error("This view is already attached to a ViewContainer!");this._appRef=t},t.prototype.attachToViewContainerRef=function(t){if(this._appRef)throw new Error("This view is already attached directly to the ApplicationRef!");this._viewContainerRef=t},t}();var vi=function(t){function e(e,n){var r=t.call(this)||this;return r._parentView=e,r._def=n,r}return Object(r.b)(e,t),e.prototype.createEmbeddedView=function(t){return new gi(or.createEmbeddedView(this._parentView,this._def,this._def.element.template,t))},Object.defineProperty(e.prototype,"elementRef",{get:function(){return new nn(er(this._parentView,this._def.nodeIndex).renderElement)},enumerable:!0,configurable:!0}),e}(ln);function yi(t,e){return new bi(t,e)}var bi=function(){function t(t,e){this.view=t,this.elDef=e}return t.prototype.get=function(t,e){void 0===e&&(e=J.THROW_IF_NOT_FOUND);var n=!!this.elDef&&0!=(33554432&this.elDef.flags);return or.resolveDep(this.view,this.elDef,n,{flags:0,token:t,tokenKey:pr(t)},e)},t}();function _i(t,e){var n=t.def.nodes[e];if(1&n.flags){var r=er(t,n.nodeIndex);return n.element.template?r.template:r.renderElement}if(2&n.flags)return tr(t,n.nodeIndex).renderText;if(20240&n.flags)return nr(t,n.nodeIndex).instance;throw new Error("Illegal state: read nodeValue for node index "+e)}var ji=function(){function t(t){this.delegate=t}return t.prototype.selectRootElement=function(t){return this.delegate.selectRootElement(t)},t.prototype.createElement=function(t,e){var n=Vr(e),r=n[0],i=n[1],o=this.delegate.createElement(i,r);return t&&this.delegate.appendChild(t,o),o},t.prototype.createViewRoot=function(t){return t},t.prototype.createTemplateAnchor=function(t){var e=this.delegate.createComment("");return t&&this.delegate.appendChild(t,e),e},t.prototype.createText=function(t,e){var n=this.delegate.createText(e);return t&&this.delegate.appendChild(t,n),n},t.prototype.projectNodes=function(t,e){for(var n=0;n<e.length;n++)this.delegate.appendChild(t,e[n])},t.prototype.attachViewAfter=function(t,e){for(var n=this.delegate.parentNode(t),r=this.delegate.nextSibling(t),i=0;i<e.length;i++)this.delegate.insertBefore(n,e[i],r)},t.prototype.detachView=function(t){for(var e=0;e<t.length;e++){var n=t[e],r=this.delegate.parentNode(n);this.delegate.removeChild(r,n)}},t.prototype.destroyView=function(t,e){for(var n=0;n<e.length;n++)this.delegate.destroyNode(e[n])},t.prototype.listen=function(t,e,n){return this.delegate.listen(t,e,n)},t.prototype.listenGlobal=function(t,e,n){return this.delegate.listen(t,e,n)},t.prototype.setElementProperty=function(t,e,n){this.delegate.setProperty(t,e,n)},t.prototype.setElementAttribute=function(t,e,n){var r=Vr(e),i=r[0],o=r[1];null!=n?this.delegate.setAttribute(t,o,n,i):this.delegate.removeAttribute(t,o,i)},t.prototype.setBindingDebugInfo=function(t,e,n){},t.prototype.setElementClass=function(t,e,n){n?this.delegate.addClass(t,e):this.delegate.removeClass(t,e)},t.prototype.setElementStyle=function(t,e,n){null!=n?this.delegate.setStyle(t,e,n):this.delegate.removeStyle(t,e)},t.prototype.invokeElementMethod=function(t,e,n){t[e].apply(t,n)},t.prototype.setText=function(t,e){this.delegate.setValue(t,e)},t.prototype.animate=function(){throw new Error("Renderer.animate is no longer supported!")},t}();function xi(t,e,n,r){return new wi(t,e,n,r)}var wi=function(){function t(t,e,n,r){this._moduleType=t,this._parent=e,this._bootstrapComponents=n,this._def=r,this._destroyListeners=[],this._destroyed=!1,function(t){for(var e=t._def,n=t._providers=new Array(e.providers.length),r=0;r<e.providers.length;r++){var i=e.providers[r];4096&i.flags||(n[r]=ii(t,i))}}(this)}return t.prototype.get=function(t,e){return void 0===e&&(e=J.THROW_IF_NOT_FOUND),ri(this,{token:t,tokenKey:pr(t),flags:0},e)},Object.defineProperty(t.prototype,"instance",{get:function(){return this.get(this._moduleType)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"componentFactoryResolver",{get:function(){return this.get(ve)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"injector",{get:function(){return this},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){if(this._destroyed)throw new Error("The ng module "+X(this.instance.constructor)+" has already been destroyed.");this._destroyed=!0,function(t,e){for(var n=t._def,r=0;r<n.providers.length;r++)if(131072&n.providers[r].flags){var i=t._providers[r];i&&i!==Zr&&i.ngOnDestroy()}}(this),this._destroyListeners.forEach(function(t){return t()})},t.prototype.onDestroy=function(t){this._destroyListeners.push(t)},t}(),Oi=pr(Ze),Ci=pr(en),ki=pr(nn),Ei=pr(pn),Si=pr(ln),Pi=pr(dn),Mi=pr(J);function Ai(t,e,n,r,i,o,s,a){var c=[];if(s)for(var u in s){var l=s[u],p=l[0],d=l[1];c[p]={flags:8,name:u,nonMinifiedName:d,ns:null,securityContext:null,suffix:null}}var h=[];if(a)for(var f in a)h.push({type:1,propName:f,target:null,eventName:a[f]});return Ii(t,e|=16384,n,r,i,i,o,c,h)}function Ti(t,e,n){return Ii(-1,t|=16,null,0,e,e,n)}function Di(t,e,n,r,i){return Ii(-1,t,e,0,n,r,i)}function Ii(t,e,n,r,i,o,s,a,c){var u=Sr(n),l=u.matchedQueries,p=u.references,d=u.matchedQueryIds;c||(c=[]),a||(a=[]),o=K(o);var h=Pr(s);return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:t,flags:e,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:l,matchedQueryIds:d,references:p,ngContentIndex:-1,childCount:r,bindings:a,bindingFlags:zr(a),outputs:c,element:null,provider:{token:i,value:o,deps:h},text:null,query:null,ngContent:null}}function Ri(t,e){for(var n=t;n.parent&&!kr(n);)n=n.parent;return Vi(n.parent,Or(n),!0,e.provider.value,e.provider.deps)}function Ni(t,e){var n=(32768&e.flags)>0,r=Vi(t,e.parent,n,e.provider.value,e.provider.deps);if(e.outputs.length)for(var i=0;i<e.outputs.length;i++){var o=e.outputs[i],s=r[o.propName].subscribe(Li(t,e.parent.nodeIndex,o.eventName));t.disposables[e.outputIndex+i]=s.unsubscribe.bind(s)}return r}function Li(t,e,n){return function(r){return xr(t,e,n,r)}}function Fi(t,e){var n=(8192&e.flags)>0,r=e.provider;switch(201347067&e.flags){case 512:return Vi(t,e.parent,n,r.value,r.deps);case 1024:return function(t,e,n,r,i){var o=i.length;switch(o){case 0:return r();case 1:return r(Bi(t,e,n,i[0]));case 2:return r(Bi(t,e,n,i[0]),Bi(t,e,n,i[1]));case 3:return r(Bi(t,e,n,i[0]),Bi(t,e,n,i[1]),Bi(t,e,n,i[2]));default:for(var s=Array(o),a=0;a<o;a++)s[a]=Bi(t,e,n,i[a]);return r.apply(void 0,s)}}(t,e.parent,n,r.value,r.deps);case 2048:return Bi(t,e.parent,n,r.deps[0]);case 256:return r.value}}function Vi(t,e,n,r,i){var o=i.length;switch(o){case 0:return new r;case 1:return new r(Bi(t,e,n,i[0]));case 2:return new r(Bi(t,e,n,i[0]),Bi(t,e,n,i[1]));case 3:return new r(Bi(t,e,n,i[0]),Bi(t,e,n,i[1]),Bi(t,e,n,i[2]));default:for(var s=new Array(o),a=0;a<o;a++)s[a]=Bi(t,e,n,i[a]);return new(r.bind.apply(r,[void 0].concat(s)))}}var zi={};function Bi(t,e,n,r,i){if(void 0===i&&(i=J.THROW_IF_NOT_FOUND),8&r.flags)return r.token;var o=t;2&r.flags&&(i=null);var s=r.tokenKey;for(s===Pi&&(n=!(!e||!e.element.componentView)),e&&1&r.flags&&(n=!1,e=e.parent);t;){if(e)switch(s){case Oi:var a=qi(t,e,n);return new ji(a.renderer);case Ci:return(a=qi(t,e,n)).renderer;case ki:return new nn(er(t,e.nodeIndex).renderElement);case Ei:return er(t,e.nodeIndex).viewContainer;case Si:if(e.element.template)return er(t,e.nodeIndex).template;break;case Pi:var c=qi(t,e,n);return new gi(c);case Mi:return yi(t,e);default:var u=(n?e.element.allProviders:e.element.publicProviders)[s];if(u){var l=nr(t,u.nodeIndex);return l||(l={instance:Fi(t,u)},t.nodes[u.nodeIndex]=l),l.instance}}n=kr(t),e=Or(t),t=t.parent}var p=o.root.injector.get(r.token,zi);return p!==zi||i===zi?p:o.root.ngModule.injector.get(r.token,i)}function qi(t,e,n){var r;if(n)r=er(t,e.nodeIndex).componentView;else for(r=t;r.parent&&!kr(r);)r=r.parent;return r}function Hi(t,e,n,r,i,o){if(32768&n.flags){var s=er(t,n.parent.nodeIndex).componentView;2&s.def.flags&&(s.state|=8)}var a=n.bindings[r].name;if(e.instance[a]=i,524288&n.flags){o=o||{};var c=t.oldValues[n.bindingIndex+r];c instanceof jn&&(c=c.wrapped),o[n.bindings[r].nonMinifiedName]=new xn(c,i,0!=(2&t.state))}return t.oldValues[n.bindingIndex+r]=i,o}function Ui(t,e){if(t.def.nodeFlags&e)for(var n=t.def.nodes,r=0,i=0;i<n.length;i++){var o=n[i],s=o.parent;for(!s&&o.flags&e&&Gi(t,i,o.flags&e,r++),0==(o.childFlags&e)&&(i+=o.childCount);s&&1&s.flags&&i===s.nodeIndex+s.childCount;)s.directChildFlags&e&&(r=Wi(t,s,e,r)),s=s.parent}}function Wi(t,e,n,r){for(var i=e.nodeIndex+1;i<=e.nodeIndex+e.childCount;i++){var o=t.def.nodes[i];o.flags&n&&Gi(t,i,o.flags&n,r++),i+=o.childCount}return r}function Gi(t,e,n,r){var i=nr(t,e);if(i){var o=i.instance;o&&(or.setCurrentNode(t,e),1048576&n&&Jn(t,512,r)&&o.ngAfterContentInit(),2097152&n&&o.ngAfterContentChecked(),4194304&n&&Jn(t,768,r)&&o.ngAfterViewInit(),8388608&n&&o.ngAfterViewChecked(),131072&n&&o.ngOnDestroy())}}function Xi(t,e,n){var r=[];for(var i in n){var o=n[i];r.push({propName:i,bindingType:o})}return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:-1,flags:t,childFlags:0,directChildFlags:0,childMatchedQueries:0,ngContentIndex:-1,matchedQueries:{},matchedQueryIds:0,references:{},childCount:0,bindings:[],bindingFlags:0,outputs:[],element:null,provider:null,text:null,query:{id:e,filterId:Er(e),bindings:r},ngContent:null}}function Yi(t){for(var e,n=t.def.nodeMatchedQueries;t.parent&&((e=t).parent&&!(32768&e.parentNodeDef.flags));){var r=t.parentNodeDef;t=t.parent;for(var i=r.nodeIndex+r.childCount,o=0;o<=i;o++){67108864&(s=t.def.nodes[o]).flags&&536870912&s.flags&&(s.query.filterId&n)===s.query.filterId&&ir(t,o).setDirty(),!(1&s.flags&&o+s.childCount<r.nodeIndex)&&67108864&s.childFlags&&536870912&s.childFlags||(o+=s.childCount)}}if(134217728&t.def.nodeFlags)for(o=0;o<t.def.nodes.length;o++){var s;134217728&(s=t.def.nodes[o]).flags&&536870912&s.flags&&ir(t,o).setDirty(),o+=s.childCount}}function Ki(t,e){var n=ir(t,e.nodeIndex);if(n.dirty){var r,i=void 0;if(67108864&e.flags){var o=e.parent.parent;i=$i(t,o.nodeIndex,o.nodeIndex+o.childCount,e.query,[]),r=nr(t,e.parent.nodeIndex).instance}else 134217728&e.flags&&(i=$i(t,0,t.def.nodes.length-1,e.query,[]),r=t.component);n.reset(i);for(var s=e.query.bindings,a=!1,c=0;c<s.length;c++){var u=s[c],l=void 0;switch(u.bindingType){case 0:l=n.first;break;case 1:l=n,a=!0}r[u.propName]=l}a&&n.notifyOnChanges()}}function $i(t,e,n,r,i){for(var o=e;o<=n;o++){var s=t.def.nodes[o],a=s.matchedQueries[r.id];if(null!=a&&i.push(Qi(t,s,a)),1&s.flags&&s.element.template&&(s.element.template.nodeMatchedQueries&r.filterId)===r.filterId){var c=er(t,o);if((s.childMatchedQueries&r.filterId)===r.filterId&&($i(t,o+1,o+s.childCount,r,i),o+=s.childCount),16777216&s.flags)for(var u=c.viewContainer._embeddedViews,l=0;l<u.length;l++){var p=u[l],d=wr(p);d&&d===c&&$i(p,0,p.def.nodes.length-1,r,i)}var h=c.template._projectedViews;if(h)for(l=0;l<h.length;l++){var f=h[l];$i(f,0,f.def.nodes.length-1,r,i)}}(s.childMatchedQueries&r.filterId)!==r.filterId&&(o+=s.childCount)}return i}function Qi(t,e,n){if(null!=n)switch(n){case 1:return er(t,e.nodeIndex).renderElement;case 0:return new nn(er(t,e.nodeIndex).renderElement);case 2:return er(t,e.nodeIndex).template;case 3:return er(t,e.nodeIndex).viewContainer;case 4:return nr(t,e.nodeIndex).instance}}function Zi(t,e){return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:-1,flags:8,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:{},matchedQueryIds:0,references:{},ngContentIndex:t,childCount:0,bindings:[],bindingFlags:0,outputs:[],element:null,provider:null,text:null,query:null,ngContent:{index:e}}}function Ji(t,e){return no(128,t,new Array(e+1))}function to(t,e){return no(32,t,new Array(e))}function eo(t,e){for(var n=Object.keys(e),r=n.length,i=new Array(r),o=0;o<r;o++){var s=n[o];i[e[s]]=s}return no(64,t,i)}function no(t,e,n){for(var r=new Array(n.length),i=0;i<n.length;i++){var o=n[i];r[i]={flags:8,name:o,ns:null,nonMinifiedName:o,securityContext:null,suffix:null}}return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:e,flags:t,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:{},matchedQueryIds:0,references:{},ngContentIndex:-1,childCount:0,bindings:r,bindingFlags:zr(r),outputs:[],element:null,provider:null,text:null,query:null,ngContent:null}}function ro(t,e,n){for(var r=new Array(n.length-1),i=1;i<n.length;i++)r[i-1]={flags:8,name:null,ns:null,nonMinifiedName:null,securityContext:null,suffix:n[i]};return{nodeIndex:-1,parent:null,renderParent:null,bindingIndex:-1,outputIndex:-1,checkIndex:t,flags:2,childFlags:0,directChildFlags:0,childMatchedQueries:0,matchedQueries:{},matchedQueryIds:0,references:{},ngContentIndex:e,childCount:0,bindings:r,bindingFlags:8,outputs:[],element:null,provider:null,text:{prefix:n[0]},query:null,ngContent:null}}function io(t,e,n){var r,i=t.renderer;r=i.createText(n.text.prefix);var o=Mr(t,e,n);return o&&i.appendChild(o,r),{renderText:r}}function oo(t,e){return(null!=t?t.toString():"")+e.suffix}function so(t,e,n,r){for(var i=0,o=0,s=0,a=0,c=0,u=null,l=null,p=!1,d=!1,h=null,f=0;f<e.length;f++){var m=e[f];if(m.nodeIndex=f,m.parent=u,m.bindingIndex=i,m.outputIndex=o,m.renderParent=l,s|=m.flags,c|=m.matchedQueryIds,m.element){var g=m.element;g.publicProviders=u?u.element.publicProviders:Object.create(null),g.allProviders=g.publicProviders,p=!1,d=!1,m.element.template&&(c|=m.element.template.nodeMatchedQueries)}if(co(u,m,e.length),i+=m.bindings.length,o+=m.outputs.length,!l&&3&m.flags&&(h=m),20224&m.flags){p||(p=!0,u.element.publicProviders=Object.create(u.element.publicProviders),u.element.allProviders=u.element.publicProviders);var v=0!=(8192&m.flags),y=0!=(32768&m.flags);!v||y?u.element.publicProviders[pr(m.provider.token)]=m:(d||(d=!0,u.element.allProviders=Object.create(u.element.publicProviders)),u.element.allProviders[pr(m.provider.token)]=m),y&&(u.element.componentProvider=m)}if(u?(u.childFlags|=m.flags,u.directChildFlags|=m.flags,u.childMatchedQueries|=m.matchedQueryIds,m.element&&m.element.template&&(u.childMatchedQueries|=m.element.template.nodeMatchedQueries)):a|=m.flags,m.childCount>0)u=m,ao(m)||(l=m);else for(;u&&f===u.nodeIndex+u.childCount;){var b=u.parent;b&&(b.childFlags|=u.childFlags,b.childMatchedQueries|=u.childMatchedQueries),l=(u=b)&&ao(u)?u.renderParent:u}}return{factory:null,nodeFlags:s,rootNodeFlags:a,nodeMatchedQueries:c,flags:t,nodes:e,updateDirectives:n||ur,updateRenderer:r||ur,handleEvent:function(t,n,r,i){return e[n].element.handleEvent(t,r,i)},bindingCount:i,outputCount:o,lastRenderRootNode:h}}function ao(t){return 0!=(1&t.flags)&&null===t.element.name}function co(t,e,n){var r=e.element&&e.element.template;if(r){if(!r.lastRenderRootNode)throw new Error("Illegal State: Embedded templates without nodes are not allowed!");if(r.lastRenderRootNode&&16777216&r.lastRenderRootNode.flags)throw new Error("Illegal State: Last root node of a template can't have embedded views, at index "+e.nodeIndex+"!")}if(20224&e.flags&&0==(1&(t?t.flags:0)))throw new Error("Illegal State: StaticProvider/Directive nodes need to be children of elements or anchors, at index "+e.nodeIndex+"!");if(e.query){if(67108864&e.flags&&(!t||0==(16384&t.flags)))throw new Error("Illegal State: Content Query nodes need to be children of directives, at index "+e.nodeIndex+"!");if(134217728&e.flags&&t)throw new Error("Illegal State: View Query nodes have to be top level nodes, at index "+e.nodeIndex+"!")}if(e.childCount){var i=t?t.nodeIndex+t.childCount:n-1;if(e.nodeIndex<=i&&e.nodeIndex+e.childCount>i)throw new Error("Illegal State: childCount of node leads outside of parent, at index "+e.nodeIndex+"!")}}function uo(t,e,n,r){var i=ho(t.root,t.renderer,t,e,n);return fo(i,t.component,r),mo(i),i}function lo(t,e,n){var r=ho(t,t.renderer,null,null,e);return fo(r,n,n),mo(r),r}function po(t,e,n,r){var i,o=e.element.componentRendererType;return i=o?t.root.rendererFactory.createRenderer(r,o):t.root.renderer,ho(t.root,i,t,e.element.componentProvider,n)}function ho(t,e,n,r,i){var o=new Array(i.nodes.length),s=i.outputCount?new Array(i.outputCount):null;return{def:i,parent:n,viewContainerParent:null,parentNodeDef:r,context:null,component:null,nodes:o,state:13,root:t,renderer:e,oldValues:new Array(i.bindingCount),disposables:s,initIndex:-1}}function fo(t,e,n){t.component=e,t.context=n}function mo(t){var e;if(kr(t)){var n=t.parentNodeDef;e=er(t.parent,n.parent.nodeIndex).renderElement}for(var r,i,o,s,a,c=t.def,u=t.nodes,l=0;l<c.nodes.length;l++){var p=c.nodes[l];or.setCurrentNode(t,l);var d=void 0;switch(201347067&p.flags){case 1:var h=Yr(t,e,p),f=void 0;if(33554432&p.flags){var m=Tr(p.element.componentView);f=or.createComponentView(t,p,m,h)}Kr(t,f,p,h),d={renderElement:h,componentView:f,viewContainer:null,template:p.element.template?(s=t,a=p,new vi(s,a)):void 0},16777216&p.flags&&(d.viewContainer=new mi(t,p,d));break;case 2:d=io(t,e,p);break;case 512:case 1024:case 2048:case 256:if(!((d=u[l])||4096&p.flags))d={instance:Fi(t,p)};break;case 16:d={instance:Ri(t,p)};break;case 16384:if(!(d=u[l]))d={instance:Ni(t,p)};if(32768&p.flags)fo(er(t,p.parent.nodeIndex).componentView,d.instance,d.instance);break;case 32:case 64:case 128:d={value:void 0};break;case 67108864:case 134217728:d=new sn;break;case 8:void 0,(o=Mr(r=t,e,i=p))&&Rr(r,i.ngContent.index,1,o,null,void 0),d=void 0}u[l]=d}Oo(t,wo.CreateViewNodes),So(t,201326592,268435456,0)}function go(t){bo(t),or.updateDirectives(t,1),Co(t,wo.CheckNoChanges),or.updateRenderer(t,1),Oo(t,wo.CheckNoChanges),t.state&=-97}function vo(t){1&t.state?(t.state&=-2,t.state|=2):t.state&=-3,Zn(t,0,256),bo(t),or.updateDirectives(t,0),Co(t,wo.CheckAndUpdate),So(t,67108864,536870912,0);var e=Zn(t,256,512);Ui(t,2097152|(e?1048576:0)),or.updateRenderer(t,0),Oo(t,wo.CheckAndUpdate),So(t,134217728,536870912,0),Ui(t,8388608|((e=Zn(t,512,768))?4194304:0)),2&t.def.flags&&(t.state&=-9),t.state&=-97,Zn(t,768,1024)}function yo(t,e,n,r,i,o,s,a,c,u,l,p,d){return 0===n?function(t,e,n,r,i,o,s,a,c,u,l,p){switch(201347067&e.flags){case 1:return P=t,A=n,T=r,D=i,I=o,R=s,N=a,L=c,F=u,V=l,z=p,B=(M=e).bindings.length,q=!1,B>0&&Qr(P,M,0,A)&&(q=!0),B>1&&Qr(P,M,1,T)&&(q=!0),B>2&&Qr(P,M,2,D)&&(q=!0),B>3&&Qr(P,M,3,I)&&(q=!0),B>4&&Qr(P,M,4,R)&&(q=!0),B>5&&Qr(P,M,5,N)&&(q=!0),B>6&&Qr(P,M,6,L)&&(q=!0),B>7&&Qr(P,M,7,F)&&(q=!0),B>8&&Qr(P,M,8,V)&&(q=!0),B>9&&Qr(P,M,9,z)&&(q=!0),q;case 2:return function(t,e,n,r,i,o,s,a,c,u,l,p){var d=!1,h=e.bindings,f=h.length;if(f>0&&yr(t,e,0,n)&&(d=!0),f>1&&yr(t,e,1,r)&&(d=!0),f>2&&yr(t,e,2,i)&&(d=!0),f>3&&yr(t,e,3,o)&&(d=!0),f>4&&yr(t,e,4,s)&&(d=!0),f>5&&yr(t,e,5,a)&&(d=!0),f>6&&yr(t,e,6,c)&&(d=!0),f>7&&yr(t,e,7,u)&&(d=!0),f>8&&yr(t,e,8,l)&&(d=!0),f>9&&yr(t,e,9,p)&&(d=!0),d){var m=e.text.prefix;f>0&&(m+=oo(n,h[0])),f>1&&(m+=oo(r,h[1])),f>2&&(m+=oo(i,h[2])),f>3&&(m+=oo(o,h[3])),f>4&&(m+=oo(s,h[4])),f>5&&(m+=oo(a,h[5])),f>6&&(m+=oo(c,h[6])),f>7&&(m+=oo(u,h[7])),f>8&&(m+=oo(l,h[8])),f>9&&(m+=oo(p,h[9]));var g=tr(t,e.nodeIndex).renderText;t.renderer.setValue(g,m)}return d}(t,e,n,r,i,o,s,a,c,u,l,p);case 16384:return f=n,m=r,g=i,v=o,y=s,b=a,_=c,j=u,x=l,w=p,O=nr(d=t,(h=e).nodeIndex),C=O.instance,k=!1,E=void 0,(S=h.bindings.length)>0&&vr(d,h,0,f)&&(k=!0,E=Hi(d,O,h,0,f,E)),S>1&&vr(d,h,1,m)&&(k=!0,E=Hi(d,O,h,1,m,E)),S>2&&vr(d,h,2,g)&&(k=!0,E=Hi(d,O,h,2,g,E)),S>3&&vr(d,h,3,v)&&(k=!0,E=Hi(d,O,h,3,v,E)),S>4&&vr(d,h,4,y)&&(k=!0,E=Hi(d,O,h,4,y,E)),S>5&&vr(d,h,5,b)&&(k=!0,E=Hi(d,O,h,5,b,E)),S>6&&vr(d,h,6,_)&&(k=!0,E=Hi(d,O,h,6,_,E)),S>7&&vr(d,h,7,j)&&(k=!0,E=Hi(d,O,h,7,j,E)),S>8&&vr(d,h,8,x)&&(k=!0,E=Hi(d,O,h,8,x,E)),S>9&&vr(d,h,9,w)&&(k=!0,E=Hi(d,O,h,9,w,E)),E&&C.ngOnChanges(E),65536&h.flags&&Jn(d,256,h.nodeIndex)&&C.ngOnInit(),262144&h.flags&&C.ngDoCheck(),k;case 32:case 64:case 128:return function(t,e,n,r,i,o,s,a,c,u,l,p){var d=e.bindings,h=!1,f=d.length;if(f>0&&yr(t,e,0,n)&&(h=!0),f>1&&yr(t,e,1,r)&&(h=!0),f>2&&yr(t,e,2,i)&&(h=!0),f>3&&yr(t,e,3,o)&&(h=!0),f>4&&yr(t,e,4,s)&&(h=!0),f>5&&yr(t,e,5,a)&&(h=!0),f>6&&yr(t,e,6,c)&&(h=!0),f>7&&yr(t,e,7,u)&&(h=!0),f>8&&yr(t,e,8,l)&&(h=!0),f>9&&yr(t,e,9,p)&&(h=!0),h){var m=rr(t,e.nodeIndex),g=void 0;switch(201347067&e.flags){case 32:g=new Array(d.length),f>0&&(g[0]=n),f>1&&(g[1]=r),f>2&&(g[2]=i),f>3&&(g[3]=o),f>4&&(g[4]=s),f>5&&(g[5]=a),f>6&&(g[6]=c),f>7&&(g[7]=u),f>8&&(g[8]=l),f>9&&(g[9]=p);break;case 64:g={},f>0&&(g[d[0].name]=n),f>1&&(g[d[1].name]=r),f>2&&(g[d[2].name]=i),f>3&&(g[d[3].name]=o),f>4&&(g[d[4].name]=s),f>5&&(g[d[5].name]=a),f>6&&(g[d[6].name]=c),f>7&&(g[d[7].name]=u),f>8&&(g[d[8].name]=l),f>9&&(g[d[9].name]=p);break;case 128:var v=n;switch(f){case 1:g=v.transform(n);break;case 2:g=v.transform(r);break;case 3:g=v.transform(r,i);break;case 4:g=v.transform(r,i,o);break;case 5:g=v.transform(r,i,o,s);break;case 6:g=v.transform(r,i,o,s,a);break;case 7:g=v.transform(r,i,o,s,a,c);break;case 8:g=v.transform(r,i,o,s,a,c,u);break;case 9:g=v.transform(r,i,o,s,a,c,u,l);break;case 10:g=v.transform(r,i,o,s,a,c,u,l,p)}}m.value=g}return h}(t,e,n,r,i,o,s,a,c,u,l,p);default:throw"unreachable"}var d,h,f,m,g,v,y,b,_,j,x,w,O,C,k,E,S;var P,M,A,T,D,I,R,N,L,F,V,z,B,q}(t,e,r,i,o,s,a,c,u,l,p,d):function(t,e,n){switch(201347067&e.flags){case 1:return function(t,e,n){for(var r=!1,i=0;i<n.length;i++)Qr(t,e,i,n[i])&&(r=!0);return r}(t,e,n);case 2:return function(t,e,n){for(var r=e.bindings,i=!1,o=0;o<n.length;o++)yr(t,e,o,n[o])&&(i=!0);if(i){var s="";for(o=0;o<n.length;o++)s+=oo(n[o],r[o]);s=e.text.prefix+s;var a=tr(t,e.nodeIndex).renderText;t.renderer.setValue(a,s)}return i}(t,e,n);case 16384:return function(t,e,n){for(var r=nr(t,e.nodeIndex),i=r.instance,o=!1,s=void 0,a=0;a<n.length;a++)vr(t,e,a,n[a])&&(o=!0,s=Hi(t,r,e,a,n[a],s));return s&&i.ngOnChanges(s),65536&e.flags&&Jn(t,256,e.nodeIndex)&&i.ngOnInit(),262144&e.flags&&i.ngDoCheck(),o}(t,e,n);case 32:case 64:case 128:return function(t,e,n){for(var r=e.bindings,i=!1,o=0;o<n.length;o++)yr(t,e,o,n[o])&&(i=!0);if(i){var s=rr(t,e.nodeIndex),a=void 0;switch(201347067&e.flags){case 32:a=n;break;case 64:for(a={},o=0;o<n.length;o++)a[r[o].name]=n[o];break;case 128:var c=n[0],u=n.slice(1);a=c.transform.apply(c,u)}s.value=a}return i}(t,e,n);default:throw"unreachable"}}(t,e,r)}function bo(t){var e=t.def;if(4&e.nodeFlags)for(var n=0;n<e.nodes.length;n++){var r=e.nodes[n];if(4&r.flags){var i=er(t,n).template._projectedViews;if(i)for(var o=0;o<i.length;o++){var s=i[o];s.state|=32,jr(s,t)}}else 0==(4&r.childFlags)&&(n+=r.childCount)}}function _o(t,e,n,r,i,o,s,a,c,u,l,p,d){return 0===n?function(t,e,n,r,i,o,s,a,c,u,l,p){var d=e.bindings.length;d>0&&br(t,e,0,n);d>1&&br(t,e,1,r);d>2&&br(t,e,2,i);d>3&&br(t,e,3,o);d>4&&br(t,e,4,s);d>5&&br(t,e,5,a);d>6&&br(t,e,6,c);d>7&&br(t,e,7,u);d>8&&br(t,e,8,l);d>9&&br(t,e,9,p)}(t,e,r,i,o,s,a,c,u,l,p,d):function(t,e,n){for(var r=0;r<n.length;r++)br(t,e,r,n[r])}(t,e,r),!1}function jo(t,e){if(ir(t,e.nodeIndex).dirty)throw sr(or.createDebugContext(t,e.nodeIndex),"Query "+e.query.id+" not dirty","Query "+e.query.id+" dirty",0!=(1&t.state))}function xo(t){if(!(128&t.state)){if(Co(t,wo.Destroy),Oo(t,wo.Destroy),Ui(t,131072),t.disposables)for(var e=0;e<t.disposables.length;e++)t.disposables[e]();!function(t){if(16&t.state){var e=wr(t);if(e){var n=e.template._projectedViews;n&&(li(n,n.indexOf(t)),or.dirtyParentQueries(t))}}}(t),t.renderer.destroyNode&&function(t){for(var e=t.def.nodes.length,n=0;n<e;n++){var r=t.def.nodes[n];1&r.flags?t.renderer.destroyNode(er(t,n).renderElement):2&r.flags?t.renderer.destroyNode(tr(t,n).renderText):(67108864&r.flags||134217728&r.flags)&&ir(t,n).destroy()}}(t),kr(t)&&t.renderer.destroy(),t.state|=128}}var wo={CreateViewNodes:0,CheckNoChanges:1,CheckNoChangesProjectedViews:2,CheckAndUpdate:3,CheckAndUpdateProjectedViews:4,Destroy:5};function Oo(t,e){var n=t.def;if(33554432&n.nodeFlags)for(var r=0;r<n.nodes.length;r++){var i=n.nodes[r];33554432&i.flags?ko(er(t,r).componentView,e):0==(33554432&i.childFlags)&&(r+=i.childCount)}}function Co(t,e){var n=t.def;if(16777216&n.nodeFlags)for(var r=0;r<n.nodes.length;r++){var i=n.nodes[r];if(16777216&i.flags)for(var o=er(t,r).viewContainer._embeddedViews,s=0;s<o.length;s++)ko(o[s],e);else 0==(16777216&i.childFlags)&&(r+=i.childCount)}}function ko(t,e){var n=t.state;switch(e){case wo.CheckNoChanges:0==(128&n)&&(12==(12&n)?go(t):64&n&&Eo(t,wo.CheckNoChangesProjectedViews));break;case wo.CheckNoChangesProjectedViews:0==(128&n)&&(32&n?go(t):64&n&&Eo(t,e));break;case wo.CheckAndUpdate:0==(128&n)&&(12==(12&n)?vo(t):64&n&&Eo(t,wo.CheckAndUpdateProjectedViews));break;case wo.CheckAndUpdateProjectedViews:0==(128&n)&&(32&n?vo(t):64&n&&Eo(t,e));break;case wo.Destroy:xo(t);break;case wo.CreateViewNodes:mo(t)}}function Eo(t,e){Co(t,e),Oo(t,e)}function So(t,e,n,r){if(t.def.nodeFlags&e&&t.def.nodeFlags&n)for(var i=t.def.nodes.length,o=0;o<i;o++){var s=t.def.nodes[o];if(s.flags&e&&s.flags&n)switch(or.setCurrentNode(t,s.nodeIndex),r){case 0:Ki(t,s);break;case 1:jo(t,s)}s.childFlags&e&&s.childFlags&n||(o+=s.childCount)}}wo[wo.CreateViewNodes]="CreateViewNodes",wo[wo.CheckNoChanges]="CheckNoChanges",wo[wo.CheckNoChangesProjectedViews]="CheckNoChangesProjectedViews",wo[wo.CheckAndUpdate]="CheckAndUpdate",wo[wo.CheckAndUpdateProjectedViews]="CheckAndUpdateProjectedViews",wo[wo.Destroy]="Destroy";var Po=!1;function Mo(){if(!Po){Po=!0;var t=Ue()?{setCurrentNode:Zo,createRootView:To,createEmbeddedView:Io,createComponentView:Ro,createNgModuleRef:No,overrideProvider:Vo,overrideComponentView:zo,clearOverrides:Bo,checkAndUpdateView:Wo,checkNoChangesView:Go,destroyView:Xo,createDebugContext:function(t,e){return new us(t,e)},handleEvent:Jo,updateDirectives:ts,updateRenderer:es}:{setCurrentNode:function(){},createRootView:Ao,createEmbeddedView:uo,createComponentView:po,createNgModuleRef:xi,overrideProvider:ur,overrideComponentView:ur,clearOverrides:ur,checkAndUpdateView:vo,checkNoChangesView:go,destroyView:xo,createDebugContext:function(t,e){return new us(t,e)},handleEvent:function(t,e,n,r){return t.def.handleEvent(t,e,n,r)},updateDirectives:function(t,e){return t.def.updateDirectives(0===e?Ho:Uo,t)},updateRenderer:function(t,e){return t.def.updateRenderer(0===e?Ho:Uo,t)}};or.setCurrentNode=t.setCurrentNode,or.createRootView=t.createRootView,or.createEmbeddedView=t.createEmbeddedView,or.createComponentView=t.createComponentView,or.createNgModuleRef=t.createNgModuleRef,or.overrideProvider=t.overrideProvider,or.overrideComponentView=t.overrideComponentView,or.clearOverrides=t.clearOverrides,or.checkAndUpdateView=t.checkAndUpdateView,or.checkNoChangesView=t.checkNoChangesView,or.destroyView=t.destroyView,or.resolveDep=Bi,or.createDebugContext=t.createDebugContext,or.handleEvent=t.handleEvent,or.updateDirectives=t.updateDirectives,or.updateRenderer=t.updateRenderer,or.dirtyParentQueries=Yi}}function Ao(t,e,n,r,i,o){return lo(Do(t,i,i.injector.get(Je),e,n),r,o)}function To(t,e,n,r,i,o){var s=i.injector.get(Je),a=Do(t,i,new hs(s),e,n),c=qo(r);return ps(Qo.create,lo,null,[a,c,o])}function Do(t,e,n,r,i){var o=e.injector.get(Qn),s=e.injector.get(jt);return{ngModule:e,injector:t,projectableNodes:r,selectorOrNode:i,sanitizer:o,rendererFactory:n,renderer:n.createRenderer(null,null),errorHandler:s}}function Io(t,e,n,r){var i=qo(n);return ps(Qo.create,uo,null,[t,e,i,r])}function Ro(t,e,n,r){var i=Fo.get(e.element.componentProvider.provider.token);return n=i||qo(n),ps(Qo.create,po,null,[t,e,n,r])}function No(t,e,n,r){return xi(t,e,n,function(t){var e=function(t){var e=!1,n=!1;if(0===Lo.size)return{hasOverrides:e,hasDeprecatedOverrides:n};return t.providers.forEach(function(t){var r=Lo.get(t.token);3840&t.flags&&r&&(e=!0,n=n||r.deprecatedBehavior)}),{hasOverrides:e,hasDeprecatedOverrides:n}}(t),n=e.hasOverrides,r=e.hasDeprecatedOverrides;if(!n)return t;return function(t){for(var e=0;e<t.providers.length;e++){var n=t.providers[e];r&&(n.flags|=4096);var i=Lo.get(n.token);i&&(n.flags=-3841&n.flags|i.flags,n.deps=Pr(i.deps),n.value=i.value)}}(t=t.factory(function(){return ur})),t}(r))}var Lo=new Map,Fo=new Map;function Vo(t){Lo.set(t.token,t)}function zo(t,e){var n=Tr(Tr(e.viewDefFactory).nodes[0].element.componentView);Fo.set(t,n)}function Bo(){Lo.clear(),Fo.clear()}function qo(t){if(0===Lo.size)return t;var e=function(t){for(var e=[],n=null,r=0;r<t.nodes.length;r++){var i=t.nodes[r];1&i.flags&&(n=i),n&&3840&i.flags&&Lo.has(i.provider.token)&&(e.push(n.nodeIndex),n=null)}return e}(t);if(0===e.length)return t;t=t.factory(function(){return ur});for(var n=0;n<e.length;n++)r(t,e[n]);return t;function r(t,e){for(var n=e+1;n<t.nodes.length;n++){var r=t.nodes[n];if(1&r.flags)return;if(3840&r.flags){var i=r.provider,o=Lo.get(i.token);o&&(r.flags=-3841&r.flags|o.flags,i.deps=Pr(o.deps),i.value=o.value)}}}}function Ho(t,e,n,r,i,o,s,a,c,u,l,p,d){var h=t.def.nodes[e];return yo(t,h,n,r,i,o,s,a,c,u,l,p,d),224&h.flags?rr(t,e).value:void 0}function Uo(t,e,n,r,i,o,s,a,c,u,l,p,d){var h=t.def.nodes[e];return _o(t,h,n,r,i,o,s,a,c,u,l,p,d),224&h.flags?rr(t,e).value:void 0}function Wo(t){return ps(Qo.detectChanges,vo,null,[t])}function Go(t){return ps(Qo.checkNoChanges,go,null,[t])}function Xo(t){return ps(Qo.destroy,xo,null,[t])}var Yo,Ko,$o,Qo={create:0,detectChanges:1,checkNoChanges:2,destroy:3,handleEvent:4};function Zo(t,e){Ko=t,$o=e}function Jo(t,e,n,r){return Zo(t,e),ps(Qo.handleEvent,t.def.handleEvent,null,[t,e,n,r])}function ts(t,e){if(128&t.state)throw cr(Qo[Yo]);return Zo(t,as(t,0)),t.def.updateDirectives(function(t,n,r){for(var i=[],o=3;o<arguments.length;o++)i[o-3]=arguments[o];var s=t.def.nodes[n];0===e?ns(t,s,r,i):rs(t,s,r,i);16384&s.flags&&Zo(t,as(t,n));return 224&s.flags?rr(t,s.nodeIndex).value:void 0},t)}function es(t,e){if(128&t.state)throw cr(Qo[Yo]);return Zo(t,cs(t,0)),t.def.updateRenderer(function(t,n,r){for(var i=[],o=3;o<arguments.length;o++)i[o-3]=arguments[o];var s=t.def.nodes[n];0===e?ns(t,s,r,i):rs(t,s,r,i);3&s.flags&&Zo(t,cs(t,n));return 224&s.flags?rr(t,s.nodeIndex).value:void 0},t)}function ns(t,e,n,r){if(yo.apply(void 0,[t,e,n].concat(r))){var i=1===n?r[0]:r;if(16384&e.flags){for(var o={},s=0;s<e.bindings.length;s++){var a=e.bindings[s],c=i[s];8&a.flags&&(o[is(a.nonMinifiedName)]=ss(c))}var u=e.parent,l=er(t,u.nodeIndex).renderElement;if(u.element.name)for(var p in o){null!=(c=o[p])?t.renderer.setAttribute(l,p,c):t.renderer.removeAttribute(l,p)}else t.renderer.setValue(l,"bindings="+JSON.stringify(o,null,2))}}}function rs(t,e,n,r){_o.apply(void 0,[t,e,n].concat(r))}function is(t){return"ng-reflect-"+(t=t.replace(/[$@]/g,"_").replace(os,function(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];return"-"+t[1].toLowerCase()}))}Qo[Qo.create]="create",Qo[Qo.detectChanges]="detectChanges",Qo[Qo.checkNoChanges]="checkNoChanges",Qo[Qo.destroy]="destroy",Qo[Qo.handleEvent]="handleEvent";var os=/([A-Z])/g;function ss(t){try{return null!=t?t.toString().slice(0,30):t}catch(t){return"[ERROR] Exception while trying to serialize the value"}}function as(t,e){for(var n=e;n<t.def.nodes.length;n++){var r=t.def.nodes[n];if(16384&r.flags&&r.bindings&&r.bindings.length)return n}return null}function cs(t,e){for(var n=e;n<t.def.nodes.length;n++){var r=t.def.nodes[n];if(3&r.flags&&r.bindings&&r.bindings.length)return n}return null}var us=function(){function t(t,e){this.view=t,this.nodeIndex=e,null==e&&(this.nodeIndex=e=0),this.nodeDef=t.def.nodes[e];for(var n=this.nodeDef,r=t;n&&0==(1&n.flags);)n=n.parent;if(!n)for(;!n&&r;)n=Or(r),r=r.parent;this.elDef=n,this.elView=r}return Object.defineProperty(t.prototype,"elOrCompView",{get:function(){return er(this.elView,this.elDef.nodeIndex).componentView||this.view},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"injector",{get:function(){return yi(this.elView,this.elDef)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"component",{get:function(){return this.elOrCompView.component},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"context",{get:function(){return this.elOrCompView.context},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"providerTokens",{get:function(){var t=[];if(this.elDef)for(var e=this.elDef.nodeIndex+1;e<=this.elDef.nodeIndex+this.elDef.childCount;e++){var n=this.elView.def.nodes[e];20224&n.flags&&t.push(n.provider.token),e+=n.childCount}return t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"references",{get:function(){var t={};if(this.elDef){ls(this.elView,this.elDef,t);for(var e=this.elDef.nodeIndex+1;e<=this.elDef.nodeIndex+this.elDef.childCount;e++){var n=this.elView.def.nodes[e];20224&n.flags&&ls(this.elView,n,t),e+=n.childCount}}return t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"componentRenderElement",{get:function(){var t=function(t){for(;t&&!kr(t);)t=t.parent;if(t.parent)return er(t.parent,Or(t).nodeIndex);return null}(this.elOrCompView);return t?t.renderElement:void 0},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"renderNode",{get:function(){return 2&this.nodeDef.flags?Cr(this.view,this.nodeDef):Cr(this.elView,this.elDef)},enumerable:!0,configurable:!0}),t.prototype.logError=function(t){for(var e,n,r=[],i=1;i<arguments.length;i++)r[i-1]=arguments[i];2&this.nodeDef.flags?(e=this.view.def,n=this.nodeDef.nodeIndex):(e=this.elView.def,n=this.elDef.nodeIndex);var o=function(t,e){for(var n=-1,r=0;r<=e;r++){var i=t.nodes[r];3&i.flags&&n++}return n}(e,n),s=-1;e.factory(function(){return++s===o?(e=t.error).bind.apply(e,[t].concat(r)):ur;var e}),s<o&&(t.error("Illegal state: the ViewDefinitionFactory did not call the logger!"),t.error.apply(t,r))},t}();function ls(t,e,n){for(var r in e.references)n[r]=Qi(t,e,e.references[r])}function ps(t,e,n,r){var i,o,s=Yo,a=Ko,c=$o;try{Yo=t;var u=e.apply(n,r);return Ko=a,$o=c,Yo=s,u}catch(t){if(yt(t)||!Ko)throw t;throw i=t,o=ds(),i instanceof Error||(i=new Error(i.toString())),ar(i,o),i}}function ds(){return Ko?new us(Ko,$o):null}var hs=function(){function t(t){this.delegate=t}return t.prototype.createRenderer=function(t,e){return new fs(this.delegate.createRenderer(t,e))},t.prototype.begin=function(){this.delegate.begin&&this.delegate.begin()},t.prototype.end=function(){this.delegate.end&&this.delegate.end()},t.prototype.whenRenderingDone=function(){return this.delegate.whenRenderingDone?this.delegate.whenRenderingDone():Promise.resolve(null)},t}(),fs=function(){function t(t){this.delegate=t}return Object.defineProperty(t.prototype,"data",{get:function(){return this.delegate.data},enumerable:!0,configurable:!0}),t.prototype.destroyNode=function(t){var e;e=yn(t),vn.delete(e.nativeNode),this.delegate.destroyNode&&this.delegate.destroyNode(t)},t.prototype.destroy=function(){this.delegate.destroy()},t.prototype.createElement=function(t,e){var n=this.delegate.createElement(t,e),r=ds();if(r){var i=new gn(n,null,r);i.name=t,bn(i)}return n},t.prototype.createComment=function(t){var e=this.delegate.createComment(t),n=ds();return n&&bn(new mn(e,null,n)),e},t.prototype.createText=function(t){var e=this.delegate.createText(t),n=ds();return n&&bn(new mn(e,null,n)),e},t.prototype.appendChild=function(t,e){var n=yn(t),r=yn(e);n&&r&&n instanceof gn&&n.addChild(r),this.delegate.appendChild(t,e)},t.prototype.insertBefore=function(t,e,n){var r=yn(t),i=yn(e),o=yn(n);r&&i&&r instanceof gn&&r.insertBefore(o,i),this.delegate.insertBefore(t,e,n)},t.prototype.removeChild=function(t,e){var n=yn(t),r=yn(e);n&&r&&n instanceof gn&&n.removeChild(r),this.delegate.removeChild(t,e)},t.prototype.selectRootElement=function(t){var e=this.delegate.selectRootElement(t),n=ds();return n&&bn(new gn(e,null,n)),e},t.prototype.setAttribute=function(t,e,n,r){var i=yn(t);if(i&&i instanceof gn){var o=r?r+":"+e:e;i.attributes[o]=n}this.delegate.setAttribute(t,e,n,r)},t.prototype.removeAttribute=function(t,e,n){var r=yn(t);if(r&&r instanceof gn){var i=n?n+":"+e:e;r.attributes[i]=null}this.delegate.removeAttribute(t,e,n)},t.prototype.addClass=function(t,e){var n=yn(t);n&&n instanceof gn&&(n.classes[e]=!0),this.delegate.addClass(t,e)},t.prototype.removeClass=function(t,e){var n=yn(t);n&&n instanceof gn&&(n.classes[e]=!1),this.delegate.removeClass(t,e)},t.prototype.setStyle=function(t,e,n,r){var i=yn(t);i&&i instanceof gn&&(i.styles[e]=n),this.delegate.setStyle(t,e,n,r)},t.prototype.removeStyle=function(t,e,n){var r=yn(t);r&&r instanceof gn&&(r.styles[e]=null),this.delegate.removeStyle(t,e,n)},t.prototype.setProperty=function(t,e,n){var r=yn(t);r&&r instanceof gn&&(r.properties[e]=n),this.delegate.setProperty(t,e,n)},t.prototype.listen=function(t,e,n){if("string"!=typeof t){var r=yn(t);r&&r.listeners.push(new fn(e,n))}return this.delegate.listen(t,e,n)},t.prototype.parentNode=function(t){return this.delegate.parentNode(t)},t.prototype.nextSibling=function(t){return this.delegate.nextSibling(t)},t.prototype.setValue=function(t,e){return this.delegate.setValue(t,e)},t}();function ms(t,e,n){return new gs(t,e,n)}var gs=function(t){function e(e,n,r){var i=t.call(this)||this;return i.moduleType=e,i._bootstrapComponents=n,i._ngModuleDefFactory=r,i}return Object(r.b)(e,t),e.prototype.create=function(t){Mo();var e=Tr(this._ngModuleDefFactory);return or.createNgModuleRef(this.moduleType,t||J.NULL,this._bootstrapComponents,e)},e}(je)}).call(e,n("../../../../webpack/buildin/global.js"))},"../../../flex-layout/esm5/flex-layout.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return Ct});var r=n("../../../core/esm5/core.js"),i=n("../../../platform-browser/esm5/platform-browser.js"),o=n("../../../../rxjs/_esm5/operators/map.js"),s=n("../../../../tslib/tslib.es6.js"),a=n("../../../common/esm5/common.js"),c=n("../../../../rxjs/_esm5/BehaviorSubject.js"),u=n("../../../../rxjs/_esm5/operators/filter.js"),l=n("../../../../rxjs/_esm5/ReplaySubject.js"),p=(new r._6("2.0.0-beta.11-9e52109"),["row","column","row-reverse","column-reverse"]);function d(t){var e=h(t);return function(t,e){void 0===e&&(e=null);return{display:"flex","box-sizing":"border-box","flex-direction":t,"flex-wrap":e||null}}(e[0],e[1])}function h(t){var e=(t=t?t.toLowerCase():"").split(" "),n=e[0],r=e[1];return p.find(function(t){return t===n})||(n=p[0]),[n,m(r)]}function f(t){var e=h(t),n=e[0];e[1];return n.indexOf("row")>-1}function m(t){if(t)switch(t.toLowerCase()){case"reverse":case"wrap-reverse":case"reverse-wrap":t="wrap-reverse";break;case"no":case"none":case"nowrap":t="nowrap";break;default:t="wrap"}return t}function g(t){for(var e in t){var n=t[e]||"";switch(e){case"display":t.display="flex"===n?["-webkit-flex","flex"]:"inline-flex"===n?["-webkit-inline-flex","inline-flex"]:n;break;case"align-items":case"align-self":case"align-content":case"flex":case"flex-basis":case"flex-flow":case"flex-grow":case"flex-shrink":case"flex-wrap":case"justify-content":t["-webkit-"+e]=n;break;case"flex-direction":n=n||"row",t["-webkit-flex-direction"]=n,t["flex-direction"]=n;break;case"order":t.order=t["-webkit-"+e]=isNaN(n)?"0":n}}return t}function v(t,e,n){var r=g(e);n.forEach(function(e){y(r,e,t)})}function y(t,e,n){Object.keys(t).sort().forEach(function(r){for(var i=0,o=Array.isArray(t[r])?t[r]:[t[r]];i<o.length;i++){var s=o[i];n.setStyle(e,r,s)}})}function b(t,e){return Object(i.h)().getStyle(t,e)}function _(t,e,n){void 0===n&&(n=!1);var r="";if(t)try{var o=r=b(t,e);n||(r=o||Object(i.h)().getComputedStyle(t).getPropertyValue(e))}catch(t){}return r?r.trim():"block"}function j(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];if(null==t)throw TypeError("Cannot convert undefined or null to object");for(var r=0,i=e;r<i.length;r++){var o=i[r];if(null!=o)for(var s in o)o.hasOwnProperty(s)&&(t[s]=o[s])}return t}var x=function(){return function(t,e,n){this.baseKey=t,this.defaultValue=e,this.inputKeys=n}}(),w=function(){function t(t,e,n){this._options=t,this._mediaMonitor=e,this._onMediaChanges=n,this._subscribers=[],this._registryMap=this._buildRegistryMap(),this._subscribers=this._configureChangeObservers()}return Object.defineProperty(t.prototype,"registryFromLargest",{get:function(){return this._registryMap.slice().reverse()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"mediaMonitor",{get:function(){return this._mediaMonitor},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activatedInputKey",{get:function(){return this._activatedInputKey||this._options.baseKey},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activatedInput",{get:function(){var t=this.activatedInputKey;return this.hasKeyValue(t)?this._lookupKeyValue(t):this._options.defaultValue},enumerable:!0,configurable:!0}),t.prototype.hasKeyValue=function(t){return void 0!==this._options.inputKeys[t]},t.prototype.destroy=function(){this._subscribers.forEach(function(t){t.unsubscribe()}),this._subscribers=[]},t.prototype._configureChangeObservers=function(){var t=this,e=[];return this._registryMap.forEach(function(n){if(t._keyInUse(n.key)){e.push(t.mediaMonitor.observe(n.alias).pipe(Object(o.a)(function(e){return(e=e.clone()).property=t._options.baseKey,e})).subscribe(function(e){t._onMonitorEvents(e)}))}}),e},t.prototype._buildRegistryMap=function(){var t=this;return this.mediaMonitor.breakpoints.map(function(e){return j({},e,{baseKey:t._options.baseKey,key:t._options.baseKey+e.suffix})}).filter(function(e){return t._keyInUse(e.key)})},t.prototype._onMonitorEvents=function(t){t.property==this._options.baseKey&&(t.value=this._calculateActivatedValue(t),this._onMediaChanges(t))},t.prototype._keyInUse=function(t){return void 0!==this._lookupKeyValue(t)},t.prototype._calculateActivatedValue=function(t){var e=this._options.baseKey+t.suffix,n=this._activatedInputKey;return n=t.matches?e:n==e?"":n,this._activatedInputKey=this._validateInputKey(n),this.activatedInput},t.prototype._validateInputKey=function(t){var e=this,n=function(t){return!e._keyInUse(t)};return n(t)&&this.mediaMonitor.activeOverlaps.some(function(r){var i=e._options.baseKey+r.suffix;return!n(i)&&(t=i,!0)}),t},t.prototype._lookupKeyValue=function(t){return this._options.inputKeys[t]},t}(),O=function(){function t(t,e,n){this._mediaMonitor=t,this._elementRef=e,this._renderer=n,this._inputMap={},this._hasInitialized=!1}return Object.defineProperty(t.prototype,"hasMediaQueryListener",{get:function(){return!!this._mqActivation},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activatedValue",{get:function(){return this._mqActivation?this._mqActivation.activatedInput:void 0},set:function(t){var e,n="baseKey";this._mqActivation&&(n=this._mqActivation.activatedInputKey,e=this._inputMap[n],this._inputMap[n]=t);var i,o=new r._0(e,t,!1);this.ngOnChanges(((i={})[n]=o,i))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"parentElement",{get:function(){return this._elementRef.nativeElement.parentNode},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"nativeElement",{get:function(){return this._elementRef.nativeElement},enumerable:!0,configurable:!0}),t.prototype._queryInput=function(t){return this._inputMap[t]},t.prototype.ngOnInit=function(){this._display=this._getDisplayStyle(),this._hasInitialized=!0},t.prototype.ngOnChanges=function(t){throw new Error("BaseFxDirective::ngOnChanges should be overridden in subclass: "+t)},t.prototype.ngOnDestroy=function(){this._mqActivation&&this._mqActivation.destroy(),this._mediaMonitor=null},t.prototype._getDefaultVal=function(t,e){var n=this._queryInput(t);return void 0!==n&&null!==n&&""!==n?n:e},t.prototype._getDisplayStyle=function(t){return void 0===t&&(t=this.nativeElement),_(t||this.nativeElement,"display")},t.prototype._getAttributeValue=function(t,e){return void 0===e&&(e=this.nativeElement),n=e||this.nativeElement,r=t,Object(i.h)().getAttribute(n,r)||"";var n,r},t.prototype._getFlowDirection=function(t,e){void 0===e&&(e=!1);var n="row";t&&(n=_(t,"flex-direction")||"row",!b(t,"flex-direction")&&e&&v(this._renderer,d(n),[t]));return n.trim()},t.prototype._applyStyleToElement=function(t,e,n){void 0===n&&(n=this.nativeElement);var r,i,o,s,a,c=n||this.nativeElement;r=this._renderer,i=c,s=e,a={},"string"==typeof(o=t)&&(a[o]=s,o=a),y(a=g(o),i,r)},t.prototype._applyStyleToElements=function(t,e){v(this._renderer,t,e||[])},t.prototype._cacheInput=function(t,e){if("object"==typeof e)for(var n in e)this._inputMap[n]=e[n];else t&&(this._inputMap[t]=e)},t.prototype._listenForMediaQueryChanges=function(t,e,n){if(!this._mqActivation){var r=new x(t,e,this._inputMap);this._mqActivation=new w(r,this._mediaMonitor,function(t){return n(t)})}return this._mqActivation},Object.defineProperty(t.prototype,"childrenNodes",{get:function(){for(var t=this.nativeElement.children,e=[],n=t.length;n--;)e[n]=t[n];return e},enumerable:!0,configurable:!0}),t.prototype.hasResponsiveAPI=function(t){return Object.keys(this._inputMap).length-(this._inputMap[t]?1:0)>0},t.prototype.hasKeyValue=function(t){return this._mqActivation.hasKeyValue(t)},Object.defineProperty(t.prototype,"hasInitialized",{get:function(){return this._hasInitialized},enumerable:!0,configurable:!0}),t}(),C=function(t){function e(e,n,r,i){var o=t.call(this,n,r,i)||this;return o._baseKey=e,o._mediaMonitor=n,o._elementRef=r,o._renderer=i,o}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"activeKey",{get:function(){var t=this._mqActivation,e=t?t.activatedInputKey:this._baseKey;return"class"===e?"klazz":e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"inputMap",{get:function(){return this._inputMap},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"mqActivation",{get:function(){return this._mqActivation},enumerable:!0,configurable:!0}),e.prototype.hasResponsiveAPI=function(){return t.prototype.hasResponsiveAPI.call(this,this._baseKey)},e.prototype.queryInput=function(t){return t?this._queryInput(t):void 0},e.prototype.cacheInput=function(t,e,n){if(void 0===n&&(n=!1),n)this._cacheInputRaw(t,e);else if(Array.isArray(e))this._cacheInputArray(t,e);else if("object"==typeof e)this._cacheInputObject(t,e);else{if("string"!=typeof e)throw new Error("Invalid class value '"+t+"' provided. Did you want to cache the raw value?");this._cacheInputString(t,e)}},e.prototype.listenForMediaQueryChanges=function(t,e,n){return this._listenForMediaQueryChanges(t,e,n)},e.prototype._cacheInputRaw=function(t,e){this._inputMap[t]=e},e.prototype._cacheInputArray=function(t,e){void 0===t&&(t=""),this._inputMap[t]=e.join(" ")},e.prototype._cacheInputObject=function(t,e){void 0===t&&(t="");var n=[];for(var r in e)e[r]&&n.push(r);this._inputMap[t]=n.join(" ")},e.prototype._cacheInputString=function(t,e){void 0===t&&(t=""),this._inputMap[t]=e},e}(O),k=new r.A("Token (@angular/flex-layout) Breakpoints"),E=function(){function t(t){this._registry=t}return Object.defineProperty(t.prototype,"items",{get:function(){return this._registry.slice()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"sortedItems",{get:function(){var t=this._registry.filter(function(t){return!0===t.overlapping}),e=this._registry.filter(function(t){return!0!==t.overlapping});return t.concat(e)},enumerable:!0,configurable:!0}),t.prototype.findByAlias=function(t){return this._registry.find(function(e){return e.alias==t})||null},t.prototype.findByQuery=function(t){return this._registry.find(function(e){return e.mediaQuery==t})||null},Object.defineProperty(t.prototype,"overlappings",{get:function(){return this._registry.filter(function(t){return 1==t.overlapping})},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"aliases",{get:function(){return this._registry.map(function(t){return t.alias})},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"suffixes",{get:function(){return this._registry.map(function(t){return t.suffix?t.suffix:""})},enumerable:!0,configurable:!0}),t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:Array,decorators:[{type:r.y,args:[k]}]}]},t}(),S=function(){function t(t,e,n,r){void 0===t&&(t=!1),void 0===e&&(e="all"),void 0===n&&(n=""),void 0===r&&(r=""),this.matches=t,this.mediaQuery=e,this.mqAlias=n,this.suffix=r}return t.prototype.clone=function(){return new t(this.matches,this.mediaQuery,this.mqAlias,this.suffix)},t}(),P=function(){function t(t,e){this._zone=t,this._document=e,this._registry=new Map,this._source=new c.a(new S(!0)),this._observable$=this._source.asObservable()}return t.prototype.isActive=function(t){var e=this._registry.get(t);return!!e&&e.matches},t.prototype.observe=function(t){return t&&this.registerQuery(t),this._observable$.pipe(Object(u.a)(function(e){return!t||e.mediaQuery===t}))},t.prototype.registerQuery=function(t){var e,n,r=this,o=void 0===(e=t)?[]:"string"==typeof e?[e]:(n={},e.filter(function(t){return!n.hasOwnProperty(t)&&(n[t]=!0)}));o.length>0&&(!function(t,e){var n=t.filter(function(t){return!M[t]});if(n.length>0){var r=n.join(", ");try{var o=Object(i.h)().createElement("style");if(Object(i.h)().setAttribute(o,"type","text/css"),!o.styleSheet){var s="/*\n  @angular/flex-layout - workaround for possible browser quirk with mediaQuery listeners\n  see http://bit.ly/2sd4HMP\n*/\n@media "+r+" {.fx-query-test{ }}";Object(i.h)().appendChild(o,Object(i.h)().createTextNode(s))}Object(i.h)().appendChild(e.head,o),n.forEach(function(t){return M[t]=o})}catch(t){console.error(t)}}}(o,this._document),o.forEach(function(t){var e=r._registry.get(t),n=function(e){r._zone.run(function(){var n=new S(e.matches,t);r._source.next(n)})};e||((e=r._buildMQL(t)).addListener(n),r._registry.set(t,e)),e.matches&&n(e)}))},t.prototype._buildMQL=function(t){return Object(i.h)().supportsDOMEvents()&&!!window.matchMedia("all").addListener?window.matchMedia(t):{matches:"all"===t||""===t,media:t,addListener:function(){},removeListener:function(){}}},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:r.L},{type:void 0,decorators:[{type:r.y,args:[a.b]}]}]},t}();var M={};function A(t,e){return j(t,e?{mqAlias:e.alias,suffix:e.suffix}:{})}var T=function(){function t(t,e){this._breakpoints=t,this._matchMedia=e,this._registerBreakpoints()}return Object.defineProperty(t.prototype,"breakpoints",{get:function(){return this._breakpoints.items.slice()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activeOverlaps",{get:function(){var t=this;return this._breakpoints.overlappings.reverse().filter(function(e){return t._matchMedia.isActive(e.mediaQuery)})},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"active",{get:function(){var t=this,e=null;this.breakpoints.reverse().forEach(function(n){""!==n.alias&&!e&&t._matchMedia.isActive(n.mediaQuery)&&(e=n)});var n=this.breakpoints[0];return e||(this._matchMedia.isActive(n.mediaQuery)?n:null)},enumerable:!0,configurable:!0}),t.prototype.isActive=function(t){var e=this._breakpoints.findByAlias(t)||this._breakpoints.findByQuery(t);return this._matchMedia.isActive(e?e.mediaQuery:t)},t.prototype.observe=function(t){var e=this._breakpoints.findByAlias(t||"")||this._breakpoints.findByQuery(t||"");return this._matchMedia.observe(e?e.mediaQuery:t).pipe(Object(o.a)(function(t){return A(t,e)}),Object(u.a)(function(t){return!e||""!==t.mqAlias}))},t.prototype._registerBreakpoints=function(){var t=this._breakpoints.sortedItems.map(function(t){return t.mediaQuery});this._matchMedia.registerQuery(t)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:E},{type:P}]},t}(),D=function(t){function e(e,n,r){var i=t.call(this,e,n,r)||this;return i._announcer=new l.a(1),i.layout$=i._announcer.asObservable(),i}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"layout",{set:function(t){this._cacheInput("layout",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutXs",{set:function(t){this._cacheInput("layoutXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutSm",{set:function(t){this._cacheInput("layoutSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutMd",{set:function(t){this._cacheInput("layoutMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutLg",{set:function(t){this._cacheInput("layoutLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutXl",{set:function(t){this._cacheInput("layoutXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutGtXs",{set:function(t){this._cacheInput("layoutGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutGtSm",{set:function(t){this._cacheInput("layoutGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutGtMd",{set:function(t){this._cacheInput("layoutGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutGtLg",{set:function(t){this._cacheInput("layoutGtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutLtSm",{set:function(t){this._cacheInput("layoutLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutLtMd",{set:function(t){this._cacheInput("layoutLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutLtLg",{set:function(t){this._cacheInput("layoutLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"layoutLtXl",{set:function(t){this._cacheInput("layoutLtXl",t)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(null!=t.layout||this._mqActivation)&&this._updateWithDirection()},e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this._listenForMediaQueryChanges("layout","row",function(t){e._updateWithDirection(t.value)}),this._updateWithDirection()},e.prototype._updateWithDirection=function(t){t=t||this._queryInput("layout")||"row",this._mqActivation&&(t=this._mqActivation.activatedInput);var e=d(t||"");this._applyStyleToElement(e),this._announcer.next(e["flex-direction"])},e.decorators=[{type:r.s,args:[{selector:"\n  [fxLayout],\n  [fxLayout.xs], [fxLayout.sm], [fxLayout.md], [fxLayout.lg], [fxLayout.xl],\n  [fxLayout.lt-sm], [fxLayout.lt-md], [fxLayout.lt-lg], [fxLayout.lt-xl],\n  [fxLayout.gt-xs], [fxLayout.gt-sm], [fxLayout.gt-md], [fxLayout.gt-lg]\n"}]}],e.ctorParameters=function(){return[{type:T},{type:r.t},{type:r.U}]},e.propDecorators={layout:[{type:r.C,args:["fxLayout"]}],layoutXs:[{type:r.C,args:["fxLayout.xs"]}],layoutSm:[{type:r.C,args:["fxLayout.sm"]}],layoutMd:[{type:r.C,args:["fxLayout.md"]}],layoutLg:[{type:r.C,args:["fxLayout.lg"]}],layoutXl:[{type:r.C,args:["fxLayout.xl"]}],layoutGtXs:[{type:r.C,args:["fxLayout.gt-xs"]}],layoutGtSm:[{type:r.C,args:["fxLayout.gt-sm"]}],layoutGtMd:[{type:r.C,args:["fxLayout.gt-md"]}],layoutGtLg:[{type:r.C,args:["fxLayout.gt-lg"]}],layoutLtSm:[{type:r.C,args:["fxLayout.lt-sm"]}],layoutLtMd:[{type:r.C,args:["fxLayout.lt-md"]}],layoutLtLg:[{type:r.C,args:["fxLayout.lt-lg"]}],layoutLtXl:[{type:r.C,args:["fxLayout.lt-xl"]}]},e}(O),I=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r)||this;return o._layout="row",i&&(o._layoutWatcher=i.layout$.subscribe(o._onLayoutChange.bind(o))),o}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"align",{set:function(t){this._cacheInput("align",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignXs",{set:function(t){this._cacheInput("alignXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignSm",{set:function(t){this._cacheInput("alignSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignMd",{set:function(t){this._cacheInput("alignMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLg",{set:function(t){this._cacheInput("alignLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignXl",{set:function(t){this._cacheInput("alignXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignGtXs",{set:function(t){this._cacheInput("alignGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignGtSm",{set:function(t){this._cacheInput("alignGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignGtMd",{set:function(t){this._cacheInput("alignGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignGtLg",{set:function(t){this._cacheInput("alignGtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLtSm",{set:function(t){this._cacheInput("alignLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLtMd",{set:function(t){this._cacheInput("alignLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLtLg",{set:function(t){this._cacheInput("alignLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLtXl",{set:function(t){this._cacheInput("alignLtXl",t)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(null!=t.align||this._mqActivation)&&this._updateWithValue()},e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this._listenForMediaQueryChanges("align","start stretch",function(t){e._updateWithValue(t.value)}),this._updateWithValue()},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._layoutWatcher&&this._layoutWatcher.unsubscribe()},e.prototype._updateWithValue=function(t){t=t||this._queryInput("align")||"start stretch",this._mqActivation&&(t=this._mqActivation.activatedInput),this._applyStyleToElement(this._buildCSS(t)),this._allowStretching(t,this._layout?this._layout:"row")},e.prototype._onLayoutChange=function(t){var e=this;this._layout=(t||"").toLowerCase(),p.find(function(t){return t===e._layout})||(this._layout="row");var n=this._queryInput("align")||"start stretch";this._mqActivation&&(n=this._mqActivation.activatedInput),this._allowStretching(n,this._layout||"row")},e.prototype._buildCSS=function(t){var e={},n=t.split(" "),r=n[0],i=n[1];switch(r){case"center":e["justify-content"]="center";break;case"space-around":e["justify-content"]="space-around";break;case"space-between":e["justify-content"]="space-between";break;case"space-evenly":e["justify-content"]="space-evenly";break;case"end":case"flex-end":e["justify-content"]="flex-end";break;case"start":case"flex-start":default:e["justify-content"]="flex-start"}switch(i){case"start":case"flex-start":e["align-items"]=e["align-content"]="flex-start";break;case"baseline":e["align-items"]="baseline";break;case"center":e["align-items"]=e["align-content"]="center";break;case"end":case"flex-end":e["align-items"]=e["align-content"]="flex-end";break;case"stretch":default:e["align-items"]=e["align-content"]="stretch"}return j(e,{display:"flex","flex-direction":this._layout||"row","box-sizing":"border-box"})},e.prototype._allowStretching=function(t,e){"stretch"==t.split(" ")[1]&&this._applyStyleToElement({"box-sizing":"border-box","max-width":f(e)?null:"100%","max-height":f(e)?"100%":null})},e.decorators=[{type:r.s,args:[{selector:"\n  [fxLayoutAlign],\n  [fxLayoutAlign.xs], [fxLayoutAlign.sm], [fxLayoutAlign.md], [fxLayoutAlign.lg],[fxLayoutAlign.xl],\n  [fxLayoutAlign.lt-sm], [fxLayoutAlign.lt-md], [fxLayoutAlign.lt-lg], [fxLayoutAlign.lt-xl],\n  [fxLayoutAlign.gt-xs], [fxLayoutAlign.gt-sm], [fxLayoutAlign.gt-md], [fxLayoutAlign.gt-lg]\n"}]}],e.ctorParameters=function(){return[{type:T},{type:r.t},{type:r.U},{type:D,decorators:[{type:r.M},{type:r.Z}]}]},e.propDecorators={align:[{type:r.C,args:["fxLayoutAlign"]}],alignXs:[{type:r.C,args:["fxLayoutAlign.xs"]}],alignSm:[{type:r.C,args:["fxLayoutAlign.sm"]}],alignMd:[{type:r.C,args:["fxLayoutAlign.md"]}],alignLg:[{type:r.C,args:["fxLayoutAlign.lg"]}],alignXl:[{type:r.C,args:["fxLayoutAlign.xl"]}],alignGtXs:[{type:r.C,args:["fxLayoutAlign.gt-xs"]}],alignGtSm:[{type:r.C,args:["fxLayoutAlign.gt-sm"]}],alignGtMd:[{type:r.C,args:["fxLayoutAlign.gt-md"]}],alignGtLg:[{type:r.C,args:["fxLayoutAlign.gt-lg"]}],alignLtSm:[{type:r.C,args:["fxLayoutAlign.lt-sm"]}],alignLtMd:[{type:r.C,args:["fxLayoutAlign.lt-md"]}],alignLtLg:[{type:r.C,args:["fxLayoutAlign.lt-lg"]}],alignLtXl:[{type:r.C,args:["fxLayoutAlign.lt-xl"]}]},e}(O),R=function(t){function e(e,n,r,i,o){var s=t.call(this,e,n,r)||this;return s._zone=o,s._layout="row",i&&(s._layoutWatcher=i.layout$.subscribe(s._onLayoutChange.bind(s))),s}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"gap",{set:function(t){this._cacheInput("gap",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapXs",{set:function(t){this._cacheInput("gapXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapSm",{set:function(t){this._cacheInput("gapSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapMd",{set:function(t){this._cacheInput("gapMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapLg",{set:function(t){this._cacheInput("gapLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapXl",{set:function(t){this._cacheInput("gapXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapGtXs",{set:function(t){this._cacheInput("gapGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapGtSm",{set:function(t){this._cacheInput("gapGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapGtMd",{set:function(t){this._cacheInput("gapGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapGtLg",{set:function(t){this._cacheInput("gapGtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapLtSm",{set:function(t){this._cacheInput("gapLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapLtMd",{set:function(t){this._cacheInput("gapLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapLtLg",{set:function(t){this._cacheInput("gapLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"gapLtXl",{set:function(t){this._cacheInput("gapLtXl",t)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(null!=t.gap||this._mqActivation)&&this._updateWithValue()},e.prototype.ngAfterContentInit=function(){var t=this;this._watchContentChanges(),this._listenForMediaQueryChanges("gap","0",function(e){t._updateWithValue(e.value)}),this._updateWithValue()},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._layoutWatcher&&this._layoutWatcher.unsubscribe(),this._observer&&this._observer.disconnect()},e.prototype._watchContentChanges=function(){var t=this;this._zone.runOutsideAngular(function(){"undefined"!=typeof MutationObserver&&(t._observer=new MutationObserver(function(e){e.some(function(t){return t.addedNodes&&t.addedNodes.length>0||t.removedNodes&&t.removedNodes.length>0})&&t._updateWithValue()}),t._observer.observe(t.nativeElement,{childList:!0}))})},e.prototype._onLayoutChange=function(t){var e=this;this._layout=(t||"").toLowerCase(),p.find(function(t){return t===e._layout})||(this._layout="row"),this._updateWithValue()},e.prototype._updateWithValue=function(t){var e=this;t=t||this._queryInput("gap")||"0",this._mqActivation&&(t=this._mqActivation.activatedInput);var n=this.childrenNodes.filter(function(t){return 1===t.nodeType&&"none"!=e._getDisplayStyle(t)}),r=n.length;if(r>0){var i=n[r-1];n=n.filter(function(t,e){return e<r-1}),this._applyStyleToElements(this._buildCSS(t),n),this._applyStyleToElements(this._buildCSS(),[i])}},e.prototype._buildCSS=function(t){void 0===t&&(t=null);var e,n={"margin-left":null,"margin-right":null,"margin-top":null,"margin-bottom":null};switch(this._layout){case"column":case"column-reverse":e="margin-bottom";break;case"row":case"row-reverse":default:e="margin-right"}return n[e]=t,n},e.decorators=[{type:r.s,args:[{selector:"\n  [fxLayoutGap],\n  [fxLayoutGap.xs], [fxLayoutGap.sm], [fxLayoutGap.md], [fxLayoutGap.lg], [fxLayoutGap.xl],\n  [fxLayoutGap.lt-sm], [fxLayoutGap.lt-md], [fxLayoutGap.lt-lg], [fxLayoutGap.lt-xl],\n  [fxLayoutGap.gt-xs], [fxLayoutGap.gt-sm], [fxLayoutGap.gt-md], [fxLayoutGap.gt-lg]\n"}]}],e.ctorParameters=function(){return[{type:T},{type:r.t},{type:r.U},{type:D,decorators:[{type:r.M},{type:r.Z}]},{type:r.L}]},e.propDecorators={gap:[{type:r.C,args:["fxLayoutGap"]}],gapXs:[{type:r.C,args:["fxLayoutGap.xs"]}],gapSm:[{type:r.C,args:["fxLayoutGap.sm"]}],gapMd:[{type:r.C,args:["fxLayoutGap.md"]}],gapLg:[{type:r.C,args:["fxLayoutGap.lg"]}],gapXl:[{type:r.C,args:["fxLayoutGap.xl"]}],gapGtXs:[{type:r.C,args:["fxLayoutGap.gt-xs"]}],gapGtSm:[{type:r.C,args:["fxLayoutGap.gt-sm"]}],gapGtMd:[{type:r.C,args:["fxLayoutGap.gt-md"]}],gapGtLg:[{type:r.C,args:["fxLayoutGap.gt-lg"]}],gapLtSm:[{type:r.C,args:["fxLayoutGap.lt-sm"]}],gapLtMd:[{type:r.C,args:["fxLayoutGap.lt-md"]}],gapLtLg:[{type:r.C,args:["fxLayoutGap.lt-lg"]}],gapLtXl:[{type:r.C,args:["fxLayoutGap.lt-xl"]}]},e}(O),N=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r)||this;return o._layout="row",i&&(o._layoutWatcher=i.layout$.subscribe(o._onLayoutChange.bind(o))),o}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"wrap",{set:function(t){this._cacheInput("wrap",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapXs",{set:function(t){this._cacheInput("wrapXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapSm",{set:function(t){this._cacheInput("wrapSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapMd",{set:function(t){this._cacheInput("wrapMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapLg",{set:function(t){this._cacheInput("wrapLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapXl",{set:function(t){this._cacheInput("wrapXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapGtXs",{set:function(t){this._cacheInput("wrapGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapGtSm",{set:function(t){this._cacheInput("wrapGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapGtMd",{set:function(t){this._cacheInput("wrapGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapGtLg",{set:function(t){this._cacheInput("wrapGtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapLtSm",{set:function(t){this._cacheInput("wrapLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapLtMd",{set:function(t){this._cacheInput("wrapLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapLtLg",{set:function(t){this._cacheInput("wrapLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"wrapLtXl",{set:function(t){this._cacheInput("wrapLtXl",t)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(null!=t.wrap||this._mqActivation)&&this._updateWithValue()},e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this._listenForMediaQueryChanges("wrap","wrap",function(t){e._updateWithValue(t.value)}),this._updateWithValue()},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._layoutWatcher&&this._layoutWatcher.unsubscribe()},e.prototype._onLayoutChange=function(t){var e=this;this._layout=(t||"").toLowerCase().replace("-reverse",""),p.find(function(t){return t===e._layout})||(this._layout="row"),this._updateWithValue()},e.prototype._updateWithValue=function(t){t=t||this._queryInput("wrap"),this._mqActivation&&(t=this._mqActivation.activatedInput),t=m(t||"wrap"),this._applyStyleToElement(this._buildCSS(t))},e.prototype._buildCSS=function(t){return{display:"flex","flex-wrap":t,"flex-direction":this.flowDirection}},Object.defineProperty(e.prototype,"flowDirection",{get:function(){var t=this;return this._layoutWatcher?this._layout:t._getFlowDirection(t.nativeElement)},enumerable:!0,configurable:!0}),e.decorators=[{type:r.s,args:[{selector:"\n  [fxLayoutWrap], [fxLayoutWrap.xs], [fxLayoutWrap.sm], [fxLayoutWrap.lg], [fxLayoutWrap.xl],\n  [fxLayoutWrap.gt-xs], [fxLayoutWrap.gt-sm], [fxLayoutWrap.gt-md], [fxLayoutWrap.gt-lg],\n  [fxLayoutWrap.lt-xs], [fxLayoutWrap.lt-sm], [fxLayoutWrap.lt-md], [fxLayoutWrap.lt-lg]\n"}]}],e.ctorParameters=function(){return[{type:T},{type:r.t},{type:r.U},{type:D,decorators:[{type:r.M},{type:r.Z}]}]},e.propDecorators={wrap:[{type:r.C,args:["fxLayoutWrap"]}],wrapXs:[{type:r.C,args:["fxLayoutWrap.xs"]}],wrapSm:[{type:r.C,args:["fxLayoutWrap.sm"]}],wrapMd:[{type:r.C,args:["fxLayoutWrap.md"]}],wrapLg:[{type:r.C,args:["fxLayoutWrap.lg"]}],wrapXl:[{type:r.C,args:["fxLayoutWrap.xl"]}],wrapGtXs:[{type:r.C,args:["fxLayoutWrap.gt-xs"]}],wrapGtSm:[{type:r.C,args:["fxLayoutWrap.gt-sm"]}],wrapGtMd:[{type:r.C,args:["fxLayoutWrap.gt-md"]}],wrapGtLg:[{type:r.C,args:["fxLayoutWrap.gt-lg"]}],wrapLtSm:[{type:r.C,args:["fxLayoutWrap.lt-sm"]}],wrapLtMd:[{type:r.C,args:["fxLayoutWrap.lt-md"]}],wrapLtLg:[{type:r.C,args:["fxLayoutWrap.lt-lg"]}],wrapLtXl:[{type:r.C,args:["fxLayoutWrap.lt-xl"]}]},e}(O);function L(t){return t.replace(/[\s]/g,"").replace(/[\/\*\+\-]/g," $& ")}var F=function(t){function e(e,n,r,i,o){var s=t.call(this,e,n,r)||this;return s._container=i,s._wrap=o,s._cacheInput("flex",""),s._cacheInput("shrink",1),s._cacheInput("grow",1),i&&(s._layoutWatcher=i.layout$.subscribe(function(t){s._onLayoutChange(t)})),s}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"shrink",{set:function(t){this._cacheInput("shrink",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"grow",{set:function(t){this._cacheInput("grow",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flex",{set:function(t){this._cacheInput("flex",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexXs",{set:function(t){this._cacheInput("flexXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexSm",{set:function(t){this._cacheInput("flexSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexMd",{set:function(t){this._cacheInput("flexMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexLg",{set:function(t){this._cacheInput("flexLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexXl",{set:function(t){this._cacheInput("flexXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexGtXs",{set:function(t){this._cacheInput("flexGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexGtSm",{set:function(t){this._cacheInput("flexGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexGtMd",{set:function(t){this._cacheInput("flexGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexGtLg",{set:function(t){this._cacheInput("flexGtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexLtSm",{set:function(t){this._cacheInput("flexLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexLtMd",{set:function(t){this._cacheInput("flexLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexLtLg",{set:function(t){this._cacheInput("flexLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"flexLtXl",{set:function(t){this._cacheInput("flexLtXl",t)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(null!=t.flex||this._mqActivation)&&this._updateStyle()},e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this._listenForMediaQueryChanges("flex","",function(t){e._updateStyle(t.value)}),this._updateStyle()},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._layoutWatcher&&this._layoutWatcher.unsubscribe()},e.prototype._onLayoutChange=function(t){this._layout=t||this._layout||"row",this._updateStyle()},e.prototype._updateStyle=function(t){var e=t||this._queryInput("flex")||"";this._mqActivation&&(e=this._mqActivation.activatedInput);var n=function(t,e,n){void 0===e&&(e="1"),void 0===n&&(n="1");var r=[e,n,t],i=t.indexOf("calc");if(i>0)r[2]=L(t.substring(i).trim()),2==(o=t.substr(0,i).trim().split(" ")).length&&(r[0]=o[0],r[1]=o[1]);else if(0==i)r[2]=L(t.trim());else{var o;r=3===(o=t.split(" ")).length?o:[e,n,t]}return r}(String(e).replace(";",""),this._queryInput("grow"),this._queryInput("shrink"));this._applyStyleToElement(this._validateValue.apply(this,n))},e.prototype._validateValue=function(t,e,n){var r,i,o=this._getFlowDirection(this.parentElement,!0).indexOf("column")>-1?"column":"row";t="0"==t?0:t,e="0"==e?0:e;var s={"max-width":null,"max-height":null,"min-width":null,"min-height":null};switch(n||""){case"":r=j(s,{flex:t+" "+e+" 0.000000001px"});break;case"initial":case"nogrow":t=0,r=j(s,{flex:"0 1 auto"});break;case"grow":r=j(s,{flex:"1 1 100%"});break;case"noshrink":e=0,r=j(s,{flex:"1 0 auto"});break;case"auto":r=j(s,{flex:t+" "+e+" auto"});break;case"none":t=0,e=0,r=j(s,{flex:"0 0 auto"});break;default:var a=String(n).indexOf("calc")>-1,c=String(n).indexOf("%")>-1&&!a;(i=a||String(n).indexOf("px")>-1||String(n).indexOf("em")>-1||String(n).indexOf("vw")>-1||String(n).indexOf("vh")>-1)||c||isNaN(n)||(n+="%"),"0px"===n&&(n="0%"),r=j(s,{"flex-grow":""+t,"flex-shrink":""+e,"flex-basis":i||this._wrap?""+n:"100%"})}var u=f(o)?"max-width":"max-height",l=f(o)?"min-width":"min-height",p=String(n).indexOf("calc")>-1||"auto"==n,d=String(n).indexOf("px")>-1||p,h=!t&&!e;return r[l]="0%"==n?0:h||d&&t?n:null,r[u]="0%"==n?0:h||!p&&e?n:null,j(r,{"box-sizing":"border-box"})},e.decorators=[{type:r.s,args:[{selector:"\n  [fxFlex],\n  [fxFlex.xs], [fxFlex.sm], [fxFlex.md], [fxFlex.lg], [fxFlex.xl],\n  [fxFlex.lt-sm], [fxFlex.lt-md], [fxFlex.lt-lg], [fxFlex.lt-xl],\n  [fxFlex.gt-xs], [fxFlex.gt-sm], [fxFlex.gt-md], [fxFlex.gt-lg],\n"}]}],e.ctorParameters=function(){return[{type:T},{type:r.t},{type:r.U},{type:D,decorators:[{type:r.M},{type:r._1}]},{type:N,decorators:[{type:r.M},{type:r._1}]}]},e.propDecorators={shrink:[{type:r.C,args:["fxShrink"]}],grow:[{type:r.C,args:["fxGrow"]}],flex:[{type:r.C,args:["fxFlex"]}],flexXs:[{type:r.C,args:["fxFlex.xs"]}],flexSm:[{type:r.C,args:["fxFlex.sm"]}],flexMd:[{type:r.C,args:["fxFlex.md"]}],flexLg:[{type:r.C,args:["fxFlex.lg"]}],flexXl:[{type:r.C,args:["fxFlex.xl"]}],flexGtXs:[{type:r.C,args:["fxFlex.gt-xs"]}],flexGtSm:[{type:r.C,args:["fxFlex.gt-sm"]}],flexGtMd:[{type:r.C,args:["fxFlex.gt-md"]}],flexGtLg:[{type:r.C,args:["fxFlex.gt-lg"]}],flexLtSm:[{type:r.C,args:["fxFlex.lt-sm"]}],flexLtMd:[{type:r.C,args:["fxFlex.lt-md"]}],flexLtLg:[{type:r.C,args:["fxFlex.lt-lg"]}],flexLtXl:[{type:r.C,args:["fxFlex.lt-xl"]}]},e}(O),V=function(t){function e(e,n,r){return t.call(this,e,n,r)||this}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"align",{set:function(t){this._cacheInput("align",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignXs",{set:function(t){this._cacheInput("alignXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignSm",{set:function(t){this._cacheInput("alignSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignMd",{set:function(t){this._cacheInput("alignMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLg",{set:function(t){this._cacheInput("alignLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignXl",{set:function(t){this._cacheInput("alignXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLtSm",{set:function(t){this._cacheInput("alignLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLtMd",{set:function(t){this._cacheInput("alignLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLtLg",{set:function(t){this._cacheInput("alignLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignLtXl",{set:function(t){this._cacheInput("alignLtXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignGtXs",{set:function(t){this._cacheInput("alignGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignGtSm",{set:function(t){this._cacheInput("alignGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignGtMd",{set:function(t){this._cacheInput("alignGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"alignGtLg",{set:function(t){this._cacheInput("alignGtLg",t)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(null!=t.align||this._mqActivation)&&this._updateWithValue()},e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this._listenForMediaQueryChanges("align","stretch",function(t){e._updateWithValue(t.value)}),this._updateWithValue()},e.prototype._updateWithValue=function(t){t=t||this._queryInput("align")||"stretch",this._mqActivation&&(t=this._mqActivation.activatedInput),this._applyStyleToElement(this._buildCSS(t))},e.prototype._buildCSS=function(t){var e={};switch(t){case"start":e["align-self"]="flex-start";break;case"end":e["align-self"]="flex-end";break;default:e["align-self"]=t}return e},e.decorators=[{type:r.s,args:[{selector:"\n  [fxFlexAlign],\n  [fxFlexAlign.xs], [fxFlexAlign.sm], [fxFlexAlign.md], [fxFlexAlign.lg], [fxFlexAlign.xl],\n  [fxFlexAlign.lt-sm], [fxFlexAlign.lt-md], [fxFlexAlign.lt-lg], [fxFlexAlign.lt-xl],\n  [fxFlexAlign.gt-xs], [fxFlexAlign.gt-sm], [fxFlexAlign.gt-md], [fxFlexAlign.gt-lg]\n"}]}],e.ctorParameters=function(){return[{type:T},{type:r.t},{type:r.U}]},e.propDecorators={align:[{type:r.C,args:["fxFlexAlign"]}],alignXs:[{type:r.C,args:["fxFlexAlign.xs"]}],alignSm:[{type:r.C,args:["fxFlexAlign.sm"]}],alignMd:[{type:r.C,args:["fxFlexAlign.md"]}],alignLg:[{type:r.C,args:["fxFlexAlign.lg"]}],alignXl:[{type:r.C,args:["fxFlexAlign.xl"]}],alignLtSm:[{type:r.C,args:["fxFlexAlign.lt-sm"]}],alignLtMd:[{type:r.C,args:["fxFlexAlign.lt-md"]}],alignLtLg:[{type:r.C,args:["fxFlexAlign.lt-lg"]}],alignLtXl:[{type:r.C,args:["fxFlexAlign.lt-xl"]}],alignGtXs:[{type:r.C,args:["fxFlexAlign.gt-xs"]}],alignGtSm:[{type:r.C,args:["fxFlexAlign.gt-sm"]}],alignGtMd:[{type:r.C,args:["fxFlexAlign.gt-md"]}],alignGtLg:[{type:r.C,args:["fxFlexAlign.gt-lg"]}]},e}(O),z={margin:0,width:"100%",height:"100%","min-width":"100%","min-height":"100%"},B=function(t){function e(e,n,r){var i=t.call(this,e,n,r)||this;return i.elRef=n,i.renderer=r,i._applyStyleToElement(z),i}return Object(s.b)(e,t),e.decorators=[{type:r.s,args:[{selector:"\n  [fxFill],\n  [fxFlexFill]\n"}]}],e.ctorParameters=function(){return[{type:T},{type:r.t},{type:r.U}]},e}(O),q=function(t){function e(e,n,r,i){var o=t.call(this,e,n,r)||this;return o._container=i,o._layout="row",o.watchParentFlow(),o}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"offset",{set:function(t){this._cacheInput("offset",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetXs",{set:function(t){this._cacheInput("offsetXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetSm",{set:function(t){this._cacheInput("offsetSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetMd",{set:function(t){this._cacheInput("offsetMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetLg",{set:function(t){this._cacheInput("offsetLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetXl",{set:function(t){this._cacheInput("offsetXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetLtSm",{set:function(t){this._cacheInput("offsetLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetLtMd",{set:function(t){this._cacheInput("offsetLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetLtLg",{set:function(t){this._cacheInput("offsetLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetLtXl",{set:function(t){this._cacheInput("offsetLtXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetGtXs",{set:function(t){this._cacheInput("offsetGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetGtSm",{set:function(t){this._cacheInput("offsetGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetGtMd",{set:function(t){this._cacheInput("offsetGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"offsetGtLg",{set:function(t){this._cacheInput("offsetGtLg",t)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(null!=t.offset||this._mqActivation)&&this._updateWithValue()},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._layoutWatcher&&this._layoutWatcher.unsubscribe()},e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this._listenForMediaQueryChanges("offset",0,function(t){e._updateWithValue(t.value)})},e.prototype.watchParentFlow=function(){var t=this;this._container&&(this._layoutWatcher=this._container.layout$.subscribe(function(e){t._onLayoutChange(e)}))},e.prototype._onLayoutChange=function(t){this._layout=t||this._layout||"row",this._updateWithValue()},e.prototype._updateWithValue=function(t){t=t||this._queryInput("offset")||0,this._mqActivation&&(t=this._mqActivation.activatedInput),this._applyStyleToElement(this._buildCSS(t))},e.prototype._buildCSS=function(t){var e=String(t).indexOf("%")>-1;return String(t).indexOf("px")>-1||e||isNaN(t)||(t+="%"),f(this._getFlowDirection(this.parentElement,!0))?{"margin-left":""+t}:{"margin-top":""+t}},e.decorators=[{type:r.s,args:[{selector:"\n  [fxFlexOffset],\n  [fxFlexOffset.xs], [fxFlexOffset.sm], [fxFlexOffset.md], [fxFlexOffset.lg], [fxFlexOffset.xl],\n  [fxFlexOffset.lt-sm], [fxFlexOffset.lt-md], [fxFlexOffset.lt-lg], [fxFlexOffset.lt-xl],\n  [fxFlexOffset.gt-xs], [fxFlexOffset.gt-sm], [fxFlexOffset.gt-md], [fxFlexOffset.gt-lg]\n"}]}],e.ctorParameters=function(){return[{type:T},{type:r.t},{type:r.U},{type:D,decorators:[{type:r.M},{type:r._1}]}]},e.propDecorators={offset:[{type:r.C,args:["fxFlexOffset"]}],offsetXs:[{type:r.C,args:["fxFlexOffset.xs"]}],offsetSm:[{type:r.C,args:["fxFlexOffset.sm"]}],offsetMd:[{type:r.C,args:["fxFlexOffset.md"]}],offsetLg:[{type:r.C,args:["fxFlexOffset.lg"]}],offsetXl:[{type:r.C,args:["fxFlexOffset.xl"]}],offsetLtSm:[{type:r.C,args:["fxFlexOffset.lt-sm"]}],offsetLtMd:[{type:r.C,args:["fxFlexOffset.lt-md"]}],offsetLtLg:[{type:r.C,args:["fxFlexOffset.lt-lg"]}],offsetLtXl:[{type:r.C,args:["fxFlexOffset.lt-xl"]}],offsetGtXs:[{type:r.C,args:["fxFlexOffset.gt-xs"]}],offsetGtSm:[{type:r.C,args:["fxFlexOffset.gt-sm"]}],offsetGtMd:[{type:r.C,args:["fxFlexOffset.gt-md"]}],offsetGtLg:[{type:r.C,args:["fxFlexOffset.gt-lg"]}]},e}(O),H=function(t){function e(e,n,r){return t.call(this,e,n,r)||this}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"order",{set:function(t){this._cacheInput("order",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderXs",{set:function(t){this._cacheInput("orderXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderSm",{set:function(t){this._cacheInput("orderSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderMd",{set:function(t){this._cacheInput("orderMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderLg",{set:function(t){this._cacheInput("orderLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderXl",{set:function(t){this._cacheInput("orderXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderGtXs",{set:function(t){this._cacheInput("orderGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderGtSm",{set:function(t){this._cacheInput("orderGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderGtMd",{set:function(t){this._cacheInput("orderGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderGtLg",{set:function(t){this._cacheInput("orderGtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderLtSm",{set:function(t){this._cacheInput("orderLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderLtMd",{set:function(t){this._cacheInput("orderLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderLtLg",{set:function(t){this._cacheInput("orderLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"orderLtXl",{set:function(t){this._cacheInput("orderLtXl",t)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(null!=t.order||this._mqActivation)&&this._updateWithValue()},e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this._listenForMediaQueryChanges("order","0",function(t){e._updateWithValue(t.value)}),this._updateWithValue()},e.prototype._updateWithValue=function(t){t=t||this._queryInput("order")||"0",this._mqActivation&&(t=this._mqActivation.activatedInput),this._applyStyleToElement(this._buildCSS(t))},e.prototype._buildCSS=function(t){return t=parseInt(t,10),{order:isNaN(t)?0:t}},e.decorators=[{type:r.s,args:[{selector:"\n  [fxFlexOrder],\n  [fxFlexOrder.xs], [fxFlexOrder.sm], [fxFlexOrder.md], [fxFlexOrder.lg], [fxFlexOrder.xl],\n  [fxFlexOrder.lt-sm], [fxFlexOrder.lt-md], [fxFlexOrder.lt-lg], [fxFlexOrder.lt-xl],\n  [fxFlexOrder.gt-xs], [fxFlexOrder.gt-sm], [fxFlexOrder.gt-md], [fxFlexOrder.gt-lg]\n"}]}],e.ctorParameters=function(){return[{type:T},{type:r.t},{type:r.U}]},e.propDecorators={order:[{type:r.C,args:["fxFlexOrder"]}],orderXs:[{type:r.C,args:["fxFlexOrder.xs"]}],orderSm:[{type:r.C,args:["fxFlexOrder.sm"]}],orderMd:[{type:r.C,args:["fxFlexOrder.md"]}],orderLg:[{type:r.C,args:["fxFlexOrder.lg"]}],orderXl:[{type:r.C,args:["fxFlexOrder.xl"]}],orderGtXs:[{type:r.C,args:["fxFlexOrder.gt-xs"]}],orderGtSm:[{type:r.C,args:["fxFlexOrder.gt-sm"]}],orderGtMd:[{type:r.C,args:["fxFlexOrder.gt-md"]}],orderGtLg:[{type:r.C,args:["fxFlexOrder.gt-lg"]}],orderLtSm:[{type:r.C,args:["fxFlexOrder.lt-sm"]}],orderLtMd:[{type:r.C,args:["fxFlexOrder.lt-md"]}],orderLtLg:[{type:r.C,args:["fxFlexOrder.lt-lg"]}],orderLtXl:[{type:r.C,args:["fxFlexOrder.lt-xl"]}]},e}(O),U=function(){function t(t){this._renderer=t}return t.prototype.setElementClass=function(t,e,n){n?this._renderer.addClass(t,e):this._renderer.removeClass(t,e)},t.prototype.setElementStyle=function(t,e,n){n?this._renderer.setStyle(t,e,n):this._renderer.removeStyle(t,e)},t.prototype.addClass=function(t,e){this._renderer.addClass(t,e)},t.prototype.removeClass=function(t,e){this._renderer.removeClass(t,e)},t.prototype.setStyle=function(t,e,n,r){this._renderer.setStyle(t,e,n,r)},t.prototype.removeStyle=function(t,e,n){this._renderer.removeStyle(t,e,n)},t.prototype.animate=function(){throw W("animate")},t.prototype.attachViewAfter=function(){throw W("attachViewAfter")},t.prototype.detachView=function(){throw W("detachView")},t.prototype.destroyView=function(){throw W("destroyView")},t.prototype.createElement=function(){throw W("createElement")},t.prototype.createViewRoot=function(){throw W("createViewRoot")},t.prototype.createTemplateAnchor=function(){throw W("createTemplateAnchor")},t.prototype.createText=function(){throw W("createText")},t.prototype.invokeElementMethod=function(){throw W("invokeElementMethod")},t.prototype.projectNodes=function(){throw W("projectNodes")},t.prototype.selectRootElement=function(){throw W("selectRootElement")},t.prototype.setBindingDebugInfo=function(){throw W("setBindingDebugInfo")},t.prototype.setElementProperty=function(){throw W("setElementProperty")},t.prototype.setElementAttribute=function(){throw W("setElementAttribute")},t.prototype.setText=function(){throw W("setText")},t.prototype.listen=function(){throw W("listen")},t.prototype.listenGlobal=function(){throw W("listenGlobal")},t}();function W(t){return new Error("The method RendererAdapter::"+t+"() has not been implemented")}var G=function(t){function e(e,n,r,i,o,s){var a=t.call(this,e,i,o)||this;return a.monitor=e,a._iterableDiffers=n,a._keyValueDiffers=r,a._ngEl=i,a._renderer=o,a._ngClassInstance=s,a._configureAdapters(),a}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"ngClassBase",{set:function(t){this._base.cacheInput("ngClass",t,!0),this._ngClassInstance.ngClass=this._base.queryInput("ngClass")},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"klazz",{set:function(t){this._base.cacheInput("class",t),this._ngClassInstance.klass=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassXs",{set:function(t){this._base.cacheInput("ngClassXs",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassSm",{set:function(t){this._base.cacheInput("ngClassSm",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassMd",{set:function(t){this._base.cacheInput("ngClassMd",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassLg",{set:function(t){this._base.cacheInput("ngClassLg",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassXl",{set:function(t){this._base.cacheInput("ngClassXl",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassLtSm",{set:function(t){this._base.cacheInput("ngClassLtSm",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassLtMd",{set:function(t){this._base.cacheInput("ngClassLtMd",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassLtLg",{set:function(t){this._base.cacheInput("ngClassLtLg",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassLtXl",{set:function(t){this._base.cacheInput("ngClassLtXl",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassGtXs",{set:function(t){this._base.cacheInput("ngClassGtXs",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassGtSm",{set:function(t){this._base.cacheInput("ngClassGtSm",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassGtMd",{set:function(t){this._base.cacheInput("ngClassGtMd",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngClassGtLg",{set:function(t){this._base.cacheInput("ngClassGtLg",t,!0)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){this._base.activeKey in t&&(this._ngClassInstance.ngClass=this._base.mqActivation.activatedInput||"")},e.prototype.ngOnInit=function(){this._configureMQListener()},e.prototype.ngDoCheck=function(){this._ngClassInstance.ngDoCheck()},e.prototype.ngOnDestroy=function(){this._base.ngOnDestroy(),this._ngClassInstance=null},e.prototype._configureAdapters=function(){if(this._base=new C("ngClass",this.monitor,this._ngEl,this._renderer),!this._ngClassInstance){var t=new U(this._renderer);this._ngClassInstance=new a.d(this._iterableDiffers,this._keyValueDiffers,this._ngEl,t)}},e.prototype._configureMQListener=function(t){var e=this;void 0===t&&(t="ngClass");var n=this._base.queryInput(t);this._base.listenForMediaQueryChanges(t,n,function(t){e._ngClassInstance.ngClass=t.value||"",e._ngClassInstance.ngDoCheck()})},e.decorators=[{type:r.s,args:[{selector:"\n    [ngClass.xs], [ngClass.sm], [ngClass.md], [ngClass.lg], [ngClass.xl],\n    [ngClass.lt-sm], [ngClass.lt-md], [ngClass.lt-lg], [ngClass.lt-xl],\n    [ngClass.gt-xs], [ngClass.gt-sm], [ngClass.gt-md], [ngClass.gt-lg]\n  "}]}],e.ctorParameters=function(){return[{type:T},{type:r.D},{type:r.E},{type:r.t},{type:r.U},{type:a.d,decorators:[{type:r.M},{type:r.Z}]}]},e.propDecorators={ngClassBase:[{type:r.C,args:["ngClass"]}],klazz:[{type:r.C,args:["class"]}],ngClassXs:[{type:r.C,args:["ngClass.xs"]}],ngClassSm:[{type:r.C,args:["ngClass.sm"]}],ngClassMd:[{type:r.C,args:["ngClass.md"]}],ngClassLg:[{type:r.C,args:["ngClass.lg"]}],ngClassXl:[{type:r.C,args:["ngClass.xl"]}],ngClassLtSm:[{type:r.C,args:["ngClass.lt-sm"]}],ngClassLtMd:[{type:r.C,args:["ngClass.lt-md"]}],ngClassLtLg:[{type:r.C,args:["ngClass.lt-lg"]}],ngClassLtXl:[{type:r.C,args:["ngClass.lt-xl"]}],ngClassGtXs:[{type:r.C,args:["ngClass.gt-xs"]}],ngClassGtSm:[{type:r.C,args:["ngClass.gt-sm"]}],ngClassGtMd:[{type:r.C,args:["ngClass.gt-md"]}],ngClassGtLg:[{type:r.C,args:["ngClass.gt-lg"]}]},e}(O),X=function(){return function(t,e,n){void 0===n&&(n=!0),this.key=t,this.value=e,this.key=n?t.replace(/['"]/g,"").trim():t.trim(),this.value=n?e.replace(/['"]/g,"").trim():e.trim(),this.value=this.value.replace(/;/,"")}}(),Y={getType:K,buildRawList:function(t,e){void 0===e&&(e=";");return String(t).trim().split(e).map(function(t){return t.trim()}).filter(function(t){return""!==t})},buildMapFromList:$,buildMapFromSet:function(t,e){var n=new Array;"set"==K(t)?t.forEach(function(t){return n.push(t)}):Object.keys(t).forEach(function(e){n.push(e+":"+t[e])});return $(n,e)}};function K(t){var e=typeof t;return"object"===e?t.constructor===Array?"array":t.constructor===Set?"set":"object":e}function $(t,e){return t.map(Q).filter(function(t){return!!t}).map(function(t){return e&&(t.value=e(t.value)),t}).reduce(Z,{})}function Q(t){var e=t.split(":"),n=e[0],r=e[1];return r?new X(n,r):null}function Z(t,e){return e.key&&(t[e.key]=e.value),t}var J=function(t){function e(e,n,r,i,o,s){var a=t.call(this,e,r,i)||this;return a.monitor=e,a._sanitizer=n,a._ngEl=r,a._renderer=i,a._differs=o,a._ngStyleInstance=s,a._configureAdapters(),a}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"ngStyleBase",{set:function(t){this._base.cacheInput("ngStyle",t,!0),this._ngStyleInstance.ngStyle=this._base.queryInput("ngStyle")},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleXs",{set:function(t){this._base.cacheInput("ngStyleXs",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleSm",{set:function(t){this._base.cacheInput("ngStyleSm",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleMd",{set:function(t){this._base.cacheInput("ngStyleMd",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleLg",{set:function(t){this._base.cacheInput("ngStyleLg",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleXl",{set:function(t){this._base.cacheInput("ngStyleXl",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleLtSm",{set:function(t){this._base.cacheInput("ngStyleLtSm",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleLtMd",{set:function(t){this._base.cacheInput("ngStyleLtMd",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleLtLg",{set:function(t){this._base.cacheInput("ngStyleLtLg",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleLtXl",{set:function(t){this._base.cacheInput("ngStyleLtXl",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleGtXs",{set:function(t){this._base.cacheInput("ngStyleGtXs",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleGtSm",{set:function(t){this._base.cacheInput("ngStyleGtSm",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleGtMd",{set:function(t){this._base.cacheInput("ngStyleGtMd",t,!0)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ngStyleGtLg",{set:function(t){this._base.cacheInput("ngStyleGtLg",t,!0)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){this._base.activeKey in t&&(this._ngStyleInstance.ngStyle=this._base.mqActivation.activatedInput||"")},e.prototype.ngOnInit=function(){this._configureMQListener()},e.prototype.ngDoCheck=function(){this._ngStyleInstance.ngDoCheck()},e.prototype.ngOnDestroy=function(){this._base.ngOnDestroy(),this._ngStyleInstance=null},e.prototype._configureAdapters=function(){if(this._base=new C("ngStyle",this.monitor,this._ngEl,this._renderer),!this._ngStyleInstance){var t=new U(this._renderer);this._ngStyleInstance=new a.e(this._differs,this._ngEl,t)}this._buildCacheInterceptor(),this._fallbackToStyle()},e.prototype._configureMQListener=function(t){var e=this;void 0===t&&(t="ngStyle");var n=this._base.queryInput(t);this._base.listenForMediaQueryChanges(t,n,function(t){e._ngStyleInstance.ngStyle=t.value||"",e._ngStyleInstance.ngDoCheck()})},e.prototype._buildCacheInterceptor=function(){var t=this,e=this._base.cacheInput.bind(this._base);this._base.cacheInput=function(n,r,i,o){void 0===i&&(i=!1),void 0===o&&(o=!0);var s=t._buildStyleMap(r);o&&(s=j({},t._base.inputMap.ngStyle,s)),e(n,s,i)}},e.prototype._buildStyleMap=function(t){var e=this,n=function(t){return e._sanitizer.sanitize(r.Y.STYLE,t)};if(t)switch(Y.getType(t)){case"string":return Y.buildMapFromList(Y.buildRawList(t),n);case"array":return Y.buildMapFromList(t,n);case"set":default:return Y.buildMapFromSet(t,n)}return t},e.prototype._fallbackToStyle=function(){this._base.queryInput("ngStyle")||(this.ngStyleBase=this._getAttributeValue("style")||"")},e.decorators=[{type:r.s,args:[{selector:"\n    [ngStyle.xs], [ngStyle.sm], [ngStyle.md], [ngStyle.lg], [ngStyle.xl],\n    [ngStyle.lt-sm], [ngStyle.lt-md], [ngStyle.lt-lg], [ngStyle.lt-xl],\n    [ngStyle.gt-xs], [ngStyle.gt-sm], [ngStyle.gt-md], [ngStyle.gt-lg]\n  "}]}],e.ctorParameters=function(){return[{type:T},{type:i.c},{type:r.t},{type:r.U},{type:r.E},{type:a.e,decorators:[{type:r.M},{type:r.Z}]}]},e.propDecorators={ngStyleBase:[{type:r.C,args:["ngStyle"]}],ngStyleXs:[{type:r.C,args:["ngStyle.xs"]}],ngStyleSm:[{type:r.C,args:["ngStyle.sm"]}],ngStyleMd:[{type:r.C,args:["ngStyle.md"]}],ngStyleLg:[{type:r.C,args:["ngStyle.lg"]}],ngStyleXl:[{type:r.C,args:["ngStyle.xl"]}],ngStyleLtSm:[{type:r.C,args:["ngStyle.lt-sm"]}],ngStyleLtMd:[{type:r.C,args:["ngStyle.lt-md"]}],ngStyleLtLg:[{type:r.C,args:["ngStyle.lt-lg"]}],ngStyleLtXl:[{type:r.C,args:["ngStyle.lt-xl"]}],ngStyleGtXs:[{type:r.C,args:["ngStyle.gt-xs"]}],ngStyleGtSm:[{type:r.C,args:["ngStyle.gt-sm"]}],ngStyleGtMd:[{type:r.C,args:["ngStyle.gt-md"]}],ngStyleGtLg:[{type:r.C,args:["ngStyle.gt-lg"]}]},e}(O),tt=["false",!1,0];function et(t){return""!==t&&("false"===t||0===t||!t)}var nt=function(t){function e(e,n,r,i){var o=t.call(this,e,r,i)||this;return o._layout=n,o.elRef=r,o.renderer=i,n&&(o._layoutWatcher=n.layout$.subscribe(function(){return o._updateWithValue()})),o}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"show",{set:function(t){this._cacheInput("show",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showXs",{set:function(t){this._cacheInput("showXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showSm",{set:function(t){this._cacheInput("showSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showMd",{set:function(t){this._cacheInput("showMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showLg",{set:function(t){this._cacheInput("showLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showXl",{set:function(t){this._cacheInput("showXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showLtSm",{set:function(t){this._cacheInput("showLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showLtMd",{set:function(t){this._cacheInput("showLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showLtLg",{set:function(t){this._cacheInput("showLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showLtXl",{set:function(t){this._cacheInput("showLtXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showGtXs",{set:function(t){this._cacheInput("showGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showGtSm",{set:function(t){this._cacheInput("showGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showGtMd",{set:function(t){this._cacheInput("showGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"showGtLg",{set:function(t){this._cacheInput("showGtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hide",{set:function(t){this._cacheInput("show",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideXs",{set:function(t){this._cacheInput("showXs",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideSm",{set:function(t){this._cacheInput("showSm",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideMd",{set:function(t){this._cacheInput("showMd",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideLg",{set:function(t){this._cacheInput("showLg",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideXl",{set:function(t){this._cacheInput("showXl",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideLtSm",{set:function(t){this._cacheInput("showLtSm",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideLtMd",{set:function(t){this._cacheInput("showLtMd",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideLtLg",{set:function(t){this._cacheInput("showLtLg",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideLtXl",{set:function(t){this._cacheInput("showLtXl",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideGtXs",{set:function(t){this._cacheInput("showGtXs",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideGtSm",{set:function(t){this._cacheInput("showGtSm",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideGtMd",{set:function(t){this._cacheInput("showGtMd",et(t))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hideGtLg",{set:function(t){this._cacheInput("showGtLg",et(t))},enumerable:!0,configurable:!0}),e.prototype._getDisplayStyle=function(){return this._layout?"flex":t.prototype._getDisplayStyle.call(this)},e.prototype.ngOnChanges=function(t){this.hasInitialized&&(null!=t.show||this._mqActivation)&&this._updateWithValue()},e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this);var n=this._getDefaultVal("show",!0);this._listenForMediaQueryChanges("show",n,function(t){e._updateWithValue(t.value)}),this._updateWithValue()},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._layoutWatcher&&this._layoutWatcher.unsubscribe()},e.prototype._updateWithValue=function(t){t=t||this._getDefaultVal("show",!0),this._mqActivation&&(t=this._mqActivation.activatedInput);var e=this._validateTruthy(t);this._applyStyleToElement(this._buildCSS(e))},e.prototype._buildCSS=function(t){return{display:t?this._display:"none"}},e.prototype._validateTruthy=function(t){return-1==tt.indexOf(t)},e.decorators=[{type:r.s,args:[{selector:"\n  [fxShow],\n  [fxShow.xs], [fxShow.sm], [fxShow.md], [fxShow.lg], [fxShow.xl],\n  [fxShow.lt-sm], [fxShow.lt-md], [fxShow.lt-lg], [fxShow.lt-xl],\n  [fxShow.gt-xs], [fxShow.gt-sm], [fxShow.gt-md], [fxShow.gt-lg],\n  [fxHide],\n  [fxHide.xs], [fxHide.sm], [fxHide.md], [fxHide.lg], [fxHide.xl],\n  [fxHide.lt-sm], [fxHide.lt-md], [fxHide.lt-lg], [fxHide.lt-xl],\n  [fxHide.gt-xs], [fxHide.gt-sm], [fxHide.gt-md], [fxHide.gt-lg]\n"}]}],e.ctorParameters=function(){return[{type:T},{type:D,decorators:[{type:r.M},{type:r.Z}]},{type:r.t},{type:r.U}]},e.propDecorators={show:[{type:r.C,args:["fxShow"]}],showXs:[{type:r.C,args:["fxShow.xs"]}],showSm:[{type:r.C,args:["fxShow.sm"]}],showMd:[{type:r.C,args:["fxShow.md"]}],showLg:[{type:r.C,args:["fxShow.lg"]}],showXl:[{type:r.C,args:["fxShow.xl"]}],showLtSm:[{type:r.C,args:["fxShow.lt-sm"]}],showLtMd:[{type:r.C,args:["fxShow.lt-md"]}],showLtLg:[{type:r.C,args:["fxShow.lt-lg"]}],showLtXl:[{type:r.C,args:["fxShow.lt-xl"]}],showGtXs:[{type:r.C,args:["fxShow.gt-xs"]}],showGtSm:[{type:r.C,args:["fxShow.gt-sm"]}],showGtMd:[{type:r.C,args:["fxShow.gt-md"]}],showGtLg:[{type:r.C,args:["fxShow.gt-lg"]}],hide:[{type:r.C,args:["fxHide"]}],hideXs:[{type:r.C,args:["fxHide.xs"]}],hideSm:[{type:r.C,args:["fxHide.sm"]}],hideMd:[{type:r.C,args:["fxHide.md"]}],hideLg:[{type:r.C,args:["fxHide.lg"]}],hideXl:[{type:r.C,args:["fxHide.xl"]}],hideLtSm:[{type:r.C,args:["fxHide.lt-sm"]}],hideLtMd:[{type:r.C,args:["fxHide.lt-md"]}],hideLtLg:[{type:r.C,args:["fxHide.lt-lg"]}],hideLtXl:[{type:r.C,args:["fxHide.lt-xl"]}],hideGtXs:[{type:r.C,args:["fxHide.gt-xs"]}],hideGtSm:[{type:r.C,args:["fxHide.gt-sm"]}],hideGtMd:[{type:r.C,args:["fxHide.gt-md"]}],hideGtLg:[{type:r.C,args:["fxHide.gt-lg"]}]},e}(O),rt=function(t){function e(e,n,r){var i=t.call(this,r,e,n)||this;return i._cacheInput("src",e.nativeElement.getAttribute("src")||""),i}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"srcBase",{set:function(t){this.cacheDefaultSrc(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcXs",{set:function(t){this._cacheInput("srcXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcSm",{set:function(t){this._cacheInput("srcSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcMd",{set:function(t){this._cacheInput("srcMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcLg",{set:function(t){this._cacheInput("srcLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcXl",{set:function(t){this._cacheInput("srcXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcLtSm",{set:function(t){this._cacheInput("srcLtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcLtMd",{set:function(t){this._cacheInput("srcLtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcLtLg",{set:function(t){this._cacheInput("srcLtLg",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcLtXl",{set:function(t){this._cacheInput("srcLtXl",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcGtXs",{set:function(t){this._cacheInput("srcGtXs",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcGtSm",{set:function(t){this._cacheInput("srcGtSm",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcGtMd",{set:function(t){this._cacheInput("srcGtMd",t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"srcGtLg",{set:function(t){this._cacheInput("srcGtLg",t)},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){var e=this;t.prototype.ngOnInit.call(this),this.hasResponsiveKeys&&this._listenForMediaQueryChanges("src",this.defaultSrc,function(){e._updateSrcFor()}),this._updateSrcFor()},e.prototype.ngOnChanges=function(){this.hasInitialized&&this._updateSrcFor()},e.prototype._updateSrcFor=function(){if(this.hasResponsiveKeys){var t=this.activatedValue||this.defaultSrc;this._renderer.setAttribute(this.nativeElement,"src",String(t))}},e.prototype.cacheDefaultSrc=function(t){this._cacheInput("src",t||"")},Object.defineProperty(e.prototype,"defaultSrc",{get:function(){return this._queryInput("src")||""},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hasResponsiveKeys",{get:function(){return Object.keys(this._inputMap).length>1},enumerable:!0,configurable:!0}),e.decorators=[{type:r.s,args:[{selector:"\n  img[src.xs],    img[src.sm],    img[src.md],    img[src.lg],   img[src.xl],\n  img[src.lt-sm], img[src.lt-md], img[src.lt-lg], img[src.lt-xl],\n  img[src.gt-xs], img[src.gt-sm], img[src.gt-md], img[src.gt-lg]\n"}]}],e.ctorParameters=function(){return[{type:r.t},{type:r.U},{type:T}]},e.propDecorators={srcBase:[{type:r.C,args:["src"]}],srcXs:[{type:r.C,args:["src.xs"]}],srcSm:[{type:r.C,args:["src.sm"]}],srcMd:[{type:r.C,args:["src.md"]}],srcLg:[{type:r.C,args:["src.lg"]}],srcXl:[{type:r.C,args:["src.xl"]}],srcLtSm:[{type:r.C,args:["src.lt-sm"]}],srcLtMd:[{type:r.C,args:["src.lt-md"]}],srcLtLg:[{type:r.C,args:["src.lt-lg"]}],srcLtXl:[{type:r.C,args:["src.lt-xl"]}],srcGtXs:[{type:r.C,args:["src.gt-xs"]}],srcGtSm:[{type:r.C,args:["src.gt-sm"]}],srcGtMd:[{type:r.C,args:["src.gt-md"]}],srcGtLg:[{type:r.C,args:["src.gt-lg"]}]},e}(O),it=[{alias:"xs",mediaQuery:"(min-width: 0px) and (max-width: 599px)"},{alias:"gt-xs",overlapping:!0,mediaQuery:"(min-width: 600px)"},{alias:"lt-sm",overlapping:!0,mediaQuery:"(max-width: 599px)"},{alias:"sm",mediaQuery:"(min-width: 600px) and (max-width: 959px)"},{alias:"gt-sm",overlapping:!0,mediaQuery:"(min-width: 960px)"},{alias:"lt-md",overlapping:!0,mediaQuery:"(max-width: 959px)"},{alias:"md",mediaQuery:"(min-width: 960px) and (max-width: 1279px)"},{alias:"gt-md",overlapping:!0,mediaQuery:"(min-width: 1280px)"},{alias:"lt-lg",overlapping:!0,mediaQuery:"(max-width: 1279px)"},{alias:"lg",mediaQuery:"(min-width: 1280px) and (max-width: 1919px)"},{alias:"gt-lg",overlapping:!0,mediaQuery:"(min-width: 1920px)"},{alias:"lt-xl",overlapping:!0,mediaQuery:"(max-width: 1920px)"},{alias:"xl",mediaQuery:"(min-width: 1920px) and (max-width: 5000px)"}],ot="(orientations: portrait) and (max-width: 599px)",st="(orientations: landscape) and (max-width: 959px)",at="(orientations: landscape) and (min-width: 960px) and (max-width: 1279px)",ct="(orientations: portrait) and (min-width: 600px) and (max-width: 839px)",ut="(orientations: portrait) and (min-width: 840px)",lt="(orientations: landscape) and (min-width: 1280px)",pt={HANDSET:ot+", "+st,TABLET:ct+" , "+at,WEB:ut+", "+lt+" ",HANDSET_PORTRAIT:""+ot,TABLET_PORTRAIT:ct+" ",WEB_PORTRAIT:""+ut,HANDSET_LANDSCAPE:st+"]",TABLET_LANDSCAPE:""+at,WEB_LANDSCAPE:""+lt},dt=[{alias:"handset",mediaQuery:pt.HANDSET},{alias:"handset.landscape",mediaQuery:pt.HANDSET_LANDSCAPE},{alias:"handset.portrait",mediaQuery:pt.HANDSET_PORTRAIT},{alias:"tablet",mediaQuery:pt.TABLET},{alias:"tablet.landscape",mediaQuery:pt.TABLET},{alias:"tablet.portrait",mediaQuery:pt.TABLET_PORTRAIT},{alias:"web",mediaQuery:pt.WEB,overlapping:!0},{alias:"web.landscape",mediaQuery:pt.WEB_LANDSCAPE,overlapping:!0},{alias:"web.portrait",mediaQuery:pt.WEB_PORTRAIT,overlapping:!0}],ht=function(){return function(){}}(),ft=function(){function t(t,e){this.breakpoints=t,this.mediaWatcher=e,this.filterOverlaps=!0,this._registerBreakPoints(),this.observable$=this._buildObservable()}return t.prototype.isActive=function(t){var e=this._toMediaQuery(t);return this.mediaWatcher.isActive(e)},t.prototype.subscribe=function(t,e,n){return this.observable$.subscribe(t,e,n)},t.prototype.asObservable=function(){return this.observable$},t.prototype._registerBreakPoints=function(){var t=this.breakpoints.sortedItems.map(function(t){return t.mediaQuery});this.mediaWatcher.registerQuery(t)},t.prototype._buildObservable=function(){var t=this,e=this;return this.mediaWatcher.observe().pipe(Object(u.a)(function(t){return!0===t.matches}),Object(u.a)(function(n){var r=t.breakpoints.findByQuery(n.mediaQuery);return!r||!(e.filterOverlaps&&r.overlapping)}),Object(o.a)(function(e){return A(e,t._findByQuery(e.mediaQuery))}))},t.prototype._findByAlias=function(t){return this.breakpoints.findByAlias(t)},t.prototype._findByQuery=function(t){return this.breakpoints.findByQuery(t)},t.prototype._toMediaQuery=function(t){var e=this._findByAlias(t)||this._findByQuery(t);return e?e.mediaQuery:t},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:E},{type:P}]},t}(),mt=/(\.|-|_)/g;function gt(t){var e=t.length>0?t.charAt(0):"",n=t.length>1?t.slice(1):"";return e.toUpperCase()+n}function vt(t){return t.forEach(function(t){t.suffix&&""!==t.suffix||(t.suffix=t.alias.replace(mt,"|").split("|").map(gt).join(""),t.overlapping=t.overlapping||!1)}),t}function yt(t,e){void 0===e&&(e=[]);var n=t.map(function(t){return j({},t)});return e.forEach(function(t){var e,r=(e=t.alias,n.reduce(function(t,n){return t||(n.alias===e?n:null)},null));r?j(r,t):n.push(t)}),vt(n)}var bt={provide:k,useFactory:function(){return vt(it)}};function _t(t,e){return{provide:k,useFactory:(n=t,r=e,r=j({},{defaults:!0,orientation:!1},r||{}),function(){var t=r.orientations?dt.concat(it):it;return r.defaults?yt(t,n||[]):yt(n)})};var n,r}var jt={provide:ht,deps:[[new r.M,new r._1,ht],P,E],useFactory:function(t,e,n){return t||new ft(n,e)}};var xt={provide:T,deps:[[new r.M,new r._1,T],E,P],useFactory:function(t,e,n){return t||new T(e,n)}},wt=function(){function t(){}return t.decorators=[{type:r.H,args:[{providers:[bt,E,P,T,jt]}]}],t.ctorParameters=function(){return[]},t}(),Ot=[D,N,R,I,F,H,q,B,V,nt,G,J,rt],Ct=function(){function t(){}return t.provideBreakPoints=function(e,n){return{ngModule:t,providers:[_t(e,n||{orientations:!1})]}},t.decorators=[{type:r.H,args:[{imports:[wt],exports:[wt].concat(Ot),declarations:Ot.slice(),providers:[xt,bt,jt]}]}],t.ctorParameters=function(){return[]},t}()},"../../../forms/esm5/forms.js":function(t,e,n){"use strict";n.d(e,"d",function(){return b}),n.d(e,"e",function(){return M}),n.d(e,"f",function(){return bt}),n.d(e,"b",function(){return Rt}),n.d(e,"a",function(){return Gt}),n.d(e,"c",function(){return d}),n.d(e,"g",function(){return m});var r=n("../../../../tslib/tslib.es6.js"),i=n("../../../core/esm5/core.js"),o=n("../../../../rxjs/_esm5/observable/forkJoin.js"),s=n("../../../../rxjs/_esm5/observable/fromPromise.js"),a=n("../../../../rxjs/_esm5/operator/map.js"),c=n("../../../platform-browser/esm5/platform-browser.js"),u=function(){function t(){}return Object.defineProperty(t.prototype,"value",{get:function(){return this.control?this.control.value:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"valid",{get:function(){return this.control?this.control.valid:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"invalid",{get:function(){return this.control?this.control.invalid:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pending",{get:function(){return this.control?this.control.pending:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this.control?this.control.disabled:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"enabled",{get:function(){return this.control?this.control.enabled:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"errors",{get:function(){return this.control?this.control.errors:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pristine",{get:function(){return this.control?this.control.pristine:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dirty",{get:function(){return this.control?this.control.dirty:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"touched",{get:function(){return this.control?this.control.touched:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"status",{get:function(){return this.control?this.control.status:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"untouched",{get:function(){return this.control?this.control.untouched:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"statusChanges",{get:function(){return this.control?this.control.statusChanges:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"valueChanges",{get:function(){return this.control?this.control.valueChanges:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"path",{get:function(){return null},enumerable:!0,configurable:!0}),t.prototype.reset=function(t){void 0===t&&(t=void 0),this.control&&this.control.reset(t)},t.prototype.hasError=function(t,e){return!!this.control&&this.control.hasError(t,e)},t.prototype.getError=function(t,e){return this.control?this.control.getError(t,e):null},t}(),l=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"formDirective",{get:function(){return null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return null},enumerable:!0,configurable:!0}),e}(u);function p(t){return null==t||0===t.length}var d=new i.A("NgValidators"),h=new i.A("NgAsyncValidators"),f=/^(?=.{1,254}$)(?=.{1,64}@)[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+(\.[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+)*@[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?(\.[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*$/,m=function(){function t(){}return t.min=function(t){return function(e){if(p(e.value)||p(t))return null;var n=parseFloat(e.value);return!isNaN(n)&&n<t?{min:{min:t,actual:e.value}}:null}},t.max=function(t){return function(e){if(p(e.value)||p(t))return null;var n=parseFloat(e.value);return!isNaN(n)&&n>t?{max:{max:t,actual:e.value}}:null}},t.required=function(t){return p(t.value)?{required:!0}:null},t.requiredTrue=function(t){return!0===t.value?null:{required:!0}},t.email=function(t){return f.test(t.value)?null:{email:!0}},t.minLength=function(t){return function(e){if(p(e.value))return null;var n=e.value?e.value.length:0;return n<t?{minlength:{requiredLength:t,actualLength:n}}:null}},t.maxLength=function(t){return function(e){var n=e.value?e.value.length:0;return n>t?{maxlength:{requiredLength:t,actualLength:n}}:null}},t.pattern=function(e){return e?("string"==typeof e?(r="^"+e+"$",n=new RegExp(r)):(r=e.toString(),n=e),function(t){if(p(t.value))return null;var e=t.value;return n.test(e)?null:{pattern:{requiredPattern:r,actualValue:e}}}):t.nullValidator;var n,r},t.nullValidator=function(t){return null},t.compose=function(t){if(!t)return null;var e=t.filter(g);return 0==e.length?null:function(t){return y((n=t,e.map(function(t){return t(n)})));var n}},t.composeAsync=function(t){if(!t)return null;var e=t.filter(g);return 0==e.length?null:function(t){var n,r,i=(n=t,r=e,r.map(function(t){return t(n)})).map(v);return a.a.call(Object(o.a)(i),y)}},t}();function g(t){return null!=t}function v(t){var e=Object(i._35)(t)?Object(s.a)(t):t;if(!Object(i._34)(e))throw new Error("Expected validator to return Promise or Observable.");return e}function y(t){var e=t.reduce(function(t,e){return null!=e?Object(r.a)({},t,e):t},{});return 0===Object.keys(e).length?null:e}var b=new i.A("NgValueAccessor"),_={provide:b,useExisting:Object(i._14)(function(){return j}),multi:!0},j=function(){function t(t,e){this._renderer=t,this._elementRef=e,this.onChange=function(t){},this.onTouched=function(){}}return t.prototype.writeValue=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"checked",t)},t.prototype.registerOnChange=function(t){this.onChange=t},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.decorators=[{type:i.s,args:[{selector:"input[type=checkbox][formControlName],input[type=checkbox][formControl],input[type=checkbox][ngModel]",host:{"(change)":"onChange($event.target.checked)","(blur)":"onTouched()"},providers:[_]}]}],t.ctorParameters=function(){return[{type:i.U},{type:i.t}]},t}(),x={provide:b,useExisting:Object(i._14)(function(){return O}),multi:!0};var w=new i.A("CompositionEventMode"),O=function(){function t(t,e,n){var r;this._renderer=t,this._elementRef=e,this._compositionMode=n,this.onChange=function(t){},this.onTouched=function(){},this._composing=!1,null==this._compositionMode&&(this._compositionMode=(r=Object(c.h)()?Object(c.h)().getUserAgent():"",!/android (\d+)/.test(r.toLowerCase())))}return t.prototype.writeValue=function(t){var e=null==t?"":t;this._renderer.setProperty(this._elementRef.nativeElement,"value",e)},t.prototype.registerOnChange=function(t){this.onChange=t},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.prototype._handleInput=function(t){(!this._compositionMode||this._compositionMode&&!this._composing)&&this.onChange(t)},t.prototype._compositionStart=function(){this._composing=!0},t.prototype._compositionEnd=function(t){this._composing=!1,this._compositionMode&&this.onChange(t)},t.decorators=[{type:i.s,args:[{selector:"input:not([type=checkbox])[formControlName],textarea[formControlName],input:not([type=checkbox])[formControl],textarea[formControl],input:not([type=checkbox])[ngModel],textarea[ngModel],[ngDefaultControl]",host:{"(input)":"_handleInput($event.target.value)","(blur)":"onTouched()","(compositionstart)":"_compositionStart()","(compositionend)":"_compositionEnd($event.target.value)"},providers:[x]}]}],t.ctorParameters=function(){return[{type:i.U},{type:i.t},{type:void 0,decorators:[{type:i.M},{type:i.y,args:[w]}]}]},t}();function C(t){return t.validate?function(e){return t.validate(e)}:t}function k(t){return t.validate?function(e){return t.validate(e)}:t}var E={provide:b,useExisting:Object(i._14)(function(){return S}),multi:!0},S=function(){function t(t,e){this._renderer=t,this._elementRef=e,this.onChange=function(t){},this.onTouched=function(){}}return t.prototype.writeValue=function(t){var e=null==t?"":t;this._renderer.setProperty(this._elementRef.nativeElement,"value",e)},t.prototype.registerOnChange=function(t){this.onChange=function(e){t(""==e?null:parseFloat(e))}},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.decorators=[{type:i.s,args:[{selector:"input[type=number][formControlName],input[type=number][formControl],input[type=number][ngModel]",host:{"(change)":"onChange($event.target.value)","(input)":"onChange($event.target.value)","(blur)":"onTouched()"},providers:[E]}]}],t.ctorParameters=function(){return[{type:i.U},{type:i.t}]},t}();function P(){throw new Error("unimplemented")}var M=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._parent=null,e.name=null,e.valueAccessor=null,e._rawValidators=[],e._rawAsyncValidators=[],e}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"validator",{get:function(){return P()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return P()},enumerable:!0,configurable:!0}),e}(u),A={provide:b,useExisting:Object(i._14)(function(){return D}),multi:!0},T=function(){function t(){this._accessors=[]}return t.prototype.add=function(t,e){this._accessors.push([t,e])},t.prototype.remove=function(t){for(var e=this._accessors.length-1;e>=0;--e)if(this._accessors[e][1]===t)return void this._accessors.splice(e,1)},t.prototype.select=function(t){var e=this;this._accessors.forEach(function(n){e._isSameGroup(n,t)&&n[1]!==t&&n[1].fireUncheck(t.value)})},t.prototype._isSameGroup=function(t,e){return!!t[0].control&&(t[0]._parent===e._control._parent&&t[1].name===e.name)},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[]},t}(),D=function(){function t(t,e,n,r){this._renderer=t,this._elementRef=e,this._registry=n,this._injector=r,this.onChange=function(){},this.onTouched=function(){}}return t.prototype.ngOnInit=function(){this._control=this._injector.get(M),this._checkName(),this._registry.add(this._control,this)},t.prototype.ngOnDestroy=function(){this._registry.remove(this)},t.prototype.writeValue=function(t){this._state=t===this.value,this._renderer.setProperty(this._elementRef.nativeElement,"checked",this._state)},t.prototype.registerOnChange=function(t){var e=this;this._fn=t,this.onChange=function(){t(e.value),e._registry.select(e)}},t.prototype.fireUncheck=function(t){this.writeValue(t)},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.prototype._checkName=function(){this.name&&this.formControlName&&this.name!==this.formControlName&&this._throwNameError(),!this.name&&this.formControlName&&(this.name=this.formControlName)},t.prototype._throwNameError=function(){throw new Error('\n      If you define both a name and a formControlName attribute on your radio button, their values\n      must match. Ex: <input type="radio" formControlName="food" name="food">\n    ')},t.decorators=[{type:i.s,args:[{selector:"input[type=radio][formControlName],input[type=radio][formControl],input[type=radio][ngModel]",host:{"(change)":"onChange()","(blur)":"onTouched()"},providers:[A]}]}],t.ctorParameters=function(){return[{type:i.U},{type:i.t},{type:T},{type:i.B}]},t.propDecorators={name:[{type:i.C}],formControlName:[{type:i.C}],value:[{type:i.C}]},t}(),I={provide:b,useExisting:Object(i._14)(function(){return R}),multi:!0},R=function(){function t(t,e){this._renderer=t,this._elementRef=e,this.onChange=function(t){},this.onTouched=function(){}}return t.prototype.writeValue=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"value",parseFloat(t))},t.prototype.registerOnChange=function(t){this.onChange=function(e){t(""==e?null:parseFloat(e))}},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.decorators=[{type:i.s,args:[{selector:"input[type=range][formControlName],input[type=range][formControl],input[type=range][ngModel]",host:{"(change)":"onChange($event.target.value)","(input)":"onChange($event.target.value)","(blur)":"onTouched()"},providers:[I]}]}],t.ctorParameters=function(){return[{type:i.U},{type:i.t}]},t}(),N={provide:b,useExisting:Object(i._14)(function(){return F}),multi:!0};function L(t,e){return null==t?""+e:(e&&"object"==typeof e&&(e="Object"),(t+": "+e).slice(0,50))}var F=function(){function t(t,e){this._renderer=t,this._elementRef=e,this._optionMap=new Map,this._idCounter=0,this.onChange=function(t){},this.onTouched=function(){},this._compareWith=i._36}return Object.defineProperty(t.prototype,"compareWith",{set:function(t){if("function"!=typeof t)throw new Error("compareWith must be a function, but received "+JSON.stringify(t));this._compareWith=t},enumerable:!0,configurable:!0}),t.prototype.writeValue=function(t){this.value=t;var e=this._getOptionId(t);null==e&&this._renderer.setProperty(this._elementRef.nativeElement,"selectedIndex",-1);var n=L(e,t);this._renderer.setProperty(this._elementRef.nativeElement,"value",n)},t.prototype.registerOnChange=function(t){var e=this;this.onChange=function(n){e.value=e._getOptionValue(n),t(e.value)}},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.prototype._registerOption=function(){return(this._idCounter++).toString()},t.prototype._getOptionId=function(t){for(var e=0,n=Array.from(this._optionMap.keys());e<n.length;e++){var r=n[e];if(this._compareWith(this._optionMap.get(r),t))return r}return null},t.prototype._getOptionValue=function(t){var e=t.split(":")[0];return this._optionMap.has(e)?this._optionMap.get(e):t},t.decorators=[{type:i.s,args:[{selector:"select:not([multiple])[formControlName],select:not([multiple])[formControl],select:not([multiple])[ngModel]",host:{"(change)":"onChange($event.target.value)","(blur)":"onTouched()"},providers:[N]}]}],t.ctorParameters=function(){return[{type:i.U},{type:i.t}]},t.propDecorators={compareWith:[{type:i.C}]},t}(),V=function(){function t(t,e,n){this._element=t,this._renderer=e,this._select=n,this._select&&(this.id=this._select._registerOption())}return Object.defineProperty(t.prototype,"ngValue",{set:function(t){null!=this._select&&(this._select._optionMap.set(this.id,t),this._setElementValue(L(this.id,t)),this._select.writeValue(this._select.value))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{set:function(t){this._setElementValue(t),this._select&&this._select.writeValue(this._select.value)},enumerable:!0,configurable:!0}),t.prototype._setElementValue=function(t){this._renderer.setProperty(this._element.nativeElement,"value",t)},t.prototype.ngOnDestroy=function(){this._select&&(this._select._optionMap.delete(this.id),this._select.writeValue(this._select.value))},t.decorators=[{type:i.s,args:[{selector:"option"}]}],t.ctorParameters=function(){return[{type:i.t},{type:i.U},{type:F,decorators:[{type:i.M},{type:i.x}]}]},t.propDecorators={ngValue:[{type:i.C,args:["ngValue"]}],value:[{type:i.C,args:["value"]}]},t}(),z={provide:b,useExisting:Object(i._14)(function(){return q}),multi:!0};function B(t,e){return null==t?""+e:("string"==typeof e&&(e="'"+e+"'"),e&&"object"==typeof e&&(e="Object"),(t+": "+e).slice(0,50))}var q=function(){function t(t,e){this._renderer=t,this._elementRef=e,this._optionMap=new Map,this._idCounter=0,this.onChange=function(t){},this.onTouched=function(){},this._compareWith=i._36}return Object.defineProperty(t.prototype,"compareWith",{set:function(t){if("function"!=typeof t)throw new Error("compareWith must be a function, but received "+JSON.stringify(t));this._compareWith=t},enumerable:!0,configurable:!0}),t.prototype.writeValue=function(t){var e,n=this;if(this.value=t,Array.isArray(t)){var r=t.map(function(t){return n._getOptionId(t)});e=function(t,e){t._setSelected(r.indexOf(e.toString())>-1)}}else e=function(t,e){t._setSelected(!1)};this._optionMap.forEach(e)},t.prototype.registerOnChange=function(t){var e=this;this.onChange=function(n){var r=[];if(n.hasOwnProperty("selectedOptions"))for(var i=n.selectedOptions,o=0;o<i.length;o++){var s=i.item(o),a=e._getOptionValue(s.value);r.push(a)}else for(i=n.options,o=0;o<i.length;o++){if((s=i.item(o)).selected){a=e._getOptionValue(s.value);r.push(a)}}e.value=r,t(r)}},t.prototype.registerOnTouched=function(t){this.onTouched=t},t.prototype.setDisabledState=function(t){this._renderer.setProperty(this._elementRef.nativeElement,"disabled",t)},t.prototype._registerOption=function(t){var e=(this._idCounter++).toString();return this._optionMap.set(e,t),e},t.prototype._getOptionId=function(t){for(var e=0,n=Array.from(this._optionMap.keys());e<n.length;e++){var r=n[e];if(this._compareWith(this._optionMap.get(r)._value,t))return r}return null},t.prototype._getOptionValue=function(t){var e=t.split(":")[0];return this._optionMap.has(e)?this._optionMap.get(e)._value:t},t.decorators=[{type:i.s,args:[{selector:"select[multiple][formControlName],select[multiple][formControl],select[multiple][ngModel]",host:{"(change)":"onChange($event.target)","(blur)":"onTouched()"},providers:[z]}]}],t.ctorParameters=function(){return[{type:i.U},{type:i.t}]},t.propDecorators={compareWith:[{type:i.C}]},t}(),H=function(){function t(t,e,n){this._element=t,this._renderer=e,this._select=n,this._select&&(this.id=this._select._registerOption(this))}return Object.defineProperty(t.prototype,"ngValue",{set:function(t){null!=this._select&&(this._value=t,this._setElementValue(B(this.id,t)),this._select.writeValue(this._select.value))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{set:function(t){this._select?(this._value=t,this._setElementValue(B(this.id,t)),this._select.writeValue(this._select.value)):this._setElementValue(t)},enumerable:!0,configurable:!0}),t.prototype._setElementValue=function(t){this._renderer.setProperty(this._element.nativeElement,"value",t)},t.prototype._setSelected=function(t){this._renderer.setProperty(this._element.nativeElement,"selected",t)},t.prototype.ngOnDestroy=function(){this._select&&(this._select._optionMap.delete(this.id),this._select.writeValue(this._select.value))},t.decorators=[{type:i.s,args:[{selector:"option"}]}],t.ctorParameters=function(){return[{type:i.t},{type:i.U},{type:q,decorators:[{type:i.M},{type:i.x}]}]},t.propDecorators={ngValue:[{type:i.C,args:["ngValue"]}],value:[{type:i.C,args:["value"]}]},t}();function U(t,e){return e.path.concat([t])}function W(t,e){var n,r,i,o,s;t||K(e,"Cannot find control with"),e.valueAccessor||K(e,"No value accessor for form control with"),t.validator=m.compose([t.validator,e.validator]),t.asyncValidator=m.composeAsync([t.asyncValidator,e.asyncValidator]),e.valueAccessor.writeValue(t.value),n=t,(r=e).valueAccessor.registerOnChange(function(t){n._pendingValue=t,n._pendingChange=!0,n._pendingDirty=!0,"change"===n.updateOn&&G(n,r)}),i=e,t.registerOnChange(function(t,e){i.valueAccessor.writeValue(t),e&&i.viewToModelUpdate(t)}),o=t,(s=e).valueAccessor.registerOnTouched(function(){o._pendingTouched=!0,"blur"===o.updateOn&&o._pendingChange&&G(o,s),"submit"!==o.updateOn&&o.markAsTouched()}),e.valueAccessor.setDisabledState&&t.registerOnDisabledChange(function(t){e.valueAccessor.setDisabledState(t)}),e._rawValidators.forEach(function(e){e.registerOnValidatorChange&&e.registerOnValidatorChange(function(){return t.updateValueAndValidity()})}),e._rawAsyncValidators.forEach(function(e){e.registerOnValidatorChange&&e.registerOnValidatorChange(function(){return t.updateValueAndValidity()})})}function G(t,e){e.viewToModelUpdate(t._pendingValue),t._pendingDirty&&t.markAsDirty(),t.setValue(t._pendingValue,{emitModelToViewChange:!1}),t._pendingChange=!1}function X(t,e){null==t&&K(e,"Cannot find control with"),t.validator=m.compose([t.validator,e.validator]),t.asyncValidator=m.composeAsync([t.asyncValidator,e.asyncValidator])}function Y(t){return K(t,"There is no FormControl instance attached to form control element with")}function K(t,e){var n;throw n=t.path.length>1?"path: '"+t.path.join(" -> ")+"'":t.path[0]?"name: '"+t.path+"'":"unspecified name attribute",new Error(e+" "+n)}function $(t){return null!=t?m.compose(t.map(C)):null}function Q(t){return null!=t?m.composeAsync(t.map(k)):null}function Z(t,e){if(!t.hasOwnProperty("model"))return!1;var n=t.model;return!!n.isFirstChange()||!Object(i._36)(e,n.currentValue)}var J=[j,R,S,F,q,D];function tt(t,e){t._syncPendingControls(),e.forEach(function(t){var e=t.control;"submit"===e.updateOn&&e._pendingChange&&(t.viewToModelUpdate(e._pendingValue),e._pendingChange=!1)})}function et(t,e){if(!e)return null;var n=void 0,r=void 0,i=void 0;return e.forEach(function(e){var o;e.constructor===O?n=e:(o=e,J.some(function(t){return o.constructor===t})?(r&&K(t,"More than one built-in value accessor matches form control with"),r=e):(i&&K(t,"More than one custom value accessor matches form control with"),i=e))}),i||(r||(n||(K(t,"No valid value accessor for form control with"),null)))}function nt(t,e){var n=t.indexOf(e);n>-1&&t.splice(n,1)}var rt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.ngOnInit=function(){this._checkParentType(),this.formDirective.addFormGroup(this)},e.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeFormGroup(this)},Object.defineProperty(e.prototype,"control",{get:function(){return this.formDirective.getFormGroup(this)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return U(this.name,this._parent)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return $(this._validators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return Q(this._asyncValidators)},enumerable:!0,configurable:!0}),e.prototype._checkParentType=function(){},e}(l),it=function(){function t(t){this._cd=t}return Object.defineProperty(t.prototype,"ngClassUntouched",{get:function(){return!!this._cd.control&&this._cd.control.untouched},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassTouched",{get:function(){return!!this._cd.control&&this._cd.control.touched},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassPristine",{get:function(){return!!this._cd.control&&this._cd.control.pristine},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassDirty",{get:function(){return!!this._cd.control&&this._cd.control.dirty},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassValid",{get:function(){return!!this._cd.control&&this._cd.control.valid},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassInvalid",{get:function(){return!!this._cd.control&&this._cd.control.invalid},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"ngClassPending",{get:function(){return!!this._cd.control&&this._cd.control.pending},enumerable:!0,configurable:!0}),t}(),ot={"[class.ng-untouched]":"ngClassUntouched","[class.ng-touched]":"ngClassTouched","[class.ng-pristine]":"ngClassPristine","[class.ng-dirty]":"ngClassDirty","[class.ng-valid]":"ngClassValid","[class.ng-invalid]":"ngClassInvalid","[class.ng-pending]":"ngClassPending"},st=function(t){function e(e){return t.call(this,e)||this}return Object(r.b)(e,t),e.decorators=[{type:i.s,args:[{selector:"[formControlName],[ngModel],[formControl]",host:ot}]}],e.ctorParameters=function(){return[{type:M,decorators:[{type:i.Z}]}]},e}(it),at=function(t){function e(e){return t.call(this,e)||this}return Object(r.b)(e,t),e.decorators=[{type:i.s,args:[{selector:"[formGroupName],[formArrayName],[ngModelGroup],[formGroup],form:not([ngNoForm]),[ngForm]",host:ot}]}],e.ctorParameters=function(){return[{type:l,decorators:[{type:i.Z}]}]},e}(it),ct="PENDING",ut="DISABLED";function lt(t){var e=dt(t)?t.validators:t;return Array.isArray(e)?$(e):e||null}function pt(t,e){var n=dt(e)?e.asyncValidators:t;return Array.isArray(n)?Q(n):n||null}function dt(t){return null!=t&&!Array.isArray(t)&&"object"==typeof t}var ht=function(){function t(t,e){this.validator=t,this.asyncValidator=e,this._onCollectionChange=function(){},this.pristine=!0,this.touched=!1,this._onDisabledChange=[]}return Object.defineProperty(t.prototype,"parent",{get:function(){return this._parent},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"valid",{get:function(){return"VALID"===this.status},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"invalid",{get:function(){return"INVALID"===this.status},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pending",{get:function(){return this.status==ct},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this.status===ut},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"enabled",{get:function(){return this.status!==ut},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dirty",{get:function(){return!this.pristine},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"untouched",{get:function(){return!this.touched},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"updateOn",{get:function(){return this._updateOn?this._updateOn:this.parent?this.parent.updateOn:"change"},enumerable:!0,configurable:!0}),t.prototype.setValidators=function(t){this.validator=lt(t)},t.prototype.setAsyncValidators=function(t){this.asyncValidator=pt(t)},t.prototype.clearValidators=function(){this.validator=null},t.prototype.clearAsyncValidators=function(){this.asyncValidator=null},t.prototype.markAsTouched=function(t){void 0===t&&(t={}),this.touched=!0,this._parent&&!t.onlySelf&&this._parent.markAsTouched(t)},t.prototype.markAsUntouched=function(t){void 0===t&&(t={}),this.touched=!1,this._pendingTouched=!1,this._forEachChild(function(t){t.markAsUntouched({onlySelf:!0})}),this._parent&&!t.onlySelf&&this._parent._updateTouched(t)},t.prototype.markAsDirty=function(t){void 0===t&&(t={}),this.pristine=!1,this._parent&&!t.onlySelf&&this._parent.markAsDirty(t)},t.prototype.markAsPristine=function(t){void 0===t&&(t={}),this.pristine=!0,this._pendingDirty=!1,this._forEachChild(function(t){t.markAsPristine({onlySelf:!0})}),this._parent&&!t.onlySelf&&this._parent._updatePristine(t)},t.prototype.markAsPending=function(t){void 0===t&&(t={}),this.status=ct,this._parent&&!t.onlySelf&&this._parent.markAsPending(t)},t.prototype.disable=function(t){void 0===t&&(t={}),this.status=ut,this.errors=null,this._forEachChild(function(t){t.disable({onlySelf:!0})}),this._updateValue(),!1!==t.emitEvent&&(this.valueChanges.emit(this.value),this.statusChanges.emit(this.status)),this._updateAncestors(!!t.onlySelf),this._onDisabledChange.forEach(function(t){return t(!0)})},t.prototype.enable=function(t){void 0===t&&(t={}),this.status="VALID",this._forEachChild(function(t){t.enable({onlySelf:!0})}),this.updateValueAndValidity({onlySelf:!0,emitEvent:t.emitEvent}),this._updateAncestors(!!t.onlySelf),this._onDisabledChange.forEach(function(t){return t(!1)})},t.prototype._updateAncestors=function(t){this._parent&&!t&&(this._parent.updateValueAndValidity(),this._parent._updatePristine(),this._parent._updateTouched())},t.prototype.setParent=function(t){this._parent=t},t.prototype.updateValueAndValidity=function(t){void 0===t&&(t={}),this._setInitialStatus(),this._updateValue(),this.enabled&&(this._cancelExistingSubscription(),this.errors=this._runValidator(),this.status=this._calculateStatus(),"VALID"!==this.status&&this.status!==ct||this._runAsyncValidator(t.emitEvent)),!1!==t.emitEvent&&(this.valueChanges.emit(this.value),this.statusChanges.emit(this.status)),this._parent&&!t.onlySelf&&this._parent.updateValueAndValidity(t)},t.prototype._updateTreeValidity=function(t){void 0===t&&(t={emitEvent:!0}),this._forEachChild(function(e){return e._updateTreeValidity(t)}),this.updateValueAndValidity({onlySelf:!0,emitEvent:t.emitEvent})},t.prototype._setInitialStatus=function(){this.status=this._allControlsDisabled()?ut:"VALID"},t.prototype._runValidator=function(){return this.validator?this.validator(this):null},t.prototype._runAsyncValidator=function(t){var e=this;if(this.asyncValidator){this.status=ct;var n=v(this.asyncValidator(this));this._asyncValidationSubscription=n.subscribe(function(n){return e.setErrors(n,{emitEvent:t})})}},t.prototype._cancelExistingSubscription=function(){this._asyncValidationSubscription&&this._asyncValidationSubscription.unsubscribe()},t.prototype.setErrors=function(t,e){void 0===e&&(e={}),this.errors=t,this._updateControlsErrors(!1!==e.emitEvent)},t.prototype.get=function(t){return e=this,r=".",null==(n=t)?null:(n instanceof Array||(n=n.split(r)),n instanceof Array&&0===n.length?null:n.reduce(function(t,e){return t instanceof mt?t.controls[e]||null:t instanceof gt?t.at(e)||null:null},e));var e,n,r},t.prototype.getError=function(t,e){var n=e?this.get(e):this;return n&&n.errors?n.errors[t]:null},t.prototype.hasError=function(t,e){return!!this.getError(t,e)},Object.defineProperty(t.prototype,"root",{get:function(){for(var t=this;t._parent;)t=t._parent;return t},enumerable:!0,configurable:!0}),t.prototype._updateControlsErrors=function(t){this.status=this._calculateStatus(),t&&this.statusChanges.emit(this.status),this._parent&&this._parent._updateControlsErrors(t)},t.prototype._initObservables=function(){this.valueChanges=new i.w,this.statusChanges=new i.w},t.prototype._calculateStatus=function(){return this._allControlsDisabled()?ut:this.errors?"INVALID":this._anyControlsHaveStatus(ct)?ct:this._anyControlsHaveStatus("INVALID")?"INVALID":"VALID"},t.prototype._anyControlsHaveStatus=function(t){return this._anyControls(function(e){return e.status===t})},t.prototype._anyControlsDirty=function(){return this._anyControls(function(t){return t.dirty})},t.prototype._anyControlsTouched=function(){return this._anyControls(function(t){return t.touched})},t.prototype._updatePristine=function(t){void 0===t&&(t={}),this.pristine=!this._anyControlsDirty(),this._parent&&!t.onlySelf&&this._parent._updatePristine(t)},t.prototype._updateTouched=function(t){void 0===t&&(t={}),this.touched=this._anyControlsTouched(),this._parent&&!t.onlySelf&&this._parent._updateTouched(t)},t.prototype._isBoxedValue=function(t){return"object"==typeof t&&null!==t&&2===Object.keys(t).length&&"value"in t&&"disabled"in t},t.prototype._registerOnCollectionChange=function(t){this._onCollectionChange=t},t.prototype._setUpdateStrategy=function(t){dt(t)&&null!=t.updateOn&&(this._updateOn=t.updateOn)},t}(),ft=function(t){function e(e,n,r){void 0===e&&(e=null);var i=t.call(this,lt(n),pt(r,n))||this;return i._onChange=[],i._applyFormState(e),i._setUpdateStrategy(n),i.updateValueAndValidity({onlySelf:!0,emitEvent:!1}),i._initObservables(),i}return Object(r.b)(e,t),e.prototype.setValue=function(t,e){var n=this;void 0===e&&(e={}),this.value=this._pendingValue=t,this._onChange.length&&!1!==e.emitModelToViewChange&&this._onChange.forEach(function(t){return t(n.value,!1!==e.emitViewToModelChange)}),this.updateValueAndValidity(e)},e.prototype.patchValue=function(t,e){void 0===e&&(e={}),this.setValue(t,e)},e.prototype.reset=function(t,e){void 0===t&&(t=null),void 0===e&&(e={}),this._applyFormState(t),this.markAsPristine(e),this.markAsUntouched(e),this.setValue(this.value,e),this._pendingChange=!1},e.prototype._updateValue=function(){},e.prototype._anyControls=function(t){return!1},e.prototype._allControlsDisabled=function(){return this.disabled},e.prototype.registerOnChange=function(t){this._onChange.push(t)},e.prototype._clearChangeFns=function(){this._onChange=[],this._onDisabledChange=[],this._onCollectionChange=function(){}},e.prototype.registerOnDisabledChange=function(t){this._onDisabledChange.push(t)},e.prototype._forEachChild=function(t){},e.prototype._syncPendingControls=function(){return!("submit"!==this.updateOn||(this._pendingDirty&&this.markAsDirty(),this._pendingTouched&&this.markAsTouched(),!this._pendingChange))&&(this.setValue(this._pendingValue,{onlySelf:!0,emitModelToViewChange:!1}),!0)},e.prototype._applyFormState=function(t){this._isBoxedValue(t)?(this.value=this._pendingValue=t.value,t.disabled?this.disable({onlySelf:!0,emitEvent:!1}):this.enable({onlySelf:!0,emitEvent:!1})):this.value=this._pendingValue=t},e}(ht),mt=function(t){function e(e,n,r){var i=t.call(this,lt(n),pt(r,n))||this;return i.controls=e,i._initObservables(),i._setUpdateStrategy(n),i._setUpControls(),i.updateValueAndValidity({onlySelf:!0,emitEvent:!1}),i}return Object(r.b)(e,t),e.prototype.registerControl=function(t,e){return this.controls[t]?this.controls[t]:(this.controls[t]=e,e.setParent(this),e._registerOnCollectionChange(this._onCollectionChange),e)},e.prototype.addControl=function(t,e){this.registerControl(t,e),this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.removeControl=function(t){this.controls[t]&&this.controls[t]._registerOnCollectionChange(function(){}),delete this.controls[t],this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.setControl=function(t,e){this.controls[t]&&this.controls[t]._registerOnCollectionChange(function(){}),delete this.controls[t],e&&this.registerControl(t,e),this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.contains=function(t){return this.controls.hasOwnProperty(t)&&this.controls[t].enabled},e.prototype.setValue=function(t,e){var n=this;void 0===e&&(e={}),this._checkAllValuesPresent(t),Object.keys(t).forEach(function(r){n._throwIfControlMissing(r),n.controls[r].setValue(t[r],{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e)},e.prototype.patchValue=function(t,e){var n=this;void 0===e&&(e={}),Object.keys(t).forEach(function(r){n.controls[r]&&n.controls[r].patchValue(t[r],{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e)},e.prototype.reset=function(t,e){void 0===t&&(t={}),void 0===e&&(e={}),this._forEachChild(function(n,r){n.reset(t[r],{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e),this._updatePristine(e),this._updateTouched(e)},e.prototype.getRawValue=function(){return this._reduceChildren({},function(t,e,n){return t[n]=e instanceof ft?e.value:e.getRawValue(),t})},e.prototype._syncPendingControls=function(){var t=this._reduceChildren(!1,function(t,e){return!!e._syncPendingControls()||t});return t&&this.updateValueAndValidity({onlySelf:!0}),t},e.prototype._throwIfControlMissing=function(t){if(!Object.keys(this.controls).length)throw new Error("\n        There are no form controls registered with this group yet.  If you're using ngModel,\n        you may want to check next tick (e.g. use setTimeout).\n      ");if(!this.controls[t])throw new Error("Cannot find form control with name: "+t+".")},e.prototype._forEachChild=function(t){var e=this;Object.keys(this.controls).forEach(function(n){return t(e.controls[n],n)})},e.prototype._setUpControls=function(){var t=this;this._forEachChild(function(e){e.setParent(t),e._registerOnCollectionChange(t._onCollectionChange)})},e.prototype._updateValue=function(){this.value=this._reduceValue()},e.prototype._anyControls=function(t){var e=this,n=!1;return this._forEachChild(function(r,i){n=n||e.contains(i)&&t(r)}),n},e.prototype._reduceValue=function(){var t=this;return this._reduceChildren({},function(e,n,r){return(n.enabled||t.disabled)&&(e[r]=n.value),e})},e.prototype._reduceChildren=function(t,e){var n=t;return this._forEachChild(function(t,r){n=e(n,t,r)}),n},e.prototype._allControlsDisabled=function(){for(var t=0,e=Object.keys(this.controls);t<e.length;t++){var n=e[t];if(this.controls[n].enabled)return!1}return Object.keys(this.controls).length>0||this.disabled},e.prototype._checkAllValuesPresent=function(t){this._forEachChild(function(e,n){if(void 0===t[n])throw new Error("Must supply a value for form control with name: '"+n+"'.")})},e}(ht),gt=function(t){function e(e,n,r){var i=t.call(this,lt(n),pt(r,n))||this;return i.controls=e,i._initObservables(),i._setUpdateStrategy(n),i._setUpControls(),i.updateValueAndValidity({onlySelf:!0,emitEvent:!1}),i}return Object(r.b)(e,t),e.prototype.at=function(t){return this.controls[t]},e.prototype.push=function(t){this.controls.push(t),this._registerControl(t),this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.insert=function(t,e){this.controls.splice(t,0,e),this._registerControl(e),this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.removeAt=function(t){this.controls[t]&&this.controls[t]._registerOnCollectionChange(function(){}),this.controls.splice(t,1),this.updateValueAndValidity(),this._onCollectionChange()},e.prototype.setControl=function(t,e){this.controls[t]&&this.controls[t]._registerOnCollectionChange(function(){}),this.controls.splice(t,1),e&&(this.controls.splice(t,0,e),this._registerControl(e)),this.updateValueAndValidity(),this._onCollectionChange()},Object.defineProperty(e.prototype,"length",{get:function(){return this.controls.length},enumerable:!0,configurable:!0}),e.prototype.setValue=function(t,e){var n=this;void 0===e&&(e={}),this._checkAllValuesPresent(t),t.forEach(function(t,r){n._throwIfControlMissing(r),n.at(r).setValue(t,{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e)},e.prototype.patchValue=function(t,e){var n=this;void 0===e&&(e={}),t.forEach(function(t,r){n.at(r)&&n.at(r).patchValue(t,{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e)},e.prototype.reset=function(t,e){void 0===t&&(t=[]),void 0===e&&(e={}),this._forEachChild(function(n,r){n.reset(t[r],{onlySelf:!0,emitEvent:e.emitEvent})}),this.updateValueAndValidity(e),this._updatePristine(e),this._updateTouched(e)},e.prototype.getRawValue=function(){return this.controls.map(function(t){return t instanceof ft?t.value:t.getRawValue()})},e.prototype._syncPendingControls=function(){var t=this.controls.reduce(function(t,e){return!!e._syncPendingControls()||t},!1);return t&&this.updateValueAndValidity({onlySelf:!0}),t},e.prototype._throwIfControlMissing=function(t){if(!this.controls.length)throw new Error("\n        There are no form controls registered with this array yet.  If you're using ngModel,\n        you may want to check next tick (e.g. use setTimeout).\n      ");if(!this.at(t))throw new Error("Cannot find form control at index "+t)},e.prototype._forEachChild=function(t){this.controls.forEach(function(e,n){t(e,n)})},e.prototype._updateValue=function(){var t=this;this.value=this.controls.filter(function(e){return e.enabled||t.disabled}).map(function(t){return t.value})},e.prototype._anyControls=function(t){return this.controls.some(function(e){return e.enabled&&t(e)})},e.prototype._setUpControls=function(){var t=this;this._forEachChild(function(e){return t._registerControl(e)})},e.prototype._checkAllValuesPresent=function(t){this._forEachChild(function(e,n){if(void 0===t[n])throw new Error("Must supply a value for form control at index: "+n+".")})},e.prototype._allControlsDisabled=function(){for(var t=0,e=this.controls;t<e.length;t++){if(e[t].enabled)return!1}return this.controls.length>0||this.disabled},e.prototype._registerControl=function(t){t.setParent(this),t._registerOnCollectionChange(this._onCollectionChange)},e}(ht),vt={provide:l,useExisting:Object(i._14)(function(){return bt})},yt=Promise.resolve(null),bt=function(t){function e(e,n){var r=t.call(this)||this;return r.submitted=!1,r._directives=[],r.ngSubmit=new i.w,r.form=new mt({},$(e),Q(n)),r}return Object(r.b)(e,t),e.prototype.ngAfterViewInit=function(){this._setUpdateStrategy()},Object.defineProperty(e.prototype,"formDirective",{get:function(){return this},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"control",{get:function(){return this.form},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return[]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"controls",{get:function(){return this.form.controls},enumerable:!0,configurable:!0}),e.prototype.addControl=function(t){var e=this;yt.then(function(){var n=e._findContainer(t.path);t.control=n.registerControl(t.name,t.control),W(t.control,t),t.control.updateValueAndValidity({emitEvent:!1}),e._directives.push(t)})},e.prototype.getControl=function(t){return this.form.get(t.path)},e.prototype.removeControl=function(t){var e=this;yt.then(function(){var n=e._findContainer(t.path);n&&n.removeControl(t.name),nt(e._directives,t)})},e.prototype.addFormGroup=function(t){var e=this;yt.then(function(){var n=e._findContainer(t.path),r=new mt({});X(r,t),n.registerControl(t.name,r),r.updateValueAndValidity({emitEvent:!1})})},e.prototype.removeFormGroup=function(t){var e=this;yt.then(function(){var n=e._findContainer(t.path);n&&n.removeControl(t.name)})},e.prototype.getFormGroup=function(t){return this.form.get(t.path)},e.prototype.updateModel=function(t,e){var n=this;yt.then(function(){n.form.get(t.path).setValue(e)})},e.prototype.setValue=function(t){this.control.setValue(t)},e.prototype.onSubmit=function(t){return this.submitted=!0,tt(this.form,this._directives),this.ngSubmit.emit(t),!1},e.prototype.onReset=function(){this.resetForm()},e.prototype.resetForm=function(t){void 0===t&&(t=void 0),this.form.reset(t),this.submitted=!1},e.prototype._setUpdateStrategy=function(){this.options&&null!=this.options.updateOn&&(this.form._updateOn=this.options.updateOn)},e.prototype._findContainer=function(t){return t.pop(),t.length?this.form.get(t):this.form},e.decorators=[{type:i.s,args:[{selector:"form:not([ngNoForm]):not([formGroup]),ngForm,[ngForm]",providers:[vt],host:{"(submit)":"onSubmit($event)","(reset)":"onReset()"},outputs:["ngSubmit"],exportAs:"ngForm"}]}],e.ctorParameters=function(){return[{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[d]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[h]}]}]},e.propDecorators={options:[{type:i.C,args:["ngFormOptions"]}]},e}(l),_t='\n    <div [formGroup]="myGroup">\n      <input formControlName="firstName">\n    </div>\n\n    In your class:\n\n    this.myGroup = new FormGroup({\n       firstName: new FormControl()\n    });',jt='\n    <div [formGroup]="myGroup">\n       <div formGroupName="person">\n          <input formControlName="firstName">\n       </div>\n    </div>\n\n    In your class:\n\n    this.myGroup = new FormGroup({\n       person: new FormGroup({ firstName: new FormControl() })\n    });',xt='\n    <div [formGroup]="myGroup">\n      <div formArrayName="cities">\n        <div *ngFor="let city of cityArray.controls; index as i">\n          <input [formControlName]="i">\n        </div>\n      </div>\n    </div>\n\n    In your class:\n\n    this.cityArray = new FormArray([new FormControl(\'SF\')]);\n    this.myGroup = new FormGroup({\n      cities: this.cityArray\n    });',wt='\n    <form>\n       <div ngModelGroup="person">\n          <input [(ngModel)]="person.name" name="firstName">\n       </div>\n    </form>',Ot='\n    <div [formGroup]="myGroup">\n       <input formControlName="firstName">\n       <input [(ngModel)]="showMoreControls" [ngModelOptions]="{standalone: true}">\n    </div>\n  ',Ct=function(){function t(){}return t.modelParentException=function(){throw new Error('\n      ngModel cannot be used to register form controls with a parent formGroup directive.  Try using\n      formGroup\'s partner directive "formControlName" instead.  Example:\n\n      '+_t+"\n\n      Or, if you'd like to avoid registering this form control, indicate that it's standalone in ngModelOptions:\n\n      Example:\n\n      "+Ot)},t.formGroupNameException=function(){throw new Error("\n      ngModel cannot be used to register form controls with a parent formGroupName or formArrayName directive.\n\n      Option 1: Use formControlName instead of ngModel (reactive strategy):\n\n      "+jt+"\n\n      Option 2:  Update ngModel's parent be ngModelGroup (template-driven strategy):\n\n      "+wt)},t.missingNameException=function(){throw new Error('If ngModel is used within a form tag, either the name attribute must be set or the form\n      control must be defined as \'standalone\' in ngModelOptions.\n\n      Example 1: <input [(ngModel)]="person.firstName" name="first">\n      Example 2: <input [(ngModel)]="person.firstName" [ngModelOptions]="{standalone: true}">')},t.modelGroupParentException=function(){throw new Error("\n      ngModelGroup cannot be used with a parent formGroup directive.\n\n      Option 1: Use formGroupName instead of ngModelGroup (reactive strategy):\n\n      "+jt+"\n\n      Option 2:  Use a regular form tag instead of the formGroup directive (template-driven strategy):\n\n      "+wt)},t}(),kt={provide:l,useExisting:Object(i._14)(function(){return Et})},Et=function(t){function e(e,n,r){var i=t.call(this)||this;return i._parent=e,i._validators=n,i._asyncValidators=r,i}return Object(r.b)(e,t),e.prototype._checkParentType=function(){this._parent instanceof e||this._parent instanceof bt||Ct.modelGroupParentException()},e.decorators=[{type:i.s,args:[{selector:"[ngModelGroup]",providers:[kt],exportAs:"ngModelGroup"}]}],e.ctorParameters=function(){return[{type:l,decorators:[{type:i.x},{type:i._1}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[d]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[h]}]}]},e.propDecorators={name:[{type:i.C,args:["ngModelGroup"]}]},e}(rt),St={provide:M,useExisting:Object(i._14)(function(){return Mt})},Pt=Promise.resolve(null),Mt=function(t){function e(e,n,r,o){var s=t.call(this)||this;return s.control=new ft,s._registered=!1,s.update=new i.w,s._parent=e,s._rawValidators=n||[],s._rawAsyncValidators=r||[],s.valueAccessor=et(s,o),s}return Object(r.b)(e,t),e.prototype.ngOnChanges=function(t){this._checkForErrors(),this._registered||this._setUpControl(),"isDisabled"in t&&this._updateDisabled(t),Z(t,this.viewModel)&&(this._updateValue(this.model),this.viewModel=this.model)},e.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeControl(this)},Object.defineProperty(e.prototype,"path",{get:function(){return this._parent?U(this.name,this._parent):[this.name]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return $(this._rawValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return Q(this._rawAsyncValidators)},enumerable:!0,configurable:!0}),e.prototype.viewToModelUpdate=function(t){this.viewModel=t,this.update.emit(t)},e.prototype._setUpControl=function(){this._setUpdateStrategy(),this._isStandalone()?this._setUpStandalone():this.formDirective.addControl(this),this._registered=!0},e.prototype._setUpdateStrategy=function(){this.options&&null!=this.options.updateOn&&(this.control._updateOn=this.options.updateOn)},e.prototype._isStandalone=function(){return!this._parent||!(!this.options||!this.options.standalone)},e.prototype._setUpStandalone=function(){W(this.control,this),this.control.updateValueAndValidity({emitEvent:!1})},e.prototype._checkForErrors=function(){this._isStandalone()||this._checkParentType(),this._checkName()},e.prototype._checkParentType=function(){!(this._parent instanceof Et)&&this._parent instanceof rt?Ct.formGroupNameException():this._parent instanceof Et||this._parent instanceof bt||Ct.modelParentException()},e.prototype._checkName=function(){this.options&&this.options.name&&(this.name=this.options.name),this._isStandalone()||this.name||Ct.missingNameException()},e.prototype._updateValue=function(t){var e=this;Pt.then(function(){e.control.setValue(t,{emitViewToModelChange:!1})})},e.prototype._updateDisabled=function(t){var e=this,n=t.isDisabled.currentValue,r=""===n||n&&"false"!==n;Pt.then(function(){r&&!e.control.disabled?e.control.disable():!r&&e.control.disabled&&e.control.enable()})},e.decorators=[{type:i.s,args:[{selector:"[ngModel]:not([formControlName]):not([formControl])",providers:[St],exportAs:"ngModel"}]}],e.ctorParameters=function(){return[{type:l,decorators:[{type:i.M},{type:i.x}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[d]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[h]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[b]}]}]},e.propDecorators={name:[{type:i.C}],isDisabled:[{type:i.C,args:["disabled"]}],model:[{type:i.C,args:["ngModel"]}],options:[{type:i.C,args:["ngModelOptions"]}],update:[{type:i.N,args:["ngModelChange"]}]},e}(M),At=function(){function t(){}return t.controlParentException=function(){throw new Error("formControlName must be used with a parent formGroup directive.  You'll want to add a formGroup\n       directive and pass it an existing FormGroup instance (you can create one in your class).\n\n      Example:\n\n      "+_t)},t.ngModelGroupException=function(){throw new Error('formControlName cannot be used with an ngModelGroup parent. It is only compatible with parents\n       that also have a "form" prefix: formGroupName, formArrayName, or formGroup.\n\n       Option 1:  Update the parent to be formGroupName (reactive form strategy)\n\n        '+jt+"\n\n        Option 2: Use ngModel instead of formControlName (template-driven strategy)\n\n        "+wt)},t.missingFormException=function(){throw new Error("formGroup expects a FormGroup instance. Please pass one in.\n\n       Example:\n\n       "+_t)},t.groupParentException=function(){throw new Error("formGroupName must be used with a parent formGroup directive.  You'll want to add a formGroup\n      directive and pass it an existing FormGroup instance (you can create one in your class).\n\n      Example:\n\n      "+jt)},t.arrayParentException=function(){throw new Error("formArrayName must be used with a parent formGroup directive.  You'll want to add a formGroup\n       directive and pass it an existing FormGroup instance (you can create one in your class).\n\n        Example:\n\n        "+xt)},t.disabledAttrWarning=function(){console.warn("\n      It looks like you're using the disabled attribute with a reactive form directive. If you set disabled to true\n      when you set up this control in your component class, the disabled attribute will actually be set in the DOM for\n      you. We recommend using this approach to avoid 'changed after checked' errors.\n       \n      Example: \n      form = new FormGroup({\n        first: new FormControl({value: 'Nancy', disabled: true}, Validators.required),\n        last: new FormControl('Drew', Validators.required)\n      });\n    ")},t}(),Tt={provide:M,useExisting:Object(i._14)(function(){return Dt})},Dt=function(t){function e(e,n,r){var o=t.call(this)||this;return o.update=new i.w,o._rawValidators=e||[],o._rawAsyncValidators=n||[],o.valueAccessor=et(o,r),o}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"isDisabled",{set:function(t){At.disabledAttrWarning()},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){this._isControlChanged(t)&&(W(this.form,this),this.control.disabled&&this.valueAccessor.setDisabledState&&this.valueAccessor.setDisabledState(!0),this.form.updateValueAndValidity({emitEvent:!1})),Z(t,this.viewModel)&&(this.form.setValue(this.model),this.viewModel=this.model)},Object.defineProperty(e.prototype,"path",{get:function(){return[]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return $(this._rawValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return Q(this._rawAsyncValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"control",{get:function(){return this.form},enumerable:!0,configurable:!0}),e.prototype.viewToModelUpdate=function(t){this.viewModel=t,this.update.emit(t)},e.prototype._isControlChanged=function(t){return t.hasOwnProperty("form")},e.decorators=[{type:i.s,args:[{selector:"[formControl]",providers:[Tt],exportAs:"ngForm"}]}],e.ctorParameters=function(){return[{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[d]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[h]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[b]}]}]},e.propDecorators={form:[{type:i.C,args:["formControl"]}],model:[{type:i.C,args:["ngModel"]}],update:[{type:i.N,args:["ngModelChange"]}],isDisabled:[{type:i.C,args:["disabled"]}]},e}(M),It={provide:l,useExisting:Object(i._14)(function(){return Rt})},Rt=function(t){function e(e,n){var r=t.call(this)||this;return r._validators=e,r._asyncValidators=n,r.submitted=!1,r.directives=[],r.form=null,r.ngSubmit=new i.w,r}return Object(r.b)(e,t),e.prototype.ngOnChanges=function(t){this._checkFormPresent(),t.hasOwnProperty("form")&&(this._updateValidators(),this._updateDomValue(),this._updateRegistrations())},Object.defineProperty(e.prototype,"formDirective",{get:function(){return this},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"control",{get:function(){return this.form},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return[]},enumerable:!0,configurable:!0}),e.prototype.addControl=function(t){var e=this.form.get(t.path);return W(e,t),e.updateValueAndValidity({emitEvent:!1}),this.directives.push(t),e},e.prototype.getControl=function(t){return this.form.get(t.path)},e.prototype.removeControl=function(t){nt(this.directives,t)},e.prototype.addFormGroup=function(t){var e=this.form.get(t.path);X(e,t),e.updateValueAndValidity({emitEvent:!1})},e.prototype.removeFormGroup=function(t){},e.prototype.getFormGroup=function(t){return this.form.get(t.path)},e.prototype.addFormArray=function(t){var e=this.form.get(t.path);X(e,t),e.updateValueAndValidity({emitEvent:!1})},e.prototype.removeFormArray=function(t){},e.prototype.getFormArray=function(t){return this.form.get(t.path)},e.prototype.updateModel=function(t,e){this.form.get(t.path).setValue(e)},e.prototype.onSubmit=function(t){return this.submitted=!0,tt(this.form,this.directives),this.ngSubmit.emit(t),!1},e.prototype.onReset=function(){this.resetForm()},e.prototype.resetForm=function(t){void 0===t&&(t=void 0),this.form.reset(t),this.submitted=!1},e.prototype._updateDomValue=function(){var t=this;this.directives.forEach(function(e){var n,r,i=t.form.get(e.path);e.control!==i&&(n=e.control,(r=e).valueAccessor.registerOnChange(function(){return Y(r)}),r.valueAccessor.registerOnTouched(function(){return Y(r)}),r._rawValidators.forEach(function(t){t.registerOnValidatorChange&&t.registerOnValidatorChange(null)}),r._rawAsyncValidators.forEach(function(t){t.registerOnValidatorChange&&t.registerOnValidatorChange(null)}),n&&n._clearChangeFns(),i&&W(i,e),e.control=i)}),this.form._updateTreeValidity({emitEvent:!1})},e.prototype._updateRegistrations=function(){var t=this;this.form._registerOnCollectionChange(function(){return t._updateDomValue()}),this._oldForm&&this._oldForm._registerOnCollectionChange(function(){}),this._oldForm=this.form},e.prototype._updateValidators=function(){var t=$(this._validators);this.form.validator=m.compose([this.form.validator,t]);var e=Q(this._asyncValidators);this.form.asyncValidator=m.composeAsync([this.form.asyncValidator,e])},e.prototype._checkFormPresent=function(){this.form||At.missingFormException()},e.decorators=[{type:i.s,args:[{selector:"[formGroup]",providers:[It],host:{"(submit)":"onSubmit($event)","(reset)":"onReset()"},exportAs:"ngForm"}]}],e.ctorParameters=function(){return[{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[d]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[h]}]}]},e.propDecorators={form:[{type:i.C,args:["formGroup"]}],ngSubmit:[{type:i.N}]},e}(l),Nt={provide:l,useExisting:Object(i._14)(function(){return Lt})},Lt=function(t){function e(e,n,r){var i=t.call(this)||this;return i._parent=e,i._validators=n,i._asyncValidators=r,i}return Object(r.b)(e,t),e.prototype._checkParentType=function(){zt(this._parent)&&At.groupParentException()},e.decorators=[{type:i.s,args:[{selector:"[formGroupName]",providers:[Nt]}]}],e.ctorParameters=function(){return[{type:l,decorators:[{type:i.M},{type:i.x},{type:i._1}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[d]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[h]}]}]},e.propDecorators={name:[{type:i.C,args:["formGroupName"]}]},e}(rt),Ft={provide:l,useExisting:Object(i._14)(function(){return Vt})},Vt=function(t){function e(e,n,r){var i=t.call(this)||this;return i._parent=e,i._validators=n,i._asyncValidators=r,i}return Object(r.b)(e,t),e.prototype.ngOnInit=function(){this._checkParentType(),this.formDirective.addFormArray(this)},e.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeFormArray(this)},Object.defineProperty(e.prototype,"control",{get:function(){return this.formDirective.getFormArray(this)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"path",{get:function(){return U(this.name,this._parent)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return $(this._validators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return Q(this._asyncValidators)},enumerable:!0,configurable:!0}),e.prototype._checkParentType=function(){zt(this._parent)&&At.arrayParentException()},e.decorators=[{type:i.s,args:[{selector:"[formArrayName]",providers:[Ft]}]}],e.ctorParameters=function(){return[{type:l,decorators:[{type:i.M},{type:i.x},{type:i._1}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[d]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[h]}]}]},e.propDecorators={name:[{type:i.C,args:["formArrayName"]}]},e}(l);function zt(t){return!(t instanceof Lt||t instanceof Rt||t instanceof Vt)}var Bt={provide:M,useExisting:Object(i._14)(function(){return qt})},qt=function(t){function e(e,n,r,o){var s=t.call(this)||this;return s._added=!1,s.update=new i.w,s._parent=e,s._rawValidators=n||[],s._rawAsyncValidators=r||[],s.valueAccessor=et(s,o),s}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"isDisabled",{set:function(t){At.disabledAttrWarning()},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){this._added||this._setUpControl(),Z(t,this.viewModel)&&(this.viewModel=this.model,this.formDirective.updateModel(this,this.model))},e.prototype.ngOnDestroy=function(){this.formDirective&&this.formDirective.removeControl(this)},e.prototype.viewToModelUpdate=function(t){this.viewModel=t,this.update.emit(t)},Object.defineProperty(e.prototype,"path",{get:function(){return U(this.name,this._parent)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"formDirective",{get:function(){return this._parent?this._parent.formDirective:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"validator",{get:function(){return $(this._rawValidators)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"asyncValidator",{get:function(){return Q(this._rawAsyncValidators)},enumerable:!0,configurable:!0}),e.prototype._checkParentType=function(){!(this._parent instanceof Lt)&&this._parent instanceof rt?At.ngModelGroupException():this._parent instanceof Lt||this._parent instanceof Rt||this._parent instanceof Vt||At.controlParentException()},e.prototype._setUpControl=function(){this._checkParentType(),this.control=this.formDirective.addControl(this),this.control.disabled&&this.valueAccessor.setDisabledState&&this.valueAccessor.setDisabledState(!0),this._added=!0},e.decorators=[{type:i.s,args:[{selector:"[formControlName]",providers:[Bt]}]}],e.ctorParameters=function(){return[{type:l,decorators:[{type:i.M},{type:i.x},{type:i._1}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[d]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[h]}]},{type:Array,decorators:[{type:i.M},{type:i.Z},{type:i.y,args:[b]}]}]},e.propDecorators={name:[{type:i.C,args:["formControlName"]}],model:[{type:i.C,args:["ngModel"]}],update:[{type:i.N,args:["ngModelChange"]}],isDisabled:[{type:i.C,args:["disabled"]}]},e}(M),Ht={provide:d,useExisting:Object(i._14)(function(){return Wt}),multi:!0},Ut={provide:d,useExisting:Object(i._14)(function(){return Gt}),multi:!0},Wt=function(){function t(){}return Object.defineProperty(t.prototype,"required",{get:function(){return this._required},set:function(t){this._required=null!=t&&!1!==t&&""+t!="false",this._onChange&&this._onChange()},enumerable:!0,configurable:!0}),t.prototype.validate=function(t){return this.required?m.required(t):null},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},t.decorators=[{type:i.s,args:[{selector:":not([type=checkbox])[required][formControlName],:not([type=checkbox])[required][formControl],:not([type=checkbox])[required][ngModel]",providers:[Ht],host:{"[attr.required]":'required ? "" : null'}}]}],t.ctorParameters=function(){return[]},t.propDecorators={required:[{type:i.C}]},t}(),Gt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.prototype.validate=function(t){return this.required?m.requiredTrue(t):null},e.decorators=[{type:i.s,args:[{selector:"input[type=checkbox][required][formControlName],input[type=checkbox][required][formControl],input[type=checkbox][required][ngModel]",providers:[Ut],host:{"[attr.required]":'required ? "" : null'}}]}],e.ctorParameters=function(){return[]},e}(Wt),Xt={provide:d,useExisting:Object(i._14)(function(){return Yt}),multi:!0},Yt=function(){function t(){}return Object.defineProperty(t.prototype,"email",{set:function(t){this._enabled=""===t||!0===t||"true"===t,this._onChange&&this._onChange()},enumerable:!0,configurable:!0}),t.prototype.validate=function(t){return this._enabled?m.email(t):null},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},t.decorators=[{type:i.s,args:[{selector:"[email][formControlName],[email][formControl],[email][ngModel]",providers:[Xt]}]}],t.ctorParameters=function(){return[]},t.propDecorators={email:[{type:i.C}]},t}(),Kt={provide:d,useExisting:Object(i._14)(function(){return $t}),multi:!0},$t=function(){function t(){}return t.prototype.ngOnChanges=function(t){"minlength"in t&&(this._createValidator(),this._onChange&&this._onChange())},t.prototype.validate=function(t){return null==this.minlength?null:this._validator(t)},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},t.prototype._createValidator=function(){this._validator=m.minLength(parseInt(this.minlength,10))},t.decorators=[{type:i.s,args:[{selector:"[minlength][formControlName],[minlength][formControl],[minlength][ngModel]",providers:[Kt],host:{"[attr.minlength]":"minlength ? minlength : null"}}]}],t.ctorParameters=function(){return[]},t.propDecorators={minlength:[{type:i.C}]},t}(),Qt={provide:d,useExisting:Object(i._14)(function(){return Zt}),multi:!0},Zt=function(){function t(){}return t.prototype.ngOnChanges=function(t){"maxlength"in t&&(this._createValidator(),this._onChange&&this._onChange())},t.prototype.validate=function(t){return null!=this.maxlength?this._validator(t):null},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},t.prototype._createValidator=function(){this._validator=m.maxLength(parseInt(this.maxlength,10))},t.decorators=[{type:i.s,args:[{selector:"[maxlength][formControlName],[maxlength][formControl],[maxlength][ngModel]",providers:[Qt],host:{"[attr.maxlength]":"maxlength ? maxlength : null"}}]}],t.ctorParameters=function(){return[]},t.propDecorators={maxlength:[{type:i.C}]},t}(),Jt={provide:d,useExisting:Object(i._14)(function(){return te}),multi:!0},te=function(){function t(){}return t.prototype.ngOnChanges=function(t){"pattern"in t&&(this._createValidator(),this._onChange&&this._onChange())},t.prototype.validate=function(t){return this._validator(t)},t.prototype.registerOnValidatorChange=function(t){this._onChange=t},t.prototype._createValidator=function(){this._validator=m.pattern(this.pattern)},t.decorators=[{type:i.s,args:[{selector:"[pattern][formControlName],[pattern][formControl],[pattern][ngModel]",providers:[Jt],host:{"[attr.pattern]":"pattern ? pattern : null"}}]}],t.ctorParameters=function(){return[]},t.propDecorators={pattern:[{type:i.C}]},t}(),ee=function(){function t(){}return t.prototype.group=function(t,e){void 0===e&&(e=null);var n=this._reduceControls(t),r=null!=e?e.validator:null,i=null!=e?e.asyncValidator:null;return new mt(n,r,i)},t.prototype.control=function(t,e,n){return new ft(t,e,n)},t.prototype.array=function(t,e,n){var r=this,i=t.map(function(t){return r._createControl(t)});return new gt(i,e,n)},t.prototype._reduceControls=function(t){var e=this,n={};return Object.keys(t).forEach(function(r){n[r]=e._createControl(t[r])}),n},t.prototype._createControl=function(t){if(t instanceof ft||t instanceof mt||t instanceof gt)return t;if(Array.isArray(t)){var e=t[0],n=t.length>1?t[1]:null,r=t.length>2?t[2]:null;return this.control(e,n,r)}return this.control(t)},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[]},t}(),ne=(new i._6("5.1.0"),[function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"form:not([ngNoForm]):not([ngNativeValidate])",host:{novalidate:""}}]}],t.ctorParameters=function(){return[]},t}(),V,H,O,S,R,j,F,q,D,st,at,Wt,$t,Zt,te,Gt,Yt]),re=[Mt,Et,bt],ie=[Dt,Rt,qt,Lt,Vt],oe=function(){function t(){}return t.decorators=[{type:i.H,args:[{declarations:ne,exports:ne}]}],t.ctorParameters=function(){return[]},t}();(function(){function t(){}t.decorators=[{type:i.H,args:[{declarations:re,providers:[T],exports:[oe,re]}]}],t.ctorParameters=function(){return[]}})(),function(){function t(){}t.decorators=[{type:i.H,args:[{declarations:[ie],providers:[ee,T],exports:[oe,ie]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/autocomplete.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return P});var r=n("../../../core/esm5/core.js"),i=n("../../../material/esm5/core.es5.js"),o=n("../../../cdk/esm5/a11y.es5.js"),s=n("../../../common/esm5/common.js"),a=n("../../../cdk/esm5/overlay.es5.js"),c=n("../../../cdk/esm5/bidi.es5.js"),u=n("../../../cdk/esm5/keycodes.es5.js"),l=n("../../../cdk/esm5/portal.es5.js"),p=n("../../../../rxjs/_esm5/operators/filter.js"),d=n("../../../../rxjs/_esm5/operators/take.js"),h=n("../../../../rxjs/_esm5/operators/switchMap.js"),f=n("../../../../rxjs/_esm5/operators/tap.js"),m=n("../../../../rxjs/_esm5/operators/delay.js"),g=n("../../../forms/esm5/forms.js"),v=n("../../../material/esm5/form-field.es5.js"),y=n("../../../../rxjs/_esm5/Subject.js"),b=n("../../../../rxjs/_esm5/observable/fromEvent.js"),_=n("../../../../rxjs/_esm5/observable/merge.js"),j=n("../../../../rxjs/_esm5/observable/of.js"),x=0,w=function(){return function(t,e){this.source=t,this.option=e}}(),O=function(){function t(t,e){this._changeDetectorRef=t,this._elementRef=e,this.showPanel=!1,this._isOpen=!1,this.displayWith=null,this.optionSelected=new r.w,this._classList={},this.id="mat-autocomplete-"+x++}return Object.defineProperty(t.prototype,"isOpen",{get:function(){return this._isOpen&&this.showPanel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"classList",{set:function(t){var e=this;t&&t.length&&(t.split(" ").forEach(function(t){return e._classList[t.trim()]=!0}),this._elementRef.nativeElement.className="")},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._keyManager=new o.c(this.options).withWrap(),this._setVisibility()},t.prototype._setScrollTop=function(t){this.panel&&(this.panel.nativeElement.scrollTop=t)},t.prototype._getScrollTop=function(){return this.panel?this.panel.nativeElement.scrollTop:0},t.prototype._setVisibility=function(){this.showPanel=!!this.options.length,this._classList["mat-autocomplete-visible"]=this.showPanel,this._classList["mat-autocomplete-hidden"]=!this.showPanel,this._changeDetectorRef.markForCheck()},t.prototype._emitSelectEvent=function(t){var e=new w(this,t);this.optionSelected.emit(e)},t.decorators=[{type:r.m,args:[{selector:"mat-autocomplete",template:'<ng-template><div class="mat-autocomplete-panel" role="listbox" [id]="id" [ngClass]="_classList" #panel><ng-content></ng-content></div></ng-template>',styles:[".mat-autocomplete-panel{min-width:112px;max-width:280px;overflow:auto;-webkit-overflow-scrolling:touch;visibility:hidden;max-width:none;max-height:256px;position:relative}.mat-autocomplete-panel:not([class*=mat-elevation-z]){box-shadow:0 5px 5px -3px rgba(0,0,0,.2),0 8px 10px 1px rgba(0,0,0,.14),0 3px 14px 2px rgba(0,0,0,.12)}.mat-autocomplete-panel.mat-autocomplete-visible{visibility:visible}.mat-autocomplete-panel.mat-autocomplete-hidden{visibility:hidden}"],encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush,exportAs:"matAutocomplete",host:{class:"mat-autocomplete"}}]}],t.ctorParameters=function(){return[{type:r.j},{type:r.t}]},t.propDecorators={template:[{type:r._7,args:[r._4]}],panel:[{type:r._7,args:["panel"]}],options:[{type:r.r,args:[i.o,{descendants:!0}]}],optionGroups:[{type:r.r,args:[i.n]}],displayWith:[{type:r.C}],optionSelected:[{type:r.N}],classList:[{type:r.C,args:["class"]}]},t}(),C=new r.A("mat-autocomplete-scroll-strategy");var k={provide:C,deps:[a.b],useFactory:function(t){return function(){return t.scrollStrategies.reposition()}}},E={provide:g.d,useExisting:Object(r._14)(function(){return S}),multi:!0};var S=function(){function t(t,e,n,r,i,o,s,a,c){this._element=t,this._overlay=e,this._viewContainerRef=n,this._zone=r,this._changeDetectorRef=i,this._scrollStrategy=o,this._dir=s,this._formField=a,this._document=c,this._panelOpen=!1,this._manuallyFloatingLabel=!1,this._escapeEventStream=new y.a,this._onChange=function(){},this._onTouched=function(){}}return t.prototype.ngOnDestroy=function(){this._destroyPanel(),this._escapeEventStream.complete()},Object.defineProperty(t.prototype,"panelOpen",{get:function(){return this._panelOpen&&this.autocomplete.showPanel},enumerable:!0,configurable:!0}),t.prototype.openPanel=function(){this._attachOverlay(),this._floatLabel()},t.prototype.closePanel=function(){this._overlayRef&&this._overlayRef.hasAttached()&&(this._overlayRef.detach(),this._closingActionsSubscription.unsubscribe()),this._resetLabel(),this._panelOpen&&(this.autocomplete._isOpen=this._panelOpen=!1,this._changeDetectorRef.detectChanges())},Object.defineProperty(t.prototype,"panelClosingActions",{get:function(){return Object(_.a)(this.optionSelections,this.autocomplete._keyManager.tabOut,this._escapeEventStream,this._outsideClickStream)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"optionSelections",{get:function(){return _.a.apply(void 0,this.autocomplete.options.map(function(t){return t.onSelectionChange}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"activeOption",{get:function(){return this.autocomplete&&this.autocomplete._keyManager?this.autocomplete._keyManager.activeItem:null},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_outsideClickStream",{get:function(){var t=this;return this._document?Object(_.a)(Object(b.a)(this._document,"click"),Object(b.a)(this._document,"touchend")).pipe(Object(p.a)(function(e){var n=e.target,r=t._formField?t._formField._elementRef.nativeElement:null;return t._panelOpen&&n!==t._element.nativeElement&&(!r||!r.contains(n))&&!!t._overlayRef&&!t._overlayRef.overlayElement.contains(n)})):Object(j.a)(null)},enumerable:!0,configurable:!0}),t.prototype.writeValue=function(t){var e=this;Promise.resolve(null).then(function(){return e._setTriggerValue(t)})},t.prototype.registerOnChange=function(t){this._onChange=t},t.prototype.registerOnTouched=function(t){this._onTouched=t},t.prototype._handleKeydown=function(t){var e=t.keyCode;if(e===u.g&&this.panelOpen)this._resetActiveItem(),this._escapeEventStream.next(),t.stopPropagation();else if(this.activeOption&&e===u.f&&this.panelOpen)this.activeOption._selectViaInteraction(),this._resetActiveItem(),t.preventDefault();else{var n=this.autocomplete._keyManager.activeItem,r=e===u.p||e===u.d;this.panelOpen||e===u.o?this.autocomplete._keyManager.onKeydown(t):r&&this.openPanel(),(r||this.autocomplete._keyManager.activeItem!==n)&&this._scrollToOption()}},t.prototype._handleInput=function(t){document.activeElement===t.target&&(this._onChange(t.target.value),this.openPanel())},t.prototype._handleFocus=function(){this._element.nativeElement.readOnly||(this._attachOverlay(),this._floatLabel(!0))},t.prototype._floatLabel=function(t){void 0===t&&(t=!1),this._formField&&"auto"===this._formField.floatLabel&&(t?this._formField._animateAndLockLabel():this._formField.floatLabel="always",this._manuallyFloatingLabel=!0)},t.prototype._resetLabel=function(){this._manuallyFloatingLabel&&(this._formField.floatLabel="auto",this._manuallyFloatingLabel=!1)},t.prototype._scrollToOption=function(){var t=this.autocomplete._keyManager.activeItemIndex||0,e=48*(t+i.o.countGroupLabelsBeforeOption(t,this.autocomplete.options,this.autocomplete.optionGroups)),n=this.autocomplete._getScrollTop();if(e<n)this.autocomplete._setScrollTop(e);else if(e+48>n+256){var r=e-256+48;this.autocomplete._setScrollTop(Math.max(0,r))}},t.prototype._subscribeToClosingActions=function(){var t=this,e=this._zone.onStable.asObservable().pipe(Object(d.a)(1)),n=this.autocomplete.options.changes.pipe(Object(f.a)(function(){return t._positionStrategy.recalculateLastPosition()}),Object(m.a)(0));return Object(_.a)(e,n).pipe(Object(h.a)(function(){return t._resetActiveItem(),t.autocomplete._setVisibility(),t.panelClosingActions}),Object(d.a)(1)).subscribe(function(e){return t._setValueAndClose(e)})},t.prototype._destroyPanel=function(){this._overlayRef&&(this.closePanel(),this._overlayRef.dispose(),this._overlayRef=null)},t.prototype._setTriggerValue=function(t){var e=this.autocomplete&&this.autocomplete.displayWith?this.autocomplete.displayWith(t):t,n=null!=e?e:"";this._formField?this._formField._control.value=n:this._element.nativeElement.value=n},t.prototype._setValueAndClose=function(t){t&&t.source&&(this._clearPreviousSelectedOption(t.source),this._setTriggerValue(t.source.value),this._onChange(t.source.value),this._element.nativeElement.focus(),this.autocomplete._emitSelectEvent(t.source)),this.closePanel()},t.prototype._clearPreviousSelectedOption=function(t){this.autocomplete.options.forEach(function(e){e!=t&&e.selected&&e.deselect()})},t.prototype._attachOverlay=function(){if(!this.autocomplete)throw Error("Attempting to open an undefined instance of `mat-autocomplete`. Make sure that the id passed to the `matAutocomplete` is correct and that you're attempting to open it after the ngAfterContentInit hook.");this._overlayRef?this._overlayRef.updateSize({width:this._getHostWidth()}):(this._portal=new l.h(this.autocomplete.template,this._viewContainerRef),this._overlayRef=this._overlay.create(this._getOverlayConfig())),this._overlayRef&&!this._overlayRef.hasAttached()&&(this._overlayRef.attach(this._portal),this._closingActionsSubscription=this._subscribeToClosingActions()),this.autocomplete._setVisibility(),this.autocomplete._isOpen=this._panelOpen=!0},t.prototype._getOverlayConfig=function(){return new a.c({positionStrategy:this._getOverlayPosition(),scrollStrategy:this._scrollStrategy(),width:this._getHostWidth(),direction:this._dir?this._dir.value:"ltr"})},t.prototype._getOverlayPosition=function(){return this._positionStrategy=this._overlay.position().connectedTo(this._getConnectedElement(),{originX:"start",originY:"bottom"},{overlayX:"start",overlayY:"top"}).withFallbackPosition({originX:"start",originY:"top"},{overlayX:"start",overlayY:"bottom"}),this._positionStrategy},t.prototype._getConnectedElement=function(){return this._formField?this._formField._connectionContainerRef:this._element},t.prototype._getHostWidth=function(){return this._getConnectedElement().nativeElement.getBoundingClientRect().width},t.prototype._resetActiveItem=function(){this.autocomplete._keyManager.setActiveItem(-1)},t.decorators=[{type:r.s,args:[{selector:"input[matAutocomplete], textarea[matAutocomplete]",host:{role:"combobox",autocomplete:"off","aria-autocomplete":"list","[attr.aria-activedescendant]":"activeOption?.id","[attr.aria-expanded]":"panelOpen.toString()","[attr.aria-owns]":"autocomplete?.id","(focusin)":"_handleFocus()","(blur)":"_onTouched()","(input)":"_handleInput($event)","(keydown)":"_handleKeydown($event)"},providers:[E]}]}],t.ctorParameters=function(){return[{type:r.t},{type:a.b},{type:r._9},{type:r.L},{type:r.j},{type:void 0,decorators:[{type:r.y,args:[C]}]},{type:c.b,decorators:[{type:r.M}]},{type:v.a,decorators:[{type:r.M},{type:r.x}]},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[s.b]}]}]},t.propDecorators={autocomplete:[{type:r.C,args:["matAutocomplete"]}]},t}(),P=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.p,a.d,i.j,s.a],exports:[O,i.p,S,i.j],declarations:[O,S],providers:[k]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/button-toggle.es5.js":function(t,e,n){"use strict";var r=n("../../../../tslib/tslib.es6.js"),i=n("../../../cdk/esm5/a11y.es5.js"),o=n("../../../cdk/esm5/coercion.es5.js"),s=n("../../../cdk/esm5/collections.es5.js"),a=n("../../../core/esm5/core.js"),c=n("../../../forms/esm5/forms.js"),u=n("../../../material/esm5/core.es5.js"),l=function(){return function(){}}(),p=Object(u.w)(l),d={provide:c.d,useExisting:Object(a._14)(function(){return m}),multi:!0},h=0,f=function(){return function(){}}(),m=function(t){function e(e){var n=t.call(this)||this;return n._changeDetector=e,n._value=null,n._name="mat-button-toggle-group-"+h++,n._vertical=!1,n._selected=null,n._controlValueAccessorChangeFn=function(){},n._onTouched=function(){},n.valueChange=new a.w,n.change=new a.w,n}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"name",{get:function(){return this._name},set:function(t){this._name=t,this._updateButtonToggleNames()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"vertical",{get:function(){return this._vertical},set:function(t){this._vertical=Object(o.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(t){this._value!=t&&(this._value=t,this.valueChange.emit(t),this._updateSelectedButtonToggleFromValue())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=t,this.value=t?t.value:null,t&&!t.checked&&(t.checked=!0)},enumerable:!0,configurable:!0}),e.prototype._updateButtonToggleNames=function(){var t=this;this._buttonToggles&&this._buttonToggles.forEach(function(e){e.name=t._name})},e.prototype._updateSelectedButtonToggleFromValue=function(){var t=this,e=null!=this._selected&&this._selected.value==this._value;if(null!=this._buttonToggles&&!e){var n=this._buttonToggles.filter(function(e){return e.value==t._value})[0];n?this.selected=n:null==this.value&&(this.selected=null,this._buttonToggles.forEach(function(t){t.checked=!1}))}},e.prototype._emitChangeEvent=function(){var t=new f;t.source=this._selected,t.value=this._value,this._controlValueAccessorChangeFn(t.value),this.change.emit(t)},e.prototype.writeValue=function(t){this.value=t,this._changeDetector.markForCheck()},e.prototype.registerOnChange=function(t){this._controlValueAccessorChangeFn=t},e.prototype.registerOnTouched=function(t){this._onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t,this._markButtonTogglesForCheck()},e.prototype._markButtonTogglesForCheck=function(){this._buttonToggles&&this._buttonToggles.forEach(function(t){return t._markForCheck()})},e.decorators=[{type:a.s,args:[{selector:"mat-button-toggle-group:not([multiple])",providers:[d],inputs:["disabled"],host:{role:"radiogroup",class:"mat-button-toggle-group","[class.mat-button-toggle-vertical]":"vertical"},exportAs:"matButtonToggleGroup"}]}],e.ctorParameters=function(){return[{type:a.j}]},e.propDecorators={_buttonToggles:[{type:a.r,args:[Object(a._14)(function(){return v})]}],name:[{type:a.C}],vertical:[{type:a.C}],value:[{type:a.C}],valueChange:[{type:a.N}],selected:[{type:a.C}],change:[{type:a.N}]},e}(p),g=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._vertical=!1,e}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"vertical",{get:function(){return this._vertical},set:function(t){this._vertical=Object(o.b)(t)},enumerable:!0,configurable:!0}),e.decorators=[{type:a.s,args:[{selector:"mat-button-toggle-group[multiple]",exportAs:"matButtonToggleGroup",inputs:["disabled"],host:{class:"mat-button-toggle-group","[class.mat-button-toggle-vertical]":"vertical",role:"group"}}]}],e.ctorParameters=function(){return[]},e.propDecorators={vertical:[{type:a.C}]},e}(p),v=function(){function t(t,e,n,r,i,o){var s=this;this._changeDetectorRef=n,this._buttonToggleDispatcher=r,this._elementRef=i,this._focusMonitor=o,this.ariaLabel="",this.ariaLabelledby=null,this._checked=!1,this._disabled=!1,this._value=null,this._isSingleSelector=!1,this._removeUniqueSelectionListener=function(){},this.change=new a.w,this.buttonToggleGroup=t,this.buttonToggleGroupMultiple=e,this.buttonToggleGroup?(this._removeUniqueSelectionListener=r.listen(function(t,e){t!=s.id&&e==s.name&&(s.checked=!1,s._changeDetectorRef.markForCheck())}),this._type="radio",this.name=this.buttonToggleGroup.name,this._isSingleSelector=!0):(this._type="checkbox",this._isSingleSelector=!1)}return Object.defineProperty(t.prototype,"inputId",{get:function(){return this.id+"-input"},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"checked",{get:function(){return this._checked},set:function(t){this._isSingleSelector&&t&&(this._buttonToggleDispatcher.notify(this.id,this.name),this._changeDetectorRef.markForCheck()),this._checked=t,t&&this._isSingleSelector&&this.buttonToggleGroup.value!=this.value&&(this.buttonToggleGroup.selected=this)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(t){this._value!=t&&(null!=this.buttonToggleGroup&&this.checked&&(this.buttonToggleGroup.value=t),this._value=t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled||null!=this.buttonToggleGroup&&this.buttonToggleGroup.disabled||null!=this.buttonToggleGroupMultiple&&this.buttonToggleGroupMultiple.disabled},set:function(t){this._disabled=Object(o.b)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){null==this.id&&(this.id="mat-button-toggle-"+h++),this.buttonToggleGroup&&this._value==this.buttonToggleGroup.value&&(this._checked=!0),this._focusMonitor.monitor(this._elementRef.nativeElement,!0)},t.prototype.focus=function(){this._inputElement.nativeElement.focus()},t.prototype._toggle=function(){this.checked=!this.checked},t.prototype._onInputChange=function(t){if(t.stopPropagation(),this._isSingleSelector){var e=this.buttonToggleGroup.selected!=this;this.checked=!0,this.buttonToggleGroup.selected=this,this.buttonToggleGroup._onTouched(),e&&this.buttonToggleGroup._emitChangeEvent()}else this._toggle();this._emitChangeEvent()},t.prototype._onInputClick=function(t){t.stopPropagation()},t.prototype._emitChangeEvent=function(){var t=new f;t.source=this,t.value=this._value,this.change.emit(t)},t.prototype.ngOnDestroy=function(){this._removeUniqueSelectionListener()},t.prototype._markForCheck=function(){this._changeDetectorRef.markForCheck()},t.decorators=[{type:a.m,args:[{selector:"mat-button-toggle",template:'<label [attr.for]="inputId" class="mat-button-toggle-label"><input #input class="mat-button-toggle-input cdk-visually-hidden" [type]="_type" [id]="inputId" [checked]="checked" [disabled]="disabled || null" [name]="name" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" (change)="_onInputChange($event)" (click)="_onInputClick($event)"><div class="mat-button-toggle-label-content"><ng-content></ng-content></div></label><div class="mat-button-toggle-focus-overlay"></div>',styles:[".mat-button-toggle-group,.mat-button-toggle-standalone{box-shadow:0 3px 1px -2px rgba(0,0,0,.2),0 2px 2px 0 rgba(0,0,0,.14),0 1px 5px 0 rgba(0,0,0,.12);position:relative;display:inline-flex;flex-direction:row;border-radius:2px;cursor:pointer;white-space:nowrap;overflow:hidden}.mat-button-toggle-vertical{flex-direction:column}.mat-button-toggle-vertical .mat-button-toggle-label-content{display:block}.mat-button-toggle-disabled .mat-button-toggle-label-content{cursor:default}.mat-button-toggle{white-space:nowrap;position:relative}.mat-button-toggle.cdk-keyboard-focused .mat-button-toggle-focus-overlay,.mat-button-toggle.cdk-program-focused .mat-button-toggle-focus-overlay{opacity:1}.mat-button-toggle-label-content{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;display:inline-block;line-height:36px;padding:0 16px;cursor:pointer}.mat-button-toggle-label-content>*{vertical-align:middle}.mat-button-toggle-focus-overlay{border-radius:inherit;pointer-events:none;opacity:0;top:0;left:0;right:0;bottom:0;position:absolute}"],encapsulation:a._10.None,preserveWhitespaces:!1,exportAs:"matButtonToggle",changeDetection:a.i.OnPush,host:{"[class.mat-button-toggle-standalone]":"!buttonToggleGroup && !buttonToggleGroupMultiple","[class.mat-button-toggle-checked]":"checked","[class.mat-button-toggle-disabled]":"disabled",class:"mat-button-toggle","[attr.id]":"id"}}]}],t.ctorParameters=function(){return[{type:m,decorators:[{type:a.M}]},{type:g,decorators:[{type:a.M}]},{type:a.j},{type:s.c},{type:a.t},{type:i.f}]},t.propDecorators={ariaLabel:[{type:a.C,args:["aria-label"]}],ariaLabelledby:[{type:a.C,args:["aria-labelledby"]}],_inputElement:[{type:a._7,args:["input"]}],id:[{type:a.C}],name:[{type:a.C}],checked:[{type:a.C}],value:[{type:a.C}],disabled:[{type:a.C}],change:[{type:a.N}]},t}();!function(){function t(){}t.decorators=[{type:a.H,args:[{imports:[u.j,i.a],exports:[m,g,v,u.j],declarations:[m,g,v],providers:[s.b]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/button.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return y});var r=n("../../../core/esm5/core.js"),i=n("../../../common/esm5/common.js"),o=n("../../../material/esm5/core.es5.js"),s=n("../../../cdk/esm5/a11y.es5.js"),a=n("../../../../tslib/tslib.es6.js"),c=n("../../../cdk/esm5/platform.es5.js"),u="accent",l=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"button[mat-button], a[mat-button]",host:{class:"mat-button"}}]}],t.ctorParameters=function(){return[]},t}(),p=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"button[mat-raised-button], a[mat-raised-button]",host:{class:"mat-raised-button"}}]}],t.ctorParameters=function(){return[]},t}(),d=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"button[mat-icon-button], a[mat-icon-button]",host:{class:"mat-icon-button"}}]}],t.ctorParameters=function(){return[]},t}(),h=function(){function t(t,e){(t||e).color=u}return t.decorators=[{type:r.s,args:[{selector:"button[mat-fab], a[mat-fab]",host:{class:"mat-fab"}}]}],t.ctorParameters=function(){return[{type:g,decorators:[{type:r.Z},{type:r.M},{type:r.y,args:[Object(r._14)(function(){return g})]}]},{type:v,decorators:[{type:r.Z},{type:r.M},{type:r.y,args:[Object(r._14)(function(){return v})]}]}]},t}(),f=function(){function t(t,e){(t||e).color=u}return t.decorators=[{type:r.s,args:[{selector:"button[mat-mini-fab], a[mat-mini-fab]",host:{class:"mat-mini-fab"}}]}],t.ctorParameters=function(){return[{type:g,decorators:[{type:r.Z},{type:r.M},{type:r.y,args:[Object(r._14)(function(){return g})]}]},{type:v,decorators:[{type:r.Z},{type:r.M},{type:r.y,args:[Object(r._14)(function(){return v})]}]}]},t}(),m=function(){return function(t){this._elementRef=t}}(),g=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i._platform=n,i._focusMonitor=r,i._isRoundButton=i._hasHostAttributes("mat-fab","mat-mini-fab"),i._isIconButton=i._hasHostAttributes("mat-icon-button"),i._focusMonitor.monitor(i._elementRef.nativeElement,!0),i}return Object(a.b)(e,t),e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef.nativeElement)},e.prototype.focus=function(){this._getHostElement().focus()},e.prototype._getHostElement=function(){return this._elementRef.nativeElement},e.prototype._isRippleDisabled=function(){return this.disableRipple||this.disabled},e.prototype._hasHostAttributes=function(){for(var t=this,e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return!!this._platform.isBrowser&&e.some(function(e){return t._getHostElement().hasAttribute(e)})},e.decorators=[{type:r.m,args:[{selector:"button[mat-button], button[mat-raised-button], button[mat-icon-button],\n             button[mat-fab], button[mat-mini-fab]",exportAs:"matButton",host:{"[disabled]":"disabled || null"},template:'<span class="mat-button-wrapper"><ng-content></ng-content></span><div matRipple class="mat-button-ripple" [class.mat-button-ripple-round]="_isRoundButton || _isIconButton" [matRippleDisabled]="_isRippleDisabled()" [matRippleCentered]="_isIconButton" [matRippleTrigger]="_getHostElement()"></div><div class="mat-button-focus-overlay"></div>',styles:[".mat-button,.mat-fab,.mat-icon-button,.mat-mini-fab,.mat-raised-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:88px;line-height:36px;padding:0 16px;border-radius:2px}[disabled].mat-button,[disabled].mat-fab,[disabled].mat-icon-button,[disabled].mat-mini-fab,[disabled].mat-raised-button{cursor:default}.cdk-keyboard-focused.mat-button .mat-button-focus-overlay,.cdk-keyboard-focused.mat-fab .mat-button-focus-overlay,.cdk-keyboard-focused.mat-icon-button .mat-button-focus-overlay,.cdk-keyboard-focused.mat-mini-fab .mat-button-focus-overlay,.cdk-keyboard-focused.mat-raised-button .mat-button-focus-overlay,.cdk-program-focused.mat-button .mat-button-focus-overlay,.cdk-program-focused.mat-fab .mat-button-focus-overlay,.cdk-program-focused.mat-icon-button .mat-button-focus-overlay,.cdk-program-focused.mat-mini-fab .mat-button-focus-overlay,.cdk-program-focused.mat-raised-button .mat-button-focus-overlay{opacity:1}.mat-button::-moz-focus-inner,.mat-fab::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-mini-fab::-moz-focus-inner,.mat-raised-button::-moz-focus-inner{border:0}.mat-fab,.mat-mini-fab,.mat-raised-button{transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1)}.mat-fab:not([class*=mat-elevation-z]),.mat-mini-fab:not([class*=mat-elevation-z]),.mat-raised-button:not([class*=mat-elevation-z]){box-shadow:0 3px 1px -2px rgba(0,0,0,.2),0 2px 2px 0 rgba(0,0,0,.14),0 1px 5px 0 rgba(0,0,0,.12)}.mat-fab:not([disabled]):active:not([class*=mat-elevation-z]),.mat-mini-fab:not([disabled]):active:not([class*=mat-elevation-z]),.mat-raised-button:not([disabled]):active:not([class*=mat-elevation-z]){box-shadow:0 5px 5px -3px rgba(0,0,0,.2),0 8px 10px 1px rgba(0,0,0,.14),0 3px 14px 2px rgba(0,0,0,.12)}[disabled].mat-fab,[disabled].mat-mini-fab,[disabled].mat-raised-button{box-shadow:none}.mat-button .mat-button-focus-overlay,.mat-icon-button .mat-button-focus-overlay{transition:none;opacity:0}.mat-button:hover .mat-button-focus-overlay{opacity:1}.mat-fab{min-width:0;border-radius:50%;width:56px;height:56px;padding:0;flex-shrink:0}.mat-fab:not([class*=mat-elevation-z]){box-shadow:0 3px 5px -1px rgba(0,0,0,.2),0 6px 10px 0 rgba(0,0,0,.14),0 1px 18px 0 rgba(0,0,0,.12)}.mat-fab:not([disabled]):active:not([class*=mat-elevation-z]){box-shadow:0 7px 8px -4px rgba(0,0,0,.2),0 12px 17px 2px rgba(0,0,0,.14),0 5px 22px 4px rgba(0,0,0,.12)}.mat-fab .mat-button-wrapper{padding:16px 0;display:inline-block;line-height:24px}.mat-mini-fab{min-width:0;border-radius:50%;width:40px;height:40px;padding:0;flex-shrink:0}.mat-mini-fab:not([class*=mat-elevation-z]){box-shadow:0 3px 5px -1px rgba(0,0,0,.2),0 6px 10px 0 rgba(0,0,0,.14),0 1px 18px 0 rgba(0,0,0,.12)}.mat-mini-fab:not([disabled]):active:not([class*=mat-elevation-z]){box-shadow:0 7px 8px -4px rgba(0,0,0,.2),0 12px 17px 2px rgba(0,0,0,.14),0 5px 22px 4px rgba(0,0,0,.12)}.mat-mini-fab .mat-button-wrapper{padding:8px 0;display:inline-block;line-height:24px}.mat-icon-button{padding:0;min-width:0;width:40px;height:40px;flex-shrink:0;line-height:40px;border-radius:50%}.mat-icon-button .mat-icon,.mat-icon-button i{line-height:24px}.mat-button,.mat-fab,.mat-icon-button,.mat-mini-fab,.mat-raised-button{color:currentColor}.mat-button .mat-button-wrapper>*,.mat-fab .mat-button-wrapper>*,.mat-icon-button .mat-button-wrapper>*,.mat-mini-fab .mat-button-wrapper>*,.mat-raised-button .mat-button-wrapper>*{vertical-align:middle}.mat-button-focus-overlay,.mat-button-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-button-focus-overlay{background-color:rgba(0,0,0,.12);border-radius:inherit;opacity:0;transition:opacity .2s cubic-bezier(.35,0,.25,1),background-color .2s cubic-bezier(.35,0,.25,1)}@media screen and (-ms-high-contrast:active){.mat-button-focus-overlay{background-color:rgba(255,255,255,.5)}}.mat-button-ripple-round{border-radius:50%;z-index:1}@media screen and (-ms-high-contrast:active){.mat-button,.mat-fab,.mat-icon-button,.mat-mini-fab,.mat-raised-button{outline:solid 1px}}"],inputs:["disabled","disableRipple","color"],encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush}]}],e.ctorParameters=function(){return[{type:r.t},{type:c.a},{type:s.f}]},e}(Object(o.u)(Object(o.w)(Object(o.v)(m)))),v=function(t){function e(e,n,r){return t.call(this,r,e,n)||this}return Object(a.b)(e,t),e.prototype._haltDisabledEvents=function(t){this.disabled&&(t.preventDefault(),t.stopImmediatePropagation())},e.decorators=[{type:r.m,args:[{selector:"a[mat-button], a[mat-raised-button], a[mat-icon-button], a[mat-fab], a[mat-mini-fab]",exportAs:"matButton, matAnchor",host:{"[attr.tabindex]":"disabled ? -1 : 0","[attr.disabled]":"disabled || null","[attr.aria-disabled]":"disabled.toString()","(click)":"_haltDisabledEvents($event)"},inputs:["disabled","disableRipple","color"],template:'<span class="mat-button-wrapper"><ng-content></ng-content></span><div matRipple class="mat-button-ripple" [class.mat-button-ripple-round]="_isRoundButton || _isIconButton" [matRippleDisabled]="_isRippleDisabled()" [matRippleCentered]="_isIconButton" [matRippleTrigger]="_getHostElement()"></div><div class="mat-button-focus-overlay"></div>',styles:[".mat-button,.mat-fab,.mat-icon-button,.mat-mini-fab,.mat-raised-button{box-sizing:border-box;position:relative;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;display:inline-block;white-space:nowrap;text-decoration:none;vertical-align:baseline;text-align:center;margin:0;min-width:88px;line-height:36px;padding:0 16px;border-radius:2px}[disabled].mat-button,[disabled].mat-fab,[disabled].mat-icon-button,[disabled].mat-mini-fab,[disabled].mat-raised-button{cursor:default}.cdk-keyboard-focused.mat-button .mat-button-focus-overlay,.cdk-keyboard-focused.mat-fab .mat-button-focus-overlay,.cdk-keyboard-focused.mat-icon-button .mat-button-focus-overlay,.cdk-keyboard-focused.mat-mini-fab .mat-button-focus-overlay,.cdk-keyboard-focused.mat-raised-button .mat-button-focus-overlay,.cdk-program-focused.mat-button .mat-button-focus-overlay,.cdk-program-focused.mat-fab .mat-button-focus-overlay,.cdk-program-focused.mat-icon-button .mat-button-focus-overlay,.cdk-program-focused.mat-mini-fab .mat-button-focus-overlay,.cdk-program-focused.mat-raised-button .mat-button-focus-overlay{opacity:1}.mat-button::-moz-focus-inner,.mat-fab::-moz-focus-inner,.mat-icon-button::-moz-focus-inner,.mat-mini-fab::-moz-focus-inner,.mat-raised-button::-moz-focus-inner{border:0}.mat-fab,.mat-mini-fab,.mat-raised-button{transform:translate3d(0,0,0);transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1)}.mat-fab:not([class*=mat-elevation-z]),.mat-mini-fab:not([class*=mat-elevation-z]),.mat-raised-button:not([class*=mat-elevation-z]){box-shadow:0 3px 1px -2px rgba(0,0,0,.2),0 2px 2px 0 rgba(0,0,0,.14),0 1px 5px 0 rgba(0,0,0,.12)}.mat-fab:not([disabled]):active:not([class*=mat-elevation-z]),.mat-mini-fab:not([disabled]):active:not([class*=mat-elevation-z]),.mat-raised-button:not([disabled]):active:not([class*=mat-elevation-z]){box-shadow:0 5px 5px -3px rgba(0,0,0,.2),0 8px 10px 1px rgba(0,0,0,.14),0 3px 14px 2px rgba(0,0,0,.12)}[disabled].mat-fab,[disabled].mat-mini-fab,[disabled].mat-raised-button{box-shadow:none}.mat-button .mat-button-focus-overlay,.mat-icon-button .mat-button-focus-overlay{transition:none;opacity:0}.mat-button:hover .mat-button-focus-overlay{opacity:1}.mat-fab{min-width:0;border-radius:50%;width:56px;height:56px;padding:0;flex-shrink:0}.mat-fab:not([class*=mat-elevation-z]){box-shadow:0 3px 5px -1px rgba(0,0,0,.2),0 6px 10px 0 rgba(0,0,0,.14),0 1px 18px 0 rgba(0,0,0,.12)}.mat-fab:not([disabled]):active:not([class*=mat-elevation-z]){box-shadow:0 7px 8px -4px rgba(0,0,0,.2),0 12px 17px 2px rgba(0,0,0,.14),0 5px 22px 4px rgba(0,0,0,.12)}.mat-fab .mat-button-wrapper{padding:16px 0;display:inline-block;line-height:24px}.mat-mini-fab{min-width:0;border-radius:50%;width:40px;height:40px;padding:0;flex-shrink:0}.mat-mini-fab:not([class*=mat-elevation-z]){box-shadow:0 3px 5px -1px rgba(0,0,0,.2),0 6px 10px 0 rgba(0,0,0,.14),0 1px 18px 0 rgba(0,0,0,.12)}.mat-mini-fab:not([disabled]):active:not([class*=mat-elevation-z]){box-shadow:0 7px 8px -4px rgba(0,0,0,.2),0 12px 17px 2px rgba(0,0,0,.14),0 5px 22px 4px rgba(0,0,0,.12)}.mat-mini-fab .mat-button-wrapper{padding:8px 0;display:inline-block;line-height:24px}.mat-icon-button{padding:0;min-width:0;width:40px;height:40px;flex-shrink:0;line-height:40px;border-radius:50%}.mat-icon-button .mat-icon,.mat-icon-button i{line-height:24px}.mat-button,.mat-fab,.mat-icon-button,.mat-mini-fab,.mat-raised-button{color:currentColor}.mat-button .mat-button-wrapper>*,.mat-fab .mat-button-wrapper>*,.mat-icon-button .mat-button-wrapper>*,.mat-mini-fab .mat-button-wrapper>*,.mat-raised-button .mat-button-wrapper>*{vertical-align:middle}.mat-button-focus-overlay,.mat-button-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-button-focus-overlay{background-color:rgba(0,0,0,.12);border-radius:inherit;opacity:0;transition:opacity .2s cubic-bezier(.35,0,.25,1),background-color .2s cubic-bezier(.35,0,.25,1)}@media screen and (-ms-high-contrast:active){.mat-button-focus-overlay{background-color:rgba(255,255,255,.5)}}.mat-button-ripple-round{border-radius:50%;z-index:1}@media screen and (-ms-high-contrast:active){.mat-button,.mat-fab,.mat-icon-button,.mat-mini-fab,.mat-raised-button{outline:solid 1px}}"],encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush}]}],e.ctorParameters=function(){return[{type:c.a},{type:s.f},{type:r.t}]},e}(g),y=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.a,o.s,o.j,s.a],exports:[g,v,f,h,o.j,l,p,d],declarations:[g,v,f,h,l,p,d]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/card.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return b});var r=n("../../../core/esm5/core.js"),i=n("../../../material/esm5/core.es5.js"),o=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"mat-card-content",host:{class:"mat-card-content"}}]}],t.ctorParameters=function(){return[]},t}(),s=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"mat-card-title, [mat-card-title], [matCardTitle]",host:{class:"mat-card-title"}}]}],t.ctorParameters=function(){return[]},t}(),a=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"mat-card-subtitle, [mat-card-subtitle], [matCardSubtitle]",host:{class:"mat-card-subtitle"}}]}],t.ctorParameters=function(){return[]},t}(),c=function(){function t(){this.align="start"}return t.decorators=[{type:r.s,args:[{selector:"mat-card-actions",exportAs:"matCardActions",host:{class:"mat-card-actions","[class.mat-card-actions-align-end]":'align === "end"'}}]}],t.ctorParameters=function(){return[]},t.propDecorators={align:[{type:r.C}]},t}(),u=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"mat-card-footer",host:{class:"mat-card-footer"}}]}],t.ctorParameters=function(){return[]},t}(),l=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-card-image], [matCardImage]",host:{class:"mat-card-image"}}]}],t.ctorParameters=function(){return[]},t}(),p=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-card-sm-image], [matCardImageSmall]",host:{class:"mat-card-sm-image"}}]}],t.ctorParameters=function(){return[]},t}(),d=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-card-md-image], [matCardImageMedium]",host:{class:"mat-card-md-image"}}]}],t.ctorParameters=function(){return[]},t}(),h=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-card-lg-image], [matCardImageLarge]",host:{class:"mat-card-lg-image"}}]}],t.ctorParameters=function(){return[]},t}(),f=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-card-xl-image], [matCardImageXLarge]",host:{class:"mat-card-xl-image"}}]}],t.ctorParameters=function(){return[]},t}(),m=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-card-avatar], [matCardAvatar]",host:{class:"mat-card-avatar"}}]}],t.ctorParameters=function(){return[]},t}(),g=function(){function t(){}return t.decorators=[{type:r.m,args:[{selector:"mat-card",exportAs:"matCard",template:'<ng-content></ng-content><ng-content select="mat-card-footer"></ng-content>',styles:[".mat-card{transition:box-shadow 280ms cubic-bezier(.4,0,.2,1);display:block;position:relative;padding:24px;border-radius:2px}.mat-card:not([class*=mat-elevation-z]){box-shadow:0 3px 1px -2px rgba(0,0,0,.2),0 2px 2px 0 rgba(0,0,0,.14),0 1px 5px 0 rgba(0,0,0,.12)}@media screen and (-ms-high-contrast:active){.mat-card{outline:solid 1px}}.mat-card-flat{box-shadow:none}.mat-card-actions,.mat-card-content,.mat-card-subtitle,.mat-card-title{display:block;margin-bottom:16px}.mat-card-actions{margin-left:-16px;margin-right:-16px;padding:8px 0}.mat-card-actions-align-end{display:flex;justify-content:flex-end}.mat-card-image{width:calc(100% + 48px);margin:0 -24px 16px -24px}.mat-card-xl-image{width:240px;height:240px;margin:-8px}.mat-card-footer{display:block;margin:0 -24px -24px -24px}.mat-card-actions .mat-button,.mat-card-actions .mat-raised-button{margin:0 4px}.mat-card-header{display:flex;flex-direction:row}.mat-card-header-text{margin:0 8px}.mat-card-avatar{height:40px;width:40px;border-radius:50%;flex-shrink:0}.mat-card-lg-image,.mat-card-md-image,.mat-card-sm-image{margin:-8px 0}.mat-card-title-group{display:flex;justify-content:space-between;margin:0 -8px}.mat-card-sm-image{width:80px;height:80px}.mat-card-md-image{width:112px;height:112px}.mat-card-lg-image{width:152px;height:152px}@media (max-width:600px){.mat-card{padding:24px 16px}.mat-card-actions{margin-left:-8px;margin-right:-8px}.mat-card-image{width:calc(100% + 32px);margin:16px -16px}.mat-card-title-group{margin:0}.mat-card-xl-image{margin-left:0;margin-right:0}.mat-card-header{margin:-8px 0 0 0}.mat-card-footer{margin-left:-16px;margin-right:-16px}}.mat-card-content>:first-child,.mat-card>:first-child{margin-top:0}.mat-card-content>:last-child:not(.mat-card-footer),.mat-card>:last-child:not(.mat-card-footer){margin-bottom:0}.mat-card-image:first-child{margin-top:-24px}.mat-card>.mat-card-actions:last-child{margin-bottom:-16px;padding-bottom:0}.mat-card-actions .mat-button:first-child,.mat-card-actions .mat-raised-button:first-child{margin-left:0;margin-right:0}.mat-card-subtitle:not(:first-child),.mat-card-title:not(:first-child){margin-top:-4px}.mat-card-header .mat-card-subtitle:not(:first-child){margin-top:-8px}.mat-card>.mat-card-xl-image:first-child{margin-top:-8px}.mat-card>.mat-card-xl-image:last-child{margin-bottom:-8px}"],encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush,host:{class:"mat-card"}}]}],t.ctorParameters=function(){return[]},t}(),v=function(){function t(){}return t.decorators=[{type:r.m,args:[{selector:"mat-card-header",template:'<ng-content select="[mat-card-avatar], [matCardAvatar]"></ng-content><div class="mat-card-header-text"><ng-content select="mat-card-title, mat-card-subtitle, [mat-card-title], [mat-card-subtitle]"></ng-content></div><ng-content></ng-content>',encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush,host:{class:"mat-card-header"}}]}],t.ctorParameters=function(){return[]},t}(),y=function(){function t(){}return t.decorators=[{type:r.m,args:[{selector:"mat-card-title-group",template:'<div><ng-content select="mat-card-title, mat-card-subtitle, [mat-card-title], [mat-card-subtitle]"></ng-content></div><ng-content select="img"></ng-content><ng-content></ng-content>',encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush,host:{class:"mat-card-title-group"}}]}],t.ctorParameters=function(){return[]},t}(),b=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.j],exports:[g,v,y,o,s,a,c,u,p,d,h,l,f,m,i.j],declarations:[g,v,y,o,s,a,c,u,p,d,h,l,f,m]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/checkbox.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return _});var r=n("../../../../tslib/tslib.es6.js"),i=n("../../../cdk/esm5/a11y.es5.js"),o=n("../../../cdk/esm5/coercion.es5.js"),s=n("../../../core/esm5/core.js"),a=n("../../../forms/esm5/forms.js"),c=n("../../../material/esm5/core.es5.js"),u=n("../../../common/esm5/common.js"),l=n("../../../cdk/esm5/observers.es5.js"),p=new s.A("mat-checkbox-click-action"),d=0,h={provide:a.d,useExisting:Object(s._14)(function(){return v}),multi:!0},f={Init:0,Checked:1,Unchecked:2,Indeterminate:3};f[f.Init]="Init",f[f.Checked]="Checked",f[f.Unchecked]="Unchecked",f[f.Indeterminate]="Indeterminate";var m=function(){return function(){}}(),g=function(){return function(t){this._elementRef=t}}(),v=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;return a._changeDetectorRef=n,a._focusMonitor=r,a._clickAction=o,a.ariaLabel="",a.ariaLabelledby=null,a._uniqueId="mat-checkbox-"+ ++d,a.id=a._uniqueId,a.labelPosition="after",a.name=null,a.change=new s.w,a.indeterminateChange=new s.w,a._rippleConfig={centered:!0,radius:25,speedFactor:1.5},a.onTouched=function(){},a._currentAnimationClass="",a._currentCheckState=f.Init,a._checked=!1,a._indeterminate=!1,a._controlValueAccessorChangeFn=function(){},a.tabIndex=parseInt(i)||0,a}return Object(r.b)(e,t),Object.defineProperty(e.prototype,"inputId",{get:function(){return(this.id||this._uniqueId)+"-input"},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(o.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"align",{get:function(){return"after"==this.labelPosition?"start":"end"},set:function(t){this.labelPosition="start"==t?"after":"before"},enumerable:!0,configurable:!0}),e.prototype.ngAfterViewInit=function(){var t=this;this._focusMonitor.monitor(this._inputElement.nativeElement,!1).subscribe(function(e){return t._onInputFocusChange(e)})},e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._inputElement.nativeElement)},Object.defineProperty(e.prototype,"checked",{get:function(){return this._checked},set:function(t){t!=this.checked&&(this._checked=t,this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"indeterminate",{get:function(){return this._indeterminate},set:function(t){var e=t!=this._indeterminate;this._indeterminate=t,e&&(this._indeterminate?this._transitionCheckState(f.Indeterminate):this._transitionCheckState(this.checked?f.Checked:f.Unchecked),this.indeterminateChange.emit(this._indeterminate))},enumerable:!0,configurable:!0}),e.prototype._isRippleDisabled=function(){return this.disableRipple||this.disabled},e.prototype._onLabelTextChange=function(){this._changeDetectorRef.markForCheck()},e.prototype.writeValue=function(t){this.checked=!!t},e.prototype.registerOnChange=function(t){this._controlValueAccessorChangeFn=t},e.prototype.registerOnTouched=function(t){this.onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t,this._changeDetectorRef.markForCheck()},e.prototype._getAriaChecked=function(){return this.checked?"true":this.indeterminate?"mixed":"false"},e.prototype._transitionCheckState=function(t){var e=this._currentCheckState,n=this._elementRef.nativeElement;e!==t&&(this._currentAnimationClass.length>0&&n.classList.remove(this._currentAnimationClass),this._currentAnimationClass=this._getAnimationClassForCheckStateTransition(e,t),this._currentCheckState=t,this._currentAnimationClass.length>0&&n.classList.add(this._currentAnimationClass))},e.prototype._emitChangeEvent=function(){var t=new m;t.source=this,t.checked=this.checked,this._controlValueAccessorChangeFn(this.checked),this.change.emit(t)},e.prototype._onInputFocusChange=function(t){this._focusRipple||"keyboard"!==t?t||(this._removeFocusRipple(),this.onTouched()):this._focusRipple=this._ripple.launch(0,0,Object(r.a)({persistent:!0},this._rippleConfig))},e.prototype.toggle=function(){this.checked=!this.checked},e.prototype._onInputClick=function(t){var e=this;t.stopPropagation(),this.disabled||"noop"===this._clickAction?this.disabled||"noop"!==this._clickAction||(this._inputElement.nativeElement.checked=this.checked,this._inputElement.nativeElement.indeterminate=this.indeterminate):(this.indeterminate&&"check"!==this._clickAction&&Promise.resolve().then(function(){e._indeterminate=!1,e.indeterminateChange.emit(e._indeterminate)}),this.toggle(),this._transitionCheckState(this._checked?f.Checked:f.Unchecked),this._emitChangeEvent())},e.prototype.focus=function(){this._focusMonitor.focusVia(this._inputElement.nativeElement,"keyboard")},e.prototype._onInteractionEvent=function(t){t.stopPropagation()},e.prototype._getAnimationClassForCheckStateTransition=function(t,e){var n="";switch(t){case f.Init:if(e===f.Checked)n="unchecked-checked";else{if(e!=f.Indeterminate)return"";n="unchecked-indeterminate"}break;case f.Unchecked:n=e===f.Checked?"unchecked-checked":"unchecked-indeterminate";break;case f.Checked:n=e===f.Unchecked?"checked-unchecked":"checked-indeterminate";break;case f.Indeterminate:n=e===f.Checked?"indeterminate-checked":"indeterminate-unchecked"}return"mat-checkbox-anim-"+n},e.prototype._removeFocusRipple=function(){this._focusRipple&&(this._focusRipple.fadeOut(),this._focusRipple=null)},e.decorators=[{type:s.m,args:[{selector:"mat-checkbox",template:'<label [attr.for]="inputId" class="mat-checkbox-layout" #label><div class="mat-checkbox-inner-container" [class.mat-checkbox-inner-container-no-side-margin]="!checkboxLabel.textContent || !checkboxLabel.textContent.trim()"><input #input class="mat-checkbox-input cdk-visually-hidden" type="checkbox" [id]="inputId" [required]="required" [checked]="checked" [attr.value]="value" [disabled]="disabled" [attr.name]="name" [tabIndex]="tabIndex" [indeterminate]="indeterminate" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" [attr.aria-checked]="_getAriaChecked()" (change)="_onInteractionEvent($event)" (click)="_onInputClick($event)"><div matRipple class="mat-checkbox-ripple" [matRippleTrigger]="label" [matRippleDisabled]="_isRippleDisabled()" [matRippleRadius]="_rippleConfig.radius" [matRippleSpeedFactor]="_rippleConfig.speedFactor" [matRippleCentered]="_rippleConfig.centered"></div><div class="mat-checkbox-frame"></div><div class="mat-checkbox-background"><svg version="1.1" focusable="false" class="mat-checkbox-checkmark" viewBox="0 0 24 24" xml:space="preserve"><path class="mat-checkbox-checkmark-path" fill="none" stroke="white" d="M4.1,12.7 9,17.6 20.3,6.3"/></svg><div class="mat-checkbox-mixedmark"></div></div></div><span class="mat-checkbox-label" #checkboxLabel (cdkObserveContent)="_onLabelTextChange()"><span style="display:none">&nbsp;</span><ng-content></ng-content></span></label>',styles:["@keyframes mat-checkbox-fade-in-background{0%{opacity:0}50%{opacity:1}}@keyframes mat-checkbox-fade-out-background{0%,50%{opacity:1}100%{opacity:0}}@keyframes mat-checkbox-unchecked-checked-checkmark-path{0%,50%{stroke-dashoffset:22.91026}50%{animation-timing-function:cubic-bezier(0,0,.2,.1)}100%{stroke-dashoffset:0}}@keyframes mat-checkbox-unchecked-indeterminate-mixedmark{0%,68.2%{transform:scaleX(0)}68.2%{animation-timing-function:cubic-bezier(0,0,0,1)}100%{transform:scaleX(1)}}@keyframes mat-checkbox-checked-unchecked-checkmark-path{from{animation-timing-function:cubic-bezier(.4,0,1,1);stroke-dashoffset:0}to{stroke-dashoffset:-22.91026}}@keyframes mat-checkbox-checked-indeterminate-checkmark{from{animation-timing-function:cubic-bezier(0,0,.2,.1);opacity:1;transform:rotate(0)}to{opacity:0;transform:rotate(45deg)}}@keyframes mat-checkbox-indeterminate-checked-checkmark{from{animation-timing-function:cubic-bezier(.14,0,0,1);opacity:0;transform:rotate(45deg)}to{opacity:1;transform:rotate(360deg)}}@keyframes mat-checkbox-checked-indeterminate-mixedmark{from{animation-timing-function:cubic-bezier(0,0,.2,.1);opacity:0;transform:rotate(-45deg)}to{opacity:1;transform:rotate(0)}}@keyframes mat-checkbox-indeterminate-checked-mixedmark{from{animation-timing-function:cubic-bezier(.14,0,0,1);opacity:1;transform:rotate(0)}to{opacity:0;transform:rotate(315deg)}}@keyframes mat-checkbox-indeterminate-unchecked-mixedmark{0%{animation-timing-function:linear;opacity:1;transform:scaleX(1)}100%,32.8%{opacity:0;transform:scaleX(0)}}.mat-checkbox-checkmark,.mat-checkbox-mixedmark{width:calc(100% - 4px)}.mat-checkbox-background,.mat-checkbox-frame{top:0;left:0;right:0;bottom:0;position:absolute;border-radius:2px;box-sizing:border-box;pointer-events:none}.mat-checkbox{transition:background .4s cubic-bezier(.25,.8,.25,1),box-shadow 280ms cubic-bezier(.4,0,.2,1);cursor:pointer}.mat-checkbox-layout{cursor:inherit;align-items:baseline;vertical-align:middle;display:inline-flex;white-space:nowrap}.mat-checkbox-inner-container{display:inline-block;height:20px;line-height:0;margin:auto;margin-right:8px;order:0;position:relative;vertical-align:middle;white-space:nowrap;width:20px;flex-shrink:0}[dir=rtl] .mat-checkbox-inner-container{margin-left:8px;margin-right:auto}.mat-checkbox-inner-container-no-side-margin{margin-left:0;margin-right:0}.mat-checkbox-frame{background-color:transparent;transition:border-color 90ms cubic-bezier(0,0,.2,.1);border-width:2px;border-style:solid}.mat-checkbox-background{align-items:center;display:inline-flex;justify-content:center;transition:background-color 90ms cubic-bezier(0,0,.2,.1),opacity 90ms cubic-bezier(0,0,.2,.1)}.mat-checkbox-checkmark{top:0;left:0;right:0;bottom:0;position:absolute;width:100%}.mat-checkbox-checkmark-path{stroke-dashoffset:22.91026;stroke-dasharray:22.91026;stroke-width:2.66667px}.mat-checkbox-mixedmark{height:2px;opacity:0;transform:scaleX(0) rotate(0)}.mat-checkbox-label-before .mat-checkbox-inner-container{order:1;margin-left:8px;margin-right:auto}[dir=rtl] .mat-checkbox-label-before .mat-checkbox-inner-container{margin-left:auto;margin-right:8px}.mat-checkbox-checked .mat-checkbox-checkmark{opacity:1}.mat-checkbox-checked .mat-checkbox-checkmark-path{stroke-dashoffset:0}.mat-checkbox-checked .mat-checkbox-mixedmark{transform:scaleX(1) rotate(-45deg)}.mat-checkbox-indeterminate .mat-checkbox-checkmark{opacity:0;transform:rotate(45deg)}.mat-checkbox-indeterminate .mat-checkbox-checkmark-path{stroke-dashoffset:0}.mat-checkbox-indeterminate .mat-checkbox-mixedmark{opacity:1;transform:scaleX(1) rotate(0)}.mat-checkbox-unchecked .mat-checkbox-background{background-color:transparent}.mat-checkbox-disabled{cursor:default}.mat-checkbox-anim-unchecked-checked .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-in-background}.mat-checkbox-anim-unchecked-checked .mat-checkbox-checkmark-path{animation:180ms linear 0s mat-checkbox-unchecked-checked-checkmark-path}.mat-checkbox-anim-unchecked-indeterminate .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-in-background}.mat-checkbox-anim-unchecked-indeterminate .mat-checkbox-mixedmark{animation:90ms linear 0s mat-checkbox-unchecked-indeterminate-mixedmark}.mat-checkbox-anim-checked-unchecked .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-out-background}.mat-checkbox-anim-checked-unchecked .mat-checkbox-checkmark-path{animation:90ms linear 0s mat-checkbox-checked-unchecked-checkmark-path}.mat-checkbox-anim-checked-indeterminate .mat-checkbox-checkmark{animation:90ms linear 0s mat-checkbox-checked-indeterminate-checkmark}.mat-checkbox-anim-checked-indeterminate .mat-checkbox-mixedmark{animation:90ms linear 0s mat-checkbox-checked-indeterminate-mixedmark}.mat-checkbox-anim-indeterminate-checked .mat-checkbox-checkmark{animation:.5s linear 0s mat-checkbox-indeterminate-checked-checkmark}.mat-checkbox-anim-indeterminate-checked .mat-checkbox-mixedmark{animation:.5s linear 0s mat-checkbox-indeterminate-checked-mixedmark}.mat-checkbox-anim-indeterminate-unchecked .mat-checkbox-background{animation:180ms linear 0s mat-checkbox-fade-out-background}.mat-checkbox-anim-indeterminate-unchecked .mat-checkbox-mixedmark{animation:.3s linear 0s mat-checkbox-indeterminate-unchecked-mixedmark}.mat-checkbox-input{bottom:0;left:50%}.mat-checkbox-ripple{position:absolute;left:-15px;top:-15px;height:50px;width:50px;z-index:1;pointer-events:none}"],exportAs:"matCheckbox",host:{class:"mat-checkbox","[id]":"id","[class.mat-checkbox-indeterminate]":"indeterminate","[class.mat-checkbox-checked]":"checked","[class.mat-checkbox-disabled]":"disabled","[class.mat-checkbox-label-before]":'labelPosition == "before"'},providers:[h],inputs:["disabled","disableRipple","color","tabIndex"],encapsulation:s._10.None,preserveWhitespaces:!1,changeDetection:s.i.OnPush}]}],e.ctorParameters=function(){return[{type:s.t},{type:s.j},{type:i.f},{type:void 0,decorators:[{type:s.g,args:["tabindex"]}]},{type:void 0,decorators:[{type:s.M},{type:s.y,args:[p]}]}]},e.propDecorators={ariaLabel:[{type:s.C,args:["aria-label"]}],ariaLabelledby:[{type:s.C,args:["aria-labelledby"]}],id:[{type:s.C}],required:[{type:s.C}],align:[{type:s.C}],labelPosition:[{type:s.C}],name:[{type:s.C}],change:[{type:s.N}],indeterminateChange:[{type:s.N}],value:[{type:s.C}],_inputElement:[{type:s._7,args:["input"]}],_ripple:[{type:s._7,args:[c.r]}],checked:[{type:s.C}],indeterminate:[{type:s.C}]},e}(Object(c.y)(Object(c.u)(Object(c.v)(Object(c.w)(g)),"accent"))),y={provide:a.c,useExisting:Object(s._14)(function(){return b}),multi:!0},b=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r.b)(e,t),e.decorators=[{type:s.s,args:[{selector:"mat-checkbox[required][formControlName],\n             mat-checkbox[required][formControl], mat-checkbox[required][ngModel]",providers:[y],host:{"[attr.required]":'required ? "" : null'}}]}],e.ctorParameters=function(){return[]},e}(a.a),_=function(){function t(){}return t.decorators=[{type:s.H,args:[{imports:[u.a,c.s,c.j,l.a,i.a],exports:[v,b,c.j],declarations:[v,b]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/chips.es5.js":function(t,e,n){"use strict";var r=n("../../../core/esm5/core.js"),i=n("../../../material/esm5/core.es5.js"),o=n("../../../../tslib/tslib.es6.js"),s=n("../../../cdk/esm5/a11y.es5.js"),a=n("../../../cdk/esm5/bidi.es5.js"),c=n("../../../cdk/esm5/coercion.es5.js"),u=n("../../../cdk/esm5/collections.es5.js"),l=n("../../../cdk/esm5/keycodes.es5.js"),p=n("../../../../rxjs/_esm5/operators/startWith.js"),d=n("../../../forms/esm5/forms.js"),h=n("../../../material/esm5/form-field.es5.js"),f=n("../../../../rxjs/_esm5/observable/merge.js"),m=n("../../../../rxjs/_esm5/Subscription.js"),g=n("../../../../rxjs/_esm5/Subject.js"),v=(function(){}(),function(){return function(t){this._elementRef=t}}()),y=Object(i.u)(Object(i.w)(v),"primary"),b=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"mat-basic-chip, [mat-basic-chip]",host:{class:"mat-basic-chip"}}]}],t.ctorParameters=function(){return[]},t}(),_=function(t){function e(e){var n=t.call(this,e)||this;return n._elementRef=e,n._selected=!1,n._selectable=!0,n._removable=!0,n._hasFocus=!1,n._onFocus=new g.a,n._onBlur=new g.a,n.selectionChange=new r.w,n.destroyed=new r.w,n.destroy=n.destroyed,n.removed=new r.w,n.onRemove=n.removed,n}return Object(o.b)(e,t),Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=Object(c.b)(t),this.selectionChange.emit({source:this,isUserInput:!1,selected:t})},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return void 0!=this._value?this._value:this._elementRef.nativeElement.textContent},set:function(t){this._value=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selectable",{get:function(){return this._selectable},set:function(t){this._selectable=Object(c.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"removable",{get:function(){return this._removable},set:function(t){this._removable=Object(c.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"ariaSelected",{get:function(){return this.selectable?this.selected.toString():null},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this.destroyed.emit({chip:this})},e.prototype.select=function(){this._selected=!0,this.selectionChange.emit({source:this,isUserInput:!1,selected:!0})},e.prototype.deselect=function(){this._selected=!1,this.selectionChange.emit({source:this,isUserInput:!1,selected:!1})},e.prototype.selectViaInteraction=function(){this._selected=!0,this.selectionChange.emit({source:this,isUserInput:!0,selected:!0})},e.prototype.toggleSelected=function(t){return void 0===t&&(t=!1),this._selected=!this.selected,this.selectionChange.emit({source:this,isUserInput:t,selected:this._selected}),this.selected},e.prototype.focus=function(){this._elementRef.nativeElement.focus(),this._onFocus.next({chip:this})},e.prototype.remove=function(){this.removable&&this.removed.emit({chip:this})},e.prototype._handleClick=function(t){this.disabled||(t.preventDefault(),t.stopPropagation(),this.focus())},e.prototype._handleKeydown=function(t){if(!this.disabled)switch(t.keyCode){case l.c:case l.b:this.remove(),t.preventDefault();break;case l.n:this.selectable&&this.toggleSelected(!0),t.preventDefault()}},e.prototype._blur=function(){this._hasFocus=!1,this._onBlur.next({chip:this})},e.decorators=[{type:r.s,args:[{selector:"mat-basic-chip, [mat-basic-chip], mat-chip, [mat-chip]",inputs:["color","disabled"],exportAs:"matChip",host:{class:"mat-chip",tabindex:"-1",role:"option","[class.mat-chip-selected]":"selected","[attr.disabled]":"disabled || null","[attr.aria-disabled]":"disabled.toString()","[attr.aria-selected]":"ariaSelected","(click)":"_handleClick($event)","(keydown)":"_handleKeydown($event)","(focus)":"_hasFocus = true","(blur)":"_blur()"}}]}],e.ctorParameters=function(){return[{type:r.t}]},e.propDecorators={selected:[{type:r.C}],value:[{type:r.C}],selectable:[{type:r.C}],removable:[{type:r.C}],selectionChange:[{type:r.N}],destroyed:[{type:r.N}],destroy:[{type:r.N}],removed:[{type:r.N}],onRemove:[{type:r.N,args:["remove"]}]},e}(y),j=function(){function t(t){this._parentChip=t}return t.prototype._handleClick=function(){this._parentChip.removable&&this._parentChip.remove()},t.decorators=[{type:r.s,args:[{selector:"[matChipRemove]",host:{class:"mat-chip-remove","(click)":"_handleClick()"}}]}],t.ctorParameters=function(){return[{type:_}]},t}(),x=function(){return function(t,e,n,r){this._defaultErrorStateMatcher=t,this._parentForm=e,this._parentFormGroup=n,this.ngControl=r}}(),w=Object(i.x)(x),O=0,C=function(){return function(t,e){this.source=t,this.value=e}}(),k=function(t){function e(e,n,i,o,s,a,c){var u=t.call(this,a,o,s,c)||this;return u._elementRef=e,u._changeDetectorRef=n,u._dir=i,u.ngControl=c,u.controlType="mat-chip-list",u._lastDestroyedIndex=null,u._chipSet=new WeakMap,u._tabOutSubscription=m.a.EMPTY,u._selectable=!0,u._multiple=!1,u._uid="mat-chip-list-"+O++,u._required=!1,u._disabled=!1,u._tabIndex=0,u._userTabIndex=null,u._onTouched=function(){},u._onChange=function(){},u._compareWith=function(t,e){return t===e},u.ariaOrientation="horizontal",u.change=new r.w,u.valueChange=new r.w,u.ngControl&&(u.ngControl.valueAccessor=u),u}return Object(o.b)(e,t),Object.defineProperty(e.prototype,"selected",{get:function(){return this.multiple?this._selectionModel.selected:this._selectionModel.selected[0]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"role",{get:function(){return this.empty?null:"listbox"},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"multiple",{get:function(){return this._multiple},set:function(t){this._multiple=Object(c.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"compareWith",{get:function(){return this._compareWith},set:function(t){this._compareWith=t,this._selectionModel&&this._initializeSelection()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(t){this.writeValue(t),this._value=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"id",{get:function(){return this._id||this._uid},set:function(t){this._id=t,this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(c.b)(t),this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"placeholder",{get:function(){return this._chipInput?this._chipInput.placeholder:this._placeholder},set:function(t){this._placeholder=t,this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"focused",{get:function(){return this.chips.some(function(t){return t._hasFocus})||this._chipInput&&this._chipInput.focused},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"empty",{get:function(){return(!this._chipInput||this._chipInput.empty)&&0===this.chips.length},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"shouldLabelFloat",{get:function(){return!this.empty||this.focused},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this.ngControl?this.ngControl.disabled:this._disabled},set:function(t){this._disabled=Object(c.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selectable",{get:function(){return this._selectable},set:function(t){this._selectable=Object(c.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"tabIndex",{set:function(t){this._userTabIndex=t,this._tabIndex=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"chipSelectionChanges",{get:function(){return f.a.apply(void 0,this.chips.map(function(t){return t.selectionChange}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"chipFocusChanges",{get:function(){return f.a.apply(void 0,this.chips.map(function(t){return t._onFocus}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"chipBlurChanges",{get:function(){return f.a.apply(void 0,this.chips.map(function(t){return t._onBlur}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"chipRemoveChanges",{get:function(){return f.a.apply(void 0,this.chips.map(function(t){return t.destroy}))},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){var t=this;this._keyManager=new s.e(this.chips).withWrap(),this._tabOutSubscription=this._keyManager.tabOut.subscribe(function(){t._tabIndex=-1,setTimeout(function(){return t._tabIndex=t._userTabIndex||0})}),this._changeSubscription=this.chips.changes.pipe(Object(p.a)(null)).subscribe(function(){t._resetChips(),t._initializeSelection(),t._updateTabIndex(),t._updateFocusForDestroyedChips()})},e.prototype.ngOnInit=function(){this._selectionModel=new u.a(this.multiple,void 0,!1),this.stateChanges.next()},e.prototype.ngDoCheck=function(){this.ngControl&&this.updateErrorState()},e.prototype.ngOnDestroy=function(){this._tabOutSubscription.unsubscribe(),this._changeSubscription&&this._changeSubscription.unsubscribe(),this._dropSubscriptions(),this.stateChanges.complete()},e.prototype.registerInput=function(t){this._chipInput=t},e.prototype.setDescribedByIds=function(t){this._ariaDescribedby=t.join(" ")},e.prototype.writeValue=function(t){this.chips&&this._setSelectionByValue(t,!1)},e.prototype.registerOnChange=function(t){this._onChange=t},e.prototype.registerOnTouched=function(t){this._onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t,this._elementRef.nativeElement.disabled=t,this.stateChanges.next()},e.prototype.onContainerClick=function(){this.focus()},e.prototype.focus=function(){this._chipInput&&this._chipInput.focused||(this.chips.length>0?(this._keyManager.setFirstItemActive(),this.stateChanges.next()):(this._focusInput(),this.stateChanges.next()))},e.prototype._focusInput=function(){this._chipInput&&this._chipInput.focus()},e.prototype._keydown=function(t){var e=t.keyCode,n=t.target,r=this._isInputEmpty(n),i=this._dir&&"rtl"==this._dir.value,o=e===(i?l.m:l.i),s=e===(i?l.i:l.m),a=e===l.b;if(r&&a)return this._keyManager.setLastItemActive(),void t.preventDefault();n&&n.classList.contains("mat-chip")&&(o?(this._keyManager.setPreviousItemActive(),t.preventDefault()):s?(this._keyManager.setNextItemActive(),t.preventDefault()):this._keyManager.onKeydown(t)),this.stateChanges.next()},e.prototype._updateTabIndex=function(){this._tabIndex=this._userTabIndex||(0===this.chips.length?-1:0)},e.prototype._updateKeyManager=function(t){var e=this.chips.toArray().indexOf(t);this._isValidIndex(e)&&(t._hasFocus&&(e<this.chips.length-1?this._keyManager.setActiveItem(e):e-1>=0&&this._keyManager.setActiveItem(e-1)),this._keyManager.activeItemIndex===e&&(this._lastDestroyedIndex=e))},e.prototype._updateFocusForDestroyedChips=function(){var t=this.chips;if(null!=this._lastDestroyedIndex&&t.length>0){var e=Math.min(this._lastDestroyedIndex,t.length-1);this._keyManager.setActiveItem(e);var n=this._keyManager.activeItem;n&&n.focus()}this._lastDestroyedIndex=null},e.prototype._isValidIndex=function(t){return t>=0&&t<this.chips.length},e.prototype._isInputEmpty=function(t){return!(!t||"input"!==t.nodeName.toLowerCase())&&!t.value},e.prototype._setSelectionByValue=function(t,e){var n=this;if(void 0===e&&(e=!0),this._clearSelection(),this.chips.forEach(function(t){return t.deselect()}),Array.isArray(t))t.forEach(function(t){return n._selectValue(t,e)}),this._sortValues();else{var r=this._selectValue(t,e);if(r){var i=this.chips.toArray().indexOf(r);e?this._keyManager.setActiveItem(i):this._keyManager.updateActiveItemIndex(i)}}},e.prototype._selectValue=function(t,e){var n=this;void 0===e&&(e=!0);var r=this.chips.find(function(e){return null!=e.value&&n._compareWith(e.value,t)});return r&&(e?r.selectViaInteraction():r.select(),this._selectionModel.select(r)),r},e.prototype._initializeSelection=function(){var t=this;Promise.resolve().then(function(){(t.ngControl||t._value)&&(t._setSelectionByValue(t.ngControl?t.ngControl.value:t._value,!1),t.stateChanges.next())})},e.prototype._clearSelection=function(t){this._selectionModel.clear(),this.chips.forEach(function(e){e!==t&&e.deselect()}),this.stateChanges.next()},e.prototype._sortValues=function(){var t=this;this._multiple&&(this._selectionModel.clear(),this.chips.forEach(function(e){e.selected&&t._selectionModel.select(e)}),this.stateChanges.next())},e.prototype._propagateChanges=function(t){var e=null;e=Array.isArray(this.selected)?this.selected.map(function(t){return t.value}):this.selected?this.selected.value:t,this._value=e,this.change.emit(new C(this,e)),this.valueChange.emit(e),this._onChange(e),this._changeDetectorRef.markForCheck()},e.prototype._blur=function(){var t=this;this.disabled||(this._chipInput?setTimeout(function(){t.focused||t._markAsTouched()}):this._markAsTouched())},e.prototype._markAsTouched=function(){this._onTouched(),this._changeDetectorRef.markForCheck(),this.stateChanges.next()},e.prototype._resetChips=function(){this._dropSubscriptions(),this._listenToChipsFocus(),this._listenToChipsSelection(),this._listenToChipsRemoved()},e.prototype._dropSubscriptions=function(){this._chipFocusSubscription&&(this._chipFocusSubscription.unsubscribe(),this._chipFocusSubscription=null),this._chipBlurSubscription&&(this._chipBlurSubscription.unsubscribe(),this._chipBlurSubscription=null),this._chipSelectionSubscription&&(this._chipSelectionSubscription.unsubscribe(),this._chipSelectionSubscription=null)},e.prototype._listenToChipsSelection=function(){var t=this;this._chipSelectionSubscription=this.chipSelectionChanges.subscribe(function(e){e.source.selected?t._selectionModel.select(e.source):t._selectionModel.deselect(e.source),t.multiple||t.chips.forEach(function(e){!t._selectionModel.isSelected(e)&&e.selected&&e.deselect()}),e.isUserInput&&t._propagateChanges()})},e.prototype._listenToChipsFocus=function(){var t=this;this._chipFocusSubscription=this.chipFocusChanges.subscribe(function(e){var n=t.chips.toArray().indexOf(e.chip);t._isValidIndex(n)&&t._keyManager.updateActiveItemIndex(n),t.stateChanges.next()}),this._chipBlurSubscription=this.chipBlurChanges.subscribe(function(e){t._blur(),t.stateChanges.next()})},e.prototype._listenToChipsRemoved=function(){var t=this;this._chipRemoveSubscription=this.chipRemoveChanges.subscribe(function(e){t._updateKeyManager(e.chip)})},e.decorators=[{type:r.m,args:[{selector:"mat-chip-list",template:'<div class="mat-chip-list-wrapper"><ng-content></ng-content></div>',exportAs:"matChipList",host:{"[attr.tabindex]":"_tabIndex","[attr.aria-describedby]":"_ariaDescribedby || null","[attr.aria-required]":"required.toString()","[attr.aria-disabled]":"disabled.toString()","[attr.aria-invalid]":"errorState","[attr.aria-multiselectable]":"multiple","[attr.role]":"role","[class.mat-chip-list-disabled]":"disabled","[class.mat-chip-list-invalid]":"errorState","[class.mat-chip-list-required]":"required","[attr.aria-orientation]":"ariaOrientation",class:"mat-chip-list","(focus)":"focus()","(blur)":"_blur()","(keydown)":"_keydown($event)"},providers:[{provide:h.b,useExisting:e}],styles:[".mat-chip-list-wrapper{display:flex;flex-direction:row;flex-wrap:wrap;align-items:baseline}.mat-chip:not(.mat-basic-chip){transition:box-shadow 280ms cubic-bezier(.4,0,.2,1);display:inline-flex;padding:7px 12px;border-radius:24px;align-items:center;cursor:default}.mat-chip:not(.mat-basic-chip)+.mat-chip:not(.mat-basic-chip){margin:0 0 0 8px}[dir=rtl] .mat-chip:not(.mat-basic-chip)+.mat-chip:not(.mat-basic-chip){margin:0 8px 0 0}.mat-form-field-prefix .mat-chip:not(.mat-basic-chip):last-child{margin-right:8px}[dir=rtl] .mat-form-field-prefix .mat-chip:not(.mat-basic-chip):last-child{margin-left:8px}.mat-chip:not(.mat-basic-chip) .mat-chip-remove.mat-icon{width:1em;height:1em}.mat-chip:not(.mat-basic-chip):focus{box-shadow:0 3px 3px -2px rgba(0,0,0,.2),0 3px 4px 0 rgba(0,0,0,.14),0 1px 8px 0 rgba(0,0,0,.12);outline:0}@media screen and (-ms-high-contrast:active){.mat-chip:not(.mat-basic-chip){outline:solid 1px}}.mat-chip-list-stacked .mat-chip-list-wrapper{display:block}.mat-chip-list-stacked .mat-chip-list-wrapper .mat-chip:not(.mat-basic-chip){display:block;margin:0;margin-bottom:8px}[dir=rtl] .mat-chip-list-stacked .mat-chip-list-wrapper .mat-chip:not(.mat-basic-chip){margin:0;margin-bottom:8px}.mat-chip-list-stacked .mat-chip-list-wrapper .mat-chip:not(.mat-basic-chip):last-child,[dir=rtl] .mat-chip-list-stacked .mat-chip-list-wrapper .mat-chip:not(.mat-basic-chip):last-child{margin-bottom:0}.mat-form-field-prefix .mat-chip-list-wrapper{margin-bottom:8px}.mat-chip-remove{margin-right:-4px;margin-left:6px;cursor:pointer}[dir=rtl] .mat-chip-remove{margin-right:6px;margin-left:-4px}input.mat-chip-input{width:150px;margin:3px;flex:1 0 150px}"],encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush}]}],e.ctorParameters=function(){return[{type:r.t},{type:r.j},{type:a.b,decorators:[{type:r.M}]},{type:d.f,decorators:[{type:r.M}]},{type:d.b,decorators:[{type:r.M}]},{type:i.d},{type:d.e,decorators:[{type:r.M},{type:r.Z}]}]},e.propDecorators={errorStateMatcher:[{type:r.C}],multiple:[{type:r.C}],compareWith:[{type:r.C}],value:[{type:r.C}],id:[{type:r.C}],required:[{type:r.C}],placeholder:[{type:r.C}],disabled:[{type:r.C}],ariaOrientation:[{type:r.C,args:["aria-orientation"]}],selectable:[{type:r.C}],tabIndex:[{type:r.C}],change:[{type:r.N}],valueChange:[{type:r.N}],chips:[{type:r.r,args:[_]}]},e}(w),E=function(){function t(t){this._elementRef=t,this.focused=!1,this._addOnBlur=!1,this.separatorKeyCodes=[l.f],this.chipEnd=new r.w,this.placeholder="",this._inputElement=this._elementRef.nativeElement}return Object.defineProperty(t.prototype,"chipList",{set:function(t){t&&(this._chipList=t,this._chipList.registerInput(this))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"addOnBlur",{get:function(){return this._addOnBlur},set:function(t){this._addOnBlur=Object(c.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"empty",{get:function(){var t=this._inputElement.value;return null==t||""===t},enumerable:!0,configurable:!0}),t.prototype._keydown=function(t){this._emitChipEnd(t)},t.prototype._blur=function(){this.addOnBlur&&this._emitChipEnd(),this.focused=!1,this._chipList.focused||this._chipList._blur(),this._chipList.stateChanges.next()},t.prototype._focus=function(){this.focused=!0,this._chipList.stateChanges.next()},t.prototype._emitChipEnd=function(t){!this._inputElement.value&&t&&this._chipList._keydown(t),(!t||this.separatorKeyCodes.indexOf(t.keyCode)>-1)&&(this.chipEnd.emit({input:this._inputElement,value:this._inputElement.value}),t&&t.preventDefault())},t.prototype._onInput=function(){this._chipList.stateChanges.next()},t.prototype.focus=function(){this._inputElement.focus()},t.decorators=[{type:r.s,args:[{selector:"input[matChipInputFor]",exportAs:"matChipInput, matChipInputFor",host:{class:"mat-chip-input mat-input-element","(keydown)":"_keydown($event)","(blur)":"_blur()","(focus)":"_focus()","(input)":"_onInput()"}}]}],t.ctorParameters=function(){return[{type:r.t}]},t.propDecorators={chipList:[{type:r.C,args:["matChipInputFor"]}],addOnBlur:[{type:r.C,args:["matChipInputAddOnBlur"]}],separatorKeyCodes:[{type:r.C,args:["matChipInputSeparatorKeyCodes"]}],chipEnd:[{type:r.N,args:["matChipInputTokenEnd"]}],placeholder:[{type:r.C}]},t}();!function(){function t(){}t.decorators=[{type:r.H,args:[{imports:[],exports:[k,_,E,j,j,b],declarations:[k,_,E,j,j,b],providers:[i.d]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/core.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return d}),n.d(e,"b",function(){return h}),n.d(e,"j",function(){return m}),n.d(e,"w",function(){return g}),n.d(e,"u",function(){return v}),n.d(e,"v",function(){return y}),n.d(e,"y",function(){return b}),n.d(e,"x",function(){return _}),n.d(e,"c",function(){return w}),n.d(e,"f",function(){return O}),n.d(e,"d",function(){return I}),n.d(e,"e",function(){return N}),n.d(e,"k",function(){return L}),n.d(e,"m",function(){return F}),n.d(e,"l",function(){return V}),n.d(e,"p",function(){return tt}),n.d(e,"h",function(){return Z}),n.d(e,"o",function(){return J}),n.d(e,"n",function(){return K}),n.d(e,"g",function(){return et}),n.d(e,"r",function(){return U}),n.d(e,"i",function(){return H}),n.d(e,"s",function(){return W}),n.d(e,"q",function(){return X}),n.d(e,"t",function(){return nt});var r=n("../../../core/esm5/core.js"),i=n("../../../cdk/esm5/bidi.es5.js"),o=n("../../../../tslib/tslib.es6.js"),s=n("../../../cdk/esm5/coercion.es5.js"),a=n("../../../../rxjs/_esm5/Subject.js"),c=n("../../../platform-browser/esm5/platform-browser.js"),u=n("../../../common/esm5/common.js"),l=n("../../../cdk/esm5/platform.es5.js"),p=n("../../../cdk/esm5/keycodes.es5.js"),d=function(){function t(){}return t.STANDARD_CURVE="cubic-bezier(0.4,0.0,0.2,1)",t.DECELERATION_CURVE="cubic-bezier(0.0,0.0,0.2,1)",t.ACCELERATION_CURVE="cubic-bezier(0.4,0.0,1,1)",t.SHARP_CURVE="cubic-bezier(0.4,0.0,0.6,1)",t}(),h=function(){function t(){}return t.COMPLEX="375ms",t.ENTERING="225ms",t.EXITING="195ms",t}(),f=new r.A("mat-sanity-checks"),m=function(){function t(t){this._sanityChecksEnabled=t,this._hasDoneGlobalChecks=!1,this._hasCheckedHammer=!1,this._document="object"==typeof document&&document?document:null,this._areChecksEnabled()&&!this._hasDoneGlobalChecks&&(this._checkDoctypeIsDefined(),this._checkThemeIsPresent(),this._hasDoneGlobalChecks=!0)}return t.prototype._areChecksEnabled=function(){return this._sanityChecksEnabled&&Object(r._16)()&&!this._isTestEnv()},t.prototype._isTestEnv=function(){return window.__karma__||window.jasmine},t.prototype._checkDoctypeIsDefined=function(){this._document&&!this._document.doctype&&console.warn("Current document does not have a doctype. This may cause some Angular Material components not to behave as expected.")},t.prototype._checkThemeIsPresent=function(){if(this._document&&"function"==typeof getComputedStyle){var t=this._document.createElement("div");t.classList.add("mat-theme-loaded-marker"),this._document.body.appendChild(t);var e=getComputedStyle(t);e&&"none"!==e.display&&console.warn("Could not find Angular Material core theme. Most Material components may not work as expected. For more info refer to the theming guide: https://material.angular.io/guide/theming"),this._document.body.removeChild(t)}},t.prototype._checkHammerIsAvailable=function(){!this._areChecksEnabled()||this._hasCheckedHammer||window.Hammer||console.warn("Could not find HammerJS. Certain Angular Material components may not work correctly."),this._hasCheckedHammer=!0},t.decorators=[{type:r.H,args:[{imports:[i.a],exports:[i.a],providers:[{provide:f,useValue:!0}]}]}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.M},{type:r.y,args:[f]}]}]},t}();function g(t){return function(t){function e(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=t.apply(this,e)||this;return r._disabled=!1,r}return Object(o.b)(e,t),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(s.b)(t)},enumerable:!0,configurable:!0}),e}(t)}function v(t,e){return function(t){function n(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var i=t.apply(this,n)||this;return i.color=e,i}return Object(o.b)(n,t),Object.defineProperty(n.prototype,"color",{get:function(){return this._color},set:function(t){var n=t||e;n!==this._color&&(this._color&&this._elementRef.nativeElement.classList.remove("mat-"+this._color),n&&this._elementRef.nativeElement.classList.add("mat-"+n),this._color=n)},enumerable:!0,configurable:!0}),n}(t)}function y(t){return function(t){function e(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=t.apply(this,e)||this;return r._disableRipple=!1,r}return Object(o.b)(e,t),Object.defineProperty(e.prototype,"disableRipple",{get:function(){return this._disableRipple},set:function(t){this._disableRipple=Object(s.b)(t)},enumerable:!0,configurable:!0}),e}(t)}function b(t,e){return void 0===e&&(e=0),function(t){function n(){for(var n=[],r=0;r<arguments.length;r++)n[r]=arguments[r];var i=t.apply(this,n)||this;return i._tabIndex=e,i}return Object(o.b)(n,t),Object.defineProperty(n.prototype,"tabIndex",{get:function(){return this.disabled?-1:this._tabIndex},set:function(t){this._tabIndex=null!=t?t:e},enumerable:!0,configurable:!0}),n}(t)}function _(t){return function(t){function e(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];var r=t.apply(this,e)||this;return r.errorState=!1,r.stateChanges=new a.a,r}return Object(o.b)(e,t),e.prototype.updateErrorState=function(){var t=this.errorState,e=this._parentFormGroup||this._parentForm,n=this.errorStateMatcher||this._defaultErrorStateMatcher,r=this.ngControl?this.ngControl.control:null,i=n.isErrorState(r,e);i!==t&&(this.errorState=i,this.stateChanges.next())},e}(t)}var j=new r.A("MAT_DATE_LOCALE"),x={provide:j,useExisting:r.F},w=function(){function t(){this._localeChanges=new a.a}return Object.defineProperty(t.prototype,"localeChanges",{get:function(){return this._localeChanges},enumerable:!0,configurable:!0}),t.prototype.deserialize=function(t){return null==t||this.isDateInstance(t)&&this.isValid(t)?t:this.invalid()},t.prototype.setLocale=function(t){this.locale=t,this._localeChanges.next()},t.prototype.compareDate=function(t,e){return this.getYear(t)-this.getYear(e)||this.getMonth(t)-this.getMonth(e)||this.getDate(t)-this.getDate(e)},t.prototype.sameDate=function(t,e){if(t&&e){var n=this.isValid(t),r=this.isValid(e);return n&&r?!this.compareDate(t,e):n==r}return t==e},t.prototype.clampDate=function(t,e,n){return e&&this.compareDate(t,e)<0?e:n&&this.compareDate(t,n)>0?n:t},t}(),O=new r.A("mat-date-formats"),C="undefined"!=typeof Intl,k={long:["January","February","March","April","May","June","July","August","September","October","November","December"],short:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],narrow:["J","F","M","A","M","J","J","A","S","O","N","D"]},E=M(31,function(t){return String(t+1)}),S={long:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],short:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],narrow:["S","M","T","W","T","F","S"]},P=/^\d{4}-\d{2}-\d{2}(?:T\d{2}:\d{2}:\d{2}(?:\.\d+)?(?:Z|(?:(?:\+|-)\d{2}:\d{2}))?)?$/;function M(t,e){for(var n=Array(t),r=0;r<t;r++)n[r]=e(r);return n}var A=function(t){function e(e){var n=t.call(this)||this;return t.prototype.setLocale.call(n,e),n.useUtcForDisplay=!("object"==typeof document&&document&&/(msie|trident)/i.test(navigator.userAgent)),n}return Object(o.b)(e,t),e.prototype.getYear=function(t){return t.getFullYear()},e.prototype.getMonth=function(t){return t.getMonth()},e.prototype.getDate=function(t){return t.getDate()},e.prototype.getDayOfWeek=function(t){return t.getDay()},e.prototype.getMonthNames=function(t){var e=this;if(C){var n=new Intl.DateTimeFormat(this.locale,{month:t});return M(12,function(t){return e._stripDirectionalityCharacters(n.format(new Date(2017,t,1)))})}return k[t]},e.prototype.getDateNames=function(){var t=this;if(C){var e=new Intl.DateTimeFormat(this.locale,{day:"numeric"});return M(31,function(n){return t._stripDirectionalityCharacters(e.format(new Date(2017,0,n+1)))})}return E},e.prototype.getDayOfWeekNames=function(t){var e=this;if(C){var n=new Intl.DateTimeFormat(this.locale,{weekday:t});return M(7,function(t){return e._stripDirectionalityCharacters(n.format(new Date(2017,0,t+1)))})}return S[t]},e.prototype.getYearName=function(t){if(C){var e=new Intl.DateTimeFormat(this.locale,{year:"numeric"});return this._stripDirectionalityCharacters(e.format(t))}return String(this.getYear(t))},e.prototype.getFirstDayOfWeek=function(){return 0},e.prototype.getNumDaysInMonth=function(t){return this.getDate(this._createDateWithOverflow(this.getYear(t),this.getMonth(t)+1,0))},e.prototype.clone=function(t){return this.createDate(this.getYear(t),this.getMonth(t),this.getDate(t))},e.prototype.createDate=function(t,e,n){if(e<0||e>11)throw Error('Invalid month index "'+e+'". Month index has to be between 0 and 11.');if(n<1)throw Error('Invalid date "'+n+'". Date has to be greater than 0.');var r=this._createDateWithOverflow(t,e,n);if(r.getMonth()!=e)throw Error('Invalid date "'+n+'" for month with index "'+e+'".');return r},e.prototype.today=function(){return new Date},e.prototype.parse=function(t){return"number"==typeof t?new Date(t):t?new Date(Date.parse(t)):null},e.prototype.format=function(t,e){if(!this.isValid(t))throw Error("NativeDateAdapter: Cannot format invalid date.");if(C){this.useUtcForDisplay&&(t=new Date(Date.UTC(t.getFullYear(),t.getMonth(),t.getDate(),t.getHours(),t.getMinutes(),t.getSeconds(),t.getMilliseconds())),e=Object(o.a)({},e,{timeZone:"utc"}));var n=new Intl.DateTimeFormat(this.locale,e);return this._stripDirectionalityCharacters(n.format(t))}return this._stripDirectionalityCharacters(t.toDateString())},e.prototype.addCalendarYears=function(t,e){return this.addCalendarMonths(t,12*e)},e.prototype.addCalendarMonths=function(t,e){var n=this._createDateWithOverflow(this.getYear(t),this.getMonth(t)+e,this.getDate(t));return this.getMonth(n)!=((this.getMonth(t)+e)%12+12)%12&&(n=this._createDateWithOverflow(this.getYear(n),this.getMonth(n),0)),n},e.prototype.addCalendarDays=function(t,e){return this._createDateWithOverflow(this.getYear(t),this.getMonth(t),this.getDate(t)+e)},e.prototype.toIso8601=function(t){return[t.getUTCFullYear(),this._2digit(t.getUTCMonth()+1),this._2digit(t.getUTCDate())].join("-")},e.prototype.deserialize=function(e){if("string"==typeof e){if(!e)return null;if(P.test(e)){var n=new Date(e);if(this.isValid(n))return n}}return t.prototype.deserialize.call(this,e)},e.prototype.isDateInstance=function(t){return t instanceof Date},e.prototype.isValid=function(t){return!isNaN(t.getTime())},e.prototype.invalid=function(){return new Date(NaN)},e.prototype._createDateWithOverflow=function(t,e,n){var r=new Date(t,e,n);return t>=0&&t<100&&r.setFullYear(this.getYear(r)-1900),r},e.prototype._2digit=function(t){return("00"+t).slice(-2)},e.prototype._stripDirectionalityCharacters=function(t){return t.replace(/[\u200e\u200f]/g,"")},e.decorators=[{type:r.z}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.M},{type:r.y,args:[j]}]}]},e}(w),T=function(){function t(){}return t.decorators=[{type:r.H,args:[{providers:[{provide:w,useClass:A},x]}]}],t.ctorParameters=function(){return[]},t}(),D={parse:{dateInput:null},display:{dateInput:{year:"numeric",month:"numeric",day:"numeric"},monthYearLabel:{year:"numeric",month:"short"},dateA11yLabel:{year:"numeric",month:"long",day:"numeric"},monthYearA11yLabel:{year:"numeric",month:"long"}}},I=(function(){function t(){}t.decorators=[{type:r.H,args:[{imports:[T],providers:[{provide:O,useValue:D}]}]}],t.ctorParameters=function(){return[]}}(),function(){function t(){}t.prototype.isErrorState=function(t,e){return!!(t&&t.invalid&&(t.dirty||e&&e.submitted))},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[]}}(),function(){function t(){}return t.prototype.isErrorState=function(t,e){return!!(t&&t.invalid&&(t.touched||e&&e.submitted))},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[]},t}()),R=new r.A("MAT_HAMMER_OPTIONS"),N=function(t){function e(e,n){var r=t.call(this)||this;return r._hammerOptions=e,r._hammer="undefined"!=typeof window?window.Hammer:null,r.events=r._hammer?["longpress","slide","slidestart","slideend","slideright","slideleft"]:[],n&&n._checkHammerIsAvailable(),r}return Object(o.b)(e,t),e.prototype.buildHammer=function(t){var e=new this._hammer(t,this._hammerOptions||void 0),n=new this._hammer.Pan,r=new this._hammer.Swipe,i=new this._hammer.Press,o=this._createRecognizer(n,{event:"slide",threshold:0},r),s=this._createRecognizer(i,{event:"longpress",time:500});return n.recognizeWith(r),e.add([r,i,n,o,s]),e},e.prototype._createRecognizer=function(t,e){for(var n=[],r=2;r<arguments.length;r++)n[r-2]=arguments[r];var i=new t.constructor(e);return n.push(t),n.forEach(function(t){return i.recognizeWith(t)}),i},e.decorators=[{type:r.z}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:r.M},{type:r.y,args:[R]}]},{type:m,decorators:[{type:r.M}]}]},e}(c.e),L=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-line], [matLine]",host:{class:"mat-line"}}]}],t.ctorParameters=function(){return[]},t}(),F=function(){function t(t,e){var n=this;this._lines=t,this._element=e,this._setLineClass(this._lines.length),this._lines.changes.subscribe(function(){n._setLineClass(n._lines.length)})}return t.prototype._setLineClass=function(t){this._resetClasses(),2===t||3===t?this._setClass("mat-"+t+"-line",!0):t>3&&this._setClass("mat-multi-line",!0)},t.prototype._resetClasses=function(){this._setClass("mat-2-line",!1),this._setClass("mat-3-line",!1),this._setClass("mat-multi-line",!1)},t.prototype._setClass=function(t,e){e?this._element.nativeElement.classList.add(t):this._element.nativeElement.classList.remove(t)},t}(),V=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[m],exports:[L,m],declarations:[L]}]}],t.ctorParameters=function(){return[]},t}(),z={FADING_IN:0,VISIBLE:1,FADING_OUT:2,HIDDEN:3};z[z.FADING_IN]="FADING_IN",z[z.VISIBLE]="VISIBLE",z[z.FADING_OUT]="FADING_OUT",z[z.HIDDEN]="HIDDEN";var B=function(){function t(t,e,n){this._renderer=t,this.element=e,this.config=n,this.state=z.HIDDEN}return t.prototype.fadeOut=function(){this._renderer.fadeOutRipple(this)},t}(),q=function(){function t(t,e,n){this._ngZone=e,this._isPointerDown=!1,this._triggerEvents=new Map,this._activeRipples=new Set,this.rippleConfig={},this.rippleDisabled=!1,n.isBrowser&&(this._containerElement=t.nativeElement,this._triggerEvents.set("mousedown",this.onMousedown.bind(this)),this._triggerEvents.set("mouseup",this.onPointerUp.bind(this)),this._triggerEvents.set("mouseleave",this.onPointerUp.bind(this)),this._triggerEvents.set("touchstart",this.onTouchStart.bind(this)),this._triggerEvents.set("touchend",this.onPointerUp.bind(this)),this.setTriggerElement(this._containerElement))}return t.prototype.fadeInRipple=function(t,e,n){var r=this;void 0===n&&(n={});var i=this._containerElement.getBoundingClientRect();n.centered&&(t=i.left+i.width/2,e=i.top+i.height/2);var o,s,a,c,u,l,p=n.radius||(o=t,s=e,a=i,c=Math.max(Math.abs(o-a.left),Math.abs(o-a.right)),u=Math.max(Math.abs(s-a.top),Math.abs(s-a.bottom)),Math.sqrt(c*c+u*u)),d=450/(n.speedFactor||1),h=t-i.left,f=e-i.top,m=document.createElement("div");m.classList.add("mat-ripple-element"),m.style.left=h-p+"px",m.style.top=f-p+"px",m.style.height=2*p+"px",m.style.width=2*p+"px",m.style.backgroundColor=n.color||null,m.style.transitionDuration=d+"ms",this._containerElement.appendChild(m),l=m,window.getComputedStyle(l).getPropertyValue("opacity"),m.style.transform="scale(1)";var g=new B(this,m,n);return g.state=z.FADING_IN,this._activeRipples.add(g),this.runTimeoutOutsideZone(function(){g.state=z.VISIBLE,n.persistent||r._isPointerDown||g.fadeOut()},d),g},t.prototype.fadeOutRipple=function(t){if(this._activeRipples.delete(t)){var e=t.element;e.style.transitionDuration="400ms",e.style.opacity="0",t.state=z.FADING_OUT,this.runTimeoutOutsideZone(function(){t.state=z.HIDDEN,e.parentNode.removeChild(e)},400)}},t.prototype.fadeOutAll=function(){this._activeRipples.forEach(function(t){return t.fadeOut()})},t.prototype.setTriggerElement=function(t){var e=this;this._triggerElement&&this._triggerEvents.forEach(function(t,n){e._triggerElement.removeEventListener(n,t)}),t&&this._ngZone.runOutsideAngular(function(){e._triggerEvents.forEach(function(e,n){return t.addEventListener(n,e)})}),this._triggerElement=t},t.prototype.onMousedown=function(t){var e=this._lastTouchStartEvent&&Date.now()<this._lastTouchStartEvent+800;this.rippleDisabled||e||(this._isPointerDown=!0,this.fadeInRipple(t.clientX,t.clientY,this.rippleConfig))},t.prototype.onTouchStart=function(t){this.rippleDisabled||(this._lastTouchStartEvent=Date.now(),this._isPointerDown=!0,this.fadeInRipple(t.touches[0].clientX,t.touches[0].clientY,this.rippleConfig))},t.prototype.onPointerUp=function(){this._isPointerDown&&(this._isPointerDown=!1,this._activeRipples.forEach(function(t){t.config.persistent||t.state!==z.VISIBLE||t.fadeOut()}))},t.prototype.runTimeoutOutsideZone=function(t,e){void 0===e&&(e=0),this._ngZone.runOutsideAngular(function(){return setTimeout(t,e)})},t}();var H=new r.A("mat-ripple-global-options"),U=function(){function t(t,e,n,r){this.radius=0,this.speedFactor=1,this._rippleRenderer=new q(t,e,n),this._globalOptions=r||{},this._updateRippleRenderer()}return t.prototype.ngOnChanges=function(t){t.trigger&&this.trigger&&this._rippleRenderer.setTriggerElement(this.trigger),this._updateRippleRenderer()},t.prototype.ngOnDestroy=function(){this._rippleRenderer.setTriggerElement(null)},t.prototype.launch=function(t,e,n){return void 0===n&&(n=this.rippleConfig),this._rippleRenderer.fadeInRipple(t,e,n)},t.prototype.fadeOutAll=function(){this._rippleRenderer.fadeOutAll()},Object.defineProperty(t.prototype,"rippleConfig",{get:function(){return{centered:this.centered,speedFactor:this.speedFactor*(this._globalOptions.baseSpeedFactor||1),radius:this.radius,color:this.color}},enumerable:!0,configurable:!0}),t.prototype._updateRippleRenderer=function(){this._rippleRenderer.rippleDisabled=this._globalOptions.disabled||this.disabled,this._rippleRenderer.rippleConfig=this.rippleConfig},t.decorators=[{type:r.s,args:[{selector:"[mat-ripple], [matRipple]",exportAs:"matRipple",host:{class:"mat-ripple","[class.mat-ripple-unbounded]":"unbounded"}}]}],t.ctorParameters=function(){return[{type:r.t},{type:r.L},{type:l.a},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[H]}]}]},t.propDecorators={trigger:[{type:r.C,args:["matRippleTrigger"]}],centered:[{type:r.C,args:["matRippleCentered"]}],disabled:[{type:r.C,args:["matRippleDisabled"]}],radius:[{type:r.C,args:["matRippleRadius"]}],speedFactor:[{type:r.C,args:["matRippleSpeedFactor"]}],color:[{type:r.C,args:["matRippleColor"]}],unbounded:[{type:r.C,args:["matRippleUnbounded"]}]},t}(),W=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[m,l.b],exports:[U,m],declarations:[U]}]}],t.ctorParameters=function(){return[]},t}(),G=function(){function t(){this.state="unchecked",this.disabled=!1}return t.decorators=[{type:r.m,args:[{encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush,selector:"mat-pseudo-checkbox",styles:[".mat-pseudo-checkbox{width:20px;height:20px;border:2px solid;border-radius:2px;cursor:pointer;display:inline-block;vertical-align:middle;box-sizing:border-box;position:relative;flex-shrink:0;transition:border-color 90ms cubic-bezier(0,0,.2,.1),background-color 90ms cubic-bezier(0,0,.2,.1)}.mat-pseudo-checkbox::after{position:absolute;opacity:0;content:'';border-bottom:2px solid currentColor;transition:opacity 90ms cubic-bezier(0,0,.2,.1)}.mat-pseudo-checkbox.mat-pseudo-checkbox-checked,.mat-pseudo-checkbox.mat-pseudo-checkbox-indeterminate{border:none}.mat-pseudo-checkbox-disabled{cursor:default}.mat-pseudo-checkbox-indeterminate::after{top:9px;left:2px;width:16px;opacity:1}.mat-pseudo-checkbox-checked::after{top:5px;left:3px;width:12px;height:5px;border-left:2px solid currentColor;transform:rotate(-45deg);opacity:1}"],template:"",host:{class:"mat-pseudo-checkbox","[class.mat-pseudo-checkbox-indeterminate]":'state === "indeterminate"',"[class.mat-pseudo-checkbox-checked]":'state === "checked"',"[class.mat-pseudo-checkbox-disabled]":"disabled"}}]}],t.ctorParameters=function(){return[]},t.propDecorators={state:[{type:r.C}],disabled:[{type:r.C}]},t}(),X=function(){function t(){}return t.decorators=[{type:r.H,args:[{exports:[G],declarations:[G]}]}],t.ctorParameters=function(){return[]},t}(),Y=0,K=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._labelId="mat-optgroup-label-"+Y++,e}return Object(o.b)(e,t),e.decorators=[{type:r.m,args:[{selector:"mat-optgroup",exportAs:"matOptgroup",template:'<label class="mat-optgroup-label" [id]="_labelId">{{ label }}</label><ng-content select="mat-option"></ng-content>',encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush,inputs:["disabled"],host:{class:"mat-optgroup",role:"group","[class.mat-optgroup-disabled]":"disabled","[attr.aria-disabled]":"disabled.toString()","[attr.aria-labelledby]":"_labelId"}}]}],e.ctorParameters=function(){return[]},e.propDecorators={label:[{type:r.C}]},e}(g(function(){return function(){}}())),$=0,Q=function(){return function(t,e){void 0===e&&(e=!1),this.source=t,this.isUserInput=e}}(),Z=new r.A("MAT_OPTION_PARENT_COMPONENT"),J=function(){function t(t,e,n,i){this._element=t,this._changeDetectorRef=e,this._parent=n,this.group=i,this._selected=!1,this._active=!1,this._disabled=!1,this._id="mat-option-"+$++,this.onSelectionChange=new r.w}return Object.defineProperty(t.prototype,"multiple",{get:function(){return this._parent&&this._parent.multiple},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"id",{get:function(){return this._id},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this.group&&this.group.disabled||this._disabled},set:function(t){this._disabled=Object(s.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disableRipple",{get:function(){return this._parent&&this._parent.disableRipple},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"active",{get:function(){return this._active},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"viewValue",{get:function(){return(this._getHostElement().textContent||"").trim()},enumerable:!0,configurable:!0}),t.prototype.select=function(){this._selected=!0,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent()},t.prototype.deselect=function(){this._selected=!1,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent()},t.prototype.focus=function(){var t=this._getHostElement();"function"==typeof t.focus&&t.focus()},t.prototype.setActiveStyles=function(){this._active||(this._active=!0,this._changeDetectorRef.markForCheck())},t.prototype.setInactiveStyles=function(){this._active&&(this._active=!1,this._changeDetectorRef.markForCheck())},t.prototype.getLabel=function(){return this.viewValue},t.prototype._handleKeydown=function(t){t.keyCode!==p.f&&t.keyCode!==p.n||(this._selectViaInteraction(),t.preventDefault())},t.prototype._selectViaInteraction=function(){this.disabled||(this._selected=!this.multiple||!this._selected,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent(!0))},t.prototype._getTabIndex=function(){return this.disabled?"-1":"0"},t.prototype._getHostElement=function(){return this._element.nativeElement},t.prototype._emitSelectionChangeEvent=function(t){void 0===t&&(t=!1),this.onSelectionChange.emit(new Q(this,t))},t.countGroupLabelsBeforeOption=function(t,e,n){if(n.length){for(var r=e.toArray(),i=n.toArray(),o=0,s=0;s<t+1;s++)r[s].group&&r[s].group===i[o]&&o++;return o}return 0},t.decorators=[{type:r.m,args:[{selector:"mat-option",exportAs:"matOption",host:{role:"option","[attr.tabindex]":"_getTabIndex()","[class.mat-selected]":"selected","[class.mat-option-multiple]":"multiple","[class.mat-active]":"active","[id]":"id","[attr.aria-selected]":"selected.toString()","[attr.aria-disabled]":"disabled.toString()","[class.mat-option-disabled]":"disabled","(click)":"_selectViaInteraction()","(keydown)":"_handleKeydown($event)",class:"mat-option"},template:'<mat-pseudo-checkbox *ngIf="multiple" class="mat-option-pseudo-checkbox" [state]="selected ? \'checked\' : \'\'" [disabled]="disabled"></mat-pseudo-checkbox><span class="mat-option-text"><ng-content></ng-content></span><div class="mat-option-ripple" mat-ripple [matRippleTrigger]="_getHostElement()" [matRippleDisabled]="disabled || disableRipple"></div>',encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush}]}],t.ctorParameters=function(){return[{type:r.t},{type:r.j},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[Z]}]},{type:K,decorators:[{type:r.M}]}]},t.propDecorators={value:[{type:r.C}],disabled:[{type:r.C}],onSelectionChange:[{type:r.N}]},t}(),tt=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[W,u.a,X],exports:[J,K],declarations:[J,K]}]}],t.ctorParameters=function(){return[]},t}(),et=new r.A("mat-label-global-options");function nt(t,e){var n=e.trim();t.style.transform=n,t.style.webkitTransform=n}},"../../../material/esm5/datepicker.es5.js":function(t,e,n){"use strict";var r=n("../../../cdk/esm5/a11y.es5.js"),i=n("../../../cdk/esm5/overlay.es5.js"),o=n("../../../common/esm5/common.js"),s=n("../../../core/esm5/core.js"),a=n("../../../material/esm5/button.es5.js"),c=n("../../../material/esm5/dialog.es5.js"),u=n("../../../material/esm5/icon.es5.js"),l=n("../../../cdk/esm5/keycodes.es5.js"),p=n("../../../material/esm5/core.es5.js"),d=n("../../../../rxjs/_esm5/operators/take.js"),h=n("../../../../rxjs/_esm5/Subject.js"),f=n("../../../cdk/esm5/bidi.es5.js"),m=n("../../../cdk/esm5/coercion.es5.js"),g=n("../../../cdk/esm5/portal.es5.js"),v=n("../../../../rxjs/_esm5/Subscription.js"),y=n("../../../forms/esm5/forms.js"),b=n("../../../material/esm5/form-field.es5.js"),_=n("../../../material/esm5/input.es5.js"),j=n("../../../../rxjs/_esm5/observable/merge.js"),x=n("../../../../rxjs/_esm5/observable/of.js");function w(t){return Error("MatDatepicker: No provider found for "+t+". You must import one of the following modules at your application root: MatNativeDateModule, MatMomentDateModule, or provide a custom implementation.")}var O=function(){function t(){this.changes=new h.a,this.calendarLabel="Calendar",this.openCalendarLabel="Open calendar",this.prevMonthLabel="Previous month",this.nextMonthLabel="Next month",this.prevYearLabel="Previous year",this.nextYearLabel="Next year",this.switchToMonthViewLabel="Change to month view",this.switchToYearViewLabel="Change to year view"}return t.decorators=[{type:s.z}],t.ctorParameters=function(){return[]},t}(),C=function(){return function(t,e,n,r){this.value=t,this.displayValue=e,this.ariaLabel=n,this.enabled=r}}(),k=function(){function t(){this.numCols=7,this.allowDisabledSelection=!1,this.activeCell=0,this.cellAspectRatio=1,this.selectedValueChange=new s.w}return t.prototype._cellClicked=function(t){(this.allowDisabledSelection||t.enabled)&&this.selectedValueChange.emit(t.value)},Object.defineProperty(t.prototype,"_firstRowOffset",{get:function(){return this.rows&&this.rows.length&&this.rows[0].length?this.numCols-this.rows[0].length:0},enumerable:!0,configurable:!0}),t.prototype._isActiveCell=function(t,e){var n=t*this.numCols+e;return t&&(n-=this._firstRowOffset),n==this.activeCell},t.decorators=[{type:s.m,args:[{selector:"[mat-calendar-body]",template:'<tr *ngIf="_firstRowOffset < labelMinRequiredCells" aria-hidden="true"><td class="mat-calendar-body-label" [attr.colspan]="numCols" [style.paddingTop.%]="50 * cellAspectRatio / numCols" [style.paddingBottom.%]="50 * cellAspectRatio / numCols">{{label}}</td></tr><tr *ngFor="let row of rows; let rowIndex = index" role="row"><td *ngIf="rowIndex === 0 && _firstRowOffset" aria-hidden="true" class="mat-calendar-body-label" [attr.colspan]="_firstRowOffset" [style.paddingTop.%]="50 * cellAspectRatio / numCols" [style.paddingBottom.%]="50 * cellAspectRatio / numCols">{{_firstRowOffset >= labelMinRequiredCells ? label : \'\'}}</td><td *ngFor="let item of row; let colIndex = index" role="gridcell" class="mat-calendar-body-cell" [tabindex]="_isActiveCell(rowIndex, colIndex) ? 0 : -1" [class.mat-calendar-body-disabled]="!item.enabled" [class.mat-calendar-body-active]="_isActiveCell(rowIndex, colIndex)" [attr.aria-label]="item.ariaLabel" [attr.aria-disabled]="!item.enabled || null" (click)="_cellClicked(item)" [style.width.%]="100 / numCols" [style.paddingTop.%]="50 * cellAspectRatio / numCols" [style.paddingBottom.%]="50 * cellAspectRatio / numCols"><div class="mat-calendar-body-cell-content" [class.mat-calendar-body-selected]="selectedValue === item.value" [class.mat-calendar-body-today]="todayValue === item.value">{{item.displayValue}}</div></td></tr>',styles:[".mat-calendar-body{min-width:224px}.mat-calendar-body-label{height:0;line-height:0;text-align:left;padding-left:4.71429%;padding-right:4.71429%}.mat-calendar-body-cell{position:relative;height:0;line-height:0;text-align:center;outline:0;cursor:pointer}.mat-calendar-body-disabled{cursor:default}.mat-calendar-body-cell-content{position:absolute;top:5%;left:5%;display:flex;align-items:center;justify-content:center;box-sizing:border-box;width:90%;height:90%;line-height:1;border-width:1px;border-style:solid;border-radius:999px}[dir=rtl] .mat-calendar-body-label{text-align:right}"],host:{class:"mat-calendar-body"},exportAs:"matCalendarBody",encapsulation:s._10.None,preserveWhitespaces:!1,changeDetection:s.i.OnPush}]}],t.ctorParameters=function(){return[]},t.propDecorators={label:[{type:s.C}],rows:[{type:s.C}],todayValue:[{type:s.C}],selectedValue:[{type:s.C}],labelMinRequiredCells:[{type:s.C}],numCols:[{type:s.C}],allowDisabledSelection:[{type:s.C}],activeCell:[{type:s.C}],cellAspectRatio:[{type:s.C}],selectedValueChange:[{type:s.N}]},t}(),E=function(){function t(t,e,n){if(this._dateAdapter=t,this._dateFormats=e,this._changeDetectorRef=n,this.selectedChange=new s.w,this._userSelection=new s.w,!this._dateAdapter)throw w("DateAdapter");if(!this._dateFormats)throw w("MAT_DATE_FORMATS");var r=this._dateAdapter.getFirstDayOfWeek(),i=this._dateAdapter.getDayOfWeekNames("narrow"),o=this._dateAdapter.getDayOfWeekNames("long").map(function(t,e){return{long:t,narrow:i[e]}});this._weekdays=o.slice(r).concat(o.slice(0,r)),this._activeDate=this._dateAdapter.today()}return Object.defineProperty(t.prototype,"activeDate",{get:function(){return this._activeDate},set:function(t){var e=this._activeDate;this._activeDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))||this._dateAdapter.today(),this._hasSameMonthAndYear(e,this._activeDate)||this._init()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(t)),this._selectedDate=this._getDateInCurrentMonth(this._selected)},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._init()},t.prototype._dateSelected=function(t){if(this._selectedDate!=t){var e=this._dateAdapter.getYear(this.activeDate),n=this._dateAdapter.getMonth(this.activeDate),r=this._dateAdapter.createDate(e,n,t);this.selectedChange.emit(r)}this._userSelection.emit()},t.prototype._init=function(){this._selectedDate=this._getDateInCurrentMonth(this.selected),this._todayDate=this._getDateInCurrentMonth(this._dateAdapter.today()),this._monthLabel=this._dateAdapter.getMonthNames("short")[this._dateAdapter.getMonth(this.activeDate)].toLocaleUpperCase();var t=this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),this._dateAdapter.getMonth(this.activeDate),1);this._firstWeekOffset=(7+this._dateAdapter.getDayOfWeek(t)-this._dateAdapter.getFirstDayOfWeek())%7,this._createWeekCells(),this._changeDetectorRef.markForCheck()},t.prototype._createWeekCells=function(){var t=this._dateAdapter.getNumDaysInMonth(this.activeDate),e=this._dateAdapter.getDateNames();this._weeks=[[]];for(var n=0,r=this._firstWeekOffset;n<t;n++,r++){7==r&&(this._weeks.push([]),r=0);var i=this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),this._dateAdapter.getMonth(this.activeDate),n+1),o=!this.dateFilter||this.dateFilter(i),s=this._dateAdapter.format(i,this._dateFormats.display.dateA11yLabel);this._weeks[this._weeks.length-1].push(new C(n+1,e[n],s,o))}},t.prototype._getDateInCurrentMonth=function(t){return t&&this._hasSameMonthAndYear(t,this.activeDate)?this._dateAdapter.getDate(t):null},t.prototype._hasSameMonthAndYear=function(t,e){return!(!t||!e||this._dateAdapter.getMonth(t)!=this._dateAdapter.getMonth(e)||this._dateAdapter.getYear(t)!=this._dateAdapter.getYear(e))},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.decorators=[{type:s.m,args:[{selector:"mat-month-view",template:'<table class="mat-calendar-table"><thead class="mat-calendar-table-header"><tr><th *ngFor="let day of _weekdays" [attr.aria-label]="day.long">{{day.narrow}}</th></tr><tr><th class="mat-calendar-table-header-divider" colspan="7" aria-hidden="true"></th></tr></thead><tbody mat-calendar-body role="grid" [label]="_monthLabel" [rows]="_weeks" [todayValue]="_todayDate" [selectedValue]="_selectedDate" [labelMinRequiredCells]="3" [activeCell]="_dateAdapter.getDate(activeDate) - 1" (selectedValueChange)="_dateSelected($event)"></tbody></table>',exportAs:"matMonthVeiw",encapsulation:s._10.None,preserveWhitespaces:!1,changeDetection:s.i.OnPush}]}],t.ctorParameters=function(){return[{type:p.c,decorators:[{type:s.M}]},{type:void 0,decorators:[{type:s.M},{type:s.y,args:[p.f]}]},{type:s.j}]},t.propDecorators={activeDate:[{type:s.C}],selected:[{type:s.C}],dateFilter:[{type:s.C}],selectedChange:[{type:s.N}],_userSelection:[{type:s.N}]},t}(),S=function(){function t(t,e,n){if(this._dateAdapter=t,this._dateFormats=e,this._changeDetectorRef=n,this.selectedChange=new s.w,!this._dateAdapter)throw w("DateAdapter");if(!this._dateFormats)throw w("MAT_DATE_FORMATS");this._activeDate=this._dateAdapter.today()}return Object.defineProperty(t.prototype,"activeDate",{get:function(){return this._activeDate},set:function(t){var e=this._activeDate;this._activeDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))||this._dateAdapter.today(),this._dateAdapter.getYear(e)!=this._dateAdapter.getYear(this._activeDate)&&this._init()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(t)),this._selectedMonth=this._getMonthInCurrentYear(this._selected)},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._init()},t.prototype._monthSelected=function(t){var e=this._dateAdapter.getNumDaysInMonth(this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),t,1));this.selectedChange.emit(this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),t,Math.min(this._dateAdapter.getDate(this.activeDate),e)))},t.prototype._init=function(){var t=this;this._selectedMonth=this._getMonthInCurrentYear(this.selected),this._todayMonth=this._getMonthInCurrentYear(this._dateAdapter.today()),this._yearLabel=this._dateAdapter.getYearName(this.activeDate);var e=this._dateAdapter.getMonthNames("short");this._months=[[0,1,2,3],[4,5,6,7],[8,9,10,11]].map(function(n){return n.map(function(n){return t._createCellForMonth(n,e[n])})}),this._changeDetectorRef.markForCheck()},t.prototype._getMonthInCurrentYear=function(t){return t&&this._dateAdapter.getYear(t)==this._dateAdapter.getYear(this.activeDate)?this._dateAdapter.getMonth(t):null},t.prototype._createCellForMonth=function(t,e){var n=this._dateAdapter.format(this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),t,1),this._dateFormats.display.monthYearA11yLabel);return new C(t,e.toLocaleUpperCase(),n,this._isMonthEnabled(t))},t.prototype._isMonthEnabled=function(t){if(!this.dateFilter)return!0;for(var e=this._dateAdapter.createDate(this._dateAdapter.getYear(this.activeDate),t,1);this._dateAdapter.getMonth(e)==t;e=this._dateAdapter.addCalendarDays(e,1))if(this.dateFilter(e))return!0;return!1},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.decorators=[{type:s.m,args:[{selector:"mat-year-view",template:'<table class="mat-calendar-table"><thead class="mat-calendar-table-header"><tr><th class="mat-calendar-table-header-divider" colspan="4"></th></tr></thead><tbody mat-calendar-body role="grid" allowDisabledSelection="true" [label]="_yearLabel" [rows]="_months" [todayValue]="_todayMonth" [selectedValue]="_selectedMonth" [labelMinRequiredCells]="2" [numCols]="4" [cellAspectRatio]="4 / 7" [activeCell]="_dateAdapter.getMonth(activeDate)" (selectedValueChange)="_monthSelected($event)"></tbody></table>',exportAs:"matYearView",encapsulation:s._10.None,preserveWhitespaces:!1,changeDetection:s.i.OnPush}]}],t.ctorParameters=function(){return[{type:p.c,decorators:[{type:s.M}]},{type:void 0,decorators:[{type:s.M},{type:s.y,args:[p.f]}]},{type:s.j}]},t.propDecorators={activeDate:[{type:s.C}],selected:[{type:s.C}],dateFilter:[{type:s.C}],selectedChange:[{type:s.N}]},t}(),P=function(){function t(t,e,n,r,i,o){var a=this;if(this._elementRef=t,this._intl=e,this._ngZone=n,this._dateAdapter=r,this._dateFormats=i,this.startView="month",this.selectedChange=new s.w,this._userSelection=new s.w,this._dateFilterForViews=function(t){return!!t&&(!a.dateFilter||a.dateFilter(t))&&(!a.minDate||a._dateAdapter.compareDate(t,a.minDate)>=0)&&(!a.maxDate||a._dateAdapter.compareDate(t,a.maxDate)<=0)},!this._dateAdapter)throw w("DateAdapter");if(!this._dateFormats)throw w("MAT_DATE_FORMATS");this._intlChanges=e.changes.subscribe(function(){return o.markForCheck()})}return Object.defineProperty(t.prototype,"startAt",{get:function(){return this._startAt},set:function(t){this._startAt=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"minDate",{get:function(){return this._minDate},set:function(t){this._minDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"maxDate",{get:function(){return this._maxDate},set:function(t){this._maxDate=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_activeDate",{get:function(){return this._clampedActiveDate},set:function(t){this._clampedActiveDate=this._dateAdapter.clampDate(t,this.minDate,this.maxDate)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_periodButtonText",{get:function(){return this._monthView?this._dateAdapter.format(this._activeDate,this._dateFormats.display.monthYearLabel).toLocaleUpperCase():this._dateAdapter.getYearName(this._activeDate)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_periodButtonLabel",{get:function(){return this._monthView?this._intl.switchToYearViewLabel:this._intl.switchToMonthViewLabel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_prevButtonLabel",{get:function(){return this._monthView?this._intl.prevMonthLabel:this._intl.prevYearLabel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_nextButtonLabel",{get:function(){return this._monthView?this._intl.nextMonthLabel:this._intl.nextYearLabel},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){this._activeDate=this.startAt||this._dateAdapter.today(),this._focusActiveCell(),this._monthView="year"!=this.startView},t.prototype.ngOnDestroy=function(){this._intlChanges.unsubscribe()},t.prototype.ngOnChanges=function(t){var e=t.minDate||t.maxDate||t.dateFilter;if(e&&!e.firstChange){var n=this.monthView||this.yearView;n&&n._init()}},t.prototype._dateSelected=function(t){this._dateAdapter.sameDate(t,this.selected)||this.selectedChange.emit(t)},t.prototype._userSelected=function(){this._userSelection.emit()},t.prototype._monthSelected=function(t){this._activeDate=t,this._monthView=!0},t.prototype._currentPeriodClicked=function(){this._monthView=!this._monthView},t.prototype._previousClicked=function(){this._activeDate=this._monthView?this._dateAdapter.addCalendarMonths(this._activeDate,-1):this._dateAdapter.addCalendarYears(this._activeDate,-1)},t.prototype._nextClicked=function(){this._activeDate=this._monthView?this._dateAdapter.addCalendarMonths(this._activeDate,1):this._dateAdapter.addCalendarYears(this._activeDate,1)},t.prototype._previousEnabled=function(){return!this.minDate||(!this.minDate||!this._isSameView(this._activeDate,this.minDate))},t.prototype._nextEnabled=function(){return!this.maxDate||!this._isSameView(this._activeDate,this.maxDate)},t.prototype._handleCalendarBodyKeydown=function(t){this._monthView?this._handleCalendarBodyKeydownInMonthView(t):this._handleCalendarBodyKeydownInYearView(t)},t.prototype._focusActiveCell=function(){var t=this;this._ngZone.runOutsideAngular(function(){t._ngZone.onStable.asObservable().pipe(Object(d.a)(1)).subscribe(function(){t._elementRef.nativeElement.querySelector(".mat-calendar-body-active").focus()})})},t.prototype._isSameView=function(t,e){return this._monthView?this._dateAdapter.getYear(t)==this._dateAdapter.getYear(e)&&this._dateAdapter.getMonth(t)==this._dateAdapter.getMonth(e):this._dateAdapter.getYear(t)==this._dateAdapter.getYear(e)},t.prototype._handleCalendarBodyKeydownInMonthView=function(t){switch(t.keyCode){case l.i:this._activeDate=this._dateAdapter.addCalendarDays(this._activeDate,-1);break;case l.m:this._activeDate=this._dateAdapter.addCalendarDays(this._activeDate,1);break;case l.p:this._activeDate=this._dateAdapter.addCalendarDays(this._activeDate,-7);break;case l.d:this._activeDate=this._dateAdapter.addCalendarDays(this._activeDate,7);break;case l.h:this._activeDate=this._dateAdapter.addCalendarDays(this._activeDate,1-this._dateAdapter.getDate(this._activeDate));break;case l.e:this._activeDate=this._dateAdapter.addCalendarDays(this._activeDate,this._dateAdapter.getNumDaysInMonth(this._activeDate)-this._dateAdapter.getDate(this._activeDate));break;case l.l:this._activeDate=t.altKey?this._dateAdapter.addCalendarYears(this._activeDate,-1):this._dateAdapter.addCalendarMonths(this._activeDate,-1);break;case l.k:this._activeDate=t.altKey?this._dateAdapter.addCalendarYears(this._activeDate,1):this._dateAdapter.addCalendarMonths(this._activeDate,1);break;case l.f:return void(this._dateFilterForViews(this._activeDate)&&(this._dateSelected(this._activeDate),this._userSelected(),t.preventDefault()));default:return}this._focusActiveCell(),t.preventDefault()},t.prototype._handleCalendarBodyKeydownInYearView=function(t){switch(t.keyCode){case l.i:this._activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,-1);break;case l.m:this._activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,1);break;case l.p:this._activeDate=this._prevMonthInSameCol(this._activeDate);break;case l.d:this._activeDate=this._nextMonthInSameCol(this._activeDate);break;case l.h:this._activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,-this._dateAdapter.getMonth(this._activeDate));break;case l.e:this._activeDate=this._dateAdapter.addCalendarMonths(this._activeDate,11-this._dateAdapter.getMonth(this._activeDate));break;case l.l:this._activeDate=this._dateAdapter.addCalendarYears(this._activeDate,t.altKey?-10:-1);break;case l.k:this._activeDate=this._dateAdapter.addCalendarYears(this._activeDate,t.altKey?10:1);break;case l.f:this._monthSelected(this._activeDate);break;default:return}this._focusActiveCell(),t.preventDefault()},t.prototype._prevMonthInSameCol=function(t){return this._dateAdapter.addCalendarMonths(t,-4)},t.prototype._nextMonthInSameCol=function(t){return this._dateAdapter.addCalendarMonths(t,4)},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.decorators=[{type:s.m,args:[{selector:"mat-calendar",template:'<div class="mat-calendar-header"><div class="mat-calendar-controls"><button mat-button class="mat-calendar-period-button" (click)="_currentPeriodClicked()" [attr.aria-label]="_periodButtonLabel">{{_periodButtonText}}<div class="mat-calendar-arrow" [class.mat-calendar-invert]="!_monthView"></div></button><div class="mat-calendar-spacer"></div><button mat-icon-button class="mat-calendar-previous-button" [disabled]="!_previousEnabled()" (click)="_previousClicked()" [attr.aria-label]="_prevButtonLabel"></button> <button mat-icon-button class="mat-calendar-next-button" [disabled]="!_nextEnabled()" (click)="_nextClicked()" [attr.aria-label]="_nextButtonLabel"></button></div></div><div class="mat-calendar-content" (keydown)="_handleCalendarBodyKeydown($event)" [ngSwitch]="_monthView" cdkMonitorSubtreeFocus><mat-month-view *ngSwitchCase="true" [activeDate]="_activeDate" [selected]="selected" [dateFilter]="_dateFilterForViews" (selectedChange)="_dateSelected($event)" (_userSelection)="_userSelected()"></mat-month-view><mat-year-view *ngSwitchDefault [activeDate]="_activeDate" [selected]="selected" [dateFilter]="_dateFilterForViews" (selectedChange)="_monthSelected($event)"></mat-year-view></div>',styles:[".mat-calendar{display:block}.mat-calendar-header{padding:8px 8px 0 8px}.mat-calendar-content{padding:0 8px 8px 8px;outline:0}.mat-calendar-controls{display:flex;margin:5% calc(33% / 7 - 16px)}.mat-calendar-spacer{flex:1 1 auto}.mat-calendar-period-button{min-width:0}.mat-calendar-arrow{display:inline-block;width:0;height:0;border-left:5px solid transparent;border-right:5px solid transparent;border-top-width:5px;border-top-style:solid;margin:0 0 0 5px;vertical-align:middle}.mat-calendar-arrow.mat-calendar-invert{transform:rotate(180deg)}[dir=rtl] .mat-calendar-arrow{margin:0 5px 0 0}.mat-calendar-next-button,.mat-calendar-previous-button{position:relative}.mat-calendar-next-button::after,.mat-calendar-previous-button::after{top:0;left:0;right:0;bottom:0;position:absolute;content:'';margin:15.5px;border:0 solid currentColor;border-top-width:2px}[dir=rtl] .mat-calendar-next-button,[dir=rtl] .mat-calendar-previous-button{transform:rotate(180deg)}.mat-calendar-previous-button::after{border-left-width:2px;transform:translateX(2px) rotate(-45deg)}.mat-calendar-next-button::after{border-right-width:2px;transform:translateX(-2px) rotate(45deg)}.mat-calendar-table{border-spacing:0;border-collapse:collapse;width:100%}.mat-calendar-table-header th{text-align:center;padding:0 0 8px 0}.mat-calendar-table-header-divider{position:relative;height:1px}.mat-calendar-table-header-divider::after{content:'';position:absolute;top:0;left:-8px;right:-8px;height:1px}"],host:{class:"mat-calendar"},exportAs:"matCalendar",encapsulation:s._10.None,preserveWhitespaces:!1,changeDetection:s.i.OnPush}]}],t.ctorParameters=function(){return[{type:s.t},{type:O},{type:s.L},{type:p.c,decorators:[{type:s.M}]},{type:void 0,decorators:[{type:s.M},{type:s.y,args:[p.f]}]},{type:s.j}]},t.propDecorators={startAt:[{type:s.C}],startView:[{type:s.C}],selected:[{type:s.C}],minDate:[{type:s.C}],maxDate:[{type:s.C}],dateFilter:[{type:s.C}],selectedChange:[{type:s.N}],_userSelection:[{type:s.N}],monthView:[{type:s._7,args:[E]}],yearView:[{type:s._7,args:[S]}]},t}(),M=0,A=new s.A("mat-datepicker-scroll-strategy");var T={provide:A,deps:[i.b],useFactory:function(t){return function(){return t.scrollStrategies.reposition()}}},D=function(){function t(){}return t.prototype.ngAfterContentInit=function(){this._calendar._focusActiveCell()},t.prototype._handleKeydown=function(t){t.keyCode===l.g&&(this.datepicker.close(),t.preventDefault(),t.stopPropagation())},t.decorators=[{type:s.m,args:[{selector:"mat-datepicker-content",template:'<mat-calendar cdkTrapFocus [id]="datepicker.id" [ngClass]="datepicker.panelClass" [startAt]="datepicker.startAt" [startView]="datepicker.startView" [minDate]="datepicker._minDate" [maxDate]="datepicker._maxDate" [dateFilter]="datepicker._dateFilter" [selected]="datepicker._selected" (selectedChange)="datepicker._select($event)" (_userSelection)="datepicker.close()"></mat-calendar>',styles:[".mat-datepicker-content{box-shadow:0 5px 5px -3px rgba(0,0,0,.2),0 8px 10px 1px rgba(0,0,0,.14),0 3px 14px 2px rgba(0,0,0,.12);display:block}.mat-datepicker-content .mat-calendar{width:296px;height:354px}.mat-datepicker-content-touch{box-shadow:0 0 0 0 rgba(0,0,0,.2),0 0 0 0 rgba(0,0,0,.14),0 0 0 0 rgba(0,0,0,.12);display:block;max-height:80vh;overflow:auto;margin:-24px}.mat-datepicker-content-touch .mat-calendar{min-width:250px;min-height:312px;max-width:750px;max-height:788px}@media all and (orientation:landscape){.mat-datepicker-content-touch .mat-calendar{width:64vh;height:80vh}}@media all and (orientation:portrait){.mat-datepicker-content-touch .mat-calendar{width:80vw;height:100vw}}"],host:{class:"mat-datepicker-content","[class.mat-datepicker-content-touch]":"datepicker.touchUi","(keydown)":"_handleKeydown($event)"},exportAs:"matDatepickerContent",encapsulation:s._10.None,preserveWhitespaces:!1,changeDetection:s.i.OnPush}]}],t.ctorParameters=function(){return[]},t.propDecorators={_calendar:[{type:s._7,args:[P]}]},t}(),I=function(){function t(t,e,n,r,i,o,a,c){if(this._dialog=t,this._overlay=e,this._ngZone=n,this._viewContainerRef=r,this._scrollStrategy=i,this._dateAdapter=o,this._dir=a,this._document=c,this.startView="month",this._touchUi=!1,this.selectedChanged=new s.w,this.openedStream=new s.w,this.closedStream=new s.w,this._opened=!1,this.id="mat-datepicker-"+M++,this._validSelected=null,this._focusedElementBeforeOpen=null,this._inputSubscription=v.a.EMPTY,this._disabledChange=new h.a,!this._dateAdapter)throw w("DateAdapter")}return Object.defineProperty(t.prototype,"startAt",{get:function(){return this._startAt||(this._datepickerInput?this._datepickerInput.value:null)},set:function(t){this._startAt=this._getValidDateOrNull(this._dateAdapter.deserialize(t))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"touchUi",{get:function(){return this._touchUi},set:function(t){this._touchUi=Object(m.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return void 0===this._disabled&&this._datepickerInput?this._datepickerInput.disabled:!!this._disabled},set:function(t){var e=Object(m.b)(t);e!==this._disabled&&(this._disabled=e,this._disabledChange.next(e))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"opened",{get:function(){return this._opened},set:function(t){t?this.open():this.close()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_selected",{get:function(){return this._validSelected},set:function(t){this._validSelected=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_minDate",{get:function(){return this._datepickerInput&&this._datepickerInput.min},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_maxDate",{get:function(){return this._datepickerInput&&this._datepickerInput.max},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_dateFilter",{get:function(){return this._datepickerInput&&this._datepickerInput._dateFilter},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this.close(),this._inputSubscription.unsubscribe(),this._disabledChange.complete(),this._popupRef&&this._popupRef.dispose()},t.prototype._select=function(t){var e=this._selected;this._selected=t,this._dateAdapter.sameDate(e,this._selected)||this.selectedChanged.emit(t)},t.prototype._registerInput=function(t){var e=this;if(this._datepickerInput)throw Error("A MatDatepicker can only be associated with a single input.");this._datepickerInput=t,this._inputSubscription=this._datepickerInput._valueChange.subscribe(function(t){return e._selected=t})},t.prototype.open=function(){if(!this._opened&&!this.disabled){if(!this._datepickerInput)throw Error("Attempted to open an MatDatepicker with no associated input.");this._document&&(this._focusedElementBeforeOpen=this._document.activeElement),this.touchUi?this._openAsDialog():this._openAsPopup(),this._opened=!0,this.openedStream.emit()}},t.prototype.close=function(){this._opened&&(this._popupRef&&this._popupRef.hasAttached()&&this._popupRef.detach(),this._dialogRef&&(this._dialogRef.close(),this._dialogRef=null),this._calendarPortal&&this._calendarPortal.isAttached&&this._calendarPortal.detach(),this._focusedElementBeforeOpen&&"function"==typeof this._focusedElementBeforeOpen.focus&&(this._focusedElementBeforeOpen.focus(),this._focusedElementBeforeOpen=null),this._opened=!1,this.closedStream.emit())},t.prototype._openAsDialog=function(){var t=this;this._dialogRef=this._dialog.open(D,{direction:this._dir?this._dir.value:"ltr",viewContainerRef:this._viewContainerRef,panelClass:"mat-datepicker-dialog"}),this._dialogRef.afterClosed().subscribe(function(){return t.close()}),this._dialogRef.componentInstance.datepicker=this},t.prototype._openAsPopup=function(){var t=this;(this._calendarPortal||(this._calendarPortal=new g.d(D,this._viewContainerRef)),this._popupRef||this._createPopup(),this._popupRef.hasAttached())||(this._popupRef.attach(this._calendarPortal).instance.datepicker=this,this._ngZone.onStable.asObservable().pipe(Object(d.a)(1)).subscribe(function(){t._popupRef.updatePosition()}));this._popupRef.backdropClick().subscribe(function(){return t.close()})},t.prototype._createPopup=function(){var t=new i.c({positionStrategy:this._createPopupPositionStrategy(),hasBackdrop:!0,backdropClass:"mat-overlay-transparent-backdrop",direction:this._dir?this._dir.value:"ltr",scrollStrategy:this._scrollStrategy(),panelClass:"mat-datepicker-popup"});this._popupRef=this._overlay.create(t)},t.prototype._createPopupPositionStrategy=function(){var t=this._datepickerInput._getPopupFallbackOffset();return this._overlay.position().connectedTo(this._datepickerInput.getPopupConnectionElementRef(),{originX:"start",originY:"bottom"},{overlayX:"start",overlayY:"top"}).withFallbackPosition({originX:"start",originY:"top"},{overlayX:"start",overlayY:"bottom"},void 0,t).withFallbackPosition({originX:"end",originY:"bottom"},{overlayX:"end",overlayY:"top"}).withFallbackPosition({originX:"end",originY:"top"},{overlayX:"end",overlayY:"bottom"},void 0,t)},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.decorators=[{type:s.m,args:[{selector:"mat-datepicker",template:"",exportAs:"matDatepicker",changeDetection:s.i.OnPush,encapsulation:s._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[{type:c.a},{type:i.b},{type:s.L},{type:s._9},{type:void 0,decorators:[{type:s.y,args:[A]}]},{type:p.c,decorators:[{type:s.M}]},{type:f.b,decorators:[{type:s.M}]},{type:void 0,decorators:[{type:s.M},{type:s.y,args:[o.b]}]}]},t.propDecorators={startAt:[{type:s.C}],startView:[{type:s.C}],touchUi:[{type:s.C}],disabled:[{type:s.C}],selectedChanged:[{type:s.N}],panelClass:[{type:s.C}],openedStream:[{type:s.N,args:["opened"]}],closedStream:[{type:s.N,args:["closed"]}],opened:[{type:s.C}]},t}(),R={provide:y.d,useExisting:Object(s._14)(function(){return F}),multi:!0},N={provide:y.c,useExisting:Object(s._14)(function(){return F}),multi:!0},L=function(){return function(t,e){this.target=t,this.targetElement=e,this.value=this.target.value}}(),F=function(){function t(t,e,n,r){var i=this;if(this._elementRef=t,this._dateAdapter=e,this._dateFormats=n,this._formField=r,this.dateChange=new s.w,this.dateInput=new s.w,this._valueChange=new s.w,this._disabledChange=new s.w,this._onTouched=function(){},this._cvaOnChange=function(){},this._validatorOnChange=function(){},this._datepickerSubscription=v.a.EMPTY,this._localeSubscription=v.a.EMPTY,this._parseValidator=function(){return i._lastValueValid?null:{matDatepickerParse:{text:i._elementRef.nativeElement.value}}},this._minValidator=function(t){var e=i._getValidDateOrNull(i._dateAdapter.deserialize(t.value));return!i.min||!e||i._dateAdapter.compareDate(i.min,e)<=0?null:{matDatepickerMin:{min:i.min,actual:e}}},this._maxValidator=function(t){var e=i._getValidDateOrNull(i._dateAdapter.deserialize(t.value));return!i.max||!e||i._dateAdapter.compareDate(i.max,e)>=0?null:{matDatepickerMax:{max:i.max,actual:e}}},this._filterValidator=function(t){var e=i._getValidDateOrNull(i._dateAdapter.deserialize(t.value));return i._dateFilter&&e&&!i._dateFilter(e)?{matDatepickerFilter:!0}:null},this._validator=y.g.compose([this._parseValidator,this._minValidator,this._maxValidator,this._filterValidator]),this._lastValueValid=!1,!this._dateAdapter)throw w("DateAdapter");if(!this._dateFormats)throw w("MAT_DATE_FORMATS");this._localeSubscription=e.localeChanges.subscribe(function(){i.value=i.value})}return Object.defineProperty(t.prototype,"matDatepicker",{set:function(t){this.registerDatepicker(t)},enumerable:!0,configurable:!0}),t.prototype.registerDatepicker=function(t){t&&(this._datepicker=t,this._datepicker._registerInput(this))},Object.defineProperty(t.prototype,"matDatepickerFilter",{set:function(t){this._dateFilter=t,this._validatorOnChange()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(t){t=this._dateAdapter.deserialize(t),this._lastValueValid=!t||this._dateAdapter.isValid(t),t=this._getValidDateOrNull(t);var e=this.value;this._value=t,this._elementRef.nativeElement.value=t?this._dateAdapter.format(t,this._dateFormats.display.dateInput):"",this._dateAdapter.sameDate(e,t)||this._valueChange.emit(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"min",{get:function(){return this._min},set:function(t){this._min=this._getValidDateOrNull(this._dateAdapter.deserialize(t)),this._validatorOnChange()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"max",{get:function(){return this._max},set:function(t){this._max=this._getValidDateOrNull(this._dateAdapter.deserialize(t)),this._validatorOnChange()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return!!this._disabled},set:function(t){var e=Object(m.b)(t);this._disabled!==e&&(this._disabled=e,this._disabledChange.emit(e))},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){var t=this;this._datepicker&&(this._datepickerSubscription=this._datepicker.selectedChanged.subscribe(function(e){t.value=e,t._cvaOnChange(e),t._onTouched(),t.dateInput.emit(new L(t,t._elementRef.nativeElement)),t.dateChange.emit(new L(t,t._elementRef.nativeElement))}))},t.prototype.ngOnDestroy=function(){this._datepickerSubscription.unsubscribe(),this._localeSubscription.unsubscribe(),this._valueChange.complete(),this._disabledChange.complete()},t.prototype.registerOnValidatorChange=function(t){this._validatorOnChange=t},t.prototype.validate=function(t){return this._validator?this._validator(t):null},t.prototype.getPopupConnectionElementRef=function(){return this._formField?this._formField.underlineRef:this._elementRef},t.prototype._getPopupFallbackOffset=function(){return this._formField?-this._formField._inputContainerRef.nativeElement.clientHeight:0},t.prototype.writeValue=function(t){this.value=t},t.prototype.registerOnChange=function(t){this._cvaOnChange=t},t.prototype.registerOnTouched=function(t){this._onTouched=t},t.prototype.setDisabledState=function(t){this.disabled=t},t.prototype._onKeydown=function(t){t.altKey&&t.keyCode===l.d&&(this._datepicker.open(),t.preventDefault())},t.prototype._onInput=function(t){var e=this._dateAdapter.parse(t,this._dateFormats.parse.dateInput);this._lastValueValid=!e||this._dateAdapter.isValid(e),e=this._getValidDateOrNull(e),this._value=e,this._cvaOnChange(e),this._valueChange.emit(e),this.dateInput.emit(new L(this,this._elementRef.nativeElement))},t.prototype._onChange=function(){this.dateChange.emit(new L(this,this._elementRef.nativeElement))},t.prototype._getValidDateOrNull=function(t){return this._dateAdapter.isDateInstance(t)&&this._dateAdapter.isValid(t)?t:null},t.decorators=[{type:s.s,args:[{selector:"input[matDatepicker]",providers:[R,N,{provide:_.a,useExisting:t}],host:{"[attr.aria-haspopup]":"true","[attr.aria-owns]":"(_datepicker?.opened && _datepicker.id) || null","[attr.min]":"min ? _dateAdapter.toIso8601(min) : null","[attr.max]":"max ? _dateAdapter.toIso8601(max) : null","[disabled]":"disabled","(input)":"_onInput($event.target.value)","(change)":"_onChange()","(blur)":"_onTouched()","(keydown)":"_onKeydown($event)"},exportAs:"matDatepickerInput"}]}],t.ctorParameters=function(){return[{type:s.t},{type:p.c,decorators:[{type:s.M}]},{type:void 0,decorators:[{type:s.M},{type:s.y,args:[p.f]}]},{type:b.a,decorators:[{type:s.M}]}]},t.propDecorators={matDatepicker:[{type:s.C}],matDatepickerFilter:[{type:s.C}],value:[{type:s.C}],min:[{type:s.C}],max:[{type:s.C}],disabled:[{type:s.C}],dateChange:[{type:s.N}],dateInput:[{type:s.N}]},t}(),V=function(){function t(t,e){this._intl=t,this._changeDetectorRef=e,this._stateChanges=v.a.EMPTY}return Object.defineProperty(t.prototype,"disabled",{get:function(){return void 0===this._disabled?this.datepicker.disabled:!!this._disabled},set:function(t){this._disabled=Object(m.b)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnChanges=function(t){t.datepicker&&this._watchStateChanges()},t.prototype.ngOnDestroy=function(){this._stateChanges.unsubscribe()},t.prototype.ngAfterContentInit=function(){this._watchStateChanges()},t.prototype._open=function(t){this.datepicker&&!this.disabled&&(this.datepicker.open(),t.stopPropagation())},t.prototype._watchStateChanges=function(){var t=this,e=this.datepicker?this.datepicker._disabledChange:Object(x.a)(),n=this.datepicker&&this.datepicker._datepickerInput?this.datepicker._datepickerInput._disabledChange:Object(x.a)();this._stateChanges.unsubscribe(),this._stateChanges=Object(j.a)(this._intl.changes,e,n).subscribe(function(){return t._changeDetectorRef.markForCheck()})},t.decorators=[{type:s.m,args:[{selector:"mat-datepicker-toggle",template:'<button mat-icon-button type="button" [attr.aria-label]="_intl.openCalendarLabel" [disabled]="disabled" (click)="_open($event)"><mat-icon><svg viewBox="0 0 24 24" width="100%" height="100%" fill="currentColor" style="vertical-align: top" focusable="false"><path d="M0 0h24v24H0z" fill="none"/><path d="M19 3h-1V1h-2v2H8V1H6v2H5c-1.11 0-1.99.9-1.99 2L3 19c0 1.1.89 2 2 2h14c1.1 0 2-.9 2-2V5c0-1.1-.9-2-2-2zm0 16H5V8h14v11zM7 10h5v5H7z"/></svg></mat-icon></button>',host:{class:"mat-datepicker-toggle"},exportAs:"matDatepickerToggle",encapsulation:s._10.None,preserveWhitespaces:!1,changeDetection:s.i.OnPush}]}],t.ctorParameters=function(){return[{type:O},{type:s.j}]},t.propDecorators={datepicker:[{type:s.C,args:["for"]}],disabled:[{type:s.C}]},t}();!function(){function t(){}t.decorators=[{type:s.H,args:[{imports:[o.a,a.a,c.b,u.a,i.d,r.a],exports:[P,k,I,D,F,V,E,S],declarations:[P,k,I,D,F,V,E,S],providers:[O,T],entryComponents:[D]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/dialog.es5.js":function(t,e,n){"use strict";n.d(e,"b",function(){return D}),n.d(e,"a",function(){return E});var r=n("../../../core/esm5/core.js"),i=n("../../../common/esm5/common.js"),o=n("../../../cdk/esm5/overlay.es5.js"),s=n("../../../cdk/esm5/portal.es5.js"),a=n("../../../cdk/esm5/a11y.es5.js"),c=n("../../../material/esm5/core.es5.js"),u=n("../../../../tslib/tslib.es6.js"),l=n("../../../cdk/esm5/bidi.es5.js"),p=n("../../../cdk/esm5/keycodes.es5.js"),d=n("../../../../rxjs/_esm5/observable/defer.js"),h=n("../../../../rxjs/_esm5/observable/of.js"),f=n("../../../../rxjs/_esm5/operators/filter.js"),m=n("../../../../rxjs/_esm5/operators/startWith.js"),g=n("../../../../rxjs/_esm5/Subject.js"),v=n("../../../animations/esm5/animations.js"),y=n("../../../../rxjs/_esm5/operators/take.js"),b=function(){return function(){this.role="dialog",this.panelClass="",this.hasBackdrop=!0,this.backdropClass="",this.disableClose=!1,this.width="",this.height="",this.maxWidth="80vw",this.data=null,this.direction="ltr",this.ariaDescribedBy=null,this.ariaLabel=null,this.autoFocus=!0}}();function _(){throw Error("Attempting to attach dialog content after content is already attached")}var j=function(t){function e(e,n,i,o){var s=t.call(this)||this;return s._elementRef=e,s._focusTrapFactory=n,s._changeDetectorRef=i,s._document=o,s._elementFocusedBeforeDialogWasOpened=null,s._state="enter",s._animationStateChanged=new r.w,s._ariaLabelledBy=null,s}return Object(u.b)(e,t),e.prototype.attachComponentPortal=function(t){return this._portalOutlet.hasAttached()&&_(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachComponentPortal(t)},e.prototype.attachTemplatePortal=function(t){return this._portalOutlet.hasAttached()&&_(),this._savePreviouslyFocusedElement(),this._portalOutlet.attachTemplatePortal(t)},e.prototype._trapFocus=function(){this._focusTrap||(this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement)),this._config.autoFocus&&this._focusTrap.focusInitialElementWhenReady()},e.prototype._restoreFocus=function(){var t=this._elementFocusedBeforeDialogWasOpened;t&&"function"==typeof t.focus&&t.focus(),this._focusTrap&&this._focusTrap.destroy()},e.prototype._savePreviouslyFocusedElement=function(){var t=this;this._document&&(this._elementFocusedBeforeDialogWasOpened=this._document.activeElement,Promise.resolve().then(function(){return t._elementRef.nativeElement.focus()}))},e.prototype._onAnimationDone=function(t){"enter"===t.toState?this._trapFocus():"exit"===t.toState&&this._restoreFocus(),this._animationStateChanged.emit(t)},e.prototype._onAnimationStart=function(t){this._animationStateChanged.emit(t)},e.prototype._startExitAnimation=function(){this._state="exit",this._changeDetectorRef.markForCheck()},e.decorators=[{type:r.m,args:[{selector:"mat-dialog-container",template:"<ng-template cdkPortalOutlet></ng-template>",styles:[".mat-dialog-container{box-shadow:0 11px 15px -7px rgba(0,0,0,.2),0 24px 38px 3px rgba(0,0,0,.14),0 9px 46px 8px rgba(0,0,0,.12);display:block;padding:24px;border-radius:2px;box-sizing:border-box;overflow:auto;outline:0;width:100%;height:100%}@media screen and (-ms-high-contrast:active){.mat-dialog-container{outline:solid 1px}}.mat-dialog-content{display:block;margin:0 -24px;padding:0 24px;max-height:65vh;overflow:auto;-webkit-overflow-scrolling:touch;-webkit-backface-visibility:hidden;backface-visibility:hidden}.mat-dialog-title{margin:0 0 20px;display:block}.mat-dialog-actions{padding:12px 0;display:flex;flex-wrap:wrap}.mat-dialog-actions:last-child{margin-bottom:-24px}.mat-dialog-actions[align=end]{justify-content:flex-end}.mat-dialog-actions[align=center]{justify-content:center}.mat-dialog-actions .mat-button+.mat-button,.mat-dialog-actions .mat-button+.mat-raised-button,.mat-dialog-actions .mat-raised-button+.mat-button,.mat-dialog-actions .mat-raised-button+.mat-raised-button{margin-left:8px}[dir=rtl] .mat-dialog-actions .mat-button+.mat-button,[dir=rtl] .mat-dialog-actions .mat-button+.mat-raised-button,[dir=rtl] .mat-dialog-actions .mat-raised-button+.mat-button,[dir=rtl] .mat-dialog-actions .mat-raised-button+.mat-raised-button{margin-left:0;margin-right:8px}"],encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.Default,animations:[Object(v.k)("slideDialog",[Object(v.h)("enter",Object(v.i)({transform:"none",opacity:1})),Object(v.h)("void",Object(v.i)({transform:"translate3d(0, 25%, 0) scale(0.9)",opacity:0})),Object(v.h)("exit",Object(v.i)({transform:"translate3d(0, 25%, 0)",opacity:0})),Object(v.j)("* => *",Object(v.e)("400ms cubic-bezier(0.25, 0.8, 0.25, 1)"))])],host:{class:"mat-dialog-container",tabindex:"-1","[attr.role]":"_config?.role","[attr.aria-labelledby]":"_config?.ariaLabel ? null : _ariaLabelledBy","[attr.aria-label]":"_config?.ariaLabel","[attr.aria-describedby]":"_config?.ariaDescribedBy || null","[@slideDialog]":"_state","(@slideDialog.start)":"_onAnimationStart($event)","(@slideDialog.done)":"_onAnimationDone($event)"}}]}],e.ctorParameters=function(){return[{type:r.t},{type:a.g},{type:r.j},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[i.b]}]}]},e.propDecorators={_portalOutlet:[{type:r._7,args:[s.c]}]},e}(s.a),x=0,w=function(){function t(t,e,n){void 0===n&&(n="mat-dialog-"+x++);var r=this;this._overlayRef=t,this._containerInstance=e,this.id=n,this.disableClose=this._containerInstance._config.disableClose,this._afterOpen=new g.a,this._afterClosed=new g.a,this._beforeClose=new g.a,e._animationStateChanged.pipe(Object(f.a)(function(t){return"done"===t.phaseName&&"enter"===t.toState}),Object(y.a)(1)).subscribe(function(){r._afterOpen.next(),r._afterOpen.complete()}),e._animationStateChanged.pipe(Object(f.a)(function(t){return"done"===t.phaseName&&"exit"===t.toState}),Object(y.a)(1)).subscribe(function(){r._overlayRef.dispose(),r._afterClosed.next(r._result),r._afterClosed.complete(),r.componentInstance=null})}return t.prototype.close=function(t){var e=this;this._result=t,this._containerInstance._animationStateChanged.pipe(Object(f.a)(function(t){return"start"===t.phaseName}),Object(y.a)(1)).subscribe(function(){e._beforeClose.next(t),e._beforeClose.complete(),e._overlayRef.detachBackdrop()}),this._containerInstance._startExitAnimation()},t.prototype.afterOpen=function(){return this._afterOpen.asObservable()},t.prototype.afterClosed=function(){return this._afterClosed.asObservable()},t.prototype.beforeClose=function(){return this._beforeClose.asObservable()},t.prototype.backdropClick=function(){return this._overlayRef.backdropClick()},t.prototype.keydownEvents=function(){return this._overlayRef.keydownEvents()},t.prototype.updatePosition=function(t){var e=this._getPositionStrategy();return t&&(t.left||t.right)?t.left?e.left(t.left):e.right(t.right):e.centerHorizontally(),t&&(t.top||t.bottom)?t.top?e.top(t.top):e.bottom(t.bottom):e.centerVertically(),this._overlayRef.updatePosition(),this},t.prototype.updateSize=function(t,e){return void 0===t&&(t="auto"),void 0===e&&(e="auto"),this._getPositionStrategy().width(t).height(e),this._overlayRef.updatePosition(),this},t.prototype._getPositionStrategy=function(){return this._overlayRef.getConfig().positionStrategy},t}(),O=new r.A("MatDialogData"),C=new r.A("mat-dialog-scroll-strategy");var k={provide:C,deps:[o.b],useFactory:function(t){return function(){return t.scrollStrategies.block()}}},E=function(){function t(t,e,n,r,i){var o=this;this._overlay=t,this._injector=e,this._scrollStrategy=r,this._parentDialog=i,this._openDialogsAtThisLevel=[],this._afterAllClosedAtThisLevel=new g.a,this._afterOpenAtThisLevel=new g.a,this.afterAllClosed=Object(d.a)(function(){return o.openDialogs.length?o._afterAllClosed:o._afterAllClosed.pipe(Object(m.a)(void 0))}),!i&&n&&n.subscribe(function(){return o.closeAll()})}return Object.defineProperty(t.prototype,"openDialogs",{get:function(){return this._parentDialog?this._parentDialog.openDialogs:this._openDialogsAtThisLevel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"afterOpen",{get:function(){return this._parentDialog?this._parentDialog.afterOpen:this._afterOpenAtThisLevel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_afterAllClosed",{get:function(){var t=this._parentDialog;return t?t._afterAllClosed:this._afterAllClosedAtThisLevel},enumerable:!0,configurable:!0}),t.prototype.open=function(t,e){var n,r=this;if(n=e,(e=Object(u.a)({},new b,n)).id&&this.getDialogById(e.id))throw Error('Dialog with id "'+e.id+'" exists already. The dialog id must be unique.');var i=this._createOverlay(e),o=this._attachDialogContainer(i,e),s=this._attachDialogContent(t,o,i,e);return this.openDialogs.push(s),s.afterClosed().subscribe(function(){return r._removeOpenDialog(s)}),this.afterOpen.next(s),s},t.prototype.closeAll=function(){for(var t=this.openDialogs.length;t--;)this.openDialogs[t].close()},t.prototype.getDialogById=function(t){return this.openDialogs.find(function(e){return e.id===t})},t.prototype._createOverlay=function(t){var e=this._getOverlayConfig(t);return this._overlay.create(e)},t.prototype._getOverlayConfig=function(t){var e=new o.c({positionStrategy:this._overlay.position().global(),scrollStrategy:this._scrollStrategy(),panelClass:t.panelClass,hasBackdrop:t.hasBackdrop,direction:t.direction,minWidth:t.minWidth,minHeight:t.minHeight,maxWidth:t.maxWidth,maxHeight:t.maxHeight});return t.backdropClass&&(e.backdropClass=t.backdropClass),e},t.prototype._attachDialogContainer=function(t,e){var n=new s.d(j,e.viewContainerRef),r=t.attach(n);return r.instance._config=e,r.instance},t.prototype._attachDialogContent=function(t,e,n,i){var o=new w(n,e,i.id);if(i.hasBackdrop&&n.backdropClick().subscribe(function(){o.disableClose||o.close()}),n.keydownEvents().pipe(Object(f.a)(function(t){return t.keyCode===p.g&&!o.disableClose})).subscribe(function(){return o.close()}),t instanceof r._4)e.attachTemplatePortal(new s.h(t,null,{$implicit:i.data,dialogRef:o}));else{var a=this._createInjector(i,o,e),c=e.attachComponentPortal(new s.d(t,void 0,a));o.componentInstance=c.instance}return o.updateSize(i.width,i.height).updatePosition(i.position),o},t.prototype._createInjector=function(t,e,n){var r=t&&t.viewContainerRef&&t.viewContainerRef.injector,i=new WeakMap;return i.set(w,e),i.set(j,n),i.set(O,t.data),i.set(l.b,{value:t.direction,change:Object(h.a)()}),new s.f(r||this._injector,i)},t.prototype._removeOpenDialog=function(t){var e=this.openDialogs.indexOf(t);e>-1&&(this.openDialogs.splice(e,1),this.openDialogs.length||this._afterAllClosed.next())},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:o.b},{type:r.B},{type:i.c,decorators:[{type:r.M}]},{type:void 0,decorators:[{type:r.y,args:[C]}]},{type:t,decorators:[{type:r.M},{type:r._1}]}]},t}();var S=0,P=function(){function t(t){this.dialogRef=t,this.ariaLabel="Close dialog"}return t.prototype.ngOnChanges=function(t){var e=t._matDialogClose||t._matDialogCloseResult;e&&(this.dialogResult=e.currentValue)},t.decorators=[{type:r.s,args:[{selector:"button[mat-dialog-close], button[matDialogClose]",exportAs:"matDialogClose",host:{"(click)":"dialogRef.close(dialogResult)","[attr.aria-label]":"ariaLabel",type:"button"}}]}],t.ctorParameters=function(){return[{type:w}]},t.propDecorators={ariaLabel:[{type:r.C,args:["aria-label"]}],dialogResult:[{type:r.C,args:["mat-dialog-close"]}],_matDialogClose:[{type:r.C,args:["matDialogClose"]}]},t}(),M=function(){function t(t){this._container=t,this.id="mat-dialog-title-"+S++}return t.prototype.ngOnInit=function(){var t=this;this._container&&!this._container._ariaLabelledBy&&Promise.resolve().then(function(){return t._container._ariaLabelledBy=t.id})},t.decorators=[{type:r.s,args:[{selector:"[mat-dialog-title], [matDialogTitle]",exportAs:"matDialogTitle",host:{class:"mat-dialog-title","[id]":"id"}}]}],t.ctorParameters=function(){return[{type:j,decorators:[{type:r.M}]}]},t.propDecorators={id:[{type:r.C}]},t}(),A=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-dialog-content], mat-dialog-content, [matDialogContent]",host:{class:"mat-dialog-content"}}]}],t.ctorParameters=function(){return[]},t}(),T=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-dialog-actions], mat-dialog-actions, [matDialogActions]",host:{class:"mat-dialog-actions"}}]}],t.ctorParameters=function(){return[]},t}(),D=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.a,o.d,s.g,a.a,c.j],exports:[j,P,M,A,T,c.j],declarations:[j,P,M,T,A],providers:[E,k],entryComponents:[j]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/expansion.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return k});var r=n("../../../common/esm5/common.js"),i=n("../../../core/esm5/core.js"),o=n("../../../cdk/esm5/collections.es5.js"),s=n("../../../cdk/esm5/accordion.es5.js"),a=n("../../../cdk/esm5/a11y.es5.js"),c=n("../../../../tslib/tslib.es6.js"),u=n("../../../cdk/esm5/coercion.es5.js"),l=n("../../../animations/esm5/animations.js"),p=n("../../../material/esm5/core.es5.js"),d=n("../../../../rxjs/_esm5/Subject.js"),h=n("../../../cdk/esm5/keycodes.es5.js"),f=n("../../../../rxjs/_esm5/operators/filter.js"),m=n("../../../../rxjs/_esm5/observable/merge.js"),g=n("../../../../rxjs/_esm5/Subscription.js"),v=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._hideToggle=!1,e.displayMode="default",e}return Object(c.b)(e,t),Object.defineProperty(e.prototype,"hideToggle",{get:function(){return this._hideToggle},set:function(t){this._hideToggle=Object(u.b)(t)},enumerable:!0,configurable:!0}),e.decorators=[{type:i.s,args:[{selector:"mat-accordion",exportAs:"matAccordion",host:{class:"mat-accordion"}}]}],e.ctorParameters=function(){return[]},e.propDecorators={hideToggle:[{type:i.C}],displayMode:[{type:i.C}]},e}(s.a),y="225ms cubic-bezier(0.4,0.0,0.2,1)",b=function(t){function e(e,n,r){return t.call(this,e,n,r)||this}return Object(c.b)(e,t),e.decorators=[{type:i.m,args:[{template:"",encapsulation:i._10.None,preserveWhitespaces:!1,changeDetection:i.i.OnPush}]}],e.ctorParameters=function(){return[{type:v},{type:i.j},{type:o.c}]},e}(s.b),_=Object(p.w)(b),j=function(t){function e(e,n,r){var i=t.call(this,e,n,r)||this;return i._hideToggle=!1,i._inputChanges=new d.a,i.accordion=e,i}return Object(c.b)(e,t),Object.defineProperty(e.prototype,"hideToggle",{get:function(){return this._hideToggle},set:function(t){this._hideToggle=Object(u.b)(t)},enumerable:!0,configurable:!0}),e.prototype._getHideToggle=function(){return this.accordion?this.accordion.hideToggle:this.hideToggle},e.prototype._hasSpacing=function(){return!!this.accordion&&"default"===(this.expanded?this.accordion.displayMode:this._getExpandedState())},e.prototype._getExpandedState=function(){return this.expanded?"expanded":"collapsed"},e.prototype.ngOnChanges=function(t){this._inputChanges.next(t)},e.prototype.ngOnDestroy=function(){t.prototype.ngOnDestroy.call(this),this._inputChanges.complete()},e.decorators=[{type:i.m,args:[{styles:[".mat-expansion-panel{transition:box-shadow 280ms cubic-bezier(.4,0,.2,1);box-sizing:content-box;display:block;margin:0;transition:margin 225ms cubic-bezier(.4,0,.2,1)}.mat-expansion-panel:not([class*=mat-elevation-z]){box-shadow:0 3px 1px -2px rgba(0,0,0,.2),0 2px 2px 0 rgba(0,0,0,.14),0 1px 5px 0 rgba(0,0,0,.12)}.mat-expanded .mat-expansion-panel-content{overflow:visible}.mat-expansion-panel-content,.mat-expansion-panel-content.ng-animating{overflow:hidden}.mat-expansion-panel-body{padding:0 24px 16px}.mat-expansion-panel-spacing{margin:16px 0}.mat-accordion .mat-expansion-panel-spacing:first-child{margin-top:0}.mat-accordion .mat-expansion-panel-spacing:last-child{margin-bottom:0}.mat-action-row{border-top-style:solid;border-top-width:1px;display:flex;flex-direction:row;justify-content:flex-end;padding:16px 8px 16px 24px}.mat-action-row button.mat-button{margin-left:8px}[dir=rtl] .mat-action-row button.mat-button{margin-left:0;margin-right:8px}"],selector:"mat-expansion-panel",exportAs:"matExpansionPanel",template:'<ng-content select="mat-expansion-panel-header"></ng-content><div [class.mat-expanded]="expanded" class="mat-expansion-panel-content" [@bodyExpansion]="_getExpandedState()" [id]="id"><div class="mat-expansion-panel-body"><ng-content></ng-content></div><ng-content select="mat-action-row"></ng-content></div>',encapsulation:i._10.None,preserveWhitespaces:!1,changeDetection:i.i.OnPush,inputs:["disabled","expanded"],outputs:["opened","closed"],host:{class:"mat-expansion-panel","[class.mat-expanded]":"expanded","[class.mat-expansion-panel-spacing]":"_hasSpacing()"},providers:[{provide:_,useExisting:Object(i._14)(function(){return e})}],animations:[Object(l.k)("bodyExpansion",[Object(l.h)("collapsed",Object(l.i)({height:"0px",visibility:"hidden"})),Object(l.h)("expanded",Object(l.i)({height:"*",visibility:"visible"})),Object(l.j)("expanded <=> collapsed",Object(l.e)(y))])]}]}],e.ctorParameters=function(){return[{type:v,decorators:[{type:i.M},{type:i.x}]},{type:i.j},{type:o.c}]},e.propDecorators={hideToggle:[{type:i.C}]},e}(_),x=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"mat-action-row",host:{class:"mat-action-row"}}]}],t.ctorParameters=function(){return[]},t}(),w=function(){function t(t,e,n,r){var i=this;this.panel=t,this._element=e,this._focusMonitor=n,this._changeDetectorRef=r,this._parentChangeSubscription=g.a.EMPTY,this._parentChangeSubscription=Object(m.a)(t.opened,t.closed,t._inputChanges.pipe(Object(f.a)(function(t){return!(!t.hideToggle&&!t.disabled)}))).subscribe(function(){return i._changeDetectorRef.markForCheck()}),n.monitor(e.nativeElement,!1)}return t.prototype._toggle=function(){this.panel.disabled||this.panel.toggle()},t.prototype._isExpanded=function(){return this.panel.expanded},t.prototype._getExpandedState=function(){return this.panel._getExpandedState()},t.prototype._getPanelId=function(){return this.panel.id},t.prototype._showToggle=function(){return!this.panel.hideToggle&&!this.panel.disabled},t.prototype._keyup=function(t){switch(t.keyCode){case h.n:case h.f:t.preventDefault(),this._toggle();break;default:return}},t.prototype.ngOnDestroy=function(){this._parentChangeSubscription.unsubscribe(),this._focusMonitor.stopMonitoring(this._element.nativeElement)},t.decorators=[{type:i.m,args:[{selector:"mat-expansion-panel-header",styles:[".mat-expansion-panel-header{display:flex;flex-direction:row;align-items:center;padding:0 24px}.mat-expansion-panel-header:focus,.mat-expansion-panel-header:hover{outline:0}.mat-expansion-panel-header.mat-expanded:focus,.mat-expansion-panel-header.mat-expanded:hover{background:inherit}.mat-expansion-panel-header:not([aria-disabled=true]){cursor:pointer}.mat-content{display:flex;flex:1;flex-direction:row;overflow:hidden}.mat-expansion-panel-header-description,.mat-expansion-panel-header-title{display:flex;flex-grow:1;margin-right:16px}[dir=rtl] .mat-expansion-panel-header-description,[dir=rtl] .mat-expansion-panel-header-title{margin-right:0;margin-left:16px}.mat-expansion-panel-header-description{flex-grow:2}.mat-expansion-indicator::after{border-style:solid;border-width:0 2px 2px 0;content:'';display:inline-block;padding:3px;transform:rotate(45deg);vertical-align:middle}"],template:'<span class="mat-content"><ng-content select="mat-panel-title"></ng-content><ng-content select="mat-panel-description"></ng-content><ng-content></ng-content></span><span [@indicatorRotate]="_getExpandedState()" *ngIf="_showToggle()" class="mat-expansion-indicator"></span>',encapsulation:i._10.None,preserveWhitespaces:!1,changeDetection:i.i.OnPush,host:{class:"mat-expansion-panel-header",role:"button","[attr.tabindex]":"panel.disabled ? -1 : 0","[attr.aria-controls]":"_getPanelId()","[attr.aria-expanded]":"_isExpanded()","[attr.aria-disabled]":"panel.disabled","[class.mat-expanded]":"_isExpanded()","(click)":"_toggle()","(keyup)":"_keyup($event)","[@expansionHeight]":"{\n        value: _getExpandedState(),\n        params: {\n          collapsedHeight: collapsedHeight,\n          expandedHeight: expandedHeight\n        }\n    }"},animations:[Object(l.k)("indicatorRotate",[Object(l.h)("collapsed",Object(l.i)({transform:"rotate(0deg)"})),Object(l.h)("expanded",Object(l.i)({transform:"rotate(180deg)"})),Object(l.j)("expanded <=> collapsed",Object(l.e)(y))]),Object(l.k)("expansionHeight",[Object(l.h)("collapsed",Object(l.i)({height:"{{collapsedHeight}}"}),{params:{collapsedHeight:"48px"}}),Object(l.h)("expanded",Object(l.i)({height:"{{expandedHeight}}"}),{params:{expandedHeight:"64px"}}),Object(l.j)("expanded <=> collapsed",Object(l.e)(y))])]}]}],t.ctorParameters=function(){return[{type:j,decorators:[{type:i.x}]},{type:i.t},{type:a.f},{type:i.j}]},t.propDecorators={expandedHeight:[{type:i.C}],collapsedHeight:[{type:i.C}]},t}(),O=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"mat-panel-description",host:{class:"mat-expansion-panel-header-description"}}]}],t.ctorParameters=function(){return[]},t}(),C=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"mat-panel-title",host:{class:"mat-expansion-panel-header-title"}}]}],t.ctorParameters=function(){return[]},t}(),k=function(){function t(){}return t.decorators=[{type:i.H,args:[{imports:[r.a,a.a,s.c],exports:[v,j,x,w,C,O],declarations:[b,v,j,x,w,C,O],providers:[o.b]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/form-field.es5.js":function(t,e,n){"use strict";n.d(e,"c",function(){return O}),n.d(e,"a",function(){return w}),n.d(e,"b",function(){return f});var r=n("../../../common/esm5/common.js"),i=n("../../../core/esm5/core.js"),o=n("../../../cdk/esm5/platform.es5.js"),s=n("../../../animations/esm5/animations.js"),a=n("../../../cdk/esm5/coercion.es5.js"),c=n("../../../../rxjs/_esm5/operators/take.js"),u=n("../../../../rxjs/_esm5/operators/startWith.js"),l=n("../../../material/esm5/core.es5.js"),p=n("../../../../rxjs/_esm5/observable/fromEvent.js"),d=0,h=function(){function t(){this.id="mat-error-"+d++}return t.decorators=[{type:i.s,args:[{selector:"mat-error",host:{class:"mat-error",role:"alert","[attr.id]":"id"}}]}],t.ctorParameters=function(){return[]},t.propDecorators={id:[{type:i.C}]},t}(),f=function(){return function(){}}();function m(t){return Error("A hint was already declared for 'align=\""+t+"\"'.")}var g=0,v=function(){function t(){this.align="start",this.id="mat-hint-"+g++}return t.decorators=[{type:i.s,args:[{selector:"mat-hint",host:{class:"mat-hint","[class.mat-right]":'align == "end"',"[attr.id]":"id","[attr.align]":"null"}}]}],t.ctorParameters=function(){return[]},t.propDecorators={align:[{type:i.C}],id:[{type:i.C}]},t}(),y=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"mat-placeholder"}]}],t.ctorParameters=function(){return[]},t}(),b=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"mat-label"}]}],t.ctorParameters=function(){return[]},t}(),_=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"[matPrefix]"}]}],t.ctorParameters=function(){return[]},t}(),j=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"[matSuffix]"}]}],t.ctorParameters=function(){return[]},t}(),x=0,w=function(){function t(t,e,n){this._elementRef=t,this._changeDetectorRef=e,this.color="primary",this._showAlwaysAnimate=!1,this._subscriptAnimationState="",this._hintLabel="",this._hintLabelId="mat-hint-"+x++,this._labelOptions=n||{},this.floatLabel=this._labelOptions.float||"auto"}return Object.defineProperty(t.prototype,"dividerColor",{get:function(){return this.color},set:function(t){this.color=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hideRequiredMarker",{get:function(){return this._hideRequiredMarker},set:function(t){this._hideRequiredMarker=Object(a.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_shouldAlwaysFloat",{get:function(){return"always"===this._floatLabel&&!this._showAlwaysAnimate},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_canLabelFloat",{get:function(){return"never"!==this._floatLabel},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"hintLabel",{get:function(){return this._hintLabel},set:function(t){this._hintLabel=t,this._processHints()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"floatPlaceholder",{get:function(){return this._floatLabel},set:function(t){this.floatLabel=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"floatLabel",{get:function(){return this._floatLabel},set:function(t){t!==this._floatLabel&&(this._floatLabel=t||this._labelOptions.float||"auto",this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){var t=this;this._validateControlChild(),this._control.controlType&&this._elementRef.nativeElement.classList.add("mat-form-field-type-"+this._control.controlType),this._control.stateChanges.pipe(Object(u.a)(null)).subscribe(function(){t._validatePlaceholders(),t._syncDescribedByIds(),t._changeDetectorRef.markForCheck()});var e=this._control.ngControl;e&&e.valueChanges&&e.valueChanges.subscribe(function(){t._changeDetectorRef.markForCheck()}),this._hintChildren.changes.pipe(Object(u.a)(null)).subscribe(function(){t._processHints(),t._changeDetectorRef.markForCheck()}),this._errorChildren.changes.pipe(Object(u.a)(null)).subscribe(function(){t._syncDescribedByIds(),t._changeDetectorRef.markForCheck()})},t.prototype.ngAfterContentChecked=function(){this._validateControlChild()},t.prototype.ngAfterViewInit=function(){this._subscriptAnimationState="enter",this._changeDetectorRef.detectChanges()},t.prototype._shouldForward=function(t){var e=this._control?this._control.ngControl:null;return e&&e[t]},t.prototype._hasPlaceholder=function(){return!(!this._control.placeholder&&!this._placeholderChild)},t.prototype._hasLabel=function(){return!!this._labelChild},t.prototype._shouldLabelFloat=function(){return this._canLabelFloat&&(this._control.shouldLabelFloat||this._control.shouldPlaceholderFloat||this._shouldAlwaysFloat)},t.prototype._hideControlPlaceholder=function(){return!this._hasLabel()||!this._shouldLabelFloat()},t.prototype._hasFloatingLabel=function(){return this._hasLabel()||this._hasPlaceholder()},t.prototype._getDisplayedMessages=function(){return this._errorChildren&&this._errorChildren.length>0&&this._control.errorState?"error":"hint"},t.prototype._animateAndLockLabel=function(){var t=this;this._hasFloatingLabel()&&this._canLabelFloat&&(this._showAlwaysAnimate=!0,this._floatLabel="always",Object(p.a)(this._label.nativeElement,"transitionend").pipe(Object(c.a)(1)).subscribe(function(){t._showAlwaysAnimate=!1}),this._changeDetectorRef.markForCheck())},t.prototype._validatePlaceholders=function(){if(this._control.placeholder&&this._placeholderChild)throw Error("Placeholder attribute and child element were both specified.")},t.prototype._processHints=function(){this._validateHints(),this._syncDescribedByIds()},t.prototype._validateHints=function(){var t,e,n=this;this._hintChildren&&this._hintChildren.forEach(function(r){if("start"==r.align){if(t||n.hintLabel)throw m("start");t=r}else if("end"==r.align){if(e)throw m("end");e=r}})},t.prototype._syncDescribedByIds=function(){if(this._control){var t=[];if("hint"===this._getDisplayedMessages()){var e=this._hintChildren?this._hintChildren.find(function(t){return"start"===t.align}):null,n=this._hintChildren?this._hintChildren.find(function(t){return"end"===t.align}):null;e?t.push(e.id):this._hintLabel&&t.push(this._hintLabelId),n&&t.push(n.id)}else this._errorChildren&&(t=this._errorChildren.map(function(t){return t.id}));this._control.setDescribedByIds(t)}},t.prototype._validateControlChild=function(){if(!this._control)throw Error("mat-form-field must contain a MatFormFieldControl.")},t.decorators=[{type:i.m,args:[{selector:"mat-input-container, mat-form-field",exportAs:"matFormField",template:'<div class="mat-input-wrapper mat-form-field-wrapper"><div class="mat-input-flex mat-form-field-flex" #connectionContainer (click)="_control.onContainerClick && _control.onContainerClick($event)"><div class="mat-input-prefix mat-form-field-prefix" *ngIf="_prefixChildren.length"><ng-content select="[matPrefix]"></ng-content></div><div class="mat-input-infix mat-form-field-infix" #inputContainer><ng-content></ng-content><span class="mat-form-field-label-wrapper mat-input-placeholder-wrapper mat-form-field-placeholder-wrapper"><label class="mat-form-field-label mat-input-placeholder mat-form-field-placeholder" [attr.for]="_control.id" [attr.aria-owns]="_control.id" [class.mat-empty]="_control.empty && !_shouldAlwaysFloat" [class.mat-form-field-empty]="_control.empty && !_shouldAlwaysFloat" [class.mat-accent]="color == \'accent\'" [class.mat-warn]="color == \'warn\'" #label *ngIf="_hasFloatingLabel()" [ngSwitch]="_hasLabel()"><ng-container *ngSwitchCase="false"><ng-content select="mat-placeholder"></ng-content>{{_control.placeholder}}</ng-container><ng-content select="mat-label" *ngSwitchCase="true"></ng-content><span class="mat-placeholder-required mat-form-field-required-marker" aria-hidden="true" *ngIf="!hideRequiredMarker && _control.required">&nbsp;*</span></label></span></div><div class="mat-input-suffix mat-form-field-suffix" *ngIf="_suffixChildren.length"><ng-content select="[matSuffix]"></ng-content></div></div><div class="mat-input-underline mat-form-field-underline" #underline><span class="mat-input-ripple mat-form-field-ripple" [class.mat-accent]="color == \'accent\'" [class.mat-warn]="color == \'warn\'"></span></div><div class="mat-input-subscript-wrapper mat-form-field-subscript-wrapper" [ngSwitch]="_getDisplayedMessages()"><div *ngSwitchCase="\'error\'" [@transitionMessages]="_subscriptAnimationState"><ng-content select="mat-error"></ng-content></div><div class="mat-input-hint-wrapper mat-form-field-hint-wrapper" *ngSwitchCase="\'hint\'" [@transitionMessages]="_subscriptAnimationState"><div *ngIf="hintLabel" [id]="_hintLabelId" class="mat-hint">{{hintLabel}}</div><ng-content select="mat-hint:not([align=\'end\'])"></ng-content><div class="mat-input-hint-spacer mat-form-field-hint-spacer"></div><ng-content select="mat-hint[align=\'end\']"></ng-content></div></div></div>',styles:[".mat-form-field{display:inline-block;position:relative;text-align:left}[dir=rtl] .mat-form-field{text-align:right}.mat-form-field-wrapper{position:relative}.mat-form-field-flex{display:inline-flex;align-items:baseline;width:100%}.mat-form-field-prefix,.mat-form-field-suffix{white-space:nowrap;flex:none}.mat-form-field-prefix .mat-icon,.mat-form-field-suffix .mat-icon{width:1em}.mat-form-field-prefix .mat-icon-button,.mat-form-field-suffix .mat-icon-button{font:inherit;vertical-align:baseline}.mat-form-field-prefix .mat-icon-button .mat-icon,.mat-form-field-suffix .mat-icon-button .mat-icon{font-size:inherit}.mat-form-field-infix{display:block;position:relative;flex:auto;min-width:0;width:180px}.mat-form-field-label-wrapper{position:absolute;left:0;box-sizing:content-box;width:100%;height:100%;overflow:hidden;pointer-events:none}.mat-form-field-label{position:absolute;left:0;font:inherit;pointer-events:none;width:100%;white-space:nowrap;text-overflow:ellipsis;overflow:hidden;transform:perspective(100px);-ms-transform:none;transform-origin:0 0;transition:transform .4s cubic-bezier(.25,.8,.25,1),color .4s cubic-bezier(.25,.8,.25,1),width .4s cubic-bezier(.25,.8,.25,1);display:none}[dir=rtl] .mat-form-field-label{transform-origin:100% 0;left:auto;right:0}.mat-form-field-can-float.mat-form-field-should-float .mat-form-field-label,.mat-form-field-empty.mat-form-field-label{display:block}.mat-form-field-autofill-control:-webkit-autofill+.mat-form-field-label-wrapper .mat-form-field-label{display:none}.mat-form-field-can-float .mat-form-field-autofill-control:-webkit-autofill+.mat-form-field-label-wrapper .mat-form-field-label{display:block;transition:none}.mat-input-server:focus+.mat-form-field-placeholder-wrapper .mat-form-field-placeholder,.mat-input-server[placeholder]:not(:placeholder-shown)+.mat-form-field-placeholder-wrapper .mat-form-field-placeholder{display:none}.mat-form-field-can-float .mat-input-server:focus+.mat-form-field-placeholder-wrapper .mat-form-field-placeholder,.mat-form-field-can-float .mat-input-server[placeholder]:not(:placeholder-shown)+.mat-form-field-placeholder-wrapper .mat-form-field-placeholder{display:block}.mat-form-field-label:not(.mat-form-field-empty){transition:none}.mat-form-field-underline{position:absolute;height:1px;width:100%}.mat-form-field-disabled .mat-form-field-underline{background-position:0;background-color:transparent}.mat-form-field-underline .mat-form-field-ripple{position:absolute;top:0;left:0;width:100%;height:2px;transform-origin:50%;transform:scaleX(.5);visibility:hidden;transition:background-color .3s cubic-bezier(.55,0,.55,.2)}.mat-form-field-invalid:not(.mat-focused) .mat-form-field-underline .mat-form-field-ripple{height:1px}.mat-focused .mat-form-field-underline .mat-form-field-ripple,.mat-form-field-invalid .mat-form-field-underline .mat-form-field-ripple{visibility:visible;transform:scaleX(1);transition:transform 150ms linear,background-color .3s cubic-bezier(.55,0,.55,.2)}.mat-form-field-subscript-wrapper{position:absolute;width:100%;overflow:hidden}.mat-form-field-label-wrapper .mat-icon,.mat-form-field-subscript-wrapper .mat-icon{width:1em;height:1em;font-size:inherit;vertical-align:baseline}.mat-form-field-hint-wrapper{display:flex}.mat-form-field-hint-spacer{flex:1 0 1em}.mat-error{display:block} .mat-input-element{font:inherit;background:0 0;color:currentColor;border:none;outline:0;padding:0;margin:0;width:100%;max-width:100%;vertical-align:bottom}.mat-input-element:-moz-ui-invalid{box-shadow:none}.mat-input-element::-ms-clear,.mat-input-element::-ms-reveal{display:none}.mat-input-element[type=date]::after,.mat-input-element[type=datetime-local]::after,.mat-input-element[type=datetime]::after,.mat-input-element[type=month]::after,.mat-input-element[type=time]::after,.mat-input-element[type=week]::after{content:' ';white-space:pre;width:1px}.mat-input-element::placeholder{transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-input-element::-moz-placeholder{transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-input-element::-webkit-input-placeholder{transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-input-element:-ms-input-placeholder{transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-form-field-hide-placeholder .mat-input-element::placeholder{color:transparent!important;transition:none}.mat-form-field-hide-placeholder .mat-input-element::-moz-placeholder{color:transparent!important;transition:none}.mat-form-field-hide-placeholder .mat-input-element::-webkit-input-placeholder{color:transparent!important;transition:none}.mat-form-field-hide-placeholder .mat-input-element:-ms-input-placeholder{color:transparent!important;transition:none}textarea.mat-input-element{resize:vertical;overflow:auto}"],animations:[Object(s.k)("transitionMessages",[Object(s.h)("enter",Object(s.i)({opacity:1,transform:"translateY(0%)"})),Object(s.j)("void => enter",[Object(s.i)({opacity:0,transform:"translateY(-100%)"}),Object(s.e)("300ms cubic-bezier(0.55, 0, 0.55, 0.2)")])])],host:{class:"mat-input-container mat-form-field","[class.mat-input-invalid]":"_control.errorState","[class.mat-form-field-invalid]":"_control.errorState","[class.mat-form-field-can-float]":"_canLabelFloat","[class.mat-form-field-should-float]":"_shouldLabelFloat()","[class.mat-form-field-hide-placeholder]":"_hideControlPlaceholder()","[class.mat-form-field-disabled]":"_control.disabled","[class.mat-focused]":"_control.focused","[class.mat-primary]":'color == "primary"',"[class.mat-accent]":'color == "accent"',"[class.mat-warn]":'color == "warn"',"[class.ng-untouched]":'_shouldForward("untouched")',"[class.ng-touched]":'_shouldForward("touched")',"[class.ng-pristine]":'_shouldForward("pristine")',"[class.ng-dirty]":'_shouldForward("dirty")',"[class.ng-valid]":'_shouldForward("valid")',"[class.ng-invalid]":'_shouldForward("invalid")',"[class.ng-pending]":'_shouldForward("pending")'},encapsulation:i._10.None,preserveWhitespaces:!1,changeDetection:i.i.OnPush}]}],t.ctorParameters=function(){return[{type:i.t},{type:i.j},{type:void 0,decorators:[{type:i.M},{type:i.y,args:[l.g]}]}]},t.propDecorators={color:[{type:i.C}],dividerColor:[{type:i.C}],hideRequiredMarker:[{type:i.C}],hintLabel:[{type:i.C}],floatPlaceholder:[{type:i.C}],floatLabel:[{type:i.C}],underlineRef:[{type:i._7,args:["underline"]}],_connectionContainerRef:[{type:i._7,args:["connectionContainer"]}],_inputContainerRef:[{type:i._7,args:["inputContainer"]}],_label:[{type:i._7,args:["label"]}],_control:[{type:i.q,args:[f]}],_placeholderChild:[{type:i.q,args:[y]}],_labelChild:[{type:i.q,args:[b]}],_errorChildren:[{type:i.r,args:[h]}],_hintChildren:[{type:i.r,args:[v]}],_prefixChildren:[{type:i.r,args:[_]}],_suffixChildren:[{type:i.r,args:[j]}]},t}(),O=function(){function t(){}return t.decorators=[{type:i.H,args:[{declarations:[h,v,w,y,_,j,b],imports:[r.a,o.b],exports:[h,v,w,y,_,j,b]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/grid-list.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return w});var r=n("../../../core/esm5/core.js"),i=n("../../../material/esm5/core.es5.js"),o=n("../../../../tslib/tslib.es6.js"),s=n("../../../cdk/esm5/bidi.es5.js");function a(t){return""+(t||"")}function c(t){return"string"==typeof t?parseInt(t,10):t}var u=function(){function t(t){this._element=t,this._rowspan=1,this._colspan=1}return Object.defineProperty(t.prototype,"rowspan",{get:function(){return this._rowspan},set:function(t){this._rowspan=c(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"colspan",{get:function(){return this._colspan},set:function(t){this._colspan=c(t)},enumerable:!0,configurable:!0}),t.prototype._setStyle=function(t,e){this._element.nativeElement.style[t]=e},t.decorators=[{type:r.m,args:[{selector:"mat-grid-tile",exportAs:"matGridTile",host:{class:"mat-grid-tile"},template:'<figure class="mat-figure"><ng-content></ng-content></figure>',styles:[".mat-grid-list{display:block;position:relative}.mat-grid-tile{display:block;position:absolute;overflow:hidden}.mat-grid-tile .mat-figure{top:0;left:0;right:0;bottom:0;position:absolute;display:flex;align-items:center;justify-content:center;height:100%;padding:0;margin:0}.mat-grid-tile .mat-grid-tile-footer,.mat-grid-tile .mat-grid-tile-header{display:flex;align-items:center;height:48px;color:#fff;background:rgba(0,0,0,.38);overflow:hidden;padding:0 16px;position:absolute;left:0;right:0}.mat-grid-tile .mat-grid-tile-footer>*,.mat-grid-tile .mat-grid-tile-header>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-tile-footer.mat-2-line,.mat-grid-tile .mat-grid-tile-header.mat-2-line{height:68px}.mat-grid-tile .mat-grid-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden}.mat-grid-tile .mat-grid-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-list-text:empty{display:none}.mat-grid-tile .mat-grid-tile-header{top:0}.mat-grid-tile .mat-grid-tile-footer{bottom:0}.mat-grid-tile .mat-grid-avatar{padding-right:16px}[dir=rtl] .mat-grid-tile .mat-grid-avatar{padding-right:0;padding-left:16px}.mat-grid-tile .mat-grid-avatar:empty{display:none}"],encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush}]}],t.ctorParameters=function(){return[{type:r.t}]},t.propDecorators={rowspan:[{type:r.C}],colspan:[{type:r.C}]},t}(),l=function(){function t(t){this._element=t}return t.prototype.ngAfterContentInit=function(){this._lineSetter=new i.m(this._lines,this._element)},t.decorators=[{type:r.m,args:[{selector:"mat-grid-tile-header, mat-grid-tile-footer",template:'<ng-content select="[mat-grid-avatar], [matGridAvatar]"></ng-content><div class="mat-grid-list-text"><ng-content select="[mat-line], [matLine]"></ng-content></div><ng-content></ng-content>',changeDetection:r.i.OnPush,encapsulation:r._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[{type:r.t}]},t.propDecorators={_lines:[{type:r.r,args:[i.k]}]},t}(),p=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"[mat-grid-avatar], [matGridAvatar]",host:{class:"mat-grid-avatar"}}]}],t.ctorParameters=function(){return[]},t}(),d=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"mat-grid-tile-header",host:{class:"mat-grid-tile-header"}}]}],t.ctorParameters=function(){return[]},t}(),h=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"mat-grid-tile-footer",host:{class:"mat-grid-tile-footer"}}]}],t.ctorParameters=function(){return[]},t}(),f=function(){function t(t,e){var n=this;this.columnIndex=0,this.rowIndex=0,this.tracker=new Array(t),this.tracker.fill(0,0,this.tracker.length),this.positions=e.map(function(t){return n._trackTile(t)})}return Object.defineProperty(t.prototype,"rowCount",{get:function(){return this.rowIndex+1},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"rowspan",{get:function(){var t=Math.max.apply(Math,this.tracker);return t>1?this.rowCount+t-1:this.rowCount},enumerable:!0,configurable:!0}),t.prototype._trackTile=function(t){var e=this._findMatchingGap(t.colspan);return this._markTilePosition(e,t),this.columnIndex=e+t.colspan,new m(this.rowIndex,e)},t.prototype._findMatchingGap=function(t){if(t>this.tracker.length)throw Error("mat-grid-list: tile with colspan "+t+' is wider than grid with cols="'+this.tracker.length+'".');var e=-1,n=-1;do{this.columnIndex+t>this.tracker.length?this._nextRow():-1!=(e=this.tracker.indexOf(0,this.columnIndex))?(n=this._findGapEndIndex(e),this.columnIndex=e+1):this._nextRow()}while(n-e<t);return e},t.prototype._nextRow=function(){this.columnIndex=0,this.rowIndex++;for(var t=0;t<this.tracker.length;t++)this.tracker[t]=Math.max(0,this.tracker[t]-1)},t.prototype._findGapEndIndex=function(t){for(var e=t+1;e<this.tracker.length;e++)if(0!=this.tracker[e])return e;return this.tracker.length},t.prototype._markTilePosition=function(t,e){for(var n=0;n<e.colspan;n++)this.tracker[t+n]=e.rowspan},t}(),m=function(){return function(t,e){this.row=t,this.col=e}}(),g=function(){function t(){this._rows=0,this._rowspan=0}return t.prototype.init=function(t,e,n,r){this._gutterSize=j(t),this._rows=e.rowCount,this._rowspan=e.rowspan,this._cols=n,this._direction=r},t.prototype.getBaseTileSize=function(t,e){return"("+t+"% - ("+this._gutterSize+" * "+e+"))"},t.prototype.getTilePosition=function(t,e){return 0===e?"0":_("("+t+" + "+this._gutterSize+") * "+e)},t.prototype.getTileSize=function(t,e){return"("+t+" * "+e+") + ("+(e-1)+" * "+this._gutterSize+")"},t.prototype.setStyle=function(t,e,n){var r=100/this._cols,i=(this._cols-1)/this._cols;this.setColStyles(t,n,r,i),this.setRowStyles(t,e,r,i)},t.prototype.setColStyles=function(t,e,n,r){var i=this.getBaseTileSize(n,r),o="ltr"===this._direction?"left":"right";t._setStyle(o,this.getTilePosition(i,e)),t._setStyle("width",_(this.getTileSize(i,t.colspan)))},t.prototype.getGutterSpan=function(){return this._gutterSize+" * ("+this._rowspan+" - 1)"},t.prototype.getTileSpan=function(t){return this._rowspan+" * "+this.getTileSize(t,1)},t.prototype.getComputedHeight=function(){return null},t}(),v=function(t){function e(e){var n=t.call(this)||this;return n.fixedRowHeight=e,n}return Object(o.b)(e,t),e.prototype.init=function(e,n,r,i){t.prototype.init.call(this,e,n,r,i),this.fixedRowHeight=j(this.fixedRowHeight)},e.prototype.setRowStyles=function(t,e){t._setStyle("top",this.getTilePosition(this.fixedRowHeight,e)),t._setStyle("height",_(this.getTileSize(this.fixedRowHeight,t.rowspan)))},e.prototype.getComputedHeight=function(){return["height",_(this.getTileSpan(this.fixedRowHeight)+" + "+this.getGutterSpan())]},e.prototype.reset=function(t){t._setListStyle(["height",null]),t._tiles.forEach(function(t){t._setStyle("top",null),t._setStyle("height",null)})},e}(g),y=function(t){function e(e){var n=t.call(this)||this;return n._parseRatio(e),n}return Object(o.b)(e,t),e.prototype.setRowStyles=function(t,e,n,r){var i=n/this.rowHeightRatio;this.baseTileHeight=this.getBaseTileSize(i,r),t._setStyle("margin-top",this.getTilePosition(this.baseTileHeight,e)),t._setStyle("padding-top",_(this.getTileSize(this.baseTileHeight,t.rowspan)))},e.prototype.getComputedHeight=function(){return["padding-bottom",_(this.getTileSpan(this.baseTileHeight)+" + "+this.getGutterSpan())]},e.prototype.reset=function(t){t._setListStyle(["padding-bottom",null]),t._tiles.forEach(function(t){t._setStyle("margin-top",null),t._setStyle("padding-top",null)})},e.prototype._parseRatio=function(t){var e=t.split(":");if(2!==e.length)throw Error('mat-grid-list: invalid ratio given for row-height: "'+t+'"');this.rowHeightRatio=parseFloat(e[0])/parseFloat(e[1])},e}(g),b=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.b)(e,t),e.prototype.setRowStyles=function(t,e){var n=100/this._rowspan,r=(this._rows-1)/this._rows,i=this.getBaseTileSize(n,r);t._setStyle("top",this.getTilePosition(i,e)),t._setStyle("height",_(this.getTileSize(i,t.rowspan)))},e.prototype.reset=function(t){t._tiles.forEach(function(t){t._setStyle("top",null),t._setStyle("height",null)})},e}(g);function _(t){return"calc("+t+")"}function j(t){return t.match(/px|em|rem/)?t:t+"px"}var x=function(){function t(t,e){this._element=t,this._dir=e,this._gutter="1px"}return Object.defineProperty(t.prototype,"cols",{get:function(){return this._cols},set:function(t){this._cols=c(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"gutterSize",{get:function(){return this._gutter},set:function(t){this._gutter=a(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"rowHeight",{set:function(t){var e=a(t);e!==this._rowHeight&&(this._rowHeight=e,this._setTileStyler(this._rowHeight))},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this._checkCols(),this._checkRowHeight()},t.prototype.ngAfterContentChecked=function(){this._layoutTiles()},t.prototype._checkCols=function(){if(!this.cols)throw Error('mat-grid-list: must pass in number of columns. Example: <mat-grid-list cols="3">')},t.prototype._checkRowHeight=function(){this._rowHeight||this._setTileStyler("1:1")},t.prototype._setTileStyler=function(t){this._tileStyler&&this._tileStyler.reset(this),"fit"===t?this._tileStyler=new b:t&&t.indexOf(":")>-1?this._tileStyler=new y(t):this._tileStyler=new v(t)},t.prototype._layoutTiles=function(){var t=this,e=new f(this.cols,this._tiles),n=this._dir?this._dir.value:"ltr";this._tileStyler.init(this.gutterSize,e,this.cols,n),this._tiles.forEach(function(n,r){var i=e.positions[r];t._tileStyler.setStyle(n,i.row,i.col)}),this._setListStyle(this._tileStyler.getComputedHeight())},t.prototype._setListStyle=function(t){t&&(this._element.nativeElement.style[t[0]]=t[1])},t.decorators=[{type:r.m,args:[{selector:"mat-grid-list",exportAs:"matGridList",template:"<div><ng-content></ng-content></div>",styles:[".mat-grid-list{display:block;position:relative}.mat-grid-tile{display:block;position:absolute;overflow:hidden}.mat-grid-tile .mat-figure{top:0;left:0;right:0;bottom:0;position:absolute;display:flex;align-items:center;justify-content:center;height:100%;padding:0;margin:0}.mat-grid-tile .mat-grid-tile-footer,.mat-grid-tile .mat-grid-tile-header{display:flex;align-items:center;height:48px;color:#fff;background:rgba(0,0,0,.38);overflow:hidden;padding:0 16px;position:absolute;left:0;right:0}.mat-grid-tile .mat-grid-tile-footer>*,.mat-grid-tile .mat-grid-tile-header>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-tile-footer.mat-2-line,.mat-grid-tile .mat-grid-tile-header.mat-2-line{height:68px}.mat-grid-tile .mat-grid-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden}.mat-grid-tile .mat-grid-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-grid-tile .mat-grid-list-text:empty{display:none}.mat-grid-tile .mat-grid-tile-header{top:0}.mat-grid-tile .mat-grid-tile-footer{bottom:0}.mat-grid-tile .mat-grid-avatar{padding-right:16px}[dir=rtl] .mat-grid-tile .mat-grid-avatar{padding-right:0;padding-left:16px}.mat-grid-tile .mat-grid-avatar:empty{display:none}"],host:{class:"mat-grid-list"},changeDetection:r.i.OnPush,encapsulation:r._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[{type:r.t},{type:s.b,decorators:[{type:r.M}]}]},t.propDecorators={_tiles:[{type:r.r,args:[u]}],cols:[{type:r.C}],gutterSize:[{type:r.C}],rowHeight:[{type:r.C}]},t}(),w=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.l,i.j],exports:[x,u,l,i.l,i.j,d,h,p],declarations:[x,u,l,d,h,p]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/icon.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return k});var r=n("../../../core/esm5/core.js"),i=n("../../../material/esm5/core.es5.js"),o=n("../../../../tslib/tslib.es6.js"),s=n("../../../../rxjs/_esm5/operators/take.js"),a=n("../../../../rxjs/_esm5/operators/catchError.js"),c=n("../../../../rxjs/_esm5/operators/tap.js"),u=n("../../../../rxjs/_esm5/operators/finalize.js"),l=n("../../../../rxjs/_esm5/operators/map.js"),p=n("../../../../rxjs/_esm5/operators/share.js"),d=n("../../../common/esm5/http.js"),h=n("../../../platform-browser/esm5/platform-browser.js"),f=n("../../../../rxjs/_esm5/observable/forkJoin.js"),m=n("../../../../rxjs/_esm5/observable/of.js"),g=n("../../../../rxjs/_esm5/observable/throw.js");function v(t){return Error('Unable to find icon with the name "'+t+'"')}function y(t){return Error("The URL provided to MatIconRegistry was not trusted as a resource URL via Angular's DomSanitizer. Attempted URL was \""+t+'".')}var b=function(){return function(t){this.url=t,this.svgElement=null}}(),_=function(){function t(t,e){this._httpClient=t,this._sanitizer=e,this._svgIconConfigs=new Map,this._iconSetConfigs=new Map,this._cachedIconsByUrl=new Map,this._inProgressUrlFetches=new Map,this._fontCssClassesByAlias=new Map,this._defaultFontSetClass="material-icons"}return t.prototype.addSvgIcon=function(t,e){return this.addSvgIconInNamespace("",t,e)},t.prototype.addSvgIconInNamespace=function(t,e,n){var r=w(t,e);return this._svgIconConfigs.set(r,new b(n)),this},t.prototype.addSvgIconSet=function(t){return this.addSvgIconSetInNamespace("",t)},t.prototype.addSvgIconSetInNamespace=function(t,e){var n=new b(e),r=this._iconSetConfigs.get(t);return r?r.push(n):this._iconSetConfigs.set(t,[n]),this},t.prototype.registerFontClassAlias=function(t,e){return void 0===e&&(e=t),this._fontCssClassesByAlias.set(t,e),this},t.prototype.classNameForFontAlias=function(t){return this._fontCssClassesByAlias.get(t)||t},t.prototype.setDefaultFontSetClass=function(t){return this._defaultFontSetClass=t,this},t.prototype.getDefaultFontSetClass=function(){return this._defaultFontSetClass},t.prototype.getSvgIconFromUrl=function(t){var e=this,n=this._sanitizer.sanitize(r.Y.RESOURCE_URL,t);if(!n)throw y(t);var i=this._cachedIconsByUrl.get(n);return i?Object(m.a)(x(i)):this._loadSvgIconFromConfig(new b(t)).pipe(Object(c.a)(function(t){return e._cachedIconsByUrl.set(n,t)}),Object(l.a)(function(t){return x(t)}))},t.prototype.getNamedSvgIcon=function(t,e){void 0===e&&(e="");var n=w(e,t),r=this._svgIconConfigs.get(n);if(r)return this._getSvgFromConfig(r);var i=this._iconSetConfigs.get(e);return i?this._getSvgFromIconSetConfigs(t,i):Object(g.a)(v(n))},t.prototype._getSvgFromConfig=function(t){return t.svgElement?Object(m.a)(x(t.svgElement)):this._loadSvgIconFromConfig(t).pipe(Object(c.a)(function(e){return t.svgElement=e}),Object(l.a)(function(t){return x(t)}))},t.prototype._getSvgFromIconSetConfigs=function(t,e){var n=this,i=this._extractIconWithNameFromAnySet(t,e);if(i)return Object(m.a)(i);var o=e.filter(function(t){return!t.svgElement}).map(function(t){return n._loadSvgIconSetFromConfig(t).pipe(Object(a.a)(function(e){var i=n._sanitizer.sanitize(r.Y.RESOURCE_URL,t.url);return console.log("Loading icon set URL: "+i+" failed: "+e),Object(m.a)(null)}),Object(c.a)(function(e){e&&(t.svgElement=e)}))});return Object(f.a)(o).pipe(Object(l.a)(function(){var r=n._extractIconWithNameFromAnySet(t,e);if(!r)throw v(t);return r}))},t.prototype._extractIconWithNameFromAnySet=function(t,e){for(var n=e.length-1;n>=0;n--){var r=e[n];if(r.svgElement){var i=this._extractSvgIconFromSet(r.svgElement,t);if(i)return i}}return null},t.prototype._loadSvgIconFromConfig=function(t){var e=this;return this._fetchUrl(t.url).pipe(Object(l.a)(function(t){return e._createSvgElementForSingleIcon(t)}))},t.prototype._loadSvgIconSetFromConfig=function(t){var e=this;return this._fetchUrl(t.url).pipe(Object(l.a)(function(t){return e._svgElementFromString(t)}))},t.prototype._createSvgElementForSingleIcon=function(t){var e=this._svgElementFromString(t);return this._setSvgAttributes(e),e},t.prototype._extractSvgIconFromSet=function(t,e){var n=t.querySelector("#"+e);if(!n)return null;var r=n.cloneNode(!0);if(r.id="","svg"===r.nodeName.toLowerCase())return this._setSvgAttributes(r);if("symbol"===r.nodeName.toLowerCase())return this._setSvgAttributes(this._toSvgElement(r));var i=this._svgElementFromString("<svg></svg>");return i.appendChild(r),this._setSvgAttributes(i)},t.prototype._svgElementFromString=function(t){var e=document.createElement("DIV");e.innerHTML=t;var n=e.querySelector("svg");if(!n)throw Error("<svg> tag not found");return n},t.prototype._toSvgElement=function(t){for(var e=this._svgElementFromString("<svg></svg>"),n=0;n<t.childNodes.length;n++)t.childNodes[n].nodeType===Node.ELEMENT_NODE&&e.appendChild(t.childNodes[n].cloneNode(!0));return e},t.prototype._setSvgAttributes=function(t){return t.getAttribute("xmlns")||t.setAttribute("xmlns","http://www.w3.org/2000/svg"),t.setAttribute("fit",""),t.setAttribute("height","100%"),t.setAttribute("width","100%"),t.setAttribute("preserveAspectRatio","xMidYMid meet"),t.setAttribute("focusable","false"),t},t.prototype._fetchUrl=function(t){var e=this;if(!this._httpClient)throw Error("Could not find HttpClient provider for use with Angular Material icons. Please include the HttpClientModule from @angular/common/http in your app imports.");var n=this._sanitizer.sanitize(r.Y.RESOURCE_URL,t);if(!n)throw y(t);var i=this._inProgressUrlFetches.get(n);if(i)return i;var o=this._httpClient.get(n,{responseType:"text"}).pipe(Object(u.a)(function(){return e._inProgressUrlFetches.delete(n)}),Object(p.a)());return this._inProgressUrlFetches.set(n,o),o},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:d.a,decorators:[{type:r.M}]},{type:h.c}]},t}();var j={provide:_,deps:[[new r.M,new r._1,_],[new r.M,d.a],h.c],useFactory:function(t,e,n){return t||new _(e,n)}};function x(t){return t.cloneNode(!0)}function w(t,e){return t+":"+e}var O=function(){return function(t){this._elementRef=t}}(),C=function(t){function e(e,n,r){var i=t.call(this,e)||this;return i._iconRegistry=n,r||e.nativeElement.setAttribute("aria-hidden","true"),i}return Object(o.b)(e,t),e.prototype._splitIconName=function(t){if(!t)return["",""];var e=t.split(":");switch(e.length){case 1:return["",e[0]];case 2:return e;default:throw Error('Invalid icon name: "'+t+'"')}},e.prototype.ngOnChanges=function(t){var e=this;if(t.svgIcon)if(this.svgIcon){var n=this._splitIconName(this.svgIcon),r=n[0],i=n[1];this._iconRegistry.getNamedSvgIcon(i,r).pipe(Object(s.a)(1)).subscribe(function(t){return e._setSvgElement(t)},function(t){return console.log("Error retrieving icon: "+t.message)})}else this._clearSvgElement();this._usingFontIcon()&&this._updateFontIconClasses()},e.prototype.ngOnInit=function(){this._usingFontIcon()&&this._updateFontIconClasses()},e.prototype._usingFontIcon=function(){return!this.svgIcon},e.prototype._setSvgElement=function(t){this._clearSvgElement(),this._elementRef.nativeElement.appendChild(t)},e.prototype._clearSvgElement=function(){for(var t=this._elementRef.nativeElement,e=t.childNodes.length,n=0;n<e;n++)t.removeChild(t.childNodes[n])},e.prototype._updateFontIconClasses=function(){if(this._usingFontIcon()){var t=this._elementRef.nativeElement,e=this.fontSet?this._iconRegistry.classNameForFontAlias(this.fontSet):this._iconRegistry.getDefaultFontSetClass();e!=this._previousFontSetClass&&(this._previousFontSetClass&&t.classList.remove(this._previousFontSetClass),e&&t.classList.add(e),this._previousFontSetClass=e),this.fontIcon!=this._previousFontIconClass&&(this._previousFontIconClass&&t.classList.remove(this._previousFontIconClass),this.fontIcon&&t.classList.add(this.fontIcon),this._previousFontIconClass=this.fontIcon)}},e.decorators=[{type:r.m,args:[{template:"<ng-content></ng-content>",selector:"mat-icon",exportAs:"matIcon",styles:[".mat-icon{background-repeat:no-repeat;display:inline-block;fill:currentColor;height:24px;width:24px}"],inputs:["color"],host:{role:"img",class:"mat-icon"},encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush}]}],e.ctorParameters=function(){return[{type:r.t},{type:_},{type:void 0,decorators:[{type:r.g,args:["aria-hidden"]}]}]},e.propDecorators={svgIcon:[{type:r.C}],fontSet:[{type:r.C}],fontIcon:[{type:r.C}]},e}(Object(i.u)(O)),k=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.j],exports:[C,i.j],declarations:[C],providers:[j]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/input.es5.js":function(t,e,n){"use strict";n.d(e,"b",function(){return y}),n.d(e,"a",function(){return h});var r=n("../../../cdk/esm5/platform.es5.js"),i=n("../../../common/esm5/common.js"),o=n("../../../core/esm5/core.js"),s=n("../../../material/esm5/form-field.es5.js"),a=n("../../../../tslib/tslib.es6.js"),c=n("../../../cdk/esm5/coercion.es5.js"),u=n("../../../forms/esm5/forms.js"),l=n("../../../material/esm5/core.es5.js"),p=n("../../../../rxjs/_esm5/Subject.js"),d=function(){function t(t,e){this._elementRef=t,this._platform=e}return Object.defineProperty(t.prototype,"minRows",{get:function(){return this._minRows},set:function(t){this._minRows=t,this._setMinHeight()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"maxRows",{get:function(){return this._maxRows},set:function(t){this._maxRows=t,this._setMaxHeight()},enumerable:!0,configurable:!0}),t.prototype._setMinHeight=function(){var t=this.minRows&&this._cachedLineHeight?this.minRows*this._cachedLineHeight+"px":null;t&&this._setTextareaStyle("minHeight",t)},t.prototype._setMaxHeight=function(){var t=this.maxRows&&this._cachedLineHeight?this.maxRows*this._cachedLineHeight+"px":null;t&&this._setTextareaStyle("maxHeight",t)},t.prototype.ngAfterViewInit=function(){this._platform.isBrowser&&this.resizeToFitContent()},t.prototype._setTextareaStyle=function(t,e){this._elementRef.nativeElement.style[t]=e},t.prototype._cacheTextareaLineHeight=function(){if(!this._cachedLineHeight){var t=this._elementRef.nativeElement,e=t.cloneNode(!1);e.rows=1,e.style.position="absolute",e.style.visibility="hidden",e.style.border="none",e.style.padding="0",e.style.height="",e.style.minHeight="",e.style.maxHeight="",e.style.overflow="hidden",t.parentNode.appendChild(e),this._cachedLineHeight=e.clientHeight,t.parentNode.removeChild(e),this._setMinHeight(),this._setMaxHeight()}},t.prototype.ngDoCheck=function(){this._platform.isBrowser&&this.resizeToFitContent()},t.prototype.resizeToFitContent=function(){if(this._cacheTextareaLineHeight(),this._cachedLineHeight){var t=this._elementRef.nativeElement,e=t.value;if(e!==this._previousValue){var n=t.placeholder;t.style.height="auto",t.style.overflow="hidden",t.placeholder="",t.style.height=t.scrollHeight+"px",t.style.overflow="",t.placeholder=n,this._previousValue=e}}},t.decorators=[{type:o.s,args:[{selector:"textarea[mat-autosize], textarea[matTextareaAutosize]",exportAs:"matTextareaAutosize",host:{rows:"1"}}]}],t.ctorParameters=function(){return[{type:o.t},{type:r.a}]},t.propDecorators={minRows:[{type:o.C,args:["matAutosizeMinRows"]}],maxRows:[{type:o.C,args:["matAutosizeMaxRows"]}]},t}();var h=new o.A("MAT_INPUT_VALUE_ACCESSOR"),f=["button","checkbox","color","file","hidden","image","radio","range","reset","submit"],m=0,g=function(){return function(t,e,n,r){this._defaultErrorStateMatcher=t,this._parentForm=e,this._parentFormGroup=n,this.ngControl=r}}(),v=function(t){function e(e,n,i,o,s,a,c){var u=t.call(this,a,o,s,i)||this;return u._elementRef=e,u._platform=n,u.ngControl=i,u._type="text",u._disabled=!1,u._required=!1,u._uid="mat-input-"+m++,u._readonly=!1,u.focused=!1,u._isServer=!1,u.stateChanges=new p.a,u.controlType="mat-input",u.placeholder="",u._neverEmptyInputTypes=["date","datetime","datetime-local","month","time","week"].filter(function(t){return Object(r.c)().has(t)}),u._inputValueAccessor=c||u._elementRef.nativeElement,u._previousNativeValue=u.value,u.id=u.id,n.IOS&&e.nativeElement.addEventListener("keyup",function(t){var e=t.target;e.value||e.selectionStart||e.selectionEnd||(e.setSelectionRange(1,1),e.setSelectionRange(0,0))}),u._isServer=!u._platform.isBrowser,u}return Object(a.b)(e,t),Object.defineProperty(e.prototype,"disabled",{get:function(){return this.ngControl?this.ngControl.disabled:this._disabled},set:function(t){this._disabled=Object(c.b)(t),this.focused&&(this.focused=!1,this.stateChanges.next())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"id",{get:function(){return this._id},set:function(t){this._id=t||this._uid},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(c.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"type",{get:function(){return this._type},set:function(t){this._type=t||"text",this._validateType(),!this._isTextarea()&&Object(r.c)().has(this._type)&&(this._elementRef.nativeElement.type=this._type)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._inputValueAccessor.value},set:function(t){t!==this.value&&(this._inputValueAccessor.value=t,this.stateChanges.next())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"readonly",{get:function(){return this._readonly},set:function(t){this._readonly=Object(c.b)(t)},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(){this.stateChanges.next()},e.prototype.ngOnDestroy=function(){this.stateChanges.complete()},e.prototype.ngDoCheck=function(){this.ngControl?this.updateErrorState():this._dirtyCheckNativeValue()},e.prototype.focus=function(){this._elementRef.nativeElement.focus()},e.prototype._focusChanged=function(t){t===this.focused||this.readonly||(this.focused=t,this.stateChanges.next())},e.prototype._onInput=function(){},e.prototype._dirtyCheckNativeValue=function(){var t=this.value;this._previousNativeValue!==t&&(this._previousNativeValue=t,this.stateChanges.next())},e.prototype._validateType=function(){if(f.indexOf(this._type)>-1)throw t=this._type,Error('Input type "'+t+"\" isn't supported by matInput.");var t},e.prototype._isNeverEmpty=function(){return this._neverEmptyInputTypes.indexOf(this._type)>-1},e.prototype._isBadInput=function(){var t=this._elementRef.nativeElement.validity;return t&&t.badInput},e.prototype._isTextarea=function(){var t=this._elementRef.nativeElement,e=this._platform.isBrowser?t.nodeName:t.name;return!!e&&"textarea"===e.toLowerCase()},Object.defineProperty(e.prototype,"empty",{get:function(){return!this._isNeverEmpty()&&!this._elementRef.nativeElement.value&&!this._isBadInput()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"shouldLabelFloat",{get:function(){return this.focused||!this.empty},enumerable:!0,configurable:!0}),e.prototype.setDescribedByIds=function(t){this._ariaDescribedby=t.join(" ")},e.prototype.onContainerClick=function(){this.focus()},e.decorators=[{type:o.s,args:[{selector:"input[matInput], textarea[matInput]",exportAs:"matInput",host:{class:"mat-input-element mat-form-field-autofill-control","[class.mat-input-server]":"_isServer","[attr.id]":"id","[placeholder]":"placeholder","[disabled]":"disabled","[required]":"required","[readonly]":"readonly","[attr.aria-describedby]":"_ariaDescribedby || null","[attr.aria-invalid]":"errorState","[attr.aria-required]":"required.toString()","(blur)":"_focusChanged(false)","(focus)":"_focusChanged(true)","(input)":"_onInput()"},providers:[{provide:s.b,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:o.t},{type:r.a},{type:u.e,decorators:[{type:o.M},{type:o.Z}]},{type:u.f,decorators:[{type:o.M}]},{type:u.b,decorators:[{type:o.M}]},{type:l.d},{type:void 0,decorators:[{type:o.M},{type:o.Z},{type:o.y,args:[h]}]}]},e.propDecorators={disabled:[{type:o.C}],id:[{type:o.C}],placeholder:[{type:o.C}],required:[{type:o.C}],type:[{type:o.C}],errorStateMatcher:[{type:o.C}],value:[{type:o.C}],readonly:[{type:o.C}]},e}(Object(l.x)(g)),y=function(){function t(){}return t.decorators=[{type:o.H,args:[{declarations:[v,d],imports:[i.a,s.c,r.b],exports:[s.c,v,d],providers:[l.d]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/list.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return D});var r=n("../../../common/esm5/common.js"),i=n("../../../core/esm5/core.js"),o=n("../../../material/esm5/core.es5.js"),s=n("../../../../tslib/tslib.es6.js"),a=n("../../../cdk/esm5/a11y.es5.js"),c=n("../../../cdk/esm5/coercion.es5.js"),u=n("../../../cdk/esm5/collections.es5.js"),l=n("../../../cdk/esm5/keycodes.es5.js"),p=n("../../../forms/esm5/forms.js"),d=function(){return function(){}}(),h=Object(o.v)(d),f=function(){return function(){}}(),m=Object(o.v)(f),g=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"mat-divider",host:{role:"separator","aria-orientation":"horizontal"}}]}],t.ctorParameters=function(){return[]},t}(),v=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(s.b)(e,t),e.decorators=[{type:i.m,args:[{selector:"mat-nav-list",exportAs:"matNavList",host:{role:"navigation",class:"mat-nav-list"},template:"<ng-content></ng-content>",styles:[".mat-subheader{display:flex;box-sizing:border-box;padding:16px;align-items:center}.mat-list .mat-subheader,.mat-nav-list .mat-subheader,.mat-selection-list .mat-subheader{margin:0}.mat-list,.mat-nav-list,.mat-selection-list{padding-top:8px;display:block}.mat-list .mat-subheader,.mat-nav-list .mat-subheader,.mat-selection-list .mat-subheader{height:48px;line-height:16px}.mat-list .mat-subheader:first-child,.mat-nav-list .mat-subheader:first-child,.mat-selection-list .mat-subheader:first-child{margin-top:-8px}.mat-list .mat-list-item,.mat-list .mat-list-option,.mat-nav-list .mat-list-item,.mat-nav-list .mat-list-option,.mat-selection-list .mat-list-item,.mat-selection-list .mat-list-option{display:block;height:48px}.mat-list .mat-list-item .mat-list-item-content,.mat-list .mat-list-option .mat-list-item-content,.mat-nav-list .mat-list-item .mat-list-item-content,.mat-nav-list .mat-list-option .mat-list-item-content,.mat-selection-list .mat-list-item .mat-list-item-content,.mat-selection-list .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list .mat-list-item .mat-list-item-content-reverse,.mat-list .mat-list-option .mat-list-item-content-reverse,.mat-nav-list .mat-list-item .mat-list-item-content-reverse,.mat-nav-list .mat-list-option .mat-list-item-content-reverse,.mat-selection-list .mat-list-item .mat-list-item-content-reverse,.mat-selection-list .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list .mat-list-item .mat-list-item-ripple,.mat-list .mat-list-option .mat-list-item-ripple,.mat-nav-list .mat-list-item .mat-list-item-ripple,.mat-nav-list .mat-list-option .mat-list-item-ripple,.mat-selection-list .mat-list-item .mat-list-item-ripple,.mat-selection-list .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list .mat-list-item.mat-list-item-avatar,.mat-list .mat-list-option.mat-list-item-avatar,.mat-nav-list .mat-list-item.mat-list-item-avatar,.mat-nav-list .mat-list-option.mat-list-item-avatar,.mat-selection-list .mat-list-item.mat-list-item-avatar,.mat-selection-list .mat-list-option.mat-list-item-avatar{height:56px}.mat-list .mat-list-item.mat-2-line,.mat-list .mat-list-option.mat-2-line,.mat-nav-list .mat-list-item.mat-2-line,.mat-nav-list .mat-list-option.mat-2-line,.mat-selection-list .mat-list-item.mat-2-line,.mat-selection-list .mat-list-option.mat-2-line{height:72px}.mat-list .mat-list-item.mat-3-line,.mat-list .mat-list-option.mat-3-line,.mat-nav-list .mat-list-item.mat-3-line,.mat-nav-list .mat-list-option.mat-3-line,.mat-selection-list .mat-list-item.mat-3-line,.mat-selection-list .mat-list-option.mat-3-line{height:88px}.mat-list .mat-list-item.mat-multi-line,.mat-list .mat-list-option.mat-multi-line,.mat-nav-list .mat-list-item.mat-multi-line,.mat-nav-list .mat-list-option.mat-multi-line,.mat-selection-list .mat-list-item.mat-multi-line,.mat-selection-list .mat-list-option.mat-multi-line{height:auto}.mat-list .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list .mat-list-option.mat-multi-line .mat-list-item-content,.mat-nav-list .mat-list-item.mat-multi-line .mat-list-item-content,.mat-nav-list .mat-list-option.mat-multi-line .mat-list-item-content,.mat-selection-list .mat-list-item.mat-multi-line .mat-list-item-content,.mat-selection-list .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list .mat-list-item .mat-list-text,.mat-list .mat-list-option .mat-list-text,.mat-nav-list .mat-list-item .mat-list-text,.mat-nav-list .mat-list-option .mat-list-text,.mat-selection-list .mat-list-item .mat-list-text,.mat-selection-list .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0 16px}.mat-list .mat-list-item .mat-list-text>*,.mat-list .mat-list-option .mat-list-text>*,.mat-nav-list .mat-list-item .mat-list-text>*,.mat-nav-list .mat-list-option .mat-list-text>*,.mat-selection-list .mat-list-item .mat-list-text>*,.mat-selection-list .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list .mat-list-item .mat-list-text:empty,.mat-list .mat-list-option .mat-list-text:empty,.mat-nav-list .mat-list-item .mat-list-text:empty,.mat-nav-list .mat-list-option .mat-list-text:empty,.mat-selection-list .mat-list-item .mat-list-text:empty,.mat-selection-list .mat-list-option .mat-list-text:empty{display:none}.mat-list .mat-list-item .mat-list-text:nth-child(2),.mat-list .mat-list-option .mat-list-text:nth-child(2),.mat-nav-list .mat-list-item .mat-list-text:nth-child(2),.mat-nav-list .mat-list-option .mat-list-text:nth-child(2),.mat-selection-list .mat-list-item .mat-list-text:nth-child(2),.mat-selection-list .mat-list-option .mat-list-text:nth-child(2){padding:0}.mat-list .mat-list-item .mat-list-avatar,.mat-list .mat-list-option .mat-list-avatar,.mat-nav-list .mat-list-item .mat-list-avatar,.mat-nav-list .mat-list-option .mat-list-avatar,.mat-selection-list .mat-list-item .mat-list-avatar,.mat-selection-list .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%}.mat-list .mat-list-item .mat-list-icon,.mat-list .mat-list-option .mat-list-icon,.mat-nav-list .mat-list-item .mat-list-icon,.mat-nav-list .mat-list-option .mat-list-icon,.mat-selection-list .mat-list-item .mat-list-icon,.mat-selection-list .mat-list-option .mat-list-icon{width:24px;height:24px;font-size:24px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list[dense],.mat-nav-list[dense],.mat-selection-list[dense]{padding-top:4px;display:block}.mat-list[dense] .mat-subheader,.mat-nav-list[dense] .mat-subheader,.mat-selection-list[dense] .mat-subheader{height:40px;line-height:8px}.mat-list[dense] .mat-subheader:first-child,.mat-nav-list[dense] .mat-subheader:first-child,.mat-selection-list[dense] .mat-subheader:first-child{margin-top:-4px}.mat-list[dense] .mat-list-item,.mat-list[dense] .mat-list-option,.mat-nav-list[dense] .mat-list-item,.mat-nav-list[dense] .mat-list-option,.mat-selection-list[dense] .mat-list-item,.mat-selection-list[dense] .mat-list-option{display:block;height:40px}.mat-list[dense] .mat-list-item .mat-list-item-content,.mat-list[dense] .mat-list-option .mat-list-item-content,.mat-nav-list[dense] .mat-list-item .mat-list-item-content,.mat-nav-list[dense] .mat-list-option .mat-list-item-content,.mat-selection-list[dense] .mat-list-item .mat-list-item-content,.mat-selection-list[dense] .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list[dense] .mat-list-item .mat-list-item-content-reverse,.mat-list[dense] .mat-list-option .mat-list-item-content-reverse,.mat-nav-list[dense] .mat-list-item .mat-list-item-content-reverse,.mat-nav-list[dense] .mat-list-option .mat-list-item-content-reverse,.mat-selection-list[dense] .mat-list-item .mat-list-item-content-reverse,.mat-selection-list[dense] .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list[dense] .mat-list-item .mat-list-item-ripple,.mat-list[dense] .mat-list-option .mat-list-item-ripple,.mat-nav-list[dense] .mat-list-item .mat-list-item-ripple,.mat-nav-list[dense] .mat-list-option .mat-list-item-ripple,.mat-selection-list[dense] .mat-list-item .mat-list-item-ripple,.mat-selection-list[dense] .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list[dense] .mat-list-item.mat-list-item-avatar,.mat-list[dense] .mat-list-option.mat-list-item-avatar,.mat-nav-list[dense] .mat-list-item.mat-list-item-avatar,.mat-nav-list[dense] .mat-list-option.mat-list-item-avatar,.mat-selection-list[dense] .mat-list-item.mat-list-item-avatar,.mat-selection-list[dense] .mat-list-option.mat-list-item-avatar{height:48px}.mat-list[dense] .mat-list-item.mat-2-line,.mat-list[dense] .mat-list-option.mat-2-line,.mat-nav-list[dense] .mat-list-item.mat-2-line,.mat-nav-list[dense] .mat-list-option.mat-2-line,.mat-selection-list[dense] .mat-list-item.mat-2-line,.mat-selection-list[dense] .mat-list-option.mat-2-line{height:60px}.mat-list[dense] .mat-list-item.mat-3-line,.mat-list[dense] .mat-list-option.mat-3-line,.mat-nav-list[dense] .mat-list-item.mat-3-line,.mat-nav-list[dense] .mat-list-option.mat-3-line,.mat-selection-list[dense] .mat-list-item.mat-3-line,.mat-selection-list[dense] .mat-list-option.mat-3-line{height:76px}.mat-list[dense] .mat-list-item.mat-multi-line,.mat-list[dense] .mat-list-option.mat-multi-line,.mat-nav-list[dense] .mat-list-item.mat-multi-line,.mat-nav-list[dense] .mat-list-option.mat-multi-line,.mat-selection-list[dense] .mat-list-item.mat-multi-line,.mat-selection-list[dense] .mat-list-option.mat-multi-line{height:auto}.mat-list[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list[dense] .mat-list-option.mat-multi-line .mat-list-item-content,.mat-nav-list[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-nav-list[dense] .mat-list-option.mat-multi-line .mat-list-item-content,.mat-selection-list[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-selection-list[dense] .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list[dense] .mat-list-item .mat-list-text,.mat-list[dense] .mat-list-option .mat-list-text,.mat-nav-list[dense] .mat-list-item .mat-list-text,.mat-nav-list[dense] .mat-list-option .mat-list-text,.mat-selection-list[dense] .mat-list-item .mat-list-text,.mat-selection-list[dense] .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0 16px}.mat-list[dense] .mat-list-item .mat-list-text>*,.mat-list[dense] .mat-list-option .mat-list-text>*,.mat-nav-list[dense] .mat-list-item .mat-list-text>*,.mat-nav-list[dense] .mat-list-option .mat-list-text>*,.mat-selection-list[dense] .mat-list-item .mat-list-text>*,.mat-selection-list[dense] .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list[dense] .mat-list-item .mat-list-text:empty,.mat-list[dense] .mat-list-option .mat-list-text:empty,.mat-nav-list[dense] .mat-list-item .mat-list-text:empty,.mat-nav-list[dense] .mat-list-option .mat-list-text:empty,.mat-selection-list[dense] .mat-list-item .mat-list-text:empty,.mat-selection-list[dense] .mat-list-option .mat-list-text:empty{display:none}.mat-list[dense] .mat-list-item .mat-list-text:nth-child(2),.mat-list[dense] .mat-list-option .mat-list-text:nth-child(2),.mat-nav-list[dense] .mat-list-item .mat-list-text:nth-child(2),.mat-nav-list[dense] .mat-list-option .mat-list-text:nth-child(2),.mat-selection-list[dense] .mat-list-item .mat-list-text:nth-child(2),.mat-selection-list[dense] .mat-list-option .mat-list-text:nth-child(2){padding:0}.mat-list[dense] .mat-list-item .mat-list-avatar,.mat-list[dense] .mat-list-option .mat-list-avatar,.mat-nav-list[dense] .mat-list-item .mat-list-avatar,.mat-nav-list[dense] .mat-list-option .mat-list-avatar,.mat-selection-list[dense] .mat-list-item .mat-list-avatar,.mat-selection-list[dense] .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%}.mat-list[dense] .mat-list-item .mat-list-icon,.mat-list[dense] .mat-list-option .mat-list-icon,.mat-nav-list[dense] .mat-list-item .mat-list-icon,.mat-nav-list[dense] .mat-list-option .mat-list-icon,.mat-selection-list[dense] .mat-list-item .mat-list-icon,.mat-selection-list[dense] .mat-list-option .mat-list-icon{width:20px;height:20px;font-size:20px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-divider{display:block;border-top-style:solid;border-top-width:1px;margin:0}.mat-nav-list a{text-decoration:none;color:inherit}.mat-nav-list .mat-list-item-content{cursor:pointer}.mat-nav-list .mat-list-item-content.mat-list-item-focus,.mat-nav-list .mat-list-item-content:hover{outline:0}.mat-list-option:not([disabled]){cursor:pointer}"],inputs:["disableRipple"],encapsulation:i._10.None,preserveWhitespaces:!1,changeDetection:i.i.OnPush}]}],e.ctorParameters=function(){return[]},e}(h),y=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(s.b)(e,t),e.decorators=[{type:i.m,args:[{selector:"mat-list",exportAs:"matList",template:"<ng-content></ng-content>",host:{class:"mat-list"},styles:[".mat-subheader{display:flex;box-sizing:border-box;padding:16px;align-items:center}.mat-list .mat-subheader,.mat-nav-list .mat-subheader,.mat-selection-list .mat-subheader{margin:0}.mat-list,.mat-nav-list,.mat-selection-list{padding-top:8px;display:block}.mat-list .mat-subheader,.mat-nav-list .mat-subheader,.mat-selection-list .mat-subheader{height:48px;line-height:16px}.mat-list .mat-subheader:first-child,.mat-nav-list .mat-subheader:first-child,.mat-selection-list .mat-subheader:first-child{margin-top:-8px}.mat-list .mat-list-item,.mat-list .mat-list-option,.mat-nav-list .mat-list-item,.mat-nav-list .mat-list-option,.mat-selection-list .mat-list-item,.mat-selection-list .mat-list-option{display:block;height:48px}.mat-list .mat-list-item .mat-list-item-content,.mat-list .mat-list-option .mat-list-item-content,.mat-nav-list .mat-list-item .mat-list-item-content,.mat-nav-list .mat-list-option .mat-list-item-content,.mat-selection-list .mat-list-item .mat-list-item-content,.mat-selection-list .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list .mat-list-item .mat-list-item-content-reverse,.mat-list .mat-list-option .mat-list-item-content-reverse,.mat-nav-list .mat-list-item .mat-list-item-content-reverse,.mat-nav-list .mat-list-option .mat-list-item-content-reverse,.mat-selection-list .mat-list-item .mat-list-item-content-reverse,.mat-selection-list .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list .mat-list-item .mat-list-item-ripple,.mat-list .mat-list-option .mat-list-item-ripple,.mat-nav-list .mat-list-item .mat-list-item-ripple,.mat-nav-list .mat-list-option .mat-list-item-ripple,.mat-selection-list .mat-list-item .mat-list-item-ripple,.mat-selection-list .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list .mat-list-item.mat-list-item-avatar,.mat-list .mat-list-option.mat-list-item-avatar,.mat-nav-list .mat-list-item.mat-list-item-avatar,.mat-nav-list .mat-list-option.mat-list-item-avatar,.mat-selection-list .mat-list-item.mat-list-item-avatar,.mat-selection-list .mat-list-option.mat-list-item-avatar{height:56px}.mat-list .mat-list-item.mat-2-line,.mat-list .mat-list-option.mat-2-line,.mat-nav-list .mat-list-item.mat-2-line,.mat-nav-list .mat-list-option.mat-2-line,.mat-selection-list .mat-list-item.mat-2-line,.mat-selection-list .mat-list-option.mat-2-line{height:72px}.mat-list .mat-list-item.mat-3-line,.mat-list .mat-list-option.mat-3-line,.mat-nav-list .mat-list-item.mat-3-line,.mat-nav-list .mat-list-option.mat-3-line,.mat-selection-list .mat-list-item.mat-3-line,.mat-selection-list .mat-list-option.mat-3-line{height:88px}.mat-list .mat-list-item.mat-multi-line,.mat-list .mat-list-option.mat-multi-line,.mat-nav-list .mat-list-item.mat-multi-line,.mat-nav-list .mat-list-option.mat-multi-line,.mat-selection-list .mat-list-item.mat-multi-line,.mat-selection-list .mat-list-option.mat-multi-line{height:auto}.mat-list .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list .mat-list-option.mat-multi-line .mat-list-item-content,.mat-nav-list .mat-list-item.mat-multi-line .mat-list-item-content,.mat-nav-list .mat-list-option.mat-multi-line .mat-list-item-content,.mat-selection-list .mat-list-item.mat-multi-line .mat-list-item-content,.mat-selection-list .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list .mat-list-item .mat-list-text,.mat-list .mat-list-option .mat-list-text,.mat-nav-list .mat-list-item .mat-list-text,.mat-nav-list .mat-list-option .mat-list-text,.mat-selection-list .mat-list-item .mat-list-text,.mat-selection-list .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0 16px}.mat-list .mat-list-item .mat-list-text>*,.mat-list .mat-list-option .mat-list-text>*,.mat-nav-list .mat-list-item .mat-list-text>*,.mat-nav-list .mat-list-option .mat-list-text>*,.mat-selection-list .mat-list-item .mat-list-text>*,.mat-selection-list .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list .mat-list-item .mat-list-text:empty,.mat-list .mat-list-option .mat-list-text:empty,.mat-nav-list .mat-list-item .mat-list-text:empty,.mat-nav-list .mat-list-option .mat-list-text:empty,.mat-selection-list .mat-list-item .mat-list-text:empty,.mat-selection-list .mat-list-option .mat-list-text:empty{display:none}.mat-list .mat-list-item .mat-list-text:nth-child(2),.mat-list .mat-list-option .mat-list-text:nth-child(2),.mat-nav-list .mat-list-item .mat-list-text:nth-child(2),.mat-nav-list .mat-list-option .mat-list-text:nth-child(2),.mat-selection-list .mat-list-item .mat-list-text:nth-child(2),.mat-selection-list .mat-list-option .mat-list-text:nth-child(2){padding:0}.mat-list .mat-list-item .mat-list-avatar,.mat-list .mat-list-option .mat-list-avatar,.mat-nav-list .mat-list-item .mat-list-avatar,.mat-nav-list .mat-list-option .mat-list-avatar,.mat-selection-list .mat-list-item .mat-list-avatar,.mat-selection-list .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%}.mat-list .mat-list-item .mat-list-icon,.mat-list .mat-list-option .mat-list-icon,.mat-nav-list .mat-list-item .mat-list-icon,.mat-nav-list .mat-list-option .mat-list-icon,.mat-selection-list .mat-list-item .mat-list-icon,.mat-selection-list .mat-list-option .mat-list-icon{width:24px;height:24px;font-size:24px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list[dense],.mat-nav-list[dense],.mat-selection-list[dense]{padding-top:4px;display:block}.mat-list[dense] .mat-subheader,.mat-nav-list[dense] .mat-subheader,.mat-selection-list[dense] .mat-subheader{height:40px;line-height:8px}.mat-list[dense] .mat-subheader:first-child,.mat-nav-list[dense] .mat-subheader:first-child,.mat-selection-list[dense] .mat-subheader:first-child{margin-top:-4px}.mat-list[dense] .mat-list-item,.mat-list[dense] .mat-list-option,.mat-nav-list[dense] .mat-list-item,.mat-nav-list[dense] .mat-list-option,.mat-selection-list[dense] .mat-list-item,.mat-selection-list[dense] .mat-list-option{display:block;height:40px}.mat-list[dense] .mat-list-item .mat-list-item-content,.mat-list[dense] .mat-list-option .mat-list-item-content,.mat-nav-list[dense] .mat-list-item .mat-list-item-content,.mat-nav-list[dense] .mat-list-option .mat-list-item-content,.mat-selection-list[dense] .mat-list-item .mat-list-item-content,.mat-selection-list[dense] .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list[dense] .mat-list-item .mat-list-item-content-reverse,.mat-list[dense] .mat-list-option .mat-list-item-content-reverse,.mat-nav-list[dense] .mat-list-item .mat-list-item-content-reverse,.mat-nav-list[dense] .mat-list-option .mat-list-item-content-reverse,.mat-selection-list[dense] .mat-list-item .mat-list-item-content-reverse,.mat-selection-list[dense] .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list[dense] .mat-list-item .mat-list-item-ripple,.mat-list[dense] .mat-list-option .mat-list-item-ripple,.mat-nav-list[dense] .mat-list-item .mat-list-item-ripple,.mat-nav-list[dense] .mat-list-option .mat-list-item-ripple,.mat-selection-list[dense] .mat-list-item .mat-list-item-ripple,.mat-selection-list[dense] .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list[dense] .mat-list-item.mat-list-item-avatar,.mat-list[dense] .mat-list-option.mat-list-item-avatar,.mat-nav-list[dense] .mat-list-item.mat-list-item-avatar,.mat-nav-list[dense] .mat-list-option.mat-list-item-avatar,.mat-selection-list[dense] .mat-list-item.mat-list-item-avatar,.mat-selection-list[dense] .mat-list-option.mat-list-item-avatar{height:48px}.mat-list[dense] .mat-list-item.mat-2-line,.mat-list[dense] .mat-list-option.mat-2-line,.mat-nav-list[dense] .mat-list-item.mat-2-line,.mat-nav-list[dense] .mat-list-option.mat-2-line,.mat-selection-list[dense] .mat-list-item.mat-2-line,.mat-selection-list[dense] .mat-list-option.mat-2-line{height:60px}.mat-list[dense] .mat-list-item.mat-3-line,.mat-list[dense] .mat-list-option.mat-3-line,.mat-nav-list[dense] .mat-list-item.mat-3-line,.mat-nav-list[dense] .mat-list-option.mat-3-line,.mat-selection-list[dense] .mat-list-item.mat-3-line,.mat-selection-list[dense] .mat-list-option.mat-3-line{height:76px}.mat-list[dense] .mat-list-item.mat-multi-line,.mat-list[dense] .mat-list-option.mat-multi-line,.mat-nav-list[dense] .mat-list-item.mat-multi-line,.mat-nav-list[dense] .mat-list-option.mat-multi-line,.mat-selection-list[dense] .mat-list-item.mat-multi-line,.mat-selection-list[dense] .mat-list-option.mat-multi-line{height:auto}.mat-list[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list[dense] .mat-list-option.mat-multi-line .mat-list-item-content,.mat-nav-list[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-nav-list[dense] .mat-list-option.mat-multi-line .mat-list-item-content,.mat-selection-list[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-selection-list[dense] .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list[dense] .mat-list-item .mat-list-text,.mat-list[dense] .mat-list-option .mat-list-text,.mat-nav-list[dense] .mat-list-item .mat-list-text,.mat-nav-list[dense] .mat-list-option .mat-list-text,.mat-selection-list[dense] .mat-list-item .mat-list-text,.mat-selection-list[dense] .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0 16px}.mat-list[dense] .mat-list-item .mat-list-text>*,.mat-list[dense] .mat-list-option .mat-list-text>*,.mat-nav-list[dense] .mat-list-item .mat-list-text>*,.mat-nav-list[dense] .mat-list-option .mat-list-text>*,.mat-selection-list[dense] .mat-list-item .mat-list-text>*,.mat-selection-list[dense] .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list[dense] .mat-list-item .mat-list-text:empty,.mat-list[dense] .mat-list-option .mat-list-text:empty,.mat-nav-list[dense] .mat-list-item .mat-list-text:empty,.mat-nav-list[dense] .mat-list-option .mat-list-text:empty,.mat-selection-list[dense] .mat-list-item .mat-list-text:empty,.mat-selection-list[dense] .mat-list-option .mat-list-text:empty{display:none}.mat-list[dense] .mat-list-item .mat-list-text:nth-child(2),.mat-list[dense] .mat-list-option .mat-list-text:nth-child(2),.mat-nav-list[dense] .mat-list-item .mat-list-text:nth-child(2),.mat-nav-list[dense] .mat-list-option .mat-list-text:nth-child(2),.mat-selection-list[dense] .mat-list-item .mat-list-text:nth-child(2),.mat-selection-list[dense] .mat-list-option .mat-list-text:nth-child(2){padding:0}.mat-list[dense] .mat-list-item .mat-list-avatar,.mat-list[dense] .mat-list-option .mat-list-avatar,.mat-nav-list[dense] .mat-list-item .mat-list-avatar,.mat-nav-list[dense] .mat-list-option .mat-list-avatar,.mat-selection-list[dense] .mat-list-item .mat-list-avatar,.mat-selection-list[dense] .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%}.mat-list[dense] .mat-list-item .mat-list-icon,.mat-list[dense] .mat-list-option .mat-list-icon,.mat-nav-list[dense] .mat-list-item .mat-list-icon,.mat-nav-list[dense] .mat-list-option .mat-list-icon,.mat-selection-list[dense] .mat-list-item .mat-list-icon,.mat-selection-list[dense] .mat-list-option .mat-list-icon{width:20px;height:20px;font-size:20px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-divider{display:block;border-top-style:solid;border-top-width:1px;margin:0}.mat-nav-list a{text-decoration:none;color:inherit}.mat-nav-list .mat-list-item-content{cursor:pointer}.mat-nav-list .mat-list-item-content.mat-list-item-focus,.mat-nav-list .mat-list-item-content:hover{outline:0}.mat-list-option:not([disabled]){cursor:pointer}"],inputs:["disableRipple"],encapsulation:i._10.None,preserveWhitespaces:!1,changeDetection:i.i.OnPush}]}],e.ctorParameters=function(){return[]},e}(h),b=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"mat-divider",host:{class:"mat-divider"}}]}],t.ctorParameters=function(){return[]},t}(),_=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"[mat-list-avatar], [matListAvatar]",host:{class:"mat-list-avatar"}}]}],t.ctorParameters=function(){return[]},t}(),j=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"[mat-list-icon], [matListIcon]",host:{class:"mat-list-icon"}}]}],t.ctorParameters=function(){return[]},t}(),x=function(){function t(){}return t.decorators=[{type:i.s,args:[{selector:"[mat-subheader], [matSubheader]",host:{class:"mat-subheader"}}]}],t.ctorParameters=function(){return[]},t}(),w=function(t){function e(e,n){var r=t.call(this)||this;return r._element=e,r._navList=n,r._isNavList=!1,r._isNavList=!!n,r}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"_hasAvatar",{set:function(t){null!=t?this._element.nativeElement.classList.add("mat-list-item-avatar"):this._element.nativeElement.classList.remove("mat-list-item-avatar")},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._lineSetter=new o.m(this._lines,this._element)},e.prototype._isRippleDisabled=function(){return!this._isNavList||this.disableRipple||this._navList.disableRipple},e.prototype._handleFocus=function(){this._element.nativeElement.classList.add("mat-list-item-focus")},e.prototype._handleBlur=function(){this._element.nativeElement.classList.remove("mat-list-item-focus")},e.prototype._getHostElement=function(){return this._element.nativeElement},e.decorators=[{type:i.m,args:[{selector:"mat-list-item, a[mat-list-item]",exportAs:"matListItem",host:{class:"mat-list-item","(focus)":"_handleFocus()","(blur)":"_handleBlur()"},inputs:["disableRipple"],template:'<div class="mat-list-item-content"><div class="mat-list-item-ripple" mat-ripple [matRippleTrigger]="_getHostElement()" [matRippleDisabled]="_isRippleDisabled()"></div><ng-content select="[mat-list-avatar], [mat-list-icon], [matListAvatar], [matListIcon]"></ng-content><div class="mat-list-text"><ng-content select="[mat-line], [matLine]"></ng-content></div><ng-content></ng-content></div>',encapsulation:i._10.None,preserveWhitespaces:!1,changeDetection:i.i.OnPush}]}],e.ctorParameters=function(){return[{type:i.t},{type:v,decorators:[{type:i.M}]}]},e.propDecorators={_lines:[{type:i.r,args:[o.k]}],_hasAvatar:[{type:i.q,args:[_]}]},e}(m),O=function(){return function(){}}(),C=Object(o.y)(Object(o.v)(Object(o.w)(O))),k=function(){return function(){}}(),E=Object(o.v)(k),S={provide:p.d,useExisting:Object(i._14)(function(){return T}),multi:!0},P=function(){return function(t,e){this.source=t,this.selected=e}}(),M=function(){return function(t,e){this.source=t,this.option=e}}(),A=function(t){function e(e,n,r){var o=t.call(this)||this;return o._element=e,o._changeDetector=n,o.selectionList=r,o._selected=!1,o._disabled=!1,o._hasFocus=!1,o.checkboxPosition="after",o.selectionChange=new i.w,o}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"disabled",{get:function(){return this.selectionList&&this.selectionList.disabled||this._disabled},set:function(t){var e=Object(c.b)(t);e!==this._disabled&&(this._disabled=e,this._changeDetector.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this.selectionList.selectedOptions.isSelected(this)},set:function(t){var e=Object(c.b)(t);e!==this._selected&&(this._setSelected(e),this.selectionList._reportValueChange())},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){var t=this;this.selected&&Promise.resolve(function(){return t.selected&&t.selectionList._reportValueChange()})},e.prototype.ngAfterContentInit=function(){this._lineSetter=new o.m(this._lines,this._element)},e.prototype.ngOnDestroy=function(){this.selectionList._removeOptionFromList(this)},e.prototype.toggle=function(){this.selected=!this.selected},e.prototype.focus=function(){this._element.nativeElement.focus()},e.prototype._isRippleDisabled=function(){return this.disabled||this.disableRipple||this.selectionList.disableRipple},e.prototype._handleClick=function(){this.disabled||(this.toggle(),this.selectionList._emitChangeEvent(this),this._emitDeprecatedChangeEvent())},e.prototype._handleFocus=function(){this._hasFocus=!0,this.selectionList._setFocusedOption(this)},e.prototype._handleBlur=function(){this._hasFocus=!1,this.selectionList.onTouched()},e.prototype._getHostElement=function(){return this._element.nativeElement},e.prototype._setSelected=function(t){t!==this._selected&&(this._selected=t,t?this.selectionList.selectedOptions.select(this):this.selectionList.selectedOptions.deselect(this),this._changeDetector.markForCheck())},e.prototype._emitDeprecatedChangeEvent=function(){this.selectionChange.emit(new P(this,this.selected))},e.decorators=[{type:i.m,args:[{selector:"mat-list-option",exportAs:"matListOption",inputs:["disableRipple"],host:{role:"option",class:"mat-list-item mat-list-option","(focus)":"_handleFocus()","(blur)":"_handleBlur()","(click)":"_handleClick()",tabindex:"-1","[class.mat-list-item-disabled]":"disabled","[class.mat-list-item-focus]":"_hasFocus","[attr.aria-selected]":"selected.toString()","[attr.aria-disabled]":"disabled.toString()"},template:'<div class="mat-list-item-content" [class.mat-list-item-content-reverse]="checkboxPosition == \'after\'" [class.mat-list-item-disabled]="disabled"><div mat-ripple class="mat-list-item-ripple" [matRippleTrigger]="_getHostElement()" [matRippleDisabled]="_isRippleDisabled()"></div><mat-pseudo-checkbox #autocheckbox [state]="selected ? \'checked\' : \'unchecked\'" [disabled]="disabled"></mat-pseudo-checkbox><div class="mat-list-text"><ng-content></ng-content></div></div>',encapsulation:i._10.None,preserveWhitespaces:!1,changeDetection:i.i.OnPush}]}],e.ctorParameters=function(){return[{type:i.t},{type:i.j},{type:T,decorators:[{type:i.M},{type:i.y,args:[Object(i._14)(function(){return T})]}]}]},e.propDecorators={_lines:[{type:i.r,args:[o.k]}],checkboxPosition:[{type:i.C}],value:[{type:i.C}],disabled:[{type:i.C}],selected:[{type:i.C}],selectionChange:[{type:i.N}]},e}(E),T=function(t){function e(e,n){var r=t.call(this)||this;return r._element=e,r.selectionChange=new i.w,r.selectedOptions=new u.a(!0),r._onChange=function(t){},r.onTouched=function(){},r.tabIndex=parseInt(n)||0,r}return Object(s.b)(e,t),e.prototype.ngAfterContentInit=function(){this._keyManager=new a.e(this.options).withWrap()},e.prototype.focus=function(){this._element.nativeElement.focus()},e.prototype.selectAll=function(){this.options.forEach(function(t){return t._setSelected(!0)}),this._reportValueChange()},e.prototype.deselectAll=function(){this.options.forEach(function(t){return t._setSelected(!1)}),this._reportValueChange()},e.prototype._setFocusedOption=function(t){this._keyManager.updateActiveItemIndex(this._getOptionIndex(t))},e.prototype._removeOptionFromList=function(t){if(t._hasFocus){var e=this._getOptionIndex(t);e>0?this._keyManager.setPreviousItemActive():0===e&&this.options.length>1&&this._keyManager.setNextItemActive()}},e.prototype._keydown=function(t){switch(t.keyCode){case l.n:this._toggleSelectOnFocusedOption(),t.preventDefault();break;default:this._keyManager.onKeydown(t)}},e.prototype._reportValueChange=function(){this.options&&this._onChange(this._getSelectedOptionValues())},e.prototype._emitChangeEvent=function(t){this.selectionChange.emit(new M(this,t))},e.prototype.writeValue=function(t){this.options&&this._setOptionsFromValues(t||[])},e.prototype.setDisabledState=function(t){this.options&&this.options.forEach(function(e){return e.disabled=t})},e.prototype.registerOnChange=function(t){this._onChange=t},e.prototype.registerOnTouched=function(t){this.onTouched=t},e.prototype._getOptionByValue=function(t){return this.options.find(function(e){return e.value===t})},e.prototype._setOptionsFromValues=function(t){var e=this;this.options.forEach(function(t){return t._setSelected(!1)}),t.map(function(t){return e._getOptionByValue(t)}).filter(Boolean).forEach(function(t){return t._setSelected(!0)})},e.prototype._getSelectedOptionValues=function(){return this.options.filter(function(t){return t.selected}).map(function(t){return t.value})},e.prototype._toggleSelectOnFocusedOption=function(){var t=this._keyManager.activeItemIndex;if(null!=t&&this._isValidIndex(t)){var e=this.options.toArray()[t];e&&(e.toggle(),this._emitChangeEvent(e),e._emitDeprecatedChangeEvent())}},e.prototype._isValidIndex=function(t){return t>=0&&t<this.options.length},e.prototype._getOptionIndex=function(t){return this.options.toArray().indexOf(t)},e.decorators=[{type:i.m,args:[{selector:"mat-selection-list",exportAs:"matSelectionList",inputs:["disabled","disableRipple","tabIndex"],host:{role:"listbox","[tabIndex]":"tabIndex",class:"mat-selection-list","(focus)":"focus()","(blur)":"onTouched()","(keydown)":"_keydown($event)","[attr.aria-disabled]":"disabled.toString()"},template:"<ng-content></ng-content>",styles:[".mat-subheader{display:flex;box-sizing:border-box;padding:16px;align-items:center}.mat-list .mat-subheader,.mat-nav-list .mat-subheader,.mat-selection-list .mat-subheader{margin:0}.mat-list,.mat-nav-list,.mat-selection-list{padding-top:8px;display:block}.mat-list .mat-subheader,.mat-nav-list .mat-subheader,.mat-selection-list .mat-subheader{height:48px;line-height:16px}.mat-list .mat-subheader:first-child,.mat-nav-list .mat-subheader:first-child,.mat-selection-list .mat-subheader:first-child{margin-top:-8px}.mat-list .mat-list-item,.mat-list .mat-list-option,.mat-nav-list .mat-list-item,.mat-nav-list .mat-list-option,.mat-selection-list .mat-list-item,.mat-selection-list .mat-list-option{display:block;height:48px}.mat-list .mat-list-item .mat-list-item-content,.mat-list .mat-list-option .mat-list-item-content,.mat-nav-list .mat-list-item .mat-list-item-content,.mat-nav-list .mat-list-option .mat-list-item-content,.mat-selection-list .mat-list-item .mat-list-item-content,.mat-selection-list .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list .mat-list-item .mat-list-item-content-reverse,.mat-list .mat-list-option .mat-list-item-content-reverse,.mat-nav-list .mat-list-item .mat-list-item-content-reverse,.mat-nav-list .mat-list-option .mat-list-item-content-reverse,.mat-selection-list .mat-list-item .mat-list-item-content-reverse,.mat-selection-list .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list .mat-list-item .mat-list-item-ripple,.mat-list .mat-list-option .mat-list-item-ripple,.mat-nav-list .mat-list-item .mat-list-item-ripple,.mat-nav-list .mat-list-option .mat-list-item-ripple,.mat-selection-list .mat-list-item .mat-list-item-ripple,.mat-selection-list .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list .mat-list-item.mat-list-item-avatar,.mat-list .mat-list-option.mat-list-item-avatar,.mat-nav-list .mat-list-item.mat-list-item-avatar,.mat-nav-list .mat-list-option.mat-list-item-avatar,.mat-selection-list .mat-list-item.mat-list-item-avatar,.mat-selection-list .mat-list-option.mat-list-item-avatar{height:56px}.mat-list .mat-list-item.mat-2-line,.mat-list .mat-list-option.mat-2-line,.mat-nav-list .mat-list-item.mat-2-line,.mat-nav-list .mat-list-option.mat-2-line,.mat-selection-list .mat-list-item.mat-2-line,.mat-selection-list .mat-list-option.mat-2-line{height:72px}.mat-list .mat-list-item.mat-3-line,.mat-list .mat-list-option.mat-3-line,.mat-nav-list .mat-list-item.mat-3-line,.mat-nav-list .mat-list-option.mat-3-line,.mat-selection-list .mat-list-item.mat-3-line,.mat-selection-list .mat-list-option.mat-3-line{height:88px}.mat-list .mat-list-item.mat-multi-line,.mat-list .mat-list-option.mat-multi-line,.mat-nav-list .mat-list-item.mat-multi-line,.mat-nav-list .mat-list-option.mat-multi-line,.mat-selection-list .mat-list-item.mat-multi-line,.mat-selection-list .mat-list-option.mat-multi-line{height:auto}.mat-list .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list .mat-list-option.mat-multi-line .mat-list-item-content,.mat-nav-list .mat-list-item.mat-multi-line .mat-list-item-content,.mat-nav-list .mat-list-option.mat-multi-line .mat-list-item-content,.mat-selection-list .mat-list-item.mat-multi-line .mat-list-item-content,.mat-selection-list .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list .mat-list-item .mat-list-text,.mat-list .mat-list-option .mat-list-text,.mat-nav-list .mat-list-item .mat-list-text,.mat-nav-list .mat-list-option .mat-list-text,.mat-selection-list .mat-list-item .mat-list-text,.mat-selection-list .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0 16px}.mat-list .mat-list-item .mat-list-text>*,.mat-list .mat-list-option .mat-list-text>*,.mat-nav-list .mat-list-item .mat-list-text>*,.mat-nav-list .mat-list-option .mat-list-text>*,.mat-selection-list .mat-list-item .mat-list-text>*,.mat-selection-list .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list .mat-list-item .mat-list-text:empty,.mat-list .mat-list-option .mat-list-text:empty,.mat-nav-list .mat-list-item .mat-list-text:empty,.mat-nav-list .mat-list-option .mat-list-text:empty,.mat-selection-list .mat-list-item .mat-list-text:empty,.mat-selection-list .mat-list-option .mat-list-text:empty{display:none}.mat-list .mat-list-item .mat-list-text:nth-child(2),.mat-list .mat-list-option .mat-list-text:nth-child(2),.mat-nav-list .mat-list-item .mat-list-text:nth-child(2),.mat-nav-list .mat-list-option .mat-list-text:nth-child(2),.mat-selection-list .mat-list-item .mat-list-text:nth-child(2),.mat-selection-list .mat-list-option .mat-list-text:nth-child(2){padding:0}.mat-list .mat-list-item .mat-list-avatar,.mat-list .mat-list-option .mat-list-avatar,.mat-nav-list .mat-list-item .mat-list-avatar,.mat-nav-list .mat-list-option .mat-list-avatar,.mat-selection-list .mat-list-item .mat-list-avatar,.mat-selection-list .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%}.mat-list .mat-list-item .mat-list-icon,.mat-list .mat-list-option .mat-list-icon,.mat-nav-list .mat-list-item .mat-list-icon,.mat-nav-list .mat-list-option .mat-list-icon,.mat-selection-list .mat-list-item .mat-list-icon,.mat-selection-list .mat-list-option .mat-list-icon{width:24px;height:24px;font-size:24px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-list[dense],.mat-nav-list[dense],.mat-selection-list[dense]{padding-top:4px;display:block}.mat-list[dense] .mat-subheader,.mat-nav-list[dense] .mat-subheader,.mat-selection-list[dense] .mat-subheader{height:40px;line-height:8px}.mat-list[dense] .mat-subheader:first-child,.mat-nav-list[dense] .mat-subheader:first-child,.mat-selection-list[dense] .mat-subheader:first-child{margin-top:-4px}.mat-list[dense] .mat-list-item,.mat-list[dense] .mat-list-option,.mat-nav-list[dense] .mat-list-item,.mat-nav-list[dense] .mat-list-option,.mat-selection-list[dense] .mat-list-item,.mat-selection-list[dense] .mat-list-option{display:block;height:40px}.mat-list[dense] .mat-list-item .mat-list-item-content,.mat-list[dense] .mat-list-option .mat-list-item-content,.mat-nav-list[dense] .mat-list-item .mat-list-item-content,.mat-nav-list[dense] .mat-list-option .mat-list-item-content,.mat-selection-list[dense] .mat-list-item .mat-list-item-content,.mat-selection-list[dense] .mat-list-option .mat-list-item-content{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;padding:0 16px;position:relative;height:inherit}.mat-list[dense] .mat-list-item .mat-list-item-content-reverse,.mat-list[dense] .mat-list-option .mat-list-item-content-reverse,.mat-nav-list[dense] .mat-list-item .mat-list-item-content-reverse,.mat-nav-list[dense] .mat-list-option .mat-list-item-content-reverse,.mat-selection-list[dense] .mat-list-item .mat-list-item-content-reverse,.mat-selection-list[dense] .mat-list-option .mat-list-item-content-reverse{display:flex;align-items:center;padding:0 16px;flex-direction:row-reverse;justify-content:space-around}.mat-list[dense] .mat-list-item .mat-list-item-ripple,.mat-list[dense] .mat-list-option .mat-list-item-ripple,.mat-nav-list[dense] .mat-list-item .mat-list-item-ripple,.mat-nav-list[dense] .mat-list-option .mat-list-item-ripple,.mat-selection-list[dense] .mat-list-item .mat-list-item-ripple,.mat-selection-list[dense] .mat-list-option .mat-list-item-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-list[dense] .mat-list-item.mat-list-item-avatar,.mat-list[dense] .mat-list-option.mat-list-item-avatar,.mat-nav-list[dense] .mat-list-item.mat-list-item-avatar,.mat-nav-list[dense] .mat-list-option.mat-list-item-avatar,.mat-selection-list[dense] .mat-list-item.mat-list-item-avatar,.mat-selection-list[dense] .mat-list-option.mat-list-item-avatar{height:48px}.mat-list[dense] .mat-list-item.mat-2-line,.mat-list[dense] .mat-list-option.mat-2-line,.mat-nav-list[dense] .mat-list-item.mat-2-line,.mat-nav-list[dense] .mat-list-option.mat-2-line,.mat-selection-list[dense] .mat-list-item.mat-2-line,.mat-selection-list[dense] .mat-list-option.mat-2-line{height:60px}.mat-list[dense] .mat-list-item.mat-3-line,.mat-list[dense] .mat-list-option.mat-3-line,.mat-nav-list[dense] .mat-list-item.mat-3-line,.mat-nav-list[dense] .mat-list-option.mat-3-line,.mat-selection-list[dense] .mat-list-item.mat-3-line,.mat-selection-list[dense] .mat-list-option.mat-3-line{height:76px}.mat-list[dense] .mat-list-item.mat-multi-line,.mat-list[dense] .mat-list-option.mat-multi-line,.mat-nav-list[dense] .mat-list-item.mat-multi-line,.mat-nav-list[dense] .mat-list-option.mat-multi-line,.mat-selection-list[dense] .mat-list-item.mat-multi-line,.mat-selection-list[dense] .mat-list-option.mat-multi-line{height:auto}.mat-list[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-list[dense] .mat-list-option.mat-multi-line .mat-list-item-content,.mat-nav-list[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-nav-list[dense] .mat-list-option.mat-multi-line .mat-list-item-content,.mat-selection-list[dense] .mat-list-item.mat-multi-line .mat-list-item-content,.mat-selection-list[dense] .mat-list-option.mat-multi-line .mat-list-item-content{padding-top:16px;padding-bottom:16px}.mat-list[dense] .mat-list-item .mat-list-text,.mat-list[dense] .mat-list-option .mat-list-text,.mat-nav-list[dense] .mat-list-item .mat-list-text,.mat-nav-list[dense] .mat-list-option .mat-list-text,.mat-selection-list[dense] .mat-list-item .mat-list-text,.mat-selection-list[dense] .mat-list-option .mat-list-text{display:flex;flex-direction:column;width:100%;box-sizing:border-box;overflow:hidden;padding:0 16px}.mat-list[dense] .mat-list-item .mat-list-text>*,.mat-list[dense] .mat-list-option .mat-list-text>*,.mat-nav-list[dense] .mat-list-item .mat-list-text>*,.mat-nav-list[dense] .mat-list-option .mat-list-text>*,.mat-selection-list[dense] .mat-list-item .mat-list-text>*,.mat-selection-list[dense] .mat-list-option .mat-list-text>*{margin:0;padding:0;font-weight:400;font-size:inherit}.mat-list[dense] .mat-list-item .mat-list-text:empty,.mat-list[dense] .mat-list-option .mat-list-text:empty,.mat-nav-list[dense] .mat-list-item .mat-list-text:empty,.mat-nav-list[dense] .mat-list-option .mat-list-text:empty,.mat-selection-list[dense] .mat-list-item .mat-list-text:empty,.mat-selection-list[dense] .mat-list-option .mat-list-text:empty{display:none}.mat-list[dense] .mat-list-item .mat-list-text:nth-child(2),.mat-list[dense] .mat-list-option .mat-list-text:nth-child(2),.mat-nav-list[dense] .mat-list-item .mat-list-text:nth-child(2),.mat-nav-list[dense] .mat-list-option .mat-list-text:nth-child(2),.mat-selection-list[dense] .mat-list-item .mat-list-text:nth-child(2),.mat-selection-list[dense] .mat-list-option .mat-list-text:nth-child(2){padding:0}.mat-list[dense] .mat-list-item .mat-list-avatar,.mat-list[dense] .mat-list-option .mat-list-avatar,.mat-nav-list[dense] .mat-list-item .mat-list-avatar,.mat-nav-list[dense] .mat-list-option .mat-list-avatar,.mat-selection-list[dense] .mat-list-item .mat-list-avatar,.mat-selection-list[dense] .mat-list-option .mat-list-avatar{flex-shrink:0;width:40px;height:40px;border-radius:50%}.mat-list[dense] .mat-list-item .mat-list-icon,.mat-list[dense] .mat-list-option .mat-list-icon,.mat-nav-list[dense] .mat-list-item .mat-list-icon,.mat-nav-list[dense] .mat-list-option .mat-list-icon,.mat-selection-list[dense] .mat-list-item .mat-list-icon,.mat-selection-list[dense] .mat-list-option .mat-list-icon{width:20px;height:20px;font-size:20px;box-sizing:content-box;border-radius:50%;padding:4px}.mat-divider{display:block;border-top-style:solid;border-top-width:1px;margin:0}.mat-nav-list a{text-decoration:none;color:inherit}.mat-nav-list .mat-list-item-content{cursor:pointer}.mat-nav-list .mat-list-item-content.mat-list-item-focus,.mat-nav-list .mat-list-item-content:hover{outline:0}.mat-list-option:not([disabled]){cursor:pointer}"],encapsulation:i._10.None,providers:[S],preserveWhitespaces:!1,changeDetection:i.i.OnPush}]}],e.ctorParameters=function(){return[{type:i.t},{type:void 0,decorators:[{type:i.g,args:["tabindex"]}]}]},e.propDecorators={options:[{type:i.r,args:[A]}],selectionChange:[{type:i.N}]},e}(C),D=function(){function t(){}return t.decorators=[{type:i.H,args:[{imports:[o.l,o.s,o.j,o.q,r.a],exports:[y,v,w,g,_,o.l,o.j,j,b,x,o.q,T,A],declarations:[y,v,w,g,_,j,b,x,T,A]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/material.es5.js":function(t,e,n){"use strict";var r=n("../../../core/esm5/core.js"),i=n("../../../material/esm5/autocomplete.es5.js"),o=n("../../../material/esm5/button.es5.js"),s=(n("../../../material/esm5/button-toggle.es5.js"),n("../../../material/esm5/card.es5.js")),a=n("../../../material/esm5/checkbox.es5.js"),c=(n("../../../material/esm5/chips.es5.js"),n("../../../material/esm5/core.es5.js"),n("../../../material/esm5/datepicker.es5.js"),n("../../../material/esm5/dialog.es5.js")),u=n("../../../material/esm5/expansion.es5.js"),l=n("../../../material/esm5/form-field.es5.js"),p=n("../../../material/esm5/grid-list.es5.js"),d=n("../../../material/esm5/icon.es5.js"),h=n("../../../material/esm5/input.es5.js"),f=n("../../../material/esm5/list.es5.js"),m=n("../../../material/esm5/menu.es5.js"),g=n("../../../material/esm5/paginator.es5.js"),v=n("../../../material/esm5/progress-bar.es5.js"),y=(n("../../../material/esm5/progress-spinner.es5.js"),n("../../../material/esm5/radio.es5.js"),n("../../../material/esm5/select.es5.js")),b=n("../../../material/esm5/sidenav.es5.js"),_=(n("../../../material/esm5/slide-toggle.es5.js"),n("../../../material/esm5/slider.es5.js"),n("../../../material/esm5/snack-bar.es5.js"),n("../../../material/esm5/sort.es5.js")),j=(n("../../../material/esm5/stepper.es5.js"),n("../../../material/esm5/table.es5.js")),x=(n("../../../material/esm5/tabs.es5.js"),n("../../../material/esm5/toolbar.es5.js")),w=n("../../../material/esm5/tooltip.es5.js");n.d(e,"a",function(){return i.a}),n.d(e,"b",function(){return o.a}),n.d(e,"c",function(){return s.a}),n.d(e,"d",function(){return a.a}),n.d(e,"e",function(){return c.b}),n.d(e,"f",function(){return u.a}),n.d(e,"g",function(){return l.c}),n.d(e,"h",function(){return p.a}),n.d(e,"i",function(){return d.a}),n.d(e,"j",function(){return h.b}),n.d(e,"k",function(){return f.a}),n.d(e,"l",function(){return m.a}),n.d(e,"m",function(){return g.a}),n.d(e,"n",function(){return v.a}),n.d(e,"o",function(){return y.a}),n.d(e,"p",function(){return b.a}),n.d(e,"q",function(){return _.a}),n.d(e,"r",function(){return j.a}),n.d(e,"s",function(){return x.a}),n.d(e,"t",function(){return w.a});new r._6("5.0.0")},"../../../material/esm5/menu.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return T});var r=n("../../../core/esm5/core.js"),i=n("../../../common/esm5/common.js"),o=n("../../../material/esm5/core.es5.js"),s=n("../../../cdk/esm5/overlay.es5.js"),a=n("../../../cdk/esm5/a11y.es5.js"),c=n("../../../cdk/esm5/keycodes.es5.js"),u=n("../../../../rxjs/_esm5/operators/startWith.js"),l=n("../../../../rxjs/_esm5/operators/switchMap.js"),p=n("../../../../rxjs/_esm5/operators/take.js"),d=n("../../../../rxjs/_esm5/observable/merge.js"),h=n("../../../../rxjs/_esm5/Subscription.js"),f=n("../../../animations/esm5/animations.js"),m=n("../../../../tslib/tslib.es6.js"),g=n("../../../../rxjs/_esm5/Subject.js"),v=n("../../../cdk/esm5/coercion.es5.js"),y=n("../../../cdk/esm5/bidi.es5.js"),b=n("../../../cdk/esm5/portal.es5.js"),_=n("../../../../rxjs/_esm5/operators/filter.js"),j=n("../../../../rxjs/_esm5/observable/of.js"),x=Object(f.k)("transformMenu",[Object(f.h)("void",Object(f.i)({opacity:0,transform:"scale(0.01, 0.01)"})),Object(f.h)("enter-start",Object(f.i)({opacity:1,transform:"scale(1, 0.5)"})),Object(f.h)("enter",Object(f.i)({transform:"scale(1, 1)"})),Object(f.j)("void => enter-start",Object(f.e)("100ms linear")),Object(f.j)("enter-start => enter",Object(f.e)("300ms cubic-bezier(0.25, 0.8, 0.25, 1)")),Object(f.j)("* => void",Object(f.e)("150ms 50ms linear",Object(f.i)({opacity:0})))]),w=Object(f.k)("fadeInItems",[Object(f.h)("showing",Object(f.i)({opacity:1})),Object(f.j)("void => *",[Object(f.i)({opacity:0}),Object(f.e)("400ms 100ms cubic-bezier(0.55, 0, 0.55, 0.2)")])]);var O=function(){return function(){}}(),C=function(t){function e(e){var n=t.call(this)||this;return n._elementRef=e,n._hovered=new g.a,n._highlighted=!1,n._triggersSubmenu=!1,n}return Object(m.b)(e,t),e.prototype.focus=function(){this._getHostElement().focus()},e.prototype.ngOnDestroy=function(){this._hovered.complete()},e.prototype._getTabIndex=function(){return this.disabled?"-1":"0"},e.prototype._getHostElement=function(){return this._elementRef.nativeElement},e.prototype._checkDisabled=function(t){this.disabled&&(t.preventDefault(),t.stopPropagation())},e.prototype._emitHoverEvent=function(){this.disabled||this._hovered.next(this)},e.prototype.getLabel=function(){var t=this._elementRef.nativeElement,e="";if(t.childNodes)for(var n=t.childNodes.length,r=0;r<n;r++)t.childNodes[r].nodeType===Node.TEXT_NODE&&(e+=t.childNodes[r].textContent);return e.trim()},e.decorators=[{type:r.m,args:[{selector:"[mat-menu-item]",exportAs:"matMenuItem",inputs:["disabled","disableRipple"],host:{role:"menuitem",class:"mat-menu-item","[class.mat-menu-item-highlighted]":"_highlighted","[class.mat-menu-item-submenu-trigger]":"_triggersSubmenu","[attr.tabindex]":"_getTabIndex()","[attr.aria-disabled]":"disabled.toString()","[attr.disabled]":"disabled || null","(click)":"_checkDisabled($event)","(mouseenter)":"_emitHoverEvent()"},changeDetection:r.i.OnPush,encapsulation:r._10.None,preserveWhitespaces:!1,template:'<ng-content></ng-content><div class="mat-menu-ripple" matRipple [matRippleDisabled]="disableRipple || disabled" [matRippleTrigger]="_getHostElement()"></div>'}]}],e.ctorParameters=function(){return[{type:r.t}]},e}(Object(o.v)(Object(o.w)(O))),k=new r.A("mat-menu-default-options"),E=function(){function t(t,e,n){this._elementRef=t,this._ngZone=e,this._defaultOptions=n,this._xPosition=this._defaultOptions.xPosition,this._yPosition=this._defaultOptions.yPosition,this._tabSubscription=h.a.EMPTY,this._classList={},this._panelAnimationState="void",this._overlapTrigger=this._defaultOptions.overlapTrigger,this.closed=new r.w,this.close=this.closed}return Object.defineProperty(t.prototype,"xPosition",{get:function(){return this._xPosition},set:function(t){"before"!==t&&"after"!==t&&function(){throw Error('x-position value must be either \'before\' or after\'.\n      Example: <mat-menu x-position="before" #menu="matMenu"></mat-menu>')}(),this._xPosition=t,this.setPositionClasses()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"yPosition",{get:function(){return this._yPosition},set:function(t){"above"!==t&&"below"!==t&&function(){throw Error('y-position value must be either \'above\' or below\'.\n      Example: <mat-menu y-position="above" #menu="matMenu"></mat-menu>')}(),this._yPosition=t,this.setPositionClasses()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"overlapTrigger",{get:function(){return this._overlapTrigger},set:function(t){this._overlapTrigger=Object(v.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"panelClass",{set:function(t){t&&t.length&&(this._classList=t.split(" ").reduce(function(t,e){return t[e]=!0,t},{}),this._elementRef.nativeElement.className="",this.setPositionClasses())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"classList",{get:function(){return this.panelClass},set:function(t){this.panelClass=t},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){var t=this;this._keyManager=new a.e(this.items).withWrap().withTypeAhead(),this._tabSubscription=this._keyManager.tabOut.subscribe(function(){return t.close.emit("keydown")})},t.prototype.ngOnDestroy=function(){this._tabSubscription.unsubscribe(),this.closed.complete()},t.prototype._hovered=function(){var t=this;return this.items?this.items.changes.pipe(Object(u.a)(this.items),Object(l.a)(function(t){return d.a.apply(void 0,t.map(function(t){return t._hovered}))})):this._ngZone.onStable.asObservable().pipe(Object(p.a)(1),Object(l.a)(function(){return t._hovered()}))},t.prototype._handleKeydown=function(t){switch(t.keyCode){case c.g:this.closed.emit("keydown"),t.stopPropagation();break;case c.i:this.parentMenu&&"ltr"===this.direction&&this.closed.emit("keydown");break;case c.m:this.parentMenu&&"rtl"===this.direction&&this.closed.emit("keydown");break;default:this._keyManager.onKeydown(t)}},t.prototype.focusFirstItem=function(){this._keyManager.setFirstItemActive()},t.prototype.resetActiveItem=function(){this._keyManager.setActiveItem(-1)},t.prototype.setPositionClasses=function(t,e){void 0===t&&(t=this.xPosition),void 0===e&&(e=this.yPosition),this._classList["mat-menu-before"]="before"===t,this._classList["mat-menu-after"]="after"===t,this._classList["mat-menu-above"]="above"===e,this._classList["mat-menu-below"]="below"===e},t.prototype.setElevation=function(t){var e="mat-elevation-z"+(2+t),n=Object.keys(this._classList).find(function(t){return t.startsWith("mat-elevation-z")});n&&n!==this._previousElevation||(this._previousElevation&&(this._classList[this._previousElevation]=!1),this._classList[e]=!0,this._previousElevation=e)},t.prototype._startAnimation=function(){this._panelAnimationState="enter-start"},t.prototype._resetAnimation=function(){this._panelAnimationState="void"},t.prototype._onAnimationDone=function(t){"enter-start"===t.toState&&(this._panelAnimationState="enter")},t.decorators=[{type:r.m,args:[{selector:"mat-menu",template:'<ng-template><div class="mat-menu-panel" [ngClass]="_classList" (keydown)="_handleKeydown($event)" (click)="closed.emit(\'click\')" [@transformMenu]="_panelAnimationState" (@transformMenu.done)="_onAnimationDone($event)" tabindex="-1" role="menu"><div class="mat-menu-content" [@fadeInItems]="\'showing\'"><ng-content></ng-content></div></div></ng-template>',styles:[".mat-menu-panel{min-width:112px;max-width:280px;overflow:auto;-webkit-overflow-scrolling:touch;max-height:calc(100vh - 48px);border-radius:2px;outline:0}.mat-menu-panel:not([class*=mat-elevation-z]){box-shadow:0 3px 1px -2px rgba(0,0,0,.2),0 2px 2px 0 rgba(0,0,0,.14),0 1px 5px 0 rgba(0,0,0,.12)}.mat-menu-panel.mat-menu-after.mat-menu-below{transform-origin:left top}.mat-menu-panel.mat-menu-after.mat-menu-above{transform-origin:left bottom}.mat-menu-panel.mat-menu-before.mat-menu-below{transform-origin:right top}.mat-menu-panel.mat-menu-before.mat-menu-above{transform-origin:right bottom}[dir=rtl] .mat-menu-panel.mat-menu-after.mat-menu-below{transform-origin:right top}[dir=rtl] .mat-menu-panel.mat-menu-after.mat-menu-above{transform-origin:right bottom}[dir=rtl] .mat-menu-panel.mat-menu-before.mat-menu-below{transform-origin:left top}[dir=rtl] .mat-menu-panel.mat-menu-before.mat-menu-above{transform-origin:left bottom}.mat-menu-panel.ng-animating{pointer-events:none}@media screen and (-ms-high-contrast:active){.mat-menu-panel{outline:solid 1px}}.mat-menu-content{padding-top:8px;padding-bottom:8px}.mat-menu-item{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;display:block;line-height:48px;height:48px;padding:0 16px;text-align:left;text-decoration:none;position:relative}.mat-menu-item[disabled]{cursor:default}[dir=rtl] .mat-menu-item{text-align:right}.mat-menu-item .mat-icon{margin-right:16px}[dir=rtl] .mat-menu-item .mat-icon{margin-left:16px;margin-right:0}.mat-menu-item .mat-icon{vertical-align:middle}.mat-menu-item-submenu-trigger{padding-right:32px}.mat-menu-item-submenu-trigger::after{width:0;height:0;border-style:solid;border-width:5px 0 5px 5px;border-color:transparent transparent transparent currentColor;content:'';display:inline-block;position:absolute;top:50%;right:16px;transform:translateY(-50%)}[dir=rtl] .mat-menu-item-submenu-trigger{padding-right:8px;padding-left:32px}[dir=rtl] .mat-menu-item-submenu-trigger::after{right:auto;left:16px;transform:rotateY(180deg) translateY(-50%)}button.mat-menu-item{width:100%}.mat-menu-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}"],changeDetection:r.i.OnPush,encapsulation:r._10.None,preserveWhitespaces:!1,animations:[x,w],exportAs:"matMenu"}]}],t.ctorParameters=function(){return[{type:r.t},{type:r.L},{type:void 0,decorators:[{type:r.y,args:[k]}]}]},t.propDecorators={xPosition:[{type:r.C}],yPosition:[{type:r.C}],templateRef:[{type:r._7,args:[r._4]}],items:[{type:r.r,args:[C]}],overlapTrigger:[{type:r.C}],panelClass:[{type:r.C,args:["class"]}],classList:[{type:r.C}],closed:[{type:r.N}],close:[{type:r.N}]},t}(),S=new r.A("mat-menu-scroll-strategy");var P={provide:S,deps:[s.b],useFactory:function(t){return function(){return t.scrollStrategies.reposition()}}},M=function(){function t(t,e,n,i,o,s,a){this._overlay=t,this._element=e,this._viewContainerRef=n,this._scrollStrategy=i,this._parentMenu=o,this._menuItemInstance=s,this._dir=a,this._overlayRef=null,this._menuOpen=!1,this._closeSubscription=h.a.EMPTY,this._positionSubscription=h.a.EMPTY,this._hoverSubscription=h.a.EMPTY,this._openedByMouse=!1,this.menuOpened=new r.w,this.onMenuOpen=this.menuOpened,this.menuClosed=new r.w,this.onMenuClose=this.menuClosed,s&&(s._triggersSubmenu=this.triggersSubmenu())}return Object.defineProperty(t.prototype,"_deprecatedMatMenuTriggerFor",{get:function(){return this.menu},set:function(t){this.menu=t},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){var t=this;this._checkMenu(),this.menu.close.subscribe(function(e){t._destroyMenu(),"click"===e&&t._parentMenu&&t._parentMenu.closed.emit(e)}),this.triggersSubmenu()&&(this._hoverSubscription=this._parentMenu._hovered().pipe(Object(_.a)(function(e){return e===t._menuItemInstance})).subscribe(function(){t._openedByMouse=!0,t.openMenu()}))},t.prototype.ngOnDestroy=function(){this._overlayRef&&(this._overlayRef.dispose(),this._overlayRef=null),this._cleanUpSubscriptions()},Object.defineProperty(t.prototype,"menuOpen",{get:function(){return this._menuOpen},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dir",{get:function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},enumerable:!0,configurable:!0}),t.prototype.triggersSubmenu=function(){return!(!this._menuItemInstance||!this._parentMenu)},t.prototype.toggleMenu=function(){return this._menuOpen?this.closeMenu():this.openMenu()},t.prototype.openMenu=function(){var t=this;this._menuOpen||(this._createOverlay().attach(this._portal),this._closeSubscription=this._menuClosingActions().subscribe(function(){return t.closeMenu()}),this._initMenu(),this.menu instanceof E&&this.menu._startAnimation())},t.prototype.closeMenu=function(){this.menu.close.emit()},t.prototype.focus=function(){this._element.nativeElement.focus()},t.prototype._destroyMenu=function(){this._overlayRef&&this.menuOpen&&(this._resetMenu(),this._closeSubscription.unsubscribe(),this._overlayRef.detach(),this.menu instanceof E&&this.menu._resetAnimation())},t.prototype._initMenu=function(){if(this.menu.parentMenu=this.triggersSubmenu()?this._parentMenu:void 0,this.menu.direction=this.dir,this._setMenuElevation(),this._setIsMenuOpen(!0),this._openedByMouse){var t=this._overlayRef.overlayElement.firstElementChild;t&&(this.menu.resetActiveItem(),t.focus())}else this.menu.focusFirstItem()},t.prototype._setMenuElevation=function(){if(this.menu.setElevation){for(var t=0,e=this.menu.parentMenu;e;)t++,e=e.parentMenu;this.menu.setElevation(t)}},t.prototype._resetMenu=function(){this._setIsMenuOpen(!1),this._openedByMouse&&this.triggersSubmenu()||this.focus(),this._openedByMouse=!1},t.prototype._setIsMenuOpen=function(t){this._menuOpen=t,this._menuOpen?this.menuOpened.emit():this.menuClosed.emit(),this.triggersSubmenu()&&(this._menuItemInstance._highlighted=t)},t.prototype._checkMenu=function(){this.menu||function(){throw Error('mat-menu-trigger: must pass in an mat-menu instance.\n\n    Example:\n      <mat-menu #menu="matMenu"></mat-menu>\n      <button [matMenuTriggerFor]="menu"></button>')}()},t.prototype._createOverlay=function(){if(!this._overlayRef){this._portal=new b.h(this.menu.templateRef,this._viewContainerRef);var t=this._getOverlayConfig();this._subscribeToPositions(t.positionStrategy),this._overlayRef=this._overlay.create(t)}return this._overlayRef},t.prototype._getOverlayConfig=function(){return new s.c({positionStrategy:this._getPosition(),hasBackdrop:!this.triggersSubmenu(),backdropClass:"cdk-overlay-transparent-backdrop",direction:this.dir,scrollStrategy:this._scrollStrategy()})},t.prototype._subscribeToPositions=function(t){var e=this;this._positionSubscription=t.onPositionChange.subscribe(function(t){var n="start"===t.connectionPair.overlayX?"after":"before",r="top"===t.connectionPair.overlayY?"below":"above";e.menu.setPositionClasses(n,r)})},t.prototype._getPosition=function(){var t="before"===this.menu.xPosition?["end","start"]:["start","end"],e=t[0],n=t[1],r="above"===this.menu.yPosition?["bottom","top"]:["top","bottom"],i=r[0],o=r[1],s=[i,o],a=s[0],c=s[1],u=[e,n],l=u[0],p=u[1],d=0;return this.triggersSubmenu()?(p=e="before"===this.menu.xPosition?"start":"end",n=l="end"===e?"start":"end",d="bottom"===i?8:-8):this.menu.overlapTrigger||(a="top"===i?"bottom":"top",c="top"===o?"bottom":"top"),this._overlay.position().connectedTo(this._element,{originX:e,originY:a},{overlayX:l,overlayY:i}).withDirection(this.dir).withOffsetY(d).withFallbackPosition({originX:n,originY:a},{overlayX:p,overlayY:i}).withFallbackPosition({originX:e,originY:c},{overlayX:l,overlayY:o},void 0,-d).withFallbackPosition({originX:n,originY:c},{overlayX:p,overlayY:o},void 0,-d)},t.prototype._cleanUpSubscriptions=function(){this._closeSubscription.unsubscribe(),this._positionSubscription.unsubscribe(),this._hoverSubscription.unsubscribe()},t.prototype._menuClosingActions=function(){var t=this,e=this._overlayRef.backdropClick(),n=this._overlayRef.detachments(),r=this._parentMenu?this._parentMenu.close:Object(j.a)(),i=this._parentMenu?this._parentMenu._hovered().pipe(Object(_.a)(function(e){return e!==t._menuItemInstance}),Object(_.a)(function(){return t._menuOpen})):Object(j.a)();return Object(d.a)(e,r,i,n)},t.prototype._handleMousedown=function(t){Object(a.j)(t)||(this._openedByMouse=!0,this.triggersSubmenu()&&t.preventDefault())},t.prototype._handleKeydown=function(t){var e=t.keyCode;this.triggersSubmenu()&&(e===c.m&&"ltr"===this.dir||e===c.i&&"rtl"===this.dir)&&this.openMenu()},t.prototype._handleClick=function(t){this.triggersSubmenu()?(t.stopPropagation(),this.openMenu()):this.toggleMenu()},t.decorators=[{type:r.s,args:[{selector:"[mat-menu-trigger-for], [matMenuTriggerFor]",host:{"aria-haspopup":"true","(mousedown)":"_handleMousedown($event)","(keydown)":"_handleKeydown($event)","(click)":"_handleClick($event)"},exportAs:"matMenuTrigger"}]}],t.ctorParameters=function(){return[{type:s.b},{type:r.t},{type:r._9},{type:void 0,decorators:[{type:r.y,args:[S]}]},{type:E,decorators:[{type:r.M}]},{type:C,decorators:[{type:r.M},{type:r.Z}]},{type:y.b,decorators:[{type:r.M}]}]},t.propDecorators={_deprecatedMatMenuTriggerFor:[{type:r.C,args:["mat-menu-trigger-for"]}],menu:[{type:r.C,args:["matMenuTriggerFor"]}],menuOpened:[{type:r.N}],onMenuOpen:[{type:r.N}],menuClosed:[{type:r.N}],onMenuClose:[{type:r.N}]},t}(),A={overlapTrigger:!0,xPosition:"after",yPosition:"below"},T=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[s.d,i.a,o.s,o.j],exports:[E,C,M,o.j],declarations:[E,C,M],providers:[P,{provide:k,useValue:A}]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/paginator.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return d});var r=n("../../../common/esm5/common.js"),i=n("../../../core/esm5/core.js"),o=n("../../../material/esm5/button.es5.js"),s=n("../../../material/esm5/select.es5.js"),a=n("../../../material/esm5/tooltip.es5.js"),c=n("../../../../rxjs/_esm5/Subject.js"),u=function(){function t(){this.changes=new c.a,this.itemsPerPageLabel="Items per page:",this.nextPageLabel="Next page",this.previousPageLabel="Previous page",this.getRangeLabel=function(t,e,n){if(0==n||0==e)return"0 of "+n;var r=t*e;return r+1+" - "+(r<(n=Math.max(n,0))?Math.min(r+e,n):r+e)+" of "+n}}return t.decorators=[{type:i.z}],t.ctorParameters=function(){return[]},t}();var l={provide:u,deps:[[new i.M,new i._1,u]],useFactory:function(t){return t||new u}},p=(function(){}(),function(){function t(t,e){var n=this;this._intl=t,this._changeDetectorRef=e,this._pageIndex=0,this._length=0,this._pageSizeOptions=[],this.page=new i.w,this._intlChanges=t.changes.subscribe(function(){return n._changeDetectorRef.markForCheck()})}return Object.defineProperty(t.prototype,"pageIndex",{get:function(){return this._pageIndex},set:function(t){this._pageIndex=t,this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"length",{get:function(){return this._length},set:function(t){this._length=t,this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pageSize",{get:function(){return this._pageSize},set:function(t){this._pageSize=t,this._updateDisplayedPageSizeOptions()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"pageSizeOptions",{get:function(){return this._pageSizeOptions},set:function(t){this._pageSizeOptions=t,this._updateDisplayedPageSizeOptions()},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){this._initialized=!0,this._updateDisplayedPageSizeOptions()},t.prototype.ngOnDestroy=function(){this._intlChanges.unsubscribe()},t.prototype.nextPage=function(){this.hasNextPage()&&(this.pageIndex++,this._emitPageEvent())},t.prototype.previousPage=function(){this.hasPreviousPage()&&(this.pageIndex--,this._emitPageEvent())},t.prototype.hasPreviousPage=function(){return this.pageIndex>=1&&0!=this.pageSize},t.prototype.hasNextPage=function(){var t=Math.ceil(this.length/this.pageSize)-1;return this.pageIndex<t&&0!=this.pageSize},t.prototype._changePageSize=function(t){var e=this.pageIndex*this.pageSize;this.pageIndex=Math.floor(e/t)||0,this.pageSize=t,this._emitPageEvent()},t.prototype._updateDisplayedPageSizeOptions=function(){this._initialized&&(this.pageSize||(this._pageSize=0!=this.pageSizeOptions.length?this.pageSizeOptions[0]:50),this._displayedPageSizeOptions=this.pageSizeOptions.slice(),-1==this._displayedPageSizeOptions.indexOf(this.pageSize)&&this._displayedPageSizeOptions.push(this.pageSize),this._displayedPageSizeOptions.sort(function(t,e){return t-e}),this._changeDetectorRef.markForCheck())},t.prototype._emitPageEvent=function(){this.page.next({pageIndex:this.pageIndex,pageSize:this.pageSize,length:this.length})},t.decorators=[{type:i.m,args:[{selector:"mat-paginator",exportAs:"matPaginator",template:'<div class="mat-paginator-container"><div class="mat-paginator-page-size"><div class="mat-paginator-page-size-label">{{_intl.itemsPerPageLabel}}</div><mat-form-field *ngIf="_displayedPageSizeOptions.length > 1" class="mat-paginator-page-size-select"><mat-select [value]="pageSize" [aria-label]="_intl.itemsPerPageLabel" (change)="_changePageSize($event.value)"><mat-option *ngFor="let pageSizeOption of _displayedPageSizeOptions" [value]="pageSizeOption">{{pageSizeOption}}</mat-option></mat-select></mat-form-field><div *ngIf="_displayedPageSizeOptions.length <= 1">{{pageSize}}</div></div><div class="mat-paginator-range-actions"><div class="mat-paginator-range-label">{{_intl.getRangeLabel(pageIndex, pageSize, length)}}</div><button mat-icon-button type="button" class="mat-paginator-navigation-previous" (click)="previousPage()" [attr.aria-label]="_intl.previousPageLabel" [matTooltip]="_intl.previousPageLabel" [matTooltipPosition]="\'above\'" [disabled]="!hasPreviousPage()"><div class="mat-paginator-increment"></div></button> <button mat-icon-button type="button" class="mat-paginator-navigation-next" (click)="nextPage()" [attr.aria-label]="_intl.nextPageLabel" [matTooltip]="_intl.nextPageLabel" [matTooltipPosition]="\'above\'" [disabled]="!hasNextPage()"><div class="mat-paginator-decrement"></div></button></div></div>',styles:[".mat-paginator-container{display:flex;align-items:center;justify-content:flex-end;min-height:56px;padding:0 8px;flex-wrap:wrap-reverse}.mat-paginator-page-size{display:flex;align-items:baseline;margin-right:8px}.mat-paginator-page-size-label{margin:0 4px}.mat-paginator-page-size-select{margin:6px 4px 0 4px;width:56px}.mat-paginator-range-label{margin:0 32px 0 24px}.mat-paginator-increment-button+.mat-paginator-increment-button{margin:0 0 0 8px}[dir=rtl] .mat-paginator-increment-button+.mat-paginator-increment-button{margin:0 8px 0 0}.mat-paginator-decrement,.mat-paginator-increment{width:8px;height:8px}.mat-paginator-decrement,[dir=rtl] .mat-paginator-increment{transform:rotate(45deg)}.mat-paginator-increment,[dir=rtl] .mat-paginator-decrement{transform:rotate(225deg)}.mat-paginator-decrement{margin-left:12px}[dir=rtl] .mat-paginator-decrement{margin-right:12px}.mat-paginator-increment{margin-left:16px}[dir=rtl] .mat-paginator-increment{margin-right:16px}.mat-paginator-range-actions{display:flex;align-items:center;min-height:48px}"],host:{class:"mat-paginator"},changeDetection:i.i.OnPush,encapsulation:i._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[{type:u},{type:i.j}]},t.propDecorators={pageIndex:[{type:i.C}],length:[{type:i.C}],pageSize:[{type:i.C}],pageSizeOptions:[{type:i.C}],page:[{type:i.N}]},t}()),d=function(){function t(){}return t.decorators=[{type:i.H,args:[{imports:[r.a,o.a,s.a,a.a],exports:[p],declarations:[p],providers:[l]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/progress-bar.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return c});var r=n("../../../core/esm5/core.js"),i=n("../../../common/esm5/common.js"),o=n("../../../material/esm5/core.es5.js"),s=function(){function t(){this.color="primary",this._value=0,this._bufferValue=0,this.mode="determinate"}return Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(t){this._value=a(t||0)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"bufferValue",{get:function(){return this._bufferValue},set:function(t){this._bufferValue=a(t||0)},enumerable:!0,configurable:!0}),t.prototype._primaryTransform=function(){return{transform:"scaleX("+this.value/100+")"}},t.prototype._bufferTransform=function(){if("buffer"==this.mode)return{transform:"scaleX("+this.bufferValue/100+")"}},t.decorators=[{type:r.m,args:[{selector:"mat-progress-bar",exportAs:"matProgressBar",host:{role:"progressbar","aria-valuemin":"0","aria-valuemax":"100","[attr.aria-valuenow]":"value","[attr.mode]":"mode","[class.mat-primary]":'color == "primary"',"[class.mat-accent]":'color == "accent"',"[class.mat-warn]":'color == "warn"',class:"mat-progress-bar"},template:'<div class="mat-progress-bar-background mat-progress-bar-element"></div><div class="mat-progress-bar-buffer mat-progress-bar-element" [ngStyle]="_bufferTransform()"></div><div class="mat-progress-bar-primary mat-progress-bar-fill mat-progress-bar-element" [ngStyle]="_primaryTransform()"></div><div class="mat-progress-bar-secondary mat-progress-bar-fill mat-progress-bar-element"></div>',styles:[".mat-progress-bar{display:block;height:5px;overflow:hidden;position:relative;transform:translateZ(0);transition:opacity 250ms linear;width:100%}.mat-progress-bar .mat-progress-bar-element,.mat-progress-bar .mat-progress-bar-fill::after{height:100%;position:absolute;width:100%}.mat-progress-bar .mat-progress-bar-background{background-repeat:repeat-x;background-size:10px 4px;display:none}.mat-progress-bar .mat-progress-bar-buffer{transform-origin:top left;transition:transform 250ms ease}.mat-progress-bar .mat-progress-bar-secondary{display:none}.mat-progress-bar .mat-progress-bar-fill{animation:none;transform-origin:top left;transition:transform 250ms ease}.mat-progress-bar .mat-progress-bar-fill::after{animation:none;content:'';display:inline-block;left:0}.mat-progress-bar[dir=rtl],[dir=rtl] .mat-progress-bar{transform:rotateY(180deg)}.mat-progress-bar[mode=query]{transform:rotateZ(180deg)}.mat-progress-bar[mode=query][dir=rtl],[dir=rtl] .mat-progress-bar[mode=query]{transform:rotateZ(180deg) rotateY(180deg)}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-fill,.mat-progress-bar[mode=query] .mat-progress-bar-fill{transition:none}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-primary,.mat-progress-bar[mode=query] .mat-progress-bar-primary{animation:mat-progress-bar-primary-indeterminate-translate 2s infinite linear;left:-145.166611%}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-primary.mat-progress-bar-fill::after,.mat-progress-bar[mode=query] .mat-progress-bar-primary.mat-progress-bar-fill::after{animation:mat-progress-bar-primary-indeterminate-scale 2s infinite linear}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-secondary,.mat-progress-bar[mode=query] .mat-progress-bar-secondary{animation:mat-progress-bar-secondary-indeterminate-translate 2s infinite linear;left:-54.888891%;display:block}.mat-progress-bar[mode=indeterminate] .mat-progress-bar-secondary.mat-progress-bar-fill::after,.mat-progress-bar[mode=query] .mat-progress-bar-secondary.mat-progress-bar-fill::after{animation:mat-progress-bar-secondary-indeterminate-scale 2s infinite linear}.mat-progress-bar[mode=buffer] .mat-progress-bar-background{animation:mat-progress-bar-background-scroll 250ms infinite linear;display:block}@keyframes mat-progress-bar-primary-indeterminate-translate{0%{transform:translateX(0)}20%{animation-timing-function:cubic-bezier(.5,0,.70173,.49582);transform:translateX(0)}59.15%{animation-timing-function:cubic-bezier(.30244,.38135,.55,.95635);transform:translateX(83.67142%)}100%{transform:translateX(200.61106%)}}@keyframes mat-progress-bar-primary-indeterminate-scale{0%{transform:scaleX(.08)}36.65%{animation-timing-function:cubic-bezier(.33473,.12482,.78584,1);transform:scaleX(.08)}69.15%{animation-timing-function:cubic-bezier(.06,.11,.6,1);transform:scaleX(.66148)}100%{transform:scaleX(.08)}}@keyframes mat-progress-bar-secondary-indeterminate-translate{0%{animation-timing-function:cubic-bezier(.15,0,.51506,.40969);transform:translateX(0)}25%{animation-timing-function:cubic-bezier(.31033,.28406,.8,.73371);transform:translateX(37.65191%)}48.35%{animation-timing-function:cubic-bezier(.4,.62704,.6,.90203);transform:translateX(84.38617%)}100%{transform:translateX(160.27778%)}}@keyframes mat-progress-bar-secondary-indeterminate-scale{0%{animation-timing-function:cubic-bezier(.15,0,.51506,.40969);transform:scaleX(.08)}19.15%{animation-timing-function:cubic-bezier(.31033,.28406,.8,.73371);transform:scaleX(.4571)}44.15%{animation-timing-function:cubic-bezier(.4,.62704,.6,.90203);transform:scaleX(.72796)}100%{transform:scaleX(.08)}}@keyframes mat-progress-bar-background-scroll{to{transform:translateX(-10px)}}"],changeDetection:r.i.OnPush,encapsulation:r._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[]},t.propDecorators={color:[{type:r.C}],value:[{type:r.C}],bufferValue:[{type:r.C}],mode:[{type:r.C}]},t}();function a(t,e,n){return void 0===e&&(e=0),void 0===n&&(n=100),Math.max(e,Math.min(n,t))}var c=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.a,o.j],exports:[s,o.j],declarations:[s]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/progress-spinner.es5.js":function(t,e,n){"use strict";var r=n("../../../core/esm5/core.js"),i=n("../../../cdk/esm5/platform.es5.js"),o=n("../../../material/esm5/core.es5.js"),s=n("../../../../tslib/tslib.es6.js"),a=n("../../../common/esm5/common.js"),c=n("../../../cdk/esm5/coercion.es5.js"),u=100,l=function(){return function(t){this._elementRef=t}}(),p=function(t){function e(e,n,r){var i=t.call(this,e)||this;i._elementRef=e,i._document=r,i._value=0,i._fallbackAnimation=!1,i._elementSize=u,i._diameter=u,i.mode="determinate",i._fallbackAnimation=n.EDGE||n.TRIDENT;var o="mat-progress-spinner-indeterminate"+(i._fallbackAnimation?"-fallback":"")+"-animation";return e.nativeElement.classList.add(o),i}return Object(s.b)(e,t),Object.defineProperty(e.prototype,"diameter",{get:function(){return this._diameter},set:function(t){this._diameter=Object(c.c)(t),this._fallbackAnimation||e.diameters.has(this._diameter)||this._attachStyleNode()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"strokeWidth",{get:function(){return this._strokeWidth||this.diameter/10},set:function(t){this._strokeWidth=Object(c.c)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return"determinate"===this.mode?this._value:0},set:function(t){this._value=Math.max(0,Math.min(100,Object(c.c)(t)))},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){(t.strokeWidth||t.diameter)&&(this._elementSize=this._diameter+Math.max(this.strokeWidth-10,0))},Object.defineProperty(e.prototype,"_circleRadius",{get:function(){return(this.diameter-10)/2},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_viewBox",{get:function(){var t=2*this._circleRadius+this.strokeWidth;return"0 0 "+t+" "+t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_strokeCircumference",{get:function(){return 2*Math.PI*this._circleRadius},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_strokeDashOffset",{get:function(){return"determinate"===this.mode?this._strokeCircumference*(100-this._value)/100:this._fallbackAnimation&&"indeterminate"===this.mode?.2*this._strokeCircumference:null},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_circleStrokeWidth",{get:function(){return this.strokeWidth/this._elementSize*100},enumerable:!0,configurable:!0}),e.prototype._attachStyleNode=function(){var t=e.styleTag;t||(t=this._document.createElement("style"),this._document.head.appendChild(t),e.styleTag=t),t&&t.sheet&&t.sheet.insertRule(this._getAnimationText(),0),e.diameters.add(this.diameter)},e.prototype._getAnimationText=function(){return"\n @keyframes mat-progress-spinner-stroke-rotate-DIAMETER {\n    0%      { stroke-dashoffset: START_VALUE;  transform: rotate(0); }\n    12.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(0); }\n    12.51%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(72.5deg); }\n    25%     { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(72.5deg); }\n\n    25.1%   { stroke-dashoffset: START_VALUE;  transform: rotate(270deg); }\n    37.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(270deg); }\n    37.51%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(161.5deg); }\n    50%     { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(161.5deg); }\n\n    50.01%  { stroke-dashoffset: START_VALUE;  transform: rotate(180deg); }\n    62.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(180deg); }\n    62.51%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(251.5deg); }\n    75%     { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(251.5deg); }\n\n    75.01%  { stroke-dashoffset: START_VALUE;  transform: rotate(90deg); }\n    87.5%   { stroke-dashoffset: END_VALUE;    transform: rotate(90deg); }\n    87.51%  { stroke-dashoffset: END_VALUE;    transform: rotateX(180deg) rotate(341.5deg); }\n    100%    { stroke-dashoffset: START_VALUE;  transform: rotateX(180deg) rotate(341.5deg); }\n  }\n".replace(/START_VALUE/g,""+.95*this._strokeCircumference).replace(/END_VALUE/g,""+.2*this._strokeCircumference).replace(/DIAMETER/g,""+this.diameter)},e.diameters=new Set([u]),e.styleTag=null,e.decorators=[{type:r.m,args:[{selector:"mat-progress-spinner",exportAs:"matProgressSpinner",host:{role:"progressbar",class:"mat-progress-spinner","[style.width.px]":"_elementSize","[style.height.px]":"_elementSize","[attr.aria-valuemin]":'mode === "determinate" ? 0 : null',"[attr.aria-valuemax]":'mode === "determinate" ? 100 : null',"[attr.aria-valuenow]":"value","[attr.mode]":"mode"},inputs:["color"],template:'<svg [style.width.px]="_elementSize" [style.height.px]="_elementSize" [attr.viewBox]="_viewBox" preserveAspectRatio="xMidYMid meet" focusable="false"><circle cx="50%" cy="50%" [attr.r]="_circleRadius" [style.animation-name]="\'mat-progress-spinner-stroke-rotate-\' + diameter" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle></svg>',styles:[".mat-progress-spinner{display:block;position:relative}.mat-progress-spinner svg{position:absolute;transform:rotate(-90deg);top:0;left:0;transform-origin:center;overflow:visible}.mat-progress-spinner circle{fill:transparent;transform-origin:center;transition:stroke-dashoffset 225ms linear}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{animation:mat-progress-spinner-linear-rotate 2s linear infinite}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition-property:stroke;animation-duration:4s;animation-timing-function:cubic-bezier(.35,0,.25,1);animation-iteration-count:infinite}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{animation:mat-progress-spinner-stroke-rotate-fallback 10s cubic-bezier(.87,.03,.33,1) infinite}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition-property:stroke}@keyframes mat-progress-spinner-linear-rotate{0%{transform:rotate(0)}100%{transform:rotate(360deg)}}@keyframes mat-progress-spinner-stroke-rotate-100{0%{stroke-dashoffset:268.60617px;transform:rotate(0)}12.5%{stroke-dashoffset:56.54867px;transform:rotate(0)}12.51%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(72.5deg)}25%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(72.5deg)}25.1%{stroke-dashoffset:268.60617px;transform:rotate(270deg)}37.5%{stroke-dashoffset:56.54867px;transform:rotate(270deg)}37.51%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(161.5deg)}50%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(161.5deg)}50.01%{stroke-dashoffset:268.60617px;transform:rotate(180deg)}62.5%{stroke-dashoffset:56.54867px;transform:rotate(180deg)}62.51%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(251.5deg)}75%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(251.5deg)}75.01%{stroke-dashoffset:268.60617px;transform:rotate(90deg)}87.5%{stroke-dashoffset:56.54867px;transform:rotate(90deg)}87.51%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(341.5deg)}100%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(341.5deg)}}@keyframes mat-progress-spinner-stroke-rotate-fallback{0%{transform:rotate(0)}25%{transform:rotate(1170deg)}50%{transform:rotate(2340deg)}75%{transform:rotate(3510deg)}100%{transform:rotate(4680deg)}}"],changeDetection:r.i.OnPush,encapsulation:r._10.None,preserveWhitespaces:!1}]}],e.ctorParameters=function(){return[{type:r.t},{type:i.a},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[a.b]}]}]},e.propDecorators={diameter:[{type:r.C}],strokeWidth:[{type:r.C}],mode:[{type:r.C}],value:[{type:r.C}]},e}(Object(o.u)(l,"primary")),d=function(t){function e(e,n,r){var i=t.call(this,e,n,r)||this;return i.mode="indeterminate",i}return Object(s.b)(e,t),e.decorators=[{type:r.m,args:[{selector:"mat-spinner",host:{role:"progressbar",mode:"indeterminate",class:"mat-spinner mat-progress-spinner","[style.width.px]":"_elementSize","[style.height.px]":"_elementSize"},inputs:["color"],template:'<svg [style.width.px]="_elementSize" [style.height.px]="_elementSize" [attr.viewBox]="_viewBox" preserveAspectRatio="xMidYMid meet" focusable="false"><circle cx="50%" cy="50%" [attr.r]="_circleRadius" [style.animation-name]="\'mat-progress-spinner-stroke-rotate-\' + diameter" [style.stroke-dashoffset.px]="_strokeDashOffset" [style.stroke-dasharray.px]="_strokeCircumference" [style.stroke-width.%]="_circleStrokeWidth"></circle></svg>',styles:[".mat-progress-spinner{display:block;position:relative}.mat-progress-spinner svg{position:absolute;transform:rotate(-90deg);top:0;left:0;transform-origin:center;overflow:visible}.mat-progress-spinner circle{fill:transparent;transform-origin:center;transition:stroke-dashoffset 225ms linear}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate]{animation:mat-progress-spinner-linear-rotate 2s linear infinite}.mat-progress-spinner.mat-progress-spinner-indeterminate-animation[mode=indeterminate] circle{transition-property:stroke;animation-duration:4s;animation-timing-function:cubic-bezier(.35,0,.25,1);animation-iteration-count:infinite}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate]{animation:mat-progress-spinner-stroke-rotate-fallback 10s cubic-bezier(.87,.03,.33,1) infinite}.mat-progress-spinner.mat-progress-spinner-indeterminate-fallback-animation[mode=indeterminate] circle{transition-property:stroke}@keyframes mat-progress-spinner-linear-rotate{0%{transform:rotate(0)}100%{transform:rotate(360deg)}}@keyframes mat-progress-spinner-stroke-rotate-100{0%{stroke-dashoffset:268.60617px;transform:rotate(0)}12.5%{stroke-dashoffset:56.54867px;transform:rotate(0)}12.51%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(72.5deg)}25%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(72.5deg)}25.1%{stroke-dashoffset:268.60617px;transform:rotate(270deg)}37.5%{stroke-dashoffset:56.54867px;transform:rotate(270deg)}37.51%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(161.5deg)}50%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(161.5deg)}50.01%{stroke-dashoffset:268.60617px;transform:rotate(180deg)}62.5%{stroke-dashoffset:56.54867px;transform:rotate(180deg)}62.51%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(251.5deg)}75%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(251.5deg)}75.01%{stroke-dashoffset:268.60617px;transform:rotate(90deg)}87.5%{stroke-dashoffset:56.54867px;transform:rotate(90deg)}87.51%{stroke-dashoffset:56.54867px;transform:rotateX(180deg) rotate(341.5deg)}100%{stroke-dashoffset:268.60617px;transform:rotateX(180deg) rotate(341.5deg)}}@keyframes mat-progress-spinner-stroke-rotate-fallback{0%{transform:rotate(0)}25%{transform:rotate(1170deg)}50%{transform:rotate(2340deg)}75%{transform:rotate(3510deg)}100%{transform:rotate(4680deg)}}"],changeDetection:r.i.OnPush,encapsulation:r._10.None,preserveWhitespaces:!1}]}],e.ctorParameters=function(){return[{type:r.t},{type:i.a},{type:void 0,decorators:[{type:r.M},{type:r.y,args:[a.b]}]}]},e}(p);!function(){function t(){}t.decorators=[{type:r.H,args:[{imports:[o.j,i.b],exports:[p,d,o.j],declarations:[p,d]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/radio.es5.js":function(t,e,n){"use strict";var r=n("../../../core/esm5/core.js"),i=n("../../../common/esm5/common.js"),o=n("../../../cdk/esm5/a11y.es5.js"),s=n("../../../cdk/esm5/collections.es5.js"),a=n("../../../cdk/esm5/overlay.es5.js"),c=n("../../../material/esm5/core.es5.js"),u=n("../../../../tslib/tslib.es6.js"),l=n("../../../cdk/esm5/coercion.es5.js"),p=0,d={provide:n("../../../forms/esm5/forms.js").d,useExisting:Object(r._14)(function(){return m}),multi:!0},h=function(){return function(){}}(),f=function(){return function(){}}(),m=function(t){function e(e){var n=t.call(this)||this;return n._changeDetector=e,n._value=null,n._name="mat-radio-group-"+p++,n._selected=null,n._isInitialized=!1,n._labelPosition="after",n._disabled=!1,n._required=!1,n._controlValueAccessorChangeFn=function(){},n.onTouched=function(){},n.change=new r.w,n}return Object(u.b)(e,t),Object.defineProperty(e.prototype,"name",{get:function(){return this._name},set:function(t){this._name=t,this._updateRadioButtonNames()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"align",{get:function(){return"after"==this.labelPosition?"start":"end"},set:function(t){this.labelPosition="start"==t?"after":"before"},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"labelPosition",{get:function(){return this._labelPosition},set:function(t){this._labelPosition="before"==t?"before":"after",this._markRadiosForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(t){this._value!=t&&(this._value=t,this._updateSelectedRadioFromValue(),this._checkSelectedRadioButton())},enumerable:!0,configurable:!0}),e.prototype._checkSelectedRadioButton=function(){this._selected&&!this._selected.checked&&(this._selected.checked=!0)},Object.defineProperty(e.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=t,this.value=t?t.value:null,this._checkSelectedRadioButton()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(l.b)(t),this._markRadiosForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(l.b)(t),this._markRadiosForCheck()},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){this._isInitialized=!0},e.prototype._touch=function(){this.onTouched&&this.onTouched()},e.prototype._updateRadioButtonNames=function(){var t=this;this._radios&&this._radios.forEach(function(e){e.name=t.name})},e.prototype._updateSelectedRadioFromValue=function(){var t=this,e=null!=this._selected&&this._selected.value==this._value;null==this._radios||e||(this._selected=null,this._radios.forEach(function(e){e.checked=t.value==e.value,e.checked&&(t._selected=e)}))},e.prototype._emitChangeEvent=function(){if(this._isInitialized){var t=new h;t.source=this._selected,t.value=this._value,this.change.emit(t)}},e.prototype._markRadiosForCheck=function(){this._radios&&this._radios.forEach(function(t){return t._markForCheck()})},e.prototype.writeValue=function(t){this.value=t,this._changeDetector.markForCheck()},e.prototype.registerOnChange=function(t){this._controlValueAccessorChangeFn=t},e.prototype.registerOnTouched=function(t){this.onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t,this._changeDetector.markForCheck()},e.decorators=[{type:r.s,args:[{selector:"mat-radio-group",exportAs:"matRadioGroup",providers:[d],host:{role:"radiogroup",class:"mat-radio-group"},inputs:["disabled"]}]}],e.ctorParameters=function(){return[{type:r.j}]},e.propDecorators={change:[{type:r.N}],_radios:[{type:r.r,args:[Object(r._14)(function(){return v}),{descendants:!0}]}],name:[{type:r.C}],align:[{type:r.C}],labelPosition:[{type:r.C}],value:[{type:r.C}],selected:[{type:r.C}],disabled:[{type:r.C}],required:[{type:r.C}]},e}(Object(c.w)(f)),g=function(){return function(t){this._elementRef=t}}(),v=function(t){function e(e,n,i,o,s){var a=t.call(this,n)||this;return a._changeDetector=i,a._focusMonitor=o,a._radioDispatcher=s,a._uniqueId="mat-radio-"+ ++p,a.id=a._uniqueId,a.change=new r.w,a._checked=!1,a._value=null,a._rippleConfig={centered:!0,radius:23,speedFactor:1.5},a._removeUniqueSelectionListener=function(){},a.radioGroup=e,a._removeUniqueSelectionListener=s.listen(function(t,e){t!=a.id&&e==a.name&&(a.checked=!1)}),a}return Object(u.b)(e,t),Object.defineProperty(e.prototype,"checked",{get:function(){return this._checked},set:function(t){var e=Object(l.b)(t);this._checked!=e&&(this._checked=e,e&&this.radioGroup&&this.radioGroup.value!=this.value?this.radioGroup.selected=this:!e&&this.radioGroup&&this.radioGroup.value==this.value&&(this.radioGroup.selected=null),e&&this._radioDispatcher.notify(this.id,this.name),this._changeDetector.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(t){this._value!=t&&(this._value=t,null!=this.radioGroup&&(this.checked||(this.checked=this.radioGroup.value==t),this.checked&&(this.radioGroup.selected=this)))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"align",{get:function(){return"after"==this.labelPosition?"start":"end"},set:function(t){this.labelPosition="start"==t?"after":"before"},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"labelPosition",{get:function(){return this._labelPosition||this.radioGroup&&this.radioGroup.labelPosition||"after"},set:function(t){this._labelPosition=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disabled",{get:function(){return this._disabled||null!=this.radioGroup&&this.radioGroup.disabled},set:function(t){this._disabled=Object(l.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required||this.radioGroup&&this.radioGroup.required},set:function(t){this._required=Object(l.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"inputId",{get:function(){return(this.id||this._uniqueId)+"-input"},enumerable:!0,configurable:!0}),e.prototype.focus=function(){this._focusMonitor.focusVia(this._inputElement.nativeElement,"keyboard")},e.prototype._markForCheck=function(){this._changeDetector.markForCheck()},e.prototype.ngOnInit=function(){this.radioGroup&&(this.checked=this.radioGroup.value===this._value,this.name=this.radioGroup.name)},e.prototype.ngAfterViewInit=function(){var t=this;this._focusMonitor.monitor(this._inputElement.nativeElement,!1).subscribe(function(e){return t._onInputFocusChange(e)})},e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._inputElement.nativeElement),this._removeUniqueSelectionListener()},e.prototype._emitChangeEvent=function(){var t=new h;t.source=this,t.value=this._value,this.change.emit(t)},e.prototype._isRippleDisabled=function(){return this.disableRipple||this.disabled},e.prototype._onInputClick=function(t){t.stopPropagation()},e.prototype._onInputChange=function(t){t.stopPropagation();var e=this.radioGroup&&this.value!=this.radioGroup.value;this.checked=!0,this._emitChangeEvent(),this.radioGroup&&(this.radioGroup._controlValueAccessorChangeFn(this.value),this.radioGroup._touch(),e&&this.radioGroup._emitChangeEvent())},e.prototype._onInputFocusChange=function(t){this._focusRipple||"keyboard"!==t?t||(this.radioGroup&&this.radioGroup._touch(),this._focusRipple&&(this._focusRipple.fadeOut(),this._focusRipple=null)):this._focusRipple=this._ripple.launch(0,0,Object(u.a)({persistent:!0},this._rippleConfig))},e.decorators=[{type:r.m,args:[{selector:"mat-radio-button",template:'<label [attr.for]="inputId" class="mat-radio-label" #label><div class="mat-radio-container"><div class="mat-radio-outer-circle"></div><div class="mat-radio-inner-circle"></div><div mat-ripple class="mat-radio-ripple" [matRippleTrigger]="label" [matRippleDisabled]="_isRippleDisabled()" [matRippleCentered]="_rippleConfig.centered" [matRippleRadius]="_rippleConfig.radius" [matRippleSpeedFactor]="_rippleConfig.speedFactor"></div></div><input #input class="mat-radio-input cdk-visually-hidden" type="radio" [id]="inputId" [checked]="checked" [disabled]="disabled" [attr.name]="name" [required]="required" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" (change)="_onInputChange($event)" (click)="_onInputClick($event)"><div class="mat-radio-label-content" [class.mat-radio-label-before]="labelPosition == \'before\'"><span style="display:none">&nbsp;</span><ng-content></ng-content></div></label>',styles:[".mat-radio-button{display:inline-block}.mat-radio-label{cursor:pointer;display:inline-flex;align-items:center;white-space:nowrap;vertical-align:middle}.mat-radio-container{box-sizing:border-box;display:inline-block;position:relative;width:20px;height:20px;flex-shrink:0}.mat-radio-outer-circle{box-sizing:border-box;height:20px;left:0;position:absolute;top:0;transition:border-color ease 280ms;width:20px;border-width:2px;border-style:solid;border-radius:50%}.mat-radio-inner-circle{border-radius:50%;box-sizing:border-box;height:20px;left:0;position:absolute;top:0;transition:transform ease 280ms,background-color ease 280ms;width:20px;transform:scale(.001)}.mat-radio-checked .mat-radio-inner-circle{transform:scale(.5)}.mat-radio-label-content{display:inline-block;order:0;line-height:inherit;padding-left:8px;padding-right:0}[dir=rtl] .mat-radio-label-content{padding-right:8px;padding-left:0}.mat-radio-label-content.mat-radio-label-before{order:-1;padding-left:0;padding-right:8px}[dir=rtl] .mat-radio-label-content.mat-radio-label-before{padding-right:0;padding-left:8px}.mat-radio-disabled,.mat-radio-disabled .mat-radio-label{cursor:default}.mat-radio-ripple{position:absolute;left:-15px;top:-15px;height:50px;width:50px;z-index:1;pointer-events:none}"],inputs:["color","disableRipple"],encapsulation:r._10.None,preserveWhitespaces:!1,exportAs:"matRadioButton",host:{class:"mat-radio-button","[class.mat-radio-checked]":"checked","[class.mat-radio-disabled]":"disabled","[attr.id]":"id","(focus)":"_inputElement.nativeElement.focus()"},changeDetection:r.i.OnPush}]}],e.ctorParameters=function(){return[{type:m,decorators:[{type:r.M}]},{type:r.t},{type:r.j},{type:o.f},{type:s.c}]},e.propDecorators={id:[{type:r.C}],name:[{type:r.C}],ariaLabel:[{type:r.C,args:["aria-label"]}],ariaLabelledby:[{type:r.C,args:["aria-labelledby"]}],checked:[{type:r.C}],value:[{type:r.C}],align:[{type:r.C}],labelPosition:[{type:r.C}],disabled:[{type:r.C}],required:[{type:r.C}],change:[{type:r.N}],_ripple:[{type:r._7,args:[c.r]}],_inputElement:[{type:r._7,args:["input"]}]},e}(Object(c.u)(Object(c.v)(g),"accent"));!function(){function t(){}t.decorators=[{type:r.H,args:[{imports:[i.a,c.s,c.j,o.a],exports:[m,v,c.j],providers:[s.b,a.e],declarations:[m,v]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/select.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return D});var r=n("../../../core/esm5/core.js"),i=n("../../../common/esm5/common.js"),o=n("../../../../tslib/tslib.es6.js"),s=n("../../../cdk/esm5/a11y.es5.js"),a=n("../../../cdk/esm5/bidi.es5.js"),c=n("../../../cdk/esm5/coercion.es5.js"),u=n("../../../cdk/esm5/collections.es5.js"),l=n("../../../cdk/esm5/keycodes.es5.js"),p=n("../../../cdk/esm5/overlay.es5.js"),d=n("../../../../rxjs/_esm5/operators/filter.js"),h=n("../../../../rxjs/_esm5/operators/take.js"),f=n("../../../../rxjs/_esm5/operators/map.js"),m=n("../../../../rxjs/_esm5/operators/startWith.js"),g=n("../../../../rxjs/_esm5/operators/takeUntil.js"),v=n("../../../forms/esm5/forms.js"),y=n("../../../material/esm5/core.es5.js"),b=n("../../../material/esm5/form-field.es5.js"),_=(n("../../../../rxjs/_esm5/Observable.js"),n("../../../../rxjs/_esm5/observable/merge.js")),j=n("../../../../rxjs/_esm5/Subject.js"),x=n("../../../animations/esm5/animations.js"),w=Object(x.k)("transformPanel",[Object(x.h)("showing",Object(x.i)({opacity:1,minWidth:"calc(100% + 32px)",transform:"scaleY(1)"})),Object(x.h)("showing-multiple",Object(x.i)({opacity:1,minWidth:"calc(100% + 64px)",transform:"scaleY(1)"})),Object(x.j)("void => *",[Object(x.i)({opacity:0,minWidth:"100%",transform:"scaleY(0)"}),Object(x.e)("150ms cubic-bezier(0.25, 0.8, 0.25, 1)")]),Object(x.j)("* => void",[Object(x.e)("250ms 100ms linear",Object(x.i)({opacity:0}))])]),O=Object(x.k)("fadeInContent",[Object(x.h)("showing",Object(x.i)({opacity:1})),Object(x.j)("void => showing",[Object(x.i)({opacity:0}),Object(x.e)("150ms 100ms cubic-bezier(0.55, 0, 0.55, 0.2)")])]);var C=0,k=new r.A("mat-select-scroll-strategy");var E={provide:k,deps:[p.b],useFactory:function(t){return function(){return t.scrollStrategies.reposition()}}},S=function(){return function(t,e){this.source=t,this.value=e}}(),P=function(){return function(t,e,n,r,i){this._elementRef=t,this._defaultErrorStateMatcher=e,this._parentForm=n,this._parentFormGroup=r,this.ngControl=i}}(),M=Object(y.y)(Object(y.w)(Object(y.x)(P))),A=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"mat-select-trigger"}]}],t.ctorParameters=function(){return[]},t}(),T=function(t){function e(e,n,i,o,s,a,c,u,l,p,d,h){var f=t.call(this,s,o,c,u,p)||this;return f._viewportRuler=e,f._changeDetectorRef=n,f._ngZone=i,f._dir=a,f._parentFormField=l,f.ngControl=p,f._scrollStrategyFactory=h,f._panelOpen=!1,f._required=!1,f._scrollTop=0,f._multiple=!1,f._compareWith=function(t,e){return t===e},f._uid="mat-select-"+C++,f._destroy=new j.a,f._triggerFontSize=0,f._onChange=function(){},f._onTouched=function(){},f._optionIds="",f._transformOrigin="top",f._panelDoneAnimating=!1,f._scrollStrategy=f._scrollStrategyFactory(),f._offsetY=0,f._positions=[{originX:"start",originY:"top",overlayX:"start",overlayY:"top"},{originX:"start",originY:"bottom",overlayX:"start",overlayY:"bottom"}],f.focused=!1,f.controlType="mat-select",f._disableRipple=!1,f.ariaLabel="",f.openedChange=new r.w,f.onOpen=f._openedStream,f.onClose=f._closedStream,f.selectionChange=new r.w,f.change=f.selectionChange,f.valueChange=new r.w,f.ngControl&&(f.ngControl.valueAccessor=f),f.tabIndex=parseInt(d)||0,f.id=f.id,f}return Object(o.b)(e,t),Object.defineProperty(e.prototype,"placeholder",{get:function(){return this._placeholder},set:function(t){this._placeholder=t,this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(c.b)(t),this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"multiple",{get:function(){return this._multiple},set:function(t){if(this._selectionModel)throw Error("Cannot change `multiple` mode of select after initialization.");this._multiple=Object(c.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"compareWith",{get:function(){return this._compareWith},set:function(t){if("function"!=typeof t)throw Error("`compareWith` must be a function.");this._compareWith=t,this._selectionModel&&this._initializeSelection()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return this._value},set:function(t){t!==this._value&&(this.writeValue(t),this._value=t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disableRipple",{get:function(){return this._disableRipple},set:function(t){this._disableRipple=Object(c.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"id",{get:function(){return this._id},set:function(t){this._id=t||this._uid,this.stateChanges.next()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"optionSelectionChanges",{get:function(){return _.a.apply(void 0,this.options.map(function(t){return t.onSelectionChange}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_openedStream",{get:function(){return this.openedChange.pipe(Object(d.a)(function(t){return t}),Object(f.a)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_closedStream",{get:function(){return this.openedChange.pipe(Object(d.a)(function(t){return!t}),Object(f.a)(function(){}))},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){this._selectionModel=new u.a(this.multiple,void 0,!1),this.stateChanges.next()},e.prototype.ngAfterContentInit=function(){var t=this;this._initKeyManager(),this.options.changes.pipe(Object(m.a)(null),Object(g.a)(this._destroy)).subscribe(function(){t._resetOptions(),t._initializeSelection()})},e.prototype.ngDoCheck=function(){this.ngControl&&this.updateErrorState()},e.prototype.ngOnChanges=function(t){t.disabled&&this.stateChanges.next()},e.prototype.ngOnDestroy=function(){this._destroy.next(),this._destroy.complete()},e.prototype.toggle=function(){this.panelOpen?this.close():this.open()},e.prototype.open=function(){var t=this;!this.disabled&&this.options&&this.options.length&&(this._triggerRect=this.trigger.nativeElement.getBoundingClientRect(),this._triggerFontSize=parseInt(getComputedStyle(this.trigger.nativeElement)["font-size"]),this._panelOpen=!0,this._calculateOverlayPosition(),this._highlightCorrectOption(),this._changeDetectorRef.markForCheck(),this._ngZone.onStable.asObservable().pipe(Object(h.a)(1)).subscribe(function(){t._triggerFontSize&&t.overlayDir.overlayRef&&t.overlayDir.overlayRef.overlayElement&&(t.overlayDir.overlayRef.overlayElement.style.fontSize=t._triggerFontSize+"px")}))},e.prototype.close=function(){this._panelOpen&&(this._panelOpen=!1,this._changeDetectorRef.markForCheck(),this.focus())},e.prototype.writeValue=function(t){this.options&&this._setSelectionByValue(t)},e.prototype.registerOnChange=function(t){this._onChange=t},e.prototype.registerOnTouched=function(t){this._onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t,this._changeDetectorRef.markForCheck(),this.stateChanges.next()},Object.defineProperty(e.prototype,"panelOpen",{get:function(){return this._panelOpen},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selected",{get:function(){return this.multiple?this._selectionModel.selected:this._selectionModel.selected[0]},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"triggerValue",{get:function(){if(this.empty)return"";if(this._multiple){var t=this._selectionModel.selected.map(function(t){return t.viewValue});return this._isRtl()&&t.reverse(),t.join(", ")}return this._selectionModel.selected[0].viewValue},enumerable:!0,configurable:!0}),e.prototype._isRtl=function(){return!!this._dir&&"rtl"===this._dir.value},e.prototype._handleKeydown=function(t){this.disabled||(this.panelOpen?this._handleOpenKeydown(t):this._handleClosedKeydown(t))},e.prototype._handleClosedKeydown=function(t){var e=t.keyCode,n=e===l.d||e===l.p;e===l.f||e===l.n||this.multiple&&n?(t.preventDefault(),this.open()):this.multiple||this._keyManager.onKeydown(t)},e.prototype._handleOpenKeydown=function(t){var e=t.keyCode;e===l.h||e===l.e?(t.preventDefault(),e===l.h?this._keyManager.setFirstItemActive():this._keyManager.setLastItemActive()):e!==l.f&&e!==l.n||!this._keyManager.activeItem?this._keyManager.onKeydown(t):(t.preventDefault(),this._keyManager.activeItem._selectViaInteraction())},e.prototype._onPanelDone=function(){this.panelOpen?(this._scrollTop=0,this.openedChange.emit(!0)):(this.openedChange.emit(!1),this._panelDoneAnimating=!1,this.overlayDir.offsetX=0,this._changeDetectorRef.markForCheck())},e.prototype._onFadeInDone=function(){this._panelDoneAnimating=this.panelOpen,this._changeDetectorRef.markForCheck()},e.prototype._onFocus=function(){this.disabled||(this.focused=!0,this.stateChanges.next())},e.prototype._onBlur=function(){this.disabled||this.panelOpen||(this.focused=!1,this._onTouched(),this._changeDetectorRef.markForCheck(),this.stateChanges.next())},e.prototype._onAttached=function(){var t=this;this.overlayDir.positionChange.pipe(Object(h.a)(1)).subscribe(function(){t._changeDetectorRef.detectChanges(),t._calculateOverlayOffsetX(),t.panel.nativeElement.scrollTop=t._scrollTop})},e.prototype._getPanelTheme=function(){return this._parentFormField?"mat-"+this._parentFormField.color:""},Object.defineProperty(e.prototype,"empty",{get:function(){return!this._selectionModel||this._selectionModel.isEmpty()},enumerable:!0,configurable:!0}),e.prototype._initializeSelection=function(){var t=this;Promise.resolve().then(function(){t._setSelectionByValue(t.ngControl?t.ngControl.value:t._value)})},e.prototype._setSelectionByValue=function(t,e){var n=this;if(void 0===e&&(e=!1),this.multiple&&t){if(!Array.isArray(t))throw Error("Value must be an array in multiple-selection mode.");this._clearSelection(),t.forEach(function(t){return n._selectValue(t,e)}),this._sortValues()}else{this._clearSelection();var r=this._selectValue(t,e);r&&this._keyManager.setActiveItem(this.options.toArray().indexOf(r))}this._changeDetectorRef.markForCheck()},e.prototype._selectValue=function(t,e){var n=this;void 0===e&&(e=!1);var i=this.options.find(function(e){try{return null!=e.value&&n._compareWith(e.value,t)}catch(t){return Object(r._16)()&&console.warn(t),!1}});return i&&(e?i._selectViaInteraction():i.select(),this._selectionModel.select(i),this.stateChanges.next()),i},e.prototype._clearSelection=function(t){this._selectionModel.clear(),this.options.forEach(function(e){e!==t&&e.deselect()}),this.stateChanges.next()},e.prototype._initKeyManager=function(){var t=this;this._keyManager=new s.c(this.options).withTypeAhead(),this._keyManager.tabOut.pipe(Object(g.a)(this._destroy)).subscribe(function(){return t.close()}),this._keyManager.change.pipe(Object(g.a)(this._destroy)).subscribe(function(){t._panelOpen&&t.panel?t._scrollActiveOptionIntoView():t._panelOpen||t.multiple||!t._keyManager.activeItem||t._keyManager.activeItem._selectViaInteraction()})},e.prototype._resetOptions=function(){var t=this;this.optionSelectionChanges.pipe(Object(g.a)(Object(_.a)(this._destroy,this.options.changes)),Object(d.a)(function(t){return t.isUserInput})).subscribe(function(e){t._onSelect(e.source),t.multiple||t.close()}),this._setOptionIds()},e.prototype._onSelect=function(t){var e=this._selectionModel.isSelected(t);this.multiple?(this._selectionModel.toggle(t),this.stateChanges.next(),e?t.deselect():t.select(),this._sortValues()):(this._clearSelection(null==t.value?void 0:t),null==t.value?this._propagateChanges(t.value):(this._selectionModel.select(t),this.stateChanges.next())),e!==this._selectionModel.isSelected(t)&&this._propagateChanges()},e.prototype._sortValues=function(){var t=this;this._multiple&&(this._selectionModel.clear(),this.options.forEach(function(e){e.selected&&t._selectionModel.select(e)}),this.stateChanges.next())},e.prototype._propagateChanges=function(t){var e=null;e=this.multiple?this.selected.map(function(t){return t.value}):this.selected?this.selected.value:t,this._value=e,this._onChange(e),this.selectionChange.emit(new S(this,e)),this.valueChange.emit(e),this._changeDetectorRef.markForCheck()},e.prototype._setOptionIds=function(){this._optionIds=this.options.map(function(t){return t.id}).join(" ")},e.prototype._highlightCorrectOption=function(){this._keyManager&&(this.empty?this._keyManager.setFirstItemActive():this._keyManager.setActiveItem(this._getOptionIndex(this._selectionModel.selected[0])))},e.prototype._scrollActiveOptionIntoView=function(){var t=this._getItemHeight(),e=this._keyManager.activeItemIndex||0,n=(e+y.o.countGroupLabelsBeforeOption(e,this.options,this.optionGroups))*t,r=this.panel.nativeElement.scrollTop;n<r?this.panel.nativeElement.scrollTop=n:n+t>r+256&&(this.panel.nativeElement.scrollTop=Math.max(0,n-256+t))},e.prototype.focus=function(){this._elementRef.nativeElement.focus()},e.prototype._getOptionIndex=function(t){return this.options.reduce(function(e,n,r){return void 0===e?t===n?r:void 0:e},void 0)},e.prototype._calculateOverlayPosition=function(){var t=this._getItemHeight(),e=this._getItemCount(),n=Math.min(e*t,256),r=e*t-n,i=this.empty?0:this._getOptionIndex(this._selectionModel.selected[0]);i+=y.o.countGroupLabelsBeforeOption(i,this.options,this.optionGroups);var o=n/2;this._scrollTop=this._calculateOverlayScroll(i,o,r),this._offsetY=this._calculateOverlayOffsetY(i,o,r),this._checkOverlayWithinViewport(r)},e.prototype._calculateOverlayScroll=function(t,e,n){var r=this._getItemHeight(),i=r*t-e+r/2;return Math.min(Math.max(0,i),n)},Object.defineProperty(e.prototype,"_ariaLabel",{get:function(){return this.ariaLabelledby?null:this.ariaLabel||this.placeholder},enumerable:!0,configurable:!0}),e.prototype._getAriaActiveDescendant=function(){return this.panelOpen&&this._keyManager&&this._keyManager.activeItem?this._keyManager.activeItem.id:null},e.prototype._calculateOverlayOffsetX=function(){var t,e=this.overlayDir.overlayRef.overlayElement.getBoundingClientRect(),n=this._viewportRuler.getViewportSize(),r=this._isRtl(),i=this.multiple?60:32;if(this.multiple)t=44;else{var o=this._selectionModel.selected[0]||this.options.first;t=o&&o.group?32:16}r||(t*=-1);var s=0-(e.left+t-(r?i:0)),a=e.right+t-n.width+(r?0:i);s>0?t+=s+8:a>0&&(t-=a+8),this.overlayDir.offsetX=t,this.overlayDir.overlayRef.updatePosition()},e.prototype._calculateOverlayOffsetY=function(t,e,n){var r,i=this._getItemHeight(),o=(i-this._triggerRect.height)/2,s=Math.floor(256/i);if(0===this._scrollTop)r=t*i;else if(this._scrollTop===n){r=(t-(this._getItemCount()-s))*i+(i-(this._getItemCount()*i-256)%i)}else r=e-i/2;return-1*r-o},e.prototype._checkOverlayWithinViewport=function(t){var e=this._getItemHeight(),n=this._viewportRuler.getViewportSize(),r=this._triggerRect.top-8,i=n.height-this._triggerRect.bottom-8,o=Math.abs(this._offsetY),s=Math.min(this._getItemCount()*e,256)-o-this._triggerRect.height;s>i?this._adjustPanelUp(s,i):o>r?this._adjustPanelDown(o,r,t):this._transformOrigin=this._getOriginBasedOnOption()},e.prototype._adjustPanelUp=function(t,e){var n=Math.round(t-e);this._scrollTop-=n,this._offsetY-=n,this._transformOrigin=this._getOriginBasedOnOption(),this._scrollTop<=0&&(this._scrollTop=0,this._offsetY=0,this._transformOrigin="50% bottom 0px")},e.prototype._adjustPanelDown=function(t,e,n){var r=Math.round(t-e);if(this._scrollTop+=r,this._offsetY+=r,this._transformOrigin=this._getOriginBasedOnOption(),this._scrollTop>=n)return this._scrollTop=n,this._offsetY=0,void(this._transformOrigin="50% top 0px")},e.prototype._getOriginBasedOnOption=function(){var t=this._getItemHeight(),e=(t-this._triggerRect.height)/2;return"50% "+(Math.abs(this._offsetY)-e+t/2)+"px 0px"},e.prototype._getItemCount=function(){return this.options.length+this.optionGroups.length},e.prototype._getItemHeight=function(){return 3*this._triggerFontSize},e.prototype.setDescribedByIds=function(t){this._ariaDescribedby=t.join(" ")},e.prototype.onContainerClick=function(){this.focus(),this.open()},Object.defineProperty(e.prototype,"shouldLabelFloat",{get:function(){return this._panelOpen||!this.empty},enumerable:!0,configurable:!0}),e.decorators=[{type:r.m,args:[{selector:"mat-select",exportAs:"matSelect",template:'<div cdk-overlay-origin class="mat-select-trigger" aria-hidden="true" (click)="toggle()" #origin="cdkOverlayOrigin" #trigger><div class="mat-select-value" [ngSwitch]="empty"><span class="mat-select-placeholder" *ngSwitchCase="true">{{placeholder || \' \'}}</span> <span class="mat-select-value-text" *ngSwitchCase="false" [ngSwitch]="!!customTrigger"><span *ngSwitchDefault>{{triggerValue}}</span><ng-content select="mat-select-trigger" *ngSwitchCase="true"></ng-content></span></div><div class="mat-select-arrow-wrapper"><div class="mat-select-arrow"></div></div></div><ng-template cdk-connected-overlay hasBackdrop backdropClass="cdk-overlay-transparent-backdrop" [scrollStrategy]="_scrollStrategy" [origin]="origin" [open]="panelOpen" [positions]="_positions" [minWidth]="_triggerRect?.width" [offsetY]="_offsetY" (backdropClick)="close()" (attach)="_onAttached()" (detach)="close()"><div #panel class="mat-select-panel {{ _getPanelTheme() }}" [ngClass]="panelClass" [@transformPanel]="multiple ? \'showing-multiple\' : \'showing\'" (@transformPanel.done)="_onPanelDone()" [style.transformOrigin]="_transformOrigin" [class.mat-select-panel-done-animating]="_panelDoneAnimating" [style.font-size.px]="_triggerFontSize"><div class="mat-select-content" [@fadeInContent]="\'showing\'" (@fadeInContent.done)="_onFadeInDone()"><ng-content></ng-content></div></div></ng-template>',styles:[".mat-select{display:inline-block;width:100%;outline:0}.mat-select-trigger{display:inline-table;cursor:pointer;position:relative;box-sizing:border-box}.mat-select-disabled .mat-select-trigger{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:default}.mat-select-value{display:table-cell;max-width:0;width:100%;overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.mat-select-value-text{white-space:nowrap;overflow:hidden;text-overflow:ellipsis}.mat-select-arrow-wrapper{display:table-cell;vertical-align:middle}.mat-select-arrow{width:0;height:0;border-left:5px solid transparent;border-right:5px solid transparent;border-top:5px solid;margin:0 4px}.mat-select-panel{min-width:112px;max-width:280px;overflow:auto;-webkit-overflow-scrolling:touch;padding-top:0;padding-bottom:0;max-height:256px;min-width:100%}.mat-select-panel:not([class*=mat-elevation-z]){box-shadow:0 5px 5px -3px rgba(0,0,0,.2),0 8px 10px 1px rgba(0,0,0,.14),0 3px 14px 2px rgba(0,0,0,.12)}@media screen and (-ms-high-contrast:active){.mat-select-panel{outline:solid 1px}}.mat-select-panel .mat-optgroup-label,.mat-select-panel .mat-option{font-size:inherit;line-height:3em;height:3em}.mat-form-field-type-mat-select:not(.mat-form-field-disabled) .mat-form-field-flex{cursor:pointer}.mat-form-field-type-mat-select .mat-form-field-label{width:calc(100% - 18px)}.mat-select-placeholder{transition:color .4s .133s cubic-bezier(.25,.8,.25,1)}.mat-form-field-hide-placeholder .mat-select-placeholder{color:transparent;transition:none}"],inputs:["disabled","tabIndex"],encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush,host:{role:"listbox","[attr.id]":"id","[attr.tabindex]":"tabIndex","[attr.aria-label]":"_ariaLabel","[attr.aria-labelledby]":"ariaLabelledby","[attr.aria-required]":"required.toString()","[attr.aria-disabled]":"disabled.toString()","[attr.aria-invalid]":"errorState","[attr.aria-owns]":"_optionIds","[attr.aria-multiselectable]":"multiple","[attr.aria-describedby]":"_ariaDescribedby || null","[attr.aria-activedescendant]":"_getAriaActiveDescendant()","[class.mat-select-disabled]":"disabled","[class.mat-select-invalid]":"errorState","[class.mat-select-required]":"required",class:"mat-select","(keydown)":"_handleKeydown($event)","(focus)":"_onFocus()","(blur)":"_onBlur()"},animations:[w,O],providers:[{provide:b.b,useExisting:e},{provide:y.h,useExisting:e}]}]}],e.ctorParameters=function(){return[{type:p.f},{type:r.j},{type:r.L},{type:y.d},{type:r.t},{type:a.b,decorators:[{type:r.M}]},{type:v.f,decorators:[{type:r.M}]},{type:v.b,decorators:[{type:r.M}]},{type:b.a,decorators:[{type:r.M}]},{type:v.e,decorators:[{type:r.Z},{type:r.M}]},{type:void 0,decorators:[{type:r.g,args:["tabindex"]}]},{type:void 0,decorators:[{type:r.y,args:[k]}]}]},e.propDecorators={trigger:[{type:r._7,args:["trigger"]}],panel:[{type:r._7,args:["panel"]}],overlayDir:[{type:r._7,args:[p.a]}],options:[{type:r.r,args:[y.o,{descendants:!0}]}],optionGroups:[{type:r.r,args:[y.n]}],panelClass:[{type:r.C}],customTrigger:[{type:r.q,args:[A]}],placeholder:[{type:r.C}],required:[{type:r.C}],multiple:[{type:r.C}],compareWith:[{type:r.C}],value:[{type:r.C}],disableRipple:[{type:r.C}],ariaLabel:[{type:r.C,args:["aria-label"]}],ariaLabelledby:[{type:r.C,args:["aria-labelledby"]}],errorStateMatcher:[{type:r.C}],id:[{type:r.C}],openedChange:[{type:r.N}],_openedStream:[{type:r.N,args:["opened"]}],_closedStream:[{type:r.N,args:["closed"]}],onOpen:[{type:r.N}],onClose:[{type:r.N}],selectionChange:[{type:r.N}],change:[{type:r.N}],valueChange:[{type:r.N}]},e}(M),D=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.a,p.d,y.p,y.j],exports:[b.c,T,A,y.p,y.j],declarations:[T,A],providers:[E,y.d]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/sidenav.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return T});var r=n("../../../cdk/esm5/a11y.es5.js"),i=n("../../../cdk/esm5/overlay.es5.js"),o=n("../../../common/esm5/common.js"),s=n("../../../core/esm5/core.js"),a=n("../../../material/esm5/core.es5.js"),c=n("../../../cdk/esm5/scrolling.es5.js"),u=n("../../../cdk/esm5/platform.es5.js"),l=n("../../../../tslib/tslib.es6.js"),p=n("../../../animations/esm5/animations.js"),d=n("../../../cdk/esm5/bidi.es5.js"),h=n("../../../cdk/esm5/coercion.es5.js"),f=n("../../../cdk/esm5/keycodes.es5.js"),m=n("../../../../rxjs/_esm5/observable/merge.js"),g=n("../../../../rxjs/_esm5/operators/filter.js"),v=n("../../../../rxjs/_esm5/operators/take.js"),y=n("../../../../rxjs/_esm5/operators/startWith.js"),b=n("../../../../rxjs/_esm5/operators/takeUntil.js"),_=n("../../../../rxjs/_esm5/operators/debounceTime.js"),j=n("../../../../rxjs/_esm5/operators/map.js"),x=n("../../../../rxjs/_esm5/Subject.js");n("../../../../rxjs/_esm5/Observable.js");function w(t){throw Error("A drawer was already declared for 'position=\""+t+"\"'")}var O=function(){return function(t,e){this.type=t,this.animationFinished=e}}(),C=new s.A("MAT_DRAWER_DEFAULT_AUTOSIZE"),k=function(){function t(t,e){this._changeDetectorRef=t,this._container=e,this._margins={left:null,right:null}}return t.prototype.ngAfterContentInit=function(){var t=this;this._container._contentMargins.subscribe(function(e){t._margins=e,t._changeDetectorRef.markForCheck()})},t.decorators=[{type:s.m,args:[{selector:"mat-drawer-content",template:"<ng-content></ng-content>",host:{class:"mat-drawer-content","[style.margin-left.px]":"_margins.left","[style.margin-right.px]":"_margins.right"},changeDetection:s.i.OnPush,encapsulation:s._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[{type:s.j},{type:S,decorators:[{type:s.y,args:[Object(s._14)(function(){return S})]}]}]},t}(),E=function(){function t(t,e,n,r,i){var o=this;this._elementRef=t,this._focusTrapFactory=e,this._focusMonitor=n,this._platform=r,this._doc=i,this._elementFocusedBeforeDrawerWasOpened=null,this._enableAnimations=!1,this._position="start",this._mode="over",this._disableClose=!1,this._opened=!1,this._animationStarted=new s.w,this._animationState="void",this.openedChange=new s.w,this.onOpen=this._openedStream,this.onClose=this._closedStream,this.onPositionChanged=new s.w,this.onAlignChanged=new s.w,this._modeChanged=new x.a,this.openedChange.subscribe(function(t){t?(o._doc&&(o._elementFocusedBeforeDrawerWasOpened=o._doc.activeElement),o._isFocusTrapEnabled&&o._focusTrap&&o._focusTrap.focusInitialElementWhenReady()):o._restoreFocus()})}return Object.defineProperty(t.prototype,"position",{get:function(){return this._position},set:function(t){(t="end"===t?"end":"start")!=this._position&&(this._position=t,this.onAlignChanged.emit(),this.onPositionChanged.emit())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"align",{get:function(){return this.position},set:function(t){this.position=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"mode",{get:function(){return this._mode},set:function(t){this._mode=t,this._modeChanged.next()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disableClose",{get:function(){return this._disableClose},set:function(t){this._disableClose=Object(h.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_openedStream",{get:function(){return this.openedChange.pipe(Object(g.a)(function(t){return t}),Object(j.a)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"openedStart",{get:function(){return this._animationStarted.pipe(Object(g.a)(function(t){return t.fromState!==t.toState&&0===t.toState.indexOf("open")}),Object(j.a)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_closedStream",{get:function(){return this.openedChange.pipe(Object(g.a)(function(t){return!t}),Object(j.a)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"closedStart",{get:function(){return this._animationStarted.pipe(Object(g.a)(function(t){return t.fromState!==t.toState&&"void"===t.toState}),Object(j.a)(function(){}))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_isFocusTrapEnabled",{get:function(){return this.opened&&"side"!==this.mode},enumerable:!0,configurable:!0}),t.prototype._restoreFocus=function(){var t=this._doc&&this._doc.activeElement;t&&this._elementRef.nativeElement.contains(t)&&(this._elementFocusedBeforeDrawerWasOpened instanceof HTMLElement?this._focusMonitor.focusVia(this._elementFocusedBeforeDrawerWasOpened,this._openedVia):this._elementRef.nativeElement.blur()),this._elementFocusedBeforeDrawerWasOpened=null,this._openedVia=null},t.prototype.ngAfterContentInit=function(){this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement),this._focusTrap.enabled=this._isFocusTrapEnabled},t.prototype.ngAfterContentChecked=function(){this._platform.isBrowser&&(this._enableAnimations=!0)},t.prototype.ngOnDestroy=function(){this._focusTrap&&this._focusTrap.destroy()},Object.defineProperty(t.prototype,"opened",{get:function(){return this._opened},set:function(t){this.toggle(Object(h.b)(t))},enumerable:!0,configurable:!0}),t.prototype.open=function(t){return this.toggle(!0,t)},t.prototype.close=function(){return this.toggle(!1)},t.prototype.toggle=function(t,e){var n=this;return void 0===t&&(t=!this.opened),void 0===e&&(e="program"),this._opened=t,t?(this._animationState=this._enableAnimations?"open":"open-instant",this._openedVia=e):(this._animationState="void",this._restoreFocus()),this._focusTrap&&(this._focusTrap.enabled=this._isFocusTrapEnabled),new Promise(function(t){n.openedChange.pipe(Object(v.a)(1)).subscribe(function(e){t(new O(e?"open":"close",!0))})})},t.prototype.handleKeydown=function(t){t.keyCode!==f.g||this.disableClose||(this.close(),t.stopPropagation())},t.prototype._onAnimationStart=function(t){this._animationStarted.emit(t)},t.prototype._onAnimationEnd=function(t){var e=t.fromState,n=t.toState;0===n.indexOf("open")&&"void"===e?this.openedChange.emit(!0):"void"===n&&0===e.indexOf("open")&&this.openedChange.emit(!1)},Object.defineProperty(t.prototype,"_width",{get:function(){return this._elementRef.nativeElement?this._elementRef.nativeElement.offsetWidth||0:0},enumerable:!0,configurable:!0}),t.decorators=[{type:s.m,args:[{selector:"mat-drawer",exportAs:"matDrawer",template:"<ng-content></ng-content>",animations:[Object(p.k)("transform",[Object(p.h)("open, open-instant",Object(p.i)({transform:"translate3d(0, 0, 0)",visibility:"visible"})),Object(p.h)("void",Object(p.i)({visibility:"hidden"})),Object(p.j)("void => open-instant",Object(p.e)("0ms")),Object(p.j)("void <=> open, open-instant => void",Object(p.e)("400ms cubic-bezier(0.25, 0.8, 0.25, 1)"))])],host:{class:"mat-drawer","[@transform]":"_animationState","(@transform.start)":"_onAnimationStart($event)","(@transform.done)":"_onAnimationEnd($event)","(keydown)":"handleKeydown($event)","[attr.align]":"null","[class.mat-drawer-end]":'position === "end"',"[class.mat-drawer-over]":'mode === "over"',"[class.mat-drawer-push]":'mode === "push"',"[class.mat-drawer-side]":'mode === "side"',tabIndex:"-1"},changeDetection:s.i.OnPush,encapsulation:s._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[{type:s.t},{type:r.g},{type:r.f},{type:u.a},{type:void 0,decorators:[{type:s.M},{type:s.y,args:[o.b]}]}]},t.propDecorators={position:[{type:s.C}],align:[{type:s.C}],mode:[{type:s.C}],disableClose:[{type:s.C}],openedChange:[{type:s.N}],_openedStream:[{type:s.N,args:["opened"]}],openedStart:[{type:s.N}],_closedStream:[{type:s.N,args:["closed"]}],closedStart:[{type:s.N}],onOpen:[{type:s.N,args:["open"]}],onClose:[{type:s.N,args:["close"]}],onPositionChanged:[{type:s.N,args:["positionChanged"]}],onAlignChanged:[{type:s.N,args:["align-changed"]}],opened:[{type:s.C}]},t}(),S=function(){function t(t,e,n,r,i){void 0===i&&(i=!1);var o=this;this._dir=t,this._element=e,this._ngZone=n,this._changeDetectorRef=r,this.backdropClick=new s.w,this._destroyed=new x.a,this._doCheckSubject=new x.a,this._contentMargins=new x.a,null!=t&&t.change.pipe(Object(b.a)(this._destroyed)).subscribe(function(){return o._validateDrawers()}),this._autosize=i}return Object.defineProperty(t.prototype,"start",{get:function(){return this._start},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"end",{get:function(){return this._end},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"autosize",{get:function(){return this._autosize},set:function(t){this._autosize=Object(h.b)(t)},enumerable:!0,configurable:!0}),t.prototype.ngAfterContentInit=function(){var t=this;this._drawers.changes.pipe(Object(y.a)(null)).subscribe(function(){t._validateDrawers(),t._drawers.forEach(function(e){t._watchDrawerToggle(e),t._watchDrawerPosition(e),t._watchDrawerMode(e)}),(!t._drawers.length||t._isDrawerOpen(t._start)||t._isDrawerOpen(t._end))&&t._updateContentMargins(),t._changeDetectorRef.markForCheck()}),this._doCheckSubject.pipe(Object(_.a)(10),Object(b.a)(this._destroyed)).subscribe(function(){return t._updateContentMargins()})},t.prototype.ngOnDestroy=function(){this._doCheckSubject.complete(),this._destroyed.next(),this._destroyed.complete()},t.prototype.open=function(){this._drawers.forEach(function(t){return t.open()})},t.prototype.close=function(){this._drawers.forEach(function(t){return t.close()})},t.prototype.ngDoCheck=function(){var t=this;this._autosize&&this._isPushed()&&this._ngZone.runOutsideAngular(function(){return t._doCheckSubject.next()})},t.prototype._watchDrawerToggle=function(t){var e=this;t._animationStarted.pipe(Object(b.a)(this._drawers.changes),Object(g.a)(function(t){return t.fromState!==t.toState})).subscribe(function(t){"open-instant"!==t.toState&&e._element.nativeElement.classList.add("mat-drawer-transition"),e._updateContentMargins(),e._changeDetectorRef.markForCheck()}),"side"!==t.mode&&t.openedChange.pipe(Object(b.a)(this._drawers.changes)).subscribe(function(){return e._setContainerClass(t.opened)})},t.prototype._watchDrawerPosition=function(t){var e=this;t&&t.onPositionChanged.pipe(Object(b.a)(this._drawers.changes)).subscribe(function(){e._ngZone.onMicrotaskEmpty.asObservable().pipe(Object(v.a)(1)).subscribe(function(){e._validateDrawers()})})},t.prototype._watchDrawerMode=function(t){var e=this;t&&t._modeChanged.pipe(Object(b.a)(Object(m.a)(this._drawers.changes,this._destroyed))).subscribe(function(){e._updateContentMargins(),e._changeDetectorRef.markForCheck()})},t.prototype._setContainerClass=function(t){t?this._element.nativeElement.classList.add("mat-drawer-opened"):this._element.nativeElement.classList.remove("mat-drawer-opened")},t.prototype._validateDrawers=function(){var t=this;this._start=this._end=null,this._drawers.forEach(function(e){"end"==e.position?(null!=t._end&&w("end"),t._end=e):(null!=t._start&&w("start"),t._start=e)}),this._right=this._left=null,null==this._dir||"ltr"==this._dir.value?(this._left=this._start,this._right=this._end):(this._left=this._end,this._right=this._start)},t.prototype._isPushed=function(){return this._isDrawerOpen(this._start)&&"over"!=this._start.mode||this._isDrawerOpen(this._end)&&"over"!=this._end.mode},t.prototype._onBackdropClicked=function(){this.backdropClick.emit(),this._closeModalDrawer()},t.prototype._closeModalDrawer=function(){[this._start,this._end].filter(function(t){return t&&!t.disableClose&&"side"!==t.mode}).forEach(function(t){return t.close()})},t.prototype._isShowingBackdrop=function(){return this._isDrawerOpen(this._start)&&"side"!=this._start.mode||this._isDrawerOpen(this._end)&&"side"!=this._end.mode},t.prototype._isDrawerOpen=function(t){return null!=t&&t.opened},t.prototype._updateContentMargins=function(){var t=this,e=0,n=0;if(this._left&&this._left.opened)if("side"==this._left.mode)e+=this._left._width;else if("push"==this._left.mode){var r=this._left._width;e+=r,n-=r}if(this._right&&this._right.opened)if("side"==this._right.mode)n+=this._right._width;else if("push"==this._right.mode){r=this._right._width;n+=r,e-=r}this._ngZone.run(function(){return t._contentMargins.next({left:e,right:n})})},t.decorators=[{type:s.m,args:[{selector:"mat-drawer-container",exportAs:"matDrawerContainer",template:'<div class="mat-drawer-backdrop" (click)="_onBackdropClicked()" [class.mat-drawer-shown]="_isShowingBackdrop()"></div><ng-content select="mat-drawer"></ng-content><ng-content select="mat-drawer-content"></ng-content><mat-drawer-content *ngIf="!_content" cdkScrollable><ng-content></ng-content></mat-drawer-content>',styles:[".mat-drawer-container{position:relative;z-index:1;box-sizing:border-box;-webkit-overflow-scrolling:touch;display:block;overflow:hidden}.mat-drawer-container[fullscreen]{top:0;left:0;right:0;bottom:0;position:absolute}.mat-drawer-container[fullscreen].mat-drawer-opened{overflow:hidden}.mat-drawer-backdrop{top:0;left:0;right:0;bottom:0;position:absolute;display:block;z-index:3;visibility:hidden}.mat-drawer-backdrop.mat-drawer-shown{visibility:visible}.mat-drawer-transition .mat-drawer-backdrop{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:background-color,visibility}@media screen and (-ms-high-contrast:active){.mat-drawer-backdrop{opacity:.5}}.mat-drawer-content{-webkit-backface-visibility:hidden;backface-visibility:hidden;position:relative;z-index:1;display:block;height:100%;overflow:auto}.mat-drawer-transition .mat-drawer-content{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:transform,margin-left,margin-right}.mat-drawer{position:relative;z-index:4;display:block;position:absolute;top:0;bottom:0;z-index:3;outline:0;box-sizing:border-box;overflow-y:auto;transform:translate3d(-100%,0,0)}.mat-drawer.mat-drawer-side{z-index:2}.mat-drawer.mat-drawer-end{right:0;transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer{transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer.mat-drawer-end{left:0;right:auto;transform:translate3d(-100%,0,0)}.mat-drawer:not(.mat-drawer-side){box-shadow:0 8px 10px -5px rgba(0,0,0,.2),0 16px 24px 2px rgba(0,0,0,.14),0 6px 30px 5px rgba(0,0,0,.12)}.mat-sidenav-fixed{position:fixed}"],host:{class:"mat-drawer-container"},changeDetection:s.i.OnPush,encapsulation:s._10.None,preserveWhitespaces:!1}]}],t.ctorParameters=function(){return[{type:d.b,decorators:[{type:s.M}]},{type:s.t},{type:s.L},{type:s.j},{type:void 0,decorators:[{type:s.y,args:[C]}]}]},t.propDecorators={_drawers:[{type:s.r,args:[E]}],_content:[{type:s.q,args:[k]}],autosize:[{type:s.C}],backdropClick:[{type:s.N}]},t}(),P=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(l.b)(e,t),e.decorators=[{type:s.m,args:[{selector:"mat-sidenav-content",template:"<ng-content></ng-content>",host:{class:"mat-drawer-content mat-sidenav-content","[style.margin-left.px]":"_margins.left","[style.margin-right.px]":"_margins.right"},changeDetection:s.i.OnPush,encapsulation:s._10.None,preserveWhitespaces:!1}]}],e.ctorParameters=function(){return[{type:s.j},{type:A,decorators:[{type:s.y,args:[Object(s._14)(function(){return A})]}]}]},e}(k),M=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._fixedInViewport=!1,e._fixedTopGap=0,e._fixedBottomGap=0,e}return Object(l.b)(e,t),Object.defineProperty(e.prototype,"fixedInViewport",{get:function(){return this._fixedInViewport},set:function(t){this._fixedInViewport=Object(h.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"fixedTopGap",{get:function(){return this._fixedTopGap},set:function(t){this._fixedTopGap=Object(h.c)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"fixedBottomGap",{get:function(){return this._fixedBottomGap},set:function(t){this._fixedBottomGap=Object(h.c)(t)},enumerable:!0,configurable:!0}),e.decorators=[{type:s.m,args:[{selector:"mat-sidenav",exportAs:"matSidenav",template:"<ng-content></ng-content>",animations:[Object(p.k)("transform",[Object(p.h)("open, open-instant",Object(p.i)({transform:"translate3d(0, 0, 0)",visibility:"visible"})),Object(p.h)("void",Object(p.i)({visibility:"hidden"})),Object(p.j)("void => open-instant",Object(p.e)("0ms")),Object(p.j)("void <=> open, open-instant => void",Object(p.e)("400ms cubic-bezier(0.25, 0.8, 0.25, 1)"))])],host:{class:"mat-drawer mat-sidenav",tabIndex:"-1","[@transform]":"_animationState","(@transform.start)":"_onAnimationStart($event)","(@transform.done)":"_onAnimationEnd($event)","(keydown)":"handleKeydown($event)","[attr.align]":"null","[class.mat-drawer-end]":'position === "end"',"[class.mat-drawer-over]":'mode === "over"',"[class.mat-drawer-push]":'mode === "push"',"[class.mat-drawer-side]":'mode === "side"',"[class.mat-sidenav-fixed]":"fixedInViewport","[style.top.px]":"fixedInViewport ? fixedTopGap : null","[style.bottom.px]":"fixedInViewport ? fixedBottomGap : null"},changeDetection:s.i.OnPush,encapsulation:s._10.None,preserveWhitespaces:!1}]}],e.ctorParameters=function(){return[]},e.propDecorators={fixedInViewport:[{type:s.C}],fixedTopGap:[{type:s.C}],fixedBottomGap:[{type:s.C}]},e}(E),A=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(l.b)(e,t),e.decorators=[{type:s.m,args:[{selector:"mat-sidenav-container",exportAs:"matSidenavContainer",template:'<div class="mat-drawer-backdrop" (click)="_onBackdropClicked()" [class.mat-drawer-shown]="_isShowingBackdrop()"></div><ng-content select="mat-sidenav"></ng-content><ng-content select="mat-sidenav-content"></ng-content><mat-sidenav-content *ngIf="!_content" cdkScrollable><ng-content></ng-content></mat-sidenav-content>',styles:[".mat-drawer-container{position:relative;z-index:1;box-sizing:border-box;-webkit-overflow-scrolling:touch;display:block;overflow:hidden}.mat-drawer-container[fullscreen]{top:0;left:0;right:0;bottom:0;position:absolute}.mat-drawer-container[fullscreen].mat-drawer-opened{overflow:hidden}.mat-drawer-backdrop{top:0;left:0;right:0;bottom:0;position:absolute;display:block;z-index:3;visibility:hidden}.mat-drawer-backdrop.mat-drawer-shown{visibility:visible}.mat-drawer-transition .mat-drawer-backdrop{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:background-color,visibility}@media screen and (-ms-high-contrast:active){.mat-drawer-backdrop{opacity:.5}}.mat-drawer-content{-webkit-backface-visibility:hidden;backface-visibility:hidden;position:relative;z-index:1;display:block;height:100%;overflow:auto}.mat-drawer-transition .mat-drawer-content{transition-duration:.4s;transition-timing-function:cubic-bezier(.25,.8,.25,1);transition-property:transform,margin-left,margin-right}.mat-drawer{position:relative;z-index:4;display:block;position:absolute;top:0;bottom:0;z-index:3;outline:0;box-sizing:border-box;overflow-y:auto;transform:translate3d(-100%,0,0)}.mat-drawer.mat-drawer-side{z-index:2}.mat-drawer.mat-drawer-end{right:0;transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer{transform:translate3d(100%,0,0)}[dir=rtl] .mat-drawer.mat-drawer-end{left:0;right:auto;transform:translate3d(-100%,0,0)}.mat-drawer:not(.mat-drawer-side){box-shadow:0 8px 10px -5px rgba(0,0,0,.2),0 16px 24px 2px rgba(0,0,0,.14),0 6px 30px 5px rgba(0,0,0,.12)}.mat-sidenav-fixed{position:fixed}"],host:{class:"mat-drawer-container mat-sidenav-container"},changeDetection:s.i.OnPush,encapsulation:s._10.None,preserveWhitespaces:!1}]}],e.ctorParameters=function(){return[]},e.propDecorators={_drawers:[{type:s.r,args:[M]}],_content:[{type:s.q,args:[P]}]},e}(S),T=function(){function t(){}return t.decorators=[{type:s.H,args:[{imports:[o.a,a.j,r.a,i.d,c.a,u.b],exports:[a.j,E,S,k,M,A,P],declarations:[E,S,k,M,A,P],providers:[{provide:C,useValue:!1}]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/slide-toggle.es5.js":function(t,e,n){"use strict";var r=n("../../../cdk/esm5/observers.es5.js"),i=n("../../../cdk/esm5/platform.es5.js"),o=n("../../../core/esm5/core.js"),s=n("../../../material/esm5/core.es5.js"),a=n("../../../platform-browser/esm5/platform-browser.js"),c=n("../../../cdk/esm5/a11y.es5.js"),u=n("../../../../tslib/tslib.es6.js"),l=n("../../../cdk/esm5/coercion.es5.js"),p=0,d={provide:n("../../../forms/esm5/forms.js").d,useExisting:Object(o._14)(function(){return m}),multi:!0},h=function(){return function(){}}(),f=function(){return function(t){this._elementRef=t}}(),m=function(t){function e(e,n,r,i,s){var a=t.call(this,e)||this;return a._platform=n,a._focusMonitor=r,a._changeDetectorRef=i,a.onChange=function(t){},a.onTouched=function(){},a._uniqueId="mat-slide-toggle-"+ ++p,a._required=!1,a._checked=!1,a.name=null,a.id=a._uniqueId,a.labelPosition="after",a.ariaLabel=null,a.ariaLabelledby=null,a.change=new o.w,a._rippleConfig={centered:!0,radius:23,speedFactor:1.5},a.tabIndex=parseInt(s)||0,a}return Object(u.b)(e,t),Object.defineProperty(e.prototype,"required",{get:function(){return this._required},set:function(t){this._required=Object(l.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"checked",{get:function(){return this._checked},set:function(t){this._checked=Object(l.b)(t),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"inputId",{get:function(){return(this.id||this._uniqueId)+"-input"},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentInit=function(){var t=this;this._slideRenderer=new g(this._elementRef,this._platform),this._focusMonitor.monitor(this._inputElement.nativeElement,!1).subscribe(function(e){return t._onInputFocusChange(e)})},e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._inputElement.nativeElement)},e.prototype._onChangeEvent=function(t){t.stopPropagation(),this._slideRenderer.dragging?this._inputElement.nativeElement.checked=this.checked:(this.checked=this._inputElement.nativeElement.checked,this._emitChangeEvent())},e.prototype._onInputClick=function(t){t.stopPropagation()},e.prototype.writeValue=function(t){this.checked=!!t},e.prototype.registerOnChange=function(t){this.onChange=t},e.prototype.registerOnTouched=function(t){this.onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t,this._changeDetectorRef.markForCheck()},e.prototype.focus=function(){this._focusMonitor.focusVia(this._inputElement.nativeElement,"keyboard")},e.prototype.toggle=function(){this.checked=!this.checked},e.prototype._onInputFocusChange=function(t){this._focusRipple||"keyboard"!==t?t||(this.onTouched(),this._focusRipple&&(this._focusRipple.fadeOut(),this._focusRipple=null)):this._focusRipple=this._ripple.launch(0,0,Object(u.a)({persistent:!0},this._rippleConfig))},e.prototype._emitChangeEvent=function(){var t=new h;t.source=this,t.checked=this.checked,this.onChange(this.checked),this.change.emit(t)},e.prototype._onDragStart=function(){this.disabled||this._slideRenderer.startThumbDrag(this.checked)},e.prototype._onDrag=function(t){this._slideRenderer.dragging&&this._slideRenderer.updateThumbPosition(t.deltaX)},e.prototype._onDragEnd=function(){var t=this;if(this._slideRenderer.dragging){var e=this._slideRenderer.dragPercentage>50;e!==this.checked&&(this.checked=e,this._emitChangeEvent()),setTimeout(function(){return t._slideRenderer.stopThumbDrag()})}},e.prototype._onLabelTextChange=function(){this._changeDetectorRef.markForCheck()},e.decorators=[{type:o.m,args:[{selector:"mat-slide-toggle",exportAs:"matSlideToggle",host:{class:"mat-slide-toggle","[id]":"id","[class.mat-checked]":"checked","[class.mat-disabled]":"disabled","[class.mat-slide-toggle-label-before]":'labelPosition == "before"'},template:'<label class="mat-slide-toggle-label" #label><div class="mat-slide-toggle-bar" [class.mat-slide-toggle-bar-no-side-margin]="!labelContent.textContent || !labelContent.textContent.trim()"><input #input class="mat-slide-toggle-input cdk-visually-hidden" type="checkbox" [id]="inputId" [required]="required" [tabIndex]="tabIndex" [checked]="checked" [disabled]="disabled" [attr.name]="name" [attr.aria-label]="ariaLabel" [attr.aria-labelledby]="ariaLabelledby" (change)="_onChangeEvent($event)" (click)="_onInputClick($event)"><div class="mat-slide-toggle-thumb-container" (slidestart)="_onDragStart()" (slide)="_onDrag($event)" (slideend)="_onDragEnd()"><div class="mat-slide-toggle-thumb"></div><div class="mat-slide-toggle-ripple" mat-ripple [matRippleTrigger]="label" [matRippleDisabled]="disableRipple || disabled" [matRippleCentered]="_rippleConfig.centered" [matRippleRadius]="_rippleConfig.radius" [matRippleSpeedFactor]="_rippleConfig.speedFactor"></div></div></div><span class="mat-slide-toggle-content" #labelContent (cdkObserveContent)="_onLabelTextChange()"><ng-content></ng-content></span></label>',styles:[".mat-slide-toggle{display:inline-block;height:24px;line-height:24px;white-space:nowrap;-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;outline:0}.mat-slide-toggle.mat-checked .mat-slide-toggle-thumb-container{transform:translate3d(16px,0,0)}.mat-slide-toggle.mat-disabled .mat-slide-toggle-label,.mat-slide-toggle.mat-disabled .mat-slide-toggle-thumb-container{cursor:default}.mat-slide-toggle-label{display:flex;flex:1;flex-direction:row;align-items:center;height:inherit;cursor:pointer}.mat-slide-toggle-label-before .mat-slide-toggle-label{order:1}.mat-slide-toggle-label-before .mat-slide-toggle-bar{order:2}.mat-slide-toggle-bar,[dir=rtl] .mat-slide-toggle-label-before .mat-slide-toggle-bar{margin-right:8px;margin-left:0}.mat-slide-toggle-label-before .mat-slide-toggle-bar,[dir=rtl] .mat-slide-toggle-bar{margin-left:8px;margin-right:0}.mat-slide-toggle-bar-no-side-margin{margin-left:0;margin-right:0}.mat-slide-toggle-thumb-container{position:absolute;z-index:1;width:20px;height:20px;top:-3px;left:0;transform:translate3d(0,0,0);transition:all 80ms linear;transition-property:transform;cursor:-webkit-grab;cursor:grab}.mat-slide-toggle-thumb-container.mat-dragging,.mat-slide-toggle-thumb-container:active{cursor:-webkit-grabbing;cursor:grabbing;transition-duration:0s}.mat-slide-toggle-thumb{height:20px;width:20px;border-radius:50%;box-shadow:0 2px 1px -1px rgba(0,0,0,.2),0 1px 1px 0 rgba(0,0,0,.14),0 1px 3px 0 rgba(0,0,0,.12)}@media screen and (-ms-high-contrast:active){.mat-slide-toggle-thumb{background:#fff;border:solid 1px #000}}.mat-slide-toggle-bar{position:relative;width:36px;height:14px;border-radius:8px}@media screen and (-ms-high-contrast:active){.mat-slide-toggle-bar{background:#fff}}.mat-slide-toggle-input{bottom:0;left:10px}.mat-slide-toggle-bar,.mat-slide-toggle-thumb{transition:all 80ms linear;transition-property:background-color;transition-delay:50ms}.mat-slide-toggle-ripple{position:absolute;top:-13px;left:-13px;height:46px;width:46px;z-index:1;pointer-events:none}"],providers:[d],inputs:["disabled","disableRipple","color","tabIndex"],encapsulation:o._10.None,preserveWhitespaces:!1,changeDetection:o.i.OnPush}]}],e.ctorParameters=function(){return[{type:o.t},{type:i.a},{type:c.f},{type:o.j},{type:void 0,decorators:[{type:o.g,args:["tabindex"]}]}]},e.propDecorators={name:[{type:o.C}],id:[{type:o.C}],labelPosition:[{type:o.C}],ariaLabel:[{type:o.C,args:["aria-label"]}],ariaLabelledby:[{type:o.C,args:["aria-labelledby"]}],required:[{type:o.C}],checked:[{type:o.C}],change:[{type:o.N}],_inputElement:[{type:o._7,args:["input"]}],_ripple:[{type:o._7,args:[s.r]}]},e}(Object(s.y)(Object(s.u)(Object(s.v)(Object(s.w)(f)),"accent"))),g=function(){function t(t,e){this.dragging=!1,e.isBrowser&&(this._thumbEl=t.nativeElement.querySelector(".mat-slide-toggle-thumb-container"),this._thumbBarEl=t.nativeElement.querySelector(".mat-slide-toggle-bar"))}return t.prototype.startThumbDrag=function(t){this.dragging||(this._thumbBarWidth=this._thumbBarEl.clientWidth-this._thumbEl.clientWidth,this._thumbEl.classList.add("mat-dragging"),this._previousChecked=t,this.dragging=!0)},t.prototype.stopThumbDrag=function(){return!!this.dragging&&(this.dragging=!1,this._thumbEl.classList.remove("mat-dragging"),Object(s.t)(this._thumbEl,""),this.dragPercentage>50)},t.prototype.updateThumbPosition=function(t){this.dragPercentage=this._getDragPercentage(t);var e=this.dragPercentage/100*this._thumbBarWidth;Object(s.t)(this._thumbEl,"translate3d("+e+"px, 0, 0)")},t.prototype._getDragPercentage=function(t){var e=t/this._thumbBarWidth*100;return this._previousChecked&&(e+=100),Math.max(0,Math.min(e,100))},t}();!function(){function t(){}t.decorators=[{type:o.H,args:[{imports:[s.s,s.j,i.b,r.a,c.a],exports:[m,s.j],declarations:[m],providers:[{provide:a.d,useClass:s.e}]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/slider.es5.js":function(t,e,n){"use strict";var r=n("../../../cdk/esm5/a11y.es5.js"),i=n("../../../cdk/esm5/bidi.es5.js"),o=n("../../../common/esm5/common.js"),s=n("../../../core/esm5/core.js"),a=n("../../../material/esm5/core.es5.js"),c=n("../../../platform-browser/esm5/platform-browser.js"),u=n("../../../../tslib/tslib.es6.js"),l=n("../../../cdk/esm5/coercion.es5.js"),p=n("../../../cdk/esm5/keycodes.es5.js"),d=n("../../../forms/esm5/forms.js"),h=n("../../../../rxjs/_esm5/Subscription.js"),f={provide:d.d,useExisting:Object(s._14)(function(){return v}),multi:!0},m=function(){return function(){}}(),g=function(){return function(t){this._elementRef=t}}(),v=function(t){function e(e,n,r,i,o){var a=t.call(this,e)||this;return a._focusMonitor=n,a._changeDetectorRef=r,a._dir=i,a._invert=!1,a._max=100,a._min=0,a._step=1,a._thumbLabel=!1,a._tickInterval=0,a._value=null,a._vertical=!1,a.change=new s.w,a.input=new s.w,a.onTouched=function(){},a._percent=0,a._isSliding=!1,a._isActive=!1,a._tickIntervalPercent=0,a._sliderDimensions=null,a._controlValueAccessorChangeFn=function(){},a._dirChangeSubscription=h.a.EMPTY,a.tabIndex=parseInt(o)||0,a}return Object(u.b)(e,t),Object.defineProperty(e.prototype,"invert",{get:function(){return this._invert},set:function(t){this._invert=Object(l.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"max",{get:function(){return this._max},set:function(t){this._max=Object(l.c)(t,this._max),this._percent=this._calculatePercentage(this._value),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"min",{get:function(){return this._min},set:function(t){this._min=Object(l.c)(t,this._min),null===this._value&&(this.value=this._min),this._percent=this._calculatePercentage(this._value),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"step",{get:function(){return this._step},set:function(t){this._step=Object(l.c)(t,this._step),this._step%1!=0&&(this._roundLabelTo=this._step.toString().split(".").pop().length),this._changeDetectorRef.markForCheck()},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"thumbLabel",{get:function(){return this._thumbLabel},set:function(t){this._thumbLabel=Object(l.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_thumbLabelDeprecated",{get:function(){return this._thumbLabel},set:function(t){this._thumbLabel=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"tickInterval",{get:function(){return this._tickInterval},set:function(t){this._tickInterval="auto"===t?"auto":"number"==typeof t||"string"==typeof t?Object(l.c)(t,this._tickInterval):0},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_tickIntervalDeprecated",{get:function(){return this.tickInterval},set:function(t){this.tickInterval=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"value",{get:function(){return null===this._value&&(this.value=this._min),this._value},set:function(t){t!==this._value&&(this._value=Object(l.c)(t,this._value||0),this._percent=this._calculatePercentage(this._value),this._changeDetectorRef.markForCheck())},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"vertical",{get:function(){return this._vertical},set:function(t){this._vertical=Object(l.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"displayValue",{get:function(){return this._roundLabelTo&&this.value&&this.value%1!=0?this.value.toFixed(this._roundLabelTo):this.value||0},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"percent",{get:function(){return this._clamp(this._percent)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_invertAxis",{get:function(){return this.vertical?!this.invert:this.invert},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_isMinValue",{get:function(){return 0===this.percent},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_thumbGap",{get:function(){return this.disabled?7:this._isMinValue&&!this.thumbLabel?this._isActive?10:7:0},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_trackBackgroundStyles",{get:function(){var t=this.vertical?"Y":"X";return{transform:"translate"+t+"("+(this._invertMouseCoords?"-":"")+this._thumbGap+"px) scale"+t+"("+(1-this.percent)+")"}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_trackFillStyles",{get:function(){var t=this.vertical?"Y":"X";return{transform:"translate"+t+"("+(this._invertMouseCoords?"":"-")+this._thumbGap+"px) scale"+t+"("+this.percent+")"}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_ticksContainerStyles",{get:function(){return{transform:"translate"+(this.vertical?"Y":"X")+"("+(this.vertical||"rtl"!=this._direction?"-":"")+this._tickIntervalPercent/2*100+"%)"}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_ticksStyles",{get:function(){var t=100*this._tickIntervalPercent,e={backgroundSize:this.vertical?"2px "+t+"%":t+"% 2px",transform:"translateZ(0) translate"+(this.vertical?"Y":"X")+"("+(this.vertical||"rtl"!=this._direction?"":"-")+t/2+"%)"+(this.vertical||"rtl"!=this._direction?"":" rotate(180deg)")};this._isMinValue&&this._thumbGap&&(e["padding"+(this.vertical?this._invertAxis?"Bottom":"Top":this._invertAxis?"Right":"Left")]=this._thumbGap+"px");return e},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_thumbContainerStyles",{get:function(){return{transform:"translate"+(this.vertical?"Y":"X")+"(-"+100*(("rtl"!=this._direction||this.vertical?this._invertAxis:!this._invertAxis)?this.percent:1-this.percent)+"%)"}},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_invertMouseCoords",{get:function(){return"rtl"!=this._direction||this.vertical?this._invertAxis:!this._invertAxis},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_direction",{get:function(){return this._dir&&"rtl"==this._dir.value?"rtl":"ltr"},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){var t=this;this._focusMonitor.monitor(this._elementRef.nativeElement,!0).subscribe(function(e){t._isActive=!!e&&"keyboard"!==e,t._changeDetectorRef.detectChanges()}),this._dir&&(this._dirChangeSubscription=this._dir.change.subscribe(function(){t._changeDetectorRef.markForCheck()}))},e.prototype.ngOnDestroy=function(){this._focusMonitor.stopMonitoring(this._elementRef.nativeElement),this._dirChangeSubscription.unsubscribe()},e.prototype._onMouseenter=function(){this.disabled||(this._sliderDimensions=this._getSliderDimensions(),this._updateTickIntervalPercent())},e.prototype._onClick=function(t){if(!this.disabled){var e=this.value;this._isSliding=!1,this._focusHostElement(),this._updateValueFromPosition({x:t.clientX,y:t.clientY}),e!=this.value&&(this._emitInputEvent(),this._emitChangeEvent())}},e.prototype._onSlide=function(t){if(!this.disabled){this._isSliding||this._onSlideStart(null),t.preventDefault();var e=this.value;this._updateValueFromPosition({x:t.center.x,y:t.center.y}),e!=this.value&&this._emitInputEvent()}},e.prototype._onSlideStart=function(t){this.disabled||this._isSliding||(this._onMouseenter(),this._isSliding=!0,this._focusHostElement(),this._valueOnSlideStart=this.value,t&&(this._updateValueFromPosition({x:t.center.x,y:t.center.y}),t.preventDefault()))},e.prototype._onSlideEnd=function(){this._isSliding=!1,this._valueOnSlideStart!=this.value&&this._emitChangeEvent(),this._valueOnSlideStart=null},e.prototype._onFocus=function(){this._sliderDimensions=this._getSliderDimensions(),this._updateTickIntervalPercent()},e.prototype._onBlur=function(){this.onTouched()},e.prototype._onKeydown=function(t){if(!this.disabled){var e=this.value;switch(t.keyCode){case p.l:this._increment(10);break;case p.k:this._increment(-10);break;case p.e:this.value=this.max;break;case p.h:this.value=this.min;break;case p.i:this._increment("rtl"==this._direction?1:-1);break;case p.p:this._increment(1);break;case p.m:this._increment("rtl"==this._direction?-1:1);break;case p.d:this._increment(-1);break;default:return}e!=this.value&&(this._emitInputEvent(),this._emitChangeEvent()),this._isSliding=!0,t.preventDefault()}},e.prototype._onKeyup=function(){this._isSliding=!1},e.prototype._increment=function(t){this.value=this._clamp((this.value||0)+this.step*t,this.min,this.max)},e.prototype._updateValueFromPosition=function(t){if(this._sliderDimensions){var e=this.vertical?this._sliderDimensions.top:this._sliderDimensions.left,n=this.vertical?this._sliderDimensions.height:this._sliderDimensions.width,r=this.vertical?t.y:t.x,i=this._clamp((r-e)/n);this._invertMouseCoords&&(i=1-i);var o=this._calculateValue(i),s=Math.round((o-this.min)/this.step)*this.step+this.min;this.value=this._clamp(s,this.min,this.max)}},e.prototype._emitChangeEvent=function(){this._controlValueAccessorChangeFn(this.value),this.change.emit(this._createChangeEvent())},e.prototype._emitInputEvent=function(){this.input.emit(this._createChangeEvent())},e.prototype._updateTickIntervalPercent=function(){if(this.tickInterval&&this._sliderDimensions)if("auto"==this.tickInterval){var t=this.vertical?this._sliderDimensions.height:this._sliderDimensions.width,e=t*this.step/(this.max-this.min),n=Math.ceil(30/e)*this.step;this._tickIntervalPercent=n/t}else this._tickIntervalPercent=this.tickInterval*this.step/(this.max-this.min)},e.prototype._createChangeEvent=function(t){void 0===t&&(t=this.value);var e=new m;return e.source=this,e.value=t,e},e.prototype._calculatePercentage=function(t){return((t||0)-this.min)/(this.max-this.min)},e.prototype._calculateValue=function(t){return this.min+t*(this.max-this.min)},e.prototype._clamp=function(t,e,n){return void 0===e&&(e=0),void 0===n&&(n=1),Math.max(e,Math.min(t,n))},e.prototype._getSliderDimensions=function(){return this._sliderWrapper?this._sliderWrapper.nativeElement.getBoundingClientRect():null},e.prototype._focusHostElement=function(){this._elementRef.nativeElement.focus()},e.prototype.writeValue=function(t){this.value=t},e.prototype.registerOnChange=function(t){this._controlValueAccessorChangeFn=t},e.prototype.registerOnTouched=function(t){this.onTouched=t},e.prototype.setDisabledState=function(t){this.disabled=t},e.decorators=[{type:s.m,args:[{selector:"mat-slider",exportAs:"matSlider",providers:[f],host:{"(focus)":"_onFocus()","(blur)":"_onBlur()","(click)":"_onClick($event)","(keydown)":"_onKeydown($event)","(keyup)":"_onKeyup()","(mouseenter)":"_onMouseenter()","(slide)":"_onSlide($event)","(slideend)":"_onSlideEnd()","(slidestart)":"_onSlideStart($event)",class:"mat-slider",role:"slider","[tabIndex]":"tabIndex","[attr.aria-disabled]":"disabled","[attr.aria-valuemax]":"max","[attr.aria-valuemin]":"min","[attr.aria-valuenow]":"value","[attr.aria-orientation]":'vertical ? "vertical" : "horizontal"',"[class.mat-slider-disabled]":"disabled","[class.mat-slider-has-ticks]":"tickInterval","[class.mat-slider-horizontal]":"!vertical","[class.mat-slider-axis-inverted]":"_invertAxis","[class.mat-slider-sliding]":"_isSliding","[class.mat-slider-thumb-label-showing]":"thumbLabel","[class.mat-slider-vertical]":"vertical","[class.mat-slider-min-value]":"_isMinValue","[class.mat-slider-hide-last-tick]":"disabled || _isMinValue && _thumbGap && _invertAxis"},template:'<div class="mat-slider-wrapper" #sliderWrapper><div class="mat-slider-track-wrapper"><div class="mat-slider-track-background" [ngStyle]="_trackBackgroundStyles"></div><div class="mat-slider-track-fill" [ngStyle]="_trackFillStyles"></div></div><div class="mat-slider-ticks-container" [ngStyle]="_ticksContainerStyles"><div class="mat-slider-ticks" [ngStyle]="_ticksStyles"></div></div><div class="mat-slider-thumb-container" [ngStyle]="_thumbContainerStyles"><div class="mat-slider-focus-ring"></div><div class="mat-slider-thumb"></div><div class="mat-slider-thumb-label"><span class="mat-slider-thumb-label-text">{{displayValue}}</span></div></div></div>',styles:[".mat-slider{display:inline-block;position:relative;box-sizing:border-box;padding:8px;outline:0;vertical-align:middle}.mat-slider-wrapper{position:absolute}.mat-slider-track-wrapper{position:absolute;top:0;left:0;overflow:hidden}.mat-slider-track-fill{position:absolute;transform-origin:0 0;transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-track-background{position:absolute;transform-origin:100% 100%;transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-ticks-container{position:absolute;left:0;top:0;overflow:hidden}.mat-slider-ticks{background-repeat:repeat;background-clip:content-box;box-sizing:border-box;opacity:0;transition:opacity .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-thumb-container{position:absolute;z-index:1;transition:transform .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-focus-ring{position:absolute;width:30px;height:30px;border-radius:50%;transform:scale(0);opacity:0;transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1),opacity .4s cubic-bezier(.25,.8,.25,1)}.cdk-keyboard-focused .mat-slider-focus-ring,.cdk-program-focused .mat-slider-focus-ring{transform:scale(1);opacity:1}.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb,.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb-label{cursor:-webkit-grab;cursor:grab}.mat-slider-sliding:not(.mat-slider-disabled) .mat-slider-thumb,.mat-slider-sliding:not(.mat-slider-disabled) .mat-slider-thumb-label,.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb-label:active,.mat-slider:not(.mat-slider-disabled) .mat-slider-thumb:active{cursor:-webkit-grabbing;cursor:grabbing}.mat-slider-thumb{position:absolute;right:-10px;bottom:-10px;box-sizing:border-box;width:20px;height:20px;border:3px solid transparent;border-radius:50%;transform:scale(.7);transition:transform .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1),border-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-thumb-label{display:none;align-items:center;justify-content:center;position:absolute;width:28px;height:28px;border-radius:50%;transition:transform .4s cubic-bezier(.25,.8,.25,1),border-radius .4s cubic-bezier(.25,.8,.25,1),background-color .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-thumb-label-text{z-index:1;opacity:0;transition:opacity .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-sliding .mat-slider-thumb-container,.mat-slider-sliding .mat-slider-track-background,.mat-slider-sliding .mat-slider-track-fill{transition-duration:0s}.mat-slider-has-ticks .mat-slider-wrapper::after{content:'';position:absolute;border-width:0;border-style:solid;opacity:0;transition:opacity .4s cubic-bezier(.25,.8,.25,1)}.mat-slider-has-ticks.cdk-focused:not(.mat-slider-hide-last-tick) .mat-slider-wrapper::after,.mat-slider-has-ticks:hover:not(.mat-slider-hide-last-tick) .mat-slider-wrapper::after{opacity:1}.mat-slider-has-ticks.cdk-focused:not(.mat-slider-disabled) .mat-slider-ticks,.mat-slider-has-ticks:hover:not(.mat-slider-disabled) .mat-slider-ticks{opacity:1}.mat-slider-thumb-label-showing .mat-slider-focus-ring{transform:scale(0);opacity:0}.mat-slider-thumb-label-showing .mat-slider-thumb-label{display:flex}.mat-slider-axis-inverted .mat-slider-track-fill{transform-origin:100% 100%}.mat-slider-axis-inverted .mat-slider-track-background{transform-origin:0 0}.mat-slider:not(.mat-slider-disabled).cdk-focused.mat-slider-thumb-label-showing .mat-slider-thumb{transform:scale(0)}.mat-slider:not(.mat-slider-disabled).cdk-focused .mat-slider-thumb-label{border-radius:50% 50% 0}.mat-slider:not(.mat-slider-disabled).cdk-focused .mat-slider-thumb-label-text{opacity:1}.mat-slider:not(.mat-slider-disabled).cdk-mouse-focused .mat-slider-thumb,.mat-slider:not(.mat-slider-disabled).cdk-program-focused .mat-slider-thumb,.mat-slider:not(.mat-slider-disabled).cdk-touch-focused .mat-slider-thumb{border-width:2px;transform:scale(1)}.mat-slider-disabled .mat-slider-focus-ring{transform:scale(0);opacity:0}.mat-slider-disabled .mat-slider-thumb{border-width:4px;transform:scale(.5)}.mat-slider-disabled .mat-slider-thumb-label{display:none}.mat-slider-horizontal{height:48px;min-width:128px}.mat-slider-horizontal .mat-slider-wrapper{height:2px;top:23px;left:8px;right:8px}.mat-slider-horizontal .mat-slider-wrapper::after{height:2px;border-left-width:2px;right:0;top:0}.mat-slider-horizontal .mat-slider-track-wrapper{height:2px;width:100%}.mat-slider-horizontal .mat-slider-track-fill{height:2px;width:100%;transform:scaleX(0)}.mat-slider-horizontal .mat-slider-track-background{height:2px;width:100%;transform:scaleX(1)}.mat-slider-horizontal .mat-slider-ticks-container{height:2px;width:100%}.mat-slider-horizontal .mat-slider-ticks{height:2px;width:100%}.mat-slider-horizontal .mat-slider-thumb-container{width:100%;height:0;top:50%}.mat-slider-horizontal .mat-slider-focus-ring{top:-15px;right:-15px}.mat-slider-horizontal .mat-slider-thumb-label{right:-14px;top:-40px;transform:translateY(26px) scale(.01) rotate(45deg)}.mat-slider-horizontal .mat-slider-thumb-label-text{transform:rotate(-45deg)}.mat-slider-horizontal.cdk-focused .mat-slider-thumb-label{transform:rotate(45deg)}.mat-slider-vertical{width:48px;min-height:128px}.mat-slider-vertical .mat-slider-wrapper{width:2px;top:8px;bottom:8px;left:23px}.mat-slider-vertical .mat-slider-wrapper::after{width:2px;border-top-width:2px;bottom:0;left:0}.mat-slider-vertical .mat-slider-track-wrapper{height:100%;width:2px}.mat-slider-vertical .mat-slider-track-fill{height:100%;width:2px;transform:scaleY(0)}.mat-slider-vertical .mat-slider-track-background{height:100%;width:2px;transform:scaleY(1)}.mat-slider-vertical .mat-slider-ticks-container{width:2px;height:100%}.mat-slider-vertical .mat-slider-focus-ring{bottom:-15px;left:-15px}.mat-slider-vertical .mat-slider-ticks{width:2px;height:100%}.mat-slider-vertical .mat-slider-thumb-container{height:100%;width:0;left:50%}.mat-slider-vertical .mat-slider-thumb-label{bottom:-14px;left:-40px;transform:translateX(26px) scale(.01) rotate(-45deg)}.mat-slider-vertical .mat-slider-thumb-label-text{transform:rotate(45deg)}.mat-slider-vertical.cdk-focused .mat-slider-thumb-label{transform:rotate(-45deg)}[dir=rtl] .mat-slider-wrapper::after{left:0;right:auto}[dir=rtl] .mat-slider-horizontal .mat-slider-track-fill{transform-origin:100% 100%}[dir=rtl] .mat-slider-horizontal .mat-slider-track-background{transform-origin:0 0}[dir=rtl] .mat-slider-horizontal.mat-slider-axis-inverted .mat-slider-track-fill{transform-origin:0 0}[dir=rtl] .mat-slider-horizontal.mat-slider-axis-inverted .mat-slider-track-background{transform-origin:100% 100%}"],inputs:["disabled","color","tabIndex"],encapsulation:s._10.None,preserveWhitespaces:!1,changeDetection:s.i.OnPush}]}],e.ctorParameters=function(){return[{type:s.t},{type:r.f},{type:s.j},{type:i.b,decorators:[{type:s.M}]},{type:void 0,decorators:[{type:s.g,args:["tabindex"]}]}]},e.propDecorators={invert:[{type:s.C}],max:[{type:s.C}],min:[{type:s.C}],step:[{type:s.C}],thumbLabel:[{type:s.C}],_thumbLabelDeprecated:[{type:s.C,args:["thumb-label"]}],tickInterval:[{type:s.C}],_tickIntervalDeprecated:[{type:s.C,args:["tick-interval"]}],value:[{type:s.C}],vertical:[{type:s.C}],change:[{type:s.N}],input:[{type:s.N}],_sliderWrapper:[{type:s._7,args:["sliderWrapper"]}]},e}(Object(a.y)(Object(a.u)(Object(a.w)(g),"accent")));!function(){function t(){}t.decorators=[{type:s.H,args:[{imports:[o.a,a.j,i.a,r.a],exports:[v,a.j],declarations:[v],providers:[{provide:c.d,useClass:a.e}]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/snack-bar.es5.js":function(t,e,n){"use strict";var r=n("../../../core/esm5/core.js"),i=n("../../../common/esm5/common.js"),o=n("../../../cdk/esm5/overlay.es5.js"),s=n("../../../cdk/esm5/portal.es5.js"),a=n("../../../cdk/esm5/a11y.es5.js"),c=n("../../../cdk/esm5/layout.es5.js"),u=n("../../../material/esm5/core.es5.js"),l=n("../../../../tslib/tslib.es6.js"),p=n("../../../../rxjs/_esm5/operators/take.js"),d=n("../../../../rxjs/_esm5/operators/takeUntil.js"),h=n("../../../animations/esm5/animations.js"),f=n("../../../../rxjs/_esm5/Subject.js"),m=function(){function t(t,e){var n=this;this._overlayRef=e,this._afterClosed=new f.a,this._afterOpened=new f.a,this._onAction=new f.a,this.containerInstance=t,this.onAction().subscribe(function(){return n.dismiss()}),t._onExit.subscribe(function(){return n._finishDismiss()})}return t.prototype.dismiss=function(){this._afterClosed.closed||this.containerInstance.exit(),clearTimeout(this._durationTimeoutId)},t.prototype.closeWithAction=function(){this._onAction.closed||(this._onAction.next(),this._onAction.complete())},t.prototype._dismissAfter=function(t){var e=this;this._durationTimeoutId=setTimeout(function(){return e.dismiss()},t)},t.prototype._open=function(){this._afterOpened.closed||(this._afterOpened.next(),this._afterOpened.complete())},t.prototype._finishDismiss=function(){this._overlayRef.dispose(),this._onAction.closed||this._onAction.complete(),this._afterClosed.next(),this._afterClosed.complete()},t.prototype.afterDismissed=function(){return this._afterClosed.asObservable()},t.prototype.afterOpened=function(){return this.containerInstance._onEnter},t.prototype.onAction=function(){return this._onAction.asObservable()},t}(),g=new r.A("MatSnackBarData"),v=function(){return function(){this.politeness="assertive",this.announcementMessage="",this.duration=0,this.direction="ltr",this.data=null,this.horizontalPosition="center",this.verticalPosition="bottom"}}(),y=function(){function t(t,e){this.snackBarRef=t,this.data=e}return t.prototype.action=function(){this.snackBarRef.closeWithAction()},Object.defineProperty(t.prototype,"hasAction",{get:function(){return!!this.data.action},enumerable:!0,configurable:!0}),t.decorators=[{type:r.m,args:[{selector:"simple-snack-bar",template:'{{data.message}} <button class="mat-simple-snackbar-action" *ngIf="hasAction" (click)="action()">{{data.action}}</button>',styles:[".mat-simple-snackbar{display:flex;justify-content:space-between;line-height:20px;opacity:1}.mat-simple-snackbar-action{-webkit-user-select:none;-moz-user-select:none;-ms-user-select:none;user-select:none;cursor:pointer;outline:0;border:none;-webkit-tap-highlight-color:transparent;background:0 0;flex-shrink:0;margin-left:48px}[dir=rtl] .mat-simple-snackbar-action{margin-right:48px;margin-left:0}"],encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush,animations:[Object(h.k)("contentFade",[Object(h.j)(":enter",[Object(h.i)({opacity:"0"}),Object(h.e)(u.b.COMPLEX+" "+u.a.STANDARD_CURVE)])])],host:{"[@contentFade]":"",class:"mat-simple-snackbar"}}]}],t.ctorParameters=function(){return[{type:m},{type:void 0,decorators:[{type:r.y,args:[g]}]}]},t}(),b=u.b.ENTERING+" "+u.a.DECELERATION_CURVE,_=u.b.EXITING+" "+u.a.ACCELERATION_CURVE,j=function(t){function e(e,n,r){var i=t.call(this)||this;return i._ngZone=e,i._elementRef=n,i._changeDetectorRef=r,i._destroyed=!1,i._onExit=new f.a,i._onEnter=new f.a,i._animationState="void",i}return Object(l.b)(e,t),e.prototype.attachComponentPortal=function(t){if(this._portalOutlet.hasAttached())throw Error("Attempting to attach snack bar content after content is already attached");var e=this._elementRef.nativeElement;return(this.snackBarConfig.panelClass||this.snackBarConfig.extraClasses)&&(this._setCssClasses(this.snackBarConfig.panelClass),this._setCssClasses(this.snackBarConfig.extraClasses)),"center"===this.snackBarConfig.horizontalPosition&&e.classList.add("mat-snack-bar-center"),"top"===this.snackBarConfig.verticalPosition&&e.classList.add("mat-snack-bar-top"),this._portalOutlet.attachComponentPortal(t)},e.prototype.attachTemplatePortal=function(){throw Error("Not yet implemented")},e.prototype.onAnimationEnd=function(t){var e=t.fromState,n=t.toState;if(("void"===n&&"void"!==e||n.startsWith("hidden"))&&this._completeExit(),n.startsWith("visible")){var r=this._onEnter;this._ngZone.run(function(){r.next(),r.complete()})}},e.prototype.enter=function(){this._destroyed||(this._animationState="visible-"+this.snackBarConfig.verticalPosition,this._changeDetectorRef.detectChanges())},e.prototype.exit=function(){return this._animationState="hidden-"+this.snackBarConfig.verticalPosition,this._onExit},e.prototype.ngOnDestroy=function(){this._destroyed=!0,this._completeExit()},e.prototype._completeExit=function(){var t=this;this._ngZone.onMicrotaskEmpty.asObservable().pipe(Object(p.a)(1)).subscribe(function(){t._onExit.next(),t._onExit.complete()})},e.prototype._setCssClasses=function(t){if(t){var e=this._elementRef.nativeElement;Array.isArray(t)?t.forEach(function(t){return e.classList.add(t)}):e.classList.add(t)}},e.decorators=[{type:r.m,args:[{selector:"snack-bar-container",template:"<ng-template cdkPortalOutlet></ng-template>",styles:[".mat-snack-bar-container{border-radius:2px;box-sizing:border-box;display:block;margin:24px;max-width:568px;min-width:288px;padding:14px 24px;transform:translateY(100%) translateY(24px)}.mat-snack-bar-container.mat-snack-bar-center{margin:0;transform:translateY(100%)}.mat-snack-bar-container.mat-snack-bar-top{transform:translateY(-100%) translateY(-24px)}.mat-snack-bar-container.mat-snack-bar-top.mat-snack-bar-center{transform:translateY(-100%)}@media screen and (-ms-high-contrast:active){.mat-snack-bar-container{border:solid 1px}}.mat-snack-bar-handset{width:100%}.mat-snack-bar-handset .mat-snack-bar-container{margin:0;max-width:inherit;width:100%}"],changeDetection:r.i.OnPush,encapsulation:r._10.None,preserveWhitespaces:!1,host:{role:"alert",class:"mat-snack-bar-container","[@state]":"_animationState","(@state.done)":"onAnimationEnd($event)"},animations:[Object(h.k)("state",[Object(h.h)("visible-top, visible-bottom",Object(h.i)({transform:"translateY(0%)"})),Object(h.j)("visible-top => hidden-top, visible-bottom => hidden-bottom",Object(h.e)(_)),Object(h.j)("void => visible-top, void => visible-bottom",Object(h.e)(b))])]}]}],e.ctorParameters=function(){return[{type:r.L},{type:r.t},{type:r.j}]},e.propDecorators={_portalOutlet:[{type:r._7,args:[s.c]}]},e}(s.a),x=function(){function t(t,e,n,r,i){this._overlay=t,this._live=e,this._injector=n,this._breakpointObserver=r,this._parentSnackBar=i,this._snackBarRefAtThisLevel=null}return Object.defineProperty(t.prototype,"_openedSnackBarRef",{get:function(){var t=this._parentSnackBar;return t?t._openedSnackBarRef:this._snackBarRefAtThisLevel},set:function(t){this._parentSnackBar?this._parentSnackBar._openedSnackBarRef=t:this._snackBarRefAtThisLevel=t},enumerable:!0,configurable:!0}),t.prototype.openFromComponent=function(t,e){var n=this,r=w(e),i=this._attach(t,r);return i.afterDismissed().subscribe(function(){n._openedSnackBarRef==i&&(n._openedSnackBarRef=null)}),this._openedSnackBarRef?(this._openedSnackBarRef.afterDismissed().subscribe(function(){i.containerInstance.enter()}),this._openedSnackBarRef.dismiss()):i.containerInstance.enter(),r.duration&&r.duration>0&&i.afterOpened().subscribe(function(){return i._dismissAfter(r.duration)}),r.announcementMessage&&this._live.announce(r.announcementMessage,r.politeness),this._openedSnackBarRef=i,this._openedSnackBarRef},t.prototype.open=function(t,e,n){void 0===e&&(e="");var r=w(n);return r.data={message:t,action:e},r.announcementMessage=t,this.openFromComponent(y,r)},t.prototype.dismiss=function(){this._openedSnackBarRef&&this._openedSnackBarRef.dismiss()},t.prototype._attachSnackBarContainer=function(t,e){var n=new s.d(j,e.viewContainerRef),r=t.attach(n);return r.instance.snackBarConfig=e,r.instance},t.prototype._attach=function(t,e){var n=this._createOverlay(e),r=this._attachSnackBarContainer(n,e),i=new m(r,n),o=this._createInjector(e,i),a=new s.d(t,void 0,o),u=r.attachComponentPortal(a);return i.instance=u.instance,this._breakpointObserver.observe(c.b.Handset).pipe(Object(d.a)(n.detachments().pipe(Object(p.a)(1)))).subscribe(function(t){t.matches?n.overlayElement.classList.add("mat-snack-bar-handset"):n.overlayElement.classList.remove("mat-snack-bar-handset")}),i},t.prototype._createOverlay=function(t){var e=new o.c;e.direction=t.direction;var n=this._overlay.position().global(),r="rtl"===t.direction,i="left"===t.horizontalPosition||"start"===t.horizontalPosition&&!r||"end"===t.horizontalPosition&&r,s=!i&&"center"!==t.horizontalPosition;return i?n.left("0"):s?n.right("0"):n.centerHorizontally(),"top"===t.verticalPosition?n.top("0"):n.bottom("0"),e.positionStrategy=n,this._overlay.create(e)},t.prototype._createInjector=function(t,e){var n=t&&t.viewContainerRef&&t.viewContainerRef.injector,r=new WeakMap;return r.set(m,e),r.set(g,t.data),new s.f(n||this._injector,r)},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:o.b},{type:a.i},{type:r.B},{type:c.a},{type:t,decorators:[{type:r.M},{type:r._1}]}]},t}();function w(t){return Object(l.a)({},new v,t)}!function(){function t(){}t.decorators=[{type:r.H,args:[{imports:[o.d,s.g,i.a,u.j,c.c],exports:[j,u.j],declarations:[j,y],entryComponents:[j,y],providers:[x,a.h]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/sort.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return b});var r=n("../../../core/esm5/core.js"),i=n("../../../../tslib/tslib.es6.js"),o=n("../../../cdk/esm5/coercion.es5.js"),s=n("../../../animations/esm5/animations.js"),a=n("../../../cdk/esm5/table.es5.js"),c=n("../../../../rxjs/_esm5/observable/merge.js"),u=n("../../../material/esm5/core.es5.js"),l=n("../../../../rxjs/_esm5/Subject.js"),p=n("../../../common/esm5/common.js");var d=function(){return function(){}}(),h=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.sortables=new Map,e._stateChanges=new l.a,e.start="asc",e._direction="",e.sortChange=new r.w,e}return Object(i.b)(e,t),Object.defineProperty(e.prototype,"direction",{get:function(){return this._direction},set:function(t){if(Object(r._16)()&&t&&"asc"!==t&&"desc"!==t)throw Error(t+" is not a valid sort direction ('asc' or 'desc').");this._direction=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disableClear",{get:function(){return this._disableClear},set:function(t){this._disableClear=Object(o.b)(t)},enumerable:!0,configurable:!0}),e.prototype.register=function(t){if(!t.id)throw Error("MatSortHeader must be provided with a unique id.");if(this.sortables.has(t.id))throw e=t.id,Error("Cannot have two MatSortables with the same id ("+e+").");var e;this.sortables.set(t.id,t)},e.prototype.deregister=function(t){this.sortables.delete(t.id)},e.prototype.sort=function(t){this.active!=t.id?(this.active=t.id,this.direction=t.start?t.start:this.start):this.direction=this.getNextSortDirection(t),this.sortChange.next({active:this.active,direction:this.direction})},e.prototype.getNextSortDirection=function(t){if(!t)return"";var e=null!=t.disableClear?t.disableClear:this.disableClear,n=function(t,e){var n=["asc","desc"];"desc"==t&&n.reverse();e||n.push("");return n}(t.start||this.start,e),r=n.indexOf(this.direction)+1;return r>=n.length&&(r=0),n[r]},e.prototype.ngOnChanges=function(){this._stateChanges.next()},e.prototype.ngOnDestroy=function(){this._stateChanges.complete()},e.decorators=[{type:r.s,args:[{selector:"[matSort]",exportAs:"matSort",inputs:["disabled: matSortDisabled"]}]}],e.ctorParameters=function(){return[]},e.propDecorators={active:[{type:r.C,args:["matSortActive"]}],start:[{type:r.C,args:["matSortStart"]}],direction:[{type:r.C,args:["matSortDirection"]}],disableClear:[{type:r.C,args:["matSortDisableClear"]}],sortChange:[{type:r.N,args:["matSortChange"]}]},e}(Object(u.w)(d));var f=function(){function t(){this.changes=new l.a,this.sortButtonLabel=function(t){return"Change sorting for "+t},this.sortDescriptionLabel=function(t,e){return"Sorted by "+t+" "+("asc"==e?"ascending":"descending")}}return t.decorators=[{type:r.z}],t.ctorParameters=function(){return[]},t}();var m={provide:f,deps:[[new r.M,new r._1,f]],useFactory:function(t){return t||new f}},g=u.b.ENTERING+" "+u.a.STANDARD_CURVE,v=function(){return function(){}}(),y=function(t){function e(e,n,r,i){var o=t.call(this)||this;if(o._intl=e,o._sort=r,o._cdkColumnDef=i,o.arrowPosition="after",!r)throw Error("MatSortHeader must be placed within a parent element with the MatSort directive.");return o._rerenderSubscription=Object(c.a)(r.sortChange,r._stateChanges,e.changes).subscribe(function(){return n.markForCheck()}),o}return Object(i.b)(e,t),Object.defineProperty(e.prototype,"disableClear",{get:function(){return this._disableClear},set:function(t){this._disableClear=Object(o.b)(t)},enumerable:!0,configurable:!0}),e.prototype.ngOnInit=function(){!this.id&&this._cdkColumnDef&&(this.id=this._cdkColumnDef.name),this._sort.register(this)},e.prototype.ngOnDestroy=function(){this._sort.deregister(this),this._rerenderSubscription.unsubscribe()},e.prototype._handleClick=function(){this._isDisabled()||this._sort.sort(this)},e.prototype._isSorted=function(){return this._sort.active==this.id&&("asc"===this._sort.direction||"desc"===this._sort.direction)},e.prototype._isDisabled=function(){return this._sort.disabled||this.disabled},e.decorators=[{type:r.m,args:[{selector:"[mat-sort-header]",exportAs:"matSortHeader",template:'<div class="mat-sort-header-container" [class.mat-sort-header-position-before]="arrowPosition == \'before\'"><button class="mat-sort-header-button" type="button" [attr.aria-label]="_intl.sortButtonLabel(id)" [attr.disabled]="_isDisabled() || null"><ng-content></ng-content></button><div *ngIf="_isSorted()" class="mat-sort-header-arrow" [@indicatorToggle]="_sort.direction"><div class="mat-sort-header-stem"></div><div class="mat-sort-header-indicator" [@indicator]="_sort.direction"><div class="mat-sort-header-pointer-left" [@leftPointer]="_sort.direction"></div><div class="mat-sort-header-pointer-right" [@rightPointer]="_sort.direction"></div><div class="mat-sort-header-pointer-middle"></div></div></div></div><span class="cdk-visually-hidden" *ngIf="_isSorted()">&nbsp;{{_intl.sortDescriptionLabel(id, _sort.direction)}}</span>',styles:[".mat-sort-header-container{display:flex;cursor:pointer}.mat-sort-header-disabled .mat-sort-header-container{cursor:default}.mat-sort-header-position-before{flex-direction:row-reverse}.mat-sort-header-button{border:none;background:0 0;display:flex;align-items:center;padding:0;cursor:inherit;outline:0;font:inherit;color:currentColor}.mat-sort-header-arrow{height:12px;width:12px;min-width:12px;margin:0 0 0 6px;position:relative;display:flex}.mat-sort-header-position-before .mat-sort-header-arrow{margin:0 6px 0 0}.mat-sort-header-stem{background:currentColor;height:10px;width:2px;margin:auto;display:flex;align-items:center}.mat-sort-header-indicator{width:100%;height:2px;display:flex;align-items:center;position:absolute;top:0;left:0;transition:225ms cubic-bezier(.4,0,.2,1)}.mat-sort-header-pointer-middle{margin:auto;height:2px;width:2px;background:currentColor;transform:rotate(45deg)}.mat-sort-header-pointer-left,.mat-sort-header-pointer-right{background:currentColor;width:6px;height:2px;transition:225ms cubic-bezier(.4,0,.2,1);position:absolute;top:0}.mat-sort-header-pointer-left{transform-origin:right;left:0}.mat-sort-header-pointer-right{transform-origin:left;right:0}"],host:{"(click)":"_handleClick()","[class.mat-sort-header-sorted]":"_isSorted()","[class.mat-sort-header-disabled]":"_isDisabled()"},encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush,inputs:["disabled"],animations:[Object(s.k)("indicator",[Object(s.h)("asc",Object(s.i)({transform:"translateY(0px)"})),Object(s.h)("desc",Object(s.i)({transform:"translateY(10px)"})),Object(s.j)("asc <=> desc",Object(s.e)(g))]),Object(s.k)("leftPointer",[Object(s.h)("asc",Object(s.i)({transform:"rotate(-45deg)"})),Object(s.h)("desc",Object(s.i)({transform:"rotate(45deg)"})),Object(s.j)("asc <=> desc",Object(s.e)(g))]),Object(s.k)("rightPointer",[Object(s.h)("asc",Object(s.i)({transform:"rotate(45deg)"})),Object(s.h)("desc",Object(s.i)({transform:"rotate(-45deg)"})),Object(s.j)("asc <=> desc",Object(s.e)(g))]),Object(s.k)("indicatorToggle",[Object(s.j)("void => asc",Object(s.e)(g,Object(s.f)([Object(s.i)({transform:"translateY(25%)",opacity:0}),Object(s.i)({transform:"none",opacity:1})]))),Object(s.j)("asc => void",Object(s.e)(g,Object(s.f)([Object(s.i)({transform:"none",opacity:1}),Object(s.i)({transform:"translateY(-25%)",opacity:0})]))),Object(s.j)("void => desc",Object(s.e)(g,Object(s.f)([Object(s.i)({transform:"translateY(-25%)",opacity:0}),Object(s.i)({transform:"none",opacity:1})]))),Object(s.j)("desc => void",Object(s.e)(g,Object(s.f)([Object(s.i)({transform:"none",opacity:1}),Object(s.i)({transform:"translateY(25%)",opacity:0})])))])]}]}],e.ctorParameters=function(){return[{type:f},{type:r.j},{type:h,decorators:[{type:r.M}]},{type:a.e,decorators:[{type:r.M}]}]},e.propDecorators={id:[{type:r.C,args:["mat-sort-header"]}],arrowPosition:[{type:r.C}],start:[{type:r.C,args:["start"]}],disableClear:[{type:r.C}]},e}(Object(u.w)(v)),b=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[p.a],exports:[h,y],declarations:[h,y],providers:[m]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/stepper.es5.js":function(t,e,n){"use strict";var r=n("../../../cdk/esm5/a11y.es5.js"),i=n("../../../cdk/esm5/portal.es5.js"),o=n("../../../cdk/esm5/stepper.es5.js"),s=n("../../../common/esm5/common.js"),a=n("../../../core/esm5/core.js"),c=n("../../../material/esm5/button.es5.js"),u=n("../../../material/esm5/core.es5.js"),l=n("../../../material/esm5/icon.es5.js"),p=n("../../../cdk/esm5/coercion.es5.js"),d=n("../../../../tslib/tslib.es6.js"),h=n("../../../../rxjs/_esm5/Subject.js"),f=n("../../../animations/esm5/animations.js"),m=n("../../../../rxjs/_esm5/operators/takeUntil.js"),g=function(t){function e(e){return t.call(this,e)||this}return Object(d.b)(e,t),e.decorators=[{type:a.s,args:[{selector:"[matStepLabel]"}]}],e.ctorParameters=function(){return[{type:a._4}]},e}(o.b),v=function(){function t(){this.changes=new h.a,this.optionalLabel="Optional"}return t.decorators=[{type:a.z}],t.ctorParameters=function(){return[]},t}(),y=function(){function t(t,e,n,r){this._intl=t,this._focusMonitor=e,this._element=n,e.monitor(n.nativeElement,!0),this._intlSubscription=t.changes.subscribe(function(){return r.markForCheck()})}return Object.defineProperty(t.prototype,"index",{get:function(){return this._index},set:function(t){this._index=Object(p.c)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"selected",{get:function(){return this._selected},set:function(t){this._selected=Object(p.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"active",{get:function(){return this._active},set:function(t){this._active=Object(p.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"optional",{get:function(){return this._optional},set:function(t){this._optional=Object(p.b)(t)},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){this._intlSubscription.unsubscribe(),this._focusMonitor.stopMonitoring(this._element.nativeElement)},t.prototype._stringLabel=function(){return this.label instanceof g?null:this.label},t.prototype._templateLabel=function(){return this.label instanceof g?this.label:null},t.prototype._getHostElement=function(){return this._element.nativeElement},t.decorators=[{type:a.m,args:[{selector:"mat-step-header",template:'<div class="mat-step-header-ripple" mat-ripple [matRippleTrigger]="_getHostElement()"></div><div [class.mat-step-icon]="icon !== \'number\' || selected" [class.mat-step-icon-not-touched]="icon == \'number\' && !selected" [ngSwitch]="icon"><span *ngSwitchCase="\'number\'">{{index + 1}}</span><mat-icon *ngSwitchCase="\'edit\'">create</mat-icon><mat-icon *ngSwitchCase="\'done\'">done</mat-icon></div><div class="mat-step-label" [class.mat-step-label-active]="active" [class.mat-step-label-selected]="selected"><ng-container *ngIf="_templateLabel()" [ngTemplateOutlet]="label.template"></ng-container><div class="mat-step-text-label" *ngIf="_stringLabel()">{{label}}</div><div class="mat-step-optional" *ngIf="optional">{{_intl.optionalLabel}}</div></div>',styles:[".mat-step-header{overflow:hidden;outline:0;cursor:pointer;position:relative}.mat-step-optional{font-size:12px}.mat-step-icon,.mat-step-icon-not-touched{border-radius:50%;height:24px;width:24px;align-items:center;justify-content:center;display:flex}.mat-step-icon .mat-icon{font-size:16px;height:16px;width:16px}.mat-step-label{display:inline-block;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;min-width:50px;vertical-align:middle}.mat-step-text-label{text-overflow:ellipsis;overflow:hidden}.mat-step-header-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}"],host:{class:"mat-step-header",role:"tab"},encapsulation:a._10.None,preserveWhitespaces:!1,changeDetection:a.i.OnPush}]}],t.ctorParameters=function(){return[{type:v},{type:r.f},{type:a.t},{type:a.j}]},t.propDecorators={icon:[{type:a.C}],label:[{type:a.C}],index:[{type:a.C}],selected:[{type:a.C}],active:[{type:a.C}],optional:[{type:a.C}]},t}(),b=function(t){function e(e,n){var r=t.call(this,e)||this;return r._errorStateMatcher=n,r}return Object(d.b)(e,t),e.prototype.isErrorState=function(t,e){var n=this._errorStateMatcher.isErrorState(t,e),r=!!(t&&t.invalid&&this.interacted);return n||r},e.decorators=[{type:a.m,args:[{selector:"mat-step",template:"<ng-template><ng-content></ng-content></ng-template>",providers:[{provide:u.d,useExisting:e}],encapsulation:a._10.None,exportAs:"matStep",preserveWhitespaces:!1,changeDetection:a.i.OnPush}]}],e.ctorParameters=function(){return[{type:_,decorators:[{type:a.y,args:[Object(a._14)(function(){return _})]}]},{type:u.d,decorators:[{type:a._1}]}]},e.propDecorators={stepLabel:[{type:a.q,args:[g]}]},e}(o.a),_=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(d.b)(e,t),e.prototype.ngAfterContentInit=function(){var t=this;this._steps.changes.pipe(Object(m.a)(this._destroyed)).subscribe(function(){return t._stateChanged()})},e.decorators=[{type:a.s,args:[{selector:"[matStepper]"}]}],e.ctorParameters=function(){return[]},e.propDecorators={_stepHeader:[{type:a._8,args:[y,{read:a.t}]}],_steps:[{type:a.r,args:[b]}]},e}(o.c),j=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(d.b)(e,t),e.decorators=[{type:a.m,args:[{selector:"mat-horizontal-stepper",exportAs:"matHorizontalStepper",template:'<div class="mat-horizontal-stepper-header-container"><ng-container *ngFor="let step of _steps; let i = index; let isLast = last"><mat-step-header class="mat-horizontal-stepper-header" (click)="step.select()" (keydown)="_onKeydown($event)" [tabIndex]="_focusIndex === i ? 0 : -1" [id]="_getStepLabelId(i)" [attr.aria-controls]="_getStepContentId(i)" [attr.aria-selected]="selectedIndex == i" [index]="i" [icon]="_getIndicatorType(i)" [label]="step.stepLabel || step.label" [selected]="selectedIndex === i" [active]="step.completed || selectedIndex === i || !linear" [optional]="step.optional"></mat-step-header><div *ngIf="!isLast" class="mat-stepper-horizontal-line"></div></ng-container></div><div class="mat-horizontal-content-container"><div *ngFor="let step of _steps; let i = index" class="mat-horizontal-stepper-content" role="tabpanel" [@stepTransition]="_getAnimationDirection(i)" [id]="_getStepContentId(i)" [attr.aria-labelledby]="_getStepLabelId(i)" [attr.aria-expanded]="selectedIndex === i"><ng-container [ngTemplateOutlet]="step.content"></ng-container></div></div>',styles:[".mat-stepper-horizontal,.mat-stepper-vertical{display:block}.mat-horizontal-stepper-header-container{white-space:nowrap;display:flex;align-items:center}.mat-stepper-horizontal-line{border-top-width:1px;border-top-style:solid;flex:auto;height:0;margin:0 -16px;min-width:32px}.mat-horizontal-stepper-header{display:flex;height:72px;overflow:hidden;align-items:center;padding:0 24px}.mat-horizontal-stepper-header .mat-step-icon,.mat-horizontal-stepper-header .mat-step-icon-not-touched{margin-right:8px;flex:none}[dir=rtl] .mat-horizontal-stepper-header .mat-step-icon,[dir=rtl] .mat-horizontal-stepper-header .mat-step-icon-not-touched{margin-right:0;margin-left:8px}.mat-vertical-stepper-header{display:flex;align-items:center;padding:24px;max-height:24px}.mat-vertical-stepper-header .mat-step-icon,.mat-vertical-stepper-header .mat-step-icon-not-touched{margin-right:12px}[dir=rtl] .mat-vertical-stepper-header .mat-step-icon,[dir=rtl] .mat-vertical-stepper-header .mat-step-icon-not-touched{margin-right:0;margin-left:12px}.mat-horizontal-stepper-content{overflow:hidden}.mat-horizontal-stepper-content[aria-expanded=false]{height:0}.mat-horizontal-content-container{overflow:hidden;padding:0 24px 24px 24px}.mat-vertical-content-container{margin-left:36px;border:0;position:relative}[dir=rtl] .mat-vertical-content-container{margin-left:0;margin-right:36px}.mat-stepper-vertical-line::before{content:'';position:absolute;top:-16px;bottom:-16px;left:0;border-left-width:1px;border-left-style:solid}[dir=rtl] .mat-stepper-vertical-line::before{left:auto;right:0}.mat-vertical-stepper-content{overflow:hidden}.mat-vertical-content{padding:0 24px 24px 24px}.mat-step:last-child .mat-vertical-content-container{border:none}"],inputs:["selectedIndex"],host:{class:"mat-stepper-horizontal","aria-orientation":"horizontal",role:"tablist"},animations:[Object(f.k)("stepTransition",[Object(f.h)("previous",Object(f.i)({transform:"translate3d(-100%, 0, 0)",visibility:"hidden"})),Object(f.h)("current",Object(f.i)({transform:"none",visibility:"visible"})),Object(f.h)("next",Object(f.i)({transform:"translate3d(100%, 0, 0)",visibility:"hidden"})),Object(f.j)("* => *",Object(f.e)("500ms cubic-bezier(0.35, 0, 0.25, 1)"))])],providers:[{provide:_,useExisting:e}],encapsulation:a._10.None,preserveWhitespaces:!1,changeDetection:a.i.OnPush}]}],e.ctorParameters=function(){return[]},e}(_),x=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(d.b)(e,t),e.decorators=[{type:a.m,args:[{selector:"mat-vertical-stepper",exportAs:"matVerticalStepper",template:'<div class="mat-step" *ngFor="let step of _steps; let i = index; let isLast = last"><mat-step-header class="mat-vertical-stepper-header" (click)="step.select()" (keydown)="_onKeydown($event)" [tabIndex]="_focusIndex == i ? 0 : -1" [id]="_getStepLabelId(i)" [attr.aria-controls]="_getStepContentId(i)" [attr.aria-selected]="selectedIndex === i" [index]="i" [icon]="_getIndicatorType(i)" [label]="step.stepLabel || step.label" [selected]="selectedIndex === i" [active]="step.completed || selectedIndex === i || !linear" [optional]="step.optional"></mat-step-header><div class="mat-vertical-content-container" [class.mat-stepper-vertical-line]="!isLast"><div class="mat-vertical-stepper-content" role="tabpanel" [@stepTransition]="_getAnimationDirection(i)" [id]="_getStepContentId(i)" [attr.aria-labelledby]="_getStepLabelId(i)" [attr.aria-expanded]="selectedIndex === i"><div class="mat-vertical-content"><ng-container [ngTemplateOutlet]="step.content"></ng-container></div></div></div></div>',styles:[".mat-stepper-horizontal,.mat-stepper-vertical{display:block}.mat-horizontal-stepper-header-container{white-space:nowrap;display:flex;align-items:center}.mat-stepper-horizontal-line{border-top-width:1px;border-top-style:solid;flex:auto;height:0;margin:0 -16px;min-width:32px}.mat-horizontal-stepper-header{display:flex;height:72px;overflow:hidden;align-items:center;padding:0 24px}.mat-horizontal-stepper-header .mat-step-icon,.mat-horizontal-stepper-header .mat-step-icon-not-touched{margin-right:8px;flex:none}[dir=rtl] .mat-horizontal-stepper-header .mat-step-icon,[dir=rtl] .mat-horizontal-stepper-header .mat-step-icon-not-touched{margin-right:0;margin-left:8px}.mat-vertical-stepper-header{display:flex;align-items:center;padding:24px;max-height:24px}.mat-vertical-stepper-header .mat-step-icon,.mat-vertical-stepper-header .mat-step-icon-not-touched{margin-right:12px}[dir=rtl] .mat-vertical-stepper-header .mat-step-icon,[dir=rtl] .mat-vertical-stepper-header .mat-step-icon-not-touched{margin-right:0;margin-left:12px}.mat-horizontal-stepper-content{overflow:hidden}.mat-horizontal-stepper-content[aria-expanded=false]{height:0}.mat-horizontal-content-container{overflow:hidden;padding:0 24px 24px 24px}.mat-vertical-content-container{margin-left:36px;border:0;position:relative}[dir=rtl] .mat-vertical-content-container{margin-left:0;margin-right:36px}.mat-stepper-vertical-line::before{content:'';position:absolute;top:-16px;bottom:-16px;left:0;border-left-width:1px;border-left-style:solid}[dir=rtl] .mat-stepper-vertical-line::before{left:auto;right:0}.mat-vertical-stepper-content{overflow:hidden}.mat-vertical-content{padding:0 24px 24px 24px}.mat-step:last-child .mat-vertical-content-container{border:none}"],inputs:["selectedIndex"],host:{class:"mat-stepper-vertical","aria-orientation":"vertical",role:"tablist"},animations:[Object(f.k)("stepTransition",[Object(f.h)("previous",Object(f.i)({height:"0px",visibility:"hidden"})),Object(f.h)("next",Object(f.i)({height:"0px",visibility:"hidden"})),Object(f.h)("current",Object(f.i)({height:"*",visibility:"visible"})),Object(f.j)("* <=> current",Object(f.e)("225ms cubic-bezier(0.4, 0.0, 0.2, 1)"))])],providers:[{provide:_,useExisting:e}],encapsulation:a._10.None,preserveWhitespaces:!1,changeDetection:a.i.OnPush}]}],e.ctorParameters=function(){return[]},e}(_),w=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(d.b)(e,t),e.decorators=[{type:a.s,args:[{selector:"button[matStepperNext]",host:{"(click)":"_stepper.next()"},providers:[{provide:o.c,useExisting:_}]}]}],e.ctorParameters=function(){return[]},e}(o.e),O=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(d.b)(e,t),e.decorators=[{type:a.s,args:[{selector:"button[matStepperPrevious]",host:{"(click)":"_stepper.previous()"},providers:[{provide:o.c,useExisting:_}]}]}],e.ctorParameters=function(){return[]},e}(o.f);!function(){function t(){}t.decorators=[{type:a.H,args:[{imports:[u.j,s.a,i.g,c.a,o.d,l.a,r.a,u.s],exports:[u.j,j,x,b,g,_,w,O,y],declarations:[j,x,b,g,_,w,O,y],providers:[v,u.d]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/table.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return w});var r=n("../../../core/esm5/core.js"),i=n("../../../../tslib/tslib.es6.js"),o=n("../../../cdk/esm5/table.es5.js"),s=n("../../../common/esm5/common.js"),a=n("../../../material/esm5/core.es5.js"),c=n("../../../../rxjs/_esm5/BehaviorSubject.js"),u=n("../../../../rxjs/_esm5/operators/combineLatest.js"),l=n("../../../../rxjs/_esm5/operators/map.js"),p=n("../../../../rxjs/_esm5/operators/startWith.js"),d=n("../../../../rxjs/_esm5/observable/empty.js"),h=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.b)(e,t),e.decorators=[{type:r.m,args:[{selector:"mat-table",exportAs:"matTable",template:o.b,styles:[".mat-table{display:block}.mat-header-row,.mat-row{display:flex;border-bottom-width:1px;border-bottom-style:solid;align-items:center;min-height:48px;padding:0 24px}.mat-cell,.mat-header-cell{flex:1;overflow:hidden;word-wrap:break-word}"],host:{class:"mat-table"},encapsulation:r._10.None,preserveWhitespaces:!1,changeDetection:r.i.OnPush}]}],e.ctorParameters=function(){return[]},e}(o.l),f=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.b)(e,t),e.decorators=[{type:r.s,args:[{selector:"[matCellDef]",providers:[{provide:o.d,useExisting:e}]}]}],e.ctorParameters=function(){return[]},e}(o.d),m=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.b)(e,t),e.decorators=[{type:r.s,args:[{selector:"[matHeaderCellDef]",providers:[{provide:o.g,useExisting:e}]}]}],e.ctorParameters=function(){return[]},e}(o.g),g=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.b)(e,t),e.decorators=[{type:r.s,args:[{selector:"[matColumnDef]",providers:[{provide:o.e,useExisting:e}]}]}],e.ctorParameters=function(){return[]},e.propDecorators={name:[{type:r.C,args:["matColumnDef"]}]},e}(o.e),v=function(t){function e(e,n){var r=t.call(this,e,n)||this;return n.nativeElement.classList.add("mat-column-"+e.cssClassFriendlyName),r}return Object(i.b)(e,t),e.decorators=[{type:r.s,args:[{selector:"mat-header-cell",host:{class:"mat-header-cell",role:"columnheader"}}]}],e.ctorParameters=function(){return[{type:o.e},{type:r.t}]},e}(o.f),y=function(t){function e(e,n){var r=t.call(this,e,n)||this;return n.nativeElement.classList.add("mat-column-"+e.cssClassFriendlyName),r}return Object(i.b)(e,t),e.decorators=[{type:r.s,args:[{selector:"mat-cell",host:{class:"mat-cell",role:"gridcell"}}]}],e.ctorParameters=function(){return[{type:o.e},{type:r.t}]},e}(o.c),b=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.b)(e,t),e.decorators=[{type:r.s,args:[{selector:"[matHeaderRowDef]",providers:[{provide:o.i,useExisting:e}],inputs:["columns: matHeaderRowDef"]}]}],e.ctorParameters=function(){return[]},e}(o.i),_=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.b)(e,t),e.decorators=[{type:r.s,args:[{selector:"[matRowDef]",providers:[{provide:o.k,useExisting:e}],inputs:["columns: matRowDefColumns","when: matRowDefWhen"]}]}],e.ctorParameters=function(){return[]},e}(o.k),j=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.b)(e,t),e.decorators=[{type:r.m,args:[{selector:"mat-header-row",template:o.a,host:{class:"mat-header-row",role:"row"},changeDetection:r.i.OnPush,encapsulation:r._10.None,exportAs:"matHeaderRow",preserveWhitespaces:!1}]}],e.ctorParameters=function(){return[]},e}(o.h),x=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(i.b)(e,t),e.decorators=[{type:r.m,args:[{selector:"mat-row",template:o.a,host:{class:"mat-row",role:"row"},changeDetection:r.i.OnPush,encapsulation:r._10.None,exportAs:"matRow",preserveWhitespaces:!1}]}],e.ctorParameters=function(){return[]},e}(o.j),w=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[o.m,s.a,a.j],exports:[h,f,m,g,v,y,j,x,b,_],declarations:[h,f,m,g,v,y,j,x,b,_]}]}],t.ctorParameters=function(){return[]},t}();!function(){function t(t){void 0===t&&(t=[]),this._renderData=new c.a([]),this._filter=new c.a(""),this.sortingDataAccessor=function(t,e){var n=t[e];return"string"!=typeof n||n.trim()?isNaN(+n)?n:+n:n},this.filterPredicate=function(t,e){var n=Object.keys(t).reduce(function(e,n){return e+t[n]},"").toLowerCase(),r=e.trim().toLowerCase();return-1!=n.indexOf(r)},this._data=new c.a(t),this._updateChangeSubscription()}Object.defineProperty(t.prototype,"data",{get:function(){return this._data.value},set:function(t){this._data.next(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"filter",{get:function(){return this._filter.value},set:function(t){this._filter.next(t)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"sort",{get:function(){return this._sort},set:function(t){this._sort=t,this._updateChangeSubscription()},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"paginator",{get:function(){return this._paginator},set:function(t){this._paginator=t,this._updateChangeSubscription()},enumerable:!0,configurable:!0}),t.prototype._updateChangeSubscription=function(){var t=this,e=this._sort?this._sort.sortChange:Object(d.a)(),n=this._paginator?this._paginator.page:Object(d.a)();this._renderChangesSubscription&&this._renderChangesSubscription.unsubscribe(),this._renderChangesSubscription=this._data.pipe(Object(u.b)(this._filter),Object(l.a)(function(e){var n=e[0];return t._filterData(n)}),Object(u.b)(e.pipe(Object(p.a)(null))),Object(l.a)(function(e){var n=e[0];return t._orderData(n)}),Object(u.b)(n.pipe(Object(p.a)(null))),Object(l.a)(function(e){var n=e[0];return t._pageData(n)})).subscribe(function(e){return t._renderData.next(e)})},t.prototype._filterData=function(t){var e=this;return this.filteredData=this.filter?t.filter(function(t){return e.filterPredicate(t,e.filter)}):t,this.paginator&&this._updatePaginator(this.filteredData.length),this.filteredData},t.prototype._orderData=function(t){var e=this;if(!this.sort||!this.sort.active||""==this.sort.direction)return t;var n=this.sort.active,r=this.sort.direction;return t.slice().sort(function(t,i){return(e.sortingDataAccessor(t,n)<e.sortingDataAccessor(i,n)?-1:1)*("asc"==r?1:-1)})},t.prototype._pageData=function(t){if(!this.paginator)return t;var e=this.paginator.pageIndex*this.paginator.pageSize;return t.slice().splice(e,this.paginator.pageSize)},t.prototype._updatePaginator=function(t){var e=this;Promise.resolve().then(function(){if(e.paginator&&(e.paginator.length=t,e.paginator.pageIndex>0)){var n=Math.ceil(e.paginator.length/e.paginator.pageSize)-1||0;e.paginator.pageIndex=Math.min(e.paginator.pageIndex,n)}})},t.prototype.connect=function(){return this._renderData},t.prototype.disconnect=function(){}}()},"../../../material/esm5/tabs.es5.js":function(t,e,n){"use strict";var r=n("../../../cdk/esm5/observers.es5.js"),i=n("../../../cdk/esm5/portal.es5.js"),o=n("../../../cdk/esm5/scrolling.es5.js"),s=n("../../../common/esm5/common.js"),a=n("../../../core/esm5/core.js"),c=n("../../../material/esm5/core.es5.js"),u=n("../../../../tslib/tslib.es6.js"),l=n("../../../../rxjs/_esm5/Subject.js"),p=n("../../../animations/esm5/animations.js"),d=n("../../../cdk/esm5/bidi.es5.js"),h=n("../../../cdk/esm5/coercion.es5.js"),f=n("../../../../rxjs/_esm5/Subscription.js"),m=n("../../../../rxjs/_esm5/observable/merge.js"),g=n("../../../cdk/esm5/keycodes.es5.js"),v=n("../../../../rxjs/_esm5/observable/of.js"),y=n("../../../cdk/esm5/platform.es5.js"),b=n("../../../../rxjs/_esm5/operators/takeUntil.js"),_=function(){function t(t,e){this._elementRef=t,this._ngZone=e}return t.prototype.alignToElement=function(t){var e=this;this.show(),"undefined"!=typeof requestAnimationFrame?this._ngZone.runOutsideAngular(function(){requestAnimationFrame(function(){return e._setStyles(t)})}):this._setStyles(t)},t.prototype.show=function(){this._elementRef.nativeElement.style.visibility="visible"},t.prototype.hide=function(){this._elementRef.nativeElement.style.visibility="hidden"},t.prototype._setStyles=function(t){var e=this._elementRef.nativeElement;e.style.left=t?(t.offsetLeft||0)+"px":"0",e.style.width=t?(t.offsetWidth||0)+"px":"0"},t.decorators=[{type:a.s,args:[{selector:"mat-ink-bar",host:{class:"mat-ink-bar"}}]}],t.ctorParameters=function(){return[{type:a.t},{type:a.L}]},t}(),j=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(u.b)(e,t),e.decorators=[{type:a.s,args:[{selector:"[mat-tab-label], [matTabLabel]"}]}],e.ctorParameters=function(){return[{type:a._4},{type:a._9}]},e}(i.b),x=function(){return function(){}}(),w=function(t){function e(e){var n=t.call(this)||this;return n._viewContainerRef=e,n.textLabel="",n._contentPortal=null,n._labelChange=new l.a,n._disableChange=new l.a,n.position=null,n.origin=null,n.isActive=!1,n}return Object(u.b)(e,t),Object.defineProperty(e.prototype,"content",{get:function(){return this._contentPortal},enumerable:!0,configurable:!0}),e.prototype.ngOnChanges=function(t){t.hasOwnProperty("textLabel")&&this._labelChange.next(),t.hasOwnProperty("disabled")&&this._disableChange.next()},e.prototype.ngOnDestroy=function(){this._disableChange.complete(),this._labelChange.complete()},e.prototype.ngOnInit=function(){this._contentPortal=new i.h(this._content,this._viewContainerRef)},e.decorators=[{type:a.m,args:[{selector:"mat-tab",template:"<ng-template><ng-content></ng-content></ng-template>",inputs:["disabled"],changeDetection:a.i.OnPush,encapsulation:a._10.None,preserveWhitespaces:!1,exportAs:"matTab"}]}],e.ctorParameters=function(){return[{type:a._9}]},e.propDecorators={templateLabel:[{type:a.q,args:[j]}],_content:[{type:a._7,args:[a._4]}],textLabel:[{type:a.C,args:["label"]}]},e}(Object(c.w)(x)),O=function(t){function e(e,n,r){var i=t.call(this,e,n)||this;return i._host=r,i}return Object(u.b)(e,t),e.prototype.ngOnInit=function(){var t=this;this._host._isCenterPosition(this._host._position)&&this.attach(this._host._content),this._centeringSub=this._host._beforeCentering.subscribe(function(e){e&&(t.hasAttached()||t.attach(t._host._content))}),this._leavingSub=this._host._afterLeavingCenter.subscribe(function(){t.detach()})},e.prototype.ngOnDestroy=function(){this._centeringSub&&!this._centeringSub.closed&&this._centeringSub.unsubscribe(),this._leavingSub&&!this._leavingSub.closed&&this._leavingSub.unsubscribe()},e.decorators=[{type:a.s,args:[{selector:"[matTabBodyHost]"}]}],e.ctorParameters=function(){return[{type:a.o},{type:a._9},{type:C,decorators:[{type:a.y,args:[Object(a._14)(function(){return C})]}]}]},e}(i.c),C=function(){function t(t,e){this._elementRef=t,this._dir=e,this._onCentering=new a.w,this._beforeCentering=new a.w,this._afterLeavingCenter=new a.w,this._onCentered=new a.w(!0)}return Object.defineProperty(t.prototype,"position",{set:function(t){this._position=t<0?"ltr"==this._getLayoutDirection()?"left":"right":t>0?"ltr"==this._getLayoutDirection()?"right":"left":"center"},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"origin",{set:function(t){if(null!=t){var e=this._getLayoutDirection();this._origin="ltr"==e&&t<=0||"rtl"==e&&t>0?"left":"right"}},enumerable:!0,configurable:!0}),t.prototype.ngOnInit=function(){"center"==this._position&&this._origin&&(this._position="left"==this._origin?"left-origin-center":"right-origin-center")},t.prototype._onTranslateTabStarted=function(t){var e=this._isCenterPosition(t.toState);this._beforeCentering.emit(e),e&&this._onCentering.emit(this._elementRef.nativeElement.clientHeight)},t.prototype._onTranslateTabComplete=function(t){this._isCenterPosition(t.toState)&&this._isCenterPosition(this._position)&&this._onCentered.emit(),this._isCenterPosition(t.fromState)&&!this._isCenterPosition(this._position)&&this._afterLeavingCenter.emit()},t.prototype._getLayoutDirection=function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},t.prototype._isCenterPosition=function(t){return"center"==t||"left-origin-center"==t||"right-origin-center"==t},t.decorators=[{type:a.m,args:[{selector:"mat-tab-body",template:'<div class="mat-tab-body-content" #content [@translateTab]="_position" (@translateTab.start)="_onTranslateTabStarted($event)" (@translateTab.done)="_onTranslateTabComplete($event)"><ng-template matTabBodyHost></ng-template></div>',styles:[".mat-tab-body-content{-webkit-backface-visibility:hidden;backface-visibility:hidden;height:100%;overflow:auto}.mat-tab-group-dynamic-height .mat-tab-body-content{overflow:hidden}"],encapsulation:a._10.None,preserveWhitespaces:!1,changeDetection:a.i.OnPush,host:{class:"mat-tab-body"},animations:[Object(p.k)("translateTab",[Object(p.h)("center, void, left-origin-center, right-origin-center",Object(p.i)({transform:"none"})),Object(p.h)("left",Object(p.i)({transform:"translate3d(-100%, 0, 0)"})),Object(p.h)("right",Object(p.i)({transform:"translate3d(100%, 0, 0)"})),Object(p.j)("* => left, * => right, left => center, right => center",Object(p.e)("500ms cubic-bezier(0.35, 0, 0.25, 1)")),Object(p.j)("void => left-origin-center",[Object(p.i)({transform:"translate3d(-100%, 0, 0)"}),Object(p.e)("500ms cubic-bezier(0.35, 0, 0.25, 1)")]),Object(p.j)("void => right-origin-center",[Object(p.i)({transform:"translate3d(100%, 0, 0)"}),Object(p.e)("500ms cubic-bezier(0.35, 0, 0.25, 1)")])])]}]}],t.ctorParameters=function(){return[{type:a.t},{type:d.b,decorators:[{type:a.M}]}]},t.propDecorators={_onCentering:[{type:a.N}],_beforeCentering:[{type:a.N}],_afterLeavingCenter:[{type:a.N}],_onCentered:[{type:a.N}],_content:[{type:a.C,args:["content"]}],position:[{type:a.C,args:["position"]}],origin:[{type:a.C,args:["origin"]}]},t}(),k=0,E=function(){return function(){}}(),S=function(){return function(t){this._elementRef=t}}(),P=function(t){function e(e,n){var r=t.call(this,e)||this;return r._changeDetectorRef=n,r._indexToSelect=0,r._tabBodyWrapperHeight=0,r._tabsSubscription=f.a.EMPTY,r._tabLabelSubscription=f.a.EMPTY,r._dynamicHeight=!1,r._selectedIndex=null,r.headerPosition="above",r.selectedIndexChange=new a.w,r.focusChange=new a.w,r.selectedTabChange=new a.w(!0),r.selectChange=r.selectedTabChange,r._groupId=k++,r}return Object(u.b)(e,t),Object.defineProperty(e.prototype,"dynamicHeight",{get:function(){return this._dynamicHeight},set:function(t){this._dynamicHeight=Object(h.b)(t)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"_dynamicHeightDeprecated",{get:function(){return this._dynamicHeight},set:function(t){this._dynamicHeight=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"selectedIndex",{get:function(){return this._selectedIndex},set:function(t){this._indexToSelect=Object(h.c)(t,null)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"backgroundColor",{get:function(){return this._backgroundColor},set:function(t){var e=this._elementRef.nativeElement;e.classList.remove("mat-background-"+this.backgroundColor),t&&e.classList.add("mat-background-"+t),this._backgroundColor=t},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentChecked=function(){var t=this,e=this._indexToSelect=Math.min(this._tabs.length-1,Math.max(this._indexToSelect||0,0));if(this._selectedIndex!=e&&null!=this._selectedIndex){var n=this._createChangeEvent(e);this.selectedTabChange.emit(n),Promise.resolve().then(function(){return t.selectedIndexChange.emit(e)})}this._tabs.forEach(function(n,r){n.position=r-e,n.isActive=r===e,null==t._selectedIndex||0!=n.position||n.origin||(n.origin=e-t._selectedIndex)}),this._selectedIndex!==e&&(this._selectedIndex=e,this._changeDetectorRef.markForCheck())},e.prototype.ngAfterContentInit=function(){var t=this;this._subscribeToTabLabels(),this._tabsSubscription=this._tabs.changes.subscribe(function(){t._subscribeToTabLabels(),t._changeDetectorRef.markForCheck()})},e.prototype.ngOnDestroy=function(){this._tabsSubscription.unsubscribe(),this._tabLabelSubscription.unsubscribe()},e.prototype._focusChanged=function(t){this.focusChange.emit(this._createChangeEvent(t))},e.prototype._createChangeEvent=function(t){var e=new E;return e.index=t,this._tabs&&this._tabs.length&&(e.tab=this._tabs.toArray()[t]),e},e.prototype._subscribeToTabLabels=function(){var t=this;this._tabLabelSubscription&&this._tabLabelSubscription.unsubscribe(),this._tabLabelSubscription=m.a.apply(void 0,this._tabs.map(function(t){return t._disableChange}).concat(this._tabs.map(function(t){return t._labelChange}))).subscribe(function(){t._changeDetectorRef.markForCheck()})},e.prototype._getTabLabelId=function(t){return"mat-tab-label-"+this._groupId+"-"+t},e.prototype._getTabContentId=function(t){return"mat-tab-content-"+this._groupId+"-"+t},e.prototype._setTabBodyWrapperHeight=function(t){if(this._dynamicHeight&&this._tabBodyWrapperHeight){var e=this._tabBodyWrapper.nativeElement;e.style.height=this._tabBodyWrapperHeight+"px",this._tabBodyWrapper.nativeElement.offsetHeight&&(e.style.height=t+"px")}},e.prototype._removeTabBodyWrapperHeight=function(){this._tabBodyWrapperHeight=this._tabBodyWrapper.nativeElement.clientHeight,this._tabBodyWrapper.nativeElement.style.height=""},e.prototype._handleClick=function(t,e,n){t.disabled||(this.selectedIndex=e.focusIndex=n)},e.prototype._getTabIndex=function(t,e){return t.disabled?null:this.selectedIndex===e?0:-1},e.decorators=[{type:a.m,args:[{selector:"mat-tab-group",exportAs:"matTabGroup",template:'<mat-tab-header #tabHeader [selectedIndex]="selectedIndex" [disableRipple]="disableRipple" (indexFocused)="_focusChanged($event)" (selectFocusedIndex)="selectedIndex = $event"><div class="mat-tab-label" role="tab" matTabLabelWrapper mat-ripple *ngFor="let tab of _tabs; let i = index" [id]="_getTabLabelId(i)" [attr.tabIndex]="_getTabIndex(tab, i)" [attr.aria-controls]="_getTabContentId(i)" [attr.aria-selected]="selectedIndex == i" [class.mat-tab-label-active]="selectedIndex == i" [disabled]="tab.disabled" [matRippleDisabled]="tab.disabled || disableRipple" (click)="_handleClick(tab, tabHeader, i)"><ng-template [ngIf]="tab.templateLabel"><ng-template [cdkPortalOutlet]="tab.templateLabel"></ng-template></ng-template><ng-template [ngIf]="!tab.templateLabel">{{tab.textLabel}}</ng-template></div></mat-tab-header><div class="mat-tab-body-wrapper" #tabBodyWrapper><mat-tab-body role="tabpanel" *ngFor="let tab of _tabs; let i = index" [id]="_getTabContentId(i)" [attr.aria-labelledby]="_getTabLabelId(i)" [class.mat-tab-body-active]="selectedIndex == i" [content]="tab.content" [position]="tab.position" [origin]="tab.origin" (_onCentered)="_removeTabBodyWrapperHeight()" (_onCentering)="_setTabBodyWrapperHeight($event)"></mat-tab-body></div>',styles:[".mat-tab-group{display:flex;flex-direction:column}.mat-tab-group.mat-tab-group-inverted-header{flex-direction:column-reverse}.mat-tab-label{height:48px;padding:0 24px;cursor:pointer;box-sizing:border-box;opacity:.6;min-width:160px;text-align:center;display:inline-flex;justify-content:center;align-items:center;white-space:nowrap;position:relative}.mat-tab-label:focus{outline:0;opacity:1}.mat-tab-label.mat-tab-disabled{cursor:default}@media (max-width:600px){.mat-tab-label{padding:0 12px}}@media (max-width:960px){.mat-tab-label{padding:0 12px}}.mat-tab-group[mat-stretch-tabs] .mat-tab-label{flex-basis:0;flex-grow:1}.mat-tab-body-wrapper{position:relative;overflow:hidden;display:flex;transition:height .5s cubic-bezier(.35,0,.25,1)}.mat-tab-body{top:0;left:0;right:0;bottom:0;position:absolute;display:block;overflow:hidden}.mat-tab-body.mat-tab-body-active{position:relative;overflow-x:hidden;overflow-y:auto;z-index:1;flex-grow:1}.mat-tab-group.mat-tab-group-dynamic-height .mat-tab-body.mat-tab-body-active{overflow-y:hidden}"],encapsulation:a._10.None,preserveWhitespaces:!1,changeDetection:a.i.OnPush,inputs:["color","disableRipple"],host:{class:"mat-tab-group","[class.mat-tab-group-dynamic-height]":"dynamicHeight","[class.mat-tab-group-inverted-header]":'headerPosition === "below"'}}]}],e.ctorParameters=function(){return[{type:a.t},{type:a.j}]},e.propDecorators={_tabs:[{type:a.r,args:[w]}],_tabBodyWrapper:[{type:a._7,args:["tabBodyWrapper"]}],dynamicHeight:[{type:a.C}],_dynamicHeightDeprecated:[{type:a.C,args:["mat-dynamic-height"]}],selectedIndex:[{type:a.C}],headerPosition:[{type:a.C}],backgroundColor:[{type:a.C}],selectedIndexChange:[{type:a.N}],focusChange:[{type:a.N}],selectedTabChange:[{type:a.N}],selectChange:[{type:a.N}]},e}(Object(c.u)(Object(c.v)(S),"primary")),M=function(){return function(){}}(),A=function(t){function e(e){var n=t.call(this)||this;return n.elementRef=e,n}return Object(u.b)(e,t),e.prototype.focus=function(){this.elementRef.nativeElement.focus()},e.prototype.getOffsetLeft=function(){return this.elementRef.nativeElement.offsetLeft},e.prototype.getOffsetWidth=function(){return this.elementRef.nativeElement.offsetWidth},e.decorators=[{type:a.s,args:[{selector:"[matTabLabelWrapper]",inputs:["disabled"],host:{"[class.mat-tab-disabled]":"disabled"}}]}],e.ctorParameters=function(){return[{type:a.t}]},e}(Object(c.w)(M)),T=function(){return function(){}}(),D=function(t){function e(e,n,r,i){var o=t.call(this)||this;return o._elementRef=e,o._changeDetectorRef=n,o._viewportRuler=r,o._dir=i,o._focusIndex=0,o._scrollDistance=0,o._selectedIndexChanged=!1,o._realignInkBar=f.a.EMPTY,o._showPaginationControls=!1,o._disableScrollAfter=!0,o._disableScrollBefore=!0,o._selectedIndex=0,o.selectFocusedIndex=new a.w,o.indexFocused=new a.w,o}return Object(u.b)(e,t),Object.defineProperty(e.prototype,"selectedIndex",{get:function(){return this._selectedIndex},set:function(t){t=Object(h.c)(t),this._selectedIndexChanged=this._selectedIndex!=t,this._selectedIndex=t,this._focusIndex=t},enumerable:!0,configurable:!0}),e.prototype.ngAfterContentChecked=function(){this._tabLabelCount!=this._labelWrappers.length&&(this._updatePagination(),this._tabLabelCount=this._labelWrappers.length,this._changeDetectorRef.markForCheck()),this._selectedIndexChanged&&(this._scrollToLabel(this._selectedIndex),this._checkScrollingControls(),this._alignInkBarToSelectedTab(),this._selectedIndexChanged=!1,this._changeDetectorRef.markForCheck()),this._scrollDistanceChanged&&(this._updateTabScrollPosition(),this._scrollDistanceChanged=!1,this._changeDetectorRef.markForCheck())},e.prototype._handleKeydown=function(t){switch(t.keyCode){case g.m:this._focusNextTab();break;case g.i:this._focusPreviousTab();break;case g.f:case g.n:this.selectFocusedIndex.emit(this.focusIndex),t.preventDefault()}},e.prototype.ngAfterContentInit=function(){var t=this,e=this._dir?this._dir.change:Object(v.a)(null),n=this._viewportRuler.change(150),r=function(){t._updatePagination(),t._alignInkBarToSelectedTab()};"undefined"!=typeof requestAnimationFrame?requestAnimationFrame(r):r(),this._realignInkBar=Object(m.a)(e,n).subscribe(r)},e.prototype.ngOnDestroy=function(){this._realignInkBar.unsubscribe()},e.prototype._onContentChanges=function(){this._updatePagination(),this._alignInkBarToSelectedTab(),this._changeDetectorRef.markForCheck()},e.prototype._updatePagination=function(){this._checkPaginationEnabled(),this._checkScrollingControls(),this._updateTabScrollPosition()},Object.defineProperty(e.prototype,"focusIndex",{get:function(){return this._focusIndex},set:function(t){this._isValidIndex(t)&&this._focusIndex!=t&&(this._focusIndex=t,this.indexFocused.emit(t),this._setTabFocus(t))},enumerable:!0,configurable:!0}),e.prototype._isValidIndex=function(t){if(!this._labelWrappers)return!0;var e=this._labelWrappers?this._labelWrappers.toArray()[t]:null;return!!e&&!e.disabled},e.prototype._setTabFocus=function(t){if(this._showPaginationControls&&this._scrollToLabel(t),this._labelWrappers&&this._labelWrappers.length){this._labelWrappers.toArray()[t].focus();var e=this._tabListContainer.nativeElement,n=this._getLayoutDirection();e.scrollLeft="ltr"==n?0:e.scrollWidth-e.offsetWidth}},e.prototype._moveFocus=function(t){if(this._labelWrappers)for(var e=this._labelWrappers.toArray(),n=this.focusIndex+t;n<e.length&&n>=0;n+=t)if(this._isValidIndex(n))return void(this.focusIndex=n)},e.prototype._focusNextTab=function(){this._moveFocus("ltr"==this._getLayoutDirection()?1:-1)},e.prototype._focusPreviousTab=function(){this._moveFocus("ltr"==this._getLayoutDirection()?-1:1)},e.prototype._getLayoutDirection=function(){return this._dir&&"rtl"===this._dir.value?"rtl":"ltr"},e.prototype._updateTabScrollPosition=function(){var t=this.scrollDistance,e="ltr"===this._getLayoutDirection()?-t:t;this._tabList.nativeElement.style.transform="translate3d("+e+"px, 0, 0)"},Object.defineProperty(e.prototype,"scrollDistance",{get:function(){return this._scrollDistance},set:function(t){this._scrollDistance=Math.max(0,Math.min(this._getMaxScrollDistance(),t)),this._scrollDistanceChanged=!0,this._checkScrollingControls()},enumerable:!0,configurable:!0}),e.prototype._scrollHeader=function(t){var e=this._tabListContainer.nativeElement.offsetWidth;this.scrollDistance+=("before"==t?-1:1)*e/3},e.prototype._scrollToLabel=function(t){var e=this._labelWrappers?this._labelWrappers.toArray()[t]:null;if(e){var n,r,i=this._tabListContainer.nativeElement.offsetWidth;"ltr"==this._getLayoutDirection()?r=(n=e.getOffsetLeft())+e.getOffsetWidth():n=(r=this._tabList.nativeElement.offsetWidth-e.getOffsetLeft())-e.getOffsetWidth();var o=this.scrollDistance,s=this.scrollDistance+i;n<o?this.scrollDistance-=o-n+60:r>s&&(this.scrollDistance+=r-s+60)}},e.prototype._checkPaginationEnabled=function(){var t=this._tabList.nativeElement.scrollWidth>this._elementRef.nativeElement.offsetWidth;t||(this.scrollDistance=0),t!==this._showPaginationControls&&this._changeDetectorRef.markForCheck(),this._showPaginationControls=t},e.prototype._checkScrollingControls=function(){this._disableScrollBefore=0==this.scrollDistance,this._disableScrollAfter=this.scrollDistance==this._getMaxScrollDistance(),this._changeDetectorRef.markForCheck()},e.prototype._getMaxScrollDistance=function(){return this._tabList.nativeElement.scrollWidth-this._tabListContainer.nativeElement.offsetWidth||0},e.prototype._alignInkBarToSelectedTab=function(){var t=this._labelWrappers&&this._labelWrappers.length?this._labelWrappers.toArray()[this.selectedIndex].elementRef.nativeElement:null;this._inkBar.alignToElement(t)},e.decorators=[{type:a.m,args:[{selector:"mat-tab-header",template:'<div class="mat-tab-header-pagination mat-tab-header-pagination-before mat-elevation-z4" aria-hidden="true" mat-ripple [matRippleDisabled]="_disableScrollBefore || disableRipple" [class.mat-tab-header-pagination-disabled]="_disableScrollBefore" (click)="_scrollHeader(\'before\')"><div class="mat-tab-header-pagination-chevron"></div></div><div class="mat-tab-label-container" #tabListContainer (keydown)="_handleKeydown($event)"><div class="mat-tab-list" #tabList role="tablist" (cdkObserveContent)="_onContentChanges()"><div class="mat-tab-labels"><ng-content></ng-content></div><mat-ink-bar></mat-ink-bar></div></div><div class="mat-tab-header-pagination mat-tab-header-pagination-after mat-elevation-z4" aria-hidden="true" mat-ripple [matRippleDisabled]="_disableScrollAfter || disableRipple" [class.mat-tab-header-pagination-disabled]="_disableScrollAfter" (click)="_scrollHeader(\'after\')"><div class="mat-tab-header-pagination-chevron"></div></div>',styles:[".mat-tab-header{display:flex;overflow:hidden;position:relative;flex-shrink:0}.mat-tab-label{height:48px;padding:0 24px;cursor:pointer;box-sizing:border-box;opacity:.6;min-width:160px;text-align:center;display:inline-flex;justify-content:center;align-items:center;white-space:nowrap;position:relative}.mat-tab-label:focus{outline:0;opacity:1}.mat-tab-label.mat-tab-disabled{cursor:default}@media (max-width:600px){.mat-tab-label{min-width:72px}}.mat-ink-bar{position:absolute;bottom:0;height:2px;transition:.5s cubic-bezier(.35,0,.25,1)}.mat-tab-group-inverted-header .mat-ink-bar{bottom:auto;top:0}.mat-tab-header-pagination{position:relative;display:none;justify-content:center;align-items:center;min-width:32px;cursor:pointer;z-index:2}.mat-tab-header-pagination-controls-enabled .mat-tab-header-pagination{display:flex}.mat-tab-header-pagination-before,.mat-tab-header-rtl .mat-tab-header-pagination-after{padding-left:4px}.mat-tab-header-pagination-before .mat-tab-header-pagination-chevron,.mat-tab-header-rtl .mat-tab-header-pagination-after .mat-tab-header-pagination-chevron{transform:rotate(-135deg)}.mat-tab-header-pagination-after,.mat-tab-header-rtl .mat-tab-header-pagination-before{padding-right:4px}.mat-tab-header-pagination-after .mat-tab-header-pagination-chevron,.mat-tab-header-rtl .mat-tab-header-pagination-before .mat-tab-header-pagination-chevron{transform:rotate(45deg)}.mat-tab-header-pagination-chevron{border-style:solid;border-width:2px 2px 0 0;content:'';height:8px;width:8px}.mat-tab-header-pagination-disabled{box-shadow:none;cursor:default}.mat-tab-label-container{display:flex;flex-grow:1;overflow:hidden;z-index:1}.mat-tab-list{flex-grow:1;position:relative;transition:transform .5s cubic-bezier(.35,0,.25,1)}.mat-tab-labels{display:flex}"],inputs:["disableRipple"],encapsulation:a._10.None,preserveWhitespaces:!1,changeDetection:a.i.OnPush,host:{class:"mat-tab-header","[class.mat-tab-header-pagination-controls-enabled]":"_showPaginationControls","[class.mat-tab-header-rtl]":"_getLayoutDirection() == 'rtl'"}}]}],e.ctorParameters=function(){return[{type:a.t},{type:a.j},{type:o.d},{type:d.b,decorators:[{type:a.M}]}]},e.propDecorators={_labelWrappers:[{type:a.r,args:[A]}],_inkBar:[{type:a._7,args:[_]}],_tabListContainer:[{type:a._7,args:["tabListContainer"]}],_tabList:[{type:a._7,args:["tabList"]}],selectedIndex:[{type:a.C}],selectFocusedIndex:[{type:a.N}],indexFocused:[{type:a.N}]},e}(Object(c.v)(T)),I=function(){return function(t){this._elementRef=t}}(),R=function(t){function e(e,n,r,i,o){var s=t.call(this,e)||this;return s._dir=n,s._ngZone=r,s._changeDetectorRef=i,s._viewportRuler=o,s._onDestroy=new l.a,s._disableRipple=!1,s}return Object(u.b)(e,t),Object.defineProperty(e.prototype,"backgroundColor",{get:function(){return this._backgroundColor},set:function(t){var e=this._elementRef.nativeElement;e.classList.remove("mat-background-"+this.backgroundColor),t&&e.classList.add("mat-background-"+t),this._backgroundColor=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disableRipple",{get:function(){return this._disableRipple},set:function(t){this._disableRipple=Object(h.b)(t),this._setLinkDisableRipple()},enumerable:!0,configurable:!0}),e.prototype.updateActiveLink=function(t){this._activeLinkChanged=this._activeLinkElement!=t,this._activeLinkElement=t,this._activeLinkChanged&&this._changeDetectorRef.markForCheck()},e.prototype.ngAfterContentInit=function(){var t=this;this._ngZone.runOutsideAngular(function(){var e=t._dir?t._dir.change:Object(v.a)(null);return Object(m.a)(e,t._viewportRuler.change(10)).pipe(Object(b.a)(t._onDestroy)).subscribe(function(){return t._alignInkBar()})}),this._setLinkDisableRipple()},e.prototype.ngAfterContentChecked=function(){this._activeLinkChanged&&(this._alignInkBar(),this._activeLinkChanged=!1)},e.prototype.ngOnDestroy=function(){this._onDestroy.next(),this._onDestroy.complete()},e.prototype._alignInkBar=function(){this._activeLinkElement&&this._inkBar.alignToElement(this._activeLinkElement.nativeElement)},e.prototype._setLinkDisableRipple=function(){var t=this;this._tabLinks&&this._tabLinks.forEach(function(e){return e.disableRipple=t.disableRipple})},e.decorators=[{type:a.m,args:[{selector:"[mat-tab-nav-bar]",exportAs:"matTabNavBar, matTabNav",inputs:["color","disableRipple"],template:'<div class="mat-tab-links" (cdkObserveContent)="_alignInkBar()"><ng-content></ng-content><mat-ink-bar></mat-ink-bar></div>',styles:[".mat-tab-nav-bar{overflow:hidden;position:relative;flex-shrink:0}.mat-tab-links{position:relative}.mat-tab-link{height:48px;padding:0 24px;cursor:pointer;box-sizing:border-box;opacity:.6;min-width:160px;text-align:center;display:inline-flex;justify-content:center;align-items:center;white-space:nowrap;vertical-align:top;text-decoration:none;position:relative;overflow:hidden}.mat-tab-link:focus{outline:0;opacity:1}.mat-tab-link.mat-tab-disabled{cursor:default}@media (max-width:600px){.mat-tab-link{min-width:72px}}.mat-ink-bar{position:absolute;bottom:0;height:2px;transition:.5s cubic-bezier(.35,0,.25,1)}.mat-tab-group-inverted-header .mat-ink-bar{bottom:auto;top:0}"],host:{class:"mat-tab-nav-bar"},encapsulation:a._10.None,preserveWhitespaces:!1,changeDetection:a.i.OnPush}]}],e.ctorParameters=function(){return[{type:a.t},{type:d.b,decorators:[{type:a.M}]},{type:a.L},{type:a.j},{type:o.d}]},e.propDecorators={_inkBar:[{type:a._7,args:[_]}],_tabLinks:[{type:a.r,args:[Object(a._14)(function(){return L}),{descendants:!0}]}],backgroundColor:[{type:a.C}]},e}(Object(c.v)(Object(c.u)(I,"primary"))),N=function(){return function(){}}(),L=function(t){function e(e,n,r,i,o,s){var a=t.call(this)||this;return a._tabNavBar=e,a._elementRef=n,a._isActive=!1,a._disableRipple=!1,a._tabLinkRipple=new c.r(n,r,i,o),a.tabIndex=parseInt(s)||0,a}return Object(u.b)(e,t),Object.defineProperty(e.prototype,"active",{get:function(){return this._isActive},set:function(t){this._isActive=t,t&&this._tabNavBar.updateActiveLink(this._elementRef)},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"disableRipple",{get:function(){return this.disabled||this._disableRipple},set:function(t){this._disableRipple=t,this._tabLinkRipple.disabled=this.disableRipple,this._tabLinkRipple._updateRippleRenderer()},enumerable:!0,configurable:!0}),e.prototype.ngOnDestroy=function(){this._tabLinkRipple.ngOnDestroy()},e.decorators=[{type:a.s,args:[{selector:"[mat-tab-link], [matTabLink]",exportAs:"matTabLink",inputs:["disabled","tabIndex"],host:{class:"mat-tab-link","[attr.aria-disabled]":"disabled.toString()","[attr.tabIndex]":"tabIndex","[class.mat-tab-disabled]":"disabled","[class.mat-tab-label-active]":"active"}}]}],e.ctorParameters=function(){return[{type:R},{type:a.t},{type:a.L},{type:y.a},{type:void 0,decorators:[{type:a.M},{type:a.y,args:[c.i]}]},{type:void 0,decorators:[{type:a.g,args:["tabindex"]}]}]},e.propDecorators={active:[{type:a.C}]},e}(Object(c.y)(Object(c.w)(N)));!function(){function t(){}t.decorators=[{type:a.H,args:[{imports:[s.a,c.j,i.g,c.s,r.a,o.a],exports:[c.j,P,j,w,R,L],declarations:[P,j,w,_,A,R,L,C,O,D],providers:[o.c]}]}],t.ctorParameters=function(){return[]}}()},"../../../material/esm5/toolbar.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return p});var r=n("../../../core/esm5/core.js"),i=n("../../../material/esm5/core.es5.js"),o=n("../../../cdk/esm5/platform.es5.js"),s=n("../../../../tslib/tslib.es6.js"),a=function(){return function(t){this._elementRef=t}}(),c=Object(i.u)(a),u=function(){function t(){}return t.decorators=[{type:r.s,args:[{selector:"mat-toolbar-row",exportAs:"matToolbarRow",host:{class:"mat-toolbar-row"}}]}],t.ctorParameters=function(){return[]},t}(),l=function(t){function e(e,n){var r=t.call(this,e)||this;return r._platform=n,r}return Object(s.b)(e,t),e.prototype.ngAfterViewInit=function(){var t=this;Object(r._16)()&&this._platform.isBrowser&&(this._checkToolbarMixedModes(),this._toolbarRows.changes.subscribe(function(){return t._checkToolbarMixedModes()}))},e.prototype._checkToolbarMixedModes=function(){this._toolbarRows.length&&([].slice.call(this._elementRef.nativeElement.childNodes).filter(function(t){return!(t.classList&&t.classList.contains("mat-toolbar-row"))}).filter(function(t){return t.nodeType!==Node.COMMENT_NODE}).some(function(t){return t.textContent.trim()})&&function(){throw Error("MatToolbar: Attempting to combine different toolbar modes. Either specify multiple `<mat-toolbar-row>` elements explicitly or just place content inside of a `<mat-toolbar>` for a single row.")}())},e.decorators=[{type:r.m,args:[{selector:"mat-toolbar",exportAs:"matToolbar",template:'<ng-content></ng-content><ng-content select="mat-toolbar-row"></ng-content>',styles:[".mat-toolbar-row,.mat-toolbar-single-row{display:flex;box-sizing:border-box;padding:0 16px;width:100%;flex-direction:row;align-items:center;white-space:nowrap}.mat-toolbar-multiple-rows{display:flex;box-sizing:border-box;flex-direction:column;width:100%}.mat-toolbar-multiple-rows{min-height:64px}.mat-toolbar-row,.mat-toolbar-single-row{height:64px}@media (max-width:600px){.mat-toolbar-multiple-rows{min-height:56px}.mat-toolbar-row,.mat-toolbar-single-row{height:56px}}"],inputs:["color"],host:{class:"mat-toolbar","[class.mat-toolbar-multiple-rows]":"this._toolbarRows.length","[class.mat-toolbar-single-row]":"!this._toolbarRows.length"},changeDetection:r.i.OnPush,encapsulation:r._10.None,preserveWhitespaces:!1}]}],e.ctorParameters=function(){return[{type:r.t},{type:o.a}]},e.propDecorators={_toolbarRows:[{type:r.r,args:[u]}]},e}(c);var p=function(){function t(){}return t.decorators=[{type:r.H,args:[{imports:[i.j,o.b],exports:[l,u,i.j],declarations:[l,u]}]}],t.ctorParameters=function(){return[]},t}()},"../../../material/esm5/tooltip.es5.js":function(t,e,n){"use strict";n.d(e,"a",function(){return O});var r=n("../../../cdk/esm5/a11y.es5.js"),i=n("../../../cdk/esm5/overlay.es5.js"),o=n("../../../cdk/esm5/platform.es5.js"),s=n("../../../common/esm5/common.js"),a=n("../../../core/esm5/core.js"),c=n("../../../material/esm5/core.es5.js"),u=n("../../../animations/esm5/animations.js"),l=n("../../../cdk/esm5/bidi.es5.js"),p=n("../../../cdk/esm5/coercion.es5.js"),d=n("../../../cdk/esm5/keycodes.es5.js"),h=n("../../../cdk/esm5/portal.es5.js"),f=n("../../../../rxjs/_esm5/operators/take.js"),m=n("../../../../rxjs/_esm5/observable/merge.js"),g=n("../../../cdk/esm5/scrolling.es5.js"),v=n("../../../../rxjs/_esm5/Subject.js"),y=20;function b(t){return Error('Tooltip position "'+t+'" is invalid.')}var _=new a.A("mat-tooltip-scroll-strategy");var j={provide:_,deps:[i.b],useFactory:function(t){return function(){return t.scrollStrategies.reposition({scrollThrottle:y})}}},x=function(){function t(t,e,n,r,i,o,s,a,c,u){var l=this;this._overlay=t,this._elementRef=e,this._scrollDispatcher=n,this._viewContainerRef=r,this._ngZone=i,this._platform=o,this._ariaDescriber=s,this._focusMonitor=a,this._scrollStrategy=c,this._dir=u,this._position="below",this._disabled=!1,this.showDelay=0,this.hideDelay=0,this._message="",this._manualListeners=new Map;var p=e.nativeElement;o.IOS?"INPUT"!==p.nodeName&&"TEXTAREA"!==p.nodeName||(p.style.webkitUserSelect=p.style.userSelect=""):(this._manualListeners.set("mouseenter",function(){return l.show()}),this._manualListeners.set("mouseleave",function(){return l.hide()}),this._manualListeners.forEach(function(t,n){return e.nativeElement.addEventListener(n,t)})),a.monitor(p,!1).subscribe(function(t){t?"program"!==t&&i.run(function(){return l.show()}):i.run(function(){return l.hide(0)})})}return Object.defineProperty(t.prototype,"position",{get:function(){return this._position},set:function(t){t!==this._position&&(this._position=t,this._tooltipInstance&&this._disposeTooltip())},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"disabled",{get:function(){return this._disabled},set:function(t){this._disabled=Object(p.b)(t),this._disabled&&this.hide(0)},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"_positionDeprecated",{get:function(){return this._position},set:function(t){this._position=t},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"message",{get:function(){return this._message},set:function(t){this._ariaDescriber.removeDescription(this._elementRef.nativeElement,this._message),this._message=null!=t?(""+t).trim():"",!this._message&&this._isTooltipVisible()?this.hide(0):(this._updateTooltipMessage(),this._ariaDescriber.describe(this._elementRef.nativeElement,this.message))},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"tooltipClass",{get:function(){return this._tooltipClass},set:function(t){this._tooltipClass=t,this._tooltipInstance&&this._setTooltipClass(this._tooltipClass)},enumerable:!0,configurable:!0}),t.prototype.ngOnDestroy=function(){var t=this;this._tooltipInstance&&this._disposeTooltip(),this._platform.IOS||(this._manualListeners.forEach(function(e,n){t._elementRef.nativeElement.removeEventListener(n,e)}),this._manualListeners.clear()),this._ariaDescriber.removeDescription(this._elementRef.nativeElement,this.message),this._focusMonitor.stopMonitoring(this._elementRef.nativeElement)},t.prototype.show=function(t){void 0===t&&(t=this.showDelay),!this.disabled&&this.message&&(this._tooltipInstance||this._createTooltip(),this._setTooltipClass(this._tooltipClass),this._updateTooltipMessage(),this._tooltipInstance.show(this._position,t))},t.prototype.hide=function(t){void 0===t&&(t=this.hideDelay),this._tooltipInstance&&this._tooltipInstance.hide(t)},t.prototype.toggle=function(){this._isTooltipVisible()?this.hide():this.show()},t.prototype._isTooltipVisible=function(){return!!this._tooltipInstance&&this._tooltipInstance.isVisible()},t.prototype._handleKeydown=function(t){this._isTooltipVisible()&&t.keyCode===d.g&&(t.stopPropagation(),this.hide(0))},t.prototype._createTooltip=function(){var t=this,e=this._createOverlay(),n=new h.d(w,this._viewContainerRef);this._tooltipInstance=e.attach(n).instance,Object(m.a)(this._tooltipInstance.afterHidden(),e.detachments()).subscribe(function(){t._tooltipInstance&&t._disposeTooltip()})},t.prototype._createOverlay=function(){var t=this,e=this._getOrigin(),n=this._getOverlayPosition(),r=this._overlay.position().connectedTo(this._elementRef,e.main,n.main).withFallbackPosition(e.fallback,n.fallback),o=this._scrollDispatcher.getAncestorScrollContainers(this._elementRef);r.withScrollableContainers(o),r.onPositionChange.subscribe(function(e){t._tooltipInstance&&(e.scrollableViewProperties.isOverlayClipped&&t._tooltipInstance.isVisible()?t.hide(0):t._tooltipInstance._setTransformOrigin(e.connectionPair))});var s=new i.c({direction:this._dir?this._dir.value:"ltr",positionStrategy:r,panelClass:"mat-tooltip-panel",scrollStrategy:this._scrollStrategy()});return this._overlayRef=this._overlay.create(s),this._overlayRef},t.prototype._disposeTooltip=function(){this._overlayRef&&(this._overlayRef.dispose(),this._overlayRef=null),this._tooltipInstance=null},t.prototype._getOrigin=function(){var t,e=!this._dir||"ltr"==this._dir.value;if("above"==this.position||"below"==this.position)t={originX:"center",originY:"above"==this.position?"top":"bottom"};else if("left"==this.position||"before"==this.position&&e||"after"==this.position&&!e)t={originX:"start",originY:"center"};else{if(!("right"==this.position||"after"==this.position&&e||"before"==this.position&&!e))throw b(this.position);t={originX:"end",originY:"center"}}var n=this._invertPosition(t.originX,t.originY);return{main:t,fallback:{originX:n.x,originY:n.y}}},t.prototype._getOverlayPosition=function(){var t,e=!this._dir||"ltr"==this._dir.value;if("above"==this.position)t={overlayX:"center",overlayY:"bottom"};else if("below"==this.position)t={overlayX:"center",overlayY:"top"};else if("left"==this.position||"before"==this.position&&e||"after"==this.position&&!e)t={overlayX:"end",overlayY:"center"};else{if(!("right"==this.position||"after"==this.position&&e||"before"==this.position&&!e))throw b(this.position);t={overlayX:"start",overlayY:"center"}}var n=this._invertPosition(t.overlayX,t.overlayY);return{main:t,fallback:{overlayX:n.x,overlayY:n.y}}},t.prototype._updateTooltipMessage=function(){var t=this;this._tooltipInstance&&(this._tooltipInstance.message=this.message,this._tooltipInstance._markForCheck(),this._ngZone.onMicrotaskEmpty.asObservable().pipe(Object(f.a)(1)).subscribe(function(){t._tooltipInstance&&t._overlayRef.updatePosition()}))},t.prototype._setTooltipClass=function(t){this._tooltipInstance&&(this._tooltipInstance.tooltipClass=t,this._tooltipInstance._markForCheck())},t.prototype._invertPosition=function(t,e){return"above"===this.position||"below"===this.position?"top"===e?e="bottom":"bottom"===e&&(e="top"):"end"===t?t="start":"start"===t&&(t="end"),{x:t,y:e}},t.decorators=[{type:a.s,args:[{selector:"[mat-tooltip], [matTooltip]",exportAs:"matTooltip",host:{"(longpress)":"show()","(keydown)":"_handleKeydown($event)","(touchend)":"hide(1500)"}}]}],t.ctorParameters=function(){return[{type:i.b},{type:a.t},{type:g.b},{type:a._9},{type:a.L},{type:o.a},{type:r.d},{type:r.f},{type:void 0,decorators:[{type:a.y,args:[_]}]},{type:l.b,decorators:[{type:a.M}]}]},t.propDecorators={position:[{type:a.C,args:["matTooltipPosition"]}],disabled:[{type:a.C,args:["matTooltipDisabled"]}],_positionDeprecated:[{type:a.C,args:["tooltip-position"]}],showDelay:[{type:a.C,args:["matTooltipShowDelay"]}],hideDelay:[{type:a.C,args:["matTooltipHideDelay"]}],message:[{type:a.C,args:["matTooltip"]}],tooltipClass:[{type:a.C,args:["matTooltipClass"]}]},t}(),w=function(){function t(t){this._changeDetectorRef=t,this._visibility="initial",this._closeOnInteraction=!1,this._transformOrigin="bottom",this._onHide=new v.a}return t.prototype.show=function(t,e){var n=this;this._hideTimeoutId&&clearTimeout(this._hideTimeoutId),this._closeOnInteraction=!0,this._position=t,this._showTimeoutId=setTimeout(function(){n._visibility="visible",n._markForCheck()},e)},t.prototype.hide=function(t){var e=this;this._showTimeoutId&&clearTimeout(this._showTimeoutId),this._hideTimeoutId=setTimeout(function(){e._visibility="hidden",e._markForCheck()},t)},t.prototype.afterHidden=function(){return this._onHide.asObservable()},t.prototype.isVisible=function(){return"visible"===this._visibility},t.prototype._setTransformOrigin=function(t){var e="X"==("above"===this._position||"below"===this._position?"Y":"X")?t.overlayX:t.overlayY;if("top"===e||"bottom"===e)this._transformOrigin=e;else if("start"===e)this._transformOrigin="left";else{if("end"!==e)throw b(this._position);this._transformOrigin="right"}},t.prototype._animationStart=function(){this._closeOnInteraction=!1},t.prototype._animationDone=function(t){var e=this,n=t.toState;"hidden"!==n||this.isVisible()||this._onHide.next(),"visible"!==n&&"hidden"!==n||Promise.resolve().then(function(){return e._closeOnInteraction=!0})},t.prototype._handleBodyInteraction=function(){this._closeOnInteraction&&this.hide(0)},t.prototype._markForCheck=function(){this._changeDetectorRef.markForCheck()},t.decorators=[{type:a.m,args:[{selector:"mat-tooltip-component",template:'<div class="mat-tooltip" [ngClass]="tooltipClass" [style.transform-origin]="_transformOrigin" [@state]="_visibility" (@state.start)="_animationStart()" (@state.done)="_animationDone($event)">{{message}}</div>',styles:[".mat-tooltip-panel{pointer-events:none!important}.mat-tooltip{color:#fff;border-radius:2px;margin:14px;max-width:250px;padding-left:8px;padding-right:8px}@media screen and (-ms-high-contrast:active){.mat-tooltip{outline:solid 1px}}"],encapsulation:a._10.None,preserveWhitespaces:!1,changeDetection:a.i.OnPush,animations:[Object(u.k)("state",[Object(u.h)("initial, void, hidden",Object(u.i)({transform:"scale(0)"})),Object(u.h)("visible",Object(u.i)({transform:"scale(1)"})),Object(u.j)("* => visible",Object(u.e)("150ms cubic-bezier(0.0, 0.0, 0.2, 1)")),Object(u.j)("* => hidden",Object(u.e)("150ms cubic-bezier(0.4, 0.0, 1, 1)"))])],host:{"[style.zoom]":'_visibility === "visible" ? 1 : null',"(body:click)":"this._handleBodyInteraction()","aria-hidden":"true"}}]}],t.ctorParameters=function(){return[{type:a.j}]},t}(),O=function(){function t(){}return t.decorators=[{type:a.H,args:[{imports:[s.a,i.d,c.j,o.b,r.a],exports:[x,w,c.j],declarations:[x,w],entryComponents:[w],providers:[j,r.b]}]}],t.ctorParameters=function(){return[]},t}()},"../../../platform-browser-dynamic/esm5/platform-browser-dynamic.js":function(t,e,n){"use strict";n.d(e,"a",function(){return w});var r,i=n("../../../compiler/esm5/compiler.js"),o=n("../../../core/esm5/core.js"),s=n("../../../common/esm5/common.js"),a=n("../../../platform-browser/esm5/platform-browser.js"),c=n("../../../../tslib/tslib.es6.js"),u=((r=new Map).set(i.j.ANALYZE_FOR_ENTRY_COMPONENTS,o.a),r.set(i.j.ElementRef,o.t),r.set(i.j.NgModuleRef,o.J),r.set(i.j.ViewContainerRef,o._9),r.set(i.j.ChangeDetectorRef,o.j),r.set(i.j.QueryList,o.S),r.set(i.j.TemplateRef,o._4),r.set(i.j.CodegenComponentFactoryResolver,o._19),r.set(i.j.ComponentFactoryResolver,o.o),r.set(i.j.ComponentFactory,o.n),r.set(i.j.ComponentRef,o.p),r.set(i.j.NgModuleFactory,o.I),r.set(i.j.createModuleFactory,o._26),r.set(i.j.moduleDef,o._37),r.set(i.j.moduleProviderDef,o._38),r.set(i.j.RegisterModuleFactoryFn,o._47),r.set(i.j.Injector,o.B),r.set(i.j.ViewEncapsulation,o._10),r.set(i.j.ChangeDetectionStrategy,o.i),r.set(i.j.SecurityContext,o.Y),r.set(i.j.LOCALE_ID,o.F),r.set(i.j.TRANSLATIONS_FORMAT,o._3),r.set(i.j.inlineInterpolate,o._31),r.set(i.j.interpolate,o._32),r.set(i.j.EMPTY_ARRAY,o._21),r.set(i.j.EMPTY_MAP,o._22),r.set(i.j.Renderer,o.T),r.set(i.j.viewDef,o._51),r.set(i.j.elementDef,o._29),r.set(i.j.anchorDef,o._24),r.set(i.j.textDef,o._49),r.set(i.j.directiveDef,o._28),r.set(i.j.providerDef,o._45),r.set(i.j.queryDef,o._46),r.set(i.j.pureArrayDef,o._41),r.set(i.j.pureObjectDef,o._43),r.set(i.j.purePipeDef,o._44),r.set(i.j.pipeDef,o._42),r.set(i.j.nodeValue,o._40),r.set(i.j.ngContentDef,o._39),r.set(i.j.unwrapValue,o._50),r.set(i.j.createRendererType2,o._27),r.set(i.j.createComponentFactory,o._25),r),l=function(){function t(){this.builtinExternalReferences=new Map,this.reflectionCapabilities=new o._23}return t.prototype.componentModuleUrl=function(t,e){var n=e.moduleId;if("string"==typeof n)return Object(i.z)(n)?n:"package:"+n;if(null!==n&&void 0!==n)throw Object(i.A)('moduleId should be a string in "'+Object(o._48)(t)+"\". See https://goo.gl/wIDDiL for more information.\nIf you're using Webpack you should inline the template and the styles, see https://goo.gl/X2J8zc.");return"./"+Object(o._48)(t)},t.prototype.parameters=function(t){return this.reflectionCapabilities.parameters(t)},t.prototype.annotations=function(t){return this.reflectionCapabilities.annotations(t)},t.prototype.propMetadata=function(t){return this.reflectionCapabilities.propMetadata(t)},t.prototype.hasLifecycleHook=function(t,e){return this.reflectionCapabilities.hasLifecycleHook(t,e)},t.prototype.resolveExternalReference=function(t){return u.get(t)||t.runtime},t}();var p=new o.A("ErrorCollector"),d={provide:o.O,useValue:"/"},h={get:function(t){throw new Error("No ResourceLoader implementation has been provided. Can't read the url \""+t+'"')}},f=new o.A("HtmlParser"),m=function(){function t(t,e,n,r,o,s,a,c,u,l){this._metadataResolver=e,this._delegate=new i.k(e,n,r,o,s,a,c,u,l,this.getExtraNgModuleProviders.bind(this)),this.injector=t}return t.prototype.getExtraNgModuleProviders=function(){return[this._metadataResolver.getProviderMetadata(new i.r(o.k,{useValue:this}))]},t.prototype.compileModuleSync=function(t){return this._delegate.compileModuleSync(t)},t.prototype.compileModuleAsync=function(t){return this._delegate.compileModuleAsync(t)},t.prototype.compileModuleAndAllComponentsSync=function(t){var e=this._delegate.compileModuleAndAllComponentsSync(t);return{ngModuleFactory:e.ngModuleFactory,componentFactories:e.componentFactories}},t.prototype.compileModuleAndAllComponentsAsync=function(t){return this._delegate.compileModuleAndAllComponentsAsync(t).then(function(t){return{ngModuleFactory:t.ngModuleFactory,componentFactories:t.componentFactories}})},t.prototype.loadAotSummaries=function(t){this._delegate.loadAotSummaries(t)},t.prototype.hasAotSummary=function(t){return this._delegate.hasAotSummary(t)},t.prototype.getComponentFactory=function(t){return this._delegate.getComponentFactory(t)},t.prototype.clearCache=function(){this._delegate.clearCache()},t.prototype.clearCacheFor=function(t){this._delegate.clearCacheFor(t)},t}(),g=[{provide:i.b,useValue:new l},{provide:i.s,useValue:h},{provide:i.l,deps:[]},{provide:i.v,useExisting:i.l},{provide:o._20,deps:[]},{provide:i.m,deps:[]},{provide:i.p,deps:[i.m]},{provide:f,useClass:i.h,deps:[]},{provide:i.i,useFactory:function(t,e,n,r,s){var a=(e=e||"")?r.missingTranslation:o.G.Ignore;return new i.i(t,e,n,a,s)},deps:[f,[new o.M,new o.y(o._2)],[new o.M,new o.y(o._3)],[i.c],[o._20]]},{provide:i.h,useExisting:i.i},{provide:i.w,deps:[i.c,i.b,i.p,i.g,i.i,o._20]},{provide:i.d,deps:[i.s,i.x,i.h,i.c]},{provide:i.a,deps:[i.c,i.h,i.o,i.e,i.q,i.v,i.g,i.d,o._20,[o.M,i.t],i.b,[o.M,p]]},d,{provide:i.u,deps:[i.x]},{provide:i.y,deps:[i.b]},{provide:i.n,deps:[i.b]},{provide:i.c,useValue:new i.c},{provide:o.k,useClass:m,deps:[o.B,i.a,i.w,i.u,i.y,i.n,i.v,i.b,i.c,o._20]},{provide:i.f,deps:[]},{provide:i.g,useExisting:i.f},{provide:i.x,deps:[o.O]},{provide:i.e,deps:[i.b]},{provide:i.q,deps:[i.b]},{provide:i.o,deps:[i.b]}],v=function(){function t(t){var e={useJit:!0,defaultEncapsulation:o._10.Emulated,missingTranslation:o.G.Warning,enableLegacyTemplate:!1};this._defaultOptions=[e].concat(t)}return t.prototype.createCompiler=function(t){void 0===t&&(t=[]);var e,n,r,s={useJit:y((e=this._defaultOptions.concat(t)).map(function(t){return t.useJit})),defaultEncapsulation:y(e.map(function(t){return t.defaultEncapsulation})),providers:(n=e.map(function(t){return t.providers}),r=[],n.forEach(function(t){return t&&r.push.apply(r,t)}),r),missingTranslation:y(e.map(function(t){return t.missingTranslation})),enableLegacyTemplate:y(e.map(function(t){return t.enableLegacyTemplate})),preserveWhitespaces:y(e.map(function(t){return t.preserveWhitespaces}))};return o.B.create([g,{provide:i.c,useFactory:function(){return new i.c({useJit:s.useJit,jitDevMode:Object(o._16)(),defaultEncapsulation:s.defaultEncapsulation,missingTranslation:s.missingTranslation,enableLegacyTemplate:s.enableLegacyTemplate,preserveWhitespaces:s.preserveWhitespaces})},deps:[]},s.providers]).get(o.k)},t}();function y(t){for(var e=t.length-1;e>=0;e--)if(void 0!==t[e])return t[e]}var b=Object(o._12)(o._17,"coreDynamic",[{provide:o.h,useValue:{},multi:!0},{provide:o.l,useClass:v,deps:[o.h]}]),_=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(c.b)(e,t),e.prototype.get=function(t){var e,n,r=new Promise(function(t,r){e=t,n=r}),i=new XMLHttpRequest;return i.open("GET",t,!0),i.responseType="text",i.onload=function(){var r=i.response||i.responseText,o=1223===i.status?204:i.status;0===o&&(o=r?200:0),200<=o&&o<=300?e(r):n("Failed to load "+t)},i.onerror=function(){n("Failed to load "+t)},i.send(),r},e.decorators=[{type:o.z}],e.ctorParameters=function(){return[]},e}(i.s),j=[a.g,{provide:o.h,useValue:{providers:[{provide:i.s,useClass:_,deps:[]}]},multi:!0},{provide:o.P,useValue:s.g}],x=function(t){function e(){var e=t.call(this)||this;if(e._cache=o._30.$templateCache,null==e._cache)throw new Error("CachedResourceLoader: Template cache was not found in $templateCache.");return e}return Object(c.b)(e,t),e.prototype.get=function(t){return this._cache.hasOwnProperty(t)?Promise.resolve(this._cache[t]):Promise.reject("CachedResourceLoader: Did not find cached template for "+t)},e}(i.s),w=(new o._6("5.1.0"),i.s,Object(o._12)(b,"browserDynamic",j))},"../../../platform-browser/esm5/animations.js":function(t,e,n){"use strict";n.d(e,"a",function(){return b});var r=n("../../../core/esm5/core.js"),i=n("../../../platform-browser/esm5/platform-browser.js"),o=n("../../../../tslib/tslib.es6.js"),s=n("../../../animations/esm5/animations.js"),a=n("../../../animations/esm5/browser.js"),c=function(t){function e(e,n){var i=t.call(this)||this;i._nextAnimationId=0;var o={id:"0",encapsulation:r._10.None,styles:[],data:{animation:[]}};return i._renderer=e.createRenderer(n.body,o),i}return Object(o.b)(e,t),e.prototype.build=function(t){var e=this._nextAnimationId.toString();this._nextAnimationId++;var n=Array.isArray(t)?Object(s.g)(t):t;return p(this._renderer,null,e,"register",[n]),new u(e,this._renderer)},e.decorators=[{type:r.z}],e.ctorParameters=function(){return[{type:r.V},{type:void 0,decorators:[{type:r.y,args:[i.b]}]}]},e}(s.b),u=function(t){function e(e,n){var r=t.call(this)||this;return r._id=e,r._renderer=n,r}return Object(o.b)(e,t),e.prototype.create=function(t,e){return new l(this._id,t,e||{},this._renderer)},e}(s.c),l=function(){function t(t,e,n,r){this.id=t,this.element=e,this._renderer=r,this.parentPlayer=null,this._started=!1,this.totalTime=0,this._command("create",n)}return t.prototype._listen=function(t,e){return this._renderer.listen(this.element,"@@"+this.id+":"+t,e)},t.prototype._command=function(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return p(this._renderer,this.element,this.id,t,e)},t.prototype.onDone=function(t){this._listen("done",t)},t.prototype.onStart=function(t){this._listen("start",t)},t.prototype.onDestroy=function(t){this._listen("destroy",t)},t.prototype.init=function(){this._command("init")},t.prototype.hasStarted=function(){return this._started},t.prototype.play=function(){this._command("play"),this._started=!0},t.prototype.pause=function(){this._command("pause")},t.prototype.restart=function(){this._command("restart")},t.prototype.finish=function(){this._command("finish")},t.prototype.destroy=function(){this._command("destroy")},t.prototype.reset=function(){this._command("reset")},t.prototype.setPosition=function(t){this._command("setPosition",t)},t.prototype.getPosition=function(){return 0},t}();function p(t,e,n,r,i){return t.setProperty(e,"@@"+n+":"+r,i)}var d=function(){function t(t,e,n){this.delegate=t,this.engine=e,this._zone=n,this._currentId=0,this._microtaskId=1,this._animationCallbacksBuffer=[],this._rendererCache=new Map,this._cdRecurDepth=0,e.onRemovalComplete=function(t,e){e&&e.parentNode(t)&&e.removeChild(t.parentNode,t)}}return t.prototype.createRenderer=function(t,e){var n=this,r=this.delegate.createRenderer(t,e);if(!(t&&e&&e.data&&e.data.animation)){var i=this._rendererCache.get(r);return i||(i=new h("",r,this.engine),this._rendererCache.set(r,i)),i}var o=e.id,s=e.id+"-"+this._currentId;return this._currentId++,this.engine.register(s,t),e.data.animation.forEach(function(e){return n.engine.registerTrigger(o,s,t,e.name,e)}),new f(this,s,r,this.engine)},t.prototype.begin=function(){this._cdRecurDepth++,this.delegate.begin&&this.delegate.begin()},t.prototype._scheduleCountTask=function(){var t=this;Zone.current.scheduleMicroTask("incremenet the animation microtask",function(){return t._microtaskId++})},t.prototype.scheduleListenerCallback=function(t,e,n){var r=this;t>=0&&t<this._microtaskId?this._zone.run(function(){return e(n)}):(0==this._animationCallbacksBuffer.length&&Promise.resolve(null).then(function(){r._zone.run(function(){r._animationCallbacksBuffer.forEach(function(t){(0,t[0])(t[1])}),r._animationCallbacksBuffer=[]})}),this._animationCallbacksBuffer.push([e,n]))},t.prototype.end=function(){var t=this;this._cdRecurDepth--,0==this._cdRecurDepth&&this._zone.runOutsideAngular(function(){t._scheduleCountTask(),t.engine.flush(t._microtaskId)}),this.delegate.end&&this.delegate.end()},t.prototype.whenRenderingDone=function(){return this.engine.whenRenderingDone()},t.decorators=[{type:r.z}],t.ctorParameters=function(){return[{type:r.V},{type:a.b},{type:r.L}]},t}(),h=function(){function t(t,e,n){this.namespaceId=t,this.delegate=e,this.engine=n,this.destroyNode=this.delegate.destroyNode?function(t){return e.destroyNode(t)}:null}return Object.defineProperty(t.prototype,"data",{get:function(){return this.delegate.data},enumerable:!0,configurable:!0}),t.prototype.destroy=function(){this.engine.destroy(this.namespaceId,this.delegate),this.delegate.destroy()},t.prototype.createElement=function(t,e){return this.delegate.createElement(t,e)},t.prototype.createComment=function(t){return this.delegate.createComment(t)},t.prototype.createText=function(t){return this.delegate.createText(t)},t.prototype.appendChild=function(t,e){this.delegate.appendChild(t,e),this.engine.onInsert(this.namespaceId,e,t,!1)},t.prototype.insertBefore=function(t,e,n){this.delegate.insertBefore(t,e,n),this.engine.onInsert(this.namespaceId,e,t,!0)},t.prototype.removeChild=function(t,e){this.engine.onRemove(this.namespaceId,e,this.delegate)},t.prototype.selectRootElement=function(t){return this.delegate.selectRootElement(t)},t.prototype.parentNode=function(t){return this.delegate.parentNode(t)},t.prototype.nextSibling=function(t){return this.delegate.nextSibling(t)},t.prototype.setAttribute=function(t,e,n,r){this.delegate.setAttribute(t,e,n,r)},t.prototype.removeAttribute=function(t,e,n){this.delegate.removeAttribute(t,e,n)},t.prototype.addClass=function(t,e){this.delegate.addClass(t,e)},t.prototype.removeClass=function(t,e){this.delegate.removeClass(t,e)},t.prototype.setStyle=function(t,e,n,r){this.delegate.setStyle(t,e,n,r)},t.prototype.removeStyle=function(t,e,n){this.delegate.removeStyle(t,e,n)},t.prototype.setProperty=function(t,e,n){"@"==e.charAt(0)&&"@.disabled"==e?this.disableAnimations(t,!!n):this.delegate.setProperty(t,e,n)},t.prototype.setValue=function(t,e){this.delegate.setValue(t,e)},t.prototype.listen=function(t,e,n){return this.delegate.listen(t,e,n)},t.prototype.disableAnimations=function(t,e){this.engine.disableAnimations(t,e)},t}(),f=function(t){function e(e,n,r,i){var o=t.call(this,n,r,i)||this;return o.factory=e,o.namespaceId=n,o}return Object(o.b)(e,t),e.prototype.setProperty=function(t,e,n){"@"==e.charAt(0)?"."==e.charAt(1)&&"@.disabled"==e?(n=void 0===n||!!n,this.disableAnimations(t,n)):this.engine.process(this.namespaceId,t,e.substr(1),n):this.delegate.setProperty(t,e,n)},e.prototype.listen=function(t,e,n){var r,i,o,s,a,c=this;if("@"==e.charAt(0)){var u=function(t){switch(t){case"body":return document.body;case"document":return document;case"window":return window;default:return t}}(t),l=e.substr(1),p="";return"@"!=l.charAt(0)&&(i=(r=l).indexOf("."),o=r.substring(0,i),s=r.substr(i+1),l=(a=[o,s])[0],p=a[1]),this.engine.listen(this.namespaceId,u,l,p,function(t){var e=t._data||-1;c.factory.scheduleListenerCallback(e,n,t)})}return this.delegate.listen(t,e,n)},e}(h);var m=function(t){function e(e,n){return t.call(this,e,n)||this}return Object(o.b)(e,t),e.decorators=[{type:r.z}],e.ctorParameters=function(){return[{type:a.a},{type:a.c}]},e}(a.b);var g=[{provide:s.b,useClass:c},{provide:a.c,useFactory:function(){return new a.f}},{provide:a.b,useClass:m},{provide:r.V,useFactory:function(t,e,n){return new d(t,e,n)},deps:[i.f,a.b,r.L]}],v=[{provide:a.a,useFactory:function(){return Object(a.g)()?new a.e:new a.d}}].concat(g),y=[{provide:a.a,useClass:a.d}].concat(g),b=function(){function t(){}return t.decorators=[{type:r.H,args:[{exports:[i.a],providers:v}]}],t.ctorParameters=function(){return[]},t}();!function(){function t(){}t.decorators=[{type:r.H,args:[{exports:[i.a],providers:y}]}],t.ctorParameters=function(){return[]}}()},"../../../platform-browser/esm5/platform-browser.js":function(t,e,n){"use strict";n.d(e,"a",function(){return $t}),n.d(e,"b",function(){return g}),n.d(e,"d",function(){return st}),n.d(e,"e",function(){return at}),n.d(e,"c",function(){return Vt}),n.d(e,"g",function(){return Xt}),n.d(e,"h",function(){return a}),n.d(e,"f",function(){return B});var r=n("../../../common/esm5/common.js"),i=n("../../../core/esm5/core.js"),o=n("../../../../tslib/tslib.es6.js"),s=null;function a(){return s}var c,u=function(t){function e(){var e=t.call(this)||this;e._animationPrefix=null,e._transitionEnd=null;try{var n=e.createElement("div",document);if(null!=e.getStyle(n,"animationName"))e._animationPrefix="";else for(var r=["Webkit","Moz","O","ms"],i=0;i<r.length;i++)if(null!=e.getStyle(n,r[i]+"AnimationName")){e._animationPrefix="-"+r[i].toLowerCase()+"-";break}var o={WebkitTransition:"webkitTransitionEnd",MozTransition:"transitionend",OTransition:"oTransitionEnd otransitionend",transition:"transitionend"};Object.keys(o).forEach(function(t){null!=e.getStyle(n,t)&&(e._transitionEnd=o[t])})}catch(t){e._animationPrefix=null,e._transitionEnd=null}return e}return Object(o.b)(e,t),e.prototype.getDistributedNodes=function(t){return t.getDistributedNodes()},e.prototype.resolveAndSetHref=function(t,e,n){t.href=null==n?e:e+"/../"+n},e.prototype.supportsDOMEvents=function(){return!0},e.prototype.supportsNativeShadowDOM=function(){return"function"==typeof document.body.createShadowRoot},e.prototype.getAnimationPrefix=function(){return this._animationPrefix?this._animationPrefix:""},e.prototype.getTransitionEnd=function(){return this._transitionEnd?this._transitionEnd:""},e.prototype.supportsAnimation=function(){return null!=this._animationPrefix&&null!=this._transitionEnd},e}(function(){function t(){this.resourceLoaderType=null}return Object.defineProperty(t.prototype,"attrToPropMap",{get:function(){return this._attrToPropMap},set:function(t){this._attrToPropMap=t},enumerable:!0,configurable:!0}),t}()),l={class:"className",innerHtml:"innerHTML",readonly:"readOnly",tabindex:"tabIndex"},p={"\b":"Backspace","\t":"Tab","":"Delete","":"Escape",Del:"Delete",Esc:"Escape",Left:"ArrowLeft",Right:"ArrowRight",Up:"ArrowUp",Down:"ArrowDown",Menu:"ContextMenu",Scroll:"ScrollLock",Win:"OS"},d={A:"1",B:"2",C:"3",D:"4",E:"5",F:"6",G:"7",H:"8",I:"9",J:"*",K:"+",M:"-",N:".",O:"/","`":"0","":"NumLock"};i._30.Node&&(c=i._30.Node.prototype.contains||function(t){return!!(16&this.compareDocumentPosition(t))});var h,f=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.b)(e,t),e.prototype.parse=function(t){throw new Error("parse not implemented")},e.makeCurrent=function(){var t;t=new e,s||(s=t)},e.prototype.hasProperty=function(t,e){return e in t},e.prototype.setProperty=function(t,e,n){t[e]=n},e.prototype.getProperty=function(t,e){return t[e]},e.prototype.invoke=function(t,e,n){var r;(r=t)[e].apply(r,n)},e.prototype.logError=function(t){window.console&&(console.error?console.error(t):console.log(t))},e.prototype.log=function(t){window.console&&window.console.log&&window.console.log(t)},e.prototype.logGroup=function(t){window.console&&window.console.group&&window.console.group(t)},e.prototype.logGroupEnd=function(){window.console&&window.console.groupEnd&&window.console.groupEnd()},Object.defineProperty(e.prototype,"attrToPropMap",{get:function(){return l},enumerable:!0,configurable:!0}),e.prototype.contains=function(t,e){return c.call(t,e)},e.prototype.querySelector=function(t,e){return t.querySelector(e)},e.prototype.querySelectorAll=function(t,e){return t.querySelectorAll(e)},e.prototype.on=function(t,e,n){t.addEventListener(e,n,!1)},e.prototype.onAndCancel=function(t,e,n){return t.addEventListener(e,n,!1),function(){t.removeEventListener(e,n,!1)}},e.prototype.dispatchEvent=function(t,e){t.dispatchEvent(e)},e.prototype.createMouseEvent=function(t){var e=this.getDefaultDocument().createEvent("MouseEvent");return e.initEvent(t,!0,!0),e},e.prototype.createEvent=function(t){var e=this.getDefaultDocument().createEvent("Event");return e.initEvent(t,!0,!0),e},e.prototype.preventDefault=function(t){t.preventDefault(),t.returnValue=!1},e.prototype.isPrevented=function(t){return t.defaultPrevented||null!=t.returnValue&&!t.returnValue},e.prototype.getInnerHTML=function(t){return t.innerHTML},e.prototype.getTemplateContent=function(t){return"content"in t&&this.isTemplateElement(t)?t.content:null},e.prototype.getOuterHTML=function(t){return t.outerHTML},e.prototype.nodeName=function(t){return t.nodeName},e.prototype.nodeValue=function(t){return t.nodeValue},e.prototype.type=function(t){return t.type},e.prototype.content=function(t){return this.hasProperty(t,"content")?t.content:t},e.prototype.firstChild=function(t){return t.firstChild},e.prototype.nextSibling=function(t){return t.nextSibling},e.prototype.parentElement=function(t){return t.parentNode},e.prototype.childNodes=function(t){return t.childNodes},e.prototype.childNodesAsList=function(t){for(var e=t.childNodes,n=new Array(e.length),r=0;r<e.length;r++)n[r]=e[r];return n},e.prototype.clearNodes=function(t){for(;t.firstChild;)t.removeChild(t.firstChild)},e.prototype.appendChild=function(t,e){t.appendChild(e)},e.prototype.removeChild=function(t,e){t.removeChild(e)},e.prototype.replaceChild=function(t,e,n){t.replaceChild(e,n)},e.prototype.remove=function(t){return t.parentNode&&t.parentNode.removeChild(t),t},e.prototype.insertBefore=function(t,e,n){t.insertBefore(n,e)},e.prototype.insertAllBefore=function(t,e,n){n.forEach(function(n){return t.insertBefore(n,e)})},e.prototype.insertAfter=function(t,e,n){t.insertBefore(n,e.nextSibling)},e.prototype.setInnerHTML=function(t,e){t.innerHTML=e},e.prototype.getText=function(t){return t.textContent},e.prototype.setText=function(t,e){t.textContent=e},e.prototype.getValue=function(t){return t.value},e.prototype.setValue=function(t,e){t.value=e},e.prototype.getChecked=function(t){return t.checked},e.prototype.setChecked=function(t,e){t.checked=e},e.prototype.createComment=function(t){return this.getDefaultDocument().createComment(t)},e.prototype.createTemplate=function(t){var e=this.getDefaultDocument().createElement("template");return e.innerHTML=t,e},e.prototype.createElement=function(t,e){return(e=e||this.getDefaultDocument()).createElement(t)},e.prototype.createElementNS=function(t,e,n){return(n=n||this.getDefaultDocument()).createElementNS(t,e)},e.prototype.createTextNode=function(t,e){return(e=e||this.getDefaultDocument()).createTextNode(t)},e.prototype.createScriptTag=function(t,e,n){var r=(n=n||this.getDefaultDocument()).createElement("SCRIPT");return r.setAttribute(t,e),r},e.prototype.createStyleElement=function(t,e){var n=(e=e||this.getDefaultDocument()).createElement("style");return this.appendChild(n,this.createTextNode(t,e)),n},e.prototype.createShadowRoot=function(t){return t.createShadowRoot()},e.prototype.getShadowRoot=function(t){return t.shadowRoot},e.prototype.getHost=function(t){return t.host},e.prototype.clone=function(t){return t.cloneNode(!0)},e.prototype.getElementsByClassName=function(t,e){return t.getElementsByClassName(e)},e.prototype.getElementsByTagName=function(t,e){return t.getElementsByTagName(e)},e.prototype.classList=function(t){return Array.prototype.slice.call(t.classList,0)},e.prototype.addClass=function(t,e){t.classList.add(e)},e.prototype.removeClass=function(t,e){t.classList.remove(e)},e.prototype.hasClass=function(t,e){return t.classList.contains(e)},e.prototype.setStyle=function(t,e,n){t.style[e]=n},e.prototype.removeStyle=function(t,e){t.style[e]=""},e.prototype.getStyle=function(t,e){return t.style[e]},e.prototype.hasStyle=function(t,e,n){var r=this.getStyle(t,e)||"";return n?r==n:r.length>0},e.prototype.tagName=function(t){return t.tagName},e.prototype.attributeMap=function(t){for(var e=new Map,n=t.attributes,r=0;r<n.length;r++){var i=n.item(r);e.set(i.name,i.value)}return e},e.prototype.hasAttribute=function(t,e){return t.hasAttribute(e)},e.prototype.hasAttributeNS=function(t,e,n){return t.hasAttributeNS(e,n)},e.prototype.getAttribute=function(t,e){return t.getAttribute(e)},e.prototype.getAttributeNS=function(t,e,n){return t.getAttributeNS(e,n)},e.prototype.setAttribute=function(t,e,n){t.setAttribute(e,n)},e.prototype.setAttributeNS=function(t,e,n,r){t.setAttributeNS(e,n,r)},e.prototype.removeAttribute=function(t,e){t.removeAttribute(e)},e.prototype.removeAttributeNS=function(t,e,n){t.removeAttributeNS(e,n)},e.prototype.templateAwareRoot=function(t){return this.isTemplateElement(t)?this.content(t):t},e.prototype.createHtmlDocument=function(){return document.implementation.createHTMLDocument("fakeTitle")},e.prototype.getDefaultDocument=function(){return document},e.prototype.getBoundingClientRect=function(t){try{return t.getBoundingClientRect()}catch(t){return{top:0,bottom:0,left:0,right:0,width:0,height:0}}},e.prototype.getTitle=function(t){return t.title},e.prototype.setTitle=function(t,e){t.title=e||""},e.prototype.elementMatches=function(t,e){return!!this.isElementNode(t)&&(t.matches&&t.matches(e)||t.msMatchesSelector&&t.msMatchesSelector(e)||t.webkitMatchesSelector&&t.webkitMatchesSelector(e))},e.prototype.isTemplateElement=function(t){return this.isElementNode(t)&&"TEMPLATE"===t.nodeName},e.prototype.isTextNode=function(t){return t.nodeType===Node.TEXT_NODE},e.prototype.isCommentNode=function(t){return t.nodeType===Node.COMMENT_NODE},e.prototype.isElementNode=function(t){return t.nodeType===Node.ELEMENT_NODE},e.prototype.hasShadowRoot=function(t){return null!=t.shadowRoot&&t instanceof HTMLElement},e.prototype.isShadowRoot=function(t){return t instanceof DocumentFragment},e.prototype.importIntoDoc=function(t){return document.importNode(this.templateAwareRoot(t),!0)},e.prototype.adoptNode=function(t){return document.adoptNode(t)},e.prototype.getHref=function(t){return t.getAttribute("href")},e.prototype.getEventKey=function(t){var e=t.key;if(null==e){if(null==(e=t.keyIdentifier))return"Unidentified";e.startsWith("U+")&&(e=String.fromCharCode(parseInt(e.substring(2),16)),3===t.location&&d.hasOwnProperty(e)&&(e=d[e]))}return p[e]||e},e.prototype.getGlobalEventTarget=function(t,e){return"window"===e?window:"document"===e?t:"body"===e?t.body:null},e.prototype.getHistory=function(){return window.history},e.prototype.getLocation=function(){return window.location},e.prototype.getBaseHref=function(t){var e=function(){if(!m&&!(m=document.querySelector("base")))return null;return m.getAttribute("href")}();return null==e?null:function(t){h||(h=document.createElement("a"));return h.setAttribute("href",t),"/"===h.pathname.charAt(0)?h.pathname:"/"+h.pathname}(e)},e.prototype.resetBaseElement=function(){m=null},e.prototype.getUserAgent=function(){return window.navigator.userAgent},e.prototype.setData=function(t,e,n){this.setAttribute(t,"data-"+e,n)},e.prototype.getData=function(t,e){return this.getAttribute(t,"data-"+e)},e.prototype.getComputedStyle=function(t){return getComputedStyle(t)},e.prototype.supportsWebAnimation=function(){return"function"==typeof Element.prototype.animate},e.prototype.performanceNow=function(){return window.performance&&window.performance.now?window.performance.now():(new Date).getTime()},e.prototype.supportsCookies=function(){return!0},e.prototype.getCookie=function(t){return Object(r.h)(document.cookie,t)},e.prototype.setCookie=function(t,e){document.cookie=encodeURIComponent(t)+"="+encodeURIComponent(e)},e}(u),m=null;var g=r.b;function v(){return!!window.history.pushState}var y=function(t){function e(e){var n=t.call(this)||this;return n._doc=e,n._init(),n}return Object(o.b)(e,t),e.prototype._init=function(){this.location=a().getLocation(),this._history=a().getHistory()},e.prototype.getBaseHrefFromDOM=function(){return a().getBaseHref(this._doc)},e.prototype.onPopState=function(t){a().getGlobalEventTarget(this._doc,"window").addEventListener("popstate",t,!1)},e.prototype.onHashChange=function(t){a().getGlobalEventTarget(this._doc,"window").addEventListener("hashchange",t,!1)},Object.defineProperty(e.prototype,"pathname",{get:function(){return this.location.pathname},set:function(t){this.location.pathname=t},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"search",{get:function(){return this.location.search},enumerable:!0,configurable:!0}),Object.defineProperty(e.prototype,"hash",{get:function(){return this.location.hash},enumerable:!0,configurable:!0}),e.prototype.pushState=function(t,e,n){v()?this._history.pushState(t,e,n):this.location.hash=n},e.prototype.replaceState=function(t,e,n){v()?this._history.replaceState(t,e,n):this.location.hash=n},e.prototype.forward=function(){this._history.forward()},e.prototype.back=function(){this._history.back()},e.decorators=[{type:i.z}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.y,args:[g]}]}]},e}(r.f),b=function(){function t(t){this._doc=t,this._dom=a()}return t.prototype.addTag=function(t,e){return void 0===e&&(e=!1),t?this._getOrCreateElement(t,e):null},t.prototype.addTags=function(t,e){var n=this;return void 0===e&&(e=!1),t?t.reduce(function(t,r){return r&&t.push(n._getOrCreateElement(r,e)),t},[]):[]},t.prototype.getTag=function(t){return t?this._dom.querySelector(this._doc,"meta["+t+"]")||null:null},t.prototype.getTags=function(t){if(!t)return[];var e=this._dom.querySelectorAll(this._doc,"meta["+t+"]");return e?[].slice.call(e):[]},t.prototype.updateTag=function(t,e){if(!t)return null;e=e||this._parseSelector(t);var n=this.getTag(e);return n?this._setMetaElementAttributes(t,n):this._getOrCreateElement(t,!0)},t.prototype.removeTag=function(t){this.removeTagElement(this.getTag(t))},t.prototype.removeTagElement=function(t){t&&this._dom.remove(t)},t.prototype._getOrCreateElement=function(t,e){if(void 0===e&&(e=!1),!e){var n=this._parseSelector(t),r=this.getTag(n);if(r&&this._containsAttributes(t,r))return r}var i=this._dom.createElement("meta");this._setMetaElementAttributes(t,i);var o=this._dom.getElementsByTagName(this._doc,"head")[0];return this._dom.appendChild(o,i),i},t.prototype._setMetaElementAttributes=function(t,e){var n=this;return Object.keys(t).forEach(function(r){return n._dom.setAttribute(e,r,t[r])}),e},t.prototype._parseSelector=function(t){var e=t.name?"name":"property";return e+'="'+t[e]+'"'},t.prototype._containsAttributes=function(t,e){var n=this;return Object.keys(t).every(function(r){return n._dom.getAttribute(e,r)===t[r]})},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.y,args:[g]}]}]},t}(),_=new i.A("TRANSITION_ID");var j=[{provide:i.c,useFactory:function(t,e,n){return function(){n.get(i.d).donePromise.then(function(){var n=a();Array.prototype.slice.apply(n.querySelectorAll(e,"style[ng-transition]")).filter(function(e){return n.getAttribute(e,"ng-transition")===t}).forEach(function(t){return n.remove(t)})})}},deps:[_,g,i.B],multi:!0}],x=function(){function t(){}return t.init=function(){Object(i._18)(new t)},t.prototype.addToWindow=function(t){i._30.getAngularTestability=function(e,n){void 0===n&&(n=!0);var r=t.findTestabilityInTree(e,n);if(null==r)throw new Error("Could not find testability for element.");return r},i._30.getAllAngularTestabilities=function(){return t.getAllTestabilities()},i._30.getAllAngularRootElements=function(){return t.getAllRootElements()};i._30.frameworkStabilizers||(i._30.frameworkStabilizers=[]),i._30.frameworkStabilizers.push(function(t){var e=i._30.getAllAngularTestabilities(),n=e.length,r=!1,o=function(e){r=r||e,0==--n&&t(r)};e.forEach(function(t){t.whenStable(o)})})},t.prototype.findTestabilityInTree=function(t,e,n){if(null==e)return null;var r=t.getTestability(e);return null!=r?r:n?a().isShadowRoot(e)?this.findTestabilityInTree(t,a().getHost(e),!0):this.findTestabilityInTree(t,a().parentElement(e),!0):null},t}(),w=function(){function t(t){this._doc=t}return t.prototype.getTitle=function(){return a().getTitle(this._doc)},t.prototype.setTitle=function(t){a().setTitle(this._doc,t)},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.y,args:[g]}]}]},t}();function O(t,e){"undefined"!=typeof COMPILED&&COMPILED||((i._30.ng=i._30.ng||{})[t]=e)}var C={ApplicationRef:i.f,NgZone:i.L},k="probe",E="coreTokens";function S(t){return Object(i._15)(t)}var P=[{provide:i.c,useFactory:function(t){return O(k,S),O(E,Object(o.a)({},C,(t||[]).reduce(function(t,e){return t[e.name]=e.token,t},{}))),function(){return S}},deps:[[i.K,new i.M]],multi:!0}],M=new i.A("EventManagerPlugins"),A=function(){function t(t,e){var n=this;this._zone=e,this._eventNameToPlugin=new Map,t.forEach(function(t){return t.manager=n}),this._plugins=t.slice().reverse()}return t.prototype.addEventListener=function(t,e,n){return this._findPluginFor(e).addEventListener(t,e,n)},t.prototype.addGlobalEventListener=function(t,e,n){return this._findPluginFor(e).addGlobalEventListener(t,e,n)},t.prototype.getZone=function(){return this._zone},t.prototype._findPluginFor=function(t){var e=this._eventNameToPlugin.get(t);if(e)return e;for(var n=this._plugins,r=0;r<n.length;r++){var i=n[r];if(i.supports(t))return this._eventNameToPlugin.set(t,i),i}throw new Error("No event manager plugin found for event "+t)},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[{type:Array,decorators:[{type:i.y,args:[M]}]},{type:i.L}]},t}(),T=function(){function t(t){this._doc=t}return t.prototype.addGlobalEventListener=function(t,e,n){var r=a().getGlobalEventTarget(this._doc,t);if(!r)throw new Error("Unsupported event target "+r+" for event "+e);return this.addEventListener(r,e,n)},t}(),D=function(){function t(){this._stylesSet=new Set}return t.prototype.addStyles=function(t){var e=this,n=new Set;t.forEach(function(t){e._stylesSet.has(t)||(e._stylesSet.add(t),n.add(t))}),this.onStylesAdded(n)},t.prototype.onStylesAdded=function(t){},t.prototype.getAllStyles=function(){return Array.from(this._stylesSet)},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[]},t}(),I=function(t){function e(e){var n=t.call(this)||this;return n._doc=e,n._hostNodes=new Set,n._styleNodes=new Set,n._hostNodes.add(e.head),n}return Object(o.b)(e,t),e.prototype._addStylesToHost=function(t,e){var n=this;t.forEach(function(t){var r=n._doc.createElement("style");r.textContent=t,n._styleNodes.add(e.appendChild(r))})},e.prototype.addHost=function(t){this._addStylesToHost(this._stylesSet,t),this._hostNodes.add(t)},e.prototype.removeHost=function(t){this._hostNodes.delete(t)},e.prototype.onStylesAdded=function(t){var e=this;this._hostNodes.forEach(function(n){return e._addStylesToHost(t,n)})},e.prototype.ngOnDestroy=function(){this._styleNodes.forEach(function(t){return a().remove(t)})},e.decorators=[{type:i.z}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.y,args:[g]}]}]},e}(D),R={svg:"http://www.w3.org/2000/svg",xhtml:"http://www.w3.org/1999/xhtml",xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"},N=/%COMP%/g,L="_nghost-%COMP%",F="_ngcontent-%COMP%";function V(t,e,n){for(var r=0;r<e.length;r++){var i=e[r];Array.isArray(i)?V(t,i,n):(i=i.replace(N,t),n.push(i))}return n}function z(t){return function(e){!1===t(e)&&(e.preventDefault(),e.returnValue=!1)}}var B=function(){function t(t,e){this.eventManager=t,this.sharedStylesHost=e,this.rendererByCompId=new Map,this.defaultRenderer=new q(t)}return t.prototype.createRenderer=function(t,e){if(!t||!e)return this.defaultRenderer;switch(e.encapsulation){case i._10.Emulated:var n=this.rendererByCompId.get(e.id);return n||(n=new G(this.eventManager,this.sharedStylesHost,e),this.rendererByCompId.set(e.id,n)),n.applyToHost(t),n;case i._10.Native:return new X(this.eventManager,this.sharedStylesHost,t,e);default:if(!this.rendererByCompId.has(e.id)){var r=V(e.id,e.styles,[]);this.sharedStylesHost.addStyles(r),this.rendererByCompId.set(e.id,this.defaultRenderer)}return this.defaultRenderer}},t.prototype.begin=function(){},t.prototype.end=function(){},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[{type:A},{type:I}]},t}(),q=function(){function t(t){this.eventManager=t,this.data=Object.create(null)}return t.prototype.destroy=function(){},t.prototype.createElement=function(t,e){return e?document.createElementNS(R[e],t):document.createElement(t)},t.prototype.createComment=function(t){return document.createComment(t)},t.prototype.createText=function(t){return document.createTextNode(t)},t.prototype.appendChild=function(t,e){t.appendChild(e)},t.prototype.insertBefore=function(t,e,n){t&&t.insertBefore(e,n)},t.prototype.removeChild=function(t,e){t&&t.removeChild(e)},t.prototype.selectRootElement=function(t){var e="string"==typeof t?document.querySelector(t):t;if(!e)throw new Error('The selector "'+t+'" did not match any elements');return e.textContent="",e},t.prototype.parentNode=function(t){return t.parentNode},t.prototype.nextSibling=function(t){return t.nextSibling},t.prototype.setAttribute=function(t,e,n,r){if(r){e=r+":"+e;var i=R[r];i?t.setAttributeNS(i,e,n):t.setAttribute(e,n)}else t.setAttribute(e,n)},t.prototype.removeAttribute=function(t,e,n){if(n){var r=R[n];r?t.removeAttributeNS(r,e):t.removeAttribute(n+":"+e)}else t.removeAttribute(e)},t.prototype.addClass=function(t,e){t.classList.add(e)},t.prototype.removeClass=function(t,e){t.classList.remove(e)},t.prototype.setStyle=function(t,e,n,r){r&i.W.DashCase?t.style.setProperty(e,n,r&i.W.Important?"important":""):t.style[e]=n},t.prototype.removeStyle=function(t,e,n){n&i.W.DashCase?t.style.removeProperty(e):t.style[e]=""},t.prototype.setProperty=function(t,e,n){U(e,"property"),t[e]=n},t.prototype.setValue=function(t,e){t.nodeValue=e},t.prototype.listen=function(t,e,n){return U(e,"listener"),"string"==typeof t?this.eventManager.addGlobalEventListener(t,e,z(n)):this.eventManager.addEventListener(t,e,z(n))},t}(),H="@".charCodeAt(0);function U(t,e){if(t.charCodeAt(0)===H)throw new Error("Found the synthetic "+e+" "+t+'. Please include either "BrowserAnimationsModule" or "NoopAnimationsModule" in your application.')}var W,G=function(t){function e(e,n,r){var i=t.call(this,e)||this;i.component=r;var o,s,a=V(r.id,r.styles,[]);return n.addStyles(a),i.contentAttr=(o=r.id,F.replace(N,o)),i.hostAttr=(s=r.id,L.replace(N,s)),i}return Object(o.b)(e,t),e.prototype.applyToHost=function(e){t.prototype.setAttribute.call(this,e,this.hostAttr,"")},e.prototype.createElement=function(e,n){var r=t.prototype.createElement.call(this,e,n);return t.prototype.setAttribute.call(this,r,this.contentAttr,""),r},e}(q),X=function(t){function e(e,n,r,i){var o=t.call(this,e)||this;o.sharedStylesHost=n,o.hostEl=r,o.component=i,o.shadowRoot=r.createShadowRoot(),o.sharedStylesHost.addHost(o.shadowRoot);for(var s=V(i.id,i.styles,[]),a=0;a<s.length;a++){var c=document.createElement("style");c.textContent=s[a],o.shadowRoot.appendChild(c)}return o}return Object(o.b)(e,t),e.prototype.nodeOrShadowRoot=function(t){return t===this.hostEl?this.shadowRoot:t},e.prototype.destroy=function(){this.sharedStylesHost.removeHost(this.shadowRoot)},e.prototype.appendChild=function(e,n){return t.prototype.appendChild.call(this,this.nodeOrShadowRoot(e),n)},e.prototype.insertBefore=function(e,n,r){return t.prototype.insertBefore.call(this,this.nodeOrShadowRoot(e),n,r)},e.prototype.removeChild=function(e,n){return t.prototype.removeChild.call(this,this.nodeOrShadowRoot(e),n)},e.prototype.parentNode=function(e){return this.nodeOrShadowRoot(t.prototype.parentNode.call(this,this.nodeOrShadowRoot(e)))},e}(q),Y="undefined"!=typeof Zone&&Zone.__symbol__||function(t){return"__zone_symbol__"+t},K=Y("addEventListener"),$=Y("removeEventListener"),Q={},Z="removeEventListener",J="__zone_symbol__propagationStopped",tt="__zone_symbol__stopImmediatePropagation",et="undefined"!=typeof Zone&&Zone[Y("BLACK_LISTED_EVENTS")];et&&(W={},et.forEach(function(t){W[t]=t}));var nt=function(t){return!!W&&W.hasOwnProperty(t)},rt=function(t){var e=Q[t.type];if(e){var n=this[e];if(n){var r=[t];if(1===n.length)return(s=n[0]).zone!==Zone.current?s.zone.run(s.handler,this,r):s.handler.apply(this,r);for(var i=n.slice(),o=0;o<i.length&&!0!==t[J];o++){var s;(s=i[o]).zone!==Zone.current?s.zone.run(s.handler,this,r):s.handler.apply(this,r)}}}},it=function(t){function e(e,n){var r=t.call(this,e)||this;return r.ngZone=n,r.patchEvent(),r}return Object(o.b)(e,t),e.prototype.patchEvent=function(){if(Event&&Event.prototype&&!Event.prototype[tt]){var t=Event.prototype[tt]=Event.prototype.stopImmediatePropagation;Event.prototype.stopImmediatePropagation=function(){this&&(this[J]=!0),t&&t.apply(this,arguments)}}},e.prototype.supports=function(t){return!0},e.prototype.addEventListener=function(t,e,n){var r=this,o=n;if(!t[K]||i.L.isInAngularZone()&&!nt(e))t.addEventListener(e,o,!1);else{var s=Q[e];s||(s=Q[e]=Y("ANGULAR"+e+"FALSE"));var a=t[s],c=a&&a.length>0;a||(a=t[s]=[]);var u=nt(e)?Zone.root:Zone.current;if(0===a.length)a.push({zone:u,handler:o});else{for(var l=!1,p=0;p<a.length;p++)if(a[p].handler===o){l=!0;break}l||a.push({zone:u,handler:o})}c||t[K](e,rt,!1)}return function(){return r.removeEventListener(t,e,o)}},e.prototype.removeEventListener=function(t,e,n){var r=t[$];if(!r)return t[Z].apply(t,[e,n,!1]);var i=Q[e],o=i&&t[i];if(!o)return t[Z].apply(t,[e,n,!1]);for(var s=!1,a=0;a<o.length;a++)if(o[a].handler===n){s=!0,o.splice(a,1);break}s?0===o.length&&r.apply(t,[e,rt,!1]):t[Z].apply(t,[e,n,!1])},e.decorators=[{type:i.z}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.y,args:[g]}]},{type:i.L}]},e}(T),ot={pan:!0,panstart:!0,panmove:!0,panend:!0,pancancel:!0,panleft:!0,panright:!0,panup:!0,pandown:!0,pinch:!0,pinchstart:!0,pinchmove:!0,pinchend:!0,pinchcancel:!0,pinchin:!0,pinchout:!0,press:!0,pressup:!0,rotate:!0,rotatestart:!0,rotatemove:!0,rotateend:!0,rotatecancel:!0,swipe:!0,swipeleft:!0,swiperight:!0,swipeup:!0,swipedown:!0,tap:!0},st=new i.A("HammerGestureConfig"),at=function(){function t(){this.events=[],this.overrides={}}return t.prototype.buildHammer=function(t){var e=new Hammer(t);e.get("pinch").set({enable:!0}),e.get("rotate").set({enable:!0});for(var n in this.overrides)e.get(n).set(this.overrides[n]);return e},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[]},t}(),ct=function(t){function e(e,n){var r=t.call(this,e)||this;return r._config=n,r}return Object(o.b)(e,t),e.prototype.supports=function(t){if(!ot.hasOwnProperty(t.toLowerCase())&&!this.isCustomEvent(t))return!1;if(!window.Hammer)throw new Error("Hammer.js is not loaded, can not bind "+t+" event");return!0},e.prototype.addEventListener=function(t,e,n){var r=this,i=this.manager.getZone();return e=e.toLowerCase(),i.runOutsideAngular(function(){var o=r._config.buildHammer(t),s=function(t){i.runGuarded(function(){n(t)})};return o.on(e,s),function(){return o.off(e,s)}})},e.prototype.isCustomEvent=function(t){return this._config.events.indexOf(t)>-1},e.decorators=[{type:i.z}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.y,args:[g]}]},{type:at,decorators:[{type:i.y,args:[st]}]}]},e}(T),ut=["alt","control","meta","shift"],lt={alt:function(t){return t.altKey},control:function(t){return t.ctrlKey},meta:function(t){return t.metaKey},shift:function(t){return t.shiftKey}},pt=function(t){function e(e){return t.call(this,e)||this}return Object(o.b)(e,t),e.prototype.supports=function(t){return null!=e.parseEventName(t)},e.prototype.addEventListener=function(t,n,r){var i=e.parseEventName(n),o=e.eventCallback(i.fullKey,r,this.manager.getZone());return this.manager.getZone().runOutsideAngular(function(){return a().onAndCancel(t,i.domEventName,o)})},e.parseEventName=function(t){var n=t.toLowerCase().split("."),r=n.shift();if(0===n.length||"keydown"!==r&&"keyup"!==r)return null;var i=e._normalizeKey(n.pop()),o="";if(ut.forEach(function(t){var e=n.indexOf(t);e>-1&&(n.splice(e,1),o+=t+".")}),o+=i,0!=n.length||0===i.length)return null;var s={};return s.domEventName=r,s.fullKey=o,s},e.getEventFullKey=function(t){var e="",n=a().getEventKey(t);return" "===(n=n.toLowerCase())?n="space":"."===n&&(n="dot"),ut.forEach(function(r){r!=n&&((0,lt[r])(t)&&(e+=r+"."))}),e+=n},e.eventCallback=function(t,n,r){return function(i){e.getEventFullKey(i)===t&&r.runGuarded(function(){return n(i)})}},e._normalizeKey=function(t){switch(t){case"esc":return"escape";default:return t}},e.decorators=[{type:i.z}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.y,args:[g]}]}]},e}(T),dt=/^(?:(?:https?|mailto|ftp|tel|file):|[^&:/?#]*(?:[/?#]|$))/gi,ht=/^data:(?:image\/(?:bmp|gif|jpeg|jpg|png|tiff|webp)|video\/(?:mpeg|mp4|ogg|webm)|audio\/(?:mp3|oga|ogg|opus));base64,[a-z0-9+\/]+=*$/i;function ft(t){return(t=String(t)).match(dt)||t.match(ht)?t:(Object(i._16)()&&a().log("WARNING: sanitizing unsafe URL value "+t+" (see http://g.co/ng/security#xss)"),"unsafe:"+t)}var mt=null,gt=null;function vt(t){for(var e={},n=0,r=t.split(",");n<r.length;n++){e[r[n]]=!0}return e}function yt(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];for(var n={},r=0,i=t;r<i.length;r++){var o=i[r];for(var s in o)o.hasOwnProperty(s)&&(n[s]=!0)}return n}var bt=vt("area,br,col,hr,img,wbr"),_t=vt("colgroup,dd,dt,li,p,tbody,td,tfoot,th,thead,tr"),jt=vt("rp,rt"),xt=yt(jt,_t),wt=yt(_t,vt("address,article,aside,blockquote,caption,center,del,details,dialog,dir,div,dl,figure,figcaption,footer,h1,h2,h3,h4,h5,h6,header,hgroup,hr,ins,main,map,menu,nav,ol,pre,section,summary,table,ul")),Ot=yt(jt,vt("a,abbr,acronym,audio,b,bdi,bdo,big,br,cite,code,del,dfn,em,font,i,img,ins,kbd,label,map,mark,picture,q,ruby,rp,rt,s,samp,small,source,span,strike,strong,sub,sup,time,track,tt,u,var,video")),Ct=yt(bt,wt,Ot,xt),kt=vt("background,cite,href,itemtype,longdesc,poster,src,xlink:href"),Et=vt("srcset"),St=vt("abbr,accesskey,align,alt,autoplay,axis,bgcolor,border,cellpadding,cellspacing,class,clear,color,cols,colspan,compact,controls,coords,datetime,default,dir,download,face,headers,height,hidden,hreflang,hspace,ismap,itemscope,itemprop,kind,label,lang,language,loop,media,muted,nohref,nowrap,open,preload,rel,rev,role,rows,rowspan,rules,scope,scrolling,shape,size,sizes,span,srclang,start,summary,tabindex,target,title,translate,type,usemap,valign,value,vspace,width"),Pt=yt(kt,Et,St),Mt=function(){function t(){this.sanitizedSomething=!1,this.buf=[]}return t.prototype.sanitizeChildren=function(t){for(var e=t.firstChild;e;)if(gt.isElementNode(e)?this.startElement(e):gt.isTextNode(e)?this.chars(gt.nodeValue(e)):this.sanitizedSomething=!0,gt.firstChild(e))e=gt.firstChild(e);else for(;e;){gt.isElementNode(e)&&this.endElement(e);var n=At(e,gt.nextSibling(e));if(n){e=n;break}e=At(e,gt.parentElement(e))}return this.buf.join("")},t.prototype.startElement=function(t){var e=this,n=gt.nodeName(t).toLowerCase();Ct.hasOwnProperty(n)?(this.buf.push("<"),this.buf.push(n),gt.attributeMap(t).forEach(function(t,n){var r,i=n.toLowerCase();Pt.hasOwnProperty(i)?(kt[i]&&(t=ft(t)),Et[i]&&(r=t,t=(r=String(r)).split(",").map(function(t){return ft(t.trim())}).join(", ")),e.buf.push(" "),e.buf.push(n),e.buf.push('="'),e.buf.push(It(t)),e.buf.push('"')):e.sanitizedSomething=!0}),this.buf.push(">")):this.sanitizedSomething=!0},t.prototype.endElement=function(t){var e=gt.nodeName(t).toLowerCase();Ct.hasOwnProperty(e)&&!bt.hasOwnProperty(e)&&(this.buf.push("</"),this.buf.push(e),this.buf.push(">"))},t.prototype.chars=function(t){this.buf.push(It(t))},t}();function At(t,e){if(e&&gt.contains(t,e))throw new Error("Failed to sanitize html because the element is clobbered: "+gt.getOuterHTML(t));return e}var Tt=/[\uD800-\uDBFF][\uDC00-\uDFFF]/g,Dt=/([^\#-~ |!])/g;function It(t){return t.replace(/&/g,"&amp;").replace(Tt,function(t){return"&#"+(1024*(t.charCodeAt(0)-55296)+(t.charCodeAt(1)-56320)+65536)+";"}).replace(Dt,function(t){return"&#"+t.charCodeAt(0)+";"}).replace(/</g,"&lt;").replace(/>/g,"&gt;")}function Rt(t){gt.attributeMap(t).forEach(function(e,n){"xmlns:ns1"!==n&&0!==n.indexOf("ns1:")||gt.removeAttribute(t,n)});for(var e=0,n=gt.childNodesAsList(t);e<n.length;e++){var r=n[e];gt.isElementNode(r)&&Rt(r)}}function Nt(t,e){try{var n=function(){if(mt)return mt;var t=(gt=a()).createElement("template");if("content"in t)return t;var e=gt.createHtmlDocument();if(null==(mt=gt.querySelector(e,"body"))){var n=gt.createElement("html",e);mt=gt.createElement("body",e),gt.appendChild(n,mt),gt.appendChild(e,n)}return mt}(),r=e?String(e):"",o=5,s=r;do{if(0===o)throw new Error("Failed to sanitize html because the input is unstable");o--,r=s,gt.setInnerHTML(n,r),t.documentMode&&Rt(n),s=gt.getInnerHTML(n)}while(r!==s);for(var c=new Mt,u=c.sanitizeChildren(gt.getTemplateContent(n)||n),l=gt.getTemplateContent(n)||n,p=0,d=gt.childNodesAsList(l);p<d.length;p++){var h=d[p];gt.removeChild(l,h)}return Object(i._16)()&&c.sanitizedSomething&&gt.log("WARNING: sanitizing HTML stripped some content (see http://g.co/ng/security#xss)."),u}catch(t){throw mt=null,t}}var Lt=new RegExp("^([-,.\"'%_!# a-zA-Z0-9]+|(?:(?:matrix|translate|scale|rotate|skew|perspective)(?:X|Y|3d)?|(?:rgb|hsl)a?|(?:repeating-)?(?:linear|radial)-gradient|(?:calc|attr))\\([-0-9.%, #a-zA-Z]+\\))$","g"),Ft=/^url\(([^)]+)\)$/;var Vt=function(){return function(){}}(),zt=function(t){function e(e){var n=t.call(this)||this;return n._doc=e,n}return Object(o.b)(e,t),e.prototype.sanitize=function(t,e){if(null==e)return null;switch(t){case i.Y.NONE:return e;case i.Y.HTML:return e instanceof qt?e.changingThisBreaksApplicationSecurity:(this.checkNotSafeValue(e,"HTML"),Nt(this._doc,String(e)));case i.Y.STYLE:return e instanceof Ht?e.changingThisBreaksApplicationSecurity:(this.checkNotSafeValue(e,"Style"),function(t){if(!(t=String(t).trim()))return"";var e=t.match(Ft);return e&&ft(e[1])===e[1]||t.match(Lt)&&function(t){for(var e=!0,n=!0,r=0;r<t.length;r++){var i=t.charAt(r);"'"===i&&n?e=!e:'"'===i&&e&&(n=!n)}return e&&n}(t)?t:(Object(i._16)()&&a().log("WARNING: sanitizing unsafe style value "+t+" (see http://g.co/ng/security#xss)."),"unsafe")}(e));case i.Y.SCRIPT:if(e instanceof Ut)return e.changingThisBreaksApplicationSecurity;throw this.checkNotSafeValue(e,"Script"),new Error("unsafe value used in a script context");case i.Y.URL:return e instanceof Gt||e instanceof Wt?e.changingThisBreaksApplicationSecurity:(this.checkNotSafeValue(e,"URL"),ft(String(e)));case i.Y.RESOURCE_URL:if(e instanceof Gt)return e.changingThisBreaksApplicationSecurity;throw this.checkNotSafeValue(e,"ResourceURL"),new Error("unsafe value used in a resource URL context (see http://g.co/ng/security#xss)");default:throw new Error("Unexpected SecurityContext "+t+" (see http://g.co/ng/security#xss)")}},e.prototype.checkNotSafeValue=function(t,e){if(t instanceof Bt)throw new Error("Required a safe "+e+", got a "+t.getTypeName()+" (see http://g.co/ng/security#xss)")},e.prototype.bypassSecurityTrustHtml=function(t){return new qt(t)},e.prototype.bypassSecurityTrustStyle=function(t){return new Ht(t)},e.prototype.bypassSecurityTrustScript=function(t){return new Ut(t)},e.prototype.bypassSecurityTrustUrl=function(t){return new Wt(t)},e.prototype.bypassSecurityTrustResourceUrl=function(t){return new Gt(t)},e.decorators=[{type:i.z}],e.ctorParameters=function(){return[{type:void 0,decorators:[{type:i.y,args:[g]}]}]},e}(Vt),Bt=function(){function t(t){this.changingThisBreaksApplicationSecurity=t}return t.prototype.toString=function(){return"SafeValue must use [property]=binding: "+this.changingThisBreaksApplicationSecurity+" (see http://g.co/ng/security#xss)"},t}(),qt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.b)(e,t),e.prototype.getTypeName=function(){return"HTML"},e}(Bt),Ht=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.b)(e,t),e.prototype.getTypeName=function(){return"Style"},e}(Bt),Ut=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.b)(e,t),e.prototype.getTypeName=function(){return"Script"},e}(Bt),Wt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.b)(e,t),e.prototype.getTypeName=function(){return"URL"},e}(Bt),Gt=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(o.b)(e,t),e.prototype.getTypeName=function(){return"ResourceURL"},e}(Bt),Xt=[{provide:i.P,useValue:r.g},{provide:i.Q,useValue:function(){f.makeCurrent(),x.init()},multi:!0},{provide:r.f,useClass:y,deps:[g]},{provide:g,useFactory:function(){return document},deps:[]}],Yt=[{provide:i.X,useExisting:Vt},{provide:Vt,useClass:zt,deps:[g]}];Object(i._12)(i._17,"browser",Xt);function Kt(){return new i.v}var $t=function(){function t(t){if(t)throw new Error("BrowserModule has already been loaded. If you need access to common directives such as NgIf and NgFor from a lazy loaded module, import CommonModule instead.")}return t.withServerTransition=function(e){return{ngModule:t,providers:[{provide:i.b,useValue:e.appId},{provide:_,useExisting:i.b},j]}},t.decorators=[{type:i.H,args:[{providers:[Yt,{provide:i.v,useFactory:Kt,deps:[]},{provide:M,useClass:it,multi:!0},{provide:M,useClass:pt,multi:!0},{provide:M,useClass:ct,multi:!0},{provide:st,useClass:at},B,{provide:i.V,useExisting:B},{provide:D,useExisting:I},I,i._5,A,P,b,w],exports:[r.a,i.e]}]}],t.ctorParameters=function(){return[{type:t,decorators:[{type:i.M},{type:i._1}]}]},t}(),Qt="undefined"!=typeof window&&window||{},Zt=function(){return function(t,e){this.msPerTick=t,this.numTicks=e}}();!function(){function t(t){this.appRef=t.injector.get(i.f)}t.prototype.timeChangeDetection=function(t){var e=t&&t.record,n=null!=Qt.console.profile;e&&n&&Qt.console.profile("Change Detection");for(var r=a().performanceNow(),i=0;i<5||a().performanceNow()-r<500;)this.appRef.tick(),i++;var o=a().performanceNow();e&&n&&Qt.console.profileEnd("Change Detection");var s=(o-r)/i;return Qt.console.log("ran "+i+" change detection cycles"),Qt.console.log(s.toFixed(2)+" ms per check"),new Zt(s,i)}}();var Jt=function(){function t(){this.store={},this.onSerializeCallbacks={}}return t.init=function(e){var n=new t;return n.store=e,n},t.prototype.get=function(t,e){return this.store[t]||e},t.prototype.set=function(t,e){this.store[t]=e},t.prototype.remove=function(t){delete this.store[t]},t.prototype.hasKey=function(t){return this.store.hasOwnProperty(t)},t.prototype.onSerialize=function(t,e){this.onSerializeCallbacks[t]=e},t.prototype.toJson=function(){for(var t in this.onSerializeCallbacks)if(this.onSerializeCallbacks.hasOwnProperty(t))try{this.store[t]=this.onSerializeCallbacks[t]()}catch(t){console.warn("Exception in onSerialize callback: ",t)}return JSON.stringify(this.store)},t.decorators=[{type:i.z}],t.ctorParameters=function(){return[]},t}();function te(t,e){var n,r,i=t.getElementById(e+"-state"),o={};if(i&&i.textContent)try{o=JSON.parse((n=i.textContent,r={"&a;":"&","&q;":'"',"&s;":"'","&l;":"<","&g;":">"},n.replace(/&[^;]+;/g,function(t){return r[t]})))}catch(t){console.warn("Exception while restoring TransferState for app "+e,t)}return Jt.init(o)}(function(){function t(){}t.decorators=[{type:i.H,args:[{providers:[{provide:Jt,useFactory:te,deps:[g,i.b]}]}]}],t.ctorParameters=function(){return[]}})(),function(){function t(){}t.all=function(){return function(t){return!0}},t.css=function(t){return function(e){return null!=e.nativeElement&&a().elementMatches(e.nativeElement,t)}},t.directive=function(t){return function(e){return-1!==e.providerTokens.indexOf(t)}}}(),new i._6("5.1.0")}});
</script><script type="text/javascript">
webpackJsonp(["main"],{"../../../../../src/$$_lazy_route_resource lazy recursive":function(t,e){function n(t){return Promise.resolve().then(function(){throw new Error("Cannot find module '"+t+"'.")})}n.keys=function(){return[]},n.resolve=n,t.exports=n,n.id="../../../../../src/$$_lazy_route_resource lazy recursive"},"../../../../../src/app/accordion/accordion.component.html":function(t,e){t.exports='<div *ngIf="data.byReg(case).length > 1">\n  <mat-expansion-panel\n  [expanded]="data.allExpandState"\n  #expPanel\n  class="mat-elevation-z0">\n     <mat-expansion-panel-header\n     class="mat-expansion-panel-header"\n     collapsedHeight="auto"\n     expandedHeight="auto"\n     style="padding:0.05vh">\n       <mat-panel-title fxLayoutAlign=\'space-between center\'>\n         <td class="test-title">\n           {{ data.byID(data.byReg(case)[0])?.title }}\n         </td>\n         <td>\n           <button mat-button style="color:black"> Info </button>\n         </td>\n       </mat-panel-title>\n     </mat-expansion-panel-header>\n       <div>\n         <p> {{ data.byID(data.byReg(case)[0])?.summary }} </p>\n       </div>\n </mat-expansion-panel>\n</div>\n  <mat-accordion\n  multi="true"\n  displayMode="flat">\n       <mat-expansion-panel\n       *ngFor="let testId of data.byReg(case); let i = index"\n       #expPanel\n       [expanded]="data.allExpandState"\n       class="mat-elevation-z0">\n          <mat-expansion-panel-header\n          class="mat-expansion-panel-header"\n          collapsedHeight="auto"\n          expandedHeight="auto"\n          style="padding:0.05vh">\n            <mat-panel-title\n            *ngIf="data.byReg(case).length === 1; else noTitle"\n            fxLayoutAlign=\'space-between center\'>\n              <td class="test-title">\n                {{ data.byID(testId)?.title }}\n              </td>\n              <td>\n                <app-result-button [testId]="testId"></app-result-button>\n              </td>\n            </mat-panel-title>\n            <ng-template #noTitle>\n              <mat-panel-title fxLayoutAlign=\'space-between center\'>\n                <td style="padding-left:1vh"> {{ data.getParam(testId, 1) }} </td>\n                <td>\n                  <app-result-button [testId]="testId"></app-result-button>\n                </td>\n              </mat-panel-title>\n            </ng-template>\n          </mat-expansion-panel-header>\n          <div fxLayout=\'column\' flexLayoutAlign=\'space-between center\'>\n            <div *ngIf="data.byReg(case).length === 1; else justMessages">\n              <p>{{ data.byID(testId)?.summary }}</p>\n              <p>{{ data.byID(testId)?.message }}</p>\n            </div>\n            <ng-template #justMessages>\n              <div>\n                <tr> {{ data.byID(testId)?.message }} </tr>\n              </div>\n            </ng-template>\n            <div *ngIf="data.reportType != \'history\'; else vegaPlot">\n              <br>\n              <textarea rows="4" style="overflow-wrap: break-word; width:100%">\n                {{ data.getString(data.byID(testId)?.data) }}\n              </textarea>\n            </div>\n            <ng-template #vegaPlot>\n              <br>\n                <div>\n                  <app-vega-plot [testId]="testId"></app-vega-plot>\n                </div>\n            </ng-template>\n          </div>\n      </mat-expansion-panel>\n  </mat-accordion>\n'},"../../../../../src/app/accordion/accordion.component.scss":function(t,e,n){(t.exports=n("../../../../css-loader/lib/css-base.js")(!1)).push([t.i,"",""]),t.exports=t.exports.toString()},"../../../../../src/app/accordion/accordion.component.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var a=n("../../../core/esm5/core.js"),s=n("../../../../../src/app/report-data.service.ts"),o=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},r=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},o([Object(a.C)(),r("design:type",String)],t.prototype,"case",void 0),t=o([Object(a.m)({selector:"app-accordion",template:n("../../../../../src/app/accordion/accordion.component.html"),styles:[n("../../../../../src/app/accordion/accordion.component.scss")],encapsulation:a._10.None}),r("design:paramtypes",[s.a])],t)}()},"../../../../../src/app/app.component.html":function(t,e){t.exports="<app-header></app-header>\n<app-dashboard></app-dashboard>\n"},"../../../../../src/app/app.component.scss":function(t,e,n){(t.exports=n("../../../../css-loader/lib/css-base.js")(!1)).push([t.i,".example-container {\n  display: -webkit-box;\n  display: -ms-flexbox;\n  display: flex;\n  -webkit-box-orient: vertical;\n  -webkit-box-direction: normal;\n      -ms-flex-direction: column;\n          flex-direction: column;\n  min-width: 300px; }\n\n.mat-table {\n  overflow: auto;\n  max-height: 500px; }\n\n.mat-header-cell.mat-sort-header-sorted {\n  color: black; }\n\n.mat-expansion-panel-content:not(.mat-expanded) {\n  height: 0px;\n  visibility: hidden; }\n\n.mat-expansion-panel-header {\n  height: 48px; }\n",""]),t.exports=t.exports.toString()},"../../../../../src/app/app.component.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var a=n("../../../core/esm5/core.js"),s=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},o=function(){function t(){this.title="app"}return t=s([Object(a.m)({selector:"app-root",template:n("../../../../../src/app/app.component.html"),styles:[n("../../../../../src/app/app.component.scss")]})],t)}()},"../../../../../src/app/app.material.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var a=n("../../../core/esm5/core.js"),s=n("../../../platform-browser/esm5/animations.js"),o=n("../../../material/esm5/material.es5.js"),r=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},i=function(){function t(){}return t=r([Object(a.H)({imports:[s.a,o.b,o.d,o.k,o.s,o.i,o.g,o.j,o.c,o.p,o.l,o.a,o.e,o.r,o.q,o.m,o.t,o.n,o.h,o.o,o.f],exports:[s.a,o.b,o.d,o.k,o.s,o.i,o.g,o.j,o.c,o.p,o.l,o.a,o.e,o.r,o.q,o.m,o.t,o.n,o.h,o.o,o.f]})],t)}()},"../../../../../src/app/app.module.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return T});var a=n("../../../platform-browser/esm5/platform-browser.js"),s=n("../../../core/esm5/core.js"),o=n("../../../../../src/app/app.material.ts"),r=n("../../../flex-layout/esm5/flex-layout.es5.js"),i=n("../../../common/esm5/http.js"),c=n("../../../../../src/app/app.component.ts"),u=n("../../../../../src/app/header/header.component.ts"),M=n("../../../../../src/app/dashboard/dashboard.component.ts"),p=n("../../../../../src/app/dashboard/universal-core/universal-core.component.ts"),d=n("../../../../../src/app/dashboard/statistics/statistics.component.ts"),l=n("../../../../../src/app/report-data.service.ts"),y=n("../../../../../src/app/keys.pipe.ts"),j=n("../../../../../src/app/dashboard/system-information/system-information.component.ts"),L=n("../../../../../src/app/accordion/accordion.component.ts"),m=n("../../../../../src/app/result-button/result-button.component.ts"),f=n("../../../../../src/app/vega-plot/vega-plot.component.ts"),N=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},x=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},T=function(){function t(t){this.reportDataService=t,this.reportDataService.loadResults()}return t.prototype.ngOnInit=function(){},t=N([Object(s.H)({declarations:[c.a,u.a,M.a,p.a,d.a,y.a,j.a,L.a,m.a,f.a],imports:[a.a,o.a,r.a,i.b],exports:[a.a,o.a],providers:[l.a],bootstrap:[c.a]}),x("design:paramtypes",[l.a])],t)}()},"../../../../../src/app/dashboard/dashboard.component.html":function(t,e){t.exports="<div\n  fxLayout='row wrap'\n  fxLayoutAlign='center stretch'\n  fxLayout.xs='column'>\n  <div\n    fxLayout='column'\n    fxFlex=\"45%\"\n    fxFlexAlign='stretch'\n    style=\"margin:0.5vmax\">\n    <mat-toolbar\n      class=\"mat-elevation-z2\"\n      color='secondary'> Scored Tests </mat-toolbar>\n      <mat-card fxFlex *ngIf='data.scoredCard'>\n        <app-universal-core [scoreData]=\"data.scoredCard\"></app-universal-core>\n      </mat-card>\n  </div>\n  <div\n    fxLayout='column'\n    fxFlex=\"45%\"\n    fxFlexAlign='stretch'\n    fxLayoutAlign='start stretch'\n    style=\"margin:0.5vmax\"\n    *ngIf='data'>\n      <mat-toolbar\n      class=\"mat-elevation-z2\"\n      color='secondary'> Unscored Tests </mat-toolbar>\n      <div\n        *ngFor=\"let card of data.statisticsCards\">\n          <app-statistics [currentCard]=\"card\"></app-statistics>\n      </div>\n      <div\n        *ngIf='data.metaData'>\n          <app-system-information></app-system-information>\n      </div>\n   </div>\n</div>\n"},"../../../../../src/app/dashboard/dashboard.component.scss":function(t,e,n){(t.exports=n("../../../../css-loader/lib/css-base.js")(!1)).push([t.i,"",""]),t.exports=t.exports.toString()},"../../../../../src/app/dashboard/dashboard.component.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var a=n("../../../core/esm5/core.js"),s=n("../../../../../src/app/report-data.service.ts"),o=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},r=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t=o([Object(a.m)({selector:"app-dashboard",template:n("../../../../../src/app/dashboard/dashboard.component.html"),styles:[n("../../../../../src/app/dashboard/dashboard.component.scss")],encapsulation:a._10.None}),r("design:paramtypes",[s.a])],t)}()},"../../../../../src/app/dashboard/statistics/statistics.component.html":function(t,e){t.exports='<mat-card fxFlex>\n  <mat-card-title>{{ currentCard.cardName }}</mat-card-title>\n    <mat-card-content>\n      <div *ngFor="let case of currentCard.associatedTests">\n         <app-accordion [case]="case"></app-accordion>\n      </div>\n    </mat-card-content>\n  </mat-card>\n'},"../../../../../src/app/dashboard/statistics/statistics.component.scss":function(t,e,n){(t.exports=n("../../../../css-loader/lib/css-base.js")(!1)).push([t.i,"",""]),t.exports=t.exports.toString()},"../../../../../src/app/dashboard/statistics/statistics.component.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return c});var a=n("../../../core/esm5/core.js"),s=n("../../../../../src/app/resultcard.model.ts"),o=n("../../../../../src/app/report-data.service.ts"),r=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},i=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},c=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},r([Object(a.C)(),i("design:type",s.a)],t.prototype,"currentCard",void 0),t=r([Object(a.m)({selector:"app-statistics",template:n("../../../../../src/app/dashboard/statistics/statistics.component.html"),styles:[n("../../../../../src/app/dashboard/statistics/statistics.component.scss")],encapsulation:a._10.None}),i("design:paramtypes",[o.a])],t)}()},"../../../../../src/app/dashboard/system-information/system-information.component.html":function(t,e){t.exports='  <mat-card fxFlex>\n    <mat-card-title>Environment</mat-card-title>\n    <mat-card-content>\n      <mat-accordion\n      multi="true"\n      displayMode="flat" >\n         <mat-expansion-panel\n         class="mat-elevation-z0"\n         hideToggle="true">\n          <mat-expansion-panel-header\n          class="mat-expansion-panel-header"\n          collapsedHeight="auto"\n          expandedHeight="auto"\n          style="padding:0.05vh">\n            <mat-panel-description fxLayoutAlign="space-between start">\n              <span> Python Version </span>\n              <span>{{data.metaData.python}}</span>\n            </mat-panel-description>\n          </mat-expansion-panel-header>\n        </mat-expansion-panel>\n        <mat-expansion-panel\n        class="mat-elevation-z0"\n        hideToggle="true">\n         <mat-expansion-panel-header\n         class="mat-expansion-panel-header"\n         collapsedHeight="auto"\n         expandedHeight="auto"\n         style="padding:0.05vh">\n           <mat-panel-description fxLayoutAlign="space-between start">\n             <span> Platform </span>\n             <span>{{data.metaData.platform}}</span>\n           </mat-panel-description>\n         </mat-expansion-panel-header>\n       </mat-expansion-panel>\n       <mat-expansion-panel\n       class="mat-elevation-z0"\n       *ngIf=\'data.metaData.packages\'\n       hideToggle="true">\n        <mat-expansion-panel-header\n        class="mat-expansion-panel-header"\n        collapsedHeight="auto"\n        expandedHeight="auto"\n        style="padding:0.05vh">\n          <mat-panel-description fxLayoutAlign="space-between start">\n            <span> Memote Version </span>\n            <span>{{data.metaData.packages.memote}}</span>\n          </mat-panel-description>\n        </mat-expansion-panel-header>\n        <div fxLayout=\'column\' flexLayoutAlign=\'space-between center\'>\n        <div> Package Versions </div>\n        <div>\n          <p></p>\n          <textarea rows="4" style="overflow-wrap: break-word; width:100%">\n          {{data.getString(data.metaData.packages)}}\n          </textarea>\n        </div>\n        </div>\n      </mat-expansion-panel>\n      </mat-accordion>\n    </mat-card-content>\n  </mat-card>\n'},"../../../../../src/app/dashboard/system-information/system-information.component.scss":function(t,e,n){(t.exports=n("../../../../css-loader/lib/css-base.js")(!1)).push([t.i,"",""]),t.exports=t.exports.toString()},"../../../../../src/app/dashboard/system-information/system-information.component.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var a=n("../../../core/esm5/core.js"),s=n("../../../../../src/app/report-data.service.ts"),o=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},r=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t=o([Object(a.m)({selector:"app-system-information",template:n("../../../../../src/app/dashboard/system-information/system-information.component.html"),styles:[n("../../../../../src/app/dashboard/system-information/system-information.component.scss")],encapsulation:a._10.None}),r("design:paramtypes",[s.a])],t)}()},"../../../../../src/app/dashboard/universal-core/universal-core.component.html":function(t,e){t.exports='<mat-card-title>Universal Core</mat-card-title>\n    <mat-card-content>\n      <div\n        fxLayout=\'row\'\n        fxLayoutAlign=\'space-between start\'\n        fxLayoutAlign.xs=\'space-between center\'\n        fxLayout.xs=\'column\'>\n        <div fxFlex> Total Score\n          <div>\n            <div style="font-size:15vh">99</div>\n          </div>\n        </div>\n        <div fxFlex> Score per Category\n        <div id="catScore">\n          TODO: Put a nice graph here!\n        </div>\n        </div>\n      </div>\n      <hr>\n      <div>\n      <div fxFlex>\n        <h2>Scored Tests</h2>\n        <div *ngFor="let section of scoreData.sections | keys">\n          <h3> {{ scoreData.sections[section].title }}</h3>\n          <div *ngFor="let case of scoreData.sections[section].cases">\n            <app-accordion [case]="case"></app-accordion>\n          </div>\n        </div>\n      </div>\n      </div>\n</mat-card-content>\n'},"../../../../../src/app/dashboard/universal-core/universal-core.component.scss":function(t,e,n){(t.exports=n("../../../../css-loader/lib/css-base.js")(!1)).push([t.i,"",""]),t.exports=t.exports.toString()},"../../../../../src/app/dashboard/universal-core/universal-core.component.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var a=n("../../../core/esm5/core.js"),s=n("../../../../../src/app/report-data.service.ts"),o=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},r=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},o([Object(a.C)(),r("design:type",Object)],t.prototype,"scoreData",void 0),t=o([Object(a.m)({selector:"app-universal-core",template:n("../../../../../src/app/dashboard/universal-core/universal-core.component.html"),styles:[n("../../../../../src/app/dashboard/universal-core/universal-core.component.scss")],encapsulation:a._10.None}),r("design:paramtypes",[s.a])],t)}()},"../../../../../src/app/header/header.component.html":function(t,e){t.exports='<mat-toolbar\n  class="mat-elevation-z9 fixed-header"\n  color=\'primary\'>\n      <img aria-label="logo" style="width: 5vw; padding-right:1vw"\n       width="48px"\n       height="48px"\n       src="data:image/svg+xml;base64, "/>\n      <span style="text-align:left; padding-right:1vw" *ngIf="data.byID(\'test_model_id_presence\')"> {{ data.byID("test_model_id_presence").data }} </span>\n      <button mat-button (click)="data.togglePanel()"> {{ !data.allExpandState ? "Expand" : "Collapse" }} All</button>\n      <button mat-button style="visibility:hidden"  [matMenuTriggerFor]="menu">Export</button>\n      <mat-menu #menu="matMenu">\n        <button mat-menu-item>Item 1</button>\n        <button mat-menu-item>Item 2</button>\n      </mat-menu>\n      <span style="flex:1 1 auto"></span>\n      <span style="text-align:right; padding-right:1vw" *ngIf="data.metaData">{{ data.metaData.timestamp | date:\'yyyy-MM-dd HH:mm\' }}</span>\n      <a\n        aria-label="GitHub Repository"\n        href="https://github.com/opencobra/memote"\n        mat-button>\n        <span class="mat-button-wrapper">\n          <img\n          width="40px"\n          height="40px"\n          src="data:image/svg+xml;base64,PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz48IURPQ1RZUEUgc3ZnIFBVQkxJQyAiLS8vVzNDLy9EVEQgU1ZHIDEuMS8vRU4iICJodHRwOi8vd3d3LnczLm9yZy9HcmFwaGljcy9TVkcvMS4xL0RURC9zdmcxMS5kdGQiPjxzdmcgdmVyc2lvbj0iMS4xIiBpZD0iTGF5ZXJfMSIgeG1sbnM9Imh0dHA6Ly93d3cudzMub3JnLzIwMDAvc3ZnIiB4bWxuczp4bGluaz0iaHR0cDovL3d3dy53My5vcmcvMTk5OS94bGluayIgeD0iMHB4IiB5PSIwcHgiIHdpZHRoPSIxNnB4IiBoZWlnaHQ9IjE2cHgiIHZpZXdCb3g9IjAgMCAxNiAxNiIgZW5hYmxlLWJhY2tncm91bmQ9Im5ldyAwIDAgMTYgMTYiIHhtbDpzcGFjZT0icHJlc2VydmUiPjx0aXRsZT5tYXJrLWdpdGh1YjwvdGl0bGU+PGRlc2M+Q3JlYXRlZCB3aXRoIFNrZXRjaC48L2Rlc2M+PGcgaWQ9Ik9jdGljb25zIj48ZyBpZD0ibWFyay1naXRodWIiPjxwYXRoIGlkPSJTaGFwZSIgZmlsbD0iI0ZGRkZGRiIgZD0iTTgsMEMzLjU4LDAsMCwzLjU4LDAsOGMwLDMuNTQsMi4yOSw2LjUzLDUuNDcsNy41OWMwLjQsMC4wNywwLjU1LTAuMTcsMC41NS0wLjM4YzAtMC4xOS0wLjAxLTAuODE5LTAuMDEtMS40OUM0LDE0LjA5LDMuNDgsMTMuMjMsMy4zMiwxMi43OGMtMC4wOS0wLjIzLTAuNDgtMC45NC0wLjgyLTEuMTNjLTAuMjgtMC4xNS0wLjY4LTAuNTIxLTAuMDEtMC41M2MwLjYzLTAuMDExLDEuMDgsMC41OCwxLjIzLDAuODE5YzAuNzIsMS4yMTEsMS44NywwLjg3MSwyLjMzLDAuNjZjMC4wNy0wLjUyLDAuMjgtMC44NjksMC41MS0xLjA2OWMtMS43OC0wLjItMy42NC0wLjg5LTMuNjQtMy45NWMwLTAuODcsMC4zMS0xLjU5LDAuODItMi4xNWMtMC4wOC0wLjItMC4zNi0xLjAyLDAuMDgtMi4xMmMwLDAsMC42Ny0wLjIxLDIuMiwwLjgyYzAuNjQtMC4xOCwxLjMyLTAuMjcsMi0wLjI3czEuMzYsMC4wOSwyLDAuMjdjMS41My0xLjA0LDIuMjAxLTAuODIsMi4yMDEtMC44MmMwLjQzOSwxLjEsMC4xNTksMS45MiwwLjA3OSwyLjEyYzAuNTExLDAuNTYsMC44MiwxLjI3LDAuODIsMi4xNWMwLDMuMDctMS44NywzLjc1LTMuNjQ5LDMuOTVjMC4yODksMC4yNSwwLjUzOSwwLjcyOSwwLjUzOSwxLjQ3OWMwLDEuMDctMC4wMSwxLjkzLTAuMDEsMi4yYzAsMC4yMSwwLjE1LDAuNDYsMC41NSwwLjM4QzEzLjcxLDE0LjUzLDE2LDExLjUzLDE2LDhDMTYsMy41OCwxMi40MiwwLDgsMEw4LDB6Ii8+PC9nPjwvZz48L3N2Zz4="/>\n        </span>\n        </a>\n</mat-toolbar>\n'},"../../../../../src/app/header/header.component.scss":function(t,e,n){(t.exports=n("../../../../css-loader/lib/css-base.js")(!1)).push([t.i,"",""]),t.exports=t.exports.toString()},"../../../../../src/app/header/header.component.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var a=n("../../../core/esm5/core.js"),s=n("../../../../../src/app/report-data.service.ts"),o=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},r=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t=o([Object(a.m)({selector:"app-header",template:n("../../../../../src/app/header/header.component.html"),styles:[n("../../../../../src/app/header/header.component.scss")],encapsulation:a._10.None}),r("design:paramtypes",[s.a])],t)}()},"../../../../../src/app/keys.pipe.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return o});var a=n("../../../core/esm5/core.js"),s=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},o=function(){function t(){}return t.prototype.transform=function(t,e){return void 0===e&&(e=null),Object.keys(t)},t=s([Object(a.R)({name:"keys",pure:!1})],t)}()},"../../../../../src/app/report-data.service.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return M});var a=n("../../../core/esm5/core.js"),s=n("../../../common/esm5/http.js"),o=n("../../../../../src/app/resultcard.model.ts"),r=n("../../../../../src/app/test-result.model.ts"),i=n("../../../../../src/app/test-history.model.ts"),c=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},u=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},M=function(){function t(t){this.http=t,this.allTests=[],this.statisticsCards=[],this.scoredTests=[],this.allExpandState=!1}return t.prototype.loadResults=function(){var t=this;switch(this.reportType=window.reportType,this.reportType){case"history":this.convertHistoryResults(window.data);break;case"snapshot":this.convertResults(window.data);break;case"diff":break;default:this.http.get("/data/testHistory.json").subscribe(function(e){t.convertHistoryResults(e)}),this.reportType="history"}},t.prototype.togglePanel=function(){this.allExpandState=!this.allExpandState},t.prototype.byID=function(t){return this.allTests.find(function(e){return e.id===t})},t.prototype.byReg=function(t){for(var e=new RegExp(t),n=[],a=0,s=this.allTests;a<s.length;a++){var o=s[a];e.test(o.id)&&n.push(o.id)}return n},t.prototype.isScored=function(t){return this.scoredTests.includes(t)},t.prototype.getParam=function(t,e){return t.split(":")[e]},t.prototype.getString=function(t){return JSON.stringify(t)},t.prototype.convertResults=function(t){for(var e=0,n=Object.keys(t.tests);e<n.length;e++){var a=n[e];if(t.tests[a].message instanceof Object)for(var s=0,o=Object.keys(t.tests[a].data);s<o.length;s++){var i=o[s],c=a+":"+i;this.allTests.push(new r.a(c,{data:t.tests[a].data[i],duration:t.tests[a].duration[i],message:t.tests[a].message[i],metric:t.tests[a].metric[i],result:t.tests[a].result[i],summary:t.tests[a].summary,title:t.tests[a].title,type:t.tests[a].format_type}))}else this.allTests.push(new r.a(a,t.tests[a]))}this.extractMetadata(t),this.distributeCardsToSections(t),this.determineScoredTests()},t.prototype.convertHistoryResults=function(t){for(var e=0,n=Object.keys(t.tests);e<n.length;e++){var a=n[e];if(t.tests[a].history instanceof Array)this.allTests.push(new i.a(a,t.tests[a]));else for(var s=0,o=Object.keys(t.tests[a].history);s<o.length;s++){var r=o[s],c=a+":"+r;this.allTests.push(new i.a(c,{history:t.tests[a].history[r],summary:t.tests[a].summary,title:t.tests[a].title,type:t.tests[a].format_type}))}}this.distributeCardsToSections(t),this.determineScoredTests()},t.prototype.extractMetadata=function(t){this.metaData=t.meta},t.prototype.distributeCardsToSections=function(t){for(var e=0,n=Object.keys(t.cards);e<n.length;e++){var a=n[e];"scored"===a?this.scoredCard=t.cards.scored:this.statisticsCards.push(new o.a(a,t.cards[a]))}},t.prototype.determineScoredTests=function(){for(var t=0,e=Object.keys(this.scoredCard.sections);t<e.length;t++){var n=e[t];if(this.scoredCard.sections[n].cases instanceof Array)for(var a=0,s=this.scoredCard.sections[n].cases;a<s.length;a++){var o=s[a];this.scoredTests.push(o)}}},t=c([Object(a.z)(),u("design:paramtypes",[s.a])],t)}()},"../../../../../src/app/result-button/result-button.component.html":function(t,e){t.exports="\x3c!-- First distinguish wether the tests have passed/failed or skipped or errorer --\x3e\n<div *ngIf=\"data.byID(testId)?.result === 'passed' || data.byID(testId)?.result === 'failed'\">\n  \x3c!-- Then determine if the test results are from scored or from unscored tests --\x3e\n  <div *ngIf=\"data.isScored(data.getParam(testId, 0)); else notScored\">\n    \x3c!-- If they are from scored tests colorize the result button --\x3e\n    <button\n    mat-raised-button\n    style=\"color:white\"\n    [ngStyle]=\"{'background-color':getColor(data.byID(testId)?.metric | number : '1.2-2')}\">\n      {{ data.byID(testId)?.metric | number : '1.2-2' }}\n    </button>\n  </div>\n  <ng-template #notScored>\n    \x3c!-- If they are from unscored tests the user has some control over how --\x3e\n    \x3c!-- they are displayed using the 'type' attribute of each test. --\x3e\n    <button\n    mat-raised-button\n    color=\"primary\"\n    style=\"color:white\">\n      \x3c!-- Hence for each test, determine now how the result ought to be formatted --\x3e\n      <div *ngIf=\"data.byID(testId)?.type === 'count'\">\n        {{ data.byID(testId)?.data.length | number : '1.'}}\n      </div>\n      <div *ngIf=\"data.byID(testId)?.type === 'number'\">\n        {{ data.byID(testId)?.data | number : '1.2-2'}}\n      </div>\n      <div *ngIf=\"data.byID(testId)?.type === 'percent'\">\n        {{ data.byID(testId)?.metric | number : '1.2-2'}}\n      </div>\n      <div *ngIf=\"data.byID(testId)?.type === 'raw'\">\n        {{ data.byID(testId)?.data }}\n      </div>\n    </button>\n  </ng-template>\n</div>\n<div *ngIf=\"data.byID(testId)?.result === 'skipped'\">\n\x3c!-- Skipped tests are marked with a gray button with black text --\x3e\n    <button\n     mat-raised-button\n     disabled\n     style=\"color:black\">\n       Skipped\n     </button>\n</div>\n<div *ngIf=\"data.byID(testId)?.result === 'error'\">\n\x3c!-- Tests with errors are marked with a black button with red text --\x3e\n    <button\n     mat-raised-button\n     style=\"color:red;background-color:black\">\n       Errored\n     </button>\n</div>\n"},"../../../../../src/app/result-button/result-button.component.scss":function(t,e,n){(t.exports=n("../../../../css-loader/lib/css-base.js")(!1)).push([t.i,"",""]),t.exports=t.exports.toString()},"../../../../../src/app/result-button/result-button.component.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return i});var a=n("../../../core/esm5/core.js"),s=n("../../../../../src/app/report-data.service.ts"),o=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},r=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},i=function(){function t(t){this.data=t}return t.prototype.ngOnInit=function(){},t.prototype.getColor=function(t){return["hsl(",(132*(1-t)).toString(10),",80%,35%)"].join("")},o([Object(a.C)(),r("design:type",String)],t.prototype,"testId",void 0),t=o([Object(a.m)({selector:"app-result-button",template:n("../../../../../src/app/result-button/result-button.component.html"),styles:[n("../../../../../src/app/result-button/result-button.component.scss")],encapsulation:a._10.None}),r("design:paramtypes",[s.a])],t)}()},"../../../../../src/app/resultcard.model.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var a=function(){return function(t,e){var n=e.title,a=e.cases;this.cardID=t,this.cardName=n,this.associatedTests=a}}()},"../../../../../src/app/test-history.model.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var a=function(){return function(t,e){var n=e.history,a=e.summary,s=e.title,o=e.type;this.id=t,this.history=n,this.summary=a,this.title=s,this.type=o}}()},"../../../../../src/app/test-result.model.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var a=function(){function t(t,e){var n=e.data,a=e.duration,s=e.message,o=e.metric,r=e.result,i=e.summary,c=e.title,u=e.type;this.id=t,this.data=n,this.duration=a,this.message=s,this.metric=o,this.errorFailsafe(r),this.summary=i,this.title=c,this.type=u}return t.prototype.errorFailsafe=function(t){"skipped"===this.result||this.data?this.result=t:this.result="error"},t}()},"../../../../../src/app/vega-plot/vega-plot.component.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return u});var a=n("../../../core/esm5/core.js"),s=n("../../../../vega-lib/index.js"),o=n("../../../../../src/app/report-data.service.ts"),r=n("../../../../../src/app/vega-plot/vega-plot.metric-spec.ts"),i=this&&this.__decorate||function(t,e,n,a){var s,o=arguments.length,r=o<3?e:null===a?a=Object.getOwnPropertyDescriptor(e,n):a;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)r=Reflect.decorate(t,e,n,a);else for(var i=t.length-1;i>=0;i--)(s=t[i])&&(r=(o<3?s(r):o>3?s(e,n,r):s(e,n))||r);return o>3&&r&&Object.defineProperty(e,n,r),r},c=this&&this.__metadata||function(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)},u=function(){function t(t,e){this.data=t,this.elementRef=e,this.nativeElement=e.nativeElement}return t.prototype.initialize=function(){this.testObject=this.data.byID(this.testId),this.type=this.testObject.type,["count","raw","number"].includes(this.type)?this.type="data":this.type="metric",r.a.data[0].values=this.testObject.history,new s.a(Object(s.b)(r.a)).renderer("svg").signal("type",this.type).initialize(this.nativeElement).hover().run().addEventListener("click",function(t,e){e&&e.datum&&e.datum.commit&&alert("Navigating to snapshort report for commit"+e.datum.commit)})},t.prototype.ngOnInit=function(){this.initialize()},i([Object(a.C)(),c("design:type",String)],t.prototype,"testId",void 0),t=i([Object(a.m)({selector:"app-vega-plot",template:"",styleUrls:[],encapsulation:a._10.None}),c("design:paramtypes",[o.a,a.t])],t)}()},"../../../../../src/app/vega-plot/vega-plot.metric-spec.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var a={$schema:"https://vega.github.io/schema/vega/v3.0.json",width:350,height:200,padding:0,signals:[{name:"shift",value:!1,on:[{events:"@legendSymbol:click, @legendLabel:click",update:"event.shiftKey",force:!0}]},{name:"clicked",value:null,on:[{events:"@legendSymbol:click, @legendLabel:click",update:"{value: datum.value}",force:!0}]},{name:"type",value:"data",description:"This keyword describes which data is displayed on the y-axis."}],data:[{name:"scores",values:[]},{name:"selected",on:[{trigger:"clicked",toggle:"clicked"}]}],scales:[{name:"x",type:"point",range:"width",domain:{data:"scores",field:"commit"}},{name:"y",type:"linear",range:"height",nice:!0,zero:!0,domain:{data:"scores",field:{signal:"type"}}},{name:"color",type:"ordinal",range:["#000000","#E69F00","#56B4E9","#009E73","#F0E442","#0072B2","#D55E00","#CC79A7"],domain:{data:"scores",field:"branch"}}],axes:[{orient:"bottom",scale:"x",title:"History",encode:{labels:{interactive:!0,update:{angle:{value:50},limit:{value:50},baseline:{value:"top"},dx:{value:3},align:{value:"left"}}}}},{orient:"left",scale:"y",title:{signal:"type === 'metric' ? 'Metric': 'Data'"}}],marks:[{type:"group",from:{facet:{name:"series",data:"scores",groupby:"branch"}},legends:[{stroke:"color",title:"Branches",padding:4,orient:"left",offset:30,encode:{symbols:{name:"legendSymbol",interactive:!0,enter:{cursor:{value:"pointer"}},update:{fill:{value:"transparent"},strokeWidth:{value:2},opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.value)",value:.7},{value:.15}],size:{value:64}}},labels:{name:"legendLabel",interactive:!0,enter:{cursor:{value:"pointer"}},update:{opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.value)",value:1},{value:.25}]}}}}],marks:[{type:"line",from:{data:"series"},encode:{enter:{x:{scale:"x",field:"commit"},y:{scale:"y",field:{signal:"type"}},stroke:{scale:"color",field:"branch"},strokeWidth:{value:2}},update:{interpolate:{value:"linear"},opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.branch)",value:1},{value:0}]}}},{type:"symbol",from:{data:"series"},encode:{enter:{x:{scale:"x",field:"commit"},y:{scale:"y",field:{signal:"type"}},fill:{scale:"color",field:"branch"},size:{value:90},tooltip:{scale:"x",field:"commit"}},update:{opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.branch)",value:1},{value:0}]},hover:{opacity:[{test:"!length(data('selected')) || indata('selected', 'value', datum.branch)",value:.5},{value:0}],cursor:{value:"pointer"}}}}]}]}},"../../../../../src/environments/environment.ts":function(t,e,n){"use strict";n.d(e,"a",function(){return a});var a={production:!1}},"../../../../../src/main.ts":function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0});var a=n("../../../core/esm5/core.js"),s=n("../../../platform-browser-dynamic/esm5/platform-browser-dynamic.js"),o=n("../../../../../src/app/app.module.ts");n("../../../../../src/environments/environment.ts").a.production&&Object(a._13)(),Object(s.a)().bootstrapModule(o.a).catch(function(t){return console.log(t)})},0:function(t,e,n){t.exports=n("../../../../../src/main.ts")},1:function(t,e){},2:function(t,e){},3:function(t,e){}},[0]);
</script></body>
</html>
